# Facilitate running of mode FILAMENT_REFINE in parallel
class FilamentRefineParallel
  NAME="Parallel runner for mode FILAMENT_REFINE"

  # Stuff we need
  require_relative "FrealixRunner.rb"

  # Initialise
  def initialize( options = {} )
    # Default values for optional arguments
    defaults = {
      :dir                =>  FrealixRefinementDirectory.new(),
      :round_number 	    => 	FrealixRefinementDirectory.new().currentRound(),
      :num_nodes		      => 	1,
      :num_threads 	      => 	1,
      :params			        =>	FrealixParameters.new(),
      :hosts              =>  ["cluster"],
      :mem_per_thread_gb	=>	1,
      :prefix             => "flx",
      :gpu_nodes_only     =>  false,
      :resume             =>  false}
    options = defaults.merge(options)
    # Instance variables
    @dir		      = options[:dir]
    @round        = options[:round_number]
    @num_nodes    = options[:num_nodes]
    @num_threads  = options[:num_threads]
    @mem_free	    = options[:mem_per_thread_gb]	# Free memory required per thread, in GB
    @params	      = options[:params]
    @avail_hosts  = options[:hosts]
    @prefix       = options[:prefix]
    @gpu_nodes_only=options[:gpu_nodes_only]
    @resume       = options[:resume]
    $LOG.debug("FilamentRefineParalle init, avail_hosts = #{@avail_hosts}")
  end

  # Do the processing for the round
  def parallelRun()
    $LOG.info("parallelRun") {"Round #{@round}: Processing with #{@num_nodes} nodes, each with #{@num_threads} threads"}
    # The subdirectory for the current & previous rounds
    round_dir = @dir.currentRoundDir
    previous_round_dir = @dir.previousRoundDir
    if (!@resume)
      # Split the .ali file
      $LOG.info("parallelRun") {"Round #{@round}: Splitting the input alignments file into #{@num_nodes}"}
      if (@round == 0)
        self.splitAli(round_dir+"/waypoints.ali",@num_nodes)
      else
        self.splitAli(previous_round_dir+"/waypoints.ali",@num_nodes)
      end
    else
      $LOG.info("parallelRun") {"Round #{@round}: Resuming interrupted processing. Will re-use split waypoints files."}
      # Let's check we have the correct number of
      # input ali files
      if (@num_nodes != countSplitAli(previous_round_dir+"/waypoints.ali"))
        $LOG.fatal("parallelRun") {"Round #{@round}: Was expecting #{@num_nodes} input ali files, but found #{countSplitAli(previous_round_dir+"/waypoints.ali")} in #{previous_round_dir}. Will not be able to resume processing."}
        exit 1
      end
    end
    # Generate the array of parameter & log files
    p = Array.new(@num_nodes,FrealixParameters)
    l = Array.new(@num_nodes,FrealixOutput)
    0.upto(@num_nodes-1) do |i|
      num = sprintf("%03i",i+1)
      # Prepare parameters for this node
      p[i]= @params.dclone
      p[i].filename = round_dir+"/frealix_parameters_"+num+".flx"
      if (@round == 0)
        p[i].rtp["INPUT_PAR_FILENAME"]  = round_dir+"/waypoints_"+num+".ali"
      else
        p[i].rtp["INPUT_PAR_FILENAME"]  = previous_round_dir+"/waypoints_"+num+".ali"
      end
      p[i].rtp["OUTPUT_PAR_FILENAME"] = round_dir+"/waypoints_"+num+".ali"
      p[i].rtp["OUTPUT_3D_FILENAME"]  = round_dir+"/3drec_"+num+".mrc"
      p[i].rtp["FINALISE_3D_RECONSTRUCTION"] = "F"
      #p[i].rtp["CALC_FSC"] = "F"
      #p[i].rtp["APPLY_FOM_FILTER"] = "F"
      # Prepare output log file
      l[i] = FrealixOutput.new(round_dir+"/frealix_output_"+num+".log")
    end
    # Prepare the frealix runner
    flx_runner = FrealixRunner.new(
      :num_jobs=>@num_nodes,
      :params=>p,
      :logs=>l,
      :hosts=>@avail_hosts,
      :threads_per_job=>@num_threads,
      :prefix=>@prefix,
      :gpu_nodes_only=>@gpu_nodes_only,
      :resume=>@resume)
    # Run frealix
    $LOG.info("parallelRun") {"Round #{@round}: Starting refinement on #{@num_nodes} nodes"}
    run_success = flx_runner.run()
    if (run_success)
      # Now that all jobs are done, we need to compile the results & cleanup
      $LOG.info("parallelRun") {"Round #{@round}: Now that all jobs are done, we need to compile the results & cleanup"}
      # First, we merge the .ali files
      $LOG.info("parallelRun") {"Round #{@round}: Merging the .ali files..."}
      self.mergeAli(round_dir+"/waypoints.ali",@num_nodes)
      # Now, we need to finalise the 3D
      $LOG.info("parallelRun") {"Round #{@round}: Finalising the 3D reconstruction..."}
      p = Array.new(1,FrealixParameters)
      l = Array.new(1,FrealixOutput)
      l[0] = FrealixOutput.new(round_dir+"/frealix_output_finalise.log")
      p[0] = @params.dclone
      p[0].filename = round_dir+"/frealix_parameters_finalise.flx"
      p[0].rtp["FREALIX_MODE"] = "FINALISE_3DS"
      p[0].rtp["INPUT_THREED_FILES"] = round_dir+"/3drec_???.mrc"
      p[0].rtp["OUTPUT_THREED_FILENAME"] = round_dir+"/3drec.mrc"
      p[0].rtp["INPUT_THREED_SELECTION"] = "1-#{@num_nodes}"
      p[0].rtp["HELICAL_MASKING"] = "T"
      p[0].rtp["CALC_FSC"] = "T"
      # Prepare the frealix runner
      flx_runner = FrealixRunner.new( :num_jobs=>1,:params=>p,
      :logs=>l,
      :hosts=>@avail_hosts,:prefix=>@prefix)
      # Run frealix
      run_success = flx_runner.run()
      if (run_success)
        # Clean up the directory
        $LOG.info("parallelRun") { "Round #{@round}: Cleaning up..."}
        subdirCleanup(round_dir)
      else
        $LOG.fatal("parallelRun") { "Finalisation job in round #{@round} was not succesful. See output above for detail."}
        exit -1
      end
      # Let's compute an FSC between the reconstruction and the previous reconstruction
      if (File.exists?(previous_round_dir+"/3drec_hm.mrc") && File.exists?(round_dir+"/3drec_hm.mrc"))
        $LOG.info("parllelRun") { "Round #{@round}: Computing FSC to previous round"}
        p = Array.new(1,FrealixParameters)
        l = Array.new(1,FrealixOutput)
        l[0] = FrealixOutput.new(round_dir+"/frealix_output_fsc.log")
        p[0] = @params.dclone
        p[0].filename = round_dir+"/frealix_parameters_fsc.flx"
        p[0].rtp["FREALIX_MODE"] = "CALC_FSC"
        p[0].rtp["INPUT_VOLUME_1"] = previous_round_dir+"/3drec_hm.mrc"
        p[0].rtp["INPUT_VOLUME_2"] = round_dir+"/3drec_hm.mrc"
        # Prepare the frealix runner
        flx_runner = FrealixRunner.new( :num_jobs=>1,:params=>p,
        :logs=>l,
        :hosts=>@avail_hosts,:prefix=>@prefix)
        # Run frealix
        run_success = flx_runner.run()
        if (run_success)
          # Parse the FSC
          fsc_values = l[0].getFSCValues()
          ave_fsc = fsc_values.reduce(:+) / fsc_values.length.to_f
          $LOG.info("parallelRun") {"Round #{@round}: average value of FSC to previous round = #{ave_fsc}"}
          $LOG.info("parallelRun") {"Round #{@round}: it looks like refinement has converged"} if (ave_fsc > 0.85) 
        else
          $LOG.error("parallelRun") { "Round #{@round}: FSC to previous round failed. See #{l[0].filename}" }
        end
      else
        $LOG.debug("parallelRun") { "Round #{@round}: Not doing FSC to previous round"}
      end
      # All done
      $LOG.info("parallelRun") {"Round #{@round}: All done."}
    else
      $LOG.fatal("parallelRun") { "Not all jobs in round #{@round} were succesful. See output above for detail."}
      exit -1
    end
  end

  # Cleanup a round subdirectory after parallel filament refinement
  def subdirCleanup(path)
    f = Dir.glob(["#{path}/3drec_???_{unweighted,ctfweights,fsc,ampl}*.mrc","#{path}/3drec_???_stats*.dat"])
    #puts "Will delete the following files:"
    #puts f
    f.each do |x|
      File.delete(x)
    end
  end

  # Split ali file
  def splitAli(ali_filename,num_output_files)
    # Check the ali file actually exists
    if (!File.exists?(ali_filename))
      $LOG.fatal("splitAli") {"File does not exist: #{ali_filename}"}
      puts caller(start=0)
      exit 1
    end
    # The different parts of the filename are dir+"/"+base+ext
    dir  = File.dirname (ali_filename)
    ext  = File.extname (ali_filename)
    base = File.basename(ali_filename,ext)
    # Set up the parameters for frealix
    params = FrealixParameters.new(filename=dir+"/frealix_parameters_split.flx")
    params.rtp = {
      "FREALIX_MODE"          =>  "SPLIT_PAR",
      "INPUT_FILENAME"        =>  ali_filename,
      "OUTPUT_ROOT_FILENAME"  =>  dir+"/"+base+"_???"+ext,
      "NUMBER_OF_GROUPS"      =>  num_output_files
    }
    # Set up the output log file
    log = FrealixOutput.new(filename=dir+"/frealix_output_split.log")
    # Set up the frealix run
    f = FrealixRunner.new(:params=>[params],:logs=>[log],:hosts=>@avail_hosts)
    # Run frealix
    success = f.run()
    exit -1 if (!success)
  end

  # Count the number of split ali files
  def countSplitAli(ali_filename)
    # The different parts of the filename are dir+"/"+base+ext
    dir  = File.dirname (ali_filename)
    ext  = File.extname (ali_filename)
    base = File.basename(ali_filename,ext)
    Dir.glob("#{dir}/#{base}_???#{ext}").count
  end

  # Merge ali files together
  def mergeAli(ali_filename,num_input_files)
    # The different parts of the filename are dir+"/"+base+ext
    dir  = File.dirname (ali_filename)
    ext  = File.extname (ali_filename)
    base = File.basename(ali_filename,ext)
    # Open the output file for reading (make sure it is overwritten)
    f = File.new(ali_filename,"w")
    # Read all the files in
    1.upto(num_input_files) do |i|
      num = sprintf("%03i",i)
      input_filename = dir+"/"+base+"_"+num+ext
      str = File.read(input_filename)
      # Strip the first line, unless we're on the first file
      if (i >1)
        a = str.split("\n")
        a.delete_at(0)
        str = a.join("\n")
        str = str+"\n"
      end
      # Append to file
      f.print(str)
    end
    f.close()
  end

end
