# Dealing with the sun grid engine
class Sge
  NAME = "SGE"

  # Initialiser method with default parameter
  def initialize(location='janelia',gpu_nodes_only=false)
    @user = ENV['USER']
    @loc  = location
    @qsub_exec = 'qsub'
    @qstat_exec = 'qstat'
    @qstat_interval = 20.0 #in seconds (was 60.0 at Brandeis)
    @gpu_nodes_only = gpu_nodes_only
    #puts "debug(Sge.initialize): gpu_nodes_only = #{gpu_nodes_only}, #{@gpu_nodes_only}"
    # Location-dependent settings
    case @loc
    when "brandeis"
      @qsub_exec = '/opt/gridengine/bin/lx26-amd64/qsub'
      @qstat_exec = '/opt/gridengine/bin/lx26-amd64/qstat'
      @qsub_options = {'-hard'=>'','-cwd'=>'','-r'=>'yes','-V'=>'','-b'=>'yes',
                       '-pe'=> 'multi 1','-v'=>'OMP_NUM_THREADS 1',
                       '-l'=>'mem_free=1G'}
    when "janelia"
      @qsub_options = { '-cwd'=>'','-r'=>'yes','-V'=>'','-b'=>'yes','-v'=>'OMP_NUM_THREADS 1','-pe'=>'batch 1' }
      if (@gpu_nodes_only) then
        @qsub_options << { '-l'=>'gpu=true'}
        puts "Janelia: will use only GPU nodes"
        p @qsub_options
      end
    else
      puts "Unknown location: #{@loc}"
      puts caller(start=0)
      exit 1
    end
    @jobid = []
    @job_stat = {} # Parsed output from qstat
    @job_stat_finished = {} # qstat-reported status for recently finished jobs
    
    @qstat_time_last = nil
    @qstat_force_refresh = false
    @num_jobs_waiting = 0	# Number of jobs queued waiting (any user)

    #
    self.loadQstatInfo
    self.getTotalNumSlots
  end

  attr_reader :total_num_slots

  # Return the string status of specified job, using qstat
  def getJobStatus(jobid)
    self.loadQstatInfo()
    if (@job_stat.has_key?(jobid))
      getJobStatus = @job_stat[jobid]
    elsif (@job_stat_finished.has_key?(jobid))
      getJobStatus = "finished"
    else
      getJobStatus = nil
    end
    return getJobStatus
  end

  # Return the number of jobs found waiting in the queue
  # (i.e. with status "qw")
  def numJobsWaiting()
    loadQstatInfo()
    $LOG.debug("numJobsWaiting") {"#{@num_jobs_waiting} jobs are waiting in the queue"}
    @num_jobs_waiting
  end

  def numSlotsRunning()
    loadQstatInfo()
    @num_slots_running
  end
  
  def numUsers()
    loadQstatInfo()
    @cluster_users.count
  end

  # Call qstat if it hasn't been called for more than a certain duration
  # Parse the results and load them into a hash
  def loadQstatInfo()
    command = "#{@qstat_exec} -u #{@user}"
    command_finished = "#{@qstat_exec} -s z -u #{@user}" # Command to list recently-finished jobs

    #$LOG.debug("loadQstatInfo") {"Entering. Qstat command = #{command}"}

    if (!which("qstat"))
      puts "qstat not available"
      puts caller(start=0)
      exit 1
    end

    # Only run the qstat command if it's been more than
    # @qstat_interval since the last time we did so
    if (!@qstat_time_last || (Time.now()-@qstat_time_last) > @qstat_interval || @qstat_force_refresh)
      $LOG.debug("loadQstatInfo") {"About to run qstat. Time last, current: #{@qstat_time_last} #{Time.now()}"}
      # Run the qstat command and read its output
      q = IO.popen(command,"r")
      qstat_out = q.readlines
      q.close
      # Let's parse the output
      @job_stat = {}


      @num_jobs_waiting = 0
      @num_slots_running = 0  # Number of slots used by jobs currently running
      @cluster_users = [] # All unique users of the cluster
      qstat_out.each do |line|
        next if (line.match("job-ID") || line.match("-----"))
        words = line.split
        @job_stat[words[0].to_i] = words[4] if (words[3] == "#{@user}")
        @num_jobs_waiting += 1 if (words[4] == 'qw')
        @num_slots_running += words[8].to_i if (words[4] == 'r')
        @cluster_users << words[3] if (!@cluster_users.include?(words[3]))
        $LOG.debug("loadQstatInfo") {"JobID, user, status, qw?: #{words[0]} #{words[3]} #{words[4]} #{(words[4] == 'qw')}"}
      end

      # Now run the command to find recently-finished jobs
      q = IO.popen(command_finished,"r")
      qstat_out = q.readlines
      q.close
      # Let's parse the output
      qstat_out.each do |line|
        next if (line.match("job-ID") || line.match("-----"))
        words = line.split
        @job_stat_finished[words[0].to_i] = words[4]  if words[3] == "#{@user}"
      end
      # Remember
      @qstat_time_last = Time.now()
      @qstat_force_refresh = false
    end
  end

  # Submit a job to the queue, return the sge job id or 0 if something went wrong
  def submitJob(executable,arguments,out,err,job_name,num_threads=1,mem_free_gb=1)
    @qsub_options['-v'] = "OMP_NUM_THREADS=#{num_threads}"
    case @loc
    when "brandeis"
      @qsub_options['-pe'] = "multi #{num_threads}"
      @qsub_options['-l'] = "mem_free=#{mem_free_gb}G"
    when "janelia"
      @qsub_options['-pe'] = "batch #{num_threads}"
      if (mem_free_gb > 7.5)
        puts "Janelia cluster only has 7.5 GB per slot"
        puts caller(start=0)
        exit 1
      end
    end
    qsub_options_string = @qsub_options.collect {|k,v| "#{k} #{v}"}.join(" ")
    File.delete(out) if File.exists?(out)
    File.delete(err) if File.exists?(err)
    command = "#{@qsub_exec} #{qsub_options_string} -N #{job_name} -e #{err} -o #{out} #{executable} #{arguments} "
    $LOG.debug("SGE command is: #{command}")
    q = IO.popen(command,"r")
    output = q.gets
    q.close
    success = (output =~ /Your job .* has been submitted/)?true:false
    if (success)
      jobid = output.split[2].to_i
      @qstat_force_refresh = true
      return jobid
    else
      return 0
    end
  end

  # Find the total number of cores available on the cluster
  def getTotalNumSlots()
    command = "#{@qstat_exec} -g c"
    # This will output something like:
    # CLUSTER QUEUE                   CQLOAD   USED  AVAIL  TOTAL aoACDS  cdsuE
    # -------------------------------------------------------------------------------
    # all.q                             0.00      0    368    372      0      4
    # guest.q                           0.00      0     99    103      0      4
    q = IO.popen(command,"r")
    output = q.gets(sep=nil)
    q.close
    case @loc
    when "brandeis"
      regex = /^all\.q\s*[\d\.]*\s*[\d\.]*\s*([\d]*).*$/
    when "janelia"
      regex = /^new\.q\s*[\d\.]*\s*[\d\.]*\s*([\d]*).*$/
    end
    m = output.match(regex)
    @total_num_slots = m[1]
  end

end
