#!/usr/bin/env ruby
#
# Wrapper script for frealix
#
# Alexis Rohou, March 2012
#
#  Copyright 2014 Howard Hughes Medical Institute.
#  All rights reserved.
#  Use is subject to Janelia Farm Research Campus Software Copyright 1.1
#  license terms ( http://license.janelia.org/license/jfrc_copyright_1_1.html )
#  
#


# We need rubylix, which defines classes to interact with frealix
require_relative "rubylix.rb"
require_relative "FrealixOutput.rb"
require_relative "FrealixParameters.rb"
require_relative "FrealixRefinementDirectory.rb"
require_relative "FrealixRunner.rb"
require_relative "FilamentRefineParallel.rb"


#
# Set up command line options
#
require 'fileutils'
require 'optparse'
#require 'pp'

#
# Some options are only enabled for me
#
if (ENV["USER"] == "alr99")
  developer_mode = true
else
  developer_mode = false
end


options = {}
defaults = {}
optparse = OptionParser.new do |opts|
  opts.banner = "Usage: flx_wrap.rb [options] parameters.flx"

  # Help message
  opts.on_tail("-h","--help","Show this message") do
    puts opts
    exit
  end

  # Number of rounds
  opts.on("-r","--rounds [R]",Integer,"Number of refinement rounds") do |r|
    options[:num_rounds] = r
  end
  defaults[:num_rounds] = 1

  # Start & finish cutoff resolution
  opts.on("-s","--start_cutoff_resolution [START_CUTOFF_RES]",Float,"Start cutoff resolution (Angstroms)") do |s|
    options[:start_cutoff_res] = s
  end
  opts.on("-f","--finish_cutoff_resolution [FINISH_CUTOFF_RES]",Float,"Finish cutoff resolution (Angstroms)") do |f|
    options[:finish_cutoff_res] = f
  end
  defaults[:start_cutoff_res] = 10.0
  defaults[:finish_cutoff_res] = 10.0

  # Target resolution
  opts.on("-t","--target_resolution [TARGET_RES]",Float,"Target resolution (Angstroms)") do |t|
    options[:target_res] = t
  end
  defaults[:target_res] = 5.0

  # Number of nodes
  opts.on("-n","--nodes [N]",Integer,"Number of nodes") do |n|
    options[:num_nodes] = n
  end
  defaults[:num_nodes] = 1
    
  # Use GPU nodes only (Janelia-specific)
  opts.on("-g","--gpu-nodes-only","Use GPU nodes only (Janelia-specific)") do |n|
    options[:gpu_nodes_only] = n
  end
  defaults[:gpu_nodes_only] = false

  # Starting model
  opts.on("-m","--model_3d [MODEL]","Starting 3D model") do |m|
    options[:starting_model_3d] = m
  end

  # Input alignment file
  opts.on("-a","--input_alignments [ALI]","Starting alignments") do |a|
    options[:starting_alignments_file] = a
  end

  # Prefix for the queue jobs
  opts.on("-p", "--prefix [P]","Prefix for SGE jobs") do |p|
    options[:prefix] = p
  end
  defaults[:prefix] = "flx"

  # Number of threads
  opts.on("--threads [T]",Integer,"Number of threads") do |t|
    options[:threads] = t
  end
  defaults[:threads] = 1

  # Free memory Gb
  opts.on("--free_memory_gb [N]",Float,"Free memory required per thread (Gb)  ") do |fm|
    options[:mem] = fm
  end

  # Additional hosts available
  #if (developer_mode)
  opts.on("-h","--hostsfile [extra_hosts.txt]","Flat text list of hosts available for processing. localhost and cluster are recognised.") do |h|
    options[:hostsfile] = h
  #end
  end

  # No cleanup not implemented

  # Apply defaults
  options = defaults.merge(options)

end.parse! # End of command-line options

# Parse the command-line. Remember there are two forms
# of the parse method. The 'parse' method simply parses
# ARGV, while the 'parse!' method parses ARGV and removes
# any options found there, as well as any parameters for
# the options. What's left is frealix parameter file.
optparse.parse!


#
# Finished getting options and parameters from the command line, let's start work
#

# Load extra available hosts into array
available_hosts = []
if (options.has_key?(:hostsfile))
  if (File.exists?(options[:hostsfile]))
    str = File.read(options[:hostsfile])
    available_hosts = str.split("\n")
  else
    $LOG.fatal("Error: file does not exist: #{options[:hostsfile]}")
    exit 1
  end
end
$LOG.debug("Available hosts = #{available_hosts}")

# Initialise the current directory as a Frealix refinement directory
flxdir = FrealixRefinementDirectory.new()

# What round will we be starting on
starting_round = flxdir.lastCompletedRound() + 1
$LOG.info("We will be starting on round #{starting_round} and doing #{options[:num_rounds]} refinement rounds using parameters from #{ARGV[0]}")
finishing_round = starting_round + options[:num_rounds] -1
finishing_round += 1 if (starting_round == 0)

# Load the base parameters
base_params = FrealixParameters.new(filename=ARGV[0])
base_params.readFromDisk()

# Define a list of required RTPs
required_rtps = ["CTFTILT_FILES","CTFFIND_FILES","FILM_FILES","FILM_SELECTION",
                 "TWIST_PER_SUBUNIT","RISE_PER_SUBUNIT","PIXEL_SIZE",
                 "HELIX_MAX_RADIUS"]
required_rtps_startup = required_rtps + [] #["INPUT_BOX_FILES"]
required_rtps_refine  = required_rtps + []

# Check the base parameters have the required RTPs
if (! base_params.checkForParameters?(required_rtps))
  $LOG.error("flx_wrap") {"The following parameters are missing from #{base_params.filename}: "}
  $LOG.error("flx_wrap") {base_params.findMissingParameters(required_rtps)}
  $LOG.error("flx_wrap") {"Would normally crash out now!"}
  #exit 0
end

#
# For round 0, we may need to run frealix mode FILAMENT_STARTUP
#
if (starting_round == 0)
  
  # If the directory already exists, it probably means that round 0
  # was started but aborted
  if (Dir.exists?(flxdir.currentRoundDir))
    $LOG.fatal("flx_wrap") {"Round 0 was started previously, but not completed. Please delete the folder and start again."}
    exit 1
  else
    # Make the directory
    current_round_dir = flxdir.mkdirCurrentRound()
  end

  # Clone the runtime parameters
  rtps_startup = base_params.dclone
  rtps_startup.filename = current_round_dir+"frealix_parameters.flx"

  # Supplement runtime parameters
  rtps_startup.rtp["OUTPUT_PARAMETER_FILE"] = current_round_dir+"waypoints.ali"
  rtps_startup.rtp["FILAMENT_TRACE_FILES"]  = current_round_dir+"trace_?????_?????.box"

  #
  # There are basically four possibilities:
  #    1. the user already has a frealix .ali file and a 3D model - just copy the files
  #    2. the user has neither - run filament_startup with CALC_3D_RECONSTRUCTION
  #    3. the user has an .ali file, but no 3D - run filament_refine mode with SKIP_REFINEMENT and copy the .ali
  #    4. the user as a 3D but no .ali file - run filament_startup without CALC_3D_RECONSTRUCTION and copy the 3D

  if (options[:starting_alignments_file] == nil)
    rtps_startup.rtp["FREALIX_MODE"] = "FILAMENT_STARTUP"
    if (options[:starting_model_3d] == nil)
      # User has neither model 3D or .ali file, we will run filament_startup with CALC_3D_RECONSTRUCTION
      rtps_startup.rtp["CALC_3D_RECONSTRUCTION"] = "T"
      rtps_startup.rtp["OUTPUT_3D_FILENAME"] = current_round_dir+"3drec.mrc"
    else
      # User has 3D but no .ali, run filament_startup without CALC_3D_RECONSTRUCTION and copy the 3D
      rtps_startup.rtp["CALC_3D_RECONSTRUCTION"] = "F"
      # Now copy the 3D
      starting_threed = options[:starting_model_3d]
      if (File.extname(starting_threed) == ".img" || File.extname(starting_threed) == ".hed")
        FileUtils.cp File.path(starting_threed)+File.basename(starting_threed)+".hed", current_round_dir+"/threed.hed"
        FileUtils.cp File.path(starting_threed)+File.basename(starting_threed)+".img", current_round_dir+"/threed.img"
      else
        FileUtils.cp starting_threed, current_round_dir+"3drec"+File.extname(starting_threed)
      end
    end
  else
    if (options[:starting_model_3d] == nil)
      # User has .ali, but no 3D, run filament_refine with SKIP_REFINEMENT
      rtps_startup.rtp["FREALIX_MODE"]        = "FILAMENT_REFINE"
      rtps_startup.rtp["SKIP_REFINEMENT"]     = "T"
      rtps_startup.rtp["MODEL_3D_AVAILABLE"]  = "F"
      rtps_startup.rtp["OUTPUT_3D_FILENAME"]  = current_round_dir+"3drec.mrc"
      FileUtils.cp options[:starting_alignments_file], current_round_dir+"waypoints.ali"
      rtps_startup.rtp["INPUT_PAR_FILENAME"]  = current_round_dir+"waypoints.ali"
      rtps_startup.rtp["FINALISE_3D_RECONSTRUCTION"] = 'T'
      options[:starting_alignments_file]      = current_round_dir+"waypoints.ali"
      rtps_refine = rtps_startup
    else
      # User has both .ali and 3D. Just copy files over
      FileUtils.cp options[:starting_alignments_file], current_round_dir+"waypoints.ali"
      rtps_startup.rtp["INPUT_PAR_FILENAME"]  = current_round_dir+"waypoints.ali"
      options[:starting_alignments_file]        = current_round_dir+"waypoints.ali"
      # Copy the starting 3D reconstruction there (if we are given an imagic file, copy both files)
      starting_threed = options[:starting_model_3d]
      if (File.extname(starting_threed) == ".img" || File.extname(starting_threed) == ".hed")
        FileUtils.cp File.path(starting_threed)+File.basename(starting_threed)+".hed", current_round_dir+"/threed.hed"
        FileUtils.cp File.path(starting_threed)+File.basename(starting_threed)+".img", current_round_dir+"/threed.img"
      else
        FileUtils.cp starting_threed, current_round_dir+"3drec"+File.extname(starting_threed)
      end
    end #if starting 3d
  end # if starting ali


  # Actually launch the job if needed
  if (options[:starting_model_3d] == nil or options[:starting_alignments_file] == nil)
    if (rtps_startup.rtp["FREALIX_MODE"] == "FILAMENT_STARTUP")
      # Check the base parameters have the required RTPs
      if (! rtps_startup.checkForParameters?(required_rtps_startup))
  	    $LOG.error("flx_wrap") {"The following parameters are missing from #{rtps_startup.filename}: "}
  	    $LOG.error("flx_wrap") { rtps_startup.findMissingParameters(required_rtps_startup)}
  	    $LOG.error("flx_wrap") {"Would normally crash out now!"}
  	    #exit 0
	    end
      # We are running in startup mode
      log = FrealixOutput.new(current_round_dir+"/frealix_output.log")
      flx_runner = FrealixRunner.new( :num_jobs=>1,:params=>[rtps_startup],
                                      :logs=>[log],:hosts=>available_hosts,
                                      :threads_per_job=>options[:threads],
                                      :mem_per_thread_gb=>options[:mem],
                                      :prefix=>options[:prefix],
                                      :gpu_nodes_only=>options[:gpu_nodes_only])

      flx_runner.run()

    else
      # Check the base parameters have the required RTPs
  	  if (! rtps_refine.checkForParameters?(required_rtps_refine))
  	    $LOG.error("flx_wrap") {"The following parameters are missing from #{rtps_refine.filename}: "}
  	    $LOG.error("flx_wrap") {rtps_refine.findMissingParameters(required_rtps_refine)}
  	    $LOG.error("flx_wrap") {"Would normally crash out now!"}
  	    #exit 0
  	  end
      # We are running in refine mode (because an .ali file was supplied)
      flx_refine_para = FilamentRefineParallel.new(
									:num_nodes=>options[:num_nodes],
									:num_threads=>options[:threads],
									:params=>rtps_startup,
									:hosts=>available_hosts,
									:mem_per_thread_gb=>options[:mem],
									:prefix=>options[:prefix],
									:gpu_nodes_only=>options[:gpu_nodes_only])
	  flx_refine_para.parallelRun()
    end
  end

  starting_round = 1
end # round 0

#
# For rounds 1+, we need to run frealix mode FILAMENT_REFINE
#
starting_round.upto(finishing_round) do |round|
  # The directories
  previous_round_dir = flxdir.previousRoundDir
  current_round_dir  = flxdir.currentRoundDir
  resume_processing = false
  if (Dir.exists?(current_round_dir))
    if (round == starting_round)
      $LOG.warn("flx_wrap") {"Will attempt to resume processing at round #{round}"}
      resume_processing = true
    else
      $LOG.fatal("flx_wrap") {"It looks like round #{round} was started previously. Please delete the folder and start again"}
      exit 1
    end
  else
    flxdir.mkdirCurrentRound()
  end
  # Prepare the runtime parameters
  rtps_refine = base_params.dclone
  rtps_refine.rtp["FREALIX_MODE"] = "FILAMENT_REFINE"
  rtps_refine.rtp["TARGET_RESOLUTION"] = options[:target_res]
  if (options.has_key?(:starting_alignments_file))
    rtps_refine.rtp["INPUT_PAR_FILENAME"] = options[:starting_alignments_file]
  else
    rtps_refine.rtp["INPUT_PAR_FILENAME"] = previous_round_dir+"/waypoints.ali"
  end
  rtps_refine.rtp["OUTPUT_PAR_FILENAME"] = current_round_dir+"/waypoints.ali"
  # For the input model 3D, we have to deal with the user option as well
  # as the fact that for round 0, we may be given a spi,img or mrc
  if (options.has_key?(:starting_model_3d) && round == starting_round)
    rtps_refine.rtp["MODEL_3D_FILENAME"] = options[:starting_model_3d]
  else
    if (round == 1)
      rtps_refine.rtp["MODEL_3D_FILENAME"] = Dir.glob([previous_round_dir+"/3drec.{hed,spi,mrc}"])[0]
    else
      rtps_refine.rtp["MODEL_3D_FILENAME"] = previous_round_dir+"/3drec.mrc"
    end
  end
  rtps_refine.rtp["OUTPUT_3D_FILENAME"]          = current_round_dir+"/3drec.mrc"
  rtps_refine.rtp["MATCHING_PROJECTIONS_FILES"]  = current_round_dir+"/matching_proj_?????.mrc"
  rtps_refine.rtp["FILAMENT_TRACE_FILES"]        = current_round_dir+"/trace_?????_?????.box"
  if (rtps_refine.rtp["CALC_3D_RECONSTRUCTION"] == 'F')
    $LOG.warn("flx_wrap") {"You specified CALC_3D_RECONSTRUCTION = F, so no 3D will be computed"}
  else
    rtps_refine.rtp["CALC_3D_RECONSTRUCTION"]      = 'T'
  end
  # Deal with target_resolution, which can be set statically by the
  # user-supplied RTPs, or dynamically here according to command-line
  # options
  if ( options.has_key?(:start_cutoff_res) && options.has_key?(:finish_cutoff_res))
    if (options[:num_rounds] == 1)
      linear_factor = 0.0
    else
      linear_factor = (round - starting_round).to_f/((options[:num_rounds])-1).to_f
    end
    rtps_refine.rtp["HIGH_RESOLUTION_CUTOFF"] =  options[:start_cutoff_res] *(1.-linear_factor) + options[:finish_cutoff_res] *linear_factor
    $LOG.info("flx_wrap") { "Round #{round}: Cutoff resolution = #{rtps_refine.rtp["HIGH_RESOLUTION_CUTOFF"]}, target resolution = #{rtps_refine.rtp["TARGET_RESOLUTION"]} A"}
  end

  # Do the actual refinement for this round
  flx_refine_para = FilamentRefineParallel.new(
								:num_nodes=>options[:num_nodes],
								:num_threads=>options[:threads],
								:params=>rtps_refine,
								:hosts=>available_hosts,
								:mem_per_thread_gb=>options[:mem],
								:prefix=>options[:prefix],
								:gpu_nodes_only=>options[:gpu_nodes_only],
								:resume=>resume_processing)
  flx_refine_para.parallelRun()


end
