#!/usr/bin/env ruby
#  select_filaments.rb
#
#  Select and modify filaments from a waypoints.ali file 
#
#  Copyright 2014 Howard Hughes Medical Institute.
#  All rights reserved.
#  Use is subject to Janelia Farm Research Campus Software Copyright 1.1
#  license terms ( http://license.janelia.org/license/jfrc_copyright_1_1.html )
#  
#

require_relative "rubylix.rb"
require_relative "FrealixFilament.rb"
require_relative "FrealixAli.rb"

require 'optparse'

options = {}
defaults = {}

optparse = OptionParser.new do |opts|
  opts.banner = "Usage: select_filaments.rb [options] input output"

  # Help message
  opts.on_tail("-h","--help","Show this message") do
    puts opts
    exit
  end

  options[:score_below] = 999.9
  opts.on("--score-below SCORE", Float, "Only filaments with scores lower than this are selected") do |b|
    options[:score_below] = b
  end

  options[:score_above] = -999.9
  opts.on("--score-above SCORE", Float, "Only filaments with scores higher than this are selected") do |b|
    options[:score_above] = b
  end

  options[:include_list_file] = nil
  opts.on("--include-list-file FILE","Only filament with IDs listed in FILE") do |file|
    options[:include_list_file] = file
  end

  options[:exclude_list_file] = nil
  opts.on("--exclude-list-file FILE","Remove filaments with IDs listed in FILE") do |file|
    options[:exclude_list_file] = file
  end

  options[:include_pos_list_file] = nil
  opts.on("--include-pos-list-file FILE","Only filament with positions listed in FILE (starting at 1)") do |file|
    options[:include_pos_list_file] = file
  end

  options[:exclude_pos_list_file] = nil
  opts.on("--exclude-pos-list-file FILE","Exclude filaments with positions listed in FILE (starting at 1)") do |file|
    options[:exclude_pos_list_file] = file
  end

  options[:refine_to_true] = nil
  opts.on("--refine-to-true","Set the refine flags for selected filaments to T (all filaments are kept in the output)") do |r|
    options[:refine_to_true] = r
  end

  options[:refine_to_false] = nil
  opts.on("--refine-to-false","Set the refine flags for selected filaments to F (all filaments are kept in the output)") do |r|
    options[:refine_to_false] = r
  end

  options[:fal_refine_to_true] = nil
    opts.on("--fal-refine-to-true","Set the refine flags for first and last waypoints of selected filaments to T") do |r|
 	options[:fal_refine_to_true] = r
  end

  options[:fal_refine_to_false] = nil
    opts.on("--fal-refine-to-false","Set the refine flags for first and last waypoints of selected filaments to F") do |r|
    options[:fal_refine_to_false] = r
  end


  options[:subunit_num_to_zero] = nil
  opts.on("--subunit-num-to-zero","Set all subunit_num values to 0.0") do |r|
    options[:subunit_num_to_zero] = r
  end

  options[:x_shift] = nil
  opts.on("--x-shift SHIFT","Shift all waypoints in X by this (Angstroms)") do |r|
	  options[:x_shift] = r
  end

  options[:y_shift] = nil
  opts.on("--y-shift SHIFT","Shift all waypoints in Y by this (Angstroms)") do |r|
	  options[:y_shift] = r
  end

end

optparse.parse!

@input_waypoints_filename = ARGV[0]
@output_waypoints_filename = ARGV[1]
@select_filaments_with_score_below = options[:score_below]
@select_filaments_with_score_above = options[:score_above]


#
# Get filaments from the waypoints file
#
ali_in = FrealixAli.new(filename=@input_waypoints_filename)
fils = FrealixFilament.filamentsFromWaypoints(ali_in.getWaypoints)
# Sort filaments by ID, so that their positions are meaningful
fils.sort!{|x,y| x.num <=> y.num}

$LOG.info() {"There are #{fils.size} filaments in input waypoints file"}


#
# If we were given a file containing a list, let's parse it
# if not, all filaments go in
#
include_list = []
if (options[:include_list_file])
  str = File.read(options[:include_list_file])
  str.split("\n").each {|n| include_list << n.to_i}
  puts "Will include only filaments with IDs: #{include_list}"
else
  fils.each {|fil| include_list << fil.num}
end

exclude_list = []
if (options[:exclude_list_file])
  str = File.read(options[:exclude_list_file])
  str.split("\n").each {|n| exclude_list << n.to_i}
  puts "Will exclude filaments with IDs: #{exclude_list}"
end

#
# If we were given a file containing a list of positions, let's parse it
# if not, all filaments go in
#
include_pos_list = []
if (options[:include_pos_list_file])
  str = File.read(options[:include_pos_list_file])
  str.split("\n").each {|n| include_pos_list << n.to_i}
  puts "Will include only filaments with positions: #{include_pos_list}"
else
  fils.each_with_index{|fil,i| include_pos_list << i+1}
end
#
exclude_pos_list = []
if (options[:exclude_pos_list_file])
  str = File.read(options[:exclude_pos_list_file])
  str.split("\n").each {|n| exclude_pos_list << n.to_i}
  puts "Will exclude filaments with positions: #{exclude_pos_list}"
end


#
# Select filaments
#
fils_out = []
fils.each_with_index { |fil,i|
	selected = (	fil.score < @select_filaments_with_score_below 	&&
			fil.score >= @select_filaments_with_score_above &&
			include_list.include?(fil.num) 			&&
			!exclude_list.include?(fil.num)			&&
			include_pos_list.include?(i+1) 			&&
			!exclude_pos_list.include?(i+1))
	if (selected)
		fils_out << fil
	end
}

#
# Zero values if requested
#
fils_out.each { |fil|
	if (options[:subunit_num_to_zero])
		fil.wps.each{|wp| wp.subunit_num = 0.0}
	end
	if (options[:refine_to_true])
		fil.wps.each{|wp| wp.refine = "T"}
	end
	if (options[:refine_to_false])
		fil.wps.each{|wp| wp.refine = "F"}
	end
	if (options[:fal_refine_to_true])
        fil.wps[0].refine = "T"
		fil.wps[-1].refine = "T"
    end
    if (options[:fal_refine_to_false])
        fil.wps[0].refine = "F"
		fil.wps[-1].refine = "F"
	end
	# X and Y shifts
	fil.wps.each{|wp| 
		wp.x += options[:x_shift].to_f if (options[:x_shift])
		wp.y += options[:y_shift].to_f if (options[:y_shift])
	}
}

#
# Write out the waypoints file
#
FrealixFilament::filamentsToAliFile(fils_out,@output_waypoints_filename)

$LOG.info() {"#{fils_out.size} filaments are included in the output ali file"}
