!>  \brief  CTF functions etc.
MODULE CTFS

    TYPE CTF_PARAMETERS
        LOGICAL ::  PIX_UNITS   =   .FALSE. !< Whether CTF parameters are currently stored in metric or pixel units
        REAL    ::  CS          =   0.0     !< Spherical aberration (mm or pix)
        REAL    ::  WL          =   0.0     !< Wavelength (Angstrom or pix)
        REAL    ::  AMPL_CONT   =   0.0     !< Fraction of amplitude contrast (from 0. to 1., usually btw .07 and .14, see Mindell 03)
        REAL    ::  DFMID1      =   0.0     !< Underfocus along first axis (Angstrom or pix) (positive values for underfocus; larger value = more underfocus)
        REAL    ::  DFMID2      =   0.0     !< Underfocus along second axis (Angstrom or pix)
        REAL    ::  ANGAST      =   0.0     !< Azimuth of first axis. 0.0 means axis is at 3 o'clock. (degrees)
    END TYPE CTF_PARAMETERS


    INTERFACE ANG_TO_PIX
        MODULE PROCEDURE CTF_ANG_TO_PIX
    END INTERFACE ANG_TO_PIX

    INTERFACE PIX_TO_ANG
        MODULE PROCEDURE CTF_PIX_TO_ANG
    END INTERFACE PIX_TO_ANG

    INTERFACE OPERATOR (.EQ.)
        MODULE PROCEDURE IS_EQUAL
    END INTERFACE

    INTERFACE OPERATOR (.NE.)
        MODULE PROCEDURE IS_NOT_EQUAL
    END INTERFACE

    INTERFACE OPERATOR (-)
        MODULE PROCEDURE CTF_MINUS_CTF
    END INTERFACE

    INTERFACE OPERATOR (+)
        MODULE PROCEDURE CTF_PLUS_CTF
    END INTERFACE

    INTERFACE OPERATOR (.LT.)
        MODULE PROCEDURE CTF_LT_CTF
    END INTERFACE

    INTERFACE ABS
        MODULE PROCEDURE CTF_ABS
    END INTERFACE ABS

CONTAINS

    SUBROUTINE CTF_UNIT_TESTS()
        CALL CTF_UNIT_TEST_1()
        CALL CTF_UNIT_TEST_2()
    END SUBROUTINE CTF_UNIT_TESTS

    SUBROUTINE CTF_UNIT_TEST_1()
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        !
        TYPE(CTF_PARAMETERS)    ::  SELF
        REAL,   ALLOCATABLE     ::  ZEROES(:)
        INTEGER                 ::  I
        REAL                    ::  CURRENT_CTF
        REAL,   PARAMETER       ::  PIXEL_SIZE = 2.4
        REAL,   PARAMETER       ::  TOL = 0.001
        !
        WRITE(*,'(A)') '**info(CTF_UNIT_TEST_1): entering'
        CALL CTF_PARAMETERS_INIT(SELF,AKV=200.0,CS=2.0,AMPL_CONT=0.07,DFMID1=3.2,DFMID2=3.2,ANGAST=0.0)
        CALL CTF_PRINT_INFO(SELF)
        CALL CTF_ANG_TO_PIX(SELF,PIXEL_SIZE=PIXEL_SIZE)
        CALL CTF_ZEROES(SELF,10,ZEROES)
        DO I=1,SIZE(ZEROES)
            CURRENT_CTF =  EVAL_CTF(SELF,BOX_SIZE=1000,RAD=1000.*ZEROES(I),ANG=0.0)
            PRINT *, I, PIXEL_SIZE/ZEROES(I),CURRENT_CTF
            IF (ABS(CURRENT_CTF) .GT. TOL) THEN
                CALL TERMINATE('Failed unit test')
            ENDIF
        ENDDO
        WRITE(*,'(A)') '**info(CTF_UNIT_TEST_1): passed'
    END SUBROUTINE CTF_UNIT_TEST_1

    SUBROUTINE CTF_UNIT_TEST_2()
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        !
        TYPE(CTF_PARAMETERS)    ::  SELF
        REAL,   ALLOCATABLE     ::  DEFS(:)
        INTEGER                 ::  I
        REAL                    ::  CURRENT_CTF
        REAL,   PARAMETER       ::  PIXEL_SIZE = 1.1
        REAL,   PARAMETER       ::  TOL = 0.99
        REAL,   PARAMETER       ::  TARGET_SPACING = 4.8
        INTEGER,PARAMETER       ::  BOX_SIZE = 10000
        REAL                    ::  TARGET_SF
        !
        WRITE(*,'(A)') '**info(CTF_UNIT_TEST_2): entering'
        WRITE(*,'(A,F0.3)') '**info(CTF_UNIT_TEST_2): target spacing (A) = ', TARGET_SPACING
        WRITE(*,'(A,F0.3)') '**info(CTF_UNIT_TEST_2): pixel size (A) = ', PIXEL_SIZE
        TARGET_SF = PIXEL_SIZE/TARGET_SPACING
        WRITE(*,'(A,F0.3)') '**info(CTF_UNIT_TEST_2): target SF (1/pixels) = ', TARGET_SF

        CALL CTF_PARAMETERS_INIT(SELF,AKV=300.0,CS=2.0,AMPL_CONT=0.07,DFMID1=3.2,DFMID2=3.2,ANGAST=0.0)
        CALL CTF_PRINT_INFO(SELF)
        CALL CTF_ANG_TO_PIX(SELF,PIXEL_SIZE=PIXEL_SIZE)
        CALL CTF_FIND_BEST_DEFS_FOR_SF(SELF,TARGET_SF,20,DEFS)
        DO I=1,SIZE(DEFS)
            SELF%DFMID1 = DEFS(I)
            SELF%DFMID2 = DEFS(I)
            CURRENT_CTF =  EVAL_CTF(SELF,BOX_SIZE=BOX_SIZE,RAD=BOX_SIZE*TARGET_SF,ANG=0.0)
            PRINT '(I6,F16.1,F10.2)', I, DEFS(I)*PIXEL_SIZE,CURRENT_CTF
            IF (ABS(CURRENT_CTF) .LT. TOL) THEN
                CALL TERMINATE('Failed unit test')
            ENDIF
        ENDDO
        WRITE(*,'(A)') '**info(CTF_UNIT_TEST_2): passed'
    END SUBROUTINE CTF_UNIT_TEST_2

    !>  \brief  Return an array with the CTF value at each radius
    SUBROUTINE CTF_TO_CURVE(CTF,CURVE,NUM_POINTS,PIXEL_SIZE)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),                   INTENT(INOUT)   ::  CTF         !< CTF object specifying CTF parameters
        REAL,                   ALLOCATABLE,    INTENT(INOUT)   ::  CURVE(:)
        INTEGER,                                INTENT(IN)      ::  NUM_POINTS
        REAL,                   OPTIONAL,       INTENT(IN)      ::  PIXEL_SIZE  !< Pixel size (required if the CTF is supplied in Angstroms)
        ! Result

        ! Private variables
        INTEGER ::  IERR, I
        ! Start work

        ! Allocate output
        IF (ALLOCATED(CURVE)) DEALLOCATE(CURVE)
        ALLOCATE(CURVE(NUM_POINTS),STAT=IERR)
        IF (IERR .NE. 0) THEN
            CALL TERMINATE('Memory allocation failed in CTF_TO_CURVE')
        ENDIF

        ! If we need the pixel size, check the user gave it
        IF ( .NOT. CTF%PIX_UNITS) THEN
            IF (PRESENT(PIXEL_SIZE)) THEN
                CALL ANG_TO_PIX(CTF,PIXEL_SIZE)
            ELSE
                WRITE(*,'(A)') '**ERROR(CTF_TO_CURVE): CTF parameters in Angstrom but no pixel size given'
                CALL TERMINATE('Fatal error in CTF_TO_CURVE')
            ENDIF
        ENDIF

        ! Loop over points in the curve
        DO I=1,SIZE(CURVE)
            CURVE(I) = EVAL_CTF(CTF,SIZE(CURVE)*2,REAL(I-1),0.)
        ENDDO

    END SUBROUTINE CTF_TO_CURVE

    !>  \brief  Initialise a CTF_PARAMETERS object
    SUBROUTINE CTF_PARAMETERS_INIT(SELF,AKV,CS,AMPL_CONT,DFMID1,DFMID2,ANGAST)
        ! Arguments
        TYPE(CTF_PARAMETERS),               INTENT(INOUT)   ::  SELF        !<  CTF parameters object
        REAL,                               INTENT(IN)      ::  AKV         !<  Acceleration voltage (kV)
        REAL,                               INTENT(IN)      ::  CS          !<  Spherical aberration (mm)
        REAL,                               INTENT(IN)      ::  AMPL_CONT   !<  Amplitude contrast (fraction, e.g. 0.07 for 7%)
        REAL,                               INTENT(IN)      ::  DFMID1      !<  Defocus (on major axis if astigmatism) (um)
        REAL,                   OPTIONAL,   INTENT(IN)      ::  DFMID2      !<  Defocus on minor axis
        REAL,                   OPTIONAL,   INTENT(IN)      ::  ANGAST      !<  Azimuth of major axis. 0.0 means axis is at 3 o'clock. (degrees)
        ! Private variables
        REAL,   PARAMETER   ::  UM_TO_A =   10000.
        ! Start work

        ! We initialise the parameters not in pixel units
        SELF%PIX_UNITS  =   .FALSE.
        SELF%WL         =   AKV_TO_WL(AKV)
        SELF%CS         =   CS
        SELF%AMPL_CONT  =   AMPL_CONT
        SELF%DFMID1     =   DFMID1 * UM_TO_A
        IF (PRESENT(DFMID2)) THEN
            SELF%DFMID2 =   DFMID2 * UM_TO_A
        ELSE
            SELF%DFMID2 =   DFMID1 * UM_TO_A
        ENDIF
        IF (PRESENT(ANGAST)) THEN
            SELF%ANGAST =   ANGAST
        ELSE
            SELF%ANGAST =   0.0
        ENDIF
    END SUBROUTINE CTF_PARAMETERS_INIT

    !>  \brief Test whether two CTFs are different
    ELEMENTAL PURE LOGICAL FUNCTION IS_NOT_EQUAL(CTF1,CTF2)
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF1
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF2
        ! Start work
        IS_NOT_EQUAL =  .NOT. IS_EQUAL(CTF1,CTF2)
    END FUNCTION IS_NOT_EQUAL

    !>  \brief Test whether two CTFs are equal
    ELEMENTAL PURE LOGICAL FUNCTION IS_EQUAL(CTF1,CTF2)
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF1
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF2
        ! Start work
        IS_EQUAL =  CTF1%CS         .EQ.    CTF2%CS         .AND.   &
                    CTF1%WL         .EQ.    CTF2%WL         .AND.   &
                    CTF1%AMPL_CONT  .EQ.    CTF2%AMPL_CONT  .AND.   &
                    CTF1%DFMID1     .EQ.    CTF2%DFMID1     .AND.   &
                    CTF1%DFMID2     .EQ.    CTF2%DFMID2     .AND.   &
                    CTF1%ANGAST     .EQ.    CTF2%ANGAST
    END FUNCTION IS_EQUAL

    !>  \brief  Return the absolute of a set of CTF parameters
    ELEMENTAL PURE FUNCTION CTF_ABS(CTF)
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF
        ! Result
        TYPE(CTF_PARAMETERS)    ::  CTF_ABS
        ! Start work
        CTF_ABS%PIX_UNITS   =   CTF%PIX_UNITS
        CTF_ABS%CS          =   ABS(CTF%CS)
        CTF_ABS%WL          =   ABS(CTF%WL)
        CTF_ABS%AMPL_CONT   =   ABS(CTF%AMPL_CONT)
        CTF_ABS%DFMID1      =   ABS(CTF%DFMID1)
        CTF_ABS%DFMID2      =   ABS(CTF%DFMID2)
        CTF_ABS%ANGAST      =   ABS(CTF%ANGAST)
    END FUNCTION CTF_ABS

    !>  \brief  Test with CTF1 is less than CTF2
    ELEMENTAL PURE LOGICAL FUNCTION CTF_LT_CTF(CTF1,CTF2) RESULT(IS_LT)
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF1
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF2
        ! Start work
        IS_LT    =  CTF1%CS         .LT.    CTF2%CS         .AND.   &
                    CTF1%WL         .LT.    CTF2%WL         .AND.   &
                    CTF1%AMPL_CONT  .LT.    CTF2%AMPL_CONT  .AND.   &
                    CTF1%DFMID1     .LT.    CTF2%DFMID1     .AND.   &
                    CTF1%DFMID2     .LT.    CTF2%DFMID2     .AND.   &
                    CTF1%ANGAST     .LT.    CTF2%ANGAST
    END FUNCTION CTF_LT_CTF

    !>  \brief  Subtract one set of CTF parameters from another
    !!
    !!  Note: we assume, but do not check, that both sets of parameters are in the same units!
    ELEMENTAL PURE FUNCTION CTF_MINUS_CTF(CTF1,CTF2)
        USE EULERANGLES, ONLY : ANGLE_DIFFERENCE
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF1
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF2
        ! Result
        TYPE(CTF_PARAMETERS)    ::  CTF_MINUS_CTF
        ! Start work
        CTF_MINUS_CTF%PIX_UNITS =   CTF1%PIX_UNITS
        CTF_MINUS_CTF%CS        =   CTF1%CS - CTF2%CS
        CTF_MINUS_CTF%WL        =   CTF1%WL - CTF2%WL
        CTF_MINUS_CTF%AMPL_CONT =   CTF1%AMPL_CONT - CTF2%AMPL_CONT
        CTF_MINUS_CTF%DFMID1    =   CTF1%DFMID1 - CTF2%DFMID1
        CTF_MINUS_CTF%DFMID2    =   CTF1%DFMID2 - CTF2%DFMID2
        CTF_MINUS_CTF%ANGAST    =   ANGLE_DIFFERENCE(DBLE(CTF1%ANGAST),DBLE(CTF2%ANGAST))
    END FUNCTION CTF_MINUS_CTF

    !>  \brief  Add one set of CTF parameters to another
    !!
    !!  Note: we assume, but do not check, that both sets of parameters are in the same units!
    ELEMENTAL PURE FUNCTION CTF_PLUS_CTF(CTF1,CTF2)
        USE EULERANGLES, ONLY : ANGLE_BTWN_0_AND_360
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF1
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF2
        ! Result
        TYPE(CTF_PARAMETERS)    ::  CTF_PLUS_CTF
        ! Start work
        CTF_PLUS_CTF%PIX_UNITS =   CTF1%PIX_UNITS
        CTF_PLUS_CTF%CS        =   CTF1%CS + CTF2%CS
        CTF_PLUS_CTF%WL        =   CTF1%WL + CTF2%WL
        CTF_PLUS_CTF%AMPL_CONT =   CTF1%AMPL_CONT + CTF2%AMPL_CONT
        CTF_PLUS_CTF%DFMID1    =   CTF1%DFMID1 + CTF2%DFMID1
        CTF_PLUS_CTF%DFMID2    =   CTF1%DFMID2 + CTF2%DFMID2
        CTF_PLUS_CTF%ANGAST    =   ANGLE_BTWN_0_AND_360(DBLE(CTF1%ANGAST) + DBLE(CTF2%ANGAST))
    END FUNCTION CTF_PLUS_CTF

    !>  \brief  Print out information describing the CTF
    SUBROUTINE CTF_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  SELF    !<  CTF object
        ! Private variables
        CHARACTER(LEN=2),       PARAMETER   ::  TEXT_WIDTH  =   '25'
        ! Start work
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')        'Pixel units:',         SELF%PIX_UNITS
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F12.2)')     'Cs:',                  SELF%CS
        WRITE(*,'(A'//TEXT_WIDTH//',1X,ES12.4)')     'Wavelength:',         SELF%WL
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.2)')      'Amplitude contrast:',  SELF%AMPL_CONT
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.2)')      'Defocus 1:',           SELF%DFMID1
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.2)')      'Defocus 2:',           SELF%DFMID2
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.2)')      'Astigmatism azimuth:', SELF%ANGAST
    END SUBROUTINE CTF_PRINT_INFO

    !>  \brief  Modify the CTF object so that its average defocus is specified by Z
    SUBROUTINE CTF_AT_Z(CTF,Z,PIXEL_SIZE)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, Z_AXIS_DIRECTION
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),               INTENT(INOUT)   ::  CTF     !<  CTF object to be modified
        REAL(KIND=8),                       INTENT(IN)      ::  Z       !<  Z-height (in pixels)
        REAL,                   OPTIONAL,   INTENT(IN)      ::  PIXEL_SIZE  !<  Pixel size (in Angstroms). Only needed if the CTF is not supplied in pixels.
        ! Private variables
        REAL    ::  ASTIG1, ASTIG2, AV_DEF
        ! Start work

        ! Make sure the CTF is in pixel units
        IF (.NOT. CTF%PIX_UNITS) THEN
            IF (PRESENT(PIXEL_SIZE)) THEN
                CALL CTF_ANG_TO_PIX(CTF,PIXEL_SIZE)
            ELSE
                WRITE(*,'(A)') '**ERROR(CTF_AT_Z): CTF not supplied in pixel units and pixel size not given'
                CALL TERMINATE('Fatal error in CTF_AT_Z')
            ENDIF
        ENDIF

        !
        AV_DEF = (CTF%DFMID1+CTF%DFMID2)/2.
        ASTIG1 = CTF%DFMID1 - AV_DEF
        ASTIG2 = CTF%DFMID2 - AV_DEF

        ! Z_AXIS_DIRECTION defines whether the Z axis goes up or down the microscope column. See comments in runtime_parmaters.f90 for more details.
        CTF%DFMID1 = Z_AXIS_DIRECTION * Z + ASTIG1
        CTF%DFMID2 = Z_AXIS_DIRECTION * Z + ASTIG2
    END SUBROUTINE CTF_AT_Z

    !>  \brief  Return the Z coordinate of a point, given its CTF information
    REAL FUNCTION Z_FROM_CTF(CTF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, Z_AXIS_DIRECTION
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF
        ! Private variables

        ! Start work

        ! Make sure the CTF is in pixel units
        IF (.NOT. CTF%PIX_UNITS) THEN
            WRITE(*,'(A)') '**ERROR(Z_FROM_CTF): CTF not supplied in pixel units'
            CALL TERMINATE('Fatal error in Z_FROM_CTF')
        ENDIF

        Z_FROM_CTF = Z_AXIS_DIRECTION * (CTF%DFMID1+CTF%DFMID2)/2.
    END FUNCTION Z_FROM_CTF

    !>  \brief Convert CTF from Angstrom & degrees to pixels & radians
    SUBROUTINE CTF_ANG_TO_PIX(CTF1,PIXEL_SIZE)
        !USE RUNTIME_PARAMETERS, ONLY: TERMINATE
        USE ROTATION_MATRICES, ONLY : RADIANS
        IMPLICIT NONE
        TYPE(CTF_PARAMETERS),   INTENT(INOUT)   ::  CTF1            !< CTF object to convert
        REAL,                   INTENT(IN)      ::  PIXEL_SIZE  !< Pixel size (Angstrom)
        ! Private variables

        ! Start work
        IF (CTF1%PIX_UNITS) THEN
            !WRITE(*,'(A)') '**ERROR(CTF_ANG_TO_PIX): trying to convert CTF parameters already converted'
            !CALL TERMINATE('Fatal error in CTF_ANG_TO_PIX')
        ELSE
            ! Do actual conversions
            CTF1%CS     =   CTF1%CS * 10.0**7 / PIXEL_SIZE      ! mm to A to pix
            CTF1%WL     =   CTF1%WL / PIXEL_SIZE                ! A to pix
            CTF1%DFMID1 =   CTF1%DFMID1 / PIXEL_SIZE            ! A to pix
            CTF1%DFMID2 =   CTF1%DFMID2 / PIXEL_SIZE            ! A to pix
            CTF1%ANGAST =   RADIANS(CTF1%ANGAST)            ! deg to rad
            CTF1%PIX_UNITS = .TRUE.
        ENDIF
    END SUBROUTINE CTF_ANG_TO_PIX

    !>  \brief Convert CTF from pixels & radians to Angstrom & degrees
    SUBROUTINE CTF_PIX_TO_ANG(CTF1,PIXEL_SIZE)
        USE RUNTIME_PARAMETERS, ONLY: TERMINATE
        USE ROTATION_MATRICES, ONLY : DEGREES
        IMPLICIT NONE
        TYPE(CTF_PARAMETERS),   INTENT(INOUT)   ::  CTF1            !< CTF object to convert
        REAL,       INTENT(IN)      ::  PIXEL_SIZE  !< Pixel size (Angstrom)
        ! Private variables

        ! Start work
        IF (.NOT. CTF1%PIX_UNITS) THEN
            WRITE(*,'(A)') '**ERROR(CTF_PIX_TO_ANG): trying to convert CTF parameters already converted'
            CALL TERMINATE('Fatal error in CTF_PIX_TO_ANG')
        ELSE
            ! Do actual conversions
            CTF1%CS     =   CTF1%CS / 10.0E7 * PIXEL_SIZE   ! pix to mm
            CTF1%WL     =   CTF1%WL * PIXEL_SIZE                ! pix to A
            CTF1%DFMID1 =   CTF1%DFMID1 * PIXEL_SIZE            ! pix to A
            CTF1%DFMID2 =   CTF1%DFMID2 * PIXEL_SIZE            ! pix to A
            CTF1%ANGAST =   DEGREES(CTF1%ANGAST)            ! rad to deg
            CTF1%PIX_UNITS = .FALSE.
        ENDIF
    END SUBROUTINE CTF_PIX_TO_ANG


    !>  \brief  Find the defocus values for which the CTF is maximal at the given target spatial frequency
    SUBROUTINE CTF_FIND_BEST_DEFS_FOR_SF(SELF,TARGET_SF,NUMBER_OF_DEFS,DEFS)
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),   INTENT(IN)      ::  SELF                !<  CTF parameters (defocus will be ignored)
        REAL,                   INTENT(IN)      ::  TARGET_SF           !<  Target spatial frequency (1/pixels)
        INTEGER,                INTENT(IN)      ::  NUMBER_OF_DEFS      !<  Number of defocus values you want
        REAL,   ALLOCATABLE,    INTENT(INOUT)   ::  DEFS(:)             !<  Defocus values (pixels)
        ! Private variables
        REAL,   ALLOCATABLE                     :: DEFS_POS(:), DEFS_NEG(:)
        INTEGER ::  I,J
        ! Start work

        ! Allocate memory for results
        IF (ALLOCATED(DEFS)) DEALLOCATE(DEFS)
        ALLOCATE(DEFS(NUMBER_OF_DEFS))
        ALLOCATE(DEFS_POS(NUMBER_OF_DEFS),DEFS_NEG(NUMBER_OF_DEFS))

        ! Find the position of the first few times the CTF gets to 1.0 and -1.0
        CALL CTF_SOLVE_FOR_ARG(SELF%AMPL_CONT,DEFS_POS,CTF_VALUE= 1.0)
        CALL CTF_SOLVE_FOR_ARG(SELF%AMPL_CONT,DEFS_NEG,CTF_VALUE=-1.0)

        ! Convert the arguments to defocus values
        CALL CTF_DF_FROM_ARG(SELF%CS,SELF%WL,DEFS_NEG(1),TARGET_SF,DEFS(1))
        DO I=2,NUMBER_OF_DEFS-2,2
            J = (I-1)+1
            CALL CTF_DF_FROM_ARG(SELF%CS,SELF%WL,DEFS_POS(I+1),TARGET_SF,DEFS(J))
            J = (I-1)+2
            CALL CTF_DF_FROM_ARG(SELF%CS,SELF%WL,DEFS_NEG(I),TARGET_SF,DEFS(J))
        ENDDO
        CALL CTF_DF_FROM_ARG(SELF%CS,SELF%WL,DEFS_POS(NUMBER_OF_DEFS),TARGET_SF,DEFS(NUMBER_OF_DEFS))


    END SUBROUTINE CTF_FIND_BEST_DEFS_FOR_SF

    !>  \brief  Find the first few frequencies where the CTF = 0.0
    !!
    !!  We find the zeroes halfway between the two axes of the elipse
    !!
    SUBROUTINE CTF_ZEROES(SELF,NUMBER_OF_ZEROES,ZEROES)
        IMPLICIT NONE
        ! Arguments
        TYPE(CTF_PARAMETERS),   INTENT(IN)      ::  SELF                !<  CTF object
        INTEGER,                INTENT(IN)      ::  NUMBER_OF_ZEROES    !<  This many zeroes will be returned
        ! Result
        REAL,   ALLOCATABLE,    INTENT(INOUT)   ::  ZEROES(:)           !<  The position of the zeroes (in 1/pixels; 0.5 is Nyquist)
        ! Private variables
        INTEGER ::  I
        ! Start work

        ! Allocate memory for results
        IF (ALLOCATED(ZEROES)) DEALLOCATE(ZEROES)
        ALLOCATE(ZEROES(NUMBER_OF_ZEROES))

        ! Find the position of the zeroes in terms of the argument to the cosine and sine in the CTF expression
        CALL CTF_SOLVE_FOR_ARG(SELF%AMPL_CONT,ZEROES)

        ! Now convert the arguments to spatial frequencies
        DO I=1,SIZE(ZEROES)
            CALL CTF_SF_FROM_ARG(SELF%CS,SELF%WL,SELF%DFMID1,SELF%DFMID2,SELF%ANGAST,ZEROES(I),SELF%ANGAST+45.0,ZEROES(I))
        ENDDO
    END SUBROUTINE CTF_ZEROES

    !>  \brief Wrapper around EVAL_CTF which takes a CTF object as argument
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    REAL FUNCTION EVAL_CTF(CTF1,BOX_SIZE,RAD,ANG,SIGN_ONLY)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF1        !< CTF object containing all relevant CTF parameters
        INTEGER,                INTENT(IN)  ::  BOX_SIZE    !< Box dimension (pixels)
        REAL,                   INTENT(IN)  ::  RAD         !< Radius at which to compute the CTF (pixels)
        REAL,                   INTENT(IN)  ::  ANG         !< Angle at which to compute the CTF (radians)
        LOGICAL,    OPTIONAL,   INTENT(IN)  ::  SIGN_ONLY   !< Return the sign of the CTF rather than its value
#ifndef SKIP_RUNTIME_CHECKS
        ! Check the CTF parameters are in pixels
        IF (.NOT. CTF1%PIX_UNITS) THEN
            WRITE(*,'(A)') '**ERROR(EVAL_CTF_WRAP): CTF parameters not supplied in appropriate units'
            CALL TERMINATE('Fatal error in EVAL_CTF_WRAP')
        ENDIF
#endif
        ! Call EVAL_CTF
        EVAL_CTF = EVAL_CTF_SLAVE(CTF1%CS,CTF1%WL,CTF1%AMPL_CONT,CTF1%DFMID1,CTF1%DFMID2,CTF1%ANGAST,BOX_SIZE,RAD,ANG,SIGN_ONLY)
    END FUNCTION EVAL_CTF

    !>  \brief Returns the CTF, based on CTFFIND3 subroutine (see Mindell 2003)
    PURE REAL FUNCTION EVAL_CTF_SLAVE(CS,WL,AMPL_CONT,DFMID1,DFMID2,ANGAST,BOX_SIZE,RAD,ANG,SIGN_ONLY)
        USE RUNTIME_PARAMETERS, ONLY: PI ! we only need PI
        IMPLICIT NONE
        ! Arguments
        REAL,                   INTENT(IN)  ::  CS          !< Spherical aberation (pixels)
        REAL,                   INTENT(IN)  ::  WL          !< Electron wavelength (pixels)
        REAL,                   INTENT(IN)  ::  AMPL_CONT   !< Fraction of amplitude contrast (from .07 to .14, see Mindell 2003)
        REAL,                   INTENT(IN)  ::  DFMID1      !< Defocus along first axis (pixels)
        REAL,                   INTENT(IN)  ::  DFMID2      !< Defocus along second axis (for astigmatic CTF, DFMID1 .NE. DFMID2) (pixels)
        REAL,                   INTENT(IN)  ::  ANGAST      !< Azimuth of first axis. 0.0 means axis is at 3 o'clock. (radians)
        INTEGER,                INTENT(IN)  ::  BOX_SIZE    !< Box dimension (pixels)
        REAL,                   INTENT(IN)  ::  RAD         !< Radius at which to compute the CTF (pixels)
        REAL,                   INTENT(IN)  ::  ANG         !< Angle at which to compute the CTF (radians)
        LOGICAL,    OPTIONAL,   INTENT(IN)  ::  SIGN_ONLY   !<  Return 1.0 or -1.0
        ! Private variables
        REAL    ::  WGH1    ! Phase contrast
        REAL    ::  WGH2    ! Amplitude contrast
        REAL    ::  CTF_ARG
        ! Start work

        ! Compute phase and amplitude contrast
        WGH1    =   SQRT(1-AMPL_CONT**2)
        WGH2    =   AMPL_CONT
        IF (RAD .EQ. 0.0) THEN
            EVAL_CTF_SLAVE = -WGH2
        ELSE
            CTF_ARG = EVAL_CTF_ARG(CS,WL,AMPL_CONT,DFMID1,DFMID2,ANGAST,BOX_SIZE,RAD,ANG)
            EVAL_CTF_SLAVE = - WGH1 * SIN(CTF_ARG) - WGH2 * COS(CTF_ARG)
        ENDIF
        IF (PRESENT(SIGN_ONLY)) THEN
            IF (SIGN_ONLY) EVAL_CTF_SLAVE = SIGN(1.0,EVAL_CTF_SLAVE)
        ENDIF
    END FUNCTION EVAL_CTF_SLAVE

    !>  \brief Returns the argument (radians) to the sine and cosine terms of the CTF
    PURE REAL FUNCTION EVAL_CTF_ARG(CS,WL,AMPL_CONT,DFMID1,DFMID2,ANGAST,BOX_SIZE,RAD,ANG) RESULT(CTF_ARG)
        USE RUNTIME_PARAMETERS, ONLY: PI ! we only need PI
        IMPLICIT NONE
        ! Arguments
        REAL,                   INTENT(IN)  ::  CS          !< Spherical aberation (pixels)
        REAL,                   INTENT(IN)  ::  WL          !< Electron wavelength (pixels)
        REAL,                   INTENT(IN)  ::  AMPL_CONT   !< Fraction of amplitude contrast (from .07 to .14, see Mindell 2003)
        REAL,                   INTENT(IN)  ::  DFMID1      !< Defocus along first axis (pixels)
        REAL,                   INTENT(IN)  ::  DFMID2      !< Defocus along second axis (for astigmatic CTF, DFMID1 .NE. DFMID2) (pixels)
        REAL,                   INTENT(IN)  ::  ANGAST      !< Azimuth of first axis. 0.0 means axis is at 3 o'clock. (radians)
        INTEGER,                INTENT(IN)  ::  BOX_SIZE    !< Box dimension (pixels)
        REAL,                   INTENT(IN)  ::  RAD         !< Radius at which to compute the CTF (pixels)
        REAL,                   INTENT(IN)  ::  ANG         !< Angle at which to compute the CTF (radians)
        ! Private variables
        REAL    ::  SF      ! Spatial frequency (pix**-1)
        REAL    ::  DF      ! Defocus at point at which we're evaluating the CTF
        REAL    ::  CCOS
        ! Start work

        IF (RAD .EQ. 0.0) THEN
            CTF_ARG = 0.0
        ELSE
            ! Convert the radius to spatial frequency (pix**-1)
            SF = RAD / REAL(BOX_SIZE)
            ! Compute the defocus
            CCOS = COS(2.0*(ANG-ANGAST))
            DF = 0.5*(DFMID1+DFMID2+CCOS*(DFMID1-DFMID2))
            ! Compute the CTF argument
            CTF_ARG = PI * WL * SF**2 * (DF - 0.5 * WL**2 * SF**2 * CS)
        ENDIF
    END FUNCTION EVAL_CTF_ARG

    !>  \brief  Given the argument to the CTF, return the spatial frequency
    !!
    !!  According to Maxima, there are two positive solutions
    PURE SUBROUTINE CTF_SF_FROM_ARG(CS,WL,DFMID1,DFMID2,ANGAST,ARG,ANG,SF)
        USE RUNTIME_PARAMETERS, ONLY: PI ! we only need PI
        IMPLICIT NONE
        ! Arguments
        REAL,                   INTENT(IN)  ::  CS          !< Spherical aberation (pixels)
        REAL,                   INTENT(IN)  ::  WL          !< Electron wavelength (pixels)
        REAL,                   INTENT(IN)  ::  DFMID1      !< Defocus along first axis (pixels)
        REAL,                   INTENT(IN)  ::  DFMID2      !< Defocus along second axis (for astigmatic CTF, DFMID1 .NE. DFMID2) (pixels)
        REAL,                   INTENT(IN)  ::  ANGAST      !< Azimuth of first axis. 0.0 means axis is at 3 o'clock. (radians)
        REAL,                   INTENT(IN)  ::  ARG         !< Radius at which to compute the CTF (pixels)
        REAL,                   INTENT(IN)  ::  ANG         !< Angle at which to compute the CTF (radians)
        REAL,                   INTENT(OUT) ::  SF          !< Spatial frequency
        ! Private variables
        REAL    ::  DF      ! Defocus
        REAL    ::  CCOS
        ! Start work

        ! Compute the defocus
        CCOS = COS(2.0*(ANG-ANGAST))
        DF = 0.5*(DFMID1+DFMID2+CCOS*(DFMID1-DFMID2))
        !
        SF = SQRT(PI)*DF*WL - SQRT(WL**2*(PI*DF**2 - 2*WL*ARG*CS) )
        SF = SF / (SQRT(PI)*WL**3*CS)
        SF = SQRT(SF)
    END SUBROUTINE CTF_SF_FROM_ARG

    !>  \brief  Given the argument to the CTF, return the defocus
    !!
    PURE SUBROUTINE CTF_DF_FROM_ARG(CS,WL,ARG,SF,DF)
        USE RUNTIME_PARAMETERS, ONLY: PI ! we only need PI
        IMPLICIT NONE
        ! Arguments
        REAL,                   INTENT(IN)  ::  CS          !< Spherical aberation (pixels)
        REAL,                   INTENT(IN)  ::  WL          !< Electron wavelength (pixels)
        REAL,                   INTENT(IN)  ::  ARG         !< CTF argument (radians)
        REAL,                   INTENT(IN)  ::  SF          !< Spatial frequency (1/pixels)
        REAL,                   INTENT(OUT) ::  DF          !< Defocus (pixels)
        ! Private variables

        ! Start work

        ! Compute the defocus
        DF = (0.5 * CS * PI * SF**4 * WL**3 + ARG) / (PI * SF**2 * WL)
    END SUBROUTINE CTF_DF_FROM_ARG

    !>  \brief  Find the CTF argument values at which CTF=0
    !!
    !!  According to Wolfram Alpha, the solutions to a*cos(t) + b*sin(t) = c are:
    !!  1)  t = 2 (atan((b-sqrt(a^2+b^2-c^2))/(a+c))+pi*n),   with n an integer
    !!  2)  t = 2 (atan((b+sqrt(a^2+b^2-c^2))/(a+c))+pi*n),   with n an integer
    !!
    !!  In our case, a = - AMPL_CONST
    !!           and b = - SQRT(1-AMPL_CONT**2)
    !!  and t is the "argument" to the ctf, which can be "converted" to a spatial frequency
    SUBROUTINE CTF_SOLVE_FOR_ARG(AMPL_CONT,SOLS,CTF_VALUE)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE RUNTIME_PARAMETERS, ONLY : PI
        IMPLICIT NONE
        ! Arguments
        REAL,                   INTENT(IN)      ::  AMPL_CONT
        REAL,   ALLOCATABLE,    INTENT(INOUT)   ::  SOLS(:)     !<  Solutions
        REAL,   OPTIONAL,       INTENT(IN)      ::  CTF_VALUE   !<  Value of the CTF at the solutions
        ! Private variables
        INTEGER ::  I,J
        REAL    ::  A,B
        REAL    ::  C
        ! Start work
        IF (.NOT. ALLOCATED(SOLS)) THEN
            WRITE(*,'(A)') '**ERROR(CTF_SOLVE_FOR_ARG): array of solutions is not allocated'
            CALL TERMINATE('Fatal error in CTF_SOLVE_FOR_ARG')
        ENDIF
        IF (SIZE(SOLS) .LT. 1) THEN
            WRITE(*,'(A,I0)') '**ERROR(CTF_SOLVE_FOR_ARG): array of solutions has unexpected size ', SIZE(SOLS)
            CALL TERMINATE('Fatal error in CTF_SOLVE_FOR_ARG')
        ENDIF
        C = 0
        IF (PRESENT(CTF_VALUE)) C = CTF_VALUE
        !
        A = -AMPL_CONT
        B = -SQRT(1-AMPL_CONT**2)
        ! Loop over the zeroes
        DO I=1,SIZE(SOLS)/2
            J = (I-1)*2+1
            SOLS(J) = 2.0*(ATAN((B-SQRT(A**2+B**2-C**2))/(A+C))+PI*REAL(I-1))
            J = (I-1)*2+2
            SOLS(J) = 2.0*(ATAN((B+SQRT(A**2+B**2-C**2))/(A+C))+PI*REAL(I))
        ENDDO
    END SUBROUTINE CTF_SOLVE_FOR_ARG

    !>  \brief  Convert acceleration voltage in kV into electron wavelength in Angstroms
    PURE REAL FUNCTION AKV_TO_WL(AKV)
        IMPLICIT NONE
        REAL, INTENT(IN) :: AKV
        AKV_TO_WL = 12.26/SQRT(1000.0*AKV+0.9784*(1000.0*AKV)**2/(10.0**6.0))
    END FUNCTION AKV_TO_WL

    !>  \brief  Convert electron wavelength in Angstroms into acceleration voltage in kV
    !!  \todo   Check carefully; not 100% sure results are correct
    PURE REAL FUNCTION WL_TO_AKV(WL)
        IMPLICIT NONE
        ! Argument
        REAL, INTENT(IN) :: WL
        ! Private variables
        REAL    ::  TMP
        ! Start work
        TMP = 0.9784 * 10.0**(-6.0)
        WL_TO_AKV = (-1. + SQRT(1.+4.*TMP*12.26**2/WL**2))/(2.*TMP)
        WL_TO_AKV = WL_TO_AKV / 1000.
    END FUNCTION WL_TO_AKV

END MODULE CTFS
