!>  \brief  Classes and methods implementing differential evolution
!!
!!  20-Feb-2012:    it turns out Hans Elmlund pointed me to DE a few months ago, but I only realised this in the last few days,
!!                  when I decided DE might be a good solution to my optimisation problem. I intend eventually to implement
!!                  Hans' improvements on the methods, but for now will start with a basic version based on the original paper
!!                  (﻿Storn, R., & Price, K. (1996). Minimizing the real functions of the ICEC’96 contest by Differential Evolution.
!!                  IN: IEEE CONFERENCE ON EVOLUTIONARY COMPUTATION, 842 - 844)
!!
!!  Another useful resource, which includes rules-of-thumbs for setting DE parameters is:
!!                  ﻿Storn, R. (1996). On the Usage of Differential Evolution for Function Optimization.
!!                  IN NAFIPS’96, 519 - 523. Retrieved from http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.71.2432
MODULE DIFFERENTIAL_EVOLUTION

    USE FGSL
    USE, INTRINSIC :: ISO_C_BINDING
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE

    PRIVATE

    PUBLIC :: DE_MINIMISER, DE_INIT, DE_RUN, DE_BEST_VALUES

    TYPE DE_INDIVIDUAL
        REAL(KIND=8),   ALLOCATABLE ::  VAL(:)
        REAL                        ::  F       =   0.7             !<  Perturbation scaling
        REAL                        ::  CR      =   0.3             !<  Crossover probability
        REAL                        ::  SCORE   =   0.0
    END TYPE

    TYPE DE_POPULATION
        TYPE(DE_INDIVIDUAL),    ALLOCATABLE ::  MEMBERS(:)          !<  Number of members in the population is noted NP in Storn & Price (1996)
    END TYPE

    !>  \brief   Definition of a minimisable function.
    ABSTRACT INTERFACE
        FUNCTION SCORING_FUNCTION(VALUES,PARAMS) BIND(C)
            IMPORT :: C_DOUBLE, C_PTR   ! Because we're in an abstract interface, we normally don't have access to all the stuff from the module, so we need to IMPORT
            TYPE(C_PTR), VALUE  ::  VALUES
            TYPE(C_PTR), VALUE  ::  PARAMS
            REAL(KIND=C_DOUBLE) ::  SCORING_FUNCTION
        END FUNCTION
    END INTERFACE

    TYPE DE_MINIMISER
        TYPE(DE_POPULATION)     ::  POPULATION
        TYPE(FGSL_RNG_TYPE)     ::  RNG_TYPE
        TYPE(FGSL_RNG)          ::  RNG
        !
        TYPE(DE_INDIVIDUAL)     ::  BEST_MEMBER
        !
        CHARACTER(LEN=50)       ::  ALGORITHM               =   'DE_BEST_2'     !<  See Storn & Price papers for algorithms
        LOGICAL                 ::  FIRST_GEN_IS_GAUSSIAN   =   .FALSE.         !<  Whether the first generation should be Gaussian-, rather than uniformly-, distributed around the starting point
        LOGICAL                 ::  EVOLVE_F                =   .FALSE.         !<  If true, F (perturbation scaling) is per-individual and is one of the parameters under evolutionary pressure
        LOGICAL                 ::  EVOLVE_CR               =   .FALSE.         !<  If true, CR (crossover probability) is per-individual and is one of the parameters under evolutionary pressur
        REAL                    ::  F_MUTATION_PROBABILITY  =   0.1             !<  Probabiltiy that F  will be mutated (only applies if EVOLVE_F)
        REAL                    ::  CR_MUTATION_PROBABILITY =   0.1             !<  Probability that CR will be mutated (only applies if EVOLVE_CR)
        REAL                    ::  PERTURBATION_SCALING    =   0.7             !<  Scaling factor for the perturbation/mutation. F in Storn & Price (1996). If EVOLVE_F, this is only the starting value.
        REAL                    ::  CROSSOVER_PROBABILITY   =   0.3             !<  Probability of mutation. Must in range [0.,1.]. CR in Storn & Price (1996). If EVOLVE_CR, this is only the starting value.
        LOGICAL,    ALLOCATABLE ::  INTEGER_CHANGES_ONLY(:)                     !<  For each dimension, restrict changes from the starting point to integer increments
        !
        INTEGER                 ::  CURRENT_GENERATION      =   0               !<  Current generation
        INTEGER                 ::  MAX_NUM_GENERATIONS     =   100             !<  Maximum number of generations
        INTEGER                 ::  NUM_FUNCTION_CALLS      =   0               !<  Keep track of the total number of function calls
        REAL                    ::  AVERAGE_SCORE           =   0.0             !<  Keep track of the average score of population members
        !
        TYPE(C_PTR)             ::  PARAMS_FOR_SCORING      !<  A pointer to an object containing all the information needed by the scoring function (other than the function's arguments)
        PROCEDURE(SCORING_FUNCTION), POINTER, NOPASS :: FUNCTION => null()  !<  A pointer to the function the minimiser aims to minimise.
                                                                            !!  NO PASS because there an assumption by default that type-bound procedures have an object of the type to which they are bound as first argument.
                                                                            !!  This default assumption is overriden by NO PASS.
        CONTAINS
            PROCEDURE   ::  NUM_DIM         =>  DE_NUM_DIM
            PROCEDURE   ::  POP_SIZE        =>  DE_POPULATION_SIZE
            PROCEDURE   ::  POPULATION_SIZE =>  DE_POPULATION_SIZE
    END TYPE


CONTAINS

        !>  \brief  Compute the score for a particular individual
        SUBROUTINE DE_SCORE_INDIVIDUAL(SELF,IND)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),             INTENT(INOUT)   ::  SELF
            TYPE(DE_INDIVIDUAL),    TARGET, INTENT(INOUT)   ::  IND
            ! Private variables
            TYPE(C_PTR) ::  VALS_C
            LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
            ! Start work

            ! Make C pointer to the individual's values
            VALS_C = C_LOC(IND%VAL)

            ! Call the scoring function
            IND%SCORE = SELF%FUNCTION(VALS_C,SELF%PARAMS_FOR_SCORING)
            SELF%NUM_FUNCTION_CALLS = SELF%NUM_FUNCTION_CALLS + 1

            IF (DEBUG) THEN
                !$OMP   CRITICAL (OMP_CRIT_DE_SCORE_IND)
                WRITE(*,'(A)',ADVANCE='NO') '**debug(DE_SCORE_INDIVIDUAL): individual information: '
                CALL DE_INDIVIDUAL_PRINT_INFO(IND)
                !$OMP   END CRITICAL (OMP_CRIT_DE_SCORE_IND)
            ENDIF
        END SUBROUTINE DE_SCORE_INDIVIDUAL

        !>  \brief  Iterate to next generation
        SUBROUTINE DE_NEW_GENERATION(SELF)
            USE OMP_LIB
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER), INTENT(INOUT)   ::  SELF
            ! Private variables
            INTEGER ::  CURRENT_MEMBER
            LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
            REAL(KIND=8)    ::  NEW_AVERAGE_SCORE
            TYPE(DE_INDIVIDUAL) ::  PERTURBED_MEMBER
            ! Start work

            ! Check the caller didn't forget to call DE_FIRST_GENERATION
            IF (SELF%CURRENT_GENERATION .EQ. 0) THEN
                WRITE(*,'(A)') '**ERROR(DE_NEW_GENERATION): first generation has not been generated yet.'
                CALL TERMINATE('Fatal error in DE_NEW_GENERATION')
            ENDIF

            !
            WRITE(*,'(A,I0)') '**info(DE_NEW_GENERATION): working on generation ', SELF%CURRENT_GENERATION+1



            !$OMP   PARALLEL        &
            !$OMP   DEFAULT (NONE)  &
            !$OMP   SHARED(NEW_AVERAGE_SCORE,SELF)        &
            !$OMP   PRIVATE(CURRENT_MEMBER,PERTURBED_MEMBER)

            ! We will compute the new average score
            NEW_AVERAGE_SCORE = 0.0D0

            ! Loop through the population members
            !$OMP   DO  REDUCTION(+:NEW_AVERAGE_SCORE)
            DO CURRENT_MEMBER=1,SIZE(SELF%POPULATION%MEMBERS)
                !IF (DEBUG .AND. OMP_GET_NUM_THREADS() .GT. 1) THEN
                !    WRITE(*,'(A,I0,A,I0,A,I0)') '**debug(DE_NEW_GENERATION): thread ', OMP_GET_THREAD_NUM()+1, ' of ', &
                !                                OMP_GET_NUM_THREADS(), ' working on member ', CURRENT_MEMBER
                !ENDIF
                ! Come up with a new challenger individual
                CALL DE_MAKE_NEW_PERTURBED_MEMBER(SELF,CURRENT_MEMBER,PERTURBED_MEMBER)
                ! Check whether the challenger beats the current individual
                CALL DE_SCORE_INDIVIDUAL(SELF,PERTURBED_MEMBER)
                ! If the challenger does score better, it will replace the current individual
                !$OMP   CRITICAL (DE_NEW_GEN)
                IF (PERTURBED_MEMBER%SCORE .LT. SELF%POPULATION%MEMBERS(CURRENT_MEMBER)%SCORE) THEN
                    SELF%POPULATION%MEMBERS(CURRENT_MEMBER) = PERTURBED_MEMBER
                    IF (DEBUG) THEN
                        WRITE(*,'(A,I0,A)', ADVANCE='NO') '**info(DE_NEW_GENERATION): member ', CURRENT_MEMBER, &
                                                            ' replaced by new individual: '
                        CALL DE_INDIVIDUAL_PRINT_INFO(PERTURBED_MEMBER)
                    ENDIF
                    ! Is the challenger so good that it beats the previous best member?
                    IF (PERTURBED_MEMBER%SCORE .LT. SELF%BEST_MEMBER%SCORE) THEN
                        SELF%BEST_MEMBER = PERTURBED_MEMBER
                    ENDIF
                ENDIF
                !$OMP   END CRITICAL (DE_NEW_GEN)
                NEW_AVERAGE_SCORE = NEW_AVERAGE_SCORE + SELF%POPULATION%MEMBERS(CURRENT_MEMBER)%SCORE
            ENDDO
            !$OMP   ENDDO

            !$OMP   END PARALLEL

            ! The average score
            NEW_AVERAGE_SCORE = NEW_AVERAGE_SCORE / REAL(SIZE(SELF%POPULATION%MEMBERS))
            SELF%AVERAGE_SCORE = NEW_AVERAGE_SCORE

            ! Increment the generation counter
            SELF%CURRENT_GENERATION = SELF%CURRENT_GENERATION + 1

            ! Output information
            WRITE(*,'(A,I0,A,F0.4)') '**info(DE_NEW_GENERATION): average score for generation ', SELF%CURRENT_GENERATION, ' = ', &
                                        SELF%AVERAGE_SCORE

        END SUBROUTINE DE_NEW_GENERATION

        !>  \brief  Generate the first generation
        SUBROUTINE DE_FIRST_GENERATION(SELF,STARTING_VALUES,SEARCH_HALF_RANGE)
            USE UI, ONLY : INTERVAL_PRINT
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER), INTENT(INOUT)   ::  SELF
            REAL(KIND=8),       INTENT(IN)      ::  STARTING_VALUES(:)
            REAL(KIND=8),       INTENT(IN)      ::  SEARCH_HALF_RANGE(:)
            ! Private variables
            INTEGER ::  CURRENT_MEMBER, CURRENT_DIM
            REAL    ::  RND
            ! Start work

            ! Check the generation number
            IF (SELF%CURRENT_GENERATION .NE. 0) THEN
                WRITE(*,'(A,I0)') '**ERROR(DE_FIRST_GENERATION): minimiser already appears to be on generation ', &
                                    SELF%CURRENT_GENERATION
                CALL TERMINATE('Fatal error in DE_FIRST_GENERATION')
            ENDIF

            SELF%AVERAGE_SCORE = 0.0

            !$OMP   PARALLEL    &
            !$OMP   DEFAULT(NONE)   &
            !$OMP   SHARED(SELF,STARTING_VALUES,SEARCH_HALF_RANGE)    &
            !$OMP   PRIVATE(CURRENT_MEMBER,CURRENT_DIM,RND)

            ! Loop over the members of the population
            !$OMP   DO
            DO CURRENT_MEMBER=1,SIZE(SELF%POPULATION%MEMBERS)
                IF (INTERVAL_PRINT(CURRENT_MEMBER,SIZE(SELF%POPULATION%MEMBERS))) THEN
                    WRITE(*,'(A,I0,A,I0)') '**info(DE_FIRST_GENERATION): working on individual ',   &
                                    CURRENT_MEMBER, ' of ', SIZE(SELF%POPULATION%MEMBERS)
                ENDIF
                ! For each dimension, we add a uniformly distributed random variable
                ! to the starting value
                DO CURRENT_DIM=1,SIZE(STARTING_VALUES)
                    IF (SELF%FIRST_GEN_IS_GAUSSIAN) THEN
                        ! We use half of the search half range as the std dev
                        ! so that ~95% of the values should be within the search
                        ! range requested by the user
                        RND = FGSL_RAN_GAUSSIAN(SELF%RNG,REAL(SEARCH_HALF_RANGE(CURRENT_DIM)/2.,KIND=FGSL_DOUBLE))
                    ELSE
                        ! Generate a random float variable between 0.0 and 1.0, excluding 1.0
                        RND = FGSL_RNG_UNIFORM(SELF%RNG)
                        ! The random values is between -half_range and +half_range
                        RND = RND*2.*SEARCH_HALF_RANGE(CURRENT_DIM) - SEARCH_HALF_RANGE(CURRENT_DIM)
                    ENDIF
                    IF (SELF%INTEGER_CHANGES_ONLY(CURRENT_DIM)) RND = ANINT(RND)
                    ! Add random value to starting value
                    SELF%POPULATION%MEMBERS(CURRENT_MEMBER)%VAL(CURRENT_DIM) = &
                            STARTING_VALUES(CURRENT_DIM) + RND
                ENDDO
                ! Score this member
                CALL DE_SCORE_INDIVIDUAL(SELF,SELF%POPULATION%MEMBERS(CURRENT_MEMBER))
                !$OMP CRITICAL (DE_FIRST_GEN)
                SELF%AVERAGE_SCORE = SELF%AVERAGE_SCORE + SELF%POPULATION%MEMBERS(CURRENT_MEMBER)%SCORE
                !$OMP END CRITICAL (DE_FIRST_GEN)
            ENDDO
            !$OMP   ENDDO


            !$OMP   END PARALLEL
            SELF%AVERAGE_SCORE = SELF%AVERAGE_SCORE / REAL(SIZE(SELF%POPULATION%MEMBERS))
            SELF%CURRENT_GENERATION = 1
        END SUBROUTINE DE_FIRST_GENERATION

        !>  \brief  Print out all members of the population (for debugging purposes)
        SUBROUTINE DE_POPULATION_PRINT(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),                 INTENT(IN)      ::  SELF
            ! Start work
            CALL TERMINATE('Not implemented yet')
        END SUBROUTINE DE_POPULATION_PRINT

        !>  \brief  Return the best values found by the minimiser
        SUBROUTINE DE_BEST_VALUES(SELF,BEST_VALUES)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),                 INTENT(IN)      ::  SELF
            REAL(KIND=8),       ALLOCATABLE,    INTENT(INOUT)   ::  BEST_VALUES(:)
            ! Private variables

            ! Start work
            BEST_VALUES = SELF%BEST_MEMBER%VAL
        END SUBROUTINE DE_BEST_VALUES

        !>  \brief  Run the DE minimiser
        SUBROUTINE DE_RUN(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),                 INTENT(INOUT)   ::  SELF
            ! Private variables
            LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
            ! Start work

            IF (DEBUG) WRITE(*,'(A)') '**debug(DE_RUN): entry'

            ! Loop over generations
            DO
                ! Generate the next generation
                CALL DE_NEW_GENERATION(SELF)

                ! Printout information about the best member
                WRITE(*,'(A,I0,A)',ADVANCE='NO') '**info(DE_RUN): the best member of generation ', SELF%CURRENT_GENERATION, ': '
                CALL DE_INDIVIDUAL_PRINT_INFO(SELF%BEST_MEMBER)

                ! Are we done
                IF (DE_MINIMISATION_DONE(SELF)) THEN
                    IF (DEBUG) WRITE(*,'(A)') '**debug(DE_RUN): minimisation done'
                    EXIT
                ENDIF
            ENDDO

            ! All done
            WRITE(*,'(A,I0)') '**info(DE_RUN): all done. Total number of function calls = ', SELF%NUM_FUNCTION_CALLS
        END SUBROUTINE DE_RUN


        !>  \brief  Print summary information about the DE minimiser
        SUBROUTINE DE_PRINT_INFO(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER), INTENT(IN)  ::  SELF
            ! Private variables

            ! Start work
            WRITE(*,'(/A)')         'Differential evolution minimiser, summary info:'
            WRITE(*,'(A,I0)')       'Population size                         = ', SELF%POP_SIZE()
            WRITE(*,'(A,I0)')       'Number of dimensions                    = ', SELF%NUM_DIM()
            WRITE(*,'(A,I0)')       'Current generation                      = ', SELF%CURRENT_GENERATION
            WRITE(*,'(A,I0)')       'Maximum # of generations                = ', SELF%MAX_NUM_GENERATIONS
            WRITE(*,'(A,I0)')       'Number of function calls                = ', SELF%NUM_FUNCTION_CALLS
            WRITE(*,'(A,L1)')       'First generation is gaussian-distributed? ', SELF%FIRST_GEN_IS_GAUSSIAN
            WRITE(*,'(A,2(L1,1X))') 'Evolve F? CR?                             ', SELF%EVOLVE_F, SELF%EVOLVE_CR
            WRITE(*,'(A)')      ' '
        END SUBROUTINE DE_PRINT_INFO

        !>  \brief  Initialise a differential evolution minimiser
        SUBROUTINE DE_INIT(SELF,NUM_DIM,FUNCTION_TO_MINIMISE,PARAMETERS,   &
                            STARTING_VALUES,SEARCH_HALF_RANGE,POP_SIZE,CROSSOVER_PROBABILITY,   &
                            EVOLVE_F,EVOLVE_CR, &
                            INTEGER_CHANGES_ONLY)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),                     INTENT(INOUT)   ::  SELF
            INTEGER,                                INTENT(IN)      ::  NUM_DIM                 !<  Number of dimensions to minimise
            PROCEDURE(SCORING_FUNCTION)                             ::  FUNCTION_TO_MINIMISE    !<  The function to minimise
            TYPE(C_PTR),                            INTENT(IN)      ::  PARAMETERS              !<  Parameters needed by the scoring function (a C pointer to an object containing...)
            REAL(KIND=8),           ALLOCATABLE,    INTENT(IN)      ::  STARTING_VALUES(:)      !<  The starting point for the minimiser
            REAL(KIND=8),           ALLOCATABLE,    INTENT(IN)      ::  SEARCH_HALF_RANGE(:)    !<  For each search dimension, the half-range which should be covered by the minimiser
            INTEGER,    OPTIONAL,                   INTENT(IN)      ::  POP_SIZE
            REAL,       OPTIONAL,                   INTENT(IN)      ::  CROSSOVER_PROBABILITY   !<  Must be in [0.,1.] range
            LOGICAL,    OPTIONAL,                   INTENT(IN)      ::  EVOLVE_F                !<  False by default. If true, F (perturbation scaling) is a per-individual parameter subject to evolution
            LOGICAL,    OPTIONAL,                   INTENT(IN)      ::  EVOLVE_CR               !<  False by default. If true, CR (crossover probability) is a per-individual parameter subject to evolution
            LOGICAL,    OPTIONAL,   ALLOCATABLE,    INTENT(IN)      ::  INTEGER_CHANGES_ONLY(:) !<  For each dimension, whether changes from the starting point should be restricted to integer increments
            ! Private variables
            INTEGER ::  I
            INTEGER ::  PPOP_SIZE
            ! Start work

            ! Population size
            PPOP_SIZE = DE_SUGGESTED_POPULATION_SIZE(SELF,NUM_DIM=NUM_DIM)
            IF (PRESENT(POP_SIZE)) PPOP_SIZE = POP_SIZE

            ! Allocate memory for the population
            ALLOCATE(SELF%POPULATION%MEMBERS(PPOP_SIZE))


            ! Allocate memory for each population member
            DO I=1,SIZE(SELF%POPULATION%MEMBERS)
                ALLOCATE(SELF%POPULATION%MEMBERS(I)%VAL(NUM_DIM))
            ENDDO

            ! F, CR under evolution?
            IF (PRESENT(EVOLVE_F)) SELF%EVOLVE_F = EVOLVE_F
            IF (PRESENT(EVOLVE_CR)) SELF%EVOLVE_CR = EVOLVE_CR


            ! Allocate memory for the perturbed member
            !ALLOCATE(SELF%PERTURBED_MEMBER%VAL(NUM_DIM))

            ! Initialise random number generator (RNG)
            SELF%RNG_TYPE = FGSL_RNG_ENV_SETUP()
            SELF%RNG = FGSL_RNG_ALLOC(SELF%RNG_TYPE)

            ! Check the range of the RNG
            IF (FGSL_RNG_MAX(SELF%RNG) .LT. SIZE(SELF%POPULATION%MEMBERS)) THEN
                WRITE(*,'(A)')  '**ERROR(DE_INIT): random number generator goes to ', FGSL_RNG_MAX(SELF%RNG), &
                                ' which is less than the population size: ', SIZE(SELF%POPULATION%MEMBERS)
                CALL TERMINATE('Fatal error in DE_INIT')
            ENDIF

            ! The function we'll be minimising
            SELF%FUNCTION => FUNCTION_TO_MINIMISE

            ! Parameters needed by the function
            SELF%PARAMS_FOR_SCORING = PARAMETERS

            ! The starting values are stored as the "best member"
            SELF%BEST_MEMBER%VAL = STARTING_VALUES

            ! The score for the starting values
            CALL DE_SCORE_INDIVIDUAL(SELF,SELF%BEST_MEMBER)
            WRITE(*,'(A,F0.5)') '**info(DE_INIT): starting score = ', SELF%BEST_MEMBER%SCORE

            ! Crossover probability
            IF (PRESENT(CROSSOVER_PROBABILITY)) SELF%CROSSOVER_PROBABILITY = CROSSOVER_PROBABILITY

            ! Give each individual its value of F & CR
            DO I=1,SIZE(SELF%POPULATION%MEMBERS)
                SELF%POPULATION%MEMBERS(I)%F = SELF%PERTURBATION_SCALING
                SELF%POPULATION%MEMBERS(I)%CR = SELF%CROSSOVER_PROBABILITY
            ENDDO

            !
            IF (ALLOCATED(SELF%INTEGER_CHANGES_ONLY)) DEALLOCATE(SELF%INTEGER_CHANGES_ONLY)
            ALLOCATE(SELF%INTEGER_CHANGES_ONLY(NUM_DIM))
            IF (PRESENT(INTEGER_CHANGES_ONLY)) THEN
                IF (SIZE(INTEGER_CHANGES_ONLY) .NE. NUM_DIM) THEN
                    WRITE(*,'(A,2(I0,1X))') '**ERROR(DE_INIT): unexpected number of integer_changes_only logicals: ', &
                                            SIZE(INTEGER_CHANGES_ONLY), NUM_DIM
                    CALL TERMINATE('Fatal error in DE_INIT')
                ENDIF
                SELF%INTEGER_CHANGES_ONLY = INTEGER_CHANGES_ONLY
            ELSE
                SELF%INTEGER_CHANGES_ONLY = .FALSE.
            ENDIF

            ! Set up the first generation
            IF (ALLOCATED(STARTING_VALUES) .AND. SIZE(STARTING_VALUES) .EQ. NUM_DIM .AND.   &
                ALLOCATED(SEARCH_HALF_RANGE) .AND. SIZE(SEARCH_HALF_RANGE) .EQ. NUM_DIM ) THEN
                WRITE(*,'(A,I0)') '**info(DE_INIT): setting up the first generation. Number of individuals = ', PPOP_SIZE
                CALL DE_FIRST_GENERATION(SELF,STARTING_VALUES,SEARCH_HALF_RANGE)
            ELSE
                WRITE(*,'(A)') '**ERROR(DE_INIT): starting values and/or search_half_range arrays not allocated or not correct size'
                WRITE(*,'(2(L1,1X,I0,1X))') ALLOCATED(STARTING_VALUES), SIZE(STARTING_VALUES),    &
                                            ALLOCATED(SEARCH_HALF_RANGE), SIZE(SEARCH_HALF_RANGE)
                CALL TERMINATE('Fatal error in DE_INIT')
            ENDIF

            !
            WRITE(*,'(A)') '**info(DE_INIT): differential evolution minimiser has been initialised. Summary:'
            CALL DE_PRINT_INFO(SELF)
        END SUBROUTINE DE_INIT

        !>  \brief  Free all memory assigned to the DE minimiser
        SUBROUTINE DE_FREE(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER), INTENT(INOUT)   ::  SELF
            ! Private variables
            INTEGER ::  I
            ! Start work

            ! Free the RNG
            CALL FGSL_RNG_FREE(SELF%RNG)

            ! Deallocate memory for member values
            DO I=1,SIZE(SELF%POPULATION%MEMBERS)
                IF (ALLOCATED(SELF%POPULATION%MEMBERS(I)%VAL)) DEALLOCATE(SELF%POPULATION%MEMBERS(I)%VAL)
            ENDDO

            ! Deallocate memory for perturbed member
            !IF (ALLOCATED(SELF%PERTURBED_MEMBER%VAL)) DEALLOCATE(SELF%PERTURBED_MEMBER%VAL)

            ! Deallocate memory for population members
            IF (ALLOCATED(SELF%POPULATION%MEMBERS)) DEALLOCATE(SELF%POPULATION%MEMBERS)

            ! Function
            NULLIFY(SELF%FUNCTION)
        END SUBROUTINE DE_FREE

        !>  \brief  Generate a new perturbed/mutated population member
        SUBROUTINE DE_MAKE_NEW_PERTURBED_MEMBER(SELF,MEMBER_TO_CHALLENGE,PERTURBED_MEMBER)
            USE PARSERS, ONLY : STRING_EQUAL
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),     INTENT(INOUT)   ::  SELF
            INTEGER,                INTENT(IN)      ::  MEMBER_TO_CHALLENGE !<  The index value for the population memeber we will be challenging with the perturbed/mutated member
            TYPE(DE_INDIVIDUAL),    INTENT(INOUT)   ::  PERTURBED_MEMBER    !<  A mutated individual
            ! Private variables
            INTEGER,    ALLOCATABLE ::  R(:)    !<  Random indices
            INTEGER ::  CURRENT_DIM
            REAL    ::  F
            ! Start work

            PERTURBED_MEMBER%SCORE = 0.0
            IF (ALLOCATED(PERTURBED_MEMBER%VAL)) THEN
                IF (SIZE(PERTURBED_MEMBER%VAL) .NE. SELF%NUM_DIM()) THEN
                    DEALLOCATE(PERTURBED_MEMBER%VAL)
                ENDIF
            ENDIF
            IF (.NOT. ALLOCATED(PERTURBED_MEMBER%VAL)) ALLOCATE(PERTURBED_MEMBER%VAL(SELF%NUM_DIM()))

            ! Set the new CR and F values for the perturbed member
            PERTURBED_MEMBER%F  =  SELF%POPULATION%MEMBERS(MEMBER_TO_CHALLENGE)%F
            PERTURBED_MEMBER%CR =  SELF%POPULATION%MEMBERS(MEMBER_TO_CHALLENGE)%CR
            IF (SELF%EVOLVE_F)  CALL INDIVIDUAL_MUTATE_F(SELF,PERTURBED_MEMBER)
            IF (SELF%EVOLVE_CR) CALL INDIVIDUAL_MUTATE_CR(SELF,PERTURBED_MEMBER)

            IF (STRING_EQUAL(SELF%ALGORITHM,'DE_RAND_1')) THEN
                !
                ! Algorithm DE/RAND/1
                ! (Storn & Price, 1996)
                !
                CALL DE_GET_MULITPLE_RANDOM_MEMBER_INDICES(SELF,NUM_TO_GET=3,RANDOM_INDICES=R,INDEX_TO_AVOID=MEMBER_TO_CHALLENGE)
                DO CURRENT_DIM=1,SIZE(SELF%BEST_MEMBER%VAL)
                    F = PERTURBED_MEMBER%F
                    IF (SELF%INTEGER_CHANGES_ONLY(CURRENT_DIM)) F = ANINT(F)
                    PERTURBED_MEMBER%VAL(CURRENT_DIM) =   &
                                                    SELF%POPULATION%MEMBERS(R(1))%VAL(CURRENT_DIM) +           &
                                                    F *                                                   &
                                                (   SELF%POPULATION%MEMBERS(R(2))%VAL(CURRENT_DIM) -      &
                                                    SELF%POPULATION%MEMBERS(R(3))%VAL(CURRENT_DIM) )
                ENDDO
            ELSE IF (STRING_EQUAL(SELF%ALGORITHM,'DE_BEST_1')) THEN
                !
                ! Algorithm DE/BEST/1
                ! (Storn & Price, 1996)
                !
                CALL DE_GET_MULITPLE_RANDOM_MEMBER_INDICES(SELF,NUM_TO_GET=2,RANDOM_INDICES=R,INDEX_TO_AVOID=MEMBER_TO_CHALLENGE)
                DO CURRENT_DIM=1,SIZE(SELF%BEST_MEMBER%VAL)
                    F = PERTURBED_MEMBER%F
                    IF (SELF%INTEGER_CHANGES_ONLY(CURRENT_DIM)) F = ANINT(F)
                    PERTURBED_MEMBER%VAL(CURRENT_DIM) =   SELF%BEST_MEMBER%VAL(CURRENT_DIM) +   &
                                                  F *           &
                                                ( SELF%POPULATION%MEMBERS(R(1))%VAL(CURRENT_DIM) -   &
                                                  SELF%POPULATION%MEMBERS(R(2))%VAL(CURRENT_DIM) )
                ENDDO
            ELSE IF (STRING_EQUAL(SELF%ALGORITHM,'DE_BEST_2')) THEN
                !
                ! Algorithm DE/BEST/2
                ! (Storn & Price, 1996)
                !
                CALL DE_GET_MULITPLE_RANDOM_MEMBER_INDICES(SELF,NUM_TO_GET=4,RANDOM_INDICES=R,INDEX_TO_AVOID=MEMBER_TO_CHALLENGE)
                DO CURRENT_DIM=1,SIZE(SELF%BEST_MEMBER%VAL)
                    F = PERTURBED_MEMBER%F
                    IF (SELF%INTEGER_CHANGES_ONLY(CURRENT_DIM)) F = ANINT(F)
                    PERTURBED_MEMBER%VAL(CURRENT_DIM) =   SELF%BEST_MEMBER%VAL(CURRENT_DIM) +   &
                                                  F *           &
                                                ( SELF%POPULATION%MEMBERS(R(1))%VAL(CURRENT_DIM) +   &
                                                  SELF%POPULATION%MEMBERS(R(2))%VAL(CURRENT_DIM) -   &
                                                  SELF%POPULATION%MEMBERS(R(3))%VAL(CURRENT_DIM) -   &
                                                  SELF%POPULATION%MEMBERS(R(4))%VAL(CURRENT_DIM) )
              ENDDO
            ELSE
                WRITE(*,'(A)') '**ERROR(DE_MAKE_NEW_PERTURBED_MEMBER): name of algorithm not recognised: ', &
                                TRIM(ADJUSTL(SELF%ALGORITHM))
                CALL TERMINATE('Fatal error in DE_MAKE_NEW_PERTURBED_MEMBER')
            ENDIF

            ! Do the crossover, so that only some of the perturbed value are kept (those in the crossover region)
            ! Note that here we think of the perturbed member as being modified by the individual we are challenging,
            ! whereas in Storn & Price (1996; Fig 2), the crossover is described in the other direction.
            CALL DE_INDIVIDUALS_CROSSOVER(SELF,PERTURBED_MEMBER,SELF%POPULATION%MEMBERS(MEMBER_TO_CHALLENGE),CR=PERTURBED_MEMBER%CR)

            ! All done
        END SUBROUTINE DE_MAKE_NEW_PERTURBED_MEMBER

        !>  \brief  Mutate the F value (perturbation scaling) for an individual, as in Brest et al (2006)
        SUBROUTINE INDIVIDUAL_MUTATE_F(SELF,IND)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),     INTENT(IN)      ::  SELF
            TYPE(DE_INDIVIDUAL),    INTENT(INOUT)   ::  IND
            ! Private variables
            REAL        ::  RND
            ! Start work
            RND = FGSL_RNG_UNIFORM(SELF%RNG)    ! Random variable, uniform distribution in [0.,1.) range
            IF (RND .LT. SELF%F_MUTATION_PROBABILITY) THEN
                RND = FGSL_RNG_UNIFORM(SELF%RNG)
                IND%F = 0.1 + RND * 0.9    ! The new value is in [0.1,1.0] range
            ENDIF
        END SUBROUTINE INDIVIDUAL_MUTATE_F

        !>  \brief  Mutate the CR value (crossover probability) for an individual, as in Brest et al (2006)
        SUBROUTINE INDIVIDUAL_MUTATE_CR(SELF,IND)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),     INTENT(IN)      ::  SELF
            TYPE(DE_INDIVIDUAL),    INTENT(INOUT)   ::  IND
            ! Private variables
            REAL        ::  RND
            ! Start work
            RND = FGSL_RNG_UNIFORM(SELF%RNG)    ! Random variable, uniform distribution in [0.,1.) range
            IF (RND .LT. SELF%CR_MUTATION_PROBABILITY) THEN
                RND = FGSL_RNG_UNIFORM(SELF%RNG)
                IND%CR = RND
            ENDIF
        END SUBROUTINE INDIVIDUAL_MUTATE_CR

        !>  \brief  Apply crossover between two individuals
        SUBROUTINE DE_INDIVIDUALS_CROSSOVER(SELF,MEMBER1,MEMBER2,CR)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),     INTENT(IN)      ::  SELF
            TYPE(DE_INDIVIDUAL),    INTENT(INOUT)   ::  MEMBER1     !<  This individual will be modified by the crossover
            TYPE(DE_INDIVIDUAL),    INTENT(IN)      ::  MEMBER2     !<  Values from this individual will be crossed into the other individual
            REAL,                   INTENT(IN)      ::  CR          !<  Crossover probability
            ! Private variables
            INTEGER ::  CROSSOVER_START
            INTEGER ::  CROSSOVER_LENGTH
            INTEGER ::  CURRENT_DIMENSION
            INTEGER ::  NUM_DIM
            INTEGER ::  I
            REAL    ::  RND
            ! Start work

            NUM_DIM = SIZE(MEMBER1%VAL)

            ! Random crossover characteristics
            CROSSOVER_START = DE_GET_RANDOM_DIMENSION_INDEX(SELF)
            CROSSOVER_LENGTH = 0
            DO
                RND = FGSL_RNG_UNIFORM(SELF%RNG)    ! Random variable, uniform distribution in [0.,1.) range
                ! The greater the crossover_probability, the more likely the crossover is to end. This is because
                ! we define the crossover as being in the opposite direction as in Storn & Price.
                ! If the crossover_probability is high, then the crossover from member 2 (the population member)
                ! to member 1 (the perturbed member) will tend to be shorter. In other words, if crossover_probability
                ! is high, the perturbed member is more different from the population member.
                IF (CROSSOVER_LENGTH .GE. NUM_DIM .OR. RND .GE. (1.0-CR)) EXIT
                CROSSOVER_LENGTH = CROSSOVER_LENGTH + 1
            ENDDO

            ! Do the crossover
            CURRENT_DIMENSION = CROSSOVER_START
            DO I=CROSSOVER_START,CROSSOVER_START+CROSSOVER_LENGTH
                CURRENT_DIMENSION = MOD(I,NUM_DIM)
                IF (CURRENT_DIMENSION .EQ. 0) CURRENT_DIMENSION = NUM_DIM
                MEMBER1%VAL(CURRENT_DIMENSION) = MEMBER2%VAL(CURRENT_DIMENSION)
            ENDDO

        END SUBROUTINE DE_INDIVIDUALS_CROSSOVER

        !>  \brief  Check whether the minimisation is done
        LOGICAL FUNCTION DE_MINIMISATION_DONE(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER), INTENT(IN)  ::  SELF
            ! Private variables
            REAL, PARAMETER ::  MIN_SEPARATION_BTW_BEST_AND_AVERAGE_SCORE = 0.002
            ! Start work
            DE_MINIMISATION_DONE = (SELF%CURRENT_GENERATION .GE. SELF%MAX_NUM_GENERATIONS) .OR. &
                                   (SELF%AVERAGE_SCORE - SELF%BEST_MEMBER%SCORE) .LE. MIN_SEPARATION_BTW_BEST_AND_AVERAGE_SCORE
        END FUNCTION DE_MINIMISATION_DONE

        !>  \brief  Come up with a default value for the number of members in the population
        !!
        !!  We use the rule of thumb in ﻿Storn, R. (1996). On the Usage of Differential Evolution for Function Optimization. IN NAFIPS’96, 519 - 523.
        INTEGER FUNCTION DE_SUGGESTED_POPULATION_SIZE(SELF,NUM_DIM)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER), INTENT(IN)  ::  SELF
            INTEGER,            INTENT(IN)  ::  NUM_DIM !<  Number of dimensions of the minimisation problem
            ! Private variables

            ! Start work
            DE_SUGGESTED_POPULATION_SIZE = NUM_DIM * 10
        END FUNCTION DE_SUGGESTED_POPULATION_SIZE

        !>  \brief  Return the population size
        PURE INTEGER FUNCTION DE_NUM_DIM(SELF)
            IMPLICIT NONE
            ! Arguments
            CLASS(DE_MINIMISER), INTENT(IN)  ::  SELF
            !
            DE_NUM_DIM = SIZE(SELF%BEST_MEMBER%VAL)
        END FUNCTION DE_NUM_DIM

        !>  \brief  Return the number of dimensions
        PURE INTEGER FUNCTION DE_POPULATION_SIZE(SELF)
            IMPLICIT NONE
            ! Arguments
            CLASS(DE_MINIMISER), INTENT(IN)  ::  SELF
            !
            DE_POPULATION_SIZE = SIZE(SELF%POPULATION%MEMBERS)
        END FUNCTION DE_POPULATION_SIZE


        !>  \brief  Pick several's members' indices at random while optionally avoiding the index value given as argument
        SUBROUTINE DE_GET_MULITPLE_RANDOM_MEMBER_INDICES(SELF,NUM_TO_GET,RANDOM_INDICES,INDEX_TO_AVOID)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),         INTENT(IN)  ::  SELF
            INTEGER,                    INTENT(IN)  ::  NUM_TO_GET                  !<  Desired number of indices
            INTEGER,    ALLOCATABLE,    INTENT(OUT) ::  RANDOM_INDICES(:)           !<  Resulting indices
            INTEGER,                    INTENT(IN)  ::  INDEX_TO_AVOID              !<  Not allowed to return this index
            ! Private variables
            INTEGER                 ::  CURRENT_OUTPUT
            LOGICAL                 ::  CHECK_ONE, CHECK_TWO
            LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
            INTEGER                 ::  I
            ! Start work

            ! Allocate memory for the result
            ALLOCATE(RANDOM_INDICES(NUM_TO_GET))

            ! We loop until all conditions have been met
            CURRENT_OUTPUT = 1
            DO
                RANDOM_INDICES(CURRENT_OUTPUT) = DE_GET_RANDOM_MEMBER_INDEX(SELF)
                CHECK_ONE = .TRUE.
                CHECK_TWO = .TRUE.
                ! Check whether the current value is different from the previous ones
                IF (CURRENT_OUTPUT .GT. 1) THEN
                    DO I=1,CURRENT_OUTPUT-1
                        IF (RANDOM_INDICES(CURRENT_OUTPUT) .EQ. RANDOM_INDICES(I)) CHECK_ONE = .FALSE.
                    ENDDO
                ENDIF
                ! Check it's different from the index we've been told to avoid
                IF (RANDOM_INDICES(CURRENT_OUTPUT) .EQ. INDEX_TO_AVOID) CHECK_TWO = .FALSE.
                ! If we were succesful, we move on to the next one
                IF (CHECK_ONE .AND. CHECK_TWO) THEN
                    IF (CURRENT_OUTPUT .EQ. NUM_TO_GET) THEN
                        ! We're done
                        EXIT
                    ELSE
                        CURRENT_OUTPUT = CURRENT_OUTPUT + 1
                    ENDIF
                ELSE
                    ! For some reason, we were not successful. Let's try again.
                ENDIF
            ENDDO

            ! Debug
            IF (DEBUG) THEN
                WRITE(*,'(A,I0)') '**debug(DE_GET_MULTIPLE_RANDOM_MEMBER_INDICES): was told to avoid this index: ', INDEX_TO_AVOID
                WRITE(*,'(A)') '**debug(DE_GET_MULTIPLE_RANDOM_MEMBER_INDICES): the following indices were selected: '
                WRITE(*,*)  RANDOM_INDICES
            ENDIF
        END SUBROUTINE DE_GET_MULITPLE_RANDOM_MEMBER_INDICES


        !>  \brief  Pick an individual's index at random
        INTEGER FUNCTION DE_GET_RANDOM_MEMBER_INDEX(SELF) RESULT(MEMBER_INDEX)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),    INTENT(IN)  ::  SELF
            ! Private variables

            ! Start work

            ! Get an integer (uniformly distributed between 0 and the integer specified in second argument)
            MEMBER_INDEX = FGSL_RNG_UNIFORM_INT(SELF%RNG,INT(SIZE(SELF%POPULATION%MEMBERS)-1,C_LONG)) + 1


        END FUNCTION DE_GET_RANDOM_MEMBER_INDEX

        !>  \brief  Pick an dimension's index at random
        INTEGER FUNCTION DE_GET_RANDOM_DIMENSION_INDEX(SELF) RESULT(DIMENSION_INDEX)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_MINIMISER),    INTENT(IN)  ::  SELF
            ! Private variables

            ! Start work

            ! Get an integer (uniformly distributed between 0 and the integer specified in second argument)
            DIMENSION_INDEX = FGSL_RNG_UNIFORM_INT(SELF%RNG,INT(SIZE(SELF%POPULATION%MEMBERS(1)%VAL)-1,C_LONG)) + 1


        END FUNCTION DE_GET_RANDOM_DIMENSION_INDEX

        !>  \brief  Assignment operation from one individual to another
        SUBROUTINE DE_INDIVIDUAL_ASSIGNMENT(IND1,IND2)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_INDIVIDUAL),    INTENT(INOUT)   ::  IND1
            TYPE(DE_INDIVIDUAL),    INTENT(IN)      ::  IND2
            ! Private variables

            ! Start work

            ! The LHS will be overwritten
            IF (ALLOCATED(IND1%VAL)) DEALLOCATE(IND1%VAL)

            ! Copy data over
            IF (ALLOCATED(IND2%VAL)) THEN
                ALLOCATE(IND1%VAL(SIZE(IND2%VAL)))
                IND1%VAL = IND2%VAL
            ENDIF
            IND1%SCORE = IND2%SCORE
            IND1%F     = IND2%F
            IND1%CR    = IND2%CR
        END SUBROUTINE DE_INDIVIDUAL_ASSIGNMENT

        !>  \brief  Print summary information about an individual
        SUBROUTINE DE_INDIVIDUAL_PRINT_INFO(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(DE_INDIVIDUAL),    INTENT(IN)  ::  SELF
            ! Private variables
            LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
            CHARACTER(LEN=50) :: FMT
            ! Start work
            WRITE(FMT,'(A,I0,A,A)') '(A,F0.4,X,2(A,F0.2,X),A,',SIZE(SELF%VAL),'(G0.2,X)',')'
            IF (DEBUG) WRITE(*,'(2A)') '**debug(DE_INDIVIDUAL_PRINT_INFO): printout format = ', TRIM(ADJUSTL(FMT))
            WRITE(*,FMT) 'Score= ', SELF%SCORE, '; F= ', SELF%F, '; CR= ', SELF%CR, '; Values= ', SELF%VAL
        END SUBROUTINE DE_INDIVIDUAL_PRINT_INFO

END MODULE DIFFERENTIAL_EVOLUTION
