!>  \brief  Differential geometry, primarily of space curves used to define filament axes
MODULE DIFFERENTIAL_GEOMETRY
    USE SPLINES, ONLY : INTERPOL_SPLINE
    USE ISO_C_BINDING, ONLY : C_PTR
    IMPLICIT NONE

    !>  \brief  An arc-length-parametrised cubic spline space curve
    TYPE SPACE_CURVE
        TYPE(INTERPOL_SPLINE)               ::  X_SPLINE                !<  Spline modelling the X coordinate of the curve
        TYPE(INTERPOL_SPLINE)               ::  Y_SPLINE                !<  Spline modelling the Y coordinate of the curve
        TYPE(INTERPOL_SPLINE)               ::  Z_SPLINE                !<  Spline modelling the Z coordinate of the curve
        REAL(KIND=8),         ALLOCATABLE   ::  ARC_LENGTH(:)           !<  At each waypoint, the arc length from the first waypoint
        REAL(KIND=8),         ALLOCATABLE   ::  TARGET_ARC_LENGTH(:)    !<  Desired arc lengths (only used occasionally)
        REAL(KIND=8)                        ::  PARAMETRISATION_ERROR   !<  Discrepency between the spline parametrisation
                                                                        !!  and the current estimates of the arc lengths
        REAL                                ::  PARAM_ERROR_TOL         !<  Parametrisation error tolerance
        INTEGER                             ::  N                       !<  Number of waypoints
        TYPE(C_PTR)                         ::  CPTR                    !<  A C-pointer pointing to itself.
                                                                        !!  This is needed when passing the object through the FGSL.
        CONTAINS
            FINAL   ::  SPACE_CURVE_DEALLOCATE
    END TYPE

    CONTAINS

    !>  \brief  Adjust the Z coordinates of the waypoints until the desired arc lengths have been achieved
    !!
    !!  The first waypoint will be left untouched
    SUBROUTINE SPACE_CURVE_ADJUST_Z_TO_MATCH_ARC_LENGTHS(SELF,TARGET_ARC_LENGTH)
        USE MINIMISERS, ONLY : DS_INIT, DS_RUN, MINIMISER_DS_FGSL, DS_MAX_ITERATIONS, DS_SET_START_STEP
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        REAL(KIND=8),       INTENT(IN)      ::  TARGET_ARC_LENGTH(:)    !<  Target arc length at each waypoint
        ! Private variables
        TYPE(MINIMISER_DS_FGSL)                 ::  DS_MINIMISER
        REAL(KIND=8),           ALLOCATABLE     ::  FIRST_STEP(:), START_VALUES(:)
        REAL(KIND=8),           ALLOCATABLE     ::  BEST_VALUES(:)
        REAL(KIND=8)                            ::  BEST_SCORE
        LOGICAL,                PARAMETER       ::  INCREMENTAL_Z = .FALSE.
        LOGICAL,				PARAMETER		::	DEBUG = .FALSE.
		! Start work

        ! Checks
        IF (SIZE(TARGET_ARC_LENGTH) .NE. SELF%N) THEN
            WRITE(*,'(A,2(I0,1X))') '**ERROR(SPACE_CURVE_ADJUST_Z_TO_GET_ARC_LENGTHS): SIZE(TARGET_ARC_LENGTH) .NE. SELF%N: ', &
                                    SIZE(TARGET_ARC_LENGTH), SELF%N
            CALL TERMINATE('Fatal error in SPACE_CURVE_ADJUST_Z_TO_GET_ARC_LENGTHS')
        ENDIF
        IF (TARGET_ARC_LENGTH(1) .NE. 0.0) THEN
            WRITE(*,'(A,G15.3)') '**ERROR(SPACE_CURVE_ADJUST_Z_TO_GET_ARC_LENGTHS): ', &
                                    ' target arc length for first waypoint non zero: ', TARGET_ARC_LENGTH(1)
            CALL TERMINATE('Fatal error in SPACE_CURVE_ADJUST_Z_TO_GET_ARC_LENGTHS')
        ENDIF

        ! Set the target arc length
        SELF%TARGET_ARC_LENGTH(:) = TARGET_ARC_LENGTH

		IF (DEBUG) THEN
			WRITE(*,'(A)') '**debug(SPACE_CURVE_ADJUST_Z_TO_MATCH_ARC_LENGTHS): on entry, space curve info:'
			CALL SPACE_CURVE_PRINT_INFO(SELF)
		ENDIF

        ! Set up the minimiser
        CALL DS_INIT(   DS_MINIMISER,SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH_WRAP,SELF%N-1,SELF%CPTR,  &
                        INITIAL_SCORE=REAL(SPACE_CURVE_DEVIATION_FROM_TARGET_ARC_LENGTH(SELF)), &
                        RANDOM = .FALSE.,MAX_NUM_ITER=1000 )

        ! Work out the initial values and the first step
        ALLOCATE(FIRST_STEP(SELF%N-1),START_VALUES(SELF%N-1))
        IF (INCREMENTAL_Z) CALL SPACE_CURVE_Z_ABS_TO_INCR(SELF)
        START_VALUES(:) = SELF%Z_SPLINE%Y(2:)
        IF (INCREMENTAL_Z) CALL SPACE_CURVE_Z_INCR_TO_ABS(SELF)
        FIRST_STEP(:)   = 10.0D0
        CALL DS_SET_START_STEP(DS_MINIMISER,START_VALUES,FIRST_STEP)

        ! Run the minimiser
        CALL DS_RUN(DS_MINIMISER,TARGET_SIZE=0.5)

        ! Get the results
        BEST_VALUES = DS_MINIMISER%CURRENT_BEST_FGSL
        BEST_SCORE = DS_MINIMISER%CURRENT_MINIMUM_FGSL

        ! Plug the results back into the space curve
        IF (INCREMENTAL_Z) CALL SPACE_CURVE_Z_ABS_TO_INCR(SELF)
        SELF%Z_SPLINE%Y(2:) = BEST_VALUES(:)
        IF (INCREMENTAL_Z) CALL SPACE_CURVE_Z_INCR_TO_ABS(SELF)
        CALL SPACE_CURVE_UPDATE_PARAMETRISATION(SELF)

        ! Write out result
		IF (DEBUG) THEN
        	WRITE(*,'(A)') '**debug(SPACE_CURVE_ADJUST_Z_TO_MATCH_ARC_LENGTHS): on exit, space curve info:'
       		CALL SPACE_CURVE_PRINT_INFO(SELF)
		ENDIF
    END SUBROUTINE SPACE_CURVE_ADJUST_Z_TO_MATCH_ARC_LENGTHS

    !>  \brief  Given Z values for the waypoints, how far are we from the target arc lengths?
   FUNCTION SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH_WRAP(VALUES,PARAMS) RESULT(DEV) BIND(C)
        USE ISO_C_BINDING
        USE FGSL, ONLY : FGSL_VECTOR, FGSL_DOUBLE, FGSL_VECTOR_ALIGN, FGSL_SUCCESS, FGSL_OBJ_C_PTR
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        TYPE(C_PTR),        VALUE       ::  VALUES  !<  C pointer to current values to try out
        TYPE(C_PTR),        VALUE       ::  PARAMS  !<  C pointer to a filament object
        ! Result
        REAL(KIND=C_DOUBLE)             ::  DEV
        ! Private variables
        TYPE(SPACE_CURVE),  POINTER     ::  SELF
        LOGICAL,            PARAMETER   ::  DEBUG   =   .FALSE.
        INTEGER                         ::  STATUS
        TYPE(FGSL_VECTOR)               ::  VAL_VEC
        REAL(FGSL_DOUBLE),  POINTER     ::  VAL_FGSL(:)
        REAL(KIND=8),       ALLOCATABLE ::  VAL(:)
        CHARACTER(LEN=6)                ::  SIZE_VAL_C
        ! Start work

        ! Unpack the space curve object
        CALL C_F_POINTER(PARAMS,SELF)

        ! Unpack the values
        CALL FGSL_OBJ_C_PTR(VAL_VEC,VALUES)
        STATUS = FGSL_VECTOR_ALIGN(VAL_FGSL,VAL_VEC)
        IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('FGSL_VECTOR_ALIGN failed')
        ALLOCATE(VAL(SIZE(VAL_FGSL)+1),STAT=STATUS)
        IF (STATUS .NE. 0) CALL TERMINATE('Memory allocation failed')
        VAL(2:) = VAL_FGSL
        VAL(1)  = SELF%Z_SPLINE%Y(1)

        ! Sanity check
        IF (SIZE(VAL) .NE. SELF%N) THEN
            WRITE(*,'(2(A,I0))') '**ERROR(SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH_WRAP): ', SIZE(VAL), &
                                 ' values retrieved, but was expecting ', SELF%N
            CALL TERMINATE('Fatal error in SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH_WRAP')
        ENDIF

        ! Get the value
        DEV = SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH(SELF,VAL)

        ! Let the user know what values we're trying
        IF (DEBUG) THEN
            WRITE(SIZE_VAL_C,'(I6)') SIZE(VAL)
            WRITE(*,'(2A,'//SIZE_VAL_C//'(F0.3,1X))') '**debug(SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH_WRAP): ', &
			                                              'tried these values: ', VAL
            WRITE(*,'(A,F0.3)') '**debug(SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH_WRAP): got this score: ', DEV
        ENDIF
    END FUNCTION SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH_WRAP

    !>  \brief  Given Z values for the waypoints, how far are we from the target arc lengths?
    REAL(KIND=8) FUNCTION SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH(SELF,Z_VALS) RESULT(DEV)
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        REAL(KIND=8),       INTENT(IN)      ::  Z_VALS(:)
        ! Private variables
        LOGICAL,                PARAMETER       ::  INCREMENTAL_Z = .FALSE.
        ! Start work
        IF (INCREMENTAL_Z) CALL SPACE_CURVE_Z_ABS_TO_INCR(SELF)
        CALL SPACE_CURVE_REPLACE_Z_COOS(SELF,Z_VALS)
        IF (INCREMENTAL_Z) CALL SPACE_CURVE_Z_INCR_TO_ABS(SELF)
        CALL SPACE_CURVE_UPDATE_PARAMETRISATION(SELF)
        DEV = SPACE_CURVE_DEVIATION_FROM_TARGET_ARC_LENGTH(SELF)
    END FUNCTION SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH

    !>  \brief  How far are we from the target arc lengths?
    PURE REAL(KIND=8) FUNCTION SPACE_CURVE_DEVIATION_FROM_TARGET_ARC_LENGTH(SELF) RESULT(DEV)
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(IN)      ::  SELF
        ! Private variables
        INTEGER ::  I
        ! Start work
        DEV = 0.0D0
        DO I=1,SELF%N
            DEV = DEV + ABS(SELF%ARC_LENGTH(I)-SELF%TARGET_ARC_LENGTH(I))
        ENDDO
    END FUNCTION SPACE_CURVE_DEVIATION_FROM_TARGET_ARC_LENGTH

    !>  \brief  Unit tests for space curve
    SUBROUTINE SPACE_CURVE_UNIT_TESTS()
        CALL SPACE_CURVE_UNIT_TEST_1()
        CALL SPACE_CURVE_UNIT_TEST_3()
        CALL SPACE_CURVE_UNIT_TEST_4()
        CALL SPACE_CURVE_UNIT_TEST_2()
    END SUBROUTINE SPACE_CURVE_UNIT_TESTS

    !>  \brief  Generate a linear space curve and check that the parametrisation error is very very low
    SUBROUTINE SPACE_CURVE_UNIT_TEST_1()
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE COOS, ONLY : WAYPOINT
        IMPLICIT NONE
        ! Private variables
        TYPE(WAYPOINT)      ::  WPS(5)
        INTEGER             ::  I
        TYPE(SPACE_CURVE)   ::  SELF
        ! Start work
        WRITE(*,'(A)') '**info(SPACE_CURVE_UNIT_TEST_1): starting'
        DO I=1,SIZE(WPS)
            WPS(I)%COO(1:3) = 100.0 * I
        ENDDO
        CALL SPACE_CURVE_INIT(SELF,WPS,KNOWN_D=.FALSE.,PARAM_ERROR_TOL=0.01)
        IF (SELF%PARAMETRISATION_ERROR .GT. 0.01) THEN
            CALL TERMINATE('Failed unit test')
        ELSE
            WRITE(*,'(A)') '**info(SPACE_CURVE_UNIT_TEST_1): passed'
        ENDIF
    END SUBROUTINE SPACE_CURVE_UNIT_TEST_1

    !>  \brief  Generate a linear space curve, run the minimiser with the modified arc lengths as target. Things shouldn't move.
    SUBROUTINE SPACE_CURVE_UNIT_TEST_2()
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE COOS, ONLY : WAYPOINT
        IMPLICIT NONE
        ! Private variables
        TYPE(WAYPOINT)      ::  WPS(5)
        INTEGER             ::  I
        TYPE(SPACE_CURVE)   ::  SELF
        REAL(KIND=8), ALLOCATABLE        ::  TARGET_AL(:)
        ! Start work
        WRITE(*,'(A)') '**info(SPACE_CURVE_UNIT_TEST_2): starting'
        DO I=1,SIZE(WPS)
            WPS(I)%COO(1:3) = 100.0 * I
        ENDDO
        CALL SPACE_CURVE_INIT(SELF,WPS,KNOWN_D=.FALSE.,PARAM_ERROR_TOL=0.01)
        TARGET_AL = SELF%ARC_LENGTH
        TARGET_AL(2) = TARGET_AL(2) + 10.0
        CALL SPACE_CURVE_ADJUST_Z_TO_MATCH_ARC_LENGTHS(SELF,TARGET_ARC_LENGTH=TARGET_AL)
        IF (SUM(ABS(SELF%TARGET_ARC_LENGTH-SELF%ARC_LENGTH)) .GT. 0.3) THEN
            CALL TERMINATE('Failed unit test')
        ELSE
            WRITE(*,'(A)') '**info(SPACE_CURVE_UNIT_TEST_2): passed'
        ENDIF
    END SUBROUTINE SPACE_CURVE_UNIT_TEST_2

    !>  \brief  Generate a linear and closed space curves and check that they have the expected total curvatures
    SUBROUTINE SPACE_CURVE_UNIT_TEST_3()
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, PI
        USE COOS, ONLY : WAYPOINT
        IMPLICIT NONE
        ! Private variables
        TYPE(WAYPOINT)      ::  WPS(50)
        INTEGER             ::  I
        TYPE(SPACE_CURVE)   ::  SELF
        REAL                ::  MEASURED_CURVATURE
        REAL,   PARAMETER   ::  CIRCLE_RADIUS = 20.0
        REAL                ::  AZI
        ! Start work
        WRITE(*,'(A)') '**info(SPACE_CURVE_UNIT_TEST_3): starting'
        ! Straight line case, total curvature should be zero
        WRITE(*,'(A)') '**info(SPACE_CURVE_UNIT_TEST_3): straight line test'
        DO I=1,SIZE(WPS)
            WPS(I)%COO(1:3) = 100.0 * I
        ENDDO
        CALL SPACE_CURVE_INIT(SELF,WPS,KNOWN_D=.FALSE.,PARAM_ERROR_TOL=0.01)
		MEASURED_CURVATURE = SPACE_CURVE_EVAL_TOTAL_CURVATURE(SELF,STEP_SIZE=0.1D0)
        WRITE(*,'(A,G15.3)') '**info(SPACE_CURVE_UNIT_TEST_3): straight line total curvature = ', MEASURED_CURVATURE
        IF (MEASURED_CURVATURE .GT. 0.01) THEN
            WRITE(*,'(A)') '**ERROR(SPACE_CURVE_UNIT_TEST_3): straight line curvature is too large'
            CALL TERMINATE('Failed unit test')
        ENDIF
        ! Circle on a plane case, total curvature should be 2*pi
        WRITE(*,'(A)') '**info(SPACE_CURVE_UNIT_TEST_3): planar circle test'
        WPS(:)%COO(3) = 100.0
        DO I=1,SIZE(WPS)
            AZI = REAL(I-1)/REAL(SIZE(WPS)-1) * 2.0 * PI
            WPS(I)%COO(1) = CIRCLE_RADIUS * COS(AZI)
            WPS(I)%COO(2) = CIRCLE_RADIUS * SIN(AZI)
            PRINT *, '**debug(SPACE_CURVE_UNIT_TEST_3): point #, azi, x, y', I, AZI, WPS(I)%COO(1:2)
        ENDDO
        CALL SPACE_CURVE_INIT(SELF,WPS,KNOWN_D=.FALSE.,PARAM_ERROR_TOL=0.01)
        MEASURED_CURVATURE = SPACE_CURVE_EVAL_TOTAL_CURVATURE(SELF,STEP_SIZE=0.1D0)
        WRITE(*,'(A,G15.3)') '**info(SPACE_CURVE_UNIT_TEST_3): circle total curvature = ', MEASURED_CURVATURE
        IF (ABS(MEASURED_CURVATURE-2*PI) .GT. 0.01) THEN
            WRITE(*,'(A,G15.3)') '**ERROR(SPACE_CURVE_UNIT_TEST_3): circle curvature is not 2*PI, diff = ', &
                                                                        ABS(MEASURED_CURVATURE-2*PI)
            CALL TERMINATE('Failed unit test')
        ENDIF
        WRITE(*,'(A)') '**info(SPACE_CURVE_UNIT_TEST_3): passed'
    END SUBROUTINE SPACE_CURVE_UNIT_TEST_3

    !>  \brief  Check that "wobbly"-looking space curves get a higher total curvature than less wobbly ones
    SUBROUTINE SPACE_CURVE_UNIT_TEST_4()
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, PI, IS_ODD
        USE COOS, ONLY : WAYPOINT
        IMPLICIT NONE
        ! Private variables
        TYPE(WAYPOINT)      ::  WPS_1(5), WPS_2(5)
        TYPE(SPACE_CURVE)   ::  SC_1, SC_2
        REAL                ::  CURVATURE_1, CURVATURE_2
        REAL,   PARAMETER   ::  CIRCLE_RADIUS = 50.0
        ! Start work
        WRITE(*,'(A)') '**info(SPACE_CURVE_UNIT_TEST_4): starting'

        ! Set up the two sets of waypoints (coos)
        WPS_1(1)%COO = [2119.290,    1934.801,  -14111.840]
        WPS_1(2)%COO = [2183.570,    1918.880,  -14100.175]
        WPS_1(3)%COO = [2240.970,    1895.530,  -14089.247]
        WPS_1(4)%COO = [2290.360,    1873.920,  -14080.311]
        WPS_1(5)%COO = [2351.010,    1857.790,  -14070.626]
        WPS_2(1)%COO = [2119.290,    1934.801,  -14111.840]
        WPS_2(2)%COO = [2183.570,    1918.880,  -14060.631]
        WPS_2(3)%COO = [2240.970,    1895.530,  -14108.365]
        WPS_2(4)%COO = [2290.360,    1873.920,  -14072.440]
		WPS_2(4)%COO = [2290.360,    1873.920,  -14372.440]
        WPS_2(5)%COO = [2351.010,    1857.790,  -14023.620]
        ! Now setup theta and psi
        WPS_1(1)%EULER%THETA = 82.
        WPS_1(2)%EULER%THETA = 82.
        WPS_1(3)%EULER%THETA = 82.
        WPS_1(4)%EULER%THETA = 83.
        WPS_1(5)%EULER%THETA = 83.
        WPS_1(1)%EULER%PSI   = -20.36
        WPS_1(2)%EULER%PSI   = -20.39
        WPS_1(3)%EULER%PSI   = -20.33
        WPS_1(4)%EULER%PSI   = -20.20
        WPS_1(5)%EULER%PSI   = -20.08
        WPS_2(1)%EULER%THETA = 82.
        WPS_2(2)%EULER%THETA = 82.
        WPS_2(3)%EULER%THETA = 82.
        WPS_2(4)%EULER%THETA = 83.
        WPS_2(5)%EULER%THETA = 83.
        WPS_2(1)%EULER%PSI   = -20.36
        WPS_2(2)%EULER%PSI   = -20.39
        WPS_2(3)%EULER%PSI   = -20.33
        WPS_2(4)%EULER%PSI   = -20.20
        WPS_2(5)%EULER%PSI   = -20.08


        CALL SPACE_CURVE_INIT(SC_1,WPS_1,KNOWN_D=.TRUE.,PARAM_ERROR_TOL=0.01)
        CALL SPACE_CURVE_INIT(SC_2,WPS_2,KNOWN_D=.TRUE.,PARAM_ERROR_TOL=0.01)
        CURVATURE_1 = SPACE_CURVE_EVAL_TOTAL_CURVATURE(SC_1,STEP_SIZE=0.1D0) * SC_1%ARC_LENGTH(SC_1%N)
        CURVATURE_2 = SPACE_CURVE_EVAL_TOTAL_CURVATURE(SC_2,STEP_SIZE=0.1D0) * SC_2%ARC_LENGTH(SC_2%N)
        CALL SPACE_CURVE_PRINT_INFO(SC_1)
        CALL SPACE_CURVE_PRINT_INFO(SC_2)
        WRITE(*,'(A,2(G15.3,1X))') '**info(SPACE_CURVE_UNIT_TEST_4): curvatures = ', CURVATURE_1,CURVATURE_2
        IF (CURVATURE_1 .GT. CURVATURE_2) THEN
            WRITE(*,'(A)') '**ERROR(SPACE_CURVE_UNIT_TEST_4): curvature got smaller when curve got wobblier'
            CALL TERMINATE('Failed unit test')
        ENDIF

    END SUBROUTINE SPACE_CURVE_UNIT_TEST_4

    !>  \brief  Update a space curve with new points (and, optionally, Euler angles)
    SUBROUTINE SPACE_CURVE_UPDATE(SELF,WAYPOINTS)
        USE COOS, ONLY : WAYPOINT
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        TYPE(WAYPOINT),     INTENT(IN)      ::  WAYPOINTS(:)    !<  Points through which the space curve travels
        ! Private variables

        ! Start work
        CALL SPACE_CURVE_GET_DATA_FROM_WAYPOINTS(SELF,WAYPOINTS,SELF%X_SPLINE%KNOWN_D)
        CALL SPACE_CURVE_UPDATE_PARAMETRISATION(SELF)
    END SUBROUTINE SPACE_CURVE_UPDATE

    !>  \brief  Initialise a space cruve
    SUBROUTINE SPACE_CURVE_INIT(SELF,WAYPOINTS,KNOWN_D,PARAM_ERROR_TOL)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE COOS, ONLY : WAYPOINT
        USE SPLINES, ONLY : SPLINE_SET_TYPE
        USE ISO_C_BINDING, ONLY : C_LOC
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  TARGET, INTENT(INOUT)   ::  SELF
        TYPE(WAYPOINT),             INTENT(IN)      ::  WAYPOINTS(:)    !<  Points through which the space curve travels
        LOGICAL,                    INTENT(IN)      ::  KNOWN_D         !<  Whether we use splines which are solved using known
                                                                        !!  1st derivatives derived from the waypoints' Euler angles
        REAL                                        ::  PARAM_ERROR_TOL !<  Tolerated parametrisation error
        ! Private variables
        INTEGER ::  MIN_NUM_WPS
        ! Start work

        ! Parametrisation error tolerance
        SELF%PARAM_ERROR_TOL = PARAM_ERROR_TOL

        ! Check size of waypoints
        IF (KNOWN_D) THEN
            MIN_NUM_WPS = 2
        ELSE
            MIN_NUM_WPS = 3
        ENDIF
        IF (SIZE(WAYPOINTS) .LT. MIN_NUM_WPS) THEN
            WRITE(*,'(A,I0,A)') '**ERROR(SPACE_CURVE_INIT): number of wps (', SIZE(WAYPOINTS), ') not sufficient'
            CALL TERMINATE('Fatal error in SPACE_CURVE_INIT')
        ENDIF

        ! Allocate memory
        CALL SPACE_CURVE_ALLOCATE(SELF,SIZE(WAYPOINTS))

        ! Set up the C pointer to self
        SELF%CPTR = C_LOC(SELF)

        ! Copy data over from waypoints
        CALL SPACE_CURVE_GET_DATA_FROM_WAYPOINTS(SELF,WAYPOINTS,KNOWN_D)

        ! First guess at arc length
        CALL SPACE_CURVE_GUESS_ARC_LENGTH(SELF)

        ! Copy arc length values to X coordinates of splines
        CALL SPACE_CURVE_ARC_LENGTH_TO_SPLINES(SELF)

        ! Set spline type
        IF (KNOWN_D) THEN
            CALL SPLINE_SET_TYPE(SELF%X_SPLINE,'KNOWN_D')
            CALL SPLINE_SET_TYPE(SELF%Y_SPLINE,'KNOWN_D')
            CALL SPLINE_SET_TYPE(SELF%Z_SPLINE,'KNOWN_D')
        ELSE
            CALL SPLINE_SET_TYPE(SELF%X_SPLINE,'PARABOLIC')
            CALL SPLINE_SET_TYPE(SELF%Y_SPLINE,'PARABOLIC')
            CALL SPLINE_SET_TYPE(SELF%Z_SPLINE,'PARABOLIC')
        ENDIF

        ! Solve the splines for the first time
        CALL SPACE_CURVE_SOLVE_SPLINES(SELF)

        ! Parametrise the space curve properly
        CALL SPACE_CURVE_UPDATE_PARAMETRISATION(SELF)

        ! Let's zero the target arc lengths
        SELF%TARGET_ARC_LENGTH = 0.0D0

    END SUBROUTINE SPACE_CURVE_INIT

    !>  \brief  Convert from absolute to relative Z values
    PURE SUBROUTINE SPACE_CURVE_Z_ABS_TO_INCR(SELF)
        IMPLICIT NONE
        ! Arguemnts
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  I
        ! Start work
        DO I=SELF%N,2,-1
            SELF%Z_SPLINE%Y(I) = SELF%Z_SPLINE%Y(I) - SELF%Z_SPLINE%Y(I-1)
        ENDDO
    END SUBROUTINE SPACE_CURVE_Z_ABS_TO_INCR

    !>  \brief  Convert from absolute to relative Z values
    PURE SUBROUTINE SPACE_CURVE_Z_INCR_TO_ABS(SELF)
        IMPLICIT NONE
        ! Arguemnts
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  I
        ! Start work
        DO I=2,SELF%N
            SELF%Z_SPLINE%Y(I) = SELF%Z_SPLINE%Y(I) + SELF%Z_SPLINE%Y(I-1)
        ENDDO
    END SUBROUTINE SPACE_CURVE_Z_INCR_TO_ABS

    !>  \brief  Replace the Z values
    PURE SUBROUTINE SPACE_CURVE_REPLACE_Z_COOS(SELF,NEW_Z)
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        REAL(KIND=8),       INTENT(IN)      ::  NEW_Z(:)
        !
        SELF%Z_SPLINE%Y(:) = NEW_Z(:)
    END SUBROUTINE SPACE_CURVE_REPLACE_Z_COOS

    !>  \brief  Copy data from waypoints to the space curve
    PURE SUBROUTINE SPACE_CURVE_GET_DATA_FROM_WAYPOINTS(SELF,WAYPOINTS,KNOWN_D)
        USE COOS, ONLY : WAYPOINT
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        TYPE(WAYPOINT),     INTENT(IN)      ::  WAYPOINTS(:)    !<  Points through which the space curve travels
        LOGICAL,            INTENT(IN)      ::  KNOWN_D         !<  Whether we use splines which are solved using known
                                                                !!  1st derivatives derived from the waypoints' Euler angles
        ! Start work
        SELF%X_SPLINE%Y(:) = WAYPOINTS(:)%COO(1)
        SELF%Y_SPLINE%Y(:) = WAYPOINTS(:)%COO(2)
        SELF%Z_SPLINE%Y(:) = WAYPOINTS(:)%COO(3)
        IF (KNOWN_D) THEN
            SELF%X_SPLINE%DY(:) = DX_FROM_THETA_PSI(WAYPOINTS(:)%EULER%THETA,WAYPOINTS(:)%EULER%PSI)
            SELF%Y_SPLINE%DY(:) = DY_FROM_THETA_PSI(WAYPOINTS(:)%EULER%THETA,WAYPOINTS(:)%EULER%PSI)
            SELF%Z_SPLINE%DY(:) = DZ_FROM_THETA(WAYPOINTS(:)%EULER%THETA)
        ENDIF
    END SUBROUTINE SPACE_CURVE_GET_DATA_FROM_WAYPOINTS

    !>  \brief  Update the arc-length parametrisation of the space curve
    SUBROUTINE SPACE_CURVE_UPDATE_PARAMETRISATION(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER                 ::  PARAMETRISATION_NUM_ITERATIONS
        REAL                    ::  PREVIOUS_PARAMETRISATION_ERROR
        INTEGER,    PARAMETER   ::  PARAMETRISATION_MAX_NUM_ITERATIONS  =   300
        INTEGER,    PARAMETER   ::  PARAMETRISATION_MIN_NUM_ITERATIONS  =   2
        !REAL(KIND=8)            ::  PARAMETRISATION_TARGET_ERROR
        !REAL(KIND=8)            ::  CURRENT_PARAMETRISATION_ERROR
        ! Start work

        ! Iterative parametrisation
        PARAMETRISATION_NUM_ITERATIONS = 0
        PREVIOUS_PARAMETRISATION_ERROR = HUGE(1.0)
        DO
            PARAMETRISATION_NUM_ITERATIONS = PARAMETRISATION_NUM_ITERATIONS + 1
            CALL SPACE_CURVE_ARC_LENGTH_TO_SPLINES(SELF)
            CALL SPACE_CURVE_SOLVE_SPLINES(SELF)
            CALL SPACE_CURVE_UPDATE_ARC_LENGTH(SELF)
            CALL SPACE_CURVE_UPDATE_PARAMETRISATION_ERROR(SELF)

            IF (    SELF%PARAMETRISATION_ERROR .LT. SELF%PARAM_ERROR_TOL &
                    .AND. PARAMETRISATION_NUM_ITERATIONS .GE. PARAMETRISATION_MIN_NUM_ITERATIONS &
                    ) THEN
                EXIT
            ENDIF
            PREVIOUS_PARAMETRISATION_ERROR = SELF%PARAMETRISATION_ERROR
            IF (PARAMETRISATION_NUM_ITERATIONS .GT. PARAMETRISATION_MAX_NUM_ITERATIONS) THEN
                WRITE(*,'(A,I4,2(A,F13.10))')   '**warning(SPACE_CURVE_UPDATE_PARAMETRISATION): spline parametrisation ',  &
                                                'did not converge after ', PARAMETRISATION_NUM_ITERATIONS, &
                                                        ' iterations. Current error = ', SELF%PARAMETRISATION_ERROR, &
                                                        '. Target error =', SELF%PARAM_ERROR_TOL
                IF (PARAMETRISATION_NUM_ITERATIONS .GT. 2*PARAMETRISATION_MAX_NUM_ITERATIONS) THEN
                    CALL TERMINATE('Fatal error in FILAMENT_PARAMETRISATION_UPDATE')
                ENDIF
            ENDIF
        ENDDO
    END SUBROUTINE SPACE_CURVE_UPDATE_PARAMETRISATION


    !>  \brief  Compute the discrepency between the parameter used by the splines and the current estimate of the arc length
    PURE SUBROUTINE SPACE_CURVE_UPDATE_PARAMETRISATION_ERROR(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  I
        ! Start work
        SELF%PARAMETRISATION_ERROR = 0.0D0
        DO I=1,SELF%N
            SELF%PARAMETRISATION_ERROR = SELF%PARAMETRISATION_ERROR + ABS(SELF%ARC_LENGTH(I)-SELF%X_SPLINE%X(I))
        ENDDO
        SELF%PARAMETRISATION_ERROR = SELF%PARAMETRISATION_ERROR / SELF%N
    END SUBROUTINE SPACE_CURVE_UPDATE_PARAMETRISATION_ERROR

    !>  \brief  Compute the arc length from first waypoint to every other waypoint in the space curve.
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE SPACE_CURVE_UPDATE_ARC_LENGTH(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE), INTENT(INOUT)   ::  SELF    !<  Filament to be updated
        ! Private variables
        INTEGER ::  I
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        CHARACTER(LEN=6) :: NWP_C
        ! Start work
        ! Start at the origin
        SELF%ARC_LENGTH(1) = 0.0D0
        DO I=2,SELF%N
            CALL SPACE_CURVE_EVAL_ARC_LENGTH(SELF,SELF%X_SPLINE%X(I),SELF%ARC_LENGTH(I))
        ENDDO
#ifndef SKIP_RUNTIME_CHECKS
        IF (DEBUG) THEN
            WRITE(NWP_C,'(I6)') SELF%N
            IF (ANY(SELF%ARC_LENGTH .NE. 0.)) THEN
                ! Sanity check
                DO I=2,SELF%N
                    IF (SELF%ARC_LENGTH(I) .LE. SELF%ARC_LENGTH(I-1)) THEN
                        WRITE(*,'(A,'//NWP_C//'(F0.3,X))') '**ERROR(SPACE_CURVE_UPDATE_ARC_LENGTH): weird arc lengths: ', &
                                SELF%ARC_LENGTH
                        CALL TERMINATE('Fatal error in SPACE_CURVE_UPDATE_ARC_LENGTH')
                    ENDIF
                ENDDO
            ENDIF
        ENDIF
#endif
    END SUBROUTINE SPACE_CURVE_UPDATE_ARC_LENGTH

    !>  \brief  Compute the arc length from first waypoint to the point specified by the coordinate T, using numerical integration.
    !!
    !!  Should only be called when T falls directly after a waypoint for which the path_length has already been determined
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE SPACE_CURVE_EVAL_ARC_LENGTH(SELF,T,ARC_LENGTH)
        USE SPLINES, ONLY : INTERPOL_SPLINE_FIND_SEG, INTERPOL_SPLINE_PRINT_INFO, INTERPOL_SPLINE_EVAL
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),      INTENT(INOUT)   ::  SELF            !<  Filament
        REAL(KIND=8),           INTENT(IN)      ::  T               !<  Parameter value for which the path length needs to be computed
        REAL(KIND=8),           INTENT(OUT)     ::  ARC_LENGTH     !<  Result. Arc length.
        ! Private variables
        REAL(KIND=8),           PARAMETER       ::  TINY_STEP = 0.0005D0    !<  Step for the integration
        INTEGER                                 ::  SEG
        REAL(KIND=8)                            ::  CURRENT_T,CURRENT_X,CURRENT_Y,CURRENT_Z,PREVIOUS_X,PREVIOUS_Y,PREVIOUS_Z
        LOGICAL,                PARAMETER       ::  DEBUG   =   .FALSE.
        ! Start work
        ! Figure out which segment of the spline we'll be working on
        SEG = INTERPOL_SPLINE_FIND_SEG(SELF%X_SPLINE,T)
		IF (DEBUG) THEN
#ifndef SKIP_RUNTIME_CHECKS
        WRITE(*,'(A,F0.3,1X,I0)') '**debug(SPACE_CURVE_EVAL_ARC_LENGTH): T, seg = ', T, SEG
        IF (SEG .EQ. 0) THEN
            WRITE(*,'(A,F0.5)') '**ERROR(SPACE_CURVE_EVAL_ARC_LENGTH): could not find spline segment for T = ', T
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%X_SPLINE)
            CALL TERMINATE('Fatal error in SPACE_CURVE_EVAL_ARC_LENGTH')
        ENDIF
#endif
        ENDIF
        ! Start at the start of that segment
        ARC_LENGTH = SELF%ARC_LENGTH(SEG)
        CURRENT_T = SELF%X_SPLINE%X(SEG)
        CURRENT_X = SELF%X_SPLINE%Y(SEG)
        CURRENT_Y = SELF%Y_SPLINE%Y(SEG)
        CURRENT_Z = SELF%Z_SPLINE%Y(SEG)

        ! Loop until we've reached the point we're interested in
        DO
            ! Keep track of where we've just been
            PREVIOUS_X = CURRENT_X
            PREVIOUS_Y = CURRENT_Y
            PREVIOUS_Z = CURRENT_Z
            ! One tiny step for man...
            CURRENT_T = CURRENT_T + TINY_STEP
            ! Check we've not stepped over the mark
            IF (CURRENT_T .GT. T) THEN
                EXIT    !   Done, get out
            ENDIF
            ! We're still walking - work out where our foot lands
            CALL INTERPOL_SPLINE_EVAL(SELF%X_SPLINE,CURRENT_T,CURRENT_X)
            CALL INTERPOL_SPLINE_EVAL(SELF%Y_SPLINE,CURRENT_T,CURRENT_Y)
            CALL INTERPOL_SPLINE_EVAL(SELF%Z_SPLINE,CURRENT_T,CURRENT_Z)
            ! Measure & accumulate the distance
            ARC_LENGTH = ARC_LENGTH + SQRT((CURRENT_X-PREVIOUS_X)**2+(CURRENT_Y-PREVIOUS_Y)**2+(CURRENT_Z-PREVIOUS_Z)**2)
        ENDDO
    END SUBROUTINE SPACE_CURVE_EVAL_ARC_LENGTH

    !>  \brief  Evaluate the total curvature along a space curve. Optionally, the total squared curvature can be returned.
    PURE FUNCTION SPACE_CURVE_EVAL_TOTAL_CURVATURE_NEW(SELF,STEP_SIZE,SQUARED_CURVATURE,IGNORE_FAL_WPS) RESULT(TOTAL_CURVATURE)
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL_DD
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),          INTENT(IN)  ::  SELF
        REAL,                       INTENT(IN)  ::  STEP_SIZE           !<  Integration step size (good default for a
                                                                        !!  filament axisis the helical rise per subunit)
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  SQUARED_CURVATURE   !<  Return the total squared curvature. Defaults to false,
                                                                        !!  in which case the total curvature (not squared)
                                                                        !!  is returned.
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  IGNORE_FAL_WPS      !<  Ignore the first and last waypoints
                                                                        !!  in the calculation. False by default.
        ! Result
        REAL(KIND=8)    ::  TOTAL_CURVATURE
        ! Private variables
        LOGICAL         ::  SSQUARED_CURVATURE
        REAL(KIND=8)    ::  CURRENT_T
        LOGICAL         ::  IIGNORE_FAL_WPS
        INTEGER         ::  FIRST_WP, LAST_WP
        ! Start work

        ! Will we be returning the squared curvature?
        SSQUARED_CURVATURE = .FALSE.
        IF (PRESENT(SQUARED_CURVATURE)) SSQUARED_CURVATURE = SQUARED_CURVATURE

        ! Ignore FAL WPS?
        IIGNORE_FAL_WPS = .FALSE.
        IF (PRESENT(IGNORE_FAL_WPS)) IIGNORE_FAL_WPS = IGNORE_FAL_WPS

        ! First and last waypoints
        IF (IIGNORE_FAL_WPS) THEN
            FIRST_WP    =   2
            LAST_WP     =   SELF%N - 1
        ELSE
            FIRST_WP    =   1
            LAST_WP     =   SELF%N
        ENDIF

        ! Total curvature
        TOTAL_CURVATURE = 0.0D0
        CURRENT_T = SELF%X_SPLINE%X(FIRST_WP)
        IF (SSQUARED_CURVATURE) THEN
!            DO
!                TOTAL_CURVATURE = TOTAL_CURVATURE + SPACE_CURVE_EVAL_CURVATURE(SELF,CURRENT_T)**2
!                CURRENT_T = CURRENT_T + STEP_SIZE
!                IF (CURRENT_T .GT. SELF%X_SPLINE%X(LAST_WP)) EXIT
!            ENDDO
        ELSE
            DO
                TOTAL_CURVATURE = TOTAL_CURVATURE + SQRT(   INTERPOL_SPLINE_EVAL_DD(SELF%X_SPLINE,CURRENT_T)**2 &
                                                        +   INTERPOL_SPLINE_EVAL_DD(SELF%Y_SPLINE,CURRENT_T)**2 &
                                                        +   INTERPOL_SPLINE_EVAL_DD(SELF%Z_SPLINE,CURRENT_T)**2 ) &
	 													* SIGN(1.0D0,SPACE_CURVE_EVAL_TORSION(SELF,CURRENT_T))
                !TOTAL_CURVATURE = TOTAL_CURVATURE + SQRT(  SPACE_CURVE_EVAL_CURVATURE(SELF,CURRENT_T)**2 &
				!                                         + SPACE_CURVE_EVAL_TORSION(SELF,CURRENT_T)**2   )
				CURRENT_T = CURRENT_T + STEP_SIZE
                IF (CURRENT_T .GT. SELF%X_SPLINE%X(LAST_WP)) EXIT
            ENDDO
        ENDIF
        TOTAL_CURVATURE = TOTAL_CURVATURE * STEP_SIZE
    END FUNCTION SPACE_CURVE_EVAL_TOTAL_CURVATURE_NEW

    !>  \brief  Evaluate the total curvature along a space curve. Optionally, the total squared curvature can be returned.
    PURE FUNCTION SPACE_CURVE_EVAL_TOTAL_CURVATURE(SELF,STEP_SIZE,SQUARED_CURVATURE,IGNORE_FAL_WPS) RESULT(TOTAL_CURVATURE)
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),          INTENT(IN)  ::  SELF
        REAL(KIND=8),               INTENT(IN)  ::  STEP_SIZE           !<  Integration step size (good default for a
                                                                        !!  filament axisis the helical rise per subunit)
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  SQUARED_CURVATURE   !<  Return the total squared curvature. Defaults to false,
                                                                        !!  in which case the total curvature (not squared)
                                                                        !!  is returned.
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  IGNORE_FAL_WPS      !<  Ignore the first and last waypoints
                                                                        !!  in the calculation. False by default.
        ! Result
        REAL(KIND=8)    ::  TOTAL_CURVATURE
        ! Private variables
        LOGICAL         ::  SSQUARED_CURVATURE
        REAL(KIND=8)    ::  CURRENT_T
        LOGICAL         ::  IIGNORE_FAL_WPS
        INTEGER         ::  FIRST_WP, LAST_WP
        ! Start work

        ! Will we be returning the squared curvature?
        SSQUARED_CURVATURE = .FALSE.
        IF (PRESENT(SQUARED_CURVATURE)) SSQUARED_CURVATURE = SQUARED_CURVATURE

        ! Ignore FAL WPS?
        IIGNORE_FAL_WPS = .FALSE.
        IF (PRESENT(IGNORE_FAL_WPS)) IIGNORE_FAL_WPS = IGNORE_FAL_WPS

        ! First and last waypoints
        IF (IIGNORE_FAL_WPS) THEN
            FIRST_WP    =   2
            LAST_WP     =   SELF%N - 1
        ELSE
            FIRST_WP    =   1
            LAST_WP     =   SELF%N
        ENDIF

        ! Total curvature
        TOTAL_CURVATURE = 0.0D0
        CURRENT_T = SELF%X_SPLINE%X(FIRST_WP)
        IF (SSQUARED_CURVATURE) THEN
            DO
                TOTAL_CURVATURE = TOTAL_CURVATURE + SPACE_CURVE_EVAL_CURVATURE(SELF,CURRENT_T)**2
                CURRENT_T = CURRENT_T + STEP_SIZE
                IF (CURRENT_T .GT. SELF%X_SPLINE%X(LAST_WP)) EXIT
            ENDDO
        ELSE
            DO
                TOTAL_CURVATURE = TOTAL_CURVATURE + SPACE_CURVE_EVAL_CURVATURE(SELF,CURRENT_T)
                CURRENT_T = CURRENT_T + STEP_SIZE
                IF (CURRENT_T .GT. SELF%X_SPLINE%X(LAST_WP)) EXIT
            ENDDO
        ENDIF
        TOTAL_CURVATURE = TOTAL_CURVATURE * STEP_SIZE
    END FUNCTION SPACE_CURVE_EVAL_TOTAL_CURVATURE

	!>	\brief	Evaluate the local torsion
	PURE FUNCTION SPACE_CURVE_EVAL_TORSION(SELF,T) RESULT(TORSION)
		USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL_D, INTERPOL_SPLINE_EVAL_DD, INTERPOL_SPLINE_EVAL_DDD
		IMPLICIT NONE
		! Arguments
		TYPE(SPACE_CURVE),  INTENT(IN)  ::  SELF
        REAL(KIND=8),       INTENT(IN)  ::  T
        ! Result
        REAL(KIND=8)    ::  TORSION
        ! Private variables
        REAL(KIND=8)    ::  DX,DY,DZ
        REAL(KIND=8)    ::  DDX,DDY,DDZ
        REAL(KIND=8)    ::  DDDX,DDDY,DDDZ
        REAL(KIND=8)    ::  A,B,C
        ! Start work

        DX  =   INTERPOL_SPLINE_EVAL_D(SELF%X_SPLINE,T)
        DY  =   INTERPOL_SPLINE_EVAL_D(SELF%Y_SPLINE,T)
        DZ  =   INTERPOL_SPLINE_EVAL_D(SELF%Z_SPLINE,T)
        DDX =   INTERPOL_SPLINE_EVAL_DD(SELF%X_SPLINE,T)
        DDY =   INTERPOL_SPLINE_EVAL_DD(SELF%Y_SPLINE,T)
        DDZ =   INTERPOL_SPLINE_EVAL_DD(SELF%Z_SPLINE,T)
        DDDX =  INTERPOL_SPLINE_EVAL_DDD(SELF%X_SPLINE,T)
        DDDY =  INTERPOL_SPLINE_EVAL_DDD(SELF%Y_SPLINE,T)
        DDDZ =  INTERPOL_SPLINE_EVAL_DDD(SELF%Z_SPLINE,T)

        A = DY*DDZ - DDY*DZ
        B = DDX*DZ - DX*DDZ
        C = DX*DDY - DDX*DY

        TORSION =     (DDDX*A + DDDY*B + DDDZ*C) &
                    / ( A**2 + B**2 + C**2 )
	END FUNCTION SPACE_CURVE_EVAL_TORSION

    !>  \brief  Evaluate the local curvature
    PURE FUNCTION SPACE_CURVE_EVAL_CURVATURE(SELF,T) RESULT(CURVATURE)
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL_D, INTERPOL_SPLINE_EVAL_DD
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(IN)  ::  SELF
        REAL(KIND=8),       INTENT(IN)  ::  T
        ! Result
        REAL(KIND=8)    ::  CURVATURE
        ! Private variables
        REAL(KIND=8)    ::  DX,DY,DZ
        REAL(KIND=8)    ::  DDX,DDY,DDZ
        LOGICAL,	PARAMETER	::	ASSUME_ARC_LENGTH_PARAMETRISATION = .FALSE.
		! Start work

		IF (.NOT. ASSUME_ARC_LENGTH_PARAMETRISATION) THEN
	        DX  =   INTERPOL_SPLINE_EVAL_D(SELF%X_SPLINE,T)
        	DY  =   INTERPOL_SPLINE_EVAL_D(SELF%Y_SPLINE,T)
    	    DZ  =   INTERPOL_SPLINE_EVAL_D(SELF%Z_SPLINE,T)
		ENDIF
        DDX =   INTERPOL_SPLINE_EVAL_DD(SELF%X_SPLINE,T)
        DDY =   INTERPOL_SPLINE_EVAL_DD(SELF%Y_SPLINE,T)
        DDZ =   INTERPOL_SPLINE_EVAL_DD(SELF%Z_SPLINE,T)

    	IF (ASSUME_ARC_LENGTH_PARAMETRISATION) THEN
		    CURVATURE = SQRT(DDX**2+DDY**2+DDZ**2)
		ELSE
			CURVATURE = SQRT((DDZ*DY-DDY*DZ)**2+(DDX*DZ-DDZ*DX)**2+(DDY*DX-DDX*DY)**2)  &
        	            /   &
            	        ((DX**2+DY**2+DZ**2)**(3/2))
		ENDIF

    END FUNCTION SPACE_CURVE_EVAL_CURVATURE

    !>  \brief  Solve space curve splines
    SUBROUTINE SPACE_CURVE_SOLVE_SPLINES(SELF)
        USE SPLINES, ONLY : INTERPOL_SPLINE_SOLVE, INTERPOL_SPLINE_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        ! Private variables

        ! Start work
		CALL INTERPOL_SPLINE_SOLVE(SELF%X_SPLINE)
		CALL INTERPOL_SPLINE_SOLVE(SELF%Y_SPLINE)
        CALL INTERPOL_SPLINE_SOLVE(SELF%Z_SPLINE)
    END SUBROUTINE SPACE_CURVE_SOLVE_SPLINES

    !>  \brief  Copy the arc length values to the "X" parameter value of each spline
    PURE SUBROUTINE SPACE_CURVE_ARC_LENGTH_TO_SPLINES(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        ! Start work
        SELF%X_SPLINE%X(:) = SELF%ARC_LENGTH(:)
        SELF%Y_SPLINE%X(:) = SELF%ARC_LENGTH(:)
        SELF%Z_SPLINE%X(:) = SELF%ARC_LENGTH(:)
    END SUBROUTINE SPACE_CURVE_ARC_LENGTH_TO_SPLINES

    !>  \brief  Guess the arc length of the space curve without using the splines. Use the chord length approximation.
    PURE SUBROUTINE SPACE_CURVE_GUESS_ARC_LENGTH(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  I
        ! Start work
        SELF%ARC_LENGTH(1) = 0.0
        DO I=2,SELF%N
            SELF%ARC_LENGTH(I) = SELF%ARC_LENGTH(I-1) &
                                 +  SQRT(   (SELF%X_SPLINE%Y(I) - SELF%X_SPLINE%Y(I-1))**2 &
                                         +  (SELF%Y_SPLINE%Y(I) - SELF%Y_SPLINE%Y(I-1))**2 &
                                         +  (SELF%Z_SPLINE%Y(I) - SELF%Z_SPLINE%Y(I-1))**2    )
        ENDDO
    END SUBROUTINE SPACE_CURVE_GUESS_ARC_LENGTH

    !>  \brief  Deallocate memory for a space curve
    SUBROUTINE SPACE_CURVE_DEALLOCATE(SELF)
        USE SPLINES, ONLY : INTERPOL_SPLINE_DEALLOCATE
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        ! Private variables

        ! Start work
        CALL INTERPOL_SPLINE_DEALLOCATE(SELF%X_SPLINE)
        CALL INTERPOL_SPLINE_DEALLOCATE(SELF%Y_SPLINE)
        CALL INTERPOL_SPLINE_DEALLOCATE(SELF%Z_SPLINE)
        IF (ALLOCATED(SELF%ARC_LENGTH)) DEALLOCATE(SELF%ARC_LENGTH)
    END SUBROUTINE SPACE_CURVE_DEALLOCATE

    !>  \brief  Allocate memory for a space curve
    SUBROUTINE SPACE_CURVE_ALLOCATE(SELF,N)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE SPLINES, ONLY : INTERPOL_SPLINE_ALLOCATE
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(INOUT)   ::  SELF
        INTEGER,            INTENT(IN)      ::  N       !<  Number of waypoints
        ! Private variables

        ! Start work
        ! Check
        IF (N .LT. 1) THEN
            WRITE(*,'(A,I0)') '**ERROR(SPACE_CURVE_ALLOCATE): invalid number of waypoints: ', N
            CALL TERMINATE('Fatal error in SPACE_CURVE_ALLOCATE')
        ENDIF
        ! Allocate
        CALL INTERPOL_SPLINE_ALLOCATE(SELF%X_SPLINE,N,3)
        CALL INTERPOL_SPLINE_ALLOCATE(SELF%Y_SPLINE,N,3)
        CALL INTERPOL_SPLINE_ALLOCATE(SELF%Z_SPLINE,N,3)
        IF (ALLOCATED(SELF%ARC_LENGTH)) DEALLOCATE(SELF%ARC_LENGTH)
        IF (ALLOCATED(SELF%TARGET_ARC_LENGTH)) DEALLOCATE(SELF%TARGET_ARC_LENGTH)
        ALLOCATE(SELF%ARC_LENGTH(N),SELF%TARGET_ARC_LENGTH(N))
        SELF%N = N
    END SUBROUTINE SPACE_CURVE_ALLOCATE

    !>  \brief  Print summary information about the space curve
    SUBROUTINE SPACE_CURVE_PRINT_INFO(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(SPACE_CURVE),  INTENT(IN)  ::  SELF
        ! Private variables
        CHARACTER(LEN=2)    :: N_C
        INTEGER             :: I
        ! Start work
        WRITE(N_C,'(I2)') SELF%N

        WRITE(*,'(A)')                  'Space curve info:'
        WRITE(*,'(2X,A,I0)')            'Number of waypoints: ', SELF%N
        WRITE(*,'(2X,2(A,F0.5),A)')     'Parametrisation error (tolerance): ', SELF%PARAMETRISATION_ERROR, &
                                                                              ' (', SELF%PARAM_ERROR_TOL, ')'
        WRITE(*,'(2X,A,F0.3)')          'Arc/chord length: ', SELF%ARC_LENGTH(SELF%N) / SPACE_CURVE_CHORD_LENGTH(SELF)
		WRITE(*,'(2X,A3,1X,12(A15,1X))') 'WP#', 'T', 'ArcLength', 'TargetArcLength', 'X', 'Y', 'Z','DX','DY','DZ','DDX','DDY','DDZ'
        DO I=1,SELF%N
            WRITE(*,'(2X,I3,1X,12(F15.3,1X))') I,  SELF%X_SPLINE%X(I), SELF%ARC_LENGTH(I), SELF%TARGET_ARC_LENGTH(I),   &
                                                SELF%X_SPLINE%Y(I), SELF%Y_SPLINE%Y(I), SELF%Z_SPLINE%Y(I), &
                                                SELF%X_SPLINE%DY(I), SELF%Y_SPLINE%DY(I), SELF%Z_SPLINE%DY(I), &
                                                SELF%X_SPLINE%DDY(I), SELF%Y_SPLINE%DDY(I), SELF%Z_SPLINE%DDY(I)
        ENDDO
        WRITE(*,'(A)') ' '
    END SUBROUTINE SPACE_CURVE_PRINT_INFO

	PURE REAL FUNCTION SPACE_CURVE_CHORD_LENGTH(SELF)
		IMPLICIT NONE
		! Arguments
		TYPE(SPACE_CURVE),	INTENT(IN)	::	SELF
		! Start work
		SPACE_CURVE_CHORD_LENGTH = SQRT( (SELF%X_SPLINE%Y(SELF%N)-SELF%X_SPLINE%Y(1))**2 &
		                                +(SELF%Y_SPLINE%Y(SELF%N)-SELF%Y_SPLINE%Y(1))**2 &
										+(SELF%Z_SPLINE%Y(SELF%N)-SELF%Z_SPLINE%Y(1))**2 )
	END FUNCTION SPACE_CURVE_CHORD_LENGTH

    !>  \brief  Return the derivatives in X,Y,Z from theta and psi
    PURE FUNCTION D_FROM_THETA_PSI(THETA,PSI) RESULT(D)
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   INTENT(IN)  ::  THETA   !<  Theta angle (degrees)
        REAL(KIND=8),   INTENT(IN)  ::  PSI     !<  Psi angle (degrees)
        ! Result
        REAL    ::  D(3)
        !
        D(1) = DX_FROM_THETA_PSI(THETA,PSI)
        D(2) = DY_FROM_THETA_PSI(THETA,PSI)
        D(3) = DZ_FROM_THETA(THETA)
    END FUNCTION D_FROM_THETA_PSI

    !>  \brief  Return the derivative in X from theta and psi
    ELEMENTAL PURE REAL FUNCTION DX_FROM_THETA_PSI(THETA,PSI) RESULT(DX)
        USE ROTATION_MATRICES, ONLY : RADIANS
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   INTENT(IN)  ::  THETA   !<  Theta angle (degrees)
        REAL(KIND=8),   INTENT(IN)  ::  PSI     !<  Psi angle (degrees)
        !
        DX = -SIN(RADIANS(THETA))*COS(RADIANS(PSI))
    END FUNCTION DX_FROM_THETA_PSI

    !>  \brief  Return the derivative in Y from theta and psi
    ELEMENTAL PURE REAL FUNCTION DY_FROM_THETA_PSI(THETA,PSI) RESULT(DY)
        USE ROTATION_MATRICES, ONLY : RADIANS
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   INTENT(IN)  ::  THETA   !<  Theta angle (degrees)
        REAL(KIND=8),   INTENT(IN)  ::  PSI     !<  Psi angle (degrees)
        !
        DY = SIN(RADIANS(THETA))*SIN(RADIANS(PSI))
    END FUNCTION DY_FROM_THETA_PSI

    !>  \brief  Return the derivative in Z from theta
    ELEMENTAL PURE REAL FUNCTION DZ_FROM_THETA(THETA) RESULT(DZ)
        USE ROTATION_MATRICES, ONLY : RADIANS
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   INTENT(IN)  ::  THETA   !<  Theta angle (degrees)
        !
        DZ = COS(RADIANS(THETA))
    END FUNCTION DZ_FROM_THETA

    PURE REAL FUNCTION PSI_FROM_DX_DY(DX,DY) RESULT(PSI)
        USE ROTATION_MATRICES, ONLY : DEGREES
        IMPLICIT NONE
        !
        REAL(KIND=8),   INTENT(IN)  ::  DX,DY
        !
        IF (DX .EQ. 0.0 .AND. DY .EQ. 0.0) THEN
            PSI = 0.0
        ELSE
            PSI = DEGREES(-ATAN2(DY,DX))
        ENDIF
    END FUNCTION PSI_FROM_DX_DY

    PURE REAL FUNCTION THETA_FROM_DX_DY_DZ(DX,DY,DZ,PSI) RESULT(THETA)
        USE ROTATION_MATRICES, ONLY : DEGREES, RADIANS
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),               INTENT(IN)  ::  DX,DY,DZ
        REAL(KIND=8),   OPTIONAL,   INTENT(IN)  ::  PSI         !<  If Psi is not supplied, it will be calculated internally
        ! Private variables
        !LOGICAL,   PARAMETER   ::  DEBUG   =   .FALSE.
        REAL(KIND=8)    ::  PPSI
        !
        ! If necessary, work out psi
        IF (PRESENT(PSI)) THEN
            PPSI = PSI
        ELSE
            PPSI = PSI_FROM_DX_DY(DX,DY)
        ENDIF

        ! For theta, we have the choice between two formulas. We will choose the one where we are less likely to deal with zeroes
        IF (ABS(SIN(RADIANS(PPSI))) .GT. ABS(COS(RADIANS(PPSI)))) THEN
            !IF (DEBUG) PRINT '(A)', '**debug(THETA_FROM_DX_DY_DZ_PSI): using sin(psi) to compute theta'
            THETA   =   DEGREES(ATAN2(DY,DZ*SIN(RADIANS(PPSI))))
        ELSE
            !IF (DEBUG) PRINT '(A)', '**debug(THETA_FROM_DX_DY_DZ_PSI): using cos(psi) to compute theta'
            THETA   =   DEGREES(ATAN2(-DX,DZ*COS(RADIANS(PPSI))))
        ENDIF

    END FUNCTION THETA_FROM_DX_DY_DZ


    !>  \brief  By convention, psi should always be between -90 and 90 degrees. This routine enforces this.
    PURE SUBROUTINE PSI_BTW_M90_90(THETA,PSI)
        USE EULERANGLES, ONLY : DEGREES_M180_TO_180
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   INTENT(INOUT)   ::  THETA, PSI
        ! Private variables
        !LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        ! By convention, psi should always be between -90 and 90
        IF (ABS(PSI) .GT. 90.) THEN
            !IF (DEBUG) PRINT '(A)', '**debug(PSI_BTW_M90_90): correcting theta & psi to enforce convention that -90<=psi<=90'
            PSI     = DEGREES_M180_TO_180(PSI - 180.)
            THETA   = DEGREES_M180_TO_180(-THETA)
        ENDIF
    END SUBROUTINE PSI_BTW_M90_90

    !>  \brief  Compute Euler angles corresponding to the derivatives given
    !!
    !!  See the "Spline Parametrisation" document for derivations of the equations
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE EULER_FROM_DERIVATIVES(DX,DY,DZ,THETA,PSI)
        USE EULERANGLES, ONLY : DEGREES_M180_TO_180
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   INTENT(IN)  ::  DX, DY, DZ
        REAL(KIND=8),   INTENT(OUT) ::  THETA, PSI
        ! Private variables
#ifndef SKIP_RUNTIME_CHECKS
        LOGICAL,    PARAMETER   ::  SANITY_CHECK        =   .FALSE.
        LOGICAL,    PARAMETER   ::  DEBUG               =   .FALSE.
        LOGICAL,    PARAMETER   ::  CORRECTION_ALERT    =   .FALSE.
#endif
        ! Start work

#ifndef SKIP_RUNTIME_CHECKS
        IF (DEBUG) PRINT '(A,3(F8.3,1X))', '**debug(EULER_FROM_DERIVATIVES):            dx, dy, dz = ', DX, DY, DZ
#endif

        PSI     =   PSI_FROM_DX_DY(DX,DY)
        THETA   =   THETA_FROM_DX_DY_DZ(DX,DY,DZ,PSI=PSI)

#ifndef SKIP_RUNTIME_CHECKS
        IF (DEBUG) PRINT '(A,2(F8.3,1X))', '**debug(EULER_FROM_DERIVATIVES): raw        psi, theta = ', PSI, THETA
#endif
        ! By convention, psi should always be between -90 and 90
        CALL PSI_BTW_M90_90(THETA,PSI)

        ! There are some cases where we may need to fix theta and/or psi
        IF (DZ .LT. 0.) THEN
            IF (ABS(THETA) .LT. 90.) THEN
#ifndef SKIP_RUNTIME_CHECKS
                IF (DEBUG .OR. CORRECTION_ALERT) PRINT '(A)', '**debug(EULER_FROM_DERIVATIVES): correcting theta (2)!'
#endif
                THETA = DEGREES_M180_TO_180(THETA - 180.)
            ELSE IF (ABS(THETA) .EQ. 90.) THEN
                ! We are probably here because DZ is almost 0.0, approaching from the - side
                ! and abs(theta) is rounded to 90. by the trig functions, but should really be slightly > 90.
#ifndef SKIP_RUNTIME_CHECKS
                IF (DEBUG .OR. CORRECTION_ALERT) WRITE(*,'(A)') '**debug(EULER_FROM_DERIVATIVES): got here (2*)'
#endif
                ! By looking at DX, we can decide whether Theta should be + or -90
                IF (DX .GT. 0.) THEN
                    THETA = -90.
                ELSE IF (DX .LT. 0.) THEN
                    THETA = 90.
                ENDIF
            ENDIF
        ELSE IF (DZ .GT. 0.) THEN
            IF (ABS(THETA) .GT. 90.) THEN
#ifndef SKIP_RUNTIME_CHECKS
                IF (DEBUG .OR. CORRECTION_ALERT) PRINT '(A)', '**debug(EULER_FROM_DERIVATIVES): correcting theta (3)!'
#endif
                THETA = DEGREES_M180_TO_180(THETA - 180.)
            ELSE IF (ABS(THETA) .EQ. 90.) THEN
                ! We are probably here because DZ is almost 0.0, approaching from the + side
                ! and abs(theta) is rounded to 90. by the trig functions, but should really be slightly < 90.
#ifndef SKIP_RUNTIME_CHECKS
                IF (DEBUG .OR. CORRECTION_ALERT) WRITE(*,'(A)') '**debug(EULER_FROM_DERIVATIVES): got here (3*)'
#endif
                ! By looking at DX, we can decide whether Theta should be + or -90
                IF (DX .GT. 0.) THEN
                    THETA = -90.
                ELSE IF (DX .LT. 0.) THEN
                    THETA = 90.
                ENDIF
            ENDIF
        ELSE IF (DZ .EQ. 0.) THEN
            IF (DX .GT. 0.) THEN
                ! Theta should be negative
#ifndef SKIP_RUNTIME_CHECKS
                IF (DEBUG .OR. CORRECTION_ALERT) PRINT '(A)', '**debug(EULER_FROM_DERIVATIVES): correcting theta (4)!'
#endif
                IF (THETA .GT. 0.) THETA = DEGREES_M180_TO_180(THETA - 180.)
#ifndef SKIP_RUNTIME_CHECKS
                IF (SANITY_CHECK) THEN
                    IF (THETA .GT. 0.) THEN
                        IF (DEBUG) PRINT *, '**debug(EULER_FROM_DERIVATIVES): theta = ', theta
                        CALL TERMINATE('Sanity check failed')
                    ENDIF
                ENDIF
#endif
            ELSE IF (DX .LT. 0.) THEN
                ! Theta should be positive
#ifndef SKIP_RUNTIME_CHECKS
                IF (DEBUG .OR. CORRECTION_ALERT) PRINT '(A)', '**debug(EULER_FROM_DERIVATIVES): correcting theta (5)!'
#endif
                IF (THETA .LT. 0.) THETA = DEGREES_M180_TO_180(THETA - 180.)
#ifndef SKIP_RUNTIME_CHECKS
                IF (SANITY_CHECK) THEN
                    IF (THETA .LT. 0.) THEN
                        CALL TERMINATE('Sanity check failed')
                    ENDIF
                ENDIF
#endif
            ELSE ! DX == 0.
                IF (DY .GT. 0.) THEN
                    THETA = -90.
                    PSI   = -90.
                ELSE IF (DY .LT. 0.) THEN
                    THETA = -90.
                    PSI   = 90.
                ELSE ! DY == 0
                    ! DX, DY and DZ are 0, this is not possible
#ifndef SKIP_RUNTIME_CHECKS
                    WRITE(*,'(A,3(F0.8,1X))') '**ERROR(EULER_FROM_DERIVATIVES): DX, DY, DZ = ', DX, DY, DZ
                    CALL TERMINATE('Fatal error in EULER_FROM_DERIVATIVES')
#endif
                ENDIF
            ENDIF
        ENDIF
#ifndef SKIP_RUNTIME_CHECKS
        IF (DEBUG) PRINT '(A,2(F8.3,1X))', '**debug(EULER_FROM_DERIVATIVES): corrected  psi, theta = ', PSI, THETA

        ! Sanity check
        IF (SANITY_CHECK .OR. DEBUG) THEN
            CALL CHECK_EULER_AND_DERIVATIVES(PSI,THETA,DX,DY,DZ)
        ENDIF
#endif
    END SUBROUTINE EULER_FROM_DERIVATIVES

    !>  \brief  Check whether the Euler angles and derivatives match (see notes from 8-Nov-2010 for derivations)
    !!
    SUBROUTINE CHECK_EULER_AND_DERIVATIVES(PSI,THETA,DX,DY,DZ)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   INTENT(IN)  ::  PSI, THETA, DX, DY, DZ
        ! Private variables
        REAL(KIND=8),       PARAMETER   ::  ALMOST_ZERO = 0.0001D0
        LOGICAL                         ::  THETA_OK, PSI_OK
        LOGICAL,            PARAMETER   ::  DEBUG = .FALSE.
        LOGICAL                         ::  DO_THE_CHECK
        INTEGER                         ::  LDX,LDY,LDZ
        ! Start work

        THETA_OK = .TRUE.
        PSI_OK = .TRUE.
        DO_THE_CHECK = .TRUE.

        IF (DEBUG) PRINT '(A,3(F8.3,1X))', '**debug(CHECK_EULER_AND_DERIVATIVES): dx, dy, dz = ', DX, DY, DZ

        ! In some cases, the numerical calculations are not stable and we need to skip the tests.
        ! Note that the check on orders of magnitude is probably overkill in the case where DZ is 0.0 (the formulas should
        ! still work in that case). Perhaps the special case of DZ == 0. should be considered separately.
        IF (ABS(DX) .LT. ALMOST_ZERO .OR. ABS(DY) .LT. ALMOST_ZERO .OR. ABS(DZ) .LT. ALMOST_ZERO) DO_THE_CHECK = .FALSE.
        IF (DO_THE_CHECK) THEN
            LDX = NINT(LOG10(ABS(DX)))
            LDY = NINT(LOG10(ABS(DY)))
            LDZ = NINT(LOG10(ABS(DZ)))
            IF (LDX-LDZ .GT. 6 .OR. LDY-LDZ .GT. 6) DO_THE_CHECK = .FALSE.
        ENDIF


        !
        ! If derivatives are very close to 0.0, the numerical calculation do not necessarily behave. Therefore, skip tests
        !
        IF (DO_THE_CHECK) THEN
            !
            ! Check psi and theta are >-180 and <=180
            !
            IF (PSI .LE. -180. .OR. PSI .GT. 180.) THEN
                WRITE(*,'(A,F8.3)') '**ERROR(FILAMENT_CHECK_EULER_AND_DERIVATIVES): PSI is out-of-range: ', PSI
                CALL TERMINATE('Fatal error in FILAMENT_CHECK_EULER_AND_DERIVATIVES')
            ENDIF
            IF (THETA .LE. -180. .OR. THETA .GT. 180.) THEN
                WRITE(*,'(A,F8.3)') '**ERROR(FILAMENT_CHECK_EULER_AND_DERIVATIVES): THETA is out-of-range: ', THETA
                CALL TERMINATE('Fatal error in FILAMENT_CHECK_EULER_AND_DERIVATIVES')
            ENDIF

            !
            ! Check theta
            !
            IF (DX .GT. ALMOST_ZERO) THEN
                IF (DZ .GT. 0.) THEN
                    IF (THETA .LE. -90. .OR. THETA .GE. 0.) THETA_OK = .FALSE.
                ELSE IF (DZ .LT. 0.) THEN
                    IF (THETA .LE. -180. .OR. THETA .GE. -90.) THETA_OK = .FALSE.
                ELSE    ! DZ == 0.
                    IF (ABS(THETA+90.) .GT. 0.) THETA_OK = .FALSE.
                ENDIF
            ELSE IF (DX .LT. 0.) THEN
                IF (DZ .GT. 0.) THEN
                    IF (THETA .LE. 0. .OR. THETA .GE. 90.) THETA_OK = .FALSE.
                ELSE IF (DZ .LT. 0.) THEN
                    IF (THETA .LE. 90. .OR. THETA .GE. 180.) THETA_OK = .FALSE.
                ELSE    ! DZ == 0.
                    IF (ABS(THETA-90.) .GT. 0.) THETA_OK = .FALSE.
                ENDIF
            ENDIF


            !
            ! Check Psi
            !
            IF (DX .LT. 0.) THEN
                IF (DY .GT. 0.) THEN
                    IF (PSI .LE. 0. .OR. PSI .GE. 90.) PSI_OK = .FALSE.
                ELSE IF (DY .LT. 0.) THEN
                    IF (PSI .LE. -90. .OR. PSI .GE. 0.) PSI_OK = .FALSE.
                ELSE ! DY == 0
                    IF (ABS(PSI) .GT. 0.) PSI_OK = .FALSE.
                ENDIF
            ELSE IF (DX .GT. 0.) THEN
                IF (DY .GT. 0.) THEN
                    IF (PSI .LE. -90. .OR. PSI .GE. 0.) PSI_OK = .FALSE.
                ELSE IF (DY .LT. 0.) THEN
                    IF (PSI .LE. 0. .OR. PSI .GE. 90.) PSI_OK = .FALSE.
                ELSE ! DY == 0
                    IF (ABS(PSI) .GT. 0.) PSI_OK = .FALSE.
                ENDIF
            ENDIF

            !
            ! Special cases where DX == 0.0
            !
            IF (DX .EQ. 0.) THEN
                IF (THETA .LT. 0. .OR. THETA .GT. 180.) THETA_OK = .FALSE.
                IF (DZ .GT. 0.) THEN
                     IF (DY .GT. 0.) THEN
                        IF (ABS(PSI-90.) .GT. 0.) PSI_OK = .FALSE.
                        IF (THETA .LE. 0.) THETA_OK = .FALSE.
                     ELSE IF (DY .LT. 0.) THEN
                        IF (ABS(PSI+90.) .GT. 0.) PSI_OK = .FALSE.
                        IF (THETA .LE. 0.) THETA_OK = .FALSE.
                     ELSE ! DY == 0
                        IF (PSI .NE. 0.) PSI_OK = .FALSE.
                     ENDIF
                ELSE IF (DZ .LT. 0.) THEN
                    IF (DY .GT. 0.) THEN
                        IF (ABS(PSI-90.) .GT. 0.) PSI_OK = .FALSE.
                        IF (THETA .GE. 180.) THETA_OK = .FALSE.
                    ELSE IF (DY .LT. 0.) THEN
                        IF (ABS(PSI+90.) .GT. 0.) PSI_OK = .FALSE.
                        IF (THETA .GE. 180.) THETA_OK = .FALSE.
                    ELSE ! DY == 0
                        IF (ABS(THETA-180.) .GT. 0.) THETA_OK = .FALSE.
                    ENDIF
                ELSE ! DZ == 0.
                    IF (DY .GT. 0.) THEN
                        IF (ABS(PSI-90.) .GT. 0.) PSI_OK = .FALSE.
                        IF (ABS(THETA-90.) .GT. 0.) THETA_OK = .FALSE.
                    ELSE IF (DY .LT. 0.) THEN
                        IF (ABS(PSI+90.) .GT. 0.) PSI_OK = .FALSE.
                        IF (ABS(THETA-90.) .GT. 0.) THETA_OK = .FALSE.
                    ELSE ! DY == 0
                        CALL TERMINATE('Ooops, all 3 derivatives are 0.0')
                    ENDIF
                ENDIF
            ENDIF


            IF (.NOT. THETA_OK) THEN
                WRITE(*,'(A)') '**ERROR(CHECK_EULER_AND_DERIVATIVES): theta angle does not match derivatives'
            ENDIF

            IF (.NOT. PSI_OK) THEN
                WRITE(*,'(A)') '**ERROR(CHECK_EULER_AND_DERIVATIVES): psi angle does not match derivatives'
            ENDIF

            !
            ! Do the signs of the derivatives match what we expected?
            !
            IF (.NOT. THETA_OK .OR. .NOT. PSI_OK) THEN
                WRITE(*,'(A)') '**ERROR(CHECK_EULER_AND_DERIVATIVES): derivatives do not have expected sign, given Euler angles'
                WRITE(*,'(A,F12.7)')    '**debug(CHECK_EULER_AND_DERIVATIVES): psi      = ', PSI
                WRITE(*,'(A,F12.7)')    '**debug(CHECK_EULER_AND_DERIVATIVES): theta    = ', THETA
                WRITE(*,'(A,F12.7)')    '**debug(CHECK_EULER_AND_DERIVATIVES): dx       = ', DX
                WRITE(*,'(A,F12.7)')    '**debug(CHECK_EULER_AND_DERIVATIVES): dy       = ', DY
                WRITE(*,'(A,F12.7)')    '**debug(CHECK_EULER_AND_DERIVATIVES): dz       = ', DZ
                CALL TERMINATE('Fatal error in CHECK_EULER_AND_DERIVATIVES')
            ENDIF
        ENDIF   ! End of check for derivatives very close to 0.0
    END SUBROUTINE CHECK_EULER_AND_DERIVATIVES

END MODULE DIFFERENTIAL_GEOMETRY
