!>  \brief  Methods to compute eigenvectors and values
MODULE EIGEN

    !USE EIGEN_NR
    USE FGSL

    CONTAINS

    !>  \brief  Compute the eigenvectors and eigenvalues for a 2D matrix
    SUBROUTINE MATRIX_EIGEN_ANALYSIS(SELF,EIGENVECTORS,EIGENVALUES)
        USE RUNTIME_PARAMETERS, ONLY: TERMINATE
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),                TARGET, INTENT(IN)      ::  SELF(:,:)
        REAL(KIND=8),   ALLOCATABLE, TARGET, INTENT(INOUT)   ::  EIGENVECTORS(:,:)
        REAL(KIND=8),   ALLOCATABLE, TARGET, INTENT(INOUT)   ::  EIGENVALUES(:)
        ! Private variables
        TYPE(FGSL_EIGEN_SYMMV_WORKSPACE)    ::  WS
        type(FGSL_MATRIX)                   ::  MATRIX, EVEC
        type(FGSL_VECTOR)                   ::  EVAL
        INTEGER(FGSL_INT)                   ::  STATUS
        INTEGER(fgsl_size_t)                ::  N_FGSL
        INTEGER                             ::  N
        LOGICAL,    PARAMETER               ::  DEBUG   =   .TRUE.
        ! Start work

        N = SIZE(SELF,1)
        N_FGSL = N

        ! Check allocation
        IF (.NOT. ALLOCATED(EIGENVECTORS)) ALLOCATE(EIGENVECTORS(N,N))
        IF (.NOT. ALLOCATED(EIGENVALUES))  ALLOCATE(EIGENVALUES(N))

        ! Check dimensions
        IF (     N .NE. SIZE(SELF,2)         .OR. N .NE. SIZE(EIGENVECTORS,1) &
            .OR. N .NE. SIZE(EIGENVECTORS,2) .OR. N .NE. SIZE(EIGENVALUES)) THEN
            WRITE(*,'(A,5(I0,1X))') '**ERROR(MATRIX_EIGEN_ANALYSIS): bad input dimensions: ', SIZE(SELF,1),SIZE(SELF,2), &
                                    SIZE(EIGENVECTORS,1),SIZE(EIGENVECTORS,2), SIZE(EIGENVALUES)
            CALL TERMINATE('Fatal error in MATRIX_EIGEN_ANALYSIS')
        ENDIF


        ! Copy data over to FGSL format
        MATRIX = FGSL_MATRIX_INIT(TYPE=1.0_FGSL_DOUBLE)
        EVEC   = FGSL_MATRIX_INIT(TYPE=1.0_FGSL_DOUBLE)
        EVAL   = FGSL_VECTOR_INIT(TYPE=1.0_FGSL_DOUBLE)
        STATUS = FGSL_MATRIX_ALIGN(SELF,N_FGSL,N_FGSL,N_FGSL,MATRIX)
        STATUS = FGSL_MATRIX_ALIGN(EIGENVECTORS,N_FGSL,N_FGSL,N_FGSL,EVEC)
        STATUS = FGSL_VECTOR_ALIGN(EIGENVALUES,N_FGSL,EVAL,N_FGSL,0_FGSL_SIZE_T,1_FGSL_SIZE_T)

        ! Allocate GSL workspace
        WS = FGSL_EIGEN_SYMMV_ALLOC(INT(N,FGSL_SIZE_T))

        !
        STATUS = FGSL_EIGEN_SYMMV(MATRIX, EVAL, EVEC, WS)
        STATUS = FGSL_EIGEN_SYMMV_SORT(EVAL,EVEC,FGSL_EIGEN_SORT_ABS_ASC)

        ! Copy results back
        !EIGENVECTORS = EVEC
        !EIGENVALUES = EVAL

        IF (DEBUG) THEN
            !NOTHING
        ENDIF

        ! Deallocate memory
        CALL FGSL_EIGEN_SYMMV_FREE(WS)
        CALL FGSL_MATRIX_FREE(MATRIX)
        CALL FGSL_MATRIX_FREE(EVEC)
        CALL FGSL_VECTOR_FREE(EVAL)
    END SUBROUTINE MATRIX_EIGEN_ANALYSIS



END MODULE EIGEN
