!>  \brief  Classes and methods dealing with Euler angles
!!          By default, we will use the Spider/Frealign convention
!!
!!  \todo   Add functions to convert between various conventions
MODULE EULERANGLES

    IMPLICIT NONE

    TYPE EULER
        REAL(KIND=8)    ::  PHI     =   0.0D0   !<  Phi angle (Spider/Frealign convention) - twist  (equiv. to Imagic's Gamma)
        REAL(KIND=8)    ::  THETA   =   0.0D0   !<  Theta angle (Spider/Frealign convention) - out-of-plane (equiv. to Imagic's Beta)
        REAL(KIND=8)    ::  PSI     =   0.0D0   !<  Psi angle, (Spider/Frealign convention) - in-plane (equiv. to Imagic's Alpha)
    END TYPE

    INTERFACE OPERATOR (.EQ.)
        MODULE PROCEDURE IS_EQUAL
    END INTERFACE

    INTERFACE OPERATOR (.NE.)
        MODULE PROCEDURE IS_NOT_EQUAL
    END INTERFACE

    INTERFACE OPERATOR (.LT.)
        MODULE PROCEDURE EULER_IS_LT
    END INTERFACE

    INTERFACE OPERATOR (+)
        MODULE PROCEDURE EULER_PLUS_EULER
    END INTERFACE

    INTERFACE OPERATOR (-)
        MODULE PROCEDURE EULER_MINUS_EULER
    END INTERFACE

    INTERFACE OPERATOR (*)
        MODULE PROCEDURE EULER_TIMES_EULER
    END INTERFACE

    INTERFACE OPERATOR (/)
        MODULE PROCEDURE EULER_DIVIDE_BY_REAL
    END INTERFACE

    INTERFACE SQRT
        MODULE PROCEDURE EULER_SQRT
    END INTERFACE SQRT

    INTERFACE ABS
        MODULE PROCEDURE EULER_ABS
    END INTERFACE ABS

    CONTAINS

    PURE SUBROUTINE EULER_INIT(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(INOUT)  ::  SELF
        ! Start work
        SELF%PHI   = 0.0D0
        SELF%THETA = 0.0D0
        SELF%PSI   = 0.0D0
    END SUBROUTINE EULER_INIT

    ELEMENTAL PURE FUNCTION EULER_DIVIDE_BY_REAL(ANGLES_IN,VAL) RESULT(ANGLES_OUT)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(IN)  ::  ANGLES_IN
        REAL,           INTENT(IN)  ::  VAL
        ! Result
        TYPE(EULER) ::  ANGLES_OUT
        ! Start work
        ANGLES_OUT%PHI      =   ANGLES_IN%PHI / VAL
        ANGLES_OUT%THETA    =   ANGLES_IN%THETA / VAL
        ANGLES_OUT%PSI      =   ANGLES_IN%PSI / VAL
    END FUNCTION EULER_DIVIDE_BY_REAL

    ELEMENTAL PURE FUNCTION EULER_PLUS_EULER(ANGLES1,ANGLES2)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(IN)  ::  ANGLES1, ANGLES2
        ! Result
        TYPE(EULER)     ::  EULER_PLUS_EULER
        ! Start work
        EULER_PLUS_EULER%PHI    =   ANGLES1%PHI + ANGLES2%PHI
        EULER_PLUS_EULER%THETA  =   ANGLES1%THETA + ANGLES2%THETA
        EULER_PLUS_EULER%PSI    =   ANGLES1%PSI + ANGLES2%PSI
    END FUNCTION EULER_PLUS_EULER

    ELEMENTAL PURE FUNCTION EULER_MINUS_EULER(ANGLES1,ANGLES2)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(IN)  ::  ANGLES1, ANGLES2
        ! Result
        TYPE(EULER)     ::  EULER_MINUS_EULER
        ! Start work
        EULER_MINUS_EULER%PHI   =   ANGLES1%PHI - ANGLES2%PHI
        EULER_MINUS_EULER%THETA =   ANGLES1%THETA - ANGLES2%THETA
        EULER_MINUS_EULER%PSI   =   ANGLES1%PSI - ANGLES2%PSI
    END FUNCTION EULER_MINUS_EULER

    ELEMENTAL PURE FUNCTION EULER_TIMES_EULER(ANGLES1,ANGLES2)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(IN)  ::  ANGLES1, ANGLES2
        ! Result
        TYPE(EULER)     ::  EULER_TIMES_EULER
        ! Start work
        EULER_TIMES_EULER%PHI   =   ANGLES1%PHI * ANGLES2%PHI
        EULER_TIMES_EULER%THETA =   ANGLES1%THETA * ANGLES2%THETA
        EULER_TIMES_EULER%PSI   =   ANGLES1%PSI * ANGLES2%PSI
    END FUNCTION EULER_TIMES_EULER

    ELEMENTAL PURE FUNCTION EULER_SQRT(ANGLES)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(IN)  ::  ANGLES
        ! Result
        TYPE(EULER)     ::  EULER_SQRT
        ! Start work
        EULER_SQRT%PHI      =   SQRT(ANGLES%PHI)
        EULER_SQRT%THETA    =   SQRT(ANGLES%THETA)
        EULER_SQRT%PSI      =   SQRT(ANGLES%PSI)
    END FUNCTION EULER_SQRT

    ELEMENTAL PURE FUNCTION EULER_ABS(ANGLES)
        IMPLICIT NONE
        ! Argumnets
        TYPE(EULER),    INTENT(IN)  ::  ANGLES
        ! Result
        TYPE(EULER)     ::  EULER_ABS
        ! Start work
        EULER_ABS%PHI       =   ABS(ANGLES%PHI)
        EULER_ABS%THETA     =   ABS(ANGLES%THETA)
        EULER_ABS%PSI       =   ABS(ANGLES%PSI)
    END FUNCTION EULER_ABS

    ELEMENTAL PURE LOGICAL FUNCTION IS_NOT_EQUAL(ANGLES1,ANGLES2)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(IN)  ::  ANGLES1
        TYPE(EULER),    INTENT(IN)  ::  ANGLES2
        ! Start work
        IS_NOT_EQUAL = .NOT. IS_EQUAL(ANGLES1,ANGLES2)
    END FUNCTION IS_NOT_EQUAL

    !>  \brief  Test whether two sets of angles are equal
    ELEMENTAL PURE LOGICAL FUNCTION IS_EQUAL(ANGLES1,ANGLES2)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(IN)  ::  ANGLES1
        TYPE(EULER),    INTENT(IN)  ::  ANGLES2
        ! Start work
        IS_EQUAL =  ANGLES1%PHI .EQ. ANGLES2%PHI .AND. ANGLES1%THETA .EQ. ANGLES2%THETA .AND. ANGLES1%PSI .EQ. ANGLES2%PSI
    END FUNCTION IS_EQUAL

    ELEMENTAL PURE LOGICAL FUNCTION EULER_IS_LT(ANGLES1,ANGLES2)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(IN)  ::  ANGLES1
        TYPE(EULER),    INTENT(IN)  ::  ANGLES2
        ! Start work
        EULER_IS_LT =   ANGLES1%PHI .LT. ANGLES2%PHI .AND. &
                        ANGLES1%THETA .LT. ANGLES2%THETA .AND. &
                        ANGLES1%PSI .LT. ANGLES2%PSI
    END FUNCTION EULER_IS_LT

    ELEMENTAL PURE FUNCTION DEGREES_M180_TO_180(DEGREES)
        IMPLICIT NONE
        ! Argument
        REAL(KIND=8),   INTENT(IN)  ::  DEGREES
        ! Result
        REAL(KIND=8)    ::  DEGREES_M180_TO_180
        ! Private variables
        ! Start work
        DEGREES_M180_TO_180 = MOD(DEGREES,360.D0)
        IF (DEGREES_M180_TO_180 .GT. 180.D0) THEN
            DEGREES_M180_TO_180 = DEGREES_M180_TO_180 - 360.D0
        ELSE IF (DEGREES_M180_TO_180 .LT. -180.D0) THEN
            DEGREES_M180_TO_180 = DEGREES_M180_TO_180 + 360.D0
        ELSE
            ! Nothing happens
        ENDIF
    END FUNCTION DEGREES_M180_TO_180

    !>  \brief  Returns the difference between two sets of Euler angles (angles1-angles2). The difference lies between -180 and +180
    PURE FUNCTION EULER_DIFFERENCE(ANGLES1,ANGLES2)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(IN)  ::  ANGLES1
        TYPE(EULER),    INTENT(IN)  ::  ANGLES2
        ! Result
        TYPE(EULER) ::  EULER_DIFFERENCE
        ! Private variables

        ! Start work
        EULER_DIFFERENCE%PHI    =   ANGLE_DIFFERENCE(ANGLES1%PHI,   ANGLES2%PHI)
        EULER_DIFFERENCE%THETA  =   ANGLE_DIFFERENCE(ANGLES1%THETA, ANGLES2%THETA)
        EULER_DIFFERENCE%PSI    =   ANGLE_DIFFERENCE(ANGLES1%PSI,   ANGLES2%PSI)
    END FUNCTION EULER_DIFFERENCE

    !>  \brief  Returns the difference between two angles (angle1-angle2). The difference lies between -180 and +180
    PURE ELEMENTAL FUNCTION ANGLE_DIFFERENCE(ANGLE1,ANGLE2)
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   INTENT(IN)  ::  ANGLE1  !<  Angle1 (in degrees)
        REAL(KIND=8),   INTENT(IN)  ::  ANGLE2  !<  Angle2 (in degrees)
        ! Result
        REAL(KIND=8)                ::  ANGLE_DIFFERENCE
        ! Private variables
        REAL(KIND=8)    ::  ANGLE1_TMP,ANGLE2_TMP
        ! Start work
        !PRINT '(A,2F9.2)', '**debug(ANGLE_DIFFERENCE): angle1, angle2 = ', ANGLE1, ANGLE2

        ! Bring everything back to the 0.0 - 360.0 interval
        ANGLE1_TMP = ANGLE_BTWN_0_AND_360(ANGLE1)
        ANGLE2_TMP = ANGLE_BTWN_0_AND_360(ANGLE2)

        !PRINT '(A,2F9.2)', '**debug(ANGLE_DIFFERENCE): angle1, angle2 (0-360) = ', ANGLE1_TMP, ANGLE2_TMP

        ! Calculate the difference
        ANGLE_DIFFERENCE = ANGLE1_TMP - ANGLE2_TMP

        !PRINT '(A,F7.2)', '**debug(ANGLE_DIFFERENCE): angle diff = ', ANGLE_DIFFERENCE

        ! If it would have been shorter to go over the 0/360 boundary,
        IF (ANGLE_DIFFERENCE .GT. 180.0) THEN
            ANGLE_DIFFERENCE = ANGLE_DIFFERENCE - 360.0
        ELSE IF (ANGLE_DIFFERENCE .LT. -180.0) THEN
            ANGLE_DIFFERENCE = ANGLE_DIFFERENCE + 360.0
        ENDIF

        !PRINT '(A,F7.2)', '**debug(ANGLE_DIFFERENCE): angle diff (-180,180) = ', ANGLE_DIFFERENCE
    END FUNCTION ANGLE_DIFFERENCE

    !>  \brief  Return angles between 0 and 360.0
    PURE FUNCTION EULER_BTWN_0_AND_360(ANGLES) RESULT(ANGLES_OUT)
        IMPLICIT NONE
        ! Arguments
        TYPE(EULER),    INTENT(IN)  ::  ANGLES
        ! Result
        TYPE(EULER) ::  ANGLES_OUT
        ! Private variables

        ! Start work
        ANGLES_OUT%PHI      =   ANGLE_BTWN_0_AND_360(ANGLES%PHI)
        ANGLES_OUT%THETA    =   ANGLE_BTWN_0_AND_360(ANGLES%THETA)
        ANGLES_OUT%PSI      =   ANGLE_BTWN_0_AND_360(ANGLES%PSI)
    END FUNCTION EULER_BTWN_0_AND_360

    !>  \brief  Add or substract 360 until the angle is .GE. 0.0 and .LT. 360.0
    !!  \todo   Old routine - double-check if correct
    PURE FUNCTION ANGLE_BTWN_0_AND_360(ANGLE)
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   INTENT(IN)  ::  ANGLE
        ! Result
        REAL(KIND=8)                ::  ANGLE_BTWN_0_AND_360
        ! Private variables

        ! Start work
        IF (ANGLE .LT. 0.0) THEN
            ANGLE_BTWN_0_AND_360 = (1-INT(ANGLE/360.0))*360.0 + ANGLE
        ELSE IF (ANGLE .GE. 360.0) THEN
            ANGLE_BTWN_0_AND_360 = ANGLE - (INT(ANGLE/360.0))*360.0
        ELSE
            ANGLE_BTWN_0_AND_360 = ANGLE
        ENDIF
    END FUNCTION ANGLE_BTWN_0_AND_360

END MODULE EULERANGLES
