!>  \brief  Classes and methods related to filament objects
!!
!!  For details on spline 3D modelling of filaments, parametrisation and related topics, see the documentation.
!!
MODULE FILAMENTS

    USE FILAMENTS_BASE

    IMPLICIT NONE


    CONTAINS

    !>  \brief  Get a set of values from a filament's waypoints.
    !!
    !!  The convention for the order of the values is that they are ordered first by parameter index
    !!  (1-3: x-z, 4-6: phi,theta,psi, 7: subunit_num) and then by waypoint #.
    SUBROUTINE VALUES_FROM_FILAMENT(SELF,VALUES,PARAMS_MASK,PARAMS_MASK_HOLD_AVERAGE,INCREMENTAL_VALUES)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_PARAMS_ABSOLUTE_TO_INCREMENTAL, FILAMENT_PARAMS_INCREMENTAL_TO_ABSOLUTE
        USE FILAMENTS_NORMAL_MODES, ONLY : FILAMENT_WPS_TO_NM_AMPLITUDES
        USE COOS, ONLY : VALUE_FROM_WAYPOINT
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF
        REAL(KIND=8),       ALLOCATABLE,    INTENT(INOUT)   ::  VALUES(:)
        LOGICAL,                            INTENT(IN)      ::  PARAMS_MASK(7)  !<  For each of the 7 parameters, whether they are included in the values array
        LOGICAL,            OPTIONAL,       INTENT(IN)      ::  PARAMS_MASK_HOLD_AVERAGE(7) !<  For each of the 7 parameters whether the average value should be left alone
        LOGICAL,            OPTIONAL,       INTENT(IN)      ::  INCREMENTAL_VALUES(7)   !<  For each parameters, whether incremental (relative to previous waypoint) rather than absolute values should be returned
        ! Private variables
        INTEGER ::  NUM_EXPECTED_VALUES
        INTEGER ::  COUNT_DIM
        INTEGER ::  CURRENT_PARAM
        INTEGER ::  CURRENT_WP
        LOGICAL ::  PPARAMS_MASK_HOLD_AVERAGE(7)
        LOGICAL ::  IINCREMENTAL_VALUES(7)
        ! Start work

        ! Check whether we'll need to hold any parameter's average constant
        IF (PRESENT(PARAMS_MASK_HOLD_AVERAGE)) THEN
            PPARAMS_MASK_HOLD_AVERAGE = PARAMS_MASK_HOLD_AVERAGE
        ELSE
            PPARAMS_MASK_HOLD_AVERAGE = .FALSE.
        ENDIF

        ! Check INCREMENTAL_VALUES flags
        IINCREMENTAL_VALUES = .FALSE.
        IF (PRESENT(INCREMENTAL_VALUES)) IINCREMENTAL_VALUES = INCREMENTAL_VALUES

        ! Allocate array
        NUM_EXPECTED_VALUES = FILAMENT_EXPECTED_NUMBER_OF_VALUES(SELF,PARAMS_MASK,PARAMS_MASK_HOLD_AVERAGE)
        IF (ALLOCATED(VALUES)) DEALLOCATE(VALUES)
        ALLOCATE(VALUES(NUM_EXPECTED_VALUES))


        ! If we need to get incremental values, let's do the conversion now
        IF (ANY(IINCREMENTAL_VALUES)) THEN
            CALL FILAMENT_PARAMS_ABSOLUTE_TO_INCREMENTAL(SELF,IINCREMENTAL_VALUES)
        ENDIF


        ! If we need to give the NM amplitudes, we need to work them out
        IF (ANY(SELF%REFINE_NORMAL_MODES)) THEN
            CALL FILAMENT_WPS_TO_NM_AMPLITUDES(SELF)
        ENDIF

        COUNT_DIM = 0
        ! Loop over the waypoints
        DO CURRENT_WP=1,SELF%NWP
            ! Loop over the parameters
            DO CURRENT_PARAM=1,7
                ! Is this parameter being refined?
                IF (PARAMS_MASK(CURRENT_PARAM)) THEN
                    ! Is this a per-wp parameter, or a rigid parameter that applies to the filament as a whole
                    IF (SELF%RIGID_PARAMS(CURRENT_PARAM)) THEN
                        IF (CURRENT_WP .EQ. 1) THEN
                            ! This is a parameter that has a single value for the whole filament
                            COUNT_DIM = COUNT_DIM + 1
                            IF (SELF%REFINE_NORMAL_MODES(CURRENT_PARAM)) THEN
                                CALL TERMINATE('Normal mode refinement not implemented with rigid parameters')
                            ELSE
                                CALL VALUE_FROM_WAYPOINT(SELF%RIGID_WP,VALUES(COUNT_DIM),CURRENT_PARAM)
                            ENDIF
                        ENDIF
                    ELSE
                        ! This is a per-wp parameter; is it being refined?
                        IF (SELF%WP(CURRENT_WP)%REFINE) THEN
                            IF (SELF%REFINE_NORMAL_MODES(CURRENT_PARAM) .AND. CURRENT_WP .EQ. 1 &
                                .AND. PPARAMS_MASK_HOLD_AVERAGE(CURRENT_PARAM)) THEN
                                CYCLE
                            ENDIF
                            COUNT_DIM = COUNT_DIM + 1
                            IF (SELF%REFINE_NORMAL_MODES(CURRENT_PARAM)) THEN
                                CALL VALUE_FROM_WAYPOINT(SELF%WP_NM(CURRENT_WP),VALUES(COUNT_DIM),CURRENT_PARAM)
                            ELSE
                                CALL VALUE_FROM_WAYPOINT(SELF%WP(CURRENT_WP),VALUES(COUNT_DIM),CURRENT_PARAM)
                            ENDIF
                        ENDIF
                    ENDIF
                ENDIF
            ENDDO
        ENDDO

        ! If we were getting incremental values, let's do the conversion back to the normal (absolute) representation
        IF (ANY(IINCREMENTAL_VALUES)) THEN
            CALL FILAMENT_PARAMS_INCREMENTAL_TO_ABSOLUTE(SELF,IINCREMENTAL_VALUES)
        ENDIF

        ! Sanity check
        IF (COUNT_DIM .NE. NUM_EXPECTED_VALUES) THEN
            WRITE(*,'(2(A,I0))') '**ERROR(VALUES_FROM_FILAMENT): was expecting ', NUM_EXPECTED_VALUES, ' but got ', COUNT_DIM
            CALL TERMINATE('Sanity check failed')
        ENDIF
    END SUBROUTINE VALUES_FROM_FILAMENT


    !>  \brief  Copy a set of values to a filament's waypoints.
    !!
    !!  The convention for the order of the values is that they are ordered first by parameter index
    !!  (1-3: x-z, 4-6: phi,theta,psi, 7: subunit_num) and then by waypoint #.
    SUBROUTINE VALUES_TO_FILAMENT(SELF,VALUES,PARAMS_MASK,PARAMS_MASK_HOLD_AVERAGE,INCREMENTAL_VALUES)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE COOS, ONLY : VALUE_TO_WAYPOINT
        USE FILAMENTS_NORMAL_MODES, ONLY : FILAMENT_NM_AMPLITUDES_TO_WPS
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_PARAMS_INCREMENTAL_TO_ABSOLUTE, FILAMENT_PARAMS_ABSOLUTE_TO_INCREMENTAL
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF
        REAL(KIND=8),       ALLOCATABLE,    INTENT(IN)      ::  VALUES(:)
        LOGICAL,                            INTENT(IN)      ::  PARAMS_MASK(7)  !<  For each of the 7 parameters, whether they are included in the values array
        LOGICAL,            OPTIONAL,       INTENT(IN)      ::  PARAMS_MASK_HOLD_AVERAGE(7) !<  For each of the 7 parameters whether the average value should be left alone
        LOGICAL,            OPTIONAL,       INTENT(IN)      ::  INCREMENTAL_VALUES(7)   !<  For each parameter, whether incremental (relative to previous waypoint) rather than absolute values are given
        ! Private variables
        INTEGER ::  NUM_EXPECTED_VALUES
        INTEGER ::  COUNT_DIM
        INTEGER ::  CURRENT_PARAM
        INTEGER ::  CURRENT_WP
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        LOGICAL ::  PPARAMS_MASK_HOLD_AVERAGE(7)
        LOGICAL ::  IINCREMENTAL_VALUES(7)
        ! Start work

        ! Check whether we'll need to hold any parameter's average constant
        IF (PRESENT(PARAMS_MASK_HOLD_AVERAGE)) THEN
            PPARAMS_MASK_HOLD_AVERAGE = PARAMS_MASK_HOLD_AVERAGE
        ELSE
            PPARAMS_MASK_HOLD_AVERAGE = .FALSE.
        ENDIF

        ! Check INCREMENTAL_VALUES flags
        IINCREMENTAL_VALUES = .FALSE.
        IF (PRESENT(INCREMENTAL_VALUES)) IINCREMENTAL_VALUES = INCREMENTAL_VALUES

        ! I don't think I can support incremental values and normal modes simultaneously.
        ! Or if I can, I need to double check the sequence of events after the values have
        ! been plugged in to the filament (see below)
        IF (ANY(SELF%REFINE_NORMAL_MODES) .AND. ANY(IINCREMENTAL_VALUES)) THEN
            WRITE(*,'(A)') '**ERROR(VALUES_TO_FILAMENT): incremental values are not supported in conjunction with normal modes'
            CALL TERMINATE('Fatal error in VALUES_TO_FILAMENT')
        ENDIF

        ! Holding a parameter's average value constant is only feasible when refining normal modes
        IF (ANY(PPARAMS_MASK_HOLD_AVERAGE .AND. PARAMS_MASK .AND. .NOT. SELF%REFINE_NORMAL_MODES)) THEN
            WRITE(*,'(A)') '**ERROR(VALUES_TO_FILAMENT): cannot hold the average value of a parameter constant '//&
                                                        'unless normal modes are being refined'
            CALL TERMINATE('Fatal error in VALUES_TO_FILAMENT')
        ENDIF

        ! Check we got the expected number of values
        NUM_EXPECTED_VALUES = FILAMENT_EXPECTED_NUMBER_OF_VALUES(SELF,PARAMS_MASK,PARAMS_MASK_HOLD_AVERAGE)
        IF (SIZE(VALUES) .NE. NUM_EXPECTED_VALUES) THEN
            WRITE(*,'(2(A,I0),A)')    '**ERROR(VALUES_TO_FILAMENT): was expecting ', NUM_EXPECTED_VALUES, &
                                    ' but got ', SIZE(VALUES), ' instead.'
            CALL TERMINATE('Fatal error in VALUES_TO_FILAMENT')
        ENDIF

        ! Convert to incremental if necessary
        IF (ANY(IINCREMENTAL_VALUES)) THEN
            CALL FILAMENT_PARAMS_ABSOLUTE_TO_INCREMENTAL(SELF,IINCREMENTAL_VALUES)
        ENDIF

        COUNT_DIM = 0
        ! Loop over the waypoints
        DO CURRENT_WP=1,SELF%NWP
            ! Loop over the parameters
            DO CURRENT_PARAM=1,7
                ! Is this parameter being refined?
                IF (PARAMS_MASK(CURRENT_PARAM)) THEN
                    ! Is this a per-wp parameter, or a rigid parameter that applies to the filament as a whole
                    IF (SELF%RIGID_PARAMS(CURRENT_PARAM)) THEN
                        IF (CURRENT_WP .EQ. 1) THEN
                            ! This is a parameter that has a single value for the whole filament
                            COUNT_DIM = COUNT_DIM + 1
                            IF (SELF%REFINE_NORMAL_MODES(CURRENT_PARAM)) THEN
                                CALL TERMINATE('Normal mode refinement not implemented with rigid parameters')
                            ELSE
                                CALL VALUE_TO_WAYPOINT(SELF%RIGID_WP,VALUES(COUNT_DIM),CURRENT_PARAM)
                            ENDIF
                        ENDIF
                    ELSE
                        ! This is a per-wp parameter; is it being refined?
                        IF (SELF%WP(CURRENT_WP)%REFINE) THEN
                            IF (SELF%REFINE_NORMAL_MODES(CURRENT_PARAM) .AND. CURRENT_WP .EQ. 1 &
                                .AND. PPARAMS_MASK_HOLD_AVERAGE(CURRENT_PARAM)) THEN
                                CYCLE
                            ENDIF
                            COUNT_DIM = COUNT_DIM + 1
                            IF (SELF%REFINE_NORMAL_MODES(CURRENT_PARAM)) THEN
                                CALL VALUE_TO_WAYPOINT(SELF%WP_NM(CURRENT_WP),VALUES(COUNT_DIM),CURRENT_PARAM)
                            ELSE
                                CALL VALUE_TO_WAYPOINT(SELF%WP(CURRENT_WP),VALUES(COUNT_DIM),CURRENT_PARAM)
                            ENDIF
                        ENDIF
                    ENDIF
                ENDIF
            ENDDO
        ENDDO

        ! If we were dealing with incremental values, we can now convert to the normal representation
        !!
        !! Should this happen before or after NUM_AMPLITUDES_TO_WPS ?
        IF (ANY(IINCREMENTAL_VALUES)) THEN
            CALL FILAMENT_PARAMS_INCREMENTAL_TO_ABSOLUTE(SELF,IINCREMENTAL_VALUES)
        ENDIF

        ! Now that we've transferred values to the WP_NMs, we need to update the WPs
        IF (ANY(SELF%REFINE_NORMAL_MODES)) THEN
            !IF (DEBUG) THEN
            !    WRITE(*,'(A)') '**debug(VALUES_TO_FILAMENT): filament info before NM_AMPLITUDES_TO_WPS'
            !    CALL FILAMENT_PRINT_INFO(SELF)
            !ENDIF
            CALL FILAMENT_NM_AMPLITUDES_TO_WPS(SELF)
            !IF (DEBUG) THEN
            !    WRITE(*,'(A)') '**debug(VALUES_TO_FILAMENT): filament info after NM_AMPLITUDES_TO_WPS'
            !    CALL FILAMENT_PRINT_INFO(SELF)
            !ENDIF
        ENDIF

        ! Sanity check
        IF (COUNT_DIM .NE. NUM_EXPECTED_VALUES) THEN
            WRITE(*,'(2(A,I0))') '**ERROR(VALUES_TO_FILAMENT): counted ', COUNT_DIM, ' dimensions, but was expecting ', &
                                NUM_EXPECTED_VALUES
            CALL TERMINATE('Sanity check failed')
        ENDIF
    END SUBROUTINE VALUES_TO_FILAMENT


    !>  \brief  Given a (filament) waypoint, create a single particle object
    SUBROUTINE WP_TO_SP(WP,SP,CTF,SP_IMAGE,SEGMENT_LENGTH,FILM_IMAGE,PIXEL_SIZE,APPLY_SHIFT_TO_IMAGE, &
                        APPLY_CTF_TO_IMAGE,CTF_FLIP_ONLY,IGNORE_SUBUNIT_NUM,&
                        ROTATE_SP_IMAGE,UNBEND,FIL,APPLY_RECT_MASK,&
                        TOO_CLOSE_TO_EDGE,SQUARE_IMAGE)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, PSF_RADIUS_PIXELS, IS_ODD
        use Images
        USE IMAGES_ARITHM, ONLY  : IMAGE_FLOAT_NORM_MASK
        USE CTFTILTS
        USE IMAGES_CTF, ONLY : IMAGE_APPLY_CTF
        USE SINGLE_PARTICLES
        USE ROTATION_MATRICES, ONLY : ROTMAT
        USE INTERPOLATION, ONLY : DEFAULT_INTERPOL_SCHEME, INTERPOLATION_SCHEME, INTERPOLATION_SCHEME_INIT
        IMPLICIT NONE
        ! Arguments
        TYPE(WAYPOINT),                     INTENT(IN)      ::  WP
        TYPE(SINGLE_PARTICLE),              INTENT(OUT)     ::  SP
        TYPE(CTF_PARAMETERS),               INTENT(INOUT)   ::  CTF                     !<  CTF parameters (the actual defocus value will be computed from the waypoint's Z coordinate)
        TYPE(IMAGE),            OPTIONAL,   INTENT(INOUT)   ::  SP_IMAGE                !<  Image of the single particle extracted from the supplied film image
        INTEGER,                OPTIONAL,   INTENT(IN)      ::  SEGMENT_LENGTH          !<  Length of filament segment to cut out, in pixels
        TYPE(IMAGE),            OPTIONAL,   INTENT(INOUT)   ::  FILM_IMAGE              !<  Film image from which to box out the SP image
        REAL,                   OPTIONAL,   INTENT(IN)      ::  PIXEL_SIZE              !<  Pixel size in Angstroms
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  APPLY_SHIFT_TO_IMAGE    !<  Apply non-integer shift to image
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  APPLY_CTF_TO_IMAGE      !<  Apply local CTF to the image before masking
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  CTF_FLIP_ONLY           !<  Multiply by sign(CTF) rather than the CTF itself
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  IGNORE_SUBUNIT_NUM      !<  Ignore the WP's subunit_num value
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  ROTATE_SP_IMAGE         !<  Rotate the SP image to align so that psi is zeroed
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  UNBEND                  !<  Unbend the image
        TYPE(FILAMENT),         OPTIONAL,   INTENT(INOUT)   ::  FIL                     !<  Filament the waypoint belongs to. Needed only when doing unbending.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  APPLY_RECT_MASK         !<  Apply a soft-edge rectangular mask
        LOGICAL,                OPTIONAL,   INTENT(OUT)     ::  TOO_CLOSE_TO_EDGE       !<  Indicate whether single particle is too close to the edge of the micrograph for processing
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  SQUARE_IMAGE            !<  Ensure the SP_IMAGE is square
        ! Private variables
        INTEGER                 ::  ICOO(3)
        REAL(KIND=8)            ::  RCOO(3)
        TYPE(CTF_PARAMETERS)    ::  CTF_LOCAL
        LOGICAL                 ::  AAPPLY_SHIFT_TO_IMAGE, AAPPLY_CTF_TO_IMAGE, IIGNORE_SUBUNIT_NUM
        LOGICAL                 ::  AAPPLY_RECT_MASK
        LOGICAL                 ::  RROTATE_SP_IMAGE, UUNBEND
        REAL(KIND=8)            ::  TMP_SUBUNIT_NUM
        REAL                    ::  ROT_TMP, ROTATION_MAT(4,4)
        TYPE(IMAGE)             ::  SP_IMAGE_ROT
        INTEGER                 ::  SP_IMAGE_DIM(2), SP_IMAGE_DIM_LARGE(2)
        REAL                    ::  AZIMUTH
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        REAL                    ::  RECT_MASK_DIM(2)
        LOGICAL                 ::  TTOO_CLOSE_TO_EDGE, SSQUARE_IMAGE
        TYPE(INTERPOLATION_SCHEME)  ::  INTERPOL
        TYPE(WINDOW)                ::  INTERPOL_WINDOW
        ! Start work

        TTOO_CLOSE_TO_EDGE = .FALSE.

        ! If we want an output sp image, we'll need the
        ! dimension and the film image
        IF (PRESENT(SP_IMAGE)) THEN
            IF (.NOT. (PRESENT(FILM_IMAGE) .AND. PRESENT(SEGMENT_LENGTH))) THEN
                WRITE(*,'(A)') '**ERROR(WP_TO_SP): need film image and segment length as arguments to be able to output SP image'
                CALL TERMINATE('Fatal error in WP_TO_SP')
            ENDIF
        ENDIF

        !
        AAPPLY_SHIFT_TO_IMAGE = .FALSE.
        IF (PRESENT(APPLY_SHIFT_TO_IMAGE))  AAPPLY_SHIFT_TO_IMAGE = APPLY_SHIFT_TO_IMAGE

        !
        AAPPLY_CTF_TO_IMAGE = .FALSE.
        IF (PRESENT(APPLY_CTF_TO_IMAGE)) AAPPLY_CTF_TO_IMAGE = APPLY_CTF_TO_IMAGE

        !
        AAPPLY_RECT_MASK = .TRUE.
        IF (PRESENT(APPLY_RECT_MASK)) AAPPLY_RECT_MASK = APPLY_RECT_MASK

        !
        IIGNORE_SUBUNIT_NUM = .FALSE.
        IF (PRESENT(IGNORE_SUBUNIT_NUM)) IIGNORE_SUBUNIT_NUM = IGNORE_SUBUNIT_NUM

        !
        RROTATE_SP_IMAGE = .FALSE.
        IF (PRESENT(ROTATE_SP_IMAGE))   RROTATE_SP_IMAGE = ROTATE_SP_IMAGE

        !
        SSQUARE_IMAGE = .FALSE.
        IF (PRESENT(SQUARE_IMAGE)) SSQUARE_IMAGE = SQUARE_IMAGE

        ! Unbending option
        UUNBEND = .FALSE.
        IF (PRESENT(UNBEND)) UUNBEND = UNBEND
        IF (UUNBEND .AND. RROTATE_SP_IMAGE) THEN
            WRITE(*,'(A)') '**info(WP_TO_SP): ROTATE_SP_IMAGE flag is redundant when unbending'
            RROTATE_SP_IMAGE = .FALSE.
        ENDIF
        IF (UUNBEND .AND. .NOT. PRESENT(FIL)) THEN
            WRITE(*,'(A)') '**ERROR(WP_TO_SP): unbending requires that you supply the filament object as an argument'
            CALL TERMINATE('Fatal error in WP_TO_SP')
        ENDIF
        IF (UUNBEND .AND. AAPPLY_CTF_TO_IMAGE) THEN
            ! For now, don't have to implement CTF correction before unbending happens
            WRITE(*,'(A)') '**warning(WP_TO_SP): CTF will be applied after unbending, which is not accurate'
        ENDIF

        ! Work out output image dimensions
        IF (PRESENT(SP_IMAGE)) THEN
            IF (AAPPLY_CTF_TO_IMAGE .AND. PSF_RADIUS_PIXELS .EQ. 0.0) THEN
                WRITE(*,'(A)') '**ERROR(WP_TO_SP): PSF_RADIUS_PIXELS has not been set. Needed to compute box size.'
                CALL TERMINATE('Fatal error in WPS_TO_SP')
            ENDIF
            IF (.NOT. PRESENT(FIL)) THEN
                WRITE(*,'(A)') '**ERROR(WP_TO_SP): need filament object to be given as argument, because need to know max radius'
                CALL TERMINATE('Fatal error in WP_TO_SP')
            ENDIF
            SP_IMAGE_DIM(1) = FILAMENT_SUGGESTED_BOX_WIDTH(FIL)
            SP_IMAGE_DIM(2) = SEGMENT_LENGTH
            IF (IS_ODD(SP_IMAGE_DIM(1))) SP_IMAGE_DIM(1) = SP_IMAGE_DIM(1) + 1
            IF (IS_ODD(SP_IMAGE_DIM(2))) SP_IMAGE_DIM(2) = SP_IMAGE_DIM(2) + 1
            IF (SSQUARE_IMAGE) SP_IMAGE_DIM(1:2) = MAX(SP_IMAGE_DIM(1),SP_IMAGE_DIM(2))
            SP_IMAGE_DIM_LARGE(1:2) = MAX(SEGMENT_LENGTH,FILAMENT_SUGGESTED_BOX_WIDTH(FIL) + CEILING(2.*PSF_RADIUS_PIXELS))
            IF (IS_ODD(SP_IMAGE_DIM_LARGE(1))) SP_IMAGE_DIM_LARGE(1) = SP_IMAGE_DIM_LARGE(1) + 1
            IF (IS_ODD(SP_IMAGE_DIM_LARGE(2))) SP_IMAGE_DIM_LARGE(2) = SP_IMAGE_DIM_LARGE(2) + 1
            IF (SSQUARE_IMAGE) SP_IMAGE_DIM_LARGE(1:2) = MAX(SP_IMAGE_DIM_LARGE(1),SP_IMAGE_DIM_LARGE(2))
            IF (DEBUG) THEN
                WRITE(*,'(A,2(I0,1X))') '**debug(WP_TO_SP): image dim       = ', SP_IMAGE_DIM
                IF (AAPPLY_CTF_TO_IMAGE) WRITE(*,'(A,2(I0,1X))') '**debug(WP_TO_SP): image dim large = ', SP_IMAGE_DIM_LARGE
            ENDIF
        ENDIF

        ! Work out rectangular mask dimensions
        IF (AAPPLY_RECT_MASK) THEN
            RECT_MASK_DIM = (/ SEGMENT_LENGTH-20., 2.*FIL%MAX_RADIUS+10./)
            IF (DEBUG) THEN
                WRITE(*,'(A,I0)') '**debug(WP_TO_SP): segment length = ', SEGMENT_LENGTH
                WRITE(*,'(A,F0.3)') '**debug(WP_TO_SP): fil max rad = ', FIL%MAX_RADIUS
                WRITE(*,'(A,2(F0.3,1X))') '**debug(WP_TO_SP): rect_mask_dim = ', RECT_MASK_DIM
            ENDIF
        ENDIF

        ! Will we be too close to the edge of the micrograph to avoid artefacts?
        ! If we're doing CTF correction, we need to be at least SP_IMAGE_DIM_LARGE / 2 away from edge of micrograph
        ! If we're doing rectangular mask, we need to be at least RECT_MASK_DIM(2)/2 away from the edge
        IF (        ANY(WP%COO(1:2) .LT. (                       REAL(SP_IMAGE_DIM_LARGE(1:2)) / 2. + 1.5))   &
            .OR.    ANY(WP%COO(1:2) .GT. (FILM_IMAGE%logical_dimensions(1:2) - (REAL(SP_IMAGE_DIM_LARGE(1:2)) / 2. + 1.5))) ) THEN
            IF (DEBUG) THEN
                WRITE(*,'(A,2(F0.3,1X))') '**debug(WP_TO_SP): too close to edge for CTF: ', WP%COO(1:2)
            ENDIF
            TTOO_CLOSE_TO_EDGE = .TRUE.
        ENDIF
        IF (AAPPLY_RECT_MASK) THEN
            IF (        ANY(WP%COO(1:2) .LT. (                       RECT_MASK_DIM(2) / 2. + 1.5))   &
                .OR.    ANY(WP%COO(1:2) .GT. (FILM_IMAGE%logical_dimensions(1:2) - (RECT_MASK_DIM(2) / 2. + 1.5))) ) THEN
                IF (DEBUG) THEN
                    WRITE(*,'(A,2(F0.3,1X))') '**debug(WP_TO_SP): too close to edge for CTF: ', WP%COO(1:2)
                ENDIF
                TTOO_CLOSE_TO_EDGE = .TRUE.
            ENDIF
        ENDIF

        ! Work out the image coordinates and non-integer shifts
        ICOO = NINT(WP%COO)
        ICOO(3) = 1
        RCOO = WP%COO - ICOO

        ! Cut out the image
        IF (UUNBEND) THEN
            CALL FILAMENT_UNBEND_SEGMENT(   FIL,WP%PATH_LENGTH,    &
                                            SEGMENT_LENGTH=SP_IMAGE_DIM(2), &
                                            SEGMENT_WIDTH=SP_IMAGE_DIM(1), &
                                            MICROGRAPH=FILM_IMAGE, &
                                            UNBENT_SEGMENT=SP_IMAGE)
        ELSE
            IF (IS_ALLOCATED(SP_IMAGE)) CALL IMAGE_DEALLOCATE(SP_IMAGE)
            SP_IMAGE%logical_dimensions = (/SP_IMAGE_DIM_LARGE(1),SP_IMAGE_DIM_LARGE(2),1/)
            CALL WINDOW_IMG(FILM_IMAGE,SP_IMAGE,ICOO)
        ENDIF

        ! Find the CTF for this waypoint
        CTF_LOCAL = CTF
        IF (.NOT. CTF_LOCAL%PIX_UNITS) THEN
            IF (.NOT. PRESENT(PIXEL_SIZE)) THEN
                WRITE(*,'(A)') '**ERROR(WP_TO_SP): CTF is not in pixel units'
                CALL TERMINATE('Fatal error in WP_TO_SP')
            ELSE
                CALL CTF_ANG_TO_PIX(CTF_LOCAL,PIXEL_SIZE)
            ENDIF
        ENDIF
        CALL CTF_AT_Z(CTF_LOCAL,WP%COO(3))

        ! Apply the CTF
        IF (AAPPLY_CTF_TO_IMAGE) THEN
            IF (IS_REAL(SP_IMAGE)) CALL FWD_FT(SP_IMAGE)
            CALL IMAGE_APPLY_CTF(SP_IMAGE,CTF_LOCAL,FLIP_ONLY=CTF_FLIP_ONLY)
        ENDIF

        ! Apply shift to image
        IF (AAPPLY_SHIFT_TO_IMAGE) THEN
            IF (IS_REAL(SP_IMAGE)) CALL FWD_FT(SP_IMAGE)
            CALL PSHIFT_IMAGE(SP_IMAGE,SHIFT_X=-REAL(RCOO(1)),SHIFT_Y=-REAL(RCOO(2)),SHIFT_Z=0.0)
            RCOO = 0.0
        ENDIF

        ! Rotate the image
        IF (RROTATE_SP_IMAGE) THEN
            ! Initialise interpolation
            INTERPOL = DEFAULT_INTERPOL_SCHEME()
            CALL INTERPOL_WINDOW%AllocateWindow(INTERPOL%WINDOW_WIDTH,COMP=.FALSE.,THREED=.FALSE.)
            IF (IS_COMP(SP_IMAGE)) CALL BWD_FT(SP_IMAGE)
            ROT_TMP = WP%EULER%PSI - 90. !! I think I needed -90 to align with the Y axis rather than the X axis. ... are my angles off? what about astig?
            IF (WP%EULER%THETA .GT. 180.) THEN
                ROT_TMP = ROT_TMP + 180.    ! Some filaments have OOP near 270 (rather than 90) because they have reverse polarity
            ENDIF
            ROTATION_MAT = ROTMAT(-ROT_TMP,0.0,0.0)
            CALL ROTATE_IMAGE(SP_IMAGE,SP_IMAGE_ROT,ROTATION_MAT,INTERPOL,INTERPOL_WINDOW)
            SP_IMAGE = SP_IMAGE_ROT
        ENDIF

        ! Let's crop the image now that we're done with CTF
        !IF (AAPPLY_CTF_TO_IMAGE) THEN
            IF (IS_COMP(SP_IMAGE)) CALL BWD_FT(SP_IMAGE)
            CALL CROP(SP_IMAGE,(/SP_IMAGE_DIM(1),SP_IMAGE_DIM(2),1/))
        !ENDIF

        ! Normalise the image
        IF (IS_COMP(SP_IMAGE)) CALL BWD_FT(SP_IMAGE)
        CALL IMAGE_FLOAT_NORM_MASK(SP_IMAGE,MASK_RADIUS=REAL(MAX(SP_IMAGE%logical_dimensions(1),SP_IMAGE%logical_dimensions(2))))

        ! We'll apply a simple soft rectangular mask
        IF (AAPPLY_RECT_MASK) THEN
            AZIMUTH = WP%EULER%PSI
            IF (UUNBEND .OR. RROTATE_SP_IMAGE) AZIMUTH=90.
            IF (.NOT. PRESENT(FIL)) CALL TERMINATE('Need filament object as argument')

            CALL COSINE_RECT(   SP_IMAGE,    &
                                RECT_DIM=RECT_MASK_DIM,  &
                                AZIMUTH=AZIMUTH,FALLOFF=5.0,WIPE_INPUT=.FALSE.,MASK_WITH_AVERAGE=.TRUE.)
        ENDIF

        ! Initialise the single particle
        TMP_SUBUNIT_NUM = WP%SUBUNIT_NUM-NINT(WP%SUBUNIT_NUM)
        IF (DEBUG) THEN
            WRITE(*,'(A,F0.6)') '**debug(WP_TO_SP): original subunit_num = ', WP%SUBUNIT_NUM
            WRITE(*,'(A,F0.6)') '**debug(WP_TO_SP): sp       subunit_num = ', TMP_SUBUNIT_NUM
        ENDIF
        IF (IIGNORE_SUBUNIT_NUM) TMP_SUBUNIT_NUM = 0.0
        CALL SP_INIT(   SP,CTF_LOCAL,ANGLES=WP%EULER,SHIFT_X=RCOO(1),SHIFT_Y=RCOO(2),    &
                        SUBUNIT_NUM=TMP_SUBUNIT_NUM,   &
                        HELIX=WP%HELIX)

        ! Were we too close to the edge?
        IF (PRESENT(TOO_CLOSE_TO_EDGE)) TOO_CLOSE_TO_EDGE = TTOO_CLOSE_TO_EDGE
    END SUBROUTINE WP_TO_SP




    !>  \brief  Calculate an unbent image of a segment of the filament centered around the specified point along its path
    SUBROUTINE FILAMENT_UNBEND_SEGMENT(SELF,CENTER_T,SEGMENT_LENGTH,SEGMENT_WIDTH,MICROGRAPH,UNBENT_SEGMENT,INTERPOL)
        USE FILAMENTS_PATH
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL_D
        use UsefulFunctions, only : IsOdd
        use Images
        use INTERPOLATION, only : INTERPOLATION_SCHEME, INTERPOLATION_SCHEME_INIT
        use UsefulFunctions, only : IsOdd
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                         INTENT(INOUT)   ::  SELF            !<  Filament object
        REAL(KIND=8),                           INTENT(IN)      ::  CENTER_T        !<  The parametrisation of the point along the filament which will be the center of the output unbent image
        INTEGER,                                INTENT(IN)      ::  SEGMENT_LENGTH  !<  Segment length, in pixels
        INTEGER,                                INTENT(IN)      ::  SEGMENT_WIDTH   !<  Segment width, in pixels
        TYPE(IMAGE),                            INTENT(IN)      ::  MICROGRAPH      !<  Film image
        TYPE(IMAGE),                            INTENT(INOUT)   ::  UNBENT_SEGMENT  !<  Output, unbent image
        TYPE(INTERPOLATION_SCHEME), OPTIONAL,   INTENT(IN)      ::  INTERPOL        !<  Interpolation scheme
        ! Private variables
        INTEGER             ::  OUTPUT_BOX_DIM(2)
        INTEGER             ::  I,J
        INTEGER             ::  ADDR_ORIGIN(3), MICROGRAPH_ORIGIN(3)
        REAL                ::  COOS_IN_OUTPUT(2), COOS_IN_MICROGRAPH(2)
        REAL(KIND=8)        ::  T_NORMAL, NORMAL_VECTOR(2), TANGENT_VECTOR(2)
        TYPE(WAYPOINT)      ::  WP_NORMAL
        TYPE(WINDOW)        ::  WIN
        REAL                ::  RVALUE
        INTEGER             ::  ADDR_MIN(3), ADDR_MAX(3)
        TYPE(INTERPOLATION_SCHEME)  ::  IINTERPOL
        ! Start work

        ! Interpolation scheme
        IF (PRESENT(INTERPOL)) THEN
            IINTERPOL = INTERPOL
        ELSE
            ! This needs to be replaced with some globally-available default interpolation scheme
            CALL INTERPOLATION_SCHEME_INIT(IINTERPOL,'LINEAR','RECT',2)
        ENDIF

        ! Check filament is set up properly
        IF (.NOT. FILAMENT_ALLOCATED(SELF)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_UNBEND_SEGMENT): input filament not allocated'
            CALL TERMINATE('Fatal error in FILAMENT_UNBEND_SEGMENT')
        ENDIF

        ! Check micrograph image is set up properly
        IF (.NOT. IS_ALLOCATED(MICROGRAPH)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_UNBEND_SEGMENT): input micrograph image not allocated'
            CALL TERMINATE('Fatal error in FILAMENT_UNBEND_SEGMENT')
        ENDIF

        ! Decide on dimensions for output (take into account PSF safety radius)
        OUTPUT_BOX_DIM(2) = SEGMENT_LENGTH
        OUTPUT_BOX_DIM(1) = SEGMENT_WIDTH
        IF (IsOdd(OUTPUT_BOX_DIM(1))) OUTPUT_BOX_DIM(1) = OUTPUT_BOX_DIM(1) + 1
        IF (IsOdd(OUTPUT_BOX_DIM(2))) OUTPUT_BOX_DIM(2) = OUTPUT_BOX_DIM(2) + 1
        WRITE(*,'(A,2(I0,1X))') '**info(FILAMENT_UNBEND_SEGMENT): unbent segment box dimensions = ', OUTPUT_BOX_DIM

        ! Check T parameter is within range
        IF (CENTER_T - OUTPUT_BOX_DIM(2)/2 .LT. SELF%T(1) .OR. CENTER_T + OUTPUT_BOX_DIM(2)/2 .GT. SELF%T(SELF%NWP)) THEN
            WRITE(*,'(A,I0,A,F0.12,A,I0,A,F0.12,A,F0.12)') '**ERROR(FILAMENT_UNBEND_SEGMENT): the segment (length ', &
                                                            OUTPUT_BOX_DIM(2) , ') centered around T (', CENTER_T,  &
                                                            ') falls beyond the ends of filament ', SELF%ID, &
                                                            ', which goes from ', SELF%T(1), ' to ', SELF%T(SELF%NWP)
            CALL TERMINATE('Fatal error in FILAMENT_UNBEND_SEGMENT')
        ENDIF

        ! Check the box size is not larger than the micrograph
        IF (MAXVAL(OUTPUT_BOX_DIM) .GT. MINVAL(MICROGRAPH%logical_dimensions(1:2))) THEN
            WRITE(*,'(A,2(I0,1X),A,I0,A,I0,A)') '**ERROR(FILAMENT_UNBEND_SEGMENT): output box size (', &
                                                OUTPUT_BOX_DIM, ') larger than input micrograph (', &
                                                MICROGRAPH%logical_dimensions(1),',', MICROGRAPH%logical_dimensions(2),')'
            CALL TERMINATE('Fatal error in FILAMENT_UNBEND_SEGMENT')
        ENDIF

        ! Allocate memory for output
        UNBENT_SEGMENT%logical_dimensions = [OUTPUT_BOX_DIM(1),OUTPUT_BOX_DIM(2),1]
        CALL IMAGE_ALLOCATE(UNBENT_SEGMENT)

        ! Address of origin
        if (unbent_segment%is_in_real_space .and. unbent_segment%origin_at_center) then
            addr_origin = unbent_segment%physical_address_of_box_center
        else
            addr_origin = [1,1,1]
        endif
        if (micrograph%is_in_real_space .and. micrograph%origin_at_center) then
            micrograph_origin = micrograph%physical_address_of_box_center
        else
            micrograph_origin = [1,1,1]
        endif

        ! Prepare the interpolating window
        CALL ALLOCATE_WINDOW(WIN,IINTERPOL%WINDOW_WIDTH,IS_COMP(MICROGRAPH),IS_THREED(MICROGRAPH))


        CALL BOUNDS_FROM_CENTER(MICROGRAPH,ADDR_MIN,ADDR_MAX)


        ! Loop over all pixels of output image
        DO J=1,UNBENT_SEGMENT%logical_dimensions(2)
            COOS_IN_OUTPUT(2) = REAL(J - ADDR_ORIGIN(2))
            ! T_NORMAL = 2nd coordinate + CENTER_T defines a point along the filament. The point we want to interpolate from lies somewhere on the normal to the filament at point T_NORMAL
            T_NORMAL = CENTER_T + COOS_IN_OUTPUT(2)

            ! Compute the coordinates of the micrograph point corresponding to T_NORMAL
            WP_NORMAL = WAYPOINT_FROM_T(SELF,T_NORMAL)

            ! The tangent vector
            TANGENT_VECTOR(1) = INTERPOL_SPLINE_EVAL_D(SELF%X_SPLINE,T_NORMAL)
            TANGENT_VECTOR(2) = INTERPOL_SPLINE_EVAL_D(SELF%Y_SPLINE,T_NORMAL)

            ! The normal vector (there are two normal vectors - I'm not sure which one is correct; if the wrong one is used, we'll probably end up mirroring around the Y axis)
            ! Option 1 of 2
            !NORMAL_VECTOR(1) = - TANGENT_VECTOR(2)
            !NORMAL_VECTOR(2) =   TANGENT_VECTOR(1)
            ! Option 2 of 2
            NORMAL_VECTOR(1) =   TANGENT_VECTOR(2)
            NORMAL_VECTOR(2) = - TANGENT_VECTOR(1)

            DO I=1,UNBENT_SEGMENT%logical_dimensions(1)
                ! Compute coordinates of current point relative to center of output image
                COOS_IN_OUTPUT(1) = REAL(I - ADDR_ORIGIN(1))

                ! The 1st coordinate defines the distance from the micrograph point to point T_NORMAL along this normal.

                ! Compute the coordinates of the micrograph point we want to interpolate from
                COOS_IN_MICROGRAPH = (WP_NORMAL%COO(1:2) - REAL(MICROGRAPH_ORIGIN(1:2))) + COOS_IN_OUTPUT(1)* NORMAL_VECTOR

                ! Interpolate a value
                WIN%CENTERED_ON(1:2) = COOS_IN_MICROGRAPH
                WIN%CENTERED_ON(3) = 0.0
                CALL EVAL_WINDOW_BOUNDS(WIN)
                IF (ANY(WIN%LBOUND(1:3) .GT. ADDR_MAX) .OR. ANY(WIN%UBOUND(1:3) .LT. ADDR_MIN)) THEN
                    CYCLE
                ENDIF
                CALL WINDOW_FROM_IMAGE_REAL(MICROGRAPH,WIN)
                CALL GET_REAL_FROM_WINDOW(RVALUE,WIN,IINTERPOL%INTERPOLATION_FUNCTION,IINTERPOL%WINDOWING_FUNCTION)

                !WRITE(*,'(A,F0.3)') '**debug(FILAMENT_UNBEND_SEGMENT): RVALUE = ', RVALUE
                ! Stick that value into the output image
                UNBENT_SEGMENT%real_values(I,J,1) = RVALUE
            ENDDO
        ENDDO
    END SUBROUTINE FILAMENT_UNBEND_SEGMENT


    !>  \brief  Print out a summary of all changes from filament 1 to filament 2
    !!  \todo   Use waypoint arithmetics (via operator overloading)
    SUBROUTINE FILAMENT_PRINT_DIFF(FIL1,FIL2)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)      ::  FIL1    !<  Filament, starting state
        TYPE(FILAMENT), INTENT(IN)      ::  FIL2    !<  Filament, final state
        ! Private variables
        INTEGER                         ::  I
        REAL                            ::  dX,dY,dZ,dR,dPhi,dTheta,dPsi,dPath,dS
        REAL                            ::  AdX,AdY,AdZ,AdR,AdPhi,AdTheta,AdPsi,AdPath,AdS,AdScore
        CHARACTER(LEN=5),   PARAMETER   ::  PARAM_LABEL(8)  =   ['X    ','Y    ','Z    ','Phi  ','Theta','Psi  ','Rise ','SubNm']
        ! Start work

        ! Check the two filaments have same number of waypoints
        IF (FIL1%NWP .NE. FIL2%NWP) THEN
            WRITE(*,'(A,2I3)') '**ERROR(FILAMENT_PRINT_DIFF): the two filaments do not have the same number of waypoints: ', &
                                FIL1%NWP, FIL2%NWP
            WRITE(*,'(A)') '**ERROR(FILAMENT_PRINT_DIFF): info for filaments 1 and 2: '
            CALL FILAMENT_PRINT_INFO(FIL1)
            CALL FILAMENT_PRINT_INFO(FIL2)
            CALL TERMINATE('Fatal error in FILAMENT_PRINT_DIFF')
        ENDIF

        ! Loop over way points
        WRITE(*,'(/A,I0)') 'Changes in filament ', FIL1%ID
        WRITE(*,'(A7,1X,5(A11,1X),3(A11,1X),(A11,1X),(A7,1X))') &
                                            'WP#','dX','dY','dZ','dR','dPathLgth','dPhi','dTheta','dPsi','dS','dScore'
        AdX     = 0.
        AdY     = 0.
        AdZ     = 0.
        AdR     = 0.
        AdPhi   = 0.
        AdTheta = 0.
        AdPsi   = 0.
        AdPath  = 0.
        AdS     = 0.
        DO I=1,FIL1%NWP
            dX      = FIL2%WP(I)%COO(1) - FIL1%WP(I)%COO(1)
            dY      = FIL2%WP(I)%COO(2) - FIL1%WP(I)%COO(2)
            dZ      = FIL2%WP(I)%COO(3) - FIL1%WP(I)%COO(3)
            dR      = SQRT(dX**2+dY**2+dZ**2)
            dPath   = FIL2%WP(I)%PATH_LENGTH - FIL1%WP(I)%PATH_LENGTH
            dPhi    = FIL2%WP(I)%EULER%PHI - FIL1%WP(I)%EULER%PHI
            dTheta  = FIL2%WP(I)%EULER%THETA - FIL1%WP(I)%EULER%THETA
            dPsi    = FIL2%WP(I)%EULER%PSI - FIL1%WP(I)%EULER%PSI
            dS      = FIL2%WP(I)%SUBUNIT_NUM - FIL1%WP(I)%SUBUNIT_NUM
            WRITE(*,'(I7,1X,5(F11.4,1X),3(F11.4,1X),1(F11.4,1X),1(A7,1X))')  I,dX,dY,dZ,dR,dPath,dPhi,dTheta,dPsi,dS,' '
            AdX     = AdX + ABS(dX)
            AdY     = AdY + ABS(dY)
            AdZ     = AdZ + ABS(dZ)
            AdR     = AdR + dR
            AdPath  = AdPath + dPath
            AdPhi   = AdPhi + ABS(dPhi)
            AdTheta = AdTheta + ABS(dTheta)
            AdPsi   = AdPsi + ABS(dPsi)
            AdS     = AdS + ABS(dS)
        ENDDO
        AdX     = AdX / FIL1%NWP
        AdY     = AdY / FIL1%NWP
        AdZ     = AdZ / FIL1%NWP
        AdR     = AdR / FIL1%NWP
        AdPath  = AdPath / FIL1%NWP
        AdPhi   = AdPhi / FIL1%NWP
        AdTheta = AdTheta / FIL1%NWP
        AdPsi   = AdPsi / FIL1%NWP
        AdS     = AdS / FIL1%NWP
        AdScore = FIL2%SCORE - FIL1%SCORE

        ! Print out averages
        WRITE(*,'(A7,1X,5(F11.4,1X),3(F11.4,1X),1(F11.4,1X),1(F7.4)/)') &
                                                                'Average',AdX,AdY,AdZ,AdR,AdPath,AdPhi,AdTheta,AdPsi,AdS,AdScore

        ! Rigid parameters
        IF (ANY(FIL1%RIGID_PARAMS)) THEN
            WRITE(*,'(A)',ADVANCE='NO') &
                                    'Changes in rigid parameters: '
            DO I=1,7
                IF (FIL1%RIGID_PARAMS(I)) THEN
                    WRITE(*,'(2A)',ADVANCE='NO') PARAM_LABEL(I), ': '
                    SELECT CASE (I)
                        CASE (1:3)
                            WRITE(*,'(G15.3)',ADVANCE='NO') FIL2%RIGID_WP%COO(I) - FIL1%RIGID_WP%COO(I)
                        CASE (4)
                            WRITE(*,'(G15.3)',ADVANCE='NO') FIL2%RIGID_WP%EULER%PHI - FIL1%RIGID_WP%EULER%PHI
                        CASE (5)
                            WRITE(*,'(G15.3)',ADVANCE='NO') FIL2%RIGID_WP%EULER%THETA - FIL1%RIGID_WP%EULER%THETA
                        CASE (6)
                            WRITE(*,'(G15.3)',ADVANCE='NO') FIL2%RIGID_WP%EULER%PSI - FIL1%RIGID_WP%EULER%PSI
                        CASE (7)
                            WRITE(*,'(G15.3)',ADVANCE='NO') FIL2%RIGID_WP%SUBUNIT_NUM - FIL1%RIGID_WP%SUBUNIT_NUM
                    END SELECT
                    WRITE(*,'(A)',ADVANCE='NO') '  '
                ENDIF
            ENDDO
            WRITE(*,'(A)') ' '
        ENDIF

    END SUBROUTINE FILAMENT_PRINT_DIFF



    !>  \brief  Write out a per-subunit trace to a CSV file on disk
    SUBROUTINE FILAMENT_TRACE_TO_DISK_CSV(SELF,FILENAME)
        USE FILAMENTS_PATH
        USE COOS
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(INOUT)   ::  SELF        !<  Filament object
        CHARACTER(LEN=*),   INTENT(IN)      ::  FILENAME    !<  Filename to write to
        ! Private variables
        TYPE(WAYPOINT), ALLOCATABLE     ::  WP(:)
        TYPE(POINT),    ALLOCATABLE     ::  COORS(:)
        REAL(KIND=8),   PARAMETER       ::  TRACE_STEP  =   1.0D0
        ! Start work
        CALL FILAMENT_GET_PATH(SELF,WP,TRACE_STEP,STEP_IN_REPEATS=.TRUE.)
        ALLOCATE(COORS(SIZE(WP)))
!#ifdef __INTEL_COMPILER
!        CALL POINT_ASSIGN(COORS,WP)
!#else
        COORS = WP
!#endif
        CALL POINTS_WRITE_TO_DISK_CSV(COORS,FILENAME)    ! I don't understand why I can't give WP as the argument here (see http://software.intel.com/en-us/forums/showthread.php?t=73564)
    END SUBROUTINE FILAMENT_TRACE_TO_DISK_CSV

    !>  \brief  Write a boxer DB containing the way points as boxes, and a trace as tiny boxes
    SUBROUTINE FILAMENT_BOXER_DB_TO_DISK(SELF,FILENAME,WITH_TRACE)
        USE FILAMENTS_PATH
        USE COOS
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(INOUT)   ::  SELF        !<  Filament object
        CHARACTER(LEN=*),   INTENT(IN)      ::  FILENAME    !<  Filename to write to
        LOGICAL,    OPTIONAL,   INTENT(IN)  ::  WITH_TRACE  !<  Include the trace as tiny boxes? (by default, .TRUE.)
        ! Private variables
        TYPE(WAYPOINT), ALLOCATABLE     ::  WP(:)
        TYPE(BOXER_DB)                  ::  DB
        REAL(KIND=8),   PARAMETER       ::  TRACE_STEP      =   1.0D0
        INTEGER,        PARAMETER       ::  TRACE_BOX_SIZE  =   0
        INTEGER,        PARAMETER       ::  WP_BOX_SIZE     =   10
        INTEGER ::  I
        LOGICAL ::  WWITH_TRACE
        ! Start work

        IF (PRESENT(WITH_TRACE)) THEN
            WWITH_TRACE = WITH_TRACE
        ELSE
            WWITH_TRACE = .TRUE.
        ENDIF

        !
        ! First, do the waypoints
        !
        CALL BOXER_DB_ALLOCATE(DB,SELF%NWP)
!#ifdef __INTEL_COMPILER
!        DO CONCURRENT (I=1:SELF%NWP)
!            CALL POINT_ASSIGN(DB%POINT(I),SELF%WP(I))
!        ENDDO
!#else
        FORALL(I=1:SELF%NWP) DB%POINT(I) = SELF%WP(I)
!#endif
        CALL BOXER_DB_SET_BOX_SIZE(DB,[WP_BOX_SIZE,WP_BOX_SIZE,WP_BOX_SIZE])
        CALL BOXER_DB_WRITE_TO_DISK(DB,FILENAME)

        !
        ! Now do the trace, if asked to do so
        !
        IF (WWITH_TRACE) THEN
            CALL FILAMENT_GET_PATH(SELF,WP,TRACE_STEP,STEP_IN_REPEATS=.TRUE.,IGNORE_SYMMETRY_MATES=.TRUE.)
            CALL BOXER_DB_ALLOCATE(DB,SIZE(WP))
!#ifdef __INTEL_COMPILER
!            DO CONCURRENT (I=1:SIZE(WP))
!                CALL POINT_ASSIGN(DB%POINT(I),WP(I))
!            ENDDO
!#else
            FORALL (I=1:SIZE(WP)) DB%POINT(I) = WP(I)
!#endif
            CALL BOXER_DB_SET_BOX_SIZE(DB,[TRACE_BOX_SIZE,TRACE_BOX_SIZE,TRACE_BOX_SIZE])
            CALL BOXER_DB_WRITE_TO_DISK(DB,FILENAME,APPEND=.TRUE.)
        ENDIF
    END SUBROUTINE FILAMENT_BOXER_DB_TO_DISK


    !>  \brief  Calcuate a 3D reconstruction from a filament
    SUBROUTINE FILAMENT_CALC_THREED(SELF,FILM_IMAGE,REC3D,SAFE_BOX_SIZE,FINALISE,FINALISED_VOL,WIPE,    &
                                    SP_MODE,SP_SEGMENT_LENGTH,SP_HELIX, &
                                    APPLY_A_BFACTOR,AVERAGE_SCORE,BFACTOR_PER_SUBUNIT)
        USE THREEDS
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_AVERAGE_WP_TO_WP_DISTANCE
        USE FILAMENTS_PATH, ONLY : FILAMENT_GET_PATH
        USE FILAMENTS_MODEL_MODIFY, ONLY : FILAMENT_CONVERT_TO_ONE_START
        USE CTFS
        USE UI, ONLY : INTERVAL_PRINT
        USE RUNTIME_PARAMETERS, ONLY : IS_EVEN
        USE OMP_LIB
        USE COOS, ONLY : WAYPOINTS_PRINT_INFO
        USE FILMS, ONLY : FILM_FIND_CTF_AT_COOS
        USE ROTATION_MATRICES, ONLY : EULER_TO_ROTMAT
        USE IMAGES_CTF, ONLY : IMAGE_APPLY_CTF
        USE IMAGES_ARITHM, ONLY : CIRC_MASK
        USE INTERPOLATION, ONLY : DEFAULT_INTERPOL_SCHEME
        use DatesAndTimes
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF                !<  Filament
        TYPE(IMAGE),                        INTENT(INOUT)   ::  FILM_IMAGE          !<  Image of the film on which the filament is. Will not be modified, but need intent INOUT for window_img routine.
        TYPE(THREED),                       INTENT(INOUT)   ::  REC3D               !<  3D reconstruction, which will be computed by inserting every helical subunit
        INTEGER,                OPTIONAL,   INTENT(IN)      ::  SAFE_BOX_SIZE       !<  Minimum safe box size (pixels) for boxing out segments, CTF flipping them etc. This can be larger than the box size used for the 3D reconstruction itself.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  FINALISE            !<  If .TRUE., a "finalised" 3D reconstruction will be returned (i.e. one that has been corrected for interpolation scheme, and where the CTF is corrected for)
        TYPE(IMAGE),            OPTIONAL,   INTENT(INOUT)   ::  FINALISED_VOL       !<  The finalised 3D reconstruction volume (only needed if FINALISE = .TRUE.)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  WIPE                !<  Whether the 3D reconstruction should be wiped at beginning
        !LOGICAL,               OPTIONAL,   INTENT(IN)      ::  IGNORE_FAL_WPS      !<  .FALSE. by default. If .TRUE., only data between the second and second-to-last waypoints are used for the reconstruction.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  SP_MODE             !<  Single particle mode. In single particle mode, the reconstruction is calculated from points near the filament's waypoint,
                                                                                    !!  with the assumption that the filament is straight near each of the waypoints.
        REAL,                   OPTIONAL,   INTENT(IN)      ::  SP_SEGMENT_LENGTH   !<  Single particle segment length (Angstroms)
        TYPE(HELICAL_SYMMETRY), OPTIONAL,   INTENT(IN)      ::  SP_HELIX            !<  Helical parameters to use for reconstruction in single-particle mode. If none given, the helical parameters found in
                                                                                    !!  the filament object or its waypoints will be used, which may not be desirable since in SP_MODE, the full-filament model does not hold.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  APPLY_A_BFACTOR     !<  Whether a B factor should be applied to the images before they are added to the 3D reconstruction. Defaults to false.
        REAL,                   OPTIONAL,   INTENT(IN)      ::  AVERAGE_SCORE       !<  Average score of filaments within the dataset. This will be used to work out what B factor to apply to the images. Defaults to 0.0.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  BFACTOR_PER_SUBUNIT !<  Apply the B factor filter per subunit, rather than per filament
        ! Private variables
        TYPE(WAYPOINT), ALLOCATABLE     ::  SUBUNITS(:)
        LOGICAL                         ::  FFINALISE
        INTEGER                         ::  I
        INTEGER                         ::  ICOO(3)
        REAL                            ::  RCOO(3)
        TYPE(IMAGE)                     ::  BOX, BOX_TMP
        TYPE(CTF_PARAMETERS)            ::  CTF
        LOGICAL                         ::  WWIPE
        !LOGICAL                         ::  IIGNORE_FAL_WPS
        INTEGER                         ::  NSR ! Non-symmetry redundancy
        LOGICAL                         ::  SSP_MODE
        INTEGER                         ::  NUM_WPS_PER_WP  ! only used in SP_MODE
        LOGICAL,    PARAMETER           ::  THREADING_ALLOWED   =   .TRUE.      !<  Allow OpenMP threading?
        LOGICAL                         ::  TTHREADING
        LOGICAL                         ::  AAPPLY_BFACTOR
        REAL                            ::  AAVERAGE_SCORE
        REAL                            ::  BFACTOR
        TYPE(FILAMENT)                  ::  SELF_ONESTART
        LOGICAL                         ::  USE_INCLUSION_THRESHOLD
        REAL                            ::  INCLUSION_THRESHOLD
        LOGICAL                         ::  INCLUDE_CURRENT_SUBUNIT
        REAL                            ::  CURRENT_SUBUNIT_SCORE
        INTEGER                         ::  NEAREST_WP
        INTEGER                         ::  NUM_SUBUNITS_INCLUDED
        INTEGER                         ::  SSAFE_BOX_SIZE
        !REAL                            ::  MATRIX(4,4)
        !TYPE(INTERPOLATION_SCHEME)      ::  INTERPOL
        LOGICAL,    PARAMETER           ::  USE_3D_SPLINE   =   .TRUE.  !<  The filament's Z info is used for working out the defocus, rather than the CTFTILT info in conjuction with (X,Y)
        LOGICAL,    PARAMETER           ::  USE_NSR =   .TRUE.  !< Whether to use non-symmetry-redundancy factor during reconstruction.
        LOGICAL                         ::  BBFACTOR_PER_SUBUNIT
        LOGICAL,    PARAMETER           ::  REAL_SPACE_2D_SHIFT_ROT = .FALSE.
        LOGICAL,    PARAMETER           ::  DEBUG = .FALSE.
        type(DateAndTime)               ::  myTime

        ! Start work

        ! Debug
        IF (DEBUG) PRINT '(A,L1)', '**debug(FILAMENT_CALC_THREED): USE_3D_SPLINE = ', USE_3D_SPLINE

        ! Check filament is fully allocated
        IF (.NOT. IS_ALLOCATED(SELF)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_CALC_THREED): filament is not fully allocated (filament information follows)'
            CALL FILAMENT_PRINT_INFO(SELF)
            CALL TERMINATE('Fatal error in FILAMENT_CALC_THREED')
        ENDIF

        ! Will we use inclusion threshold?
        USE_INCLUSION_THRESHOLD = ALLOCATED(SELF%WP_SCORE)
        WRITE(*,'(A)') '**debug(FILAMENT_CALC_THREED): overriding use_inclusion_threshold to F'
        USE_INCLUSION_THRESHOLD = .FALSE.

        ! Output a finalised 3D reconstruction?
        IF (PRESENT(FINALISE)) THEN
            FFINALISE = FINALISE
        ELSE
            FFINALISE = .FALSE.
        ENDIF

        ! Wipe?
        IF (PRESENT(WIPE)) THEN
            WWIPE = WIPE
        ELSE
            WWIPE = .FALSE.
        ENDIF

        ! Finalised 3D volume?
        IF (FFINALISE) THEN
            IF (.NOT. PRESENT(FINALISED_VOL)) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_CALC_THREED): finalised_vol argument not supplied, but FINALISE = .TRUE.'
                CALL TERMINATE('Fatal error in FILAMENT_CALC_THREED')
            ENDIF
        ENDIF

        ! Ignore first and last (FAL) way points (WPS)
        !IF (PRESENT(IGNORE_FAL_WPS)) THEN
        !   IIGNORE_FAL_WPS = IGNORE_FAL_WPS
        !ELSE
        !   IIGNORE_FAL_WPS = .FALSE.
        !ENDIF

        ! Single-particle mode
        SSP_MODE = .FALSE.
        IF (PRESENT(SP_MODE)) SSP_MODE = SP_MODE


        ! Apply B factor?
        AAPPLY_BFACTOR = .FALSE.
        IF (PRESENT(APPLY_A_BFACTOR)) AAPPLY_BFACTOR = APPLY_A_BFACTOR


        ! B factor per subunit?
        BBFACTOR_PER_SUBUNIT = .FALSE.
        IF (PRESENT(BFACTOR_PER_SUBUNIT)) BBFACTOR_PER_SUBUNIT = BFACTOR_PER_SUBUNIT

        ! Average filament score ?
        IF (PRESENT(AVERAGE_SCORE)) THEN
            AAVERAGE_SCORE = AVERAGE_SCORE
        ELSE
            IF (AAPPLY_BFACTOR) THEN
                WRITE(*,'(2A)') '**warning(FILAMENT_CALC_THREED): you asked to apply a B factor, but did not supply ',   &
                                'an average score. Will assume an average score of 0.0.'
            ENDIF
            AAVERAGE_SCORE = 0.0
        ENDIF

        ! B factor to be used during 3D reconstruction
        IF (AAPPLY_BFACTOR ) THEN
            IF (BBFACTOR_PER_SUBUNIT) THEN
                WRITE(*,'(A,I0,A)') '**info(FILAMENT_CALC_THREED): the subunits in filament ', SELF%ID, &
                                    ' will be added with B factors proportional to their local score'
            ELSE
                BFACTOR = BFACTOR_FROM_SCORE(SELF%SCORE,AAVERAGE_SCORE)
                WRITE(*,'(A,I0,A,F0.4,A,F0.2,A)') '**info(FILAMENT_CALC_THREED): filament ', SELF%ID, ', with a score of ', &
                                                    SELF%SCORE, ', will be added to the 3D reconstruction with a B factor of ', &
                                                    BFACTOR, ' pixels^2'
            ENDIF
        ELSE
            BFACTOR = 0.0
        ENDIF

        ! Let the user know if we're doing single-particle mode
        IF (SSP_MODE) THEN
            WRITE(*,'(3A)') '**info(FILAMENT_CALC_THREED): single-particle mode. Only subunits near filament waypoints will',    &
                            ' be included in the reconstruction, and their parameters will be computed by ',    &
                            'linear extrapolation from each waypoint.'
        ELSE
            WRITE(*,'(2A)') '**info(FILAMENT_CALC_THREED): normal filament mode. All subunits of the filament will be included', &
                            ' in the reconstruction.'
        ENDIF

        ! Let the user know if we're ignoring first and last wps
        IF (.NOT. SELF%FAL_WPS_IN_THREED) THEN
            WRITE(*,'(2A)') '**info(FILAMENT_CALC_THREED): only data between second and second-to-last waypoints will be',   &
                            ' included in the 3D reconstruction'
        ENDIF

        ! Check the threed object was initialised
        IF (.NOT. IS_ALLOCATED(REC3D)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_CALC_THREED): 3D reconstruction object was not initialised'
            CALL TERMINATE('Fatal error in FILAMENT_CALC_THREED')
        ENDIF

        ! Wipe the threed
        IF (WWIPE) THEN
            CALL THREED_WIPE(REC3D)
        ENDIF

        ! What will be the non-symmetry redundancy (NSR) caused by the insertion of the repeats?
        ! We approximate this as the number of repeats that fit in the reconstruction volume
        IF (USE_NSR) THEN
            IF (SELF%HELIX%PIXEL_UNITS) THEN
                NSR = NINT(REAL(REC3D%logical_dimensions(1)) / SELF%HELIX%RISE)
            ELSE
                NSR = NINT(REAL(REC3D%logical_dimensions(1)) / (SELF%HELIX%RISE/SELF%FILM%PIXEL_SIZE))
            ENDIF
            CALL THREED_SET_NSR(REC3D,NSR)
            IF (DEBUG) PRINT '(A,I0)', '**debug(FILAMENT_CALC_THREED): non-symmetry redundancy set to ', NSR
        ENDIF

        ! We have the option to only include in the reconstruction those subunits in parts of the filament
        ! where the local score is above a threshold
        IF (USE_INCLUSION_THRESHOLD) THEN
            INCLUSION_THRESHOLD = FILAMENT_THRESHOLD_FOR_INCLUSION_INTO_THREED(SELF)
            WRITE(*,'(A,F0.3)') '**info(FILAMENT_CALC_THREED): threshold for inclusion in the 3D set to ', INCLUSION_THRESHOLD
        ELSE
            INCLUSION_THRESHOLD = -1.0
        ENDIF

        ! Obtain the coordinates and orientation of all helical subunits
        IF (SSP_MODE) THEN
            ! This option does not work with microtubules yet
            IF (SELF%MICROTUBULE) THEN
                WRITE(*,'(2A)') '**ERROR(FILAMENT_CALC_THREED): single-particle 3D reconstruction not enabled yet',  &
                                ' from microtubule filaments'
                CALL TERMINATE('Fatal error from FILAMENT_CALC_THREED')
            ENDIF
            IF (.NOT. SELF%HELIX%PIXEL_UNITS) THEN
                CALL TERMINATE('Need helical parameters in pixels')
            ENDIF
            ! Work out a suitable number of subunits for reconstruction from each waypoint
            IF (PRESENT(SP_SEGMENT_LENGTH)) THEN
                NUM_WPS_PER_WP = NINT(    FILAMENT_AVERAGE_WP_TO_WP_DISTANCE(SELF)                   /SELF%HELIX%RISE)
            ELSE
                NUM_WPS_PER_WP = NINT(MIN(FILAMENT_AVERAGE_WP_TO_WP_DISTANCE(SELF),SP_SEGMENT_LENGTH)/SELF%HELIX%RISE)
            ENDIF
            IF (IS_EVEN(NUM_WPS_PER_WP)) NUM_WPS_PER_WP = NUM_WPS_PER_WP + 1
            WRITE(*,'(A,I0,A)')         '**info(FILAMENT_CALC_THREED): ', NUM_WPS_PER_WP, &
                                        ' subunits per filament waypoint will be used for the 3D reconstruction'
            WRITE(*,'(A,2(F0.2,A))')    '**info(FILAMENT_CALC_THREED): this number of subunits is equivalent to ', &
                                        NUM_WPS_PER_WP * SELF%HELIX%RISE, ' pixels, or ', &
                                        NUM_WPS_PER_WP * SELF%HELIX%RISE * SELF%FILM%PIXEL_SIZE, ' Angstroms.'
        ENDIF
        SELF_ONESTART = SELF
        IF (SELF%HELIX%NUM_STARTS .NE. 1) THEN
            WRITE(*,'(A,I0,A)') '**info(FILAMENT_CALC_THREED): filament is currently described by a ', SELF%HELIX%NUM_STARTS, &
                                '-start helical operator. Will be converted to 1-start helical operator so',    &
                                ' that subunits are explicitely descibed.'
            CALL FILAMENT_CONVERT_TO_ONE_START(SELF_ONESTART)
        ENDIF

        ! Get array of waypoints, one per subunit
        CALL FILAMENT_GET_PATH(         SELF_ONESTART, SUBUNITS, &
                                        STEP=1.0D0, &
                                        STEP_IN_REPEATS=.TRUE.,IGNORE_FAL_WPS=.NOT. SELF_ONESTART%FAL_WPS_IN_THREED,    &
                                        SP_MODE=SP_MODE,NUM_WPS_PER_WP=NUM_WPS_PER_WP,HELIX=SP_HELIX)


        ! Debug: only get a few positions to save time for testing
        !PRINT *, '**debug(FILAMENT_CALC_THREED): not doing every subunit - testing'
        !SUBUNITS = FILAMENT_GET_PATH(SELF,100.0,STEP_IN_PIXELS=.TRUE.)


        ! Safe box size?
        IF (PRESENT(SAFE_BOX_SIZE)) THEN
            SSAFE_BOX_SIZE = SAFE_BOX_SIZE
        ELSE
            SSAFE_BOX_SIZE = REC3D%logical_dimensions(1)
        ENDIF
        WRITE(*,'(A,(I0,1X))') '**debug(FILAMENT_CALC_THREED): box dim = ', SSAFE_BOX_SIZE

        ! Debug: printout detailled path info
        CALL WAYPOINTS_PRINT_INFO(SUBUNITS,PHI_MOD_360=.TRUE.)

        ! Avoid nested parallelisation
        IF (OMP_GET_ACTIVE_LEVEL() .GT. 0 .AND. THREADING_ALLOWED) THEN
            TTHREADING = .FALSE.
        ELSE
            TTHREADING = .TRUE.
        ENDIF

        ! Count how many subunits are actually added to the 3D reconstruction
        NUM_SUBUNITS_INCLUDED = 0

        PRINT '(A,I0,A,I0,A)', '**info(FILAMENT_CALC_THREED): will reconstruct filament ', SELF%ID ,&
                                ' with ', SIZE(SUBUNITS), ' subunits'

        !$OMP   PARALLEL    IF(TTHREADING)              &
        !$OMP               DEFAULT(NONE)               &
        !$OMP               PRIVATE(I,ICOO,RCOO,CTF)    &
        !$OMP               PRIVATE(BOX,BOX_TMP)        &
        !$OMP               FIRSTPRIVATE(BFACTOR)       &
        !$OMP               SHARED(SUBUNITS,FILM_IMAGE,SELF,REC3D,TTHREADING,SSAFE_BOX_SIZE,myTime) &
        !$OMP               PRIVATE(CURRENT_SUBUNIT_SCORE,INCLUDE_CURRENT_SUBUNIT,NEAREST_WP,BBFACTOR_PER_SUBUNIT) &
        !$OMP               SHARED(INCLUSION_THRESHOLD,USE_INCLUSION_THRESHOLD,AAPPLY_BFACTOR,AAVERAGE_SCORE) &
        !$OMP               REDUCTION(+:NUM_SUBUNITS_INCLUDED)

        ! Check that the number of threads is reasonable - good speed up only occurs with interpolation schemes which use large supports
        IF (TTHREADING) THEN
            !$OMP   SINGLE
            IF (OMP_GET_NUM_THREADS() .GT. THREED_GET_WINDOW_WIDTH(REC3D) / 2) THEN
                WRITE(*,'(A,I0,A,I0,A)') '**WARNING(FILAMENT_CALC_THREED): threading with ', OMP_GET_NUM_THREADS(), &
               ' threads may not offer great speed-ups when using an interpolation scheme with kernel support (window width) of ', &
                THREED_GET_WINDOW_WIDTH(REC3D), ' pixels. In extreme cases, you may actually experience a slow-down.'
            ENDIF
            !$OMP   END SINGLE  NOWAIT
        ENDIF


        ! OpenMP PRIVATE does not set up derived type objects properly, so let's do it explicitely here
        BOX%IN_MEMORY        = .FALSE.
        BOX%PLANNED          = .FALSE.
        BOX%logical_dimensions(3)           = 1
        BOX_TMP%IN_MEMORY    = .FALSE.
        BOX_TMP%PLANNED      = .FALSE.
        BOX_TMP%logical_dimensions(3)       = 1

        BOX%logical_dimensions(1:2)         = SSAFE_BOX_SIZE
        BOX_TMP%logical_dimensions(1:2)     = SSAFE_BOX_SIZE

        ! This variable keeps track of the nearest WP, which is worked out by the function to find the local score
        NEAREST_WP = 1


        ! Loop over the helical subunits
        !$OMP   DO
        DO I=1,SIZE(SUBUNITS)
            IF (INTERVAL_PRINT(I,SIZE(SUBUNITS))) THEN
                WRITE(*,'(3A,I4,A,I4)',ADVANCE='NO') '**info(FILAMENT_CALC_THREED): ', myTime%DateAndTimeAsString(), &
                                                    ': working on subunit ', I, ' of ', SIZE(SUBUNITS)
                IF (TTHREADING) THEN
                    WRITE(*,'(A,I0,A,I0)') ' on thread ', OMP_GET_THREAD_NUM()+1, ' of ', OMP_GET_NUM_THREADS()
                ELSE
                    WRITE(*,'(A)') ' '
                ENDIF
            ENDIF
            ! Find the score of the current subunit, if needed
            IF (USE_INCLUSION_THRESHOLD .OR. (AAPPLY_BFACTOR .AND. BBFACTOR_PER_SUBUNIT)) THEN
                ! Find the local score
                CURRENT_SUBUNIT_SCORE = FILAMENT_FIND_LOCAL_SCORE(SELF,SUBUNITS(I)%PATH_LENGTH,NEAREST_WP)
            ENDIF
            ! Will we include the current subunit
            IF (USE_INCLUSION_THRESHOLD) THEN
                INCLUDE_CURRENT_SUBUNIT = CURRENT_SUBUNIT_SCORE .GT. INCLUSION_THRESHOLD
                IF (DEBUG) THEN
                WRITE(*,'(A,I0,A,F0.2,A,F0.3,A,L1)') &
                                '**debug(FILAMENT_CALC_THREED): subunit ', I, ' at path length ', SUBUNITS(I)%PATH_LENGTH,   &
                                ' has a score of ', CURRENT_SUBUNIT_SCORE, '. Is this above threshold? ', &
                                INCLUDE_CURRENT_SUBUNIT
                ENDIF
            ELSE
                INCLUDE_CURRENT_SUBUNIT = .TRUE.
            ENDIF
            IF (INCLUDE_CURRENT_SUBUNIT) THEN
                NUM_SUBUNITS_INCLUDED = NUM_SUBUNITS_INCLUDED + 1
                ! Calculate the appropriate CTF parameters
                IF (USE_3D_SPLINE) THEN
                    ! The average CTF for the film
                    CTF = SELF%FILM%CTFTILT%CTF
                    ! Convert CTF parameters to pixel units
                    IF (.NOT. CTF%PIX_UNITS) CALL ANG_TO_PIX(CTF,SELF%FILM%PIXEL_SIZE)
                    ! The CTF at Z (but with astigmatism)
                    CALL CTF_AT_Z(CTF,SUBUNITS(I)%COO(3))
                ELSE
                    CTF = FILM_FIND_CTF_AT_COOS(SELF%FILM,SUBUNITS(I)%COO(1),SUBUNITS(I)%COO(2))
                    ! Convert CTF parameters to pixel units
                    IF (.NOT. CTF%PIX_UNITS) CALL ANG_TO_PIX(CTF,SELF%FILM%PIXEL_SIZE)
                ENDIF
                ! Get integer coordinates & remainders
                ICOO = NINT(SUBUNITS(I)%COO)
                RCOO = SUBUNITS(I)%COO - ICOO
                ! Window image
                BOX%is_in_real_space = .true.
                BOX%ORIGIN_AT_CENTER = .TRUE.
                CALL WINDOW_IMG(FILM_IMAGE,BOX,ICOO)
                IF (IMAGE_IS_BLANK(BOX)) WRITE(*,'(A)') '**warning(FILAMENT_CALC_THREED): blank image boxed from micrograph'
                ! Do the translation shift in Fourier space, don't do in-plane rotatoin
                CALL FWD_FT(BOX)
                CALL PSHIFT_IMAGE(BOX,-RCOO(1),-RCOO(2),0.0)

                ! Compute the subunit's B factor
                IF (AAPPLY_BFACTOR .AND. BBFACTOR_PER_SUBUNIT) THEN
                    BFACTOR = BFACTOR_FROM_SCORE(CURRENT_SUBUNIT_SCORE,AAVERAGE_SCORE)
                    IF (BFACTOR .LT. -500.) THEN
                        BFACTOR = -500.
                        WRITE(*,'(A,I0,A,I0,A,F0.3,A)') '**warning(FILAMENT_CALC_THREED): subunit ', I, ' of ', SIZE(SUBUNITS), &
                                                    ' had a high score and its B factor was limited to ', BFACTOR, ' A^2'
                    ENDIF
                    WRITE(*,'(A,I0,A,I0,A,F0.3,A)') '**debug(FILAMENT_CALC_THREED): subunit ', I, ' of ', SIZE(SUBUNITS), &
                                                    ' will be added with a B factor of ', BFACTOR, ' A^2'
                ENDIF
                ! Add slice to the 3D reconstruction
                CALL THREED_ADD_SLICE(  REC3D,BOX,CTF,SUBUNITS(I)%EULER,BFACTOR=BFACTOR/SELF%FILM%PIXEL_SIZE**2,  &
                                        APPLY_CTF_TO_SLICE=.TRUE.,SLICE_PSI_ZERO=.FALSE. &
                                        ,ADD_TO_THREED_NUM=1+INT(2.*REAL(I-1)/REAL(SIZE(SUBUNITS))) &
                                        )
                ! Debug: check for NaNs
                IF (DEBUG) THEN
                    IF (THREED_HAS_NAN(REC3D)) THEN
                        WRITE(*,'(A,I0,A,I0)') '**debug(FILAMENT_CALC_THREED): 3D reconstruction has at least one NaN value '// &
                                        'after inserting slice ', I , ' of ', SIZE(SUBUNITS)
                    ENDIF
                ENDIF
            ENDIF
        ENDDO
        !$OMP   END DO
        CALL IMAGE_DEALLOCATE(BOX)
        !$OMP   END PARALLEL

        WRITE(*,'(3A,I0,A,I0,A)') '**info(FILAMENT_CALC_THREED): ', myTime%DateAndTimeAsString(), ': ', &
                                    NUM_SUBUNITS_INCLUDED, ' of ', SIZE(SUBUNITS),  &
                                    ' subunits have been added to the 3D reconstruction'

        ! If required, compute a weighted / corrected 3D reconstruction volume
        IF (FFINALISE) THEN
            CALL THREED_EVAL_RECONSTRUCTION(REC3D,FINALISED_VOL)
        ENDIF

    END SUBROUTINE FILAMENT_CALC_THREED

    !>  \brief  Find the local score at a point along the filament specified by its path length
    REAL FUNCTION FILAMENT_FIND_LOCAL_SCORE(SELF,PATH_LENGTH,NEAREST_WP_LAST_TIME)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)      ::  SELF                    !<  Filament
        REAL(KIND=8),   INTENT(IN)      ::  PATH_LENGTH             !<  Ordinate at which to evaluate the score
        INTEGER,        INTENT(INOUT)   ::  NEAREST_WP_LAST_TIME    !<  The index of the nearest wapoint the last time
                                                                    !!  this function was called. Set this to 1 the first
                                                                    !!  time you call this function.
        ! Private variables
        INTEGER         ::  NEAREST_WP, NEAREST_PREVIOUS_WP
        INTEGER         ::  CURRENT_WP
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        ! Check the score was actually computed
        IF (.NOT. ALLOCATED(SELF%WP_SCORE)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_FIND_LOCAL_SC0RE): score has not yet been computed'
            CALL TERMINATE('Fatal error in FILAMENT_FIND_LOCAL_SCORE')
        ENDIF

        IF (DEBUG) THEN
            WRITE(*,'(A,I0)') '**debug(FILAMENT_FIND_LOCAL_SC0RE): nearest WP last time = ',  NEAREST_WP_LAST_TIME
        ENDIF

        NEAREST_WP = 0
        NEAREST_PREVIOUS_WP = 0
        ! Iterate from previous guess to end of filament
        DO CURRENT_WP=MAX(1,NEAREST_WP_LAST_TIME-1),SIZE(SELF%WP_SCORE) - 1
            IF (    PATH_LENGTH .GE. SELF%WP_SCORE(CURRENT_WP)%PATH_LENGTH .AND. &
                    PATH_LENGTH .LE. SELF%WP_SCORE(CURRENT_WP+1)%PATH_LENGTH ) THEN
                NEAREST_PREVIOUS_WP = CURRENT_WP
            ENDIF
        ENDDO
        ! If we still haven't found it, look from start of filament to previous guess
        IF (NEAREST_PREVIOUS_WP .EQ. 0) THEN
            IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_FIND_LOCAL_SC0RE): searching from start of filament'
            DO CURRENT_WP=1,MAX(1,NEAREST_WP_LAST_TIME-2)
                IF (    PATH_LENGTH .GE. SELF%WP_SCORE(CURRENT_WP)%PATH_LENGTH .AND. &
                        PATH_LENGTH .LE. SELF%WP_SCORE(CURRENT_WP+1)%PATH_LENGTH ) THEN
                    NEAREST_PREVIOUS_WP = CURRENT_WP
                ENDIF
            ENDDO
        ENDIF
        ! If we still haven't found it, look after the last wp_score, but before the last wp
        IF (NEAREST_PREVIOUS_WP .EQ. 0) THEN
            IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_FIND_LOCAL_SC0RE): searching at end of filament'
            IF (    PATH_LENGTH .GE. SELF%WP_SCORE(SIZE(SELF%WP_SCORE))%PATH_LENGTH .AND. &
                    PATH_LENGTH .LE. SELF%WP(SELF%NWP)%PATH_LENGTH) THEN
                NEAREST_PREVIOUS_WP = SIZE(SELF%WP_SCORE)
            ENDIF
        ENDIF
        ! Now we know which interval our point is in, find out which WP it's nearest to
        IF (NEAREST_PREVIOUS_WP .EQ. 0) THEN
            WRITE(*,'(A,F0.3,A)') '**ERROR(FILAMENT_FIND_LOCAL_SCORE): point with path_length ', PATH_LENGTH, &
                            ' does not appear to lie on filament '
            CALL FILAMENT_PRINT_INFO(SELF)
            CALL TERMINATE('Fatal error in FILAMENT_FIND_LOCAL_SCORE')
        ELSE
            IF (    PATH_LENGTH - SELF%WP_SCORE(NEAREST_PREVIOUS_WP)%PATH_LENGTH .LE. &
                    SELF%WP_SCORE(NEAREST_PREVIOUS_WP+1)%PATH_LENGTH - PATH_LENGTH ) THEN
                NEAREST_WP = NEAREST_PREVIOUS_WP
            ELSE
                NEAREST_WP = NEAREST_PREVIOUS_WP+1
            ENDIF
        ENDIF

        IF (NEAREST_WP .EQ. 0) THEN
            ! If we haven't found the nearest WP by now and we haven't already crashed out, something
            ! strange is happening.
            CALL TERMINATE('Programming error in FILAMENT_FIND_LOCAL_SCORE')
        ELSE
            NEAREST_WP_LAST_TIME = NEAREST_WP    ! Save this for the next time the function is called
            FILAMENT_FIND_LOCAL_SCORE = SELF%WP_SCORE(NEAREST_WP)%SCORE
        ENDIF
    END FUNCTION FILAMENT_FIND_LOCAL_SCORE


    !>  \brief  Compute a thresold for inclusion into the 3D reconstruction
    REAL FUNCTION FILAMENT_THRESHOLD_FOR_INCLUSION_INTO_THREED(SELF) RESULT(THRESHOLD)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF    !<  Filament
        ! Private variables
        REAL,   PARAMETER   ::  NUM_STDEV = 1.0
        REAL(KIND=8)    ::  AVE,STD
        ! Start work

        ! Get statistics for filament score
        IF (ALLOCATED(SELF%WP_SCORE)) THEN
            CALL FILAMENT_SEGMENT_SCORE_STATS(SELF,AVE,STD)
            THRESHOLD = AVE - NUM_STDEV*STD
        ELSE
            ! This case arises when the scoring function has never been called,
            ! for example during filament_startup
            THRESHOLD = -1.0
        ENDIF
    END FUNCTION FILAMENT_THRESHOLD_FOR_INCLUSION_INTO_THREED





    !>  \brief  Compute average and std dev of segment scores
    SUBROUTINE FILAMENT_SEGMENT_SCORE_STATS(SELF,AVE,STD)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF    !<  Filament
        REAL(KIND=8),   INTENT(OUT) ::  AVE     !<  Average of scores
        REAL(KIND=8),   INTENT(OUT) ::  STD     !<  Standard deviation of scores
        ! Private variables
        INTEGER ::  CURRENT_WP
        ! Start work
        AVE = 0.0D0
        STD = 0.0D0
        IF (ALLOCATED(SELF%WP_SCORE)) THEN
            DO CURRENT_WP=1,SIZE(SELF%WP_SCORE)
                AVE = AVE + SELF%WP_SCORE(CURRENT_WP)%SCORE
                STD = STD + SELF%WP_SCORE(CURRENT_WP)%SCORE**2
            ENDDO
            STD = (STD - (AVE**2/SIZE(SELF%WP_SCORE)))/SIZE(SELF%WP_SCORE)
            STD = SQRT(STD)
            AVE = AVE / SIZE(SELF%WP_SCORE)
        ELSE
            WRITE(*,'(A)') '**ERROR(FILAMENT_SEGMENT_SCORE_STATS): wp_score array not allocated. Scoring probably not done.'
            CALL TERMINATE('Fatal error in FILAMENT_SEGMENT_SCORE_STATS')
        ENDIF
    END SUBROUTINE FILAMENT_SEGMENT_SCORE_STATS



    !>  \brief  Waypoints with non-integer subunit_num cannot be compared to the 3D model, which always has subunit_num=0.0 at origin.
    !!          This subroutine works out how the model would need to be rotated and shifted to be compared to that waypoint.
    !!
    !!  See notes of December 7th, 2011, for details
    !!
    SUBROUTINE FILAMENT_MODEL_PHASING_FROM_WP(SELF,WP,MODEL_Z_SHIFT,LOCAL_H_SYM)
        USE ROTATION_MATRICES
        USE FILAMENTS_MODEL_BASE, ONLY : WP_COMPUTE_SUBUNIT_NUM_PHASING
        USE HELICES, ONLY : HELICAL_SYMMETRY_UNIT_CONVERT
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                 INTENT(INOUT)   ::  SELF
        TYPE(WAYPOINT),                 INTENT(INOUT)   ::  WP              !<  Waypoint. Its phi value will be modified to account for the non-integer subunit_num value.
        REAL(KIND=8),                   INTENT(OUT)     ::  MODEL_Z_SHIFT   !<  This Z shift (pixels) should be applied to the model before calculating a projection
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  LOCAL_H_SYM     !<  Whether the model 3D should be phased (shifted & rotated) using local helical paramters. Defaults to true.
        ! Private variables
        REAL                    ::  SUBUNIT_NUM_SHIFT
        REAL                    ::  SUBUNIT_NUM_PHI
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        LOGICAL                 ::  LLOCAL_H_SYM
        ! Start work

        ! Is the model symmetrised using the local helical symmetry?
        IF (PRESENT(LOCAL_H_SYM)) THEN
            LLOCAL_H_SYM = LOCAL_H_SYM
        ELSE
            LLOCAL_H_SYM = .TRUE.
        ENDIF

        ! Check helical symmetry is in pixel units
        IF (.NOT. SELF%HELIX%PIXEL_UNITS) CALL HELICAL_SYMMETRY_UNIT_CONVERT(SELF%HELIX,SELF%FILM%PIXEL_SIZE)

        ! Work out additional shift due to model phasing by subunit_num
        SUBUNIT_NUM_SHIFT = WP_COMPUTE_SUBUNIT_NUM_PHASING(WP) ! In units of subunit_num

        ! Working out the phi rotation
        IF (LLOCAL_H_SYM) THEN
            SUBUNIT_NUM_PHI = SUBUNIT_NUM_SHIFT * WP%HELIX%TWIST
        ELSE
            SUBUNIT_NUM_PHI = SUBUNIT_NUM_SHIFT * SELF%HELIX%TWIST
        ENDIF

        ! Convert from subunit number to axial distance
        IF (LLOCAL_H_SYM) THEN
            SUBUNIT_NUM_SHIFT = SUBUNIT_NUM_SHIFT * WP%HELIX%RISE
        ELSE
            SUBUNIT_NUM_SHIFT = SUBUNIT_NUM_SHIFT * SELF%HELIX%RISE
        ENDIF

        ! This is the amount we would need to translate the model along Z
        MODEL_Z_SHIFT   = - SUBUNIT_NUM_SHIFT

        ! This is the amount we would need to rotate the model around Z
        SUBUNIT_NUM_PHI = - SUBUNIT_NUM_PHI

        ! Rather than rotate the model itself, we'll just extract the slice at a different angle
        WP%EULER%PHI = WP%EULER%PHI + SUBUNIT_NUM_PHI

     END SUBROUTINE FILAMENT_MODEL_PHASING_FROM_WP



    !>  \brief  Modify a waypoint's phi and (x,y) coordinates so that it can be compared
    !!          directly to the 3D model with canonical helical symmetry (i.e. it will coincide
    !!          exactly with a helical subunit)
    !!
    !!  See also documentation for FILAMENT_COMPUTE_WP_PHASING.
    !!
    !!  Note that with high curvature, it might be better to adjust the T value for the waypoint rather than do a XY shift as in here, so that the curvature might be better accounted for.
    !!  Because this may require more major re-writes, and because I anticipate the phase shifts to be small, I am implementing things like this for now (May/June 2011).
    SUBROUTINE FILAMENT_APPLY_PHASING_TO_WP_OLD(SELF,WP,LOCAL_H_SYM,NEG)
        USE ROTATION_MATRICES
        USE FILAMENTS_MODEL_BASE, ONLY : WP_COMPUTE_SUBUNIT_NUM_PHASING
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE HELICES, ONLY : HELICAL_SYMMETRY_UNIT_CONVERT
		IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                 INTENT(INOUT)   ::  SELF
        TYPE(WAYPOINT),                 INTENT(INOUT)   ::  WP
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  LOCAL_H_SYM !<  Whether the model 3D we are phasing against is h-symmetrised for every waypoint. Defaults to false.
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  NEG
        ! Private variables
        REAL                    ::  SUBUNIT_NUM_SHIFT
        REAL                    ::  SUBUNIT_NUM_PHI
        REAL                    ::  SUBUNIT_NUM_SHIFT_XY(2)
        REAL                    ::  MAT(4,4)
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        INTEGER                 ::  CALC_NEG
        LOGICAL                 ::  LLOCAL_H_SYM
        LOGICAL,    PARAMETER   ::  SANITY_CHECK    =   .FALSE.
        REAL                    ::  EXPECTED_SHIFT_XY(2)
        ! Start work

        ! Is the model symmetrised using the local helical symmetry?
        IF (PRESENT(LOCAL_H_SYM)) THEN
            LLOCAL_H_SYM = LOCAL_H_SYM
        ELSE
            LLOCAL_H_SYM = .FALSE.
        ENDIF

        ! Are we calculating the negative?
        CALC_NEG = 1
        IF (PRESENT(NEG)) THEN
            IF (NEG) THEN
                CALC_NEG = -1
            ELSE
                CALC_NEG = 1
            ENDIF
        ENDIF

        ! Check helical symmetry is in pixel units
        IF (.NOT. SELF%HELIX%PIXEL_UNITS) CALL HELICAL_SYMMETRY_UNIT_CONVERT(SELF%HELIX,SELF%FILM%PIXEL_SIZE)

        ! Work out additional shift due to model phasing by subunit_num
        SUBUNIT_NUM_SHIFT = WP_COMPUTE_SUBUNIT_NUM_PHASING(WP) * REAL(CALC_NEG) ! In units of subunit_num

        ! Working out the phi rotation
        IF (LLOCAL_H_SYM) THEN
            SUBUNIT_NUM_PHI = SUBUNIT_NUM_SHIFT * WP%HELIX%TWIST
        ELSE
            SUBUNIT_NUM_PHI = SUBUNIT_NUM_SHIFT * SELF%HELIX%TWIST
        ENDIF

        ! Convert from subunit number to axial distance
        IF (LLOCAL_H_SYM) THEN
            SUBUNIT_NUM_SHIFT = SUBUNIT_NUM_SHIFT * WP%HELIX%RISE
        ELSE
            SUBUNIT_NUM_SHIFT = SUBUNIT_NUM_SHIFT * SELF%HELIX%RISE
        ENDIF

        ! Convert axial distance to XY in the projection
        ! The safest way to do this might be with an actual rotation matrix used by the projection routines.
        ! We think of the shift as a point on the Z axis, and we multiply it by the rotation matrix
        MAT = EULER_TO_ROTMAT(WP%EULER%PHI,WP%EULER%THETA,WP%EULER%PSI,REVERSE=.FALSE.)
        SUBUNIT_NUM_SHIFT_XY(1:2) = MAT(1:2,3) * SUBUNIT_NUM_SHIFT
        IF (DEBUG) THEN
            WRITE(*,'(A,4(G15.3,1X))') '**debug(FILAMENT_APPLY_PHASING_TO_WP): subunit_num_shift, subunit_num_shift_x, '//&
                                    'subunit_num_shift_y, subunit_num_phi = ', &
                                    SUBUNIT_NUM_SHIFT / WP%HELIX%RISE, SUBUNIT_NUM_SHIFT_XY(1:2), SUBUNIT_NUM_PHI
        ENDIF

        ! Sanity check (see notes 1-Dec-2011 for derivation) - minus sign is because clockwise positive convention
        ! 23-Jan: doesn't appear to work correctly
        IF (SANITY_CHECK) THEN
            EXPECTED_SHIFT_XY(1) = COS(-RADIANS(WP%EULER%PSI)) * COS(-RADIANS(90.-WP%EULER%THETA)) * SUBUNIT_NUM_SHIFT
            EXPECTED_SHIFT_XY(2) = SIN(-RADIANS(WP%EULER%PSI)) * COS(-RADIANS(90.-WP%EULER%THETA)) * SUBUNIT_NUM_SHIFT
            ! Check the shifts are as expected
            IF (ABS(SUBUNIT_NUM_SHIFT_XY(1) - EXPECTED_SHIFT_XY(1)) .GT. 0.001 .OR. &
                ABS(SUBUNIT_NUM_SHIFT_XY(2) - EXPECTED_SHIFT_XY(2)) .GT. 0.001) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_APPLY_PHASING_TO_WP): Sanity check failed.'
                WRITE(*,'(A,1(F0.5,1X))') '**ERROR(FILAMENT_APPLY_PHASING_TO_WP):        subunit_num = ', WP%SUBUNIT_NUM
                WRITE(*,'(A,2(F0.5,1X))') '**ERROR(FILAMENT_APPLY_PHASING_TO_WP): expected x y shift = ', EXPECTED_SHIFT_XY
                WRITE(*,'(A,2(F0.5,1X))') '**ERROR(FILAMENT_APPLY_PHASING_TO_WP):   actual x y shift = ', SUBUNIT_NUM_SHIFT_XY
                CALL TERMINATE('Failed sanity check in FILAMENT_APPLY_PHASING_TO_WP')
            ENDIF
        ENDIF

        ! Apply the shift and the rotation
        WP%COO(1:2) = WP%COO(1:2) - SUBUNIT_NUM_SHIFT_XY(1:2)
        WP%EULER%PHI = WP%EULER%PHI - SUBUNIT_NUM_PHI

        !
        WP%SUBUNIT_NUM = NINT(WP%SUBUNIT_NUM)
    END SUBROUTINE FILAMENT_APPLY_PHASING_TO_WP_OLD




    !>  \brief  Resample waypoints along the path of a filament
    SUBROUTINE FILAMENT_RESAMPLE_WAYPOINTS(SELF,STEP_SIZE,STEP_IN_REPEATS,STEP_IN_PHI)
        USE FILAMENTS_PATH, ONLY : FILAMENT_GET_PATH_SUBUNIT_STEPS, FILAMENT_GET_PATH_PIXEL_STEPS, FILAMENT_GET_PATH_CROSSOVERS
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        USE COOS, ONLY : WAYPOINTS_PRINT_INFO, POINT, WAYPOINT, WAYPOINTS_REMOVE_TOO_CLOSE
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
		IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF            !<  Filament object
        REAL(KIND=8),               INTENT(IN)      ::  STEP_SIZE       !<  Desired distance from waypoint to waypoint, in pixels
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  STEP_IN_REPEATS !<  The step size was given in numbers of repeats
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  STEP_IN_PHI     !<  The step size was given in phi degrees
        ! Private variables
        TYPE(FILAMENT)              ::  FIL_TMP
        TYPE(WAYPOINT), ALLOCATABLE ::  WPS_TMP(:)
        LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
        INTEGER                     ::  MIN_NUM_WPS
        LOGICAL                     ::  IGNORE_FAL_SEGMENTS, SSTEP_IN_REPEATS, SSTEP_IN_PHI
        ! Start work

        ! Will we ignore the first and last segments?
        IGNORE_FAL_SEGMENTS = .NOT. SELF%FAL_WPS_IN_SCORE .OR. .NOT. SELF%FAL_WPS_IN_THREED
        IF (IGNORE_FAL_SEGMENTS) THEN
            MIN_NUM_WPS = 5
        ELSE
            MIN_NUM_WPS = 3
        ENDIF

        SSTEP_IN_REPEATS = .FALSE.
        IF (PRESENT(STEP_IN_REPEATS)) SSTEP_IN_REPEATS = STEP_IN_REPEATS
        SSTEP_IN_PHI     = .FALSE.
        IF (PRESENT(STEP_IN_PHI)) SSTEP_IN_PHI = STEP_IN_PHI

        ! Info
        IF (SSTEP_IN_REPEATS) THEN
            WRITE(*,'(A,I0,A,F0.2,A)') '**info(FILAMENT_RESAMPLE_WAYPOINTS): will resample waypoints along the path of filament ', &
                                        SELF%ID, ' so that waypoints are ', STEP_SIZE, ' repeats apart'
        ELSE IF (SSTEP_IN_PHI) THEN
            WRITE(*,'(A,I0,A,F0.2,A)') '**info(FILAMENT_RESAMPLE_WAYPOINTS): will resample waypoints along the path of filament ', &
                                        SELF%ID, ' so that waypoints are ', STEP_SIZE, ' degrees apart in phi (twist)'
        ELSE
            WRITE(*,'(A,I0,A,F0.2,A)') '**info(FILAMENT_RESAMPLE_WAYPOINTS): will resample waypoints along the path of filament ', &
                                        SELF%ID, ' so that waypoints are ', STEP_SIZE, ' pixels apart'
        ENDIF

        ! Debug
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_RESAMPLE_WAYPOINTS): filament info on entry: '
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

        ! Get new waypoints
        IF (SSTEP_IN_REPEATS) THEN
            CALL    FILAMENT_GET_PATH_SUBUNIT_STEPS(SELF,WPS_TMP,INT(STEP_SIZE),IGNORE_FAL_WPS=IGNORE_FAL_SEGMENTS)
        ELSE IF (SSTEP_IN_PHI) THEN
            ! Turn ignore_fal_wps off becauseoften the 2nd wp has phi=3 degrees (or some small value)
            ! and we end up losing the first crossover
            CALL    FILAMENT_GET_PATH_CROSSOVERS(   SELF,WPS_TMP,STEP=1,IGNORE_FAL_WPS=IGNORE_FAL_SEGMENTS .AND. .FALSE.,   &
                                                    IGNORE_SYMMETRY_MATES=.TRUE., &
                                                    PHI_STEP_SIZE=STEP_SIZE, &
                                                    INCLUDE_FAL_WPS=IGNORE_FAL_SEGMENTS)
        ELSE
            CALL    FILAMENT_GET_PATH_PIXEL_STEPS(  SELF,STEP_SIZE,WPS_TMP,   &
                                                    IGNORE_FAL_WPS=IGNORE_FAL_SEGMENTS, &
                                                    INCLUDE_FINISH_WP=.TRUE.,   &
                                                    INCLUDE_FAL_WPS=IGNORE_FAL_SEGMENTS)
        ENDIF

        CALL WAYPOINTS_REMOVE_TOO_CLOSE(WPS_TMP)

		IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_RESAMPLE_WAYPOINTS): resampled waypoints: '
            CALL WAYPOINTS_PRINT_INFO(WPS_TMP)
        ENDIF

        IF (SIZE(WPS_TMP) .GE. MIN_NUM_WPS) THEN
            ! Initialise new filament
            CALL FILAMENT_INIT( FIL_TMP,waypoints_in=WPS_TMP,                                   &
                                FILM_IN                     =   SELF%FILM,                      &
                                MOULD                       =   SELF)


            ! Deallocate the original filament
            CALL FILAMENT_DEALLOCATE(SELF)

            ! Copy the new filament over
            FIL_TMP%SCORE = SELF%SCORE
            SELF = FIL_TMP

            ! Make sure the model is up-to-date
            CALL FILAMENT_UPDATE_MODEL(SELF)

            ! Now that the model (and therefore the arc lengths) have been updated, we need to checks that the waypoints
            ! are not too close to each other
            CALL FILAMENT_REMOVE_WAYPOINTS_TOO_CLOSE(SELF)
        ELSE
            WRITE(*,'(A,I0,A)') '**warning(FILAMENT_RESAMPLE_WAYPOINTS): resampling would have led to less than ',MIN_NUM_WPS,&
                            ' waypoints. Will do nothing.'
        ENDIF

        ! Diagnostics
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_RESAMPLE_WAYPOINTS): filament info on exit: '
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF
    END SUBROUTINE FILAMENT_RESAMPLE_WAYPOINTS

    !>  \brief  Remove waypoints which are too close to each other
    SUBROUTINE FILAMENT_REMOVE_WAYPOINTS_TOO_CLOSE(SELF)
        USE COOS, ONLY : WAYPOINTS_REMOVE_TOO_CLOSE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Private variables
        TYPE(WAYPOINT), ALLOCATABLE ::  WPS_TMP(:)
        TYPE(FILAMENT)              ::  SELF_TMP
        ! Start work
        WPS_TMP = SELF%WP
        CALL FILAMENT_INIT(SELF_TMP,waypoints_in=SELF%WP,MOULD=SELF)
        CALL WAYPOINTS_REMOVE_TOO_CLOSE(WPS_TMP)
        CALL FILAMENT_INIT(SELF,waypoints_in=WPS_TMP,MOULD=SELF_TMP)
    END SUBROUTINE FILAMENT_REMOVE_WAYPOINTS_TOO_CLOSE



    !>  \brief  Initialise a filament by giving it a set of coordinates
    !!
    !!
    !!  As of 9-Jun-2010, option ADD_EXTRA_FAL_WPS does not appear to work on wobbly filaments and will probably not be refined/tested further. As of 20-Jul-2010, this is even more so,
    !!  because I do not intend to fully test the 3D spline upgrade in combination with this option. Sep-2011: significant fixes to the spline extrapolation routines may mean guessing
    !!  FAL waypoins works again, but I have not tested this. Jul-2012: add_extra_fal_wps should work.
    !!
    !!  \todo   Simplify this subroutine or fragment it. See for example mode_estimate_pl_slave for 4 lines to initialise a simple filament without helix.
    SUBROUTINE FILAMENT_INIT(   SELF, &
                                points_in, waypoints_in, &
                                FILM_IN,CROSSOVERS,HELIX,ID,ADD_EXTRA_FAL_WPS,FAL_WPS_ARE_CROSSOVERS,MAX_RADIUS,    &
                                MICROTUBULE,MT_PF_NUMBER,MT_RISE,REFINE_PARAMS,RIGID_PARAMS,RIGID_WP,   &
                                CONSTANT_RISE_PER_SUBUNIT,CONSTANT_TWIST_PER_LENGTH,    &
                                WPS_AS_SPS,PERSISTENCE_LENGTH,TORSION_PERSISTENCE_LENGTH,STRETCH_PERSISTENCE_LENGTH,        &
                                FAL_WPS_IN_SCORE,FAL_WPS_IN_THREED,   &
                                REFINE_NORMAL_MODES,AVERAGE_Z_TO_CTFTILT_PLANE,USE_HERMITE_SPLINES, &
                                FRAGMENTED_PHI, FRAGMENTED_SUBUNIT_NUM, AMYLOID_BETA,MOULD)
		USE COOS
		USE FILMS, ONLY : FILM_FIND_Z_FROM_XY, FILM_PRINT_INFO
		USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
		USE FILAMENTS_NORMAL_MODES, ONLY : FILAMENT_WPS_TO_NM_AMPLITUDES, FILAMENT_UPDATE_NMS
		USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_UPDATE_WP_RES, FILAMENT_FIT_SPLINES_LINEAR, FILAMENT_EVAL_TOTAL_CURVATURE
		USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_SPLINES_ARE_MONOTONIC, FILAMENT_RELAX_TWIST_PER_SUBUNIT, FILAMENT_RELAX_RISE_PER_SUBUNIT
		USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_CHECK_SUBUNIT_NUM, FILAMENT_PRINT_INFO, FILAMENT_GUESS_WPS_PHI
		USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_CHECK_SUBUNIT_NUM_INCREASES
		USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_RIGID_WP, FILAMENT_UPDATE_AVERAGE_HELIX_TWIST, FILAMENT_UPDATE_WP_HELIX
		USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_AVERAGE_HELIX_RISE
		USE FILAMENTS_MODEL_MODIFY, ONLY : FILAMENT_SHIFT_AVERAGE_Z_TO_CTFTILT_PLANE
		USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),         TARGET,     INTENT(INOUT)   ::  SELF                                !<  Filament object to populate
        type(point),            optional,   INTENT(IN)      ::  points_in(:)                        !<  Array of points
        type(waypoint),         optional,   intent(in)      ::  waypoints_in(:)                     !<  Array of waypoints
        TYPE(FILM),             OPTIONAL,   INTENT(IN)      ::  FILM_IN                             !<  A film object (already initiated) from which the film properties will be inherited
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  CROSSOVERS                          !<  Whether the points given are crossovers (defaults to .FALSE.). If so and no angles are given, they will be set to 0/180/360/...
        TYPE(HELICAL_SYMMETRY), OPTIONAL,   INTENT(IN)      ::  HELIX                               !<  (Average) helical parameters for the filament.
        INTEGER,                OPTIONAL,   INTENT(IN)      ::  ID
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  ADD_EXTRA_FAL_WPS                   !<  Add extra waypoints before/after the first-and-last (FAL) provided waypoints (WPS).
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  FAL_WPS_ARE_CROSSOVERS              !<  First and last (FAL) waypoints (WPS) are also crossovers, i.e. their twist angles are known.
                                                                                                    !!  By default, all WPS are crossovers (if CROSSOVERS was .TRUE. of course), so this option is true.
        REAL,                   OPTIONAL,   INTENT(IN)      ::  MAX_RADIUS                          !<  Maximum filament radius (pixels)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  MICROTUBULE                         !<  Whether the filament is a microtubule
        INTEGER,                OPTIONAL,   INTENT(IN)      ::  MT_PF_NUMBER                        !<  In the case of microtubules, the number of protofilaments
        INTEGER,                OPTIONAL,   INTENT(IN)      ::  MT_RISE                             !<  How much further along the filament (distance measured in tubulin subunits) the end of a MT_PF_NUMBER-element "washer"
                                                                                                    !!  is compared to its beginning. The 13-pf microtobules have a value of 3.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  REFINE_PARAMS(7)                    !<  Which of the 7 degrees of freedom to refine per way point (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi,7.Subunit_num)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  RIGID_PARAMS(7)                     !<  Whether to keep the following parameters fixed throughout the filament: 1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta, 6.Psi, 7-H.Rise, 8-H.Twist.
                                                                                                    !!  If all are set to .TRUE., the filament is a rigid rod, treated as a single particle.
        TYPE(WAYPOINT),         OPTIONAL,   INTENT(IN)      ::  RIGID_WP                            !<  Rigid waypoint
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  CONSTANT_RISE_PER_SUBUNIT           !<  Whether we'll assume the rise-per-subunit to be constant (i.e. no cumulative axial disorder)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  CONSTANT_TWIST_PER_LENGTH           !<  Wether we should assume a constant twist per unit length along the filament
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  WPS_AS_SPS                          !<  Whether waypoints are treated as independent single particles
        REAL,                   OPTIONAL,   INTENT(IN)      ::  PERSISTENCE_LENGTH                  !<  Persistence length (pixels)
        REAL,                   OPTIONAL,   INTENT(IN)      ::  TORSION_PERSISTENCE_LENGTH          !<  Torsional persistence length (pixels)
        REAL,                   OPTIONAL,   INTENT(IN)      ::  STRETCH_PERSISTENCE_LENGTH          !<  Stretching persistence length (pixels)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  FAL_WPS_IN_SCORE                    !<  If false, the parts of the filament between the first and the second as well as between the penultimate and
                                                                                                    !!  last waypoints will not be included in the scoring function
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  FAL_WPS_IN_THREED                   !<  If false, the parts of the filament between the first and the second as well as between the penultimate and
                                                                                                    !!  last waypoints will not be included in the 3D reconstruction
        !LOGICAL,                OPTIONAL,   INTENT(IN)      ::  REFINE_IN_FOURIER_SPACE             !<  Whether the filament model or its Fourier transform should be refined
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  REFINE_NORMAL_MODES(7)              !<  Whether normal mode amplitudes should be refined, instead of the real-space filament model
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  AVERAGE_Z_TO_CTFTILT_PLANE          !<  Shift the filament in Z so that its average lies on the plane described by the CTFTILT plane
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  USE_HERMITE_SPLINES                 !<  Use Hermite splines (i.e. use theta,psi to constrain the space curve)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  FRAGMENTED_PHI                      !<  There are discontinuities in the phi function between waypoints
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  FRAGMENTED_SUBUNIT_NUM              !<  There are discontinuities in the subunit_num function between waypoints
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  AMYLOID_BETA                        !<  The filament is an amyloid beta fibril
        TYPE(FILAMENT),         OPTIONAL,   INTENT(IN)      ::  MOULD                               !<  Another filament object from which properties will be copied unless they are specified by other arguments
                                                                                                    !!  (allocatable components, i.e. waypoints, splines and films, are not copied over). Cannot be the same as SELF.
        ! Private variables
        INTEGER                 ::  I
        LOGICAL                 ::  AADD_EXTRA_FAL_WPS
        LOGICAL                 ::  ESTIMATE_PHI
        INTEGER                 ::  MIN_NUM_POINTS
        LOGICAL                 ::  Z_UNKNOWN
        INTEGER                 ::  TWIST_SIGN
        LOGICAL                 ::  WWPS_AS_SPS
        LOGICAL                 ::  FAL_WPS_PHI_TBD !<  Whether the first-and-last waypoints' phi angles are to be determined
        LOGICAL,    PARAMETER   ::  DEBUG           = .FALSE.
        LOGICAL,    PARAMETER   ::  RRELAX_HELIX    = .TRUE.
        INTEGER                 ::  FIRST_XOVER_WP,LAST_XOVER_WP
        LOGICAL                 ::  KNOWN_PSI_THETA
        LOGICAL                 ::  AAVERAGE_Z_TO_CTFTILT_PLANE
        TYPE(WAYPOINT), ALLOCATABLE :: WP_BAK(:)
        CHARACTER(LEN=6)        ::  NWP_C
        type(point),    allocatable ::  points(:)
        type(waypoint), allocatable ::  waypoints(:)
        REAL                    ::  TOTAL_CURVATURE
        INTEGER                 ::  ID_COPY
        IF (PRESENT(ID)) ID_COPY = ID


        ! Start work

        !
        ! By default
        !
        ESTIMATE_PHI = .FALSE.
        FAL_WPS_PHI_TBD = .FALSE.
        AAVERAGE_Z_TO_CTFTILT_PLANE = .FALSE.
        IF (PRESENT(AVERAGE_Z_TO_CTFTILT_PLANE)) AAVERAGE_Z_TO_CTFTILT_PLANE = AVERAGE_Z_TO_CTFTILT_PLANE
        CALL FILAMENT_PROPERTIES_RESET(SELF)


        !
        ! Do we have a mould?
        !
        IF (PRESENT(MOULD)) CALL FILAMENT_COPY_PROPERTIES(SELF,MOULD)


        !
        ! If optional parameters were given, heed them
        !
        IF (PRESENT(ID))                        SELF%ID                         = ID_COPY
        IF (PRESENT(WPS_AS_SPS))                WWPS_AS_SPS                     = WPS_AS_SPS
        IF (PRESENT(CROSSOVERS))                SELF%WPS_ARE_CROSSOVERS         = CROSSOVERS
        IF (PRESENT(FAL_WPS_ARE_CROSSOVERS)) THEN
                                                SELF%FAL_WPS_ARE_CROSSOVERS     = FAL_WPS_ARE_CROSSOVERS
        ELSE
            IF (.NOT. PRESENT(MOULD))           SELF%FAL_WPS_ARE_CROSSOVERS     = SELF%WPS_ARE_CROSSOVERS
        ENDIF
        IF (PRESENT(FAL_WPS_IN_SCORE))          SELF%FAL_WPS_IN_SCORE           = FAL_WPS_IN_SCORE
        IF (PRESENT(FAL_WPS_IN_THREED))         SELF%FAL_WPS_IN_THREED          = FAL_WPS_IN_THREED
        IF (PRESENT(MICROTUBULE))               SELF%MICROTUBULE                = MICROTUBULE
        IF (PRESENT(AMYLOID_BETA))              SELF%AMYLOID_BETA               = AMYLOID_BETA
        IF (PRESENT(REFINE_PARAMS))             SELF%REFINE_PARAMS              = REFINE_PARAMS
        IF (PRESENT(RIGID_PARAMS)) THEN
                                                SELF%RIGID_PARAMS               = RIGID_PARAMS
        ELSE
            IF (.NOT. PRESENT(MOULD)) THEN
                WRITE(*,'(2A)') '**info(FILAMENT_INIT): no rigid parameters supplied, overriding with flexible space curve ',    &
                                '& flexible helical parameters'
                SELF%RIGID_PARAMS = .FALSE.
            ENDIF
        ENDIF
        IF (PRESENT(RIGID_WP))                  SELF%RIGID_WP = RIGID_WP
        IF (PRESENT(CONSTANT_RISE_PER_SUBUNIT)) SELF%CONSTANT_RISE_PER_SUBUNIT  = CONSTANT_RISE_PER_SUBUNIT
        IF (PRESENT(CONSTANT_TWIST_PER_LENGTH)) SELF%CONSTANT_TWIST_PER_LENGTH  = CONSTANT_TWIST_PER_LENGTH
        IF (PRESENT(FILM_IN))                   SELF%FILM                       = FILM_IN
        IF (PRESENT(PERSISTENCE_LENGTH))        SELF%PERSISTENCE_LENGTH         = PERSISTENCE_LENGTH
        IF (PRESENT(TORSION_PERSISTENCE_LENGTH))SELF%TORSION_PERSISTENCE_LENGTH = TORSION_PERSISTENCE_LENGTH
        IF (PRESENT(STRETCH_PERSISTENCE_LENGTH))SELF%STRETCH_PERSISTENCE_LENGTH = STRETCH_PERSISTENCE_LENGTH
        IF (PRESENT(MAX_RADIUS)) THEN
                                                SELF%MAX_RADIUS                 = MAX_RADIUS
        ELSE
            IF (.NOT. PRESENT(MOULD)) THEN
                WRITE(*,'(A)') '**warning(FILAMENT_INIT): maximum filament radius not given'
                !CALL TERMINATE('Debug FILAMENT_INIT, no max radius given')
            ENDIF
        ENDIF
        IF (PRESENT(REFINE_NORMAL_MODES))       SELF%REFINE_NORMAL_MODES        = REFINE_NORMAL_MODES
        IF (PRESENT(USE_HERMITE_SPLINES))       SELF%USE_HERMITE_SPLINES        = USE_HERMITE_SPLINES
        IF (PRESENT(FRAGMENTED_PHI))            SELF%FRAGMENTED_PHI             = FRAGMENTED_PHI
        IF (PRESENT(FRAGMENTED_SUBUNIT_NUM))    SELF%FRAGMENTED_SUBUNIT_NUM     = FRAGMENTED_SUBUNIT_NUM

        IF (DEBUG) THEN
            WRITE(*,'(A,8L1)') '**debug(FILAMENT_INIT): rigid parameters: ', SELF%RIGID_PARAMS
            WRITE(*,'(A,2(I0,1X))') '**debug(FILAMENT_INIT): filament, film id = ', SELF%ID, SELF%FILM%ID
        ENDIF


        ! How many points do we need to define a filament?
        IF (.NOT. (SELF%FAL_WPS_IN_SCORE .AND. SELF%FAL_WPS_IN_THREED)) THEN
            MIN_NUM_POINTS = 4 ! If the first and last waypoints are not crossovers, later down the line we'll want to reconstruct from the 2nd waypoint to the 2nd-to-last waypoint
        ELSE
            MIN_NUM_POINTS = 3
        ENDIF

        !
        if (present(points_in)) then

            ! Copy input points
            allocate(points(size(points_in)))
            points = points_in
            ! Remove duplicates
            call points_remove_duplicates(points)
            ! Check we have the required number of points to define the filament. If we do, remember.
            if (size(points) .eq. 1) then
                write(*,'(a,i0,a,i0,a,i0,a)') '**error(filament_init): insufficient number of unique input points (',&
                                                size(points), &
                                                ') for filament ', self%id, '. ',min_num_points,'+ points needed'
                call terminate('fatal error in filament_init')
            else if (size(points) .lt. min_num_points) then
                write(*,'(a,i0,a,i0,a)') '**warning(filament_init): insufficient number of unique input points (', &
                                                size(points), &
                                                ') for filament ', self%id, '. adding points by linear interpolation.'
                call points_add_by_linear_interpolation(points)
            endif
            ! Remember how many waypoints we have in the filament
            SELF%NWP = SIZE(POINTS)

        else if (present(waypoints_in)) then

            ! Copy input waypoints
            allocate(waypoints(size(waypoints_in)))
            waypoints = waypoints_in
            ! Remove duplicates
            call waypoints_remove_duplicates(waypoints)
            ! Check we have the required number of points to define the filament. If we do, remember.
            if (size(waypoints) .eq. 1) then
                write(*,'(a,i0,a,i0,a,i0,a)') '**error(filament_init): insufficient number of unique input points (',&
                                                size(waypoints), &
                                                ') for filament ', self%id, '. ',min_num_points,'+ points needed'
                call terminate('fatal error in filament_init')
            else if (size(waypoints) .lt. min_num_points) then
                write(*,'(a,i0,a,i0,a)') '**warning(filament_init): insufficient number of unique input points (',size(waypoints),&
                                                ') for filament ', self%id, '. adding points by linear interpolation.'
                call waypoints_add_by_linear_interpolation(waypoints)
            endif
            ! Remember how many waypoints we have in the filament
            SELF%NWP = SIZE(WAYPOINTS)
        else
            write(*,'(a)') '**ERROR(FILAMENT_INIT): need to supply either points or waypoints'
            call TERMINATE('Fatal error in FILAMENT_INIT')
        endif ! end of test for presence of points_in or waypoints_in

        ! We need this for some write format statements
        WRITE(NWP_C,'(I6)') SELF%NWP

        ! If we're not dealing with crossovers, but the user specified FAL_WPS_NOT_XOVERS, something is wrong.
        IF (.NOT. SELF%WPS_ARE_CROSSOVERS) THEN
            IF (SELF%FAL_WPS_ARE_CROSSOVERS) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_INIT): FAL waypoints can be crossovers only if the others are too'
                CALL TERMINATE('Fatal error in FILAMENT_INIT')
            ENDIF
        ENDIF

        ! Are we adding extra waypoints?
        IF (PRESENT(ADD_EXTRA_FAL_WPS)) THEN
            AADD_EXTRA_FAL_WPS = ADD_EXTRA_FAL_WPS
        ELSE
            AADD_EXTRA_FAL_WPS = .FALSE.
        ENDIF


        !
        ! Microtubule-specific stuff
        !
        IF (SELF%MICROTUBULE) THEN
            IF (PRESENT(MT_PF_NUMBER)) THEN
                SELF%MT_PF_NUMBER = MT_PF_NUMBER
            ENDIF
            IF (.NOT. PRESENT(MT_PF_NUMBER) .AND. SELF%MT_PF_NUMBER == 0) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_INIT): microtubule number of protofilaments not supplied'
                CALL TERMINATE('Fatal error in FILAMENT_INIT')
            ENDIF
            IF (PRESENT(MT_RISE)) THEN
                SELF%MT_RISE = MT_RISE
            ENDIF
            IF (.NOT. PRESENT(MT_RISE) .AND. SELF%MT_RISE == 0) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_INIT): microtubule rise not supplied'
                CALL TERMINATE('Fatal error in FILAMENT_INIT')
            ENDIF
        ENDIF


        !
        ! If we were given helical parameters, copy them over
        !
        IF (PRESENT(HELIX)) THEN
            SELF%HELIX = HELIX
!            SELF%WP%HELIX = HELIX   ! Is it correct to do this here? May be better off computing this later...
!                                    ! E.g. in FILAMENT_UPDATE_MODEL?
            IF (SELF%RIGID_PARAMS(7)) THEN
                SELF%RIGID_WP%HELIX = HELIX ! Is it correct to do this here?
                                            ! It might be more genrenal to do this in FILAMENT_UPDATE_MODEL by taking
                                            ! the average of the waypoint helical parameters, or some sort of average
                                            ! over the whole filament?
            ENDIF
        ELSE
            IF (.NOT. PRESENT(MOULD)) THEN
                WRITE(*,'(A)') '**info(FILAMENT_INIT): helical parameters not supplied'
                CALL TERMINATE('Debug stop - no helical parameters?')
            ENDIF
        ENDIF

        ! If crossovers, we can't assume constant twist per subunit!
        IF (SELF%WPS_ARE_CROSSOVERS) THEN
            SELF%CONSTANT_TWIST_PER_LENGTH = .FALSE.
        ENDIF

        !
        ! What is the sign of the twist?
        !
        IF (MOD(SELF%HELIX%TWIST,180.D0) .GT. 0.D0) THEN
            TWIST_SIGN = 1
        ELSE
            TWIST_SIGN = -1
        ENDIF
        IF (DEBUG) WRITE(*,'(A,F8.2,1X,I2)') '**debug(FILAMENT_INIT): twist, twist_sign = ', SELF%HELIX%TWIST, TWIST_SIGN


        !
        ! Check whether we're given points (just coordinates) or waypoints (coordinates & angles)
        !
        IF (PRESENT(WAYPOINTS_IN)) THEN
            !
            ! Waypoints were supplied, which contain angle as well as X,Y and Z coordinates etc.
            !

            IF (DEBUG) THEN
                WRITE(*,'(A)') '**debug(FILAMENT_INIT): waypoints info on entry:'
                CALL WAYPOINTS_PRINT_INFO(WAYPOINTS)
            ENDIF

            ! Remove waypoints which are too close together (in terms of subunit num)
            ALLOCATE(WP_BAK(SIZE(WAYPOINTS)),SOURCE=WAYPOINTS)
            WP_BAK = WAYPOINTS
            CALL WAYPOINTS_REMOVE_TOO_CLOSE(WP_BAK)
            IF (DEBUG) THEN
                WRITE(*,'(A)') '**debug(FILAMENT_INIT): waypoints after remove_too_close:'
                CALL WAYPOINTS_PRINT_INFO(WP_BAK)
            ENDIF

            ! Allocate memory for the filament
            IF (AADD_EXTRA_FAL_WPS) THEN
                ! Allocate memory
                CALL FILAMENT_ALLOCATE(SELF,SIZE(WP_BAK)+2)
                ! Copy coordinates over
                FORALL(I=1:SIZE(WP_BAK))SELF%WP(I+1) = WP_BAK(I)
            ELSE
                ! Allocate memory
                CALL FILAMENT_ALLOCATE(SELF,SIZE(WP_BAK))
                ! Copy coordinates over
                FORALL(I=1:SELF%NWP) SELF%WP(I) = WP_BAK(I)
            ENDIF


            ! Check that the supplied waypoints are indeed crossovers, if that's what the arguments suggest
            IF (SELF%WPS_ARE_CROSSOVERS) THEN
                IF (SELF%FAL_WPS_ARE_CROSSOVERS) THEN
                    FIRST_XOVER_WP = 1
                    LAST_XOVER_WP = SELF%NWP
                ELSE
                    FIRST_XOVER_WP = 2
                    LAST_XOVER_WP = SELF%NWP - 1
                ENDIF
                DO I=FIRST_XOVER_WP,LAST_XOVER_WP
                    IF (.NOT. WAYPOINT_IS_CROSSOVER(WAYPOINTS(I))) THEN
                        WRITE(*,'(A,I0,A,F0.5)')    '**ERROR(FILAMENT_INIT): you supplied waypoint ', I, &
                                                    ' saying it was a crossover, but it has a phi angle of ', &
                                                     WAYPOINTS(I)%EULER%PHI
                        CALL TERMINATE('Fatal error in FILAMENT_INIT')
                    ENDIF
                ENDDO
            ENDIF


            ! Carry all information from waypoints to filament
            IF (AADD_EXTRA_FAL_WPS) THEN
                DO I=1,SIZE(WAYPOINTS)
                    SELF%WP(I+1) = WP_BAK(I)
                ENDDO
            ELSE
                SELF%WP = WP_BAK
            ENDIF
            SELF%WP%HELIX = SELF%HELIX

            !
            ! Do we know any of the Z coordinates? (we expect that we know Z coordinates,
            ! but if we don't, we'll guess them from the film information)
            !
            IF (ANY(SELF%WP%COO(3) .NE. 0.0)) THEN
                Z_UNKNOWN = .FALSE.
            ELSE
                WRITE(*,'(A)') '**warning(FILAMENT_INIT): unexpectedly, supplied waypoints had Z coordinates = 0.0. '//&
                                    'Perhaps you supplied an old parameter file?'
                WRITE(*,'(A)') '**info(FILAMENT_INIT): Z coordinates will be guessed for each waypoint using its X,Y '//&
                                    'coordinates and the CTFTILT/CTFFIND information'
                Z_UNKNOWN = .TRUE.
            ENDIF
            !
            ! If necessary, work out Z coordinates from film properties (the film may have CTFTILT information, for example)
            !
            IF (Z_UNKNOWN) THEN
                WRITE(*,'(A)') '**info(FILAMENT_INIT): assigning Z coordinates to waypoints:'
                IF (PRESENT(FILM_IN)) THEN
                    IF (ANY(SELF%FILM%logical_dimensions .NE. 0)) THEN
                        IF (DEBUG) THEN
                            WRITE(*,'(A)') '**debug(FILAMENT_INIT): about to compute Z coordinates using this film info:'
                            CALL FILM_PRINT_INFO(SELF%FILM)
                        ENDIF
                        DO I=1,SELF%NWP
                            SELF%WP(I)%COO(3) = FILM_FIND_Z_FROM_XY(SELF%FILM,SELF%WP(I)%COO(1),SELF%WP(I)%COO(2))
                        ENDDO
                        WRITE(*,'(A,'//NWP_C//'(F15.3,1X))') 'Z values: ', SELF%WP(:)%COO(3)
                    ELSE
                        WRITE(*,'(2A)') '**ERROR(FILAMENT_INIT): cannot calculate Z coordinates ', &
                                        'of points with unknown film dimensions'
                        CALL TERMINATE('Fatal error in FILAMENT_INIT')
                    ENDIF
                ELSE
                    WRITE(*,'(A)') '**ERROR(FILAMENT_INIT): cannot calculate Z coordinates of points without a film object'
                    CALL TERMINATE('Fatal error in FILAMENT_INIT')
                ENDIF
            ENDIF

            IF (DEBUG) THEN
                WRITE(*,'(A)') '**debug(FILAMENT_INIT): before calling update model: '
                CALL FILAMENT_PRINT_INFO(SELF)
            ENDIF

            !
            ! Trigger a spline fit (known_d) (if we're adding extra first and last WPS, this will also calculate their coordinates)
            !
            KNOWN_PSI_THETA = ALL(SELF%WP%EULER%PSI .EQ. 0.0) .AND. ALL(SELF%WP%EULER%THETA .EQ. 0.0)
            KNOWN_PSI_THETA = .NOT. KNOWN_PSI_THETA
            CALL FILAMENT_UPDATE_MODEL(SELF,&!KNOWN_PSI=KNOWN_PSI_THETA,KNOWN_THETA=KNOWN_PSI_THETA,  &
                                        GUESS_FAL_WP_PARAMS=AADD_EXTRA_FAL_WPS,WPS_AS_SPS=WWPS_AS_SPS)

            IF (DEBUG) THEN
                WRITE(*,'(A)') '**debug(FILAMENT_INIT): after calling update model: '
                CALL FILAMENT_PRINT_INFO(SELF)
            ENDIF

        ELSE IF (PRESENT(POINTS_IN)) THEN
            !
            ! In the default case, only coordinates (no angles) were given
            !
            IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_INIT): coordinates (not waypoints) were supplied'


            !
            ! If the user wants extra first/last wps, they need to give us helical paramters. Otherwise,
            ! we can't guess what the Phi angles are for the waypoints we're adding.
            !
            IF (AADD_EXTRA_FAL_WPS .OR. .NOT. SELF%WPS_ARE_CROSSOVERS) THEN
                IF (.NOT. PRESENT(HELIX)) THEN
                    WRITE(*,'(2A)') '**ERROR(FILAMENT_INIT): cannot guess first/last waypoints ',&
                                    'parameters without helical parameters'
                    CALL TERMINATE('Fatal error in FILAMENT_INIT')
                ENDIF
            ENDIF


            !
            ! Allocate memory for the filament object
            !
            IF (AADD_EXTRA_FAL_WPS) THEN

                ! Allocate memory
                CALL FILAMENT_ALLOCATE(SELF,SIZE(POINTS)+2)
                ! Copy coordinates over
                DO  I=1,SIZE(POINTS)
                    self%wp(i+1) = points(i)
                ENDDO
            ELSE
                ! Allocate memory
                CALL FILAMENT_ALLOCATE(SELF,SIZE(POINTS))
                ! Copy coordinates over
                DO  I=1,SIZE(POINTS)
                    self%wp(i) = points(i)
                ENDDO
            ENDIF


            ! Do we know any of the Z coordinates? (since users pick waypoints on 2D micrographs we don't expect so, but let's check)
            IF (ANY(SELF%WP(:)%COO(3) .NE. 0.0)) THEN
                Z_UNKNOWN = .FALSE.
                WRITE(*,'(A)') '**WARNING(FILAMENT_INIT): unexpectedly, supplied points (which were not waypoints) '//&
                                'had non-zero Z coordinates. Programming error?'
                WRITE(*,'(A,'//NWP_C//'(F0.3,1X))') 'Z values: ', SELF%WP(:)%COO(3)
            ELSE
                Z_UNKNOWN = .TRUE.
            ENDIF

            ! If necessary, work out Z coordinates from film properties (the film may have CTFTILT information, for example)
            IF (Z_UNKNOWN) THEN
                !PRINT *, '**debug(FILAMENT_INIT): assigning Z coordinates to waypoints'
                IF (PRESENT(FILM_IN)) THEN
                    DO I=1,SELF%NWP
                        SELF%WP(I)%COO(3) = FILM_FIND_Z_FROM_XY(SELF%FILM,SELF%WP(I)%COO(1),SELF%WP(I)%COO(2))
                    ENDDO
                ELSE
                    WRITE(*,'(A)') '**ERROR(FILAMENT_INIT): cannot calculate Z coordinates of points without a film object'
                    CALL TERMINATE('Fatal error in FILAMENT_INIT')
                ENDIF
            ENDIF

            !
            ! Work out Euler angle PHI (twist) for the waypoints
            !
            IF (AADD_EXTRA_FAL_WPS) THEN
                WRITE(*,'(2A)') '**warning(FILAMENT_INIT): option to add extra first-and-last waypoints',    &
                                ' is experimental and seldom tested'
                ! Phi (twist) angle: if crossovers, give first way-point 0 then add/subtract 180 to each subsequent way point
                IF (SELF%WPS_ARE_CROSSOVERS) THEN
                    SELF%WP(2)%EULER%PHI = 0.0
                    DO I=3,SELF%NWP-1
                        SELF%WP(I)%EULER%PHI = SELF%WP(I-1)%EULER%PHI + 180.0 * TWIST_SIGN
                    ENDDO
                ELSE
                    SELF%WP(1:SELF%NWP)%EULER%PHI = 0.0
                    IF (.NOT. PRESENT(HELIX)) THEN
                        WRITE(*,'(A)') '**WARNING(FILAMENT_INIT): helical parameters not provided, not crossovers. ',   &
                                        'Will not be able to assign Phi angles.'
                        CALL TERMINATE('Unable to assign phi angles')
                    ELSE
                        ESTIMATE_PHI = .TRUE.
                    ENDIF
                ENDIF
            ELSE ! We're not adding extra points
                IF (SELF%WPS_ARE_CROSSOVERS) THEN
                    IF (TWIST_SIGN .EQ. -1) THEN
                        PRINT '(A)', '**info(FILAMENT_INIT): assuming right-handed helix with crossovers, '//&
                                        'so that first xover has phi(1)=0.0, and each subsequent xover has phi(i+1)'//&
                                        ' = phi(i)-180. If this is not the case, results will be wrong.'
                    ELSE
                        PRINT '(A)', '**info(FILAMENT_INIT): assuming left-handed helix with crossovers, '//&
                                        'so that first xover has phi(1)=0.0, and each subsequent xover has phi(i+1)'//&
                                        ' = phi(i)+180. If this is not the case, results will be wrong.'
                    ENDIF
                    IF (SELF%FAL_WPS_ARE_CROSSOVERS) THEN
                        WRITE(*,'(A)') '**info(FILAMENT_INIT): all waypoints are crossovers'
                        ! Phi (twist) angle: if crossovers, give first way-point 0 then add (or subtract, if left-handed) 180 to each subsequent way point
                        SELF%WP(1)%EULER%PHI = 0.0
                        DO I=2,SELF%NWP
                            SELF%WP(I)%EULER%PHI = SELF%WP(I-1)%EULER%PHI + 180.0 * TWIST_SIGN
                        ENDDO
                    ELSE
                        WRITE(*,'(A)') '**info(FILAMENT_INIT): first and last waypoints are not crossovers. All other are.'
                        SELF%WP(2)%EULER%PHI = 0.0
                        DO I=3,SELF%NWP-1
                            SELF%WP(I)%EULER%PHI = SELF%WP(I-1)%EULER%PHI + 180.0 * TWIST_SIGN
                        ENDDO
                        IF (DEBUG) WRITE(*,'(2A)') '**debug(FILAMENT_INIT): crossover waypoints have phi already, but the',  &
                                                 ' FAL WPS phi values will have to be determined later by filament_update_model'
                        FAL_WPS_PHI_TBD = .TRUE.
                    ENDIF
                ELSE
                    SELF%WP(1:SELF%NWP)%EULER%PHI = 0.0
                    IF (.NOT. PRESENT(HELIX)) THEN
                        WRITE(*,'(2A)') '**WARNING(FILAMENT_INIT): helical parameters not provided, not crossovers. ',   &
                                        'Will not be able to assign Phi angles.'
                        CALL TERMINATE('Unable to assign phi angles')
                    ELSE
                        ESTIMATE_PHI = .TRUE.
                    ENDIF
                ENDIF
            ENDIF

            ! Update the filament model
            CALL FILAMENT_UPDATE_MODEL( SELF,GUESS_FAL_WP_PARAMS=AADD_EXTRA_FAL_WPS, &
                                        WPS_AS_SPS=WWPS_AS_SPS,GUESS_FAL_WP_EULER=FAL_WPS_PHI_TBD)

            ! In case we haven't been able to work out PHI yet, do so now (note that if Phi is one of the rigid parameters, FILAMNET_UPDATE_MODEL will have taken care of it
            IF (ESTIMATE_PHI .AND. .NOT. SELF%RIGID_PARAMS(4)) THEN
                ! Debug
                IF (DEBUG) THEN
                    WRITE (*,'(A)') '**debug(FILAMENT_INIT): filament info before guessing WPS_PHI'
                    CALL FILAMENT_PRINT_INFO(SELF)
                ENDIF
                ! Now that we have the splines, we can make a guess at the phi angles
                CALL FILAMENT_GUESS_WPS_PHI(SELF)
            ENDIF

            !
            ! Since we now have a first estimate for PSI & THETA, we'll fix it and we can refine it later
            !
            SELF%X_SPLINE%KNOWN_D = .TRUE.
            SELF%Y_SPLINE%KNOWN_D = .TRUE.
            SELF%Z_SPLINE%KNOWN_D = .TRUE.

        else
            call terminate('Fatal error in filament_init: need either points or waypoints')
        ENDIF  ! End of selection as to whether user supplied waypoints or regular points


        !
        IF (AAVERAGE_Z_TO_CTFTILT_PLANE) THEN
            CALL FILAMENT_SHIFT_AVERAGE_Z_TO_CTFTILT_PLANE(SELF)
        ENDIF


        ! Now that we have proper splines and path length information, let's refine our helical parameters per WP,
        ! unless the helical parameters and/or phi are to be kept rigid
        IF (PRESENT(HELIX) .OR. PRESENT(MOULD)) THEN
            IF (.NOT. (SELF%RIGID_PARAMS(4) .OR. SELF%RIGID_PARAMS(7)) .AND. .NOT. WWPS_AS_SPS) THEN
                ! Work out the rise per subunit for each segment.
                CALL FILAMENT_UPDATE_WP_HELIX(SELF)
                ! Work out the average twist per subunit. This is useful later on when a reference with appropriate twist needs to be generated.
                CALL FILAMENT_UPDATE_AVERAGE_HELIX_TWIST(SELF)
            ENDIF
        ENDIF

        IF (DEBUG) THEN
            WRITE(*,'(A,'//NWP_C//'L1)') '**debug(FILAMENT_INIT): final self%wp(:)%helix%pixel_units = ', &
                                        self%wp(:)%helix%pixel_units
        ENDIF

        ! Update the rigid waypoint
        CALL FILAMENT_UPDATE_RIGID_WP(SELF)

        ! Check the subunit_num values
        CALL FILAMENT_CHECK_SUBUNIT_NUM(SELF)

        ! Fit the splines & compute residuals
        CALL FILAMENT_FIT_SPLINES_LINEAR(SELF)
        CALL FILAMENT_UPDATE_WP_RES(SELF)

        ! Compute the normal modes
        CALL FILAMENT_UPDATE_NMS(SELF)
        CALL FILAMENT_WPS_TO_NM_AMPLITUDES(SELF)

        ! Check the total curvature - it should have a reasonable value
        TOTAL_CURVATURE = FILAMENT_EVAL_TOTAL_CURVATURE(SELF,SQUARED_CURVATURE=.TRUE.)
        IF (DEBUG) WRITE(*,'(A,G15.4)') '**debug(FILAMENT_INIT): total curvature = ', TOTAL_CURVATURE
        IF (TOTAL_CURVATURE .GT. 5.0) THEN
            WRITE(*,'(A,G15.4)') '**WARNING(FILAMENT_INIT): filament appears to have extreme curvature: ', TOTAL_CURVATURE
            WRITE(*,'(A)') '**WARNING(FILAMENT_INIT): maybe there is something funny about the input coordinates?'
            !CALL FILAMENT_PRINT_INFO(SELF)
            !CALL TERMINATE('Fatal error in FILAMENT_INIT')
        ENDIF


        ! Iteratively relax the helix and update the average helical parameters for the filament
        DO I=1,2
            IF (DEBUG) THEN
                WRITE(*,'(A,I0,A)') '**debug(FILAMENT_INIT): relaxing helix, iteration ', I, ' of 2, filament info: '
                CALL FILAMENT_PRINT_INFO(SELF)
            ENDIF
            IF (RRELAX_HELIX) THEN
                !WRITE(*,'(A)') '**info(FILAMENT_INIT): relaxing helix (experimental)'
                !CALL FILAMENT_RELAX_HELIX(SELF)
                IF (.NOT. (SELF%FAL_WPS_IN_THREED .OR. SELF%FAL_WPS_IN_SCORE)) THEN
                    WRITE(*,'(2A)') '**info(FILAMENT_INIT): relaxing rise & twist for first and last segments,', &
                                    ' since they are not used in refinement or 3D reconstruction'
                    CALL FILAMENT_RELAX_RISE_PER_SUBUNIT(SELF,FAL_ONLY=.TRUE.)
                    CALL FILAMENT_RELAX_TWIST_PER_SUBUNIT(SELF,FAL_ONLY=.TRUE.,ALLOW_ANY_CHANGE=.TRUE.)
                ENDIF
                IF (SELF%FRAGMENTED_SUBUNIT_NUM) THEN
                    WRITE(*,'(A)') '**info(FILAMENT_INIT): relaxing rise per subunit (experimental)'
                    CALL FILAMENT_RELAX_RISE_PER_SUBUNIT(SELF)
                ENDIF
                IF (SELF%FRAGMENTED_PHI) THEN
                    WRITE(*,'(A)') '**info(FILAMENT_INIT): relaxing twist per subunit (experimental)'
                    CALL FILAMENT_RELAX_TWIST_PER_SUBUNIT(SELF)
                ENDIF
            ENDIF

            IF (.NOT. SELF%RIGID_PARAMS(7) .AND. .NOT. WWPS_AS_SPS .AND. .NOT. SELF%FRAGMENTED_SUBUNIT_NUM) THEN
                WRITE(*,'(A)') '**debug(FILAMENT_INIT): updating helix rise'
                CALL FILAMENT_UPDATE_AVERAGE_HELIX_RISE(SELF)
            ENDIF

            IF (.NOT. SELF%RIGID_PARAMS(4) .AND. .NOT. WWPS_AS_SPS .AND. .NOT. SELF%FRAGMENTED_PHI) THEN
                WRITE(*,'(A)') '**debug(FILAMENT_INIT): updating average helix twist'
                CALL FILAMENT_UPDATE_AVERAGE_HELIX_TWIST(SELF)
            ENDIF
            CALL FILAMENT_UPDATE_WP_HELIX(SELF)
        ENDDO

        ! TODO HERE: do the same (update filament average) for the rise per subunit?

        ! Check the subunit_num values, they should increase wp-to-wp
        CALL FILAMENT_CHECK_SUBUNIT_NUM_INCREASES(SELF)

        ! Check that the relevant splines are monotonic
        IF (.NOT. FILAMENT_SPLINES_ARE_MONOTONIC(SELF)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_INIT): phi & subunit_num splines are not both monotonic'
            CALL FILAMENT_PRINT_INFO(SELF)
            CALL TERMINATE('Fatal error in FILAMENT_INIT')
        ENDIF

    END SUBROUTINE FILAMENT_INIT

END MODULE FILAMENTS
