!>  \brief  Filament class and basic house-keeping routines
MODULE FILAMENTS_BASE

    USE COOS, ONLY : WAYPOINT
    USE SPLINES, ONLY : INTERPOL_SPLINE
    USE FILMS, ONLY : FILM
    USE HELICES, ONLY : HELICAL_SYMMETRY
    USE ISO_C_BINDING, ONLY : C_PTR

    IMPLICIT NONE

    !>  \brief  Filament class
    TYPE FILAMENT
        INTEGER                     ::  NWP                                     !<  Number of waypoints defined for the filament
        TYPE(WAYPOINT), ALLOCATABLE ::  WP(:)                                   !<  Array of waypoints
        TYPE(FILM)                  ::  FILM                                    !<  Film on which the filament is imaged (this is where we'll get all the CTF info from)

        ! Space curve model of the filament
        REAL(KIND=8),   ALLOCATABLE ::  T(:)                                    !<  For each waypoint, the corresponding value of the T parameter. The T parameter should be a good approximation of the path length.
        TYPE(INTERPOL_SPLINE)       ::  PHI_SPLINE                              !<  Spline modelling the PHI (twist) angle along the filament, as a function of T
        TYPE(INTERPOL_SPLINE)       ::  X_SPLINE                                !<  Spline modelling the X position of the filament, as a function of T
        TYPE(INTERPOL_SPLINE)       ::  Y_SPLINE                                !<  Spline modelling the Y position of the filament, as a function of T
        TYPE(INTERPOL_SPLINE)       ::  Z_SPLINE                                !<  Spline modelling the Z position of the filament, as a function of T
        TYPE(INTERPOL_SPLINE)       ::  S_SPLINE                                !<  Spline modelling the helical subunit number along the filament, as a function of T
        LOGICAL                     ::  REFINE_PARAMS(7)            =   .FALSE. !<  Which of the 7 degrees of freedom to refine per way point (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi,7.SubunitNumber)
        LOGICAL                     ::  USE_HERMITE_SPLINES         =   .TRUE.  !<  USE SPLINESs where the first derivative is used as a constraint (aka KNOWN_D in parts of Frealix).
                                                                                !!  In other words, psi & theta are not used to solve the splines, they are computed from them.

        ! Linear fit of the splines and their residuals
        REAL(KIND=8),   ALLOCATABLE ::  T_LINEAR(:)                             !<  The T values (ordinates) that were used when computing the linear fit.
        REAL(KIND=8)                ::  X_LINEAR(2)                 =   0.0D0   !<  Coefficients of the linear fit of the spline. The first number is multiplicative (slope), the second additive (offset).
        REAL(KIND=8)                ::  Y_LINEAR(2)                 =   0.0D0
        REAL(KIND=8)                ::  Z_LINEAR(2)                 =   0.0D0
        REAL(KIND=8)                ::  S_LINEAR(2)                 =   0.0D0
        REAL(KIND=8)                ::  PHI_LINEAR(2)               =   0.0D0
        TYPE(WAYPOINT), ALLOCATABLE ::  WP_RES(:)                               !<  These waypoints  store the residuals between the linear fits to the splines and the actual WP values

        ! Rigid-body modelling of the filament
        LOGICAL                     ::  RIGID_PARAMS(7)             =   .FALSE. !<  Whether to keep the following parameters fixed throughout the filament: 1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta, 6.Psi, 7-H.Rise.
                                                                                !!  If all are set to .TRUE., the filament is a rigid rod, treated as a single particle.
        TYPE(WAYPOINT)              ::  RIGID_WP                                !<  A special waypoint object storing values for the parameters which are to be kept rigid.

        ! Normal modes of the filament
        LOGICAL                     ::  REFINE_NORMAL_MODES(7)      =   .FALSE. !<  Whether normal mode amplitudes, rather than the model itself, should be refined
        REAL(KIND=8),   ALLOCATABLE ::  NM_EIGENVECTORS(:,:)                    !<  Eigenvectors. The elements of each vector are in the first dimension and the vectors are indexed in the second dimension.
        REAL(KIND=8),   ALLOCATABLE ::  NM_EIGENVALUES(:)                       !<  Eigenvalues
        TYPE(WAYPOINT), ALLOCATABLE ::  WP_NM(:)                                !<  Array of waypoints storing normal mode (NM) amplitudes

        ! Subunit_num-related flags
        LOGICAL                     ::  CONSTANT_RISE_PER_SUBUNIT   =   .TRUE.  !<  Whether to assume that the rise-per-subunit is constant.
        LOGICAL                     ::  CONSTANT_TWIST_PER_LENGTH   =   .TRUE.  !<  Whether to assume that the twist per unit length is constant.

        ! Book-keeping
        LOGICAL,        PRIVATE     ::  IN_MEMORY                   =   .FALSE. !<  Whether memory is already allocated
        INTEGER                     ::  ID                                      !<  ID number, from the parameter file

        ! Description of the helix
        TYPE(HELICAL_SYMMETRY)      ::  HELIX                                   !<  A descriptor for the helical symmetry the filament obeys on average.
        REAL                        ::  MAX_RADIUS                  =   0.0     !<  Maximum radius of the filament (pixels)
        LOGICAL                     ::  FRAGMENTED_SUBUNIT_NUM      =   .FALSE. !<  Introduce breaks/discontinuities in the subunit_num function between waypoints.
        LOGICAL                     ::  FRAGMENTED_PHI              =   .FALSE. !<  Introduce breaks/discontinuities in the phi function between waypoints.

        ! Amyloid-beta-specific components
        LOGICAL                     ::  AMYLOID_BETA                =   .FALSE. !<  Special flag for amyloid beta filaments
        LOGICAL                     ::  WPS_ARE_CROSSOVERS          =   .FALSE. !<  Waypoints are crossovers
        LOGICAL                     ::  FAL_WPS_ARE_CROSSOVERS      =   .FALSE. !<  If waypoints are crossoves, this flag can be used to indicate that the first and last waypoints are in fact not crossovers.

        ! Flags concerning the treatment of the first and last waypoints (FAL = first and last; WPS = waypoints)
        LOGICAL                     ::  FAL_WPS_IN_SCORE            =   .TRUE.  !<  If false, the parts of the filament between the first and the second as well as between the penultimate and last waypoints will not be included in the scoring function
        LOGICAL                     ::  FAL_WPS_IN_THREED           =   .TRUE.  !<  If false, the parts of the filament between the first and the second as well as between the penultimate and last waypoints will not be included in the 3D reconstruction

        ! Microtubule-specific components
        LOGICAL                     ::  MICROTUBULE                 =   .FALSE. !<  Special flag for microtubule filaments
        INTEGER                     ::  MT_PF_NUMBER                =   0       !<  For microtubules only: number of protofilaments (e.g. 13)
        INTEGER                     ::  MT_RISE                     =   0       !<  For microtubules only: how much further along the filament (distance measured in tubulin subunits) the end
                                                                                !!  of a MT_PF_NUMBER-element "washer" is compared to its beginning. The 13-pf microtobules have a value of 3.

        ! Mechanical properties
        REAL                        ::  PERSISTENCE_LENGTH          =   0.0     !<  The persistence length of the filament, in pixels
        REAL                        ::  TORSION_PERSISTENCE_LENGTH  =   0.0     !<  The torsional persistence length of the filament, in pixels
        REAL                        ::  STRETCH_PERSISTENCE_LENGTH  =   0.0     !<  The stretching/compression persistence length of the filament, in pixels.

        ! Scoring
        REAL                        ::  SCORE                       =   0.0     !<  Score for the filament to see ways to calculate the score, check the filament_model_comparison module.
        TYPE(WAYPOINT), ALLOCATABLE ::  WP_SCORE(:)                             !<  The array of waypoints used by the scoring function

        !
        TYPE(C_PTR)                 ::  CPTR                                    !<  A C-pointer pointing to itself. This is needed when passing the object through the FGSL.
        !
        CONTAINS
            FINAL       ::  FILAMENT_DEALLOCATE
        !   PROCEDURE   ::  ALLOCATE            =>  FILAMENT_ALLOCATE
        !   PROCEDURE   ::  ALLOCATED           =>  FILAMENT_ALLOCATED
        !   PROCEDURE   ::  DEALLOCATE          =>  FILAMENT_DEALLOCATE
        !   PROCEDURE   ::  COOS_FIL_TO_SPLINE  =>  COOS_FIL_TO_SPLINE
        !   PROCEDURE   ::  COOS_SPLINE_TO_FIL  =>  COOS_SPLINE_TO_FIL
        !   PROCEDURE   ::  GET_PATH            =>  FILAMENT_GET_PATH
        !   PROCEDURE   ::  UPDATE_SPLINE       =>  FILAMENT_UPDATE_SPLINE
        !   PROCEDURE   ::  INIT                =>  FILAMENT_INIT
        !   PROCEDURE   ::  PRINT_INFO          =>  FILAMENT_PRINT_INFO
    END TYPE

    INTERFACE IS_ALLOCATED
        MODULE PROCEDURE FILAMENT_ALLOCATED
    END INTERFACE

    !INTERFACE ALLOCATE
    !   MODULE PROCEDURE FILAMENT_ALLOCATE
    !END INTERFACE

    !INTERFACE DEALLOCATE
    !   MODULE PROCEDURE FILAMENT_DEALLOCATE
    !END INTERFACE

    INTERFACE ASSIGNMENT(=)
        MODULE PROCEDURE FILAMENT_ASSIGN
    END INTERFACE

    REAL,   PARAMETER   ::  FILAMENT_FT_PADDING = 1.0


    CONTAINS

    !>  \brief  Return reasonable value for box width, in pixels
    PURE FUNCTION FILAMENT_SUGGESTED_BOX_WIDTH(SELF) RESULT(WIDTH)
        USE RUNTIME_PARAMETERS, ONLY : IS_ODD
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        ! Result
        INTEGER ::  WIDTH
        ! Start work
        WIDTH = 3.*SELF%MAX_RADIUS
        IF (IS_ODD(WIDTH)) WIDTH = WIDTH+1
    END FUNCTION FILAMENT_SUGGESTED_BOX_WIDTH

    !>  \brief  Split an array of filament objects into an array of smaller arrays of filament objects. This is useful when
    !!          distributing processing of filaments over nodes etc.
    !!
    !!          Each filament is assigned to a group. Group #1 should be processed first.
    SUBROUTINE FILAMENT_ARRAY_SPLIT(FILAMENTS_IN,NUM_OUTPUT_GROUPS,GROUP_MEMBERSHIP,MEMBERS_PER_GROUP)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(IN)      ::  FILAMENTS_IN(:)         !<  Array of filaments
        INTEGER,                            INTENT(IN)      ::  NUM_OUTPUT_GROUPS       !<  Number of groups of filaments to be defined
        INTEGER,        ALLOCATABLE,        INTENT(INOUT)   ::  GROUP_MEMBERSHIP(:)     !<  Output array assigning a group to each filament
        INTEGER,        ALLOCATABLE,        INTENT(INOUT)   ::  MEMBERS_PER_GROUP(:)    !<  Number of filaments in each group
        ! Private variables
        INTEGER ::  NUM_INPUT_FILS
        INTEGER ::  IERR
        INTEGER ::  FILS_PER_GROUP, FILS_IN_CURRENT_GROUP, CURRENT_GROUP, GROUPS_WITH_MORE_FILS
        INTEGER ::  I
        INTEGER ::  NWP_ARRAY(SIZE(FILAMENTS_IN))
        LOGICAL ::  TO_BE_REFINED(SIZE(FILAMENTS_IN))
        INTEGER ::  GROUP_CAPACITY(NUM_OUTPUT_GROUPS)
        INTEGER ::  NWP_PER_GROUP(NUM_OUTPUT_GROUPS)
        INTEGER ::  LENGTH_OF_LONGEST_FIL, INDEX_OF_LONGEST_FIL
        INTEGER ::  LENGTH_OF_SHORTEST_FIL, INDEX_OF_SHORTEST_FIL
        ! Start work

        ! The number of filaments in the array
        NUM_INPUT_FILS = SIZE(FILAMENTS_IN)
        WRITE(*,'(A,I0,A)') '**info(FILAMENT_ARRAY_SPLIT): there are ', NUM_INPUT_FILS, ' input filaments'

        !  Cannot have more output groups than input filaments
        IF (NUM_OUTPUT_GROUPS .GT. NUM_INPUT_FILS) THEN
            WRITE(*,'(A,I0,A,I0,A)') '**ERROR(FILAMENT_ARRAY_SPLIT): number of output groups (', NUM_OUTPUT_GROUPS, &
                                    ') greater than number of input filaments (', NUM_INPUT_FILS, ')'
            CALL TERMINATE('Fatal error in FILAMENT_ARRAY_SPLIT')
        ENDIF

        ! The number of waypoints in each filament
        NWP_ARRAY(1:NUM_INPUT_FILS) = FILAMENTS_IN(1:NUM_INPUT_FILS)%NWP

        ! Whether filaments are set to be refined
        FORALL(I=1:NUM_INPUT_FILS)
            TO_BE_REFINED(I) = ANY(FILAMENTS_IN(I)%WP%REFINE)
        END FORALL

        ! Allocate memory to group membership array
        ALLOCATE(GROUP_MEMBERSHIP(NUM_INPUT_FILS),MEMBERS_PER_GROUP(NUM_OUTPUT_GROUPS),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(FILAMENT_ARRAY_SPLIT): memory allocation error ', IERR
            CALL TERMINATE('Fatal error in FILAMENT_ARRAY_SPLIT')
        ENDIF

        ! Initialise group membership array
        GROUP_MEMBERSHIP = 0

        ! Work out a number of filaments to go into each group.
        FILS_PER_GROUP = INT(NUM_INPUT_FILS/NUM_OUTPUT_GROUPS)
        GROUPS_WITH_MORE_FILS = MOD(NUM_INPUT_FILS,NUM_OUTPUT_GROUPS)

        ! The "group capacity" array keeps track for each group of how many members we're looking for
        GROUP_CAPACITY(:) = FILS_PER_GROUP
        IF (GROUPS_WITH_MORE_FILS .GT. 0) THEN
            !GROUP_CAPACITY(1:GROUPS_WITH_MORE_FILS) = FILS_PER_GROUP + 1
            GROUP_CAPACITY(NUM_OUTPUT_GROUPS-GROUPS_WITH_MORE_FILS+1:NUM_OUTPUT_GROUPS) = FILS_PER_GROUP + 1
            WRITE(*,'(4(A,I0),A)') '**info(FILAMENT_ARRAY_SPLIT): splitting into ', GROUPS_WITH_MORE_FILS, &
                                    ' groups of ', FILS_PER_GROUP + 1, ' and ', NUM_OUTPUT_GROUPS - GROUPS_WITH_MORE_FILS, &
                                    ' groups of ', FILS_PER_GROUP, ' filaments'
        ELSE
            WRITE(*,'(2(A,I0),A)') '**info(FILAMENT_ARRAY_SPLIT): splitting into ', NUM_OUTPUT_GROUPS, ' groups of ', &
                                    FILS_PER_GROUP, ' filaments'
        ENDIF


        ! Loop until we have assigned membership to every filament
        CURRENT_GROUP = 1
        FILS_IN_CURRENT_GROUP = 0
        NWP_PER_GROUP = 0
        DO WHILE (ANY(GROUP_MEMBERSHIP .EQ. 0))

            ! If the current group is full, move on to the next one
            IF (FILS_IN_CURRENT_GROUP .EQ. GROUP_CAPACITY(CURRENT_GROUP)) THEN
                CURRENT_GROUP = CURRENT_GROUP + 1
                FILS_IN_CURRENT_GROUP = 0
            ENDIF

            ! Find the longest filament still available, which has at least one waypoint set to be refined
            IF (ANY(GROUP_MEMBERSHIP .EQ. 0 .AND. TO_BE_REFINED)) THEN
                LENGTH_OF_LONGEST_FIL   =   MAXVAL(NWP_ARRAY,       MASK = (GROUP_MEMBERSHIP .EQ. 0 .AND. TO_BE_REFINED))
                INDEX_OF_LONGEST_FIL    =   MAXLOC(NWP_ARRAY, 1,    MASK = (GROUP_MEMBERSHIP .EQ. 0 .AND. TO_BE_REFINED))
            ELSE
                LENGTH_OF_LONGEST_FIL   =   MAXVAL(NWP_ARRAY,       MASK =  GROUP_MEMBERSHIP .EQ. 0)
                INDEX_OF_LONGEST_FIL    =   MAXLOC(NWP_ARRAY, 1,    MASK =  GROUP_MEMBERSHIP .EQ. 0)
            ENDIF

            ! Assign it to the current group
            GROUP_MEMBERSHIP(INDEX_OF_LONGEST_FIL) = CURRENT_GROUP
            FILS_IN_CURRENT_GROUP = FILS_IN_CURRENT_GROUP+1
            NWP_PER_GROUP(CURRENT_GROUP) = NWP_PER_GROUP(CURRENT_GROUP) + FILAMENTS_IN(INDEX_OF_LONGEST_FIL)%NWP

            ! If the current group is full, move on to the next one
            IF (FILS_IN_CURRENT_GROUP .EQ. GROUP_CAPACITY(CURRENT_GROUP)) THEN
                CURRENT_GROUP = CURRENT_GROUP + 1
                FILS_IN_CURRENT_GROUP = 0
                CYCLE
            ENDIF

            ! Fill the group with small filaments
            DO WHILE (FILS_IN_CURRENT_GROUP .LT. GROUP_CAPACITY(CURRENT_GROUP))
                ! Find the shortest filament still available. If possible use a filament that's not going to be refined
                IF (ANY(GROUP_MEMBERSHIP .EQ. 0 .AND. .NOT. TO_BE_REFINED)) THEN
                    LENGTH_OF_SHORTEST_FIL  =   MINVAL(NWP_ARRAY,       MASK = (GROUP_MEMBERSHIP .EQ. 0 .AND. .NOT. TO_BE_REFINED))
                    INDEX_OF_SHORTEST_FIL   =   MINLOC(NWP_ARRAY, 1,    MASK = (GROUP_MEMBERSHIP .EQ. 0 .AND. .NOT. TO_BE_REFINED))
                ELSE
                    LENGTH_OF_SHORTEST_FIL  =   MINVAL(NWP_ARRAY,       MASK =  GROUP_MEMBERSHIP .EQ. 0)
                    INDEX_OF_SHORTEST_FIL   =   MINLOC(NWP_ARRAY, 1,    MASK =  GROUP_MEMBERSHIP .EQ. 0)
                ENDIF

                ! Assign it to the current group
                GROUP_MEMBERSHIP(INDEX_OF_SHORTEST_FIL) = CURRENT_GROUP
                FILS_IN_CURRENT_GROUP = FILS_IN_CURRENT_GROUP + 1
                NWP_PER_GROUP(CURRENT_GROUP) = NWP_PER_GROUP(CURRENT_GROUP) + FILAMENTS_IN(INDEX_OF_SHORTEST_FIL)%NWP
            ENDDO

        ENDDO

        !
        ! Diagnostics
        !
        WRITE(*,'(3(A20,1X))') 'Group #', '# of filaments', 'total # of waypoints'
        DO I=1,NUM_OUTPUT_GROUPS
            WRITE(*,'(3(I20,1X))') I, GROUP_CAPACITY(I),NWP_PER_GROUP(I)
        ENDDO

        MEMBERS_PER_GROUP = GROUP_CAPACITY

    END SUBROUTINE FILAMENT_ARRAY_SPLIT


    !>  \brief  Given an array of waypoints, populate a filament object
    SUBROUTINE FILAMENT_FROM_WPS(SELF,WPS,ID,FILM_ID,PIXEL_SIZE,SCORE)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                 INTENT(INOUT)   ::  SELF
        TYPE(WAYPOINT), ALLOCATABLE,    INTENT(IN)      ::  WPS(:)      !<  Array of waypoints
        INTEGER,        OPTIONAL,       INTENT(IN)      ::  ID          !<  Filament ID
        INTEGER,        OPTIONAL,       INTENT(IN)      ::  FILM_ID     !<  Film ID
        REAL,           OPTIONAL,       INTENT(IN)      ::  PIXEL_SIZE  !<  Film pixel size, in Angstroms
        REAL,           OPTIONAL,       INTENT(IN)      ::  SCORE       !<  Filament score
        ! Private variables

        ! Start work

        IF (.NOT. ALLOCATED(WPS)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_FROM_WPS): input array of waypoints not allocated'
            CALL TERMINATE('Fatal error in FILAMENT_FROM_WPS')
        ENDIF

        ! Copy the waypoints over
        IF (ALLOCATED(SELF%WP)) DEALLOCATE(SELF%WP)
        ALLOCATE(SELF%WP(SIZE(WPS)))
        SELF%WP = WPS
        SELF%NWP = SIZE(WPS)

        ! Other properties
        IF (PRESENT(ID))            SELF%ID = ID
        IF (PRESENT(FILM_ID))       SELF%FILM%ID = FILM_ID
        IF (PRESENT(PIXEL_SIZE))    SELF%FILM%PIXEL_SIZE = PIXEL_SIZE
        IF (PRESENT(SCORE))         SELF%SCORE = SCORE

    END SUBROUTINE FILAMENT_FROM_WPS

    !>  \brief  Subroutine which overloads the assignment operator (=)
    !!
    !!  This subroutine does not work for "in-place" arguments, e.g. a = a
    SUBROUTINE FILAMENT_ASSIGN(FIL_LHS,FIL_RHS)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)      ::  FIL_RHS     !<  Filament object on the right-hand-side of the assignment
        TYPE(FILAMENT),     INTENT(OUT)     ::  FIL_LHS     !<  Filament object on the left-hand-side of the assignment
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work
        ! Allocate memory
        IF (IS_ALLOCATED(FIL_LHS) .AND. (FIL_LHS%NWP .NE. FIL_RHS%NWP)) CALL FILAMENT_DEALLOCATE(FIL_LHS)
        IF (.NOT. IS_ALLOCATED(FIL_LHS)) CALL FILAMENT_ALLOCATE(FIL_LHS,FIL_RHS%NWP)
        ! Copy contents over
        CALL FILAMENT_COPY_PROPERTIES(FIL_LHS,FIL_RHS)
        IF (ALLOCATED(FIL_RHS%WP)) THEN
            FIL_LHS%WP(:)                   =   FIL_RHS%WP
        ELSE
            WRITE(*,'(2A)') '**ERROR(FILAMENT_ASSIGN): attempt to copy non-allocated array ',    &
                            'of waypoints from RHS to LHS of assignment'
            CALL TERMINATE('Fatal error in FILAMENT_ASSIGN')
        ENDIF

        FIL_LHS%FILM                        =   FIL_RHS%FILM
        IF (ALLOCATED(FIL_RHS%T)) THEN
            FIL_LHS%T(:)                    =   FIL_RHS%T
        ELSE
            IF (DEBUG) WRITE(*,'(2A)') '**debug(FILAMENT_ASSIGN): attempt to copy non-allocated array of ', &
                                        'T values from RHS to LHS of assignment'
        ENDIF
        FIL_LHS%PHI_SPLINE                  =   FIL_RHS%PHI_SPLINE
        FIL_LHS%X_SPLINE                    =   FIL_RHS%X_SPLINE
        FIL_LHS%Y_SPLINE                    =   FIL_RHS%Y_SPLINE
        FIL_LHS%Z_SPLINE                    =   FIL_RHS%Z_SPLINE
        FIL_LHS%S_SPLINE                    =   FIL_RHS%S_SPLINE
        IF (ALLOCATED(FIL_RHS%T_LINEAR)) THEN
            FIL_LHS%T_LINEAR                =   FIL_RHS%T_LINEAR
        ENDIF
        FIL_LHS%X_LINEAR                    =   FIL_RHS%X_LINEAR
        FIL_LHS%Y_LINEAR                    =   FIL_RHS%Y_LINEAR
        FIL_LHS%Z_LINEAR                    =   FIL_RHS%Z_LINEAR
        FIL_LHS%S_LINEAR                    =   FIL_RHS%S_LINEAR
        FIL_LHS%PHI_LINEAR                  =   FIL_RHS%PHI_LINEAR
        IF (ALLOCATED(FIL_RHS%WP_RES)) THEN
            FIL_LHS%WP_RES(:)               =   FIL_RHS%WP_RES
        ENDIF

        FIL_LHS%IN_MEMORY                   =   FIL_RHS%IN_MEMORY

        IF (ALLOCATED(FIL_RHS%WP_SCORE)) THEN
            IF (.NOT. ALLOCATED(FIL_LHS%WP_SCORE)) ALLOCATE(FIL_LHS%WP_SCORE(SIZE(FIL_RHS%WP_SCORE)))
            FIL_LHS%WP_SCORE(:)             =   FIL_RHS%WP_SCORE
        ENDIF

        ! Deal with normal modes

        IF (ALLOCATED(FIL_RHS%NM_EIGENVECTORS)) THEN
            FIL_LHS%NM_EIGENVECTORS(:,:)    =   FIL_RHS%NM_EIGENVECTORS
        ENDIF
        IF (ALLOCATED(FIL_RHS%NM_EIGENVALUES)) THEN
            FIL_LHS%NM_EIGENVALUES(:)       =   FIL_RHS%NM_EIGENVALUES
        ENDIF
        IF (ALLOCATED(FIL_RHS%WP_NM)) THEN
            FIL_LHS%WP_NM(:)                =   FIL_RHS%WP_NM
        ENDIF
    END SUBROUTINE FILAMENT_ASSIGN

    !>  \brief  Copy proerpties of the filament that do not relate to the waypoints themselves
    !!          or other allocatable components. In other words, this can be used between filaments
    !!          with different numbers of waypoints
    PURE SUBROUTINE FILAMENT_COPY_PROPERTIES(FIL_LHS,FIL_RHS)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)      ::  FIL_RHS     !<  Filament object on the right-hand-side of the assignment
        TYPE(FILAMENT),     INTENT(INOUT)   ::  FIL_LHS     !<  Filament object on the left-hand-side of the assignment
        ! Start work
        FIL_LHS%NWP                         =   FIL_RHS%NWP
        FIL_LHS%REFINE_PARAMS               =   FIL_RHS%REFINE_PARAMS
        FIL_LHS%USE_HERMITE_SPLINES         =   FIL_RHS%USE_HERMITE_SPLINES
        FIL_LHS%X_LINEAR                    =   FIL_RHS%X_LINEAR
        FIL_LHS%Y_LINEAR                    =   FIL_RHS%Y_LINEAR
        FIL_LHS%Z_LINEAR                    =   FIL_RHS%Z_LINEAR
        FIL_LHS%S_LINEAR                    =   FIL_RHS%S_LINEAR
        FIL_LHS%PHI_LINEAR                  =   FIL_RHS%PHI_LINEAR
        FIL_LHS%RIGID_PARAMS                =   FIL_RHS%RIGID_PARAMS
        FIL_LHS%RIGID_WP                    =   FIL_RHS%RIGID_WP
        FIL_LHS%REFINE_NORMAL_MODES         =   FIL_RHS%REFINE_NORMAL_MODES
        FIL_LHS%CONSTANT_RISE_PER_SUBUNIT   =   FIL_RHS%CONSTANT_RISE_PER_SUBUNIT
        FIL_LHS%CONSTANT_TWIST_PER_LENGTH   =   FIL_RHS%CONSTANT_TWIST_PER_LENGTH
        FIL_LHS%ID                          =   FIL_RHS%ID
        FIL_LHS%HELIX                       =   FIL_RHS%HELIX
        FIL_LHS%MAX_RADIUS                  =   FIL_RHS%MAX_RADIUS
        FIL_LHS%FRAGMENTED_SUBUNIT_NUM      =   FIL_RHS%FRAGMENTED_SUBUNIT_NUM
        FIL_LHS%FRAGMENTED_PHI              =   FIL_RHS%FRAGMENTED_PHI
        FIL_LHS%AMYLOID_BETA                =   FIL_RHS%AMYLOID_BETA
        FIL_LHS%WPS_ARE_CROSSOVERS          =   FIL_RHS%WPS_ARE_CROSSOVERS
        FIL_LHS%FAL_WPS_ARE_CROSSOVERS      =   FIL_RHS%FAL_WPS_ARE_CROSSOVERS
        FIL_LHS%FAL_WPS_IN_SCORE            =   FIL_RHS%FAL_WPS_IN_SCORE
        FIL_LHS%FAL_WPS_IN_THREED           =   FIL_RHS%FAL_WPS_IN_THREED
        FIL_LHS%MICROTUBULE                 =   FIL_RHS%MICROTUBULE
        FIL_LHS%MT_PF_NUMBER                =   FIL_RHS%MT_PF_NUMBER
        FIL_LHS%MT_RISE                     =   FIL_RHS%MT_RISE
        FIL_LHS%PERSISTENCE_LENGTH          =   FIL_RHS%PERSISTENCE_LENGTH
        FIL_LHS%TORSION_PERSISTENCE_LENGTH  =   FIL_RHS%TORSION_PERSISTENCE_LENGTH
        FIL_LHS%STRETCH_PERSISTENCE_LENGTH  =   FIL_RHS%STRETCH_PERSISTENCE_LENGTH
        FIL_LHS%SCORE                       =   FIL_RHS%SCORE
    END SUBROUTINE FILAMENT_COPY_PROPERTIES

    SUBROUTINE FILAMENT_PROPERTIES_RESET(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Start work
        SELF%REFINE_PARAMS               =  .FALSE.
        SELF%USE_HERMITE_SPLINES         =  .TRUE.
        SELF%RIGID_PARAMS                =  .FALSE.
        !SELF%RIGID_WP                    =
        SELF%CONSTANT_RISE_PER_SUBUNIT   =  .TRUE.
        SELF%CONSTANT_TWIST_PER_LENGTH   =  .TRUE.
        SELF%SCORE                       =  0.0
        SELF%ID                          =  0
        !SELF%HELIX                       =
        SELF%MAX_RADIUS                  =  0.0
        SELF%FRAGMENTED_SUBUNIT_NUM      =  .FALSE.
        SELF%FRAGMENTED_PHI              =  .FALSE.
        SELF%AMYLOID_BETA                =  .FALSE.
        SELF%WPS_ARE_CROSSOVERS          =  .FALSE.
        SELF%FAL_WPS_ARE_CROSSOVERS      =  .FALSE.
        SELF%MICROTUBULE                 =  .FALSE.
        SELF%MT_PF_NUMBER                =  0
        SELF%MT_RISE                     =  0
        SELF%PERSISTENCE_LENGTH          =  0.0
        SELF%TORSION_PERSISTENCE_LENGTH  =  0.0
        SELF%STRETCH_PERSISTENCE_LENGTH  =  0.0
        SELF%FAL_WPS_IN_SCORE            =  .TRUE.
        SELF%FAL_WPS_IN_THREED           =  .TRUE.
        SELF%REFINE_NORMAL_MODES         =  .FALSE.
        SELF%X_LINEAR                    =  0.0
        SELF%Y_LINEAR                    =  0.0
        SELF%Z_LINEAR                    =  0.0
        SELF%S_LINEAR                    =  0.0
        SELF%PHI_LINEAR                  =  0.0
    END SUBROUTINE FILAMENT_PROPERTIES_RESET


    !>  \brief  Allocate memory for filament - need to know the number of way points
    SUBROUTINE FILAMENT_ALLOCATE(SELF,N)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE SPLINES, ONLY : ALLOCATE_INTERPOL_SPLINE
        USE, INTRINSIC  ::  ISO_C_BINDING
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), TARGET, INTENT(INOUT)   ::  SELF
        INTEGER,                INTENT(IN)      ::  N       !<  Number of way points in the filament
        ! Private variables
        INTEGER ::  IERR
        LOGICAL,    PARAMETER   ::  DEBUG    =  .FALSE.
        ! Start work
        IF (SELF%IN_MEMORY) THEN
            PRINT '(A)', '**WARNING(FILAMENT_ALLOCATE): already allocated, will deallocate first'
            CALL FILAMENT_DEALLOCATE(SELF)
        ENDIF
        SELF%NWP = N
        IF (ALLOCATED(SELF%WP)) THEN
            IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_ALLOCATE): filament was not allocated, yet had to deallocate waypoints'
            DEALLOCATE(SELF%WP)
        ENDIF
        ALLOCATE(SELF%WP(N),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_ALLOCATE): allocation of waypoints failed'
            CALL TERMINATE('Fatal error in FILAMENT_ALLOCATE')
        ENDIF
        ! Allocate & initialise the T array for spline parameter
        ALLOCATE(SELF%T(N),SELF%T_LINEAR(N),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_ALLOCATE): allocation of parameter array failed'
            CALL TERMINATE('Fatal error in FILAMENT_ALLOCATE')
        ENDIF
        SELF%T = 0.0D0
        SELF%T_LINEAR = 0.0D0
        ! Allocate the splines
        CALL ALLOCATE_INTERPOL_SPLINE(SELF%PHI_SPLINE,N,3)
        CALL ALLOCATE_INTERPOL_SPLINE(SELF%X_SPLINE,N,3)
        CALL ALLOCATE_INTERPOL_SPLINE(SELF%Y_SPLINE,N,3)
        CALL ALLOCATE_INTERPOL_SPLINE(SELF%Z_SPLINE,N,3)
        CALL ALLOCATE_INTERPOL_SPLINE(SELF%S_SPLINE,N,3)
        SELF%IN_MEMORY = .TRUE.
        ! Allocate NM data
        IF (ALLOCATED(SELF%NM_EIGENVECTORS)) DEALLOCATE(SELF%NM_EIGENVECTORS)
        IF (ALLOCATED(SELF%NM_EIGENVALUES)) DEALLOCATE(SELF%NM_EIGENVALUES)
        IF (ALLOCATED(SELF%WP_NM)) DEALLOCATE(SELF%WP_NM)
        IF (ALLOCATED(SELF%WP_RES)) DEALLOCATE(SELF%WP_RES)
        ALLOCATE(SELF%NM_EIGENVECTORS(N,N),SELF%NM_EIGENVALUES(N))
        ALLOCATE(SELF%WP_NM(N))
        ALLOCATE(SELF%WP_RES(N))
        SELF%NM_EIGENVECTORS = 0.0D0
        SELF%NM_EIGENVALUES  = 0.0D0
        ! The C pointer is setup
        SELF%CPTR = C_LOC(SELF)
    END SUBROUTINE FILAMENT_ALLOCATE

    !>  \brief  Deallocate memory from filament
    SUBROUTINE FILAMENT_DEALLOCATE(SELF)
        USE SPLINES, ONLY : INTERPOL_SPLINE_DEALLOCATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work
        IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_DEALLOCATE): entry'
        IF (SELF%IN_MEMORY) THEN
            IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_DEALLOCATE): deallocating'
            IF (ALLOCATED(SELF%T)) DEALLOCATE(SELF%T)
            IF (ALLOCATED(SELF%T_LINEAR)) DEALLOCATE(SELF%T_LINEAR)
            ! Deallocate splines
            IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_DEALLOCATE): deallocating splines'
            CALL INTERPOL_SPLINE_DEALLOCATE(SELF%PHI_SPLINE)
            CALL INTERPOL_SPLINE_DEALLOCATE(SELF%X_SPLINE)
            CALL INTERPOL_SPLINE_DEALLOCATE(SELF%Y_SPLINE)
            CALL INTERPOL_SPLINE_DEALLOCATE(SELF%Z_SPLINE)
            CALL INTERPOL_SPLINE_DEALLOCATE(SELF%S_SPLINE)
            ! Deallocate waypoints
            IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_DEALLOCATE): deallocating waypoints'
            DEALLOCATE(SELF%WP)
            !
            IF (ALLOCATED(SELF%WP_NM)) DEALLOCATE(SELF%WP_NM)
            IF (ALLOCATED(SELF%WP_RES)) DEALLOCATE(SELF%WP_RES)
            IF (ALLOCATED(SELF%NM_EIGENVECTORS)) DEALLOCATE(SELF%NM_EIGENVECTORS)
            IF (ALLOCATED(SELF%NM_EIGENVALUES)) DEALLOCATE(SELF%NM_EIGENVALUES)
            !
            SELF%IN_MEMORY = .FALSE.
        ELSE
            !WRITE(*,'(A)') '**ERROR(FILAMENT_DEALLOCATE): attempt to deallocate object which was not allocated'
            !CALL TERMINATE('Fatal error in FILAMENT_DEALLOCATE')
        ENDIF

        IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_DEALLOCATE): exit'

    END SUBROUTINE FILAMENT_DEALLOCATE

    PURE LOGICAL FUNCTION PSI_THETA_NON_ZERO(SELF)
        IMPLICIT NONE
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        PSI_THETA_NON_ZERO = ANY(SELF%WP%EULER%PSI .NE. 0.0) .OR. ANY(SELF%WP%EULER%THETA .NE. 0.0)
    END FUNCTION PSI_THETA_NON_ZERO

    PURE LOGICAL FUNCTION THETA_NON_ZERO(SELF)
        IMPLICIT NONE
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        THETA_NON_ZERO = ANY(SELF%WP%EULER%THETA .NE. 0.0)
    END FUNCTION THETA_NON_ZERO

    PURE LOGICAL FUNCTION FILAMENT_ALLOCATED(SELF)
        USE SPLINES, ONLY : INTERPOL_SPLINE_ALLOCATED
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        ! Start work
        FILAMENT_ALLOCATED =    SELF%IN_MEMORY .AND. &
                                SELF%X_SPLINE%IN_MEMORY .AND. &
                                SELF%Y_SPLINE%IN_MEMORY .AND. &
                                SELF%Z_SPLINE%IN_MEMORY .AND. &
                                SELF%PHI_SPLINE%IN_MEMORY .AND. &
                                SELF%S_SPLINE%IN_MEMORY .AND. &
                                ALLOCATED(SELF%WP) .AND. &
                                ALLOCATED(SELF%WP_RES) .AND. &
                                ALLOCATED(SELF%T) .AND. &
                                ALLOCATED(SELF%T_LINEAR) .AND. &
                                INTERPOL_SPLINE_ALLOCATED(SELF%X_SPLINE) .AND. &
                                INTERPOL_SPLINE_ALLOCATED(SELF%Y_SPLINE) .AND. &
                                INTERPOL_SPLINE_ALLOCATED(SELF%Z_SPLINE) .AND. &
                                INTERPOL_SPLINE_ALLOCATED(SELF%S_SPLINE) .AND. &
                                ALLOCATED(SELF%NM_EIGENVECTORS) .AND. &
                                ALLOCATED(SELF%NM_EIGENVALUES) .AND. &
                                ALLOCATED(SELF%WP_NM)
    END FUNCTION FILAMENT_ALLOCATED

    !>  \brief  Check whether the space curve is being refined via its normal modes
    PURE LOGICAL FUNCTION FILAMENT_REFINE_AXIS_NMS(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        ! Start work
        FILAMENT_REFINE_AXIS_NMS = ALL(SELF%REFINE_NORMAL_MODES(1:3)) .AND. ALL(SELF%REFINE_NORMAL_MODES(5:6))
    END FUNCTION FILAMENT_REFINE_AXIS_NMS

    !>  \brief  Work out the total number of dimensions of the search space
    INTEGER FUNCTION FILAMENT_EXPECTED_NUMBER_OF_VALUES(SELF,PARAMS_MASK,PARAMS_MASK_HOLD_AVERAGE) RESULT(NUM_DIM_TOTAL)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(IN)  ::  SELF
        LOGICAL,                            INTENT(IN)  ::  PARAMS_MASK(7)  !<  For each of the 7 parameters, whether they are included in the values array
        LOGICAL,            OPTIONAL,       INTENT(IN)  ::  PARAMS_MASK_HOLD_AVERAGE(7) !<  For each of the 7 parameters whether the average value should be left alone
        ! Private variables
        INTEGER ::  NUM_DIM_PER_WP, NUM_WP, NUM_DIM_RIGID
        ! Start work
        NUM_DIM_PER_WP  =   COUNT(PARAMS_MASK .AND. .NOT. SELF%RIGID_PARAMS)    ! Number of parameters to refine per waypoint
        NUM_WP          =   COUNT(SELF%WP%REFINE)                                           ! Number of way points we're refining
        NUM_DIM_RIGID   =   COUNT(PARAMS_MASK .AND. SELF%RIGID_PARAMS)
        NUM_DIM_TOTAL   =   NUM_DIM_PER_WP * NUM_WP + NUM_DIM_RIGID
        IF (PRESENT(PARAMS_MASK_HOLD_AVERAGE)) THEN
            NUM_DIM_TOTAL = NUM_DIM_TOTAL - COUNT(PARAMS_MASK_HOLD_AVERAGE .AND. PARAMS_MASK)
        ENDIF
    END FUNCTION FILAMENT_EXPECTED_NUMBER_OF_VALUES



    !>  \brief  Print nicely all data contained in the Filament object
    SUBROUTINE FILAMENT_PRINT_INFO(SELF,VERBOSE)
        USE COOS, ONLY : WAYPOINTS_PRINT_INFO
        USE HELICES, ONLY : HELICAL_SYMMETRY_PRINT_INFO
        use splines, only : interpol_spline_print_info
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(IN)      ::  SELF
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  VERBOSE !<  Print out more details
        ! Private variables
        INTEGER ::  I
        CHARACTER(LEN=5),   PARAMETER   ::  PARAM_LABEL(7)  =   (/  'X    ',  &
                                                                    'Y    ',  &
                                                                    'Z    ',  &
                                                                    'Phi  ',  &
                                                                    'Theta',  &
                                                                    'Psi  ',  &
                                                                    'SubNm'/)
        LOGICAL ::  VVERBOSE
        ! Start work

        VVERBOSE = .FALSE.
        IF (PRESENT(VERBOSE)) VVERBOSE = VERBOSE


        !
        WRITE(*,'(/A,I0,A)')                    'Filament info for filament ', SELF%ID, ':'
        IF (VVERBOSE) WRITE(*,'(A,L1)')         '  allocated? ', FILAMENT_ALLOCATED(SELF)
        IF (VVERBOSE) WRITE(*,'(A,I0)')         '  number of way points = ', SELF%NWP
        WRITE(*,'(A,I0,A,I0,A)')                '  on micrograph number (location in file) ', SELF%FILM%ID, &
                                                                                        ' (',SELF%FILM%LOCATION_IN_FILE,')'
        IF (VVERBOSE) WRITE(*,'(A,4(L1,1X))')   '  known_d (X,Y,Z,Phi)? ', SELF%X_SPLINE%KNOWN_D,SELF%Y_SPLINE%KNOWN_D, &
                                                                            SELF%Z_SPLINE%KNOWN_D, SELF%PHI_SPLINE%KNOWN_D
        WRITE(*,'(A,L1)')                       '  amyloid beta? ', SELF%AMYLOID_BETA
        IF (VVERBOSE) WRITE(*,'(A,2L2)')        '  WPs, FAL WPs are crossovers? ', SELF%WPS_ARE_CROSSOVERS, &
                                                                                    SELF%FAL_WPS_ARE_CROSSOVERS
        WRITE(*,'(A,2L2)')                      '  FAL WPs in scoring function? 3D rec? ', SELF%FAL_WPS_IN_SCORE, &
                                                                                           SELF%FAL_WPS_IN_THREED
        IF (VVERBOSE) WRITE(*,'(A,7(L1,1X))')   '  refine normal mode amplitudes? ', SELF%REFINE_NORMAL_MODES
        IF (VVERBOSE) WRITE(*,'(A,L1)')         '  constant rise per subunit? ', SELF%CONSTANT_RISE_PER_SUBUNIT
        IF (VVERBOSE) WRITE(*,'(A,L1)')         '  constant twist per length? ', SELF%CONSTANT_TWIST_PER_LENGTH
        IF (VVERBOSE) WRITE(*,'(A,7(L1,1X))')   '  rigid parameters? ', SELF%RIGID_PARAMS
        IF (VVERBOSE) WRITE(*,'(A,L1)')         '  use Hermite splines for the space curve? ', SELF%USE_HERMITE_SPLINES
        IF (VVERBOSE) WRITE(*,'(A,2(L1,1X))')   '  fragmented phi? subunit_num? ', SELF%FRAGMENTED_PHI, SELF%FRAGMENTED_SUBUNIT_NUM
        WRITE(*,'(A)',ADVANCE='NO')             '  average helical parameters: '
        CALL HELICAL_SYMMETRY_PRINT_INFO(SELF%HELIX)
        IF (VVERBOSE) WRITE(*,'(A,F0.3)')                     '  max radius (pixels) = ', SELF%MAX_RADIUS
        WRITE(*,'(A,3(F0.3,1X))')               '  persistence lengths (curvature torsion stretch) (pixels) = ', &
                                                    SELF%PERSISTENCE_LENGTH, SELF%TORSION_PERSISTENCE_LENGTH, &
                                                    SELF%STRETCH_PERSISTENCE_LENGTH
        WRITE(*,'(A,F0.4)')                     '  score = ', SELF%SCORE
        IF (SELF%MICROTUBULE) THEN
            WRITE(*,'(A,I0,1X,I0)')             '  microtubule pf number; rise ', SELF%MT_PF_NUMBER, SELF%MT_RISE
        ENDIF
        IF (ANY(SELF%RIGID_PARAMS)) THEN
            IF (VVERBOSE) WRITE(*,'(A)',ADVANCE='NO') &
                                                '  rigid parameters: '
            DO I=1,7
                IF (SELF%RIGID_PARAMS(I)) THEN
                    IF (VVERBOSE)  WRITE(*,'(2A)',ADVANCE='NO') PARAM_LABEL(I), ': '
                    SELECT CASE (I)
                        CASE (1:3)
                            IF (VVERBOSE) WRITE(*,'(F0.3)',ADVANCE='NO') SELF%RIGID_WP%COO(I)
                        CASE (4)
                            IF (VVERBOSE) WRITE(*,'(F0.3)',ADVANCE='NO') SELF%RIGID_WP%EULER%PHI
                        CASE (5)
                            IF (VVERBOSE) WRITE(*,'(F0.3)',ADVANCE='NO') SELF%RIGID_WP%EULER%THETA
                        CASE (6)
                            IF (VVERBOSE) WRITE(*,'(F0.3)',ADVANCE='NO') SELF%RIGID_WP%EULER%PSI
                        CASE (7)
                            IF (VVERBOSE) WRITE(*,'(F0.3)',ADVANCE='NO') SELF%RIGID_WP%SUBUNIT_NUM
                    END SELECT
                    IF (VVERBOSE) WRITE(*,'(A)',ADVANCE='NO') '  '
                ENDIF
            ENDDO
            IF (VVERBOSE) WRITE(*,'(A)') ' '
        ENDIF
        CALL WAYPOINTS_PRINT_INFO(SELF%WP)
        WRITE(*,'(A/)')             ' '
        IF (ALLOCATED(SELF%WP_NM)) THEN
            IF (VVERBOSE) WRITE(*,'(A,I0)') 'Normal mode amplitudes for filament ', SELF%ID
            IF (VVERBOSE) CALL WAYPOINTS_PRINT_INFO(SELF%WP_NM)
        ELSE
            IF (VVERBOSE) WRITE(*,'(A)') 'Normal modes not set up'
        ENDIF
        !NM_AMP_SUM = FILAMENT_NM_AMPLITUDES_WEIGHTED_SUM(SELF)
        !WRITE(*,'(A,F0.3)') 'Weighted sum of NM amplitudes = ', NM_AMP_SUM
        !WRITE(*,'(A,2(F0.3,1X))') 'Linear fit coefficients (X): ', SELF%X_LINEAR
        !WRITE(*,'(A,2(F0.3,1X))') 'Linear fit coefficients (Y): ', SELF%Y_LINEAR
        !WRITE(*,'(A,2(F0.3,1X))') 'Linear fit coefficients (Z): ', SELF%Z_LINEAR
        !WRITE(*,'(A,2(F0.3,1X))') 'Linear fit coefficients (Phi): ', SELF%PHI_LINEAR
        !WRITE(*,'(A,2(F0.3,1X))') 'Linear fit coefficients (Subunit_num): ', SELF%S_LINEAR
        IF (ALLOCATED(SELF%WP_RES)) THEN
            IF (VVERBOSE) WRITE(*,'(A)') 'Linear fit residuals'
            IF (VVERBOSE) CALL WAYPOINTS_PRINT_INFO(SELF%WP_RES)
        ELSE
            IF (VVERBOSE) WRITE(*,'(A)') 'Residuals not set up'
        ENDIF

        ! Print out spline information
        if (vverbose) then
            write(*,'(a)') 'Spline X'
            call interpol_spline_print_info(self%x_spline)
            write(*,'(a)') 'Spline Y'
            call interpol_spline_print_info(self%y_spline)
            write(*,'(a)') 'Spline Z'
            call interpol_spline_print_info(self%z_spline)
            write(*,'(a)') 'Spline Phi'
            call interpol_spline_print_info(self%phi_spline)
            write(*,'(a)') 'Spline S'
            call interpol_spline_print_info(self%s_spline)
            write(*,'(A//)') ' '
        endif

    END SUBROUTINE FILAMENT_PRINT_INFO

    !>  \brief  The T parameter is set to the current estimate of the arc length, plus the appropriate offset
    PURE SUBROUTINE FILAMENT_SET_T_FROM_PATH_LENGTH(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Private variables
        ! Start work
        SELF%T = SELF%WP%PATH_LENGTH
    END SUBROUTINE FILAMENT_SET_T_FROM_PATH_LENGTH

END MODULE FILAMENTS_BASE
