!>  \brief  Methods for the filament class which deal with geometry, euler angles, measurements etc.
!!
!!          Generally, these are methods which don't have to do with image processing.
MODULE FILAMENTS_GEOMETRY
    USE FILAMENTS_BASE, ONLY : FILAMENT
    IMPLICIT NONE

    CONTAINS

    !>  \brief  Change the subunit_num parameters from the usual values to incremental values
    !!
    !!  The "normal" representation has each waypoint's subunit_num value indicate directly the
    !!  value of the subunit_num function at that point of the space curve.
    !!  Alternatively, in the "incremental" representation, each waypoint's subunit_num value
    !!  is the difference in subunit_num between that waypoint and its preceding waypoint.
    !!  The filament's first waypoint is the only one which is unchanged by the transformation.
    !!  \todo deprecate and consolidate with the more general "params" subroutine
    PURE SUBROUTINE FILAMENT_SUBUNIT_NUM_ABSOLUTE_TO_INCREMENTAL(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_WP
        ! Start work

        ! Loop over waypoints
        DO CURRENT_WP=SELF%NWP,2,-1
            SELF%WP(CURRENT_WP)%SUBUNIT_NUM = SELF%WP(CURRENT_WP)%SUBUNIT_NUM - &
                                            SELF%WP(CURRENT_WP-1)%SUBUNIT_NUM
        ENDDO
    END SUBROUTINE FILAMENT_SUBUNIT_NUM_ABSOLUTE_TO_INCREMENTAL

    !>  \brief  Change the subunit_num representation from incremental to absolute
    PURE SUBROUTINE FILAMENT_SUBUNIT_NUM_INCREMENTAL_TO_ABSOLUTE(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_WP
        ! Start work

        ! Loop over waypoints
        DO CURRENT_WP=2,SELF%NWP
            SELF%WP(CURRENT_WP)%SUBUNIT_NUM = SELF%WP(CURRENT_WP)%SUBUNIT_NUM + &
                                            SELF%WP(CURRENT_WP-1)%SUBUNIT_NUM
        ENDDO
    END SUBROUTINE FILAMENT_SUBUNIT_NUM_INCREMENTAL_TO_ABSOLUTE

    !>  \brief  Change the filament parameter representation from incremental to absolute
    PURE SUBROUTINE FILAMENT_PARAMS_INCREMENTAL_TO_ABSOLUTE(SELF,PARAM_MASK)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        LOGICAL,        INTENT(IN)      ::  PARAM_MASK(7)
        ! Private variables
        INTEGER ::  CURRENT_WP
        ! Start work

        ! Loop over waypoints
        DO CURRENT_WP=2,SELF%NWP
            IF (PARAM_MASK(1)) THEN
                SELF%WP(CURRENT_WP)%COO(1) = SELF%WP(CURRENT_WP)%COO(1) + &
                                             SELF%WP(CURRENT_WP-1)%COO(1)
            ENDIF
            IF (PARAM_MASK(2)) THEN
                SELF%WP(CURRENT_WP)%COO(2) = SELF%WP(CURRENT_WP)%COO(2) + &
                                             SELF%WP(CURRENT_WP-1)%COO(2)
            ENDIF
            IF (PARAM_MASK(3)) THEN
                SELF%WP(CURRENT_WP)%COO(3) = SELF%WP(CURRENT_WP)%COO(3) + &
                                             SELF%WP(CURRENT_WP-1)%COO(3)
            ENDIF
            IF (PARAM_MASK(4)) THEN
                SELF%WP(CURRENT_WP)%EULER%PHI = SELF%WP(CURRENT_WP)%EULER%PHI + &
                                                SELF%WP(CURRENT_WP-1)%EULER%PHI
            ENDIF
            IF (PARAM_MASK(5)) THEN
                SELF%WP(CURRENT_WP)%EULER%THETA = SELF%WP(CURRENT_WP)%EULER%THETA + &
                                                  SELF%WP(CURRENT_WP-1)%EULER%THETA
            ENDIF
            IF (PARAM_MASK(6)) THEN
                SELF%WP(CURRENT_WP)%EULER%PSI = SELF%WP(CURRENT_WP)%EULER%PSI + &
                                                SELF%WP(CURRENT_WP-1)%EULER%PSI
            ENDIF
            IF (PARAM_MASK(7)) THEN
                SELF%WP(CURRENT_WP)%SUBUNIT_NUM = SELF%WP(CURRENT_WP)%SUBUNIT_NUM + &
                                                  SELF%WP(CURRENT_WP-1)%SUBUNIT_NUM
            ENDIF
        ENDDO
    END SUBROUTINE FILAMENT_PARAMS_INCREMENTAL_TO_ABSOLUTE

    PURE SUBROUTINE FILAMENT_PARAMS_ABSOLUTE_TO_INCREMENTAL(SELF,PARAM_MASK)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        LOGICAL,        INTENT(IN)      ::  PARAM_MASK(7)   !<  Which parameters should be switched
        ! Private variables
        INTEGER ::  CURRENT_WP
        ! Start work

        ! Loop over waypoints
        DO CURRENT_WP=SELF%NWP,2,-1
            IF (PARAM_MASK(1)) THEN
                SELF%WP(CURRENT_WP)%COO(1) = SELF%WP(CURRENT_WP)%COO(1) - &
                                             SELF%WP(CURRENT_WP-1)%COO(1)
            ENDIF
            IF (PARAM_MASK(2)) THEN
                SELF%WP(CURRENT_WP)%COO(2) = SELF%WP(CURRENT_WP)%COO(2) - &
                                             SELF%WP(CURRENT_WP-1)%COO(2)
            ENDIF
            IF (PARAM_MASK(3)) THEN
                SELF%WP(CURRENT_WP)%COO(3) = SELF%WP(CURRENT_WP)%COO(3) - &
                                             SELF%WP(CURRENT_WP-1)%COO(3)
            ENDIF
            IF (PARAM_MASK(4)) THEN
                SELF%WP(CURRENT_WP)%EULER%PHI = SELF%WP(CURRENT_WP)%EULER%PHI - &
                                                SELF%WP(CURRENT_WP-1)%EULER%PHI
            ENDIF
            IF (PARAM_MASK(5)) THEN
                SELF%WP(CURRENT_WP)%EULER%THETA = SELF%WP(CURRENT_WP)%EULER%THETA - &
                                                SELF%WP(CURRENT_WP-1)%EULER%THETA
            ENDIF
            IF (PARAM_MASK(6)) THEN
                SELF%WP(CURRENT_WP)%EULER%PSI = SELF%WP(CURRENT_WP)%EULER%PSI - &
                                                SELF%WP(CURRENT_WP-1)%EULER%PSI
            ENDIF
            IF (PARAM_MASK(7)) THEN
                SELF%WP(CURRENT_WP)%SUBUNIT_NUM = SELF%WP(CURRENT_WP)%SUBUNIT_NUM - &
                                                  SELF%WP(CURRENT_WP-1)%SUBUNIT_NUM
            ENDIF
        ENDDO
    END SUBROUTINE FILAMENT_PARAMS_ABSOLUTE_TO_INCREMENTAL


    !>  \brief  Return the average (X,Y,Z) coordinates of the filament
    FUNCTION FILAMENT_AVERAGE_WP(SELF) RESULT(WP_AVE)
        USE COOS, ONLY : WAYPOINTS_AVE_STD, WAYPOINT
        IMPLICIT NONE
        ! Argument
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        ! Result
        TYPE(WAYPOINT)  ::  WP_AVE
        ! Private variables
        TYPE(WAYPOINT)  ::  WPS(2)
        ! Start work
        CALL WAYPOINTS_AVE_STD(SELF%WP,WPS)
        WP_AVE = WPS(1)
    END FUNCTION FILAMENT_AVERAGE_WP

    !>  \brief  For each spline, compute the residuals
    SUBROUTINE FILAMENT_UPDATE_WP_RES(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Private variables
        REAL(KIND=8)    ::  THETA_AVE, PSI_AVE
        ! Start work

        ! Safety check
        IF (ALL (SELF%T_LINEAR .EQ. 0.0D0)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_UPDATE_WP_RES): t_linear not setup'
            CALL TERMINATE('Fatal error in FILAMENT_UDPATE_WP_RES. Programming error.')
        ENDIF

        SELF%WP_RES%COO(1)      = SELF%WP%COO(1)      - SELF%X_LINEAR(1)   * SELF%T_LINEAR   - SELF%X_LINEAR(2)
        SELF%WP_RES%COO(2)      = SELF%WP%COO(2)      - SELF%Y_LINEAR(1)   * SELF%T_LINEAR   - SELF%Y_LINEAR(2)
        SELF%WP_RES%COO(3)      = SELF%WP%COO(3)      - SELF%Z_LINEAR(1)   * SELF%T_LINEAR   - SELF%Z_LINEAR(2)
        SELF%WP_RES%SUBUNIT_NUM = SELF%WP%SUBUNIT_NUM - SELF%S_LINEAR(1)   * SELF%T_LINEAR   - SELF%S_LINEAR(2)
        SELF%WP_RES%EULER%PHI   = SELF%WP%EULER%PHI   - SELF%PHI_LINEAR(1) * SELF%T_LINEAR   - SELF%PHI_LINEAR(2)

        ! They're not strictly residuals (since there is no spline), but we can fill the values for theta & psi with deviations
        ! from the mean
        PSI_AVE   = SUM(SELF%WP%EULER%PSI)   / REAL(SELF%NWP,KIND=8)
        THETA_AVE = SUM(SELF%WP%EULER%THETA) / REAL(SELF%NWP,KIND=8)
        SELF%WP_RES%EULER%THETA = SELF%WP%EULER%THETA - THETA_AVE
        SELF%WP_RES%EULER%PSI   = SELF%WP%EULER%PSI   - PSI_AVE
    END SUBROUTINE FILAMENT_UPDATE_WP_RES

    !>  \brief  Re-compute waypoint values from the residuals
    SUBROUTINE FILAMENT_UPDATE_WPS_FROM_WP_RES(SELF,AXIS,PHI,SUBUNIT_NUM)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  AXIS        !<  Update axis properties (X,Y,Z)
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  PHI         !<  Update phi
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  SUBUNIT_NUM !<  Update subunit_num
        ! Private variables
        LOGICAL ::  AAXIS, PPHI, SSUBUNIT_NUM
        ! Start work
        AAXIS = .TRUE.
        IF (PRESENT(AXIS)) AAXIS = AXIS
        PPHI = .TRUE.
        IF (PRESENT(PHI)) PPHI = PHI
        SSUBUNIT_NUM = .TRUE.
        IF (PRESENT(SUBUNIT_NUM)) SSUBUNIT_NUM = SUBUNIT_NUM
        !
        IF (AAXIS) THEN
            SELF%WP%COO(1)      = SELF%WP_RES%COO(1)      + SELF%X_LINEAR(1)   * SELF%T_LINEAR   + SELF%X_LINEAR(2)
            SELF%WP%COO(2)      = SELF%WP_RES%COO(2)      + SELF%Y_LINEAR(1)   * SELF%T_LINEAR   + SELF%Y_LINEAR(2)
            SELF%WP%COO(3)      = SELF%WP_RES%COO(3)      + SELF%Z_LINEAR(1)   * SELF%T_LINEAR   + SELF%Z_LINEAR(2)
        ENDIF
        IF (SSUBUNIT_NUM) THEN
            SELF%WP%SUBUNIT_NUM = SELF%WP_RES%SUBUNIT_NUM + SELF%S_LINEAR(1)   * SELF%T_LINEAR   + SELF%S_LINEAR(2)
        ENDIF
        IF (PPHI) THEN
            SELF%WP%EULER%PHI   = SELF%WP_RES%EULER%PHI   + SELF%PHI_LINEAR(1) * SELF%T_LINEAR   + SELF%PHI_LINEAR(2)
        ENDIF
    END SUBROUTINE FILAMENT_UPDATE_WPS_FROM_WP_RES


    !>  \brief  Compute a linear fit of the splines
    SUBROUTINE FILAMENT_FIT_SPLINES_LINEAR(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)  ::  SELF
        ! Private variables
        INTEGER ::  I
        REAL(KIND=8)    ::  S, ST, STT, SX, SY, SZ, SS, SPHI
        REAL(KIND=8)    ::  STX, STY, STZ, STS, STPHI
        REAL(KIND=8)    ::  D
        LOGICAL         ::  ALL_SPLINE_ORDINATES_MATCH
        ! Start work

        ! Keep a record of the ordinates
        SELF%T_LINEAR = SELF%T  !SELF%X_SPLINE%X

        ! Sanity check. The ordinates should be the same for all the splines
        ALL_SPLINE_ORDINATES_MATCH =    ALL(SELF%T_LINEAR .EQ. SELF%X_SPLINE%X) .AND. &
                                        ALL(SELF%T_LINEAR .EQ. SELF%Y_SPLINE%X) .AND. &
                                        ALL(SELF%T_LINEAR .EQ. SELF%Z_SPLINE%X) .AND. &
                                        ALL(SELF%T_LINEAR .EQ. SELF%S_SPLINE%X) .AND. &
                                        ALL(SELF%T_LINEAR .EQ. SELF%PHI_SPLINE%X )
        IF (.NOT. ALL_SPLINE_ORDINATES_MATCH) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_FIT_SPLINES_LINEAR): not all splines have the same parameter values'
            CALL TERMINATE('Fatal error in FILAMENT_FIT_SPLINES_LINEAR. Programming error.')
        ENDIF

        ! See eq 15.2.4 and onwards of Numerical recipes
        S        = REAL(SELF%X_SPLINE%N,KIND=8)
        ST       = 0.0D0
        STT      = 0.0D0
        SX       = 0.0D0
        SY       = 0.0D0
        SZ       = 0.0D0
        SS       = 0.0D0
        SPHI     = 0.0D0
        STX      = 0.0D0
        STY      = 0.0D0
        STZ      = 0.0D0
        STS      = 0.0D0
        STPHI    = 0.0D0
        DO I=1,SELF%X_SPLINE%N
            ! Accumulate sums
            ST       =   ST      + SELF%T_LINEAR(I)
            STT      =   STT     + SELF%T_LINEAR(I)**2
            SX       =   SX      + SELF%X_SPLINE%Y(I)
            SY       =   SY      + SELF%Y_SPLINE%Y(I)
            SZ       =   SZ      + SELF%Z_SPLINE%Y(I)
            SS       =   SS      + SELF%S_SPLINE%Y(I)
            SPHI     =   SPHI    + SELF%PHI_SPLINE%Y(I)
            STX      =   STX     + SELF%X_SPLINE%X(I) *   SELF%X_SPLINE%Y(I)
            STY      =   STY     + SELF%X_SPLINE%X(I) *   SELF%Y_SPLINE%Y(I)
            STZ      =   STZ     + SELF%X_SPLINE%X(I) *   SELF%Z_SPLINE%Y(I)
            STS      =   STS     + SELF%X_SPLINE%X(I) *   SELF%S_SPLINE%Y(I)
            STPHI    =   STPHI   + SELF%X_SPLINE%X(I) *   SELF%PHI_SPLINE%Y(I)
        ENDDO

        D = S*STT - ST**2
        !
        SELF%X_LINEAR(1)   = ( S*STX    - ST*SX    ) / D
        SELF%X_LINEAR(2)   = ( STT*SX   - ST*STX   ) / D

        SELF%Y_LINEAR(1)   = ( S*STY    - ST*SY    ) / D
        SELF%Y_LINEAR(2)   = ( STT*SY   - ST*STY   ) / D

        SELF%Z_LINEAR(1)   = ( S*STZ    - ST*SZ    ) / D
        SELF%Z_LINEAR(2)   = ( STT*SZ   - ST*STZ   ) / D

        SELF%S_LINEAR(1)   = ( S*STS    - ST*SS    ) / D
        SELF%S_LINEAR(2)   = ( STT*SS   - ST*STS   ) / D

        SELF%PHI_LINEAR(1) = ( S*STPHI  - ST*SPHI  ) / D
        SELF%PHI_LINEAR(2) = ( STT*SPHI - ST*STPHI ) / D

    END SUBROUTINE FILAMENT_FIT_SPLINES_LINEAR

    !>  \brief  Find the maximal normal distance for the filament
    SUBROUTINE FILAMENT_MAXIMAL_NORMAL_DISTANCE(SELF,MAXIMAL_NORMAL_DISTANCE)
        USE COOS, ONLY : WAYPOINT_WAYPOINT_DISTANCE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        REAL,           INTENT(OUT)     ::  MAXIMAL_NORMAL_DISTANCE
        ! Private variables
        INTEGER         ::  START_WP, FINISH_WP
        REAL            ::  CHORD_LENGTH, CHORD_VECTOR(3)
        REAL(KIND=8)    ::  T
        REAL            ::  NORMAL_DISTANCE
        INTEGER         ::  I
        ! Start work

        ! Initialise the output
        MAXIMAL_NORMAL_DISTANCE = 0.0

        ! Define the chord axis
        IF (SELF%FAL_WPS_IN_THREED .AND. SELF%FAL_WPS_IN_SCORE) THEN
            START_WP = 1
            FINISH_WP = SELF%NWP
        ELSE
            START_WP = 2
            FINISH_WP = SELF%NWP - 1
        ENDIF

        ! Chord length
        CHORD_VECTOR = SELF%WP(FINISH_WP)%COO - SELF%WP(START_WP)%COO
        CHORD_LENGTH = SQRT(CHORD_VECTOR(1)**2 + CHORD_VECTOR(2)**2 + CHORD_VECTOR(3)**2)

        ! Loop along the filament (1 pixel per step) and look for the longest normal distance
        DO I=1,INT(SELF%T(FINISH_WP)-SELF%T(START_WP))+1
            T = SELF%T(START_WP) + DBLE(I)
            NORMAL_DISTANCE =  FILAMENT_EVAL_NORMAL_DISTANCE(SELF,SELF%WP(START_WP)%COO,CHORD_VECTOR,CHORD_LENGTH,T)
            IF (NORMAL_DISTANCE .GT. MAXIMAL_NORMAL_DISTANCE) THEN
                MAXIMAL_NORMAL_DISTANCE = NORMAL_DISTANCE
            ENDIF
        ENDDO



    END SUBROUTINE FILAMENT_MAXIMAL_NORMAL_DISTANCE

    !>  \brief  Evaluate the mid-point normal distance for the filament, in 3D and in 2D projection
    SUBROUTINE FILAMENT_MID_POINT_NORMAL_DISTANCE(SELF,MID_POINT_NORMAL_DIST,MID_POINT_NORMAL_DIST_2D)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(INOUT)   ::  SELF
        REAL,               INTENT(OUT)     ::  MID_POINT_NORMAL_DIST
        REAL,               INTENT(OUT)     ::  MID_POINT_NORMAL_DIST_2D
        ! Private variables
        INTEGER ::  START_WP, FINISH_WP
        REAL    ::  CHORD_LENGTH, CHORD_VECTOR(3)
        ! Start work

        ! Define the first and last waypoints we're considering
        IF (SELF%FAL_WPS_IN_THREED .AND. SELF%FAL_WPS_IN_SCORE) THEN
            START_WP = 1
            FINISH_WP = SELF%NWP
        ELSE
            START_WP = 2
            FINISH_WP = SELF%NWP - 1
        ENDIF

        !
        CHORD_VECTOR = SELF%WP(FINISH_WP)%COO - SELF%WP(START_WP)%COO
        CHORD_LENGTH = SQRT(CHORD_VECTOR(1)**2+CHORD_VECTOR(2)**2+CHORD_VECTOR(3)**2)

        !
        MID_POINT_NORMAL_DIST    = FILAMENT_EVAL_NORMAL_DISTANCE(SELF,SELF%WP(START_WP)%COO,CHORD_VECTOR,CHORD_LENGTH,          &
                                                                 T=(SELF%T(FINISH_WP)-SELF%T(START_WP))/2+SELF%T(START_WP))
        MID_POINT_NORMAL_DIST_2D = FILAMENT_EVAL_NORMAL_DISTANCE(SELF,SELF%WP(START_WP)%COO,CHORD_VECTOR,CHORD_LENGTH,          &
                                                                 T=(SELF%T(FINISH_WP)-SELF%T(START_WP))/2+SELF%T(START_WP),     &
                                                                 TWO_D=.TRUE.)

    END SUBROUTINE FILAMENT_MID_POINT_NORMAL_DISTANCE

    REAL FUNCTION FILAMENT_EVAL_NORMAL_DISTANCE(SELF,START_COO,CHORD_VECTOR,CHORD_LENGTH,T,TWO_D) RESULT(NORMAL_DISTANCE)
        USE COOS, ONLY  :   WAYPOINT
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),         INTENT(INOUT)   ::  SELF                !<  Filament object
        REAL(KIND=8),           INTENT(IN)      ::  START_COO(3)        !<  Coordinates of the first waypoint considered
        REAL,                   INTENT(IN)      ::  CHORD_VECTOR(3)     !<  Chord vector
        REAL,                   INTENT(IN)      ::  CHORD_LENGTH        !<  Chord length
        REAL(KIND=8),           INTENT(IN)      ::  T
        LOGICAL,    OPTIONAL,   INTENT(IN)      ::  TWO_D               !<  Return the 2D normal distance
        ! Private variables
        REAL(KIND=8)                    ::  COO(3)
        REAL                            ::  WP_VECTOR(3)
        REAL                            ::  CROSS_PRODUCT(3)
        LOGICAL                         ::  TTWO_D
        ! Start work

        TTWO_D = .FALSE.
        IF (PRESENT(TWO_D)) TTWO_D = TWO_D

        !
        CALL INTERPOL_SPLINE_EVAL(SELF%X_SPLINE,T,COO(1))
        CALL INTERPOL_SPLINE_EVAL(SELF%Y_SPLINE,T,COO(2))
        CALL INTERPOL_SPLINE_EVAL(SELF%Z_SPLINE,T,COO(3))

        WP_VECTOR = COO - START_COO

        ! Compute the cross product
        ! u = chord
        ! v = wp
        CROSS_PRODUCT(1) =  CHORD_VECTOR(2) * WP_VECTOR(3) - CHORD_VECTOR(3) * WP_VECTOR(2)
        CROSS_PRODUCT(2) =  CHORD_VECTOR(3) * WP_VECTOR(1) - CHORD_VECTOR(1) * WP_VECTOR(3)
        CROSS_PRODUCT(3) =  CHORD_VECTOR(1) * WP_VECTOR(2) - CHORD_VECTOR(2) * WP_VECTOR(1)

        !
        IF (TTWO_D) THEN
            NORMAL_DISTANCE = SQRT(CROSS_PRODUCT(1)**2 + CROSS_PRODUCT(2)**2)  &
                                    / CHORD_LENGTH
        ELSE
            NORMAL_DISTANCE = SQRT(CROSS_PRODUCT(1)**2 + CROSS_PRODUCT(2)**2 + CROSS_PRODUCT(3)**2)  &
                                    / CHORD_LENGTH
        ENDIF
    END FUNCTION FILAMENT_EVAL_NORMAL_DISTANCE


    !>  \brief  Return the chord length of the filament
    !!
    !!  \todo   Ignore FAL WPS?
    SUBROUTINE FILAMENT_CHORD_LENGTH(SELF,CHORD_LENGTH_3D,CHORD_LENGTH_2D)
        USE COOS, ONLY : WAYPOINT_WAYPOINT_DISTANCE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        REAL,           INTENT(OUT) ::  CHORD_LENGTH_3D
        REAL,           INTENT(OUT) ::  CHORD_LENGTH_2D
        ! Private variables
        INTEGER ::  START_WP, FINISH_WP
        ! Start work
        ! Define the chord axisSELF%NWP
        IF (SELF%FAL_WPS_IN_THREED .AND. SELF%FAL_WPS_IN_SCORE) THEN
            START_WP = 1
            FINISH_WP = SELF%NWP
        ELSE
            START_WP = 2
            FINISH_WP = SELF%NWP - 1
        ENDIF
        CHORD_LENGTH_3D = WAYPOINT_WAYPOINT_DISTANCE(SELF%WP(START_WP),SELF%WP(FINISH_WP))
        CHORD_LENGTH_2D = SQRT(     (SELF%WP(FINISH_WP)%COO(1)-SELF%WP(START_WP)%COO(1))**2 &
                                +   (SELF%WP(FINISH_WP)%COO(2)-SELF%WP(START_WP)%COO(2))**2)
    END SUBROUTINE FILAMENT_CHORD_LENGTH

    !>  \brief  Return the arc length of the filament (i.e. the distance along the helical axis)
    PURE REAL FUNCTION FILAMENT_ARC_LENGTH(SELF,IGNORE_FAL_WPS)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(IN)  ::  SELF
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  IGNORE_FAL_WPS      !<  Ignore the first and last waypoints
        ! Private variables
        LOGICAL ::  IIGNORE_FAL_WPS
        ! Start work
        IIGNORE_FAL_WPS = .FALSE.
        IF (PRESENT(IGNORE_FAL_WPS)) IIGNORE_FAL_WPS = IGNORE_FAL_WPS

        IF (IIGNORE_FAL_WPS) THEN
            FILAMENT_ARC_LENGTH = SELF%WP(SELF%NWP-1)%PATH_LENGTH - SELF%WP(2)%PATH_LENGTH
        ELSE
            FILAMENT_ARC_LENGTH = SELF%WP(SELF%NWP  )%PATH_LENGTH - SELF%WP(1)%PATH_LENGTH
        ENDIF
    END FUNCTION FILAMENT_ARC_LENGTH

    !>  \broef  Find the average changes in phi and subunit_num per unit length
    subroutine filament_eval_average_dphi_ds(self,average_dphi,average_ds,ignore_fal_wps)
        use splines, only : interpol_spline_eval_d
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        real(kind=8),               intent(out)     ::  average_dphi
        real(kind=8),               intent(out)     ::  average_ds
        logical,        optional,   intent(in)      ::  ignore_fal_wps
        ! Private variables
        logical         ::  iignore_fal_wps
        integer         ::  first_wp,last_wp
        real(kind=8)    ::  step_size, current_t
        integer         ::  num_steps
        ! Ignore FAL?
        iignore_fal_wps = .not. self%FAL_WPS_IN_SCORE
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        ! First and last waypoint
        if (iignore_fal_wps) then
            first_wp = 2
            last_wp  = self%nwp -1
        else
            first_wp = 1
            last_wp  = self%nwp
        endif

        ! Step size for search
        if (self%helix%pixel_units) then
            step_size = self%helix%rise
        else
            step_size = self%helix%rise / self%film%pixel_size
        endif

        ! Search the filament
        current_t = self%t(first_wp)
        average_dphi = 0.0d0
        average_ds   = 0.0d0
        num_steps    = 0
        do
            num_steps = num_steps + 1
            !
            average_dphi = average_dphi + interpol_spline_eval_d(self%phi_spline,current_t)
            average_ds   = average_ds   + interpol_spline_eval_d(self%s_spline,current_t)
            !
            current_t = current_t + step_size
            if (current_t .gt. self%t(last_wp)) exit
        enddo
        average_dphi = average_dphi / real(num_steps,kind=8)
        average_ds   = average_ds   / real(num_steps,kind=8)
    end subroutine filament_eval_average_dphi_ds


    !>  \brief  Find the largest local curvature of a filament.
    subroutine filament_eval_maximum_curvature_overtwist_stretch(self,max_curvature,max_overtwist,max_stretch,ignore_fal_wps)
        ! Arguments
        type(filament),             intent(inout)   ::  self
        real(kind=8),               intent(out)     ::  max_curvature
        real(kind=8),               intent(out)     ::  max_overtwist
        real(kind=8),               intent(out)     ::  max_stretch
        logical,        optional,   intent(in)      ::  ignore_fal_wps      !<  ignore the first and last segments
        ! Private variables
        logical         ::  iignore_fal_wps
        integer         ::  first_wp,last_wp
        real(kind=8)    ::  step_size, current_t, current_c, current_o, current_s
        real(kind=8)    ::  average_dphi, average_ds
        logical, parameter  ::  debug = .false.
        ! Start work

        if (debug) write(*,'(a)') '**debug(filament_eval_maximum_curvature_overtwist_stretch): entering'

        ! Ignore FAL?
        iignore_fal_wps = .not. self%FAL_WPS_IN_SCORE
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        ! First and last waypoint
        if (iignore_fal_wps) then
            first_wp = 2
            last_wp  = self%nwp -1
        else
            first_wp = 1
            last_wp  = self%nwp
        endif

        ! Step size for search
        if (self%helix%pixel_units) then
            step_size = self%helix%rise
        else
            step_size = self%helix%rise / self%film%pixel_size
        endif

        if (debug) write(*,'(a)') '**debug(filament_eval_maximum_curvature_overtwist_stretch): about to get average dphi ds'

        !
        call filament_eval_average_dphi_ds(self,average_dphi,average_ds,ignore_fal_wps=iignore_fal_wps)

        if (debug) write(*,'(a)') '**debug(filament_eval_maximum_curvature_overtwist_stretch): about to loop to find max'

        ! Search the filament
        current_t = self%t(first_wp)
        max_curvature = 0.0d0
        max_overtwist = 0.0d0
        max_stretch   = 0.0d0
        do
            current_c = filament_eval_curvature(self,current_t)
            current_o = abs(filament_eval_overtwist(self,current_t,average_dphi))
            current_s = abs(filament_eval_stretch(self,current_t,average_ds))
            if (current_c .gt. max_curvature) max_curvature = current_c
            if (current_o .gt. max_overtwist) max_overtwist = current_o
            if (current_s .gt. max_stretch  ) max_stretch   = current_s
            current_t = current_t + step_size
            if (current_t .gt. self%t(last_wp)) exit
        enddo

        if (debug) write(*,'(a)') '**debug(filament_eval_maximum_curvature_overtwist_stretch): exiting'
    end subroutine filament_eval_maximum_curvature_overtwist_stretch


    !>  \brief  Find the largest local curvature of a filament.
    subroutine filament_eval_average_curvature_overtwist_stretch(self,ave_curvature,ave_overtwist,ave_stretch,ignore_fal_wps)
        ! Arguments
        type(filament),             intent(inout)   ::  self
        real(kind=8),               intent(out)     ::  ave_curvature
        real(kind=8),               intent(out)     ::  ave_overtwist
        real(kind=8),               intent(out)     ::  ave_stretch
        logical,        optional,   intent(in)      ::  ignore_fal_wps      !<  ignore the first and last segments
        ! Private variables
        logical         ::  iignore_fal_wps
        integer         ::  first_wp,last_wp
        real(kind=8)    ::  step_size, current_t
        real(kind=8)    ::  average_dphi, average_ds
        integer         ::  num_steps
        ! Start work

        ! Ignore FAL?
        iignore_fal_wps = .not. self%FAL_WPS_IN_SCORE
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        ! First and last waypoint
        if (iignore_fal_wps) then
            first_wp = 2
            last_wp  = self%nwp -1
        else
            first_wp = 1
            last_wp  = self%nwp
        endif

        ! Step size for search
        if (self%helix%pixel_units) then
            step_size = self%helix%rise
        else
            step_size = self%helix%rise / self%film%pixel_size
        endif

        !
        call filament_eval_average_dphi_ds(self,average_dphi,average_ds,ignore_fal_wps=iignore_fal_wps)

        ! Search the filament
        current_t = self%t(first_wp)
        ave_curvature = 0.0d0
        ave_overtwist = 0.0d0
        ave_stretch   = 0.0d0
        num_steps     = 0
        do
            num_steps = num_steps + 1

            ave_curvature = ave_curvature + filament_eval_curvature(self,current_t)
            ave_overtwist = ave_overtwist + abs(filament_eval_overtwist(self,current_t,average_dphi))
            ave_stretch   = ave_stretch   + abs(filament_eval_stretch(self,current_t,average_ds))

            current_t = current_t + step_size
            if (current_t .gt. self%t(last_wp)) exit
        enddo
        ave_curvature = ave_curvature / real(num_steps,kind=8)
        ave_overtwist = ave_overtwist / real(num_steps,kind=8)
        ave_stretch   = ave_stretch   / real(num_steps,kind=8)
    end subroutine filament_eval_average_curvature_overtwist_stretch

    !>  \brief  Evaluate the total curvature along a filament. Optionally, the total squared curvature can be returned.
    FUNCTION FILAMENT_EVAL_AVERAGE_CURVATURE(SELF,IGNORE_FAL_WPS) RESULT(AVERAGE_CURVATURE)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  IGNORE_FAL_WPS      !<  Whether to ignore the first and last waypoints in the calculation
        ! Result
        REAL(KIND=8)    ::  AVERAGE_CURVATURE
        ! Private variables
        REAL(KIND=8)    ::  STEP_SIZE
        REAL(KIND=8)    ::  CURRENT_T
        LOGICAL         ::  IIGNORE_FAL_WPS
        INTEGER         ::  FIRST_WP, LAST_WP
        INTEGER         ::  COUNT_STEPS
        ! Start work

        ! Ignore FAL WPS?
        IF (PRESENT(IGNORE_FAL_WPS)) THEN
            IIGNORE_FAL_WPS = IGNORE_FAL_WPS
        ELSE
            IIGNORE_FAL_WPS = .NOT. SELF%FAL_WPS_IN_SCORE
        ENDIF

        ! First and last waypoints
        IF (IIGNORE_FAL_WPS) THEN
            FIRST_WP    =   2
            LAST_WP     =   SELF%NWP - 1
        ELSE
            FIRST_WP    =   1
            LAST_WP     =   SELF%NWP
        ENDIF

        ! The step size for the integration
        IF (SELF%HELIX%PIXEL_UNITS) THEN
            STEP_SIZE = SELF%HELIX%RISE
        ELSE
            STEP_SIZE = SELF%HELIX%RISE/SELF%FILM%PIXEL_SIZE
        ENDIF

        ! Total curvature
        AVERAGE_CURVATURE = 0.0D0
        COUNT_STEPS = 0
        CURRENT_T = SELF%T(FIRST_WP)
        DO
            AVERAGE_CURVATURE = AVERAGE_CURVATURE + FILAMENT_EVAL_CURVATURE(SELF,CURRENT_T)
            CURRENT_T = CURRENT_T + STEP_SIZE
            COUNT_STEPS = COUNT_STEPS + 1
            IF (CURRENT_T .GT. SELF%T(LAST_WP)) EXIT
        ENDDO
        AVERAGE_CURVATURE = AVERAGE_CURVATURE / COUNT_STEPS
    END FUNCTION FILAMENT_EVAL_AVERAGE_CURVATURE

    !>  \brief  Evaluate the total change of direction of the filament in radians, i.e. as though all the curvature was "flattened" into a single arc of circle
    REAL FUNCTION FILAMENT_EVAL_CHANGE_OF_DIRECTION(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF
        ! Private variables
        ! Start work
        FILAMENT_EVAL_CHANGE_OF_DIRECTION = FILAMENT_EVAL_TOTAL_CURVATURE(SELF,SQUARED_CURVATURE=.FALSE.)
    END FUNCTION FILAMENT_EVAL_CHANGE_OF_DIRECTION


    !>  \brief  Evaluate the total curvature along a filament. Optionally, the total squared curvature can be returned.
    FUNCTION FILAMENT_EVAL_TOTAL_CURVATURE(SELF,SQUARED_CURVATURE,IGNORE_FAL_WPS) RESULT(TOTAL_CURVATURE)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  SQUARED_CURVATURE   !<  If .TRUE., the total squared curvature will be returned. Defaults to false, in which case the total curvature (not squared) is returned.
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  IGNORE_FAL_WPS      !<  Whether to ignore the first and last waypoints in the calculation. False by default.
        ! Result
        REAL(KIND=8)    ::  TOTAL_CURVATURE
        ! Private variables
        LOGICAL         ::  SSQUARED_CURVATURE
        REAL(KIND=8)    ::  STEP_SIZE
        REAL(KIND=8)    ::  CURRENT_T
        LOGICAL         ::  IIGNORE_FAL_WPS
        INTEGER         ::  FIRST_WP, LAST_WP
        ! Start work
        ! Will we be returning the squared curvature?
        IF (PRESENT(SQUARED_CURVATURE)) THEN
            SSQUARED_CURVATURE = SQUARED_CURVATURE
        ELSE
            SSQUARED_CURVATURE = .TRUE.
        ENDIF

        ! Ignore FAL WPS?
        IF (PRESENT(IGNORE_FAL_WPS)) THEN
            IIGNORE_FAL_WPS = IGNORE_FAL_WPS
        ELSE
            IIGNORE_FAL_WPS = .NOT. SELF%FAL_WPS_IN_SCORE
        ENDIF

        ! First and last waypoints
        IF (IIGNORE_FAL_WPS) THEN
            FIRST_WP    =   2
            LAST_WP     =   SELF%NWP - 1
        ELSE
            FIRST_WP    =   1
            LAST_WP     =   SELF%NWP
        ENDIF

        ! The step size for the integration
        IF (SELF%HELIX%PIXEL_UNITS) THEN
            STEP_SIZE = SELF%HELIX%RISE
        ELSE
            STEP_SIZE = SELF%HELIX%RISE/SELF%FILM%PIXEL_SIZE
        ENDIF

        ! Total curvature
        TOTAL_CURVATURE = 0.0D0
        CURRENT_T = SELF%T(FIRST_WP)
        IF (SSQUARED_CURVATURE) THEN
            DO
                TOTAL_CURVATURE = TOTAL_CURVATURE + FILAMENT_EVAL_CURVATURE(SELF,CURRENT_T)**2
                CURRENT_T = CURRENT_T + STEP_SIZE
                IF (CURRENT_T .GT. SELF%T(LAST_WP)) EXIT
            ENDDO
        ELSE
            DO
                TOTAL_CURVATURE = TOTAL_CURVATURE + FILAMENT_EVAL_CURVATURE(SELF,CURRENT_T)
                CURRENT_T = CURRENT_T + STEP_SIZE
                IF (CURRENT_T .GT. SELF%T(LAST_WP)) EXIT
            ENDDO
        ENDIF
        TOTAL_CURVATURE = TOTAL_CURVATURE * STEP_SIZE
    END FUNCTION FILAMENT_EVAL_TOTAL_CURVATURE

    !>  \brief  Evaluate the local curvature
    FUNCTION FILAMENT_EVAL_CURVATURE(SELF,T) RESULT(CURVATURE)
        USE SPLINES, ONLY : INTERPOL_SPLINE_HORNER_EVAL_ALL
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        REAL(KIND=8),   INTENT(IN)      ::  T
        ! Result
        REAL(KIND=8)    ::  CURVATURE
        ! Private variables
        REAL(KIND=8)    ::  X,Y,Z
        REAL(KIND=8)    ::  DX,DY,DZ
        REAL(KIND=8)    ::  DDX,DDY,DDZ
		LOGICAL,	PARAMETER	::	ASSUME_ARC_LENGTH_PARAMETRISATION = .FALSE.
        ! Start work


        CALL INTERPOL_SPLINE_HORNER_EVAL_ALL(SELF%X_SPLINE,T,X,DX,DDX)
        CALL INTERPOL_SPLINE_HORNER_EVAL_ALL(SELF%Y_SPLINE,T,Y,DY,DDY)
        CALL INTERPOL_SPLINE_HORNER_EVAL_ALL(SELF%Z_SPLINE,T,Z,DZ,DDZ)

		IF (ASSUME_ARC_LENGTH_PARAMETRISATION) THEN
			CURVATURE = SQRT(DDX**2+DDY**2+DDZ**2)
		ELSE
        	CURVATURE = SQRT((DDZ*DY-DDY*DZ)**2+(DDX*DZ-DDZ*DX)**2+(DDY*DX-DDX*DY)**2)  &
                    	/   &
                    	!(DX**2+DY**2+DZ**2)**(3/2)
                    	SQRT((DX**2+DY**2+DZ**2))**3
		ENDIF

    END FUNCTION FILAMENT_EVAL_CURVATURE

    !>  \brief  Evaluate the curvature vector, which should be normal to the tangent vector.
    !!
    !!  With an ideal arc-length parametrised space curve, this would just be the second derivative vector.
    !!  However, small errors creep in when the space curve is not exactly arc-length parametrised. This aims to correct for those errors.
    !!  See notes 131107 for derivation
    !!
    !!  \todo make more efficient
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE FILAMENT_EVAL_CURVATURE_VECTOR(SELF,T,DX,DY,DZ,DDX,DDY,DDZ,KX,KY,KZ)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        REAL(KIND=8),   INTENT(IN)      ::  T
        REAL(KIND=8),   INTENT(IN)      ::  DX,DY,DZ    !<  Tangent (first derivative) vector
        REAL(KIND=8),   INTENT(IN)      ::  DDX,DDY,DDZ !<  Second derivative vector
        REAL(KIND=8),   INTENT(OUT)     ::  KX,KY,KZ    !<  Curvature vector
        ! Private variables
        REAL(KIND=8)    ::  T_NORM
        REAL(KIND=8)    ::  TXDT_X,TXDT_Y,TXDT_Z,TXDT_NORM
        !REAL(KIND=8)    ::  TDT
        REAL(KIND=8)    ::  CURV
        LOGICAL,PARAMETER   ::  SANITY_CHECK = .TRUE.
        ! Start work

        ! The norm of the first and second derivatives
        T_NORM    = DSQRT(DX**2 +DY**2 +DZ**2)
        !DT_NORM   = SQRT(DDX**2+DDY**2+DDZ**2)

        ! Cross product of first and second derivative vectors
        TXDT_X = DY*DDZ-DZ*DDY
        TXDT_Y = DZ*DDX-DX*DDZ
        TXDT_Z = DX*DDY-DY*DDX
        IF (SANITY_CHECK) TXDT_NORM = DSQRT(TXDT_X**2 + TXDT_Y**2 + TXDT_Z**2)

        ! Curvature vector (alt)
        KX = TXDT_X / T_NORM**3
        KY = TXDT_Y / T_NORM**3
        KZ = TXDT_Z / T_NORM**3

#ifndef SKIP_RUNTIME_CHECKS
        ! Sanity check - |K| should be equal to curvature
        IF (SANITY_CHECK) THEN
            CURV      = FILAMENT_EVAL_CURVATURE(SELF,T)
            IF (ABS(DSQRT(KX**2+KY**2+KZ**2)-CURV) .GT. 0.0001) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_EVAL_CURVATURE_VECTOR): norm of output not as expected'
                PRINT *, 'TXDT_X,Y,Z              = ', TXDT_X, TXDT_Y, TXDT_Z
                PRINT *, 'T_NORM**3               = ', T_NORM**3
                PRINT *, 'SQRT(KX**2+KY**2+KZ**2) = ', DSQRT(KX**2+KY**2+KZ**2)
                PRINT *, 'SQRT(TXDT_X**2+TXDT_Y**2+TXDT_Z**2) / T_NORM**3'
                PRINT *, '                        = ', DSQRT(TXDT_X**2+TXDT_Y**2+TXDT_Z**2) / T_NORM**3
                PRINT *, 'Straight from derivs    = ', DSQRT((DY*DDZ-DZ*DDY)**2+(DZ*DDX-DX*DDZ)**2+(DX*DDY-DY*DDX)**2) &
                                                        / DSQRT(DX**2 +DY**2 +DZ**2)**3
                PRINT *, 'CURV                    = ', CURV
                PRINT *, 'DIFF                    = ', DSQRT(KX**2+KY**2+KZ**2) - CURV
                CALL TERMINATE('Failed sanity check in FILAMENT_EVAL_CURVATURE_VECTOR')
            ENDIF
        ENDIF
#endif

    END SUBROUTINE FILAMENT_EVAL_CURVATURE_VECTOR


    !>  \brief  Evaluate the two principal curvatures
    !!
    !!  Note that arc-length parametrisation is assumed
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE FILAMENT_EVAL_PRINCIPAL_CURVATURES(SELF,T,CURVATURE_A,CURVATURE_B)
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL_D, INTERPOL_SPLINE_EVAL_DD
        USE ROTATION_MATRICES
        USE DIFFERENTIAL_GEOMETRY, ONLY : EULER_FROM_DERIVATIVES
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL, INTERPOL_SPLINE_PRINT_INFO, INTERPOL_SPLINE_HORNER_EVAL_ALL
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FLOAT_COMPARISONS, ONLY : DOUBLES_ARE_EQUAL
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF        !<  Filament object
        REAL(KIND=8),   INTENT(IN)      ::  T           !<  Arc length at which to evaluate the curvature
        REAL(KIND=8),   INTENT(OUT)     ::  CURVATURE_A !<  Curvature along the first principal axis
        REAL(KIND=8),   INTENT(OUT)     ::  CURVATURE_B !<  Curvature along the second principal axis
        ! Private variables
        LOGICAL,    PARAMETER   ::  SANITY_CHECK   = .TRUE.
        LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
        REAL(KIND=8)            ::  X,Y,Z, DX,DY,DZ, DDX,DDY,DDZ, PHI,THETA,PSI
!        REAL(KIND=8)            ::  CDX,CDY,CDZ,CDDX,CDDY,CDDZ
        TYPE(ROTATION_MATRIX)   ::  MATRIX
        REAL(KIND=8)            ::  COOS_IN_REC(4), COOS_IN_LAB(4)
        REAL(KIND=8)            ::  CURV
        REAL(KIND=8)            ::  D_DD_DOT_PRODUCT, D_DD_DOT_PRODUCT_XY
        REAL(KIND=8)            ::  KX,KY,KZ
        ! Start work

!        IF (SELF%FRAGMENTED_PHI) THEN
!            WRITE(*,'(A)') '**ERROR(FILAMENT_EVAL_PRINCIPAL_CURVATURES): only continuous description implemented for now.'
!            CALL TERMINATE('Fatal error FILAMENT_EVAL_PRINCIPAL_CURVATURES. Feature not yet implemented.')
!        ENDIF

        ! Let's get the derivatives
        CALL INTERPOL_SPLINE_HORNER_EVAL_ALL(SELF%X_SPLINE,T,X,DX,DDX)
        CALL INTERPOL_SPLINE_HORNER_EVAL_ALL(SELF%Y_SPLINE,T,Y,DY,DDY)
        CALL INTERPOL_SPLINE_HORNER_EVAL_ALL(SELF%Z_SPLINE,T,Z,DZ,DDZ)


        ! Let's get the curvature vector K
        CALL FILAMENT_EVAL_CURVATURE_VECTOR(SELF,T,DX,DY,DZ,DDX,DDY,DDZ,KX,KY,KZ)


#ifndef SKIP_RUNTIME_CHECKS
        ! First sanity check
        ! Let's check that the K and D vectors are orthgonal
        IF (SANITY_CHECK) THEN
            D_DD_DOT_PRODUCT    = (DX*KX + DY*KY + DZ*KZ)
            D_DD_DOT_PRODUCT_XY = (DX*KX + DY*KY           )
            IF (ABS(D_DD_DOT_PRODUCT) .GT. 0.00001) THEN
                WRITE(*,'(2A,G10.3)') '**ERROR(FILAMENT_EVAL_PRINCIPAL_CURVATURES): D and K vectors are not orthogonal. ', &
                                        'Dot product = ', D_DD_DOT_PRODUCT
                WRITE(*,'(A,G10.3)')  '**ERROR(FILAMENT_EVAL_PRINCIPAL_CURVATURES): angle = ', &
                                        DEGREES(ACOS(D_DD_DOT_PRODUCT/CURV))
                WRITE(*,'(A,G10.3)')  '**ERROR(FILAMENT_EVAL_PRINCIPAL_CURVATURES): angle (XY only) = ', &
                                        DEGREES(ACOS(D_DD_DOT_PRODUCT_XY/CURV))
            ENDIF
        ENDIF
#endif

        ! Let's get the rotation matrix to go from the lab frame to the filament reconstruction frame
        CALL EULER_FROM_DERIVATIVES(DX,DY,DZ,THETA,PSI)
        CALL INTERPOL_SPLINE_EVAL(SELF%PHI_SPLINE,T,PHI)
        MATRIX%RMATRIX  =   EULER_TO_ROTMAT(PHI,THETA,PSI,REVERSE=.TRUE.)

        ! Affine-transform the curvature (second derivative) vector to the filament frame of reference
        COOS_IN_LAB     =   (/KX,KY,KZ,1.0D0/)
        COOS_IN_REC     =   MATMUL(MATRIX%RMATRIX,COOS_IN_LAB)

        ! Now we have our result
        CURVATURE_A     =   COOS_IN_REC(1)
        CURVATURE_B     =   COOS_IN_REC(2)


#ifndef SKIP_RUNTIME_CHECKS
        IF (SANITY_CHECK) THEN
            CURV = FILAMENT_EVAL_CURVATURE(SELF,T)
            IF (DEBUG) PRINT '(A,G10.3)', '**debug(FILAMENT_EVAL_PRINCIPAL_CURVATURES): curvature = ', CURV
            IF (ABS(COOS_IN_REC(3)) .GT. 0.1D0 * CURV) THEN
                WRITE(*,'(A,G10.3)') '**ERROR(FILAMENT_EVAL_PRINCIPAL_CURVATURES): coos_in_rec(3) should be 0.0, but it is ', &
                                    COOS_IN_REC(3)
                CALL TERMINATE('Failed sanity check in FIALMENT_EVAL_PRINCIPAL_CURVATURES')
            ENDIF
            IF (ABS(SQRT(CURVATURE_A**2+CURVATURE_B**2) - CURV) .GT. 0.01*CURV) THEN
            !IF (.NOT. DOUBLES_ARE_EQUAL(SQRT(CURVATURE_A**2+CURVATURE_B**2),CURV)) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_EVAL_PRINCIPAL_CURVATURES): failed sanity check:'
                WRITE(*,'(A,G10.3)') 'curvature_a                         = ', CURVATURE_A
                WRITE(*,'(A,G10.3)') 'curvature_b                         = ', CURVATURE_B
                WRITE(*,'(A,G10.3)') 'curvature_c (should be 0.0)         = ', COOS_IN_REC(3)
                WRITE(*,'(A,G10.3)') 'sqrt(curvature_a**2+curvature_b**2) = ', SQRT(CURVATURE_A**2+CURVATURE_B**2)
                WRITE(*,'(A,G10.3)') 'curvature                           = ', CURV
                CALL TERMINATE('Fatal error in FILAMENT_EVAL_PRINCIPAL_CURVATURES')
            ENDIF
        ENDIF
#endif
    END SUBROUTINE FILAMENT_EVAL_PRINCIPAL_CURVATURES


    !>  \brief  Evaluate the total (integral) squared axial curvature, helical overtwist and stretching along a filament,
    !!          in reciprocal pixels.
    SUBROUTINE FILAMENT_EVAL_TOTAL_CURVATURE_OVERTWIST_STRETCH(SELF,CURVATURE,OVERTWIST,STRETCH,IGNORE_FAL_WPS)
        USE RUNTIME_PARAMETERS, ONLY : DPI
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF
        REAL(KIND=8),               INTENT(OUT)     ::  CURVATURE           !<  Total squared curvature integral (1/pixel)
        REAL(KIND=8),               INTENT(OUT)     ::  OVERTWIST           !<  Total squared overtwist integral (1/pixel)
        REAL(KIND=8),               INTENT(OUT)     ::  STRETCH             !<  Total squared stretch integral (1/pixel)
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  IGNORE_FAL_WPS      !<  Whether to ignore the first and last waypoints in the calculation. False by default.
        ! Result
        ! Private variables
        REAL(KIND=8)    ::  STEP_SIZE
        REAL(KIND=8)    ::  CURRENT_T
        LOGICAL         ::  IIGNORE_FAL_WPS
        INTEGER         ::  FIRST_WP, LAST_WP
        REAL(KIND=8)    ::  AVERAGE_DPHI    !<  Average change in phi per unit path length
        REAL(KIND=8)    ::  AVERAGE_DS      !<  Average change in subunit_num per unit path length
        ! Start work


        ! Ignore FAL WPS?
        IF (PRESENT(IGNORE_FAL_WPS)) THEN
            IIGNORE_FAL_WPS = IGNORE_FAL_WPS
        ELSE
            IIGNORE_FAL_WPS = .NOT. SELF%FAL_WPS_IN_SCORE
        ENDIF

        ! First and last waypoints
        IF (IIGNORE_FAL_WPS) THEN
            FIRST_WP    =   2
            LAST_WP     =   SELF%NWP - 1
        ELSE
            FIRST_WP    =   1
            LAST_WP     =   SELF%NWP
        ENDIF

        IF (LAST_WP .LE. FIRST_WP) THEN
            CURVATURE = 0.0D0
            OVERTWIST = 0.0D0
            STRETCH   = 0.0D0
            WRITE(*,'(A)') '**warning(FILAMENT_EVAL_TOTAL_CURVATURE_OVERTWIST_STRETCH): cannot compute. filament too short? '
        ELSE
        ! The step size for the integration
        IF (SELF%HELIX%PIXEL_UNITS) THEN
            STEP_SIZE = MAX(SELF%HELIX%RISE,1.0D0)
        ELSE
            STEP_SIZE = MAX(SELF%HELIX%RISE/SELF%FILM%PIXEL_SIZE,1.0D0)
        ENDIF

        ! The average change in phi per pixel along the filament
        !AVERAGE_DPHI = (SELF%WP(LAST_WP)%EULER%PHI - SELF%WP(FIRST_WP)%EULER%PHI) / (SELF%T(LAST_WP) - SELF%T(FIRST_WP))

        ! The average change in subunit_num per pixel along the filament
        !AVERAGE_DS = (SELF%WP(LAST_WP)%SUBUNIT_NUM - SELF%WP(FIRST_WP)%SUBUNIT_NUM) / (SELF%T(LAST_WP) - SELF%T(FIRST_WP))

        call filament_eval_average_dphi_ds(self,average_dphi,average_ds,ignore_fal_wps=iignore_fal_wps)

        ! Total squared torsion
        CURVATURE = 0.0D0
        OVERTWIST = 0.0D0
        STRETCH   = 0.0D0

        CURRENT_T = SELF%T(FIRST_WP)
        DO
            OVERTWIST = OVERTWIST + FILAMENT_EVAL_OVERTWIST(SELF,CURRENT_T,AVERAGE_DPHI)**2
            CURVATURE = CURVATURE + FILAMENT_EVAL_CURVATURE(SELF,CURRENT_T)**2
            STRETCH   = STRETCH   + FILAMENT_EVAL_STRETCH(SELF,CURRENT_T,AVERAGE_DS)**2
            CURRENT_T = CURRENT_T + STEP_SIZE
            IF (CURRENT_T .GT. SELF%T(LAST_WP)) EXIT
        ENDDO
        ! Convert to radians
        OVERTWIST = OVERTWIST / 180.0D0**2 * DPI**2
        ! The integral
        CURVATURE = CURVATURE * STEP_SIZE
        OVERTWIST = OVERTWIST * STEP_SIZE
        STRETCH   = STRETCH   * STEP_SIZE
        ENDIF
    END SUBROUTINE FILAMENT_EVAL_TOTAL_CURVATURE_OVERTWIST_STRETCH

    !>  \brief  Evaluate the local torsion at some point along the filament
    !!
    !!  Actually, we're evaluating the over/undertwist of the helical assembly (not the axial space curve's torsion).
    !!  The result is returned in degrees.
    PURE FUNCTION FILAMENT_EVAL_OVERTWIST(SELF,T,AVERAGE_TWIST_PER_DISTANCE) RESULT(TORSION)
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL_D
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        REAL(KIND=8),   INTENT(IN)  ::  T
        REAL(KIND=8),   INTENT(IN)  ::  AVERAGE_TWIST_PER_DISTANCE  !<  Average change in phi per pixel along the filament
        ! Result
        REAL(KIND=8)    ::  TORSION
        ! Private variables

        ! Start work
        TORSION = INTERPOL_SPLINE_EVAL_D(SELF%PHI_SPLINE,T) - AVERAGE_TWIST_PER_DISTANCE
    END FUNCTION FILAMENT_EVAL_OVERTWIST

    PURE FUNCTION FILAMENT_EVAL_STRETCH(SELF,T,AVERAGE_DS_PER_DISTANCE) RESULT(STRETCH)
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL_D
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        REAL(KIND=8),   INTENT(IN)  ::  T
        REAL(KIND=8),   INTENT(IN)  ::  AVERAGE_DS_PER_DISTANCE     !<  Average change in subunit_num per unit distance along filament
        ! Result
        REAL(KIND=8)    ::  STRETCH
        ! Private variables

        ! Start work
        STRETCH = INTERPOL_SPLINE_EVAL_D(SELF%S_SPLINE,T) - AVERAGE_DS_PER_DISTANCE
    END FUNCTION FILAMENT_EVAL_STRETCH


    !>  \brief  Evaluate the local helical twist at a point along the filament
    PURE REAL FUNCTION FILAMENT_EVAL_HELICAL_TWIST(SELF,T,RISE) RESULT(TWIST)
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL_D
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(IN)  ::  SELF
        REAL(KIND=8),               INTENT(IN)  ::  T
        REAL(KIND=8),   OPTIONAL,   INTENT(IN)  ::  RISE
        ! Private variables

        ! Start work
        IF (PRESENT(RISE)) THEN
            TWIST = INTERPOL_SPLINE_EVAL_D(SELF%PHI_SPLINE,T) * RISE
        ELSE
            TWIST = INTERPOL_SPLINE_EVAL_D(SELF%PHI_SPLINE,T) / INTERPOL_SPLINE_EVAL_D(SELF%S_SPLINE,T)
        ENDIF
    END FUNCTION FILAMENT_EVAL_HELICAL_TWIST

    !> \brief   Evaluate the local helical rise at a point along the filament
    PURE REAL FUNCTION FILAMENT_EVAL_HELICAL_RISE(SELF,T) RESULT(RISE)
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL_D
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        REAL(KIND=8),   INTENT(IN)  ::  T
        ! Private variables

        ! Start work
        RISE = 1. / INTERPOL_SPLINE_EVAL_D(SELF%S_SPLINE,T)
    END FUNCTION FILAMENT_EVAL_HELICAL_RISE

    !>  \brief  Compute the path length from first waypoint to the point specified by the coordinate T.
    !!
    !!  Should only be called when T falls directly after a waypoint for which the path_length has already been determined
#if defined (SKIP_RUNTIME_CHECKS)
    PURE &
#endif
    SUBROUTINE FILAMENT_EVAL_PATH_LENGTH(SELF,T,PATH_LENGTH)
        USE SPLINES, ONLY : INTERPOL_SPLINE_FIND_SEG, INTERPOL_SPLINE_PRINT_INFO, INTERPOL_SPLINE_EVAL
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),         INTENT(INOUT)   ::  SELF            !<  Filament
        REAL(KIND=8),           INTENT(IN)      ::  T               !<  Parameter value for which the path length needs to be computed
        REAL(KIND=8),           INTENT(OUT)     ::  PATH_LENGTH     !<  Result. Arc length in pixels.
        ! Private variables
        REAL(KIND=8),           PARAMETER       ::  TINY_STEP = 0.0005D0    !<  Step for the integration (pixels)
        INTEGER                                 ::  SEG
        REAL(KIND=8)                            ::  CURRENT_T,CURRENT_X,CURRENT_Y,CURRENT_Z,PREVIOUS_X,PREVIOUS_Y,PREVIOUS_Z
        LOGICAL,                PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work
        ! Figure out which segment of the spline we'll be working on
        SEG = INTERPOL_SPLINE_FIND_SEG(SELF%X_SPLINE,T)
        IF (DEBUG) THEN
#ifndef SKIP_RUNTIME_CHECKS
        IF (SEG .EQ. 0) THEN
            WRITE(*,'(A,F0.5)') '**ERROR(FILAMENT_EVAL_PATH_LENGTH): could not find spline segment for T = ', T
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%X_SPLINE)
            CALL TERMINATE('Fatal error in FILAMENT_EVAL_PATH_LENGTH')
        ENDIF
#endif
        ENDIF
        ! Start at the start of that segment
        PATH_LENGTH = SELF%WP(SEG)%PATH_LENGTH
        CURRENT_T = SELF%T(SEG)
        CURRENT_X = SELF%X_SPLINE%Y(SEG)
        CURRENT_Y = SELF%Y_SPLINE%Y(SEG)
        CURRENT_Z = SELF%Z_SPLINE%Y(SEG)

        !IF (DEBUG) THEN
        !   ! Sanity check
        !   IF (SEG .GT. 1 .AND. SELF%WP(SEG)%PATH_LENGTH .EQ. 0.0) THEN
        !       WRITE(*,'(A,I0,A)') '**ERROR(FILAMENT_EVAL_PATH_LENGTH): the path length of the nearest previous waypoint (', SEG, ') was 0.0, but this should only be true of the first waypoint of a filament!'
        !       CALL TERMINATE('Fatal error in FILAMENT_EVAL_PATH_LENGTH')
        !   ENDIF
        !ENDIF

        ! Loop until we've reached the point we're interested in
        DO
            ! Keep track of where we've just been
            PREVIOUS_X = CURRENT_X
            PREVIOUS_Y = CURRENT_Y
            PREVIOUS_Z = CURRENT_Z
            ! One tiny step for man...
            CURRENT_T = CURRENT_T + TINY_STEP
            ! Check we've not stepped over the mark
            IF (CURRENT_T .GT. T) THEN
                EXIT    !   Done, get out
            ENDIF
            ! We're still walking - work out where our foot lands
            CALL INTERPOL_SPLINE_EVAL(SELF%X_SPLINE,CURRENT_T,CURRENT_X)
            CALL INTERPOL_SPLINE_EVAL(SELF%Y_SPLINE,CURRENT_T,CURRENT_Y)
            CALL INTERPOL_SPLINE_EVAL(SELF%Z_SPLINE,CURRENT_T,CURRENT_Z)
            ! Measure & accumulate the distance
            PATH_LENGTH = PATH_LENGTH + SQRT((CURRENT_X-PREVIOUS_X)**2+(CURRENT_Y-PREVIOUS_Y)**2+(CURRENT_Z-PREVIOUS_Z)**2)
        ENDDO
    END SUBROUTINE FILAMENT_EVAL_PATH_LENGTH

    !>  \brief  Return the nearest waypoint for the input arc length
    PURE SUBROUTINE NEAREST_WP_FROM_T(SELF,T,NEAREST_WP,DIST)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(IN)      ::  SELF
        REAL(KIND=8),                       INTENT(IN)      ::  T
        INTEGER,                            INTENT(OUT)     ::  NEAREST_WP  !<  Index of nearest waypoint
        REAL(KIND=8),                       INTENT(OUT)     ::  DIST        !<  (Signed) distance to nearest waypoint
        ! Private variables
        INTEGER ::  ITMP(1)
        REAL(KIND=8) :: DISTANCES(SELF%NWP)
        ! Start work
        DISTANCES = T - SELF%T
        ITMP = MINLOC(ABS(DISTANCES))
        NEAREST_WP = ITMP(1)
        DIST = T - SELF%T(NEAREST_WP)
    END SUBROUTINE NEAREST_WP_FROM_T


    !>  \brief  Return the nearest previous waypoint for the input path length
    INTEGER FUNCTION NEAREST_PREVIOUS_WP_FROM_T(SELF,T) RESULT(PREVIOUS_WP)
        USE SPLINES, ONLY : INTERPOL_SPLINE_FIND_SEG, INTERPOL_SPLINE_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(IN)      ::  SELF
        REAL(KIND=8),                       INTENT(IN)      ::  T
        ! Private variables

        ! Start work
        PREVIOUS_WP = INTERPOL_SPLINE_FIND_SEG(SELF%X_SPLINE,T)
        IF(PREVIOUS_WP .EQ. 0) THEN
            WRITE(*,'(A,F0.3,A)') '**WARNING(NEAREST_PREVIOUS_WP_FROM_T): could not find segment within which input T value (', &
                                    T, ') falls. Maybe the splines are not set up properly?'
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%X_SPLINE)
            !CALL TERMINATE('Fatal error in NEAREST_PREVIOUS_WP_FROM_T')
        ENDIF
    END FUNCTION NEAREST_PREVIOUS_WP_FROM_T




    !>  \brief  Returns the straight-line distance in pixels from waypoint WP_NUM to the nearest waypoint
    REAL FUNCTION FILAMENT_WP_DISTANCE_TO_NEAREST_WP(FIL,WP_NUM)
        USE COOS, ONLY : WAYPOINT_WAYPOINT_DISTANCE
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)  ::  FIL     !<  Filament
        INTEGER,            INTENT(IN)  ::  WP_NUM  !<  The number of the waypoint we're interested in
        ! Private variables
        INTEGER ::  WP_BEFORE, WP_AFTER, I
        REAL    ::  TMP_DIST
        REAL,   PARAMETER   ::  LARGE_VALUE = 999999.99
        ! Start work

        ! Define the two neighbouring waypoints
        IF (WP_NUM .EQ. 1) THEN
            WP_BEFORE = 2
            WP_AFTER = 2
        ELSE IF (WP_NUM .EQ. FIL%NWP) THEN
            WP_BEFORE = FIL%NWP - 1
            WP_AFTER = FIL%NWP - 1
        ELSE IF (WP_NUM .GT. 1 .AND. WP_NUM .LT. FIL%NWP) THEN
            WP_BEFORE = WP_NUM-1
            WP_AFTER = WP_NUM+1
        ELSE
            WRITE(*,'(3(A,I3),A)') '**ERROR(FILAMENT_WP_DISTANCE_TO_NEAREST_WP): waypoint ', WP_NUM, &
                                    ' does not exist (filament ', FIL%ID, ' has ', FIL%NWP, ' waypoints).'
            CALL TERMINATE('Fatal error in FILAMENT_WP_DISTANCE_TO_NEAREST_WP')
        ENDIF

        ! Loop over the neighbours to find the nearest one
        FILAMENT_WP_DISTANCE_TO_NEAREST_WP = LARGE_VALUE
        DO I = WP_BEFORE,WP_AFTER,2
            TMP_DIST = WAYPOINT_WAYPOINT_DISTANCE(FIL%WP(WP_NUM),FIL%WP(I))
            IF (TMP_DIST .LT. FILAMENT_WP_DISTANCE_TO_NEAREST_WP) THEN
                FILAMENT_WP_DISTANCE_TO_NEAREST_WP = TMP_DIST
            ENDIF
        ENDDO

        ! If we didn't find a value, crash out
        IF (FILAMENT_WP_DISTANCE_TO_NEAREST_WP .EQ. LARGE_VALUE) THEN
            WRITE(*,'(A,F12.2)') '**ERROR(FILAMENT_WO_DISTANCE_TO_NEAREST_WP): did not find a distance-to-neighbour less than ', &
                                    LARGE_VALUE
            CALL TERMINATE('Fatal error in FILAMENT_WP_DISTANCE_TO_NEAREST_WP')
        ENDIF
    END FUNCTION FILAMENT_WP_DISTANCE_TO_NEAREST_WP

    !>  \brief  Compute the average defocus of the filament
    PURE REAL FUNCTION FILAMENT_AVERAGE_DEFOCUS(SELF,IGNORE_FAL_WPS) RESULT(AVERAGE_DEFOCUS)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(IN)  ::  SELF            !<  Filament object
        LOGICAL,       OPTIONAL,    INTENT(IN)  ::  IGNORE_FAL_WPS  !<  Whether to ignore the first and last waypoints in the calculation. False by default.
        ! Private variables
        LOGICAL ::  IIGNORE_FAL_WPS
        INTEGER ::  FIRST_WP, LAST_WP, CURRENT_WP
        ! Start work

        ! Ignore FAL WPS?
        IF (PRESENT(IGNORE_FAL_WPS)) THEN
            IIGNORE_FAL_WPS = IGNORE_FAL_WPS
        ELSE
            IIGNORE_FAL_WPS = .FALSE.
        ENDIF

        ! First and last waypoints
        IF (IIGNORE_FAL_WPS) THEN
            FIRST_WP    =   2
            LAST_WP     =   SELF%NWP - 1
        ELSE
            FIRST_WP    =   1
            LAST_WP     =   SELF%NWP
        ENDIF

        AVERAGE_DEFOCUS = 0.0
        DO CURRENT_WP=FIRST_WP,LAST_WP
            AVERAGE_DEFOCUS = AVERAGE_DEFOCUS + SELF%WP(CURRENT_WP)%COO(3)
        ENDDO

        AVERAGE_DEFOCUS = AVERAGE_DEFOCUS / REAL(LAST_WP-FIRST_WP+1)

    END FUNCTION FILAMENT_AVERAGE_DEFOCUS

    !>  \brief  Return the difference between the min and max Z (defocus) of waypoints in the filament
    PURE REAL FUNCTION FILAMENT_DEFOCUS_RANGE(SELF,IGNORE_FAL_WPS) RESULT(DEFOCUS_RANGE)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(IN)  ::  SELF            !<  Filament object
        LOGICAL,       OPTIONAL,    INTENT(IN)  ::  IGNORE_FAL_WPS  !<  Whether to ignore the first and last waypoints in the calculation. False by default.
        ! Private variables
        REAL    ::  DEFOCUS_MIN, DEFOCUS_MAX
        LOGICAL ::  IIGNORE_FAL_WPS
        INTEGER ::  FIRST_WP, LAST_WP, CURRENT_WP
        ! Start work

        ! Ignore FAL WPS?
        IF (PRESENT(IGNORE_FAL_WPS)) THEN
            IIGNORE_FAL_WPS = IGNORE_FAL_WPS
        ELSE
            IIGNORE_FAL_WPS = .FALSE.
        ENDIF

        ! First and last waypoints
        IF (IIGNORE_FAL_WPS) THEN
            FIRST_WP    =   2
            LAST_WP     =   SELF%NWP - 1
        ELSE
            FIRST_WP    =   1
            LAST_WP     =   SELF%NWP
        ENDIF

        !
        DEFOCUS_MIN =  9999999.99
        DEFOCUS_MAX = -9999999.99
        DO CURRENT_WP=FIRST_WP,LAST_WP
            IF (SELF%WP(CURRENT_WP)%COO(3) .GT. DEFOCUS_MAX) DEFOCUS_MAX = SELF%WP(CURRENT_WP)%COO(3)
            IF (SELF%WP(CURRENT_WP)%COO(3) .LT. DEFOCUS_MIN) DEFOCUS_MIN = SELF%WP(CURRENT_WP)%COO(3)
        ENDDO
        DEFOCUS_RANGE = ABS(DEFOCUS_MAX-DEFOCUS_MIN)
    END FUNCTION FILAMENT_DEFOCUS_RANGE

    !>  \brief  Return the average waypoint-to-waypoint distance (along the path of the filament)
    FUNCTION FILAMENT_AVERAGE_WP_TO_WP_DISTANCE(SELF,IGNORE_FAL_WPS) RESULT(DIST_AVE)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(IN)  ::  SELF            !<  Filament object
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  IGNORE_FAL_WPS  !<  Whether to ignore the first and last waypoints in the calculation. False by default.
        ! Result
        REAL    ::  DIST_AVE    !<  Average distance from one waypoint to the next (pixels)
        ! Private variables
        LOGICAL ::  IIGNORE_FAL_WPS
        INTEGER ::  FIRST_WP, LAST_WP
        ! Start work

        ! Ignore FAL WPS?
        IF (PRESENT(IGNORE_FAL_WPS)) THEN
            IIGNORE_FAL_WPS = IGNORE_FAL_WPS
        ELSE
            IIGNORE_FAL_WPS = .FALSE.
        ENDIF

        ! First and last waypoints
        IF (IIGNORE_FAL_WPS) THEN
            FIRST_WP    =   2
            LAST_WP     =   SELF%NWP - 1
        ELSE
            FIRST_WP    =   1
            LAST_WP     =   SELF%NWP
        ENDIF

        ! Check there are enough waypoints
        IF (LAST_WP-FIRST_WP+1 .LT. 2) THEN
            WRITE(*,'(A,I0,A)') '**ERROR(FILAMENT_AVERAGE_WP_TO_WP_DISTANCE): filament only has ', &
                                SELF%NWP, ' waypoints. Not enough'
            WRITE(*,'(A,L1)')   '**ERROR(FILAMENT_AVERAGE_WP_TO_WP_DISTANCE): ignore first and last? ',  IIGNORE_FAL_WPS
            CALL TERMINATE('Fatal error in FILAMENT_AVERAGE_WP_TO_WP_DISTANCE')
        ENDIF

        ! Compute the average distance
        DIST_AVE = SELF%WP(LAST_WP)%PATH_LENGTH - SELF%WP(FIRST_WP)%PATH_LENGTH
        DIST_AVE = DIST_AVE / (LAST_WP-FIRST_WP)
    END FUNCTION FILAMENT_AVERAGE_WP_TO_WP_DISTANCE

        !>  \brief  Given a filament with Theta and Psi Euler angles set for each waypoint, compute the spline derivatives at each point
    PURE SUBROUTINE FILAMENT_DERIVATIVES_FROM_EULER(SELF)
        USE DIFFERENTIAL_GEOMETRY, ONLY : DX_FROM_THETA_PSI, DY_FROM_THETA_PSI, DZ_FROM_THETA
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF    !<  Filament object
        ! Private variables
        INTEGER ::  I
        ! Start work
        FORALL(I=1:SELF%NWP)
            SELF%X_SPLINE%DY(I) = DX_FROM_THETA_PSI(SELF%WP(I)%EULER%THETA,SELF%WP(I)%EULER%PSI)
            SELF%Y_SPLINE%DY(I) = DY_FROM_THETA_PSI(SELF%WP(I)%EULER%THETA,SELF%WP(I)%EULER%PSI)
            SELF%Z_SPLINE%DY(I) = DZ_FROM_THETA(SELF%WP(I)%EULER%THETA)
        END FORALL
    END SUBROUTINE FILAMENT_DERIVATIVES_FROM_EULER



    !>  \brief  Given a filament with solved X,Y and Z splines, compute the Theta and Psi Euler angles at each waypoint
    !!
    !!  See the "Spline Parametrisation" document for derivations of the equations
    SUBROUTINE FILAMENT_EULER_FROM_DERIVATIVES(SELF,KNOWN_PSI,KNOWN_THETA)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE DIFFERENTIAL_GEOMETRY, ONLY : EULER_FROM_DERIVATIVES
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),         INTENT(INOUT)   ::  SELF        !<  Filament object
        LOGICAL,    OPTIONAL,   INTENT(IN)      ::  KNOWN_PSI   !<  If true, we won't recalculate psi
        LOGICAL,    OPTIONAL,   INTENT(IN)      ::  KNOWN_THETA !<  If true, we won't recalculate theta
        ! Private variables
        LOGICAL ::  KKNOWN_PSI, KKNOWN_THETA
        INTEGER ::  I
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        ! Defaults are false
        IF (PRESENT(KNOWN_THETA)) THEN
            KKNOWN_THETA = KNOWN_THETA
        ELSE
            KKNOWN_THETA = .FALSE.
        ENDIF
        IF (PRESENT(KNOWN_PSI)) THEN
            KKNOWN_PSI = KNOWN_PSI
        ELSE
            KKNOWN_PSI = .FALSE.
        ENDIF

        ! If we already know the angles, why are we even calling this subroutine?
        IF (KKNOWN_PSI .AND. KKNOWN_THETA) THEN
            WRITE(*,'(2A)') '**ERROR(FILAMENT_EULER_FROM_DERIVATIVES): psi and theta are already known, ',   &
                            'no need to call this subroutine'
            CALL TERMINATE('Programming error')
        ENDIF

        IF (KKNOWN_PSI .OR. KKNOWN_THETA) THEN
            CALL TERMINATE('Not coded yet - is this needed?')
        ENDIF

        !
        ! Loop over waypoints
        !
        DO I=1,SELF%NWP
            CALL EULER_FROM_DERIVATIVES(    SELF%X_SPLINE%DY(I),        &
                                            SELF%Y_SPLINE%DY(I),        &
                                            SELF%Z_SPLINE%DY(I),        &
                                            SELF%WP(I)%EULER%THETA,         &
                                            SELF%WP(I)%EULER%PSI    )
        ENDDO

    END SUBROUTINE FILAMENT_EULER_FROM_DERIVATIVES



    !>  \brief  Ensure that psi and theta are as smooth as possible, even if it
    !!          means breaking the convention of -90<=psi<=90
    PURE SUBROUTINE FILAMENT_PSI_THETA_SMOOTH(SELF)
        USE EULERANGLES, ONLY : DEGREES_M180_TO_180
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)      ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_WP
        LOGICAL ::  KEEP_THETA_POSITIVE, THETA_POSITIVE
        ! Start work

        ! We will keep the sign of theta constant
        KEEP_THETA_POSITIVE = (SELF%WP(1)%EULER%THETA .GT. 0.0)

        DO CURRENT_WP=1,SELF%NWP
            THETA_POSITIVE = (SELF%WP(CURRENT_WP)%EULER%THETA .GT. 0.0)
            IF (    (THETA_POSITIVE .AND. .NOT. KEEP_THETA_POSITIVE) &
                .OR.(.NOT. THETA_POSITIVE .AND. KEEP_THETA_POSITIVE) ) THEN
                ! We need to switch the sign of Theta
                SELF%WP(CURRENT_WP)%EULER%THETA = - SELF%WP(CURRENT_WP)%EULER%THETA
                SELF%WP(CURRENT_WP)%EULER%PSI   = DEGREES_M180_TO_180(SELF%WP(CURRENT_WP)%EULER%PSI + 180.)
            ENDIF
        ENDDO
    END SUBROUTINE FILAMENT_PSI_THETA_SMOOTH

    !>  \brief  Ensure that psi and theta are as smooth as possible, even if it
    !!          means breaking the convention of -90<=psi<=90
    PURE SUBROUTINE WAYPOINTS_PSI_THETA_SMOOTH(WPS)
        USE EULERANGLES, ONLY : DEGREES_M180_TO_180
        USE COOS, ONLY : WAYPOINT
        IMPLICIT NONE
        ! Arguments
        TYPE(WAYPOINT), INTENT(INOUT)      ::  WPS(:)
        ! Private variables
        INTEGER ::  CURRENT_WP
        LOGICAL ::  KEEP_THETA_POSITIVE, THETA_POSITIVE
        ! Start work

        ! We will keep the sign of theta constant
        KEEP_THETA_POSITIVE = (WPS(1)%EULER%THETA .GT. 0.0)

        DO CURRENT_WP=1,SIZE(WPS)
            THETA_POSITIVE = (WPS(CURRENT_WP)%EULER%THETA .GT. 0.0)
            IF (    (THETA_POSITIVE .AND. .NOT. KEEP_THETA_POSITIVE) &
                .OR.(.NOT. THETA_POSITIVE .AND. KEEP_THETA_POSITIVE) ) THEN
                ! We need to switch the sign of Theta
                WPS(CURRENT_WP)%EULER%THETA = - WPS(CURRENT_WP)%EULER%THETA
                WPS(CURRENT_WP)%EULER%PSI   = DEGREES_M180_TO_180(WPS(CURRENT_WP)%EULER%PSI + 180.)
            ENDIF
        ENDDO
    END SUBROUTINE WAYPOINTS_PSI_THETA_SMOOTH

    !>  \brief  Enforce convention of -90<=psi<=90 on all waypoints of the filament
    PURE SUBROUTINE FILAMENT_PSI_BTW_M90_90(SELF)
        USE DIFFERENTIAL_GEOMETRY, ONLY : PSI_BTW_M90_90
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_WP
        ! Start work
        DO CURRENT_WP = 1, SELF%NWP
            CALL PSI_BTW_M90_90(SELF%WP(CURRENT_WP)%EULER%THETA,SELF%WP(CURRENT_WP)%EULER%PSI)
        ENDDO
    END SUBROUTINE FILAMENT_PSI_BTW_M90_90





    !>  \brief  Solve the filament axis' splines, based on the current values for T
    SUBROUTINE FILAMENT_SOLVE_AXIS_SPLINES(SELF)
        USE FILAMENTS_BASE, ONLY : PSI_THETA_NON_ZERO
        USE SPLINES
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
        LOGICAL ::  DERIVATIVES_FROM_EULER
        ! Start work

        ! Initialise the "X" values for the splines
        SELF%X_SPLINE%X     = SELF%T
        SELF%Y_SPLINE%X     = SELF%T
        SELF%Z_SPLINE%X     = SELF%T

        ! Initialise the "Y" values for the splines
        SELF%X_SPLINE%Y     = SELF%WP%COO(1)
        SELF%Y_SPLINE%Y     = SELF%WP%COO(2)
        SELF%Z_SPLINE%Y     = SELF%WP%COO(3)

        ! If we know the Euler angles at waypoints, they can force the (x,y,z) splines' derivatives.
        DERIVATIVES_FROM_EULER = SELF%USE_HERMITE_SPLINES .AND. PSI_THETA_NON_ZERO(SELF)
        IF (DERIVATIVES_FROM_EULER) THEN
            CALL FILAMENT_DERIVATIVES_FROM_EULER(SELF)
            CALL SPLINE_SET_TYPE(SELF%X_SPLINE,'KNOWN_D')
            CALL SPLINE_SET_TYPE(SELF%Y_SPLINE,'KNOWN_D')
            CALL SPLINE_SET_TYPE(SELF%Z_SPLINE,'KNOWN_D')
        ELSE
            CALL SPLINE_SET_TYPE(SELF%X_SPLINE,'PARABOLIC')
            CALL SPLINE_SET_TYPE(SELF%Y_SPLINE,'PARABOLIC')
            CALL SPLINE_SET_TYPE(SELF%Z_SPLINE,'PARABOLIC')
        ENDIF

        ! Now we can actually solve the splines
        CALL SOLVE_INTERPOL_SPLINE(SELF%X_SPLINE)
        CALL SOLVE_INTERPOL_SPLINE(SELF%Y_SPLINE)
        CALL SOLVE_INTERPOL_SPLINE(SELF%Z_SPLINE)

        ! Update the Euler angles if needed
        IF (.NOT. DERIVATIVES_FROM_EULER) CALL FILAMENT_EULER_FROM_DERIVATIVES(SELF)

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_SOLVE_SPLINES): X spline on exit'
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%X_SPLINE)
        ENDIF
    END SUBROUTINE FILAMENT_SOLVE_AXIS_SPLINES

    !>  \brief  Solve the filament's helix's splines, based on the current values for T
    SUBROUTINE FILAMENT_SOLVE_HELIX_SPLINES(SELF)
        USE SPLINES
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF
        ! Private variables

        ! Start work


        ! Initialise the "X" values for the splines
        SELF%S_SPLINE%X     = SELF%T
        SELF%PHI_SPLINE%X   = SELF%T

        ! Initialise the "Y" values for the splines
        SELF%S_SPLINE%Y     = SELF%WP%SUBUNIT_NUM
        SELF%PHI_SPLINE%Y   = SELF%WP%EULER%PHI

        ! The phi & s splines are always natural, since we don't have a good way to measure local helical parameters
        ! Until 20130909, parabolic splines were used here
        CALL SPLINE_SET_TYPE(SELF%PHI_SPLINE,'NATURAL')
        CALL SPLINE_SET_TYPE(SELF%S_SPLINE,'NATURAL')

        ! Now we can actually solve the splines
        CALL SOLVE_INTERPOL_SPLINE(SELF%S_SPLINE)
        CALL SOLVE_INTERPOL_SPLINE(SELF%PHI_SPLINE)

    END SUBROUTINE FILAMENT_SOLVE_HELIX_SPLINES


    !>  \brief  Solve the filament's splines, based on the current values for T
    SUBROUTINE FILAMENT_SOLVE_SPLINES(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF
        ! Private variables

        ! Start work

        ! Update the helical axis splines
        CALL FILAMENT_SOLVE_AXIS_SPLINES(SELF)

        ! Update the helical splines
        CALL FILAMENT_SOLVE_HELIX_SPLINES(SELF)

    END SUBROUTINE FILAMENT_SOLVE_SPLINES

    !>  \brief Check whether the phi & subunit_num spline segments are monotonic
    LOGICAL FUNCTION FILAMENT_SPLINES_ARE_MONOTONIC(SELF)
        USE SPLINES, ONLY : INTERPOL_SPLINE_IS_MONOTONIC
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        ! Private variable

        ! Start work
        FILAMENT_SPLINES_ARE_MONOTONIC =        INTERPOL_SPLINE_IS_MONOTONIC(SELF%PHI_SPLINE)   &
                                        .AND.   INTERPOL_SPLINE_IS_MONOTONIC(SELF%S_SPLINE)
    END FUNCTION FILAMENT_SPLINES_ARE_MONOTONIC

    !>  \brief  Check whether waypoints' path lengths increase along the filament
    PURE LOGICAL FUNCTION FILAMENT_WP_PATH_LENGTHS_INCREASE(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_WP
        ! Start work
        FILAMENT_WP_PATH_LENGTHS_INCREASE = .TRUE.
        DO CURRENT_WP=2,SELF%NWP
            IF (SELF%WP(CURRENT_WP)%PATH_LENGTH .LE. SELF%WP(CURRENT_WP-1)%PATH_LENGTH) FILAMENT_WP_PATH_LENGTHS_INCREASE = .FALSE.
        ENDDO
    END FUNCTION FILAMENT_WP_PATH_LENGTHS_INCREASE

    !>  \brief  Evaluate deviations from canonical helical parameters
    SUBROUTINE FILAMENT_DEVIATIONS_FROM_CANONICAL_HELIX(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)  ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_WP
        REAL    ::  DIST_TO_NEXT_WP
        REAL    ::  EXPECTED_CHANGE_IN_PHI, ACTUAL_CHANGE_IN_PHI
        REAL    ::  EXPECTED_CHANGE_IN_SUBUNIT_NUM, ACTUAL_CHANGE_IN_SUBUNIT_NUM
        ! Start work

        ! Measure the total curvature - high curvature probably means something is wrong (probably in the Z direction)
        WRITE(*,'(A,G15.4)') '**info(FILAMENT_DEVIATIONS_FROM_CANONICAL_HELIX): total curvature = ', &
                             FILAMENT_EVAL_TOTAL_CURVATURE(SELF)

        !
        ! Since we have a path length estimate for each waypoint, we can see whether the phi angles agree with each other
        ! (assuming constant canonical helical parameters)
        !
        IF (SELF%HELIX%TWIST .NE. 0.) THEN
            DO CURRENT_WP=1,SELF%NWP-1
                ! Work out the distance to the next waypoint
                DIST_TO_NEXT_WP = SELF%WP(CURRENT_WP+1)%PATH_LENGTH - SELF%WP(CURRENT_WP)%PATH_LENGTH
                ! Work out what change in phi this should correspond to
                EXPECTED_CHANGE_IN_PHI = DIST_TO_NEXT_WP * SELF%HELIX%TWIST / SELF%HELIX%RISE
                ! What change in phi do we actually have?
                ACTUAL_CHANGE_IN_PHI = SELF%WP(CURRENT_WP+1)%EULER%PHI - SELF%WP(CURRENT_WP)%EULER%PHI
                ! Expected change in subunit_num
                EXPECTED_CHANGE_IN_SUBUNIT_NUM = DIST_TO_NEXT_WP / SELF%HELIX%RISE
                ! Actual change in subunit_num
                ACTUAL_CHANGE_IN_SUBUNIT_NUM = SELF%WP(CURRENT_WP+1)%SUBUNIT_NUM - SELF%WP(CURRENT_WP)%SUBUNIT_NUM
                ! How far off are we?
                WRITE(*,'(A,I0,A,I0,A,F0.2,A)') &
                                '**info(FILAMENT_DEVIATIONS_FROM_CANONICAL_HELIX): subunit_num change between waypoints ', &
                                CURRENT_WP, ' and ', CURRENT_WP+1, &
                                ' is ', ACTUAL_CHANGE_IN_SUBUNIT_NUM - EXPECTED_CHANGE_IN_SUBUNIT_NUM, &
                                ' subunits off compared to expected value'
                WRITE(*,'(2(A,I0),2(A,F0.4),A)',ADVANCE='NO')    &
                                            '**info(FILAMENT_DEVIATIONS_FROM_CANONICAL_HELIX): phi change between waypoints ', &
                                            CURRENT_WP, ' and ', CURRENT_WP+1, &
                                            ' is ', ACTUAL_CHANGE_IN_PHI/EXPECTED_CHANGE_IN_PHI, &  ! Watch out - division by zero if helix%twist = 0.0
                                            ' times the expected change, corresponding to a difference of ', &
                                            (ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI), ' degrees'
                IF (SELF%HELIX%TWIST .NE. 0.0) THEN
                    WRITE(*,'(A,F0.4,A)') ' or ', (ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI)/SELF%HELIX%TWIST, ' subunits'
                ELSE
                    WRITE(*,'(A)') ' '
                ENDIF
            ENDDO
        ELSE
            ! Helical twist is 0.0, therefore we can't really check phi angles
        ENDIF

        !
        ! When a filament was refined in SP mode against a reference with canonical helical symmetry, we can infer the subunit number
        ! of each waypoint from its phi angle. We can then see if the s_spline looks linear. If it doesn't then there are inconsistencies.
        !
    END SUBROUTINE FILAMENT_DEVIATIONS_FROM_CANONICAL_HELIX

    !>  \brief
    REAL FUNCTION FILAMENT_DISCREPENCY_BTW_WPS_PHI_AND_CANONICAL_HELIX(SELF) RESULT(TOTAL_ERROR)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)  ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_WP
        ! Start work
        TOTAL_ERROR = 0.0
        DO CURRENT_WP=1,SELF%NWP-1
            TOTAL_ERROR = TOTAL_ERROR + FILAMENT_WP_ERROR_IN_CHANGE_IN_PHI(SELF,CURRENT_WP)
        ENDDO
    END FUNCTION FILAMENT_DISCREPENCY_BTW_WPS_PHI_AND_CANONICAL_HELIX

    !>  \brief  Work out how far off the following waypoint's phi is from what is predicted by the current waypoint's phi and the arc length distance to the next wp
    REAL FUNCTION FILAMENT_WP_ERROR_IN_CHANGE_IN_PHI(SELF,WP_NUM) RESULT(ERROR_IN_CHANGE_IN_PHI)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)  ::  SELF
        INTEGER,            INTENT(IN)  ::  WP_NUM      !<  The error will be computed between this waypoint and the next
        ! Private variables
        REAL    ::  DIST_TO_NEXT_WP, EXPECTED_CHANGE_IN_PHI, ACTUAL_CHANGE_IN_PHI
        ! Start work
        IF (WP_NUM .GE. SELF%NWP .OR. WP_NUM .LT. 1) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_WP_ERROR_IN_CHANGE_IN_PHI): waypoint number ', WP_NUM, &
                            ' either does not exist or is the last waypoint in the filament'
            CALL TERMINATE('Fatal programming error')
        ENDIF
        ! Work out the distance to the next waypoint
        DIST_TO_NEXT_WP = SELF%WP(WP_NUM+1)%PATH_LENGTH - SELF%WP(WP_NUM)%PATH_LENGTH
        ! Work out what change in phi this should correspond to
        EXPECTED_CHANGE_IN_PHI = DIST_TO_NEXT_WP * SELF%HELIX%TWIST / SELF%HELIX%RISE
        ! What change in phi do we actually have?
        ACTUAL_CHANGE_IN_PHI = SELF%WP(WP_NUM+1)%EULER%PHI - SELF%WP(WP_NUM)%EULER%PHI
        ! How far off are we?
        ERROR_IN_CHANGE_IN_PHI = (ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI)
    END FUNCTION FILAMENT_WP_ERROR_IN_CHANGE_IN_PHI

    !>  \brief  Numerically tweak the filament model to get as close as possible to a relaxed helix
    SUBROUTINE FILAMENT_RELAX_HELIX(SELF,FORCE_INTG_SUBUNIT_NUM,IGNORE_SUBUNIT_PHASING,FORBID_UNDERTWIST,FORBID_STRETCHING)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF                    !<  Filament
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  FORCE_INTG_SUBUNIT_NUM  !<  Subunit_num values can be forced to integers
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  IGNORE_SUBUNIT_PHASING  !<  Do not preserve subunit_num phasing along the model
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  FORBID_UNDERTWIST       !<  Do not allow less phi change between waypoints than the canonical helix would give
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  FORBID_STRETCHING       !<  Do not allow less subunits between waypoints than the canonical helix would give
        ! Private variables

        ! Start work

        ! First, sort out the subunit_num values
        CALL FILAMENT_RELAX_RISE_PER_SUBUNIT(SELF,FORCE_INTG_SUBUNIT_NUM,IGNORE_SUBUNIT_PHASING,FORBID_STRETCHING)

        ! Now we can modify phi so that the twist per subunit is relaxed
        CALL FILAMENT_RELAX_TWIST_PER_SUBUNIT(SELF,IGNORE_SUBUNIT_PHASING,FORBID_UNDERTWIST)


    END SUBROUTINE FILAMENT_RELAX_HELIX

    !>  \brief  Numerically tweak subunit_num values to get the rise per subunit as close to canonical as possible
    SUBROUTINE FILAMENT_RELAX_RISE_PER_SUBUNIT(SELF,FORCE_INTG_SUBUNIT_NUM,IGNORE_SUBUNIT_NUM_PHASING,FORBID_STRETCHING,FAL_ONLY)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE SPLINES, ONLY : INTERPOL_SPLINE_IS_MONOTONIC
        USE FILAMENTS_BASE, ONLY : FILAMENT_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  FORCE_INTG_SUBUNIT_NUM      !<  Subunit_num values can be forced to integers
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  IGNORE_SUBUNIT_NUM_PHASING  !<  Don't preserve subunit_num phasing along the filament
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  FORBID_STRETCHING           !<  Do not allow less subunits between waypoints than the canonical helix would give
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  FAL_ONLY                    !<  First and last segments only will be relaxed
        ! Private variables
        LOGICAL,    PARAMETER   ::  FFORCE_INTG_SUBUNIT_NUM_DEFAULT = .FALSE.
        LOGICAL                 ::  FFORCE_INTG_SUBUNIT_NUM
        INTEGER                 ::  CURRENT_WP
        REAL(KIND=8)            ::  DIST_TO_NEXT_WP
        REAL(KIND=8)            ::  EXPECTED_CHANGE_IN_SUBUNIT_NUM, ACTUAL_CHANGE_IN_SUBUNIT_NUM
        REAL(KIND=8)            ::  SUBUNIT_NUM_TO_ADD
        LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
        LOGICAL ::  IIGNORE_SUBUNIT_NUM_PHASING
        LOGICAL ::  FFORBID_STRETCHING
        LOGICAL ::  FFAL_ONLY
        INTEGER ::  WP_INCR

        ! Start work
        IIGNORE_SUBUNIT_NUM_PHASING = .FALSE.
        IF (PRESENT(IGNORE_SUBUNIT_NUM_PHASING)) IIGNORE_SUBUNIT_NUM_PHASING = IGNORE_SUBUNIT_NUM_PHASING

        FFORCE_INTG_SUBUNIT_NUM = FFORCE_INTG_SUBUNIT_NUM_DEFAULT
        IF (PRESENT(FORCE_INTG_SUBUNIT_NUM)) FFORCE_INTG_SUBUNIT_NUM = FORCE_INTG_SUBUNIT_NUM

        FFORBID_STRETCHING = .FALSE.
        IF (PRESENT(FORBID_STRETCHING)) FFORBID_STRETCHING = FORBID_STRETCHING

        FFAL_ONLY = .FALSE.
        IF (PRESENT(FAL_ONLY)) FFAL_ONLY = FAL_ONLY

        WP_INCR = 1
        IF (FFAL_ONLY) WP_INCR = SELF%NWP-2

        IF (FFORBID_STRETCHING .AND. IIGNORE_SUBUNIT_NUM_PHASING) THEN
            WRITE(*,'(A)') '**WARNING(FILAMENT_RELAX_RISE_PER_SUBUNIT): non-sensical options'
            WRITE(*,'(2A)') '**WARNING(FILAMENT_RELAX_RISE_PER_SUBUNIT): with IGNORE_SUBUNIT_NUM_PHASING, there can be no', &
                            ' stretching or compression, so the FORBID_STRETCHING flag is unnecessary or meaningless'
            CALL TERMINATE('Fatal error in FILAMENT_RELAX_RISE_PER_SUBUNIT. Programming error.') ! This is to catch programming error - the conditions isn't really fatal
        ENDIF

        !
        DO CURRENT_WP = 1, SELF%NWP-1, WP_INCR
            ! Distance to next wp
            DIST_TO_NEXT_WP = SELF%WP(CURRENT_WP+1)%PATH_LENGTH - SELF%WP(CURRENT_WP)%PATH_LENGTH
            ! Expected change in subunit_num
            EXPECTED_CHANGE_IN_SUBUNIT_NUM = DIST_TO_NEXT_WP / SELF%HELIX%RISE
            ! Actual change in subunit_num
            ACTUAL_CHANGE_IN_SUBUNIT_NUM = SELF%WP(CURRENT_WP+1)%SUBUNIT_NUM - SELF%WP(CURRENT_WP)%SUBUNIT_NUM
            !
            IF (DEBUG) THEN
                WRITE(*,'(2(A,I0),A,2(F0.3,1X))')    '**debug(FILAMENT_RELAX_RISE_PER_SUBUNIT): between waypoints ', &
                                                    CURRENT_WP, ' and ', CURRENT_WP+1, &
                                                    ' expected, actual change in subunit_num = ', &
                                                    EXPECTED_CHANGE_IN_SUBUNIT_NUM, ACTUAL_CHANGE_IN_SUBUNIT_NUM
            ENDIF
            !
            IF (IIGNORE_SUBUNIT_NUM_PHASING) THEN
                SUBUNIT_NUM_TO_ADD =      EXPECTED_CHANGE_IN_SUBUNIT_NUM - ACTUAL_CHANGE_IN_SUBUNIT_NUM
            ELSE
                IF (FFORBID_STRETCHING) THEN
                    SUBUNIT_NUM_TO_ADD = CEILING(EXPECTED_CHANGE_IN_SUBUNIT_NUM - ACTUAL_CHANGE_IN_SUBUNIT_NUM)
                ELSE
                    SUBUNIT_NUM_TO_ADD = NINT(EXPECTED_CHANGE_IN_SUBUNIT_NUM - ACTUAL_CHANGE_IN_SUBUNIT_NUM)
                ENDIF
            ENDIF
            IF (CURRENT_WP .GT. SELF%NWP/2) THEN
                ! Debug
                IF (DEBUG .AND. SUBUNIT_NUM_TO_ADD .NE. 0.) THEN
                    WRITE(*,'(A,F0.5,A,I0,A,I0)') '**debug(FILAMENT_RELAX_RISE_PER_SUBUNIT): adding ', &
                                    SUBUNIT_NUM_TO_ADD, ' subunits to waypoints ', CURRENT_WP+1, ' to ', SELF%NWP
                ENDIF
                ! Correct values
                SELF%WP(CURRENT_WP+1:SELF%NWP)%SUBUNIT_NUM      =   SELF%WP(CURRENT_WP+1:SELF%NWP)%SUBUNIT_NUM + &
                                                                    SUBUNIT_NUM_TO_ADD
            ELSE
                ! Debug
                IF (DEBUG .AND. SUBUNIT_NUM_TO_ADD .NE. 0.) THEN
                    WRITE(*,'(A,F0.5,A,I0,A,I0)') '**debug(FILAMENT_RELAX_RISE_PER_SUBUNIT): subtracting ', &
                                    SUBUNIT_NUM_TO_ADD, ' subunits to waypoints ', 1, ' to ', CURRENT_WP
                ENDIF
                ! Correct values
                SELF%WP(1:CURRENT_WP)%SUBUNIT_NUM      =   SELF%WP(1:CURRENT_WP)%SUBUNIT_NUM - &
                                                                    SUBUNIT_NUM_TO_ADD
            ENDIF
        END DO
        IF (FFORCE_INTG_SUBUNIT_NUM) THEN
            SELF%WP%SUBUNIT_NUM = NINT(SELF%WP%SUBUNIT_NUM)
        ENDIF

        ! Update the relevant splines
        CALL FILAMENT_SOLVE_HELIX_SPLINES(SELF)

        ! The spline should be monotonic
        IF (DEBUG) THEN
        IF (.NOT. INTERPOL_SPLINE_IS_MONOTONIC(SELF%S_SPLINE)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_RELAX_RISE_PER_SUBUNIT): subunit_num spline is not monotonic'
            CALL FILAMENT_PRINT_INFO(SELF)
            CALL TERMINATE('Fatal error in FILAMENT_RELAX_RISE_PER_SUBUNIT')
        ENDIF
        ENDIF

    END SUBROUTINE FILAMENT_RELAX_RISE_PER_SUBUNIT

    !>  \brief  Add or subtract helical turns to get as close as possible to a canonical twist per subunit
    SUBROUTINE FILAMENT_RELAX_TWIST_PER_SUBUNIT(SELF,IGNORE_SUBUNIT_NUM_PHASING,FORBID_UNDERTWIST,FAL_ONLY,ALLOW_ANY_CHANGE)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_BASE, ONLY : FILAMENT_PRINT_INFO
        USE SPLINES, ONLY : INTERPOL_SPLINE_IS_MONOTONIC
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF                        !< Filament
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  IGNORE_SUBUNIT_NUM_PHASING  !<  Do not preserve phasing of subunits
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  FORBID_UNDERTWIST           !<  Do not allow less phi change between waypoints than the canonical helix would give
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  FAL_ONLY                    !<  First and last segments only will be relaxed
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  ALLOW_ANY_CHANGE            !<  Allow a change of any value in phi (by default only 360.0/axial_sym is allowed)
        ! Private variables
        INTEGER ::  CURRENT_WP
        REAL(KIND=8)    ::  SUBUNITS_TO_NEXT_WP
        REAL(KIND=8)    ::  EXPECTED_CHANGE_IN_PHI, ACTUAL_CHANGE_IN_PHI
        REAL(KIND=8)    ::  DPHI
        REAL(KIND=8)    ::  PHI_TO_ADD
        REAL(KIND=8)    ::  ALLOWED_CHANGE
        LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
        LOGICAL ::  IIGNORE_SUBUNIT_NUM_PHASING
        LOGICAL ::  FFORBID_UNDERTWIST
        LOGICAL ::  FFAL_ONLY
        INTEGER ::  WP_INCR
        LOGICAL ::  AALLOW_ANY_CHANGE
        LOGICAL ::  SAFE_TO_ADD
        ! Start work

        !IF (DEBUG) THEN
        !    WRITE(*,'(A)') '**debug(FILAMENT_RELAX_TWIST_PER_SUBUNIT): filament info on entry'
        !    CALL FILAMENT_PRINT_INFO(SELF)
        !ENDIF

        IIGNORE_SUBUNIT_NUM_PHASING = .FALSE.
        IF (PRESENT(IGNORE_SUBUNIT_NUM_PHASING)) IIGNORE_SUBUNIT_NUM_PHASING = IGNORE_SUBUNIT_NUM_PHASING

        FFORBID_UNDERTWIST = .FALSE.
        IF (PRESENT(FORBID_UNDERTWIST)) FFORBID_UNDERTWIST = FORBID_UNDERTWIST

        IF (FFORBID_UNDERTWIST .AND. IIGNORE_SUBUNIT_NUM_PHASING) THEN
            WRITE(*,'(A)') '**WARNING(FILAMENT_RELAX_TWIST_PER_SUBUNIT): non-sensical options'
            WRITE(*,'(2A)') '**WARNING(FILAMENT_RELAX_TWIST_PER_SUBUNIT): with IGNORE_SUBUNIT_NUM_PHASING, there can be no', &
                            ' stretching or compression, so the FFORBID_UNDERTWIST flag is unnecessary or meaningless'
            CALL TERMINATE('Fatal error in FILAMENT_RELAX_TWIST_PER_SUBUNIT. Programming error.') ! This is to catch programming error - the conditions isn't really fatal
        ENDIF

        FFAL_ONLY = .FALSE.
        IF (PRESENT(FAL_ONLY)) FFAL_ONLY = FAL_ONLY

        WP_INCR = 1
        IF (FFAL_ONLY) WP_INCR = SELF%NWP-2

        AALLOW_ANY_CHANGE = .FALSE.
        IF (PRESENT(ALLOW_ANY_CHANGE)) AALLOW_ANY_CHANGE = ALLOW_ANY_CHANGE

        ! Work out the allowed change
        IF (AALLOW_ANY_CHANGE) THEN
            ALLOWED_CHANGE = 0.001D0
        ELSE
            ALLOWED_CHANGE = 360.0D0 / SELF%HELIX%AXIAL_SYMMETRY
        ENDIF


        ! Loop over waypoints
        DO CURRENT_WP = 1, SELF%NWP-1, WP_INCR
            ! Number of subunits to the next waypoint
            SUBUNITS_TO_NEXT_WP             = SELF%WP(CURRENT_WP+1)%SUBUNIT_NUM - SELF%WP(CURRENT_WP)%SUBUNIT_NUM
            ! Expected change
            EXPECTED_CHANGE_IN_PHI          = SUBUNITS_TO_NEXT_WP * SELF%HELIX%TWIST
            ! Actual change
            ACTUAL_CHANGE_IN_PHI            = SELF%WP(CURRENT_WP+1)%EULER%PHI - SELF%WP(CURRENT_WP)%EULER%PHI
            !
            IF (DEBUG) THEN
                WRITE(*,'(2A,I0,A,I0,A,2(F0.2,1X))')    '**debug(FILAMENT_RELAX_TWIST_PER_SUBUNIT): expected, actual ',&
                                                        'change between waypoints ', &
                                                        CURRENT_WP, ' and ', CURRENT_WP+1, ': ', &
                                                        EXPECTED_CHANGE_IN_PHI, ACTUAL_CHANGE_IN_PHI
            ENDIF
            !
            DPHI = ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI
            IF (IIGNORE_SUBUNIT_NUM_PHASING) THEN
                PHI_TO_ADD = - DPHI
            ELSE
                IF (FFORBID_UNDERTWIST) THEN
                    PHI_TO_ADD = -ALLOWED_CHANGE*(CEILING(DPHI/ALLOWED_CHANGE))
                ELSE
                    PHI_TO_ADD = -ALLOWED_CHANGE*(NINT(DPHI/ALLOWED_CHANGE))
                ENDIF
            ENDIF



            ! Correct values
            IF (CURRENT_WP .GT. SELF%NWP/2) THEN
                SAFE_TO_ADD =   (SELF%HELIX%TWIST .GT. 0.0  .AND.                                                   &
                                (SELF%WP(CURRENT_WP+1)%EULER%PHI + PHI_TO_ADD .GT. SELF%WP(CURRENT_WP)%EULER%PHI))  &
                                .OR.                                                                                &
                                (SELF%HELIX%TWIST .LT. 0.0  .AND.                                                   &
                                (SELF%WP(CURRENT_WP+1)%EULER%PHI + PHI_TO_ADD .LT. SELF%WP(CURRENT_WP)%EULER%PHI))
                IF ( SAFE_TO_ADD ) THEN
                    ! Debug
                    IF (DEBUG .AND. PHI_TO_ADD .NE. 0.) THEN
                        WRITE(*,'(A,F0.2,A,I0,A,I0)') '**debug(FILAMENT_RELAX_TWIST_PER_SUBUNIT): adding ', &
                                        PHI_TO_ADD, ' twist degrees to waypoints ', CURRENT_WP+1, ' to ', SELF%NWP
                    ENDIF
                    SELF%WP(CURRENT_WP+1:SELF%NWP)%EULER%PHI    =   SELF%WP(CURRENT_WP+1:SELF%NWP)%EULER%PHI + &
                                                                    PHI_TO_ADD
                ENDIF
            ELSE
                SAFE_TO_ADD =   (SELF%HELIX%TWIST .GT. 0.0  .AND.                                                   &
                                (SELF%WP(CURRENT_WP)%EULER%PHI + PHI_TO_ADD .LT. SELF%WP(CURRENT_WP+1)%EULER%PHI))  &
                                .OR.                                                                                &
                                (SELF%HELIX%TWIST .LT. 0.0  .AND.                                                   &
                                (SELF%WP(CURRENT_WP)%EULER%PHI + PHI_TO_ADD .GT. SELF%WP(CURRENT_WP+1)%EULER%PHI))
                IF ( SAFE_TO_ADD ) THEN
                    ! Debug
                    IF (DEBUG .AND. PHI_TO_ADD .NE. 0.) THEN
                        WRITE(*,'(A,F0.2,A,I0,A,I0)') '**debug(FILAMENT_RELAX_TWIST_PER_SUBUNIT):  subtracting ', &
                                        PHI_TO_ADD, ' twist degrees to waypoints ', 1, ' to ', CURRENT_WP
                    ENDIF
                    SELF%WP(1:CURRENT_WP)%EULER%PHI             =   SELF%WP(1:CURRENT_WP)%EULER%PHI - &
                                                                    PHI_TO_ADD
                ENDIF
            ENDIF
        END DO

        ! Update the relevant splines
        CALL FILAMENT_SOLVE_HELIX_SPLINES(SELF)

        ! The spline should be monotonic
        IF (DEBUG) THEN
        IF (.NOT. FILAMENT_SPLINES_ARE_MONOTONIC(SELF)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_RELAX_TWIST_PER_SUBUNIT): phi & subunit_num splines are not both monotonic'
            CALL FILAMENT_PRINT_INFO(SELF)
            CALL TERMINATE('Fatal error in FILAMENT_RELAX_TWIST_PER_SUBUNIT')
        ENDIF
        ENDIF

    END SUBROUTINE FILAMENT_RELAX_TWIST_PER_SUBUNIT

    !>  \brief  Add or subtract helical turns to get closer to a relaxed helix
    SUBROUTINE FILAMENT_ADD_HELICAL_TURNS_TO_RELAX(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)  ::  SELF    !< Filament
        ! Private variables
        INTEGER ::  CURRENT_WP
        REAL    ::  DIST_TO_NEXT_WP
        REAL    ::  EXPECTED_CHANGE_IN_SUBUNIT_NUM, ACTUAL_CHANGE_IN_SUBUNIT_NUM
        REAL    ::  EXPECTED_CHANGE_IN_PHI, ACTUAL_CHANGE_IN_PHI
        REAL    ::  DPHI
        REAL    ::  SUBUNIT_NUM_TO_ADD, PHI_TO_ADD
        LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
        ! Start work

        ! Loop over waypoints
        DO CURRENT_WP = 1, SELF%NWP-1
            ! Distance to next wp
            DIST_TO_NEXT_WP = SELF%WP(CURRENT_WP+1)%PATH_LENGTH - SELF%WP(CURRENT_WP)%PATH_LENGTH
            ! Expected change
            EXPECTED_CHANGE_IN_SUBUNIT_NUM = DIST_TO_NEXT_WP / SELF%HELIX%RISE
            EXPECTED_CHANGE_IN_PHI         = EXPECTED_CHANGE_IN_SUBUNIT_NUM * SELF%HELIX%TWIST
            ! Actual change
            ACTUAL_CHANGE_IN_SUBUNIT_NUM = SELF%WP(CURRENT_WP+1)%SUBUNIT_NUM - SELF%WP(CURRENT_WP)%SUBUNIT_NUM
            ACTUAL_CHANGE_IN_PHI         = SELF%WP(CURRENT_WP+1)%EULER%PHI - SELF%WP(CURRENT_WP)%EULER%PHI
            !
            DPHI = ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI
            PHI_TO_ADD = 360.*(NINT(DPHI/360.))
            SUBUNIT_NUM_TO_ADD = NINT(PHI_TO_ADD / SELF%HELIX%TWIST)      ! I would have thought a NINT here would make sense
                                                                    ! but testing so far does not confirm this (29-feb-2012)

            ! Debug
            IF (DEBUG .AND. PHI_TO_ADD .NE. 0.) THEN
!                WRITE(*,'(A,F0.2,A,F0.1,A,I0,A,I0)') '**debug(FILAMENT_ADD_HELICAL_TURNS_TO_RELAX): adding ', &
!                                SUBUNIT_NUM_TO_ADD, ' subunits and ', &
!                                PHI_TO_ADD, ' twist degrees to waypoints ', CURRENT_WP+1, ' to ', SELF%NWP
                WRITE(*,'(A,F0.1,A,I0,A,I0)') '**debug(FILAMENT_ADD_HELICAL_TURNS_TO_RELAX): adding ', &
                                PHI_TO_ADD, ' twist degrees to waypoints ', CURRENT_WP+1, ' to ', SELF%NWP
            ENDIF

            ! Correct values
!            SELF%WP(CURRENT_WP+1:SELF%NWP)%SUBUNIT_NUM  =   SELF%WP(CURRENT_WP+1:SELF%NWP)%SUBUNIT_NUM + &
!                                                            SUBUNIT_NUM_TO_ADD
            SELF%WP(CURRENT_WP+1:SELF%NWP)%EULER%PHI    =   SELF%WP(CURRENT_WP+1:SELF%NWP)%EULER%PHI + &
                                                            PHI_TO_ADD
        END DO
    END SUBROUTINE FILAMENT_ADD_HELICAL_TURNS_TO_RELAX

    !>  \brief  Add or subtract integer multiples of 360 degrees to waypoints' phi values to minimise discrepencies with canonical helical symmetry
    !!
    !!  It is assumed the filament model is up-to-date (in particular the path lengths)
    SUBROUTINE FILAMENT_WPS_ADD_SUB_PHI_360(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(INOUT)   ::  SELF    !<  Filament object. Only WPS' phi values will be changed
        ! Private variables
        INTEGER ::  CURRENT_WP
        REAL    ::  ERROR_IN_CHANGE_IN_PHI
        ! Start work

        !
        ! Since we have a path length estimate for each waypoint, we can see whether the phi angles agree with each other
        ! (assuming constant canonical helical parameters). If there is a disagreement which is over 180 degrees, add/subtract
        ! n x 360deg
        !
        DO CURRENT_WP=1,SELF%NWP-1
            ! How far off are we?
            ERROR_IN_CHANGE_IN_PHI = FILAMENT_WP_ERROR_IN_CHANGE_IN_PHI(SELF,CURRENT_WP)
            ! If we're more than 180 degrees off, then we can improve things by adding/subtracting 360
            IF (ABS(ERROR_IN_CHANGE_IN_PHI) .GT. 180.) THEN
                SELF%WP(CURRENT_WP+1)%EULER%PHI = SELF%WP(CURRENT_WP+1)%EULER%PHI &
                                                - 360. * SIGN(1.,ERROR_IN_CHANGE_IN_PHI) * NINT(ABS(ERROR_IN_CHANGE_IN_PHI)/360.)
            ENDIF
        ENDDO

    END SUBROUTINE FILAMENT_WPS_ADD_SUB_PHI_360

    !>  \brief  Compute the ideal path_lengths for the waypoints, taking their phi values to be correct
    !!          and aiming for canonical helical symmetry
    FUNCTION FILAMENT_IDEAL_PATH_LENGTHS_FOR_CANONICAL_HELIX(SELF) RESULT(PATH_LENGTHS)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF                !<  Filament object
        ! Result
        REAL,           ALLOCATABLE ::  PATH_LENGTHS(:)
        ! Private varibles
        INTEGER ::  CURRENT_WP
        ! Start work

        ! Allocate & initialise results
        ALLOCATE(PATH_LENGTHS(SELF%NWP))
        PATH_LENGTHS(1) = 0.0

        ! Loop over waypoints
        IF (SELF%HELIX%TWIST .NE. 0.0) THEN
            DO CURRENT_WP=2,SELF%NWP
                PATH_LENGTHS(CURRENT_WP) =  PATH_LENGTHS(CURRENT_WP-1) + &
                                            ( SELF%WP(CURRENT_WP)%EULER%PHI - SELF%WP(CURRENT_WP-1)%EULER%PHI ) * &
                                            SELF%HELIX%RISE / SELF%HELIX%TWIST
            ENDDO
        ELSE
            PATH_LENGTHS = 0.0
        ENDIF
    END FUNCTION FILAMENT_IDEAL_PATH_LENGTHS_FOR_CANONICAL_HELIX

    !>  \brief  The discrepency between the actual path lengths and the path lengths which would accomodate the canonical helical
    !!          symmetry given the waypoints' phi values
    REAL FUNCTION FILAMENT_DEVIATION_FROM_IDEAL_PATH_LENGTHS(SELF,IDEAL_PATH_LENGTHS) RESULT(DEV)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(IN)  ::  SELF                    !<  Filament object
        REAL,   ALLOCATABLE,    OPTIONAL,   INTENT(IN)  ::  IDEAL_PATH_LENGTHS(:)   !<  Desired path length values for the filament's waypoints
        ! Private variables
        INTEGER         ::  CURRENT_WP
        REAL(KIND=8)    ::  DDEV
        REAL,   ALLOCATABLE ::  IIDEAL_PATH_LENGTHS(:)
        ! Start work

        ! Ideal path lengths
        IF (PRESENT(IDEAL_PATH_LENGTHS)) THEN
            IIDEAL_PATH_LENGTHS = IDEAL_PATH_LENGTHS
        ELSE
            IIDEAL_PATH_LENGTHS = FILAMENT_IDEAL_PATH_LENGTHS_FOR_CANONICAL_HELIX(SELF)
        ENDIF

        ! Accumulate sum of deviations
        DDEV = 0.0D0
        DO CURRENT_WP=2,SELF%NWP
            DDEV = DDEV + ABS(SELF%WP(CURRENT_WP)%PATH_LENGTH - IIDEAL_PATH_LENGTHS(CURRENT_WP))
        ENDDO
        DEV = DDEV
    END FUNCTION FILAMENT_DEVIATION_FROM_IDEAL_PATH_LENGTHS

END MODULE FILAMENTS_GEOMETRY
