MODULE FILAMENTS_MICROTUBULES
    USE FILAMENTS_BASE
    IMPLICIT NONE

    CONTAINS

    !>  \brief  Check whether the microtubule is expected to have a seam
    !!  \todo   Implement logic to check arbitrary microtubule geometry
    LOGICAL FUNCTION MICROTUBULE_HAS_SEAM(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)  ::  SELF    !<  Filament object. Must be a microtubule
        ! Private variables

        ! Start work
        IF (.NOT. SELF%MICROTUBULE) THEN
            WRITE(*,'(A)') '**ERROR(MICROTUBULE_HAS_SEAM): filament is not a microtubule'
            CALL TERMINATE('Fatal error in MICROTUBULE_HAS_SEAM')
        ENDIF

        ! I do not have time to implement all possible geometries for now,
        ! so I'm just implementing the particular case I need during development.
        IF (SELF%MT_PF_NUMBER .EQ. 13 .AND. SELF%MT_RISE .EQ. 3) THEN
            MICROTUBULE_HAS_SEAM = .TRUE.
        ELSE
            WRITE(*,'(A)') '**ERROR(MICROTUBULE_HAS_SEAM): feature not implemented yet.'
            CALL TERMINATE('Fatal error in MICROTUBULE_HAS_SEAM')
        ENDIF
    END FUNCTION MICROTUBULE_HAS_SEAM

    !>  \brief  Shift the waypoints of a microtubule filament so that protofilament 1 has been moved PF_SHIFT protofilaments over
    SUBROUTINE MICROTUBULE_SHIFT_PF(SELF,PF_SHIFT)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_SOLVE_SPLINES
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(INOUT)   ::  SELF        !<  Microtubule filament
        INTEGER,            INTENT(IN)      ::  PF_SHIFT    !<  By how many protofilaments should this microtubule be shifted
        ! Private variables
        INTEGER         ::  CURRENT_WP
        REAL            ::  CHANGE_IN_PHI!, CHANGE_IN_T
        !TYPE(WAYPOINT)  ::  WP_TMP
        ! Start work

        ! Check filament is a microtubule
        IF (.NOT. SELF%MICROTUBULE) THEN
            WRITE(*,'(A)') '**ERROR(MICROTUBULE_SHIFT_PF): filament is not a microtubule'
            CALL TERMINATE('Fatal error in MICROTUBULE_SHIFT_PF')
        ENDIF

        ! Check the PF shift value makes sense
        IF (PF_SHIFT .GT. SELF%MT_PF_NUMBER) THEN
            WRITE(*,'(A,2(I0,A))') '**ERROR(MICROTUBULE_SHIFT_PF): cannot shift a ', SELF%MT_PF_NUMBER, '-pf microtubules by ', &
                                    PF_SHIFT, ' protofilaments'
            CALL TERMINATE('Fatal error in MICROTUBULE_SHIFT_PF')
        ENDIF

        ! Loop over waypoints
        DO CURRENT_WP=1,SELF%NWP

            ! Work out changes in T and Phi
            !CHANGE_IN_T = MICROTUBULE_LOCAL_WASHER_RISE_PER_DIMER(SELF,SELF%T(CURRENT_WP)) * REAL(PF_SHIFT)
            CHANGE_IN_PHI = MICROTUBULE_LOCAL_WASHER_TWIST_PER_DIMER(SELF,SELF%T(CURRENT_WP)) * REAL(PF_SHIFT)

            ! Apply change in Phi
            SELF%WP(CURRENT_WP)%EULER%PHI   = SELF%WP(CURRENT_WP)%EULER%PHI - CHANGE_IN_PHI

            ! Apply change in subunit_num (integer subunit_num indicates dimer of PF1)
            SELF%WP(CURRENT_WP)%SUBUNIT_NUM = SELF%WP(CURRENT_WP)%SUBUNIT_NUM - REAL(PF_SHIFT) / REAL(SELF%MT_PF_NUMBER)

        ENDDO

        ! Update splines
        CALL FILAMENT_SOLVE_SPLINES(SELF)   !,USE_PSI_THETA_CONSTRAINTS=.TRUE.) ! Commented out 29-Nov, with introduction of USE_HERMITE_SPLINES property of filaments
    END SUBROUTINE MICROTUBULE_SHIFT_PF

    !>  \brief  Compute the angular distance from one protofilament (tubulin dimer) to the next within a washer at the given
    !!          abscissa along the microtubule filament
    REAL FUNCTION MICROTUBULE_LOCAL_WASHER_TWIST_PER_DIMER(SELF,T) RESULT(TWIST_PER_DIMER)
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_HELICAL_TWIST
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)      ::  SELF        !<  Microtubule filament
        REAL(KIND=8),       INTENT(IN)      ::  T           !<  The abscissa along the filament at which the twist will be computed
        ! Private variables
        REAL    ::  PF1_TO_PF1_TWIST
        ! Start work

        ! What's the twist from PF1 of one washer to PF1 of the next?
        PF1_TO_PF1_TWIST = FILAMENT_EVAL_HELICAL_TWIST(SELF,T) + 360.0

        !
        TWIST_PER_DIMER = PF1_TO_PF1_TWIST/REAL(SELF%MT_PF_NUMBER)
    END FUNCTION MICROTUBULE_LOCAL_WASHER_TWIST_PER_DIMER

    !>  \brief  Compute the distance from one protofilament (tubulin dimer) to the next within a washer at the given
    !!          abscissa along the microtubule filament
    REAL FUNCTION MICROTUBULE_LOCAL_WASHER_RISE_PER_DIMER(SELF,T) RESULT(RISE_PER_DIMER)
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_HELICAL_RISE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)      ::  SELF        !<  Microtubule filament
        REAL(KIND=8),       INTENT(IN)      ::  T           !<  The abscissa along the filament at which the rise will be computed
        ! Private variables
        REAL    ::  LOCAL_RISE
        ! Start work

        ! Work out local rise from PF1 of one washer to PF1 of the next
        LOCAL_RISE = FILAMENT_EVAL_HELICAL_RISE(SELF,T)

        !
        RISE_PER_DIMER = LOCAL_RISE/REAL(SELF%MT_PF_NUMBER)
    END FUNCTION MICROTUBULE_LOCAL_WASHER_RISE_PER_DIMER

    !>  \brief  Given a microtubule symmetry, return the number of the protofilament which can be reconstructed correctly
    INTEGER FUNCTION MICROTUBULE_GET_GOOD_PF(SELF) RESULT(GOOD_PF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, IS_EVEN
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)  ::  SELF    !<  Filament object. Must be a microtubule.
        ! Private variables

        ! Start work

        ! Check the filament is a microtubule
        IF (.NOT. SELF%MICROTUBULE) THEN
            WRITE(*,'(A)') '**ERROR(MICROTUBULE_GET_GOOD_PF): filament is not a microtubule'
            CALL TERMINATE('Fatal error in MICROTUBULE_GET_GOOD_PF')
        ENDIF

        ! Work out which protofilaments we will aim for
        IF (IS_EVEN(SELF%MT_PF_NUMBER)) THEN
            ! Even number of protofilaments
            GOOD_PF = SELF%MT_PF_NUMBER/2
        ELSE
            ! Odd number of protofilaments
            GOOD_PF = (SELF%MT_PF_NUMBER-1)/2 + 1
        ENDIF
    END FUNCTION MICROTUBULE_GET_GOOD_PF

    !>  \brief  Impose microtubule symmetry in real space. This is necessary when calculating
    !!          a reconstruction of a microtubule with a seam, in which case only one protofilament is reconstructed correctly
    SUBROUTINE IMAGE_IMPOSE_MICROTUBULE_SYMMETRY(SELF,VOL)
        USE EULERANGLES
        USE INTERPOLATION
        USE ROTATION_MATRICES, ONLY : RADIANS
        use Images
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(IN)      ::  SELF
        TYPE(IMAGE),                        INTENT(INOUT)   ::  VOL             !<  Volume to be symmetrised
        ! Private variables
        INTEGER                 ::  GOOD_PF
        INTEGER                 ::  I,J,K, II,JJ,KK, II2,JJ2,KK2
        REAL                    ::  MAX_RAD_SQ
        INTEGER                 ::  CURRENT_PF
        REAL                    ::  SEGMENT_PHI
        INTEGER                 ::  ORIGIN(3)
        REAL                    ::  SEGMENT_Z
        REAL                    ::  Z_SHIFT
        LOGICAL,    PARAMETER   ::  SANITY_CHECK = .FALSE.
        LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
        REAL                    ::  PHI_ROT
        REAL                    ::  COO(3)
        REAL                    ::  DIST_FROM_EDGE_OF_PF
        ! Private variables to do with interpolation etc
        TYPE(WINDOW)            ::  WIN
        TYPE(INTERPOLATION_SCHEME)  ::  INTERPOL_SCHEME
        ! Start work

        ! Check the volume is in real space
        IF (.not. VOL%is_in_real_space) THEN
            WRITE(*,'(A)') '**info(IMAGE_IMPOSE_MICROTUBULE_SYMMETRY): volume will be back-Fourier-transformed to real space'
            CALL BWD_FT(VOL)
        ENDIF

        ! Prepare the window which will be used for interpolation
        CALL WIN%AllocateWindow(2,VOL%IsComplex(),VOL%IsAVolume())

        ! Prepare an interpolation scheme
        CALL INTERPOLATION_SCHEME_INIT(INTERPOL_SCHEME,'LINEAR','RECT',2)

        ! Get address of origin (centre) voxel
        if (vol%is_in_real_space .and. vol%origin_at_center) then
            origin = vol%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif

        ! Square the maximum filament radius
        MAX_RAD_SQ = SELF%MAX_RADIUS**2
        IF (DEBUG) PRINT '(A,F0.4)', '**debug(IMAGE_IMPOSE_MICROTUBULE_SYMMETRY): SELF%MAX_RADIUS = ', SELF%MAX_RADIUS

        ! The phi extent of each segment
        SEGMENT_PHI = 360./REAL(SELF%MT_PF_NUMBER)

        ! The Z-shift from one protofilament to the next
        SEGMENT_Z = 1. / REAL(SELF%MT_PF_NUMBER) * REAL(SELF%MT_RISE) / 2. * SELF%HELIX%RISE

        ! Which is the good PF?
        GOOD_PF = MICROTUBULE_GET_GOOD_PF(SELF)
        IF (DEBUG) PRINT '(A,I0)', '**debug(IMAGE_IMPOSE_MICROTUBULE_SYMMETRY): GOOD_PF = ', GOOD_PF

        ! Loop over output voxels
        K = ORIGIN(3)
        KK = 0
        DO J=1,VOL%logical_dimensions(2)
            JJ = (J-ORIGIN(2))
            JJ2 = JJ**2
            DO I=1,VOL%logical_dimensions(1)
                II = (I-ORIGIN(1))
                II2 = II**2
                ! If beyond the maximum cylinder radius of the filament, or if at X,Y origin, skip
                IF (II2+JJ2 .GT. MAX_RAD_SQ .OR. II2+JJ2 .EQ. 0) CYCLE
                ! Which PF are we on currently?
                IF (DEBUG) PRINT '(A,3(F0.3,1X))', '**debug(IMAGE_IMPOSE_MICROTUBULE_SYMMETRY): COO = ', REAL(II),REAL(JJ),REAL(KK)
                CURRENT_PF = MICROTUBULE_PF_FROM_COOS(SELF,[REAL(II),REAL(JJ),REAL(KK)],SEGMENT_PHI)
                IF (DEBUG) PRINT '(A,I0)', '**debug(IMAGE_IMPOSE_MICROTUBULE_SYMMETRY): CURRENT_PF = ', CURRENT_PF
                ! Now we know which protofilament we're in, we can copy the appropriate value to the output volume
                IF (CURRENT_PF .EQ. GOOD_PF) THEN
                    ! unchanged
                    CYCLE
                ELSE
                    ! we need to copy values from the good protofilament
                    PHI_ROT = - SEGMENT_PHI * REAL(CURRENT_PF-GOOD_PF)  ! + or -?
                    Z_SHIFT = - SEGMENT_Z   * REAL(CURRENT_PF-GOOD_PF)  ! + or -?
                    IF (DEBUG) THEN
                        PRINT '(A,I0,1X,F0.3)', '**debug(IMAGE_IMPOSE_MICROTUBULE_SYMMETRY): got here with CURRENT_PF, PHI_ROT = ',&
                                             CURRENT_PF, PHI_ROT
                    ENDIF
                    ! The coordinates from which we're copying
                    COO(1) =  COS(RADIANS(PHI_ROT)) * REAL(II) + SIN(RADIANS(PHI_ROT)) * REAL(JJ)
                    COO(2) = -SIN(RADIANS(PHI_ROT)) * REAL(II) + COS(RADIANS(PHI_ROT)) * REAL(JJ)
                    ! Sanity check - we should now be within the good PF
                    IF (SANITY_CHECK) THEN
                        CURRENT_PF = MICROTUBULE_PF_FROM_COOS(SELF,COO,SEGMENT_PHI)
                        IF (CURRENT_PF .NE. GOOD_PF) THEN
                            ! It appears that we are copying from the wrong PF. But let's first check whether we're actually very close
                            ! to the edge of the good PF, in which case we might just be victim of rounding errors.
                            DIST_FROM_EDGE_OF_PF = MICROTUBULE_PF_FROM_COOS_DISTANCE(SELF,COO,SEGMENT_PHI)
                            IF (DEBUG) THEN
                                PRINT '(A,F0.2)', '**debug(IMAGE_IMPOSE_MICROTUBULE_SYMMETRY): distance from edge = ', &
                                                    DIST_FROM_EDGE_OF_PF
                            ENDIF
                            IF (ABS(DIST_FROM_EDGE_OF_PF) .GT. 0.0005 .OR. ABS(CURRENT_PF-GOOD_PF) .GT. 1) THEN
                                WRITE(*,'(2A,I0,A)') '**ERROR(IMAGE_IMPOSE_MICROTUBULE_SYMMETRY): trying to copy ',  &
                                                    'from wrong PF segment (', CURRENT_PF, ')'
                                CALL TERMINATE('Fatal error in IMAGE_IMPOSE_MICROTUBULE_SYMMETRY')
                            ENDIF
                        ENDIF
                    ENDIF
                    ! Loop over that column of voxels
                    DO K=1,VOL%logical_dimensions(3)
                        KK = (K-ORIGIN(3))
                        KK2 = KK**2
                        ! Check whether the Z shift takes us beyond the edge of the volume
                        IF (K + Z_SHIFT .LT. 1 .OR. K + Z_SHIFT .GT. VOL%logical_dimensions(3)) THEN   ! + or - ?
                            CYCLE
                        ENDIF
                        ! The coordinates from which we're copying
                        COO(3) =  REAL(KK) + Z_SHIFT
                        ! Center the window around the coordinates
                        WIN%CENTERED_ON = COO
                        ! Define the bounds of the window over the volume
                        CALL EVAL_WINDOW_BOUNDS(WIN)
                        ! Copy image data to the window, get interpolated value from window
                        CALL WINDOW_FROM_IMAGE_REAL(VOL,WIN)
                        CALL GET_REAL_FROM_WINDOW(VOL%real_values(I,J,K),WIN,&
                                                    INTERPOL_SCHEME%INTERPOLATION_FUNCTION,&
                                                    INTERPOL_SCHEME%WINDOWING_FUNCTION)
                    ENDDO
                ENDIF
            ENDDO
        ENDDO
    END SUBROUTINE IMAGE_IMPOSE_MICROTUBULE_SYMMETRY

    !>  \brief  Find out which protofilament pie slice the coordinates lie in
    PURE INTEGER FUNCTION MICROTUBULE_PF_FROM_COOS(SELF,COOS,SEGMENT_PHI) RESULT(PF)
        USE ROTATION_MATRICES, ONLY : DEGREES
        USE RUNTIME_PARAMETERS, ONLY : MT_PHI_FIRST_PF
        USE EULERANGLES, ONLY : ANGLE_BTWN_0_AND_360
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)  ::  SELF        !<  Microtubule filament
        REAL,               INTENT(IN)  ::  COOS(3)     !<  Coordinates (relative to origin at center of volume)
        REAL,               INTENT(IN)  ::  SEGMENT_PHI !<  The Phi extent of each segment
        ! Private variables
        REAL(KIND=8)                    ::  PHI
        REAL(KIND=8)                    ::  PHI_DIFF
        !LOGICAL,   PARAMETER           ::  DEBUG = .TRUE.
        ! Start work
        !IF (DEBUG) PRINT '(A,4(F0.3,X))', '**debug(MICROTUBULE_PF_FROM_COOS): COOS, SEGMENT_PHI  =', COOS, SEGMENT_PHI
        ! Which Phi are we at right now?
        PHI = DEGREES(-ATAN2(COOS(2),COOS(1)))
        ! How far is that from the center of the PF1 slice of the pie? (come up with a positive difference angle between 0 and 360.)
        PHI_DIFF = ANGLE_BTWN_0_AND_360(PHI - MT_PHI_FIRST_PF)
        !IF (DEBUG) PRINT '(A,2(F0.3,X))', '**debug(MICROTUBULE_PF_FROM_COOS): PHI, PHI_DIFF = ', PHI, PHI_DIFF
        !  Now we can work out which protofilament pie slice we're in
        IF (PHI_DIFF .LE. SEGMENT_PHI/2. .OR. PHI_DIFF .GE. (REAL(SELF%MT_PF_NUMBER)-0.5)*SEGMENT_PHI) THEN
            PF = 1
        ELSE
            PF = 2 + INT(PHI_DIFF/SEGMENT_PHI-0.5)
        ENDIF
        !IF (DEBUG) PRINT '(A,I0,X,F0.6,X,F0.6)', '**debug(MICROTUBULE_PF_FROM_COOS): PF = ', PF, PHI_DIFF/SEGMENT_PHI, MOD(PHI_DIFF,SEGMENT_PHI)-0.5*SEGMENT_PHI
    END FUNCTION MICROTUBULE_PF_FROM_COOS

    !>  \brief  Return the angular distance (degrees) of the coordinates from the assigned PF segment
    PURE REAL FUNCTION MICROTUBULE_PF_FROM_COOS_DISTANCE(SELF,COOS,SEGMENT_PHI)
        USE ROTATION_MATRICES, ONLY : DEGREES
        USE RUNTIME_PARAMETERS, ONLY : MT_PHI_FIRST_PF
        USE EULERANGLES, ONLY : ANGLE_BTWN_0_AND_360
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)  ::  SELF        !<  Microtubule filament
        REAL,               INTENT(IN)  ::  COOS(3)     !<  Coordinates (relative to origin at center of volume)
        REAL,               INTENT(IN)  ::  SEGMENT_PHI !<  The Phi extent of each segment
        ! Private variables
        REAL(KIND=8)                    ::  PHI
        REAL                            ::  PHI_DIFF
        LOGICAL,            PARAMETER   ::  DEBUG = .TRUE.
        ! Start work

        ! Which Phi are we at right now?
        PHI = DEGREES(-ATAN2(COOS(2),COOS(1)))
        ! How far is that from the center of the PF1 slice of the pie? (come up with a positive difference angle between 0 and 360.)
        PHI_DIFF = ANGLE_BTWN_0_AND_360(PHI - MT_PHI_FIRST_PF)

        ! Result
        MICROTUBULE_PF_FROM_COOS_DISTANCE = MOD(PHI_DIFF,SEGMENT_PHI)-0.5*SEGMENT_PHI
    END FUNCTION MICROTUBULE_PF_FROM_COOS_DISTANCE

END MODULE FILAMENTS_MICROTUBULES



