!>  \brief  Minimisation methods for the filament class
MODULE FILAMENTS_MINIMISATION
    USE FILAMENTS_BASE
    IMPLICIT NONE

    CONTAINS

    !>  \brief  Minimise deviations from canonical helix by moving waypoints in Z direction
    SUBROUTINE FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX(SELF)
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_RELAX_HELIX, FILAMENT_RELAX_RISE_PER_SUBUNIT
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_IDEAL_PATH_LENGTHS_FOR_CANONICAL_HELIX
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_WPS_Z_FROM_WPS_THETA_AND_RIGID_Z, FILAMENT_UPDATE_MODEL
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_EXTRA_INFO
        USE FILAMENTS, ONLY : FILAMENT_PRINT_INFO, FILAMENT_PRINT_DIFF
        USE DIFFERENTIAL_GEOMETRY, ONLY : SPACE_CURVE, SPACE_CURVE_INIT, SPACE_CURVE_ADJUST_Z_TO_MATCH_ARC_LENGTHS
        USE DIFFERENTIAL_GEOMETRY, ONLY : SPACE_CURVE_EVAL_TOTAL_CURVATURE_NEW, SPACE_CURVE_PRINT_INFO
		USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_TOTAL_CURVATURE_OVERTWIST_STRETCH, FILAMENT_DEVIATIONS_FROM_CANONICAL_HELIX
		USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_TOTAL_CURVATURE
		USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF            !<  Filament object
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG                  = .TRUE.
        LOGICAL,    PARAMETER   ::  ALL_WPS_SIMULTANEOUSLY = .FALSE.
        LOGICAL,    PARAMETER   ::  USE_SPACE_CURVES       = .TRUE.
        INTEGER,    PARAMETER   ::  REFINE_WPS_GROUP_SIZE  = 3
        INTEGER,    PARAMETER   ::  REFINE_WPS_GROUP_OLAP  = 2
        INTEGER,    PARAMETER   ::  NUMBER_OF_ITERATIONS   = 2
        INTEGER                 ::  REFINE_WPS_FIRST, REFINE_WPS_LAST
        TYPE(FILAMENT)          ::  FIL_BAK
        INTEGER                 ::  CURRENT_ITERATION
        TYPE(SPACE_CURVE)       ::  CURRENT_SPACE_CURVE
        REAL(KIND=8), ALLOCATABLE :: IDEAL_PATH_LENGTHS(:)
        REAL(KIND=8)            ::  CURVATURE, OVERTWIST, STRETCH
        REAL(KIND=8)            ::  CURVATURE_START, CURVATURE_FINISH
        ! Start work

        WRITE(*,'(A)') '**info(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): attempting Z coo adjustments to minimise '// &
                                                                                'discrepency from canonical helical symmetry'
        ! Keep track of our starting point
        FIL_BAK = SELF
        CALL SPACE_CURVE_INIT(  CURRENT_SPACE_CURVE,SELF%WP,KNOWN_D=.TRUE.,PARAM_ERROR_TOL=0.01)
        CALL SPACE_CURVE_PRINT_INFO(CURRENT_SPACE_CURVE)
        CURVATURE_START = SPACE_CURVE_EVAL_TOTAL_CURVATURE_NEW(CURRENT_SPACE_CURVE,STEP_SIZE=0.1,SQUARED_CURVATURE=.FALSE.)
        IF (DEBUG) THEN
            CALL FILAMENT_PRINT_INFO(SELF)
            CALL FILAMENT_PRINT_EXTRA_INFO(SELF)
        ENDIF
        ! We will only modify the Z coordinates
        SELF%REFINE_PARAMS = .FALSE.
        SELF%REFINE_PARAMS(3) = .TRUE.
        ! We move the Z coordinates so as to minimise vertical curvature. We can do this since we know the
        ! out-of-plane angle (theta) at each waypoint and the avereage Z of the filament (rigid_wp%coo(3))
        IF (DEBUG) THEN
            PRINT *, '**debug(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): rigid wp Z = ', SELF%RIGID_WP%COO(3)
            CALL FILAMENT_EVAL_TOTAL_CURVATURE_OVERTWIST_STRETCH(SELF,CURVATURE,OVERTWIST,STRETCH,  &
                                                     IGNORE_FAL_WPS=.NOT. (SELF%FAL_WPS_IN_SCORE .AND. SELF%FAL_WPS_IN_THREED))
            WRITE(*,'(3A,3(G15.4,1X))')  '**debug(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): ',   &
                                        'before minimising Z curvature, ', &
                                        'total curvature, overtwist, stretch (1/pixels) : ', CURVATURE, OVERTWIST, STRETCH

        ENDIF
        CALL FILAMENT_UPDATE_WPS_Z_FROM_WPS_THETA_AND_RIGID_Z(SELF)
        CALL FILAMENT_UPDATE_MODEL(SELF)
        IF (DEBUG) THEN
            PRINT *, '**debug(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): rigid wp Z = ', SELF%RIGID_WP%COO(3)
            CALL FILAMENT_EVAL_TOTAL_CURVATURE_OVERTWIST_STRETCH(SELF,CURVATURE,OVERTWIST,STRETCH,  &
                                                     IGNORE_FAL_WPS=.NOT. (SELF%FAL_WPS_IN_SCORE .AND. SELF%FAL_WPS_IN_THREED))
            WRITE(*,'(3A,3(G15.4,1X))')  '**debug(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): ',   &
                                        'after minimising Z curvature, ', &
                                        'total curvature, overtwist, stretch (1/pixels) : ', CURVATURE, OVERTWIST, STRETCH

            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

        ! No we can relax the helix again, but only by adding subunits or turns between waypoints, so that if anything it
        ! will have overtwist and compression.
        ! That way, the minimiser will later be able to add arc length between waypoints to "accomodate" these extra
        ! subunits or turns and relieve the overtwist and compression.
        ! On the other hand, if we let the helix relaxer subtract subunits or turns between waypoints,
        ! the minimiser will not be able to do anything about it, since we already have the waypoints as close
        ! together as possible
        DO CURRENT_ITERATION=1,NUMBER_OF_ITERATIONS
            IF (CURRENT_ITERATION==1) THEN
                CALL FILAMENT_RELAX_HELIX(SELF,FORBID_UNDERTWIST=.TRUE.,FORBID_STRETCHING=.TRUE.)
            ELSE
                CALL FILAMENT_RELAX_HELIX(SELF,FORBID_UNDERTWIST=.FALSE.,FORBID_STRETCHING=.FALSE.)
            ENDIF
            CALL FILAMENT_UPDATE_MODEL(SELF)
            IF (DEBUG) THEN
                CALL FILAMENT_PRINT_INFO(SELF)
                CALL FILAMENT_PRINT_EXTRA_INFO(SELF)
            ENDIF
            ! What are our ideal path lengths?
            IDEAL_PATH_LENGTHS = FILAMENT_IDEAL_PATH_LENGTHS_FOR_CANONICAL_HELIX(SELF)
            print *, '**debug(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): ideal path lengths = ', IDEAL_PATH_LENGTHS
			IF (ANY(IDEAL_PATH_LENGTHS .LT. 0.0)) THEN
				WRITE(*,'(A)') '**ERROR(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): weird values for ideal path lengths'
				print *, '**debug(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): ideal path lengths = ', IDEAL_PATH_LENGTHS
				CALL FILAMENT_PRINT_INFO(SELF)
				CALL TERMINATE('Fatal error in FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX')
			ENDIF


            ! We will refine waypoints' Z coordinates.
            ! Optionally, we do this by groups of waypoints to take it easy on the minimiser and
            ! hopefully save time too
            REFINE_WPS_FIRST = 0
            REFINE_WPS_LAST  = 0
            DO
                IF (ALL_WPS_SIMULTANEOUSLY) THEN
                    REFINE_WPS_FIRST = 1
                    REFINE_WPS_LAST  = SELF%NWP
                ELSE
                    ! The first and last waypoints which will be refined this time
                    REFINE_WPS_FIRST = REFINE_WPS_FIRST + REFINE_WPS_GROUP_SIZE - REFINE_WPS_GROUP_OLAP
                    REFINE_WPS_LAST  = REFINE_WPS_FIRST + REFINE_WPS_GROUP_SIZE - 1
                    ! Take care of edge conditions
                    IF (REFINE_WPS_LAST .GT. SELF%NWP) REFINE_WPS_LAST = SELF%NWP
                ENDIF
                WRITE(*,'(2A,I0,A,I0)') '**info(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): tweaking Z coordinates for ', &
                                        'waypoints ', REFINE_WPS_FIRST, ' to ', REFINE_WPS_LAST
                IF (USE_SPACE_CURVES) THEN
                    CALL SPACE_CURVE_INIT(  CURRENT_SPACE_CURVE,SELF%WP(REFINE_WPS_FIRST:REFINE_WPS_LAST),    &
                                            KNOWN_D=.TRUE.,PARAM_ERROR_TOL=0.01)

                    CALL SPACE_CURVE_ADJUST_Z_TO_MATCH_ARC_LENGTHS(CURRENT_SPACE_CURVE,             &
                       TARGET_ARC_LENGTH=    IDEAL_PATH_LENGTHS(REFINE_WPS_FIRST:REFINE_WPS_LAST)   &
                                            -IDEAL_PATH_LENGTHS(REFINE_WPS_FIRST))
                    ! Plug the values back in to the filament
                    SELF%WP(REFINE_WPS_FIRST:REFINE_WPS_LAST)%COO(3) = CURRENT_SPACE_CURVE%Z_SPLINE%Y(:)
                    CALL FILAMENT_UPDATE_MODEL(SELF)
                ELSE
                    ! Set the refine flags
                    SELF%WP%REFINE = .FALSE.
                    SELF%WP(REFINE_WPS_FIRST:REFINE_WPS_LAST)%REFINE = .TRUE.
                    ! Work on the current group of waypoints
                    CALL FILAMENT_MINIMISE( SELF,         &
                                            MINIMISE_CURVATURE=.FALSE.,     &
                                            MINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS=.TRUE.  )
                ENDIF
                ! Check whether this was the last group of waypoints we're to process
                IF (REFINE_WPS_LAST .EQ. SELF%NWP) THEN
                    EXIT
                ENDIF
             ENDDO
            CALL FILAMENT_RELAX_RISE_PER_SUBUNIT(SELF,FORCE_INTG_SUBUNIT_NUM=.TRUE.)
        ENDDO
        CALL FILAMENT_UPDATE_MODEL(SELF)
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): before relax helix, filament info:'
            CALL FILAMENT_PRINT_INFO(SELF)
            CALL FILAMENT_PRINT_EXTRA_INFO(SELF)
        ENDIF

        CALL FILAMENT_RELAX_HELIX(SELF,FORBID_UNDERTWIST=.FALSE.,FORBID_STRETCHING=.FALSE.)
        CALL FILAMENT_UPDATE_MODEL(SELF)
        CALL SPACE_CURVE_INIT(  CURRENT_SPACE_CURVE,SELF%WP,KNOWN_D=.TRUE.,PARAM_ERROR_TOL=0.01)
        CALL SPACE_CURVE_PRINT_INFO(CURRENT_SPACE_CURVE)
        CURVATURE_FINISH = SPACE_CURVE_EVAL_TOTAL_CURVATURE_NEW(CURRENT_SPACE_CURVE,STEP_SIZE=0.1,SQUARED_CURVATURE=.FALSE.)
        !IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): on exit, filament info:'
            CALL FILAMENT_PRINT_INFO(SELF)
            CALL FILAMENT_PRINT_EXTRA_INFO(SELF)
            WRITE(*,'(A)') '**debug(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): on exit, filament changes:'
            CALL FILAMENT_PRINT_DIFF(FIL_BAK,SELF)
            WRITE(*,'(A,2(F0.2,1X))') '**debug(FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX): curvature start, finish: ', &
                                    CURVATURE_START, CURVATURE_FINISH
        !ENDIF

    END SUBROUTINE FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX

    !>  \brief  Minimise physical characteristics of the filament, e.g. curvature, torsion etc.
    !!  \todo   Use case construct and character string argument to specify what to minimise
    SUBROUTINE FILAMENT_MINIMISE(SELF,  MINIMISE_CURVATURE,MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX,  &
                                        MINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS)
        USE MINIMISERS, ONLY : DS_INIT, DS_RUN, MINIMISER_DS_FGSL, DS_MAX_ITERATIONS, DS_SET_START_STEP
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS, ONLY : VALUES_TO_FILAMENT, VALUES_FROM_FILAMENT
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_DISCREPENCY_BTW_WPS_PHI_AND_CANONICAL_HELIX
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_DEVIATION_FROM_IDEAL_PATH_LENGTHS
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_TOTAL_CURVATURE
        USE CONJUGATE_GRADIENT, ONLY : CG_INIT, CG_MINIMIZER, CG_RUN
        USE, INTRINSIC :: ISO_C_BINDING
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)       ::  SELF                !<  Filament object
        LOGICAL,            OPTIONAL,       INTENT(IN)          ::  MINIMISE_CURVATURE  !<  Whether the curvature should be minimised. Defaults to true.
        LOGICAL,            OPTIONAL,       INTENT(IN)          ::  MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX
        LOGICAL,            OPTIONAL,       INTENT(IN)          ::  MINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS
        ! Private variables
        LOGICAL                                 ::  MMINIMISE_CURVATURE
        LOGICAL                                 ::  MMINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX
        LOGICAL                                 ::  MMINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS
        CHARACTER(LEN=2)                        ::  NUM_DIM_C
        ! Private variables for the minimiser
        TYPE(MINIMISER_DS_FGSL)                 ::  DS_MINIMISER
        TYPE(CG_MINIMIZER)                      ::  CG_MINIM
        INTEGER                                 ::  NUM_DIM, IERR
        REAL(KIND=8),           ALLOCATABLE     ::  BEST_VALUES(:)
        REAL(KIND=8)                            ::  BEST_SCORE
        REAL                                    ::  DS_TARGET_SIZE                  !<  Target size of the simplex.
        LOGICAL,                PARAMETER       ::  DEBUG   =   .FALSE.
        REAL(KIND=8),           ALLOCATABLE     ::  FIRST_STEP(:), START_VALUES(:), ACCURACY(:)
        LOGICAL,				PARAMETER		::	PATH_LENGTHS_CG = .FALSE.
		! Start work


        ! What will we be minimising?
        MMINIMISE_CURVATURE = .TRUE.
        IF (PRESENT(MINIMISE_CURVATURE)) MMINIMISE_CURVATURE = MINIMISE_CURVATURE
        MMINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX = .FALSE.
        IF (PRESENT(MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX)) THEN
            MMINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX = MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX
        ENDIF
        MMINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS = .FALSE.
        IF (PRESENT(MINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS)) THEN
            MMINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS = MINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS
        ENDIF

        IF (COUNT([ MMINIMISE_CURVATURE,MMINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX,   &
                    MMINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS]) .GT. 1               ) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_MINIMISE): can only minimise one property at a time!'
            CALL TERMINATE('Fatal programming error')
        ENDIF

        ! How many dimensions will the minimiser work with?
        NUM_DIM = COUNT(SELF%REFINE_PARAMS) * COUNT(SELF%WP%REFINE)

!        ! Sanity check
!        IF (MMINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS) THEN
!            IF (NUM_DIM .NE. SELF%NWP) THEN
!                WRITE(*,'(A,I0,1X,I0)') '**ERROR(FILAMENT_MINIMISE): num_dim .ne. self%nwp: ', NUM_DIM, SELF%NWP
!                WRITE(*,'(A,7(L1,1X))') '**ERROR(FILAMENT_MINIMISE): refine_params: ', SELF%REFINE_PARAMS
!                CALL TERMINATE('Failed sanity check in FILAMENT_MINIMISE')
!            ENDIF
!        ENDIF

        ! Allocate memory for the result and the start / first step values
        ALLOCATE(BEST_VALUES(NUM_DIM),START_VALUES(NUM_DIM),FIRST_STEP(NUM_DIM),ACCURACY(NUM_DIM),STAT=IERR)
        IF (IERR .NE. 0) CALL TERMINATE('Memory allocation failed')

        ! Set the starting position and the first step for the DS
        IF (MMINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS) THEN
            !START_VALUES = SELF%WP%COO(3)
            CALL VALUES_FROM_FILAMENT(SELF,START_VALUES,SELF%REFINE_PARAMS, &
                                        INCREMENTAL_VALUES=[.FALSE.,.FALSE.,.TRUE.,.FALSE.,.FALSE.,.FALSE.,.FALSE.])
            FIRST_STEP = 20.0
            ACCURACY=0.5
        ELSE
            CALL TERMINATE('Not implemented properly')
            START_VALUES = SELF%WP(1:3)%COO(3)
            FIRST_STEP = 10.0
            ACCURACY=0.5
        ENDIF

        ! Work out a reasonable value for the DS target size
        WRITE(*,'(A)') '**warning(FILAMENT_MINIMISE): setting the ds target size to arbitrary value of 0.1'
        DS_TARGET_SIZE = 0.5

        ! Initialise the downhill simplex minimiser
        IF (MMINIMISE_CURVATURE) THEN
            CALL DS_INIT(   DS_MINIMISER,FILAMENT_EVAL_TOTAL_CURVATURE_WRAP,NUM_DIM,SELF%CPTR,  &
                            INITIAL_SCORE=REAL(FILAMENT_EVAL_TOTAL_CURVATURE(SELF)) )
        ELSE IF (MMINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS) THEN
			IF (.NOT. PATH_LENGTHS_CG) THEN
            	CALL DS_INIT(   DS_MINIMISER,FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP,NUM_DIM,SELF%CPTR,  &
                	            INITIAL_SCORE=-FILAMENT_DEVIATION_FROM_IDEAL_PATH_LENGTHS(SELF), &
							   RANDOM = .FALSE.	)
			ELSE
            !PRINT *, 'START_VALUES  = ', START_VALUES
            !PRINT *, 'ACCURACY      = ', ACCURACY
            	CALL CG_INIT(   CG_MINIM,NUM_DIM,FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP,SELF%CPTR,  &
            	                STARTING_VALUES=START_VALUES, &
            	                ACCURACY=ACCURACY )
			ENDIF
        ELSE IF (MMINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX) THEN
            CALL DS_INIT(   DS_MINIMISER,FIL_DEV_FROM_CANON_HELIX_WRAP,NUM_DIM,SELF%CPTR,   &
                            INITIAL_SCORE=FILAMENT_DISCREPENCY_BTW_WPS_PHI_AND_CANONICAL_HELIX(SELF) )
        ENDIF



        ! debug
        !print *, '**debug(filament_minimise): ds_max_iterations = ', ds_max_iterations
        !print *, '**debug(filament_minimise): self pointer associated? = ', c_associated(self%cptr)
        !print *, '**debug(filament_minimise): self refine_params  = ', self%refine_params


        IF (MMINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS .AND. PATH_LENGTHS_CG) THEN
            CALL CG_RUN(CG_MINIM)
            BEST_VALUES = SELF%WP%COO(3)
            BEST_SCORE = FILAMENT_DEVIATION_FROM_IDEAL_PATH_LENGTHS(SELF)
        ELSE
            !print *, '**debug(filament_minimise): self refine_params 2a= ', self%refine_params
            !call c_f_pointer(self%cptr,tmpfil)
            !print *, '**debug(filament_minimise): self refine_params 2b=', tmpfil%refine_params
            CALL DS_SET_START_STEP(DS_MINIMISER,START_VALUES,FIRST_STEP)
            !print *, '**debug(filament_minimise): self refine_params 3= ', self%refine_params

            ! Run the minimiser
            CALL DS_RUN(DS_MINIMISER,DS_TARGET_SIZE)
            ! Remember the best values
            BEST_VALUES =   DS_MINIMISER%CURRENT_BEST_FGSL
            BEST_SCORE  =   DS_MINIMISER%CURRENT_MINIMUM_FGSL
        ENDIF



        IF (DEBUG) THEN
           WRITE(NUM_DIM_C,'(I2)') NUM_DIM
           WRITE(*,'(A,'//NUM_DIM_C//'(F0.3,1X))') '**debug(FILAMENT_MINIMISE): best values = ', BEST_VALUES
           WRITE(*,'(A,F0.3)') '**debug(FILAMENT_MINIMISE): best score = ', BEST_SCORE
        ENDIF

        ! Set the filament to the best values
        CALL VALUES_TO_FILAMENT(SELF,BEST_VALUES,SELF%REFINE_PARAMS,    &
                                INCREMENTAL_VALUES=[.FALSE.,.FALSE.,.TRUE.,.FALSE.,.FALSE.,.FALSE.,.FALSE.])
        CALL FILAMENT_UPDATE_MODEL(SELF)

        IF (DEBUG) THEN
           WRITE(*,'(A)') '**debug(FILAMENT_MINIMISE): filament info at end:'
           CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF
    END SUBROUTINE FILAMENT_MINIMISE

    FUNCTION FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP(VALUES,PARAMS) BIND(C)
        USE, INTRINSIC  ::  ISO_C_BINDING
        USE FGSL, ONLY : FGSL_VECTOR, FGSL_DOUBLE, FGSL_VECTOR_ALIGN, FGSL_SUCCESS, FGSL_OBJ_C_PTR
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_DEVIATION_FROM_IDEAL_PATH_LENGTHS
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS, ONLY : VALUES_TO_FILAMENT
        IMPLICIT NONE
        ! Arguments
        TYPE(C_PTR),        VALUE       ::  VALUES  !<  C pointer to current values to try out
        TYPE(C_PTR),        VALUE       ::  PARAMS  !<  C pointer to a filament object
        ! Result
        REAL(KIND=C_DOUBLE)             ::  FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP
        ! Private variables
        INTEGER                         ::  STATUS
        TYPE(FGSL_VECTOR)               ::  VAL_VEC
        REAL(FGSL_DOUBLE),  POINTER     ::  VAL_FGSL(:)
        REAL(KIND=8),       ALLOCATABLE ::  VAL(:)
        REAL(KIND=8),       POINTER     ::  VAL_PTR(:)
        TYPE(FILAMENT),     POINTER     ::  SELF
        LOGICAL,            PARAMETER   ::  DEBUG   =   .FALSE.
        CHARACTER(LEN=6)                ::  SIZE_VAL_C
        LOGICAL,            PARAMETER   ::  INPUT_VALS_ARE_FGSL = .TRUE.
        INTEGER                         ::  NUM_DIM
        ! Start work

        ! "Unpack" the filament object
        CALL C_F_POINTER(PARAMS,SELF)
        IF (DEBUG) PRINT *, '**debug(FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP): REFINE_PARAMS = ', SELF%REFINE_PARAMS
        NUM_DIM = COUNT(SELF%REFINE_PARAMS)*SELF%NWP

        ! Unpack the values to try
        IF (INPUT_VALS_ARE_FGSL) THEN
            CALL FGSL_OBJ_C_PTR(VAL_VEC,VALUES)
            STATUS = FGSL_VECTOR_ALIGN(VAL_FGSL,VAL_VEC)
            IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('FGSL_VECTOR_ALIGN failed')
            ALLOCATE(VAL(SIZE(VAL_FGSL)),STAT=STATUS)
            IF (STATUS .NE. 0) CALL TERMINATE('Memory allocation failed')
            VAL = VAL_FGSL
        ELSE
            CALL C_F_POINTER(VALUES,VAL_PTR,SHAPE=[NUM_DIM])
            ALLOCATE(VAL(NUM_DIM))
            VAL = VAL_PTR
        ENDIF

        ! Let the user know what values we're trying
        IF (DEBUG) THEN
            WRITE(SIZE_VAL_C,'(I6)') SIZE(VAL)
            WRITE(*,'(A,'//SIZE_VAL_C//'(F0.3,1X))') '**debug(FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP): trying these values: ', VAL
        ENDIF

        ! Assign those values to the waypoints
        CALL VALUES_TO_FILAMENT(SELF,VAL,SELF%REFINE_PARAMS,    &
                                INCREMENTAL_VALUES=[.FALSE.,.FALSE.,.TRUE.,.FALSE.,.FALSE.,.FALSE.,.FALSE.])

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP): filament info after values_to_filament:'
            CALL FILAMENT_PRINT_INFO(SELF)
            WRITE(*,'(A)') '**debug(FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP): about to call filament_update_model'
        ENDIF

        ! Update the filament model
        CALL FILAMENT_UPDATE_MODEL(SELF)

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP): filament info after update model: '
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

        !
        FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP = FILAMENT_DEVIATION_FROM_IDEAL_PATH_LENGTHS(SELF)

        IF (DEBUG) THEN
            WRITE(*,'(A,(G15.6))') '**debug(FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP): got this score: ', &
                                    FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP
        ENDIF
    END FUNCTION FIL_DEV_FROM_CANON_PATH_LENGTHS_WRAP



    FUNCTION FIL_DEV_FROM_CANON_HELIX_WRAP(VALUES,PARAMS) BIND(C)
        USE, INTRINSIC  ::  ISO_C_BINDING
        USE FGSL, ONLY : FGSL_VECTOR, FGSL_DOUBLE, FGSL_VECTOR_ALIGN, FGSL_SUCCESS, FGSL_OBJ_C_PTR
		USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_DISCREPENCY_BTW_WPS_PHI_AND_CANONICAL_HELIX
		USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
		USE RUNTIME_PARAMETERS, ONLY : TERMINATE
		USE FILAMENTS, ONLY : VALUES_TO_FILAMENT
        IMPLICIT NONE
        ! Arguments
        TYPE(C_PTR),        VALUE       ::  VALUES  !<  C pointer to current values to try out
        TYPE(C_PTR),        VALUE       ::  PARAMS  !<  C pointer to a filament object
        ! Result
        REAL(KIND=C_DOUBLE)             ::  FIL_DEV_FROM_CANON_HELIX_WRAP
        ! Private variables
        INTEGER                         ::  STATUS
        TYPE(FGSL_VECTOR)               ::  VAL_VEC
        REAL(FGSL_DOUBLE),  POINTER     ::  VAL_FGSL(:)
        REAL(KIND=8),       ALLOCATABLE ::  VAL(:)
        TYPE(FILAMENT),     POINTER     ::  SELF
        LOGICAL,            PARAMETER   ::  DEBUG   =   .FALSE.
        CHARACTER(LEN=6)                ::  SIZE_VAL_C
        ! Start work

        ! "Unpack" the filament object
        CALL C_F_POINTER(PARAMS,SELF)

        ! Unpack the values to try
        CALL FGSL_OBJ_C_PTR(VAL_VEC,VALUES)
        STATUS = FGSL_VECTOR_ALIGN(VAL_FGSL,VAL_VEC)
        IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('FGSL_VECTOR_ALIGN failed')
        ALLOCATE(VAL(SIZE(VAL_FGSL)),STAT=STATUS)
        IF (STATUS .NE. 0) CALL TERMINATE('Memory allocation failed')
        VAL = VAL_FGSL

        ! Let the user know what values we're trying
        IF (DEBUG) THEN
            WRITE(SIZE_VAL_C,'(I6)') SIZE(VAL)
            WRITE(*,'(A,'//SIZE_VAL_C//'(F0.3,X))') '**debug(FIL_DEV_FROM_CANON_HELIX_WRAP): trying these values: ', VAL
        ENDIF

        ! Assign those values to the waypoints
        CALL VALUES_TO_FILAMENT(SELF,VAL,SELF%REFINE_PARAMS)

        ! Update the filament model
        CALL FILAMENT_UPDATE_MODEL(SELF)

        !
        FIL_DEV_FROM_CANON_HELIX_WRAP = FILAMENT_DISCREPENCY_BTW_WPS_PHI_AND_CANONICAL_HELIX(SELF)

        IF (DEBUG) THEN
            WRITE(*,'(A,(G15.6))') '**debug(FIL_DEV_FROM_CANON_HELIX_WRAP): got this score: ', FIL_DEV_FROM_CANON_HELIX_WRAP
        ENDIF
    END FUNCTION FIL_DEV_FROM_CANON_HELIX_WRAP

!   FUNCTION FILAMENT_EVAL_PHASING_STDEV_WRAP(VALUES,PARAMS) BIND(C)
!       USE, INTRINSIC  ::  ISO_C_BINDING
!       USE RUNTIME_PARAMETERS
!       IMPLICIT NONE
!       ! Arguments
!       TYPE(C_PTR),        VALUE       ::  VALUES  !<  C pointer to current values to try out
!       TYPE(C_PTR),        VALUE       ::  PARAMS  !<  C pointer to a filament object
!       ! Result
!       REAL(KIND=C_DOUBLE)             ::  FILAMENT_EVAL_PHASING_STDEV_WRAP
!       ! Private variables
!       INTEGER                         ::  STATUS
!       TYPE(FGSL_VECTOR)               ::  VAL_VEC
!       REAL(FGSL_DOUBLE),  POINTER     ::  VAL_FGSL(:)
!       REAL(KIND=8),       ALLOCATABLE ::  VAL(:)
!       TYPE(FILAMENT),     POINTER     ::  SELF
!       REAL(KIND=8)                    ::  AVE_STD_PHASING(2)
!       LOGICAL,            PARAMETER   ::  DEBUG   =   .TRUE.
!       ! Start work
!
!       ! "Unpack" the filament object
!       CALL C_F_POINTER(PARAMS,SELF)
!
!       ! Unpack the values to try
!       CALL FGSL_OBJ_C_PTR(VAL_VEC,VALUES)
!       STATUS = FGSL_VECTOR_ALIGN(VAL_FGSL,VAL_VEC)
!       IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('FGSL_VECTOR_ALIGN failed')
!       ALLOCATE(VAL(SIZE(VAL_FGSL)),STAT=STATUS)
!       IF (STATUS .NE. 0) CALL TERMINATE('Memory allocation failed')
!       VAL = VAL_FGSL
!
!       ! Let the user know what values we're trying
!       IF (DEBUG) THEN
!           WRITE(*,'(A,<SIZE(VAL)>(F0.3,X))') '**debug(FILAMENT_EVAL_PHASING_STDEV_WRAP): trying these values: ', VAL
!       ENDIF
!
!       ! Assign those values to the waypoints
!       CALL VALUES_TO_FILAMENT(SELF,VAL,SELF%REFINE_PARAMS)
!
!       ! Update the filament model
!       CALL FILAMENT_UPDATE_MODEL(SELF)
!
!       ! Compute the average and standard deviation phasing
!       AVE_STD_PHASING = FILAMENT_COMPUTE_AVE_STD_WP_PHASING(SELF)
!       FILAMENT_EVAL_PHASING_STDEV_WRAP = AVE_STD_PHASING(2)
!
!       IF (DEBUG) THEN
!           WRITE(*,'(A,(G0.6))') '**debug(FILAMENT_EVAL_PHASING_STDEV_WRAP): got this score: ', FILAMENT_EVAL_PHASING_STDEV_WRAP
!       ENDIF
!   END FUNCTION FILAMENT_EVAL_PHASING_STDEV_WRAP


    !>  \brief  Wrapper to be called by the DS minimiser
    FUNCTION FILAMENT_EVAL_TOTAL_CURVATURE_WRAP(VALUES,PARAMS) BIND(C)
        USE, INTRINSIC  ::  ISO_C_BINDING
        USE FGSL, ONLY : FGSL_VECTOR, FGSL_DOUBLE, FGSL_OBJ_C_PTR, FGSL_VECTOR_ALIGN, FGSL_SUCCESS
		USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_TOTAL_CURVATURE
		USE RUNTIME_PARAMETERS, ONLY : TERMINATE
		USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
		USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
		USE FILAMENTS, ONLY : VALUES_TO_FILAMENT
        IMPLICIT NONE
        ! Arguments
        TYPE(C_PTR),        VALUE       ::  VALUES  !<  C pointer to current values to try out
        TYPE(C_PTR),        VALUE       ::  PARAMS  !<  C pointer to a filament object
        ! Result
        REAL(KIND=C_DOUBLE)             ::  FILAMENT_EVAL_TOTAL_CURVATURE_WRAP
        ! Private variables
        INTEGER                         ::  STATUS
        TYPE(FGSL_VECTOR)               ::  VAL_VEC
        REAL(FGSL_DOUBLE),  POINTER     ::  VAL_FGSL(:)
        REAL(KIND=8),       ALLOCATABLE ::  VAL(:)
        TYPE(FILAMENT),     POINTER     ::  SELF
        LOGICAL,            PARAMETER   ::  DEBUG   =   .TRUE.
        CHARACTER(LEN=6)                ::  SIZE_VAL_C
        ! Start work

        ! "Unpack" the filament object
        CALL C_F_POINTER(PARAMS,SELF)

        ! Unpack the values to try
        CALL FGSL_OBJ_C_PTR(VAL_VEC,VALUES)
        STATUS = FGSL_VECTOR_ALIGN(VAL_FGSL,VAL_VEC)
        IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('FGSL_VECTOR_ALIGN failed')
        ALLOCATE(VAL(SIZE(VAL_FGSL)),STAT=STATUS)
        IF (STATUS .NE. 0) CALL TERMINATE('Memory allocation failed')

        ! Let the user know what values we're trying
        IF (DEBUG) THEN
            WRITE(SIZE_VAL_C,'(I6)') SIZE(VAL)
            WRITE(*,'(A,'//SIZE_VAL_C//'(F0.3,X))') '**debug(FILAMENT_EVAL_TOTAL_CURVATURE_WRAP): trying these values: ', VAL
        ENDIF

        ! Assign those values to the waypoints
        CALL VALUES_TO_FILAMENT(SELF,VAL,SELF%REFINE_PARAMS)

        ! Update the filament model
        CALL FILAMENT_UPDATE_MODEL(SELF)

        ! Compute the total curvature
        FILAMENT_EVAL_TOTAL_CURVATURE_WRAP = FILAMENT_EVAL_TOTAL_CURVATURE(SELF)

        IF (DEBUG) THEN
            WRITE(*,'(A,(G15.6))') '**debug(FILAMENT_EVAL_TOTAL_CURVATURE_WRAP): got this score: ', &
                                    FILAMENT_EVAL_TOTAL_CURVATURE_WRAP
        ENDIF
    END FUNCTION FILAMENT_EVAL_TOTAL_CURVATURE_WRAP

END MODULE FILAMENTS_MINIMISATION
