!>  \brief  Methods dealing with the internal representation of the physical model for a filament
MODULE FILAMENTS_MODEL_BASE
    USE FILAMENTS_BASE
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE

    CONTAINS



    !>  \brief  Print out extra information about the filament
    SUBROUTINE FILAMENT_PRINT_EXTRA_INFO(SELF)
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_TOTAL_CURVATURE, FILAMENT_EVAL_CHANGE_OF_DIRECTION
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_TOTAL_CURVATURE_OVERTWIST_STRETCH, FILAMENT_DEVIATIONS_FROM_CANONICAL_HELIX
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_IDEAL_PATH_LENGTHS_FOR_CANONICAL_HELIX, FILAMENT_DEVIATION_FROM_IDEAL_PATH_LENGTHS
        USE FILAMENTS_PARAMETRISATION, ONLY : FILAMENT_PARAMETRISATION_ERROR
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_MAXIMAL_NORMAL_DISTANCE, FILAMENT_CHORD_LENGTH, FILAMENT_MID_POINT_NORMAL_DISTANCE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)  ::  SELF
        ! Private variables
        REAL             ::  HELICAL_STATS(4), PARAM_ERROR, RIGID_T, XOVER_STATS(2)
        REAL             ::  MAX_NORMAL_DIST, CHORD_LENGTH_3D,CHORD_LENGTH_2D
        REAL             ::  MID_POINT_NORMAL_DIST_3D, MID_POINT_NORMAL_DIST_2D
        REAL(KIND=8)     ::  CURVATURE, OVERTWIST,STRETCH
        CHARACTER(LEN=2) ::  NWPC
        ! Start work
        WRITE(NWPC,'(I2)') SELF%NWP
        HELICAL_STATS = FILAMENT_EVAL_HELICAL_PARAMETERS_STATS(SELF)
        PARAM_ERROR = FILAMENT_PARAMETRISATION_ERROR(SELF)
        RIGID_T = FILAMENT_GET_T_FOR_RIGID_WP(SELF)
        WRITE(*,'(/A,I0,A)')                 'Extra info for filament ', SELF%ID, ':'
        WRITE(*,'(A,F0.6,1X,F0.3,1X,F0.3)')  '  parametrisation error, t(1), t(rigid): ', PARAM_ERROR, SELF%T(1), RIGID_T
        WRITE(*,'(A,G15.4)')                 '  total curvature (1/pixels) = ', FILAMENT_EVAL_TOTAL_CURVATURE(SELF)
        WRITE(*,'(A,G15.4)')                 '  total change of direction (rad) = ', FILAMENT_EVAL_CHANGE_OF_DIRECTION(SELF)
        WRITE(*,'(2(A,G15.4),A)')            '  average (stdev) twist per subunit (degrees): ', HELICAL_STATS(1), ' (', &
                                                                                                HELICAL_STATS(2), ')'
        WRITE(*,'(2(A,G15.4),A)')            '  average (stdev) rise per subunit (pixels): ', HELICAL_STATS(3), ' (',    &
                                                                                            HELICAL_STATS(4), ')'
        WRITE(*,'(A,'//NWPC//'(F0.3,1X))')   '  ideal path lengths for canonical helix: ', &
                                                            FILAMENT_IDEAL_PATH_LENGTHS_FOR_CANONICAL_HELIX(SELF)
        WRITE(*,'(A,F0.3)')                  '  deviation from ideal path lenghts for canonical helix: ', &
                                                            FILAMENT_DEVIATION_FROM_IDEAL_PATH_LENGTHS(SELF)
        IF (SELF%AMYLOID_BETA) THEN
            XOVER_STATS = FILAMENT_CROSSOVER_STATS(SELF)
            WRITE(*,'(2(A,G15.4),A)')        '  average (stdev) crossover-to-crossover distance (pixels): ', XOVER_STATS(1), &
                                                                                                    ' (', XOVER_STATS(2), ')'
        ENDIF
        CALL FILAMENT_EVAL_TOTAL_CURVATURE_OVERTWIST_STRETCH(SELF,CURVATURE,OVERTWIST,STRETCH,  &
                                                     IGNORE_FAL_WPS=.NOT. (SELF%FAL_WPS_IN_SCORE .AND. SELF%FAL_WPS_IN_THREED))
        WRITE(*,'(A,3(G15.4,1X))')            '  total curvature, overtwist, stretch (1/pixels) : ', CURVATURE, OVERTWIST, STRETCH
        CALL FILAMENT_MAXIMAL_NORMAL_DISTANCE(SELF,MAX_NORMAL_DIST)
        CALL FILAMENT_MID_POINT_NORMAL_DISTANCE(SELF,MID_POINT_NORMAL_DIST_3D,MID_POINT_NORMAL_DIST_2D)
        CALL FILAMENT_CHORD_LENGTH(SELF,CHORD_LENGTH_3D,CHORD_LENGTH_2D)
        WRITE(*,'(A,2(F0.3,1X))')             '  chord length 3D, 2D (pixels): ', CHORD_LENGTH_3D, CHORD_LENGTH_2D
        WRITE(*,'(A,2(F0.3,1X))')             '  mid-point normal distance 3D, 2D (pixels): ',  MID_POINT_NORMAL_DIST_3D, &
                                                                                                MID_POINT_NORMAL_DIST_2D
        WRITE(*,'(A,F0.3)')                   '  maximal normal distance (pixels): ', MAX_NORMAL_DIST
        CALL FILAMENT_DEVIATIONS_FROM_CANONICAL_HELIX(SELF)
    END SUBROUTINE FILAMENT_PRINT_EXTRA_INFO

    !>  \brief  Evaluate the average helical twist and rise per subunit and the standard deviation
    FUNCTION FILAMENT_EVAL_HELICAL_PARAMETERS_STATS(SELF,IGNORE_FAL_WPS) RESULT(STATS)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, SAFE_FOR_SQRT
        USE SPLINES, ONLY : INTERPOL_SPLINE_FIND_ROOT, INTERPOL_SPLINE_EVAL
        USE COOS, ONLY : WAYPOINT
        USE FILAMENTS_PATH, ONLY : FILAMENT_GET_PATH_SUBUNIT_STEPS
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  IGNORE_FAL_WPS
        ! Result
        REAL(KIND=8)    ::  STATS(4)    !<  The average and standard deviation are returned first and second respectively, with 1&2 for the twist and 3&4 for the rise
        ! Private variables
        !REAL(KIND=8)    ::  CURRENT_S   ! subunit number we're on
        !REAL(KIND=8)    ::  CURRENT_T   ! parameter value
        !REAL(KIND=8)    ::  FINISH_S
        REAL(KIND=8)    ::  TWIST_SUM, TWIST_SUM2, RISE_SUM, RISE_SUM2
        !REAL(KIND=8)    ::  CURRENT_TWIST, CURRENT_RISE
        !INTEGER         ::  S_COUNT
        !LOGICAL         ::  IIGNORE_FAL_WPS
        TYPE(WAYPOINT), ALLOCATABLE ::  S_PATH(:)
        INTEGER         ::  I
        ! Start work
        CALL FILAMENT_GET_PATH_SUBUNIT_STEPS(SELF,S_PATH,STEP=1,IGNORE_FAL_WPS=IGNORE_FAL_WPS)

		TWIST_SUM = 0.0D0
		TWIST_SUM2 = 0.0D0
		RISE_SUM = 0.0D0
		RISE_SUM2 = 0.0D0

        DO I=1,SIZE(S_PATH)
            ! Keep running sum and sum of squares
            TWIST_SUM = TWIST_SUM + S_PATH(I)%HELIX%TWIST
            TWIST_SUM2 = TWIST_SUM2 + S_PATH(I)%HELIX%TWIST**2
            RISE_SUM = RISE_SUM + S_PATH(I)%HELIX%RISE
            RISE_SUM2 = RISE_SUM2 + S_PATH(I)%HELIX%RISE**2
        ENDDO

        ! Averages
        STATS(1) = TWIST_SUM / DBLE(SIZE(S_PATH))
        TWIST_SUM2 = TWIST_SUM2 / DBLE(SIZE(S_PATH))
        STATS(3) = RISE_SUM / DBLE(SIZE(S_PATH))
        RISE_SUM2 = RISE_SUM2 / DBLE(SIZE(S_PATH))
        ! Variance
        STATS(2) = TWIST_SUM2 - STATS(1)**2
        STATS(4) = RISE_SUM2 - STATS(3)**2
        ! Standard deviation
        CALL SAFE_FOR_SQRT(STATS(2),NEGATIVE_TO_ZERO=.TRUE.)
        CALL SAFE_FOR_SQRT(STATS(4),NEGATIVE_TO_ZERO=.TRUE.)
        STATS(2) = SQRT(STATS(2))
        STATS(4) = SQRT(STATS(4))

    END FUNCTION FILAMENT_EVAL_HELICAL_PARAMETERS_STATS


    !>  \brief  Compute the average and standard deviation of crossover-to-crossover distances
    FUNCTION FILAMENT_CROSSOVER_STATS(SELF) RESULT(STATS)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_PATH, ONLY : FILAMENT_GET_PATH_CROSSOVERS
        USE COOS, ONLY : WAYPOINTS_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(INOUT)   ::  SELF    !<  Filament object
        ! Result
        REAL    ::  STATS(2)
        ! Private variables
        REAL, ALLOCATABLE                   ::  DISTANCES(:)
        REAL(KIND=8)                        ::  SUM_DISTANCES, SUM_SQ_DISTANCES
        TYPE(WAYPOINT),     ALLOCATABLE     ::  XOVERS(:)
        INTEGER                             ::  CURRENT_XOVER
        INTEGER                             ::  IERR
        LOGICAL,            PARAMETER       ::  DEBUG = .FALSE.
        ! Start work

        ! Get the crossovers
        IF (SELF%WPS_ARE_CROSSOVERS) THEN
            IF (SELF%FAL_WPS_ARE_CROSSOVERS) THEN
                XOVERS = SELF%WP
            ELSE
                XOVERS = SELF%WP(2:SELF%NWP-1)
            ENDIF
        ELSE
            CALL FILAMENT_GET_PATH_CROSSOVERS(SELF,XOVERS,STEP=1)
        ENDIF

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_CROSSOVER_STATS): xover info:'
            CALL WAYPOINTS_PRINT_INFO(XOVERS)
        ENDIF

        ! Allocate memory for xover-to-xover distances
        IF (ALLOCATED(DISTANCES)) DEALLOCATE(DISTANCES)
        ALLOCATE(DISTANCES(SIZE(XOVERS)-1),STAT=IERR)
        IF (IERR .NE. 0) CALL TERMINATE('Failed memory allocation in FILAMENT_CROSSOVER_STATS')

        ! Compute distances
        SUM_DISTANCES = 0.0D0
        SUM_SQ_DISTANCES = 0.0D0
        DO CURRENT_XOVER=1,SIZE(XOVERS)-1
            DISTANCES(CURRENT_XOVER)= XOVERS(CURRENT_XOVER+1)%PATH_LENGTH - XOVERS(CURRENT_XOVER)%PATH_LENGTH
            SUM_DISTANCES = SUM_DISTANCES + DISTANCES(CURRENT_XOVER)
            SUM_SQ_DISTANCES = SUM_SQ_DISTANCES + DISTANCES(CURRENT_XOVER)**2
        ENDDO

        ! Average
        IF (SIZE(DISTANCES) .NE. 0) THEN
            STATS(1) = SUM_DISTANCES / SIZE(DISTANCES)
            STATS(2) = SUM_SQ_DISTANCES / SIZE(DISTANCES) - STATS(1)**2
            IF (STATS(2) .GE. 0.0D0) THEN
                STATS(2) = SQRT( STATS(2) )
            ELSE
                STATS(2) = 0.0D0
            ENDIF
        ELSE
            WRITE(*,'(A,I0)') '**WARNING(FILAMENT_CROSSOVER_STATS): not enough crossovers. Number of intervals = ', SIZE(DISTANCES)
            !CALL TERMINATE('Fatal error in FILAMENT_CROSSOVER_STATS')
        ENDIF
    END FUNCTION FILAMENT_CROSSOVER_STATS






    !>  \brief  When the helical operator describes an n-start helix with m-fold symmetry,
    !           the number of subunits should be a multiple of n*m. This subroutine checks this is the case.
    SUBROUTINE FILAMENT_CHECK_SUBUNIT_NUM(SELF)
        USE HELICES, ONLY : HELICAL_SYMMETRY_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(IN)  ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_WP
        REAL    ::  ERROR_RELATIVE_TO_PATH_LENGTH
        REAL    ::  ERROR_RELATIVE_TO_PHI
        REAL    ::  AVERAGE_ERROR
        REAL    ::  EXPECTED_FROM_PATH_LENGTH
        REAL    ::  EXPECTED_FROM_PHI
        REAL    ::  SUBUNIT_NUM_FROM_START
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        ! Loop over the waypoints
        ERROR_RELATIVE_TO_PATH_LENGTH   =   0.0
        ERROR_RELATIVE_TO_PHI           =   0.0
        DO CURRENT_WP=2,SELF%NWP
            ! Number of subunits from 1st to current waypoint
            SUBUNIT_NUM_FROM_START = SELF%WP(CURRENT_WP)%SUBUNIT_NUM - SELF%WP(1)%SUBUNIT_NUM
            ! According to path_length, expected number of subunits (i.e. assuming constant rise per subunit)
            EXPECTED_FROM_PATH_LENGTH =     WP_SUBUNIT_NUM_FROM_PATH_LENGTH(SELF%WP(CURRENT_WP),SELF%HELIX) &
                                        -   WP_SUBUNIT_NUM_FROM_PATH_LENGTH(SELF%WP(1),SELF%HELIX)
            ! According to phi, expected number of subunits (i.e. assuming constant twist per subunit)
            IF (SELF%HELIX%TWIST .EQ. 0.0) THEN
                ! If the twist is 0.0 (can happen, for example when talking about pf1 of a 13-pf MT), we
                ! just can't estimate subunit_num
                EXPECTED_FROM_PHI = SUBUNIT_NUM_FROM_START
            ELSE
                EXPECTED_FROM_PHI =     WP_SUBUNIT_NUM_FROM_PHI(SELF%WP(CURRENT_WP),SELF%HELIX) &
                                    -   WP_SUBUNIT_NUM_FROM_PHI(SELF%WP(1),SELF%HELIX)
            ENDIF
            ! Debug output
            IF (DEBUG) THEN
                WRITE(*,'(A,I0,A,I0,A,3(F0.3,1X))')  '**debug(FILAMENT_CHECK_SUBUNIT_NUM): waypoint ', &
                                                    CURRENT_WP, ' of ', SELF%NWP, &
                                                    ': subunit_num, expected from path length, expected from phi: ', &
                                                    SUBUNIT_NUM_FROM_START, EXPECTED_FROM_PATH_LENGTH,  EXPECTED_FROM_PHI
            ENDIF
            ! Accumulate average error
            ERROR_RELATIVE_TO_PATH_LENGTH   =   ERROR_RELATIVE_TO_PATH_LENGTH + EXPECTED_FROM_PATH_LENGTH / SUBUNIT_NUM_FROM_START
            ERROR_RELATIVE_TO_PHI           =   ERROR_RELATIVE_TO_PHI         + EXPECTED_FROM_PHI / SUBUNIT_NUM_FROM_START
        ENDDO
        ! Average -fold error
        ERROR_RELATIVE_TO_PATH_LENGTH   =   ERROR_RELATIVE_TO_PATH_LENGTH / (SELF%NWP-1) - 1.
        ERROR_RELATIVE_TO_PHI           =   ERROR_RELATIVE_TO_PHI         / (SELF%NWP-1) - 1.

        ! Average of the two average errors
        AVERAGE_ERROR = (ERROR_RELATIVE_TO_PATH_LENGTH + ERROR_RELATIVE_TO_PHI) / 2.

        IF (DEBUG) WRITE(*,'(A,F0.2,A)') '**debug(FILAMENT_CHECK_SUBUNIT_NUM): average error = ', AVERAGE_ERROR, '-fold'

        IF (NINT(AVERAGE_ERROR) .NE. 0) THEN
            WRITE(*,'(A,F0.2,A)') '**info(FILAMENT_CHECK_SUBUNIT_NUM): it looks like the subunit_num values are ',  &
                                             AVERAGE_ERROR, '-fold out compared to expected values based on helical operator'
            WRITE(*,'(A)') '**info(FILAMENT_CHECK_SUBUNIT_NUM): perhaps the number of starts and/or symmetries in ',    &
                                        'the helical operator were not taken into account when estimating subunit_num values'
            CALL HELICAL_SYMMETRY_PRINT_INFO(SELF%HELIX)
            CALL FILAMENT_PRINT_INFO(SELF)
            !CALL TERMINATE('Testing')
        ENDIF


    END SUBROUTINE FILAMENT_CHECK_SUBUNIT_NUM

    !>  \brief  The subunit_num should increase from waypoint to waypoint, if not crash out
    SUBROUTINE FILAMENT_CHECK_SUBUNIT_NUM_INCREASES(SELF)
        IMPLICIT NONE
        TYPE(FILAMENT), INTENT(IN)  ::  SELF    !<  Filament
        ! Private variables
        INTEGER ::  CURRENT_WP
        ! Start work
        DO CURRENT_WP = 2, SELF%NWP
            IF ( SELF%WP(CURRENT_WP)%SUBUNIT_NUM .LT. SELF%WP(CURRENT_WP-1)%SUBUNIT_NUM) THEN
                WRITE(*, '(A)') '**ERROR(FILAMENT_CHECK_SUBUNIT_NUM_INCREASES): subunit_num values should increase'
                CALL FILAMENT_PRINT_INFO(SELF)
                CALL TERMINATE('Fatal error in FILAMENT_CHECK_SUBUNIT_NUM_INCREASES')
            ENDIF
        END DO
    END SUBROUTINE FILAMENT_CHECK_SUBUNIT_NUM_INCREASES


    !>  \brief  Based on the waypoint's phi, return the subunit_num predicted by the helical parameters, assuming
    !!          constant twist per subunit.
    !!
    !!  Note that by convention, waypoints' "subunit_num" value is the number of subunits along the 1-start helix, without
    !!  counting symmetry mates.
    !!
    !!  The helical parameters given could either be the canonical or local ones
    PURE REAL FUNCTION WP_SUBUNIT_NUM_FROM_PHI(SELF,HELIX) RESULT(SUBUNIT_NUM)
        IMPLICIT NONE
        ! Arguments
        TYPE(WAYPOINT),         INTENT(IN)  ::  SELF
        TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  HELIX
        ! Private variables

        ! Start work
        SUBUNIT_NUM = SELF%EULER%PHI / HELIX%TWIST
    END FUNCTION WP_SUBUNIT_NUM_FROM_PHI

    !>  \brief  Based on the waypoint's path length, return the subunit_num predicted by the helical parameters,
    !!          assuming constant rise per subunit.
    !!
    !!  The helical parameters given could either be the canonical or local ones
    PURE REAL FUNCTION WP_SUBUNIT_NUM_FROM_PATH_LENGTH(SELF,HELIX) RESULT(SUBUNIT_NUM)
        IMPLICIT NONE
        ! Arguments
        TYPE(WAYPOINT),         INTENT(IN)  ::  SELF
        TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  HELIX
        ! Private variables

        ! Start work
        SUBUNIT_NUM = SELF%PATH_LENGTH / HELIX%RISE
    END FUNCTION WP_SUBUNIT_NUM_FROM_PATH_LENGTH

    !>  \brief  Compute the subunit_num shift to the nearest integer subunit number for the given waypoint.
    !!
    !!  By convention, the reference 3D model has the origin subunit at phi=0.0, so that s=0.0. Any rotation around the helical axis (phi angle)
    !!  still leaves a subunit exactly at the origin of a projection image. Therefore, if a waypoint as a non-integer subunit_num value, the
    !!  reference volume will need to be "phase shifted" to match it, i.e. a screw operation will need to be applied to the reference.
    !!  To work out the smallest possible screw operation that would need to be performed, we first need to work out the shift in subunit_num
    !!  value to the nereast integer subunit.
    PURE REAL FUNCTION WP_COMPUTE_SUBUNIT_NUM_PHASING(WP) RESULT(SUBUNIT_NUM_SHIFT)
        IMPLICIT NONE
        ! Arguments
        TYPE(WAYPOINT),                 INTENT(IN)      ::  WP
        ! Private variables
        ! Start work
        SUBUNIT_NUM_SHIFT = WP%SUBUNIT_NUM - NINT(WP%SUBUNIT_NUM)
    END FUNCTION WP_COMPUTE_SUBUNIT_NUM_PHASING



    !>  \brief  Return the T parameter at which we conceptually place the FIL_AS_SP_WP
    PURE FUNCTION FILAMENT_GET_T_FOR_RIGID_WP(SELF) RESULT(T)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(IN)  ::  SELF
        ! Result
        REAL(KIND=8)                            ::  T
        ! Private variables

        ! Start work

        ! We chose to "position" the virtual rigid waypoint at the midpoint of the space curve
        T = SELF%T(1) + (SELF%T(SELF%NWP) - SELF%T(1)) / 2.
    END FUNCTION FILAMENT_GET_T_FOR_RIGID_WP

    !>  \brief  Estimate the Phi angles of waypoints in the absence of a model, but given helical parameters and T parameters for each waypoints
    !!
    SUBROUTINE FILAMENT_GUESS_WPS_PHI(SELF)
        USE HELICES, ONLY : HELICAL_SYMMETRY_UNIT_CONVERT
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(INOUT)       ::  SELF    !<  Filament whose waypoints are to be updated with new phi values
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        INTEGER                 ::  TWIST_SIGN, I
        ! Start work

        ! Check helical parameters are in pixel units
        IF (.NOT. SELF%HELIX%PIXEL_UNITS) THEN
            !WRITE(*,'(A)') '**ERROR(FILAMENT_ESTIMATE_WPS_PHI): filament helical parameters not given in pixel units!'
            !CALL TERMINATE('Fatal error in FILAMENT_GUESS_WPS_PHI')
            CALL HELICAL_SYMMETRY_UNIT_CONVERT(SELF%HELIX,SELF%FILM%PIXEL_SIZE)
        ENDIF

        ! Check waypoints have T values
        IF (ALL(SELF%T .EQ. 0.0)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_ESTIMATE_WPS_PHI): filament was not parametrised - all T values are 0'
            CALL TERMINATE('Fatal error in FILAMENT_GUESS_WPS_PHI')
        ENDIF

        !
        ! What is the sign of the twist?
        !
        IF (MOD(SELF%HELIX%TWIST,180.D0) .GT. 0.D0) THEN
            TWIST_SIGN = 1
        ELSE
            TWIST_SIGN = -1
        ENDIF
        IF (DEBUG) WRITE(*,'(A,F8.2,1X,I2)') '**debug(FILAMENT_GUESS_WPS_PHI): twist, twist_sign = ', SELF%HELIX%TWIST, TWIST_SIGN

        IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_GUESS_WPS_PHI): guessing phi values for waypoints'
        IF (SELF%WPS_ARE_CROSSOVERS) THEN
            IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_GUESS_WPS_PHI): waypoints are crossovers'
            IF (SELF%FAL_WPS_ARE_CROSSOVERS) THEN
                SELF%WP(1)%EULER%PHI = 0.
                DO I=2,SELF%NWP
                    SELF%WP(I)%EULER%PHI = SELF%WP(I-1)%EULER%PHI + 180.0 * TWIST_SIGN
                ENDDO
            ELSE
                IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_GUESS_WPS_PHI): first and last waypoints are not crossovers'
                SELF%WP(2)%EULER%PHI = 0.
                DO I=3,SELF%NWP-1
                    SELF%WP(I)%EULER%PHI = SELF%WP(I-1)%EULER%PHI + 180.0 * TWIST_SIGN
                ENDDO
                CALL FILAMENT_GUESS_FAL_WPS(SELF,GUESS_FAL_EULER_ONLY=.TRUE.)!,KNOWN_PSI=SELF%X_SPLINE%KNOWN_D .AND. SELF%Y_SPLINE%KNOWN_D,KNOWN_THETA=SELF%Z_SPLINE%KNOWN_D)
            ENDIF
        ELSE
            SELF%WP%EULER%PHI   =   SELF%T / SELF%HELIX%RISE * SELF%HELIX%TWIST
        ENDIF

    END SUBROUTINE FILAMENT_GUESS_WPS_PHI


    !>  \brief  Find the integer subunit_num value nearest to the given waypoint, making sure that subunit is within the filament
    INTEGER FUNCTION FILAMENT_FIND_NEAREST_SUBUNIT_TO_WAYPOINT(SELF,WP) RESULT(NEAREST_SUBUNIT_NUM)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                 INTENT(IN)      ::  SELF
        TYPE(WAYPOINT),                 INTENT(IN)      ::  WP
        ! Private variables

        ! Start work
        NEAREST_SUBUNIT_NUM = NINT(WP%SUBUNIT_NUM)
        NEAREST_SUBUNIT_NUM = MAX(NEAREST_SUBUNIT_NUM,CEILING(SELF%WP(1)%SUBUNIT_NUM))
        NEAREST_SUBUNIT_NUM = MIN(NEAREST_SUBUNIT_NUM,FLOOR(SELF%WP(SELF%NWP)%SUBUNIT_NUM))
    END FUNCTION FILAMENT_FIND_NEAREST_SUBUNIT_TO_WAYPOINT

    !>  \brief  For each waypoint in a filament, "guess"/estimate the subunit number
    !!
    !!  The KEEP_RIGID_SUBUNIT_NUM option allows the subunit number values of each waypoints to be recalculated assuming constant rise per subunit while
    !!  (for example) refining the rigid subunit_num value.
    SUBROUTINE FILAMENT_GUESS_WPS_SUBUNIT_NUM(  SELF,FROM_RIGID_SUBUNIT_NUM,  &
                                                CONSTANT_RISE_PER_SUBUNIT,CONSTANT_TWIST_PER_SUBUNIT,INTG_SUBUNIT_NUM   )
        USE HELICES, ONLY : HELICAL_SYMMETRY_UNIT_CONVERT
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                 INTENT(INOUT)   ::  SELF                        !<  Filament whose waypoints are to be updated with new SUBUNIT_NUM values
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  FROM_RIGID_SUBUNIT_NUM      !<  If true, the subunit_num values will be recalculated for each waypoint while keeping the rigid waypoint's subunit_num value unchanged. Defaults to true.
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  CONSTANT_RISE_PER_SUBUNIT   !<  If true, we assume the rise per subunit is constant. Defaults to true.
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  CONSTANT_TWIST_PER_SUBUNIT  !<  If true, we assume the twist per subunit is constant. Defaults to ???.
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  INTG_SUBUNIT_NUM            !<  If true, the subunit_num value for each waypoint will necessarily be integer
        ! Private variables
        LOGICAL                 ::  FFROM_RIGID_SUBUNIT_NUM
        LOGICAL                 ::  CCONSTANT_RISE_PER_SUBUNIT
        LOGICAL                 ::  CCONSTANT_TWIST_PER_SUBUNIT
        LOGICAL                 ::  IINTG_SUBUNIT_NUM
        INTEGER                 ::  CURRENT_WP
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        !REAL                    ::  DIST_TO_NEXT_WP, EXPECTED_CHANGE_IN_PHI, ACTUAL_CHANGE_IN_PHI
        ! Start work

        ! Will we be leaving the rigid subunit_num value unchanged?
        IF (PRESENT(FROM_RIGID_SUBUNIT_NUM)) THEN
            FFROM_RIGID_SUBUNIT_NUM = FROM_RIGID_SUBUNIT_NUM
        ELSE
            FFROM_RIGID_SUBUNIT_NUM = .TRUE.
        ENDIF

        ! Constant rise per subunit?
        IF (PRESENT(CONSTANT_RISE_PER_SUBUNIT)) THEN
            CCONSTANT_RISE_PER_SUBUNIT = CONSTANT_RISE_PER_SUBUNIT
        ELSE
            CCONSTANT_RISE_PER_SUBUNIT = SELF%CONSTANT_RISE_PER_SUBUNIT
        ENDIF

        ! Constant twist per subunit
        IF (PRESENT(CONSTANT_TWIST_PER_SUBUNIT)) THEN
            CCONSTANT_TWIST_PER_SUBUNIT = CONSTANT_TWIST_PER_SUBUNIT
        ELSE
            CCONSTANT_TWIST_PER_SUBUNIT = SELF%CONSTANT_RISE_PER_SUBUNIT .AND. SELF%CONSTANT_TWIST_PER_LENGTH
        ENDIF

        ! Will we assign integer values for subunit_num?
        ! This can be useful for example when assigning subunit_num values to waypoints which were previously refined as single particles
        IF (PRESENT(INTG_SUBUNIT_NUM)) THEN
            IINTG_SUBUNIT_NUM = INTG_SUBUNIT_NUM
        ELSE
            IINTG_SUBUNIT_NUM = .FALSE.
        ENDIF


        ! It is not possible to garantee constant rise per subunit and integer subunit_num values together, since this subroutine cannot move waypoints
        IF (CCONSTANT_RISE_PER_SUBUNIT .AND.  IINTG_SUBUNIT_NUM) THEN
            WRITE(*,'(3A)') '**warning(FILAMENT_GUESS_WPS_SUBUNIT_NUM): it is not possible to find integer subunit_num', &
                            ' values while forcing constant rise per subunit. Integer subunit_num values will be',  &
                            ' returned while keeping as close as possible to a constant rise per subunit.'
        ENDIF

        ! Check the helix is in pixel units
        IF (.NOT. SELF%HELIX%PIXEL_UNITS) CALL HELICAL_SYMMETRY_UNIT_CONVERT(SELF%HELIX,SELF%FILM%PIXEL_SIZE)


        ! Start the actual work

        IF (CCONSTANT_TWIST_PER_SUBUNIT .AND. SELF%HELIX%TWIST .NE. 0.) THEN
            IF (DEBUG) WRITE(*,'(2A)') '**debug(FILAMENT_GUESS_WPS_SUBUNIT_NUM): calculating subunit_num values for all',   &
                                        ' waypoints assuming constant twist per subunit'
            DO CURRENT_WP=1,SELF%NWP
                SELF%WP(CURRENT_WP)%SUBUNIT_NUM = WP_SUBUNIT_NUM_FROM_PHI(SELF%WP(CURRENT_WP),SELF%HELIX)
            ENDDO
        ELSE IF (CCONSTANT_RISE_PER_SUBUNIT .OR. ALL(SELF%WP%SUBUNIT_NUM .EQ. 0.0)) THEN
            IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_GUESS_WPS_SUBUNIT_NUM): calculating subunit_num values for all',    &
                                    ' waypoints assuming constant rise per subunit'
            DO CURRENT_WP=1,SELF%NWP
                SELF%WP(CURRENT_WP)%SUBUNIT_NUM = WP_SUBUNIT_NUM_FROM_PATH_LENGTH(SELF%WP(CURRENT_WP),SELF%HELIX)
            ENDDO
        ELSE
            !WRITE(*,'(A)') '**ERROR(FILAMENT_UPDATE_WPS_SUBUNIT_NUM): neither constant rise nor constant twist per subunit was specified'
            !CALL TERMINATE('Fatal error in FILAMENT_UPDATE_WPS_SUBUNIT_NUM')
            !DO CURRENT_WP=1,SELF%NWP-1
            !    ! Work out the distance to the next waypoint
            !    DIST_TO_NEXT_WP = SELF%WP(CURRENT_WP+1)%PATH_LENGTH - SELF%WP(CURRENT_WP)%PATH_LENGTH
            !    ! Work out what change in phi this should correspond to
            !    EXPECTED_CHANGE_IN_PHI = DIST_TO_NEXT_WP * SELF%HELIX%TWIST / SELF%HELIX%RISE
            !    ! What change in phi do we actually have?
            !    ACTUAL_CHANGE_IN_PHI = SELF%WP(CURRENT_WP+1)%EULER%PHI - SELF%WP(CURRENT_WP)%EULER%PHI
            !
            !    !
            !    SELF%WP(CURRENT_WP+1:SELF%NWP)%SUBUNIT_NUM =  -NINT((ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI)/SELF%HELIX%TWIST) + &
            !                                        SELF%WP(CURRENT_WP)%SUBUNIT_NUM
            !    PRINT *, '**DEBUG(): change in subunit_num = ', -NINT((ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI)/SELF%HELIX%TWIST)
            !ENDDO
        ENDIF

        IF (IINTG_SUBUNIT_NUM) THEN
            IF (FFROM_RIGID_SUBUNIT_NUM) THEN
                WRITE(*,'(A)') '**warning(FILAMENT_GUESS_WPS_SUBUNIT_NUM): rouding of all subunit_num values will mean',    &
                                ' that the rigid subunit_num value is not necessarily preserved'
            ENDIF
            IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_GUESS_WPS_SUBUNIT_NUM): rounding all subunit_num values to nearest integer'
            SELF%WP%SUBUNIT_NUM = ANINT(SELF%WP%SUBUNIT_NUM)
            ! Check against two subsequent equal values
            DO CURRENT_WP=1,SELF%NWP-1
                IF (SELF%WP(CURRENT_WP)%SUBUNIT_NUM .EQ. SELF%WP(CURRENT_WP+1)%SUBUNIT_NUM) THEN
                    ! See if we can move it one back
                    IF (CURRENT_WP .EQ. 1) THEN
                        SELF%WP(CURRENT_WP)%SUBUNIT_NUM = SELF%WP(CURRENT_WP)%SUBUNIT_NUM - 1.0
                    ELSE IF (SELF%WP(CURRENT_WP-1)%SUBUNIT_NUM .LT. SELF%WP(CURRENT_WP)%SUBUNIT_NUM - 1.5) THEN
                        SELF%WP(CURRENT_WP)%SUBUNIT_NUM = SELF%WP(CURRENT_WP)%SUBUNIT_NUM - 1.0
                    ! See if we can move the next one forward
                    ELSE IF (CURRENT_WP .EQ. SELF%NWP - 1) THEN
                        SELF%WP(SELF%NWP)%SUBUNIT_NUM = SELF%WP(SELF%NWP)%SUBUNIT_NUM + 1.0
                    ELSE IF (SELF%WP(CURRENT_WP+2)%SUBUNIT_NUM .GT. SELF%WP(CURRENT_WP+1)%SUBUNIT_NUM + 1.5) THEN
                        SELF%WP(CURRENT_WP+1)%SUBUNIT_NUM = SELF%WP(CURRENT_WP+1)%SUBUNIT_NUM + 1
                    ELSE
                        WRITE(*,'(A,I0,A,I0,A,F0.3)') '**warning(FILAMENT_GUESS_WPS_SUBUNIT_NUM): waypoints ',     &
                                                        CURRENT_WP, ' and ', CURRENT_WP+1, &
                                                        ' have the same subunit_num value: ', SELF%WP(CURRENT_WP)%SUBUNIT_NUM
                    ENDIF
                ENDIF
            ENDDO
        ENDIF

        IF (FFROM_RIGID_SUBUNIT_NUM) THEN
            IF (DEBUG) WRITE(*,'(3A,F0.3,A)') '**debug(FILAMENT_GUESS_WPS_SUBUNIT_NUM): recalculating subunit_num values',   &
                                                ' for all waypoints to take into account subunit_num value',    &
                                                ' of rigid waypoint (', SELF%RIGID_WP%SUBUNIT_NUM, ')'
            SELF%WP%SUBUNIT_NUM = SELF%WP%SUBUNIT_NUM - (SUM(SELF%WP%SUBUNIT_NUM) / REAL(SELF%NWP)) + SELF%RIGID_WP%SUBUNIT_NUM
        ENDIF

    END SUBROUTINE FILAMENT_GUESS_WPS_SUBUNIT_NUM





    !>  \brief  Guess the coordinates and/or angles of the first and last waypoints of a filament.
    SUBROUTINE FILAMENT_GUESS_FAL_WPS(SELF,GUESS_FAL_EULER_ONLY)
        USE SPLINES
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_SOLVE_SPLINES
        USE DIFFERENTIAL_GEOMETRY, ONLY : EULER_FROM_DERIVATIVES
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF                    !<  Filament object
        LOGICAL,        INTENT(IN)      ::  GUESS_FAL_EULER_ONLY    !<  The coordinates for the FAL WPs are already known, we only need to guess their Euler angles. By default .FALSE.
        ! Private variables
        TYPE(INTERPOL_SPLINE)   ::  TMP_X_SPLINE, TMP_Y_SPLINE, TMP_Z_SPLINE, TMP_S_SPLINE, TMP_PHI_SPLINE
        INTEGER                 ::  I
        REAL(KIND=8)            ::  THETA_TMP, PSI_TMP
        LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
		LOGICAL						::	HERMITE_BAK
        ! Start work


        ! This will only work with >= 5 waypoints
        IF (SELF%NWP .LT. 5) THEN
           WRITE(*,'(A,I0)') '**ERROR(FILAMENT_GUESS_FAL_WPS): need at least 5 waypoints, but only have ', SELF%NWP
           CALL TERMINATE('Fatal error in FILAMENT_GUESS_FAL_WPS')
        ENDIF

        ! Initialise the "X" values for the splines
        SELF%X_SPLINE%X     = SELF%T
        SELF%Y_SPLINE%X     = SELF%T
        SELF%Z_SPLINE%X     = SELF%T
        SELF%S_SPLINE%X     = SELF%T
        SELF%PHI_SPLINE%X   = SELF%T

        ! We assume that we haven't solved the splines yet and therefore don't have weights to calculate the X,Y,Z values
        ! One solution is to create temporary spline objects without the extra first/last waypoints, solve those,
        ! and then work out the coos for the extra first/last.
        WRITE(*,'(2A)') '**warning(FILAMENT_GUESS_FAL_WPS): options to guess FAL WP parameters are experimental',    &
                            ' - please check results carefully'
        ! Create temporary spline objects
        IF (.NOT. GUESS_FAL_EULER_ONLY) THEN
            CALL ALLOCATE_INTERPOL_SPLINE(TMP_X_SPLINE,   SELF%NWP-2,   SELF%X_SPLINE%K)
            CALL ALLOCATE_INTERPOL_SPLINE(TMP_Y_SPLINE,   SELF%NWP-2,   SELF%Y_SPLINE%K)
            CALL ALLOCATE_INTERPOL_SPLINE(TMP_Z_SPLINE,   SELF%NWP-2,   SELF%Z_SPLINE%K)
            CALL ALLOCATE_INTERPOL_SPLINE(TMP_S_SPLINE,   SELF%NWP-2,   SELF%S_SPLINE%K)
        ENDIF
        CALL ALLOCATE_INTERPOL_SPLINE(TMP_PHI_SPLINE, SELF%NWP-2,   SELF%PHI_SPLINE%K)
        ! Copy contents of filament splines to temporary splines (except of course first and last)
        IF (.NOT. GUESS_FAL_EULER_ONLY) THEN
            TMP_X_SPLINE%X(1:SELF%NWP-2)     = SELF%T(2:SELF%NWP-1)
            TMP_Y_SPLINE%X(1:SELF%NWP-2)     = SELF%T(2:SELF%NWP-1)
            TMP_Z_SPLINE%X(1:SELF%NWP-2)     = SELF%T(2:SELF%NWP-1)
            TMP_S_SPLINE%X(1:SELF%NWP-2)     = SELF%T(2:SELF%NWP-1)
            TMP_X_SPLINE%Y(1:SELF%NWP-2)     = SELF%WP(2:SELF%NWP-1)%COO(1)
            TMP_Y_SPLINE%Y(1:SELF%NWP-2)     = SELF%WP(2:SELF%NWP-1)%COO(2)
            TMP_Z_SPLINE%Y(1:SELF%NWP-2)     = SELF%WP(2:SELF%NWP-1)%COO(3)
            TMP_S_SPLINE%Y(1:SELF%NWP-2)     = SELF%WP(2:SELF%NWP-1)%SUBUNIT_NUM
        ENDIF
        TMP_PHI_SPLINE%X(1:SELF%NWP-2)   = SELF%T(2:SELF%NWP-1)
        TMP_PHI_SPLINE%Y(1:SELF%NWP-2)   = SELF%WP(2:SELF%NWP-1)%EULER%PHI
        ! Set type of temporary splines
        IF (.NOT. GUESS_FAL_EULER_ONLY) THEN
            CALL SPLINE_SET_TYPE(TMP_X_SPLINE,  'PARABOLIC')
            CALL SPLINE_SET_TYPE(TMP_Y_SPLINE,  'PARABOLIC')
            CALL SPLINE_SET_TYPE(TMP_Z_SPLINE,  'PARABOLIC')
            CALL SPLINE_SET_TYPE(TMP_S_SPLINE,  'PARABOLIC')
        ENDIF
        CALL SPLINE_SET_TYPE(TMP_PHI_SPLINE,'PARABOLIC')
        ! Solve temporary splines
        IF (.NOT. GUESS_FAL_EULER_ONLY) THEN
            CALL SOLVE_INTERPOL_SPLINE(TMP_X_SPLINE)
            CALL SOLVE_INTERPOL_SPLINE(TMP_Y_SPLINE)
            CALL SOLVE_INTERPOL_SPLINE(TMP_Z_SPLINE)
            CALL SOLVE_INTERPOL_SPLINE(TMP_S_SPLINE)
        ENDIF
        CALL SOLVE_INTERPOL_SPLINE(TMP_PHI_SPLINE)
        ! Copy the results back to the filament's splines
        !DO I=2,SELF%X_SPLINE%N-1
        DO I=2,SELF%NWP-1
            IF (.NOT. GUESS_FAL_EULER_ONLY) THEN
                ! X spline
                SELF%X_SPLINE%X(I)                          =   TMP_X_SPLINE%X(I-1)
                SELF%X_SPLINE%Y(I)                          =   TMP_X_SPLINE%Y(I-1)
                SELF%X_SPLINE%DY(I)                         =   TMP_X_SPLINE%DY(I-1)
                SELF%X_SPLINE%DDY(I)                        =   TMP_X_SPLINE%DDY(I-1)
                SELF%X_SPLINE%W(I,1:SELF%X_SPLINE%K+1)      =   TMP_X_SPLINE%W(I-1,1:SELF%X_SPLINE%K+1)
                ! Y spline
                SELF%Y_SPLINE%X(I)                          =   TMP_Y_SPLINE%X(I-1)
                SELF%Y_SPLINE%Y(I)                          =   TMP_Y_SPLINE%Y(I-1)
                SELF%Y_SPLINE%DY(I)                         =   TMP_Y_SPLINE%DY(I-1)
                SELF%Y_SPLINE%DDY(I)                        =   TMP_Y_SPLINE%DDY(I-1)
                SELF%Y_SPLINE%W(I,1:SELF%Y_SPLINE%K+1)      =   TMP_Y_SPLINE%W(I-1,1:SELF%Y_SPLINE%K+1)
                ! Z spline
                SELF%Z_SPLINE%X(I)                          =   TMP_Z_SPLINE%X(I-1)
                SELF%Z_SPLINE%Y(I)                          =   TMP_Z_SPLINE%Y(I-1)
                SELF%Z_SPLINE%DY(I)                         =   TMP_Z_SPLINE%DY(I-1)
                SELF%Z_SPLINE%DDY(I)                        =   TMP_Z_SPLINE%DDY(I-1)
                SELF%Z_SPLINE%W(I,1:SELF%Z_SPLINE%K+1)      =   TMP_Z_SPLINE%W(I-1,1:SELF%Z_SPLINE%K+1)
                ! S spline
                SELF%S_SPLINE%X(I)                          =   TMP_S_SPLINE%X(I-1)
                SELF%S_SPLINE%Y(I)                          =   TMP_S_SPLINE%Y(I-1)
                SELF%S_SPLINE%DY(I)                         =   TMP_S_SPLINE%DY(I-1)
                SELF%S_SPLINE%DDY(I)                        =   TMP_S_SPLINE%DDY(I-1)
                SELF%S_SPLINE%W(I,1:SELF%Z_SPLINE%K+1)      =   TMP_S_SPLINE%W(I-1,1:SELF%Z_SPLINE%K+1)
            ENDIF
            ! PHI spline
            SELF%PHI_SPLINE%X(I)                        =   TMP_PHI_SPLINE%X(I-1)
            SELF%PHI_SPLINE%Y(I)                        =   TMP_PHI_SPLINE%Y(I-1)
            SELF%PHI_SPLINE%DY(I)                       =   TMP_PHI_SPLINE%DY(I-1)
            SELF%PHI_SPLINE%DDY(I)                      =   TMP_PHI_SPLINE%DDY(I-1)
            SELF%PHI_SPLINE%W(I,1:SELF%PHI_SPLINE%K+1)  =   TMP_PHI_SPLINE%W(I-1,1:SELF%PHI_SPLINE%K+1)
        ENDDO
        ! Extrapolate FAL waypoint coordinates
        IF (.NOT. GUESS_FAL_EULER_ONLY) THEN
            ! X
            CALL INTERPOL_SPLINE_EVAL(SELF%X_SPLINE,DBLE(SELF%T(1))       ,SELF%WP(1)%COO(1),       FIRST_AND_LAST_UNKNOWN=.TRUE.)
            CALL INTERPOL_SPLINE_EVAL(SELF%X_SPLINE,DBLE(SELF%T(SELF%NWP)),SELF%WP(SELF%NWP)%COO(1),FIRST_AND_LAST_UNKNOWN=.TRUE.)
            ! Y
            CALL INTERPOL_SPLINE_EVAL(SELF%Y_SPLINE,DBLE(SELF%T(1))       ,SELF%WP(1)%COO(2), FIRST_AND_LAST_UNKNOWN=.TRUE.)
            CALL INTERPOL_SPLINE_EVAL(SELF%Y_SPLINE,DBLE(SELF%T(SELF%NWP)),SELF%WP(SELF%NWP)%COO(2),FIRST_AND_LAST_UNKNOWN=.TRUE.)
            ! Z
            CALL INTERPOL_SPLINE_EVAL(SELF%Z_SPLINE,DBLE(SELF%T(1))       ,SELF%WP(1)%COO(3), FIRST_AND_LAST_UNKNOWN=.TRUE.)
            CALL INTERPOL_SPLINE_EVAL(SELF%Z_SPLINE,DBLE(SELF%T(SELF%NWP)),SELF%WP(SELF%NWP)%COO(3),FIRST_AND_LAST_UNKNOWN=.TRUE.)
            ! S
            CALL INTERPOL_SPLINE_EVAL(SELF%S_SPLINE,DBLE(SELF%T(1))       ,SELF%WP(1)%SUBUNIT_NUM, FIRST_AND_LAST_UNKNOWN=.TRUE.)
            CALL INTERPOL_SPLINE_EVAL(SELF%S_SPLINE,DBLE(SELF%T(SELF%NWP)),SELF%WP(SELF%NWP)%SUBUNIT_NUM,&
                                                                                                    FIRST_AND_LAST_UNKNOWN=.TRUE.)
        ENDIF
        ! phi
        !WRITE(*,'(A)') '**debug(FILAMENT_GUESS_FAL_WPS): about to guess PHI values. Here is the spline'
        !CALL INTERPOL_SPLINE_PRINT_INFO(SELF%PHI_SPLINE)
        CALL INTERPOL_SPLINE_EVAL(SELF%PHI_SPLINE,DBLE(SELF%T(1))       ,SELF%WP(1)%EULER%PHI, FIRST_AND_LAST_UNKNOWN=.TRUE.)
        CALL INTERPOL_SPLINE_EVAL(SELF%PHI_SPLINE,DBLE(SELF%T(SELF%NWP)),SELF%WP(SELF%NWP)%EULER%PHI,FIRST_AND_LAST_UNKNOWN=.TRUE.)

        ! Re-solve all the splines
        ! Have to make them parabolic, otherwise the derivatives at the new first and last
        ! points will not be computed
        !CALL SPLINE_SET_TYPE(SELF%X_SPLINE,  'PARABOLIC')
        !CALL SPLINE_SET_TYPE(SELF%Y_SPLINE,  'PARABOLIC')
        !CALL SPLINE_SET_TYPE(SELF%Z_SPLINE,  'PARABOLIC')
        HERMITE_BAK = SELF%USE_HERMITE_SPLINES
		SELF%USE_HERMITE_SPLINES = .FALSE.
		CALL FILAMENT_SOLVE_SPLINES(SELF)
		SELF%USE_HERMITE_SPLINES = HERMITE_BAK
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_GUESS_FAL_WPS): spline info before guess theta & psi'
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%X_SPLINE)
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%Y_SPLINE)
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%Z_SPLINE)
        ENDIF

        ! If we know the Psi and Theta angles for the other waypoints, we probably mean to estimate them for the FAL WPs too
        IF (PSI_THETA_NON_ZERO(SELF)) THEN
			WRITE(*,'(A)') '**debug(FILAMENT_GUESS_FAL_WPS): will estimate psi & theta for new FAL WPs'
            DO I=1,SELF%NWP,SELF%NWP-1
                CALL EULER_FROM_DERIVATIVES(SELF%X_SPLINE%DY(I),SELF%Y_SPLINE%DY(I),SELF%Z_SPLINE%DY(I),THETA_TMP,PSI_TMP)
                !IF (KNOWN_THETA)    SELF%WP(I)%EULER%THETA  =   THETA_TMP
                !IF (KNOWN_PSI)      SELF%WP(I)%EULER%PSI    =   PSI_TMP
                SELF%WP(I)%EULER%THETA  =   THETA_TMP
                SELF%WP(I)%EULER%PSI    =   PSI_TMP
            ENDDO
			HERMITE_BAK = SELF%USE_HERMITE_SPLINES
			SELF%USE_HERMITE_SPLINES = .FALSE.
            CALL FILAMENT_SOLVE_SPLINES(SELF)
			SELF%USE_HERMITE_SPLINES = HERMITE_BAK
            !CALL SPLINE_SET_TYPE(SELF%X_SPLINE,  'KNOWN_D')
            !CALL SPLINE_SET_TYPE(SELF%Y_SPLINE,  'KNOWN_D')
            !CALL SPLINE_SET_TYPE(SELF%Z_SPLINE,  'KNOWN_D')
		ELSE
			WRITE(*,'(2A)') '**debug(FILAMENT_GUESS_FAL_WPS): psi & theta are not known for any point, ',    &
			                'will not estimate them for new FAL WPs'
		ENDIF

    END SUBROUTINE FILAMENT_GUESS_FAL_WPS



END MODULE FILAMENTS_MODEL_BASE
