!>  \brief   Routines to modify the filament's model
MODULE FILAMENTS_MODEL_MODIFY
    USE FILAMENTS_BASE
    IMPLICIT NONE

    CONTAINS

    !>  \brief  Reverse the polarity of a filament (in-place)
    !!  \todo   Details - see comments
    !!
    !!  \todo   Update to work with SUBUNIT_NUM
    !!  \todo   Fix this - this may require some proper thinking on what the proper thing to do is. The problem is that after flipping,
    !!          filaments need several runs of the minimiser to find the alignment parameters that they can find quickly if the model is
    !!          rotated manually by 180 degrees (see notes 11-Mar-2011)
    SUBROUTINE FILAMENT_FLIP_POLARITY(SELF)
        USE EULERANGLES
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        USE FILAMENTS_NORMAL_MODES, ONLY : FILAMENT_UPDATE_NMS, FILAMENT_WPS_TO_NM_AMPLITUDES
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_UPDATE_WP_RES, FILAMENT_FIT_SPLINES_LINEAR
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE SPLINES, ONLY : INTERPOL_SPLINE_PRINT_INFO

        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),         INTENT(INOUT)   ::  SELF    !<  Filament object
        ! Private variables
        TYPE(FILAMENT)          ::  TMPFIL
        INTEGER                 ::  I, NWP
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        IF (DEBUG) THEN
            WRITE(*,'(/A)') '**debug(FILAMENT_FLIP_POLARITY): filament before flipping:'
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

        ! Number of waypoints
        NWP = SELF%NWP

        IF (DEBUG) THEN
            WRITE(*,*) '**debug(): HERE 5. T values = ', self%T
        ENDIF

        ! Allocate memory for tempfil
        CALL FILAMENT_ALLOCATE(TMPFIL,SELF%NWP)

        ! Copy the filament info over
        TMPFIL = SELF

        IF (DEBUG) THEN
            WRITE(*,*) '**debug(): HERE 4. T values = ', tmpfil%T
        ENDIF

        ! We need to reverse the order of the waypoints and define the colinear theta/psi
        ! Loop over the input filament's waypoints
        DO I=1,NWP
            ! Deal with the order of the waypoints
            TMPFIL%WP(NWP-I+1)              =   SELF%WP(I)
            TMPFIL%T(NWP-I+1)               =   SELF%T(NWP) - SELF%T(I)
            TMPFIL%WP(NWP-I+1)%PATH_LENGTH  =   SELF%WP(NWP)%PATH_LENGTH - SELF%WP(I)%PATH_LENGTH

            ! Deal with phi
            ! The line below was inspired by frealix1, but I'm not sure it makes sense now.
            !TMPFIL%WP(NWP-I+1)%EULER%PHI   =   TMPFIL%WP(NWP-I+1)%EULER%PHI * -1.
            ! For now, we set phi = 0.0, and we rely on filament_update_model to do the right thing
            TMPFIL%WP(NWP-I+1)%EULER%PHI    =   0.0

            ! Deal with theta
            TMPFIL%WP(NWP-I+1)%EULER%THETA  =   SELF%WP(I)%EULER%THETA - 180.0
            TMPFIL%WP(NWP-I+1)%EULER%THETA  =   DEGREES_M180_TO_180(TMPFIL%WP(NWP-I+1)%EULER%THETA)

            ! Deal with psi (note the convention in frealix that -90<psi<90)
            TMPFIL%WP(NWP-I+1)%EULER%PSI    =   SELF%WP(I)%EULER%PSI
            IF (DEBUG) THEN
                IF (TMPFIL%WP(NWP-I+1)%EULER%PSI .GT. 90. .OR. TMPFIL%WP(NWP-I+1)%EULER%PSI .LT. -90.) THEN
                    WRITE(*,'(A,I0,A,I0,A,F0.3,A)') '**WARNING(FILAMENT_FLIP_POLARITY): input waypoint ', I, &
                                                    ' (new waypoint #', NWP-I+1, ') has been given psi = ', &
                                                    TMPFIL%WP(NWP-I+1)%EULER%PSI, ', which is outside -90 to 90'
                ENDIF
            ENDIF
            ! The line below shouldn't be necessary, but on the other hand it shouldn't hurt either
            TMPFIL%WP(NWP-I+1)%EULER%PSI    =   DEGREES_M180_TO_180(TMPFIL%WP(NWP-I+1)%EULER%PSI)

            ! Set subunit number to 0.0 (filament_update_model will set good values later)
            TMPFIL%WP(NWP-I+1)%SUBUNIT_NUM  =   0.0

        ENDDO


        IF (DEBUG) THEN
            WRITE(*,*) '**debug(): HERE 1. T values = ', TMPFIL%T
        ENDIF

        ! Copy the flipped filament over
        SELF = TMPFIL

        IF (DEBUG) THEN
            WRITE(*,*) '**debug(): HERE 2. T values = ', self%T
        ENDIF

        ! Update splines
        CALL FILAMENT_UPDATE_MODEL(SELF,FORCE_UPDATE_RIGID_WP=.TRUE.) !,KNOWN_PSI=.TRUE.,KNOWN_THETA=.TRUE.

        ! Sanity check
        IF (DEBUG) THEN
            IF (SELF%WP(1)%PATH_LENGTH .NE. 0.) THEN
                WRITE(*,'(A,F0.3)') '**ERROR(FILAMENT_FLIP_POLARITY): path_length of first waypoint is not 0.0: ', &
                                    SELF%WP(1)%PATH_LENGTH
                CALL TERMINATE('Fatal error in FILAMENT_FLIP_POLARITY')
            ENDIF
        ENDIF

        ! We may need to update the normal modes & residuals
        CALL FILAMENT_FIT_SPLINES_LINEAR(SELF)
        CALL FILAMENT_UPDATE_WP_RES(SELF)
        CALL FILAMENT_UPDATE_NMS(SELF)
        CALL FILAMENT_WPS_TO_NM_AMPLITUDES(SELF)
        CALL FILAMENT_UPDATE_MODEL(SELF)

        !
        ! Diagnostics
        !
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_FLIP_POLARITY): filament after flipping:'
            CALL FILAMENT_PRINT_INFO(SELF)
            WRITE(*,'(A)') '**debug(FILAMENT_FLIP_POLARITY): splines after flipping:'
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%X_SPLINE)
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%Y_SPLINE)
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%Z_SPLINE)
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%PHI_SPLINE)
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%S_SPLINE)
        ENDIF

    END SUBROUTINE FILAMENT_FLIP_POLARITY

    !>  \brief  Set up the filament so that it follows the 1-start operator
    !!
    !!  This is useful when one wants to make sure the positions of each helical subunits are explicitely described,
    !!  for example before starting 3D reconstruciton
    SUBROUTINE FILAMENT_CONVERT_TO_ONE_START(SELF)
        USE HELICES, ONLY : HELIX_TO_ONE_START
        USE FILAMENTS_PATH, ONLY : WAYPOINT_FROM_T
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_HELICAL_TWIST, FILAMENT_SOLVE_SPLINES
        USE SPLINES, ONLY : INTERPOL_SPLINE_FIND_ROOT
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(INOUT)   ::  SELF    !<  Input filament
        ! Private variables
        INTEGER                     ::  CURRENT_WP
        TYPE(FILAMENT)              ::  FIL
        LOGICAL,        PARAMETER   ::  DEBUG       =   .FALSE.
        REAL                        ::  TMP, TWIST_PER_SUB_AT_WP1
        TYPE(WAYPOINT)              ::  FIRST_SUBUNIT_AFTER_WP1
        REAL(KIND=8)                ::  T_FIRST_SUBUNIT_AFTER_WP1
        ! Start work

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_CONVERT_TO_ONE_START): filament info on entry: '
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

        ! First, copy the filament over
        FIL = SELF

        ! Adjust phi values
        DO CURRENT_WP = 2, FIL%NWP
            FIL%WP(CURRENT_WP)%EULER%PHI = FIL%WP(CURRENT_WP)%EULER%PHI + 360. * REAL((FIL%WP(CURRENT_WP)%SUBUNIT_NUM &
                                                                        - FIL%WP(1)%SUBUNIT_NUM)/ FIL%HELIX%NUM_STARTS)
            !FIL%WP(CURRENT_WP)%EULER%PHI = FIL%WP(CURRENT_WP)%EULER%PHI + 360. * INT((FIL%WP(CURRENT_WP)%EULER%PHI - FIL%WP(CURRENT_WP-1)%EULER%PHI) / FIL%HELIX%TWIST)
        ENDDO
        ! Find nearest subunit after the first waypoint
        T_FIRST_SUBUNIT_AFTER_WP1 = INTERPOL_SPLINE_FIND_ROOT(FIL%S_SPLINE,DBLE(CEILING(FIL%WP(1)%SUBUNIT_NUM)))
        FIRST_SUBUNIT_AFTER_WP1 = WAYPOINT_FROM_T(FIL,T_FIRST_SUBUNIT_AFTER_WP1)
        ! In subunits, how far the first waypoint is to the nearest subunit
        TMP = FIL%WP(1)%SUBUNIT_NUM - FIRST_SUBUNIT_AFTER_WP1%SUBUNIT_NUM
        ! Local twist per subunit at first waypoint
        TWIST_PER_SUB_AT_WP1 = FILAMENT_EVAL_HELICAL_TWIST(FIL,FIL%T(1))
        ! Convert the twist per subunit into 1-start equivalent
        TWIST_PER_SUB_AT_WP1 = (360. + TWIST_PER_SUB_AT_WP1) / REAL(FIL%HELIX%NUM_STARTS)
        !
        FIL%WP%EULER%PHI = FIL%WP%EULER%PHI - FIL%WP(1)%EULER%PHI + FIRST_SUBUNIT_AFTER_WP1%EULER%PHI + TMP * TWIST_PER_SUB_AT_WP1

        ! Diagnostics
        !IF (DEBUG) THEN
        !    WRITE(*,'(A)') '**debug(FILAMENT_CONVERT_TO_ONE_START): filament after adjusting phi values: '
        !    CALL FILAMENT_PRINT_INFO(FIL)
        !ENDIF


        ! Update phi spline
        CALL FILAMENT_SOLVE_SPLINES(FIL)

        ! Convert the helical operator to its 1-start equivalent
        CALL HELIX_TO_ONE_START(FIL%HELIX)

        ! Update model
        CALL FILAMENT_UPDATE_MODEL(FIL)

        ! Diagnostics
        !IF (DEBUG) THEN
        !    WRITE(*,'(A)') '**debug(FILAMENT_CONVERT_TO_ONE_START): filament after updating model: '
        !    CALL FILAMENT_PRINT_INFO(FIL)
        !ENDIF

        ! Copy the filament back
        SELF = FIL

        ! Diagnostics
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_CONVERT_TO_ONE_START): filament info on exit: '
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

    END SUBROUTINE FILAMENT_CONVERT_TO_ONE_START

    !>  \brief  Shift a filament in the Z direction so that it coincides with the CTFTILT plane
    SUBROUTINE FILAMENT_SHIFT_AVERAGE_Z_TO_CTFTILT_PLANE(SELF)
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_AVERAGE_WP
        USE FILMS, ONLY : FILM_FIND_Z_FROM_XY
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),             INTENT(INOUT)   ::  SELF            !<  Filament object
        ! Private variables
        TYPE(WAYPOINT)  ::  AVE_WP
        REAL(KIND=4)    ::  SHIFT(6)
        REAL            ::  NEW_AVE_Z
        LOGICAL,        PARAMETER   ::  DEBUG   =   .TRUE.
        ! Start work

        ! Compute the average coordinates of the filament
        AVE_WP =  FILAMENT_AVERAGE_WP(SELF)

        ! Compute the desired new average Z from the CTFTILT information
        NEW_AVE_Z = FILM_FIND_Z_FROM_XY(SELF%FILM,AVE_WP%COO(1),AVE_WP%COO(2))

        ! Apply the shift
        SHIFT = (/ 0.0,0.0,NEW_AVE_Z-REAL(AVE_WP%COO(3),KIND=4),0.0,0.0,0.0  /)
        IF (DEBUG) WRITE(*,'(A,F0.3)') '**debug(FILAMENT_SHIFT_AVERAGE_Z_TO_CTFTILT_PLANE): shifting filament in Z by ', SHIFT(3)
        CALL FILAMENT_SHIFT_ALL_WPS(SELF,SHIFT)
    END SUBROUTINE FILAMENT_SHIFT_AVERAGE_Z_TO_CTFTILT_PLANE


    !>  \brief  Straighten filament to a straigt line with constant helical parameters
    SUBROUTINE FILAMENT_STRAIGHTEN(SELF)
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_UPDATE_WPS_FROM_WP_RES, FILAMENT_RELAX_HELIX
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        USE COOS, ONLY : WAYPOINTS_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     INTENT(INOUT)   ::  SELF
        ! Private variables
        LOGICAL,        PARAMETER   ::  DEBUG    = .FALSE.
        INTEGER                     ::  I
        LOGICAL,        PARAMETER   ::  SANITY_CHECK = .TRUE.
        REAL(KIND=8),   PARAMETER   ::  SANITY_TOL = 0.001D0
        INTEGER                     ::  ITER
        INTEGER,        PARAMETER   ::  NUM_ITER = 10
        LOGICAL                     ::  HERMITE_BACKUP
        ! Start work



        ! First, straighten the axis
        DO ITER=1,NUM_ITER
        SELF%WP_RES%COO(1)      = 0.0D0
        SELF%WP_RES%COO(2)      = 0.0D0
        SELF%WP_RES%COO(3)      = 0.0D0
        SELF%WP_RES%SUBUNIT_NUM = 0.0D0
        SELF%WP_RES%EULER%PHI   = 0.0D0
        CALL FILAMENT_UPDATE_WPS_FROM_WP_RES(SELF)
        HERMITE_BACKUP = SELF%USE_HERMITE_SPLINES
        SELF%USE_HERMITE_SPLINES = .FALSE.
        CALL FILAMENT_UPDATE_MODEL(SELF,REDO_LINEAR_FIT=.TRUE.) !,KNOWN_PSI=.FALSE.,KNOWN_THETA=.FALSE.
        SELF%USE_HERMITE_SPLINES = HERMITE_BACKUP
        ENDDO

        IF (SANITY_CHECK) THEN
            DO I=1,SELF%NWP
                ! All the linear residuals should be 0.0
                IF (ANY (ABS(SELF%WP_RES(I)%COO)          .GT. SANITY_TOL)) THEN
                    WRITE(*,'(A,I0,A,3(L1))') '**ERROR(FILAMENT_STRAIGHTEN): axis sanity check failed for waypoint ', I, &
                                                '. Residuals not 0 (xyz): ', ABS(SELF%WP_RES(I)%COO) .GT. SANITY_TOL
                    WRITE(*,'(3(F0.8,1X))') SELF%WP_RES(I)%COO
                    CALL WAYPOINTS_PRINT_INFO(SELF%WP_RES)
                    CALL TERMINATE('Failed sanity check in FILAMENT_STRAIGHTEN')
                ENDIF
            ENDDO
        ENDIF

        ! Second, relax the helix
        CALL FILAMENT_RELAX_HELIX(SELF,IGNORE_SUBUNIT_PHASING=.TRUE.)
        CALL FILAMENT_UPDATE_MODEL(SELF,REDO_LINEAR_FIT=.TRUE.)!,KNOWN_PSI=.FALSE.,KNOWN_THETA=.FALSE.)
        SELF%WP_RES%SUBUNIT_NUM = 0.0D0
        SELF%WP_RES%EULER%PHI   = 0.0D0
        CALL FILAMENT_UPDATE_WPS_FROM_WP_RES(SELF)
        CALL FILAMENT_UPDATE_MODEL(SELF,REDO_LINEAR_FIT=.TRUE.)!,KNOWN_PSI=.FALSE.,KNOWN_THETA=.FALSE.)


        IF (SANITY_CHECK) THEN
            DO I=1,SELF%NWP
                ! All the linear residuals should be 0.0
                IF (ANY (ABS(SELF%WP_RES(I)%COO)          .GT. SANITY_TOL) .OR. &
                         ABS(SELF%WP_RES(I)%SUBUNIT_NUM)  .GT. SANITY_TOL .OR. &
                         ABS(SELF%WP_RES(I)%EULER%PHI)    .GT. SANITY_TOL .OR. &
                         ABS(SELF%WP_RES(I)%EULER%THETA)  .GT. SANITY_TOL .OR. &
                         ABS(SELF%WP_RES(I)%EULER%PSI)    .GT. SANITY_TOL) THEN
                    WRITE(*,'(A,I0,A)') '**ERROR(FILAMENT_STRAIGHTEN): helix sanity check failed for waypoint ', &
                                        I, '. Residuals not 0: '
                    CALL WAYPOINTS_PRINT_INFO(SELF%WP_RES)
                    CALL TERMINATE('Failed sanity check in FILAMENT_STRAIGHTEN')
                ENDIF
            ENDDO
        ENDIF

        ! Diagnostic
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_STRAIGHTEN): straightened filament info'
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

    END SUBROUTINE FILAMENT_STRAIGHTEN


    !>  \brief  Apply an out-of-plane tilt to the whole filament as a rigid body about its "midpoint". Waypoints keep the same X,Y coordinates.
    !!
    !!  Note: for extreme simplification during inital coding: 1. the "midpoint" will be the median waypoint of the filament, 2. the filament
    !!  will be assumed to be completely straight.
    !!
    !!  \todo   Add a check for how straight the filament really is?
    !!
    SUBROUTINE FILAMENT_APPLY_TILT(SELF,TILT,WPS_AS_SPS)
        USE ROTATION_MATRICES, ONLY : RADIANS
        USE RUNTIME_PARAMETERS, ONLY : IS_EVEN
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                 INTENT(INOUT)   ::  SELF                !<  Filament object
        REAL,                           INTENT(IN)      ::  TILT                !<  Tilt angle to be applied. In degrees
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  WPS_AS_SPS          !<  Whether waypoints are being treated as single particles. False by default
        ! Private variables
        INTEGER                 ::  WP_MIDPOINT
        INTEGER                 ::  I
        REAL                    ::  DIST_TO_MIDPOINT, Z_ADD
        LOGICAL                 ::  WWPS_AS_SPS
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        ! Waypoints as single particles?
        IF (PRESENT(WPS_AS_SPS)) THEN
            WWPS_AS_SPS = WPS_AS_SPS
        ELSE
            WWPS_AS_SPS = .FALSE.
        ENDIF

        ! Diagnostics
        IF (DEBUG) THEN
            WRITE(*,'(/A)') '**debug(FILAMENT_APPLY_TILT): filament state before tilt is applied:'
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

        IF (SELF%RIGID_PARAMS(5)) THEN
            SELF%RIGID_WP%EULER%THETA = SELF%RIGID_WP%EULER%THETA + TILT
            CALL FILAMENT_UPDATE_MODEL(SELF)
        ELSE
            !
            ! Find the "midpoint" about which the rotation will happen
            ! Different logic depending on whether there is an odd or even number of waypoints in the filament
            !
            IF (IS_EVEN(SELF%NWP)) THEN
                ! Even number of waypoints
                WP_MIDPOINT = SELF%NWP/2 + 1
            ELSE
                ! Odd number of waypoints
                WP_MIDPOINT = (SELF%NWP-1)/2 + 1
            ENDIF

            !
            ! Loop over the waypoints
            !
            DO I=1,SELF%NWP
                IF (I .NE. WP_MIDPOINT) THEN
                    ! Compute straight-line distance from current waypoint to mid waypoint
                    DIST_TO_MIDPOINT = SQRT(    (SELF%WP(I)%COO(1)-SELF%WP(WP_MIDPOINT)%COO(1))**2 &
                                              + (SELF%WP(I)%COO(2)-SELF%WP(WP_MIDPOINT)%COO(2))**2 &
                                              + (SELF%WP(I)%COO(3)-SELF%WP(WP_MIDPOINT)%COO(3))**2 )
                    ! Compute the additional Z
                    ! The tilt is applied clockwise-positive around the Y axis (following Spider, Frealign convention, bah)
                    ! The X axis is the direction of the filament, and the coordinate system is right-handed
                    Z_ADD = TAN(RADIANS(TILT)) * DIST_TO_MIDPOINT
                    IF ( I .LT. WP_MIDPOINT) THEN
                        Z_ADD = - Z_ADD
                    ENDIF
                    ! Apply the change in Z
                    CALL FILAMENT_SHIFT_WP(SELF,I,[0.,0.,Z_ADD,0.,0.,0.],UPDATE_SPLINES=.FALSE.)
                ENDIF
                ! In the single-particle case, we update the theta (out-of-plane) Euler angle "manually"
                IF (WWPS_AS_SPS) THEN
                    SELF%WP(I)%EULER%THETA = SELF%WP(I)%EULER%THETA + TILT
                ENDIF
            ENDDO

            ! The Euler angles will be re-calculated
            SELF%WP%EULER%THETA = 0.0
            SELF%WP%EULER%PSI   = 0.0
            SELF%X_SPLINE%KNOWN_D = .FALSE.
            SELF%Y_SPLINE%KNOWN_D = .FALSE.
            SELF%Z_SPLINE%KNOWN_D = .FALSE.

            ! Update the model
            CALL FILAMENT_UPDATE_MODEL(SELF)
            SELF%X_SPLINE%KNOWN_D = .TRUE.
            SELF%Y_SPLINE%KNOWN_D = .TRUE.
            SELF%Z_SPLINE%KNOWN_D = .TRUE.
        ENDIF

        ! Diagnostics
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_APPLY_TILT): filament state after tilt is applied:'
            CALL FILAMENT_PRINT_INFO(SELF)
            WRITE(*,'(A)') ' '
        ENDIF

    END SUBROUTINE FILAMENT_APPLY_TILT



    !>  \brief  Apply the same shift (in X,Y,Z,phi,theta and psi) to all waypoints of a filament
    !!
    !!  \todo   Update to work with SUBUNIT_NUM
    SUBROUTINE FILAMENT_SHIFT_ALL_WPS(SELF,SHIFT,FILAMENT_COORDINATES,UPDATE_MODEL)
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_SOLVE_HELIX_SPLINES, FILAMENT_FIT_SPLINES_LINEAR, FILAMENT_UPDATE_WP_RES
        USE FILAMENTS_NORMAL_MODES, ONLY : FILAMENT_WPS_TO_NM_AMPLITUDES
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),         INTENT(INOUT)   ::  SELF                    !<  Filament object
        REAL,                   INTENT(IN)      ::  SHIFT(6)                !<  Shift to be applied in 1.X, 2.Y, 3.Z/defocus, 4.Phi, 5.Theta, 6.Psi
        LOGICAL,    OPTIONAL,   INTENT(IN)      ::  FILAMENT_COORDINATES    !<  If .TRUE., X refers to filament axis, Y refers to perpendicular. If .FALSE., which is the default, X and Y refer to shifts in the film's coordinate system.
        LOGICAL,    OPTIONAL,   INTENT(IN)      ::  UPDATE_MODEL            !<  If .TRUE., update splines after having shifted the waypoints. Defaults to .TRUE.
        ! Private variables
        LOGICAL ::  UUPDATE_MODEL
        INTEGER ::  I
        ! Start work

        ! Will we update splines once we've shifted all the waypoints?
        IF (PRESENT(UPDATE_MODEL)) THEN
            UUPDATE_MODEL = UPDATE_MODEL
        ELSE
            UUPDATE_MODEL = .TRUE.
        ENDIF

        ! Shift the waypoints
        DO I=1,SELF%NWP
            CALL FILAMENT_SHIFT_WP(SELF,I,SHIFT,FILAMENT_COORDINATES,UPDATE_SPLINES=.FALSE.)
        END DO

        ! Update splines if we are meant to
        IF (UUPDATE_MODEL) THEN
            CALL FILAMENT_UPDATE_MODEL(SELF,FORCE_UPDATE_RIGID_WP=.TRUE.)
            CALL FILAMENT_FIT_SPLINES_LINEAR(SELF)
            CALL FILAMENT_UPDATE_WP_RES(SELF)
            CALL FILAMENT_WPS_TO_NM_AMPLITUDES(SELF)
            CALL FILAMENT_UPDATE_MODEL(SELF,FORCE_UPDATE_RIGID_WP=.TRUE.)
        ENDIF
    END SUBROUTINE FILAMENT_SHIFT_ALL_WPS

    !>  \brief  Apply a shift to a filament's way point. The shift is in X,Y,Z/defocus,phi,theta and psi. Translation can be applied in regular coordinate system,
    !!          or with X being along the filament axis and Y perpendicular to it.
    !!
    !!  \todo   Update to work with SUBUNIT_NUM
    SUBROUTINE FILAMENT_SHIFT_WP(SELF,WP_NUM,SHIFT,FILAMENT_COORDINATES,UPDATE_SPLINES)
        USE ROTATION_MATRICES, ONLY : RADIANS
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Argument
        TYPE(FILAMENT),         INTENT(INOUT)   ::  SELF                    !<  Filament
        INTEGER,                INTENT(IN)      ::  WP_NUM                  !<  Number of the way point to be shifted
        REAL,                   INTENT(IN)      ::  SHIFT(6)                !<  Shift to be applied in 1.X, 2.Y, 3.Z/defocus, 4.Phi, 5.Theta, 6.Psi
        LOGICAL,    OPTIONAL,   INTENT(IN)      ::  FILAMENT_COORDINATES    !<  If .TRUE., X refers to filament axis, Y refers to perpendicular. If .FALSE., which is the default, X and Y refer to shifts in the film's coordinate system.
        LOGICAL,    OPTIONAL,   INTENT(IN)      ::  UPDATE_SPLINES          !<  If .TRUE., update splines after having shifted the waypoint. Defaults to .TRUE.
        ! Private variables
        LOGICAL ::  FFILAMENT_COORDINATES, UUPDATE_SPLINES
        REAL    ::  SSHIFT(6), IPA
        ! Start work
        ! Check the way point number
        IF (WP_NUM .GT. SELF%NWP) THEN
            WRITE(*,'(A,I0,A,I0)') '**ERROR(FILAMENT_SHIFT_WP): waypoint ', WP_NUM, &
                                    ' does not exist in this filament - number of way points = ', SELF%NWP
            CALL TERMINATE('Fatal error in FILAMENT_SHIFT_WP')
        ENDIF
        ! Are we thinking relative to film or filament?
        IF (PRESENT(FILAMENT_COORDINATES)) THEN
            FFILAMENT_COORDINATES = FILAMENT_COORDINATES
        ELSE
            FFILAMENT_COORDINATES = .FALSE.
        ENDIF
        ! Will we update splines once we've shifted the waypoint?
        IF (PRESENT(UPDATE_SPLINES)) THEN
            UUPDATE_SPLINES = UPDATE_SPLINES
        ELSE
            UUPDATE_SPLINES = .TRUE.
        ENDIF

        ! If necessary, convert the shifts to film coordinates.
        ! This can be done by looking up the PSI (in-plane) angle at the relevant way point.
        IF (FFILAMENT_COORDINATES) THEN
            IPA = RADIANS(SELF%WP(WP_NUM)%EULER%PSI)
            SSHIFT(1) = COS(IPA)*SHIFT(1) + SIN(IPA)*SHIFT(2)
            SSHIFT(2) = SIN(IPA)*SHIFT(1) + COS(IPA)*SHIFT(2)
        ELSE
            SSHIFT(1) = SHIFT(1)
            SSHIFT(2) = SHIFT(2)
        ENDIF
        ! The Z shift is applied directly
        SSHIFT(3) = SHIFT(3)
        ! The angular shifts will be applied directly
        SSHIFT(4:6) = SHIFT(4:6)
        ! Apply the shifts
        SELF%WP(WP_NUM)%COO(1:3) = SELF%WP(WP_NUM)%COO(1:3) + SSHIFT(1:3)
        SELF%WP(WP_NUM)%EULER%PHI = SELF%WP(WP_NUM)%EULER%PHI + SSHIFT(4)
        SELF%WP(WP_NUM)%EULER%THETA = SELF%WP(WP_NUM)%EULER%THETA + SSHIFT(5)
        SELF%WP(WP_NUM)%EULER%PSI = SELF%WP(WP_NUM)%EULER%PSI + SSHIFT(6)
        ! Diagnostics
        !PRINT '(A,6F12.3)', '**debug(FILAMENT_SHIFT_WP): applying shifts: ', SSHIFT

        !
        ! Update splines if we're meant to
        !
        IF (UUPDATE_SPLINES) THEN
            CALL FILAMENT_UPDATE_MODEL(SELF)
        ENDIF
    END SUBROUTINE FILAMENT_SHIFT_WP

   !>  \brief  Move a filament's waypoints to the nearest subunit position
    SUBROUTINE FILAMENT_MOVE_WPS_TO_NEAREST_SUBUNIT(SELF)
        USE FILAMENTS_PATH, ONLY : WAYPOINT_FROM_T
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_FIND_NEAREST_SUBUNIT_TO_WAYPOINT
        USE SPLINES, ONLY : INTERPOL_SPLINE_FIND_ROOT
        USE COOS, ONLY : WP_PRINT_INFO
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_SOLVE_SPLINES
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                 INTENT(INOUT)   ::  SELF                        !<  Filament whose waypoints will be modified
        ! Private variables
        INTEGER                 ::  CURRENT_WP, NEAREST_SUBUNIT
        REAL(KIND=8)            ::  NEAREST_SUBUNIT_T
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        DO CURRENT_WP=1,SELF%NWP
            IF (DEBUG) THEN
                WRITE(*,'(A,I0,A,I0,A)') '**debug(FILAMENT_MOVE_WPS_TO_NEAREST_SUBUNIT): waypoint ', CURRENT_WP, ' of ', &
                                            SELF%NWP, ' before change: '
                CALL WP_PRINT_INFO(SELF%WP(CURRENT_WP))
            ENDIF
            ! Find the T parameter for the nearest subunit
            NEAREST_SUBUNIT   = FILAMENT_FIND_NEAREST_SUBUNIT_TO_WAYPOINT(SELF,SELF%WP(CURRENT_WP))
            NEAREST_SUBUNIT_T = INTERPOL_SPLINE_FIND_ROOT(SELF%S_SPLINE,DBLE(NEAREST_SUBUNIT))
            ! Replace the waypoint
            SELF%WP(CURRENT_WP) = WAYPOINT_FROM_T(SELF,NEAREST_SUBUNIT_T)
            IF (DEBUG) THEN
                WRITE(*,'(A,I0,A,I0,A)') '**debug(FILAMENT_MOVE_WPS_TO_NEAREST_SUBUNIT): waypoint ', CURRENT_WP, ' of ', &
                                            SELF%NWP, ' after change: '
                CALL WP_PRINT_INFO(SELF%WP(CURRENT_WP))
            ENDIF
        ENDDO

        ! Update all splines
        CALL FILAMENT_SOLVE_SPLINES(SELF)

    END SUBROUTINE FILAMENT_MOVE_WPS_TO_NEAREST_SUBUNIT

    !>  \brief  Add disorder to a filament's waypoints
    SUBROUTINE FILAMENT_ADD_DISORDER(SELF,CUMULATIVE_DISORDER_STD,NON_CUMULATIVE_DISORDER_STD)
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
        USE COOS, ONLY : WAYPOINTS_ADD_DISORDER
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        REAL,                               INTENT(IN)      ::  CUMULATIVE_DISORDER_STD(7)      !<  For each parameter (X,Y,Z,phi,theta,psi,subunit_num), the standard deviation of the cumulative disorder. In units of pixels, degrees and subunits.
        REAL,                               INTENT(IN)      ::  NON_CUMULATIVE_DISORDER_STD(7)  !<  For each parameter, the standard deviation of the non-cumulative disorder
        ! Private variables

        ! Start work
        CALL WAYPOINTS_ADD_DISORDER(SELF%WP,CUMULATIVE_DISORDER_STD,NON_CUMULATIVE_DISORDER_STD)
        CALL FILAMENT_UPDATE_MODEL(SELF)
    END SUBROUTINE FILAMENT_ADD_DISORDER



    !>  \brief  Add new waypoints to the ends of the filament
    SUBROUTINE FILAMENT_ADD_FAL_WPS(SELF)
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_GUESS_FAL_WPS, FILAMENT_PRINT_INFO, FILAMENT_SET_T_FROM_PATH_LENGTH
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
        USE FILAMENTS_NORMAL_MODES, ONLY : FILAMENT_UPDATE_NMS, FILAMENT_WPS_TO_NM_AMPLITUDES
        USE FILAMENTS_GEOMETRY
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Private variables
        TYPE(FILAMENT) :: TMP_FIL
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_ADD_FAL_WPS): filament info on entry'
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

        ! Backup the filament
        TMP_FIL = SELF

        ! Deallocate & rallocate filament memory
        CALL FILAMENT_DEALLOCATE(SELF)
        CALL FILAMENT_ALLOCATE(SELF,TMP_FIL%NWP+2)

        ! Restore WPs
        SELF%WP(2:SELF%NWP-1) = TMP_FIL%WP

        ! Give path length values for the first and last waypoints
        WRITE(*,'(A)') '**info(FILAMENT_ADD_FAL_WPS): adding waypoints at beginning and end of the filament'
        SELF%WP(1)%PATH_LENGTH        = SELF%WP(2)%PATH_LENGTH - &
                                        0.5*(SELF%WP(3)%PATH_LENGTH - SELF%WP(2)%PATH_LENGTH)
        SELF%WP(SELF%NWP)%PATH_LENGTH = SELF%WP(SELF%NWP-1)%PATH_LENGTH +   &
                                        0.5*(SELF%WP(SELF%NWP-1)%PATH_LENGTH - SELF%WP(SELF%NWP-2)%PATH_LENGTH)

        ! Ensure the first waypoint has path_length = 0.0
        SELF%WP%PATH_LENGTH = SELF%WP%PATH_LENGTH - SELF%WP(1)%PATH_LENGTH

        ! Deal with the T array
        CALL FILAMENT_SET_T_FROM_PATH_LENGTH(SELF)

        ! Get values for the FAL waypoints
        CALL FILAMENT_GUESS_FAL_WPS(SELF,GUESS_FAL_EULER_ONLY=.FALSE.)!,KNOWN_PSI=.TRUE.,KNOWN_THETA=.TRUE.)

        ! The full splines can now be solved
        CALL FILAMENT_SOLVE_SPLINES(SELF)!,KNOWN_PSI_THETA=.TRUE.) ! Commented out 29-Nov-2012 with introduction of USE_HERMITE_SPLINES flag

        ! Update the full model
        CALL FILAMENT_UPDATE_MODEL(SELF)

        ! Update the normal modes & residuals
        ! Fit the splines & compute residuals
        CALL FILAMENT_FIT_SPLINES_LINEAR(SELF)
        CALL FILAMENT_UPDATE_WP_RES(SELF)
        ! Compute the normal modes
        CALL FILAMENT_UPDATE_NMS(SELF)
        CALL FILAMENT_WPS_TO_NM_AMPLITUDES(SELF)

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_ADD_FAL_WPS): filament info on exit'
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

    END SUBROUTINE FILAMENT_ADD_FAL_WPS



END MODULE FILAMENTS_MODEL_MODIFY
