!>  \brief  Methods to deal with normal modes
MODULE FILAMENTS_NORMAL_MODES
    USE FILAMENTS_BASE
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE


    CONTAINS

    !>  \brief  Return a weighted sum of normal mode amplitudes. The amplitudes are weighted by the mode's eigenvalues.
    FUNCTION FILAMENT_NM_AMPLITUDES_WEIGHTED_SUM(SELF) RESULT(AMP_SUM)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        ! Result
        REAL(KIND=8)    ::  AMP_SUM
        ! Private variables
        INTEGER ::  CURRENT_MODE
        LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
        ! Relative weights
        REAL(KIND=8)   ::  REL_WEIGHT_XYZ
        REAL(KIND=8)   ::  REL_WEIGHT_PSITHETA
        REAL(KIND=8)   ::  REL_WEIGHT_SPHI
        REAL(KIND=8)    ::  DUMMY
        ! Start work
        AMP_SUM = 0.0D0

        ! User-set values
        REL_WEIGHT_XYZ      =   1.0D0
        REL_WEIGHT_PSITHETA =   3.0D0
        REL_WEIGHT_SPHI     =   1.0D0

        ! Normalise so that the total power is maintained (assume equal power in each mode)
        DUMMY = (3.*REL_WEIGHT_XYZ + 2.*REL_WEIGHT_PSITHETA + 2.*REL_WEIGHT_SPHI) / 7.
        REL_WEIGHT_XYZ = REL_WEIGHT_XYZ / DUMMY
        REL_WEIGHT_PSITHETA = REL_WEIGHT_PSITHETA / DUMMY
        REL_WEIGHT_SPHI = REL_WEIGHT_SPHI / DUMMY


        ! Loop over each mode, ignore the first mode which has no associated energy
        DO CURRENT_MODE = 2,SELF%NWP
            AMP_SUM = AMP_SUM + SELF%NM_EIGENVALUES(CURRENT_MODE) * (             &
                                    REL_WEIGHT_XYZ      * ABS(SELF%WP_NM(CURRENT_MODE)%COO(1))      +   &
                                    REL_WEIGHT_XYZ      * ABS(SELF%WP_NM(CURRENT_MODE)%COO(2))      +   &
                                    REL_WEIGHT_XYZ      * ABS(SELF%WP_NM(CURRENT_MODE)%COO(3))      +   &
                                    REL_WEIGHT_SPHI     * ABS(SELF%WP_NM(CURRENT_MODE)%EULER%PHI)   +   &
                                    REL_WEIGHT_PSITHETA * ABS(SELF%WP_NM(CURRENT_MODE)%EULER%THETA) +   &
                                    REL_WEIGHT_PSITHETA * ABS(SELF%WP_NM(CURRENT_MODE)%EULER%PSI)   +   &
                                    REL_WEIGHT_SPHI     * ABS(SELF%WP_NM(CURRENT_MODE)%SUBUNIT_NUM)     &
                                                                                    )
        ENDDO

        IF (DEBUG) THEN
            WRITE(*,'(A,I0,A,F0.3)') '**debug(FILAMENT_NM_AMPLITUDES_WEIGHTED_SUM): weighted sum of NM amplitudes for filament ', &
                                        SELF%ID, ' : ', AMP_SUM
        ENDIF
    END FUNCTION FILAMENT_NM_AMPLITUDES_WEIGHTED_SUM

    !>  \brief  Convert NM amplitudes to waypoint data
    SUBROUTINE FILAMENT_NM_AMPLITUDES_TO_WPS(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_PSI_BTW_M90_90, FILAMENT_UPDATE_WPS_FROM_WP_RES
        IMPLICIT NONE
        ! Aruguments
        TYPE(FILAMENT), INTENT(INOUT)  ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_WP, CURRENT_MODE
        ! Start work

        ! Sanity checks
        IF (SIZE(SELF%NM_EIGENVECTORS,2) .NE. SELF%NWP) THEN
            CALL TERMINATE('Failed sanity check')
        ENDIF

        ! The X coordinates along the filament can be recovered
        ! by doing a sum of the eigenvectors weighted by the
        ! amplitudes stored in the X coordinates of the NM WPs
        DO CURRENT_WP=1,SELF%NWP
            ! Reset
            SELF%WP_RES(CURRENT_WP)%COO         = 0.0D0
            SELF%WP_RES(CURRENT_WP)%EULER%PHI   = 0.0D0
            SELF%WP(CURRENT_WP)%EULER%THETA     = 0.0D0
            SELF%WP(CURRENT_WP)%EULER%PSI       = 0.0D0
            SELF%WP_RES(CURRENT_WP)%SUBUNIT_NUM = 0.0D0
            DO CURRENT_MODE = 1, SIZE(SELF%NM_EIGENVECTORS,2)
                IF (FILAMENT_REFINE_AXIS_NMS(SELF)) THEN
                    SELF%WP_RES(CURRENT_WP)%COO(1:3)    = SELF%WP_RES(CURRENT_WP)%COO(1:3) + &
                                                            SELF%WP_NM(CURRENT_MODE)%COO(1:3) * &
                                                            SELF%NM_EIGENVECTORS(CURRENT_WP,CURRENT_MODE)
                    SELF%WP(CURRENT_WP)%EULER%THETA     = SELF%WP(CURRENT_WP)%EULER%THETA + &
                                                            SELF%WP_NM(CURRENT_MODE)%EULER%THETA * &
                                                            SELF%NM_EIGENVECTORS(CURRENT_WP,CURRENT_MODE)
                    SELF%WP(CURRENT_WP)%EULER%PSI       = SELF%WP(CURRENT_WP)%EULER%PSI + &
                                                            SELF%WP_NM(CURRENT_MODE)%EULER%PSI * &
                                                            SELF%NM_EIGENVECTORS(CURRENT_WP,CURRENT_MODE)
                ENDIF
                ! The helical parameters
                IF (SELF%REFINE_NORMAL_MODES(4)) THEN
                    SELF%WP_RES(CURRENT_WP)%EULER%PHI   = SELF%WP_RES(CURRENT_WP)%EULER%PHI + &
                                                            SELF%WP_NM(CURRENT_MODE)%EULER%PHI * &
                                                            SELF%NM_EIGENVECTORS(CURRENT_WP,CURRENT_MODE)
                ENDIF
                IF (SELF%REFINE_NORMAL_MODES(7)) THEN
                    SELF%WP_RES(CURRENT_WP)%SUBUNIT_NUM = SELF%WP_RES(CURRENT_WP)%SUBUNIT_NUM + &
                                                            SELF%WP_NM(CURRENT_MODE)%SUBUNIT_NUM * &
                                                            SELF%NM_EIGENVECTORS(CURRENT_WP,CURRENT_MODE)
                ENDIF
            ENDDO
        ENDDO

        ! Now we need to convert residuals to actual waypoint data for the relevant dimensions (x,y,z,phi,s)
        CALL FILAMENT_UPDATE_WPS_FROM_WP_RES(SELF,  AXIS=FILAMENT_REFINE_AXIS_NMS(SELF),    &
                                                    PHI=SELF%REFINE_NORMAL_MODES(4),        &
                                                    SUBUNIT_NUM=SELF%REFINE_NORMAL_MODES(7))

        ! We need to make sure psi & theta adhere to the convention that -90<=psi<=90
        CALL FILAMENT_PSI_BTW_M90_90(SELF)

    END SUBROUTINE FILAMENT_NM_AMPLITUDES_TO_WPS


    !>  \brief  Convert waypoint data to NM amplitudes
    SUBROUTINE FILAMENT_WPS_TO_NM_AMPLITUDES(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_PSI_THETA_SMOOTH, FILAMENT_PSI_BTW_M90_90
        USE COOS, ONLY : WP_PRINT_INFO
        IMPLICIT NONE
        ! Aruguments
        TYPE(FILAMENT), INTENT(INOUT)  ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_WP, CURRENT_MODE
        LOGICAL,    PARAMETER   ::  DEBUG    = .FALSE.
        ! Start work

        ! Sanity checks
        IF (SIZE(SELF%NM_EIGENVECTORS,2) .NE. SELF%NWP) THEN
            WRITE(*,'(A,3(I0,1X))') '**ERROR(FILAMENT_WPS_TO_NM_AMPLITUDES): eigenvectors dim, nwp = ', &
                                                                       SIZE(SELF%NM_EIGENVECTORS,1), &
                                                                       SIZE(SELF%NM_EIGENVECTORS,2), &
                                                                       SELF%NWP
            CALL TERMINATE('Failed sanity check')
        ENDIF

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_WPS_TO_NM_AMPLITUDES): NMs follow'
            CALL FILAMENT_PRINT_NMS(SELF)
        ENDIF

        ! Allocate WP_NM array if necessary
        IF (.NOT. ALLOCATED(SELF%WP_NM)) THEN
            ALLOCATE(SELF%WP_NM(SELF%NWP))
        ENDIF

        ! Make sure psi & theta are as smooth as possible for our NM calculation
        CALL FILAMENT_PSI_THETA_SMOOTH(SELF)

        !
        DO CURRENT_MODE=1,SIZE(SELF%NM_EIGENVECTORS,2)
            !
            SELF%WP_NM(CURRENT_MODE)%COO = 0.0D0
            SELF%WP_NM(CURRENT_MODE)%EULER%PHI = 0.0D0
            SELF%WP_NM(CURRENT_MODE)%EULER%THETA = 0.0D0
            SELF%WP_NM(CURRENT_MODE)%EULER%PSI = 0.0D0
            SELF%WP_NM(CURRENT_MODE)%SUBUNIT_NUM = 0.0D0
            !
            DO CURRENT_WP=1,SELF%NWP
                SELF%WP_NM(CURRENT_MODE)%COO = SELF%WP_NM(CURRENT_MODE)%COO + &
                                            SELF%WP_RES(CURRENT_WP)%COO * SELF%NM_EIGENVECTORS(CURRENT_WP,CURRENT_MODE)
                SELF%WP_NM(CURRENT_MODE)%EULER%PHI = SELF%WP_NM(CURRENT_MODE)%EULER%PHI + &
                                            SELF%WP_RES(CURRENT_WP)%EULER%PHI * SELF%NM_EIGENVECTORS(CURRENT_WP,CURRENT_MODE)
                SELF%WP_NM(CURRENT_MODE)%EULER%THETA = SELF%WP_NM(CURRENT_MODE)%EULER%THETA + &
                                            SELF%WP(CURRENT_WP)%EULER%THETA * SELF%NM_EIGENVECTORS(CURRENT_WP,CURRENT_MODE)
                SELF%WP_NM(CURRENT_MODE)%EULER%PSI = SELF%WP_NM(CURRENT_MODE)%EULER%PSI + &
                                            SELF%WP(CURRENT_WP)%EULER%PSI * SELF%NM_EIGENVECTORS(CURRENT_WP,CURRENT_MODE)
                SELF%WP_NM(CURRENT_MODE)%SUBUNIT_NUM = SELF%WP_NM(CURRENT_MODE)%SUBUNIT_NUM + &
                                            SELF%WP_RES(CURRENT_WP)%SUBUNIT_NUM * SELF%NM_EIGENVECTORS(CURRENT_WP,CURRENT_MODE)
            ENDDO
            IF (DEBUG) THEN
                WRITE(*,'(A,I0)') '**debug(FILAMENT_WPS_TO_NM_AMPLITUDES): NM amplitudes for mode ', CURRENT_MODE
                CALL WP_PRINT_INFO(SELF%WP_NM(CURRENT_MODE))
            ENDIF
        ENDDO

        ! Now that we are done, we can return psi & theta to their original values,
        ! which adhere to the convention that -90<=psi<=90
        CALL FILAMENT_PSI_BTW_M90_90(SELF)
    END SUBROUTINE FILAMENT_WPS_TO_NM_AMPLITUDES


    !>  \brief  Print out normal modes for the filament
    SUBROUTINE FILAMENT_PRINT_NMS(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        ! Private variables
        INTEGER ::  I
        CHARACTER(LEN=64) :: LINE_FORMAT
        ! Start work

        ! Set up the printout format
        LINE_FORMAT = ' '
        WRITE(LINE_FORMAT,'(A,I0,A)') '(I4,X,G  10.3,X,', SIZE(SELF%NM_EIGENVECTORS,1), '(F8.3,X))'


        ! Do the printing
        WRITE(*,'(A,I0)') 'Normal modes for filament ', SELF%ID
        WRITE(*,'(3(A,1X))') 'NM #', 'Eigenvalue', 'Eigenvector'
        DO I=1,SIZE(SELF%NM_EIGENVALUES)
            WRITE(*,LINE_FORMAT) I, SELF%NM_EIGENVALUES(I), SELF%NM_EIGENVECTORS(:,I)
        ENDDO

    END SUBROUTINE FILAMENT_PRINT_NMS

    !>  \brief  Sort eigenvectors by increasing eigenvalue (aka eigenfrequency)
    SUBROUTINE FILAMENT_SORT_NMS(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER, ALLOCATABLE ::  POSITION_OF_LOWEST_EIGENVALUE(:)
        INTEGER ::  I
        REAL, ALLOCATABLE    ::  CURRENT_EIGENVECTOR(:)
        REAL    ::CURRENT_EIGENVALUE
        ! Start work

        ! Loop over the vectors / values
        DO I=1,SIZE(SELF%NM_EIGENVALUES)-1
            ! Find the position of lowest eigenvalue
            POSITION_OF_LOWEST_EIGENVALUE = MINLOC(SELF%NM_EIGENVALUES(I:SIZE(SELF%NM_EIGENVALUES))) + I - 1
            ! Move the eigenvectors and value to the current position
            ! First we copy to temp variables
            CURRENT_EIGENVALUE  = SELF%NM_EIGENVALUES(POSITION_OF_LOWEST_EIGENVALUE(1))
            CURRENT_EIGENVECTOR = SELF%NM_EIGENVECTORS(:,POSITION_OF_LOWEST_EIGENVALUE(1))
            ! Then we clear space for them
            SELF%NM_EIGENVALUES(POSITION_OF_LOWEST_EIGENVALUE(1)) = SELF%NM_EIGENVALUES(I)
            SELF%NM_EIGENVECTORS(:,POSITION_OF_LOWEST_EIGENVALUE(1)) = SELF%NM_EIGENVECTORS(:,I)
            ! Now we place the vector/value in their correct, sorted position
            SELF%NM_EIGENVALUES(I) = CURRENT_EIGENVALUE
            SELF%NM_EIGENVECTORS(:,I) = CURRENT_EIGENVECTOR
        ENDDO
    END SUBROUTINE FILAMENT_SORT_NMS

    !>  \brief  Update normal modes for the filament
    SUBROUTINE FILAMENT_UPDATE_NMS(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Start work
        CALL FILAMENT_NMS_FROM_PATH_LENGTHS(SELF)
        !CALL FILAMENT_SORT_NMS(SELF) ! Not necessary, because the FSGSL eigen analysis routine sorted the vectors already
    END SUBROUTINE FILAMENT_UPDATE_NMS

    !>  \brief  Compute eigenvectors & eigenvalues (i.e. normal modes) from the waypoints' path lengths
    !!
    !!  Following method suggested by Niko (11-Feb-2012)
    SUBROUTINE FILAMENT_NMS_FROM_PATH_LENGTHS(SELF)
        USE EIGEN
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE ! DEBUG
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF    !<  Filament object
        ! Private variables
        INTEGER ::  I,J,K
        REAL(KIND=8)    ::  E((SELF%NWP-1)*2,SELF%NWP)
        REAL(KIND=8)    ::  A(SELF%NWP,SELF%NWP)
        REAL(KIND=8)    ::  X(SELF%NWP-1)
        !REAL(KIND=8)    ::  D(SELF%NWP), O(SELF%NWP)
        !REAL(KIND=8)    ::  Z(SELF%NWP,SELF%NWP)
        REAL(KIND=8)    ::  S
        REAL(KIND=8),   PARAMETER   ::  Y1 = 1.0D0
        REAL(KIND=8),   PARAMETER   ::  OVERALL_RESTORING_CONSTANT  =   0.0D0
        LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
        LOGICAL,        PARAMETER   ::  USE_FGSL = .TRUE.
        CHARACTER(LEN=6)    ::  NWP_M, NWP_C
        ! Start work
        IF (DEBUG) THEN
            WRITE(NWP_M,'(I6)') SELF%NWP-1
            WRITE(NWP_C,'(I6)') SELF%NWP
        ENDIF
        ! The X array contains the path-length distances from
        ! waypoint to waypoint
        DO I=2,SELF%NWP
            X(I-1) = 1./(SELF%WP(I)%PATH_LENGTH - SELF%WP(I-1)%PATH_LENGTH)
        ENDDO

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_NMS_FROM_PATH_LENGTHS): coupling values:'
            WRITE(*,'('//NWP_M//'(F0.2,1X))') (X(I),I=1,SELF%NWP-1)
        ENDIF


        ! Set up E matrix (coupling matrix)
        E = 0.0D0
        DO I=1,SELF%NWP-1
            E(I,I)   = -1.0D0
            E(I+SELF%NWP-1,I) = -1.0D0
            E(I,I+1) =  1.0D0
        ENDDO

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_NMS_FROM_PATH_LENGTHS): coupling matrix:'
            DO I=1,SELF%NWP-1
                WRITE(*,'('//NWP_M//'(F7.3,1X))')  (E(J,I),J=1,SELF%NWP-1)
            ENDDO
        ENDIF

        ! Find the eigenvectors & values for the matrix
        IF (USE_FGSL) THEN
            ! Set up A matrix (eigen matrix)
            DO I=1,SELF%NWP
                DO J=1,SELF%NWP
                    S=0.0D0
                    DO K=1,SELF%NWP-1
                        S=S+Y1*E(K,I)*E(K,J)*X(K)
                        S=S+OVERALL_RESTORING_CONSTANT*E(K+SELF%NWP-1,I)*E(K+SELF%NWP-1,J)
                    ENDDO
                    A(I,J) = S
                ENDDO
            ENDDO

            IF (DEBUG) THEN
                WRITE(*,'(A)') '**debug(FILAMENT_NMS_FROM_PATH_LENGTHS): eigen matrix:'
                DO I=1,SELF%NWP
                    WRITE(*,'('//NWP_C//'(F7.3,1X))') (A(J,I),J=1,SELF%NWP)
                ENDDO
            ENDIF

            CALL MATRIX_EIGEN_ANALYSIS(A,SELF%NM_EIGENVECTORS,SELF%NM_EIGENVALUES)
            SELF%NM_EIGENVECTORS = TRANSPOSE(SELF%NM_EIGENVECTORS)
            IF (DEBUG) THEN
                PRINT *, 'SELF%NM_EIGENVALUES NEW', SELF%NM_EIGENVALUES
                PRINT *, 'SELF%NM_EIGENVECTORS NEW', SELF%NM_EIGENVECTORS
            ENDIF
        ENDIF




        IF (DEBUG) THEN
            PRINT *, '**DEBUG(FILAMENTS_NORMAL_MODES): EIGENVECTORS'
            PRINT *, SELF%NM_EIGENVECTORS
            PRINT *, '**DEBUG(FILAMENTS_NORMAL_MODES): END EIGENVECTORS'
            !CALL FILAMENT_PRINT_NMS(SELF)
            !CALL FILAMENT_WPS_TO_NM_AMPLITUDES(SELF)
            !WRITE(*,'(A)') '**DEBUG(FILAMENT_NMS_FROM_PATH_LENGTHS): WP NM array:'
            !CALL WAYPOINTS_PRINT_INFO(SELF%WP_NM)
            !CALL FILAMENT_NM_AMPLITUDES_TO_WPS(SELF)
            !WRITE(*,'(A)') '**DEBUG(FILAMENT_NMS_FROM_PATH_LENGTHS): reconstructed WP array:'
            !CALL WAYPOINTS_PRINT_INFO(SELF%WP)
            !CALL TERMINATE('DEBUG')
        ENDIF

    END SUBROUTINE FILAMENT_NMS_FROM_PATH_LENGTHS
END MODULE FILAMENTS_NORMAL_MODES
