!>  \brief  Routines to deal with the arc length parametrisation of the filament's axial space curve
MODULE FILAMENTS_PARAMETRISATION
    USE FILAMENTS_BASE
    IMPLICIT NONE

    CONTAINS

    !>  \brief  Iteratively refine the parametrisation of the splines until the target error tolerance is reached
    SUBROUTINE FILAMENT_PARAMETRISATION_UPDATE(SELF,TOLERANCE)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_SOLVE_SPLINES
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF
        !LOGICAL,            OPTIONAL,       INTENT(IN)      ::  USE_PSI_THETA_CONSTRAINTS   !<  Whether the psi & theta angles are known for each waypoint and should be used as constraints when solving splines
        REAL,               OPTIONAL,       INTENT(IN)      ::  TOLERANCE                   !<  Parametrisation error tolerance (pixels)
        ! Private variables
        INTEGER                 ::  PARAMETRISATION_NUM_ITERATIONS
        REAL                    ::  PREVIOUS_PARAMETRISATION_ERROR
        INTEGER,    PARAMETER   ::  PARAMETRISATION_MAX_NUM_ITERATIONS  =   300
        INTEGER,    PARAMETER   ::  PARAMETRISATION_MIN_NUM_ITERATIONS  =   2
        REAL(KIND=8)            ::  PARAMETRISATION_TARGET_ERROR
        REAL(KIND=8)            ::  CURRENT_PARAMETRISATION_ERROR
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_PARAMETRISATION_UPDATE): filament info on entry:'
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

        !IF (PRESENT(USE_PSI_THETA_CONSTRAINTS)) CALL TERMINATE('USE_PSI_THETA_CONSTAINTS is deprecated here')

        ! Work out tolerance
        IF (PRESENT(TOLERANCE)) THEN
            PARAMETRISATION_TARGET_ERROR = DBLE(TOLERANCE)
        ELSE
            IF (SELF%HELIX%RISE .EQ. 0.0) THEN
                PARAMETRISATION_TARGET_ERROR = 0.1D0
                WRITE(*,'(2A)') '**warning(FILAMENT_PARAMETRISATION_UPDATE): using parametrisation tolerance of 0.1 ', &
                                'because helical parameters were not supplied. This may not be sufficient in some cases'
            ELSE
                IF (SELF%HELIX%PIXEL_UNITS) THEN
                    PARAMETRISATION_TARGET_ERROR = MIN(SELF%HELIX%RISE / 100.D0,0.1D0)
                ELSE
                    WRITE(*,'(2A)') '**ERROR(FILAMENT_PARAMETRISATION_UPDATE): helical symmetry needs to be in ',  &
                                    'pixel units to work out tolerance for parametrisation'
                    CALL TERMINATE('Fatal error in FILAMENT_PARAMETRISATION_UPDATE')
                ENDIF
            ENDIF
        ENDIF

        IF (.NOT. ANY(SELF%WP%PATH_LENGTH .NE. 0.0)) THEN
            CALL FILAMENT_UPDATE_PATH_LENGTH(SELF)
        ENDIF

        ! Iterative parametrisation
        PARAMETRISATION_NUM_ITERATIONS = 0
        PREVIOUS_PARAMETRISATION_ERROR = HUGE(1.0)
        DO
            PARAMETRISATION_NUM_ITERATIONS = PARAMETRISATION_NUM_ITERATIONS + 1
            CALL FILAMENT_SET_T_FROM_PATH_LENGTH(SELF)
            IF (DEBUG) THEN
                WRITE(*,'(A,I0)') '**debug(FILAMENT_PARAMETRISATION_UPDATE): (before solve_splines) arc length ',   &
                                    'parametrisation, iteration #', PARAMETRISATION_NUM_ITERATIONS
                WRITE(*,'(A)') '**debug(FILAMENT_PARAMETRISATION_UPDATE): filament info:'
                CALL FILAMENT_PRINT_INFO(SELF)
            ENDIF
            CALL FILAMENT_SOLVE_SPLINES(SELF)   !,USE_PSI_THETA_CONSTRAINTS=USE_PSI_THETA_CONSTRAINTS) ! Commented out 29-Nov-2012 with introduction of USE_HERMITE_SPLINES parameter
            IF (DEBUG) THEN
                WRITE(*,'(A,I0)') '**debug(FILAMENT_PARAMETRISATION_UPDATE): (after solve_splines) arc ',   &
                                'length parametrisation, iteration #', PARAMETRISATION_NUM_ITERATIONS
                WRITE(*,'(A)') '**debug(FILAMENT_PARAMETRISATION_UPDATE): filament info:'
                CALL FILAMENT_PRINT_INFO(SELF)
            ENDIF
            CURRENT_PARAMETRISATION_ERROR = FILAMENT_PARAMETRISATION_ERROR(SELF)
            IF (DEBUG) THEN
                WRITE(*,'(A,I0,A,F0.11,A)') '**debug(FILAMENT_PARAMETRISATION_UPDATE): arc length parametrisation', &
                                            ', iteration #', PARAMETRISATION_NUM_ITERATIONS, '; parametrisation error = ', &
                                            CURRENT_PARAMETRISATION_ERROR, ' pixels'
                !WRITE(*,'(A)') '**debug(FILAMENT_PARAMETRISATION_UPDATE): filament info:'
                !CALL FILAMENT_PRINT_INFO(SELF)
            ENDIF
            IF (    CURRENT_PARAMETRISATION_ERROR .LT. PARAMETRISATION_TARGET_ERROR &
                    .AND. PARAMETRISATION_NUM_ITERATIONS .GE. PARAMETRISATION_MIN_NUM_ITERATIONS &
                    !.AND. CURRENT_PARAMETRISATION_ERROR .LE. PREVIOUS_PARAMETRISATION_ERROR    &
                    ) THEN
                EXIT
            ENDIF
            PREVIOUS_PARAMETRISATION_ERROR = CURRENT_PARAMETRISATION_ERROR
            IF (PARAMETRISATION_NUM_ITERATIONS .GT. PARAMETRISATION_MAX_NUM_ITERATIONS) THEN
                WRITE(*,'(A,I4,2(A,F13.10))')   '**warning(FILAMENT_PARAMETRISATION_UPDATE): spline parametrisation ',  &
                                                'did not converge after ', PARAMETRISATION_NUM_ITERATIONS, &
                                                        ' iterations. Current error = ', FILAMENT_PARAMETRISATION_ERROR(SELF), &
                                                        '. Target error =', PARAMETRISATION_TARGET_ERROR
                IF (PARAMETRISATION_NUM_ITERATIONS .GT. 2*PARAMETRISATION_MAX_NUM_ITERATIONS) THEN
                    CALL TERMINATE('Fatal error in FILAMENT_PARAMETRISATION_UPDATE')
                ENDIF
            ENDIF
        ENDDO

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_PARAMETRISATION_UPDATE): filament info on exit:'
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF

    END SUBROUTINE FILAMENT_PARAMETRISATION_UPDATE



    !>  \brief  Measure the discrepency between the filament's spline parametrisation and the current estimates of the arc lengths between waypoints
    FUNCTION FILAMENT_PARAMETRISATION_ERROR(SELF) RESULT(ERROR)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE, INTRINSIC :: IEEE_ARITHMETIC, ONLY : IEEE_IS_NAN
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF
        ! Result
        REAL(KIND=8)                    ::  ERROR   !< Average over waypoints of the difference (pixels) between the parameter value and the path length
        ! Private variables
        INTEGER ::  I
        LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
        REAL(KIND=8) :: DUMMY
        ! Start work

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_PARAMETRISATION_ERROR): filament info on entry'
            CALL FILAMENT_PRINT_INFO(SELF)
            DUMMY = SELF%WP_NM(2)%COO(1)
        ENDIF

        ! Update the path lengths (this is done by walking in tiny steps along the parameter and
        ! integrating the path lengths)
        CALL FILAMENT_UPDATE_PATH_LENGTH(SELF)

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_PARAMETRISATION_ERROR): filament info after update path length'
            CALL FILAMENT_PRINT_INFO(SELF)
            IF (SELF%WP_NM(2)%COO(1) .NE. DUMMY) THEN
                PRINT *, SELF%WP_NM(2)%COO(1), DUMMY
                CALL TERMINATE ('BUG')
            ENDIF
        ENDIF

        ! Sanity check. First waypoint should have parameter and path_length of 0.0
        IF (SELF%WP(1)%PATH_LENGTH .NE. 0.0) THEN
            WRITE(*,'(A,F13.8)') '**ERROR(FILAMENT_PARAMETRISATION_ERROR): non-zero path length for first waypoint: ', &
                                SELF%WP(1)%PATH_LENGTH
            CALL TERMINATE('Fatal error in FILAMENT_PARAMETRISATION_ERROR')
        ENDIF

        ! Compute error
        ERROR = 0.0D0
        DO I=1,SELF%NWP
            ERROR = ERROR + ABS(SELF%T(I)-SELF%T(1)-SELF%WP(I)%PATH_LENGTH)
        ENDDO
        ERROR = ERROR / SELF%NWP

        !
        IF (DEBUG) THEN
            IF (IEEE_IS_NAN(ERROR)) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_PARAMETRISATION_ERROR): NaN will be returned'
                CALL FILAMENT_PRINT_INFO(SELF)
                CALL TERMINATE('Fatal error in FILAMENT_PARAMETRISATION_ERROR')
            ENDIF
        ENDIF

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_PARAMETRISATION_ERROR): filament info on exit'
            CALL FILAMENT_PRINT_INFO(SELF)
        ENDIF
    END FUNCTION FILAMENT_PARAMETRISATION_ERROR

    !>  \brief  For each waypoint of the filament, compute an estimate of the parameter value for spline interpolation
    SUBROUTINE FILAMENT_PARAMETRISATION_INITIAL_GUESS(SELF,GUESS_FAL)
        !USE FILAMENTS_DEPRECATED, ONLY: COOS_FIL_TO_SPLINE
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_WP_PATH_LENGTHS_INCREASE
        USE COOS, ONLY : WAYPOINT_WAYPOINT_DISTANCE, WAYPOINTS_PRINT_INFO
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF            !<  Filament
        LOGICAL,            OPTIONAL,       INTENT(IN)      ::  GUESS_FAL       !<  The coordinates for first and last waypoints are not known. Their parameters will be "guessed".
        ! Private variables
        CHARACTER(LEN=*),   PARAMETER   ::  PARAM_SCHEME    =   'CHORD_LENGTH'  !<  Parametrisation scheme to use. See "Spline Parametrisation" document for detailled descriptions. Possible values: MINIMAL, CHORD, CUMULATIVE_CHORD, CENTRIPETAL
        INTEGER                         ::  I
        INTEGER                         ::  FIRST_WP, LAST_WP
        LOGICAL                         ::  GGUESS_FAL
        LOGICAL,            PARAMETER   ::  DEBUG           = .FALSE.
        CHARACTER(LEN=6)                ::  NWP_C
        ! Start work

        ! Check: this subroutine is only really meant for filaments that don't already have path-length estimates or whose path lengths are messed up somehow
        IF (ANY(SELF%WP%PATH_LENGTH .NE. 0.0) .AND. FILAMENT_WP_PATH_LENGTHS_INCREASE(SELF)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_INITIAL_PARAMETRISATION): filament already appears to have path-length estimates'
            WRITE(NWP_C,'(I6)') SELF%NWP
            IF (DEBUG) THEN
                WRITE(*,'(/A4,1X,'//NWP_C//'(I8,1X))'  ) 'WP #',(I,I=1,SELF%NWP)
                WRITE(*,'(A4,1X,'//NWP_C//'(F8.3,1X)/)') 'WP T',SELF%T(:)
            ENDIF
            CALL TERMINATE('Fatal error in FILAMENT_INITIAL_PARAMETRISATION')
        ENDIF

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_PARAMETRISATION_INITIAL_GUESS): waypoints on entry'
            CALL WAYPOINTS_PRINT_INFO(SELF%WP)
        ENDIF

        ! Are we ignoring first and last waypoints?
        IF (PRESENT(GUESS_FAL)) THEN
            GGUESS_FAL = GUESS_FAL
        ELSE
            GGUESS_FAL = .FALSE.
        ENDIF

        ! Work out which are the first and last waypoints to which we will be assigning parameter values
        IF (GGUESS_FAL) THEN
            FIRST_WP = 2
            LAST_WP  = SELF%NWP -1
        ELSE
            FIRST_WP = 1
            LAST_WP  = SELF%NWP
        ENDIF

        ! Parametrise
        SELECT CASE (PARAM_SCHEME)
            !CASE ('MINIMAL')
            !    DO I=FIRST_WP,LAST_WP
            !        TMP(1:3) = COOS_FIL_TO_SPLINE(SELF,SELF%WP(I)%COO(1:3),IGNORE_FIRST_WP=GGUESS_FAL)
            !        SELF%T(I) = TMP(1)
            !    ENDDO
            CASE ('CHORD_LENGTH')
                SELF%T(FIRST_WP)    =   0.0
                DO I=FIRST_WP,LAST_WP-1
                    SELF%T(I+1) = SELF%T(I) + WAYPOINT_WAYPOINT_DISTANCE(SELF%WP(I),SELF%WP(I+1))
                ENDDO
            CASE ('CENTRIPETAL')
                SELF%T(FIRST_WP)    =   0.0
                DO I=FIRST_WP,LAST_WP-1
                    SELF%T(I+1) = SELF%T(I) + SQRT(WAYPOINT_WAYPOINT_DISTANCE(SELF%WP(I),SELF%WP(I+1)))
                ENDDO
            CASE DEFAULT
                WRITE(*,'(A)') '**ERROR(FILAMENT_INITIAL_PARAMETRISATION): unknown parametrisation scheme: ', &
                                TRIM(ADJUSTL(PARAM_SCHEME))
                CALL TERMINATE('Fatal error in FILAMENT_INITIAL_PARAMETRISATION')
        END SELECT

        ! If we need to make guesses for FAL WP
        IF (GGUESS_FAL) THEN
            SELF%T(1)           =   SELF%T(2) - (SELF%T(3)-SELF%T(2))*0.5
            SELF%T(SELF%NWP)    =   SELF%T(SELF%NWP-1) + (SELF%T(SELF%NWP-1)-SELF%T(SELF%NWP-2))*0.5
        ENDIF
        DO I=1,SELF%NWP
            SELF%T(I) = SELF%T(I) + SELF%T(1)
        ENDDO

        ! Sanity checks
        IF (SELF%T(1) .NE. 0.0) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_INITIAL_PARAMETRISATION): sanity check failed, SELF%T(1) = ', SELF%T(1)
            CALL TERMINATE('Fatal error in FILAMENT_INITIAL_PARAMETRISATION')
        ENDIF

        ! Diagnostic message
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_INITIAL_PARAMETRISATION): initial parameters: '
            WRITE(*,'(A4,1X,'//NWP_C//'(I8,1X))'  ) 'WP #',(I,I=1,SELF%NWP)
            WRITE(*,'(A4,1X,'//NWP_C//'(F8.3,1X))') 'WP T',SELF%T(:)
        ENDIF
    END SUBROUTINE FILAMENT_PARAMETRISATION_INITIAL_GUESS

    !>  \brief  Compute the path length from first waypoint to every other waypoint in the filament.
#if defined(SKIP_RUNTIME_CHECKS)
    PURE &
#endif
    SUBROUTINE FILAMENT_UPDATE_PATH_LENGTH(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_PATH_LENGTH
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF    !<  Filament to be updated
        ! Private variables
        INTEGER ::  I
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        CHARACTER(LEN=6) :: NWP_C
        ! Start work
        ! Start at the origin
        SELF%WP(1)%PATH_LENGTH = 0.0
        DO I=2,SELF%NWP
            CALL FILAMENT_EVAL_PATH_LENGTH(SELF,SELF%T(I),SELF%WP(I)%PATH_LENGTH)
        ENDDO
        IF (DEBUG) THEN
#ifndef SKIP_RUNTIME_CHECKS
            WRITE(NWP_C,'(I6)') SELF%NWP
            IF (ANY(SELF%WP%PATH_LENGTH .NE. 0.)) THEN
                ! Sanity check
                DO I=2,SELF%NWP
                    IF (SELF%WP(I)%PATH_LENGTH .LE. SELF%WP(I-1)%PATH_LENGTH) THEN
                        WRITE(*,'(A,'//NWP_C//'(F0.3,X))') '**ERROR(FILAMENT_UPDATE_PATH_LENGTH): weird path lengths: ', &
                                SELF%WP%PATH_LENGTH
                        CALL TERMINATE('Fatal error in FILAMENT_UPDATE_PATH_LENGTH')
                    ENDIF
                ENDDO
            ENDIF
#endif
        ENDIF
    END SUBROUTINE FILAMENT_UPDATE_PATH_LENGTH



END MODULE FILAMENTS_PARAMETRISATION
