!>  \brief  Methods for the filament class related to paths
!!
!!  Paths are arrays of waypoints which lie along the helical path of a filament.
!!
MODULE FILAMENTS_PATH
    USE FILAMENTS_BASE
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE

    CONTAINS

    !>  \brief  Return a waypoint (with X,Y,Z coordinates and Euler angles) of the filament SELF at T
    !!
    FUNCTION WAYPOINT_FROM_T(SELF,T) RESULT(WP)
        USE FLOAT_COMPARISONS, ONLY : DBLES_ARE_DIFFERENT
        USE FILAMENTS_GEOMETRY, ONLY : NEAREST_PREVIOUS_WP_FROM_T, FILAMENT_EVAL_HELICAL_RISE, FILAMENT_EVAL_HELICAL_TWIST, &
                                        FILAMENT_EVAL_CURVATURE, NEAREST_WP_FROM_T, &
                                        FILAMENT_EVAL_OVERTWIST, FILAMENT_EVAL_STRETCH, &
                                        FILAMENT_EVAL_PRINCIPAL_CURVATURES
        USE DIFFERENTIAL_GEOMETRY, ONLY : EULER_FROM_DERIVATIVES
        USE SPLINES, ONLY : INTERPOL_SPLINE_EVAL, INTERPOL_SPLINE_HORNER_EVAL_ALL
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF            !<  Filament
        REAL(KIND=8),                       INTENT(IN)      ::  T               !<  Spline parameter (assumed to be == path length + constant)
        ! Result
        TYPE(WAYPOINT)  ::  WP
        ! Private variables
        INTEGER                 ::  WP_BEFORE, WP_NEAREST
        REAL(KIND=8)            ::  DDX,DDY,DDZ
        REAL(KIND=8)            ::  TT, DIST_TO_NEAREST_WP
        LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
        REAL(KIND=8)            ::  AVERAGE_TWIST_PER_DISTANCE, AVERAGE_DS_PER_DISTANCE
        INTEGER                 ::  FIRST_WP, LAST_WP
        REAL(KIND=8)            ::  C_A, C_B
        ! Start work

        IF (DEBUG) WRITE(*,'(A,F0.6)') '**debug(WAYPOINT_FROM_PATH_LENGTH): input parameter value = ', T

        ! Copy value over
        TT = T

        ! Find between which filament waypoints we are
        WP_BEFORE = NEAREST_PREVIOUS_WP_FROM_T(SELF,TT)
        IF (WP_BEFORE .EQ. 0) THEN
            WRITE(*,'(A,F0.12,A,I0,A,F0.12,A,F0.12)') '**WARNING(WAYPOINT_FROM_T): T (', TT, &
                                                        ') falls beyond the ends of filament ', SELF%ID, &
                                                        ', which goes from ', SELF%T(1), ' to ', SELF%T(SELF%NWP)
            WRITE(*,'(A,G15.6)') '**debug(WAYPOINT_FROM_T): the spacing is ', SPACING(TT)
            !CALL TERMINATE('Fatal error in WAYPOINT_FROM_T')

            ! It could be that we are dealing with a simple rounding error. Let's check and keep going.
            IF (.NOT. DBLES_ARE_DIFFERENT(TT,SELF%T(1))) THEN
                TT = SELF%T(1)
            ENDIF
        ENDIF

        ! Evaluate the splines and their derivatives
        CALL INTERPOL_SPLINE_HORNER_EVAL_ALL(SELF%X_SPLINE,TT,WP%COO(1),WP%DCOO(1),DDX)
        CALL INTERPOL_SPLINE_HORNER_EVAL_ALL(SELF%Y_SPLINE,TT,WP%COO(2),WP%DCOO(2),DDY)
        CALL INTERPOL_SPLINE_HORNER_EVAL_ALL(SELF%Z_SPLINE,TT,WP%COO(3),WP%DCOO(3),DDZ)
        IF (DEBUG) THEN
            WRITE(*,'(A,3(F0.3,1X))') '**debug(WAYPOINT_FROM_PATH_LENGTH):  x, y, z = ', WP%COO(1), WP%COO(2), WP%COO(3)
            WRITE(*,'(A,3(F0.3,1X))') '**debug(WAYPOINT_FROM_PATH_LENGTH): dx,dy,dz = ', WP%DCOO(1), WP%DCOO(2), WP%DCOO(3)
        ENDIF

        ! Work out the Euler angles
        CALL EULER_FROM_DERIVATIVES(WP%DCOO(1),WP%DCOO(2),WP%DCOO(3),WP%EULER%THETA,WP%EULER%PSI)


        ! Helical parameters
        IF (SELF%FRAGMENTED_PHI .OR. SELF%FRAGMENTED_SUBUNIT_NUM) THEN
            CALL NEAREST_WP_FROM_T(SELF,TT,WP_NEAREST,DIST_TO_NEAREST_WP)
            WP%HELIX            =   SELF%WP(WP_NEAREST)%HELIX
            IF (SELF%FRAGMENTED_PHI) THEN
                WP%EULER%PHI    =   SELF%WP(WP_NEAREST)%EULER%PHI + DIST_TO_NEAREST_WP * SELF%WP(WP_NEAREST)%HELIX%TWIST   &
                                                                                 / SELF%WP(WP_NEAREST)%HELIX%RISE
            ELSE
                CALL INTERPOL_SPLINE_EVAL(SELF%PHI_SPLINE,TT,WP%EULER%PHI)
            ENDIF
            IF (SELF%FRAGMENTED_SUBUNIT_NUM) THEN
                WP%SUBUNIT_NUM  =   SELF%WP(WP_NEAREST)%SUBUNIT_NUM + DIST_TO_NEAREST_WP / SELF%WP(WP_NEAREST)%HELIX%RISE
            ELSE
                CALL INTERPOL_SPLINE_EVAL(SELF%S_SPLINE,TT,WP%SUBUNIT_NUM)
            ENDIF
        ELSE
            CALL INTERPOL_SPLINE_EVAL(SELF%PHI_SPLINE,TT,WP%EULER%PHI)
            ! Subunit number
            CALL INTERPOL_SPLINE_EVAL(SELF%S_SPLINE,TT,WP%SUBUNIT_NUM)
            !
            WP%HELIX        =   SELF%WP(WP_BEFORE)%HELIX
            WP%HELIX%RISE   =   FILAMENT_EVAL_HELICAL_RISE(SELF,TT)
            WP%HELIX%TWIST  =   FILAMENT_EVAL_HELICAL_TWIST(SELF,TT)
        ENDIF

        ! Compute local curvature
        WP%CURVATURE    =   FILAMENT_EVAL_CURVATURE(SELF,TT)
        ! Compute the local curvatures along the major axes
        !! TODO: do only this, then compute the curvature as the norm of the two
        CALL FILAMENT_EVAL_PRINCIPAL_CURVATURES(SELF,TT,C_A,C_B)
        WP%CURVATURE_A = C_A
        WP%CURVATURE_B = C_B
        ! For the overtwist and overstretch, we need the averate twist and rise per distance along the filament
        ! TODO: don't do this here, it should only be done once per filament - this is a waste!
        ! Also, fix the first and last wp
        !> \todo: don't do this here
        FIRST_WP = 1
        LAST_WP = SELF%NWP
        AVERAGE_TWIST_PER_DISTANCE  =   &
                            (SELF%WP(LAST_WP)%EULER%PHI - SELF%WP(FIRST_WP)%EULER%PHI) / (SELF%T(LAST_WP) - SELF%T(FIRST_WP))
        AVERAGE_DS_PER_DISTANCE     =   &
                            (SELF%WP(LAST_WP)%SUBUNIT_NUM - SELF%WP(FIRST_WP)%SUBUNIT_NUM) / (SELF%T(LAST_WP) - SELF%T(FIRST_WP))
        WP%OVERTWIST    =   FILAMENT_EVAL_OVERTWIST(SELF,TT,AVERAGE_TWIST_PER_DISTANCE)
        WP%STRETCH      =   FILAMENT_EVAL_STRETCH(SELF,TT,AVERAGE_DS_PER_DISTANCE)

        WP%REFINE       =   SELF%WP(WP_BEFORE)%REFINE .OR. SELF%WP(WP_BEFORE+1)%REFINE
        WP%PATH_LENGTH  =   TT-SELF%T(1)

        IF (DEBUG) THEN
            WRITE(*,'(A,3(F0.3,1X))') '**debug(WAYPOINT_FROM_PATH_LENGTH): phi, theta, psi = ', &
                                        WP%EULER%PHI, WP%EULER%THETA, WP%EULER%PSI
        ENDIF

    END FUNCTION WAYPOINT_FROM_T



    !>  \brief  Return an array of waypoints which describes the path of the filament in the region of its waypoints.
    !!  We make the assumption that near the waypoints, the filament is perfectly straight.
    FUNCTION FILAMENT_GET_PATH_NEAR_WPS(SELF,STEP,NUM_WPS_PER_WP,IGNORE_FAL_WPS,HELIX,STEP_IN_REPEATS,INTG_REPEATS) RESULT(PATH)
        USE RUNTIME_PARAMETERS, ONLY : IS_EVEN
        USE HELICES, ONLY : HELICAL_SYMMETRY_UNIT_CONVERT, HELICAL_SYMMETRY_PRINT_INFO
        USE DIFFERENTIAL_GEOMETRY, ONLY : D_FROM_THETA_PSI
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF                    !<  Filament
        REAL(KIND=8),                       INTENT(IN)      ::  STEP                    !<  Step size (pixels)
        INTEGER,                            INTENT(IN)      ::  NUM_WPS_PER_WP          !<  Each incoming WP will "generate" NUM_WPS_PER_WP waypoints in the output path, including itself. This number does not include symmetry-related views.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  IGNORE_FAL_WPS          !<  If .TRUE., the path will extend form the second to the second-to-last waypoint. .FALSE. by default.
        TYPE(HELICAL_SYMMETRY), OPTIONAL,   INTENT(IN)      ::  HELIX                   !<  Helical parameters. If not given as argument, the helical parameters found in the filament's waypoints will be used instead.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  STEP_IN_REPEATS         !<  Whether the step argument was given in repeats
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  INTG_REPEATS            !<  Return waypoints that coincide exactly with helical repeats
        ! Result
        TYPE(WAYPOINT),         ALLOCATABLE                 ::  PATH(:)                 !<  Array of waypoints
        ! Private variables
        LOGICAL                 ::  IIGNORE_FAL_WPS
        INTEGER                 ::  CURRENT_WP_NUM, FIRST_WP, LAST_WP
        REAL(KIND=8)            ::  DIST_TO_INPUT_WP
        INTEGER                 ::  NUM_OUTPUT_WPS, CURRENT_OUTPUT_WP
        INTEGER                 ::  IERR
        INTEGER                 ::  I
        REAL(KIND=8)            ::  D(3)
        INTEGER                 ::  AXIAL_SYM
        LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
        REAL                    ::  CURRENT_TWIST, CURRENT_RISE
        REAL(KIND=8)            ::  SSTEP_IN_PIXELS
        INTEGER                 ::  CURRENT_WP_OUTPUT_WPS_START, CURRENT_WP_OUTPUT_WPS_FINISH
        LOGICAL                 ::  IINTG_REPEATS, SSTEP_IN_REPEATS
        TYPE(WAYPOINT)          ::  CURRENT_WP
        REAL(KIND=8)            ::  CURRENT_T
        ! Start work

        ! Not coded for microtubules yet
        IF (SELF%MICROTUBULE) THEN
            CALL TERMINATE('Not coded for microtubules yet')
        ENDIF

        ! Not coded for symmetry yet
        IF (SELF%HELIX%PERPENDICULAR_SYMMETRY .GT. 1) THEN
            CALL TERMINATE('Not coded for perpendicular symmetry yet')
        ENDIF
        IF (SELF%HELIX%AXIAL_SYMMETRY .GT. 1) THEN
            WRITE(*,'(2A)') '**warning(FILAMENT_GET_PATH_NEAR_WPS): this routine has not yet been tested ',  &
                            'with axial symmetry (remove this warning when it has)'
        ENDIF

        ! NUM_WPS_PER_WP is assumed to be odd
        IF (IS_EVEN(NUM_WPS_PER_WP)) THEN
            WRITE(*,'(A,I0)') '**ERROR(FILAMENT_GET_PATH_NEAR_WPS): NUM_WPS_PER_WP is not odd: ', NUM_WPS_PER_WP
            CALL TERMINATE('Fatal error in FILAMENT_GET_PATH_NEAR_WPS')
        ENDIF

        ! We need the input waypoints' helical parameters to be in pixel units
        DO I=1,SELF%NWP
            IF (.NOT. SELF%WP(I)%HELIX%PIXEL_UNITS) THEN
                CALL HELICAL_SYMMETRY_UNIT_CONVERT(SELF%WP(I)%HELIX,PIXEL_SIZE=SELF%FILM%PIXEL_SIZE)
            ENDIF
        ENDDO

        ! Ignore first and last waypoints?
        IIGNORE_FAL_WPS = .FALSE.
        IF (PRESENT(IGNORE_FAL_WPS)) IIGNORE_FAL_WPS = IGNORE_FAL_WPS


        ! Work out the start and finish waypoints
        IF (IIGNORE_FAL_WPS) THEN
            FIRST_WP    = 2
            LAST_WP     = SELF%NWP - 1
        ELSE
            FIRST_WP    = 1
            LAST_WP     = SELF%NWP
        ENDIF

        ! If helical parameters are supplied, make sure they're in pixel units
        IF (PRESENT(HELIX)) THEN
            IF (.NOT. HELIX%PIXEL_UNITS) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_GET_PATH_NEAR_WPS): helical parameters not in pixel units'
                CALL TERMINATE('Fatal error in FILAMENT_GET_PATH_NEAR_WPS')
            ENDIF
            WRITE(*,'(A)',ADVANCE='NO') '**info(FILAMENT_GET_PATH_NEAR_WPS): using the following helical parameters: '
            CALL HELICAL_SYMMETRY_PRINT_INFO(HELIX)
        ELSE
            WRITE(*,'(2A)') '**WARNING(FILAMENT_GET_PATH_NEAR_WPS): local helical parameters will be used, ',&
                            'which may not be what you want, particularly if you are refining waypoints as single particles'
        ENDIF

        ! Step in repeats?
        SSTEP_IN_REPEATS = .FALSE.
        IF (PRESENT(STEP_IN_REPEATS)) SSTEP_IN_REPEATS = STEP_IN_REPEATS

        ! Return waypoints that coincide with subunits?
        IINTG_REPEATS = .FALSE.
        IF (PRESENT(INTG_REPEATS)) IINTG_REPEATS = INTG_REPEATS
        IF (IINTG_REPEATS .AND. .NOT. SSTEP_IN_REPEATS) THEN
            WRITE(*,'(2A)') '**ERROR(FILAMENT_GET_PATH_NEAR_WPS): option to return waypoints on ',  &
                            'repeats only available when step size given in repeats'
            CALL TERMINATE('Feature not yet implemented')
        ENDIF

        ! Allocate memory for the output
        NUM_OUTPUT_WPS  =   (LAST_WP-FIRST_WP+1-1)*NUM_WPS_PER_WP*SELF%HELIX%AXIAL_SYMMETRY + 1
        ALLOCATE(PATH(NUM_OUTPUT_WPS),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(FILAMENT_GET_PATH_NEAR_WPS): memory allocation failed with error ', IERR
            CALL TERMINATE('Fatal error in FILAMENT_GET_PATH_NEAR_WPS')
        ENDIF


        ! Loop over (input) waypoints
        CURRENT_OUTPUT_WP   =   0
        DO CURRENT_WP_NUM = FIRST_WP,LAST_WP
            !
            IF (IINTG_REPEATS) THEN
                CURRENT_T = SELF%T(CURRENT_WP_NUM)
                CURRENT_T = FILAMENT_FIND_T_OF_NEAREST_SUBUNIT(SELF,CURRENT_T)
                CURRENT_WP = WAYPOINT_FROM_T(SELF,CURRENT_T)
            ELSE
                CURRENT_WP = SELF%WP(CURRENT_WP_NUM)
            ENDIF
            ! Work out the twist and rise to use near this waypoint
            IF (PRESENT(HELIX)) THEN
                CURRENT_TWIST = HELIX%TWIST
                CURRENT_RISE  = HELIX%RISE
            ELSE
                CURRENT_TWIST = CURRENT_WP%HELIX%TWIST
                CURRENT_RISE  = CURRENT_WP%HELIX%RISE
            ENDIF
            ! Debug: printout helical parameters at current waypoint
            IF (DEBUG .AND. .NOT. PRESENT(HELIX)) THEN
                WRITE(*,'(A,I0,A,I0,A)',ADVANCE='NO') '**debug(FILAMENT_GET_PATH_NEAR_WPS): filament ', SELF%ID, &
                                                        ', waypoint ', CURRENT_WP_NUM, ', helical parameters: '
                CALL HELICAL_SYMMETRY_PRINT_INFO(CURRENT_WP%HELIX)
            ENDIF
            ! If the step size was given in helical subunits, convert to pixels
            SSTEP_IN_PIXELS = STEP
            IF (SSTEP_IN_REPEATS) SSTEP_IN_PIXELS = STEP * CURRENT_RISE

            ! The derivatives at the current input waypoint
            D = D_FROM_THETA_PSI(CURRENT_WP%EULER%THETA,CURRENT_WP%EULER%PSI)

            ! If we're on the first or last waypoint, we don't extend further
            IF (CURRENT_WP_NUM .EQ. FIRST_WP) THEN
                CURRENT_WP_OUTPUT_WPS_START = (NUM_WPS_PER_WP-1)/2 + 1
            ELSE
                CURRENT_WP_OUTPUT_WPS_START = 1
            ENDIF
            IF (CURRENT_WP_NUM .EQ. LAST_WP) THEN
                CURRENT_WP_OUTPUT_WPS_FINISH = (NUM_WPS_PER_WP-1)/2 + 1
            ELSE
                CURRENT_WP_OUTPUT_WPS_FINISH = NUM_WPS_PER_WP
            ENDIF

            IF (DEBUG) THEN
                WRITE(*,'(A,I0,A,2(I0,1X))') '**debug(FILAMENT_GET_PATH_NEAR_WPS): on waypoint ', CURRENT_WP_NUM, &
                                            ' the start and finish output wp are: ', CURRENT_WP_OUTPUT_WPS_START, &
                                            CURRENT_WP_OUTPUT_WPS_FINISH
            ENDIF

            ! Loop over the output waypoints
            DO I=CURRENT_WP_OUTPUT_WPS_START,CURRENT_WP_OUTPUT_WPS_FINISH
                ! Work out distance (pixels) from current output WP to current input WP. Negative if output waypoint is before input waypoint on the filament.
                DIST_TO_INPUT_WP    =   SSTEP_IN_PIXELS * REAL(I - ((NUM_WPS_PER_WP-1)/2+1))
                DO AXIAL_SYM=1,SELF%HELIX%AXIAL_SYMMETRY
                    ! The output WP we're currently working on
                    CURRENT_OUTPUT_WP   =   CURRENT_OUTPUT_WP + 1
                    ! Theta and psi are the same as the input waypoint
                    PATH(CURRENT_OUTPUT_WP)%EULER%THETA =   CURRENT_WP%EULER%THETA
                    PATH(CURRENT_OUTPUT_WP)%EULER%PSI   =   CURRENT_WP%EULER%PSI
                    ! The coordinates only need to be worked out for the first, they can then be copied to the axial-symmetry-related views.
                    ! The phi angle needs to be computed for all waypoints
                    IF (AXIAL_SYM .EQ. 1) THEN
                        ! Work out the coordinates of the output waypoint
                        PATH(CURRENT_OUTPUT_WP)%COO(1:3)    =   DIST_TO_INPUT_WP * D(1:3) + CURRENT_WP%COO(1:3)
                        ! Phi needs to be worked out
                        PATH(CURRENT_OUTPUT_WP)%EULER%PHI   =   CURRENT_WP%EULER%PHI &
                                                                + DIST_TO_INPUT_WP * CURRENT_TWIST / CURRENT_RISE
                    ELSE
                        ! Just copy the coordinates over
                        PATH(CURRENT_OUTPUT_WP)%COO(1:3)    =   PATH(CURRENT_OUTPUT_WP-AXIAL_SYM+1)%COO(1:3)
                        ! Phi needs to be worked out for the axial-symmetry-related views
                        PATH(CURRENT_OUTPUT_WP)%EULER%PHI   =   PATH(CURRENT_OUTPUT_WP-AXIAL_SYM+1)%EULER%PHI + &
                                                                REAL(AXIAL_SYM-1)*360./SELF%HELIX%AXIAL_SYMMETRY
                    ENDIF
                    ! Helical parameters are the same
                    PATH(CURRENT_OUTPUT_WP)%HELIX       =   CURRENT_WP%HELIX
                    PATH(CURRENT_OUTPUT_WP)%HELIX%TWIST =   CURRENT_TWIST
                    PATH(CURRENT_OUTPUT_WP)%HELIX%RISE  =   CURRENT_RISE
                    ! Refine flag
                    PATH(CURRENT_OUTPUT_WP)%REFINE      =   CURRENT_WP%REFINE
                    ! Path length reset to 0.0
                    PATH(CURRENT_OUTPUT_WP)%PATH_LENGTH =   0.0
                ENDDO
            ENDDO
        ENDDO
    END FUNCTION FILAMENT_GET_PATH_NEAR_WPS



    !>  \brief  Return an array of waypoints along the filament's axial space curve, with specified step size in pixels
    !!  \todo   Rename IGNORE_FAL_WPS to IGNORE_FAL_SEGMENTS
    SUBROUTINE FILAMENT_GET_PATH_PIXEL_STEPS(   SELF,STEP,PATH,    &
                                                IGNORE_FAL_WPS,INCLUDE_FINISH_WP,   &
                                                START_WP,FINISH_WP,INCLUDE_EXTRA_WP,INCLUDE_FAL_WPS, &
                                                START_T,FINISH_T,STRICT_START_FINISH_T)
        USE COOS, ONLY : WAYPOINTS_GROW_ARRAY, WAYPOINTS_PRINT_INFO
        USE SPLINES, ONLY : INTERPOL_SPLINE_FIND_ROOT
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF                    !<  Filament object
        REAL(KIND=8),                       INTENT(IN)      ::  STEP                    !<  Step size (pixels)
        TYPE(WAYPOINT),         ALLOCATABLE,INTENT(OUT)     ::  PATH(:)                 !<  Resulting array of waypoints
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  IGNORE_FAL_WPS          !<  Whether to ignore segments between the first and second and penultimate and last waypoints
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  INCLUDE_FINISH_WP       !<  Whether the last "active" waypoint should be included in the output path. Defaults to false.
        INTEGER,                OPTIONAL,   INTENT(IN)      ::  START_WP                !<  Return waypoints only between this and the FINISH_WP waypoints
        INTEGER,                OPTIONAL,   INTENT(IN)      ::  FINISH_WP               !<  Return waypoints only between START_WP and this waypoint
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  INCLUDE_EXTRA_WP        !<  Add an extra waypoint at the end, which may be beyond the finish_wp
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  INCLUDE_FAL_WPS         !<  Even if the first and last segments were ignored, the FAL WPS can be included in the output path. Defaults to FALSE.
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  START_T                 !<  Overrides start_wp. Defines the starting point of the path.
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  FINISH_T                !<  Overrides finish_wp. Defines the finishing point of the path.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  STRICT_START_FINISH_T   !<  Ensure that no waypoint will be allowed beyond those T values. By default, there is some flexibility there.

        ! Private variables
        LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
        REAL(KIND=8)                ::  CURRENT_T, SSTART_T, FFINISH_T
        INTEGER                     ::  CURRENT_OUTPUT_WP
        INTEGER                     ::  SSTART_WP, FFINISH_WP
        LOGICAL                     ::  IIGNORE_FAL_WPS
        LOGICAL                     ::  IINCLUDE_FINISH_WP
        LOGICAL                     ::  LAST_OUTPUT_WP
        LOGICAL                     ::  IINCLUDE_EXTRA_WP
        LOGICAL                     ::  IINCLUDE_FAL_WPS
        TYPE(WAYPOINT), ALLOCATABLE ::  PATH_TMP(:)
        LOGICAL,        PARAMETER   ::  ROUND_TO_NEAREST_SUBUNIT = .FALSE.
        INTEGER                     ::  PREDICTED_NUM_OUTPUT_WPS
        LOGICAL                     ::  SSTRICT_START_FINISH_T
        ! Start work

        ! Ignore first and last waypoints?
        IIGNORE_FAL_WPS = .FALSE.
        IF (PRESENT(IGNORE_FAL_WPS)) IIGNORE_FAL_WPS = IGNORE_FAL_WPS


        ! Will we force the finish WP to be included in the output path?
        IINCLUDE_FINISH_WP = .FALSE.
        IF (PRESENT(INCLUDE_FINISH_WP)) IINCLUDE_FINISH_WP = INCLUDE_FINISH_WP

        ! Will we add one output waypoint past the finish_wp
        IINCLUDE_EXTRA_WP = .FALSE.
        IF (PRESENT(INCLUDE_EXTRA_WP)) IINCLUDE_EXTRA_WP = INCLUDE_EXTRA_WP

        ! Include first and last waypoints (even if the first and last segments are ignored)
        IINCLUDE_FAL_WPS = .FALSE.
        IF (PRESENT(INCLUDE_FAL_WPS)) IINCLUDE_FAL_WPS = INCLUDE_FAL_WPS

        ! Work out the start and finish waypoints for the path
        IF (IIGNORE_FAL_WPS) THEN
            SSTART_WP   = 2
            FFINISH_WP  = SELF%NWP - 1
        ELSE
            SSTART_WP   = 1
            FFINISH_WP  = SELF%NWP
        ENDIF

        !IF (DEBUG) THEN
        !   WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_PIXEL_STEPS): filament info: '
        !   CALL FILAMENT_PRINT_INFO(SELF)
        !ENDIF

        ! The user may have supplied their own START and FINISH WPs, which will override
        IF (PRESENT(START_WP)) THEN
            IF (START_WP .EQ. 1 .AND. IIGNORE_FAL_WPS) THEN
                WRITE(*,'(A)') '**warning(FILAMENT_GET_PATH_PIXEL_STEPS): you asked for path starting at waypoint 1 '//&
                                'but specified that FAL WPs should be ignored. Will start at waypoint 2.'
                SSTART_WP = 2
            ELSE
                SSTART_WP = START_WP
            ENDIF
        ENDIF
        IF (PRESENT(FINISH_WP)) THEN
            IF (FINISH_WP .EQ. SELF%NWP .AND. IIGNORE_FAL_WPS) THEN
                WRITE(*,'(A)') '**warning(FILAMENT_GET_PATH_PIXEL_STEPS): you asked for path finishing at last waypoint '//&
                                'but specified that FAL WPs should be ignored. Will finish at penultimate waypoint.'
                FFINISH_WP = SELF%NWP-1
            ELSE
                FFINISH_WP = FINISH_WP
            ENDIF
        ENDIF

        ! The user may have supplied their own start and finish Ts, which will override start and finish wp
        SSTART_T = SELF%T(SSTART_WP)
        IF (PRESENT(START_T)) SSTART_T = START_T
        FFINISH_T = SELF%T(FFINISH_WP)
        IF (PRESENT(FINISH_T)) FFINISH_T = FINISH_T

        ! Start, finish T strictly enforced?
        SSTRICT_START_FINISH_T = .FALSE.
        IF (PRESENT(STRICT_START_FINISH_T)) SSTRICT_START_FINISH_T = STRICT_START_FINISH_T

        ! Sanity check: finish T must be greater than or equal to start T
        IF (SSTART_T .GT. FFINISH_T) THEN
            WRITE(*,'(A,2(F0.6,1X))') '**ERROR(FILAMENT_GET_PATH_PIXEL_STEPS): sstart_t > ffinish_t: ', SSTART_T, FFINISH_T
            CALL TERMINATE('Fatal error in FILAMENT_GET_PATH_PIXEL_STEPS')
        ENDIF


        ! Allocate memory for the path
        IF (ALLOCATED(PATH)) DEALLOCATE(PATH)

        IF (DEBUG) THEN
            WRITE(*,'(A,3(F0.10,1X))') '**debug(FILAMENT_GET_PATH_PIXEL_STEPS): start_t, finish_t, step: ', &
                                                SSTART_T, FFINISH_T, STEP
            ! We can predict how many waypoints we expect (ignoring for now FAL waypoints we may have been asked to add anyway)
            PREDICTED_NUM_OUTPUT_WPS = INT((FFINISH_T-SSTART_T)/STEP) + 1
            WRITE(*,'(A,I0)') '**debug(FILAMENT_GET_PATH_PIXEL_STEPS): predicted number of output waypoints = ', &
                                    PREDICTED_NUM_OUTPUT_WPS
        ENDIF

        CURRENT_T           =   SSTART_T
        ! We may need to explicitely add the first waypoint
        IF (IIGNORE_FAL_WPS .AND. IINCLUDE_FAL_WPS) THEN
            CALL WAYPOINTS_GROW_ARRAY(PATH,INCREMENT=1)
            IF (ALLOCATED(PATH_TMP)) DEALLOCATE(PATH_TMP)
            ALLOCATE(PATH_TMP(1))
            PATH_TMP(1) = SELF%WP(1)
            PATH = PATH_TMP
            CURRENT_OUTPUT_WP   =   1
            IF (DEBUG) PREDICTED_NUM_OUTPUT_WPS = PREDICTED_NUM_OUTPUT_WPS + 1
        ELSE
            CURRENT_OUTPUT_WP   =   0
        ENDIF

        ! Start looping
        LAST_OUTPUT_WP      =   .FALSE.
        DO
            IF (CURRENT_T .GT. FFINISH_T) THEN
                IF (SSTRICT_START_FINISH_T) EXIT
                IF (IINCLUDE_FINISH_WP) THEN
                    CURRENT_T = FFINISH_T
                    LAST_OUTPUT_WP = .TRUE.
                ELSE IF (IINCLUDE_EXTRA_WP) THEN
                    IF (CURRENT_T .LT. SELF%T(SELF%NWP)) THEN
                        LAST_OUTPUT_WP = .TRUE.
                    ELSE
                        EXIT
                    ENDIF
                ELSE IF (CURRENT_T-FFINISH_T .LT. STEP*0.5d0 .AND. CURRENT_T .LT. SELF%T(SELF%NWP)) THEN
                    LAST_OUTPUT_WP = .TRUE.
                ELSE
                    ! If we're past the last waypoint and we've not been asked to include the last waypoint in the path, exit without adding
                    EXIT
                ENDIF
            ELSE IF (CURRENT_T .EQ. FFINISH_T) THEN
                LAST_OUTPUT_WP = .TRUE.
                EXIT
            ENDIF
            IF (ROUND_TO_NEAREST_SUBUNIT) THEN
                CURRENT_T = FILAMENT_FIND_T_OF_NEAREST_SUBUNIT(SELF,CURRENT_T)
            ENDIF
            ! Add a waypoint to the end of the output path, without symmetry related views
            CALL ADD_WAYPOINT_TO_PATH(PATH,SELF,CURRENT_T)
            CURRENT_OUTPUT_WP = CURRENT_OUTPUT_WP + 1
            IF (LAST_OUTPUT_WP) EXIT
            CURRENT_T = CURRENT_T + STEP
        ENDDO

        ! We may need to explicitely add the last waypoint
        IF (IIGNORE_FAL_WPS .AND. IINCLUDE_FAL_WPS) THEN
            PATH_TMP = PATH
            CALL WAYPOINTS_GROW_ARRAY(PATH_TMP,INCREMENT=1)
            PATH_TMP(SIZE(PATH_TMP)) = SELF%WP(SIZE(SELF%WP))
            PATH = PATH_TMP
            CURRENT_OUTPUT_WP = CURRENT_OUTPUT_WP + 1
        ENDIF

        ! There are cases where the two last waypoints may be in the same place, which is wasteful and can cause problems elsewhere
        IF (SIZE(PATH) .GT. 1) THEN
            IF (PATH(SIZE(PATH))%PATH_LENGTH - PATH(SIZE(PATH)-1)%PATH_LENGTH .LE. 0.1D0) THEN
                ! Remove the penultimate point in the path
                PATH_TMP = PATH
                DEALLOCATE(PATH)
                ALLOCATE(PATH(SIZE(PATH_TMP)-1))
                DO CURRENT_OUTPUT_WP=1,SIZE(PATH_TMP)-2
                    PATH(CURRENT_OUTPUT_WP) = PATH_TMP(CURRENT_OUTPUT_WP)
                ENDDO
                ! The last point
                PATH(SIZE(PATH)) = PATH_TMP(SIZE(PATH_TMP))
            ENDIF
        ENDIF

        IF (DEBUG) THEN
            WRITE(*,'(A,I0,A)') '**debug(FILAMENT_GET_PATH_PIXEL_STEPS): full filament path with ', SIZE(PATH), ' waypoints'
            CALL WAYPOINTS_PRINT_INFO(PATH)
            WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_PIXEL_STEPS): all done'
        ENDIF

        ! Sanity check
        IF (SIZE(PATH) .NE. CURRENT_OUTPUT_WP) THEN
            WRITE (*,'(A,2(I0,1X))') '**ERROR(FILAMENT_GET_PATH_PIXEL_STEPS): failed sanity check: ', SIZE(PATH), CURRENT_OUTPUT_WP
            CALL TERMINATE('Fatal error in FILAMENT_GET_PATH_PIXEL_STEPS')
        ENDIF

    END SUBROUTINE FILAMENT_GET_PATH_PIXEL_STEPS

    !>  \brief  Find the position of the nearest subunit to the point at T along the filament
    FUNCTION FILAMENT_FIND_T_OF_NEAREST_SUBUNIT(SELF,T) RESULT(T_OF_NEAREST_SUBUNIT)
        USE SPLINES, ONLY : INTERPOL_SPLINE_FIND_ROOT, INTERPOL_SPLINE_EVAL
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(INOUT)   ::  SELF    !< Filament object
        REAL(KIND=8),   INTENT(IN)      ::  T
        ! Result
        REAL(KIND=8)                    ::  T_OF_NEAREST_SUBUNIT
        ! Private variables
        REAL(KIND=8)                    ::  SUBUNIT_NUM, NEAREST_SUBUNIT
        ! Start work

        ! What subunit_num are we on?
        CALL INTERPOL_SPLINE_EVAL(SELF%S_SPLINE,T,SUBUNIT_NUM)

        ! What's the nearest subunit?
        NEAREST_SUBUNIT = ANINT(SUBUNIT_NUM,KIND=8)

        ! We don't allow ourselves to go beyond the first or last subunit
        NEAREST_SUBUNIT = MAX(NEAREST_SUBUNIT,DBLE(CEILING(SELF%S_SPLINE%Y(1))))
        NEAREST_SUBUNIT = MIN(NEAREST_SUBUNIT,DBLE(FLOOR(SELF%S_SPLINE%Y(SELF%NWP))))

        ! We look for the T corresponding to the nearest integer of subunit_num
        T_OF_NEAREST_SUBUNIT = INTERPOL_SPLINE_FIND_ROOT(SELF%S_SPLINE,NEAREST_SUBUNIT)
    END FUNCTION FILAMENT_FIND_T_OF_NEAREST_SUBUNIT

    !>  \brief  Return an array of waypoints along the filament's axial space curve, with the step size specified in subunits. By default, all symmetry-related subunits are returned
    !!  \todo   This subroutine could be made much shorter by calling the newer INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE, as is done in FILAMENT_GET_PATH_CROSSOVERS
    SUBROUTINE FILAMENT_GET_PATH_SUBUNIT_STEPS(SELF,PATH,STEP,IGNORE_FAL_WPS,IGNORE_SYMMETRY_MATES,START_T,FINISH_T)
        USE PROCFS, ONLY : PRINT_MEMORY_USAGE_STATS
        USE SPLINES, ONLY : INTERPOL_SPLINE_FIND_ROOT, INTERPOL_SPLINE_PRINT_INFO, INTERPOL_SPLINE_EVAL
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF                    !<  Filament object
        TYPE(WAYPOINT),      ALLOCATABLE,   INTENT(INOUT)   ::  PATH(:)                 !<  Array of waypoints
        INTEGER,                            INTENT(IN)      ::  STEP                    !<  Step size (subunits)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  IGNORE_FAL_WPS          !<  Whether to ignore segments between the first and second and penultimate and last waypoints
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  IGNORE_SYMMETRY_MATES   !<  False by default. If true, symmetry-related subunits (and in the microtubule case, tubulin dimers from the other protofilaments) will not be included
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  START_T                 !<  The point on the filament at which we should start looking for waypoints. Overrides IGNORE_FAL_WPS.
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  FINISH_T                !<  The point on the filament at which we should stop looking for waypoints. Override IGNORE_FAL_WPS.
        ! Private variables
        LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
        REAL(KIND=8)                ::  CURRENT_T, SSTART_T, FFINISH_T, NEXT_T, PREVIOUS_T
        REAL(KIND=8)                ::  CURRENT_S, START_S, FINISH_S, NEXT_S, PREVIOUS_S
        INTEGER                     ::  START_WP, FINISH_WP
        INTEGER                     ::  CURRENT_OUTPUT_WP
        LOGICAL                     ::  IIGNORE_FAL_WPS
        ! Start work

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): entry'
        ENDIF

        ! Ignore first and last waypoints?
        IF (PRESENT(IGNORE_FAL_WPS)) THEN
            IIGNORE_FAL_WPS = IGNORE_FAL_WPS
        ELSE
            IIGNORE_FAL_WPS = .FALSE.
        ENDIF

        ! Work out the start and finish waypoints for the path
        IF (IIGNORE_FAL_WPS) THEN
            START_WP    = 2
            FINISH_WP   = SELF%NWP - 1
        ELSE
            START_WP    = 1
            FINISH_WP   = SELF%NWP
        ENDIF

        !
        IF (PRESENT(IGNORE_SYMMETRY_MATES)) THEN
            IF (IGNORE_SYMMETRY_MATES) THEN
                WRITE(*,'(A)') '**warning(FILAMENT_GET_PATH_SUBUNIT_STEPS): option to ignore symmetry mates is not implemented yet'
                !CALL TERMINATE('Not implemented')
            ENDIF
        ENDIF


        ! Allocate memory for the path
        IF (ALLOCATED(PATH)) DEALLOCATE(PATH)

        !IF (DEBUG) THEN
        !   WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): filament info upon entry'
        !   CALL FILAMENT_PRINT_INFO(SELF)
        !ENDIF


        ! Define the starting point for our search
        IF (PRESENT(START_T)) THEN
            SSTART_T        =   START_T
        ELSE
            ! We start from the first waypoint, find the nearest subunit, then start the walk from there, along the full filament till the last waypoint
            SSTART_T        =   SELF%T(START_WP)
        ENDIF
        CALL INTERPOL_SPLINE_EVAL(SELF%S_SPLINE,SSTART_T,START_S)
        START_S             =   CEILING(START_S)
        SSTART_T            =   INTERPOL_SPLINE_FIND_ROOT(SELF%S_SPLINE,START_S)
        IF (SSTART_T .LT. SELF%T(START_WP)) THEN
            WRITE(*,'(A,G15.6)') '**warning(FILAMENT_GET_PATH_SUBUNIT_STEPS): start_t is before T for first waypoint, by ', &
                                    SELF%T(START_WP) - SSTART_T
            IF (SELF%T(START_WP) - SSTART_T .LT. 1.0) THEN
                WRITE(*,'(2A)') '**warning(FILAMENT_GET_PATH_SUBUNIT_STEPS): this is probably due to rounding errors ',  &
                                '(check the difference above is very small). Overriding.'
                SSTART_T = SELF%T(START_WP)
            ELSE
                WRITE(*,'(A)') '**ERROR(FILAMENT_GET_PATH_SUBUNIT_STEPS): this is not a tolerable error. Crashing out.'
                CALL TERMINATE('Fatal error in FILAMENT_GET_PATH_SUBUNIT_STEPS')
            ENDIF
        ENDIF

        ! Define the end point for our search
        IF (PRESENT(FINISH_T)) THEN
            FFINISH_T       =   FINISH_T
        ELSE
            FFINISH_T       =   SELF%T(FINISH_WP)
        ENDIF
        CALL INTERPOL_SPLINE_EVAL(SELF%S_SPLINE,FFINISH_T,FINISH_S)

        IF (DEBUG) THEN
            WRITE(*,'(A,2(F0.3,1X))') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): start, finish T: ', SSTART_T, FFINISH_T
            WRITE(*,'(A,2(F0.3,1X))') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): start, finish S: ', START_S, FINISH_S
            WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): S spline info:'
            CALL INTERPOL_SPLINE_PRINT_INFO(SELF%S_SPLINE)
        ENDIF

        ! If the helix is fragmented (i.e. with discontinuities between waypoints, we need to call another function
        IF (SELF%FRAGMENTED_PHI .OR. SELF%FRAGMENTED_SUBUNIT_NUM) THEN
            CALL FILAMENT_GET_PATH_SUBUNIT_STEPS_FRAG_HELIX(SELF,START_WP,FINISH_WP,PATH)
        ELSE
            ! The helix is not fragmented
            ! Initialise the search
            CURRENT_T           =   SSTART_T
            CURRENT_S           =   START_S
            NEXT_S              =   CURRENT_S + DBLE(STEP)
            NEXT_T              =   INTERPOL_SPLINE_FIND_ROOT(SELF%S_SPLINE,NEXT_S)
            PREVIOUS_S          =   CURRENT_S - DBLE(STEP)
            CURRENT_OUTPUT_WP   =   0

            ! Memory usage before the search starts
            IF (DEBUG) THEN
                WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): memory usage before search'
                CALL PRINT_MEMORY_USAGE_STATS()
            ENDIF

            ! Start the search
            DO
                IF (DEBUG) WRITE(*,'(A,F0.3)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): S is = ', CURRENT_S

                ! Update previous, current and next S values
                PREVIOUS_S = CURRENT_S
                CURRENT_S = NEXT_S
                NEXT_S = NEXT_S + DBLE(STEP)

                ! Check whether the S values are still within range
                IF (NEXT_S .GT. FINISH_S ) THEN
                    EXIT
                ELSE IF (PREVIOUS_S .LT. START_S) THEN
                    WRITE(*,'(A)') '**ERROR(FILAMENT_GET_PATH_SUBUNIT_STEPS): should not reach this point'
                    CALL TERMINATE('Programming error')
                    CYCLE
                ENDIF

                IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): here'
                ! Update previous, current and next T values
                PREVIOUS_T = CURRENT_T
                CURRENT_T = NEXT_T
                IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): finding next_t'
                NEXT_T = INTERPOL_SPLINE_FIND_ROOT(SELF%S_SPLINE,NEXT_S)
                IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): found next_t'
                ! Check whether T values are still within range
                IF (DEBUG) WRITE(*,'(A,I0,A,2(F0.3,1X))')    '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): output waypoint ', &
                                                            CURRENT_OUTPUT_WP+1, ' S, T = ', CURRENT_S, CURRENT_T
                IF (CURRENT_T .GT. FFINISH_T .OR. (SELF%MICROTUBULE .AND. NEXT_T .GT. FFINISH_T)) THEN
                    IF (FINISH_S - CURRENT_S .GT. 0.01) THEN
                        WRITE(*,'(A,G15.4,2(A,F0.3),A)') &
                                        '**WARNING(FILAMENT_GET_PATH_SUBUNIT_STEPS): current_t is beyond finish_t by ', &
                                            CURRENT_T-FFINISH_T, ' but current_s (', CURRENT_S, &
                                            ') is less than finish_s (', FINISH_S, ')'
                    ENDIF
                    EXIT
                ELSE IF (CURRENT_T .LT. SSTART_T) THEN  ! This can happen with rounding errors
                    CURRENT_T = SSTART_T
                ENDIF
                ! Keep track of how many waypoints we'll be adding to path (this only works well with non-microtubules)
                IF (SELF%MICROTUBULE) THEN
                    CURRENT_OUTPUT_WP = CURRENT_OUTPUT_WP + SELF%MT_PF_NUMBER
                ELSE
                    CURRENT_OUTPUT_WP = CURRENT_OUTPUT_WP + SELF%HELIX%AXIAL_SYMMETRY * SELF%HELIX%PERPENDICULAR_SYMMETRY
                ENDIF
                ! Add waypoints to path
                IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): calling add_waypoint_to_path'
                CALL ADD_WAYPOINT_TO_PATH(    PATH,SELF,CURRENT_T,SYMMETRY_MATES=.TRUE.,  &
                                                T_FOR_NEXT_NEW_WP=NEXT_T,T_FOR_PREVIOUS_WP=PREVIOUS_T)
                IF (DEBUG) WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): add_waypoint_to_path returned'


                !IF (DEBUG) THEN
                !    WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): only returning first subunit'
                !    EXIT
                !ENDIF
            ENDDO

            ! Sanity check (only works for non-microtubules, because with microtubules some of the dimers in the first or last washer
            ! can be beyond the end of the filament)
            IF (.NOT. SELF%MICROTUBULE) THEN
                IF (SIZE(PATH) .NE. CURRENT_OUTPUT_WP ) THEN
                    WRITE (*,'(A,2(I0,1X))') '**ERROR(FILAMENT_GET_PATH_SUBUNIT_STEPS): failed sanity check: ', &
                                            SIZE(PATH), CURRENT_OUTPUT_WP
                    CALL TERMINATE('Fatal error in FILAMENT_GET_PATH_SUBUNIT_STEPS')
                ENDIF
            ENDIF
        ENDIF   ! End of check for fragmented helix

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_SUBUNIT_STEPS): returning'
        ENDIF


    END SUBROUTINE FILAMENT_GET_PATH_SUBUNIT_STEPS


    !>  \brief  Return an array of waypoints along the filament's axial space curve, one at each subunit.
    !!          Use when fragmented_helix. All symmetry-related subunits are returned
    SUBROUTINE FILAMENT_GET_PATH_SUBUNIT_STEPS_FRAG_HELIX(SELF,START_WP,FINISH_WP,PATH)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF            !<  Filament object
        INTEGER,                            INTENT(IN)      ::  START_WP        !<  First waypoint
        INTEGER,                            INTENT(IN)      ::  FINISH_WP       !<  Last waypoint
        TYPE(WAYPOINT), ALLOCATABLE,        INTENT(INOUT)   ::  PATH(:)         !<  Array of waypoints
        ! Private variables
        REAL(KIND=8),   PARAMETER   ::  SUBUNIT_NUM_TOL = 0.001D0
        LOGICAL,        PARAMETER   ::  CHECK_SUBUNIT_NUM = .TRUE.
        INTEGER                     ::  CURRENT_ORIGIN_S, SUBUNITS_TO_ORIGIN, CURRENT_WP
        REAL(KIND=8)                ::  CURRENT_START_T, CURRENT_FINISH_T, CURRENT_ORIGIN_T, CURRENT_T
        ! Private variables for debugging
		INTEGER						::	SUBUNIT_COUNT, SUBUNIT_COUNT_CURRENT_WP
		! Start work

		SUBUNIT_COUNT = 0

        ! Loop over filament waypoints
        DO CURRENT_WP=START_WP,FINISH_WP

            ! Work out the start and finish T for the current wp
            CALL FILAMENT_T_NEIGHBOURHOOD_OF_WP(SELF,CURRENT_WP,CURRENT_START_T,CURRENT_FINISH_T)
            IF (CURRENT_WP .EQ. START_WP .OR. CURRENT_WP .EQ. FINISH_WP) CURRENT_START_T = SELF%T(CURRENT_WP)

            ! Work out the T value for the central subunit for the current segment
            IF (CURRENT_WP .EQ. START_WP) THEN
                CURRENT_ORIGIN_S = CEILING(SELF%WP(CURRENT_WP)%SUBUNIT_NUM)
            ELSE IF (CURRENT_WP .EQ. FINISH_WP) THEN
                CURRENT_ORIGIN_S =   FLOOR(SELF%WP(CURRENT_WP)%SUBUNIT_NUM)
            ELSE
                CURRENT_ORIGIN_S =    NINT(SELF%WP(CURRENT_WP)%SUBUNIT_NUM)
            ENDIF
            CURRENT_ORIGIN_T =     SELF%T(CURRENT_WP)   &
                                + (CURRENT_ORIGIN_S-SELF%WP(CURRENT_WP)%SUBUNIT_NUM) * SELF%WP(CURRENT_WP)%HELIX%RISE

            ! Work out the T value for the first subunit of the current segment
            SUBUNITS_TO_ORIGIN =   - FLOOR((CURRENT_ORIGIN_T-CURRENT_START_T)/SELF%WP(CURRENT_WP)%HELIX%RISE)
            CURRENT_T          =     CURRENT_ORIGIN_T     &
                                  +  SUBUNITS_TO_ORIGIN*SELF%WP(CURRENT_WP)%HELIX%RISE

			! Perhaps because of floating point arithm, we may end up just the wrong side of the discontinuity
			CURRENT_T          =     MAX(CURRENT_T,CURRENT_START_T+0.0001D0)
			IF (CURRENT_T .GT. CURRENT_FINISH_T) CYCLE

			SUBUNIT_COUNT_CURRENT_WP = 0

			! Loop over subunits
            DO

                ! Add a waypoint at the current T
                CALL ADD_WAYPOINT_TO_PATH(    PATH,SELF,CURRENT_T,SYMMETRY_MATES=.TRUE.)

				SUBUNIT_COUNT = SUBUNIT_COUNT + 1
				SUBUNIT_COUNT_CURRENT_WP = SUBUNIT_COUNT_CURRENT_WP + 1

				! Check we're on a subunit exactly. If not, print out useful debug info.
                IF ( ABS(PATH(SIZE(PATH))%SUBUNIT_NUM-NINT(PATH(SIZE(PATH))%SUBUNIT_NUM)) .GT. SUBUNIT_NUM_TOL) THEN
                    WRITE(*,'(2A,F0.5)') '**ERROR(FILAMENT_GET_PATH_SUBUNIT_STEPS_FRAG_HELIX): ',   &
                                            'current output subunit_num is not integer: ', PATH(SIZE(PATH))%SUBUNIT_NUM
                    WRITE(*,'(A,I0)') '**debug(): SUBUNIT_COUNT = ', SUBUNIT_COUNT
					WRITE(*,'(A,I0)') '**debug(): SUBUNIT_COUNT_CURRENT_WP = ', SUBUNIT_COUNT_CURRENT_WP
					WRITE(*,'(A,F0.5)') '**debug(): current_t = ', CURRENT_T
					WRITE(*,'(A,F0.5)') '**debug(): distance to current_wp = ', SELF%T(CURRENT_WP) - CURRENT_T
					WRITE(*,'(A,2(F0.5,3X))') '**debug(): discontinuities around current_wp = ', &
																		0.5*(SELF%T(CURRENT_WP)-SELF%T(CURRENT_WP-1)), &
				                                                        0.5*(SELF%T(CURRENT_WP+1)-SELF%T(CURRENT_WP))
					WRITE(*,'(A,2(F0.5,3X))') '**debug(): current_start_t, current_finish_t = ', &
													CURRENT_START_T, CURRENT_FINISH_T
					CALL TERMINATE('Fatal error in FILAMENT_GET_PATH_SUBUNIT_STEPS_FRAG_HELIX. Failed sanity check')
                ENDIF

                ! Increment T
                SUBUNITS_TO_ORIGIN = SUBUNITS_TO_ORIGIN + 1
                CURRENT_T = CURRENT_ORIGIN_T + SUBUNITS_TO_ORIGIN*SELF%WP(CURRENT_WP)%HELIX%RISE

                ! Check we are still in the neighbourhood of the current waypoint
                IF (CURRENT_T .GT. CURRENT_FINISH_T) THEN
                    EXIT
                ENDIF

            ENDDO

        ENDDO

    END SUBROUTINE FILAMENT_GET_PATH_SUBUNIT_STEPS_FRAG_HELIX

    !>  \brief  Define the neighbourhood of the current waypoint (i.e. halfway to the previous and next waypoints
    PURE SUBROUTINE FILAMENT_T_NEIGHBOURHOOD_OF_WP(SELF,WP_NUM,START_T,FINISH_T)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        INTEGER,        INTENT(IN)  ::  WP_NUM
        REAL(KIND=8),   INTENT(OUT) ::  START_T
        REAL(KIND=8),   INTENT(OUT) ::  FINISH_T
        ! Private variables

        ! Start work
        IF (WP_NUM .EQ. 1) THEN
            START_T  = SELF%T(1)
        ELSE
            START_T  = SELF%T(WP_NUM) - (SELF%T(WP_NUM)-SELF%T(WP_NUM-1))*0.5D0
        ENDIF

        IF (WP_NUM .EQ. SELF%NWP) THEN
            FINISH_T = SELF%T(SELF%NWP)
        ELSE
            FINISH_T = SELF%T(WP_NUM) + (SELF%T(WP_NUM+1)-SELF%T(WP_NUM))*0.5D0
        ENDIF
    END SUBROUTINE FILAMENT_T_NEIGHBOURHOOD_OF_WP


    !>  \brief  Return an array of waypoints along the filament's axial space curve, with the step size specified in numbers of crossovers.
    SUBROUTINE FILAMENT_GET_PATH_CROSSOVERS(SELF,PATH,STEP,IGNORE_FAL_WPS,IGNORE_SYMMETRY_MATES,PHI_STEP_SIZE,INCLUDE_FAL_WPS)
        USE SPLINES, ONLY : INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF                    !<  Filament object
        TYPE(WAYPOINT),         ALLOCATABLE,INTENT(INOUT)   ::  PATH(:)                 !<  Output array of waypoints
        INTEGER,                            INTENT(IN)      ::  STEP                    !<  Step size (number of crossovers). Call with STEP=1 to get every crossover.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  IGNORE_FAL_WPS          !<  Whether to ignore segments between the first and second and penultimate and last waypoints
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  IGNORE_SYMMETRY_MATES   !<  False by default. If true, symmetry-related subunits (and in the microtubule case, tubulin dimers from the other protofilaments) will not be included
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  PHI_STEP_SIZE           !<  Phi (twist) step size. Defaults to 180.0 for crossovers.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  INCLUDE_FAL_WPS         !<  Even if the first and last segments were ignored, the FAL WPS can be included in the output path. Defaults to FALSE.
        ! Private variables
        LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
        LOGICAL                     ::  IIGNORE_FAL_WPS, IIGNORE_SYMMETRY_MATES
        INTEGER                     ::  START_WP, FINISH_WP, WP_COUNT
        REAL(KIND=8),   ALLOCATABLE ::  T(:)
        REAL(KIND=8)                ::  START_T, FINISH_T
        REAL(KIND=8)                ::  PPHI_STEP_SIZE
        LOGICAL                     ::  IINCLUDE_FAL_WPS
        ! Start work

        ! Ignore first and last waypoints?
        IIGNORE_FAL_WPS = .FALSE.
        IF (PRESENT(IGNORE_FAL_WPS)) IIGNORE_FAL_WPS = IGNORE_FAL_WPS

        IIGNORE_SYMMETRY_MATES=.TRUE.
        IF (PRESENT(IGNORE_SYMMETRY_MATES)) IIGNORE_SYMMETRY_MATES = IGNORE_SYMMETRY_MATES

        IINCLUDE_FAL_WPS=.FALSE.
        IF (PRESENT(INCLUDE_FAL_WPS)) IINCLUDE_FAL_WPS = INCLUDE_FAL_WPS

        ! Phi step size
        PPHI_STEP_SIZE = 180.0D0
        IF (PRESENT(PHI_STEP_SIZE)) PPHI_STEP_SIZE = PHI_STEP_SIZE

        ! Work out the start and finish waypoints for the path
        IF (IIGNORE_FAL_WPS) THEN
            START_WP    = 2
            FINISH_WP   = SELF%NWP - 1
        ELSE
            START_WP    = 1
            FINISH_WP   = SELF%NWP
        ENDIF

        START_T             =   SELF%T(START_WP)
        FINISH_T            =   SELF%T(FINISH_WP)

        ! Allocate memory for the path
        IF (ALLOCATED(PATH)) DEALLOCATE(PATH)

        IF (IINCLUDE_FAL_WPS) THEN
            CALL ADD_WAYPOINT_TO_PATH(PATH,SELF,SELF%T(1),SYMMETRY_MATES=.NOT. IIGNORE_SYMMETRY_MATES)
        ENDIF

        !IF (DEBUG) THEN
        !   WRITE(*,'(A)') '**debug(FILAMENT_GET_PATH_CROSSOVERS): filament info upon entry'
        !   CALL FILAMENT_PRINT_INFO(SELF)
        !ENDIF

        ! Find T values at which phi is a multiple of PPHI_STEP_SIZE
        T = INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE(SELF%PHI_SPLINE,PPHI_STEP_SIZE)

        ! Loop over the T values and find corresponding waypoints
        DO WP_COUNT=1,SIZE(T)
            IF (T(WP_COUNT) .GE. START_T .AND. T(WP_COUNT) .LE. FINISH_T) THEN
                CALL ADD_WAYPOINT_TO_PATH(PATH,SELF,T(WP_COUNT),SYMMETRY_MATES=.NOT. IIGNORE_SYMMETRY_MATES)
            ENDIF
        ENDDO

        IF (IINCLUDE_FAL_WPS) THEN
            CALL ADD_WAYPOINT_TO_PATH(PATH,SELF,SELF%T(SELF%NWP),SYMMETRY_MATES=.NOT. IIGNORE_SYMMETRY_MATES)
        ENDIF

    END SUBROUTINE FILAMENT_GET_PATH_CROSSOVERS



    !>  \brief  Return an array of coordinates which describes the path of a filament, with specified step size (in Angstroems unless specified otherwise)
    SUBROUTINE FILAMENT_GET_PATH(   SELF,PATH,STEP,STEP_IN_PIXELS,STEP_IN_REPEATS,IGNORE_FAL_WPS,SP_MODE,NUM_WPS_PER_WP,HELIX,&
                                    START_T,FINISH_T,START_WP,FINISH_WP,INCLUDE_EXTRA_WP,IGNORE_SYMMETRY_MATES)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  SELF                    !<  Filament
        TYPE(WAYPOINT),         ALLOCATABLE,INTENT(INOUT)   ::  PATH(:)                 !<  Array of waypoints
        REAL(KIND=8),                       INTENT(IN)      ::  STEP                    !<  Step size (pixels)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  STEP_IN_PIXELS          !<  Specifies (if .TRUE.) that the step is given in pixels
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  STEP_IN_REPEATS         !<  .FALSE. by default. If .TRUE., the STEP is interpreted as being in number of helical subunits, rather than pixels or Angstroms.
                                                                                        !!  So e.g. if STEP=1.0, a path with every subunit will be returned.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  IGNORE_FAL_WPS          !<  If .TRUE., the path will extend form the second to the second-to-last waypoint. .FALSE. by default.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  SP_MODE                 !<  Single particle mode. In single particle mode, it is assumed the waypoints were processed independently
                                                                                        !!  from each other, so we return waypoints in the output path which are near
                                                                                        !!  the input waypoints, with the assumption that the filament is straight near the waypoints.
        INTEGER,                OPTIONAL,   INTENT(IN)      ::  NUM_WPS_PER_WP          !<  In single particle mode, the number of output waypoints per waypoint in the input filament
        TYPE(HELICAL_SYMMETRY), OPTIONAL,   INTENT(IN)      ::  HELIX                   !<  Overriding helical parameters. If not given, helical parameters found in the filament object will be used.
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  START_T                 !<  Overrides IGNORE_FAL_WPS. Sets the point on the filament at which the search for waypoints will be started.
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  FINISH_T                !<  Overrides IGNORE_FAL_WPS. Sets the point on the filament at which the search for waypoints will be ended.
        INTEGER,                OPTIONAL,   INTENT(IN)      ::  START_WP                !<  Path will be restricted to start at this waypoint
        INTEGER,                OPTIONAL,   INTENT(IN)      ::  FINISH_WP               !<  Path will be restricted to finish at/before this waypoint
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  INCLUDE_EXTRA_WP        !<  Return one last waypoint which will be past the finish input WP (but will not be beyond the end of the filament)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  IGNORE_SYMMETRY_MATES   !<  If true, symmetry-related subunits (and in the microtubule case, tubulin dimers from the other protofilaments) will not be included
        ! Private variables
        REAL(KIND=8)                ::  SSTEP_IN_PIXELS         !   The step size, in pixels
        LOGICAL                     ::  SSTEP_IN_REPEATS
        LOGICAL                     ::  IIGNORE_FAL_WPS                         !   Whether to ignore the first and last (FAL) waypoints (WPS)
        LOGICAL,    PARAMETER       ::  FULL_FILAMENT_PATH          =   .TRUE.
        LOGICAL,    PARAMETER       ::  PATH_FROM_EACH_WAYPOINT     =   .FALSE.
        LOGICAL,    PARAMETER       ::  DEBUG   =   .FALSE.
        LOGICAL                     ::  SSP_MODE
        !TYPE(WAYPOINT), ALLOCATABLE ::  PATH_CHECK(:) ! For debugging purposes


        !
        ! Start work
        !

        ! This subroutine does not handle perpendicular helical symmetries yet
        IF (SELF%HELIX%PERPENDICULAR_SYMMETRY .GT. 1) THEN
            WRITE(*,'(A)') '**warning(FILAMENT_GET_PATH): helical symmetries with perpendicular symmetries not ',   &
                            'tested yet. Please check results carefully & remove this warning when done.'
        ENDIF

        ! Is the steps in number of repeats?
        IF (PRESENT(STEP_IN_REPEATS)) THEN
            SSTEP_IN_REPEATS = STEP_IN_REPEATS
        ELSE
            SSTEP_IN_REPEATS = .FALSE.
        ENDIF

        ! Ignore first and last waypoints?
        IF (PRESENT(IGNORE_FAL_WPS)) THEN
            IIGNORE_FAL_WPS = IGNORE_FAL_WPS
        ELSE
            IIGNORE_FAL_WPS = .FALSE.
        ENDIF


        !
        ! Find the step size in pixels. If the user gave STEP_IN_PIXELS=.TRUE., then we don't need to do anything.
        !
        IF (SSTEP_IN_REPEATS) THEN
            ! If the step was given in number of repeats, we first need to update the helical rise information for each segment of the filament
            !CALL FILAMENT_UPDATE_WP_HELIX(SELF) ! Not needed anymore, with the new S spline
            IF (PRESENT(STEP_IN_PIXELS)) THEN
                IF (STEP_IN_PIXELS) THEN
                    WRITE(*,'(A)') '**ERROR(FILAMENT_GET_PATH): step can be given either in number of ',    &
                                    'repeats, or in Angstroms or pixels, not both'
                    CALL TERMINATE('Fatal error in FILAMENT_GET_PATH')
                ENDIF
            ENDIF
        ELSE
            SSTEP_IN_PIXELS = STEP * SELF%FILM%PIXEL_SIZE
            IF (PRESENT(STEP_IN_PIXELS)) THEN
                IF (STEP_IN_PIXELS) THEN
                    SSTEP_IN_PIXELS = STEP
                ENDIF
            ENDIF
        ENDIF

        ! START_WP and FINISH_WP are only implemented with pixel step sizes
        IF (PRESENT(START_WP) .AND. SSTEP_IN_REPEATS) THEN
            WRITE(*,'(A)') '**warning(FILAMENT_GET_PATH): start_wp optional argument only supported for pixel step sizes. '//&
                            'Will be ignored.'
        ENDIF
        IF (PRESENT(FINISH_WP) .AND. SSTEP_IN_REPEATS) THEN
            WRITE(*,'(A)') '**warning(FILAMENT_GET_PATH): finish_wp optional argument only supported for pixel step sizes. '//&
                            'Will be ignored.'
        ENDIF

        !
        IF (PRESENT(INCLUDE_EXTRA_WP) .AND. SSTEP_IN_REPEATS) THEN
            WRITE(*,'(A)') '**warning(FILAMENT_GET_PATH): include_extra_waypoint option is only supported for pixel step sizes. '//&
                            'Will be ignored.'
        ENDIF

        ! Are we in single particle mode?
        SSP_MODE = .FALSE.
        IF (PRESENT(SP_MODE))  SSP_MODE = SP_MODE

        ! If we were given start_t and finish_t, make sure they are sensical
        IF (PRESENT(START_T) .AND. PRESENT(FINISH_T)) THEN
            IF (START_T .GT. FINISH_T) THEN
                WRITE(*,'(A,2(F0.6,1X))') '**ERROR(FILAMENT_GET_PATH): optional arguments start_t and finish_t ',    &
                                        'do not make sense; start_t > finish_t: ', START_T, FINISH_T
                CALL TERMINATE('Fatal error in FILAMENT_GET_PATH')
            ENDIF
        ENDIF


        ! If we're in single particle mode, we need to know how many output WPs we want per input WP
        IF (SSP_MODE) THEN
            IF (.NOT. PRESENT(NUM_WPS_PER_WP)) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_GET_PATH): in SP mode, need to know how many output WPS you want per input WP'
                CALL TERMINATE('Fatal error in FILAMENT_GET_PATH - programming error')
            ENDIF
        ENDIF

        IF (SSP_MODE) THEN
            IF (SSTEP_IN_REPEATS) THEN
                PATH = FILAMENT_GET_PATH_NEAR_WPS(  SELF,STEP,NUM_WPS_PER_WP,IGNORE_FAL_WPS=IIGNORE_FAL_WPS,  &
                                                    HELIX=HELIX,STEP_IN_REPEATS=SSTEP_IN_REPEATS,INTG_REPEATS=.TRUE.)
                !WRITE(*,'(A)') '**debug: below SP path:'
                !CALL WAYPOINTS_PRINT_INFO(PATH )
                ! Debug - let's look at how different the path would have been in FF mode
                !PATH_CHECK = FILAMENT_GET_PATH_SUBUNIT_STEPS( SELF,NINT(STEP),IGNORE_FAL_WPS=IIGNORE_FAL_WPS,  &
                !                                            START_T=START_T,FINISH_T=FINISH_T,  &
                !                                            IGNORE_SYMMETRY_MATES=IGNORE_SYMMETRY_MATES)
                !WRITE(*,'(A)') '**debug: below FF path:'
                !CALL WAYPOINTS_PRINT_INFO(PATH_CHECK)
                !PATH_CHECK = PATH - PATH_CHECK
                !WRITE(*,'(A)') '**debug: below difference between SP path and FF path:'
                !CALL WAYPOINTS_PRINT_INFO(PATH_CHECK)
            ELSE
                PATH = FILAMENT_GET_PATH_NEAR_WPS(  SELF,SSTEP_IN_PIXELS,NUM_WPS_PER_WP,IGNORE_FAL_WPS=IIGNORE_FAL_WPS, &
                                                    HELIX=HELIX,STEP_IN_REPEATS=SSTEP_IN_REPEATS)
            ENDIF
        ELSE
            IF (PRESENT(HELIX)) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENT_GET_PATH): helical parameters override is only intended for single-particle mode'
                CALL TERMINATE('Fatal error in FILAMENT_GET_PATH')
            ENDIF

            IF (FULL_FILAMENT_PATH) THEN
                ! In this mode, we start from the first waypoint, and walk along the full filament till the last waypoint
                IF (SSTEP_IN_REPEATS) THEN
                    CALL FILAMENT_GET_PATH_SUBUNIT_STEPS(   SELF,PATH,NINT(STEP),IGNORE_FAL_WPS=IIGNORE_FAL_WPS,  &
                                                            START_T=START_T,FINISH_T=FINISH_T,  &
                                                            IGNORE_SYMMETRY_MATES=IGNORE_SYMMETRY_MATES)
                ELSE
                    CALL FILAMENT_GET_PATH_PIXEL_STEPS(     SELF,SSTEP_IN_PIXELS,PATH,IIGNORE_FAL_WPS,  &
                                                            START_WP=START_WP,FINISH_WP=FINISH_WP,  &
                                                            INCLUDE_EXTRA_WP=INCLUDE_EXTRA_WP, &
                                                            START_T=START_T,FINISH_T=FINISH_T)
                ENDIF
            ELSE IF (PATH_FROM_EACH_WAYPOINT) THEN
                ! In this mode, we start from each waypoint and travel along the path until
                ! we're halfway to the next/previous waypoint
                CALL TERMINATE('Not implemented yet')
            ELSE
                WRITE(*,'(A)') '**ERROR(FILAMENT_GET_PATH): oops, programming error'
                CALL TERMINATE('Fatal error in FILAMENT_GET_PATH')
            ENDIF
        ENDIF   ! End of test for SP_MODE

    END SUBROUTINE FILAMENT_GET_PATH

    !>  \brief  Add waypoint (and optionally its symmetry-mates) to the end of a path
    SUBROUTINE ADD_WAYPOINT_TO_PATH(OLD_PATH,SELF,T_FOR_NEW_WP,SYMMETRY_MATES,T_FOR_NEXT_NEW_WP,T_FOR_PREVIOUS_WP)
        USE FILAMENTS_MICROTUBULES, ONLY : MICROTUBULE_GET_GOOD_PF
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_EVAL_HELICAL_RISE
        USE COOS, ONLY : WAYPOINTS_GROW_ARRAY, WP_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(WAYPOINT), ALLOCATABLE,    INTENT(INOUT)   ::  OLD_PATH(:)
        TYPE(FILAMENT),                 INTENT(INOUT)   ::  SELF                !<  Filament object
        REAL(KIND=8),                   INTENT(IN)      ::  T_FOR_NEW_WP        !<  The arc length parameter (T) for the waypoint(s) to be added
        LOGICAL,        OPTIONAL,       INTENT(IN)      ::  SYMMETRY_MATES      !<  If .TRUE., symmetry-mates (tubulin dimers in the microtubule case) will also be added to the path. False by default.
        REAL(KIND=8),   OPTIONAL,       INTENT(IN)      ::  T_FOR_NEXT_NEW_WP   !<  The arc length parameter (T) for the waypoint which will be added next (needed only in the MT case, where it helps determine the end of the "washer")
        REAL(KIND=8),   OPTIONAL,       INTENT(IN)      ::  T_FOR_PREVIOUS_WP   !<  The T parameter for the waypoints which was last added (needed for the MT case)
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG                   =   .FALSE.
        LOGICAL                 ::  SSYMMETRY_MATES
        INTEGER                 ::  CURRENT_OUTPUT_WP
        INTEGER                 ::  PERP_SYM, AXIAL_SYM
        ! Microtubule-specific private variables
        LOGICAL,    PARAMETER   ::  IGNORE_SEAM_DIMERS      =   .FALSE.
        REAL                    ::  WASHER_TOTAL_PHI
        REAL                    ::  ADDITIONAL_T, ADDITIONAL_PHI
        TYPE(WAYPOINT)          ::  PF1_WP,NEXT_PF1_WP, WP
        REAL(KIND=8)            ::  PF1_T,NEXT_PF1_T
        INTEGER                 ::  CURRENT_PF,FIRST_PF,LAST_PF,GOOD_PF
        INTEGER                 ::  NUM_DIMERS_BEYOND_FAL_WPS
        REAL                    ::  LOCAL_HELICAL_RISE
        TYPE(WAYPOINT), ALLOCATABLE ::  NEW_PATH(:)
        ! Start work

        ! Will we be adding symmetry mates too?
        IF (PRESENT(SYMMETRY_MATES)) THEN
            SSYMMETRY_MATES = SYMMETRY_MATES
        ELSE
            SSYMMETRY_MATES = .FALSE.
        ENDIF

        ! If we're dealing with microtubules, we need to know the T parameter for where the next washer starts
        IF (SELF%MICROTUBULE .AND. SSYMMETRY_MATES) THEN
            IF (.NOT. PRESENT(T_FOR_NEXT_NEW_WP)) THEN
                WRITE(*,'(A)') '**ERROR(ADD_WAYPOINT_TO_PATH): the filament is a microtubule, so you need to specify ', &
                                'the T parameter for the following waypoint'
                CALL TERMINATE('Fatal error in ADD_WAYPOINT_TO_PATH')
            ENDIF
        ENDIF

        ! How many waypoints already in the old path?
        IF (ALLOCATED(OLD_PATH)) THEN
            CURRENT_OUTPUT_WP = SIZE(OLD_PATH)
        ELSE
            CURRENT_OUTPUT_WP = 0
        ENDIF

        IF (DEBUG) WRITE(*,'(A,I0)') '**debug(ADD_WAYPOINT_TO_PATH): on entry, current_output_wp = ', CURRENT_OUTPUT_WP

        !
        IF (SSYMMETRY_MATES) THEN
            IF (SELF%MICROTUBULE) THEN
                ! Check we have the required info
                IF (.NOT. PRESENT(T_FOR_NEXT_NEW_WP) .OR. .NOT. PRESENT(T_FOR_PREVIOUS_WP)) THEN
                    WRITE(*,'(A)') '**ERROR(ADD_WAYPOINT_TO_PATH): when dealing with microtubule, need to know the position'// &
                                    ' preceding and following waypoints'
                    CALL TERMINATE('Fatal error in ADD_WAYPOINT_TO_PATH. Programming error.')
                ENDIF
                ! In the case of microtubules, the symmetry-mates are defined by the helical symmetry operator of the "washer"
                GOOD_PF     =   MICROTUBULE_GET_GOOD_PF(SELF)   ! For 13-pf mts, this is pf 7
                NEXT_PF1_T  =   T_FOR_NEW_WP+(T_FOR_NEXT_NEW_WP-T_FOR_NEW_WP)/REAL(SELF%MT_PF_NUMBER)*REAL(GOOD_PF)
                NEXT_PF1_WP =   WAYPOINT_FROM_T(SELF,NEXT_PF1_T)
                PF1_T       =   T_FOR_PREVIOUS_WP+(T_FOR_NEW_WP-T_FOR_PREVIOUS_WP)/REAL(SELF%MT_PF_NUMBER)*REAL(GOOD_PF)
                PF1_WP      =   WAYPOINT_FROM_T(SELF,PF1_T)
                ! We may need to ignore the dimers either side of the seam
                IF (IGNORE_SEAM_DIMERS) THEN
                    FIRST_PF = 2
                    LAST_PF = SELF%MT_PF_NUMBER - 1
                ELSE
                    FIRST_PF = 1
                    LAST_PF = SELF%MT_PF_NUMBER
                ENDIF
                NUM_DIMERS_BEYOND_FAL_WPS = 0
                IF (ALLOCATED(OLD_PATH)) NEW_PATH = OLD_PATH
                ! Total PHI angle change from the start of the current washer to the start of the next washer
                WASHER_TOTAL_PHI = 360. + NEXT_PF1_WP%EULER%PHI - PF1_WP%EULER%PHI
                ! Local helical rise
                LOCAL_HELICAL_RISE = FILAMENT_EVAL_HELICAL_RISE(SELF,T_FOR_NEW_WP)
                ! Loop over the dimers / protofilaments
                DO CURRENT_PF=FIRST_PF,LAST_PF
                    ! Additional T, to place origin at level of desired PF dimer
                    ADDITIONAL_T = -REAL(CURRENT_PF-GOOD_PF)/REAL(SELF%MT_PF_NUMBER) * REAL(SELF%MT_RISE) / 2.0 * LOCAL_HELICAL_RISE
                    ! Sanity check
                    IF (ABS(ADDITIONAL_T) .GT. ABS(LOCAL_HELICAL_RISE * REAL(SELF%MT_RISE)/2.0)) THEN   ! ABS - not sure
                        WRITE(*,'(A,F0.6,A,F0.6,A)') '**ERROR(ADD_WAYPOINT_TO_PATH): additional_t (', ADDITIONAL_T, &
                                                    ') > local washer rise (', LOCAL_HELICAL_RISE* REAL(SELF%MT_RISE)/2.0, ')'
                        CALL TERMINATE('Fatal error in ADD_WAYPOINT_TO_PATH')
                    ENDIF
                    ! Additional twist/phi, to bring current PF to correct PHI
                    ! I suspect WASHER_TOTAL_PHI here should be replaced by 360, since we're getting the waypoint from T anyway, which means we're already compensating for the any twist of the microtubule
                    ADDITIONAL_PHI = -REAL(CURRENT_PF-GOOD_PF) * WASHER_TOTAL_PHI/REAL(SELF%MT_PF_NUMBER)
                    ! Get waypoint at T
                    IF (T_FOR_NEW_WP+ADDITIONAL_T .GT. SELF%T(SELF%NWP) .OR. T_FOR_NEW_WP+ADDITIONAL_T .LT. SELF%T(1)) THEN
                        NUM_DIMERS_BEYOND_FAL_WPS = NUM_DIMERS_BEYOND_FAL_WPS + 1
                        IF (DEBUG) WRITE(*,'(A)') '**debug(ADD_WAYPOINT_TO_PATH): dimer beyond ends of filament'
                    ELSE
                        IF (DEBUG) PRINT '(A,F8.3,1X,F8.3)', '**debug(ADD_WAYPOINT_TO_PATH): additional_t, additional_phi = ', &
                                                            ADDITIONAL_T, ADDITIONAL_PHI
                        WP = WAYPOINT_FROM_T(SELF,T_FOR_NEW_WP+ADDITIONAL_T)
                        ! Add appropriate phi angle to waypoint
                        WP%EULER%PHI = PF1_WP%EULER%PHI + ADDITIONAL_PHI
                        ! Add waypoint to output path
                        CURRENT_OUTPUT_WP = CURRENT_OUTPUT_WP + 1
                        IF (DEBUG) THEN
                            WRITE(*,'(A)',ADVANCE='NO') '**debug(ADD_WAYPOINT_TO_PATH): adding this wp: '
                            CALL WP_PRINT_INFO(WP)
                        ENDIF
                        CALL WAYPOINTS_GROW_ARRAY(NEW_PATH,INCREMENT=1)
                        NEW_PATH(CURRENT_OUTPUT_WP) = WP
                    ENDIF
                ENDDO
                IF (DEBUG) THEN
                    WRITE(*,'(A,I0,A)') '**debug(ADD_WAYPOINT_TO_PATH): ', NUM_DIMERS_BEYOND_FAL_WPS ,&
                                        ' dimers fell beyond the ends of the filament'
                ENDIF
            ELSE
                ! Symmetry mates, but not microtubule
                IF (ALLOCATED(OLD_PATH)) then
                    if (allocated(new_path)) deallocate(new_path)
                    allocate(new_path(size(old_path)))
                    NEW_PATH = OLD_PATH
                endif
                CALL WAYPOINTS_GROW_ARRAY(NEW_PATH,INCREMENT=SELF%HELIX%PERPENDICULAR_SYMMETRY*SELF%HELIX%AXIAL_SYMMETRY)
                IF (DEBUG) WRITE(*,'(A,I0)') '**debug(ADD_WAYPOINT_TO_PATH): size(new_path) after grow_array: ', SIZE(NEW_PATH)
                DO PERP_SYM=1,SELF%HELIX%PERPENDICULAR_SYMMETRY
                    DO AXIAL_SYM=1,SELF%HELIX%AXIAL_SYMMETRY
                        CURRENT_OUTPUT_WP                       =   CURRENT_OUTPUT_WP + 1
                        NEW_PATH(CURRENT_OUTPUT_WP)             =   WAYPOINT_FROM_T(SELF,T_FOR_NEW_WP)
                        NEW_PATH(CURRENT_OUTPUT_WP)%EULER%THETA =   NEW_PATH(CURRENT_OUTPUT_WP)%EULER%THETA + REAL(PERP_SYM-1)*180.
                        NEW_PATH(CURRENT_OUTPUT_WP)%EULER%PHI   =   NEW_PATH(CURRENT_OUTPUT_WP)%EULER%PHI + &
                                                                    REAL(AXIAL_SYM-1)*360./SELF%HELIX%AXIAL_SYMMETRY
                    ENDDO
                ENDDO
            ENDIF   ! End of test for microtubule
        ELSE
            ! No symmetry-mates to be added
            IF (ALLOCATED(OLD_PATH)) then
                if (allocated(new_path)) deallocate(new_path)
                allocate(new_path(size(old_path)))
                NEW_PATH = OLD_PATH
            endif
            CALL WAYPOINTS_GROW_ARRAY(NEW_PATH)
            NEW_PATH(SIZE(NEW_PATH))    =   WAYPOINT_FROM_T(SELF,T_FOR_NEW_WP)
        ENDIF   ! End of test for symmetry mates

        !
        if (allocated(old_path)) deallocate(old_path)
        allocate(old_path(size(new_path)))
        OLD_PATH = NEW_PATH
    END SUBROUTINE ADD_WAYPOINT_TO_PATH

    !>  \brief  Define the range of parameter values in the neighbourhood of waypoints to be refined. This range can
    !!          then be used to define the region the scoring function should examine.
    SUBROUTINE FILAMENT_T_INTERVAL_FOR_SCORING_FUNCTION(SELF,START_T,FINISH_T,START_WP,FINISH_WP)
        USE COOS, ONLY : WAYPOINTS_TO_BE_REFINED_ARE_CONTIGUOUS, WAYPOINTS_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), INTENT(IN)  ::  SELF
        REAL(KIND=8),   INTENT(OUT) ::  START_T
        REAL(KIND=8),   INTENT(OUT) ::  FINISH_T
        INTEGER,        INTENT(OUT) ::  START_WP
        INTEGER,        INTENT(OUT) ::  FINISH_WP
        ! Private variables
        INTEGER ::  I
        INTEGER ::  FIRST_REFINE_WP, LAST_REFINE_WP
        LOGICAL,    PARAMETER   ::  HALFWAY_TO_NEXT_WP = .FALSE.
        ! Start work
        ! Check that waypoints to be refined are contiguous
        IF (.NOT. WAYPOINTS_TO_BE_REFINED_ARE_CONTIGUOUS(SELF%WP)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENT_T_INTERVAL_FOR_SCORING_FUNCTION): waypoints to be refined are not contiguous'
            CALL WAYPOINTS_PRINT_INFO(SELF%WP)
            CALL TERMINATE('Fatal error in FILAMENT_T_INTERVAL_FOR_SCORING_FUNCTION')
        ENDIF

        !print *, '**DEBUG(): fal_wps_in_score = ', SELF%FAL_WPS_IN_SCORE

        ! Work out the first and last waypoints which are being refined
        FIRST_REFINE_WP = 0
        IF (SELF%WP(1)%REFINE) FIRST_REFINE_WP = 1
        LAST_REFINE_WP  = 0
        IF (SELF%WP(SELF%NWP)%REFINE) LAST_REFINE_WP = SELF%NWP
        DO I=2,SELF%NWP
            IF (SELF%WP(I)%REFINE .AND. .NOT. SELF%WP(I-1)%REFINE .AND. FIRST_REFINE_WP .EQ. 0) FIRST_REFINE_WP = I
        ENDDO
        DO I=1,SELF%NWP-1
            IF (SELF%WP(I)%REFINE .AND. .NOT. SELF%WP(I+1)%REFINE .AND. LAST_REFINE_WP .EQ. 0) LAST_REFINE_WP = I
        ENDDO

		! In case all waypoints are set to F, we probably want to score the whole filament
		FIRST_REFINE_WP = 1
		LAST_REFINE_WP  = SELF%NWP

        !PRINT *, '**DEBUG(): FIRST, LAST REFINE = ', FIRST_REFINE_WP, LAST_REFINE_WP

        IF (HALFWAY_TO_NEXT_WP) THEN
            ! We go halfway to the previous/next waypoints
            ! Work out the first and last waypoints for the scoring function
            START_WP = 0
            FINISH_WP = 0
            IF (SELF%FAL_WPS_IN_SCORE) THEN
                START_WP = FIRST_REFINE_WP
                FINISH_WP = LAST_REFINE_WP
            ELSE
                START_WP = MAX(FIRST_REFINE_WP,2)
                START_WP = MIN(START_WP,SELF%NWP-1)
                FINISH_WP = MAX(LAST_REFINE_WP,2)
                FINISH_WP = MIN(FINISH_WP,SELF%NWP-1)
            ENDIF

            !PRINT *, '**DEBUG(): START, FINISH WP = ', START_WP, FINISH_WP

            ! Work out the T parameters
            START_T = 0.0D0
            FINISH_T = 0.0D0

            IF (START_WP .EQ. 1) THEN
                START_T = SELF%T(1)
            ELSE
                START_T = SELF%T(START_WP) - (SELF%T(START_WP)-SELF%T(START_WP-1)) * 0.5D0
            ENDIF

            IF (FINISH_WP .EQ. SELF%NWP) THEN
                FINISH_T = SELF%T(SELF%NWP)
            ELSE
                FINISH_T = SELF%T(FINISH_WP) + (SELF%T(FINISH_WP+1)-SELF%T(FINISH_WP)) * 0.5D0
            ENDIF

            !PRINT *, '**DEBUG(): START_T, FINISH_T = ', START_T, FINISH_T

            ! If FAL_WPs are not in scoring function, make sure T is limited to that also
            IF (.NOT. SELF%FAL_WPS_IN_SCORE) THEN
                START_T = MAX(START_T,SELF%T(2))
                FINISH_T = MIN(FINISH_T,SELF%T(SELF%NWP-1))
            ENDIF

            ! When we have FRAGMENT_HELIX, being exactly half-way between waypoints can cause numerical instabilities,
            ! because of the discontinuity in the helix that happens at this point
            START_T = START_T + 0.1
        ELSE
            ! We go all the way to the previous & next waypoints
            START_WP = 0
            FINISH_WP = 0
            IF (SELF%FAL_WPS_IN_SCORE) THEN
                START_WP = MAX(FIRST_REFINE_WP-1,1)
                FINISH_WP = MIN(LAST_REFINE_WP+1,SELF%NWP)
            ELSE
                START_WP = MAX(FIRST_REFINE_WP-1,2)
                FINISH_WP = MIN(LAST_REFINE_WP+1,SELF%NWP-1)
            ENDIF
            START_T = SELF%T(START_WP)
            FINISH_T = SELF%T(FINISH_WP)
        ENDIF

        !PRINT *, '**DEBUG(): START_T, FINISH_T = ', START_T, FINISH_T

    END SUBROUTINE FILAMENT_T_INTERVAL_FOR_SCORING_FUNCTION

END MODULE FILAMENTS_PATH
