!>  \brief  Classes and methods to deal with EM films / negatives
!!  \todo   Rename film -> micrograph everywhere
MODULE FILMS

    use runtime_parameters, only    :   filename_len
    use ctftilts

    type film
        integer                         ::  id                          !<  film id
        character(len=filename_len)     ::  filename                    !<  filename
        integer                         ::  location_in_file = 1        !<  Location of the film image within the image file
        type(ctftilt)                   ::  ctftilt                     !<  object containing ctf info for this film
        integer                         ::  logical_dimensions(2)       !<  dimensions (pixels)
        real                            ::  pixel_size  =   0.0         !<  pixel size (angstroems)
        !contains
        !   procedure   ::  find_ctf_at_coos    =>  find_ctf_at_coos
        !   procedure   ::  init    =>  film_init
    end type film

    CONTAINS

    !>  \brief  Print out summary information about a film object
    SUBROUTINE FILM_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILM),                     INTENT(IN)      ::  SELF
        ! Private variables

        ! Start work
        WRITE(*,'(A,I0,A)')     '**info(FILM_PRINT_INFO): information for film ', self%id, ' follows:'
        WRITE(*,'(2A)')         '**info(FILM_PRINT_INFO): file name             : ', TRIM(ADJUSTL(SELF%FILENAME))
        WRITE(*,'(A,I0)')       '**info(FILM_PRINT_INFO): loc. in file          : ', self%location_in_file
        WRITE(*,'(A,2(I0,1X))') '**info(FILM_PRINT_INFO): dimensions (pixels)   : ', self%logical_dimensions
        WRITE(*,'(A,F0.2)')     '**info(FILM_PRINT_INFO): pixel size (Angstroms): ', self%pixel_size
        WRITE(*,'(A)')          '**info(FILM_PRINT_INFO): CTF tilt info         : '
        CALL CTFTILT_PRINT_INFO(SELF%CTFTILT)
    END SUBROUTINE FILM_PRINT_INFO

    !>  \brief  Initialise a film object - the dimensions are worked out by opening the image file
    !!  \todo   Add warning if the pixel size given as input does not match the pixel size specified in the CTFTILT/CTFFIND output
    SUBROUTINE FILM_INIT(SELF,ID,FILENAME,PIXEL_SIZE,CTFTILT_FILENAME,CTFFIND_FILENAME,FILM_IMAGE_AVAILABLE,location_within_file)
        USE ImageFiles
        USE PARSERS, ONLY : FILE_EXISTS
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILM),                     INTENT(INOUT)   ::  SELF
        INTEGER,                        INTENT(IN)      ::  ID                      !<  Film ID
        CHARACTER(LEN=*),               INTENT(IN)      ::  FILENAME                !<  Film filename
        REAL,                           INTENT(IN)      ::  PIXEL_SIZE              !<  Pixel size (Angstroms)
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  CTFFIND_FILENAME        !<  Output from CTFFIND
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  CTFTILT_FILENAME        !<  Output from CTFTILT
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  FILM_IMAGE_AVAILABLE    !<  Whether the actual film image is available somewhere disk
        integer,            optional,   intent(in)      ::  location_within_file    !<  Location of the film image within the file
        ! Private variables
        TYPE(IMAGEFILE) ::  imgfile
        INTEGER ::  IDUMMY(3)
        LOGICAL ::  FFILM_IMAGE_AVAILABLE
        LOGICAL, PARAMETER  ::  DEBUG = .FALSE.

        ! Start work

        ! We assume the film image is available
        FFILM_IMAGE_AVAILABLE = .TRUE.
        IF (PRESENT(FILM_IMAGE_AVAILABLE)) FFILM_IMAGE_AVAILABLE = FILM_IMAGE_AVAILABLE

        !
        ! Remember the filename & the ID
        !
        SELF%ID = ID
        SELF%FILENAME = TRIM(ADJUSTL(FILENAME))
        self%location_in_file = 1
        if (present(location_within_file)) self%location_in_file = location_within_file

        !
        ! Initialise a file object so that we can check the file exists, and get its dimensions
        !
        IF (FFILM_IMAGE_AVAILABLE) THEN
            CALL IMGFILE%INIT(SELF%FILENAME)
            IF (DEBUG) CALL IMGFILE%PrintInfo()
            IDUMMY = IMGFILE%GetDimensions()
            SELF%logical_dimensions = IDUMMY(1:2)
            if (self%location_in_file .gt. imgfile%GetStackSize()) then
                write(*,'(2(a,i0),3a)') '**ERROR(film_init): location in file (', self%location_in_file, &
                                        ') > number of locations in the file (', imgfile%GetStackSize(), &
                                        '; ', imgfile%GetFilename(), ')'
                call terminate('Fatal error in film_init')
            endif
            CALL IMGFILE%CLOSE()
        ELSE
            IF (FILE_EXISTS(SELF%FILENAME)) THEN
                WRITE(*,'(2A)') '**WARNING(FILM_INIT): was told this file did not exist, but it does: ', &
                                TRIM(ADJUSTL(SELF%FILENAME))
            ELSE
                WRITE(*,'(A)') '**info(FILM_INIT): film image not available'
            ENDIF
            SELF%logical_dimensions = 0
        ENDIF


        !
        ! Set the pixel size
        !
        SELF%PIXEL_SIZE = PIXEL_SIZE

        !
        ! Get the CTFFIND / CTFTILT information.
        ! If both are present, the CTFTILT information is used.
        !
        IF (PRESENT(CTFFIND_FILENAME)) THEN
            IF (PRESENT(CTFTILT_FILENAME)) THEN
                WRITE(*,'(A)') '**WARNING(FILM_INIT): both CTFFIND and CTFTILT info was given. CTFTILT will take precedence.'
            ENDIF
            CALL CTFTILT_READ_FROM_CTFFIND(SELF%CTFTILT,CTFFIND_FILENAME)
        ENDIF
        IF (PRESENT(CTFTILT_FILENAME)) THEN
            CALL CTFTILT_READ_FROM_CTFTILT(SELF%CTFTILT,CTFTILT_FILENAME)
        ENDIF

        !
        ! Done
        !
        IF (DEBUG) CALL FILM_PRINT_INFO(SELF)
    END SUBROUTINE FILM_INIT

    !>  \brief  Return the CTF (underfocus) parameters at a point on a film
    FUNCTION FILM_FIND_CTF_AT_COOS(SELF,X,Y) RESULT(CTF)
        USE ROTATION_MATRICES, ONLY : RADIANS,DEGREES   ! Conversion routines
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILM),     INTENT(IN)  ::  SELF    ! This used to be CLASS, but that caused problems with openmp
        REAL(KIND=8),   INTENT(IN)  ::  X,Y
        ! Result
        TYPE(CTF_PARAMETERS)        ::  CTF
        ! Private variables
        REAL    ::  N1,N2   ! Tilt axis normal
        REAL    ::  CX,CY   ! Film center coordinates
        REAL    ::  DX,DY   ! Distance from coos to the center
        REAL    ::  DF      ! Distance (in defocus axis) between center of film and coos
        LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
        ! Start work

        ! We assume the CTF is in Angstroms, not pixels
        IF (SELF%CTFTILT%CTF%PIX_UNITS) THEN
            WRITE(*,'(A)') '**ERROR(FILM_FIND_CTF_AT_COOS): CTF should be in Angstroms/degrees, not pixels/radians'
            CALL TERMINATE('Fatal error in FILM_FIND_CTF_AT_COOS')
        ENDIF

        IF (.NOT. ANY(SELF%logical_dimensions .NE. 0)) THEN
            WRITE(*,'(A,2(I0,1X))') '**ERROR(FILM_FIND_CTF_AT_COOS): unknown dimensions for film: ', SELF%logical_dimensions
            CALL TERMINATE('Fatal error in FILM_FIND_CTF_AT_COOS')
        ENDIF


        !PRINT *, '**debug(FIND_CTF_AT_COOS): central CTF = ', SELF%CTFTILT%CTF
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FILM_FIND_CTF_AT_COOS): film info on entry:'
            CALL FILM_PRINT_INFO(SELF)
        ENDIF
        CTF = SELF%CTFTILT%CTF  ! This copies the central CTF parameters
        ! Do the calculation according to CTFTILT documentation
        N1 = -SIN(RADIANS(SELF%CTFTILT%TILT_AXIS))
        N2 =  COS(RADIANS(SELF%CTFTILT%TILT_AXIS))
        CX =  SELF%logical_dimensions(1) / 2.0
        CY =  SELF%logical_dimensions(2) / 2.0
        DX =  CX - X
        DY =  CY - Y
        DF = (N1*DX+N2*DY)*SELF%PIXEL_SIZE*TAN(RADIANS(SELF%CTFTILT%TILT_ANGLE))
        IF (DEBUG) THEN
            PRINT '(A,2(F8.4,1X))', '**debug(FILM_FIND_CTF_AT_COOS): N1, N2 = ', N1, N2
            PRINT '(A,2(F9.2,1X))', '**debug(FILM_FIND_CTF_AT_COOS): DX, DY = ', DX, DY
            PRINT '(A,F9.2)', '**debug(FILM_FIND_CTF_AT_COOS): DF = ', DF
        ENDIF
        CTF%DFMID1 = SELF%CTFTILT%CTF%DFMID1 + DF
        CTF%DFMID2 = SELF%CTFTILT%CTF%DFMID2 + DF
    END FUNCTION FILM_FIND_CTF_AT_COOS

    !>  \brief  Given X,Y coordinates, return the Z coordinate (defocus, in pixels)
    FUNCTION FILM_FIND_Z_FROM_XY(SELF,X,Y) RESULT(Z)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, Z_AXIS_DIRECTION
        IMPLICIT NONE
        ! Arguments
        TYPE(FILM),         INTENT(IN)  ::  SELF    !<  FILM object
        REAL(KIND=8),       INTENT(IN)  ::  X       !<  X coordinates (in pixels)
        REAL(KIND=8),       INTENT(IN)  ::  Y
        ! Result
        REAL                            ::  Z
        ! Private variables
        TYPE(CTF_PARAMETERS)            ::  CTF
        LOGICAL,    PARAMETER           ::  DEBUG = .FALSE.
        ! Start work

        ! We assume the CTF is in Angstroms, not pixels
        IF (SELF%CTFTILT%CTF%PIX_UNITS) THEN
            WRITE(*,'(A)') '**ERROR(FILM_FIND_Z_FROM_XY): CTF should be in Angstroms, not pixels'
            CALL TERMINATE('Fatal error in FILM_FIND_Z_FROM_XY')
        ENDIF

        !CTF = FILM_FIND_CTF_AT_COOS(SELF,X,Y)
        CTF = FILM_FIND_CTF_AT_COOS(SELF,X,Y)
        IF (DEBUG) THEN
            WRITE(*,'(A,2(F0.2,1X))') '**debug(FILM_FIND_Z_FROM_XY): central CTF: '
            CALL CTFTILT_PRINT_INFO(SELF%CTFTILT)
            WRITE(*,'(A,2(F0.2,1X))') '**debug(FILM_FIND_Z_FROM_XY): x, y = ', X, Y
            PRINT '(A,F0.2)', '**debug(FILM_FIND_Z_FROM_XY): CTF%DFMID1 = ', CTF%DFMID1
        ENDIF
        Z = (CTF%DFMID1 + CTF%DFMID2)/2.0 ! In Angstroms
        ! Z_AXIS_DIRECTION defines whether the Z axis goes up or down the microscope column. See comments in runtime_parmaters.f90 for more details.
        Z =  Z_AXIS_DIRECTION * Z / SELF%PIXEL_SIZE
    END FUNCTION FILM_FIND_Z_FROM_XY

END MODULE FILMS
