!>  \brief  Classes and methods to do with comparing filaments and models
MODULE FMCS_BASE
!DEC$ NOOPTIMIZE
! Turn off optimisations for ifort, because there is a bug with ifort 13.1.2.183 Build 20130514 when -ipo compilation option
! is used on this file

    USE FILAMENTS
    USE HELICES
    USE THREEDS
    USE INTERPOLATION
    USE PROJECTION_CACHES
    USE, INTRINSIC  ::  ISO_C_BINDING

    IMPLICIT NONE

    !>  \brief  Structure containing all elements necessary to do a "scoring" of a filament w.r.t a model 3D
    !!
    !!  MODEL3DVOL_ORIGINAL is the pointer to the input model. The MODEL3DVOL is modified (hextend/mask/etc), always
    !!  starting from the original. To avoid cumulative interpolation artefacts.
    !!
    TYPE FILAMENT_MODEL_COMPARISON
        TYPE(FILAMENT)                      ::  FIL                                     !<  Filament to be compared
        TYPE(FILAMENT)                      ::  FIL_ORIGINAL                            !<  Filament in its original state. Whenever the minimiser (for example) wants to try new
                                                                                        !!  parameter values for the filament, we should ensure its start from this state and overwrites the relevant parameters.
        TYPE(IMAGE),            POINTER     ::  FILM_IMAGE              =>  NULL()      !<  Image object containing the film on which the filament is
        TYPE(THREED),           POINTER     ::  THREED_ORIGINAL         =>  NULL()      !<  Original 3D reconstruction (not finalised - i.e. separate CTF^2 etc.)
        TYPE(THREED)                        ::  FIL_THREED                              !<  In some cases, we keep the reconstruction of the filament in its pre-refinement state so that it can be subtracted from the reference before refinement starts
        TYPE(IMAGE),            POINTER     ::  MODEL3DVOL_ORIGINAL     =>  NULL()      !<  Model 3D to compare to
        TYPE(IMAGE)                         ::  MODEL3DVOL_PRIVATE                      !<  Model 3D to compare to (prepared, i.e. helical symmetry applied, masked, ...) (private version)
        TYPE(IMAGE),            POINTER     ::  MODEL3DVOL              =>  NULL()      !<  Model 3D to compare to (prepared, i.e. helical symmetry applied, masked, ...) (either points to private version or to shared version)
        TYPE(HELICAL_SYMMETRY)              ::  HELIX                                   !<  Helical parameters of the model
        INTEGER                             ::  BOX_SIZE                                !<  Box size to cut out & measure correlation with
        INTEGER                             ::  THREED_SIZE                             !<  Box size for the 3D reconstruction
        REAL(KIND=8)                        ::  STEP                                    !<  The step size (in pixels). At every step, a projection will be correlated with corresponding segment of the filament.
        LOGICAL                             ::  STEP_IN_PIXELS                          !<  Specifies (if .TRUE.) that the step is given in pixels
        LOGICAL                             ::  GAVE_WARNING_ABOUT_STEP_SIZE = .FALSE.  !<  Whether we've already printed out a warning about step size
        REAL                                ::  TARGET_RESOLUTION       =   0.0         !<  Target resolution (A) for the refinement
        REAL                                ::  FREQ_MAX                =   0.5         !<  Highest frequency (1/pixel units, 0.5 is Nyquist) to use when scoring using correlation functions.
        REAL                                ::  FREQ_MIN                =   0.0         !<  Lowest frequency (1/pixel units, 0.5 is Nyquist) to use when scoring using correlation functions.
        REAL                                ::  MODEL3DVOL_BFACTOR      =   0.0         !<  B factor (A^2) to be applied to the reference volume before computing the score
        REAL                                ::  ASYM_UNIT_MW            =   0.0         !<  Expected molecular weight of the asymmetric unit (in Daltons)
        !
        REAL                                ::  MODEL3DVOL_MASK_RAD     =   0.0         !<  The radius (pixels) of the mask imposed on the prepared 3D reference volume
        LOGICAL                             ::  MODEL3DVOL_READY        =   .FALSE.     !<  Keeps track of whether the model 3D is ready (i.e. helical symmetry imposed, masked etc.)
        LOGICAL                             ::  MODEL3DVOL_IS_SHARED    =   .FALSE.     !<  The prepared 3D reference volume be shared between FMC instances
        !
        TYPE(C_PTR)                         ::  CPTR                                    !<  A C-pointer pointing to itself. This is needed when passing the object through the FGSL.
        ! Which kind of minimizer will we use?
        LOGICAL                             ::  USE_DOWNHILL_SIMPLEX        =   .TRUE.  !<  Use the downhill simplex minimiser
        LOGICAL                             ::  USE_CONJUGATE_GRADIENT      =   .FALSE. !<  Use conjugate gradient rather than downhill simplex
        LOGICAL                             ::  USE_DIFFERENTIAL_EVOLUTION  =   .FALSE. !<  Use differential evolution rather than downhill simplex
        ! The components below define behaviour when optimising the filament/model comparison
        LOGICAL                             ::  REFINE_WPS_SIMULTANEOUSLY   =   .TRUE.  !<  All waypoints are refined simultaneously by the minimiser
        LOGICAL                             ::  REFINE_PARAMS(7)        =   .FALSE.     !<  Which of the 7 degrees of freedom to refine per way point (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi,7.SubunitNumber)
        LOGICAL                             ::  HOLD_PARAM_AVERAGE(7)   =   .FALSE.     !<  For each parameter, whether its average value should be maintained / held fixed
        LOGICAL                             ::  REFINE_PARAM_INCREMENTS(7)  =   .FALSE. !<  If false, that parameter's absolute value at each waypoint is refined. If true, the increment from the previous waypoint is refined.
        LOGICAL                             ::  INTEGER_CHANGES_ONLY(7) =   .FALSE.     !<  If true, values in that dimension can only be changed by integer amounts
        REAL                                ::  PARAM_SCALE(7)          =   0.0         !<  For each parameter, a scaling factor to ensure that all parameters have comparable values
        REAL                                ::  NM_PARAM_SCALE(7)       =   1.0         !<  When using NM minimisation, it makes sense to scale values so that by the time they are "translated" to real space they are on the scale the user meant for them
        REAL                                ::  SA_MAX_STEP(7)          =   0.0         !<  The maximum step to take along each direction / for each parameter during the random walks of simulated annealing (SA). X is along filament axis, Y is perpendicular.
        REAL                                ::  SEARCH_ACCURACY(7)      =   0.0         !<  The desired accuracy of the search result (pixels, degrees).
        REAL                                ::  SEARCH_HALF_RANGE(7)    =   0.0         !<  The half range (degrees, pixels) of exhaustive search (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi,7.SubunitNumber)
        REAL                                ::  SEARCH_SAMPLING(7)      =   0.0         !<  The sampling step size (pixels, degrees) during exhaustive search (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi,7.SubunitNumber)
        INTEGER                             ::  NUM_MINIM_RUNS          =   1           !<  The number of times the minimiser should be run. Only meant for dowhill simplex (DS), which can be started
                                                                                        !!  several times from the same position but with a random first step.
        LOGICAL                             ::  DS_RANDOM_START_CONF    =   .TRUE.      !<  Whether the starting setup of the DS should be randomised (see GSL documentation for gsl_multimin_fminimizer_nmsimplex2rand
                                                                                        !!  at http://www.gnu.org/software/gsl/manual/html_node/Multimin-Algorithms-without-Derivatives.html)
        LOGICAL                             ::  MINIM_RANDOM_START_POS  =   .FALSE.     !<  Randomise the starting position of the minimiser. The randomisation uses Gaussian with sigma = search_half_range
        INTEGER                             ::  DS_MAX_NUM_ITER                         !<  Maximum number of iterations of the DS algorithm
        !
        LOGICAL                             ::  MODEL_AVE_CYL           =   .FALSE.     !<  Whether the model 3D should be cylindrically averaged
        !
        LOGICAL                             ::  REAL_SPACE_PROJ         =   .TRUE.      !<  Whether to use real space projection. If .FALSE., reciprocal space slice extraction is used instead (which is faster).
        LOGICAL                             ::  REAL_SPACE_BOX_SHIFT    =   .TRUE.      !<  After boxing out image from micrograph, shift it in real space
        LOGICAL                             ::  SUBTRACT_FIL_FROM_REF   =   .FALSE.     !<  Whether the original state of the filament should be subtracted from the 3D model, to reduce bias to original state.
        LOGICAL                             ::  SINGLE_FILAMENT_REFINEMENT = .FALSE.    !<  Refine the filament against a reconstruction of itself (which will be computed at FMC_INIT time)
        INTEGER                             ::  ID_OF_SUBTRACTED_FIL    =   0           !<  Keep track of which filament was subtracted from the 3D reference
        INTEGER                             ::  PATH_OVERSAMPLING       =   1           !<  On average, each pixel will be used this number of times in the scoring function. 1 by default, which means the step size is the same as the filament diameter.
        TYPE(INTERPOLATION_SCHEME)          ::  INTERPOL                                !<  Interpolation scheme
        ! Components below are for single-particle processing of filaments
        LOGICAL                             ::  WPS_AS_SPS              =   .FALSE.     !<  When set to .TRUE., the waypoints become independent single particles, and parts of the filament between waypoints are ignored.
        REAL                                ::  SP_SEGMENT_LENGTH       =   0.0         !<  In Angstroms. Single particles are defined as segments of this length along the filament. Only used if WPS_AS_SPS is .TRUE..
        ! Tune the way the score is calculated
        LOGICAL                             ::  USE_CC3                 =   .FALSE.     !<  Whether to use CC3. If not, CCLIN will be used.
        LOGICAL                             ::  CUBED_AVERAGE           =   .TRUE.      !<  Whether to use cubed average in an attempt to downweigh low SNR segments. If not and we're not in WPS_AS_SPS, a whole filament normalisation will be used
        LOGICAL                             ::  CURVATURE_RESTRAINT     =   .FALSE.     !<  Whether curvature should be used as a restraint on the scoring function
        LOGICAL                             ::  TORSION_RESTRAINT       =   .FALSE.
        LOGICAL                             ::  STRETCH_RESTRAINT       =   .FALSE.
        LOGICAL                             ::  CROSS_BETA_FILTER       =   .FALSE.     !<  Apply a Fourier mask so that the low-frequencies and the expected 4.8A reflections are used and the rest is masked out
        ! Work area memory
        INTEGER                             ::  NUM_WORK_AREAS          =   1           !<  Number of work areas (more than 1 work areas are needed when threading)
        TYPE(WINDOW), ALLOCATABLE           ::  INTERP_WIN_3D(:)                        !<  Window image for interpolation when computing 3D -> 2D projections
        TYPE(WINDOW), ALLOCATABLE           ::  INTERP_WIN_BOX_SHIFT(:)                 !<  Window image for interpolation when shifting the image boxed from micrograph
        TYPE(IMAGE) , ALLOCATABLE           ::  BOX_CTF(:)                              !<  CTF image
        TYPE(IMAGE) , ALLOCATABLE           ::  BOX_AMPL(:)                             !<  Fourier amplitudes of the box cut out from the micrograph
        TYPE(IMAGE) , ALLOCATABLE           ::  AMPL_MASK(:)                            !<  Fourier-space mask used by the scoring function
        TYPE(IMAGE) , ALLOCATABLE           ::  PROJ(:)                                 !<  Projection image
        TYPE(IMAGE) , ALLOCATABLE           ::  MASK_ENV(:)                             !<  Envelope mask (computed from projection)
        TYPE(IMAGE) , ALLOCATABLE           ::  BOX(:)                                  !<  Image boxed out from the micrograph
        TYPE(IMAGE) , ALLOCATABLE           ::  BOX_TMP(:)                              !<  Temporary work image.
        ! Projection caching
        LOGICAL                             ::  CACHE_PROJECTIONS       =   .FALSE.     !<  Whether to keep a cache of projections
        TYPE(EULER)                         ::  CACHE_EULER_TOL                         !<  Euler angles will be considered identical within this tolerance
        TYPE(PROJECTION_CACHE)              ::  PROJ_CACHE                              !<  Cache of projections
        ! Keep track of how many times the minimiser called the scoring function
        INTEGER                             ::  CALLS_FROM_MINIMISER    =   0           !<  Number of times the scoring function was called by the minimiser
        ! Keep track of whether the FMC object was initialised
        LOGICAL                             ::  INITIALISED             =   .FALSE.     !<  Whether the FMC object has already been initialised
    END TYPE FILAMENT_MODEL_COMPARISON

    !INTERFACE ASSIGNMENT(=)
    !    MODULE PROCEDURE FMC_ASSIGNMENT
    !END INTERFACE

    CONTAINS


    !>  \brief  Deallocate work memory for the FMC object
    SUBROUTINE FMC_DEALLOCATE_WORK_MEMORY(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  I
        ! Start work
        IF (FMC_WORK_MEMORY_ARRAYS_ALLOCATED(SELF)) THEN
            DO I=1,SELF%NUM_WORK_AREAS
                CALL DEALLOCATE_WINDOW(SELF%INTERP_WIN_3D(I))
                CALL DEALLOCATE_WINDOW(SELF%INTERP_WIN_BOX_SHIFT(I))
                CALL IMAGE_DEALLOCATE(SELF%BOX_CTF(I))
                CALL IMAGE_DEALLOCATE(SELF%BOX_AMPL(I))
                CALL IMAGE_DEALLOCATE(SELF%AMPL_MASK(I))
                CALL IMAGE_DEALLOCATE(SELF%PROJ(I))
                CALL IMAGE_DEALLOCATE(SELF%MASK_ENV(I))
                CALL IMAGE_DEALLOCATE(SELF%BOX(I))
                CALL IMAGE_DEALLOCATE(SELF%BOX_TMP(I))
            ENDDO
            CALL FMC_DEALLOCATE_WORK_MEMORY_ARRAYS(SELF)
        ENDIF
    END SUBROUTINE FMC_DEALLOCATE_WORK_MEMORY

    PURE LOGICAL FUNCTION FMC_WORK_MEMORY_ARRAYS_ALLOCATED(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)   ::SELF
        ! Start work
        FMC_WORK_MEMORY_ARRAYS_ALLOCATED =  ALLOCATED(SELF%INTERP_WIN_3D       )  .AND. &
                                            ALLOCATED(SELF%INTERP_WIN_BOX_SHIFT)  .AND. &
                                            ALLOCATED(SELF%BOX_CTF             )  .AND. &
                                            ALLOCATED(SELF%BOX_AMPL            )  .AND. &
                                            ALLOCATED(SELF%AMPL_MASK           )  .AND. &
                                            ALLOCATED(SELF%PROJ                )  .AND. &
                                            ALLOCATED(SELF%MASK_ENV            )  .AND. &
                                            ALLOCATED(SELF%BOX                 )  .AND. &
                                            ALLOCATED(SELF%BOX_TMP             )
    END FUNCTION FMC_WORK_MEMORY_ARRAYS_ALLOCATED

    SUBROUTINE FMC_DEALLOCATE_WORK_MEMORY_ARRAYS(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::SELF
        ! Start work
        IF (ALLOCATED(SELF%INTERP_WIN_3D       )) DEALLOCATE(SELF%INTERP_WIN_3D)
        IF (ALLOCATED(SELF%INTERP_WIN_BOX_SHIFT)) DEALLOCATE(SELF%INTERP_WIN_BOX_SHIFT)
        IF (ALLOCATED(SELF%BOX_CTF             )) DEALLOCATE(SELF%BOX_CTF)
        IF (ALLOCATED(SELF%BOX_AMPL            )) DEALLOCATE(SELF%BOX_AMPL)
        IF (ALLOCATED(SELF%AMPL_MASK           )) DEALLOCATE(SELF%AMPL_MASK)
        IF (ALLOCATED(SELF%PROJ                )) DEALLOCATE(SELF%PROJ)
        IF (ALLOCATED(SELF%MASK_ENV            )) DEALLOCATE(SELF%MASK_ENV)
        IF (ALLOCATED(SELF%BOX                 )) DEALLOCATE(SELF%BOX)
        IF (ALLOCATED(SELF%BOX_TMP             )) DEALLOCATE(SELF%BOX_TMP)
    END SUBROUTINE FMC_DEALLOCATE_WORK_MEMORY_ARRAYS

    SUBROUTINE FMC_ALLOCATE_WORK_MEMORY_ARRAYS(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::SELF
        ! Start work
        ALLOCATE(SELF%INTERP_WIN_3D       (SELF%NUM_WORK_AREAS))
        ALLOCATE(SELF%INTERP_WIN_BOX_SHIFT(SELF%NUM_WORK_AREAS))
        ALLOCATE(SELF%BOX_CTF             (SELF%NUM_WORK_AREAS))
        ALLOCATE(SELF%BOX_AMPL            (SELF%NUM_WORK_AREAS))
        ALLOCATE(SELF%AMPL_MASK           (SELF%NUM_WORK_AREAS))
        ALLOCATE(SELF%PROJ                (SELF%NUM_WORK_AREAS))
        ALLOCATE(SELF%MASK_ENV            (SELF%NUM_WORK_AREAS))
        ALLOCATE(SELF%BOX                 (SELF%NUM_WORK_AREAS))
        ALLOCATE(SELF%BOX_TMP             (SELF%NUM_WORK_AREAS))
    END SUBROUTINE FMC_ALLOCATE_WORK_MEMORY_ARRAYS

    !>  \brief  Set the number of work areas. This may need to memory reallocation if necessary.
    SUBROUTINE FMC_SET_NUM_WORK_AREAS(SELF,N)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::  SELF
        INTEGER,                            INTENT(IN)      ::  N
        ! Private variables

        ! Start work
        IF (SELF%NUM_WORK_AREAS .EQ. N) THEN
            ! Nothing to do
        ELSE
            CALL FMC_DEALLOCATE_WORK_MEMORY(SELF)
            SELF%NUM_WORK_AREAS = N
            CALL FMC_ALLOCATE_WORK_MEMORY(SELF)
        ENDIF
    END SUBROUTINE FMC_SET_NUM_WORK_AREAS


    !>  \brief  Allocate work memory for the FMC object
    SUBROUTINE FMC_ALLOCATE_WORK_MEMORY(SELF)
        use images
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  I
        ! Start work

        CALL FMC_DEALLOCATE_WORK_MEMORY(SELF)
        CALL FMC_ALLOCATE_WORK_MEMORY_ARRAYS(SELF)

        ! Allocate the interpol windows
        DO I=1,SELF%NUM_WORK_AREAS
            CALL SELF%INTERP_WIN_3D(I)%AllocateWindow(          SELF%INTERPOL%WINDOW_WIDTH, &
                                                                COMP=.NOT. SELF%REAL_SPACE_PROJ,&
                                                                THREED=.TRUE.)
            CALL SELF%INTERP_WIN_BOX_SHIFT(I)%AllocateWindow(   SELF%INTERPOL%WINDOW_WIDTH, &
                                                                COMP=.NOT.SELF%REAL_SPACE_BOX_SHIFT,    &
                                                                THREED=.FALSE.)
            ! Allocate images
            SELF%BOX_CTF(I)%logical_dimensions = (/SELF%BOX_SIZE,SELF%BOX_SIZE,1/)
            CALL IMAGE_ALLOCATE(SELF%BOX_CTF(I))
            CALL IMAGE_ALLOCATE(SELF%BOX_AMPL(I),SELF%BOX_CTF(I))
            CALL IMAGE_ALLOCATE(SELF%AMPL_MASK(I),SELF%BOX_CTF(I))
            CALL IMAGE_ALLOCATE(SELF%PROJ(I),SELF%BOX_CTF(I))
            CALL IMAGE_ALLOCATE(SELF%MASK_ENV(I),SELF%BOX_CTF(I))
            CALL IMAGE_ALLOCATE(SELF%BOX(I),SELF%BOX_CTF(I))
            CALL IMAGE_ALLOCATE(SELF%BOX_TMP(I),SELF%BOX_CTF(I))
        ENDDO
    END SUBROUTINE FMC_ALLOCATE_WORK_MEMORY

    !>  \brief  Copy all non-allocatable properties from one FMC to another (images, volumes, Cptr
    !!          and initialisation state are not copied)
    SUBROUTINE FMC_COPY_PROPERTIES(DEST,SRC)
        IMPLICIT NONE
        ! Argumnets
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::  DEST    !<  Copy properties into this object
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)      ::  SRC     !<  Copy properties from this object
        ! Private variables

        ! Start work
        DEST%HELIX                              =   SRC%HELIX
        DEST%BOX_SIZE                           =   SRC%BOX_SIZE
        DEST%THREED_SIZE                        =   SRC%THREED_SIZE
        DEST%STEP                               =   SRC%STEP
        DEST%STEP_IN_PIXELS                     =   SRC%STEP_IN_PIXELS
        DEST%GAVE_WARNING_ABOUT_STEP_SIZE       =   SRC%GAVE_WARNING_ABOUT_STEP_SIZE
        DEST%TARGET_RESOLUTION                  =   SRC%TARGET_RESOLUTION
        DEST%FREQ_MAX                           =   SRC%FREQ_MAX
        DEST%FREQ_MIN                           =   SRC%FREQ_MIN
        DEST%MODEL3DVOL_BFACTOR                 =   SRC%MODEL3DVOL_BFACTOR
        DEST%ASYM_UNIT_MW                       =   SRC%ASYM_UNIT_MW
        DEST%MODEL3DVOL_MASK_RAD                =   SRC%MODEL3DVOL_MASK_RAD
        DEST%MODEL3DVOL_READY                   =   SRC%MODEL3DVOL_READY
        DEST%MODEL3DVOL_IS_SHARED               =   SRC%MODEL3DVOL_IS_SHARED
        DEST%USE_DOWNHILL_SIMPLEX               =   SRC%USE_DOWNHILL_SIMPLEX
        DEST%USE_CONJUGATE_GRADIENT             =   SRC%USE_CONJUGATE_GRADIENT
        DEST%USE_DIFFERENTIAL_EVOLUTION         =   SRC%USE_DIFFERENTIAL_EVOLUTION
        DEST%REFINE_WPS_SIMULTANEOUSLY          =   SRC%REFINE_WPS_SIMULTANEOUSLY
        DEST%REFINE_PARAMS                      =   SRC%REFINE_PARAMS
        DEST%HOLD_PARAM_AVERAGE                 =   SRC%HOLD_PARAM_AVERAGE
        DEST%REFINE_PARAM_INCREMENTS            =   SRC%REFINE_PARAM_INCREMENTS
        DEST%INTEGER_CHANGES_ONLY               =   SRC%INTEGER_CHANGES_ONLY
        DEST%PARAM_SCALE                        =   SRC%PARAM_SCALE
        DEST%NM_PARAM_SCALE                     =   SRC%NM_PARAM_SCALE
        DEST%SA_MAX_STEP                        =   SRC%SA_MAX_STEP
        DEST%SEARCH_ACCURACY                    =   SRC%SEARCH_ACCURACY
        DEST%SEARCH_HALF_RANGE                  =   SRC%SEARCH_HALF_RANGE
        DEST%SEARCH_SAMPLING                    =   SRC%SEARCH_SAMPLING
        DEST%NUM_MINIM_RUNS                     =   SRC%NUM_MINIM_RUNS
        DEST%DS_RANDOM_START_CONF               =   SRC%DS_RANDOM_START_CONF
        DEST%MINIM_RANDOM_START_POS             =   SRC%MINIM_RANDOM_START_POS
        DEST%DS_MAX_NUM_ITER                    =   SRC%DS_MAX_NUM_ITER
        DEST%MODEL_AVE_CYL                      =   SRC%MODEL_AVE_CYL
        DEST%REAL_SPACE_PROJ                    =   SRC%REAL_SPACE_PROJ
        DEST%REAL_SPACE_BOX_SHIFT               =   SRC%REAL_SPACE_BOX_SHIFT
        DEST%SUBTRACT_FIL_FROM_REF              =   SRC%SUBTRACT_FIL_FROM_REF
        DEST%SINGLE_FILAMENT_REFINEMENT         =   SRC%SINGLE_FILAMENT_REFINEMENT
        DEST%ID_OF_SUBTRACTED_FIL               =   SRC%ID_OF_SUBTRACTED_FIL
        DEST%PATH_OVERSAMPLING                  =   SRC%PATH_OVERSAMPLING
        DEST%INTERPOL                           =   SRC%INTERPOL
        DEST%WPS_AS_SPS                         =   SRC%WPS_AS_SPS
        DEST%SP_SEGMENT_LENGTH                  =   SRC%SP_SEGMENT_LENGTH
        DEST%USE_CC3                            =   SRC%USE_CC3
        DEST%CUBED_AVERAGE                      =   SRC%CUBED_AVERAGE
        DEST%CURVATURE_RESTRAINT                =   SRC%CURVATURE_RESTRAINT
        DEST%TORSION_RESTRAINT                  =   SRC%TORSION_RESTRAINT
        DEST%STRETCH_RESTRAINT                  =   SRC%STRETCH_RESTRAINT
        DEST%CROSS_BETA_FILTER                  =   SRC%CROSS_BETA_FILTER
        DEST%NUM_WORK_AREAS                     =   SRC%NUM_WORK_AREAS
        DEST%CACHE_PROJECTIONS                  =   SRC%CACHE_PROJECTIONS
        DEST%CACHE_EULER_TOL                    =   SRC%CACHE_EULER_TOL
        !DEST%PROJ_CACHE                         =   SRC%PROJ_CACHE
        DEST%CALLS_FROM_MINIMISER               =   SRC%CALLS_FROM_MINIMISER
    END SUBROUTINE FMC_COPY_PROPERTIES


    !>  \brief  Print useful summary info about the FMC object
    SUBROUTINE FMC_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)  ::  SELF    !<  Filament-model comparison object
        ! Private variables
        CHARACTER(LEN=2),    PARAMETER   ::  TEXT_WIDTH  =   '38'
        ! Start work

        ! Print header line
        WRITE(*,'(/A)')                             'Filament-model comparison parameters'
        ! Print information
        WRITE(*,'(A'//TEXT_WIDTH//',1X,I0)')            'Filament #',                          SELF%FIL%ID
        WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')             'Film filename',                       TRIM(ADJUSTL(SELF%FIL%FILM%FILENAME))
        WRITE(*,'(A'//TEXT_WIDTH//',1X,I0)')            'Box size',                            SELF%BOX_SIZE
        IF (SELF%STEP_IN_PIXELS) THEN
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.2)')          'Step size (pixels)',                  SELF%STEP
        ELSE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.2)')          'Step size (A)',                       SELF%STEP
        ENDIF
        WRITE(*,'(A'//TEXT_WIDTH//',1X,I0)')            'Path oversampling factor',            SELF%PATH_OVERSAMPLING
        WRITE(*,'(A'//TEXT_WIDTH//',1X,I0)')            'Reconstruction volume size (pixels)', SELF%THREED_SIZE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Real space projections?',             SELF%REAL_SPACE_PROJ
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Subtract starting state from 3D ref?',SELF%SUBTRACT_FIL_FROM_REF
        IF (SELF%SUBTRACT_FIL_FROM_REF) THEN
            WRITE(*,'(A'//TEXT_WIDTH//',1X,I0)')        'Subtracted filament #',               SELF%ID_OF_SUBTRACTED_FIL
        ENDIF
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Refine filament against itself?',          SELF%SUBTRACT_FIL_FROM_REF
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(A9,1X))')      'Search paramameters',                 'X','Y','Z/def',&
                                                                                              'Phi','Theta','Psi','SubunitNum'
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(L9,1X))')      'Refine parameter?',                        SELF%REFINE_PARAMS
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(L9,1X))')      'Hold parameter average value?',            SELF%HOLD_PARAM_AVERAGE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(L9,1X))')      'Rigid parameter?',                         SELF%FIL%RIGID_PARAMS
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(L9,1X))')      'Refine parameter wp-to-wp increments?',    SELF%REFINE_PARAM_INCREMENTS
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(L9,1X))')      'Change parameter by integer steps only?',  SELF%INTEGER_CHANGES_ONLY
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(F9.3,1X))')    'Search half-range (pixels, degrees)',      SELF%SEARCH_HALF_RANGE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(F9.3,1X))')    'Search accuracy  (pixels, degrees)',       SELF%SEARCH_ACCURACY
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(F9.3,1X))')    'Parameter scales',                         SELF%PARAM_SCALE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(F9.3,1X))')    'NM parameters scaling factor',             SELF%NM_PARAM_SCALE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.5,A,F0.5,A)') 'Target resolution (A)',                    SELF%TARGET_RESOLUTION
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.5,A,F0.5,A)') 'Cuton frequency, resolution',              SELF%FREQ_MIN, '/pixel, ', &
                                                                                       SELF%FIL%FILM%PIXEL_SIZE/SELF%FREQ_MIN, 'A'
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.5,A,F0.5,A)') 'Cutoff frequency, resolution',             SELF%FREQ_MAX, '/pixel, ',  &
                                                                                       SELF%FIL%FILM%PIXEL_SIZE/SELF%FREQ_MAX, 'A'
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Cross-beta Fourier filter?',               SELF%CROSS_BETA_FILTER
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.2)')          'Asymmetric unit MW (Da)',                  SELF%ASYM_UNIT_MW
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.2)')          'B factor applied to reference (A^2)',      SELF%MODEL3DVOL_BFACTOR
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Waypoints as single particles?',           SELF%WPS_AS_SPS
        IF (SELF%WPS_AS_SPS) THEN
            WRITE(*,'(A'//TEXT_WIDTH//',1X,F0.2)')      'Single particle segment length (A)',       SELF%SP_SEGMENT_LENGTH
        ENDIF
        WRITE(*,'(A'//TEXT_WIDTH//',3(1X,L1))')         'DS? CG? DE?',                              SELF%USE_DOWNHILL_SIMPLEX, &
                                                                                                    SELF%USE_CONJUGATE_GRADIENT, &
                                                                                                    SELF%USE_DIFFERENTIAL_EVOLUTION
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Refine waypoints simultaneously?',         SELF%REFINE_WPS_SIMULTANEOUSLY
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Randomised DS start configuration?',       SELF%DS_RANDOM_START_CONF
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Randomised minim. start position?',        SELF%MINIM_RANDOM_START_POS
        WRITE(*,'(A'//TEXT_WIDTH//',1X,I0)')            'Number of (DS) minimisation runs',         SELF%NUM_MINIM_RUNS
        WRITE(*,'(A'//TEXT_WIDTH//',1X,I0)')            'Max number of DS iterations per run',      SELF%DS_MAX_NUM_ITER
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Use weighted correlation coefficient?',    SELF%USE_CC3
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Use average of cubed segment scores?',     SELF%CUBED_AVERAGE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Curvature restraint?',                     SELF%CURVATURE_RESTRAINT
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')            'Cache projections?',                       SELF%CACHE_PROJECTIONS
        ! Add a blank line
        CALL FMC_SUITABLE_FOR_EARLY_SEARCH(SELF)
        WRITE(*,'(A)')                          ' '
    END SUBROUTINE FMC_PRINT_INFO


    !>  \brief  Work out our tolerance to errors in the helical rise
    REAL FUNCTION FMC_HELICAL_RISE_TOLERANCE(SELF) RESULT(TOL)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)  ::  SELF            !<  Filament-model comparison object
        ! Private variables
        REAL    ::  NUM_SUBS_IN_SEGMENT
        ! Start work

        NUM_SUBS_IN_SEGMENT = FMC_NUM_SUBUNITS_PER_SEGMENT(SELF)
        TOL = SELF%SEARCH_ACCURACY(7) / NUM_SUBS_IN_SEGMENT
    END FUNCTION FMC_HELICAL_RISE_TOLERANCE

    !>  \brief  Work out our tolerance to errors in the helical rise
    REAL FUNCTION FMC_HELICAL_TWIST_TOLERANCE(SELF) RESULT(TOL)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)  ::  SELF            !<  Filament-model comparison object
        ! Private variables
        REAL    ::  NUM_SUBS_IN_SEGMENT
        ! Start work

        NUM_SUBS_IN_SEGMENT = FMC_NUM_SUBUNITS_PER_SEGMENT(SELF)
        TOL = SELF%SEARCH_ACCURACY(4) / NUM_SUBS_IN_SEGMENT
    END FUNCTION FMC_HELICAL_TWIST_TOLERANCE

    !>  \brief  The number of helical subunits in each segment used by the scoring function
    REAL FUNCTION FMC_NUM_SUBUNITS_PER_SEGMENT(SELF) RESULT(NUM_SUBS)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)  ::  SELF            !<  Filament-model comparison object
        ! Start work
        IF (SELF%WPS_AS_SPS) THEN
            ! If we are dealing with single particles, we can only tolerate errors
            ! where the last subunit in the SP segment is within accuracy
            NUM_SUBS = SELF%SP_SEGMENT_LENGTH / SELF%FIL%FILM%PIXEL_SIZE * SELF%HELIX%RISE
        ELSE
            IF (SELF%STEP_IN_PIXELS) THEN
                NUM_SUBS = SELF%STEP * SELF%HELIX%RISE
            ELSE
                NUM_SUBS = SELF%STEP / SELF%FIL%FILM%PIXEL_SIZE * SELF%HELIX%RISE
            ENDIF
        ENDIF
    END FUNCTION FMC_NUM_SUBUNITS_PER_SEGMENT

    !>  \brief  Determine whether the FMC parameters are suitable for an early search & refinement. If not, give feedback to user.
    SUBROUTINE FMC_SUITABLE_FOR_EARLY_SEARCH(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)  ::  SELF            !<  Filament-model comparison object
        ! Private variables
        REAL    ::  MIN_PHI
        ! Start work

        ! Establish the minimum values for search half-ranges which would be adequate in an initial search & refine
        MIN_PHI = SELF%HELIX%TWIST / 2.

        ! Check whether the search half-ranges are sufficient
        IF (SELF%SEARCH_HALF_RANGE(4) .GT. MIN_PHI) THEN

        ELSE
            WRITE(*,'(A,F0.3,A,F0.3)') '**info(FMC_SUITABLE_FOR_EARLY_SEARCH): search half-range for phi angle (', &
                                        SELF%SEARCH_HALF_RANGE(4), ') may not be sufficient for an early round '// &
                                        'of search & refinement. A recommended value would be ', MIN_PHI
        ENDIF
    END SUBROUTINE FMC_SUITABLE_FOR_EARLY_SEARCH

    !>  \brief  The FMC object keeps a reconstruction of the filament - update it.
    SUBROUTINE FMC_UPDATE_FIL_THREED(SELF)
        USE IMAGEFILES ! Debug only
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON), TARGET,      INTENT(INOUT)   ::  SELF
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        TYPE(IMAGE)             ::  DBG_3DVOL
        ! Start work
        ! Initialise a 3D reconstruction object
        IF (DEBUG) WRITE(*,'(A,3(I0,1X))') '**debug(FMC_UPDATE_FIL_THREED): entry, threed_size = ', SELF%THREED_SIZE
        SELF%FIL_THREED%logical_dimensions = SELF%THREED_SIZE
        CALL THREED_INIT(SELF%FIL_THREED)
        WRITE(*,'(A)') '**info(FMC_UPDATE_FIL_THREED): calculating reconstruction for current state of filament'
        ! Reconstruct the starting state of the filament
        IF (ASSOCIATED(SELF%THREED_ORIGINAL)) THEN
            CALL THREED_ASSIGNMENT(SELF%FIL_THREED,SELF%THREED_ORIGINAL)
            IF (.NOT. SELF%FIL_THREED%FOM_FILTER .AND. .NOT. SELF%FIL_THREED%COMPUTE_FSC) THEN
                CALL THREED_TO_SINGLE(SELF%FIL_THREED)
            ENDIF
            CALL THREED_WIPE(SELF%FIL_THREED)
        ENDIF
        CALL FILAMENT_CALC_THREED(SELF%FIL,SELF%FILM_IMAGE,SELF%FIL_THREED,FINALISE=.FALSE.,SP_MODE=SELF%WPS_AS_SPS,    &
                                  SP_SEGMENT_LENGTH=SELF%SP_SEGMENT_LENGTH)

        IF (SELF%SINGLE_FILAMENT_REFINEMENT) SELF%MODEL3DVOL_READY = .FALSE.

        IF (DEBUG) THEN
            CALL THREED_EVAL_RECONSTRUCTION(SELF%FIL_THREED,DBG_3DVOL)
            CALL WRITE_TO_DISK(DBG_3DVOL,'dbg_fil3d.mrc')
        ENDIF
    END SUBROUTINE FMC_UPDATE_FIL_THREED

    !>  \brief  Replace the FMC's filament object and update the FMC step size (since it depends on the filament deformations)
    SUBROUTINE FMC_NEW_FILAMENT(SELF,FIL,REPLACE_ORIGINAL,UPDATE_FIL_THREED)
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    TARGET, INTENT(INOUT)   ::  SELF
        TYPE(FILAMENT),                             INTENT(INOUT)   ::  FIL
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  REPLACE_ORIGINAL        !<  Also replace the "fil_original" in the FMC object
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  UPDATE_FIL_THREED       !<  Update the FMC's fil_threed with the new filament's 3D
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
        LOGICAL                 ::  RREPLACE_ORIGINAL, UUPDATE_FIL_THREED
        ! Start work

        RREPLACE_ORIGINAL = .TRUE.
        IF (PRESENT(REPLACE_ORIGINAL)) RREPLACE_ORIGINAL = REPLACE_ORIGINAL

        UUPDATE_FIL_THREED = .FALSE.
        IF (PRESENT(UPDATE_FIL_THREED)) UUPDATE_FIL_THREED = UPDATE_FIL_THREED


        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FMC_NEW_FILAMENT): filament info upon entry'
            CALL FILAMENT_PRINT_INFO(FIL)
        ENDIF

        !CALL FILAMENT_INIT(SELF%FIL,FIL%WP, FILM_IN=FIL%FILM, MOULD=FIL)
        CALL FILAMENT_ASSIGN(SELF%FIL,FIL)
        IF (RREPLACE_ORIGINAL) CALL FILAMENT_ASSIGN(SELF%FIL_ORIGINAL,FIL)

        SELF%REFINE_PARAMS          =   SELF%FIL%REFINE_PARAMS

        IF (UUPDATE_FIL_THREED) CALL FMC_UPDATE_FIL_THREED(SELF)

        CALL FMC_UPDATE_BOX_SIZE(SELF)
        CALL FMC_UPDATE_STEP_SIZE(SELF,FIL)

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FMC_NEW_FILAMENT): FMC filament info on exit: '
            CALL FILAMENT_PRINT_INFO(SELF%FIL)
        ENDIF
    END SUBROUTINE FMC_NEW_FILAMENT

    !>  \brief  FMC assignment (deep copy, except for pointers obviously)
	!!	\todo	Deal with projection cache
    SUBROUTINE FMC_ASSIGNMENT(LHS,RHS)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    TARGET, INTENT(INOUT)   ::  LHS
        TYPE(FILAMENT_MODEL_COMPARISON),            INTENT(IN)      ::  RHS
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_ASSIGNMENT): entry'
        CALL FMC_DEALLOCATE(LHS)
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_ASSIGNMENT): about to copy properties'
		CALL FMC_COPY_PROPERTIES(LHS,RHS)
		IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_ASSIGNMENT): about to assign filament'
        LHS%FIL                    =   RHS%FIL
        LHS%FIL_ORIGINAL           =   RHS%FIL_ORIGINAL
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_ASSIGNMENT): pointer to film_image'
        LHS%FILM_IMAGE             =>  RHS%FILM_IMAGE
        LHS%THREED_ORIGINAL        =>  RHS%THREED_ORIGINAL
        LHS%MODEL3DVOL_ORIGINAL    =>  RHS%MODEL3DVOL_ORIGINAL
        IF (RHS%MODEL3DVOL_IS_SHARED) THEN
            IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_ASSIGNMENT): model3dvol_shared pointing'
            LHS%MODEL3DVOL         =>  RHS%MODEL3DVOL
        ELSE
            IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_ASSIGNMENT): model3dvol_private assignment'
            LHS%MODEL3DVOL_PRIVATE =   RHS%MODEL3DVOL_PRIVATE
            NULLIFY(LHS%MODEL3DVOL)
            LHS%MODEL3DVOL => LHS%MODEL3DVOL_PRIVATE
        ENDIF
        IF (RHS%FIL_THREED%IN_MEMORY) CALL THREED_ASSIGNMENT(LHS%FIL_THREED,RHS%FIL_THREED)
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_ASSIGNMENT): c_loc call'
        LHS%CPTR                   =   C_LOC(LHS)
        CALL FMC_ALLOCATE_WORK_MEMORY(LHS)
        LHS%INITIALISED            =   RHS%INITIALISED
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_ASSIGNMENT): all done'
    END SUBROUTINE FMC_ASSIGNMENT

    !>  \brief  Deallocate all the memory allocated to an FMC object
    SUBROUTINE FMC_DEALLOCATE(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::  SELF
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_DEALLOCATE): entry'
        CALL FILAMENT_DEALLOCATE(SELF%FIL)
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_DEALLOCATE): about to dealloc fil_original'
        CALL FILAMENT_DEALLOCATE(SELF%FIL_ORIGINAL)
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_DEALLOCATE): about to nullify'
        NULLIFY(SELF%FILM_IMAGE)
        NULLIFY(SELF%THREED_ORIGINAL)
        NULLIFY(SELF%MODEL3DVOL_ORIGINAL)
        NULLIFY(SELF%MODEL3DVOL)
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_DEALLOCATE): done nullify'
        IF (SELF%MODEL3DVOL_IS_SHARED) THEN
            ! Nothing to do
        ELSE
            CALL IMAGE_DEALLOCATE(SELF%MODEL3DVOL_PRIVATE)
        ENDIF
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_DEALLOCATE): about to clean proj cache'
        CALL PROJECTION_CACHE_CLEAN(SELF%PROJ_CACHE)
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_DEALLOCATE): about to dealloc threed'
        CALL THREED_DEALLOCATE(SELF%FIL_THREED)
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_DEALLOCATE): about to dealloc work mem'
        CALL FMC_DEALLOCATE_WORK_MEMORY(SELF)
        !NULLIFY(SELF%CPTR)
        SELF%INITIALISED = .FALSE.
        IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_DEALLOCATE): exit'
    END SUBROUTINE FMC_DEALLOCATE

    !>  \brief  Prepare a filament-model comparison object
    SUBROUTINE FMC_INIT(    SELF,FIL,FILM_IMAGE,MODEL3DVOL,HELIX,                       &
                            TARGET_RESOLUTION,                                          &
                            INTERPOL,                                                   &
                            ASYM_UNIT_MW,REAL_SPACE_PROJ,                               &
                            SUBTRACT_FIL_FROM_REF,MODEL3D,WPS_AS_SPS,SP_SEGMENT_LENGTH, &
                            SEARCH_HALF_RANGE_OVERRIDE,                                 &
                            NUM_MINIM_RUNS,DS_RANDOM_START_CONF,MINIM_RANDOM_START_POS, &
                            DS_MAX_NUM_ITER,                                            &
                            PATH_OVERSAMPLING,                                          &
                            LOW_RESOLUTION_CUTON,HIGH_RESOLUTION_CUTOFF,                &
                            USE_CC3,CUBED_AVERAGE,                                      &
                            CURVATURE_RESTRAINT,TORSION_RESTRAINT,STRETCH_RESTRAINT,    &
                            MODEL_AVE_CYL,CACHE_PROJECTIONS,                            &
                            MODEL3DVOL_BFACTOR,HOLD_PARAM_AVERAGE,                      &
                            USE_CONJUGATE_GRADIENT,                                     &
                            USE_DIFFERENTIAL_EVOLUTION,                                 &
                            REFINE_PARAM_INCREMENTS,INTEGER_CHANGES_ONLY,               &
                            CROSS_BETA_FILTER,REFINE_WPS_SIMULTANEOUSLY,                &
                            MODEL3DVOL_IS_SHARED,SINGLE_FILAMENT_REFINEMENT)
        USE, INTRINSIC  ::  ISO_C_BINDING
        USE EULERANGLES
        USE IMAGES_COMPARISON
        USE MINIMISERS, ONLY : SEARCH_ACCURACY_TO_DS_SIZE_RATIO, DS_MAX_ITERATIONS
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    TARGET, INTENT(INOUT)   ::  SELF
        TYPE(FILAMENT),                             INTENT(INOUT)   ::  FIL                             !<  Filament object
        TYPE(IMAGE),        TARGET,                 INTENT(IN)      ::  FILM_IMAGE                      !<  Image object containing the film on which the filament is
        TYPE(IMAGE),        TARGET,                 INTENT(IN)      ::  MODEL3DVOL                      !<  3D model to compare to
        TYPE(HELICAL_SYMMETRY),                     INTENT(IN)      ::  HELIX                           !<  Helical parameters for the model
        REAL,                                       INTENT(IN)      ::  TARGET_RESOLUTION               !<  The target resolution (in Angstrom) of the refinement.
        TYPE(INTERPOLATION_SCHEME),                 INTENT(IN)      ::  INTERPOL                        !<  Interpolation scheme
        REAL,                       OPTIONAL,       INTENT(IN)      ::  ASYM_UNIT_MW                    !<  Expected molecular weight (Daltons) of the asymmetric subunit
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  REAL_SPACE_PROJ                 !<  Whether to use real space projection. If .FALSE., reciprocal space slice extraction is used instead (which is faster).
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  SUBTRACT_FIL_FROM_REF           !<  Whether the original state of the filament should be subtracted from the 3D model, to reduce bias to original state. This is done at FMC_INIT time.
        TYPE(THREED),       TARGET, OPTIONAL,       INTENT(IN)      ::  MODEL3D                         !<  3D reconstruction which will be used as a reference
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  WPS_AS_SPS                      !<  If .TRUE., filament waypoints will be treated as independent single particles
        REAL,                       OPTIONAL,       INTENT(IN)      ::  SP_SEGMENT_LENGTH               !<  In Angstroms. Single particles are defined as segments of this length along the filament.
                                                                                                        !!  Only used if WPS_AS_SPS is .TRUE.. 700A for TMV according to Sachse.
        REAL,                       OPTIONAL,       INTENT(IN)      ::  SEARCH_HALF_RANGE_OVERRIDE(7)   !<  For each refinement dimension (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi), 0.0 if default search half-range
                                                                                                        !!  should be used, or other value to override the default.
        INTEGER,                    OPTIONAL,       INTENT(IN)      ::  NUM_MINIM_RUNS                  !<  The number of times the minimiser should be run. Only meant for dowhill simplex (DS), which can be started
                                                                                                        !!  several times from the same position but with a random first step.
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  DS_RANDOM_START_CONF            !<  Randomise the starting setup of the DS (see GSL documentation for gsl_multimin_fminimizer_nmsimplex2rand
                                                                                                        !!  at http://www.gnu.org/software/gsl/manual/html_node/Multimin-Algorithms-without-Derivatives.html)
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  MINIM_RANDOM_START_POS          !<  Randomise the starting position of the minimiser
        INTEGER,                    OPTIONAL,       INTENT(IN)      ::  DS_MAX_NUM_ITER                 !<  Maximum number of iterations of the DS
        INTEGER,                    OPTIONAL,       INTENT(IN)      ::  PATH_OVERSAMPLING               !<  On average, each pixel will be used this number of times in the scoring function. 1 by default,
                                                                                                        !!  which means the step size is the same as the filament diameter.
        REAL,                       OPTIONAL,       INTENT(IN)      ::  LOW_RESOLUTION_CUTON            !<  The lowest resolution (in Angstroms) to be used by the scoring functions. Defaults to 1000A.
        REAL,                       OPTIONAL,       INTENT(IN)      ::  HIGH_RESOLUTION_CUTOFF          !<  The highest resolution (in Angstroms) to be used by the scoring functions.
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  USE_CC3                         !<  Whether we will use CC3, rather than CCLIN
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  CUBED_AVERAGE                   !<  Whether we will use cubed averaging of segment scores
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  CURVATURE_RESTRAINT             !<  Whether filament curvature should be used as a constraint for the scoring function
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  TORSION_RESTRAINT               !<  Whether filament torsion should be used as a constraint for the scoring function
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  STRETCH_RESTRAINT               !<  Whether filament stretch should be used as a constraint for the scoring function
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  MODEL_AVE_CYL                   !<  Whether the model 3D should be cylindrically averaged
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  CACHE_PROJECTIONS               !<  Whether projections should be cache
        REAL,                       OPTIONAL,       INTENT(IN)      ::  MODEL3DVOL_BFACTOR              !<  B factor to be applied to the reference volume before use in scoring function (in A^2)
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  HOLD_PARAM_AVERAGE(7)           !<  Whether to hold a parameter's average value
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  USE_CONJUGATE_GRADIENT          !<  Whether conjugate gradient (rather than downhill simplex) should be used as a maximiser
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  USE_DIFFERENTIAL_EVOLUTION      !<  Whether differential evolution (rather than downhill simplex) should be used as a maximiser
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  REFINE_PARAM_INCREMENTS(7)      !<  If false, that parameter's absolute value at each waypoint is refined. If true, the increment from the previous waypoint is refined.
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  INTEGER_CHANGES_ONLY(7)         !<  For each dimension, only integer changes in values will be allowed if this flag is set
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  CROSS_BETA_FILTER               !<  Apply a Fourier-space mask to let the 4.8A expected reflection through, as well as low-frequencies
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  REFINE_WPS_SIMULTANEOUSLY       !<  All waypoints are refined by the minimiser at the same time
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  MODEL3DVOL_IS_SHARED            !<  The model3dvol reference should be shared between instances
        LOGICAL,                    OPTIONAL,       INTENT(IN)      ::  SINGLE_FILAMENT_REFINEMENT      !<  Refine each filament against a reconstruction of itself only (not the average reconstruction of all filaments)
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
        TYPE(CTF_PARAMETERS)    ::  CTF_TOL
        ! Start work

        WRITE(*,'(A)') '<fmc_init>'
        WRITE(*,'(A)') '**info(FMC_INIT): initialising a filament-model comparison object'

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FMC_INIT): filament info on entry:'
            CALL FILAMENT_PRINT_INFO(FIL,VERBOSE=.TRUE.)
        ENDIF

        ! Will we be keeping a private copy of the model3d volume?
        IF (PRESENT(MODEL3DVOL_IS_SHARED)) SELF%MODEL3DVOL_IS_SHARED = MODEL3DVOL_IS_SHARED
        WRITE(*,'(A,1X,L1)') '**debug(FMC_INIT): SELF%MODEL3DVOL_IS_SHARED = ', SELF%MODEL3DVOL_IS_SHARED

        ! Required data
        SELF%FILM_IMAGE             =>  FILM_IMAGE
        SELF%MODEL3DVOL_ORIGINAL    =>  MODEL3DVOL
        IF (SELF%MODEL3DVOL_IS_SHARED) THEN
            SELF%MODEL3DVOL         =>  MODEL3DVOL
        ELSE
            SELF%MODEL3DVOL_PRIVATE =   MODEL3DVOL
            SELF%MODEL3DVOL         =>  SELF%MODEL3DVOL_PRIVATE
        ENDIF

        ! Will we be subtracting the filament (in its original state) from the model 3D reconstruction?
        SELF%SUBTRACT_FIL_FROM_REF = .FALSE.
        IF (PRESENT(SUBTRACT_FIL_FROM_REF)) SELF%SUBTRACT_FIL_FROM_REF = SUBTRACT_FIL_FROM_REF

        ! Will we be refining each filament against itself?
        SELF%SINGLE_FILAMENT_REFINEMENT = .FALSE.
        IF (PRESENT(SINGLE_FILAMENT_REFINEMENT)) SELF%SINGLE_FILAMENT_REFINEMENT = SINGLE_FILAMENT_REFINEMENT

        ! Subtracting the original state of the filament from the shared model3d should be possible
        ! as long as we don't then use the same FMC to process another filament.
        ! But for now, I will not spend time to make sure this works, so I disallow it.
        IF (SELF%SUBTRACT_FIL_FROM_REF .AND. SELF%MODEL3DVOL_IS_SHARED) THEN
            WRITE(*,'(A)') '**ERROR(FMC_INIT): SUBTRACT_FIL_FROM_REF and MODEL3DVOL_IS_SHARED are incompatible'
            CALL TERMINATE('Fatal error in FMC_INIT. Feature not (yet) supported.')
        ENDIF

        ! Check for non-sensical option combination
        IF (SELF%SINGLE_FILAMENT_REFINEMENT .AND. SELF%SUBTRACT_FIL_FROM_REF) THEN
            ! we would end up with a blank reference!
            WRITE(*,'(A)') '**ERROR(FMC_INIT): SINGLE_FILAMENT_REFINEMENT and SUBTRACT_FIL_FROM_REF are incompatible'
            CALL TERMINATE('Fatal error in FMC_INIT')
        ENDIF


        ! If we were given an original 3D reconstruction
        IF (PRESENT(MODEL3D)) SELF%THREED_ORIGINAL        =>  MODEL3D

        ! If we're going to be subtracting the filament, we need to original 3D reconstruction
        ! Make sure the user gave us the original (not finalised) 3D reconstruction
        IF (SELF%SUBTRACT_FIL_FROM_REF) THEN
            IF (.NOT. PRESENT(MODEL3D)) THEN
                WRITE(*,'(2A)') '**ERROR(FMC_INIT): non-finalised model 3D reconstruction is needed ',  &
                                'but was not supplied as an argument'
                CALL TERMINATE('Fatal error in FMC_INIT')
            ENDIF
        ENDIF

        ! Path oversampling?
        ! (this is done here because the step size calculation needs it, and that's done in FMC_NEW_FILAMENT)
        IF (PRESENT(PATH_OVERSAMPLING)) SELF%PATH_OVERSAMPLING = PATH_OVERSAMPLING

        ! Will we be treating waypoints as independent single particles?
        SELF%WPS_AS_SPS = .FALSE.
        IF (PRESENT(WPS_AS_SPS)) SELF%WPS_AS_SPS = WPS_AS_SPS
        ! If we're treating waypoints as single particles, we need to know how big the single particles need to be
        IF (SELF%WPS_AS_SPS) THEN
            IF (PRESENT(SP_SEGMENT_LENGTH)) THEN
                SELF%SP_SEGMENT_LENGTH = SP_SEGMENT_LENGTH
            ELSE
                WRITE(*,'(2A)') '**ERROR(FMC_INIT): when treating waypoints as single particles, ', &
                                    'the SP_SEGMENT_LENGTH parameter must be supplied'
                CALL TERMINATE('Fatal error in FMC_INIT')
            ENDIF
        ENDIF

        ! Deal with the interpolation scheme. First check it was setup properly, then copy it over to FMC
        IF (INTERPOLATION_SCHEME_CHECK_SETUP(INTERPOL)) THEN
            SELF%INTERPOL = INTERPOL
        ELSE
            WRITE(*,'(A)') '**ERROR(FMC_INIT): interpolation scheme object not setup!'
            CALL TERMINATE('Fatal error in FMC_INIT')
        ENDIF

        ! Target resolution of the refinement
        SELF%TARGET_RESOLUTION      =   TARGET_RESOLUTION

        ! The size of the 3D reconstruction
        SELF%THREED_SIZE            =   FMC_THREED_SIZE(SELF%INTERPOL,FIL%MAX_RADIUS)

        ! Setup the filament within the FMC object
        CALL FMC_NEW_FILAMENT(SELF,FIL,UPDATE_FIL_THREED=SELF%SINGLE_FILAMENT_REFINEMENT .OR. SELF%SUBTRACT_FIL_FROM_REF)
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FMC_INIT): after fmc_new_filament, filament info:'
            CALL FILAMENT_PRINT_INFO(SELF%FIL,VERBOSE=.TRUE.)
            WRITE(*,'(A)') '**debug(FMC_INIT): after fmc_new_filament, fil_original info:'
            CALL FILAMENT_PRINT_INFO(SELF%FIL_ORIGINAL,VERBOSE=.TRUE.)
        ENDIF

        ! We'll need to know the filament's maximum radius
        IF (SELF%FIL%MAX_RADIUS .EQ. 0.0) THEN
            WRITE(*,'(A)') '**ERROR(FMC_INIT): filament object does not specify maximum filament radius, which is needed.'
            CALL TERMINATE('Fatal error in FMC_INIT')
        ENDIF


        IF (DEBUG) PRINT '(A,I0,1X,F0.3)', '**debug(FMC_INIT): box_size, step = ', SELF%BOX_SIZE, SELF%STEP




        ! Compute the reconstruction of the filament from the original state, so that it can later
        ! be subtracted from the reference 3D
        ! Commented out because should be done by FMC_NEW_FILAMENT
        !IF (SELF%SUBTRACT_FIL_FROM_REF .OR. SELF%SINGLE_FILAMENT_REFINEMENT)  THEN
        !    CALL FMC_UPDATE_FIL_THREED(SELF)
        !ENDIF


        !
        ! What is the desired accuracy of the search
        !
        SELF%SEARCH_ACCURACY(1:6)   =   ACCURACY_FROM_RESOLUTION(   SELF%TARGET_RESOLUTION, &
                                                                    SELF%FIL%MAX_RADIUS*FIL%FILM%PIXEL_SIZE,    &
                                                                    SELF%THREED_SIZE,   &
                                                                    SELF%FIL%FILM%CTFTILT%CTF%WL,   &
                                                                    SELF%FIL%FILM%PIXEL_SIZE,   &
                                                                    AKV_IS_WL=.TRUE.    )
        PRINT '(A,F0.5,A)', '**warning(FMC_INIT): hard setting of subunit_num search accuracy at 0.1 (was ', &
                            SELF%SEARCH_ACCURACY(7), ')'
        SELF%SEARCH_ACCURACY(7) =   0.1

        ! Copy helical parameters over to the FMC object
        SELF%HELIX              =   HELIX

        ! Compute scaling factors for the search parameters
        CALL FMC_UPDATE_PARAM_SCALE(SELF)

        ! Maximum and minimum radii in reciprocal space, i.e. minimum and maximum resolution during refinement,
        ! in reciprocal pixels (Nyquist is 0.5)
        ! high_resolution_cutoff and target_resolution are in Angstroms
        IF (PRESENT(HIGH_RESOLUTION_CUTOFF)) THEN
            SELF%FREQ_MAX       =   SELF%FIL%FILM%PIXEL_SIZE / HIGH_RESOLUTION_CUTOFF
        ELSE
            SELF%FREQ_MAX       =   SELF%FIL%FILM%PIXEL_SIZE / SELF%TARGET_RESOLUTION
        ENDIF
        SELF%FREQ_MAX = MIN(SELF%FREQ_MAX,0.5)

        IF (PRESENT(LOW_RESOLUTION_CUTON)) THEN
            SELF%FREQ_MIN       =   SELF%FIL%FILM%PIXEL_SIZE / LOW_RESOLUTION_CUTON
        ELSE
            SELF%FREQ_MIN       =   1 / (10.*SELF%FIL%MAX_RADIUS)
        ENDIF

        ! We start off with a raw 3D model (i.e. not symmetrised/masked)
        SELF%MODEL3DVOL_READY   =   .FALSE.


        ! The refine_params flags are inherited from the filament object, in fmc_new_filament
        ! If we're refining theta per waypoint, but we're not refining Z, we probably mean to have Z rigid
        IF (SELF%REFINE_PARAMS(5)   .AND. .NOT. SELF%FIL%RIGID_PARAMS(5) &
                                    .AND. .NOT. SELF%REFINE_PARAMS(3)   &
                                    .AND. .NOT. SELF%FIL%RIGID_PARAMS(3)) THEN
            WRITE(*,'(3A)') '**warning(FMC_INIT): you are refining theta per-waypoint and you are not refining Z; ', &
                            'it is recommended you set Z to rigid so that individual waypoint Z values are updated ',   &
                            'in concert with their theta values'
            WRITE(*,'(A,7L1)') '**warning(FMC_INIT): refine_params = ', SELF%FIL%REFINE_PARAMS
            WRITE(*,'(A,7L1)') '**warning(FMC_INIT): rigid_params = ', SELF%FIL%RIGID_PARAMS
            IF (.NOT. SELF%WPS_AS_SPS) CALL TERMINATE('Bad choice of options')
        ENDIF


        ! The parameters for exhaustive searches (and the initial DS step)
        SELF%SEARCH_HALF_RANGE = SELF%SEARCH_ACCURACY * 2.
        SELF%SEARCH_SAMPLING   = SELF%SEARCH_ACCURACY / 4.

        ! If the user supplied overrides, apply them
        IF (PRESENT(SEARCH_HALF_RANGE_OVERRIDE)) THEN
            WHERE (SEARCH_HALF_RANGE_OVERRIDE .NE. 0.0)
                SELF%SEARCH_HALF_RANGE = SEARCH_HALF_RANGE_OVERRIDE
            END WHERE
        ENDIF

        ! Real space projection?
        IF (PRESENT(REAL_SPACE_PROJ)) SELF%REAL_SPACE_PROJ = REAL_SPACE_PROJ

        ! Do we know the expected molecular weight?
        IF (PRESENT(ASYM_UNIT_MW)) SELF%ASYM_UNIT_MW = ASYM_UNIT_MW


        ! Which minimiser will we use?
        IF (PRESENT(USE_CONJUGATE_GRADIENT)) SELF%USE_CONJUGATE_GRADIENT = USE_CONJUGATE_GRADIENT
        IF (PRESENT(USE_DIFFERENTIAL_EVOLUTION)) SELF%USE_DIFFERENTIAL_EVOLUTION = USE_DIFFERENTIAL_EVOLUTION
        IF (SELF%USE_DIFFERENTIAL_EVOLUTION .AND. SELF%USE_CONJUGATE_GRADIENT) THEN
            WRITE(*,'(A)') '**ERROR(FMC_INIT): cannot have both differential evolution and conjugate gradient minimization'
            CALL TERMINATE('Fatal error in FMC_INIT')
        ENDIF
        SELF%USE_DOWNHILL_SIMPLEX = .NOT. (SELF%USE_DIFFERENTIAL_EVOLUTION .OR. SELF%USE_CONJUGATE_GRADIENT)

        ! Deal with DS options for random & multiple starting conditions
        IF (PRESENT(NUM_MINIM_RUNS)) SELF%NUM_MINIM_RUNS = NUM_MINIM_RUNS
        IF (PRESENT(DS_RANDOM_START_CONF)) SELF%DS_RANDOM_START_CONF = DS_RANDOM_START_CONF
        IF (PRESENT(MINIM_RANDOM_START_POS)) SELF%MINIM_RANDOM_START_POS = MINIM_RANDOM_START_POS
        SELF%DS_MAX_NUM_ITER = DS_MAX_ITERATIONS
        IF (PRESENT(DS_MAX_NUM_ITER)) SELF%DS_MAX_NUM_ITER = DS_MAX_NUM_ITER

        ! It only makes sense to have more than one DS run if the starting conditions are randomised
        IF (SELF%NUM_MINIM_RUNS .GT. 1) THEN
            IF (.NOT. (SELF%DS_RANDOM_START_CONF .OR. SELF%MINIM_RANDOM_START_POS)) THEN
                WRITE(*,'(2A)') '**ERROR(FMC_INIT): without a randomised minimiser start, it does not make sense to ask ',  &
                                'for several minimisation runs'
                CALL TERMINATE('Fatal error in FMC_INIT')
            ENDIF
        ENDIF

        ! Use CC3?
        IF (PRESENT(USE_CC3)) SELF%USE_CC3 = USE_CC3

        ! Cubed average
        IF (PRESENT(CUBED_AVERAGE)) SELF%CUBED_AVERAGE = CUBED_AVERAGE

        ! Restraints
        IF (PRESENT(CURVATURE_RESTRAINT)) SELF%CURVATURE_RESTRAINT = CURVATURE_RESTRAINT
        IF (PRESENT(TORSION_RESTRAINT)) SELF%TORSION_RESTRAINT = TORSION_RESTRAINT
        IF (PRESENT(STRETCH_RESTRAINT)) SELF%STRETCH_RESTRAINT = STRETCH_RESTRAINT

        ! Cylindrical averaging
        IF (PRESENT(MODEL_AVE_CYL)) SELF%MODEL_AVE_CYL  =   MODEL_AVE_CYL

        ! Number of work areas
        SELF%NUM_WORK_AREAS = 1

        ! Projection caching
        IF (PRESENT(CACHE_PROJECTIONS)) SELF%CACHE_PROJECTIONS = CACHE_PROJECTIONS

        IF (SELF%CACHE_PROJECTIONS) THEN
            ! We use the search accuracy for keeping track of projection's Euler angles
            SELF%CACHE_EULER_TOL%PHI = SELF%SEARCH_ACCURACY(4) / 2.
            SELF%CACHE_EULER_TOL%THETA = SELF%SEARCH_ACCURACY(5) / 2.
            SELF%CACHE_EULER_TOL%PSI = SELF%SEARCH_ACCURACY(6) / 2.
            !
            CTF_TOL%PIX_UNITS   = .FALSE.
            CTF_TOL%CS          = 99999.
            CTF_TOL%WL          = 99999.
            CTF_TOL%AMPL_CONT   = 99999.
            CTF_TOL%DFMID1      = 1000.
            CTF_TOL%DFMID2      = 1000.
            CTF_TOL%ANGAST      = 360.
            CALL PROJECTION_CACHE_INIT( SELF%PROJ_CACHE,    &
                                        SELF%CACHE_EULER_TOL,   &
                                        SELF%SEARCH_ACCURACY(1)/SEARCH_ACCURACY_TO_DS_SIZE_RATIO,   &
                                        CTF_TOL)
            WRITE(*,'(A)') '**debug(FMC_INIT): projection cache info follows'
            CALL PROJECTION_CACHE_PRINT_INFO(SELF%PROJ_CACHE)
        ENDIF

        ! NM parameter scaling
        SELF%NM_PARAM_SCALE = 1.0
        WHERE (SELF%FIL%REFINE_NORMAL_MODES) SELF%NM_PARAM_SCALE = MAXVAL(ABS(SELF%FIL%NM_EIGENVECTORS))

        ! B factor
        SELF%MODEL3DVOL_BFACTOR = 0.0
        IF (PRESENT(MODEL3DVOL_BFACTOR)) SELF%MODEL3DVOL_BFACTOR = MODEL3DVOL_BFACTOR

        ! Hold parameter averages?
        IF (PRESENT(HOLD_PARAM_AVERAGE)) SELF%HOLD_PARAM_AVERAGE = HOLD_PARAM_AVERAGE
        IF (ANY(SELF%HOLD_PARAM_AVERAGE .AND. SELF%REFINE_PARAMS .AND. .NOT. SELF%FIL%REFINE_NORMAL_MODES)) THEN
            WRITE(*,'(A)') '**ERROR(FMC_INIT): cannot hold the average value of a parameter unless refining normal modes'
            CALL TERMINATE('Fatal error in FMC_INIT')
        ENDIF

        ! Deal with parameters whose wp-on-wp increments are refined
        IF (PRESENT(REFINE_PARAM_INCREMENTS)) SELF%REFINE_PARAM_INCREMENTS = REFINE_PARAM_INCREMENTS
        IF (ANY(SELF%REFINE_PARAM_INCREMENTS(1:6))) THEN
            WRITE(*,'(A)') '**ERROR(FMC_INIT): refining parameter increments is only supported for subunit_num'
            CALL TERMINATE('Fatal error in FMC_INIT. Programming error.')
        ENDIF

        ! Dimensions for which only integer value changes are allowed
        IF (PRESENT(INTEGER_CHANGES_ONLY)) SELF%INTEGER_CHANGES_ONLY = INTEGER_CHANGES_ONLY
        IF (ANY(SELF%INTEGER_CHANGES_ONLY(1:6))) THEN
            WRITE(*,'(A)') '**ERROR(FMC_INIT): integer-only refinement is only supported for subunit_num'
            CALL TERMINATE('Fatal error in FMC_INIT. Programming error.')
        ENDIF

        ! Cross-beta filter?
        IF (PRESENT(CROSS_BETA_FILTER)) SELF%CROSS_BETA_FILTER = CROSS_BETA_FILTER

        ! Refine waypoints simultaneously?
        IF (PRESENT(REFINE_WPS_SIMULTANEOUSLY)) SELF%REFINE_WPS_SIMULTANEOUSLY = REFINE_WPS_SIMULTANEOUSLY


        ! Allocate work memory
        CALL FMC_ALLOCATE_WORK_MEMORY(SELF)


        ! The C pointer is setup
        SELF%CPTR = C_LOC(SELF)

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FMC_INIT): filament info on exit:'
            CALL FILAMENT_PRINT_INFO(SELF%FIL)
        ENDIF

        !
        ! Debug override
        !
        !WRITE(*,'(A)') '**WARNING(FMC_INIT): overriding search parameters!!!'
        !SELF%SEARCH_HALFRANGE = [4.,4.,0.,0.,0.,2.5]
        !WRITE(*,'(A)') '**WARNING(FMC_INIT): overriding box and 3D dimensions!!!'
        !SELF%THREED_SIZE = 106
        !SELF%BOX_SIZE = 106

        ! FMC object has been initialised, remember this
        SELF%INITIALISED = .TRUE.

        WRITE(*,'(A)') '</fmc_init>'
    END SUBROUTINE FMC_INIT

    !>  \brief   Compute appropriate scale factors for the search parameters, to make them "comparable" to each other
    SUBROUTINE FMC_UPDATE_PARAM_SCALE(SELF)
        USE IMAGES_COMPARISON
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::  SELF
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
        ! Start work

        ! Check the maximum radius of the object is known
        IF (SELF%FIL%MAX_RADIUS .EQ. 0.0) THEN
            WRITE(*,'(A)') '**ERROR(FMC_UPDATE_PARAM_SCALE): need to know the maximum radius of the filament'
            CALL TERMINATE('Fatal error in FMC_UPDATE_PARAM_SCALE')
        ENDIF

        ! Compute the scale factors
        SELF%PARAM_SCALE(1:6) =  PARAM_SCALING_FROM_ACCURACY(SELF%SEARCH_ACCURACY(1:6),SELF%FIL%MAX_RADIUS)
        ! The scaling for the subunit_num search is the inverse of the rise per subunit
        IF (.NOT. SELF%HELIX%PIXEL_UNITS) THEN
            PRINT *, '**DEBUG(FMC_UPDATE_PARAM_SCALE): rise before conversion = ', SELF%HELIX%RISE
            CALL HELICAL_SYMMETRY_UNIT_CONVERT(SELF%HELIX,SELF%FIL%FILM%PIXEL_SIZE)
        ENDIF
        SELF%PARAM_SCALE(7) = 1 / SELF%HELIX%RISE
        WRITE(*,'(A)') '**warning(FMC_UPDATE_PARAM_SCALE): setting scale for subunit_num to 1.0'
        SELF%PARAM_SCALE(7) = 1.0

    END SUBROUTINE FMC_UPDATE_PARAM_SCALE

    !>  \brief  Return an appropriate box size
    INTEGER FUNCTION FMC_BOX_SIZE(INTERPOL,MAX_RAD)
        USE RUNTIME_PARAMETERS, only : PSF_RADIUS_PIXELS
        use UsefulFunctions, only : IsOdd
        IMPLICIT NONE
        ! Arguments
        TYPE(INTERPOLATION_SCHEME),         INTENT(IN)  ::  INTERPOL
        REAL,                               INTENT(IN)  ::  MAX_RAD     !<  Maximum radius of the object (filament) (pixels)
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        IF (DEBUG) PRINT '(A,F0.4)', '**debug(FMC_BOX_SIZE): interpol ROI_SAFE_RADIUS = ', INTERPOL%ROI_SAFE_RADIUS

        ! Given the filament's maximum radius, work out the minimum safe dimension for the box
        FMC_BOX_SIZE = CEILING(MAX(2.*(MAX_RAD+PSF_RADIUS_PIXELS),MAX_RAD/INTERPOL%ROI_SAFE_RADIUS))
        ! We prefer even dimensions
        IF (IsOdd(FMC_BOX_SIZE)) FMC_BOX_SIZE = FMC_BOX_SIZE + 1
        ! Debug
        IF (DEBUG) PRINT '(A,F0.3,1X,F0.3,1X,I0)', '**debug(FMC_BOX_SIZE): MAX_RAD, PSF_RADIUS_PIXELS, BOX_SIZE = ', &
                                                    MAX_RAD, PSF_RADIUS_PIXELS,FMC_BOX_SIZE
        ! Sanity check
        IF (FMC_BOX_SIZE .LT. 2 .OR. FMC_BOX_SIZE .GT. 4096) THEN
            WRITE(*,'(A,I0)') '**ERROR(FMC_BOX_SIZE): oops, box size falls outside expected range: ', FMC_BOX_SIZE
            CALL TERMINATE('Failed sanity check in FMC_BOX_SIZE. Programming error.')
        ENDIF
    END FUNCTION FMC_BOX_SIZE

    !>  \brief  Update the FMC's step size
    subroutine fmc_update_step_size(self,fil)
        use runtime_parameters, only : default_falloff_pixels
        implicit none
        ! Arguments
        type(filament_model_comparison),    intent(inout)   ::  self
        type(filament),                     intent(inout)   ::  fil     !<  Filament the FMC object will be used to score
        ! Private variables
        logical,    parameter   ::  debug   =   .false.
        real    ::  max_step, max_step_wp
        ! Start work
        if (debug) then
            WRITE(*,'(A)') '**debug(FMC_UPDATE_STEP_SIZE): working out a step size'
        endif
        max_step = FMC_MAX_STEP_SIZE(SELF,fil)
        if (debug) then
            write(*,'(a,f0.3,1x,f0.3)') '**debug(fmc_update_step_size): max_step_size, step_size_from_wp_to_wp_dist = ', &
                                                max_step, fmc_step_size_from_wp_to_wp_distance(self)
            write(*,'(a,i0)') '**debug(fmc_update_step_size): path oversampling = ', self%path_oversampling
        endif
        max_step_wp         =   fmc_step_size_from_wp_to_wp_distance(self)
        self%step           =   min(max_step,max_step_wp)
        self%step           =   self%step / real(self%path_oversampling)
        self%step_in_pixels =   .true.
        ! The minimum step size is default_falloff_pixels, since this is the width of the cosine falloff
        ! of the rectangular mask during the scoring function
        if (self%step .lt. default_falloff_pixels) then
            if (.not. self%gave_warning_about_step_size) then
            write(*,'(3a,f0.2,a)') '**warning(fmc_update_step_size): scoring function step size is too small. ', &
                        'This is probably because the filament is significantly deformed relative to the pixel size. ', &
                        'Resetting to the minimum step size of ', default_falloff_pixels, '. This warning will not be repeated.'
            endif
            self%step = default_falloff_pixels
            self%gave_warning_about_step_size = .true.
        endif
    end subroutine fmc_update_step_size

    !>  \brief  Update the FMC's box size
    SUBROUTINE FMC_UPDATE_BOX_SIZE(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::  SELF
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
        REAL                    ::  MIN_SAFE_BOX_SIZE
        ! Start work
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FMC_UPDATE_BOX_SIZE): working out a box size'
        ENDIF
        !
        ! Work out box size
        !
        IF (SELF%WPS_AS_SPS) THEN
            MIN_SAFE_BOX_SIZE = FMC_BOX_SIZE(SELF%INTERPOL,SELF%SP_SEGMENT_LENGTH/2./SELF%FIL%FILM%PIXEL_SIZE)
        ELSE
            ! We'll need to know the filament's maximum radius
            IF (SELF%FIL%MAX_RADIUS .EQ. 0.0) THEN
                WRITE(*,'(A)') '**ERROR(FMC_INIT): filament object does not specify maximum filament radius, which is needed.'
                CALL TERMINATE('Fatal error in FMC_INIT')
            ENDIF
            MIN_SAFE_BOX_SIZE = FMC_BOX_SIZE(SELF%INTERPOL,SELF%FIL%MAX_RADIUS)
        ENDIF
        SELF%BOX_SIZE       =   MIN_SAFE_BOX_SIZE
    END SUBROUTINE FMC_UPDATE_BOX_SIZE




    !>  \brief  Return an appropriate 3D reconstruction size
    PURE INTEGER FUNCTION FMC_THREED_SIZE(INTERPOL,MAX_RAD)
        use RUNTIME_PARAMETERS, only : PSF_RADIUS_PIXELS
        use UsefulFunctions, only : IsOdd
        IMPLICIT NONE
        ! Arguments
        TYPE(INTERPOLATION_SCHEME),         INTENT(IN)  ::  INTERPOL    !<  Interpolation scheme used in 3D reconstruction
        REAL,                               INTENT(IN)  ::  MAX_RAD     !<  Maximum radius of the object (filament) (pixels)
        ! Private variables

        ! Start work
        FMC_THREED_SIZE = MAX(  CEILING(MAX_RAD/INTERPOL%ROI_SAFE_RADIUS),  &
                                CEILING(2.*(MAX_RAD+PSF_RADIUS_PIXELS))     &
                                )

        ! We prefer even dimensions
        IF (IsOdd(FMC_THREED_SIZE)) FMC_THREED_SIZE = FMC_THREED_SIZE + 1
    END FUNCTION FMC_THREED_SIZE

    !>  \brief  Work out the maximum appropriate step size for the FMC scoring function
    real function fmc_max_step_size(self,fil)
        use filaments_geometry, only : filament_eval_average_curvature_overtwist_stretch
        use filaments_geometry, only : filament_eval_maximum_curvature_overtwist_stretch
		use runtime_parameters, only :  terminate, default_falloff_pixels
		use rotation_matrices, only : radians
        implicit none
        ! arguments
        type(filament_model_comparison),    intent(inout)  ::  self
        type(filament),                     intent(inout)  ::  fil     !<  filament the fmc will be used to score
        ! private variables
        REAL    ::  RAD
        real(kind=8)        ::  curvature, overtwist, stretch
        real(kind=4)        ::  curvature_max_step_size, overtwist_max_step_size, stretch_max_step_size
        logical, parameter  ::  limit_to_helical_pitch = .false.
        logical, parameter  ::  use_maxima = .true.    ! Use the maximum curvature, overtwist and stretch rather than average
        logical, parameter  ::  debug = .false.
        real,    parameter  ::  phase_error_tolerance = 12.0 ! Fraction of pi (e.g. 12.0 means tolerance is pi/12)
        ! Start work

        if (debug) write(*,'(a)') '**debug(fmc_max_step_size): entering'

        ! At most, we want the segment size to be such that the
        ! rectangular mask's corners stay safely away from the edge
        ! of the box, where safely is DEFAULT_FALLOFF_PIXELS
        ! In other words, the maximum step size is the length of the rectangle
        ! of width 2*SELF%FIL%MAX_RADIUS, which fits within a circle
        ! of radius 0.5*BOX_SIZE-DEFAULT_FALLOFF_PIXELS
        rad = 0.5*self%box_size-default_falloff_pixels
        fmc_max_step_size = 2. * sqrt(rad**2 - (fil%max_radius-default_falloff_pixels)**2 )

        fmc_max_step_size = min(fmc_max_step_size,self%box_size-3.*default_falloff_pixels-2.)

        ! Given a target resolution (frequency f) and a target phase residual
        ! at that resolution (%THETA_t) and assuming a small curvature (c),
        ! one can work out the maximum segment length L_max.
        ! See notes of 131213.
        !
        ! L_max = sqrt{(4 %THETA_t)/(c f %pi)}
        !
        ! Below, we set f = 1/4 = .25 /pixel and %THETA_t = pi/12
        !
        ! Similar reasoning applies to the overtwist and stretch. See notes of 140301.
        if (use_maxima) then
            call filament_eval_maximum_curvature_overtwist_stretch(fil,curvature,overtwist,stretch)
        else
            call filament_eval_average_curvature_overtwist_stretch(fil,curvature,overtwist,stretch)
        endif

        if (debug) write(*,'(a)') '**debug(fmc_max_step_size): worked out ave/max curvature'

        !
        curvature_max_step_size = sqrt( (4.*pi/phase_error_tolerance) /                                   &
                                    (max(real(curvature,kind=4),0.0005e0)/self%target_resolution*fil%film%pixel_size*pi))
        fmc_max_step_size = min(fmc_max_step_size,curvature_max_step_size)

        if (debug) write(*,'(a)') '**debug(fmc_max_step_size): about to do overtwist'

        if (.not. fil%fragmented_phi) then
            overtwist_max_step_size = (pi/phase_error_tolerance) /                                                &
                                    (max(real(radians(overtwist),kind=4),0.01e0)*fil%max_radius/ &
                                    self%target_resolution*fil%film%pixel_size*pi)
            fmc_max_step_size = min(fmc_max_step_size, overtwist_max_step_size)
        else
            overtwist_max_step_size = 0.0e0
        endif

        if (debug) write(*,'(a)') '**debug(fmc_max_step_size): about to do stretch'


        if (.not. fil%fragmented_subunit_num) then
            stretch_max_step_size = (pi/phase_error_tolerance) /                                            &
                                (max(real(stretch,kind=4),0.01e0)/self%target_resolution*fil%film%pixel_size*pi)
            fmc_max_step_size = min(fmc_max_step_size, stretch_max_step_size)
        else
            stretch_max_step_size = 0.0e0
        endif

        if (debug) then
			write(*,'(a,6(g15.5,1x))') '**debug(FMC_MAX_STEP_SIZE): curvature, overtwist, stretch, max step size = ', &
                            curvature, overtwist, stretch, curvature_max_step_size, overtwist_max_step_size, stretch_max_step_size
        endif
        ! We probably don't want to the step size to be greater than the helical pitch
        if (limit_to_helical_pitch) then
            if (.not. fil%helix%pixel_units) then
                call terminate('need helical parameters in pixels, not angstroms')
            endif
            fmc_max_step_size = min(    fmc_max_step_size, real(helix_pitch(fil%helix)) )
        endif

        ! Also, we will keep the step size to be .le. the filament diameter. This ensures that CTF artefacts are properly avoided
        ! when calculating projections.
        ! This is because the 3D model size has to be determined before the step size is known and it is set to
        ! >= the diameter + 2*psf_radius. Therefore we don't want our segment length to be > diameter, because otherwise the psf_radius
        ! doesn't protect us sufficiently
        fmc_max_step_size = min(fmc_max_step_size,2.0*(fil%max_radius+default_falloff_pixels))
    end function fmc_max_step_size

    !>  \brief  Work out a step size based on the average wp-to-wp distance in the filament
    REAL FUNCTION FMC_STEP_SIZE_FROM_WP_TO_WP_DISTANCE(SELF) RESULT(STEP_SIZE)
        USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_AVERAGE_WP_TO_WP_DISTANCE
		IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)  ::  SELF
        ! Private variables

        ! Start work
        STEP_SIZE = FILAMENT_AVERAGE_WP_TO_WP_DISTANCE(SELF%FIL,.NOT. SELF%FIL%FAL_WPS_IN_SCORE) / 4.
    END FUNCTION FMC_STEP_SIZE_FROM_WP_TO_WP_DISTANCE




    !>  \brief  Work out an order in which to process waypoints
    FUNCTION FMC_GET_WP_ORDER(SELF,START_FROM_HIGHEST_SCORE) RESULT(WP_ORDER)
        use UsefulFunctions, only : IsEven
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)  ::  SELF
        LOGICAL,            OPTIONAL,       INTENT(IN)  ::  START_FROM_HIGHEST_SCORE  !< The first waypoint will be the one with the highest score, rather than the mid waypoint
        ! Result
        INTEGER, ALLOCATABLE    ::  WP_ORDER(:)
        ! Private variables
        LOGICAL ::  SSTART_FROM_MID_WP, SSTART_FROM_HIGHEST_SCORE
        INTEGER ::  I,J,MID_WP
        INTEGER ::  IERR
        INTEGER ::  START_WP
        REAL    ::  HIGHEST_SCORE
        ! Start work

        ! How are we going to decide which WP is first?
        SSTART_FROM_HIGHEST_SCORE = .FALSE.
        IF (PRESENT(START_FROM_HIGHEST_SCORE)) SSTART_FROM_HIGHEST_SCORE = START_FROM_HIGHEST_SCORE
        SSTART_FROM_MID_WP = .NOT. SSTART_FROM_HIGHEST_SCORE

        ! If we need to know the scores, check they're there
        IF (SSTART_FROM_HIGHEST_SCORE) THEN
            IF (ALL(SELF%FIL%WP%SCORE .EQ. 0.0)) THEN
                WRITE(*,'(2A)') '**ERROR(FMC_GET_WP_ORDER): cannot start from highest scoring waypoints ',  &
                                'because waypoint scores are unknown'
                CALL TERMINATE('Fatal error in FMC_GET_WP_ORDER')
            ENDIF
        ENDIF


        ! Allocate memory for output array
        IF (ALLOCATED(WP_ORDER)) DEALLOCATE(WP_ORDER)
        ALLOCATE(WP_ORDER(SELF%FIL%NWP+1),STAT=IERR)
        IF (IERR .NE. 0) THEN
            CALL TERMINATE('Memory allocation failed in FMC_GET_WP_ORDER')
        ENDIF

        ! Work out the starting waypoint
        IF (SSTART_FROM_MID_WP) THEN
            IF (IsEven(SELF%FIL%NWP)) THEN
                MID_WP = SELF%FIL%NWP/2 + 1
            ELSE
                MID_WP = (SELF%FIL%NWP-1)/2 + 1
            ENDIF
            START_WP = MID_WP
        ELSE IF (SSTART_FROM_HIGHEST_SCORE) THEN
            HIGHEST_SCORE = -HUGE(1.0)
            DO I=1,SELF%FIL%NWP
                IF (SELF%FIL%WP(I)%SCORE .GT. HIGHEST_SCORE) THEN
                    HIGHEST_SCORE = SELF%FIL%WP(I)%SCORE
                    START_WP = I
                ENDIF
            ENDDO
        ELSE
            START_WP = 1
        ENDIF

        ! Add waypoint positions to the list
        J=1
        DO I=START_WP,SELF%FIL%NWP
            WP_ORDER(J) = I
            J = J + 1
        ENDDO
        IF (START_WP .GT. 1) THEN
            DO I=START_WP-1,1,-1
                WP_ORDER(J) = I
                J = J + 1
            ENDDO
        ENDIF
        WP_ORDER(SELF%FIL%NWP+1) = START_WP
    END FUNCTION FMC_GET_WP_ORDER




    !>  \brief  Prepare arrays containing the starting positions and first step useful for the DS minimiser
    SUBROUTINE FMC_START_STEP_VALUES(FMC,VALUES,STEP,VALUES_TO_FIL,DIM_IS_SUBUNIT_NUM)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),            INTENT(INOUT)   ::  FMC
        REAL(KIND=8),   ALLOCATABLE,                INTENT(INOUT)   ::  VALUES(:)
        REAL(KIND=8),   ALLOCATABLE,                INTENT(INOUT)   ::  STEP(:)
        LOGICAL,                        OPTIONAL,   INTENT(IN)      ::  VALUES_TO_FIL           !<  .FALSE. by default. If .TRUE. the values from the VALUES array are plugged back into the filament object (and nothing happens to the STEP array)
        LOGICAL,        ALLOCATABLE,    OPTIONAL,   INTENT(INOUT)   ::  DIM_IS_SUBUNIT_NUM(:)   !<  For each dimension which corresponds to subunit_num, true.
        ! Private variables
        INTEGER                 ::  NUM_DIM_TOTAL, COUNT_DIM
        LOGICAL                 ::  VVALUES_TO_FIL
        INTEGER                 ::  IERR, I, J
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        LOGICAL,    PARAMETER   ::  SKIP_SCALING = .FALSE.   ! For debug purposes
        CHARACTER(LEN=6)        ::  SIZE_VAL_C
        ! Start work

        IF (PRESENT(VALUES_TO_FIL)) THEN
            VVALUES_TO_FIL = VALUES_TO_FIL
        ELSE
            VVALUES_TO_FIL = .FALSE.
        ENDIF

        ! Work out how many dimensions
        NUM_DIM_TOTAL   =   FMC_NUM_DIM_TOTAL(FMC)

        ! If necessary, allocate memory to values and step arrays
        IF (.NOT. VVALUES_TO_FIL) THEN
            IF (ALLOCATED(VALUES)) THEN
                IF (SIZE(VALUES) .NE. NUM_DIM_TOTAL) DEALLOCATE(VALUES)
            ENDIF
			IF ( .NOT. ALLOCATED(VALUES) ) THEN
				ALLOCATE(VALUES(NUM_DIM_TOTAL),STAT=IERR)
				IF (IERR .NE. 0) THEN
	                WRITE(*,'(A,I0)') '**ERROR(FMC_START_STEP_VALUES): memory allocation failed with error ', IERR
	                CALL TERMINATE('Fatal error in FMC_START_STEP_VALUES')
                ENDIF
			ENDIF
			IF ( ALLOCATED(STEP) ) THEN
			    IF ( SIZE(STEP) .NE. NUM_DIM_TOTAL ) DEALLOCATE(STEP)
			ENDIF
			IF (.NOT. ALLOCATED(STEP)) THEN
				ALLOCATE(STEP(NUM_DIM_TOTAL),STAT=IERR)
				IF (IERR .NE. 0) THEN
					WRITE(*,'(A,I0)') '**ERROR(FMC_START_STEP_VALUES): memory allocation failed with error ', IERR
					CALL TERMINATE('Fatal error in FMC_START_STEP_VALUES')
				ENDIF
			ENDIF
        ENDIF

        IF (PRESENT(DIM_IS_SUBUNIT_NUM)) THEN
            IF (VVALUES_TO_FIL) CALL TERMINATE('Not implemented')
            IF (ALLOCATED(DIM_IS_SUBUNIT_NUM)) DEALLOCATE(DIM_IS_SUBUNIT_NUM)
            ALLOCATE(DIM_IS_SUBUNIT_NUM(NUM_DIM_TOTAL))
        ENDIF

        ! Copy data to or from the filament, with appropriate scaling
        IF (VVALUES_TO_FIL) THEN
            ! Scale values appropriately before copying to filament
            IF (SKIP_SCALING) THEN
                WRITE(*,'(A)') '**WARNING(FMC_START_STEP_VALUES): not scaling values'
            ELSE
                CALL FMC_SCALE_VALUES(FMC,VALUES,VVALUES_TO_FIL)
            ENDIF
            CALL VALUES_TO_FILAMENT(    FMC%FIL,VALUES,FMC%REFINE_PARAMS,FMC%HOLD_PARAM_AVERAGE,    &
                                        INCREMENTAL_VALUES=FMC%REFINE_PARAM_INCREMENTS  )
        ELSE
            CALL VALUES_FROM_FILAMENT(  FMC%FIL,VALUES,FMC%REFINE_PARAMS,FMC%HOLD_PARAM_AVERAGE,    &
                                        INCREMENTAL_VALUES=FMC%REFINE_PARAM_INCREMENTS)
            IF (DEBUG) THEN
                WRITE(SIZE_VAL_C,'(I6)') SIZE(VALUES)
                WRITE(*,'(A)') '**debug(FMC_START_STEP_VALUES): values from filament: '
                WRITE(*,'(A,'//SIZE_VAL_C//'(F0.3,X))') '**debug(FMC_START_STEP_VALUES):  ', VALUES
            ENDIF

            ! Scale values appropriately after reading from filament
            IF (SKIP_SCALING) THEN
                WRITE(*,'(A)') '**WARNING(FMC_START_STEP_VALUES): not scaling values'
            ELSE
                CALL FMC_SCALE_VALUES(FMC,VALUES,VVALUES_TO_FIL)
            ENDIF

            ! Calculate step size for real-space refinement
            COUNT_DIM = 0
            ! Loop over the waypoints
            DO I=1,FMC%FIL%NWP
                ! Loop over the parameters
                DO J=1,7
                    ! Is this parameter being refined?
                    IF (FMC%REFINE_PARAMS(J)) THEN
                        ! Is this a per-wp parameter, or a rigid parameter that applies to the filament as a whole
                        IF (FMC%FIL%RIGID_PARAMS(J)) THEN
                            IF (I .EQ. 1) THEN
                                ! This is a parameter that has a single value for the whole filament
                                COUNT_DIM = COUNT_DIM + 1
                                IF (.NOT. VVALUES_TO_FIL) THEN
                                    ! The first step for that dimension
                                    STEP(COUNT_DIM) = DBLE( FMC%SEARCH_HALF_RANGE(J) / FMC%PARAM_SCALE(J))
                                ENDIF
                                IF (PRESENT(DIM_IS_SUBUNIT_NUM)) DIM_IS_SUBUNIT_NUM(COUNT_DIM) = J .EQ. 7
                            ENDIF
                        ELSE
                            ! This is a per-wp parameter; is it being refined?
                            IF (FMC%FIL%WP(I)%REFINE) THEN
                                IF (FMC%FIL%REFINE_NORMAL_MODES(J) .AND. I .EQ. 1 .AND. FMC%HOLD_PARAM_AVERAGE(J)) CYCLE
                                COUNT_DIM = COUNT_DIM + 1
                                IF (.NOT. VVALUES_TO_FIL) THEN
                                    STEP(COUNT_DIM) = DBLE( FMC%SEARCH_HALF_RANGE(J) / FMC%PARAM_SCALE(J))
                                ENDIF
                                IF (PRESENT(DIM_IS_SUBUNIT_NUM)) DIM_IS_SUBUNIT_NUM(COUNT_DIM) = J .EQ. 7
                            ENDIF
                        ENDIF
                    ENDIF
                ENDDO
            ENDDO
        ENDIF

    END SUBROUTINE FMC_START_STEP_VALUES

    !>  \brief  Apply FMC scaling factors to a set of values
    SUBROUTINE FMC_SCALE_VALUES(FMC,VALUES,VALUES_TO_FIL)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::  FMC
        REAL(KIND=8),   ALLOCATABLE,        INTENT(INOUT)   ::  VALUES(:)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  VALUES_TO_FIL   !<  .FALSE. by default. If .TRUE. the values from the VALUES array are plugged back into the filament object (and nothing happens to the STEP array)
        ! Private variables

        ! Start work
        INTEGER ::  NUM_DIM_TOTAL, COUNT_DIM
        LOGICAL ::  VVALUES_TO_FIL
        INTEGER ::  I, J
        ! Start work

        IF (PRESENT(VALUES_TO_FIL)) THEN
            VVALUES_TO_FIL = VALUES_TO_FIL
        ELSE
            VVALUES_TO_FIL = .FALSE.
        ENDIF

        ! Work out how many dimensions
        NUM_DIM_TOTAL   =   FMC_NUM_DIM_TOTAL(FMC)

        ! Check allocation of values array
        IF (ALLOCATED(VALUES)) THEN
            IF (SIZE(VALUES) .NE. NUM_DIM_TOTAL) THEN
                WRITE(*,'(2(A,I0))') '**ERROR(FMC_SCALE_VALUES): was expecting ', NUM_DIM_TOTAL, ' values, but got ', SIZE(VALUES)
                CALL TERMINATE('Fatal error in FMC_SCALE_VALUES')
            ENDIF
        ELSE
            WRITE(*,'(A)') '**ERROR(FMC_SCALE_VALUES): array not allocated on entry'
            CALL TERMINATE('Fatal error in FMC_SCALE_VALUES')
        ENDIF

        COUNT_DIM = 0
        ! Loop over the waypoints
        DO I=1,FMC%FIL%NWP
            ! Loop over the parameters
            DO J=1,7
                ! Is this parameter being refined?
                IF (FMC%REFINE_PARAMS(J)) THEN
                    ! Is this a per-wp parameter, or a rigid parameter that applies to the filament as a whole
                    IF (FMC%FIL%RIGID_PARAMS(J)) THEN
                        IF (I .EQ. 1) THEN
                            ! This is a parameter that has a single value for the whole filament
                            COUNT_DIM = COUNT_DIM + 1
                            IF (VVALUES_TO_FIL) THEN
                                VALUES(COUNT_DIM) = VALUES(COUNT_DIM) * FMC%PARAM_SCALE(J) / FMC%NM_PARAM_SCALE(J)
                            ELSE
                                VALUES(COUNT_DIM) = VALUES(COUNT_DIM) / FMC%PARAM_SCALE(J) * FMC%NM_PARAM_SCALE(J)
                            ENDIF
                        ENDIF
                    ELSE
                        ! This is a per-wp parameter; is it being refined?
                        IF (FMC%FIL%WP(I)%REFINE) THEN
                            IF (FMC%FIL%REFINE_NORMAL_MODES(J) .AND. I .EQ. 1 .AND. FMC%HOLD_PARAM_AVERAGE(J)) CYCLE
                            COUNT_DIM = COUNT_DIM + 1
                            IF (VVALUES_TO_FIL) THEN
                                VALUES(COUNT_DIM) = VALUES(COUNT_DIM) * FMC%PARAM_SCALE(J) / FMC%NM_PARAM_SCALE(J)
                            ELSE
                                VALUES(COUNT_DIM) = VALUES(COUNT_DIM) / FMC%PARAM_SCALE(J) * FMC%NM_PARAM_SCALE(J)
                            ENDIF
                        ENDIF
                    ENDIF
                ENDIF
            ENDDO
        ENDDO

        ! Sanity check
        IF (COUNT_DIM .NE. NUM_DIM_TOTAL) THEN
            WRITE(*,'(A,I0,A,I0)') '**ERROR(FMC_SCALE_VALUES): expected ', NUM_DIM_TOTAL,  ' values, but got ', COUNT_DIM
            CALL TERMINATE('Sanity check failed')
        ENDIF
    END SUBROUTINE FMC_SCALE_VALUES


    !>  \brief  Work out the total number of dimensions of the search space
    INTEGER FUNCTION FMC_NUM_DIM_TOTAL(SELF) RESULT(NUM_DIM_TOTAL)
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)  ::  SELF
        ! Private variables

        ! Start work
        NUM_DIM_TOTAL = FILAMENT_EXPECTED_NUMBER_OF_VALUES(SELF%FIL,SELF%REFINE_PARAMS,SELF%HOLD_PARAM_AVERAGE)
    END FUNCTION FMC_NUM_DIM_TOTAL


    !>  \brief  Copy a value to a waypoint
    SUBROUTINE FMC_VALUE_TO_WP(SELF,WP,PARAM_INDEX,VALUE)
        USE COOS, ONLY : VALUE_TO_WAYPOINT
        IMPLICIT NONE
        ! Argument
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::  SELF
        TYPE(WAYPOINT),                     INTENT(INOUT)   ::  WP
        INTEGER,                            INTENT(IN)      ::  PARAM_INDEX     !<  Parameter index. 1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta, 6.Psi, 7.Subunit_num
        REAL(KIND=8),                       INTENT(IN)      ::  VALUE           !<  Value to be copied over
        ! Private variables
        REAL(KIND=8)    ::  VVALUE
        ! Start work

        ! Apply parameter scaling
        VVALUE = VALUE * SELF%PARAM_SCALE(PARAM_INDEX)

        ! Copy value to waypoint
        CALL VALUE_TO_WAYPOINT(WP,VVALUE,PARAM_INDEX)
    END SUBROUTINE FMC_VALUE_TO_WP

    !>  \brief  Get a value from a waypoint
    SUBROUTINE FMC_VALUE_FROM_WP(SELF,WP,PARAM_INDEX,VALUE)
        USE COOS, ONLY : VALUE_FROM_WAYPOINT
        IMPLICIT NONE
        ! Argument
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(IN)      ::  SELF
        TYPE(WAYPOINT),                     INTENT(IN)      ::  WP
        INTEGER,                            INTENT(IN)      ::  PARAM_INDEX     !<  Parameter index. 1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta, 6.Psi, 7.Subunit_num
        REAL(KIND=8),                       INTENT(OUT)     ::  VALUE           !<  Value to be copied over
        ! Private variables

        ! Start work

        CALL VALUE_FROM_WAYPOINT(WP,VALUE,PARAM_INDEX)

        ! Apply parameter scaling
        VALUE = VALUE / SELF%PARAM_SCALE(PARAM_INDEX)

    END SUBROUTINE FMC_VALUE_FROM_WP

    !>  \brief  Print information on the current state of the filament model comparison (fmc) object
    !>  \todo   Code this
    SUBROUTINE FMC_PRINT(FMC_PTR) BIND(C)
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(C_PTR),    VALUE   ::  FMC_PTR
        ! Private variables
        TYPE(FILAMENT_MODEL_COMPARISON),    POINTER ::  FMC
        ! Start work
        ! Convert the C pointer to an F pointer
        CALL C_F_POINTER(FMC_PTR,FMC)
        !
        !
        CALL FILAMENT_PRINT_INFO(FMC%FIL)
        !FMC%FIL%PRINT_INFO()
    END SUBROUTINE FMC_PRINT

END MODULE FMCS_BASE
