!>  \brief  Functions to evaluate the simiarity between a filament and a model
MODULE FMCS_EVALUATION

    USE FMCS_BASE

    CONTAINS


    !>  \brief  Prepare the helical model so that it can be used for the comparison routines. This includes cropping/padding to appropriate dimensions, imposing helical symmetry, degridding and masking.
    !!          Note that this is only done when necessary.
    SUBROUTINE FMC_MODEL3D_PREP(FMC,THREED_SIZE,MASK_RAD)
        USE IMAGES_HELICES
        USE IMAGES_ARITHM
        USE MASKS
        USE HELICES
        USE PROCFS, ONLY : PRINT_MEMORY_USAGE_STATS
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    TARGET,     INTENT(INOUT)   ::  FMC         !<  Filament model comparison (FMC) object
        INTEGER,                            OPTIONAL,   INTENT(IN)      ::  THREED_SIZE !<  If present, this argument will force the model3D to have this size.
        REAL,                               OPTIONAL,   INTENT(IN)      ::  MASK_RAD    !<  Radius of spherical mask to impose on the 3D model. Pixels.
        ! Private variables
        !INTEGER                        ::  NUM_CENTRAL_SLICES
        LOGICAL                     ::  DIFFERENT_HELIX
        LOGICAL,    PARAMETER       ::  DEBUG   =   .FALSE.
        INTEGER                     ::  TTHREED_SIZE
        LOGICAL                     ::  SUBTRACT_FIL_FROM_REF
        REAL                        ::  MMASK_RAD
        LOGICAL                     ::  NEW_MASK
        LOGICAL, PARAMETER          ::  SKIP_HEXTEND =  .FALSE.  ! For debug purposes, to save time
        LOGICAL, PARAMETER          ::  USE_NEW_HEXTEND = .TRUE.
        REAL                        ::  RISE_TOL, TWIST_TOL
        ! Start work

        ! Memory footprint diag
        !IF (DEBUG) THEN
        !    WRITE(*,'(A)') '**debug(FMC_MODEL3D_PREP): memory usage on entry'
        !    CALL PRINT_MEMORY_USAGE_STATS()
        !ENDIF

        ! Check the model is cubic
        IF (.NOT. IS_SQUARE(FMC%MODEL3DVOL_PRIVATE)) THEN
            WRITE(*,'(A,3(I0,1X))') '**ERROR(FMC_MODEL3D_PREP): sorry, 3D model must be in a cubic box. Image supplied is ', &
                                    FMC%MODEL3DVOL_PRIVATE%logical_dimensions
            CALL TERMINATE('Fatal error in FMC_MODEL3D_PREP')
        ENDIF

        ! What size are we looking for
        IF (PRESENT(THREED_SIZE)) THEN
            TTHREED_SIZE = THREED_SIZE
        ELSE
            TTHREED_SIZE = FMC%BOX_SIZE
            !TTHREED_SIZE = FMC%THREED_SIZE  ! 1-SEP-2011 TEST
        ENDIF

        !
        ! Are the filament's helical parameters significantly different from the FMC's helical parameters?
        !
        RISE_TOL            =   FMC_HELICAL_RISE_TOLERANCE(FMC)
        TWIST_TOL           =   FMC_HELICAL_TWIST_TOLERANCE(FMC)
        DIFFERENT_HELIX     =   ABS(FMC%HELIX%TWIST - FMC%FIL%HELIX%TWIST) .GT. TWIST_TOL .OR. &
                                ABS(FMC%HELIX%RISE  - FMC%FIL%HELIX%RISE)  .GT. RISE_TOL


        ! Will we need to subtract a new filament from the reference 3D reconstruction?
        SUBTRACT_FIL_FROM_REF = .FALSE.
        IF (FMC%SUBTRACT_FIL_FROM_REF) THEN
            IF (FMC%ID_OF_SUBTRACTED_FIL .EQ. 0 .OR. FMC%ID_OF_SUBTRACTED_FIL .NE. FMC%FIL%ID) THEN
                SUBTRACT_FIL_FROM_REF = .TRUE.
            ENDIF
        ENDIF

        !
        ! What mask radius will we use?
        !
        IF (PRESENT(MASK_RAD)) THEN
            MMASK_RAD = MASK_RAD
        ELSE
            IF (FMC%WPS_AS_SPS) THEN
                MMASK_RAD = FMC%SP_SEGMENT_LENGTH / 2.0 / FMC%FIL%FILM%PIXEL_SIZE
            ELSE
                MMASK_RAD = FMC%FIL%MAX_RADIUS
            ENDIF
        ENDIF

        !
        ! Are we using a new mask radius?
        !
        NEW_MASK = .FALSE.
        IF (ABS(MMASK_RAD - FMC%MODEL3DVOL_MASK_RAD) .GT. 0.5) THEN
            NEW_MASK = .TRUE.
        ENDIF

        !
        ! Do we need to do anything?
        !
        IF ( FMC%MODEL3DVOL_READY .AND. TTHREED_SIZE .EQ. FMC%MODEL3DVOL_PRIVATE%logical_dimensions(1) .AND. &
            .NOT. DIFFERENT_HELIX .AND. .NOT. SUBTRACT_FIL_FROM_REF .AND. .NOT. NEW_MASK) THEN
            ! Everything is ready
            RETURN
        ELSE
            ! Let the user know what we're doing
            WRITE(*,'(2A)') '**info(FMC_MODEL3D_PREP): preparing the model 3D volume for use as a reference ',   &
                            'in scoring function (resizing, imposing helical symmetry, masking etc.)'
            IF (DEBUG) THEN
                WRITE(*,'(A,5L1)') '**debug(FMC_MODEL3D_PREP): logicals = ', FMC%MODEL3DVOL_READY, &
                        TTHREED_SIZE .EQ. FMC%MODEL3DVOL_PRIVATE%logical_dimensions(1), .NOT. DIFFERENT_HELIX, &
                         .NOT. SUBTRACT_FIL_FROM_REF, .NOT. NEW_MASK
            ENDIF
            !PRINT '(A,I4)', '**debug(FMC_MODEL3D_PREP): preparing model 3D (pad/crop, helical sym, masking), box size = ', FMC%BOX_SIZE
            FMC%MODEL3DVOL_PRIVATE = FMC%MODEL3DVOL_ORIGINAL

            ! Reset the projection cache
            IF (FMC%CACHE_PROJECTIONS) THEN
                CALL PROJECTION_CACHE_CLEAN(FMC%PROJ_CACHE)
            ENDIF

            ! Print info regarding the helical symmetrisation
            IF (DIFFERENT_HELIX) THEN
                PRINT '(4(A,F0.4))', '**info(FMC_MODEL3D_PREP): using the filament-specific helical parameters of ', &
                                    FMC%FIL%HELIX%TWIST, ' and ', FMC%FIL%HELIX%RISE, &
                                        ' rather than ', FMC%HELIX%TWIST, ' and ', FMC%HELIX%RISE
                FMC%HELIX = FMC%FIL%HELIX
            ELSE
                PRINT '(5(A,F0.4))', '**info(FMC_MODEL3D_PREP): the FMC helical parameters (',FMC%HELIX%TWIST,   &
                                        ' and ', FMC%HELIX%RISE, &
                                       ') are close enough to this filament''s helical parameters (',FMC%FIL%HELIX%TWIST,    &
                                       ' and ', FMC%FIL%HELIX%RISE, ') and is left unchanged at ', FMC%HELIX%TWIST
            ENDIF

            ! Memory footprint diag
            !IF (DEBUG) THEN
            !    WRITE(*,'(A)') '**debug(FMC_MODEL3D_PREP): memory usage before crop/pad'
            !    CALL PRINT_MEMORY_USAGE_STATS()
            !ENDIF

            !
            ! Check whether padding or cropping necessary
            !
            !PRINT *, '**debug(): before crop/pad FMC%MODEL3DVOL%TYPE = ', FMC%MODEL3DVOL%TYPE
            !PRINT *, '**debug(): before crop/pad FMC%MODEL3DVOL_ORIGINAL%TYPE = ', FMC%MODEL3DVOL_ORIGINAL%TYPE
            IF (FMC%BOX_SIZE .LT. FMC%MODEL3DVOL_PRIVATE%logical_dimensions(1)) THEN
                PRINT '(A,I3)', '**info(FMC_MODEL3D_PREP): cropping input 3D to ', TTHREED_SIZE
                CALL CROP(FMC%MODEL3DVOL_PRIVATE,[TTHREED_SIZE,TTHREED_SIZE,TTHREED_SIZE])
            ELSE IF (FMC%BOX_SIZE .GT. FMC%MODEL3DVOL_PRIVATE%logical_dimensions(1)) THEN
                PRINT '(2(A,I0))', '**info(FMC_MODEL3D_PREP): padding input 3D from ', &
                                    FMC%MODEL3DVOL_PRIVATE%logical_dimensions(1) , &
                                    ' to ', TTHREED_SIZE
                CALL PAD(FMC%MODEL3DVOL_PRIVATE,[TTHREED_SIZE,TTHREED_SIZE,TTHREED_SIZE])
            ENDIF

            !
            ! Subtract the filament from the model 3D reconstruction
            !
            IF (SUBTRACT_FIL_FROM_REF) THEN
                IF (FMC%SUBTRACT_FIL_FROM_REF) THEN
                    WRITE(*,'(A,I0,A)') '**info(FMC_MODEL3D_PREP): subtracting current state of filament ', FMC%FIL%ID, &
                                        ' from the reference 3D reconstruction'
                    ! Subtract the filament reconstruction from the model 3D reconstruction
                    WRITE(*,'(A)') '**info(FMC_MODEL3D_PREP): subtracting current reconstruction from reference 3D'
                    CALL THREED_ADD(FMC%THREED_ORIGINAL,FMC%FIL_THREED,FMC%FIL_THREED,SUBTRACT=.TRUE.)
                    ! Finalise this, and store as MODEL3DVOL
                    !CALL WRITE_TO_DISK(FMC%MODEL3DVOL,'dbg_model3d.mrc')
                    !CALL THREED_WRITE_TO_DISK(FMC%THREED_ORIGINAL,'dbg_threed.mrc')
                    CALL THREED_EVAL_RECONSTRUCTION(FMC%FIL_THREED,FMC%MODEL3DVOL_PRIVATE)
                    !CALL WRITE_TO_DISK(FMC%MODEL3DVOL_PRIVATE,'dbg_model3d_sub.mrc')
                    !CALL THREED_WRITE_TO_DISK(FMC%FIL_THREED,'dbg_filthreed.mrc')
                    ! Remember that we subtracted from this filament
                    FMC%ID_OF_SUBTRACTED_FIL = FMC%FIL%ID
                    !IF (DEBUG) CALL WRITE_TO_DISK(FMC%MODEL3DVOL_PRIVATE, 'dbg_model3dprep_aftersubtract.mrc')
                ELSE
                    WRITE(*,'(A)') '**ERROR(FMC_MODEL3D_PREP): bug - it should be impossible for SUBTRACT_FIL_FROM_REF '//&
                                    'to be true without FMC%SUBTRACT_FIL_FROM_REF!'
                    CALL TERMINATE('Fatal error in FMC_MODEL3D_PREP')
                ENDIF
            ENDIF

            ! Use the current filament's 3D reconstruction as a reference for refinement
            IF (FMC%SINGLE_FILAMENT_REFINEMENT) THEN
                WRITE(*,'(A)') '**info(FMC_MODEL3D_PREP): evaluating single filament reconstruction'
                CALL THREED_EVAL_RECONSTRUCTION(FMC%FIL_THREED,FMC%MODEL3DVOL_PRIVATE)
                IF (DEBUG) CALL WRITE_TO_DISK(FMC%MODEL3DVOL_PRIVATE, 'dbg_model3dprep_fil3d.mrc')
            ENDIF



            ! Memory footprint diag
            !IF (DEBUG) THEN
            !    WRITE(*,'(A)') '**debug(FMC_MODEL3D_PREP): memory usage before HEXTEND'
            !    CALL PRINT_MEMORY_USAGE_STATS()
            !ENDIF

            IF (FMC%MODEL_AVE_CYL) THEN
                IF (DEBUG) CALL WRITE_TO_DISK(FMC%MODEL3DVOL_PRIVATE, 'dbg_model3dprep_beforecyl.mrc')
                WRITE(*,'(A)') '**info(FMC_MODEL3D_PREP): imposing cylindrical averaging on volume'
                CALL IMAGE_CYL_AVERAGE(FMC%MODEL3DVOL_PRIVATE)
                IF (DEBUG) CALL WRITE_TO_DISK(FMC%MODEL3DVOL_PRIVATE, 'dbg_model3dprep_cyl.mrc')
            ELSE
                !
                ! Impose helical symmetry
                !
                IF (DEBUG) CALL WRITE_TO_DISK(FMC%MODEL3DVOL_PRIVATE, 'dbg_model3dprep_nomask_nohextend.mrc')
                IF (SKIP_HEXTEND) THEN
                    WRITE(*,'(A)') '**DEBUG(FMC_MODEL3D_PREP): NOT IMPOSING H SYMMETRY'
                ELSE
                    WRITE(*,'(A)',ADVANCE='NO') '**info(FMC_MODEL3D_PREP): imposing helical symmetry on model 3D: '
                    IF (USE_NEW_HEXTEND) THEN
                        CALL IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW(FMC%MODEL3DVOL_PRIVATE,FMC%HELIX,FMC%INTERPOL,  &
                                                                MAX_RADIUS=FMC%FIL%MAX_RADIUS)
                    ELSE
                        CALL IMAGE_IMPOSE_HELICAL_SYMMETRY( FMC%MODEL3DVOL_PRIVATE,FMC%HELIX,FMC%INTERPOL,  &
                                                            MAX_RADIUS=FMC%FIL%MAX_RADIUS,PROGRESS_REPORT=.TRUE.)
                    ENDIF
                ENDIF

                IF (DEBUG) CALL WRITE_TO_DISK(FMC%MODEL3DVOL_PRIVATE, 'dbg_model3dprep_nomask.mrc')
                !IF (DEBUG) CALL TERMINATE('DEBUG FMC_MODEL3D_PREP')

                ! Memory footprint diag
                !IF (DEBUG) THEN
                !    WRITE(*,'(A)') '**debug(FMC_MODEL3D_PREP): memory usage after HEXTEND'
                !    CALL PRINT_MEMORY_USAGE_STATS()
                !ENDIF

                !
                ! Apply B factor
                !
                IF (FMC%MODEL3DVOL_BFACTOR .NE. 0.0) THEN
                    !WRITE(*,'(A,F0.3,A)') '**info(FMC_MODEL3D_PREP): applying B factor (', FMC%MODEL3DVOL_BFACTOR,') to model 3D'
                    IF (FMC%MODEL3DVOL_PRIVATE%is_in_real_space) CALL FWD_FT(FMC%MODEL3DVOL_PRIVATE)
                    ! We want the falloff to start at the target resolution, and end 5 pixels later
                    WRITE(*,'(A,F0.2,A,2(F0.2,1X))') '**info(FMC_MODEL3D_PREP): applying B factor of ', FMC%MODEL3DVOL_BFACTOR, &
                                                    ' Angstroems^2, with cosine-edged low-pas filter with parameters: ', &
                                                    MIN(0.5,FMC%FREQ_MAX + (5.0/REAL(FMC%THREED_SIZE))), (5.0/REAL(FMC%THREED_SIZE))
                    CALL APPLY_BFACTOR( FMC%MODEL3DVOL_PRIVATE,FMC%MODEL3DVOL_BFACTOR/FMC%FIL%FILM%PIXEL_SIZE**2,   &
                                        LP_COSINE=(FMC%FREQ_MAX .LT. 0.5),                                     &
                                        LP_COSINE_FREQ=MIN(0.5,FMC%FREQ_MAX + (5.0/REAL(FMC%THREED_SIZE))),    &
                                        LP_COSINE_FALLOFF=(5.0/REAL(FMC%THREED_SIZE)))
                    IF (.not. FMC%MODEL3DVOL_PRIVATE%is_in_real_space) CALL BWD_FT(FMC%MODEL3DVOL_PRIVATE)
                ENDIF

                !
                ! Masking (envelope of the filament)
                !
                WRITE(*,'(A)') '**info(FMC_MODEL3D_PREP): calculating and applying mask to model 3D'
                !PRINT *, '**DEBUG(): NOT H MASKING'
                CALL IMAGE_HELIX_MASK(FMC%MODEL3DVOL_PRIVATE,FMC%HELIX,FMC%ASYM_UNIT_MW,FMC%FIL%FILM%PIXEL_SIZE,FMC%FIL%MAX_RADIUS)

                ! Memory footprint diag
                !IF (DEBUG) THEN
                !    WRITE(*,'(A)') '**debug(FMC_MODEL3D_PREP): memory usage after masking'
                !    CALL PRINT_MEMORY_USAGE_STATS()
                !ENDIF
            ENDIF

            IF (DEBUG) call write_to_disk(FMC%MODEL3DVOL_PRIVATE,'dbg_before_correct3d.mrc')

            !
            ! Degridding
            !
            ! If the projection will happen in real space, the correction needs to happen in Fourier
            ! space, and vice versa.
            !
            IF (FMC%REAL_SPACE_PROJ) THEN
                IF (FMC%MODEL3DVOL_PRIVATE%is_in_real_space) CALL FWD_FT(FMC%MODEL3DVOL_PRIVATE)
            ELSE
                IF (.not. FMC%MODEL3DVOL_PRIVATE%is_in_real_space) CALL BWD_FT(FMC%MODEL3DVOL_PRIVATE)
            ENDIF

            WRITE(*,'(A)') '**info(FMC_MODEL3D_PREP): dividing by reciprocal of interpolation function (aka (de)gridding)'
            CALL CORRECT_3D(    FMC%MODEL3DVOL_PRIVATE, &
                                FMC%INTERPOL%INTERPOLATION_INSTR_FUNCTION,  &
                                FMC%INTERPOL%WINDOWING_INSTR_FUNCTION,      &
                                FMC%INTERPOL%WINDOW_WIDTH,                  &
                                REVERSE=.FALSE. )

            IF (.not. FMC%MODEL3DVOL_PRIVATE%is_in_real_space) CALL BWD_FT(FMC%MODEL3DVOL_PRIVATE)
            IF (DEBUG) call write_to_disk(FMC%MODEL3DVOL_PRIVATE,'dbg_after_correct3d.mrc')


            ! Memory footprint diag
            !IF (DEBUG) THEN
            !    WRITE(*,'(A)') '**debug(FMC_MODEL3D_PREP): memory usage after correct_3d'
            !    CALL PRINT_MEMORY_USAGE_STATS()
            !ENDIF

            !
            ! Masking (cosine-edged sphere) (commented out 5-sep-2011)
            !
            !IF (DEBUG) WRITE(*,'(A,F0.3)') '**info(FMC_MODEL3D_PREP): applying cosine disc with radius (pixels) ', MMASK_RAD+2.*FMC%MODEL3DVOL_PRIVATE%logical_dimensions(1)/2.0*DEFAULT_FALLOFF
            !CALL COSINE_DISC(FMC%MODEL3DVOL_PRIVATE,MMASK_RAD+2.*FMC%MODEL3DVOL_PRIVATE%logical_dimensions(1)/2.0*DEFAULT_FALLOFF,FMC%MODEL3DVOL_PRIVATE%logical_dimensions(1)/2.0*DEFAULT_FALLOFF,WIPE_INPUT=.FALSE.)
            !FMC%MODEL3DVOL_MASK_RAD = MMASK_RAD

            !
            ! Debug dump
            !
            IF (DEBUG) THEN
                !PRINT *, '**debug(FMC_MODEL3D_PREP): may dump prep 3d to disk ', FMC%FIL%ID
                !IF (FMC%FIL%ID .EQ. 1) THEN
                    PRINT '(A)', '**debug(FMC_MODEL3D_PREP): dumping prepared 3D to disk'
                    CALL WRITE_TO_DISK(FMC%MODEL3DVOL_PRIVATE, 'dbg_model3dprep.mrc')
                    !CALL TERMINATE ('debug')
                !ENDIF
            ENDIF

            !
            ! We may need to go to Fourier space and/or apply a phase shift
            !
            IF (.NOT. FMC%REAL_SPACE_PROJ) THEN
                IF (FMC%MODEL3DVOL_PRIVATE%is_in_real_space) CALL FWD_FT(FMC%MODEL3DVOL_PRIVATE)
                IF (FMC%MODEL3DVOL_PRIVATE%ORIGIN_AT_CENTER) CALL SHIFT_REAL_SPACE_ORIGIN(FMC%MODEL3DVOL_PRIVATE)
            ENDIF

            !
            ! The model is now ready
            !
            FMC%MODEL3DVOL_READY = .TRUE.
            WRITE(*,'(A)') '**info(FMC_MODEL3D_PREP): all done'
        ENDIF
    END SUBROUTINE FMC_MODEL3D_PREP



    !>  \brief  Energy function to be called by the FGSL SA minimiser
    FUNCTION FMC_EVAL_SA_ENERGY(FMC_PTR) BIND(C)
        USE, INTRINSIC  ::  ISO_C_BINDING
        IMPLICIT NONE
        ! Arguments
        TYPE(C_PTR),    VALUE   ::  FMC_PTR !<  Pointer to filament model comparison object
        ! Result
        REAL(KIND=C_DOUBLE)     ::  FMC_EVAL_SA_ENERGY
        ! Private variables
        TYPE(FILAMENT_MODEL_COMPARISON),    POINTER ::  FMC
        ! Start work
        CALL C_F_POINTER(FMC_PTR,FMC)
        FMC_EVAL_SA_ENERGY = - FMC_EVAL(FMC,FMC%FIL,DETAILLED_SCORE_INFO=.FALSE.)
    END FUNCTION FMC_EVAL_SA_ENERGY

    !>  \brief  Wrapper to be called by the FGSL DS minimiser
    !!
    !!  The values and parameters are passed as C pointers so that this subroutine can be called by the (F)GSL.
    !!  The parameter passed is in fact a FILAMENT_MODEL_COMPARISON object, which contains all the information needed
    !!  to do the comparison.
    FUNCTION FMC_EVAL_WRAP(VALUES,PARAMS) BIND(C)
        USE, INTRINSIC :: ISO_C_BINDING
        USE FGSL, ONLY : FGSL_DOUBLE, FGSL_VECTOR, FGSL_INT, FGSL_OBJ_C_PTR, FGSL_VECTOR_ALIGN, FGSL_SUCCESS
		USE RUNTIME_PARAMETERS, ONLY : TERMINATE
		USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(C_PTR), VALUE  ::  VALUES  !<  C pointer to current values to try out
        TYPE(C_PTR), VALUE  ::  PARAMS  !<  C pointer to a FMC object
        ! Result
        REAL(KIND=C_DOUBLE) ::  FMC_EVAL_WRAP   !<  Result of the function
        ! Private variables
        TYPE(FILAMENT_MODEL_COMPARISON),    POINTER     ::  FMC
        REAL(FGSL_DOUBLE),                  POINTER     ::  VAL_FGSL(:)
        REAL(KIND=8),                       ALLOCATABLE ::  VAL(:)
        REAL(KIND=8),                       POINTER     ::  VAL_PTR(:)
        TYPE(FGSL_VECTOR)                               ::  VAL_VEC
        INTEGER                                         ::  EXPECTED_NUM_VALUES
        INTEGER(FGSL_INT)                               ::  STATUS
        INTEGER                                         ::  NUM_DIM_TOTAL, IERR
        LOGICAL,                            PARAMETER   ::  DEBUG           =   .FALSE.
        LOGICAL,                            PARAMETER   ::  SANITY_CHECK    =   .FALSE.
        REAL                                            ::  DUMMY
        TYPE(FILAMENT)                                  ::  FIL
        LOGICAL,                            PARAMETER   ::  SKIP_SCALING    =   .FALSE. ! For debugging purposes
        CHARACTER(LEN=6)                                ::  SIZE_VAL_C
        ! Start work


        CALL C_F_POINTER(PARAMS,FMC)
        ! Work out how many dimensions
        NUM_DIM_TOTAL = FMC_NUM_DIM_TOTAL(FMC)
        EXPECTED_NUM_VALUES = NUM_DIM_TOTAL


        IF (FMC%USE_DOWNHILL_SIMPLEX) THEN
            ! Get FGSL_DOUBLE reals from the incoming C pointer
            CALL FGSL_OBJ_C_PTR(VAL_VEC,VALUES)
            STATUS = FGSL_VECTOR_ALIGN(VAL_FGSL,VAL_VEC)
            IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('FGSL_VECTOR_ALIGN failed in FMC_EVAL_WRAP')
            ! Check we got the expected number of variables
            IF (SIZE(VAL_FGSL) .NE. EXPECTED_NUM_VALUES) THEN
                WRITE(*,'(A,2(I0,1X))') '**ERROR(FMC_EVAL_WRAP): did not get expected number of variables ', &
                                    SIZE(VAL_FGSL), EXPECTED_NUM_VALUES
                CALL TERMINATE('Fatal error in FMC_EVAL_WRAP')
            ENDIF

            ! Convert from FGSL to REAL
            ALLOCATE(VAL(EXPECTED_NUM_VALUES),STAT=IERR)
            IF (IERR .NE. 0) THEN
                WRITE(*,'(A,I0)') '**ERROR(FMC_EVAL_WRAP): allocation of real values array failed with error', IERR
                CALL TERMINATE('Fatal error in FMC_EVAL_WRAP')
            ENDIF
            VAL = VAL_FGSL
        ELSE
            CALL C_F_POINTER(VALUES,VAL_PTR,SHAPE=(/EXPECTED_NUM_VALUES/))
            VAL = VAL_PTR
            IF (SIZE(VAL) .NE. EXPECTED_NUM_VALUES) THEN
                WRITE(*,'(A,2(I0,1X))') '**ERROR(FMC_EVAL_WRAP): did not get expected number of variables ', &
                                SIZE(VAL), EXPECTED_NUM_VALUES
                CALL TERMINATE('Fatal error in FMC_EVAL_WRAP')
            ENDIF

        ENDIF

        !IF (DEBUG) THEN
        !   WRITE(*,'(A,<SIZE(VAL)>(G0.2,X))') '**debug(FMC_EVAL_WRAP): trying these values: ', VAL
        !   WRITE(*,'(A,<SIZE(VAL)>(G0.5,X))') '**debug(FMC_EVAL_WRAP): trying these values: ', VAL
        !ENDIF

        !IF (DEBUG) THEN
        !   WRITE(*,'(A)') '**debug(FMC_EVAL_WRAP): filament info before values_to_fil'
        !   CALL FILAMENT_PRINT_INFO(FMC%FIL)
        !ENDIF

        ! Restore filament to original state
        FIL = FMC%FIL_ORIGINAL

        !IF (DEBUG) THEN
        !   WRITE(*,'(A)') '**debug(FMC_EVAL_WRAP): filament info before values_to_fil'
        !   CALL FILAMENT_PRINT_INFO(FIL)
        !ENDIF

        ! Assign the new positions & angles to the filament's waypoints
        IF (SKIP_SCALING) THEN
            WRITE(*,'(A)') '**WARNING(FMC_EVAL_WRAP): not scaling values'
        ELSE
            CALL FMC_SCALE_VALUES(FMC,VAL,VALUES_TO_FIL=.TRUE.)
        ENDIF
        CALL VALUES_TO_FILAMENT(FIL,VAL,FMC%REFINE_PARAMS,FMC%HOLD_PARAM_AVERAGE,INCREMENTAL_VALUES=FMC%REFINE_PARAM_INCREMENTS)


        IF (DEBUG) THEN
           WRITE(*,'(A)') '**debug(FMC_EVAL_WRAP): filament info after values_to_fil, before scoring routine'
           CALL FILAMENT_PRINT_DIFF(FMC%FIL_ORIGINAL,FIL)
           CALL FILAMENT_PRINT_INFO(FIL)
        ENDIF

        ! We can now call the actual function that does the scoring
        FMC_EVAL_WRAP = - FMC_EVAL(FMC,FIL,DETAILLED_SCORE_INFO=.FALSE.,WRITE_FRC_PLOTS=.FALSE.)

        ! Remember we called the scoring function
        FMC%CALLS_FROM_MINIMISER = FMC%CALLS_FROM_MINIMISER + 1

        IF (DEBUG) THEN
            WRITE(SIZE_VAL_C,'(I6)') SIZE(VAL)
            WRITE(*,'(A,I0,A,'//SIZE_VAL_C//'(F0.3,1X),A,F0.6,A)') '**debug(FMC_EVAL_WRAP): call # ', FMC%CALLS_FROM_MINIMISER, &
                                                             ' , values (score): ', VAL , '( ', -FMC_EVAL_WRAP, ' )'
        ENDIF

        IF (SANITY_CHECK) THEN
            DUMMY = - FMC_EVAL(FMC,FIL,DETAILLED_SCORE_INFO=.FALSE.,THREADING=.FALSE.)
            IF (DEBUG) WRITE(*,'(A,F0.6)') '**debug(FMC_EVAL_WRAP): got this score (2nd time, no threading): ', DUMMY
            IF ((DUMMY - FMC_EVAL_WRAP) .GT. 0.000001) THEN
                WRITE(*,'(A,F0.6,A,F0.6,A,G15.3)') '**ERROR(FMC_EVAL_WRAP): non-theaded score (', DUMMY, &
                                                    ') different from threaded score (', FMC_EVAL_WRAP, &
                                                    '). Difference = ', DUMMY - FMC_EVAL_WRAP
                CALL TERMINATE('Fatal error in FMC_EVAL_WRAP')
            ENDIF
        ENDIF

        !PRINT *, '**debug(FILAMENT_EVAL_MODEL_AGREEMENT_WRAP): done'
    END FUNCTION FMC_EVAL_WRAP

    !>  \brief  Score the comparison between the filament and the model
    REAL FUNCTION FMC_EVAL(SELF,FIL,WRITE_STACKS,STACKS_ROOTNAME,DETAILLED_SCORE_INFO,THREADING,WRITE_FRC_PLOTS,DONT_USE_CACHE)
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),    INTENT(INOUT)   ::  SELF
        TYPE(FILAMENT),                     INTENT(INOUT)   ::  FIL
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  WRITE_STACKS            !<  If .TRUE., write stacks to disk of segmented filament and matching projections
        CHARACTER(LEN=*),       OPTIONAL,   INTENT(IN)      ::  STACKS_ROOTNAME         !<  Root file name to use when writing stacks out.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  DETAILLED_SCORE_INFO    !<  Whether to print out detailled info on the score at each point of path along the filament. Defaults to false.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  THREADING               !<  If .FALSE., OpenMp threading will be switched off. If .TRUE., OpenMP threading will be determined by the enviroment etc. as usual.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  WRITE_FRC_PLOTS         !<  Defaults to false. If true, will attempt to plot the segment FRCs to disk as a png file.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  DONT_USE_CACHE          !<  Make sure that the projection cache is not used, even if available
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        INTEGER                 ::  NUMBER_OF_CALLS, CURRENT_CALL
        REAL                    ::  SCORE_CHECK
        TYPE(FILAMENT)          ::  FIL_CHECK(2)
        ! Start work

        ! Are we going to call the scoring routine twice, for testing purposes?
        IF (DEBUG) THEN
            NUMBER_OF_CALLS = 2
        ELSE
            NUMBER_OF_CALLS = 1
        ENDIF
        ! Call the scoring subroutine
        DO CURRENT_CALL = 1, NUMBER_OF_CALLS
            IF (DEBUG) THEN
                WRITE(*,'(A,I0,A,I0)') '**debug(FMC_EVAL): filament info before call ', CURRENT_CALL, ' of ', NUMBER_OF_CALLS
                CALL FILAMENT_PRINT_INFO(SELF%FIL)
            ENDIF
            FMC_EVAL = FMC_EVAL_SLAVE(             SELF,                                        &
                                                   FIL,                                         &
                                                   WRITE_STACKS=WRITE_STACKS,                   &
                                                   STACKS_ROOTNAME=STACKS_ROOTNAME,             &
                                                   DETAILLED_SCORE_INFO=DETAILLED_SCORE_INFO,   &
                                                   THREADING=THREADING,                         &
                                                   WRITE_FRC_PLOTS=WRITE_FRC_PLOTS,             &
                                                   DONT_USE_CACHE=DONT_USE_CACHE)
            IF (NUMBER_OF_CALLS .GT. 1 .AND. CURRENT_CALL .EQ. 1) SCORE_CHECK = FMC_EVAL
            IF (DEBUG) THEN
                WRITE(*,'(A,I0,A,I0)') '**debug(FMC_EVAL): filament info after call ', CURRENT_CALL, ' of ', NUMBER_OF_CALLS
                CALL FILAMENT_PRINT_INFO(SELF%FIL)
                FIL_CHECK(CURRENT_CALL) = SELF%FIL
            ENDIF
        ENDDO

        ! Sanity check
        IF (DEBUG) THEN
            IF (ABS(SCORE_CHECK - FMC_EVAL) .GT. .00001) THEN
                WRITE(*,'(A,2(G15.6,1X))') '**ERROR(FMC_EVAL): the scoring routine was called twice, with different results: ', &
                                        FMC_EVAL, SCORE_CHECK
                CALL FILAMENT_PRINT_DIFF(FIL_CHECK(1),FIL_CHECK(2))
                CALL TERMINATE('Failed sanity check in FMC_EVAL')
            ENDIF
        ENDIF
    END FUNCTION FMC_EVAL


    !>  \brief  Measure how well a model matches a filament by measuring the total correlation of the model's projection
    !!          along the filament's path
    !!  \todo   Move masks generation to outside this subroutine?
    FUNCTION FMC_EVAL_SLAVE(    SELF,FIL,       &
                                WRITE_STACKS,STACKS_ROOTNAME,DETAILLED_SCORE_INFO,                  &
                                THREADING,                  &
                                WRITE_FRC_PLOTS,DONT_USE_CACHE)                                    &
                                    RESULT(SCORE)
        USE IMAGES_COMPARISON, ONLY : IMAGE_PREPARE, PROJECTION_PREPARE, CCLIN_NUM_DENOMS_RINGS, CC3_FINALISE, CCLIN_NUM_DENOMS
		USE IMAGES_COMPARISON, ONLY : CC3, CCLIN, CCLIN_NUM_DENOMS_FROM_RINGS
        USE IMAGES_CTF, ONLY : CTF_TO_IMG
        USE OMP_LIB
        USE ROTATION_MATRICES, ONLY : IDENTITY_ROTMAT
        USE FSCS, ONLY : COMPUTE_FSC
        USE PARSERS, ONLY : STRING_EQUAL, EXTENSION_FROM_FILENAME, IS_BLANK
#ifdef HAVE_PLPLOT
        USE PLOTS, ONLY : SERIES_TO_PNG   ! For plotting the FRCs
#endif
        USE FILAMENTS_MODEL_UPDATE, ONLY : FILAMENT_UPDATE_MODEL
		USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_DEFOCUS_RANGE, FILAMENT_EVAL_TOTAL_CURVATURE_OVERTWIST_STRETCH
		USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_SPLINES_ARE_MONOTONIC
		USE FILAMENTS_PATH, ONLY : FILAMENT_GET_PATH, FILAMENT_T_INTERVAL_FOR_SCORING_FUNCTION, ADD_WAYPOINT_TO_PATH
		USE FILAMENTS_NORMAL_MODES, ONLY : FILAMENT_NM_AMPLITUDES_WEIGHTED_SUM
		USE COOS, ONLY : WAYPOINTS_PRINT_INFO, WP_PRINT_INFO
		USE FILMS, ONLY : FILM_FIND_CTF_AT_COOS
		USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
		USE MASKS, ONLY : FOURIER_MASK_INIT_AMYLOID
		USE IMAGES
		USE RUNTIME_PARAMETERS, only : DEFAULT_FALLOFF_PIXELS, SAFE_FOR_SQRT, FILENAME_LEN
		use ImageFiles
		USE, INTRINSIC :: IEEE_ARITHMETIC, ONLY : IEEE_IS_NAN
		use filaments_geometry, only : NEAREST_WP_FROM_T
!#ifdef __INTEL_COMPILER
!        USE COOS, ONLY : POINT_ASSIGN
!#endif
		IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT_MODEL_COMPARISON),INTENT(INOUT)   ::  SELF                    !<  Filament-model comparison object
        TYPE(FILAMENT),                 INTENT(INOUT)   ::  FIL                     !<  Filament object
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  WRITE_STACKS            !<  If .TRUE., write stacks to disk of segmented filament and matching projections
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  STACKS_ROOTNAME         !<  Root file name to use when writing stacks out.
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  DETAILLED_SCORE_INFO    !<  Whether to print out detailled info on the score at each point of path along the filament. Defaults to false.
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  THREADING               !<  If .FALSE., no threading will be used, even if the OpenMP environment is set up
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  WRITE_FRC_PLOTS         !<  Whether to write plots of the FRCs to disk (png files). Defaults to false.
        LOGICAL,            OPTIONAL,   INTENT(IN)      ::  DONT_USE_CACHE          !<  Even if it is available, do not use the projection cache.
        ! Result
        REAL                                            ::  SCORE
        ! Private variables
        REAL                            ::  SCORE_LOCAL                             ! Local, normalised score
        REAL(KIND=8)                    ::  SCORE_ND_LOCAL(3), SCORE_ND_WHOLE(3)    ! ND stands for numerator denominator
        REAL(KIND=8)                    ::  SCORE_ND_WHOLE_PRIVATE(3)
        REAL(KIND=8),       ALLOCATABLE ::  CC3_ND_RINGS_LOCAL(:,:), CC3_ND_RINGS_WHOLE(:,:)    ! ND stands for numerator denominator
        REAL(KIND=8),       ALLOCATABLE ::  CC3_ND_RINGS_WHOLE_PRIVATE(:,:)
        INTEGER                         ::  I, J, IERR
        INTEGER                         ::  ICOO(3)                                             ! integer coordinates
        REAL(KIND=8)                    ::  RCOO(3)                                             ! remainder coordinates
        TYPE(CTF_PARAMETERS)            ::  CTF
        LOGICAL                         ::  WWRITE_STACKS, WWRITE_FRC_PLOTS, DDETAILLED_SCORE_INFO
        LOGICAL                         ::  CCALC_FRC
        CHARACTER(LEN=FILENAME_LEN)     ::  BOXES_FILENAME, PROJS_FILENAME, AMPLS_FILENAME, AMPL_MASKS_FILENAME
        REAL                            ::  MATRIX(4,4)
        INTEGER                         ::  NUM_LOCAL_THREADS                                   ! Number of openmp threads
        INTEGER                         ::  LOCAL_THREAD_NUM                                    ! Current thread number
        INTEGER                         ::  LOC_COUNT
        INTEGER                         ::  NUM_WPS
        CHARACTER(LEN=6)                ::  NUM_WPS_C                                           ! Used in format specs
        LOGICAL                         ::  TTHREADING                                          ! Whether we will thread this scoring function
        CHARACTER(LEN=FILENAME_LEN)     ::  TMP_STRING
        !INTEGER                         ::  LLUN_BOX, LLUN_PROJ                                ! Unit numbers for files
        INTEGER                         ::  MINIM_NWP
        REAL,               PARAMETER   ::  SOFT_EDGE               =   DEFAULT_FALLOFF_PIXELS  !<  The width of the soft-edge falloff for masking (as a fraction of maximum inner radius)
        LOGICAL                         ::  WWHOLE_FILAMENT_NORM                                !<  If true, the score is the correlation is calculated over the whole filament, rather than an average of normalised CC of segments
        LOGICAL,            PARAMETER   ::  ENVELOPE_MASK                           =   .TRUE.  !<  If true, the input images (boxes) and projections are masked in real space using a binarised (& softened) version of the matching projection (before CTF)
        LOGICAL,            PARAMETER   ::  GET_CTF_FROM_Z                          =   .TRUE.  !<  If true, the Z coordinate of a waypoint is used to get the required CTF.
                                                                                                !!  If .FALSE., the X,Y coordinates, in conjunction with the CTFTILT information is used instead.
        LOGICAL,            PARAMETER   ::  DEBUG                                   =   .FALSE.
        LOGICAL,            PARAMETER   ::  NORMALISED_SCORING_FUNCTION             =   .TRUE.  !<  Whether a normalised CC should be used as a scoring function
        LOGICAL,            PARAMETER   ::  SINGLE_CTF_PER_FILAMENT_IF_REASONABLE   =   .FALSE. !<  To save time, one can use a single CTF function for the whole filament. This can be
                                                                                                !!  safely done if the filament's Z does not vary more than desired defocus accuracy.
        LOGICAL,			PARAMETER	::	WWRITE_AMPL_STACK			            =	.FALSE.
		LOGICAL,            PARAMETER   ::  RETURN_MINUS_ONE_IF_NOT_MONOTONIC       =   .TRUE.  !<  If the phi & subunit_num splines are not monotonic, there is something funny going, so let's return a bad score. (this may give a performance hit)
		TYPE(IMAGEFILE)                 ::  BOXES_FILE
        TYPE(IMAGEFILE)                 ::  PROJS_FILE
		TYPE(IMAGEFILE)					::	AMPLS_FILE
		TYPE(IMAGEFILE)                 ::  AMPL_MASKS_FILE
        LOGICAL                         ::  BOX_IS_BLANK, RETURN_MINUS_ONE
        REAL                            ::  OBJ_MAX_RAD
        LOGICAL                         ::  FIRST_TIME                                      ! helps keep track when doing cc3 calculation
        REAL                            ::  RECT_DIM(2)
        LOGICAL                         ::  DDONT_USE_CACHE
        LOGICAL                         ::  OONE_CTF_FOR_WHOLE_FILAMENT
        REAL                            ::  PROJ_MAX_RAD
        ! Variables for FRC
        REAL,           ALLOCATABLE     ::  FRC_TMP(:), FRC(:,:), FRC_TMP_TWOD(:,:)
        LOGICAL,        allocatable     ::  WP_HAS_SCORE_WP_NEARBY(:)
        REAL(KIND=8)                    ::  DUMMY
#if defined(HAVE_PLPLOT)
        CHARACTER(LEN=FILENAME_LEN)     ::  FRCS_FILENAME
#endif
        ! Variables for phase-shifting due to subunit_num
        TYPE(WAYPOINT)                  ::  WP_PHASED
        REAL(KIND=8)                    ::  MODEL_Z_SHIFT
        REAL,           ALLOCATABLE     ::  CTF_CURVE(:), CTF_CURVE_SUM(:), CTF_CURVE_SUM_PRIVATE(:)
        ! Curvature, torsion & stretch penalties, maximum-likelihood-like
        LOGICAL                         ::  ESTIMATE_NOISE_VARIANCE
        REAL                            ::  NUMBER_OF_PIXELS_IN_SCORING_FUNCTION
        REAL                            ::  ESTIMATED_NOISE_VARIANCE   ! see notes of 9-Jul-2011 and Sigworth 1998
        REAL                            ::  ESTIMATED_SIGNAL_VARIANCE   ! see notes of 9-Jul-2011 and Sigworth 1998
        REAL(KIND=8)                    ::  CURVATURE, TORSION, STRETCH
        REAL                            ::  CURVATURE_PENALTY
        REAL                            ::  PENALTY_FACTOR
        REAL                            ::  TORSION_PENALTY
        REAL                            ::  STRETCH_PENALTY
        INTEGER                         ::  START_WP, FINISH_WP ! If only a few waypoints are being refined, it may not make sense to keep
        REAL(KIND=8)                    ::  START_T, FINISH_T
        LOGICAL,        PARAMETER       ::  OONLY_SCORE_REFINING_NEIGHBOURHOOD = .TRUE.
        !
        LOGICAL,        PARAMETER       ::  APPLY_Z_SHIFT_IN_3D     =   .TRUE.  !<  False makes projection caching more efficient.
        ! Penalty for NM amplitudes
        LOGICAL,        PARAMETER       ::  NM_AMP_PENALTY          =   .FALSE.
        REAL,           PARAMETER       ::  NM_AMP_PENALTY_FACTOR   =   0.5
        REAL                            ::  NM_AMP_WEIGHTED_SUM



        ! Start work

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FMC_EVAL_SLAVE): filament info on entry:'
            CALL FILAMENT_PRINT_INFO(FIL)
        ENDIF

        ! Will we print out scoring info along the path?
        DDETAILLED_SCORE_INFO = .FALSE.
        IF (PRESENT(DETAILLED_SCORE_INFO)) DDETAILLED_SCORE_INFO = DETAILLED_SCORE_INFO
        IF (DDETAILLED_SCORE_INFO) THEN
            WRITE(*,'(A)') '**info(FMC_EVAL_SLAVE): detailled scoring information follows'
        ENDIF


        ! Check the interpolation object was setup properly
        IF (.NOT. INTERPOLATION_SCHEME_CHECK_SETUP(SELF%INTERPOL)) THEN
            WRITE(*,'(A)') '**ERROR(FMC_EVAL_SLAVE): interpolation scheme object not setup!'
            CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
        ENDIF


        ! Will we be writing stacks to disk?
        WWRITE_STACKS = .FALSE.
        IF (PRESENT(WRITE_STACKS)) WWRITE_STACKS = WRITE_STACKS

        ! Will we be writing FRC plots to disk?
        WWRITE_FRC_PLOTS = .FALSE.
        IF (PRESENT(WRITE_FRC_PLOTS)) WWRITE_FRC_PLOTS = WRITE_FRC_PLOTS

        ! Is the user overriding the use of projection cache?
        DDONT_USE_CACHE = .FALSE.
        IF (PRESENT(DONT_USE_CACHE)) DDONT_USE_CACHE = DONT_USE_CACHE

        ! Let the user know if a cache is available but will not be used
        IF (SELF%CACHE_PROJECTIONS .AND. DDONT_USE_CACHE) THEN
            WRITE(*,'(A)') '**info(FMC_EVAL_SLAVE): the projection cache will not be used, projections will be (re)computed'
        ENDIF

        ! If we're writing stacks to disk, what filenames will we use?
        IF (WWRITE_STACKS) THEN
            IF (PRESENT(STACKS_ROOTNAME)) THEN
                TMP_STRING = TRIM(ADJUSTL(STACKS_ROOTNAME))
                IF (STRING_EQUAL(EXTENSION_FROM_FILENAME(TMP_STRING),'   ')) THEN
                    BOXES_FILENAME = TRIM(ADJUSTL(TMP_STRING))//'_boxes.mrc'
                    PROJS_FILENAME = TRIM(ADJUSTL(TMP_STRING))//'_projs.mrc'
					AMPLS_FILENAME = TRIM(ADJUSTL(TMP_STRING))//'_boxes_ampl.mrc'
					AMPL_MASKS_FILENAME = TRIM(ADJUSTL(TMP_STRING))//'_boxes_ampl_mask.mrc'
                ELSE
                    BOXES_FILENAME = TMP_STRING(1:LEN_TRIM(TMP_STRING)-4)//'_boxes'//TMP_STRING(LEN_TRIM(TMP_STRING)-3:)
                    PROJS_FILENAME = TMP_STRING(1:LEN_TRIM(TMP_STRING)-4)//'_projs'//TMP_STRING(LEN_TRIM(TMP_STRING)-3:)
					AMPLS_FILENAME = TMP_STRING(1:LEN_TRIM(TMP_STRING)-4)//'_boxes_ampl'//TMP_STRING(LEN_TRIM(TMP_STRING)-3:)
			    AMPL_MASKS_FILENAME = TMP_STRING(1:LEN_TRIM(TMP_STRING)-4)//'_boxes_ampl_mask'//TMP_STRING(LEN_TRIM(TMP_STRING)-3:)
                ENDIF
            ELSE
                BOXES_FILENAME = 'boxes.mrc'
                PROJS_FILENAME = 'projs.mrc'
				AMPLS_FILENAME = 'boxes_ampl.mrc'
				AMPL_MASKS_FILENAME = 'boxes_ampl_mask.mrc'
            ENDIF
        ENDIF

        ! If we're writing stacks to disk, initialise the files
        IF (WWRITE_STACKS) THEN
            CALL BOXES_FILE%INIT(BOXES_FILENAME,DIM_X=SELF%BOX_SIZE,DIM_Y=SELF%BOX_SIZE,DIM_Z=1)
            CALL PROJS_FILE%INIT(PROJS_FILENAME,DIM_X=SELF%BOX_SIZE,DIM_Y=SELF%BOX_SIZE,DIM_Z=1)
			IF (WWRITE_AMPL_STACK) THEN
			CALL AMPLS_FILE%INIT(AMPLS_FILENAME,DIM_X=SELF%BOX_SIZE,DIM_Y=SELF%BOX_SIZE,DIM_Z=1)
			CALL AMPL_MASKS_FILE%INIT(AMPL_MASKS_FILENAME,DIM_X=SELF%BOX_SIZE,DIM_Y=SELF%BOX_SIZE,DIM_Z=1)
			ENDIF
		ENDIF


        ! Are we calculating FRC between image and matching projection?
        CCALC_FRC = DDETAILLED_SCORE_INFO
        IF (CCALC_FRC) THEN
            IF (DEBUG) WRITE(*,'(A)') '**info(FMC_EVAL_SLAVE): disabling FRC computations, to reduce output verbosity'
            CCALC_FRC = .FALSE.
        ENDIF

        ! If waypoints are independent single particles, it makes no sense to do whole-filament normalisation of the score
        WWHOLE_FILAMENT_NORM = .NOT. SELF%WPS_AS_SPS


        ! Cubed average implies normalisation within each segment
        IF (SELF%CUBED_AVERAGE .AND. WWHOLE_FILAMENT_NORM) THEN
            IF (DEBUG) WRITE(*,'(A)') '**debug(FMC_EVAL_SLAVE): switching whole-filament normalisation off, '//&
                                        'since you asked for average of cubed segments scores'
            WWHOLE_FILAMENT_NORM = .FALSE.
        ENDIF

        ! Cubed averaging cannot be done with whole-filament normalisation
        IF (WWHOLE_FILAMENT_NORM) THEN
            IF (SELF%CUBED_AVERAGE) THEN
                WRITE(*,'(A)') '**ERROR(FMC_EVAL_SLAVE): options WWHOLE_FILAMENT_NORM and CUBED_AVERAGE are incompatible'
                CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
            ENDIF
        ENDIF

        ! Are we estimating the noise variance? (we'll only do so if we need to)
        ESTIMATE_NOISE_VARIANCE = .FALSE.

        ! If we want to apply the curvature penalty, ...
        IF (SELF%CURVATURE_RESTRAINT) THEN
            ! ... we need to have an estimate of the persistence length
            IF (FIL%PERSISTENCE_LENGTH == 0.0) THEN
                WRITE(*,'(A)') '**ERROR(FMC_EVAL_SLAVE): curvature penalties cannot be applied without'//&
                                                        ' an estimate of the filament persistence length'
                CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
            ENDIF
            ! ... we need to be using linear cross-correlation, since we don't know how to implement MLL restraints a la Sigworth with weighted correlation coefficients
            IF (SELF%USE_CC3) THEN
                WRITE(*,'(A)') '**ERROR(FMC_EVAL_SLAVE): ML-like curvature restraints are not implemented '//&
                                'in conjunction with weigted correlation coefficient'
                CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE. Feature not implemented.')
            ENDIF
            ! ... we need to estimate the noise variance
            ESTIMATE_NOISE_VARIANCE = .TRUE.
        ENDIF

        ! We only know how to estimate noise variance if we're doing whole-filament normalisation
        IF (ESTIMATE_NOISE_VARIANCE .AND. .NOT. WWHOLE_FILAMENT_NORM) THEN
             WRITE(*,'(A)') '**ERROR(FMC_EVAL_SLAVE): have not implemented noise variance estimator in case where '//&
                            'whole-filament normalisation is not activated'
             CALL TERMINATE('Feature not yet implemented in FMC_EVAL_SLAVE')
        ENDIF

        ! Check the filament has enough waypoints
        IF (.NOT. FIL%FAL_WPS_IN_SCORE) THEN
            MINIM_NWP = 4
        ELSE
            MINIM_NWP = 3
        ENDIF

        IF (FIL%NWP .LT. MINIM_NWP) THEN
            WRITE(*,'(3(A,I0),A)') '**ERROR(FMC_EVAL_SLAVE): filament ', FIL%ID, ' only has ', FIL%NWP, &
                                    ' waypoints. A minimum of ', MINIM_NWP, ' are needed.'
            CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
        ENDIF

        ! Diagnostics
        IF (WWRITE_STACKS) THEN
            PRINT '(4A)', '**info(FMC_EVAL_SLAVE): dumping stacks ', TRIM(ADJUSTL(BOXES_FILENAME)), &
                                                            ' and ', TRIM(ADJUSTL(PROJS_FILENAME))
        ENDIF

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FMC_EVAL_SLAVE): filament info before updating model:'
            CALL FILAMENT_PRINT_INFO(FIL)
        ENDIF



        !
        ! Make sure our filament model is up-to-date
        !
        CALL FILAMENT_UPDATE_MODEL(FIL,WPS_AS_SPS=SELF%WPS_AS_SPS)

        !
        ! Now that the filament model has been updated, we know the curvature, overtwist etc, so we can update the step size
        ! for the scoring function
        !
        CALL FMC_UPDATE_STEP_SIZE(SELF,FIL)


        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FMC_EVAL_SLAVE): filament info after model was updated:'
            CALL FILAMENT_PRINT_INFO(FIL)
        ENDIF



        ! Will we return minus one?
        RETURN_MINUS_ONE = .FALSE.
        IF (RETURN_MINUS_ONE_IF_NOT_MONOTONIC) THEN
            IF (.NOT. FILAMENT_SPLINES_ARE_MONOTONIC(FIL)) THEN
                WRITE(*,'(A)') '**WARNING(FMC_EVAL_SLAVE): non-monotonic phi or subunit_num splines. Will return -1.0.'
                RETURN_MINUS_ONE = .TRUE.
            ENDIF
        ENDIF

        ! debug
        IF (DDETAILLED_SCORE_INFO) THEN
            WRITE(*,'(A)') '**info(FMC_EVAL_SLAVE): changes relative to original filament'
            CALL FILAMENT_PRINT_DIFF(SELF%FIL_ORIGINAL,FIL)
        ENDIF

        ! Will we use a single CTF for the whole filament, or one per segment?
        IF (SINGLE_CTF_PER_FILAMENT_IF_REASONABLE) THEN
            IF (FILAMENT_DEFOCUS_RANGE(FIL,IGNORE_FAL_WPS=.NOT.FIL%FAL_WPS_IN_SCORE) .LT. SELF%SEARCH_ACCURACY(6)) THEN
                !IF (DEBUG) THEN
                    WRITE(*,'(A)') '**debug(FMC_EVAL_SLAVE): defocus variations along filament less than desired Z accuracy. '//&
                                    'Will compute only one CTF for the whole filament.'
                !ENDIF
                OONE_CTF_FOR_WHOLE_FILAMENT = .TRUE.
            ELSE
                OONE_CTF_FOR_WHOLE_FILAMENT = .FALSE.
            ENDIF
        ELSE
            OONE_CTF_FOR_WHOLE_FILAMENT = .FALSE.
        ENDIF


        ! If only one or a few waypoints are to be refined, we should only get a path between neighbouring waypoints
        ! Note that we assume the waypoints which are to be refined are contiguous
        IF (OONLY_SCORE_REFINING_NEIGHBOURHOOD) THEN
            CALL FILAMENT_T_INTERVAL_FOR_SCORING_FUNCTION(FIL,START_T,FINISH_T,START_WP,FINISH_WP)
        ELSE
            IF (.NOT. FIL%FAL_WPS_IN_SCORE) THEN
                START_WP = 2
                FINISH_WP = FIL%NWP-1
            ELSE
                START_WP = 1
                FINISH_WP = FIL%NWP
            ENDIF
            ! Convert to T values
            START_T = FIL%T(START_WP)
            FINISH_T = FIL%T(FINISH_WP)
        ENDIF

		IF (DEBUG) THEN
			WRITE(*,'(A,2(F0.4,1X))') '**debug(FMC_EVAL_SLAVE): scoring only in refinement neighbourhood. Start, finish T = ', &
                                                                                                        START_T, FINISH_T
		ENDIF

        ! Sanity check
        IF (START_T .GT. FINISH_T) THEN
            WRITE(*,'(A,2(F0.6,1X))') '**ERROR(FMC_EVAL_SLAVE): sanity check failed: start_t is greater than finish_t: ', &
                                    START_T, FINISH_T
            CALL TERMINATE('Failed sanity check in FMC_EVAL_SLAVE')
        ENDIF

        ! Get path
        IF (SELF%WPS_AS_SPS) THEN
            FIL%WP_SCORE =  FIL%WP(START_WP:FINISH_WP)
        ELSE
            CALL FILAMENT_GET_PATH( FIL,FIL%WP_SCORE,SELF%STEP,SELF%STEP_IN_PIXELS,START_T=START_T,FINISH_T=FINISH_T)
        ENDIF
        NUM_WPS = SIZE(FIL%WP_SCORE)
        WRITE(NUM_WPS_C,'(I6)') NUM_WPS

        ! If any of the parameters are not full-filament, we need to ensure there is at least one scoring waypoint per refinement waypoint.
        IF (FIL%FRAGMENTED_PHI .OR. FIL%FRAGMENTED_SUBUNIT_NUM) then
            ALLOCATE(WP_HAS_SCORE_WP_NEARBY(FIL%NWP))
            WP_HAS_SCORE_WP_NEARBY = .FALSE.
            IF (.NOT. FIL%FAL_WPS_IN_SCORE) THEN
                ! We don't really need the first and last waypoint
                WP_HAS_SCORE_WP_NEARBY(1) = .TRUE.
                WP_HAS_SCORE_WP_NEARBY(FIL%NWP) = .TRUE.
            ENDIF
            DO I=1,SIZE(FIL%WP_SCORE)
                CALL NEAREST_WP_FROM_T(FIL,FIL%WP_SCORE(I)%PATH_LENGTH,J,DUMMY)
                WP_HAS_SCORE_WP_NEARBY(J) = .TRUE.
            ENDDO
            IF (.NOT. ALL(WP_HAS_SCORE_WP_NEARBY)) then
                IF ( ALL(WP_HAS_SCORE_WP_NEARBY(1:fil%nwp-1) .AND. FIL%FAL_WPS_IN_SCORE)) then
                    ! Only the last waypoint doesn't have a scoring waypoint near it
                    WRITE(*,'(A)') '**info/debug(FMC_EVAL_SLAVE): adding scoring waypoint at last refinement waypoint'
                    CALL ADD_WAYPOINT_TO_PATH(FIL%WP_SCORE,FIL,FIL%T(FIL%NWP))
                ELSE IF (ALL(WP_HAS_SCORE_WP_NEARBY(1:fil%nwp-2) .AND. .NOT. FIL%FAL_WPS_IN_SCORE)) THEN
                    CALL ADD_WAYPOINT_TO_PATH(FIL%WP_SCORE,FIL,FIL%T(FIL%NWP-1))
                ELSE
                    WRITE(*,'(A)') '**ERROR(FMC_EVAL_SLAVE): not all refinement waypoints have a scoring waypoint nearby'
                    WRITE(*,*) WP_HAS_SCORE_WP_NEARBY
                    CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
                ENDIF
            ENDIF
        ENDIF


        ! Get the FT of the model if we're working with slice extraction rather than projection
        IF (.NOT. SELF%REAL_SPACE_PROJ) THEN
            IF (SELF%MODEL3DVOL_PRIVATE%is_in_real_space) THEN
                WRITE(*,'(A)') '**ERROR(FMC_EVAL_SLAVE): model 3D must be in Fourier space'
                CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
            ENDIF
            IF (SELF%MODEL3DVOL_PRIVATE%ORIGIN_AT_CENTER) THEN
                WRITE(*,'(2A)') '**ERROR(FMC_EVAL_SLAVE): model 3D must have its real-space origin at the first pixel. ',    &
                                'Apply phase shift'
                CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
            ENDIF
        ENDIF

        ! Maximum radius of object, in pixels
        IF (SELF%WPS_AS_SPS) THEN
            OBJ_MAX_RAD = MAX(SELF%SP_SEGMENT_LENGTH/FIL%FILM%PIXEL_SIZE/2.0,FIL%MAX_RADIUS)
        ELSE
            OBJ_MAX_RAD = FIL%MAX_RADIUS
        ENDIF


        ! Allocate CTF_CURVE_SUM
        IF (CCALC_FRC)  THEN
            IF (ALLOCATED(CTF_CURVE_SUM)) DEALLOCATE(CTF_CURVE_SUM)
            ALLOCATE(CTF_CURVE_SUM(SELF%BOX_SIZE/2+1))
            CTF_CURVE_SUM = 0.0
        ENDIF


        !
        !   Move along the path to calculate the score
        !

        ! We will be keeping track of how many pixels contributed to the scoring function
        NUMBER_OF_PIXELS_IN_SCORING_FUNCTION = 0.0

        ! Dimensions of rectangular mask to be imposed on boxes & projections
        RECT_DIM = (/ MIN(REAL(SELF%STEP,KIND=4),2.*OBJ_MAX_RAD)+SOFT_EDGE,(OBJ_MAX_RAD+SOFT_EDGE)*2./)
        IF (SELF%WPS_AS_SPS) RECT_DIM(1) = SELF%SP_SEGMENT_LENGTH/SELF%FIL%FILM%PIXEL_SIZE+SOFT_EDGE
        PROJ_MAX_RAD = SQRT((RECT_DIM(1)*.5)**2 + (RECT_DIM(2)*.5)**2)


        !
        ! Decide whether we're going to use threading
        !
        TTHREADING = .FALSE.
        NUM_LOCAL_THREADS = 1
#ifdef FMC_EVAL_THREAD
        ! We won't do threading if we are already within a parallel region (for example if we're being
        ! called by FMC_BF_TILT)
        IF ( .NOT. OMP_IN_PARALLEL()) THEN
            IF (OMP_GET_MAX_THREADS() .GT. 1) THEN
                TTHREADING = .TRUE.
            ENDIF
        ENDIF
        ! If the user specified we should not do threading with the relevant argument,
        ! we definitely won't do threading
        IF (TTHREADING .AND. PRESENT(THREADING)) THEN
            IF (.NOT. THREADING) TTHREADING = .FALSE.
        ENDIF

        !
        ! OpenMP parallel region
        !
        !$OMP   PARALLEL    IF(TTHREADING)  &
        !$OMP               DEFAULT(NONE)   &
        !$OMP               PRIVATE(I,ICOO,RCOO,LOCAL_THREAD_NUM,SCORE_LOCAL,BOX_IS_BLANK,CC3_ND_RINGS_LOCAL)            &
        !$OMP               PRIVATE(MATRIX,CTF,WP_PHASED)                            &
        !$OMP               SHARED(SELF,FIL,LOC_COUNT)  &
        !$OMP               SHARED(SCORE,SCORE_ND_WHOLE,CC3_ND_RINGS_WHOLE)           &
        !$OMP               SHARED(RETURN_MINUS_ONE,WWHOLE_FILAMENT_NORM)   &
        !$OMP               SHARED(BOXES_FILENAME,DDETAILLED_SCORE_INFO,NUM_LOCAL_THREADS) &
        !$OMP               SHARED(WWRITE_STACKS,PROJS_FILENAME,BOXES_FILE,PROJS_FILE,AMPLS_FILE,AMPL_MASKS_FILE,OBJ_MAX_RAD) &
        !$OMP               SHARED(FRC,CCALC_FRC,IERR)  &
        !$OMP               PRIVATE(FRC_TMP) &
        !$OMP               SHARED(CTF_CURVE_SUM) &
        !$OMP               PRIVATE(CTF_CURVE,CTF_CURVE_SUM_PRIVATE) &
        !$OMP               SHARED(NUMBER_OF_PIXELS_IN_SCORING_FUNCTION,ESTIMATE_NOISE_VARIANCE,ESTIMATED_NOISE_VARIANCE,ESTIMATED_SIGNAL_VARIANCE)&
        !$OMP               SHARED(START_WP,FINISH_WP) &
        !$OMP               PRIVATE(MODEL_Z_SHIFT) &
        !$OMP               SHARED(DDONT_USE_CACHE,RECT_DIM) &
        !$OMP               SHARED(NUM_WPS,OONE_CTF_FOR_WHOLE_FILAMENT,PROJ_MAX_RAD,TTHREADING) &
        !$OMP               PRIVATE(SCORE_ND_LOCAL,CC3_ND_RINGS_WHOLE_PRIVATE,FIRST_TIME,SCORE_ND_WHOLE_PRIVATE)
#endif


        ! See how many threads have been started (note that single has an implicit barrier on exit)s
#ifdef FMC_EVAL_THREAD
        !$OMP   SINGLE
#endif
        IF (TTHREADING) NUM_LOCAL_THREADS = OMP_GET_NUM_THREADS()
        IF (NUM_LOCAL_THREADS .GT. 1 .AND. DDETAILLED_SCORE_INFO) THEN
            WRITE(*,'(A,I0,A)') '**info(FMC_EVAL_SLAVE): using ', NUM_LOCAL_THREADS, ' OpenMP threads.'
        ENDIF
        ! Initialise shared variables
        LOC_COUNT           = 1
        SCORE               = 0.0
        SCORE_ND_WHOLE      = [0.0,0.0,0.0]

        ! Ensure that the FMC object has the necessary workspace memory allocated
        CALL FMC_SET_NUM_WORK_AREAS(SELF,NUM_LOCAL_THREADS)
        IF (DEBUG) THEN
            WRITE(*,'(A,I0)') '**debug(FMC_EVAL_SLAVE): num local threads = ', NUM_LOCAL_THREADS
            WRITE(*,'(A,I0)') '**debug(FMC_EVAL_SLAVE): num work areas    = ', SELF%NUM_WORK_AREAS
        ENDIF

#ifdef FMC_EVAL_THREAD
        !$OMP   END SINGLE
#endif

        ! Initialise private variables
        SCORE_LOCAL         = 0.0
        SCORE_ND_LOCAL      = [0.0,0.0,0.0]
        MATRIX              = IDENTITY_ROTMAT()
        FIRST_TIME          = .TRUE. ! This is true for the first location encountered by a thread
        LOCAL_THREAD_NUM    = 1

        ! Make private copies of arrays that need to be "reduced"
        IF (CCALC_FRC) CTF_CURVE_SUM_PRIVATE = CTF_CURVE_SUM
        SCORE_ND_WHOLE_PRIVATE = SCORE_ND_WHOLE

        ! Loop over the scoring waypoints
#ifdef FMC_EVAL_THREAD
        !$OMP   DO  REDUCTION(+:NUMBER_OF_PIXELS_IN_SCORING_FUNCTION) &
        !$OMP       REDUCTION(+:SCORE)
#endif
        DO I=1,NUM_WPS

#ifdef FMC_EVAL_THREAD
            LOCAL_THREAD_NUM = OMP_GET_THREAD_NUM()+1

            IF (DEBUG) THEN
                PRINT '(4(A,I0))', '**debug(FMC_EVAL_SLAVE): thread ', LOCAL_THREAD_NUM,  &
                                                            ' of ', NUM_LOCAL_THREADS,    &
                                                            ' is on step ', I,      &
                                                            ' of ', NUM_WPS
            ENDIF
#endif

            ! Because the waypoint doesn't necessarily have an integer subunit_num value,
            ! we'll need to shift & rotate the reference to accomodate for the non-integer part of subunit_num
            ! Below, the phi angle of the waypoint is modified, and we also compute a Z shift which will need to be
            ! applied to the 3D model before projection
!#ifdef __INTEL_COMPILER
!            CALL POINT_ASSIGN(WP_PHASED,FIL%WP_SCORE(I))
!#else
            WP_PHASED = FIL%WP_SCORE(I)
!#endif
            IF (APPLY_Z_SHIFT_IN_3D) THEN
                CALL FILAMENT_MODEL_PHASING_FROM_WP(FIL,WP_PHASED,MODEL_Z_SHIFT=MODEL_Z_SHIFT,LOCAL_H_SYM=.TRUE.)
            ELSE
                ! Rather than apply the Z shift to the 3D model before projection, we can apply an (x,y) shift
                ! to the 2D projection or, equivalently, shift the center of the box over the micrograph.
                CALL FILAMENT_APPLY_PHASING_TO_WP_OLD(FIL,WP_PHASED,NEG=.FALSE.)
                MODEL_Z_SHIFT = 0.0
            ENDIF


            ! Get integer coordinates & remainders
            ICOO(1:2) = NINT(WP_PHASED%COO(1:2))
            ICOO(3) = 1
            RCOO = WP_PHASED%COO - ICOO


            ! Window image
            ! (maybe this should be windowed bigger initially to recover information spread by PSF/CTF?)
            ! (no, we're imposing a small-ish mask later on instead)
            SELF%BOX(LOCAL_THREAD_NUM)%is_in_real_space = .true.
            SELF%BOX(LOCAL_THREAD_NUM)%ORIGIN_AT_CENTER = .TRUE.
            CALL WINDOW_IMG(SELF%FILM_IMAGE,SELF%BOX(LOCAL_THREAD_NUM),ICOO)

            ! Check the image isn't blank
            BOX_IS_BLANK = IMAGE_IS_BLANK(SELF%BOX(LOCAL_THREAD_NUM))
            IF (BOX_IS_BLANK) THEN
                WRITE(*,'(A)') '**warning(FMC_EVAL_SLAVE): boxed image is blank!'
                WRITE(*,'(2A)') '**WARNING(FMC_EVAL_SLAVE): image boxed from micrograph was blank',    &
                                    '. Will return a score of -1.0'
                RETURN_MINUS_ONE = .TRUE.
            ENDIF

            ! If necessary, deal with CTF
            IF (OONE_CTF_FOR_WHOLE_FILAMENT) THEN
                CALL TERMINATE('Not implemented yet')
            ELSE
                ! Get the appropriate CTF parameters
                IF (GET_CTF_FROM_Z) THEN
                    CTF = FIL%FILM%CTFTILT%CTF
                    CALL CTF_ANG_TO_PIX(CTF,FIL%FILM%PIXEL_SIZE)
                    CALL CTF_AT_Z(CTF,FIL%WP_SCORE(I)%COO(3))
                ELSE
                    CTF = FILM_FIND_CTF_AT_COOS(FIL%FILM,FIL%WP_SCORE(I)%COO(1),FIL%WP_SCORE(I)%COO(2))
                ENDIF

                ! Compute the CTF image
                IF (DEBUG) WRITE(*,'(A,I0)') '**debug(FMC_EVAL_SLAVE): ctf_to_img on thread      ', LOCAL_THREAD_NUM
                IF (DEBUG) WRITE(*,'(A,L1)') '**debug(FMC_EVAL_SLAVE): box_ctf allocated?        ', &
                                                                                  IMAGE_ALLOCATED(SELF%BOX_CTF(LOCAL_THREAD_NUM))
                CALL CTF_TO_IMG(CTF,SELF%BOX_CTF(LOCAL_THREAD_NUM),FIL%FILM%PIXEL_SIZE)
                IF (DEBUG) WRITE(*,'(A,I0)') '**debug(FMC_EVAL_SLAVE): ctf_to_img on thread done ', LOCAL_THREAD_NUM
                IF (CCALC_FRC) THEN
                    CALL CTF_TO_CURVE(CTF,CTF_CURVE,SELF%BOX_SIZE/2+1,FIL%FILM%PIXEL_SIZE)
                    CTF_CURVE_SUM_PRIVATE = CTF_CURVE_SUM_PRIVATE + CTF_CURVE
                ENDIF
            ENDIF

            ! If we're caching projections, check whether we've already computed a projection similar enough to this one
            IF (SELF%CACHE_PROJECTIONS .AND. .NOT. DDONT_USE_CACHE) THEN
                CALL PROJECTION_PREPARE(SELF%MODEL3DVOL_PRIVATE,SELF%PROJ(LOCAL_THREAD_NUM),&
                                        SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions, &
                                        WP_PHASED%EULER,SELF%BOX_CTF(LOCAL_THREAD_NUM),   &
                                        PROJ_MAX_RAD=PROJ_MAX_RAD,SHIFT_X=RCOO(1),SHIFT_Y=RCOO(2),   &
                                        INTERPOL=SELF%INTERPOL,   &
                                        WIN=SELF%INTERP_WIN_3D(LOCAL_THREAD_NUM), &
                                        ENV_MASK=SELF%MASK_ENV(LOCAL_THREAD_NUM),REAL_SPACE_PROJ=SELF%REAL_SPACE_PROJ,   &
                                        VOL_SHIFT_Z=MODEL_Z_SHIFT,APPLY_DISC_MASK=.FALSE.,  &
                                        !PROJ_CACHE=SELF%PROJ_CACHE, &
                                        FREQ_MAX=SELF%FREQ_MAX, &
                                        RECT_MASK_WIDTH=FIL%MAX_RADIUS*2.)
            ELSE
                CALL PROJECTION_PREPARE(SELF%MODEL3DVOL_PRIVATE,SELF%PROJ(LOCAL_THREAD_NUM),&
                                        SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions, &
                                        WP_PHASED%EULER,SELF%BOX_CTF(LOCAL_THREAD_NUM),   &
                                        PROJ_MAX_RAD=PROJ_MAX_RAD,SHIFT_X=RCOO(1),SHIFT_Y=RCOO(2),   &
                                        INTERPOL=SELF%INTERPOL,   &
                                        WIN=SELF%INTERP_WIN_3D(LOCAL_THREAD_NUM), &
                                        ENV_MASK=SELF%MASK_ENV(LOCAL_THREAD_NUM),REAL_SPACE_PROJ=SELF%REAL_SPACE_PROJ,     &
                                        VOL_SHIFT_Z=MODEL_Z_SHIFT,APPLY_DISC_MASK=.FALSE.,  &
                                        FREQ_MAX=SELF%FREQ_MAX, &
                                        RECT_MASK_WIDTH=FIL%MAX_RADIUS*2.)
            ENDIF

            IF (IMAGE_IS_BLANK(SELF%PROJ(LOCAL_THREAD_NUM))) WRITE(*,'(A)') '**WARNING(FMC_EVAL_SLAVE): blank projection'

            ! Prepare the image
            IF (DEBUG) CALL WRITE_TO_DISK(SELF%BOX(LOCAL_THREAD_NUM),'dbg_box_before_prep.mrc')
            CALL IMAGE_PREPARE( SELF%BOX(LOCAL_THREAD_NUM),SELF%BOX_TMP(LOCAL_THREAD_NUM),SELF%BOX_CTF(LOCAL_THREAD_NUM), &
                                OBJ_MAX_RAD,ENV_MASK=SELF%MASK_ENV(LOCAL_THREAD_NUM), &
                                REAL_SPACE_SHIFT=SELF%REAL_SPACE_BOX_SHIFT, &
                                INTERPOL=SELF%INTERPOL,INTERPOL_WINDOW=SELF%INTERP_WIN_BOX_SHIFT(LOCAL_THREAD_NUM),   &
                                APPLY_DISC_MASK=.FALSE.)

            ! Apply rectangular masks to both
            IF (.not. SELF%PROJ(LOCAL_THREAD_NUM)%is_in_real_space) CALL BWD_FT(SELF%PROJ(LOCAL_THREAD_NUM))
            IF (DEBUG) CALL WRITE_TO_DISK(SELF%PROJ(LOCAL_THREAD_NUM),'dbg_proj_1.mrc')
            CALL COSINE_RECT(   SELF%PROJ(LOCAL_THREAD_NUM),RECT_DIM,REAL(WP_PHASED%EULER%PSI), &
                                FALLOFF=SOFT_EDGE,WIPE_INPUT=.FALSE.,MASK_WITH_AVERAGE_EDGE_VALUE=.TRUE.)
            IF (.not. SELF%BOX(LOCAL_THREAD_NUM)%is_in_real_space) CALL BWD_FT(SELF%BOX(LOCAL_THREAD_NUM))
            IF (DEBUG) CALL WRITE_TO_DISK(SELF%BOX(LOCAL_THREAD_NUM),'dbg_box_1.mrc')
            CALL COSINE_RECT(   SELF%BOX(LOCAL_THREAD_NUM) ,RECT_DIM,REAL(WP_PHASED%EULER%PSI), &
                                FALLOFF=SOFT_EDGE,WIPE_INPUT=.FALSE.,MASK_WITH_AVERAGE_EDGE_VALUE=.TRUE.)
            IF (DEBUG) CALL WRITE_TO_DISK(SELF%PROJ(LOCAL_THREAD_NUM),'dbg_proj_2.mrc')
            IF (DEBUG) CALL WRITE_TO_DISK(SELF%BOX(LOCAL_THREAD_NUM),'dbg_box_2.mrc')

            IF (ESTIMATE_NOISE_VARIANCE) THEN
                ! From the projection mask, we can work out how many pixels will contribute to the scoring function
                CALL COSINE_RECT(   SELF%MASK_ENV(LOCAL_THREAD_NUM) ,RECT_DIM,REAL(WP_PHASED%EULER%PSI), &
                                    FALLOFF=SOFT_EDGE,WIPE_INPUT=.FALSE.,MASK_WITH_AVERAGE=.FALSE.)
                NUMBER_OF_PIXELS_IN_SCORING_FUNCTION =  NUMBER_OF_PIXELS_IN_SCORING_FUNCTION &
                                                        + SUM_REAL(SELF%MASK_ENV(LOCAL_THREAD_NUM))
            ENDIF

            ! The projection and the image need to be in Fourier space before we measure correlation
            IF (SELF%PROJ(LOCAL_THREAD_NUM)%is_in_real_space) CALL FWD_FT(SELF%PROJ(LOCAL_THREAD_NUM))
            IF (SELF%BOX(LOCAL_THREAD_NUM)%is_in_real_space) CALL FWD_FT(SELF%BOX(LOCAL_THREAD_NUM))

            ! Prepare a Fourier mask for the amyloid case
            ! TODO: move this to FMC_INIT or somewhere similar
            IF (SELF%CROSS_BETA_FILTER) THEN
                CALL FOURIER_MASK_INIT_AMYLOID(         SELF%AMPL_MASK(LOCAL_THREAD_NUM),                   &
                                                        SELF%BOX(LOCAL_THREAD_NUM),FIL%FILM%PIXEL_SIZE,     &
                                                        FIL%MAX_RADIUS  *   FIL%FILM%PIXEL_SIZE,            &
                                                        FIL%FILM%PIXEL_SIZE /   SELF%FREQ_MAX,              &
                                                        WP_PHASED%EULER)
            ENDIF

            ! Measure correlation
            IF (DEBUG) WRITE(*,'(A,F0.3,A,L1)') '**debug(FMC_EVAL_SLAVE): about to calc correlation with FREQ_MAX = ', &
                                                SELF%FREQ_MAX, ' and whole-filament norm = ', WWHOLE_FILAMENT_NORM
            IF (.NOT. NORMALISED_SCORING_FUNCTION) THEN
                WRITE(*,'(A)') '**ERROR(FMC_EVAL_SLAVE): unnormalised scoring function not implemented yet'
                CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
            ENDIF
            IF (WWHOLE_FILAMENT_NORM) THEN
                IF (SELF%USE_CC3) THEN

                    IF (SELF%CROSS_BETA_FILTER) THEN
                        CC3_ND_RINGS_LOCAL = CCLIN_NUM_DENOMS_RINGS(    SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),  &
                                                                        FOURIER_MASK=SELF%AMPL_MASK(LOCAL_THREAD_NUM)     )
                    ELSE
                        CC3_ND_RINGS_LOCAL = CCLIN_NUM_DENOMS_RINGS(    SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),   &
                                                     RAD_MAX=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MAX,&
                                                     RAD_MIN=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MIN )
                    ENDIF
                    IF (FIRST_TIME) THEN
                        ! The first time around, CC3_ND_RINGS_WHOLE hasn't yet been allocated yet, so we can't have it on the rhs of the assignment
                        CC3_ND_RINGS_WHOLE_PRIVATE = CC3_ND_RINGS_LOCAL
                    ELSE
                        CC3_ND_RINGS_WHOLE_PRIVATE = CC3_ND_RINGS_WHOLE_PRIVATE + CC3_ND_RINGS_LOCAL
                    ENDIF
                    FIRST_TIME = .FALSE.
                    SCORE_LOCAL = CC3_FINALISE(     CC3_ND_RINGS_LOCAL,  &
                                                    SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),   &
                                                    RAD_MAX=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MAX,   &
                                                    RAD_MIN=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MIN)
                    IF (ESTIMATE_NOISE_VARIANCE) THEN
                        SCORE_ND_LOCAL = CCLIN_NUM_DENOMS_FROM_RINGS(CC3_ND_RINGS_LOCAL)
                        SCORE_ND_WHOLE_PRIVATE = SCORE_ND_WHOLE_PRIVATE + SCORE_ND_LOCAL
                    ENDIF
                ELSE ! We're not using CC3
                    IF (SELF%CROSS_BETA_FILTER) THEN
                        CALL CCLIN_NUM_DENOMS(  SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),   &
                                                            SCORE_ND_LOCAL, &
															FOURIER_MASK=SELF%AMPL_MASK(LOCAL_THREAD_NUM)   &
                                                            )
                    ELSE
                        CALL CCLIN_NUM_DENOMS(  SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),   &
                                                         SCORE_ND_LOCAL, &
										            RAD_MAX=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MAX,   &
                                                    RAD_MIN=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MIN)
                    ENDIF
                    SCORE_ND_WHOLE_PRIVATE = SCORE_ND_WHOLE_PRIVATE + SCORE_ND_LOCAL
                    IF (SCORE_ND_LOCAL(2) .LE. 0.0D0 .OR. SCORE_ND_LOCAL(3) .LE. 0.0D0) THEN
                        ! I believe this only happens when there is a blank box
                        SCORE_LOCAL = -999.0D0
                        IF (.NOT. BOX_IS_BLANK) THEN
                            ! Something must have gone wrong, let's dump some diagnostics and crash out
                            IF (.not. SELF%BOX(LOCAL_THREAD_NUM)%is_in_real_space) CALL BWD_FT(SELF%BOX(LOCAL_THREAD_NUM))
                            IF (.not. SELF%PROJ(LOCAL_THREAD_NUM)%is_in_real_space) CALL BWD_FT(SELF%PROJ(LOCAL_THREAD_NUM))
                            CALL WRITE_TO_DISK(SELF%BOX(LOCAL_THREAD_NUM),'dbg_box.mrc')
                            CALL WRITE_TO_DISK(SELF%PROJ(LOCAL_THREAD_NUM),'dbg_proj.mrc')
                            CALL WRITE_TO_DISK(SELF%MASK_ENV(LOCAL_THREAD_NUM),'dbg_mask_env.mrc')
                            CALL WP_PRINT_INFO(WP_PHASED)
                            !! Let's trigger a FPE, so that we get a core dump
                            !RCOO(1) = SQRT(-1.0/0.0)
                            WRITE(*,'(A)') '**ERROR(FMC_EVAL_SLAVE): this is strange - '
                            WRITE(*,'(A)') '**ERROR(FMC_EVAL_SLAVE): I was getting CC denominator terms <= 0.0 '//&
                                                                    'yet the box isn''t blank. '//&
                                                                    'See dbg_*mrc files dumped to current directory.'
                            CALL TERMINATE('Thought box was blank')
                        ELSE
                            WRITE(*,'(A)') '**warning(FMC_EVAL_SLAVE): strange values because of blank box'
                        ENDIF
                    ELSE
                        SCORE_LOCAL = REAL(SCORE_ND_LOCAL(1)/SQRT(SCORE_ND_LOCAL(2)*SCORE_ND_LOCAL(3)))
                    ENDIF
                ENDIF
                !PRINT *, '**debug(FMC_EVAL_SLAVE): SCORE_LOCAL = ', SCORE_LOCAL
                !IF (SCORE_LOCAL .NE. CCLIN(BOX,PROJ,RAD_MAX=NINT(REAL(BOX%logical_dimensions(1))/2.0*RAD_MAX))) THEN
                !   WRITE(*,'(A,2(F10.4,X))') '**ERROR(FMC_EVAL_SLAVE): SCORE_LOCAL, CCLIN = ', SCORE_LOCAL, CCLIN(BOX,PROJ,RAD_MAX=NINT(REAL(BOX%logical_dimensions(1))/2.0*RAD_MAX))
                !   CALL TERMINATE('Failed sanity check')
                !ENDIF
            ELSE ! We're not doing whole-filament normalisation
                IF (SELF%USE_CC3) THEN
                    IF (SELF%CROSS_BETA_FILTER) THEN
                        SCORE_LOCAL = CC3(  SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),   &
                                            FOURIER_MASK=SELF%AMPL_MASK(LOCAL_THREAD_NUM)   &
                                            )
                    ELSE
                        SCORE_LOCAL = CC3(    SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),   &
                                            RAD_MAX=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MAX,   &
                                            RAD_MIN=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MIN)
                    ENDIF
                ELSE
                    IF (SELF%CROSS_BETA_FILTER) THEN
                        SCORE_LOCAL = CCLIN(    SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),   &
                                                FOURIER_MASK=SELF%AMPL_MASK(LOCAL_THREAD_NUM)   &
                                            )
                    ELSE
                        SCORE_LOCAL = CCLIN(    SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),   &
                                                RAD_MAX=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MAX,   &
                                                RAD_MIN=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MIN)
                    ENDIF
                ENDIF
                IF (SELF%CUBED_AVERAGE) THEN
                    SCORE = SCORE + SCORE_LOCAL**3
                ELSE
                    SCORE = SCORE + SCORE_LOCAL
                ENDIF
            ENDIF

            ! Remember the local correlation score
            FIL%WP_SCORE(I)%SCORE = SCORE_LOCAL
            IF (SELF%WPS_AS_SPS) FIL%WP(I)%SCORE = SCORE_LOCAL

            !
            ! Check for NaN
            !
            IF (IEEE_IS_NAN(SCORE_LOCAL)) THEN
                ! Check whether box is blank at this point
                CALL BWD_FT(SELF%BOX(LOCAL_THREAD_NUM))
                BOX_IS_BLANK = IMAGE_IS_BLANK(SELF%BOX(LOCAL_THREAD_NUM))
                IF (BOX_IS_BLANK) THEN
                    WRITE(*,'(2A)') '**WARNING(FMC_EVAL_SLAVE): image boxed from micrograph was blank, ',    &
                                    'leading to a non-sensical score. Will return a score of -1.0'
                    RETURN_MINUS_ONE = .TRUE.
                ELSE
                    WRITE(*,'(2A,F8.4)') '**ERROR(FMC_EVAL_SLAVE): intermediate score is not a number ,',    &
                                        '(but boxed image was not blank): ', SCORE_LOCAL
                    IF (WWHOLE_FILAMENT_NORM) THEN
                        WRITE(*,'(A,3(F8.4,1X))') '**debug(FMC_EVAL_SLAVE): numerator, denominator1, denominator2: ', &
                                                SCORE_ND_LOCAL(1:3)
                    ENDIF
                    WRITE(*,*) '**debug(FMC_EVAL_SLAVE): CTF = ', CTF
                    WRITE(*,'(A,3(I6,1X))') '**debug(FMC_EVAL_SLAVE): ICOO = ', ICOO(1:3)
                    WRITE(*,'(A,3(F7.4,1X))') '**debug(FMC_EVAL_SLAVE): RCOO = ', RCOO(1:3)
                    WRITE(*,'(A)') '**debug(FMC_EVAL_SLAVE): dumping box and projection to disk as dbg_box.mrc and dbg_proj.mrc'
                    CALL BWD_FT(SELF%BOX(LOCAL_THREAD_NUM))
                    CALL BWD_FT(SELF%PROJ(LOCAL_THREAD_NUM))
                    CALL WRITE_TO_DISK(SELF%BOX(LOCAL_THREAD_NUM),'dbg_box.mrc')
                    CALL WRITE_TO_DISK(SELF%PROJ(LOCAL_THREAD_NUM),'dbg_proj.mrc')
                    CALL WRITE_TO_DISK(SELF%BOX_CTF(LOCAL_THREAD_NUM),'dbg_box_ctf.mrc')
                    CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
                ENDIF
            ELSE
                IF (BOX_IS_BLANK .AND. SCORE_LOCAL .NE. -999.) THEN
                    WRITE(*,'(A,ES12.3)') '**ERROR(FMC_EVAL_SLAVE): the image boxed from micrograph was blank, but the score was', &
                                        ' still a number: ', SCORE_LOCAL
                    CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
                ENDIF
            ENDIF

            ! If requested, write to stack
            IF (WWRITE_STACKS) THEN
                IF (.not. SELF%BOX(LOCAL_THREAD_NUM)%is_in_real_space) CALL BWD_FT(SELF%BOX(LOCAL_THREAD_NUM))
                IF (.not. SELF%PROJ(LOCAL_THREAD_NUM)%is_in_real_space) CALL BWD_FT(SELF%PROJ(LOCAL_THREAD_NUM))
#ifdef FMC_EVAL_THREAD
                !$OMP   CRITICAL        (WRITE_STACKS_TO_DISK_OMP_CRIT)
#endif
                !PRINT '(A,I3)', '**debug(FMC_EVAL_SLAVE): dumping box/proj number ', LOC_COUNT
                call self%box(LOCAL_THREAD_NUM)%WriteToDisk(BOXES_FILE,wanted_location=LOC_COUNT)
                call self%proj(LOCAL_THREAD_NUM)%WriteToDisk(PROJS_FILE,wanted_location=LOC_COUNT)
				IF (WWRITE_AMPL_STACK) THEN
					CALL AMPL(SELF%BOX(LOCAL_THREAD_NUM),SELF%BOX_AMPL(LOCAL_THREAD_NUM))
					call self%box_ampl(local_thread_num)%WriteToDisk(ampls_file,wanted_location=loc_count)
					SELF%BOX_AMPL = SELF%AMPL_MASK(LOCAL_THREAD_NUM)
					CALL AMPL_FROM_HALF_TO_FULL(SELF%BOX_AMPL(LOCAL_THREAD_NUM))
					call self%box_ampl(local_thread_num)%WriteToDisk(ampl_masks_file,wanted_location=loc_count)
				ENDIF
				LOC_COUNT = LOC_COUNT + 1
#ifdef FMC_EVAL_THREAD
                !$OMP   END CRITICAL    (WRITE_STACKS_TO_DISK_OMP_CRIT)
#endif
            ENDIF

            ! If request, compute FRC between box and proj
            IF (CCALC_FRC) THEN
                CALL COMPUTE_FSC(SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),FRC_TMP)
                IF (.NOT. ALLOCATED(FRC)) THEN
#ifdef FMC_EVAL_THREAD
                    !$OMP   CRITICAL    (FMC_FRC_ACCUM_OMP_CRIT)
#endif
                    IF (.NOT. ALLOCATED(FRC)) THEN
                        ALLOCATE(FRC(SIZE(FRC_TMP),NUM_WPS+2),STAT=IERR)
                        IF (IERR .NE. 0) THEN
                            WRITE(*,'(A,I0)') '**ERROR(FMC_EVAL_SLAVE): memory allocation for FRC array failed with error ', IERR
                            CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
                        ENDIF
                        FRC = 0.0
                    ENDIF
#ifdef FMC_EVAL_THREAD
                    !$OMP   END CRITICAL    (FMC_FRC_ACCUM_OMP_CRIT)
#endif
                ENDIF

                FRC(:,I) = FRC_TMP(:)
            ENDIF
        ENDDO   ! End loop over path waypoints
#ifdef FMC_EVAL_THREAD
        !$OMP   ENDDO
#endif

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(FMC_EVAL_SLAVE): filament info after loop over waypoints:'
            CALL FILAMENT_PRINT_INFO(FIL)
        ENDIF

        !
        ! Let's do our manual array reductions
        !
#ifdef FMC_EVAL_THREAD
        !$OMP   CRITICAL        (FMC_EVAL_MANUAL_REDUCTIONS)
#endif
        IF (CCALC_FRC) THEN
            CTF_CURVE_SUM = CTF_CURVE_SUM + CTF_CURVE_SUM_PRIVATE
        ENDIF
        IF (SELF%USE_CC3) THEN
            CC3_ND_RINGS_WHOLE = CC3_ND_RINGS_WHOLE + CC3_ND_RINGS_WHOLE_PRIVATE
        ENDIF
        SCORE_ND_WHOLE = SCORE_ND_WHOLE + SCORE_ND_WHOLE_PRIVATE
#ifdef FMC_EVAL_THREAD
        !$OMP   END CRITICAL    (FMC_EVAL_MANUAL_REDUCTIONS)
        ! Make sure we've done all the array reductions before moving on
        !$OMP BARRIER
#endif


        !
        ! Let's finalise the score
        !
#ifdef FMC_EVAL_THREAD
        !$OMP   SINGLE
#endif
        IF (ESTIMATE_NOISE_VARIANCE) THEN
            IF (RETURN_MINUS_ONE) THEN
                ! This only happens when boxes are blank (negative coordinates etc),
                ! i.e. there is no signal
                ESTIMATED_NOISE_VARIANCE  = 10000000.0E0
                ESTIMATED_SIGNAL_VARIANCE = 0.0E0
            ELSE
                IF (DEBUG) THEN
                    WRITE(*,'(A,F0.2)') '**debug(FMC_EVAL_SLAVE): number of pixels in scoring function = ', &
                                        NUMBER_OF_PIXELS_IN_SCORING_FUNCTION
                    WRITE(*,'(A,I0)') '**debug(FMC_EVAL_SLAVE): total number of pixels = ', &
											NUM_WPS*SELF%BOX(1)%logical_dimensions(1)*SELF%BOX(1)%logical_dimensions(2)
                ENDIF
                ESTIMATED_NOISE_VARIANCE = REAL(SCORE_ND_WHOLE(2) - (SCORE_ND_WHOLE(1)**2/SCORE_ND_WHOLE(3)))  &
                                                        / NUMBER_OF_PIXELS_IN_SCORING_FUNCTION
                ESTIMATED_SIGNAL_VARIANCE = REAL(SCORE_ND_WHOLE(1)/SCORE_ND_WHOLE(3)**2)**2 ! alpha, the scaling factor for the reference projection, squared
                ESTIMATED_SIGNAL_VARIANCE = ESTIMATED_SIGNAL_VARIANCE * SCORE_ND_WHOLE(2)**2 / NUMBER_OF_PIXELS_IN_SCORING_FUNCTION
                IF (ESTIMATED_NOISE_VARIANCE .LT. 0.0) THEN
                    WRITE(*,'(A,F0.4)') '**ERROR(FMC_EVAL_SLAVE): negative variance per pixel! ', ESTIMATED_NOISE_VARIANCE
                    CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
                ENDIF
                IF (NORMALISED_SCORING_FUNCTION) THEN
                    ! Since we are actually using a normalised cross-correlation for the scoring function,
                    ! and not just the cross-product of the noisy image with the reference as in Sigworth 1998 Eq 21,
                    ! we modify the estimate of the variance to include the normalisation.
                    ! See notes of 8-July-2011 for derivation.
                    !ESTIMATED_NOISE_VARIANCE_PER_PIXEL =   ESTIMATED_NOISE_VARIANCE_PER_PIXEL / &
                    !                                       (SCORE_ND_WHOLE(1) * SQRT(SCORE_ND_WHOLE(2)/SCORE_ND_WHOLE(3)))
                    ! New derivation, 9-July-2011:
                    ESTIMATED_NOISE_VARIANCE  =   ESTIMATED_NOISE_VARIANCE / &
                                                 (SQRT(SCORE_ND_WHOLE(2))*SQRT(SCORE_ND_WHOLE(3)))
                    ESTIMATED_SIGNAL_VARIANCE =   ESTIMATED_SIGNAL_VARIANCE / &
                                                 (SQRT(SCORE_ND_WHOLE(2))*SQRT(SCORE_ND_WHOLE(3)))
                ENDIF
                !IF (ESTIMATED_NOISE_VARIANCE_PER_PIXEL .LT. 0.0) THEN
                !   WRITE(*,'(A,F0.4)') '**ERROR(FMC_EVAL_SLAVE): negative variance per pixel after normalisation! ', ESTIMATED_NOISE_VARIANCE_PER_PIXEL
                !   CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
                !ENDIF
            ENDIF
        ENDIF
        IF (RETURN_MINUS_ONE) THEN
            SCORE = -1.0
        ELSE
            IF (WWHOLE_FILAMENT_NORM) THEN
                IF (SELF%USE_CC3) THEN
                    SCORE = CC3_FINALISE(   CC3_ND_RINGS_WHOLE,SELF%BOX(LOCAL_THREAD_NUM),SELF%PROJ(LOCAL_THREAD_NUM),    &
                                            RAD_MAX=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MAX,   &
                                            RAD_MIN=REAL(SELF%BOX(LOCAL_THREAD_NUM)%logical_dimensions(1))*SELF%FREQ_MIN)
                ELSE
                    SCORE = REAL(SCORE_ND_WHOLE(1)/SQRT(SCORE_ND_WHOLE(2)*SCORE_ND_WHOLE(3)))
                ENDIF
            ELSE
                IF (SELF%CUBED_AVERAGE) THEN
                    ! Version 1 (average at the end)
                    !SCORE = (  &
                    !       EXP(LOG(ABS(SCORE))/3.0)    &   ! cubic root
                    !       * SCORE / ABS(SCORE)    &       ! recover sign
                    !       )   &
                    !       / SIZE(PATH)                    ! average
                    ! Version 2
                    SCORE = SCORE / NUM_WPS              ! average
                    SCORE = (   &
                            EXP(LOG(ABS(SCORE))/3.0)    &   ! cubic root
                            * SCORE / ABS(SCORE)    &       ! recover sign
                            )
                ELSE
                    SCORE = SCORE/NUM_WPS
                ENDIF
            ENDIF
        ENDIF
#ifdef FMC_EVAL_THREAD
        !$OMP   END SINGLE
        ! Deallocate objects which were private and allocated dynamically within the parallel region
        IF (ALLOCATED(CC3_ND_RINGS_LOCAL)) DEALLOCATE(CC3_ND_RINGS_LOCAL)
        IF (ALLOCATED(FRC_TMP)) DEALLOCATE(FRC_TMP)
        IF (ALLOCATED(CTF_CURVE)) DEALLOCATE(CTF_CURVE)
        IF (ALLOCATED(CTF_CURVE_SUM_PRIVATE)) DEALLOCATE(CTF_CURVE_SUM_PRIVATE)
        IF (ALLOCATED(CC3_ND_RINGS_WHOLE_PRIVATE)) DEALLOCATE(CC3_ND_RINGS_WHOLE_PRIVATE)
        !$OMP SINGLE
#endif
        ! This is actually a shared object, so we only deallocate it one of the threads
        IF (ALLOCATED(CC3_ND_RINGS_WHOLE)) DEALLOCATE(CC3_ND_RINGS_WHOLE)
#ifdef FMC_EVAL_THREAD
        !$OMP END SINGLE

        ! The end of the parallel region is here, because CC3_ND_RINGS_WHOLE is only allocated within it,
        ! and is needed for the score calculate, above
        !$OMP   END PARALLEL
#endif

        ! The detailled scoring info can be printed out now
        IF (DDETAILLED_SCORE_INFO) THEN
            PRINT *
            PRINT '(A,I0,A)', '**info(FMC_EVAL_SLAVE): detailled score information along path of filament ', FIL%ID, ':'
            IF (OONLY_SCORE_REFINING_NEIGHBOURHOOD) THEN
                WRITE(*,'(3(A,F0.4),A)')    '(region from T = ', START_T, ' to T = ', FINISH_T, &
                                            ' used to compute score, in steps of ', SELF%STEP ,')'
            ENDIF
            CALL WAYPOINTS_PRINT_INFO(FIL%WP_SCORE)
            WRITE(*,'(A,F0.5)') '**info(FMC_EVAL_SLAVE): score = ', SCORE
        ENDIF


        !
        ! Check for NaN
        !
        IF (IEEE_IS_NAN(SCORE)) THEN
            WRITE(*,'(A,F0.4)') '**ERROR(FMC_EVAL_SLAVE): final score is not a number: ', SCORE
            CALL TERMINATE('Fatal error in FMC_EVAL_SLAVE')
        ENDIF

        IF (DEBUG) PRINT '(A,F0.5)', '**debug(FMC_EVAL_SLAVE): score = ', SCORE


        !
        ! Apply curvature, overtwist and/or stretching restraints
        !
        IF (SELF%CURVATURE_RESTRAINT .OR. SELF%TORSION_RESTRAINT .OR. SELF%STRETCH_RESTRAINT) THEN
            ! Diagnostic output
            IF (DDETAILLED_SCORE_INFO .OR. DEBUG) THEN
                WRITE(*,'(A,F0.4)') '**info(FMC_EVAL_SLAVE): score before penalties = ', SCORE
                WRITE(*,'(A,F0.4)') '**info(FMC_EVAL_SLAVE): estimated SNR (Bershad & Rockmore estimator) = ', &
                                      SCORE / (1.0-SCORE) ! Bershad & Rockmore 1974
                WRITE(*,'(2(A,G15.4),A)') '**info(FMC_EVAL_SLAVE): With an estimated noise variance per pixel of ', &
                            ESTIMATED_NOISE_VARIANCE, ' , (estimated SNR ', &
                            ESTIMATED_SIGNAL_VARIANCE / ESTIMATED_NOISE_VARIANCE, &
                            !SCORE_ND_WHOLE(1)**2 / ESTIMATED_NOISE_VARIANCE_PER_PIXEL, &
                            '), the following restraints lead to score penalties: '
                WRITE(*,'(4(A20,1X))') 'Property', 'Total along filament', 'Persistence length', 'Penalty'
            ENDIF
            ! For testing purposes, we can either do this properly or use a fudge factor
            IF (ESTIMATE_NOISE_VARIANCE) THEN
                PENALTY_FACTOR = ESTIMATED_NOISE_VARIANCE
            ELSE
                PENALTY_FACTOR = 0.1
            ENDIF
            ! Compute the squared curvature, overtwist and stretch integrated over the filament's length
            CALL FILAMENT_EVAL_TOTAL_CURVATURE_OVERTWIST_STRETCH(   FIL,  &
                                                                    CURVATURE,TORSION,STRETCH,   &
                                                                    IGNORE_FAL_WPS=.NOT.FIL%FAL_WPS_IN_SCORE)
            ! Compute the penalties
            CURVATURE_PENALTY = CURVATURE * 0.5 * PENALTY_FACTOR * FIL%PERSISTENCE_LENGTH
            TORSION_PENALTY   = TORSION   * 0.5 * PENALTY_FACTOR * FIL%TORSION_PERSISTENCE_LENGTH
            STRETCH_PENALTY   = STRETCH   * 0.5 * PENALTY_FACTOR * FIL%STRETCH_PERSISTENCE_LENGTH
            ! Debug
            IF (CURVATURE_PENALTY .GT. 100.0) THEN
                WRITE(*,'(A,G17.10)') '**debug/warning(FMC_EVAL_SLAVE): very large curvature penalty = ', CURVATURE_PENALTY
                !CALL TERMINATE('Strange curvature penalty value')
            ENDIF
            ! Apply the penalties
            SCORE = SCORE - CURVATURE_PENALTY - TORSION_PENALTY - STRETCH_PENALTY
            ! Diagnostic output
            IF (DDETAILLED_SCORE_INFO .OR. DEBUG) THEN
                WRITE(*,'(A20,1X,3(G20.4,1X))') 'Curvature', CURVATURE, FIL%PERSISTENCE_LENGTH,         CURVATURE_PENALTY
                WRITE(*,'(A20,1X,3(G20.4,1X))') 'Overtwist', TORSION,   FIL%TORSION_PERSISTENCE_LENGTH, TORSION_PENALTY
                WRITE(*,'(A20,1X,3(G20.4,1X))') 'Stetch',    STRETCH,   FIL%STRETCH_PERSISTENCE_LENGTH, STRETCH_PENALTY
                WRITE(*,'(A)') ' '
            ENDIF
        ENDIF

        !
        ! Apply NM amplitude penalty
        !
        IF (NM_AMP_PENALTY) THEN
            NM_AMP_WEIGHTED_SUM = FILAMENT_NM_AMPLITUDES_WEIGHTED_SUM(FIL)
            WRITE(*,'(A,F0.4)') '**debug(FMC_EVAL_SLAVE): score before penalty = ', SCORE
            WRITE(*,'(A,F0.4)') '**debug(FMC_EVAL_SLAVE): penalty = ', NM_AMP_PENALTY_FACTOR * NM_AMP_WEIGHTED_SUM
            SCORE = SCORE - NM_AMP_PENALTY_FACTOR * NM_AMP_WEIGHTED_SUM
            WRITE(*,'(A,F0.4)') '**debug(FMC_EVAL_SLAVE): score after penalty = ', SCORE
        ENDIF

        ! The filament keeps track of its own score
        FIL%SCORE = SCORE
        IF (.NOT. SELF%WPS_AS_SPS) FIL%WP%SCORE = SCORE

        ! Close files
        IF (WWRITE_STACKS) THEN
            IF (DEBUG) THEN
                PRINT '(A)', '**debug(FMC_EVAL_SLAVE): closing these files'
                CALL BOXES_FILE%PrintInfo()
                CALL PROJS_FILE%PrintInfo()
            ENDIF
            CALL BOXES_FILE%CLOSE()
            CALL PROJS_FILE%CLOSE()
			IF (WWRITE_AMPL_STACK) CALL AMPLS_FILE%CLOSE()
			IF (WWRITE_AMPL_STACK) CALL AMPL_MASKS_FILE%CLOSE()
        ENDIF

        ! Calculate the average and stdev FRCs and print them out
        IF (CCALC_FRC) THEN
            CTF_CURVE_SUM = ABS(CTF_CURVE_SUM/NUM_WPS)
            WRITE(*,'(/A)') 'FRC between images and matching projections used in the scoring function'
            WRITE(*,'(A5,1X,'//NUM_WPS_C//'(A3,1X,I3,X),3(1X,A6,1X))') 'Shell', ('WP#', I,I=1,NUM_WPS), 'Ave.', 'StdDev','AveCTF'
            DO I=1,SIZE(FRC,1)
                FRC(I,NUM_WPS+1) = SUM( FRC(I,1:NUM_WPS)   ) / NUM_WPS ! average
                FRC(I,NUM_WPS+2) = SUM( FRC(I,1:NUM_WPS)**2) / NUM_WPS ! average of squares
                FRC(I,NUM_WPS+2) = FRC(I,NUM_WPS+2) - FRC(I,NUM_WPS+1)**2  ! population variance
                CALL SAFE_FOR_SQRT(FRC(I,NUM_WPS+2),NEGATIVE_TO_ZERO=.TRUE.)
                FRC(I,NUM_WPS+2) = SQRT(FRC(I,NUM_WPS+2)) ! standard deviation
                WRITE(*,'(I5,X,'//NUM_WPS_C//'(F7.3,1X),2(F7.3,1X),F7.3)') I, (FRC(I,J),J=1,NUM_WPS+2), CTF_CURVE_SUM(I)
            ENDDO
            IF (WWRITE_FRC_PLOTS) THEN
                ! If the PLplot library is available, we'll also plot the FRCs and output a PNG image of the plot
#if defined(HAVE_PLPLOT)
                WRITE(FRCS_FILENAME,'(A,I0,A)') 'frcs_',FIL%ID,'.png'
                IF (PRESENT(STACKS_ROOTNAME)) THEN
                    IF (.NOT. IS_BLANK(STACKS_ROOTNAME)) THEN
                        !PRINT '(2A)', 'STACKS_ROOTNAME = ', STACKS_ROOTNAME
                        TMP_STRING = TRIM(ADJUSTL(STACKS_ROOTNAME))
                        IF (STRING_EQUAL(EXTENSION_FROM_FILENAME(TMP_STRING),'   ')) THEN
                            FRCS_FILENAME = TMP_STRING//'_frcs.png'
                        ELSE
                            FRCS_FILENAME = TMP_STRING(1:LEN(TMP_STRING)-4)//'_frcs.png'
                        ENDIF
                    ENDIF
                ENDIF
                WRITE(TMP_STRING,'(A,I0,A)') 'FRCs between segments of filament ', FIL%ID, ' and matching projections'
                WRITE(*,'(3A)') '**debug(FMC_EVAL_SLAVE): writing FRCs plot to image ', TRIM(ADJUSTL(FRCS_FILENAME)), '.'
                CALL SERIES_TO_PNG( FRCS_FILENAME,FRC(:,1:NUM_WPS),  &
                                    PLOT_TITLE=TMP_STRING,  &
                                    Y_MIN=0.,Y_MAX=1.,  &
                                    X_TITLE='Shell',Y_TITLE='FRC')

                ! Now write out average / std for the FRCs
                WRITE(FRCS_FILENAME,'(A,I0,A)') 'frcs_avestd_',FIL%ID,'.png'
                IF (PRESENT(STACKS_ROOTNAME)) THEN
                    IF (.NOT. IS_BLANK(STACKS_ROOTNAME)) THEN
                        TMP_STRING = TRIM(ADJUSTL(STACKS_ROOTNAME))
                        IF (STRING_EQUAL(EXTENSION_FROM_FILENAME(TMP_STRING),'   ')) THEN
                            FRCS_FILENAME = TMP_STRING//'_frcs_avestd.png'
                        ELSE
                            FRCS_FILENAME = TMP_STRING(1:LEN(TMP_STRING)-4)//'_frcs_avestd.png'
                        ENDIF
                    ENDIF
                ENDIF
                WRITE(TMP_STRING,'(A,I0,A)') 'FRCs between segments of filament ', FIL%ID, ' and matching projections'
                WRITE(*,'(3A)') '**debug(FMC_EVAL_SLAVE): writing FRCs ave/std plot to image ', TRIM(ADJUSTL(FRCS_FILENAME)), '.'
                FRC_TMP_TWOD = FRC(:,NUM_WPS+1:NUM_WPS+2)
                CALL SERIES_TO_PNG( FRCS_FILENAME,FRC_TMP_TWOD, &
                                    PLOT_TITLE=TMP_STRING,  &
                                    Y_MIN=0.,Y_MAX=1.,  &
                                    X_TITLE='Shell',Y_TITLE='FRC',CURVE_TITLES=['Average','StdDev '],   &
                                    OVERLAY_YS=CTF_CURVE_SUM)
#else
                ! PLplot is not available
                WRITE(*,'(A)') '**debug(FMC_EVAL_SLAVE): plotting library not available, will not write FRCs to image'
#endif
            ENDIF   ! End of test for FRC plotting
            WRITE(*,'(A)') ''
        ENDIF


        ! Explicit deallocations
        IF (ALLOCATED(CC3_ND_RINGS_WHOLE)) DEALLOCATE(CC3_ND_RINGS_WHOLE)
        IF (ALLOCATED(CC3_ND_RINGS_LOCAL)) DEALLOCATE(CC3_ND_RINGS_LOCAL)
        IF (ALLOCATED(CTF_CURVE)) DEALLOCATE(CTF_CURVE)
        IF (ALLOCATED(CTF_CURVE_SUM)) DEALLOCATE(CTF_CURVE_SUM)
        IF (ALLOCATED(FRC_TMP)) DEALLOCATE(FRC_TMP)
        IF (ALLOCATED(FRC)) DEALLOCATE(FRC)
        IF (ALLOCATED(FRC_TMP_TWOD)) DEALLOCATE(FRC_TMP_TWOD)

        IF (DDETAILLED_SCORE_INFO) THEN
            WRITE(*,'(A)') '**info(FMC_EVAL_SLAVE): all done'
        ENDIF


    END FUNCTION FMC_EVAL_SLAVE




END MODULE FMCS_EVALUATION
