!>  \brief  Module to interact with frealign parameter files
!!
MODULE FREALIGN_PARAMETERS

    CONTAINS

    !>  \brief  Write a frealign parameter file from an array of single particles
    SUBROUTINE SP_TO_FREALIGN_PAR(SP,FILENAME,MAG,PIXEL_SIZE)
        USE SINGLE_PARTICLES
        USE RUNTIME_PARAMETERS, ONLY : AVAILABLE_LUN, LUN_IS_NOW_AVAILABLE, TERMINATE
        USE ROTATION_MATRICES, ONLY : DEGREES
        USE PARSERS, ONLY : FILE_EXISTS
        IMPLICIT NONE
        ! Arguments
        TYPE(SINGLE_PARTICLE),  ALLOCATABLE,    INTENT(IN)  ::  SP(:)       !<  Array of single particles
        CHARACTER(LEN=*),                       INTENT(IN)  ::  FILENAME    !<  Filename of frealign parameter file to written out
        REAL,                                   INTENT(IN)  ::  MAG         !<  The magnification to be written to the Frealign parameter file
        REAL,                                   INTENT(IN)  ::  PIXEL_SIZE  !<  Pixel size in Angstroms
        ! Private variables
        INTEGER ::  PAR_LUN !<  Logical unit number for our parameter file
        INTEGER ::  IOS     !<  IO status code
        INTEGER ::  CURRENT_SP
        LOGICAL,    PARAMETER :: DEBUG = .FALSE.
        ! Start work

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(SP_TO_FREALIGN_PAR): SP info on entry: '
            CALL SPS_PRINT_INFO(SP)
        ENDIF

        ! Open the file
        PAR_LUN = AVAILABLE_LUN()

        ! Check whether the file already exists, in which case we'll overwrite it
        IF (FILE_EXISTS(FILENAME)) THEN
            WRITE(*,'(2A)') '**WARNING(SP_TO_FREALIGN_PAR): parameter file already exists, will be overwritten : ', &
                                TRIM(ADJUSTL(FILENAME))
        ENDIF
        OPEN(UNIT=PAR_LUN,FILE=TRIM(ADJUSTL(FILENAME)),STATUS='REPLACE',IOSTAT=IOS)
        IF (IOS .NE. 0) THEN
            WRITE(*,'(3A,I0)') '**ERROR(SP_TO_FREALIGN_PAR): file opening failed for file ', TRIM(ADJUSTL(FILENAME)), &
                                ' with error ', IOS
            CALL TERMINATE('Fatal error in SP_TO_FREALIGN_PAR')
        ENDIF

        ! Write the header
        WRITE(PAR_LUN,'(A)')'C           PSI   THETA     PHI     SHX     SHY    MAG   FILM      DF1      DF2  ANGAST  PRESA   DPRES'

        ! Loop over the single particles
        IF (DEBUG) WRITE(*,'(A)') '**debug(SP_TO_FREALIGN_PAR): about to start looping'
        DO CURRENT_SP=1,SIZE(SP)
            IF (ABS(SP(CURRENT_SP)%SUBUNIT_NUM - NINT(SP(CURRENT_SP)%SUBUNIT_NUM)) .GT. 0.001) THEN
                WRITE(*,'(A,I0,A,F0.3,A)') '**ERROR(SP_TO_FREALIGN_PAR): SP # ', CURRENT_SP,    &
                                ' has non-integer subunit_num of ', SP(CURRENT_SP)%SUBUNIT_NUM, &
                                '. Frealign parameter file cannot store this information. Retry with integer subunit_num values'
                CALL TERMINATE('Fatal error in SP_TO_FREALIGN_PAR')
            ENDIF
            IF (.NOT. SP(CURRENT_SP)%CTF%PIX_UNITS) THEN
                WRITE(*,'(A)') '**ERROR(SP_TO_FREALIGN_PAR): CTF needs to be given in pixel / radians units'
                CALL TERMINATE('Fatal error in SP_TO_FREALIGN_PAR')
            ENDIF
            WRITE(PAR_LUN,'(I7,1X,3(F7.2,1X),2(F7.2,1X),F7.0,1X,I5,1X,2(F8.1,1X),F7.2,1X,F6.2,1X,F7.2)') &
                                    CURRENT_SP, &
                                    SP(CURRENT_SP)%ANGLES%PSI, SP(CURRENT_SP)%ANGLES%THETA, MOD(SP(CURRENT_SP)%ANGLES%PHI,360.D0), &
                                    SP(CURRENT_SP)%SHIFT_X, SP(CURRENT_SP)%SHIFT_Y, &
                                    MAG, SP(CURRENT_SP)%FILM_NUM, &
                                    SP(CURRENT_SP)%CTF%DFMID1 * PIXEL_SIZE, SP(CURRENT_SP)%CTF%DFMID2 * PIXEL_SIZE, &
                                    DEGREES(SP(CURRENT_SP)%CTF%ANGAST), &
                                    DEGREES(ACOS(SP(CURRENT_SP)%SCORE)), 0.0
        ENDDO

        ! Close the file
        IF (DEBUG) WRITE(*,'(A)') '**debug(SP_TO_FREALIGN_PAR): about to close output file'
        CLOSE(UNIT=PAR_LUN)
        CALL LUN_IS_NOW_AVAILABLE(PAR_LUN)
        IF (DEBUG) WRITE(*,'(A)') '**debug(SP_TO_FREALIGN_PAR): about to return'
    END SUBROUTINE SP_TO_FREALIGN_PAR

    !>  \brief  Parse a Frealign parameter file and return an array of single particles
    FUNCTION FREALIGN_PAR_TO_SP(FILENAME) RESULT(SP)
        USE SINGLE_PARTICLES
        USE PARSERS
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        CHARACTER(LEN=*),       INTENT(IN)  ::  FILENAME    !<  Filename of Frealign parameter file
        ! Result
        TYPE(SINGLE_PARTICLE),  ALLOCATABLE ::  SP(:)       !<  Output array of single-particle objects
        !  Private variables
        INTEGER ::  NUM_SP, IERR
        REAL,   ALLOCATABLE ::  TMP(:)
        INTEGER, ALLOCATABLE :: ITMP(:)
        ! Start work
        ! Check the file exists & find out number of lines
        IF (.NOT. FILE_EXISTS(FILENAME)) THEN
            WRITE(*,'(A,A)') '**ERROR(FREALIGN_PAR_TO_SP): file does not exist: ', TRIM(FILENAME)
            CALL TERMINATE('Fatal error in FREALIGN_PAR_TO_SP')
        ENDIF
        ! Number of particles
        NUM_SP = NUMBER_OF_LINES_IN_FILE(FILENAME,IGNORE_COMMENTS=.TRUE.)
        ! Allocate
        ALLOCATE(SP(NUM_SP),TMP(NUM_SP),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(FREALIGN_PAR_TO_SP): memory allocation failed with error # ', IERR
            CALL TERMINATE('Fatal error in FREALIGN_PAR_TO_SP')
        ENDIF
        ! Read data in
        CALL GET_INTS_FROM_COLUMN(FILENAME,1,ITMP)
        SP%LOC_IN_STACK = ITMP
        ! The Euler angles and the shifts are those that need to be applied to the 3D to obtain the single-particle,
        ! which is counter-intuitive to me, so I convert them here straight away!
        CALL GET_REALS_FROM_COLUMN(FILENAME,2,TMP)
        SP%ANGLES%PSI = TMP
        CALL GET_REALS_FROM_COLUMN(FILENAME,3,TMP)
        SP%ANGLES%THETA = TMP
        CALL GET_REALS_FROM_COLUMN(FILENAME,4,TMP)
        SP%ANGLES%PHI = TMP
        ! Need to apply negative shifts to align particles with (unshifted) reference
        CALL GET_REALS_FROM_COLUMN(FILENAME,5,TMP)
        SP%SHIFT_X    = -TMP
        CALL GET_REALS_FROM_COLUMN(FILENAME,6,TMP)
        SP%SHIFT_Y    = -TMP
        CALL GET_REALS_FROM_COLUMN(FILENAME,8,TMP)
        SP%FILM_NUM   = TMP
        CALL GET_REALS_FROM_COLUMN(FILENAME,9,TMP)
        SP%CTF%DFMID1 = TMP
        CALL GET_REALS_FROM_COLUMN(FILENAME,10,TMP)
        SP%CTF%DFMID2 = TMP
        CALL GET_REALS_FROM_COLUMN(FILENAME,11,TMP)
        SP%CTF%ANGAST = TMP
        CALL GET_REALS_FROM_COLUMN(FILENAME,12,TMP)
        SP%SCORE      = TMP

    END FUNCTION FREALIGN_PAR_TO_SP


END MODULE FREALIGN_PARAMETERS
