!>  \brief  Module to interact with frealix parameter files
!!
!!
!!  There are several versions of the frealix parameter files. For full details, see functions FREALIX_PAR_HEADER, :
!!  - version 1 is from the frealign-like first incarnation of frealix, written by Grigorieff, Sindelar and Rohou
!!  - version 2 is from the modern incarnation, until 13-July-2010
!!  - version 3 is new as of July/August 2010. It incorporates a Z coordinate and a path length for each waypoint as stores all coordinates in Angstroems rather than pixels
!!  - version 4 is new as of April 2011. It incorporates the subunit number, and may store coordinates in pixels or Angstroems
!!
MODULE FREALIX_PARAMETERS

    ! The format of the frealix1 parameter file
    CHARACTER(LEN=*), PARAMETER   ::  FLX1_PAR_FORMAT                     =   &
                                    '(I6,1X,3(F7.2,1X),2(F9.2,1X),(F7.1,1X),(I5,1X),2(F8.1,1X),'//&
                                    '2(F7.2,1X),(E12.5,1X),2(F7.2,1X),(I8,1X),F7.2)'
    CHARACTER(LEN=6), PARAMETER   ::  FLX1_PAR_HEADER(17)                 =   &
                                    ['WP#   ','AIP   ','A-OOP ','AA    ','SHX   ','SHY   ','MAG   ',&
                                    'FILM  ','DF1   ','DF2   ','ANGAST','OCC   ','LogP  ','SCORE ','CHANGE','IFIL  ','dR    ']
    CHARACTER(LEN=*), PARAMETER   ::  FLX1_PAR_HEADER_FORMAT              =   &
                                    '(A6,1X,3(A7,1X),2(A9,1X),(A7,1X),(A5,1X),2(A8.1X),2(A7.1X),(A12,1X),2(A7,1X),(A8,1X),(A7))'

    ! The format for the frealix2 parameter file
    CHARACTER(LEN=*), PARAMETER   ::  FLX2_PAR_FORMAT                     =   '(3(I6,1X),2(F11.3,1X),3(F12.3,1X),(F8.3,1X),L7,1X)'
    CHARACTER(LEN=*), PARAMETER   ::  FLX2_PAR_HEADER_FORMAT              =   '(3(A6,1X),2(A11,1X),  3(A12,1X),  (A8,1X),  A7,1X)'

    ! The current latest version of the frealix2 parameter file
    INTEGER,          PARAMETER   ::  FLX2_PAR_LASTEST_VERSION            =   4
    ! If the parameter file does not specify which version it is, this version will be assumed
    INTEGER,          PARAMETER   ::  FLX2_PAR_VERSION_IF_NOT_SPECIFIED   =   2




    CONTAINS

    !>  \brief  Return the header for the relevant frealix parameter file version
    FUNCTION FREALIX_PAR_HEADER(FLX_PAR_VERSION,UNITS)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        INTEGER,                            INTENT(IN)  ::  FLX_PAR_VERSION
        CHARACTER(*),       OPTIONAL,       INTENT(IN)  ::  UNITS               !<  For versions 4 and above, we can be in Angstroems or pixels. The default is Angstroms, but this can be overriden here. If the first letter is P, the header will indicate pixels
        ! Result
        CHARACTER(LEN=:),   DIMENSION(:),   ALLOCATABLE ::  FREALIX_PAR_HEADER
        ! Private variables
        CHARACTER(LEN=:),   DIMENSION(:),   ALLOCATABLE ::  FLX_PAR_HEADERS_1,FLX_PAR_HEADERS_2,FLX_PAR_HEADERS_3, FLX_PAR_HEADERS_4
        CHARACTER(LEN=10)                               ::  UUNITS
        ! Start work


        IF (PRESENT(UNITS)) THEN
            IF (FLX_PAR_VERSION .LT. 4) THEN
                WRITE(*,'(A,I0,A)') '**WARNING(FREALIX_PAR_HEADER): specification of units will be ignored, since version ', &
                                    FLX_PAR_VERSION, ' is in use'
                CALL TERMINATE('Debug')
            ENDIF
            IF (UNITS(1:1) .EQ. 'P') THEN
                UUNITS = 'Pixels'
            ELSE
                UUNITS = 'Angstr'
            ENDIF
        ELSE
            UUNITS = 'Angstr'
        ENDIF

        ! Here are all the frealix parameter file headers
        FLX_PAR_HEADERS_1 = [CHARACTER(6)  ::   'WP#   ','AIP   ','A-OOP ','AA    ','SHX   ','SHY   ','MAG   ','FILM  ',&
                                                'DF1   ','DF2   ','ANGAST','OCC   ','LogP  ','SCORE ','CHANGE','IFIL  ','dR    ']
        FLX_PAR_HEADERS_2 = [CHARACTER(9)  ::   '!Fil.#   ','WP #     ', 'Film #   ', 'X        ','Y        ',&
                                                'Phi      ','Theta    ', 'Psi      ', 'Score    ','Refine?  ',&
                                                'par_vers.','2        ']
        FLX_PAR_HEADERS_3 = [CHARACTER(10) ::   '!Fil.#    ','WP #      ','Film #    ', &
                                                'X         ','Y         ','Z         ',&
                                                'PathLength','Phi       ','Theta     ',&
                                                'Psi       ','Score     ','Refine?   ',&
                                                'par_vers. ','3         ']
        FLX_PAR_HEADERS_4 = [CHARACTER(10) ::   '!Fil.#    ','WP #      ','Film #    ',&
                                                'X         ','Y         ','Z         ',&
                                                'PathLength','Phi       ','Theta     ',&
                                                'Psi       ','Subunit # ','Score     ',&
                                                'Refine?   ','Units     ',UUNITS,      &
                                                'par_vers. ','4         ']

        ! Return the relevant one
        !FREALIX_PAR_HEADER(:) = FLX_PAR_HEADERS(:,FLX_PAR_VERSION)
        SELECT CASE(FLX_PAR_VERSION)
            CASE(1)
                ALLOCATE(CHARACTER(LEN=6) :: FREALIX_PAR_HEADER(17))
                FREALIX_PAR_HEADER = FLX_PAR_HEADERS_1
            CASE(2)
                ALLOCATE(CHARACTER(LEN=7) :: FREALIX_PAR_HEADER(12))
                FREALIX_PAR_HEADER = FLX_PAR_HEADERS_2
            CASE(3)
                ALLOCATE(CHARACTER(LEN=10) :: FREALIX_PAR_HEADER(14))
                FREALIX_PAR_HEADER = FLX_PAR_HEADERS_3
            CASE(4)
                ALLOCATE(CHARACTER(LEN=10) :: FREALIX_PAR_HEADER(17))
                FREALIX_PAR_HEADER = FLX_PAR_HEADERS_4
            CASE DEFAULT
                WRITE(*,'(A,I0)') '**ERROR(FREALIX_PAR_HEADER): unknown parameter file format: ', FLX_PAR_VERSION
                CALL TERMINATE('Fatal error in FREALIX_PAR_HEADER')
        END SELECT
    END FUNCTION FREALIX_PAR_HEADER

    !>  \brief  Return the header format string for the relevant frealix parameter file version
    FUNCTION FREALIX_PAR_HEADER_FORMAT(FLX_PAR_VERSION)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, LINE_LEN
        IMPLICIT NONE
        ! Arguments
        INTEGER,    INTENT(IN)  ::  FLX_PAR_VERSION
        ! Result
        CHARACTER(LEN=LINE_LEN) ::  FREALIX_PAR_HEADER_FORMAT
        ! Private variables
        CHARACTER(LEN=LINE_LEN) ::  FLX_PAR_HEADER_FORMATS_1,FLX_PAR_HEADER_FORMATS_2
        CHARACTER(LEN=LINE_LEN) ::  FLX_PAR_HEADER_FORMATS_3,FLX_PAR_HEADER_FORMATS_4
        ! Start work

        ! Here are all the frealix parameter file header formats
        FLX_PAR_HEADER_FORMATS_1    = '(A6,1X,3(A7,1X),2(A9,1X),(A7,1X),(A5,1X),2(A8.1X),2(A7.1X),(A12,1X),2(A7,1X),(A8,1X),(A7))'
        FLX_PAR_HEADER_FORMATS_2    = '(3(A6,1X),2(A11,1X),  3(A12,1X),             (A8,1X),  A7,1X,             A11,1X,A1)'
        FLX_PAR_HEADER_FORMATS_3    = '(3(A6,1X),4(A11,1X),  3(A12,1X),             (A8,1X),  A7,1X,             A11,1X,A1)'
        FLX_PAR_HEADER_FORMATS_4    = '(3(A6,1X),4(A11,1X),  3(A12,1X),  1(A11,1X), (A8,1X),  A7,1X, A5,1X,A6,1X,A11,1X,A1)'

        ! Return the relevant one
        SELECT CASE(FLX_PAR_VERSION)
            CASE(1)
                FREALIX_PAR_HEADER_FORMAT = FLX_PAR_HEADER_FORMATS_1
            CASE(2)
                FREALIX_PAR_HEADER_FORMAT = FLX_PAR_HEADER_FORMATS_2
            CASE(3)
                FREALIX_PAR_HEADER_FORMAT = FLX_PAR_HEADER_FORMATS_3
            CASE(4)
                FREALIX_PAR_HEADER_FORMAT = FLX_PAR_HEADER_FORMATS_4
            CASE DEFAULT
                WRITE(*,'(A,I0)') '**ERROR(FREALIX_PAR_HEADER_FORMAT): unknown parameter file format: ', FLX_PAR_VERSION
                CALL TERMINATE('Fatal error in FREALIX_PAR_HEADER_FORMAT')
        END SELECT
    END FUNCTION FREALIX_PAR_HEADER_FORMAT

    !>  \brief  Return the header format string for the relevant frealix parameter file version
    FUNCTION FREALIX_PAR_FORMAT(FLX_PAR_VERSION)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, LINE_LEN
        IMPLICIT NONE
        ! Arguments
        INTEGER,    INTENT(IN)  ::  FLX_PAR_VERSION
        ! Result
        CHARACTER(LEN=LINE_LEN) ::  FREALIX_PAR_FORMAT
        ! Private variables
        CHARACTER(LEN=LINE_LEN) ::  FLX_PAR_FORMATS_1,FLX_PAR_FORMATS_2,FLX_PAR_FORMATS_3,FLX_PAR_FORMATS_4
        ! Start work

        ! Here are all the frealix parameter file header formats
        FLX_PAR_FORMATS_1   = '(I6,1X,3(F7.2,1X),2(F9.2,1X),(F7.1,1X),(I5,1X),2(F8.1,1X),'//&
                              '2(F7.2,1X),(E12.5,1X),2(F7.2,1X),(I8,1X),F7.2)'
        FLX_PAR_FORMATS_2   = '(3(I6,1X),2(F11.3,1X),3(F12.3,1X),            (F8.3,1X),L7,1X)'
        FLX_PAR_FORMATS_3   = '(3(I6,1X),4(F11.3,1X),3(F12.3,1X),            (F8.3,1X),L7,1X)'
        FLX_PAR_FORMATS_4   = '(3(I6,1X),4(F11.3,1X),3(F12.3,1X),1(F11.3,1X),(F8.3,1X),L7,1X)'

        ! Return the relevant one
        SELECT CASE(FLX_PAR_VERSION)
            CASE(1)
                FREALIX_PAR_FORMAT = FLX_PAR_FORMATS_1
            CASE(2)
                FREALIX_PAR_FORMAT = FLX_PAR_FORMATS_2
            CASE(3)
                FREALIX_PAR_FORMAT = FLX_PAR_FORMATS_3
            CASE(4)
                FREALIX_PAR_FORMAT = FLX_PAR_FORMATS_4
            CASE DEFAULT
                WRITE(*,'(A,I0)') '**ERROR(FREALIX_PAR_FORMAT): unknown parameter file format: ', FLX_PAR_VERSION
                CALL TERMINATE('Fatal error in FREALIX_PAR_FORMAT')
        END SELECT
    END FUNCTION FREALIX_PAR_FORMAT

    !>  \brief  Work out the version of a frealix2 parameter file
    !!
    !!  The version number is found at the end of the first line of the parameter file. The penultimate word in that line should be par_version and the last work should be an integer number.
    !!  If the version number is not specified there, this function returns the FLX2_PAR_VERSION_IF_NOT_SPECIFIED parameter
    INTEGER FUNCTION FREALIX_PAR_FIND_VERSION(PAR_FILENAME)
        USE RUNTIME_PARAMETERS
        USE PARSERS
        IMPLICIT NONE
        ! Arguments
        CHARACTER(LEN=*),               INTENT(IN)  ::  PAR_FILENAME            !<  Filename of parameter file
        ! Private variables
        CHARACTER(LEN=LINE_LEN) ::  LINE
        CHARACTER(LEN=LINE_LEN), ALLOCATABLE    ::  WORDS(:)
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        ! Read the first line
        CALL FILE_READ_LINE_NUM(PAR_FILENAME,1,LINE)
        IF (DEBUG) THEN
            PRINT *, '**debug(FREALIX_PAR_FIND_VERSION): filename = ', TRIM(ADJUSTL(PAR_FILENAME))
            PRINT *, '**debug(FREALIX_PAR_FIND_VERSION): line     = ', LINE
        ENDIF

        ! It should be a comment
        IF (.NOT. IS_COMMENT(LINE)) THEN
            WRITE(*,'(4A)') '**ERROR(FREALIX_PAR_FIND_VERSION): first line of file ', TRIM(ADJUSTL(PAR_FILENAME)), &
                                ' is not a comment :'
            WRITE(*,'(A)') TRIM(ADJUSTL(LINE))
            CALL TERMINATE('Fatal error in FREALIX_PAR_FIND_VERSION')
        ENDIF

        ! Split the line
        WORDS = SPLIT(LINE)
        !PRINT '(2A)', '**debug(FREALIX_PAR_FIND_VERSION): LINE = ', TRIM(ADJUSTL(LINE))
        !PRINT '(A,I3)', '**debug(FREALIX_PAR_FIND_VERSION): # of words: ', SIZE(WORDS)

        ! The penultimate word should be par_version. If it isn't, we'll assume that the version just isn't specified.
        IF (.NOT. STRING_EQUAL(WORDS(SIZE(WORDS)-1),'par_vers.')) THEN
            WRITE(*,'(3A,I2)') '**info(FREALIX_PAR_FIND_VERSION): unexpected penultimate word (not par_vers.): ', &
                                TRIM(ADJUSTL(WORDS(SIZE(WORDS)-1))), &
                                '. Will revert to default parameter file version ', FLX2_PAR_VERSION_IF_NOT_SPECIFIED
            !CALL TERMINATE('Fatal error in FREALIX2_PAR_FIND_VERSION')
            FREALIX_PAR_FIND_VERSION = FLX2_PAR_VERSION_IF_NOT_SPECIFIED
            RETURN
        ENDIF

        ! The last word should be the version number, an integer
        LINE = WORDS(SIZE(WORDS))
        IF (.NOT. IS_INTG(LINE)) THEN
            WRITE(*,'(A)') '**ERROR(FREALIX_PAR_FIND_VERSION): last word is not an integer :', TRIM(ADJUSTL(LINE))
            CALL TERMINATE('Fatal error in FREALIX_PAR_FIND_VERSION')
        ENDIF

        ! We're almost done
        READ(LINE,'(I2)') FREALIX_PAR_FIND_VERSION

    END FUNCTION FREALIX_PAR_FIND_VERSION

    !>  \brief  For versions 4 and above, frealix parameter files can store coordinates in pixels or Angstroms. Use this function to find out which is in use.
    !!
    FUNCTION FREALIX_PAR_FIND_UNITS(PAR_FILENAME) RESULT(UNITS)
        USE RUNTIME_PARAMETERS
        USE PARSERS
        IMPLICIT NONE
        ! Arguments
        CHARACTER(LEN=*),               INTENT(IN)  ::  PAR_FILENAME            !<  Filename of parameter file
        ! Results
        CHARACTER(LEN=1)    ::  UNITS
        ! Private variables
        CHARACTER(LEN=LINE_LEN) ::  LINE
        CHARACTER(LEN=LINE_LEN), ALLOCATABLE    ::  WORDS(:)
        INTEGER ::  PAR_VERSION
        ! Start work

        ! Find out the version of the parameter file first
        PAR_VERSION = FREALIX_PAR_FIND_VERSION(PAR_FILENAME)

        IF (PAR_VERSION .LT. 3) THEN
            UNITS = 'P'
        ELSE IF (PAR_VERSION .EQ. 3) THEN
            UNITS = 'A'
        ELSE
            ! For versions 4 and above, we need to check the header line

            ! Read the first line
            CALL FILE_READ_LINE_NUM(PAR_FILENAME,1,LINE)

            ! It should be a comment
            IF (.NOT. IS_COMMENT(LINE)) THEN
                WRITE(*,'(4A)') '**ERROR(FREALIX_PAR_FIND_UNITS): first line of file ', TRIM(ADJUSTL(PAR_FILENAME)), &
                                    ' is not a comment :'
                WRITE(*,'(A)') TRIM(ADJUSTL(LINE))
                CALL TERMINATE('Fatal error in FREALIX_PAR_FIND_UNITS')
            ENDIF

            ! Split the line
            WORDS = SPLIT(LINE)

            ! The fourth word to last should be "Units". If it isn't, we'll assume that the version just isn't specified.
            IF (.NOT. STRING_EQUAL(WORDS(SIZE(WORDS)-3),'Units')) THEN
                WRITE(*,'(2A)') '**info(FREALIX_PAR_FIND_UNITS): unexpected penultimate word (not Units): ', &
                                TRIM(ADJUSTL(WORDS(SIZE(WORDS)-3)))
                CALL TERMINATE('Fatal error in FREALIX_PAR_FIND_UNITS')
            ENDIF

            ! The third word to last should be either 'Pixels' or 'Angstr'
            SELECT CASE (WORDS(SIZE(WORDS)-2)(1:1))
                CASE ('P')
                    UNITS = 'P'
                CASE ('A')
                    UNITS = 'A'
                CASE DEFAULT
                    WRITE(*,'(2A)') '**ERROR(FREALIX_PAR_FIND_UNITS): unit specification not recognised: ', &
                                    TRIM(ADJUSTL(WORDS(SIZE(WORDS)-2)))
                    CALL TERMINATE('Fatal error in FREALIX_PAR_FIND_UNITS')
            END SELECT
        ENDIF ! End of check for frealix par version

    END FUNCTION FREALIX_PAR_FIND_UNITS


    !>  \brief  Apply a binning factor to a frealix2 parameter file
    !!  \todo   Make this work with other version of frealix parameter files
    SUBROUTINE FREALIX2_PAR_BIN(INPUT_PAR_FILENAME,OUTPUT_PAR_FILENAME,BIN_FACTOR,UNBIN)
        !USE FILAMENTS
        USE PARSERS
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, AVAILABLE_LUN, LUN_IS_NOW_AVAILABLE
		IMPLICIT NONE
        ! Arguments
        CHARACTER(LEN=*),               INTENT(IN)  ::  INPUT_PAR_FILENAME      !<  Filename for input parameter file
        CHARACTER(LEN=*),               INTENT(IN)  ::  OUTPUT_PAR_FILENAME     !<  Filename for output parameter file
        INTEGER,                        INTENT(IN)  ::  BIN_FACTOR              !<  Binning factor
        LOGICAL,            OPTIONAL,   INTENT(IN)  ::  UNBIN                   !<  If .TRUE., unbinning, rather than binning, is done
        ! Private variables
        LOGICAL ::  UUNBIN
        REAL    ::  BBIN_FACTOR
        !TYPE(FILAMENT),    ALLOCATABLE ::  FILS(:)
        INTEGER,    ALLOCATABLE ::  FIL_NUM(:),WP_NUM(:),FILM_NUM(:)
        REAL,       ALLOCATABLE ::  X(:),Y(:),PHI(:),THETA(:),PSI(:),SCORE(:)
        LOGICAL,    ALLOCATABLE ::  REFINE(:)
        INTEGER::   PAR_LUN
        INTEGER                 ::  I, IOS
        ! Start work

        PAR_LUN = AVAILABLE_LUN()

        !
        ! This only deals with frealix2 parameter files (i.e. version 2 format) -TODO: make it work on all formats (note: nothing to do for version 3 since coordinates are stored in Angstroems)
        !
        IF (FREALIX_PAR_FIND_VERSION(INPUT_PAR_FILENAME) .NE. 2) THEN
            WRITE(*,'(2A)') '**ERROR(FREALIX2_PAR_BIN): subroutine not ready for this format of parameter file: ', &
                            FREALIX_PAR_FIND_VERSION(INPUT_PAR_FILENAME)
            CALL TERMINATE('Fatal error in FREALIX2_PAR_BIN')
        ENDIF

        !
        ! Are we binning or unbinning?
        !
        IF (PRESENT(UNBIN)) THEN
            UUNBIN = UNBIN
        ELSE
            UUNBIN = .FALSE.
        ENDIF

        ! Check whether the output file already exists, in which case we'll overwrite it
        IF (FILE_EXISTS(OUTPUT_PAR_FILENAME)) THEN
            WRITE(*,'(2A)') '**WARNING(FREALIX2_PAR_BIN): parameter file already exists, will be overwritten : ', &
                            TRIM(ADJUSTL(OUTPUT_PAR_FILENAME))
        ENDIF
        OPEN(UNIT=PAR_LUN,FILE=TRIM(ADJUSTL(OUTPUT_PAR_FILENAME)),STATUS='REPLACE',IOSTAT=IOS)
        IF (IOS .NE. 0) THEN
            WRITE(*,'(3A,I3)') '**ERROR(FREALIX2_PAR_BIN): file opening failed for file ', TRIM(ADJUSTL(OUTPUT_PAR_FILENAME)), &
                                ' with error ', IOS
            !STOP 'Fatal error in FREALIX2_PAR_BIN'
            CALL TERMINATE()
        ENDIF

        !
        ! Read in all the information from the input file
        !
        CALL GET_INTS_FROM_COLUMN       (INPUT_PAR_FILENAME,1,FIL_NUM)
        CALL GET_INTS_FROM_COLUMN       (INPUT_PAR_FILENAME,2,WP_NUM)
        CALL GET_INTS_FROM_COLUMN       (INPUT_PAR_FILENAME,3,FILM_NUM)
        CALL GET_REALS_FROM_COLUMN      (INPUT_PAR_FILENAME,4,X)
        CALL GET_REALS_FROM_COLUMN      (INPUT_PAR_FILENAME,5,Y)
        CALL GET_REALS_FROM_COLUMN      (INPUT_PAR_FILENAME,6,PHI)
        CALL GET_REALS_FROM_COLUMN      (INPUT_PAR_FILENAME,7,THETA)
        CALL GET_REALS_FROM_COLUMN      (INPUT_PAR_FILENAME,8,PSI)
        CALL GET_REALS_FROM_COLUMN      (INPUT_PAR_FILENAME,9,SCORE)
        CALL GET_LOGICALS_FROM_COLUMN   (INPUT_PAR_FILENAME,10,REFINE)

        !
        ! Do the binning
        !
        ! The first pixel is (1,1,1), not (0,0,0), so we have to be careful.
        !
        IF (UUNBIN) THEN
            BBIN_FACTOR = 1.0/REAL(BIN_FACTOR)
        ELSE
            BBIN_FACTOR = REAL(BIN_FACTOR)
        ENDIF
        !PRINT *, '**debug(FREALIX2_PAR_BIN): binning by ', BBIN_FACTOR
        X = (X-1.0)/BBIN_FACTOR + 1.0
        Y = (Y-1.0)/BBIN_FACTOR + 1.0

        !
        ! Write modified info to output parameter file
        !
        WRITE(PAR_LUN,FLX2_PAR_HEADER_FORMAT) '!Fil.#', 'WP #', 'Film #', 'X','Y','Phi','Theta','Psi','Score','Refine?'
        DO I=1,SIZE(WP_NUM)
            WRITE(PAR_LUN,FLX2_PAR_FORMAT)  FIL_NUM(I), WP_NUM(I), FILM_NUM(I),     &
                                            X(I), Y(I),                             &
                                            PHI(I), THETA(I), PSI(I),               &
                                            SCORE(I),                               &
                                            REFINE(I)
        ENDDO

        CLOSE(UNIT=PAR_LUN)
        CALL LUN_IS_NOW_AVAILABLE(PAR_LUN)

    END SUBROUTINE FREALIX2_PAR_BIN

    !>  \brief  Dump the current state of the filament waypoints to a frealix parameter file, using the latest version of the parameter file format
    SUBROUTINE FILAMENTS_TO_FREALIX_PAR(FILS,PAR_FILENAME,PAR_VERSION,UNITS,SUBUNITS)
        USE FILAMENTS_BASE
		USE FILAMENTS_PATH, ONLY  : FILAMENT_GET_PATH
        USE PARSERS
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        USE RUNTIME_PARAMETERS, ONLY: TERMINATE, AVAILABLE_LUN, LUN_IS_NOW_AVAILABLE
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     ALLOCATABLE,    INTENT(INOUT)   ::  FILS(:)             !<  Array of filaments
        CHARACTER(LEN=*),                   INTENT(IN)      ::  PAR_FILENAME        !<  Filename for parameter file
        INTEGER,            OPTIONAL,       INTENT(IN)      ::  PAR_VERSION         !<  Version of parameter file format to use. Defaults to latest.
        CHARACTER(LEN=1),   OPTIONAL,       INTENT(IN)      ::  UNITS               !<  Either 'A' or 'P', for Angstroems or pixels. Defaults to Angstroems
        LOGICAL,            OPTIONAL,       INTENT(IN)      ::  SUBUNITS            !<  Output one line per helical asymetric unit, rather than one line per waypoint
        ! Private variables
        INTEGER ::  IOS ! IO status
        INTEGER ::  PAR_LUN
        INTEGER ::  FIL_NUM, WP_NUM, FILM_NUM
        REAL    ::  PIXEL_SIZE
        INTEGER ::  PPAR_VERSION
        CHARACTER(LEN=:), DIMENSION(:), ALLOCATABLE :: PPAR_HEADER
        INTEGER :: I
        LOGICAL, PARAMETER :: DEBUG = .FALSE.
        CHARACTER(LEN=1)    ::  UUNITS
        REAL    ::  TEMP_COO(2)
        LOGICAL ::  SSUBUNITS
        TYPE(WAYPOINT), ALLOCATABLE ::  WPS(:)
        ! Start work

        ! Was a parameter file format version specified?
        IF (PRESENT(PAR_VERSION)) THEN
            PPAR_VERSION = PAR_VERSION
        ELSE
            PPAR_VERSION = FLX2_PAR_LASTEST_VERSION
        ENDIF

        ! Were the units specified?
        IF (PRESENT(UNITS)) THEN
            ! Units are only availble for parameter file versions 4 and above
            IF (PPAR_VERSION .LT. 4) THEN
                WRITE(*,'(A)') '**ERROR(FILAMENTS_TO_FREALIX_PAR): specification of units only availble for ',  &
                                'parameter file formats 4 and later'
                CALL TERMINATE('Fatal error in FILAMENTS_TO_FREALIX_PAR')
            ENDIF
            UUNITS = UNITS
        ELSE
            UUNITS = 'A'
        ENDIF

        ! Subunits?
        SSUBUNITS=.FALSE.
        IF (PRESENT(SUBUNITS)) SSUBUNITS = SUBUNITS


        PAR_LUN = AVAILABLE_LUN()

        ! Check whether the file already exists, in which case we'll overwrite it
        IF (FILE_EXISTS(PAR_FILENAME)) THEN
            WRITE(*,'(2A)') '**WARNING(FILAMENTS_TO_FREALIX_PAR): parameter file already exists, will be overwritten : ', &
                            TRIM(ADJUSTL(PAR_FILENAME))
        ENDIF
        OPEN(UNIT=PAR_LUN,FILE=TRIM(ADJUSTL(PAR_FILENAME)),STATUS='REPLACE',IOSTAT=IOS)
        IF (IOS .NE. 0) THEN
            WRITE(*,'(3A,I0)') '**ERROR(FILAMENTS_TO_FREALIX2_PAR): file opening failed for file ', TRIM(ADJUSTL(PAR_FILENAME)),&
                                 ' with error ', IOS
            CALL TERMINATE('Fatal error in FILAMENTS_TO_FREALIX_PAR')
        ENDIF

        ! Write header to file
        PPAR_HEADER = FREALIX_PAR_HEADER(PPAR_VERSION,UNITS=UUNITS)
        WRITE(PAR_LUN,FREALIX_PAR_HEADER_FORMAT(PPAR_VERSION)) (TRIM(PPAR_HEADER(I)),I=1,UBOUND(PPAR_HEADER,1))


        ! Loop over the filaments
        DO FIL_NUM=1,SIZE(FILS)
            FILM_NUM = FILS(FIL_NUM)%FILM%ID
            IF (DEBUG) THEN
                !PRINT '(2A)', '**debug(FILAMENTS_TO_FREALIX_PAR): format = ', FREALIX_PAR_FORMAT(PPAR_VERSION)
                !PRINT '(A,I)', '**debug(FILAMENTS_TO_FREALIX_PAR): format num = ', PPAR_VERSION
                WRITE(*,'(A)') '**debug(FILAMENTS_TO_FREALIX_PAR): about to add the following filament:'
                CALL FILAMENT_PRINT_INFO(FILS(FIL_NUM))
            ENDIF
            IF (ALLOCATED(WPS)) DEALLOCATE(WPS)
            IF (SSUBUNITS) THEN
                IF (FILS(FIL_NUM)%HELIX%NUM_STARTS .NE. 1) CALL TERMINATE('Conversion to one start not handled here yet')
                CALL FILAMENT_GET_PATH(     FILS(FIL_NUM),WPS,  &
                                            STEP=1.0D0*REAL(FILS(FIL_NUM)%HELIX%AXIAL_SYMMETRY,KIND=8)&
                                                      *REAL(FILS(FIL_NUM)%HELIX%PERPENDICULAR_SYMMETRY,KIND=8), &
                                            STEP_IN_REPEATS=.TRUE.,&
                                            IGNORE_FAL_WPS=.NOT. FILS(FIL_NUM)%FAL_WPS_IN_THREED)
            ELSE
                WPS = FILS(FIL_NUM)%WP
            ENDIF
            DO WP_NUM=1,SIZE(WPS)
                PIXEL_SIZE = FILS(FIL_NUM)%FILM%PIXEL_SIZE
                SELECT CASE(PPAR_VERSION)
                    CASE (2)
                        WRITE(PAR_LUN,FREALIX_PAR_FORMAT(PPAR_VERSION))     FILS(FIL_NUM)%ID, WP_NUM, FILM_NUM,    &
                                                                            WPS(WP_NUM)%COO(1),                    &
                                                                            WPS(WP_NUM)%COO(2),                    &
                                                                            WPS(WP_NUM)%EULER%PHI,                 &
                                                                            WPS(WP_NUM)%EULER%THETA,               &
                                                                            WPS(WP_NUM)%EULER%PSI,                 &
                                                                            FILS(FIL_NUM)%SCORE,                   &
                                                                            WPS(WP_NUM)%REFINE
                    CASE (3)
                        WRITE(PAR_LUN,FREALIX_PAR_FORMAT(PPAR_VERSION))     FILS(FIL_NUM)%ID, WP_NUM, FILM_NUM,    &
                                                                            (WPS(WP_NUM)%COO(1)-1.0)*PIXEL_SIZE,   &   ! -1.0 because the origin is at (1,1) pixel or (0,0) Angstroems
                                                                            (WPS(WP_NUM)%COO(2)-1.0)*PIXEL_SIZE,   &
                                                                            WPS(WP_NUM)%COO(3)*PIXEL_SIZE,         &
                                                                            WPS(WP_NUM)%PATH_LENGTH*PIXEL_SIZE,    &
                                                                            WPS(WP_NUM)%EULER%PHI,                 &
                                                                            WPS(WP_NUM)%EULER%THETA,               &
                                                                            WPS(WP_NUM)%EULER%PSI,                 &
                                                                            FILS(FIL_NUM)%SCORE,                   &
                                                                            WPS(WP_NUM)%REFINE
                    CASE (4)
                        ! Need to deal with pixel / angstroem units
                        IF (UUNITS .EQ. 'P') THEN
                            PIXEL_SIZE = 1.0
                            TEMP_COO(1:2) = WPS(WP_NUM)%COO(1:2)
                        ELSE
                            TEMP_COO(1:2) = WPS(WP_NUM)%COO(1:2)-1.0       ! -1.0 because the origin is at (1,1) pixel or (0,0) Angstroems
                        ENDIF
                        IF (DEBUG) THEN
                            WRITE(*,'(3A,G15.3)')    '**debug(FILAMENTS_TO_FREALIX_PAR): unit is ', UUNITS, &
                                                    ', therefore will use pixel size of ', PIXEL_SIZE
                            PRINT *, '**DEBUG() FMT = ', FREALIX_PAR_FORMAT(PPAR_VERSION)
                        ENDIF
                        WRITE(PAR_LUN,FREALIX_PAR_FORMAT(PPAR_VERSION))     FILS(FIL_NUM)%ID, WP_NUM, FILM_NUM,    &
                                                                            TEMP_COO(1)*PIXEL_SIZE,                &
                                                                            TEMP_COO(2)*PIXEL_SIZE,                &
                                                                            WPS(WP_NUM)%COO(3)*PIXEL_SIZE,         &
                                                                            WPS(WP_NUM)%PATH_LENGTH*PIXEL_SIZE,    &
                                                                            WPS(WP_NUM)%EULER%PHI,                 &
                                                                            WPS(WP_NUM)%EULER%THETA,               &
                                                                            WPS(WP_NUM)%EULER%PSI,                 &
                                                                            WPS(WP_NUM)%SUBUNIT_NUM,               &
                                                                            WPS(WP_NUM)%SCORE,                     &
                                                                            WPS(WP_NUM)%REFINE
                    CASE DEFAULT
                        WRITE(*,'(A)') '**ERROR(FILAMENTS_TO_FREALIX_PAR): unsupported parameter file format: ', PPAR_VERSION
                        CALL TERMINATE('Fatal error in FILAMENTS_TO_FREALIX_PAR')
                END SELECT
            ENDDO
        ENDDO

        CLOSE(UNIT=PAR_LUN)
        CALL LUN_IS_NOW_AVAILABLE(PAR_LUN)

    END SUBROUTINE FILAMENTS_TO_FREALIX_PAR

    !>  \brief  Write a frealix alignment file describing the input array of waypoints
    SUBROUTINE WAYPOINTS_TO_FREALIX_ALI(WP,PAR_FILENAME,PIXEL_SIZE,FIL_ID,FILM_NUM,SCORE)
        USE RUNTIME_PARAMETERS, ONLY : AVAILABLE_LUN, LUN_IS_NOW_AVAILABLE, TERMINATE
        USE PARSERS, ONLY : FILE_EXISTS
        USE COOS
        IMPLICIT NONE
        ! Arguments
        TYPE(WAYPOINT),     ALLOCATABLE,    INTENT(IN)  ::  WP(:)
        CHARACTER(LEN=*),                   INTENT(IN)  ::  PAR_FILENAME
        REAL,                               INTENT(IN)  ::  PIXEL_SIZE
        INTEGER,            OPTIONAL,       INTENT(IN)  ::  FIL_ID
        INTEGER,            OPTIONAL,       INTENT(IN)  ::  FILM_NUM
        REAL,               OPTIONAL,       INTENT(IN)  ::  SCORE
        ! Private variables
        INTEGER ::  PAR_LUN, IOS,I, WP_NUM, FFIL_ID, FFILM_NUM, PPAR_VERSION
        REAL    ::  SSCORE, TEMP_COO(3)
        CHARACTER(LEN=:), DIMENSION(:), ALLOCATABLE :: PPAR_HEADER
        CHARACTER(LEN=1) :: UUNITS
        ! Start work

        PAR_LUN = AVAILABLE_LUN()

        UUNITS = 'A'
        PPAR_VERSION = 4
        FFIL_ID = 1
        FFILM_NUM = 1
        SSCORE = 0.0
        IF (PRESENT(FIL_ID)) FFIL_ID = FIL_ID
        IF (PRESENT(FILM_NUM)) FFILM_NUM = FILM_NUM
        IF (PRESENT(SCORE)) SSCORE = SCORE

        ! Check whether the file already exists, in which case we'll overwrite it
        IF (FILE_EXISTS(PAR_FILENAME)) THEN
            WRITE(*,'(2A)') '**WARNING(WAYPOINTS_TO_FREALIX_ALI): parameter file already exists, will be overwritten : ', &
                            TRIM(ADJUSTL(PAR_FILENAME))
        ENDIF
        OPEN(UNIT=PAR_LUN,FILE=TRIM(ADJUSTL(PAR_FILENAME)),STATUS='REPLACE',IOSTAT=IOS)
        IF (IOS .NE. 0) THEN
            WRITE(*,'(3A,I0)') '**ERROR(WAYPOINTS_TO_FREALIX_ALI): file opening failed for file ', &
                                TRIM(ADJUSTL(PAR_FILENAME)), ' with error ', IOS
            CALL TERMINATE('Fatal error in WAYPOINTS_TO_FREALIX_ALI')
        ENDIF

        ! Write header to file
        PPAR_HEADER = FREALIX_PAR_HEADER(PPAR_VERSION,UNITS=UUNITS)
        WRITE(PAR_LUN,FREALIX_PAR_HEADER_FORMAT(PPAR_VERSION)) (TRIM(PPAR_HEADER(I)),I=1,UBOUND(PPAR_HEADER,1))

        DO WP_NUM=1,SIZE(WP)
            ! Need to deal with pixel / angstroem units
            TEMP_COO(1:2) = WP(WP_NUM)%COO(1:2)-1.0       ! -1.0 because the origin is at (1,1) pixel or (0,0) Angstroems
            WRITE(PAR_LUN,FREALIX_PAR_FORMAT(PPAR_VERSION))     FFIL_ID, WP_NUM, FFILM_NUM,           &
                                                                TEMP_COO(1)*PIXEL_SIZE,               &
                                                                TEMP_COO(2)*PIXEL_SIZE,               &
                                                                WP(WP_NUM)%COO(3)*PIXEL_SIZE,         &
                                                                WP(WP_NUM)%PATH_LENGTH*PIXEL_SIZE,    &
                                                                WP(WP_NUM)%EULER%PHI,                 &
                                                                WP(WP_NUM)%EULER%THETA,               &
                                                                WP(WP_NUM)%EULER%PSI,                 &
                                                                WP(WP_NUM)%SUBUNIT_NUM,               &
                                                                SSCORE,                               &
                                                                WP(WP_NUM)%REFINE
        ENDDO

        CLOSE(UNIT=PAR_LUN)
        CALL LUN_IS_NOW_AVAILABLE(PAR_LUN)
    END SUBROUTINE WAYPOINTS_TO_FREALIX_ALI


    !>  \brief  Read a frealix parameter file (specified version and return an array of filament objects, not initialised)
    SUBROUTINE FILAMENTS_FROM_FREALIX_PAR_NOINIT(FILS,PAR_FILENAME,PIXEL_SIZE)
        USE FILAMENTS_BASE
		USE RUNTIME_PARAMETERS, ONLY : INT_UNIQUE, TERMINATE
		USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
		USE COOS, ONLY : WAYPOINTS_PRINT_INFO
        USE PARSERS
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT),     ALLOCATABLE, TARGET,   INTENT(INOUT)   ::  FILS(:)
        CHARACTER(LEN=*),                          INTENT(IN)      ::  PAR_FILENAME                        !<  Filename for parameter file
        REAL,               OPTIONAL,              INTENT(IN)      ::  PIXEL_SIZE                          !<  Pixel size (Angstroms). Required for frealix par format 3+.
        ! Result

        ! Private variables
        INTEGER                     ::  FLX_PAR_VERSION
        INTEGER,        ALLOCATABLE ::  FILM_NUM(:), FIL_NUM(:)
        REAL(KIND=8),   ALLOCATABLE ::  X(:), Y(:), Z(:), PATH_LENGTH(:), PHI(:), THETA(:), PSI(:), SUBUNIT_NUM(:)
        REAL(KIND=4),   ALLOCATABLE ::  SCORE(:)
        LOGICAL,        ALLOCATABLE ::  REFINE(:)
        INTEGER,        ALLOCATABLE ::  TMP_INT(:)
        TYPE(WAYPOINT), ALLOCATABLE ::  WP(:)
        INTEGER                     ::  NUM_OF_FILS
        INTEGER                     ::  I, J, J_START, IERR
        INTEGER                     ::  NUM_OF_WPS, CURRENT_WP
        LOGICAL                     ::  PAR_IN_ANGSTROMS
        LOGICAL,        PARAMETER   ::  DEBUG   = .FALSE.
        ! Start work

        ! Check the parameter file exists
        IF (.NOT. FILE_EXISTS(PAR_FILENAME)) THEN
            WRITE(*,'(A,A)') '**ERROR(FREALIX_PAR_TO_FILAMENTS): file does not exist: ', TRIM(ADJUSTL(PAR_FILENAME))
            CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS')
        ENDIF

        ! Which version of frealix parameter file?
        FLX_PAR_VERSION = FREALIX_PAR_FIND_VERSION(PAR_FILENAME)

        ! Is the parameter file in Angstroms?
        PAR_IN_ANGSTROMS = FREALIX_PAR_FIND_UNITS(PAR_FILENAME) .EQ. 'A'

        IF (DEBUG) THEN
            WRITE(*,'(A,I0,1X,L1)') '**debug(FILAMENTS_FROM_FREALIX_PAR_NOINIT): version, Angstroms? = ', &
                                    FLX_PAR_VERSION, PAR_IN_ANGSTROMS
        ENDIF

        ! If required, check the pixel size was given
        IF (PAR_IN_ANGSTROMS) THEN
            IF (PRESENT(PIXEL_SIZE)) THEN
                !PPIXEL_SIZE  = PIXEL_SIZE
            ELSE
                WRITE(*,'(2A)') '**ERROR(FREALIX_PAR_TO_FILAMENTS_NOINIT): pixel size is required when reading in ', &
                                'frealix parameter files versions 3+'
                CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
            ENDIF
        ENDIF


        ! Read in data
        IF (FLX_PAR_VERSION .EQ. 1) THEN
            WRITE(*,'(2A)') '**ERROR(FREALIX_PAR_TO_FILAMENTS_NOINIT): unable to deal with frealix parameter ',  &
                            'file of version 1 yet. Sorry.'
            CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
        ELSE IF (FLX_PAR_VERSION .EQ. 2) THEN
            CALL GET_INTS_FROM_COLUMN(PAR_FILENAME,1,FIL_NUM)
            CALL GET_INTS_FROM_COLUMN(PAR_FILENAME,3,FILM_NUM)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,4,X)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,5,Y)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,6,PHI)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,7,THETA)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,8,PSI)
            CALL GET_REALS_FROM_COLUMN(PAR_FILENAME,9,SCORE)
            CALL GET_LOGICALS_FROM_COLUMN(PAR_FILENAME,10,REFINE)
        ELSE IF (FLX_PAR_VERSION .EQ. 3 .OR. FLX_PAR_VERSION .EQ. 4) THEN
            CALL GET_INTS_FROM_COLUMN(PAR_FILENAME,1,FIL_NUM)
            CALL GET_INTS_FROM_COLUMN(PAR_FILENAME,3,FILM_NUM)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,4,X)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,5,Y)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,6,Z)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,7,PATH_LENGTH)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,8,PHI)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,9,THETA)
            CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,10,PSI)
            IF (FLX_PAR_VERSION .EQ. 3) THEN
                CALL GET_REALS_FROM_COLUMN(PAR_FILENAME,11,SCORE)
                CALL GET_LOGICALS_FROM_COLUMN(PAR_FILENAME,12,REFINE)
            ELSE IF (FLX_PAR_VERSION .EQ. 4) THEN
                CALL GET_DBLES_FROM_COLUMN(PAR_FILENAME,11,SUBUNIT_NUM)
                CALL GET_REALS_FROM_COLUMN(PAR_FILENAME,12,SCORE)
                CALL GET_LOGICALS_FROM_COLUMN(PAR_FILENAME,13,REFINE)
            ENDIF
        ELSE
            WRITE(*,'(A,I0)') '**ERROR(FREALIX_PAR_TO_FILAMENTS_NOINIT): unknown frealix parameter file format: ', FLX_PAR_VERSION
            CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
        ENDIF

        ! Check we have the expected number of lines etc.
        NUM_OF_WPS  =   NUMBER_OF_LINES_IN_FILE(PAR_FILENAME,IGNORE_COMMENTS=.TRUE.)
        IF (SIZE(FIL_NUM) .NE. NUM_OF_WPS) THEN
            WRITE(*,'(2(A,I6),2A)') '**ERROR(FREALIX_PAR_TO_FILAMENTS_NOINIT): found ', SIZE(FIL_NUM), &
                                    ' values in column1, but counted ', NUM_OF_WPS, ' lines in file ', TRIM(PAR_FILENAME)
            CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
        ENDIF

        ! Work out number of filaments
        CALL GET_INTS_FROM_COLUMN(PAR_FILENAME,1,FIL_NUM)
        TMP_INT = INT_UNIQUE(FIL_NUM)
        NUM_OF_FILS = SIZE(TMP_INT)



        ! Allocate filament array
        IF (ALLOCATED(FILS)) DEALLOCATE(FILS)
        ALLOCATE(FILS(NUM_OF_FILS),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(FREALIX_PAR_TO_FILAMENTS): filament array memory allocation failed with error ', IERR
            CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS')
        ENDIF

        ! Loop over the filaments to find out how many waypoints are in each and then copy the data over into the filament objects
        CURRENT_WP = 1
        J = 0   ! This is the current line/waypoint
        DO I=1,NUM_OF_FILS
            ! At the beginning of the iteration, J points to the last waypoint of the previous filament
            J = J+1
            J_START = J

            ! For the current filament, work out how many waypoints are defined in the parameter file
            ! If we're on the last waypoint, of course we don't need to do this
            IF (J .LT. NUM_OF_WPS) THEN
                ! Let's increment J until we reach the last waypoint in the current filament, I
                DO
                    IF (J .EQ. NUM_OF_WPS) THEN
                        ! We reached the last way point
                        EXIT
                    ELSE IF (FIL_NUM(J+1) .EQ. FIL_NUM(J)) THEN
                        ! We are not yet at last way point of current filament
                        J = J +1
                    ELSE
                        ! We are at last waypoint of current filament
                        EXIT
                    ENDIF
                ENDDO
            ELSE
                ! We're on the last waypoint. This is strange, cos it implies the last way point is "alone"
                WRITE(*,'(A)') '**ERROR(FREALIX_PAR_TO_FILAMENTS_NOINIT): last way point appears to be alone'
                CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
            ENDIF

            ! So now J is set to the position of the last way point of the current filament

            ! Diagnostic message
            IF (DEBUG) THEN
                PRINT '(A,I3,A,I2,A)', '**debug(FREALIX_PAR_TO_FILAMENTS_NOINIT): filament ', FIL_NUM(J),&
                                             ' has ', J-J_START+1, ' way points'
            ENDIF

            ! Allocate memory to an array of way points
            IF (ALLOCATED(WP)) DEALLOCATE(WP)
            ALLOCATE(WP(J-J_START+1),STAT=IERR)
            IF (IERR .NE. 0) THEN
                WRITE(*,'(A,I0)') '**ERROR(FREALIX_PAR_TO_FILAMENTS_NOINIT): memory allocation of temporary ',   &
                                    'waypoint array failed with err ', IERR
                CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
            ENDIF



            ! Set up the film - we can do this by parsing either CTFTILT or CTFFIND
            ! If the user has given us a CTFTILT file, we use that.
!           CALL FILENAME_FROM_WILD(FILM_FILENAME_WILD,FILM_NUM(J_START),FILM_FILENAME)
!           IF (CTFTILT_FILE) THEN
!               CALL FILENAME_FROM_WILD(CTFTILT_FILENAME_WILD,FILM_NUM(J_START),CTFTILT_FILENAME)
!               CALL FILM_INIT(FILM1,FILM_FILENAME,PIXEL_SIZE,CTFTILT_FILENAME=CTFTILT_FILENAME)
!           ELSE
!               CALL FILENAME_FROM_WILD(CTFFIND_FILENAME_WILD,FILM_NUM(J_START),CTFFIND_FILENAME)
!               CALL FILM_INIT(FILM1,FILM_FILENAME,PIXEL_SIZE,CTFFIND_FILENAME=CTFFIND_FILENAME)
!           ENDIF

            ! Populate the waypoints
            SELECT CASE (FLX_PAR_VERSION)
                CASE (1)
                    WRITE(*,'(2A)') '**ERROR(FREALIX_PAR_TO_FILAMENTS_NOINIT): unable to deal with frealix ',    &
                                    'parameter file of version 1 yet. Sorry.'
                    CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
                CASE (2)
                    WP(1:J-J_START+1)%COO(1)        =   X(J_START:J)
                    WP(1:J-J_START+1)%COO(2)        =   Y(J_START:J)
                    WP(1:J-J_START+1)%COO(3)        =   0.0
                    WP(1:J-J_START+1)%EULER%PHI     =   PHI(J_START:J)
                    WP(1:J-J_START+1)%EULER%THETA   =   THETA(J_START:J)
                    WP(1:J-J_START+1)%EULER%PSI     =   PSI(J_START:J)
                    WP(1:J-J_START+1)%REFINE        =   REFINE(J_START:J)
                CASE (3)
                    WP(1:J-J_START+1)%COO(1)        =   X(J_START:J)
                    WP(1:J-J_START+1)%COO(2)        =   Y(J_START:J)
                    WP(1:J-J_START+1)%COO(3)        =   Z(J_START:J)
                    WP(1:J-J_START+1)%PATH_LENGTH   =   PATH_LENGTH(J_START:J)/PIXEL_SIZE
                    IF (PAR_IN_ANGSTROMS) THEN
                        WP(1:J-J_START+1)%COO(1)        =   WP(1:J-J_START+1)%COO(1)/PIXEL_SIZE +1.0    ! +1.0 because the origin is at pixel (1,1) in corner
                        WP(1:J-J_START+1)%COO(2)        =   WP(1:J-J_START+1)%COO(2)/PIXEL_SIZE +1.0
                        WP(1:J-J_START+1)%COO(3)        =   WP(1:J-J_START+1)%COO(3)/PIXEL_SIZE
                        WP(1:J-J_START+1)%PATH_LENGTH   =   WP(1:J-J_START+1)%PATH_LENGTH/PIXEL_SIZE
                    ENDIF
                    WP(1:J-J_START+1)%EULER%PHI     =   PHI(J_START:J)
                    WP(1:J-J_START+1)%EULER%THETA   =   THETA(J_START:J)
                    WP(1:J-J_START+1)%EULER%PSI     =   PSI(J_START:J)
                    WP(1:J-J_START+1)%REFINE        =   REFINE(J_START:J)
                CASE (4)
                    WP(1:J-J_START+1)%COO(1)        =   X(J_START:J)
                    WP(1:J-J_START+1)%COO(2)        =   Y(J_START:J)
                    WP(1:J-J_START+1)%COO(3)        =   Z(J_START:J)
                    WP(1:J-J_START+1)%PATH_LENGTH   =   PATH_LENGTH(J_START:J)
                    IF (PAR_IN_ANGSTROMS) THEN
                        WP(1:J-J_START+1)%COO(1)        =   WP(1:J-J_START+1)%COO(1)/PIXEL_SIZE +1.0    ! +1.0 because the origin is at pixel (1,1) in corner
                        WP(1:J-J_START+1)%COO(2)        =   WP(1:J-J_START+1)%COO(2)/PIXEL_SIZE +1.0
                        WP(1:J-J_START+1)%COO(3)        =   WP(1:J-J_START+1)%COO(3)/PIXEL_SIZE
                        WP(1:J-J_START+1)%PATH_LENGTH   =   WP(1:J-J_START+1)%PATH_LENGTH/PIXEL_SIZE
                    ENDIF
                    WP(1:J-J_START+1)%EULER%PHI     =   PHI(J_START:J)
                    WP(1:J-J_START+1)%EULER%THETA   =   THETA(J_START:J)
                    WP(1:J-J_START+1)%EULER%PSI     =   PSI(J_START:J)
                    WP(1:J-J_START+1)%REFINE        =   REFINE(J_START:J)
                    WP(1:J-J_START+1)%SUBUNIT_NUM   =   SUBUNIT_NUM(J_START:J)
                CASE DEFAULT
                    WRITE(*,'(A,I0)') '**ERROR(FREALIX_PAR_TO_FILAMENTS): unknown frealix parameter file format: ', FLX_PAR_VERSION
                    CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS')
            END SELECT

            ! Sanity check
            IF (ANY(WP(1:J-J_START+1)%COO(1) .LT. 0.0) .OR. &
                ANY(WP(1:J-J_START+1)%COO(2) .LT. 0.0) ) THEN
                WRITE(*,'(A,I0,A)') '**WARNING(FREALIX_PAR_TO_FILAMENTS_NOINIT): at least one of the waypoints '//&
                                    'read from file for filament ', FIL_NUM(J_START), ' has negative coordinates: '
                CALL WAYPOINTS_PRINT_INFO(WP(1:J-J_START+1))
                !CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
            ENDIF

            IF (DEBUG) THEN
                CALL WAYPOINTS_PRINT_INFO(WP)
            ENDIF

            ! Now that the waypoints are ready, populate a filament object
            CALL FILAMENT_FROM_WPS(FILS(I),WP,ID=FIL_NUM(J_START),FILM_ID=FILM_NUM(J_START),&
                                                PIXEL_SIZE=PIXEL_SIZE,SCORE=SCORE(J_START))


            ! Diagnostics
            IF (DEBUG) THEN
                WRITE(*,'(A)') '**debug(FILAMENTS_FROM_FREALIX_PAR_NOINIT): filament info on return:'
                CALL FILAMENT_PRINT_INFO(FILS(I))
            ENDIF

        ENDDO ! End of loop over filaments

    END SUBROUTINE FILAMENTS_FROM_FREALIX_PAR_NOINIT

    !>  \brief  Read a frealix parameter file (specified version and return an array of initialised filament objects)
    SUBROUTINE FILAMENTS_FROM_FREALIX_PAR(  FILS,PAR_FILENAME,FILM_FILENAME_WILD,PIXEL_SIZE,HELIX,              &
                                            CTFTILT_FILENAME_WILD,CTFFIND_FILENAME_WILD,                        &
                                            WPS_ARE_CROSSOVERS,FAL_WPS_ARE_CROSSOVERS,                          &
                                            MAX_RADIUS,FLX_PAR_VERSION,                                         &
                                            MICROTUBULE,MT_PF_NUMBER,MT_RISE,                                   &
                                            RIGID_PARAMS,CONSTANT_RISE_PER_SUBUNIT,CONSTANT_TWIST_PER_LENGTH,   &
                                            WPS_AS_SPS,REFINE_PARAMS,PERSISTENCE_LENGTH,                        &
                                            TORSION_PERSISTENCE_LENGTH,STRETCH_PERSISTENCE_LENGTH,              &
                                            FAL_WPS_IN_SCORE,FAL_WPS_IN_THREED,                                 &
                                            REFINE_NORMAL_MODES,FILMS_EXIST,                                    &
                                            AVERAGE_Z_TO_CTFTILT_PLANE,USE_HERMITE_SPLINES,                     &
                                            FRAGMENTED_PHI,FRAGMENTED_SUBUNIT_NUM,AMYLOID_BETA,                 &
                                            SHIFT_FILS_AND_SET_FILMS_DIM)
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_EXTRA_INFO
		USE FILAMENTS_BASE
		USE FILAMENTS, ONLY : FILAMENT_INIT
		USE FILMS, ONLY : FILM, FILM_INIT
		USE HELICES, ONLY : HELICAL_SYMMETRY
        USE PARSERS
        USE COOS, ONLY : WAYPOINTS_PRINT_INFO
        USE RUNTIME_PARAMETERS, ONLY : FILENAME_LEN, TERMINATE, IS_ODD
        IMPLICIT NONE
        ! Arguments
        TYPE(FILAMENT), ALLOCATABLE, TARGET,   INTENT(INOUT)::  FILS(:)
        CHARACTER(LEN=*),                      INTENT(IN)   ::  PAR_FILENAME                        !<  Filename for parameter file
        CHARACTER(LEN=*),                      INTENT(IN)   ::  FILM_FILENAME_WILD                  !<  Filename for the films, with ? wildcard
        REAL,                                  INTENT(IN)   ::  PIXEL_SIZE                          !<  Pixel size (in Angstroms)
        TYPE(HELICAL_SYMMETRY),                INTENT(IN)   ::  HELIX                               !<  (Average) helical parameters for the filaments.
        CHARACTER(LEN=*),          OPTIONAL,   INTENT(IN)   ::  CTFTILT_FILENAME_WILD               !<  Filename for the CTFTILT results, with ? wildcard
        CHARACTER(LEN=*),          OPTIONAL,   INTENT(IN)   ::  CTFFIND_FILENAME_WILD               !<  Filename for the CTFFIND results, with ? wildcard
        !INTEGER,                  OPTIONAL,   INTENT(IN)   ::  NUM_SUBUNITS_BTWN_WPS_MULITPLE_OF   !<  The number of subunits between successive waypoints should be a multiple of this integer number, if given. If not given, defaults to 1.
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  WPS_ARE_CROSSOVERS                  !<  Whether the way points in the parameter file define crossover positions
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  FAL_WPS_ARE_CROSSOVERS              !<  Whether the first and last waypoints of each filaments are also crossovers
        REAL,                      OPTIONAL,   INTENT(IN)   ::  MAX_RADIUS                          !<  Filament's maximum outer radius (pixels)
        INTEGER,                   OPTIONAL,   INTENT(IN)   ::  FLX_PAR_VERSION                     !<  Frealix parameter file format version. Will override whatever version the parameter file may say it is.
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  MICROTUBULE                         !<  Whether the filaments are microtubules
        INTEGER,                   OPTIONAL,   INTENT(IN)   ::  MT_PF_NUMBER                        !<  If microtubules, how many protofilaments
        INTEGER,                   OPTIONAL,   INTENT(IN)   ::  MT_RISE                             !<  Microtubule: How much further along the filament (distance measured in tubulin subunits)
                                                                                                    !!  the end of a MT_PF_NUMBER-element "washer" is compared to its beginning. The 13-pf microtobules have a value of 3.
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  RIGID_PARAMS(7)                     !<  Flags for parameters to be held rigid (i.e. per-filament rather than per-waypoint)
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  CONSTANT_RISE_PER_SUBUNIT           !<  Whether the rise per subunit is assumed to be constant
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  CONSTANT_TWIST_PER_LENGTH           !<  Whether the twist per unit length along the filament is assumed to be constant
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  WPS_AS_SPS                          !<  Whether we are treating waypoints (wps) as single particles (sps)
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  REFINE_PARAMS(7)                    !<  Which parameters will be allowed to be refined
        REAL,                      OPTIONAL,   INTENT(IN)   ::  PERSISTENCE_LENGTH                  !<  Filament persistence length (pixels)
        REAL,                      OPTIONAL,   INTENT(IN)   ::  TORSION_PERSISTENCE_LENGTH          !<  Torsional persistence length (pixels)
        REAL,                      OPTIONAL,   INTENT(IN)   ::  STRETCH_PERSISTENCE_LENGTH          !<  Stretch persistence length (pixels)
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  FAL_WPS_IN_SCORE                    !<  If false, the parts of the filament between the first and the second as well as between the penultimate
                                                                                                    !!  and last waypoints will not be included in the scoring function
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  FAL_WPS_IN_THREED                   !<  If false, the parts of the filament between the first and the second as well as between the penultimate
                                                                                                    !!  and last waypoints will not be included in the 3D reconstruction
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  REFINE_NORMAL_MODES(7)              !<  Whether normal mode amplitudes should be refined, rather than the filament model directly
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  FILMS_EXIST                         !<  In some cases, we want to initialise filaments without having access to the micrograph images
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  AVERAGE_Z_TO_CTFTILT_PLANE          !<  Shift the filament in Z so that its average lies on the plane described by the CTFTILT plane
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  USE_HERMITE_SPLINES                 !<  Use Hermite splines (where psi, theta and therefore the tangent at waypoints constrain the splines)
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  FRAGMENTED_PHI                      !<  Phi function is broken between waypoints, but the axial space curve is still continuous
        LOGICAL,                   OPTIONAL,   INTENT(IN)   ::  FRAGMENTED_SUBUNIT_NUM              !<  Subunit_num function is broken between waypoints, but the axial space curve is still continuous
        LOGICAL,                    OPTIONAL,   INTENT(IN)  ::  AMYLOID_BETA                        !<  The filaments are amyloid beta fibrils
        LOGICAL,                    OPTIONAL,   INTENT(IN)  ::  SHIFT_FILS_AND_SET_FILMS_DIM        !<  Shift filaments and set film dimensions (useful when films not available, e.g. when making synthetic micrographs)
        ! Private variables
        LOGICAL                     ::  CTFTILT_FILE
        LOGICAL                     ::  CCROSSOVERS
        TYPE(WAYPOINT), ALLOCATABLE ::  WP(:)
        INTEGER                     ::  I,J,CURRENT_WP
        CHARACTER(LEN=FILENAME_LEN) ::  FILM_FILENAME, CTFTILT_FILENAME, CTFFIND_FILENAME
        TYPE(FILM)                  ::  FILM1
        LOGICAL                     ::  FFILMS_EXIST, SSHIFT_FILS_AND_SET_FILMS_DIM
        REAL(KIND=8),   PARAMETER   ::  MICROGRAPH_MARGIN = 600 ! In Angstroms
        LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        IF (PRESENT(FLX_PAR_VERSION)) THEN
            WRITE(*,'(A)') '**ERROR(FILAMENTS_FROM_FREALIX_PAR): optional dummy argument FLX_PAR_VERSION not implemented yet'
            CALL TERMINATE('Feature not implemented yet')
        ENDIF

        !  Read in the parameter file and populate the filament objects without initialising them, and without film data
        CALL FILAMENTS_FROM_FREALIX_PAR_NOINIT(FILS,PAR_FILENAME,PIXEL_SIZE=PIXEL_SIZE)

        ! Check the user gave us either a CTFFIND or a CTFTILT results file
        IF (.NOT. PRESENT(CTFTILT_FILENAME_WILD) .AND. .NOT. PRESENT(CTFFIND_FILENAME_WILD)) THEN
            WRITE(*,'(A)') '**ERROR(FREALIX_PAR_TO_FILAMENTS): we need either CTFFIND or CTFTILT results for each film'
            CALL TERMINATE('Fatal error in FREALIX_PAR_TO_FILAMENTS')
        ELSE IF (PRESENT(CTFTILT_FILENAME_WILD) .AND. PRESENT(CTFFIND_FILENAME_WILD)) THEN
            WRITE(*,'(2A)') '**INFO(FREALIX_PAR_TO_FILAMENTS): both CTFFIND and CTFTILT were given - ',  &
                            'CTFTILT will be used, CTFFIND ignored'
            CTFTILT_FILE = .TRUE.
        ELSE IF (PRESENT(CTFTILT_FILENAME_WILD)) THEN
            CTFTILT_FILE = .TRUE.
        ELSE
            CTFTILT_FILE = .FALSE.
        ENDIF

        ! Crossovers?
        CCROSSOVERS = .FALSE.
        IF (PRESENT(WPS_ARE_CROSSOVERS)) CCROSSOVERS = WPS_ARE_CROSSOVERS

        ! Films exist?
        FFILMS_EXIST = .TRUE.
        IF (PRESENT(FILMS_EXIST)) FFILMS_EXIST = FILMS_EXIST

        SSHIFT_FILS_AND_SET_FILMS_DIM = .FALSE.
        IF (PRESENT(SHIFT_FILS_AND_SET_FILMS_DIM)) SSHIFT_FILS_AND_SET_FILMS_DIM = SHIFT_FILS_AND_SET_FILMS_DIM

        ! Loop over the filaments
        CURRENT_WP = 1
        J = 0   ! This is the current line/waypoint
        DO I=1,SIZE(FILS)

            ! Set up the film - we can do this by parsing either CTFTILT or CTFFIND
            ! If the user has given us a CTFTILT file, we use that.
            IF (FILENAME_CONTAINS_QMARKS(FILM_FILENAME_WILD)) THEN
                CALL FILENAME_FROM_WILD(FILM_FILENAME_WILD,FILS(I)%FILM%ID,FILM_FILENAME)
                FILS(I)%FILM%location_in_file = 1
            ELSE
                FILM_FILENAME = FILM_FILENAME_WILD
                FILS(I)%FILM%location_in_file = FILS(I)%FILM%id
            ENDIF
            IF (CTFTILT_FILE) THEN
                CALL FILENAME_FROM_WILD(CTFTILT_FILENAME_WILD,FILS(I)%FILM%ID,CTFTILT_FILENAME)
                CALL FILM_INIT(FILM1,FILS(I)%FILM%ID,FILM_FILENAME,PIXEL_SIZE,  &
                                CTFTILT_FILENAME=CTFTILT_FILENAME, &
                                FILM_IMAGE_AVAILABLE=FFILMS_EXIST, &
                                LOCATION_WITHIN_FILE=FILS(I)%FILM%location_in_file)
            ELSE
                CALL FILENAME_FROM_WILD(CTFFIND_FILENAME_WILD,FILS(I)%FILM%ID,CTFFIND_FILENAME)
                CALL FILM_INIT(FILM1,FILS(I)%FILM%ID,FILM_FILENAME,PIXEL_SIZE,  &
                                CTFFIND_FILENAME=CTFFIND_FILENAME,  &
                                FILM_IMAGE_AVAILABLE=FFILMS_EXIST, &
                                LOCATION_WITHIN_FILE=FILS(I)%FILM%location_in_file)
            ENDIF

            IF (DEBUG) THEN
                WRITE(*,'(A)') '**debug(FILAMENTS_FROM_FREALIX_PAR): filament info before init'
                CALL FILAMENT_PRINT_INFO(FILS(I))
            ENDIF

            IF (FFILMS_EXIST .OR. SSHIFT_FILS_AND_SET_FILMS_DIM) THEN
                ! Copy the waypoint info into a separate waypoints array
                IF (ALLOCATED(WP)) DEALLOCATE(WP)
                ALLOCATE(WP(SIZE(FILS(I)%WP)))
                WP = FILS(I)%WP

                ! Move the filament so its min x,y are near the origin, then set the micrograph
                ! dimensions to be so that the filament just fits in (with additional margin)
                IF (SSHIFT_FILS_AND_SET_FILMS_DIM) THEN
                    WP%COO(1) = WP%COO(1) - MINVAL(WP%COO(1)) + MICROGRAPH_MARGIN/PIXEL_SIZE
                    WP%COO(2) = WP%COO(2) - MINVAL(WP%COO(2)) + MICROGRAPH_MARGIN/PIXEL_SIZE
                    FILM1%logical_dimensions(1) = MAXVAL(WP%COO(1)) + MICROGRAPH_MARGIN/PIXEL_SIZE
                    FILM1%logical_dimensions(2) = MAXVAL(WP%COO(2)) + MICROGRAPH_MARGIN/PIXEL_SIZE
                    IF (IS_ODD(FILM1%logical_dimensions(1))) FILM1%logical_dimensions(1) = FILM1%logical_dimensions(1) + 1
                    IF (IS_ODD(FILM1%logical_dimensions(2))) FILM1%logical_dimensions(2) = FILM1%logical_dimensions(2) + 1
                    FILM1%logical_dimensions(1:2) = MAXVAL(FILM1%logical_dimensions)! We have to do this temporarily to make it square, to accomodate for crappy restriction on IMAGE_APPLY_CTF
                ENDIF

                IF (DEBUG) THEN
                    WRITE(*,'(A)') '**debug(FILAMENTS_FROM_FREALIX_PAR): waypoints to be used by filament_init:'
                    CALL WAYPOINTS_PRINT_INFO(WP)
                ENDIF


                ! Initialise a filament object
                CALL FILAMENT_INIT( FILS(I),WAYPOINTS_IN=WP,FILM_IN=FILM1,HELIX=HELIX,CROSSOVERS=CCROSSOVERS, &
                                FAL_WPS_ARE_CROSSOVERS=FAL_WPS_ARE_CROSSOVERS,&
                                ID=FILS(I)%ID,MAX_RADIUS=MAX_RADIUS,    &
                                MICROTUBULE=MICROTUBULE,MT_PF_NUMBER=MT_PF_NUMBER,MT_RISE=MT_RISE,RIGID_PARAMS=RIGID_PARAMS,    &
                                CONSTANT_RISE_PER_SUBUNIT=CONSTANT_RISE_PER_SUBUNIT,    &
                                CONSTANT_TWIST_PER_LENGTH=CONSTANT_TWIST_PER_LENGTH,    &
                                WPS_AS_SPS=WPS_AS_SPS,REFINE_PARAMS=REFINE_PARAMS,PERSISTENCE_LENGTH=PERSISTENCE_LENGTH,    &
                                TORSION_PERSISTENCE_LENGTH=TORSION_PERSISTENCE_LENGTH,  &
                                STRETCH_PERSISTENCE_LENGTH=STRETCH_PERSISTENCE_LENGTH,  &
                                FAL_WPS_IN_SCORE=FAL_WPS_IN_SCORE,FAL_WPS_IN_THREED=FAL_WPS_IN_THREED,  &
                                REFINE_NORMAL_MODES=REFINE_NORMAL_MODES,    &
                                AVERAGE_Z_TO_CTFTILT_PLANE=AVERAGE_Z_TO_CTFTILT_PLANE,USE_HERMITE_SPLINES=USE_HERMITE_SPLINES, &
                                FRAGMENTED_PHI=FRAGMENTED_PHI,FRAGMENTED_SUBUNIT_NUM=FRAGMENTED_SUBUNIT_NUM,    &
                                AMYLOID_BETA=AMYLOID_BETA)
            ELSE
                WRITE(*,'(A)') '**debug(FILAMENTS_FROM_FREALIX_PAR): no film, filament not initialised'
            ENDIF

            ! Diagnostics: print filament info
            WRITE(*,'(A)') '**info(FILAMENTS_FROM_FREALIX_PAR): filament info on return:'
            CALL FILAMENT_PRINT_INFO(FILS(I))
            CALL FILAMENT_PRINT_EXTRA_INFO(FILS(I))

        ENDDO

        WRITE(*,'(A)') '**info(FILAMENTS_FROM_FREALIX_PAR): all done'

    END SUBROUTINE FILAMENTS_FROM_FREALIX_PAR


    !>  \brief  Parse a Frealix1 parameter file and return an array of filament objects
    !!  Either CTFFIND or CTFTILT results must be available, for all films
    FUNCTION FREALIX1_PAR_TO_FILAMENTS( FILENAME,FILM_FILENAME_WILD,PIXEL_SIZE,HELIX,   &
                                        CTFTILT_FILENAME_WILD,CTFFIND_FILENAME_WILD,CROSSOVERS) RESULT(FILS)
        USE FILAMENTS_BASE
        USE FILAMENTS, ONLY : FILAMENT_INIT
        USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
        USE PARSERS
        USE RUNTIME_PARAMETERS, ONLY : FILENAME_LEN, TERMINATE
        USE FILMS, ONLY : FILM, FILM_INIT
        IMPLICIT NONE
        ! Arguments
        CHARACTER(LEN=*),               INTENT(IN)  ::  FILENAME                            !<  Filename of frealix1 parameter file
        CHARACTER(LEN=*),               INTENT(IN)  ::  FILM_FILENAME_WILD                  !<  Filename for the films, with ? wildcard
        REAL,                           INTENT(IN)  ::  PIXEL_SIZE                          !<  Pixel size (in Angstroms)
        TYPE(HELICAL_SYMMETRY),         INTENT(IN)  ::  HELIX                               !<  (Average) helical parameters for the filaments.
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)  ::  CTFTILT_FILENAME_WILD               !<  Filename for the CTFTILT results, with ? wildcard
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)  ::  CTFFIND_FILENAME_WILD               !<  Filename for the CTFFIND results, with ? wildcard
        !INTEGER,           OPTIONAL,   INTENT(IN)  ::  NUM_SUBUNITS_BTWN_WPS_MULITPLE_OF   !<  The number of subunits between successive waypoints should be a multiple of this integer number, if given. If not given, defaults to 1.
        LOGICAL,            OPTIONAL,   INTENT(IN)  ::  CROSSOVERS                          !<  Whether the way points in the parameter file describe crossovers

        ! Result
        TYPE(FILAMENT), ALLOCATABLE ::  FILS(:) !<  Output array of filaments
        ! Private variables
        INTEGER,        ALLOCATABLE ::  FIL_NUMS(:)
        INTEGER                     ::  NUM_OF_FILS, NUM_OF_WPS, I, J, J_START, CURRENT_WP, IERR
        REAL,           ALLOCATABLE ::  TMP(:)
        INTEGER,        ALLOCATABLE ::  ITMP(:), IFIL(:), FILM_NUM(:)
        TYPE(WAYPOINT), ALLOCATABLE ::  WP(:)
        REAL,           ALLOCATABLE ::  PHI(:), THETA(:), PSI(:), SHX(:), SHY(:)
        CHARACTER(LEN=FILENAME_LEN) ::  FILM_FILENAME, CTFTILT_FILENAME, CTFFIND_FILENAME
        LOGICAL                     ::  CTFTILT_FILE
        TYPE(FILM)                  ::  FILM1
        !INTEGER                        ::  NNUM_SUBUNITS_BTWN_WPS_MULITPLE_OF
        LOGICAL                     ::  CCROSSOVERS
        ! Start work

        ! Check the parameter file exists
        IF (.NOT. FILE_EXISTS(FILENAME)) THEN
            WRITE(*,'(A,A)') '**ERROR(FREALIX1_PAR_TO_FIL): file does not exist: ', TRIM(FILENAME)
            STOP 'Fatal error in FREALIX1_PAR_TO_FIL'
        ENDIF

        ! Check the user gave us either a CTFTILT or a CTFTILT results file
        IF (.NOT. PRESENT(CTFTILT_FILENAME_WILD) .AND. .NOT. PRESENT(CTFFIND_FILENAME_WILD)) THEN
            WRITE(*,'(A)') '**ERROR(FREALIX1_PAR_TO_FIL): we need either CTFFIND or CTFTILT results for each film'
            STOP 'Fatal error in FREALIX1_PAR_TO_FIL'
        ELSE IF (PRESENT(CTFTILT_FILENAME_WILD) .AND. PRESENT(CTFFIND_FILENAME_WILD)) THEN
            WRITE(*,'(2A)') '**INFO(FREALIX1_PAR_TO_FIL): both CTFFIND and CTFTILT were given - ',   &
                            'CTFTILT will be used, CTFFIND ignored'
            CTFTILT_FILE = .TRUE.
        ELSE IF (PRESENT(CTFTILT_FILENAME_WILD)) THEN
            CTFTILT_FILE = .TRUE.
        ELSE
            CTFTILT_FILE = .FALSE.
        ENDIF

        !! NUM_SUBUNITS_BTWN_WPS_MULITPLE_OF
        !IF (PRESENT(NUM_SUBUNITS_BTWN_WPS_MULITPLE_OF)) THEN
        !   NNUM_SUBUNITS_BTWN_WPS_MULITPLE_OF = NUM_SUBUNITS_BTWN_WPS_MULITPLE_OF
        !ELSE
        !   NNUM_SUBUNITS_BTWN_WPS_MULITPLE_OF = 1
        !ENDIF

        IF (PRESENT(CROSSOVERS)) THEN
            CCROSSOVERS = CROSSOVERS
        ELSE
            CCROSSOVERS = .FALSE.
        ENDIF

        ! Work out number of filaments
        CALL GET_INTS_FROM_COLUMN(FILENAME,16,FIL_NUMS)
        NUM_OF_FILS = MAXVAL(FIL_NUMS)

        ! Work out number of waypoints
        NUM_OF_WPS  =   NUMBER_OF_LINES_IN_FILE(FILENAME,IGNORE_COMMENTS=.TRUE.)

        ! Allocate the output array and temporary arrays
        IF (ALLOCATED(FILS)) DEALLOCATE(FILS)
        ALLOCATE(FILS(NUM_OF_FILS),TMP(NUM_OF_WPS),ITMP(NUM_OF_WPS),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(FREALIX1_PAR_TO_FIL): memory allocation failed with error ', IERR
            CALL TERMINATE('Fatal error in FREALIX1_PAR_TO_FIL')
        ENDIF

        ! Read in the columns that we'll need
        CALL GET_REALS_FROM_COLUMN(FILENAME,2,PSI)
        CALL GET_REALS_FROM_COLUMN(FILENAME,3,THETA)
        CALL GET_REALS_FROM_COLUMN(FILENAME,4,PHI)
        CALL GET_REALS_FROM_COLUMN(FILENAME,5,SHX)
        CALL GET_REALS_FROM_COLUMN(FILENAME,6,SHY)
        CALL GET_INTS_FROM_COLUMN(FILENAME,8,FILM_NUM)
        CALL GET_INTS_FROM_COLUMN(FILENAME,16,IFIL)

        IF (SIZE(IFIL) .NE. NUM_OF_WPS) THEN
            WRITE(*,'(A,I6,A,I6,A)') '**ERROR(FREALIX1_PAR_TO_FIL): oops, number of lines (', NUM_OF_WPS ,&
                                    ') not equal to number of IFIL values read in (',SIZE(IFIL),')'
            CALL TERMINATE('Fatal error in FREALIX1_PAR_TO_FIL')
        ENDIF

        ! Loop over the filaments
        CURRENT_WP = 1
        J = 0   ! This is the current line/waypoint
        DO I=1,NUM_OF_FILS
            ! At the beginning of the iteration, J points to the last waypoint of the previous filament
            J = J+1
            J_START = J

            ! For the current filament, work out how many waypoints are defined in the parameter file
            ! If we're on the last waypoint, of course we don't need to do this
            IF (J .LT. NUM_OF_WPS) THEN
                ! Let's increment J until we reach the last waypoint in the current filament, I
                DO
                    IF (IFIL(J+1) .EQ. IFIL(J)) THEN
                        ! We are not yet at last way point of current filament
                        J = J +1
                    ELSE
                        ! We are at last waypoint of current filament
                        EXIT
                    ENDIF
                ENDDO
            ELSE
                ! We're on the last waypoint. This is strange, cos it implies the last way point is "alone"
                WRITE(*,'(A)') '**ERROR(FREALIX1_PAR_TO_FIL): last way point appears to be alone'
                STOP 'Fatal error in FREALIX1_PAR_TO_FIL'
            ENDIF

            ! So now J is set to the position of the last way point of the current filament

            ! Diagnostic message
            PRINT '(A,I3,A,I2,A)', '**debug(FREALIX1_PAR_TO_FIL): filament ', IFIL(J), ' has ', J-J_START+1, ' way points'

            ! Allocate memory to an array of way points
            IF (ALLOCATED(WP)) DEALLOCATE(WP)
            ALLOCATE(WP(J-J_START+1),STAT=IERR)
            IF (IERR .NE. 0) THEN
                WRITE(*,'(A,I0)') '**ERROR(FREALIX1_PAR_TO_FIL): memory allocation of temporary waypoint array failed with err ', &
                                    IERR
                CALL TERMINATE('Fatal error in FREALIX1_PAR_TO_FIL')
            ENDIF

            ! Populate the waypoints
            WP(1:J-J_START+1)%COO(1)        =   SHX(J_START:J)
            WP(1:J-J_START+1)%COO(2)        =   SHY(J_START:J)
            WP(1:J-J_START+1)%EULER%PHI     =   PHI(J_START:J)
            WP(1:J-J_START+1)%EULER%THETA   =   THETA(J_START:J)
            WP(1:J-J_START+1)%EULER%PSI     =   PSI(J_START:J)

            ! Set up the film - we can do this by parsing either CTFTILT or CTFFIND
            ! If the user has given us a CTFTILT file, we use that.
            CALL FILENAME_FROM_WILD(FILM_FILENAME_WILD,FILM_NUM(J_START),FILM_FILENAME)
            IF (CTFTILT_FILE) THEN
                CALL FILENAME_FROM_WILD(CTFTILT_FILENAME_WILD,FILM_NUM(J_START),CTFTILT_FILENAME)
                CALL FILM_INIT(FILM1,FILM_NUM(J_START),FILM_FILENAME,PIXEL_SIZE,CTFTILT_FILENAME=CTFTILT_FILENAME)
            ELSE
                CALL FILENAME_FROM_WILD(CTFFIND_FILENAME_WILD,FILM_NUM(J_START),CTFFIND_FILENAME)
                CALL FILM_INIT(FILM1,FILM_NUM(J_START),FILM_FILENAME,PIXEL_SIZE,CTFFIND_FILENAME=CTFFIND_FILENAME)
            ENDIF

            ! Initialise a filament object
            CALL FILAMENT_INIT(FILS(I),WAYPOINTS_IN=WP,FILM_IN=FILM1,HELIX=HELIX,CROSSOVERS=CCROSSOVERS)

            ! Diagnostics: print filament info
            CALL FILAMENT_PRINT_INFO(FILS(I))

        ENDDO

    END FUNCTION FREALIX1_PAR_TO_FILAMENTS

END MODULE FREALIX_PARAMETERS
