MODULE FRSCS

CONTAINS
!>  \brief  Compute the Filtered Real-Space Correlation between two images over a mask.
SUBROUTINE FRSC(IMG1,IMG2,MASK)
    USE IMAGES
    USE IMAGES_COMPARISON
    IMPLICIT NONE
    ! Arguments
    TYPE(IMAGE),                INTENT(INOUT)   ::  IMG1
    TYPE(IMAGE),                INTENT(INOUT)   ::  IMG2
    TYPE(IMAGE),    OPTIONAL,   INTENT(IN)      ::  MASK    !<  Correlation will be calculated between pixels where mask .ne. 0.0. If not supplied, assumed 1.0 everywhere.
    ! Private variables
    TYPE(IMAGE) ::  MMASK
    TYPE(IMAGE) ::  IMG1_FIL, IMG2_FIL
    TYPE(IMAGE) ::  FILTER
    REAL        ::  TMP
    INTEGER     ::  RAD
    REAL,   ALLOCATABLE ::  CORR(:)
    ! Start work

    ! Check dimensions
    IF (.NOT. IMG1%HasSameDimsAs(IMG2)) THEN
        WRITE(*,'(A,3I4,A,3I4)') '**ERROR(FRSC): the two images do not have same dimensions: ', &
                                IMG1%logical_dimensions, '&', IMG2%logical_dimensions
        CALL TERMINATE('Fatal error in FRSC')
    ENDIF
    ! Check images are square
    IF (.NOT. IMG1%IsSquare()) THEN
        WRITE(*,'(A,4I4)') '**ERROR(FRSC): the images are not square/cube: ', IMG1%logical_dimensions
        CALL TERMINATE('Fatal error in FRSC')
    ENDIF
    ! Check type
    IF (.not. img1%is_in_real_space .OR. .not. img2%is_in_real_space) THEN
        CALL TERMINATE('Fatal error in FRSC')
    ENDIF
    ! If no mask supplied, we assume the calculation should include all voxels
    CALL IMAGE_ALLOCATE(MMASK,IMG1)
    IF (PRESENT(MASK)) THEN
        IF (.NOT. IMG1%HasSameDimsAs(MASK)) THEN
            WRITE(*,'(A,3I4,A,3I4)') '**ERROR(FRSC): the mask does not have same dimensions as the images: ', &
                        IMG1%logical_dimensions, '&', MASK%logical_dimensions
            CALL TERMINATE('Fatal error in FRSC')
        ENDIF
        MMASK%real_values = MASK%real_values
    ELSE
        MMASK%real_values = 1.0
    ENDIF

    ! Allocate memory for the filter image
    CALL IMAGE_ALLOCATE(FILTER,IMG1)
    FILTER%ORIGIN_AT_CENTER = .FALSE.

    ! Print out info regarding the mask
    TMP = COUNT(MMASK%real_values(1:IMG1%logical_dimensions(1),1:IMG1%logical_dimensions(2),1:IMG1%logical_dimensions(3)) &
                .NE. 0.0)*100/ &
                (IMG1%logical_dimensions(1)*IMG1%logical_dimensions(2)*IMG1%logical_dimensions(3))
    WRITE(*,'(A,F6.2,A)') '**debug(FRSC): ', TMP, '% of voxels will be used in correlation calculation'
    ! Correlation without filtering
    WRITE(*,'(A,F8.5)') '**debug(FRSC): the correlation coefficient without filtering is ', CORRELATION(IMG1,IMG2,MMASK,.TRUE.)

    ! Allocate memory for the result
    ALLOCATE(CORR(0:IMG1%logical_dimensions(1)/2-1))

    ! Filter
    DO RAD=0,(IMG1%logical_dimensions(1)/2)-1
        ! Generate a disk for the filter
        CALL HARD_DISC(FILTER,REAL(RAD))
        ! Apply the filter
        IMG1_FIL = IMG1
        IMG2_FIL = IMG2
        CALL FWD_FT(IMG1_FIL)
        CALL FWD_FT(IMG2_FIL)
        CALL APPLY_FILTER_IMG(IMG1_FIL,FILTER)
        CALL APPLY_FILTER_IMG(IMG2_FIL,FILTER)
        CALL BWD_FT(IMG1_FIL)
        CALL BWD_FT(IMG2_FIL)
        ! Calculate real-space normalised correlation over the mask
        CORR(RAD) = CORRELATION(IMG1_FIL,IMG2_FIL,MMASK,.TRUE.)
        WRITE(*,'(A,I3,A,F8.5)') '**debug(FRSC): FRSC(',RAD,') = ', CORR(RAD)
!       IF (RAD == IMG1%logical_dimensions(1)/4) THEN
!           WRITE(*,*) '**debug(FRSC): dumping files for RAD = ', RAD
!           CALL WRITE_TO_DISK(IMG1_FIL,'junk1')
!           CALL WRITE_TO_DISK(IMG2_FIL,'junk2')
!           CALL WRITE_TO_DISK(FILTER,'junk3')
!       ENDIF
    ENDDO
END SUBROUTINE FRSC
END MODULE FRSCS
