!>  \brief  Classes and methods to deal with helical symmetries, helical objects etc.
!!
MODULE HELICES

    IMPLICIT NONE
    PRIVATE
    PUBLIC  ::  HELICAL_SYMMETRY, HELICAL_SYMMETRY_INIT, HELICAL_SYMMETRY_PRINT_INFO, HELIX_PRINT_INFO,HELICAL_SYMMETRY_UNIT_CONVERT
    PUBLIC  ::  HELIX_TO_ONE_START,HELIX_PITCH, HELIX_GUESS_SUBUNIT_HEIGHT
    PUBLIC  ::  HELIX_PHI_FROM_XY, HELIX_XY_FROM_PHI_R
    PUBLIC  ::  HELICAL_SYMMETRY_RESET, HELICAL_SYMMETRY_PRINT_EXTRA_INFO
    PUBLIC  ::  HELIX_UNIT_TESTS
    PUBLIC	::	HELIX_POINT_IS_IN_ASYM, HELIX_POINT_TO_ASYM
    PUBLIC  ::  HELIX_FIND_LAYER_LINES
	PUBLIC  ::  OPERATOR(+), OPERATOR(-), OPERATOR(*), OPERATOR(/), SQRT


    ! Tolerances for the boundaries of the asymmetric unit
    REAL(KIND=8),   PARAMETER   ::  Z_TOL = EPSILON(0.0D0) * 10.0D0
    REAL(KIND=8),   PARAMETER   ::  PHI_TOL = EPSILON(0.0D0) * 10.0D0

    !>  \brief  Describes helical lattice parameters
    !!          For an explanation of how helical objects can be described as rolled-up 2D crystal lattices, see
    !!          Chapter 12 "Electron crystallography of helical structures" by David DeRosier, in
    !!          "Electron crystallography of biological macromolecules" (Eds. Glaeser, Downing, DeRosier, Chiu & Frank)
    !!
    TYPE    ::  HELICAL_LATTICE
        PRIVATE
        LOGICAL ::  PIXEL_UNITS = .FALSE.   !<  Whether the lattice parameters are currently in pixel units
        REAL    ::  A                       !<  Unit cell dimension 1
        REAL    ::  B                       !<  Unit cell dimension 2
        REAL    ::  ANGLE                   !<  Unit cell angle
        INTEGER ::  CIRCUM_VECTOR(2)        !<  Circumferential vector
    END TYPE

    !>  \brief  Describes a helical symmetry
    !!          Again, for explanations, I recommend the DeRosier chapter in that book
    TYPE    ::  HELICAL_SYMMETRY
        LOGICAL         ::  PIXEL_UNITS             =   .FALSE.     !<  Whether the rise parameter is in pixels or Angstroms
        REAL(KIND=8)    ::  RISE                    =   0.0D0       !<  The rise (in pixels/Angstroms), i.e. the translation along the helical axis to get from one subunit to the next
        REAL(KIND=8)    ::  TWIST                   =   0.0D0       !<  The twist (in degrees), i.e. the rotation around the helical axis to get from one subunit to the next (positive angle implies left-handed helix)
        INTEGER         ::  AXIAL_SYMMETRY          =   1           !<  Axial symmetry. If AXIAL_SYMMETRY = N, the helix is N-start.
        INTEGER         ::  PERPENDICULAR_SYMMETRY  =   1           !<  Symmetry with axis perpendicular to helix axis. This can either be 1 or 2. 2 means the helix is not polar, like DNA for example.
        INTEGER         ::  NUM_STARTS              =   1           !<  Indicate which start number the helical rise and twist refer to. It may make more sense or be more intuitive to describe helices not with the 1-start helix in some situations.
        REAL(KIND=8)    ::  ASYM_UNIT_WIDTH         =   0.0D0       !<  The "width" of the asymmetric unit, in degrees
        REAL(KIND=8)    ::  ASYM_UNIT_HEIGHT        =   0.0D0       !<  The "height" of the asymmetric unit, in pixels
        LOGICAL         ::  STACK                   =   .FALSE.     !<  The helical assembly is a stack rather than a true helix
        ! Description of the equivalent helix with the smallest twist
        INTEGER         ::  START_MIN_TWIST         =   0           !<  The start number with the smallest twist
        REAL(KIND=8)    ::  RISE_MIN_TWIST          =   0.0D0
        REAL(KIND=8)    ::  TWIST_MIN               =   0.0D0
        !
        INTEGER         ::  START_MIN_RISE          =   0
        REAL(KIND=8)    ::  RISE_MIN                =   0.0D0
        REAL(KIND=8)    ::  TWIST_MIN_RISE          =   0.0D0
        !
        REAL(KIND=8)    ::  CONV_MAT(2,2)           =   0.0D0       !<  Matrix to convert (phi,z) pairs into (X,Y), the distances to the fastest and 1-start helices, in units of asymetric units. See notes of 130618 and their edit on 130620/1.
    END TYPE

    INTERFACE OPERATOR (+)
        MODULE PROCEDURE HELIX_PLUS_HELIX
    END INTERFACE

    INTERFACE OPERATOR (-)
        MODULE PROCEDURE HELIX_MINUS_HELIX
    END INTERFACE

    INTERFACE OPERATOR (*)
        MODULE PROCEDURE HELIX_TIMES_HELIX
    END INTERFACE

    INTERFACE OPERATOR (/)
        MODULE PROCEDURE HELIX_DIVIDE_BY_REAL
    END INTERFACE

    INTERFACE SQRT
        MODULE PROCEDURE HELIX_SQRT
    END INTERFACE

    INTERFACE HELIX_PRINT_INFO
        MODULE PROCEDURE HELICAL_SYMMETRY_PRINT_INFO
    END INTERFACE

    INTERFACE HELIX_PRINT_EXTRA_INFO
        MODULE PROCEDURE HELICAL_SYMMETRY_PRINT_EXTRA_INFO
    END INTERFACE

    CONTAINS

        SUBROUTINE HELIX_UNIT_TESTS()
            IMPLICIT NONE
            !
            CALL HELIX_UNIT_TEST_2()
            CALL HELIX_UNIT_TEST_1()
        END SUBROUTINE HELIX_UNIT_TESTS

        SUBROUTINE HELIX_UNIT_TEST_1()
			USE ROTATION_MATRICES, ONLY : RADIANS
			USE RUNTIME_PARAMETERS, ONLY : TERMINATE
            IMPLICIT NONE
            ! Private variables
            TYPE(HELICAL_SYMMETRY), ALLOCATABLE  ::  SELF(:)
            INTEGER ::  I, J
            INTEGER ::  A,B,C
            INTEGER, PARAMETER :: VOLDIM = 128
            INTEGER, PARAMETER :: ORI(3) = [51,51,51]
            REAL(KIND=8)    ::  X,Y,Z, Z_NEW
            REAL(KIND=8)    ::  PHI, PHI_NEW
            ! Start work
            WRITE(*,'(A)') '**info(HELIX_UNIT_TEST_1): entering'
            ALLOCATE(SELF(4))
            ! Let's pretend we're dealing with TMV
            CALL HELICAL_SYMMETRY_INIT(SELF(1),RISE=1.403,TWIST=-22.03,PIXEL_SIZE=1.163,NUM_STARTS=1,STACK=.FALSE.)
            ! Let's pretend we're dealing with Abeta (true helix)
            CALL HELICAL_SYMMETRY_INIT(SELF(2),RISE=2.4,TWIST=180.385,PIXEL_SIZE=1.163,NUM_STARTS=1,STACK=.FALSE.)
            CALL HELICAL_SYMMETRY_INIT(SELF(3),RISE=2.4,TWIST=-179.615,PIXEL_SIZE=1.163,NUM_STARTS=1,STACK=.FALSE.)
            ! Let's pretend we're dealing with Abeta (stack)
            CALL HELICAL_SYMMETRY_INIT(SELF(4),RISE=4.8,TWIST=0.77,PIXEL_SIZE=1.163,NUM_STARTS=1,STACK=.FALSE.)
            DO J=1,SIZE(SELF)
                WRITE(*,'(//A,I0,A,I0)') '**info(HELIX_UNIT_TEST_1): testing helical symmetry ', J , ' of ', SIZE(SELF)
                CALL HELIX_PRINT_INFO(SELF(J))
                CALL HELIX_PRINT_EXTRA_INFO(SELF(J))
                ! Print out the conversion matrix
                WRITE(*,'(A)') '**info(HELIX_UNIT_TEST_1): conversion matrix'
                PRINT '(/2(F15.6,1X) )', SELF(J)%CONV_MAT(1,1), SELF(J)%CONV_MAT(1,2)
                PRINT '( 2(F15.6,1X)/)', SELF(J)%CONV_MAT(2,1), SELF(J)%CONV_MAT(2,2)
                WRITE(*,'(A)') '**info(HELIX_UNIT_TEST_1): all distance values below should be integer'
                ! Loop over subunits
                DO I=1,40
                    Z   = REAL(I-1,KIND=8)*SELF(J)%RISE
                    PHI = REAL(I-1,KIND=8)*SELF(J)%TWIST
                    PHI = MOD(PHI,360.0D0)+360.0D0
                    PHI = HELIX_PHI_TO_C1(SELF(J),PHI)
                    CALL HELIX_POINT_DIST_TO_ORIGIN(SELF(J),Z,PHI,Y,X)
                    PRINT '(I3,1X,4(F11.3,1X))', I-1, Z, PHI, Y, X
                    IF (ABS(Y-NINT(Y)) .GT. 0.001 .OR. ABS(X-NINT(X)) .GT. 0.001) THEN
                        WRITE(*,'(A)') '**ERROR(HELIX_UNIT_TEST_1): bad distance values - should be integer'
                        CALL TERMINATE('Failed HELIX_UNIT_TEST_1')
                    ENDIF
                ENDDO
                WRITE(*,'(A)') '**info(HELIX_UNIT_TEST_1): checking helix_point_to_asym'
                ! Checking that we can loop through a volume and do symmetrisation
                DO C=1,VOLDIM
                    Z = REAL(C-ORI(3),KIND=8)
                    DO B=1,VOLDIM
                        Y = REAL(B-ORI(2),KIND=8)
                        DO A=1,VOLDIM
                            X = REAL(A-ORI(1),KIND=8)
                            !
                            PHI = HELIX_PHI_FROM_XY(X,Y)
                            CALL HELIX_POINT_TO_ASYM(SELF(J),Z,PHI,Z_NEW,PHI_NEW)
                        ENDDO
                    ENDDO
                ENDDO
            ENDDO   ! End of loop over symmetries

            WRITE(*,'(A)') '**info(HELIX_UNIT_TEST_1): all done'
        END SUBROUTINE HELIX_UNIT_TEST_1


        SUBROUTINE HELIX_UNIT_TEST_2()
            IMPLICIT NONE
            ! Private variables
            TYPE(HELICAL_SYMMETRY)  ::  SELF
            INTEGER,    ALLOCATABLE ::  N(:)
            REAL,       ALLOCATABLE ::  Z(:)
            INTEGER                 ::  I
            REAL,       PARAMETER   ::  PIXEL_SIZE = 1.163
            ! Start work
            WRITE(*,'(A)') '**info(HELIX_UNIT_TEST_2): entering'
            CALL HELICAL_SYMMETRY_INIT(SELF,RISE=1.403,TWIST=-22.03,PIXEL_SIZE=PIXEL_SIZE,NUM_STARTS=1,STACK=.FALSE.)
            !CALL HELICAL_SYMMETRY_INIT(SELF,RISE=2.4,TWIST=179.7,PIXEL_SIZE=PIXEL_SIZE,NUM_STARTS=1,STACK=.FALSE.)
            CALL HELIX_FIND_LAYER_LINES(SELF,Z,N)

            WRITE(*,'(A6,1X,2(A7,1X),A9)') 'Line #', 'Z', '1/Z', 'Order (n)'
            DO I=1,SIZE(N)
                IF (ABS(Z(I)) .LE. 0.5) THEN
                IF (Z(I) .NE. 0.) THEN
                    WRITE(*,'(I6,1X,2(F7.2,1X),I9)') I, Z(I)/PIXEL_SIZE, 1./Z(I)*PIXEL_SIZE, N(I)
                ELSE
                    WRITE(*,'(I6,1X,2(F7.2,1X),I9)') I, Z(I)/PIXEL_SIZE,               999., N(I)
                ENDIF
                ENDIF

            ENDDO
            WRITE(*,'(A)') '**info(HELIX_UNIT_TEST_2): all done'

        END SUBROUTINE HELIX_UNIT_TEST_2

        !>  \brief  Return the layer line positions (in Z) and their order (n)
        !!
        !!  We use the DeRosier "selection rule":
        !!
        !!  rise_per_subunit*Z - n*twist_per_subunit/2pi = l
        !!
        !!  so that Z = (l + n*twist_per_subunit/2pi)/rise_per_subunit
        !!
        !!  where Z is in reciprocal pixels,
        !!  rise_per_subunit is in pixels,
        !!  n and l are integers,
        !!  twist_per_subunit is in radians
        SUBROUTINE HELIX_FIND_LAYER_LINES(SELF,ZS,NS)
            USE RUNTIME_PARAMETERS, ONLY: PI
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY),     INTENT(IN)      ::  SELF            !<  Helical operator
            REAL,       ALLOCATABLE,    INTENT(INOUT)   ::  ZS(:)
            INTEGER,    ALLOCATABLE,    INTENT(INOUT)   ::  NS(:)
            ! Private variables
            INTEGER,    PARAMETER   ::  L_MAX = 200
            INTEGER,    PARAMETER   ::  N_MAX = 46
            INTEGER                 ::  I,L,N
            REAL                    ::  TWIST_RAD
            ! Start work

            ! Allocate output arrays
            IF (ALLOCATED(ZS)) DEALLOCATE(ZS)
            IF (ALLOCATED(NS)) DEALLOCATE(NS)
            ALLOCATE(ZS(L_MAX*N_MAX*2),NS(L_MAX*N_MAX*2))

            ! The twist per subunit in radians
            TWIST_RAD = SELF%TWIST / 180. * PI

            ! Get values of Z and N
            I=0

            DO L=0,L_MAX-1
                DO N=0,N_MAX-1
                    I     = I+1
                    ZS(I) = (L + N*TWIST_RAD/(2.*PI))/SELF%RISE
                    NS(I) = N
                    I     = I+1
                    ZS(I) = -ZS(I-1)
                    NS(I) = -NS(I-1)
                ENDDO
            ENDDO

        END SUBROUTINE HELIX_FIND_LAYER_LINES


        !>  \brief  Move by an integer number of subunits to land in the original asymetric unit
        !!
        !!  See notes of 130618 and their edit on 130620/1 for details, nomenclature etc
#ifdef SKIP_RUNTIME_CHECKS
        PURE &
#endif
        SUBROUTINE HELIX_POINT_TO_ASYM(SELF,Z,PHI,Z_NEW,PHI_NEW)
#ifndef SKIP_RUNTIME_CHECKS
            USE RUNTIME_PARAMETERS, ONLY : TERMINATE
#endif
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(IN)      ::  SELF            !<  Helical operator
            REAL(KIND=8),           INTENT(IN)      ::  Z,PHI           !<  Coordinates of point
            REAL(KIND=8),           INTENT(OUT)     ::  Z_NEW,PHI_NEW   !<  New coordinates
            ! Private variables
            REAL(KIND=8)    ::  Y,X
            INTEGER         ::  K,H
#ifndef SKIP_RUNTIME_CHECKS
			REAL(KIND=8)	::	YN,XN
#endif
			! Start work

            ! First, take advantage of any axial symmetry
            CALL HELIX_POINT_TO_C1(SELF,Z,PHI,Z_NEW,PHI_NEW)

			H = 0
			K = 0

            CALL HELIX_POINT_DIST_TO_ORIGIN(SELF,Z_NEW,PHI_NEW,Y,X)
            ! Move towards the 1-start axis using the fastest helix (smallest twist)
            IF (ABS(Y) .GT. 0.5) H = -CEILING(Y-0.5)
            ! Move towards the fastest helix axis using the 1-start helix
            IF (ABS(X) .GT. 0.5) K = -CEILING(X-0.5)

            ! Apply the move
            Z_NEW   = Z_NEW   + H*SELF%RISE_MIN_TWIST + K*SELF%RISE_MIN
            PHI_NEW = PHI_NEW + H*SELF%TWIST_MIN      + K*SELF%TWIST_MIN_RISE

            ! Make sure we're back in the asym range of angles (-180,180 for C1)
            PHI_NEW =  HELIX_PHI_TO_C1(SELF,PHI_NEW)

#ifndef SKIP_RUNTIME_CHECKS
            ! Sanity check
            IF (.NOT. HELIX_POINT_IS_IN_ASYM(SELF,Z_NEW,PHI_NEW)) THEN
				CALL HELIX_POINT_DIST_TO_ORIGIN(SELF,Z_NEW,PHI_NEW,YN,XN)
				IF (ABS(YN)-0.5D0 .GT. 0.01 .OR. ABS(XN)-0.5D0 .GT. 0.01) THEN
                WRITE(*,'(A)') '**ERROR(HELIX_POINT_TO_ASYM): failed sanity check'
                CALL HELIX_PRINT_INFO(SELF)
                CALL HELIX_PRINT_EXTRA_INFO(SELF)
                WRITE(*,'(A,2(F0.3,1X))') 'Z, Phi on entry: ', Z, PHI
                WRITE(*,'(A,2(F0.3,1X))') 'Y, X on entry  : ', Y, X
				WRITE(*,'(A,2(F0.3,1X))') 'Z, Phi on exit : ', Z_NEW, PHI_NEW
				WRITE(*,'(A,2(F0.3,1X))') 'Y, X on exit   : ', YN, XN
				WRITE(*,'(A,2(I0,1X))')   'K, H           : ', K, H
				WRITE(*,*) ''
				CALL TERMINATE('Fatal error in HELIX_POINT_TO_ASYM')
				ENDIF
            ENDIF
#endif
        END SUBROUTINE HELIX_POINT_TO_ASYM

        !>  \brief  Move a point using axial symmetry operators to the C1 region
        PURE SUBROUTINE HELIX_POINT_TO_C1(SELF,Z,PHI,Z_NEW,PHI_NEW)
            USE RUNTIME_PARAMETERS, ONLY : TERMINATE
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(IN)      ::  SELF            !<  Helical operator
            REAL(KIND=8),           INTENT(IN)      ::  Z,PHI           !<  Coordinates of point
            REAL(KIND=8),           INTENT(OUT)     ::  Z_NEW,PHI_NEW   !<  New coordinates
            ! Private variables

            ! Start work
            Z_NEW = Z
            PHI_NEW = HELIX_PHI_TO_C1(SELF,PHI)
        END SUBROUTINE HELIX_POINT_TO_C1

        PURE FUNCTION HELIX_PHI_TO_C1(SELF,PHI)
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(IN)      ::  SELF        !<  Helical operator
            REAL(KIND=8),           INTENT(IN)      ::  PHI         !<  Coordinate of point
            ! Result
            REAL(KIND=8)                            ::  HELIX_PHI_TO_C1
            ! Start work
            ! Private variables
            REAL(KIND=8)    ::  C1_PHI_LIMIT, C1_RANGE
            INTEGER         ::  N
            ! Start work

            C1_RANGE     = 360.0D0 / REAL(SELF%AXIAL_SYMMETRY,KIND=8)
            C1_PHI_LIMIT = C1_RANGE * 0.5D0
            N = CEILING((ABS(PHI)-C1_PHI_LIMIT)/C1_RANGE) * INT(SIGN(1.0D0,PHI)) * (-1)
            HELIX_PHI_TO_C1 = PHI + N * C1_RANGE

        END FUNCTION HELIX_PHI_TO_C1

        !>  \brief  Check whether a point is within the original asymetric unit
        PURE LOGICAL FUNCTION HELIX_POINT_IS_IN_ASYM(SELF,Z,PHI)
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(IN)      ::  SELF    !<  Helical operator
            REAL(KIND=8),           INTENT(IN)      ::  Z,PHI   !<  Coordinates of point
            ! Private variables
            REAL(KIND=8)    ::  Y,X
            REAL(KIND=8),   PARAMETER   ::  TOL = 0.000D0
            ! Start work
            CALL HELIX_POINT_DIST_TO_ORIGIN(SELF,Z,PHI,Y,X)
            HELIX_POINT_IS_IN_ASYM = (ABS(Y) .LE. 0.5D0+TOL) .AND. (ABS(X) .LE. 0.5D0+TOL)
        END FUNCTION HELIX_POINT_IS_IN_ASYM

        !>  \brief  Compute the distance
        PURE SUBROUTINE  HELIX_POINT_DIST_TO_ORIGIN(SELF,Z,PHI,Y,X)
            USE ROTATION_MATRICES, ONLY : RADIANS
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(IN)      ::  SELF    !<  Helical operator
            REAL(KIND=8),           INTENT(IN)      ::  Z,PHI   !<  Coordinates of point
            REAL(KIND=8),           INTENT(OUT)     ::  Y      !<  Distance to the "1-start" axis, in unit cells
            REAL(KIND=8),           INTENT(OUT)     ::  X      !<  Distance to the "fastest start" axis, in unit cells
            ! Private variables

            ! Start work

            X = RADIANS(PHI)*SELF%CONV_MAT(1,1)+Z*SELF%CONV_MAT(1,2)
            Y = RADIANS(PHI)*SELF%CONV_MAT(2,1)+Z*SELF%CONV_MAT(2,2)

        END SUBROUTINE HELIX_POINT_DIST_TO_ORIGIN


        !>  \brief  Update the conversion matrix
        SUBROUTINE HELIX_UPDATE_CONVERSION_MATRIX(SELF)
            USE ROTATION_MATRICES, ONLY : RADIANS
            USE RUNTIME_PARAMETERS, ONLY : TERMINATE
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(INOUT)   ::  SELF    !<  Helical operator
            ! Private variables
			REAL(KIND=8)	::	DET
            ! Start work

            !
            SELF%CONV_MAT(1,1) =  SELF%RISE_MIN_TWIST
            SELF%CONV_MAT(2,1) = -SELF%RISE_MIN
            SELF%CONV_MAT(1,2) = -RADIANS(SELF%TWIST_MIN)
            SELF%CONV_MAT(2,2) =  RADIANS(SELF%TWIST_MIN_RISE)

			DET = RADIANS(SELF%TWIST_MIN_RISE)*SELF%RISE_MIN_TWIST - RADIANS(SELF%TWIST_MIN)*SELF%RISE_MIN

			IF (DET .EQ. 0.0) THEN
				WRITE(*,'(A)') '**ERROR(HELIX_UPDATE_CONVERSION_MATRIX): singular matrix'
				CALL HELICAL_SYMMETRY_PRINT_INFO(SELF)
				CALL HELICAL_SYMMETRY_PRINT_EXTRA_INFO(SELF)
				WRITE(*,'(A,2(F0.3,1X))') 'twist, rise of S-start = ', SELF%TWIST_MIN_RISE, SELF%RISE_MIN
				WRITE(*,'(A,2(F0.3,1X))') 'twist, rise of F-start = ', SELF%TWIST_MIN, SELF%RISE_MIN_TWIST
				CALL TERMINATE('Fatal error in HELIX_UPDATE_CONVERSION_MATRIX')
			ELSE
				SELF%CONV_MAT = SELF%CONV_MAT / DET
			ENDIF

        END SUBROUTINE HELIX_UPDATE_CONVERSION_MATRIX

        !>  \brief  Convenience function, should probably be deprecated
        PURE FUNCTION HELIX_GUESS_SUBUNIT_HEIGHT(SELF)
            IMPLICIT NONE
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  SELF
            REAL(KIND=8) :: HELIX_GUESS_SUBUNIT_HEIGHT
            HELIX_GUESS_SUBUNIT_HEIGHT = SELF%ASYM_UNIT_HEIGHT
        END FUNCTION HELIX_GUESS_SUBUNIT_HEIGHT

        !>  \brief Guess the "height" of a subunit, i.e. its extent along the helical axis
        PURE SUBROUTINE HELIX_GUESS_SUBUNIT_HEIGHT_WIDTH(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(INOUT)  ::  SELF
            ! Private variables
            INTEGER ::  START_WITH_MIN_TWIST,START_WITH_MAX_TWIST
            TYPE(HELICAL_SYMMETRY)  ::  H_TEMP
            ! Start work

            H_TEMP = SELF
            IF (SELF%STACK) THEN
                SELF%ASYM_UNIT_HEIGHT = SELF%RISE
                SELF%ASYM_UNIT_WIDTH = 360.0D0/REAL(SELF%AXIAL_SYMMETRY,KIND=8)
            ELSE
                IF (SELF%AXIAL_SYMMETRY .GT. 1) THEN
                    ! Not sure I have the correct algorithm when axial symmetry is involved
                    CALL HELIX_TO_ONE_START(H_TEMP)
                    SELF%ASYM_UNIT_HEIGHT = H_TEMP%RISE
                    SELF%ASYM_UNIT_WIDTH = 360.0D0/REAL(SELF%AXIAL_SYMMETRY,KIND=8)
                ELSE
                    ! Find the start numbers which gives the smallest and largest twist per subunit
                    CALL HELIX_FIND_STARTS_WITH_MIN_MAX_TWIST(SELF,START_WITH_MIN_TWIST,START_WITH_MAX_TWIST)

                    ! Convert to that start number
                    CALL HELIX_CHANGE_NUM_STARTS(H_TEMP,START_WITH_MIN_TWIST)

                    ! The rise of that start helix is our guess
                    SELF%ASYM_UNIT_HEIGHT = H_TEMP%RISE

                    !
                    SELF%TWIST_MIN = H_TEMP%TWIST
                    SELF%RISE_MIN_TWIST = H_TEMP%RISE

                    !
                    !CALL HELIX_CHANGE_NUM_STARTS(H_TEMP,START_WITH_MAX_TWIST)
                    CALL HELIX_CHANGE_NUM_STARTS(H_TEMP,1)

                    ! The twist of that start helix is our guess
                    SELF%ASYM_UNIT_WIDTH = H_TEMP%TWIST
                ENDIF
            ENDIF
        END SUBROUTINE HELIX_GUESS_SUBUNIT_HEIGHT_WIDTH

        !>  \brief  Work out the helical parameters with fastest rise (slowest twist) within the family of equivalent start helices
        PURE SUBROUTINE HELIX_UPDATE_FASTEST_HELIX(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(INOUT)  ::  SELF
            ! Private variables
            INTEGER ::  START_WITH_MAX_TWIST
            TYPE(HELICAL_SYMMETRY)  ::  H_TEMP
            ! Start work

            IF (SELF%STACK) THEN
                H_TEMP = SELF
                CALL HELIX_CHANGE_NUM_STARTS(H_TEMP,1)
                SELF%TWIST_MIN = H_TEMP%TWIST
                SELF%RISE_MIN_TWIST = H_TEMP%RISE
            ELSE
                ! Find the start numbers which gives the smallest and largest twist per subunit
                CALL HELIX_FIND_STARTS_WITH_MIN_MAX_TWIST(SELF,SELF%START_MIN_TWIST,START_WITH_MAX_TWIST)

                ! Convert to that start number
                H_TEMP = SELF
                CALL HELIX_CHANGE_NUM_STARTS(H_TEMP,SELF%START_MIN_TWIST)

                SELF%TWIST_MIN = H_TEMP%TWIST
                SELF%RISE_MIN_TWIST = H_TEMP%RISE
            ENDIF
        END SUBROUTINE HELIX_UPDATE_FASTEST_HELIX

        !>  \brief  Work out the helical parameters with the slowest rise within the family of equivalent start helices
        PURE SUBROUTINE HELIX_UPDATE_SLOWEST_HELIX(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(INOUT)  ::  SELF
            ! Private variables
            TYPE(HELICAL_SYMMETRY)  ::  H_TEMP
            ! Start work
            IF (SELF%AXIAL_SYMMETRY .GT. 1 .OR. SELF%STACK) THEN
                SELF%START_MIN_RISE = 0
                SELF%TWIST_MIN_RISE = 360.0D0 / SELF%AXIAL_SYMMETRY
                SELF%RISE_MIN = 0.0D0
            ELSE
                SELF%START_MIN_RISE = 1
                H_TEMP = SELF
                CALL HELIX_CHANGE_NUM_STARTS(H_TEMP,1)
                SELF%TWIST_MIN_RISE = H_TEMP%TWIST
                SELF%RISE_MIN = H_TEMP%RISE
            ENDIF
        END SUBROUTINE HELIX_UPDATE_SLOWEST_HELIX

        !>  \brief  Find the number of starts such that the twist-per-subunit is minimised
        PURE SUBROUTINE HELIX_FIND_STARTS_WITH_MIN_MAX_TWIST(SELF,START_MIN,START_MAX)
            USE EULERANGLES, ONLY : DEGREES_M180_TO_180
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  SELF            !<  Helical operator
            INTEGER,                INTENT(OUT) ::  START_MIN       !<  Start with the smallest twist per subunit
            INTEGER,                INTENT(OUT) ::  START_MAX       !<  Start with the largest twist per subunit
            ! Private variables
            TYPE(HELICAL_SYMMETRY)  ::  H_TEMP_ONESTART, H_TEMP
            INTEGER ::  MAX_NUM_STARTS
            INTEGER ::  CURRENT_NUM_STARTS
            REAL(KIND=8)    ::  MIN_TWIST_PER_SUBUNIT, MAX_TWIST_PER_SUBUNIT
            ! Start work
            H_TEMP_ONESTART = SELF
            CALL HELIX_TO_ONE_START(H_TEMP_ONESTART)
            ! Work out how many starts would take the twist to > 360.
            MAX_NUM_STARTS = FLOOR(360.D0 / ABS(H_TEMP_ONESTART%TWIST) / SELF%AXIAL_SYMMETRY)
            ! Loop over all possible N starts
            MIN_TWIST_PER_SUBUNIT = 360.0D0
            START_MIN = 0
            MAX_TWIST_PER_SUBUNIT = 0.0D0
            START_MAX = 0
            DO CURRENT_NUM_STARTS=1,MAX_NUM_STARTS
                H_TEMP = H_TEMP_ONESTART
                CALL HELIX_CHANGE_NUM_STARTS(H_TEMP,CURRENT_NUM_STARTS)
                IF (ABS(H_TEMP%TWIST) .LT. MIN_TWIST_PER_SUBUNIT) THEN
                    MIN_TWIST_PER_SUBUNIT = ABS(H_TEMP%TWIST)
                    START_MIN = CURRENT_NUM_STARTS
                ENDIF
                IF (ABS(H_TEMP%TWIST) .GT. MAX_TWIST_PER_SUBUNIT) THEN
                    MAX_TWIST_PER_SUBUNIT = ABS(H_TEMP%TWIST)
                    START_MAX = CURRENT_NUM_STARTS
                ENDIF
            ENDDO
        END SUBROUTINE HELIX_FIND_STARTS_WITH_MIN_MAX_TWIST

        !>  \brief  Given a helical symmetry operator, modify it so that it describes the N-start equivalent
        PURE SUBROUTINE HELIX_CHANGE_NUM_STARTS(SELF,N)
            USE RUNTIME_PARAMETERS, ONLY : TERMINATE
            USE EULERANGLES, ONLY : DEGREES_M180_TO_180
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(INOUT)   ::  SELF
            INTEGER,                INTENT(IN)      ::  N       !<  Number of starts for the helical symmetry after modification
            ! Private variables

            ! Start work
            IF (SELF%NUM_STARTS .EQ. N) THEN
                ! Nothing to do
            ELSE
                IF (.NOT. SELF%NUM_STARTS .EQ. 1) THEN
                    CALL HELIX_TO_ONE_START(SELF)
                ENDIF
                ! The operator now describes the 1-start helix. Let's go to N
                IF (N .GT. 1) THEN
                    SELF%TWIST      =   MOD(DBLE(N)*SELF%TWIST,360.D0)
                    SELF%TWIST      =   DEGREES_M180_TO_180(SELF%TWIST)
                    SELF%RISE       =   DBLE(N) * SELF%RISE
                    SELF%NUM_STARTS =   N
                ELSE
                    ! Nothing to do
                ENDIF
            ENDIF
        END SUBROUTINE HELIX_CHANGE_NUM_STARTS

        !>  \brief  Return the 1-start helical operator corresponding to the n-start helical operator given as input
        PURE SUBROUTINE HELIX_TO_ONE_START(SELF)
            USE EULERANGLES, ONLY : DEGREES_M180_TO_180
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(INOUT)   ::  SELF
            ! Private variables

            ! Start work
            SELF%TWIST      =   MOD((360.D0 + SELF%TWIST) / DBLE(SELF%NUM_STARTS),360.D0)
            SELF%TWIST      =   DEGREES_M180_TO_180(SELF%TWIST)
            SELF%RISE       =   SELF%RISE / DBLE(SELF%NUM_STARTS)
            SELF%NUM_STARTS =   1
        END SUBROUTINE HELIX_TO_ONE_START

        !>  \brief  Compute the phi angle (in degrees) given X,Y, the coordinates (from the center; in pixels) of a point
        PURE FUNCTION HELIX_PHI_FROM_XY(X,Y) RESULT(PHI)
            USE ROTATION_MATRICES, ONLY : DEGREES
            USE RUNTIME_PARAMETERS, ONLY : H_ASYM_UNIT_CENTER
            IMPLICIT NONE
            ! Arguments
            REAL(KIND=8),   INTENT(IN)  ::  X,Y
            ! Result
            REAL(KIND=8)    ::  PHI
            ! Private variables

            ! Start work
            IF (X .EQ. 0.0 .AND. Y .EQ. 0.0) THEN
                PHI = 0.0
            ELSE
                PHI = DEGREES(  ATAN2(  Y*(-H_ASYM_UNIT_CENTER(1))+X*H_ASYM_UNIT_CENTER(2), &
                                        X*( H_ASYM_UNIT_CENTER(1))+Y*H_ASYM_UNIT_CENTER(2)))
            ENDIF
        END FUNCTION HELIX_PHI_FROM_XY

        !>  \brief  Given polar coordinates (phi in degrees, radius in pixels), output the XY coordinates (in pixels from center)
        PURE SUBROUTINE HELIX_XY_FROM_PHI_R(PHI,R,X,Y)
            USE ROTATION_MATRICES, ONLY : RADIANS
            USE RUNTIME_PARAMETERS, ONLY : H_ASYM_UNIT_CENTER
            IMPLICIT NONE
            ! Arguments
            REAL(KIND=8),   INTENT(IN)  ::  PHI, R
            REAL(KIND=8),   INTENT(OUT) ::  X,Y
            !
            X =  COS(RADIANS(PHI)) * H_ASYM_UNIT_CENTER(1) * R + SIN(RADIANS(PHI)) * H_ASYM_UNIT_CENTER(2) * R
            Y = -SIN(RADIANS(PHI)) * H_ASYM_UNIT_CENTER(1) * R + COS(RADIANS(PHI)) * H_ASYM_UNIT_CENTER(2) * R
        END SUBROUTINE HELIX_XY_FROM_PHI_R



        !>  \brief  Compute the pitch of a helix, given the helical operator
        !!
        !!  The pitch of the helix is the distance along the helical axis over which the helix rotates by 360 degrees
        PURE FUNCTION HELIX_PITCH(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  SELF
            ! Result
            REAL(KIND=8)    ::  HELIX_PITCH
            ! Private variables

            ! Start work
            IF (SELF%TWIST .EQ. 0.0) THEN
                HELIX_PITCH = ABS(SELF%RISE)
            ELSE
                HELIX_PITCH = ABS(SELF%RISE * 360.0D0 / SELF%TWIST)
            ENDIF
        END FUNCTION HELIX_PITCH

        PURE ELEMENTAL FUNCTION HELIX_DIVIDE_BY_REAL(HELIX_IN,VAL) RESULT(HELIX_OUT)
            IMPLICIT NONE
            !
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  HELIX_IN
            REAL,                   INTENT(IN)  ::  VAL
            !
            TYPE(HELICAL_SYMMETRY)  ::  HELIX_OUT
            !
            HELIX_OUT%PIXEL_UNITS   =   HELIX_IN%PIXEL_UNITS
            HELIX_OUT%RISE          =   HELIX_IN%RISE / VAL
            HELIX_OUT%TWIST         =   HELIX_IN%TWIST / VAL
        END FUNCTION HELIX_DIVIDE_BY_REAL

        PURE ELEMENTAL FUNCTION HELIX_PLUS_HELIX(HELIX1,HELIX2)
            IMPLICIT NONE
            !
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  HELIX1, HELIX2
            !
            TYPE(HELICAL_SYMMETRY)  ::  HELIX_PLUS_HELIX
            !
            !CALL CHECK_HELIX_ARITHMETIC_IS_POSSIBLE(HELIX1,HELIX2)
            HELIX_PLUS_HELIX%PIXEL_UNITS    =   HELIX1%PIXEL_UNITS
            HELIX_PLUS_HELIX%RISE           =   HELIX1%RISE + HELIX2%RISE
            HELIX_PLUS_HELIX%TWIST          =   HELIX1%TWIST + HELIX2%TWIST
        END FUNCTION HELIX_PLUS_HELIX

        PURE ELEMENTAL FUNCTION HELIX_MINUS_HELIX(HELIX1,HELIX2)
            IMPLICIT NONE
            !
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  HELIX1, HELIX2
            !
            TYPE(HELICAL_SYMMETRY)  ::  HELIX_MINUS_HELIX
            !
            !CALL CHECK_HELIX_ARITHMETIC_IS_POSSIBLE(HELIX1,HELIX2)
            HELIX_MINUS_HELIX%PIXEL_UNITS   =   HELIX1%PIXEL_UNITS
            HELIX_MINUS_HELIX%RISE          =   HELIX1%RISE - HELIX2%RISE
            HELIX_MINUS_HELIX%TWIST         =   HELIX1%TWIST - HELIX2%TWIST
        END FUNCTION HELIX_MINUS_HELIX

        PURE ELEMENTAL FUNCTION HELIX_TIMES_HELIX(HELIX1,HELIX2)
            IMPLICIT NONE
            !
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  HELIX1, HELIX2
            !
            TYPE(HELICAL_SYMMETRY)  ::  HELIX_TIMES_HELIX
            !
            !CALL CHECK_HELIX_ARITHMETIC_IS_POSSIBLE(HELIX1,HELIX2)
            HELIX_TIMES_HELIX%PIXEL_UNITS   =   HELIX1%PIXEL_UNITS
            HELIX_TIMES_HELIX%RISE          =   HELIX1%RISE * HELIX2%RISE
            HELIX_TIMES_HELIX%TWIST         =   HELIX1%TWIST * HELIX2%TWIST
        END FUNCTION HELIX_TIMES_HELIX

        PURE ELEMENTAL FUNCTION HELIX_SQRT(HELIX)
            IMPLICIT NONE
            !
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  HELIX
            !
            TYPE(HELICAL_SYMMETRY)  ::  HELIX_SQRT
            !
            HELIX_SQRT = HELIX
            HELIX_SQRT%RISE = SQRT(HELIX%RISE)
            HELIX_SQRT%TWIST = SQRT(HELIX%TWIST)
        END FUNCTION HELIX_SQRT


        !>  \brief  Check that two sets of helical parameters can be combined using arithmetics. If not, crash out.
        SUBROUTINE CHECK_HELIX_ARITHMETIC_IS_POSSIBLE(HELIX1,HELIX2)
            USE RUNTIME_PARAMETERS, ONLY : TERMINATE
            IMPLICIT NONE
            !
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  HELIX1, HELIX2
            !
            IF (      HELIX1%PIXEL_UNITS .EQV. HELIX2%PIXEL_UNITS &
                .AND. HELIX1%AXIAL_SYMMETRY .EQ. HELIX2%AXIAL_SYMMETRY &
                .AND. HELIX1%PERPENDICULAR_SYMMETRY .EQ. HELIX2%PERPENDICULAR_SYMMETRY ) THEN
                ! We don't do anything
            ELSE
                WRITE(*,'(2A)') '**ERROR(CHECK_HELIX_ARITHMETIC_IS_POSSIBLE): cannot add helical parameters unless ',    &
                                'they are in the same units and have the same axial & perpendicular symmetries'
                CALL HELICAL_SYMMETRY_PRINT_INFO(HELIX1)
                CALL HELICAL_SYMMETRY_PRINT_INFO(HELIX2)
                CALL TERMINATE('Fatal error in CHECK_HELIX_ARITHMETIC_IS_POSSIBLE')
            ENDIF
        END SUBROUTINE CHECK_HELIX_ARITHMETIC_IS_POSSIBLE


        !>  \brief  Print out helical symmetry parameters to disk
        SUBROUTINE HELICAL_SYMMETRY_PRINT_INFO(SELF)
            USE RUNTIME_PARAMETERS, ONLY : WORD_LEN
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  SELF
            ! Private variables
            CHARACTER(LEN=WORD_LEN)::   RISE_UNIT
            ! Start work
            IF (SELF%PIXEL_UNITS) THEN
                RISE_UNIT  = ' pixels'
            ELSE
                RISE_UNIT  = ' Angstr.'
            ENDIF
            IF (SELF%STACK) THEN
                WRITE(*,'(A,I0,A,F0.6,1X,A,A,F0.6,A)', ADVANCE='NO')  'Stack (', &
                                                                    SELF%AXIAL_SYMMETRY, '-fold symmetry) with rise of ', &
                                                                    SELF%RISE, TRIM(ADJUSTL(RISE_UNIT)), &
                                                                    ' and twist of ', SELF%TWIST, ' degrees.'
            ELSE
                WRITE(*,'(I0,A,I0,A,F0.6,1X,A,A,F0.6,A)', ADVANCE='NO')  SELF%NUM_STARTS, '-start helix (', &
                                                                    SELF%AXIAL_SYMMETRY, '-fold symmetry) with rise of ', &
                                                                    SELF%RISE, TRIM(ADJUSTL(RISE_UNIT)), &
                                                                    ' and twist of ', SELF%TWIST, ' degrees.'
            ENDIF
            IF (SELF%PERPENDICULAR_SYMMETRY .EQ. 2) THEN
                WRITE(*,'(A)') ' Non-polar.'
            ELSE
                WRITE(*,'(A)') ' Polar.'
            ENDIF
        END SUBROUTINE HELICAL_SYMMETRY_PRINT_INFO

        !>  \brief  Print extra info about the helical operator
        SUBROUTINE HELICAL_SYMMETRY_PRINT_EXTRA_INFO(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY), INTENT(IN)  ::  SELF
            ! Private variables
            ! Start work
            WRITE(*,'(A,2(I0,1X))') 'Starts fastest, slowest = ', SELF%START_MIN_TWIST, SELF%START_MIN_RISE
            WRITE(*,'(A,2(F0.3,1X))') 'rise, twist of fastest helix: ', SELF%RISE_MIN_TWIST, SELF%TWIST_MIN
            WRITE(*,'(A,2(F0.3,1X))') 'rise, twist of slowest helix: ', SELF%RISE_MIN, SELF%TWIST_MIN_RISE
        END SUBROUTINE HELICAL_SYMMETRY_PRINT_EXTRA_INFO

        !>  \brief  Reset all values
        PURE ELEMENTAL SUBROUTINE HELICAL_SYMMETRY_RESET(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY),     INTENT(OUT)     ::  SELF
            ! Start work
            SELF%PIXEL_UNITS = .FALSE.
            SELF%RISE = 0.0D0
            SELF%TWIST = 0.0D0
            SELF%AXIAL_SYMMETRY = 1
            SELF%PERPENDICULAR_SYMMETRY = 1
            SELF%NUM_STARTS = 1
            SELF%ASYM_UNIT_HEIGHT = 0.0D0
            SELF%ASYM_UNIT_WIDTH = 0.0D0
            SELF%STACK = .FALSE.
        END SUBROUTINE HELICAL_SYMMETRY_RESET

        !>  \brief  Initialise a helical_symmetry object, given relevant parameters
        !!
        !!  If a pixel size is given, it is assumed the parameters were given in Angstroms & degrees, and they will be converted to pixels & radians
        SUBROUTINE HELICAL_SYMMETRY_INIT(SELF,RISE,TWIST,AXIAL_SYM,PERPENDICULAR_SYM,PIXEL_SIZE,NUM_STARTS,STACK)
            USE ROTATION_MATRICES, ONLY : RADIANS
            USE RUNTIME_PARAMETERS, ONLY : TERMINATE
            USE EULERANGLES, ONLY : DEGREES_M180_TO_180
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY),     INTENT(INOUT)   ::  SELF                !<  Helical symmetry parameters to initialise
            REAL,                       INTENT(IN)      ::  RISE                !<  The rise (in pixels/Angstroms), i.e. the translation along the helical axis to get from one subunit to the next
            REAL,                       INTENT(IN)      ::  TWIST               !<  The twist (in radians/degrees), i.e. the rotation around the helical axis to get from one subunit to the next
            INTEGER,    OPTIONAL,       INTENT(IN)      ::  AXIAL_SYM           !<  Axial symmetry. If AXIAL_SYMMETRY = N, the helix is N-start.
            INTEGER,    OPTIONAL,       INTENT(IN)      ::  PERPENDICULAR_SYM   !<  Symmetry with axis perpendicular to helix axis. This can either be 1 or 2. 2 means the helix is not polar, like DNA for example.
            REAL,       OPTIONAL,       INTENT(IN)      ::  PIXEL_SIZE          !<  Pixel size (in Angstrom). If given, it is assumed the rise were given in Angstroms. If not, it is assume the rise was given in pixels.
            INTEGER,    OPTIONAL,       INTENT(IN)      ::  NUM_STARTS          !<  The helical parameters describes n NUM_START-helix
            LOGICAL,    OPTIONAL,       INTENT(IN)      ::  STACK               !<  This a stack rather than a true helix
            ! Private variables
            INTEGER ::  AAXIAL_SYM
            INTEGER ::  PPERPENDICULAR_SYM
            ! Start work
            ! Rise must be positive
            IF (RISE .LE. 0.) THEN
                WRITE(*,'(A,F0.6)') '**ERROR(HELICAL_SYMMETRY_INIT): rise is not positive: ', RISE
                CALL TERMINATE('Fatal error in HELICAL_SYMMETRY_INIT')
            ENDIF
            ! Deal with optional parameters
            IF (PRESENT(AXIAL_SYM)) THEN
                AAXIAL_SYM = AXIAL_SYM
            ELSE
                AAXIAL_SYM = 1  ! if the user didn't specify a helical symmetrym we assume no symmetry, i.e. a 1-start helix
            ENDIF
            IF (PRESENT(PERPENDICULAR_SYM)) THEN
                PPERPENDICULAR_SYM = PERPENDICULAR_SYM
            ELSE
                PPERPENDICULAR_SYM = 1  ! if the user didn't specify a symmetry here, we assume no symmetry, i.e. a polar filament
            ENDIF
            ! Perpendicular symmetry can only be 1 or 2
            IF (PPERPENDICULAR_SYM .LT. 1 .OR. PPERPENDICULAR_SYM .GT. 2) THEN
                WRITE(*,'(2A,I0)') '**ERROR(HELICAL_SYMMETRY_INIT): a helix cannot have this symmetry perpendicular', &
                                    ' to the helical axis : ', PPERPENDICULAR_SYM
                CALL TERMINATE('Fatal error in HELICAL_SYMMETRY_INIT')
            ENDIF
            ! Number of starts
            IF (PRESENT(NUM_STARTS)) THEN
                IF (NUM_STARTS .GE. 1) THEN
                    SELF%NUM_STARTS = NUM_STARTS
                ELSE
                    WRITE(*,'(A)') '**ERROR(HELICAL_SYMMETRY_INIT): nonsensical number of starts given as argument: ', NUM_STARTS
                    CALL TERMINATE('Fatal error in HELICAL_SYMMETRY_INIT')
                ENDIF
            ELSE
                SELF%NUM_STARTS = 1
            ENDIF
            ! Set up the values
            IF (PRESENT(PIXEL_SIZE)) THEN
                SELF%RISE = RISE / PIXEL_SIZE
                SELF%PIXEL_UNITS = .TRUE.
            ELSE
                SELF%RISE = RISE
                SELF%PIXEL_UNITS = .FALSE.
            ENDIF
            SELF%TWIST          =   MOD(TWIST,360.E0)
            SELF%TWIST          =   DEGREES_M180_TO_180(SELF%TWIST)
            SELF%AXIAL_SYMMETRY = AAXIAL_SYM
            SELF%PERPENDICULAR_SYMMETRY = PPERPENDICULAR_SYM
            SELF%STACK = .FALSE.
            IF (PRESENT(STACK)) SELF%STACK = STACK
            IF (SELF%STACK .AND. SELF%NUM_STARTS .GT. 1) THEN
                WRITE(*,'(A)') '**ERROR(HELICAL_SYMMETRY_INIT): stacks with >1 starts are not supported - do they exist?'
                CALL TERMINATE('Fatal error in HELICAL_SYMMETRY_INIT')
            ENDIF


            !
            CALL HELIX_UPDATE_FASTEST_HELIX(SELF)
            CALL HELIX_UPDATE_SLOWEST_HELIX(SELF)
            CALL HELIX_UPDATE_CONVERSION_MATRIX(SELF)

            ! Work out the extent of the asymmetric unit (to be deprecated)
            CALL HELIX_GUESS_SUBUNIT_HEIGHT_WIDTH(SELF)

        END SUBROUTINE HELICAL_SYMMETRY_INIT

        !>  \brief  Given a pixel size, convert the helical parameters to/from pixel/Angstrom units
        PURE SUBROUTINE HELICAL_SYMMETRY_UNIT_CONVERT(SELF,PIXEL_SIZE)
            IMPLICIT NONE
            ! Arguments
            TYPE(HELICAL_SYMMETRY),     INTENT(INOUT)   ::  SELF
            REAL,                       INTENT(IN)      ::  PIXEL_SIZE  !<  Pixel size in Angstroms
            ! Private variables
            ! Start work
            IF (SELF%PIXEL_UNITS) THEN
                SELF%RISE = SELF%RISE * PIXEL_SIZE
            ELSE
                SELF%RISE = SELF%RISE / PIXEL_SIZE
            ENDIF
            SELF%PIXEL_UNITS = .NOT. SELF%PIXEL_UNITS
        END SUBROUTINE HELICAL_SYMMETRY_UNIT_CONVERT


        !>  \brief  Given lattice parameters, calculate & return the corresponding helical symmetry
        !!  \todo   Write this function! I have already worked out how to do this when the lattice is assumed to be square (ANGLE = 90 deg),
        !!          but I haven't had time to generalise to any lattice. (see /gusr/alr99/linux/notes/notes-20090510-helical-parameters.ods)
!       FUNCTION HELICAL_SYMMETRY_FROM_LATTICE RESULT(H_SYM)
!           IMPLICIT NONE
!           ! Arguments
!
!           ! Result
!           TYPE(HELICAL_SYMMETRY)  ::  H_SYM
!           ! Private variables
!
!           ! Start work
!
!       END FUNCTION HELICAL_SYMMETRY_FROM_LATTICE



END MODULE HELICES
