!>  \brief Module to perform arithmetic operations on images
!!
!!
MODULE IMAGES_ARITHM
    USE IMAGES
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE

    !INTERFACE OPERATOR(+)
    !   MODULE PROCEDURE ADD_IMAGES_FUNC
    !END INTERFACE

    !INTERFACE OPERATOR(*)
    !   MODULE PROCEDURE MULTIPLY_IMAGES_FUNC
    !END INTERFACE

    !INTERFACE OPERATOR(/)
    !   MODULE PROCEDURE DIVIDE_IMAGES_FUNC
    !END INTERFACE


    CONTAINS

    !>  \brief  Compute the average of a stack of images
    SUBROUTINE IMAGES_AVERAGE(SELF,AVE)
        use Images
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    ALLOCATABLE,    INTENT(IN)      ::  SELF(:)
        TYPE(IMAGE),                    INTENT(INOUT)   ::  AVE
        ! Private variables
        INTEGER ::  I
        ! Start work
        IF (.NOT. ALLOCATED(SELF)) THEN
            WRITE(*,'(A)') '**ERROR(IMAGES_AVERAGE): input array of images is not allocated'
            CALL TERMINATE('Fatal error in IMAGES_AVERAGE')
        ENDIF

        ! Allocate memory for the average image
        IF (ave%IsAllocated()) CALL IMAGE_DEALLOCATE(AVE)
        call ave%Allocate(mould=self(1))

        ! Loop over the input images
        DO I=1,SIZE(SELF)
            IF ( .NOT. ave%HasSameDimsAs(self(i)) ) THEN
                WRITE(*,'(A,I0,A,2(3(I0,1X),A))') '**ERROR(IMAGES_AVERAGE): dimensions of input image ', I, ' (', &
                                                  SELF(I)%logical_dimensions, &
                                                  ') are not compatible with the dimensions of the average image (',  &
                                                  AVE%logical_dimensions, ')'
                CALL TERMINATE('Fatal error in IMAGES_AVERAGE')
            ENDIF
            AVE%real_values = AVE%real_values + SELF(I)%real_values
        ENDDO
        AVE%real_values = AVE%real_values / REAL(SIZE(SELF))
    END SUBROUTINE IMAGES_AVERAGE

    !>  \brief  Float an image average to 0.0, normalise its standard deviation to 1.0 and mask the outside to zero
    SUBROUTINE IMAGE_FLOAT_NORM_MASK(SELF,MASK_RADIUS)
        USE IMAGES
        USE RUNTIME_PARAMETERS, ONLY : SAFE_FOR_SQRT
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),            INTENT(INOUT)   ::  SELF
        REAL,       OPTIONAL,   INTENT(IN)      ::  MASK_RADIUS !<  Ignore pixels beyond this radius (in pixels). If mask_radius is very large or 0.0, there will be no masking.
        ! Private variables
        INTEGER ::  I,J,K, ORIGIN(3), X,Y,Z, VOX_COUNT
        REAL    ::  RAD_TMP
        REAL(KIND=8)    ::  SUM_1, SUM_2
        REAL    ::  RAD_MASK, STD, AVE
        ! Start work

        IF (.NOT. self%IsReal()) THEN
            WRITE(*,'(A)') '**ERROR(IMAGE_FLOAT_NORM_MASK): input image should be real'
            CALL IMAGE_PRINT_INFO(SELF)
            CALL TERMINATE('Fatal error in IMAGE_FLOAT_NORM_MASK')
        ENDIF

        ! The radius of the mask
        RAD_MASK = REAL(SELF%logical_dimensions(1)) * 0.5 * 0.9
        IF (PRESENT(MASK_RADIUS)) RAD_MASK = MASK_RADIUS
        IF (RAD_MASK .EQ. 0.0) THEN
            RAD_MASK = 1.0 + MAXVAL(SELF%logical_dimensions)
        ENDIF
        RAD_MASK = RAD_MASK**2

        SUM_1 = 0.0D0
        SUM_2 = 0.0D0
        VOX_COUNT = 0
        if (self%is_in_real_space .and. self%origin_at_center) then
            ORIGIN = self%physical_address_of_box_center
        else
            ORIGIN = [1,1,1]
        endif


        ! Loop over voxels to get stats
        DO K=1,SELF%logical_dimensions(3)
            Z = (K-ORIGIN(3))**2
            DO J=1,SELF%logical_dimensions(2)
                Y = (J-ORIGIN(2))**2
                DO I=1,SELF%logical_dimensions(1)
                    X = (I-ORIGIN(1))**2
                    RAD_TMP = REAL(X+Y+Z)
                    IF (RAD_TMP .LT. RAD_MASK) THEN
                        SUM_1 = SUM_1 + SELF%real_values(I,J,K)
                        SUM_2 = SUM_2 + SELF%real_values(I,J,K)**2
                        VOX_COUNT = VOX_COUNT + 1
                    ENDIF
                ENDDO
            ENDDO
        ENDDO

        STD = (SUM_2 - (SUM_1**2/VOX_COUNT))/VOX_COUNT
        CALL SAFE_FOR_SQRT(STD,NEGATIVE_TO_ZERO=.TRUE.)
        !IF (STD .EQ. 0.0) CALL WRITE_TO_DISK(SELF,'dbg_std_zero.mrc')
        STD = SQRT(STD)
        AVE = SUM_1 / VOX_COUNT

        ! Loop over voxels to modify values
        DO K=1,SELF%logical_dimensions(3)
            Z = (K-ORIGIN(3))**2
            DO J=1,SELF%logical_dimensions(2)
                Y = (J-ORIGIN(2))**2
                DO I=1,SELF%logical_dimensions(1)
                    X = (I-ORIGIN(1))**2
                    RAD_TMP = REAL(X+Y+Z)
                    IF (RAD_TMP .LT. RAD_MASK) THEN
                        IF (STD .NE. 0.0) SELF%real_values(I,J,K) = (SELF%real_values(I,J,K)-AVE)/STD
                    ELSE
                        SELF%real_values(I,J,K) = 0.0
                    ENDIF
                ENDDO
            ENDDO
        ENDDO


    END SUBROUTINE IMAGE_FLOAT_NORM_MASK

    !>  \brief  Compute the polar moment of inertia of a structure based on the input 2D image, which is assumed to be a binarised version of a slice through the structure.
    !!
    !!  The polar moment of inertia is a quantity which relates the torque applied to an object and the object's resistance to that torque. See http://en.wikipedia.org/wiki/Polar_moment_of_inertia
    !!  It is calculated here following Sachse, Grigorieff & Fandrich (2010): Nanoscale Flexibility Parameters of Alzheimer Amyloid Fibrils Determined by Electron Cryo-Microscopy, DOI: 10.1002/anie.200904781
    !!
    !!  \todo   Decide whether it's a good
    REAL FUNCTION POLAR_MOMENT_OF_INERTIA(IMG,PIXEL_SIZE)
        USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(IN)  ::  IMG         !<  Input image. Must be 2D (i.e. DIM(3) == 1) and binarised (i.e. only 1.0's and 0.0's in RVALUE)
        REAL,           INTENT(IN)  ::  PIXEL_SIZE  !<  Pixel size (in Angstroms)
        ! Private variables
        INTEGER ::  I,J
        REAL(KIND=8)    ::  SUM
        REAL(KIND=8)    ::  PIXEL_SIZE_2
        ! Start work

        ! Check that the image is 2D
        IF (IMG%logical_dimensions(3) .NE. 1) THEN
            WRITE(*,'(A,I4)') '**ERROR(POLAR_MOMENT_OF_INERTIA): input image should be 2D, but DIM(3) = ', IMG%logical_dimensions(3)
            CALL TERMINATE('Fatal error in POLAR_MOMENT_OF_INERTIA')
        ENDIF

        ! Check that the image contains only 0s and 1s
        IF (.NOT. img%IsBinarised()) THEN
            WRITE(*,'(A)') '**ERROR(POLAR_MOMENT_OF_INERTIA): input image should be binarised'
            CALL TERMINATE('Fatal error in POLAR_MOMENT_OF_INERTIA')
        ENDIF

        ! Initialise
        POLAR_MOMENT_OF_INERTIA = 0.0
        SUM                     = 0.0D0
        PIXEL_SIZE_2            = DBLE(PIXEL_SIZE)**2

        ! Loop over the image
        DO J=1,IMG%logical_dimensions(2)
            DO I=1,IMG%logical_dimensions(1)
                IF (IMG%real_values(I,J,1) .GT. 0.0) THEN
                    SUM = SUM + DBLE(IMG%ComputeRadiusFromPhysicalAddress(I,J,1)**2)*PIXEL_SIZE_2
                ENDIF
            ENDDO
        ENDDO

        ! Result is ready
        POLAR_MOMENT_OF_INERTIA = REAL(SUM)
    END FUNCTION POLAR_MOMENT_OF_INERTIA

    !>  \brief  Compute the center of mass of an image
    !!
    !!  I lifted the idea for this from EMAN2.06's calc_max_location_wrap_intp routine
    SUBROUTINE IMAGE_CENTER_OF_MASS(SELF,COM_COO,LBOUNDS,UBOUNDS)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(IN)  ::  SELF
        REAL,                       INTENT(OUT) ::  COM_COO(3)  !<  Coordinates of the center of mass
        INTEGER,        OPTIONAL,   INTENT(IN)  ::  LBOUNDS(3)  !<  Image address lower bounds for the center of mass computation
        INTEGER,        OPTIONAL,   INTENT(IN)  ::  UBOUNDS(3)  !<  Image address upper bounds
        ! Private variables
        INTEGER ::  LLBOUNDS(3), UUBOUNDS(3)
        INTEGER ::  I,J,K
        REAL(KIND=8)    ::  COO(3), S
        REAL(KIND=8)    ::  VAL
        ! Start work

        ! Meant for real images only
        IF (.NOT. self%IsReal()) THEN
            CALL TERMINATE('Only implemented for real images')
        ENDIF

        ! Bounds for the computation
        LLBOUNDS = (/1,1,1/)
        UUBOUNDS = SELF%logical_dimensions
        IF (PRESENT(LBOUNDS)) LLBOUNDS = LBOUNDS
        IF (PRESENT(UBOUNDS)) UUBOUNDS = UBOUNDS

        LLBOUNDS = MAX(LLBOUNDS,1)
        UUBOUNDS = MIN(UUBOUNDS,SELF%logical_dimensions)

        COO = 0.0D0
        S   = 0.0D0
        DO K=LLBOUNDS(3),UUBOUNDS(3)
            DO J=LLBOUNDS(2),UUBOUNDS(2)
                DO I=LLBOUNDS(1),UUBOUNDS(1)
                    VAL = DBLE(SELF%real_values(I,J,K))
                    COO(1) = COO(1) + DBLE(I)*VAL
                    COO(2) = COO(2) + DBLE(J)*VAL
                    COO(3) = COO(3) + DBLE(K)*VAL
                    S = S + VAL
                ENDDO
            ENDDO
        ENDDO

        COM_COO = COO / S

    END SUBROUTINE IMAGE_CENTER_OF_MASS

    !>  \brief  Return the moment of inertia along the requested dimension. See Sachse,Grigorieff&Fandrich (2010) for details.
    REAL FUNCTION MOMENT_OF_INERTIA(IMG,DIM,PIXEL_SIZE)
        USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(IN)  ::  IMG             !<  Input image. Must be binarised (i.e. only 0.0's and 1.0's), REAL and 2D.
        INTEGER,        INTENT(IN)  ::  DIM             !<  The dimension along which to calculate the moment of inertia. 1 or 2
        REAL,           INTENT(IN)  ::  PIXEL_SIZE      !<  The pixel size (Angstroms)
        ! Private variables.
        INTEGER         ::  I,J
        REAL(KIND=8)    ::  SUM
        REAL(KIND=8)    ::  PIXEL_SIZE_2
        INTEGER         ::  ORIG(3)
        ! Start work

        ! Check that the image is 2D
        IF (IMG%logical_dimensions(3) .NE. 1) THEN
            WRITE(*,'(A,I4)') '**ERROR(MOMENT_OF_INERTIA): input image should be 2D, but DIM(3) = ', IMG%logical_dimensions(3)
            CALL TERMINATE('Fatal error in MOMENT_OF_INERTIA')
        ENDIF

        ! Check that the image contains only 0s and 1s
        IF (.NOT. img%IsBinarised()) THEN
            WRITE(*,'(A)') '**ERROR(MOMENT_OF_INERTIA): input image should be binarised'
            CALL TERMINATE('Fatal error in MOMENT_OF_INERTIA')
        ENDIF

        ! Initialise
        MOMENT_OF_INERTIA       = 0.0
        SUM                     = 0.0D0
        PIXEL_SIZE_2            = DBLE(PIXEL_SIZE)**2
        if (img%is_in_real_space .and. img%origin_at_center) then
            ORIG = img%physical_address_of_box_center
        else
            ORIG = [1,1,1]
        endif

        ! Loop over the image
        DO J=1,IMG%logical_dimensions(2)
            DO I=1,IMG%logical_dimensions(1)
                IF (IMG%real_values(I,J,1) .GT. 0.0) THEN
                    IF (DIM .EQ. 1) THEN
                        SUM = SUM + DBLE((I-ORIG(1))**2)*PIXEL_SIZE_2
                    ELSE
                        SUM = SUM + DBLE((J-ORIG(2))**2)*PIXEL_SIZE_2
                    ENDIF
                ENDIF
            ENDDO
        ENDDO

        ! Result is ready
        MOMENT_OF_INERTIA = REAL(SUM)
    END FUNCTION MOMENT_OF_INERTIA

    !>  \brief  Swap the X and Y axes. This results in a mirror about the X=Y plane.
    SUBROUTINE SWAP_XY(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(INOUT)   ::  SELF    !<  Image object
        ! Private variables
        TYPE(IMAGE) ::  OUTPUT_IMAGE
        INTEGER     ::  I,J,K
        ! Start work

        ! Check input image is allocated
        IF (.NOT. self%IsAllocated()) THEN
            WRITE(*,'(A)') '**ERROR(SWAP_XY): input image not allocated'
            CALL TERMINATE('Fatal error in SWAP_XY')
        ENDIF

        ! Only coding for real images, to save time
        IF (.not. self%is_in_real_space) THEN
            WRITE(*,'(A)') '**ERROR(SWAP_XY): this feature only implemented for real images'
            CALL TERMINATE('Fatal error in SWAP_XY')
        ENDIF

        !
        ! Set up the output image
        !
        OUTPUT_IMAGE%logical_dimensions(3)     =   SELF%logical_dimensions(3)
        OUTPUT_IMAGE%logical_dimensions(2)     =   SELF%logical_dimensions(1)
        OUTPUT_IMAGE%logical_dimensions(1)     =   SELF%logical_dimensions(2)
        OUTPUT_IMAGE%is_in_real_space           =   SELF%is_in_real_space

        !
        ! Allocate the output image
        !
        call output_image%Allocate()

        !
        ! Get values for output image
        !
        DO K=1,SELF%logical_dimensions(3)
            DO J=1,SELF%logical_dimensions(2)
                DO I=1,SELF%logical_dimensions(1)
                    OUTPUT_IMAGE%real_values(J,I,K)  =   SELF%real_values(I,J,K)
                ENDDO
            ENDDO
        ENDDO

        ! Reallocate input image
        call self%Allocate(mould=OUTPUT_IMAGE)

        ! Copy results overs
        SELF%real_values = OUTPUT_IMAGE%real_values

        ! Deallocate temporary image object
        CALL IMAGE_DEALLOCATE(OUTPUT_IMAGE)
    END SUBROUTINE SWAP_XY

    !>  \brief  Zero-float an image (ie. remove average pixel value from all values in real space, or zero central pixel in reciprocal space)
    SUBROUTINE ZERO_FLOAT(IMG)
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(INOUT)   ::  IMG
        ! Private variables
        REAL(KIND=8)    ::  AVERAGE
        INTEGER ::  I,J,K
        ! Start work
        IF (img%is_in_real_space) THEN
            ! Evaluate the average
            AVERAGE = 0.0D0
            DO K=1,IMG%logical_dimensions(3)
                DO J=1,IMG%logical_dimensions(2)
                    DO I=1,IMG%logical_dimensions(1)
                        AVERAGE = AVERAGE + DBLE(IMG%real_values(I,J,K))
                    ENDDO
                ENDDO
            ENDDO
            AVERAGE = AVERAGE / DBLE(IMG%logical_dimensions(3)*IMG%logical_dimensions(2)*IMG%logical_dimensions(1))
            ! Substract average from all pixels
            IMG%real_values = IMG%real_values - REAL(AVERAGE)
        ELSE IF (.not. img%is_in_real_space) THEN
            IMG%complex_values(1,1,1) = (0.0,0.0)
        ENDIF
    END SUBROUTINE

    !>  \brief  Binarise an image. Pixels with values above supplied threshold are set to 1.0, others are set to 0.0
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE BINARISE(IMG,THRESHOLD)
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(INOUT)   ::  IMG         !<  Image to be binarised (overwritten)
        REAL,           INTENT(IN)      ::  THRESHOLD   !<  Threshold. Pixels with values .GE. THRESHOLD are set to 1.0. Others are set to 0.0
        ! Private variables
        INTEGER ::  I,J,K
        ! Start work

#ifndef SKIP_RUNTIME_CHECKS
        ! Only meant for real images
        IF (.NOT. img%IsReal()) THEN
            WRITE(*,'(A)') '**ERROR(BINARISE): image type not supported: '
            CALL TERMINATE('Fatal error in BINARISE')
        ENDIF
#endif

        ! Loop over voxel values
        DO K=1,IMG%logical_dimensions(3)
            DO J=1,IMG%logical_dimensions(2)
                DO I=1,IMG%logical_dimensions(1)
                    IF (IMG%real_values(I,J,K)-THRESHOLD .GE. TINY(THRESHOLD)) THEN
                        IMG%real_values(I,J,K) = 1.0
                    ELSE
                        IMG%real_values(I,J,K) = 0.0
                    ENDIF
                ENDDO
            ENDDO
        ENDDO

    END SUBROUTINE BINARISE



    !>  \brief  Convolve two real images by mutliplying their FT
    SUBROUTINE FAST_CONVOLUTION(IMG_IN1,IMG_IN2,IMG_OUT)
        USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(INOUT)   ::  IMG_IN1
        TYPE(IMAGE),    INTENT(INOUT)   ::  IMG_IN2
        TYPE(IMAGE),    INTENT(INOUT)   ::  IMG_OUT
        ! Private variables

        ! Start work
        ! Check both images are real
        IF (.NOT. (img_in1%IsSameTypeAs(img_in2) .AND. img_in1%is_in_real_space)) THEN
            WRITE(*,'(A)') '**ERROR(FAST_CONVOLUTION): invalid image types'
            CALL TERMINATE('Fatal error in FAST_CONVOLUTION')
        ENDIF
        ! Check both images have same dimensions
        IF (.NOT. img_in1%HasSameDimsAs(img_in2)) THEN
            WRITE(*,'(A,2(3(I3,1X),A))') '**ERROR(FAST_CONVOLUTION): input images do not have same dimensions: ', &
                                        IMG_IN1%logical_dimensions, ' & ', IMG_IN2%logical_dimensions
            CALL TERMINATE('Fatal error in FAST_CONVOLUTION')
        ENDIF
        ! Allocate output image if necessary
        IF (.NOT. (img_in1%HasSameDimsAs(img_out) .AND. img_out%IsAllocated())) THEN
            call img_out%Allocate(mould=img_in1)
        ENDIF
        ! Prepare the output image, which will initially be complex
        IMG_OUT%is_in_real_space = .false.
        IMG_OUT%ORIGIN_AT_CENTER = .FALSE.
        ! Fourier transform both inputs
        call img_in1%ForwardFT(scale=.true.)
        call img_in2%ForwardFT(scale=.true.)
        ! Multiply
        IMG_OUT%complex_values = IMG_IN1%complex_values * IMG_IN2%complex_values / &
                                REAL(IMG_OUT%logical_dimensions(1) * IMG_OUT%logical_dimensions(2) * IMG_OUT%logical_dimensions(3))
        ! Center
        CALL SHIFT_REAL_SPACE_ORIGIN(IMG_OUT)
        ! Reverse FT the output
        call img_out%BackwardFT(scale=.true.)
    END SUBROUTINE FAST_CONVOLUTION



    !>  \brief  Impose of soft cylindrical mask on image
    !!  \todo   Make this subroutine more versatile, so it can deal with fractional radius better, and do softness better (see eg. MOD_FILTERS)
    SUBROUTINE CYL_MASK(IMG,RAD,SOFT,RAD_IN,MASK_WITH_AVERAGE_INSIDE,ZERO_FLOAT_INSIDE)
        USE INTERPOLATION    ! We need this for the Cosine / Hann fall off function
        USE IMAGES
        IMPLICIT NONE
        TYPE(IMAGE),    INTENT(INOUT)   ::  IMG     !<  Image to be masked
        REAL,           INTENT(IN)      ::  RAD     !<  Radius of cylinder (pixels)
        INTEGER,        INTENT(IN)      ::  SOFT    !<  Width of soft fall-off
        REAL, OPTIONAL, INTENT(IN)      ::  RAD_IN  !<  Inner radius (pixels). If not zero, a hollow cylinder is calculated
        LOGICAL, OPTIONAL, INTENT(IN)   ::  MASK_WITH_AVERAGE_INSIDE    !<  Mask value is average of densities within the cylinder
        LOGICAL, OPTIONAL, INTENT(IN)   ::  ZERO_FLOAT_INSIDE
        ! Private variables
        INTEGER ::  I,J,K
        REAL    ::  RAD_TMP, RAD_START, RAD_IN_START
        REAL    ::  X,Y
        INTEGER ::  ORIGIN(3)
		REAL	::	RRAD_IN
		REAL(KIND=8)    ::  MASK_VALUE
		LOGICAL ::  MMASK_WITH_AVERAGE_INSIDE
		LOGICAL  :: ZZERO_FLOAT_INSIDE
		REAL(KIND=8) :: VOXEL_COUNT


		RRAD_IN = 0.0
		IF (PRESENT(RAD_IN)) RRAD_IN = RAD_IN

        RAD_START = RAD-SOFT/2.0
        RAD_IN_START = RRAD_IN-SOFT/2.0

        if (img%is_in_real_space .and. img%origin_at_center) then
            ORIGIN = img%physical_address_of_box_center
        else
            ORIGIN = [1,1,1]
        endif


        MMASK_WITH_AVERAGE_INSIDE = .FALSE.
        IF (PRESENT(MASK_WITH_AVERAGE_INSIDE)) MMASK_WITH_AVERAGE_INSIDE = MASK_WITH_AVERAGE_INSIDE

        ZZERO_FLOAT_INSIDE = .FALSE.
        IF (PRESENT(ZERO_FLOAT_INSIDE)) ZZERO_FLOAT_INSIDE = ZERO_FLOAT_INSIDE

        ! Determine mask value
        IF (MMASK_WITH_AVERAGE_INSIDE .OR. ZZERO_FLOAT_INSIDE) THEN
            MASK_VALUE = 0.0D0
            VOXEL_COUNT = 0.0D0
            DO K=1,IMG%logical_dimensions(3)
            DO J=1,IMG%logical_dimensions(2)
                Y = REAL(J-ORIGIN(2))**2
                DO I=1,IMG%logical_dimensions(1)
                    X = REAL(I-ORIGIN(1))**2
                    RAD_TMP = SQRT(X+Y)
                    IF (RAD_TMP .GE. RAD_START) THEN
                        MASK_VALUE = MASK_VALUE + IMG%real_values(I,J,K) * HANN(RAD_TMP-RAD_START,SOFT)
                        VOXEL_COUNT = VOXEL_COUNT + HANN(RAD_TMP-RAD_START,SOFT)
                    ELSE IF (RAD_TMP .LE. RRAD_IN) THEN
                        IF (RAD_TMP .GE. RAD_IN_START) THEN
                            MASK_VALUE = MASK_VALUE + IMG%real_values(I,J,K) * (HANN(RAD_TMP-RAD_IN_START,SOFT)-1.0)*(-1.0)
                            VOXEL_COUNT = VOXEL_COUNT + (HANN(RAD_TMP-RAD_IN_START,SOFT)-1.0)*(-1.0)
                        ELSE
                            !
                        ENDIF
                    ENDIF
                ENDDO
            ENDDO
            ENDDO
            MASK_VALUE = MASK_VALUE / VOXEL_COUNT
        ELSE
            MASK_VALUE = 0.0E0
        ENDIF

        IF (ZZERO_FLOAT_INSIDE) THEN
            IMG%real_values = IMG%real_values - MASK_VALUE
            MASK_VALUE = 0.0D0
        ENDIF


        DO J=1,IMG%logical_dimensions(2)
            Y = REAL(J-ORIGIN(2))**2
            DO I=1,IMG%logical_dimensions(1)
                X = REAL(I-ORIGIN(1))**2
                RAD_TMP = SQRT(X+Y)
                IF (RAD_TMP .GE. RAD_START) THEN
                    IMG%real_values(I,J,1:IMG%logical_dimensions(3)) = IMG%real_values(I,J,1:IMG%logical_dimensions(3)) * &
                                                                            HANN(RAD_TMP-RAD_START,SOFT) + &
                                                                            (HANN(RAD_TMP-RAD_START,SOFT)-1.0)*(-1.0)*MASK_VALUE
                ELSE IF (RAD_TMP .LE. RRAD_IN) THEN
                    IF (RAD_TMP .GE. RAD_IN_START) THEN
                        IMG%real_values(I,J,1:IMG%logical_dimensions(3)) = IMG%real_values(I,J,1:IMG%logical_dimensions(3)) * &
                                                                            (HANN(RAD_TMP-RAD_IN_START,SOFT)-1.0)*(-1.0) + &
                                                                            HANN(RAD_TMP-RAD_IN_START,SOFT) * MASK_VALUE
                    ELSE
                        IMG%real_values(I,J,1:IMG%logical_dimensions(3)) = MASK_VALUE
                    ENDIF
                ENDIF
            ENDDO
        ENDDO

    END SUBROUTINE CYL_MASK

    !>  \brief  Impose a sharp circular mask on image. Pixels at a distance .GE. RAD from the origin will be set to 0.0
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE CIRC_MASK(IMG, RAD,INVERSE)
        USE IMAGES
        IMPLICIT NONE
        TYPE(IMAGE),    INTENT(INOUT)   ::  IMG
        REAL,           INTENT(IN)      ::  RAD !< Any voxels at this distance or greater from origin will be zeroed. In pixels (for real images) or reciprocal pixels (for complex images; so that 0.5 is Nyquist)
        LOGICAL,    OPTIONAL,   INTENT(IN)  ::  INVERSE !<  Invert the mask (center is masked out
        ! Private variables
        INTEGER ::  I,J,K
        REAL    ::  COOS(3)
        REAL    ::  FREQ
        INTEGER ::  ORIGIN(3)
        INTEGER ::  X,Y,Z
        REAL    ::  RAD_TMP, RAD_SQ
        LOGICAL :: IINVERSE
        ! Start work
#ifndef SKIP_RUNTIME_CHECKS
        IF (.not. img%is_in_real_space) THEN
            IF (RAD .GT. 0.5) THEN
                WRITE(*,'(A,F0.3)') '**ERROR(CIRC_MASK): masking complex image with radius > 0.5, i.e. beyond Nyquist: ', RAD
                CALL TERMINATE('Fatal error in CIRC_MASK')  ! Remove this eventually!
            ENDIF
        ENDIF
#endif

        IINVERSE = .FALSE.
        IF (PRESENT(INVERSE)) IINVERSE = INVERSE

        if (img%is_in_real_space .and. img%origin_at_center) then
            ORIGIN = img%physical_address_of_box_center
        else
            ORIGIN = [1,1,1]
        endif

        RAD_SQ = RAD**2

        IF (img%IsReal()) THEN
            DO K=1,IMG%logical_dimensions(3)
                Z = (K-ORIGIN(3))**2
                DO J=1,IMG%logical_dimensions(2)
                    Y = (J-ORIGIN(2))**2
                    DO I=1,IMG%logical_dimensions(1)
                        X = (I-ORIGIN(1))**2
                        RAD_TMP = REAL(X+Y+Z)
                        IF (IINVERSE) THEN
                            IF (RAD_TMP .LE. RAD_SQ) THEN
                                IMG%real_values(I,J,K) = 0.0
                            ENDIF
                        ELSE
                            IF (RAD_TMP .GE. RAD_SQ) THEN
                                IMG%real_values(I,J,K) = 0.0
                            ENDIF
                        ENDIF
                    ENDDO
                ENDDO
            ENDDO
        ELSE
            DO K=1,IMG%logical_dimensions(3)
                ! The distance from the origin to the current pixel, in reciprocal pixel units
                COOS(3) = COMP_COO_FROM_PHYS_INDEX_DIM2OR3(IMG%logical_dimensions(3),K)**2
                DO J=1,IMG%logical_dimensions(2)
                    COOS(2) = COMP_COO_FROM_PHYS_INDEX_DIM2OR3(IMG%logical_dimensions(2),J)**2
                    DO I=1,HIGHEST_INDEX_DIM1(IMG%logical_dimensions(1))
                        COOS(1) = COMP_COO_FROM_PHYS_INDEX_DIM1(IMG%logical_dimensions(1),I)**2
                        ! Compute frequency, in units of reciprocal pixels
                        FREQ = SQRT(SUM(COOS))
                        IF (IINVERSE) THEN
                            IF (FREQ .LE. RAD) THEN
                                IMG%complex_values(I,J,K) = (0.0,0.0)
                            ENDIF
                        ELSE
                            IF (FREQ .GE. RAD) THEN
                                IMG%complex_values(I,J,K) = (0.0,0.0)
                            ENDIF
                        ENDIF
                    ENDDO
                ENDDO
            ENDDO
        ENDIF

    END SUBROUTINE CIRC_MASK

    FUNCTION MULTIPLY_IMAGES_FUNC(IMG1,IMG2) RESULT(IMGOUT)
        IMPLICIT NONE
        TYPE(IMAGE), INTENT(IN)     ::  IMG1, IMG2
        TYPE(IMAGE) ::  IMGOUT
        CALL MULTIPLY_IMAGES(IMG1,IMG2,IMGOUT)
    END FUNCTION MULTIPLY_IMAGES_FUNC

    FUNCTION DIVIDE_IMAGES_FUNC(IMG1,IMG2) RESULT(IMGOUT)
        IMPLICIT NONE
        TYPE(IMAGE), INTENT(IN)     ::  IMG1, IMG2
        TYPE(IMAGE) ::  IMGOUT
        CALL DIVIDE_IMAGES(IMG1,IMG2,IMGOUT)
    END FUNCTION DIVIDE_IMAGES_FUNC

    FUNCTION ADD_IMAGES_FUNC(IMG1,IMG2) RESULT(IMGOUT)
        IMPLICIT NONE
        TYPE(IMAGE), INTENT(IN)     ::  IMG1, IMG2
        TYPE(IMAGE) ::  IMGOUT
        CALL ADD_IMAGES(IMG1,IMG2,IMGOUT)
    END FUNCTION ADD_IMAGES_FUNC

    !>  \brief  Multiply pixel values in an image by a constant
    SUBROUTINE IMAGE_MULTIPLY_BY_CONSTANT(SELF,CONSTANT)
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(INOUT)   ::  SELF    !<  Image object whose pixel/voxel values are to be multiplied
        REAL,           INTENT(IN)      ::  CONSTANT
        ! Private variables

        ! Start work
        IF (self%is_in_real_space) THEN
            SELF%real_values = SELF%real_values * CONSTANT
        ELSE
            SELF%complex_values = SELF%complex_values * CONSTANT
        ENDIF
    END SUBROUTINE IMAGE_MULTIPLY_BY_CONSTANT

    !>  \brief Multiply IMG1 and IMG2, using real or complex arithmetics, depending on the type of the images.
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE MULTIPLY_IMAGES(IMG1,IMG2,IMGOUT)
        IMPLICIT NONE
        TYPE(IMAGE), INTENT(IN)     ::  IMG1,IMG2
        TYPE(IMAGE), INTENT(INOUT)  ::  IMGOUT

#ifndef SKIP_RUNTIME_CHECKS
        ! Check both images have the same type
        IF ( .NOT. IMG1%IsSameTypeAs(IMG2)) THEN
            WRITE(*,*) '**ERROR (MULTIPLY_IMAGES): cannot multiply images of different types'
            CALL TERMINATE('Fatal error in MULTIPLY_IMAGES')
        ENDIF

        ! Check both images have the same dimensions
        IF ( .NOT. IMG1%HasSameDimsAs(IMG2)) THEN
            WRITE(*,'(2(A,3(I4,1X)))') '**ERROR (MULTIPLY_IMAGES): cannot multiply images of different dimensions: ', &
                                        IMG1%logical_dimensions, ';',IMG2%logical_dimensions
            CALL TERMINATE('Fatal error in MULTIPLY_IMAGES')
        ENDIF

        ! Check both input images have memory allocated
        IF ( (.NOT. IMG1%IN_MEMORY) .OR. (.NOT. IMG2%IN_MEMORY)) THEN
            WRITE(*,*) '**ERROR (MULTIPLY_IMAGES): one or both input images do not have allocated memory'
            CALL TERMINATE('Fatal error in MULTIPLY_IMAGES')
        ENDIF
#endif

        ! Do the multiplication
        IF ( .not. IMG1%is_in_real_space) THEN
            IMGOUT%complex_values = IMG1%complex_values * IMG2%complex_values
        ELSE
            IMGOUT%real_values = IMG1%real_values * IMG2%real_values
        ENDIF
    END SUBROUTINE MULTIPLY_IMAGES

    !>  \brief  Apply a mask to an image, with option to level the mask so that it corresponds to the average of the image within the mask
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE IMAGE_APPLY_MASK(SELF,MASK,MASK_WITH_AVERAGE,MASK_WITH_AVERAGE_OUTSIDE)
        USE IMAGEFILES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(INOUT)   ::  SELF                !<  Image to be masked
        TYPE(IMAGE),                INTENT(IN)      ::  MASK                !<  Mask image. Expected to go from 0. to 1.
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  MASK_WITH_AVERAGE   !<  Whether the image should be masked with a value corresponding to its average within the mask
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  MASK_WITH_AVERAGE_OUTSIDE   !<  Whether the image should be masked with a value corresponding to its average outside the mask
        ! Private variables
        LOGICAL ::  MMASK_WITH_AVERAGE, MMASK_WITH_AVERAGE_OUTSIDE
        INTEGER ::  I,J,K
        REAL(KIND=8)    ::  MASK_VALUE, NUM_ELEMS
        ! Start work

#ifndef SKIP_RUNTIME_CHECKS
        ! Check the mask is kosher
        IF (.NOT. IMAGE_SUITABLE_AS_MASK(MASK)) THEN
            WRITE(*,'(A)') '**ERROR(IMAGE_APPLY_MASK): mask image is not a mask'
            CALL WRITE_TO_DISK(MASK,'dbg_mask_pb.mrc')
            CALL TERMINATE('Fatal error in IMAGE_APPLY_MASK')
        ENDIF
#endif

        ! Will we mask with average? (rather than 0.0)
        MMASK_WITH_AVERAGE = .TRUE.
        IF (PRESENT(MASK_WITH_AVERAGE)) MMASK_WITH_AVERAGE = MASK_WITH_AVERAGE
        MMASK_WITH_AVERAGE_OUTSIDE = .FALSE.
        IF (PRESENT(MASK_WITH_AVERAGE_OUTSIDE)) MMASK_WITH_AVERAGE_OUTSIDE = MASK_WITH_AVERAGE_OUTSIDE

#ifndef SKIP_RUNTIME_CHECKS
        IF (MMASK_WITH_AVERAGE .AND. MMASK_WITH_AVERAGE_OUTSIDE) then
            CALL TERMINATE('Fatal error in IMAGE_APPLY_MASK. Arguments incompatible.')
        endif
#endif


        IF (MMASK_WITH_AVERAGE) THEN
            ! Work out the average within the mask
            MASK_VALUE = 0.0D0
            NUM_ELEMS  = 0.0D0
            DO K=1,SELF%logical_dimensions(3)
                DO J=1,SELF%logical_dimensions(2)
                    DO I=1,SELF%logical_dimensions(1)
                        MASK_VALUE = MASK_VALUE + MASK%real_values(I,J,K) * SELF%real_values(I,J,K)
                        NUM_ELEMS  = NUM_ELEMS  + MASK%real_values(I,J,K)
                    ENDDO
                ENDDO
            ENDDO
            MASK_VALUE = MASK_VALUE / NUM_ELEMS
            ! Apply the mask
            SELF%real_values = SELF%real_values * MASK%real_values + (1.-MASK%real_values)*MASK_VALUE
        ELSE IF (MMASK_WITH_AVERAGE_OUTSIDE) then
            ! Work out the average within the mask
            MASK_VALUE = 0.0D0
            NUM_ELEMS  = 0.0D0
            DO K=1,SELF%logical_dimensions(3)
                DO J=1,SELF%logical_dimensions(2)
                    DO I=1,SELF%logical_dimensions(1)
                        MASK_VALUE = MASK_VALUE + (1.0-MASK%real_values(I,J,K)) * SELF%real_values(I,J,K)
                        NUM_ELEMS  = NUM_ELEMS  + (1.0-MASK%real_values(I,J,K))
                    ENDDO
                ENDDO
            ENDDO
            MASK_VALUE = MASK_VALUE / NUM_ELEMS
            ! Apply the mask
            SELF%real_values = SELF%real_values * MASK%real_values + (1.-MASK%real_values)*MASK_VALUE
        ELSE
            CALL MULTIPLY_IMAGES(SELF,MASK,SELF)
        ENDIF

    END SUBROUTINE IMAGE_APPLY_MASK

    !>  \brief  Check whether the image can be used as a mask
    LOGICAL FUNCTION IMAGE_SUITABLE_AS_MASK(SELF)
        ! Arguments
        TYPE(IMAGE),    INTENT(IN)  ::  SELF    !<  Image to check
        !

        ! Start work
        IMAGE_SUITABLE_AS_MASK = .TRUE.
        ! Mask cannot be blank
        IF      (SELF%IsBlank()) THEN
            IMAGE_SUITABLE_AS_MASK = .FALSE.
        ENDIF
        ! Mask can only have values between 0. and 1.
        IF      (ANY(SELF%real_values(1:SELF%logical_dimensions(1),1:SELF%logical_dimensions(2),1:SELF%logical_dimensions(3)) &
                                                                                                                    .GT. 1.0)) THEN
            IMAGE_SUITABLE_AS_MASK = .FALSE.
        ELSE IF (ANY(SELF%real_values(1:SELF%logical_dimensions(1),1:SELF%logical_dimensions(2),1:SELF%logical_dimensions(3)) &
                                                                                                                    .LT. 0.0)) THEN
            IMAGE_SUITABLE_AS_MASK = .FALSE.
        ENDIF

    END FUNCTION IMAGE_SUITABLE_AS_MASK

    !>  \brief Divides IMG1 by IMG2 using appropriate arithmetics
    SUBROUTINE DIVIDE_IMAGES(IMG1,IMG2,IMGOUT)
        IMPLICIT NONE
        TYPE(IMAGE), INTENT(IN)     ::  IMG1    !< Numerator
        TYPE(IMAGE), INTENT(IN)     ::  IMG2    !< Denominator
        TYPE(IMAGE), INTENT(INOUT)  ::  IMGOUT  !< Result

        INTEGER ::  ITEMP

        ! Check both images have the same type
        IF ( .NOT. IMG1%IsSameTypeAs(IMG2)) THEN
            WRITE(*,*) '**ERROR (DIVIDE_IMAGES): cannot divide images of different types'
            STOP 'Fatal error in DIVIDE_IMAGES'
        ENDIF

        ! Check both images have the same dimensions
        IF ( .NOT. IMG1%HasSameDimsAs(IMG2)) THEN
            WRITE(*,*) '**ERROR (DIVIDE_IMAGES): cannot divide images of different dimensions'
            STOP 'Fatal error in DIVIDE_IMAGES'
        ENDIF

        ! Check both input images have memory allocated
        IF ( (.NOT. IMG1%IN_MEMORY) .OR. (.NOT. IMG2%IN_MEMORY)) THEN
            WRITE(*,*) '**ERROR (DIVIDE_IMAGES): one or both input images do not have allocated memory'
            STOP 'Fatal error in DIVIDE_IMAGES'
        ENDIF

        ! Count the number of zeroes in IMG2
        IF ( .not. IMG2%is_in_real_space) THEN
            ITEMP = COUNT(IMG2%complex_values .EQ. (0.0,0.0))
        ELSE
            ITEMP = COUNT(IMG2%real_values(1:IMG2%logical_dimensions(1),:,:) .EQ. 0.0)
        ENDIF

        ! If there are any zeroes, crash out
        IF (ITEMP .GT. 0) THEN
            WRITE(*,'(A,I9,A,I3,A)')    '**ERROR: attempt to divide by an image with ', ITEMP, ' zeroes (',     &
                                        INT(100*REAL(ITEMP)/    &
                                        (IMG2%logical_dimensions(1)/2*IMG2%logical_dimensions(2)*IMG2%logical_dimensions(3))),'%)'
            CALL TERMINATE('Fatal error in DIVIDE_IMAGES')
        ENDIF

        ! Now we can do the actual division
        IF (.not. IMG2%is_in_real_space) THEN
            IMGOUT%complex_values = IMG1%complex_values / IMG2%complex_values
        ELSE
            IMGOUT%real_values(1:IMG2%logical_dimensions(1),:,:) = IMG1%real_values(1:IMG2%logical_dimensions(1),:,:) / &
                                                                   IMG2%real_values(1:IMG2%logical_dimensions(1),:,:)
        ENDIF
    END SUBROUTINE DIVIDE_IMAGES

    !>  \brief  The output image contains pixel values from input image 1 or input image 2, whichever is greater. It is therefore a "greater than" image, hence the name of this subroutine.
    SUBROUTINE GT_IMAGE(IMG1,IMG2,IMGOUT)
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(IN)      ::  IMG1
        TYPE(IMAGE),    INTENT(IN)      ::  IMG2
        TYPE(IMAGE),    INTENT(INOUT)   ::  IMGOUT
        ! Private variables

        ! Start work

        ! Input images must be of same type & real
        IF (.NOT. IMG1%IsSameTypeAs(IMG2)) THEN
            WRITE(*,'(A)') '**ERROR(GT_IMAGE): input images are of different types'
            CALL TERMINATE('Fatal error in GT_IMAGE')
        ENDIF
        IF (.not. IMG1%is_in_real_space) THEN
            WRITE(*,'(2A)') '**ERROR(GT_IMAGE): can only computer greater than ',   &
                            'image with real inputs'
            CALL TERMINATE('Fatal error in GT_IMAGE')
        ENDIF


        ! Input images must be have same dimensions
        IF (.NOT. IMG1%HasSameDimsAs(IMG2)) THEN
            WRITE(*,'(2(A,3(I5,1X)))') '**ERROR(GT_IMAGE): input image dimensions are not identical: ', &
                                        IMG1%logical_dimensions, ' and ', IMG2%logical_dimensions
            CALL TERMINATE('Fatal error in GT_IMAGE')
        ENDIF

        ! Output image allocation
        IF (IMGOUT%IN_MEMORY) THEN
            IF (.NOT. IMG1%HasSameDimsAs(IMGOUT)) THEN
                WRITE(*,'(2A)') '**WARNING(GT_IMAGE): output image was allocated with wrong dimensions. ',   &
                                'Programming error? Reallocating.'
                call imgout%Allocate(img1)
            ENDIF
        ELSE
            call imgout%Allocate(img1)
        ENDIF

        ! Compute output image
        WHERE(IMG1%real_values .GT. IMG2%real_values)
            IMGOUT%real_values = IMG1%real_values
        ELSEWHERE
            IMGOUT%real_values = IMG2%real_values
        END WHERE

    END SUBROUTINE GT_IMAGE

    !>  \brief Add values from IMG1 to those of IMG2, and return the result
    SUBROUTINE ADD_IMAGES(IMG1,IMG2,IMGOUT)


        IMPLICIT NONE

        TYPE(IMAGE), INTENT(IN)     ::  IMG1
        TYPE(IMAGE), INTENT(IN)     ::  IMG2
        TYPE(IMAGE), INTENT(INOUT)  ::  IMGOUT

        ! Check both images have the same type
        IF ( .NOT. IMG1%IsSameTypeAs(IMG2) ) THEN
            WRITE(*,*) '**WARNING: adding two images of different types '
        ENDIF

        ! Check both images have memory allocated
        IF ( .NOT. (IMG1%IN_MEMORY .AND. IMG2%IN_MEMORY) ) THEN
            WRITE(*,*) ' **ERROR: trying to add images which do not have allocated memory'
            STOP ' Fatal error in ADD_IMAGES'
        ENDIF

        IF (.NOT. IMG1%IN_MEMORY)       WRITE(*,*) ' **ERROR: IMG1 not allocated'
        IF (.NOT. IMG2%IN_MEMORY)       WRITE(*,*) ' **ERROR: IMG2 not allocated'
        IF (.NOT. IMGOUT%IN_MEMORY)     WRITE(*,*) ' **ERROR: IMGOUT not allocated'



        IMGOUT%real_values = IMG1%real_values + IMG2%real_values


    END SUBROUTINE ADD_IMAGES

END MODULE IMAGES_ARITHM
