!>  \brief  Module with interfaces to various subroutines to compare images
MODULE IMAGES_COMPARISON
    USE IMAGES
    USE IMAGES_ARITHM, ONLY : CIRC_MASK, BINARISE

    IMPLICIT NONE

    !>  \brief  This defines what a function which accumulates scores between images should look like
!   ABSTRACT INTERFACE
!       FUNCTION SCORE_ACCUMULAT0R(IMG1,IMG2,MIN_FREQ_RAD,MAX_FREQ_RAD,B_FACTOR,ABSOLUTE)
!           TYPE(IMAGE)                 ::  IMG1
!           TYPE(IMAGE)                 ::  IMG2
!           INTEGER,        OPTIONAL    ::  MIN_FREQ_RAD, MAX_FREQ_RAD, ABSOLUTE
!           REAL,           OPTIONAL    ::  B_FACTOR
!           REAL(KIND=8),   ALLOCATABLE ::  SCORE_ACCUMULAT0R
!       END FUNCTION SCORE_ACCUMULAT0R
!   END INTERFACE

    !>  \brief  This defines what a function which accumulates scores between images should look like
!   ABSTRACT INTERFACE
!       FUNCTION SCORE_FINALISER(NUM_DENOM_RINGS,IMG1,IMG2,MIN_FREQ_RAD,MAX_FREQ_RAD,B_FACTOR,ABSOLUTE)
!
!       END FUNCTION SCORE_FINALISER
!   END INTERFACE

    !>  \brief  A similarity measure object defines all that is needed to know to compare two images (or set of images...) and obtain a quantitative assessment of their similarity
    TYPE SIMILARITY_MEASURE
        LOGICAL ::  PER_FREQUENCY   =   .FALSE. !<  Some similarity measures discriminate between frequency bands (i.e. they are weighted / filtered), others (like the traditional, linear cross-correlation coefficient) do not
        REAL    ::  MIN_FREQ, MAX_FREQ
        LOGICAL ::  ARBITRARY_FILTER    =   .FALSE. !<  Whether some arbitrary filter should be used (for example, to up-weigh layer lines etc)

    END TYPE SIMILARITY_MEASURE

    TYPE RESTRAINT

    END TYPE RESTRAINT

	INTERFACE SCORE_GOT_WORSE
		MODULE PROCEDURE SCORE_GOT_WORSE_SINGLE
		MODULE PROCEDURE SCORE_GOT_WORSE_DOUBLE
	END INTERFACE

    CONTAINS

    !>  \brief  Do the two images have exactly the same values?
    pure logical function images_are_equal(img1,img2)
        implicit none
        ! Arguments
        type(image),    intent(in)  ::  img1,img2
        ! Start work
        images_are_equal =  &
            all(img1%real_values(1:img1%logical_dimensions(1),1:img1%logical_dimensions(2),1:img1%logical_dimensions(3)) .eq. &
                img2%real_values(1:img2%logical_dimensions(1),1:img2%logical_dimensions(2),1:img2%logical_dimensions(3)))
        images_are_equal =  images_are_equal .and. img1%HasSameDimsAs(img1) .and. img1%IsSameTypeAs(img2)
    end function images_are_equal

    !>  \brief  Has the score become significantly worse?
    PURE LOGICAL FUNCTION SCORE_GOT_WORSE_SINGLE(SCORE_START,SCORE_FINISH)
    	IMPLICIT NONE
		! Arguments
		REAL(KIND=4),   INTENT(IN)  ::  SCORE_START, SCORE_FINISH
		! Private variables
		REAL,   PARAMETER   ::  TOL_MUL = 0.01
		REAL,   PARAMETER   ::  TOL_ADD = 0.01
		REAL                ::  SCORE_START_MOD
		! Start work
		SCORE_START_MOD = MIN(SCORE_START-TOL_ADD,SCORE_START-ABS(TOL_MUL*SCORE_START))
		SCORE_GOT_WORSE_SINGLE = SCORE_FINISH .LT. SCORE_START_MOD
	END FUNCTION SCORE_GOT_WORSE_SINGLE

	PURE LOGICAL FUNCTION SCORE_GOT_WORSE_DOUBLE(SCORE_START,SCORE_FINISH)
		IMPLICIT NONE
		REAL(KIND=8),	INTENT(IN)	::	SCORE_START,SCORE_FINISH
		SCORE_GOT_WORSE_DOUBLE = SCORE_GOT_WORSE_SINGLE(REAL(SCORE_START,KIND=4),REAL(SCORE_FINISH,KIND=4))
	END FUNCTION SCORE_GOT_WORSE_DOUBLE


    !>  \brief  Prepare a 2D image for a similarity measure
    !!  \todo   Modify the envelope masking so that we mask to the average within the mask, not 0. Alternatively, make sure the projection is zero floated.
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE IMAGE_PREPARE(   IMG,IMG_TMP,CTF_IMAGE,OBJ_MAX_RAD,    &
                                SHIFT_X,SHIFT_Y,ENV_MASK,REAL_SPACE_SHIFT,   &
                                INTERPOL,INTERPOL_WINDOW,   &
                                APPLY_DISC_MASK)
        Use Images
        USE CTFS
        USE IMAGES_ARITHM
        USE IMAGES_CTF
        USE INTERPOLATION
        USE ROTATION_MATRICES
        use RUNTIME_PARAMETERS, only : DEFAULT_FALLOFF_PIXELS
#ifndef SKIP_RUNTIME_CHECKS
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
#endif
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                            INTENT(INOUT)   ::  IMG                 !<  2D image. Will be modified (shifted, CTF-corrected, masked...)
        TYPE(IMAGE),                            INTENT(INOUT)   ::  IMG_TMP             !<  2D work image. May be overwritten etc. Should be allocated to be same size as other images.
        TYPE(IMAGE),                            INTENT(IN)      ::  CTF_IMAGE           !<  CTF image
        REAL,                                   INTENT(IN)      ::  OBJ_MAX_RAD         !<  Maximum radius of the object (in pixels)
        REAL,                       OPTIONAL,   INTENT(IN)      ::  SHIFT_X             !<  Shift in X to be applied to 2D image
        REAL,                       OPTIONAL,   INTENT(IN)      ::  SHIFT_Y             !<  Shift in Y to be applied to 2D image
        TYPE(IMAGE),                OPTIONAL,   INTENT(IN)      ::  ENV_MASK            !<  A soft-edged, envelope mask to be applied to the image (usually obtained from a 3D projection)
        LOGICAL,                    OPTIONAL,   INTENT(IN)      ::  REAL_SPACE_SHIFT    !<  Whether the image will be shifted in real-space or in Fourier space (using a phase shift). Defaults to .FALSE.
        TYPE(INTERPOLATION_SCHEME), OPTIONAL,   INTENT(IN)      ::  INTERPOL            !<  Interpolation scheme to use for real-space shift.
        TYPE(WINDOW),               OPTIONAL,   INTENT(INOUT)   ::  INTERPOL_WINDOW     !<  Interpolation window.
        LOGICAL,                    OPTIONAL,   INTENT(IN)      ::  APPLY_DISC_MASK     !<  Whether a soft-edged disk aperture should be applied to the projection
        ! Private variables
        REAL                    ::  SSHIFT_X, SSHIFT_Y
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        LOGICAL                 ::  AAPPLY_DISC_MASK
        ! Private variables below are only used when doing real space shift
        LOGICAL                 ::  RREAL_SPACE_SHIFT
        REAL                    ::  MATRIX(4,4)
        INTEGER					::	I,J,K
		! Start work

        ! Real-space shift?
        IF (PRESENT(REAL_SPACE_SHIFT)) THEN
            RREAL_SPACE_SHIFT = REAL_SPACE_SHIFT
        ELSE
            RREAL_SPACE_SHIFT = .FALSE.
        ENDIF

#ifndef SKIP_RUNTIME_CHECKS
        ! If real space shifts, check interpolation scheme was supplied
        IF (RREAL_SPACE_SHIFT) THEN
            IF (.NOT. PRESENT(INTERPOL) .OR. .NOT. PRESENT(INTERPOL_WINDOW)) THEN
                WRITE(*,'(2A)') '**ERROR(IMAGE_PREPARE): real-space shift requires an',  &
                                ' interpolation scheme to be supplied as argument'
                CALL TERMINATE('Fatal error in IMAGE_PREPARE')
            ENDIF
        ENDIF
#endif

        ! Will we apply a disk mask?
        IF (PRESENT(APPLY_DISC_MASK)) THEN
            AAPPLY_DISC_MASK = APPLY_DISC_MASK
        ELSE
            AAPPLY_DISC_MASK = .TRUE.
        ENDIF

        ! The 2D image needs to be in reciprocal space
        IF (img%is_in_real_space) CALL FWD_FT(IMG)


        ! Apply the CTF to the image
        FORALL(I=1:MAXIMUM_COMPLEX_ADDRESS_DIM1(IMG),J=1:IMG%logical_dimensions(2),K=1:IMG%logical_dimensions(3))
			IMG%complex_values(I,J,K)  =   IMG%complex_values(I,J,K)  * CTF_IMAGE%real_values(I,J,K)
		END FORALL

        ! Apply mask to image to avoid artefacts during shifting
        IF (.not. img%is_in_real_space) CALL BWD_FT(IMG)
        CALL IMG%ApplyCosineDisc(REAL(IMG%logical_dimensions(1)/2)-6.0,6.0,WIPE_INPUT=.FALSE.,MASK_WITH_AVERAGE=.TRUE.)

        ! Apply (X,Y) shift to image
        IF (PRESENT(SHIFT_X) .OR. PRESENT(SHIFT_Y)) THEN
            ! Work out what the shift should be
            SSHIFT_X = 0.0
            SSHIFT_Y = 0.0
            IF (PRESENT(SHIFT_X)) SSHIFT_X = SHIFT_X
            IF (PRESENT(SHIFT_Y)) SSHIFT_Y = SHIFT_Y
            IF (SSHIFT_X .NE. 0.0 .OR. SSHIFT_Y .NE. 0.0) THEN
                ! Apply the shift, either in real space or Fourier space
                IF (RREAL_SPACE_SHIFT) THEN
                    MATRIX = IDENTITY_ROTMAT()
                    MATRIX(1,4) = -SSHIFT_X
                    MATRIX(2,4) = -SSHIFT_Y
                    IF (.not. img%is_in_real_space) CALL BWD_FT(IMG)
                    CALL ROTATE_IMAGE(IMG,IMG_TMP,MATRIX,INTERPOL,INTERPOL_WINDOW)
                    IMG%real_values = IMG_TMP%real_values
                ELSE
                    IF (img%is_in_real_space) CALL FWD_FT(IMG)
                    CALL PSHIFT_IMAGE(IMG,SSHIFT_X,SSHIFT_Y,0.0)
                ENDIF
            ENDIF
        ENDIF

        ! Apply envelope mask (calculated from the 3D) to the image, if the mask was supplied
        IF (PRESENT(ENV_MASK)) THEN
            IF (.not. img%is_in_real_space)  CALL BWD_FT(IMG)
#ifndef SKIP_RUNTIME_CHECKS
            IF (DEBUG) WRITE(*,'(A)') '**debug(IMAGE_PREPARE): applying envelope mask'
#endif
            CALL IMAGE_APPLY_MASK(IMG,ENV_MASK,MASK_WITH_AVERAGE=.FALSE.,MASK_WITH_AVERAGE_OUTSIDE=.TRUE.)
        ENDIF

        ! Apply a circular mask
        IF (AAPPLY_DISC_MASK) THEN
#ifndef SKIP_RUNTIME_CHECKS
            IF (DEBUG) PRINT '(A,F0.3,A,I0,A)', '**debug(IMAGE_PREPARE): obj_max_rad, box size = ', &
                                                 OBJ_MAX_RAD, ' , ',  IMG%logical_dimensions(1), ' pixels'
#endif
            IF (.not. img%is_in_real_space) CALL BWD_FT(IMG)
            CALL IMG%ApplyCosineDisc(OBJ_MAX_RAD+DEFAULT_FALLOFF_PIXELS,DEFAULT_FALLOFF_PIXELS,WIPE_INPUT=.FALSE.)
        ENDIF
    END SUBROUTINE IMAGE_PREPARE

    !>  \brief  Start the preparation of a projection from a 3D image. Only do the projection / slice extraction.
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE PROJECTION_PREPARE_RAW(  VOL,PROJ,ANGLES,INTERP,WIN,REAL_SPACE_PROJ,VOL_SHIFT_Z,FREQ_MAX,    &
                                        SHIFT_X,SHIFT_Y,    &
                                        HELICAL_SUBUNIT_ONLY,HELIX,PITCH)
        USE EULERANGLES
        USE IMAGES_THREED
        USE INTERPOLATION
        USE IMAGES
        USE HELICES, ONLY : HELICAL_SYMMETRY
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                        INTENT(IN)      ::  VOL                 !<  3D volume from which a projection will be computed
        TYPE(IMAGE),                        INTENT(INOUT)   ::  PROJ                !<  Computed projection
        TYPE(EULER),                        INTENT(IN)      ::  ANGLES              !<  Euler angles at which the projection should be computed
        TYPE(INTERPOLATION_SCHEME),         INTENT(IN)      ::  INTERP              !<  Interpolation scheme
        TYPE(WINDOW),           TARGET,     INTENT(INOUT)   ::  WIN                 !<  Interpolating window
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  REAL_SPACE_PROJ     !<  Whether the projection should be computed in real space
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  VOL_SHIFT_Z         !<  The volume will be shifted along its Z axis before projection by this amount (in pixels).
        REAL,                   OPTIONAL,   INTENT(IN)      ::  FREQ_MAX            !<  Maximum frequency (0.5 is Nyquist) of interest
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  SHIFT_X             !<  Output projection is shifted by this amount in first dimension (pixels)
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  SHIFT_Y             !<  Output projection is shifted by this amount in second dimension (pixels)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  HELICAL_SUBUNIT_ONLY!<  Only the helical subunit is projected
        TYPE(HELICAL_SYMMETRY), OPTIONAL,   INTENT(IN)      ::  HELIX               !<  Helical operator (only needed when helical_subunit_only)
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  PITCH               !<  Helical pitch (only needed when helical_subunit_only)
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        LOGICAL                 ::  RREAL_SPACE_PROJ
        REAL(KIND=8)            ::  FFREQ_MAX
        ! Start work

        ! Real space projection?
        IF (PRESENT(REAL_SPACE_PROJ)) THEN
            RREAL_SPACE_PROJ = REAL_SPACE_PROJ
        ELSE
            RREAL_SPACE_PROJ = .FALSE.
        ENDIF

        ! Maximum frequency
        IF (PRESENT(FREQ_MAX)) THEN
            FFREQ_MAX = DBLE(FREQ_MAX)
#ifndef SKIP_RUNTIME_CHECKS
            IF (RREAL_SPACE_PROJ) THEN
                WRITE(*,'(2A)') '**warning(PROJECTION_PREPARE_RAW): maximum frequency not implemented',  &
                                ' for real-space projection. Ignored.'
                !CALL TERMINATE('Fatal error in PROJECTION_PREPARE_RAW')
            ENDIF
#endif
        ELSE
            FFREQ_MAX = 0.5D0
        ENDIF

#ifndef SKIP_RUNTIME_CHECKS
        IF (DEBUG) WRITE(*,'(A,L1)') '**debug(PROJECTION_PREPARE_RAW): real space projection? ', RREAL_SPACE_PROJ
#endif

        ! Get a projection from the volume, either via real space projection or Fourier space slice extraction
        IF (RREAL_SPACE_PROJ) THEN
#ifndef SKIP_RUNTIME_CHECKS
            IF (.NOT. VOL%ORIGIN_AT_CENTER) THEN
                WRITE(*,'(A)') '**ERROR(PROJECTION_PREPARE_RAW): input volume must have origin at center'
                CALL TERMINATE('Fatal error in PROJECTION_PREPARE_RAW')
            ENDIF
            IF (.not. vol%is_in_real_space) THEN
                WRITE(*,'(A)') '**ERROR(PROJECTION_PREPARE_RAW): input volume must be in real space'
                CALL TERMINATE('Fatal error in PROJECTION_PREPARE_RAW')
            ENDIF
#endif
            CALL VOLUME_PROJECT_LEAN(VOL,PROJ,ANGLES%PHI,ANGLES%THETA,ANGLES%PSI,INTERP,WIN,SUPER_LEAN=.FALSE.,   &
                                    SHIFT_X=SHIFT_X,SHIFT_Y=SHIFT_Y,Z_SHIFT=VOL_SHIFT_Z,    &
                                    HELICAL_SUBUNIT_ONLY=HELICAL_SUBUNIT_ONLY,HELIX=HELIX,PITCH=PITCH)
        ELSE
            ! Extract a slice from the 3D volume
#ifndef SKIP_RUNTIME_CHECKS
            IF (vol%is_in_real_space) THEN
                WRITE(*,'(A)') '**ERROR(PROJECTION_PREPARE_RAW): input volume must be in Fourier space'
                CALL TERMINATE('Fatal error in PROJECTION_PREPARE_RAW')
            ENDIF

            IF (VOL%ORIGIN_AT_CENTER) THEN
                WRITE(*,'(A)') '**ERROR(PROJECTION_PREPARE_RAW): input volume must not have origin at center'
                CALL TERMINATE('Fatal error in PROJECTION_PREPARE_RAW')
            ENDIF
#endif
            CALL EXTRACT_SLICE( VOL,PROJ,ANGLES%PHI,ANGLES%THETA,ANGLES%PSI,INTERP,WIN,Z_SHIFT=VOL_SHIFT_Z,&
                                RAD_MAX=DBLE(VOL%logical_dimensions(1))*FFREQ_MAX,X_SHIFT=SHIFT_X,Y_SHIFT=SHIFT_Y)
#ifndef SKIP_RUNTIME_CHECKS
            ! DEBUG
            IF (DEBUG) THEN
                IF (PROJ%IsComplex()) CALL PROJ%BackwardFT()
                CALL WRITE_TO_DISK(PROJ,'dbg_proj_beforeoriginshift.mrc')
            ENDIF
#endif
            ! Apply shift to the projection (to move origin to center)
            CALL SHIFT_REAL_SPACE_ORIGIN(PROJ)
#ifndef SKIP_RUNTIME_CHECKS
            ! DEBUG
            IF (DEBUG) THEN
                IF (PROJ%IsComplex()) CALL PROJ%BackwardFT()
                CALL WRITE_TO_DISK(PROJ,'dbg_proj_afteroriginshift.mrc')
            ENDIF
#endif
        ENDIF

    END SUBROUTINE PROJECTION_PREPARE_RAW

    !>  \brief  Compute an envelope mask from a projection
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE ENV_MASK_FROM_PROJ(PROJ,ENV_MASK,REAL_SPACE_PROJ,PROJ_MAX_RAD,RECT_DIM,RECT_AZIMUTH)
        USE MASKS
        USE RUNTIME_PARAMETERS, only : DEFAULT_FALLOFF_PIXELS
        USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),            INTENT(IN)      ::  PROJ            !<  Projection image
        TYPE(IMAGE),            INTENT(INOUT)   ::  ENV_MASK        !<  Envelope mask
        LOGICAL,                INTENT(IN)      ::  REAL_SPACE_PROJ !<  Whether the projection was calculated in real space
        REAL,                   INTENT(IN)      ::  PROJ_MAX_RAD    !<  In pixels, the maximum radius of interest
        REAL,       OPTIONAL,   INTENT(IN)      ::  RECT_DIM(2)     !<  Pixels outside this rectangle will be ignored
        REAL,       OPTIONAL,   INTENT(IN)      ::  RECT_AZIMUTH    !<  Azimuth for the rectangle
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        REAL                    ::  ENV_MASK_THRESH
        REAL,       PARAMETER   ::  SOFT_EDGE       =   DEFAULT_FALLOFF_PIXELS  !<  The width of the soft-edge falloff for masking
        ! Start work
        ENV_MASK%real_values = PROJ%real_values


        IF (ENV_MASK%is_in_real_space) CALL FWD_FT(ENV_MASK)
        CALL ENV_MASK%ApplyBandPassFilter(0.0,0.05,COSINE_FALLOFF=0.0)
        IF (.not. ENV_MASK%is_in_real_space) CALL BWD_FT(ENV_MASK)
#ifndef SKIP_RUNTIME_CHECKS
        IF (DEBUG) CALL WRITE_TO_DISK(ENV_MASK,'dbg_env_mask_lp.mrc')
#endif
        IF (PRESENT(RECT_DIM)) THEN
#ifndef SKIP_RUNTIME_CHECKS
            IF (.NOT. PRESENT(RECT_AZIMUTH)) THEN
                WRITE(*,'(A)') '**ERROR(ENV_MASK_FROM_PROJ): rect dim, but not azimuth, supplied.'
                CALL TERMINATE('Fatal error in ENV_MASK_FROM_PROJ. Programming error.')
            ENDIF
#endif
            IF (ENV_MASK%IsComplex()) CALL ENV_MASK%BackwardFT()
            CALL ENV_MASK%ApplyRectMask(RECT_DIM=RECT_DIM,AZIMUTH=RECT_AZIMUTH,WIPE_INPUT=.FALSE.)
#ifndef SKIP_RUNTIME_CHECKS
            IF (DEBUG) CALL WRITE_TO_DISK(ENV_MASK,'dbg_env_mask_lp_rectmasked.mrc')
#endif
        ENDIF
        CALL CIRC_MASK(ENV_MASK,PROJ_MAX_RAD+1.)
#ifndef SKIP_RUNTIME_CHECKS
        IF (DEBUG) CALL WRITE_TO_DISK(ENV_MASK,'dbg_env_mask_lp_disc.mrc')
#endif
        IF (REAL_SPACE_PROJ) THEN
            ENV_MASK_THRESH = 10.*PROJ%logical_dimensions(1)*TINY(1.0)
        ELSE
            ENV_MASK_THRESH = SQRT(ENV_MASK%ReturnVariance())
        ENDIF
        CALL BINARISE(ENV_MASK,ENV_MASK_THRESH)
        CALL CIRC_MASK(ENV_MASK,PROJ_MAX_RAD+1.)
        CALL MASK_ADD_COSINE_FALLOFF(ENV_MASK,SOFT_EDGE)
#ifndef SKIP_RUNTIME_CHECKS
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(PROJECTION_PREPARE): dumping envelope mask to dbg_env_mask.mrc'
            CALL WRITE_TO_DISK(ENV_MASK,'dbg_env_mask.mrc')
        ENDIF
#endif

    END SUBROUTINE ENV_MASK_FROM_PROJ

    !>  \brief  Prepare a projection from a 3D image for a similarity measure
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE PROJECTION_PREPARE(  VOL,PROJ,PROJ_DIM,ANGLES,CTF_IMAGE,PROJ_MAX_RAD,SHIFT_X,SHIFT_Y,INTERPOL,WIN,    &
                                    ENV_MASK,REAL_SPACE_PROJ,VOL_SHIFT_Z,APPLY_DISC_MASK,   &
                                    !PROJ_CACHE, &
                                    FREQ_MAX,&
                                    CTF_SQUARED,MASK_WITH_ZEROS,APPLY_CTF,HELICAL_SUBUNIT_ONLY,HELIX,PITCH,RECT_MASK_WIDTH)
        USE EULERANGLES
        USE INTERPOLATION
        USE IMAGES
        USE MASKS
        USE IMAGES_CTF
        USE PROJECTION_CACHES
        USE HELICES, ONLY : HELICAL_SYMMETRY
        USE RUNTIME_PARAMETERS, only  : DEFAULT_FALLOFF_PIXELS
        USE IMAGES_ARITHM, only : IMAGE_APPLY_MASK
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                        INTENT(IN)      ::  VOL                 !<  3D volume from which a projection image will be computed
        TYPE(IMAGE),                        INTENT(INOUT)   ::  PROJ                !<  Output projection image
        INTEGER,                            INTENT(IN)      ::  PROJ_DIM(3)         !<  Desired dimensions for the projection
        TYPE(EULER),                        INTENT(IN)      ::  ANGLES              !<  Euler angles at which the projection should be computed
        TYPE(IMAGE),                        INTENT(IN)      ::  CTF_IMAGE           !<  CTF parameters
        REAL,                               INTENT(IN)      ::  PROJ_MAX_RAD        !<  Maximum radius of the projection (pixels). Pixels beyond this radius can be set to zero, for example.
        REAL(KIND=8),                       INTENT(IN)      ::  SHIFT_X             !<  The projection will be shifted by this amount (pixels) in the first dimension
        REAL(KIND=8),                       INTENT(IN)      ::  SHIFT_Y             !<  Shift applied to the projection in the second dimension
        TYPE(INTERPOLATION_SCHEME),         INTENT(IN)      ::  INTERPOL            !<  Interpolation scheme
        TYPE(WINDOW),           TARGET,     INTENT(INOUT)   ::  WIN                 !<  Interpolating window
        TYPE(IMAGE),            OPTIONAL,   INTENT(INOUT)   ::  ENV_MASK            !<  If this argument is given, a envelope mask will be calculated from the projection. This can later be used to mask a noisy image.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  REAL_SPACE_PROJ     !<  Whether to do real space projection. By default, this is false and Fourier space slice extraction is done instead.
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  VOL_SHIFT_Z         !<  The volume will be shifted along its Z axis before projection by this amount (in pixels).
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  APPLY_DISC_MASK     !<  Whether a soft-edged disk aperture should be applied to the projection
        !TYPE(PROJECTION_CACHE), OPTIONAL,   INTENT(INOUT)   ::  PROJ_CACHE          !<  If present, the projection will be added to this cache
        REAL,                   OPTIONAL,   INTENT(IN)      ::  FREQ_MAX            !<  Maximum frequency (0.5 is Nyquist) of interest
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  CTF_SQUARED         !<  For testing purposes, one can multiply by the CTF rather than the CTF^2
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  MASK_WITH_ZEROS     !<  For testing purposes, can mask with 0.0 rather than average value. May be broken. Default is to mask with 0.0.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  APPLY_CTF           !<  For testing purposes, the CTF can be switched off.
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  HELICAL_SUBUNIT_ONLY!<  Only the helical subunit is projected
        TYPE(HELICAL_SYMMETRY), OPTIONAL,   INTENT(IN)      ::  HELIX               !<  Helical operator (only needed when helical_subunit_only)
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  PITCH               !<  Helical pitch (only needed when helical_subunit_only)
        REAL,                   OPTIONAL,   INTENT(IN)      ::  RECT_MASK_WIDTH     !<  Useful when projecting helical structure. The width of the rectangle mask to be applied
        ! Private variables
        REAL,       PARAMETER   ::  SOFT_EDGE       =   DEFAULT_FALLOFF_PIXELS  !<  The width of the soft-edge falloff for masking
        LOGICAL,    PARAMETER   ::  DEBUG           =   .FALSE.
        LOGICAL                 ::  RREAL_SPACE_PROJ
        LOGICAL                 ::  AAPPLY_DISC_MASK
        !INTEGER                 ::  PROJ_POS_IN_CACHE
        LOGICAL                 ::  COMPUTE_PROJECTION
        LOGICAL                 ::  CCTF_SQUARED
        LOGICAL                 ::  AAPPLY_CTF
        LOGICAL,    PARAMETER   ::  AAPPLY_ENV_MASK =   .TRUE.
        INTEGER					::	I,J,K
		! Start work

        ! Real space projection?
        RREAL_SPACE_PROJ = .FALSE.
        IF (PRESENT(REAL_SPACE_PROJ))  RREAL_SPACE_PROJ = REAL_SPACE_PROJ

        ! Will we apply a disk mask?
        AAPPLY_DISC_MASK = .TRUE.
        IF (PRESENT(APPLY_DISC_MASK)) AAPPLY_DISC_MASK = APPLY_DISC_MASK

        ! Will we apply the CTF to the projection
        AAPPLY_CTF = .TRUE.
        IF (PRESENT(APPLY_CTF)) AAPPLY_CTF = APPLY_CTF

        ! Squared CTF?
        CCTF_SQUARED = .TRUE.
        IF (PRESENT(CTF_SQUARED)) CCTF_SQUARED = CTF_SQUARED

#ifndef SKIP_RUNTIME_CHECKS
        IF (DEBUG) WRITE(*,'(A,L1)') '**debug(PROJECTION_PREPARE): real space projection? ', RREAL_SPACE_PROJ

        ! Check the CTF image has the correct dimensions
        IF (.NOT. PROJ%HasSameDimsAs(CTF_IMAGE)) THEN
            WRITE(*,'(A)') '**ERROR(PROJECTION_PREPARE): projection and CTF images don''t have same dims'
            CALL IMAGE_PRINT_INFO(PROJ)
            CALL IMAGE_PRINT_INFO(CTF_IMAGE)
            CALL TERMINATE('Fatal error in PROJECTION_PREPARE')
        ENDIF
#endif

        COMPUTE_PROJECTION = .TRUE.
        !IF (PRESENT(PROJ_CACHE)) THEN
        !    ! Check whether the projection has already been computed and is stored in cache
        !    PROJ_POS_IN_CACHE = PROJECTION_CACHE_FIND_PROJ(PROJ_CACHE,ANGLES,REAL(VOL_SHIFT_Z),CTF)
        !    IF (PROJ_POS_IN_CACHE .GT. 0) COMPUTE_PROJECTION = .FALSE.
        !ENDIF

        !IF (.NOT. COMPUTE_PROJECTION) THEN
        !    ! A similar-enough projection was already computed, so just fetch it
        !    IF (DEBUG) WRITE(*,'(A)') '**debug(PROJECTION_PREPARE): using cached projection'
        !    CALL PROJECTION_CACHE_RETRIEVE_PROJ(PROJ_CACHE,PROJ_POS_IN_CACHE,PROJ,ENV_MASK)
        !ELSE
            !
            ! We need to compute the projection (and when it's ready, add it to the cache if we were given one)
            !

#ifndef SKIP_RUNTIME_CHECKS
            IF (DEBUG) THEN
                write(*,'(a,3(i0,1x))') '**debug(PROJECTION_PREPARE): projection dim dummy argument: ', PROJ_DIM
                WRITE(*,'(A)') '**debug(PROJECTION_PREPARE): VOL info before prepare_raw:'
                CALL IMAGE_PRINT_INFO(VOL)
                WRITE(*,'(A)') '**debug(PROJECTION_PREPARE): PROJ info before prepare_raw:'
                CALL IMAGE_PRINT_INFO(PROJ)
            ENDIF
#endif
            CALL PROJECTION_PREPARE_RAW(VOL,PROJ,ANGLES,INTERPOL,WIN,REAL_SPACE_PROJ,VOL_SHIFT_Z=VOL_SHIFT_Z,FREQ_MAX=FREQ_MAX, &
                                        SHIFT_X=SHIFT_X,SHIFT_Y=SHIFT_Y,&
                                        HELICAL_SUBUNIT_ONLY=HELICAL_SUBUNIT_ONLY,HELIX=HELIX,PITCH=PITCH)

#ifndef SKIP_RUNTIME_CHECKS
            IF (DEBUG) THEN
                WRITE(*,'(A)') '**debug(PROJECTION_PREPARE): PROJ info after prepare_raw:'
                CALL IMAGE_PRINT_INFO(PROJ)
            ENDIF

            IF (DEBUG) THEN
                IF (PROJ%IsBlank()) THEN
                    WRITE(*,'(A)') '**ERROR(PROJECTION_PREPARE): projection is blank'
                    CALL TERMINATE('Fatal error in PROJECTION_PREPARE')
                ENDIF
            ENDIF

            IF (DEBUG) THEN
                IF (.not. proj%is_in_real_space) CALL BWD_FT(PROJ)
                CALL WRITE_TO_DISK(PROJ,'dbg_proj_raw.mrc')
                CALL IMAGE_PRINT_INFO(PROJ)
                CALL FWD_FT(PROJ)
            ENDIF
#endif

            ! Resize projection if necessary
            IF (.not. proj%is_in_real_space) CALL BWD_FT(PROJ)
#ifndef SKIP_RUNTIME_CHECKS
            IF (DEBUG) THEN
                IF (PROJ%IsBlank()) THEN
                    WRITE(*,'(A)') '**ERROR(PROJECTION_PREPARE): projection is blank'
                    CALL TERMINATE('Fatal error in PROJECTION_PREPARE')
                ENDIF
            ENDIF
            IF (ANY(PROJ%logical_dimensions .NE. PROJ_DIM)) THEN
                WRITE(*,'(A,3(I0,1X),A,3(I0,1X),A)')    '**WARNING(PROJECTION_PREPARE): projection (', PROJ%logical_dimensions, &
                                                    ') will be resized to  ', PROJ_DIM, ', which will probably lead to artefacts'
            ENDIF
            CALL RESIZE(PROJ,PROJ_DIM)
            IF (DEBUG)  CALL WRITE_TO_DISK(PROJ,'dbg_proj_beforectf_resize.mrc')
#endif

            ! If we're going to need it later, calculate an envelope mask from the projection,
            ! before applying the CTF
            IF (PRESENT(ENV_MASK)) THEN
                IF (PRESENT(RECT_MASK_WIDTH)) THEN
                    CALL ENV_MASK_FROM_PROJ(PROJ,ENV_MASK,RREAL_SPACE_PROJ,PROJ_MAX_RAD,    &
                                        RECT_DIM=(/REAL(PROJ%logical_dimensions(1)),RECT_MASK_WIDTH/),RECT_AZIMUTH=REAL(ANGLES%PSI))
                ELSE
                    CALL ENV_MASK_FROM_PROJ(PROJ,ENV_MASK,RREAL_SPACE_PROJ,PROJ_MAX_RAD)
                ENDIF
#ifndef SKIP_RUNTIME_CHECKS
                IF (DEBUG)  CALL WRITE_TO_DISK(ENV_MASK,'dbg_envmask.mrc')
#endif
            ENDIF

#ifndef SKIP_RUNTIME_CHECKS
            IF (DEBUG) THEN
                IF (.not. proj%is_in_real_space) CALL BWD_FT(PROJ)
                CALL WRITE_TO_DISK(PROJ,'dbg_proj_beforectf.mrc')
            ENDIF

            IF (DEBUG) THEN
                IF (PROJ%IsBlank()) THEN
                    WRITE(*,'(A)') '**ERROR(PROJECTION_PREPARE): projection is blank'
                    CALL TERMINATE('Fatal error in PROJECTION_PREPARE')
                ENDIF
            ENDIF
#endif
            ! Apply a soft cosine-edged disc mask so that CTF multiplication creates less ripples
            if (.not. proj%is_in_real_space) call bwd_ft(proj)
            CALL PROJ%ApplyCosineDisc(real(PROJ%logical_dimensions(2))/2.0-2.0,proj%logical_dimensions(2)*0.2,&
                                WIPE_INPUT=.FALSE.,MASK_WITH_AVERAGE=.FALSE.)

#ifndef SKIP_RUNTIME_CHECKS
            IF (DEBUG) THEN
                IF (.not. proj%is_in_real_space) CALL BWD_FT(PROJ)
                CALL WRITE_TO_DISK(PROJ,'dbg_proj_beforectf_cos.mrc')
            ENDIF
#endif

            ! Apply the CTF to the projection
            IF (proj%is_in_real_space) CALL FWD_FT(PROJ)
            IF (AAPPLY_CTF) THEN
                !CALL IMAGE_APPLY_CTF(PROJ,CTF,CTF_SQUARED=CTF_SQUARED)
                IF (CCTF_SQUARED) THEN
                    FORALL(I=1:MAXIMUM_COMPLEX_ADDRESS_DIM1(PROJ),J=1:PROJ%logical_dimensions(2),K=1:PROJ%logical_dimensions(3))
						PROJ%complex_values(I,J,K) =   PROJ%complex_values(I,J,K) * CTF_IMAGE%real_values(I,J,K)**2
					END FORALL
                ELSE
					FORALL(I=1:MAXIMUM_COMPLEX_ADDRESS_DIM1(PROJ),J=1:PROJ%logical_dimensions(2),K=1:PROJ%logical_dimensions(3))
                    	PROJ%complex_values(I,J,K) =   PROJ%complex_values(I,J,K) * CTF_IMAGE%real_values(I,J,K)
					END FORALL
                ENDIF
            ENDIF

#ifndef SKIP_RUNTIME_CHECKS
            IF (DEBUG) THEN
                IF (.not. proj%is_in_real_space) CALL BWD_FT(PROJ)
                CALL WRITE_TO_DISK(PROJ,'dbg_proj_afterctf.mrc')
                CALL FWD_FT(PROJ)
            ENDIF
#endif

            ! Apply the envelope mask if calculated
            IF (PRESENT(ENV_MASK) .AND. AAPPLY_ENV_MASK) THEN
                IF (.not. proj%is_in_real_space) CALL BWD_FT(PROJ)
                IF (PRESENT(MASK_WITH_ZEROS)) THEN
                    CALL IMAGE_APPLY_MASK(PROJ,ENV_MASK,MASK_WITH_AVERAGE=.NOT.MASK_WITH_ZEROS)
                ELSE
                    CALL IMAGE_APPLY_MASK(PROJ,ENV_MASK,MASK_WITH_AVERAGE=.FALSE.)
                ENDIF
            ENDIF

#ifndef SKIP_RUNTIME_CHECKS
            IF (DEBUG) THEN
                IF (.not. proj%is_in_real_space) CALL BWD_FT(PROJ)
                CALL WRITE_TO_DISK(PROJ,'dbg_beforedisc.mrc')
            ENDIF

            IF (DEBUG) THEN
                IF (PROJ%IsBlank()) THEN
                    WRITE(*,'(A)') '**ERROR(PROJECTION_PREPARE): projection is blank'
                    CALL TERMINATE('Fatal error in PROJECTION_PREPARE')
                ENDIF
            ENDIF
#endif

            ! Apply a circular mask
            IF (AAPPLY_DISC_MASK) THEN
                IF (.not. proj%is_in_real_space) CALL BWD_FT(PROJ)
#ifndef SKIP_RUNTIME_CHECKS
                IF (DEBUG) WRITE(*,'(A,L1,1X,L1)') '**debug(PROJECTION_PREPARE): about to apply disc mask. '//&
                                                'PROJ%ORIGIN_AT_CENTER, real space = ', PROJ%ORIGIN_AT_CENTER, PROJ%is_in_real_space
                IF (DEBUG) CALL WRITE_TO_DISK(PROJ,'dbg_proj_before_disc.mrc')
#endif
                CALL PROJ%ApplyCosineDisc(PROJ_MAX_RAD+SOFT_EDGE,SOFT_EDGE,WIPE_INPUT=.FALSE.,MASK_WITH_AVERAGE=.FALSE.)
#ifndef SKIP_RUNTIME_CHECKS
                IF (DEBUG) CALL WRITE_TO_DISK(PROJ,'dbg_proj_after_disc.mrc')
#endif
            ENDIF

            ! If we were given a projection cache, add the newly-computed projection to it
            !IF (PRESENT(PROJ_CACHE)) THEN
            !    CALL PROJECTION_CACHE_ADD_PROJ(PROJ_CACHE,ANGLES,REAL(VOL_SHIFT_Z),CTF,PROJ,ENV_MASK)
            !ENDIF
        !ENDIF   !   End of test of whether projection needs to be computed

    END SUBROUTINE PROJECTION_PREPARE

    !>  \brief  Calculate the cross-correlation function between two images by conjugate-multiplying their FT. If they
    !!          are real, they'll be FT'ed back and forth internally
    !!
    SUBROUTINE FAST_CORRELATION(IMG_IN1,IMG_IN2,IMG_OUT,FOURIER_MASK)
        use Images
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(INOUT)   ::  IMG_IN1
        TYPE(IMAGE),                INTENT(INOUT)   ::  IMG_IN2
        TYPE(IMAGE),                INTENT(INOUT)   ::  IMG_OUT
        TYPE(IMAGE),    OPTIONAL,   INTENT(IN)      ::  FOURIER_MASK    !<  Real image. The FT of the CC function is multiplied by this.
        ! Private variables
        LOGICAL ::  DO_FTS
        ! Start work
        ! Check both images are of same type
        IF (.NOT. IMG_IN1%IsSameTypeAs(IMG_IN2) ) THEN
            CALL TERMINATE('Fatal error in FAST_CORRELATION')
        ENDIF
        ! Check both images have same dimensions
        IF (.NOT. IMG_IN1%HasSameDimsAs(IMG_IN2)) THEN
            WRITE(*,'(A,2(3(I3,1X),A))') '**ERROR(FAST_CORRELATION): input images do not have same dimensions: ', &
                                        IMG_IN1%logical_dimensions, ' & ', IMG_IN2%logical_dimensions
            CALL TERMINATE('Fatal error in FAST_CORRELATION')
        ENDIF
        ! Check whether we'll have to do FTs internally
        DO_FTS = IMG_IN1%is_in_real_space
        ! Allocate output image if necessary
        IF (.NOT. (IMG_IN1%HasSameDimsAs(IMG_OUT) .AND. IMG_OUT%IsAllocated())) THEN
            CALL IMAGE_ALLOCATE(IMG_OUT,IMG_IN1)
        ENDIF
        ! Prepare the output image, which will initially be complex
        IMG_OUT%is_in_real_space = .false.
        IMG_OUT%ORIGIN_AT_CENTER = .FALSE.
        ! Fourier transform both inputs if necessary
        IF (DO_FTS) THEN
            CALL FWD_FT(IMG_IN1)
            CALL FWD_FT(IMG_IN2)
        ENDIF
        ! Multiply
        IMG_OUT%complex_values = IMG_IN1%complex_values * CONJG(IMG_IN2%complex_values)
        ! Apply Fourier mask
        IF (PRESENT(FOURIER_MASK)) THEN
            ! Check the dimensions are as expected
            IF (        FOURIER_MASK%logical_dimensions(1) .NE. MAXIMUM_COMPLEX_ADDRESS_DIM1(IMG_OUT) &
                .OR.    ANY(FOURIER_MASK%logical_dimensions(2:3) .NE. IMG_OUT%logical_dimensions(2:3)) ) THEN
                WRITE(*,'(A,2(3(I0,1X),1X))') '**ERROR(FAST_CORRELATION): mismatch between image and Fourier mask dimensions: ', &
                                                FOURIER_MASK%logical_dimensions, IMG_OUT%logical_dimensions
                CALL TERMINATE('Fatal error in FAST_CORRELATION')
            ENDIF
            ! Apply the mask
            IMG_OUT%complex_values(1:MAXIMUM_COMPLEX_ADDRESS_DIM1(IMG_OUT),:,:) = &
                                           IMG_OUT%complex_values(1:MAXIMUM_COMPLEX_ADDRESS_DIM1(IMG_OUT),:,:) * &
                                           FOURIER_MASK%real_values(1:FOURIER_MASK%logical_dimensions(1),:,:)
        ENDIF
        ! Center
        CALL SHIFT_REAL_SPACE_ORIGIN(IMG_OUT)
        ! Reverse FT the output
        CALL BWD_FT(IMG_OUT)
        ! The scaling needs to be adjusted
        IMG_OUT%real_values = IMG_OUT%real_values * PRODUCT(IMG_OUT%logical_dimensions)
        ! If necessary, return the input images to real space
        IF (DO_FTS) THEN
            CALL BWD_FT(IMG_IN1)
            CALL BWD_FT(IMG_IN2)
        ENDIF
    END SUBROUTINE FAST_CORRELATION


    !>  \brief  Find the shift and rotation Euler angles that can be applied to image1 to superimpose it onto image2
    !!  \todo   Peak lists
    SUBROUTINE FIND_SHIFT_EULER(IMG1,IMG2,SHIFT,ROTATION,INTEGER_SHIFT,ROTATION_SAMPLING,ROTATION_HALF_RANGE,ROTATION_FIRST,    &
                                SHIFT_ONLY,MAX_RAD_SHIFT,HORIZONTAL_ONLY,VERTICAL_ONLY,FOURIER_MASK)
        USE IMAGES
        USE EULERANGLES
        USE ROTATION_MATRICES
        USE INTERPOLATION
        USE PEAKS
        USE IMAGES_ARITHM
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(INOUT)   ::  IMG1                    !<  Image 1
        TYPE(IMAGE),                INTENT(INOUT)   ::  IMG2                    !<  Image 2
        REAL,                       INTENT(OUT)     ::  SHIFT(3)                !<  The translational shift
        TYPE(EULER),                INTENT(OUT)     ::  ROTATION                !<  The Euler angles defining the rotation to be applied
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  INTEGER_SHIFT           !<  If .TRUE., we will look for a shift of an integer number of pixels. .TRUE. is the default.
        TYPE(EULER),    OPTIONAL,   INTENT(IN)      ::  ROTATION_SAMPLING       !<  The sampling step (degrees) to be used in exhaustive search of rotations. Default is 5.0 degrees for each Euler angle.
        TYPE(EULER),    OPTIONAL,   INTENT(IN)      ::  ROTATION_HALF_RANGE     !<  The half-range (degrees) for the exhaustive search in each Euler angle. Default is 10.0 degrees for each Euler angle.
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  ROTATION_FIRST          !<  Whether the rotation should be applied to image 1 before the translation to align it to image 2. .TRUE. by default
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  SHIFT_ONLY              !<  If .TRUE., no rotation will be attempted
        REAL,           OPTIONAL,   INTENT(IN)      ::  MAX_RAD_SHIFT           !<  The maximum radial shift (in pixels) allowed in the alignment
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  HORIZONTAL_ONLY         !<  Do horizontal alignment only (search along 1st dimension only)
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  VERTICAL_ONLY           !<  Do vertical alignment only (search along 2nd dimension only)
        TYPE(IMAGE),    OPTIONAL,   INTENT(IN)      ::  FOURIER_MASK            !<  Real-space image - non-zero voxels will be used in the correlation calculation
        ! Private variables
        LOGICAL     ::  IINTEGER_SHIFT
        TYPE(EULER) ::  RROTATION_SAMPLING
        TYPE(EULER) ::  RROTATION_HALF_RANGE
        LOGICAL     ::  RROTATION_FIRST
        REAL        ::  CURRENT_PHI, CURRENT_THETA, CURRENT_PSI
        TYPE(IMAGE) ::  IMG_ROT
        REAL        ::  MATRIX(4,4)
        TYPE(PEAK)  ::  CCF_PEAK, CCF_PEAK_BEST
        TYPE(IMAGE) ::  CCF
        INTEGER     ::  ORIGIN(3)
        REAL        ::  CURRENT_CCC, BEST_CCC
        TYPE(IMAGE) ::  MASK
        REAL        ::  CURRENT_SHIFT(3)
        LOGICAL, PARAMETER     ::  REAL_SPACE_ROTATION = .TRUE.
        LOGICAL     ::  SSHIFT_ONLY
        TYPE(INTERPOLATION_SCHEME)  ::  INTERPOL
        TYPE(WINDOW)                ::  INTERPOL_WINDOW
        !LOGICAL        ::  HHORIZONTAL_ONLY,VVERTICAL_ONLY
        ! Check optional arguments
        IF (PRESENT(INTEGER_SHIFT)) THEN
            IINTEGER_SHIFT = INTEGER_SHIFT
        ELSE
            IINTEGER_SHIFT = .TRUE.
        ENDIF
        IF (PRESENT(SHIFT_ONLY)) THEN
            SSHIFT_ONLY = SHIFT_ONLY
        ELSE
            SSHIFT_ONLY = .FALSE.
        ENDIF

        ! Initialise interpolation
        INTERPOL = DEFAULT_INTERPOL_SCHEME()
        CALL INTERPOL_WINDOW%AllocateWindow(INTERPOL%WINDOW_WIDTH,.NOT.REAL_SPACE_ROTATION,IMG1%IsAVolume())

        ! Set the rotational search parameters
        IF (SSHIFT_ONLY) THEN
            RROTATION_SAMPLING%PHI      = 1.0
            RROTATION_SAMPLING%THETA    = 1.0
            RROTATION_SAMPLING%PSI      = 1.0
            RROTATION_HALF_RANGE%PHI    = 0.0
            RROTATION_HALF_RANGE%THETA  = 0.0
            RROTATION_HALF_RANGE%PSI    = 0.0
        ELSE
            IF (PRESENT(ROTATION_SAMPLING)) THEN
                RROTATION_SAMPLING = ROTATION_SAMPLING
            ELSE
                RROTATION_SAMPLING%PHI = 5.0
                RROTATION_SAMPLING%THETA = 5.0
                RROTATION_SAMPLING%PSI = 5.0
            ENDIF
            IF (PRESENT(ROTATION_HALF_RANGE)) THEN
                RROTATION_HALF_RANGE = ROTATION_HALF_RANGE
            ELSE
                RROTATION_HALF_RANGE%PHI = 10.0
                RROTATION_HALF_RANGE%THETA = 10.0
                RROTATION_HALF_RANGE%PSI = 10.0
            ENDIF
        ENDIF

        IF (PRESENT(ROTATION_FIRST)) THEN
            RROTATION_FIRST = ROTATION_FIRST
        ELSE
            RROTATION_FIRST = .TRUE.
        ENDIF

        ! We don't have a peak finding routine that returns non-integer peak coordinates yet
        IF (.NOT. IINTEGER_SHIFT) THEN
            WRITE(*,'(2A)') '**ERROR(FIND_SHIFT_EULER): We dont have a peak finding routine',    &
                            ' that returns non-integer peak coordinates yet'
            CALL TERMINATE('Fatal error in FIND_SHIFT_EULER')
        ENDIF

        ! Start work

        ! For now, we only deal with 2D images, to save on coding time!
        IF (IMG1%logical_dimensions(3) .NE. 1) THEN
            WRITE(*,'(A)') '**ERROR(FIND_SHIFT_EULER): only 2D images allowed, for now'
            CALL TERMINATE('Fatal error in FIND_SHIFT_EULER')
        ENDIF

        ! For now, we only do rotation_first, to save on coding time!
        IF (.NOT. RROTATION_FIRST) THEN
            WRITE(*,'(A)') '**ERROR(FIND_SHIFT_EULER): only rotation first allowed, for now'
            CALL TERMINATE('Fatal error in FIND_SHIFT_EULER')
        ENDIF


        ! Allocate memory to mask
        CALL IMAGE_ALLOCATE(MASK,IMG1)
        MASK%real_values = 1.0

        ! Work out the origin
        if (img1%is_in_real_space .and. img1%origin_at_center) then
            origin = img1%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif

        ! Fourier transform
        IF (.NOT. REAL_SPACE_ROTATION) THEN
            CALL FWD_FT(IMG1)
            CALL FWD_FT(IMG2)
        ENDIF

        ! Allocate memory to rotated image
        CALL IMAGE_ALLOCATE(IMG_ROT,IMG1)

        CURRENT_PHI = 0.0
        CURRENT_THETA = 0.0
        CURRENT_PSI = - RROTATION_HALF_RANGE%PSI
        CCF_PEAK_BEST%COO = [0.0,0.0,0.0]
        CCF_PEAK_BEST%rvalue = -999999.9999
        BEST_CCC = -1.0
        DO WHILE (CURRENT_PSI .LE. RROTATION_HALF_RANGE%PSI)
            ! Apply rotation (ROTATE_IMAGE needs the matrix that transforms output into input)
            MATRIX = EULER_TO_ROTMAT(DBLE(CURRENT_PHI),DBLE(CURRENT_THETA),DBLE(CURRENT_PSI),REVERSE=.TRUE.)
            IF (.NOT. REAL_SPACE_ROTATION) THEN
                CALL SHIFT_REAL_SPACE_ORIGIN(IMG1)
            ENDIF
            CALL ROTATE_IMAGE(IMG1,IMG_ROT,MATRIX,INTERPOL,INTERPOL_WINDOW)
            IF (.NOT. REAL_SPACE_ROTATION) THEN
                CALL SHIFT_REAL_SPACE_ORIGIN(IMG1)
                CALL SHIFT_REAL_SPACE_ORIGIN(IMG_ROT)
            ENDIF

            !CALL BWD_FT(IMG_ROT)
            !CALL WRITE_TO_DISK(IMG_ROT,'debug_rot_before_cc.img')
            !CALL FWD_FT(IMG_ROT)

            ! Calculate the cross-correlation function between the two images
            CALL FAST_CORRELATION(IMG_ROT,IMG2,CCF,FOURIER_MASK=FOURIER_MASK)

            !
            !PRINT *, '**debug(FIND_SHIFT_EULER): dumping CCF'
            !CALL WRITE_TO_DISK(CCF,'dbg_ccf.img')

            ! Find the peak
            IF (IINTEGER_SHIFT) THEN
                !PRINT *, '**debug(FIND_SHIFT_EULER): going to do a search with vertical_only =', VERTICAL_ONLY
                CCF_PEAK = PEAK_FROM_IMAGE_INTG(CCF,MAX_RAD=MAX_RAD_SHIFT,X_ONLY=HORIZONTAL_ONLY,Y_ONLY=VERTICAL_ONLY)
            ELSE
                ! No function for this yet
                ! 2013: there is a function for this now - just need to call it here
            ENDIF

            ! Do a normalised cross-correlation coefficient calculation to find the peak height
            CURRENT_SHIFT = - (CCF_PEAK%COO - ORIGIN)
            !PRINT '(A,3(I6,X))', '**debug(FIND_SHIFT_EULER): origin       = ', ORIGIN
            !PRINT '(A,3(F8.3,X))', '**debug(FIND_SHIFT_EULER): CCF_PEAK%COO = ', CCF_PEAK%COO
            !PRINT '(A,2(F8.3,X))', '**debug(FIND_SHIFT_EULER): shifting by ', CURRENT_SHIFT(1), CURRENT_SHIFT(2)
            IF ( REAL_SPACE_ROTATION) THEN
                MATRIX(4,1:3) = CURRENT_SHIFT(1:3)
                CALL ROTATE_IMAGE(IMG1,IMG_ROT,MATRIX,INTERPOL,INTERPOL_WINDOW)
            ELSE
                CALL PSHIFT_IMAGE(IMG_ROT,CURRENT_SHIFT(1),CURRENT_SHIFT(2),0.0)
            ENDIF


            !CALL BWD_FT(IMG_ROT)
            !CALL WRITE_TO_DISK(IMG_ROT,'debug_rot_after_cc2.img')
            !CALL FWD_FT(IMG_ROT)

            CURRENT_CCC = CORRELATION(IMG_ROT,IMG2,MASK,NORMALISE=.TRUE.)
            !IF (CURRENT_PSI .EQ. -2.0) THEN
            !   IF (.NOT. REAL_SPACE_ROTATION) THEN
            !       CALL BWD_FT(IMG_ROT)
            !       CALL BWD_FT(IMG2)
            !   ENDIF
            !   CALL WRITE_TO_DISK(CCF,'debug_ccf.img')
            !   CALL WRITE_TO_DISK(IMG_ROT,'debug_rot.img')
            !   CALL WRITE_TO_DISK(IMG2,'debug_ref.img')
            !   IF (.NOT. REAL_SPACE_ROTATION) THEN
            !       CALL FWD_FT(IMG2)
            !       CALL FWD_FT(IMG_ROT)
            !   ENDIF
            !ENDIF

            !PRINT '(A,3(F6.3,X))', '**debug(FIND_SHIFT_EULER): CURRENT_PSI, peak value, CCC: ', CURRENT_PSI, CCF_PEAK%real_values, CURRENT_CCC

            !CALL TERMINATE()

            ! Is this the best peak we've seen so far?
            !IF (CCF_PEAK%real_values .GT. CCF_PEAK_BEST%real_values) THEN
            IF (CURRENT_CCC .GT. BEST_CCC) THEN
                !PRINT '(A,F6.3,A)', '**debug(FIND_SHIFT_EULER): CURRENT_PSI of ', CURRENT_PSI, ' beats current max.'
                CCF_PEAK_BEST  = CCF_PEAK
                ROTATION%PHI   = CURRENT_PHI
                ROTATION%THETA = CURRENT_THETA
                ROTATION%PSI   = CURRENT_PSI
                SHIFT = - (CCF_PEAK_BEST%COO - ORIGIN)
                BEST_CCC = CURRENT_CCC
            ENDIF

            ! Increment psi for the next iteration of the search
            CURRENT_PSI = CURRENT_PSI + RROTATION_SAMPLING%PSI
        ! End of loop over PSI
        ENDDO

        !CALL TERMINATE()

    END SUBROUTINE FIND_SHIFT_EULER

    !>  \brief  Find the translation shift (X,Y,Z) which can be applied to image1 to superimpose it onto image2.
    FUNCTION FIND_SHIFT(IMG1,IMG2,INTEGER_SHIFT,Z_SHIFT_ONLY,MIN_RADIUS) RESULT(SHIFT)
        !USE IMAGES
        USE PEAKS
        USE IMAGES_ARITHM
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),            INTENT(INOUT)   ::  IMG1            !<  Image 1
        TYPE(IMAGE),            INTENT(INOUT)   ::  IMG2            !<  Image 2
        LOGICAL,    OPTIONAL,   INTENT(IN)      ::  INTEGER_SHIFT   !<  If .TRUE., we will look for a shift of an integer number of pixels. .TRUE. is the default.
        LOGICAL,    OPTIONAL,   INTENT(IN)      ::  Z_SHIFT_ONLY    !<  Only look for shift along the 3rd axis
        REAL,       OPTIONAL,   INTENT(IN)      ::  MIN_RADIUS      !<  Only look for peaks further or at this radius distance from origin (pixels)
        ! Result
        REAL        ::  SHIFT(3)
        ! Private variables
        LOGICAL     ::  IINTEGER_SHIFT
        TYPE(IMAGE) ::  CCF ! Cross-correlation function
        TYPE(PEAK)  ::  CCF_PEAK
        LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
        INTEGER     ::  ORI(3), I
        ! Start work

        ! Integer shift?
        IF (PRESENT(INTEGER_SHIFT)) THEN
            IINTEGER_SHIFT = INTEGER_SHIFT
        ELSE
            IINTEGER_SHIFT = .TRUE.
        ENDIF

        ! We don't have a peak finding routine that returns non-integer peak coordinates yet
        IF (.NOT. IINTEGER_SHIFT) THEN
            WRITE(*,'(A)') '**ERROR(FIND_SHIFT): We dont have a peak finding routine that returns non-integer peak coordinates yet'
            CALL TERMINATE('Fatal error in FIND_SHIFT')
        ENDIF

        ! Maybe add filtering step here?

        ! Calculate the cross-correlation function between the two images
        CALL FAST_CORRELATION(IMG1,IMG2,CCF)
        IF (DEBUG) THEN
            CALL WRITE_TO_DISK(CCF,'dbg_ccf.mrc')
        ENDIF
        ! Find the peak
        IF (IINTEGER_SHIFT) THEN
            CCF_PEAK = PEAK_FROM_IMAGE_INTG(CCF,MIN_RAD=MIN_RADIUS)
        ELSE
            ! No function for this yet
        ENDIF
        !
        DO I=1,3
            ORI(I) = INDEX_OF_CENTRAL_PIXEL_FROM_DIM(IMG1%logical_dimensions(I))
        ENDDO
        SHIFT = CCF_PEAK%COO - ORI    ! Maybe this needs a minus?
    END FUNCTION FIND_SHIFT

    !>  \brief  For 6 search parameters (x,y,z,phi,theta,psi), given a target accuracy, return scaling factors to make the parameter values comparable to each other
    FUNCTION PARAM_SCALING_FROM_ACCURACY(ACCURACY,OBJ_MAX_RADIUS)
        USE RUNTIME_PARAMETERS, ONLY: TERMINATE
        USE ROTATION_MATRICES, ONLY : RADIANS
        IMPLICIT NONE
        ! Arguments
        REAL,       INTENT(IN)      ::  ACCURACY(6)     !<  The desired accuracy for each parameter
        REAL,       INTENT(IN)      ::  OBJ_MAX_RADIUS  !<  The maximum radius of the object (pixels)
        ! Result
        REAL    ::  PARAM_SCALING_FROM_ACCURACY(6)
        ! Private variables
        LOGICAL,    PARAMETER       ::  DEBUG   =   .FALSE.
        ! Start work

        ! Check that the search accuracy is known
        IF (.NOT. ANY(ACCURACY .NE. 0.0)) THEN
            WRITE(*,'(2A)') '**ERROR(PARAM_SCALING_FROM_ACCURACY): desired search accuracy unknown', &
                            ' - cannot compute scaling factors for search parameters'
            CALL TERMINATE('Fatal error in PARAM_SCALING_FROM_ACCURACY')
        ENDIF

        ! Compute the scale factors
        PARAM_SCALING_FROM_ACCURACY(1:3)    =   ACCURACY(1:3) / ACCURACY(1)
        PARAM_SCALING_FROM_ACCURACY(4:6)    =   RADIANS(ACCURACY(4:6)) * OBJ_MAX_RADIUS / ACCURACY(1)

        ! Debug
        IF (DEBUG) THEN
            WRITE(*,'(A,6(F8.3,1X))') '**debug(PARAM_SCALING_FROM_ACCURACY): search accuracy = ', ACCURACY
            WRITE(*,'(A,6(F8.3,1X))') '**debug(PARAM_SCALING_FROM_ACCURACY): param. scaling  = ', PARAM_SCALING_FROM_ACCURACY
        ENDIF

    END FUNCTION PARAM_SCALING_FROM_ACCURACY

    !>  \brief  Return the desired accuracy in the 6 degrees of freedom (x,y,z,phi,theta,psi) to obtain the target resolution
    !! For details of derivation & meanings, see documentation.
    !! Note that these are just ballpark figures.
    !!
    !!  \todo   Thighten up definitions: what exactly are we returning here? What phase error is targeted? Maybe relate to Jensen paper.
    FUNCTION ACCURACY_FROM_RESOLUTION(RESOLUTION,OBJ_MAX_RAD,BOX_SIZE,AKV,PIXEL_SIZE,AKV_IS_WL,TARGET_PHASE_ERROR) RESULT (ACCURACY)
        USE CTFS
        USE ROTATION_MATRICES, ONLY : DEGREES, RADIANS
        use Globals, only : pi
        IMPLICIT NONE
        ! Arguments
        REAL,                   INTENT(IN)  ::  RESOLUTION          !<  Target resolution (in A)
        REAL,                   INTENT(IN)  ::  OBJ_MAX_RAD         !<  Object maximum radius (in A)
        INTEGER,                INTENT(IN)  ::  BOX_SIZE            !<  Reconstruction volume dimension (in pixels). Assumption that the volume is cubic and can be specified by just one value.
        REAL,                   INTENT(IN)  ::  AKV                 !<  Acceleration voltage, in kV, unless AKV_IS_WL is given as .TRUE., in which case this argument is actually the wavelength (Angstroms).
        REAL,                   INTENT(IN)  ::  PIXEL_SIZE          !<  Pixel size in Angstrom.
        LOGICAL,    OPTIONAL,   INTENT(IN)  ::  AKV_IS_WL           !<  If TRUE, the AKV argument has already been converted to wavelength.
        REAL,       OPTIONAL,   INTENT(IN)  ::  TARGET_PHASE_ERROR  !<  The target phase error (in degrees). Defaults to a sensible value (see below).
        ! Result
        REAL                                ::  ACCURACY(6) !<  Accuracy required to achieve the resolution given as argument. In pixels and degrees.
        ! Private variables
        !REAL   ::  PPIXEL_SIZE
        REAL,   PARAMETER       ::  OBJECT_DIM_FUDGE = 1.0
        REAL                    ::  WL
        LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
        LOGICAL,    PARAMETER   ::  Z_AS_XY =   .TRUE.
        REAL,       PARAMETER   ::  DEFAULT_TARGET_PHASE_ERROR = 45.0
        REAL                    ::  TTARGET_PHASE_ERROR
        ! Start work

        ! Did the user actually supply the wavelength rather than the acceleration voltage?
        IF (PRESENT(AKV_IS_WL)) THEN
            IF (AKV_IS_WL) THEN
                WL = AKV
            ELSE
                WL = AKV_TO_WL(AKV)
            ENDIF
        ELSE
            WL = AKV_TO_WL(AKV)
        ENDIF

        ! What is our target phase error?
        IF (PRESENT(TARGET_PHASE_ERROR)) THEN
            TTARGET_PHASE_ERROR = TARGET_PHASE_ERROR
        ELSE
            TTARGET_PHASE_ERROR = DEFAULT_TARGET_PHASE_ERROR
        ENDIF

        ! X, Y
        !ACCURACY(1:2)      = RESOLUTION / 2.
        ACCURACY(1:2)       = RADIANS(TTARGET_PHASE_ERROR) / ( (1./RESOLUTION) * 2. * PI )
        ACCURACY(1:2)       = ACCURACY(1:2) / PIXEL_SIZE

        ! Z (defocus)
        IF (Z_AS_XY) THEN
            WRITE(*,'(A)') '**warning(ACCURACY_FROM_RESOLUTION): Z treated as X and Y, not as a defocus'
            ACCURACY(3)         = RADIANS(TTARGET_PHASE_ERROR) / ( (1./RESOLUTION) * 2. * PI )
            ACCURACY(3)         = ACCURACY(3) / PIXEL_SIZE
        ELSE
            ACCURACY(3)         = RADIANS(TTARGET_PHASE_ERROR) / (PI*WL*(1/RESOLUTION)**2)
            ACCURACY(3)         = ACCURACY(3) / PIXEL_SIZE
        ENDIF

        ! Euler angles
        !ACCURACY(4:6)  = DEGREES(ATAN(RESOLUTION/(PIXEL_SIZE*2.*REAL(BOX_SIZE)*OBJECT_DIM_FUDGE)))
        ACCURACY(4:6)   =   DEGREES(RADIANS(TTARGET_PHASE_ERROR) / ( (1./RESOLUTION) * 2. * PI *OBJ_MAX_RAD))
        !ACCURACY(4:6)      = DEGREES(RESOLUTION/OBJ_MAX_RAD)/2.


        ! Since each tolerance was worked out independently but they combine in effect, we need to tighten things up
        ACCURACY = ACCURACY / 6.0

        !
        ! Print out information for the user
        !
        !$OMP   SINGLE
        WRITE(*,'(/2A)') '**info(ACCURACY_FROM_RESOLUTION): Below are important parameters used to', &
                                ' determine the required alignment accuracies'
        WRITE(*,'(2X,A,F8.3,1X,A)')    'Target resolution             = ', RESOLUTION, 'A'
        WRITE(*,'(2X,A,F8.3,1X,A)')    'Target phase error            = ', TTARGET_PHASE_ERROR, 'degrees'
        WRITE(*,'(2X,A,I8,  1X,A)')    'Reconstruction box size       = ', BOX_SIZE, 'pixels'
        WRITE(*,'(2X,A,F8.0,1X,A)')    'Object maximum radius         = ', OBJ_MAX_RAD, 'A'
        WRITE(*,'(2X,A,F8.0,1X,A)')    'Acceleration voltage          = ', WL_TO_AKV(WL), 'kV'
        WRITE(*,'(2X,A,F8.3,1X,A)')    'Pixel/voxel size              = ', PIXEL_SIZE, 'A'
        WRITE(*,'(2X,A,F8.3,1X,A)')    'Required accuracy (X,Y)       = ', ACCURACY(1), 'pixels'
        WRITE(*,'(2X,A,F8.3,1X,A)')    'Required accuracy (Z)         = ', ACCURACY(3), 'pixels'
        WRITE(*,'(2X,A,F8.3,1X,A)')    'Required accuracy (rotations) = ', ACCURACY(4), 'degrees'
        WRITE(*,'(A)') ' '
        !$OMP   END SINGLE  NOWAIT
    END FUNCTION ACCURACY_FROM_RESOLUTION



    !>  \brief  Print FSC to terminal
    SUBROUTINE PRINT_FSC(FSC)
        IMPLICIT NONE
        REAL,   ALLOCATABLE,    INTENT(IN)  ::  FSC(:)
        ! Private variables
        INTEGER ::  I
        ! Start work
        WRITE(*,'(A)') ' '
        WRITE(*,'(A5,2X,3X,A3)') 'Shell', 'FSC'
        DO I=1,SIZE(FSC)
            WRITE(*,'(1X,I4,2X,F6.3)')   I-1, FSC(I)
        ENDDO
    END SUBROUTINE PRINT_FSC

    !>  \brief  Linear cross-correlation coefficient
    REAL FUNCTION CCLIN(IMG1,IMG2,RAD_MIN,RAD_MAX,RBFACT,FOURIER_MASK)
        !USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(IN)  ::  IMG1            !<  Noisy image
        TYPE(IMAGE),                INTENT(IN)  ::  IMG2            !<  Reference image
        REAL,           OPTIONAL,   INTENT(IN)  ::  RAD_MIN         !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
        REAL,           OPTIONAL,   INTENT(IN)  ::  RAD_MAX         !<  Maximum radius
        REAL,           OPTIONAL,   INTENT(IN)  ::  RBFACT          !<  B-factor (if not given, assumed to be 0.0)
        TYPE(IMAGE),    OPTIONAL,   INTENT(IN)  ::  FOURIER_MASK    !<  A real image which defines a Fourier-space mask.
        ! Private variables
        REAL(KIND=8)    ::  TMP(3)
        ! Start work
        CALL  CCLIN_NUM_DENOMS(IMG1,IMG2,TMP,RAD_MIN,RAD_MAX,RBFACT,FOURIER_MASK)
        ! Cross-correlation coefficient
        CCLIN = REAL(TMP(1)/SQRT(TMP(2)*TMP(3)))
    END FUNCTION CCLIN


    !>  \brief  Linear cross-correlation coefficient (returns numerator and denominator components separately) (denominator components are running sums of squares)
	SUBROUTINE CCLIN_NUM_DENOMS(IMG1,IMG2,NUM_DENOMS,RAD_MIN,RAD_MAX,RBFACT,FOURIER_MASK)
        !USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(IN)  ::  IMG1            !<  Noisy image
        TYPE(IMAGE),                INTENT(IN)  ::  IMG2            !<  Reference image
        REAL,           OPTIONAL,   INTENT(IN)  ::  RAD_MIN         !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
        REAL,           OPTIONAL,   INTENT(IN)  ::  RAD_MAX         !<  Maximum radius
        REAL,           OPTIONAL,   INTENT(IN)  ::  RBFACT          !<  B-factor (if not given, assumed to be 0.0)
        TYPE(IMAGE),    OPTIONAL,   INTENT(IN)  ::  FOURIER_MASK    !<  A real image which defines a Fourier-space mask.
        ! Result
        REAL(KIND=8)     ,INTENT(OUT)                   ::  NUM_DENOMS(3)
        ! Private variables
        REAL(KIND=8),   ALLOCATABLE ::  TMP(:,:)
        ! Start work

        TMP = CCLIN_NUM_DENOMS_RINGS(IMG1,IMG2,RAD_MIN,RAD_MAX,RBFACT,FOURIER_MASK)


        NUM_DENOMS    =   CCLIN_NUM_DENOMS_FROM_RINGS(TMP)

	END SUBROUTINE CCLIN_NUM_DENOMS

    FUNCTION CCLIN_NUM_DENOMS_FROM_RINGS(RINGS)
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   ALLOCATABLE ::  RINGS(:,:)      !<  For each spatial frequency (indexed in the second dimension), the numerator and two denominator terms of the cross-correlation
        ! Result
        REAL(KIND=8)                ::  CCLIN_NUM_DENOMS_FROM_RINGS(3)
        ! Private variables

        ! Start work
        ! Check allocation of input
        IF (.NOT. ALLOCATED(RINGS)) THEN
            WRITE(*,'(A)') '**ERROR(CCLIN_NUM_DENOMS_FROM_RINGS): input array not allocated'
            CALL TERMINATE('Fatal error in CCLIN_NUM_DENOMS_FROM_RINGS')
        ENDIF
        ! Check dimension of input
        IF (SIZE(RINGS,1) .NE. 3) THEN
            WRITE(*,'(A,I0)') '**ERROR(CCLIN_NUM_DENOMS_FROM_RINGS): unexpected size of input array in first dimension: ',  &
                                SIZE(RINGS,1)
            CALL TERMINATE('Fatal error in CCLIN_NUM_DENOMS_FROM_RINGS')
        ENDIF

        CCLIN_NUM_DENOMS_FROM_RINGS(1) = SUM(RINGS(1,:))
        CCLIN_NUM_DENOMS_FROM_RINGS(2) = SUM(RINGS(2,:))
        CCLIN_NUM_DENOMS_FROM_RINGS(3) = SUM(RINGS(3,:))

    END FUNCTION CCLIN_NUM_DENOMS_FROM_RINGS


    !>  \brief  Linear cross-correlation coefficient (returns numerator and denominator components separately) (denominator components are running sums of squares)
    !!
    !!          Only those voxels which are between the min and max radius and which are non-zero in the Fourier mask
    !!          (if given) are used in the computation
    FUNCTION CCLIN_NUM_DENOMS_RINGS(IMG1,IMG2,RAD_MIN,RAD_MAX,RBFACT,FOURIER_MASK) RESULT (CCLIN_NDR)
        USE RUNTIME_PARAMETERS, ONLY : IS_EVEN, TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(IN)  ::  IMG1            !<  Noisy image
        TYPE(IMAGE),                INTENT(IN)  ::  IMG2            !<  Reference image
        REAL,           OPTIONAL,   INTENT(IN)  ::  RAD_MIN         !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
        REAL,           OPTIONAL,   INTENT(IN)  ::  RAD_MAX         !<  Maximum radius
        REAL,           OPTIONAL,   INTENT(IN)  ::  RBFACT          !<  B-factor (if not given, assumed to be 0.0)
        TYPE(IMAGE),    OPTIONAL,   INTENT(IN)  ::  FOURIER_MASK    !<  A real image which defines a Fourier-space mask.
                                                                    !!  This supersedes the min/max radius. Any positions at which the mask is not 0.0
                                                                    !!  will be used in the computation.
        ! Result
        REAL(KIND=8),   ALLOCATABLE         ::  CCLIN_NDR(:,:)
        ! Private variables
        INTEGER                     ::  I,J,K, IRAD
        REAL                        ::  RAD, RRAD_MIN, RRAD_MAX
        REAL(KIND=8)                ::  RRBFACT
        LOGICAL                     ::  DNF(3)
        INTEGER                     ::  IERR
        REAL,       ALLOCATABLE     ::  WGT(:)
        LOGICAL,    PARAMETER       ::  IIGNORE_FRIEDEL_MATES   =   .TRUE.  ! Some Friedel mates are described explicelty by the FFTW array, we should not count them twice!
        INTEGER                     ::  X,Y,Z
        LOGICAL                     ::  INCLUDE_THIS_VOXEL
        ! Start work

#ifndef SKIP_RUNTIME_CHECKS
        ! Check both images are complex
        IF (.NOT. (img1%IsSameTypeAs(IMG2) .AND. .not. IMG1%is_in_real_space)) THEN
            CALL TERMINATE('Fatal error in CCLIN_NUM_DENOMS_RINGS')
        ENDIF
        ! Check images have same size
        IF (.NOT. img1%HasSameDimsAs(img2)) THEN
            WRITE(*,'(A,2(3(I0,1X),1X))') '**ERROR(CCLIN_NUM_DENOMS_RINGS): images not same dimensions. ', &
                                            IMG1%logical_dimensions, IMG2%logical_dimensions
            CALL TERMINATE('Fatal error in CCLIN_NUM_DENOMS_RINGS')
        ENDIF
        ! Hardcoded for even dimension
        IF (.NOT. IMG1%HasEvenDims()) THEN
            WRITE(*,'(A,3(I0,1X))') '**ERROR(CCLIN_NUM_DENOMS_RINGS): input images dimensions are not even: ', &
                                            IMG1%logical_dimensions
            CALL TERMINATE('Fatal error in CCLIN_NUM_DENOMS_RINGS')
        ENDIF
        ! Hardcoded for cubic images
        IF (.NOT. img1%IsSquare()) THEN
            CALL TERMINATE('Hardcoded for square/cubic images')
        ENDIF
#endif

        ! Low & high res. limits
        RRAD_MIN = 1.0
        IF (PRESENT(RAD_MIN)) RRAD_MIN = RAD_MIN
        RRAD_MIN = MAX(1.0,RRAD_MIN)

        RRAD_MAX = REAL(IMG1%logical_dimensions(1))/2.0
        IF (PRESENT(RAD_MAX)) RRAD_MAX = RAD_MAX
        RRAD_MAX = MIN(RRAD_MAX,REAL(IMG1%logical_dimensions(1))/2.)

        ! B-factor
        RRBFACT = 0.0D0
        IF (PRESENT(RBFACT)) RRBFACT = DBLE(RBFACT)

        ! Allocate memory for results
        ALLOCATE(CCLIN_NDR(3,IMG1%logical_dimensions(1)/2),WGT(IMG1%logical_dimensions(1)/2),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(CCLIN_NUM_DENOMS_RINGS): memory allocation for results failed with error ', IERR
            CALL TERMINATE('Fatal error in CCLIN_NUM_DENOMS_RINGS')
        ENDIF

        ! Initialise
        CCLIN_NDR = 0.0D0

        ! Calculate the B-factor values to be applied at each radius
        IF (RRBFACT .EQ. 0.) THEN
            WGT = 1.0
        ELSE
            DO I=1,IMG1%logical_dimensions(1)/2
                WGT(I) = EXP(-RRBFACT*(I-1)**2)
            ENDDO
        ENDIF

        ! Loop to collect cross-correlation info
        CCLIN_NDR(1,:) = 0.0D0
        CCLIN_NDR(2,:) = 0.0D0
        CCLIN_NDR(3,:) = 0.0D0
        DNF = [.TRUE.,.FALSE.,.FALSE.]
        DO I=1,HIGHEST_INDEX_DIM1(IMG1%logical_dimensions(1))
            X = (I-1)**2
            DO J=1,IMG1%logical_dimensions(2)
                IF (IIGNORE_FRIEDEL_MATES .AND. I == 1 .AND. J .GT. 1 .AND. &
                    J .LT. LOWEST_INDEX_FOR_NEGATIVE_FREQUENCIES(IMG1%logical_dimensions(2))) CYCLE    ! Skip voxels whose Friedel mates are explicit in the FFTW description
                Y = (J-NEAREST_ORIGIN(J,IMG1%logical_dimensions(2)))**2
                DO K=1,IMG1%logical_dimensions(3)
                    IF (IIGNORE_FRIEDEL_MATES .AND. I == 1 .AND. J == 1 .AND. K .GT. 1 .AND.    &
                        K .LT. LOWEST_INDEX_FOR_NEGATIVE_FREQUENCIES(IMG1%logical_dimensions(3))) CYCLE
                    Z = (K-NEAREST_ORIGIN(K,IMG1%logical_dimensions(3)))**2

                    RAD = SQRT(REAL(X+Y+Z))
                    IRAD = NINT(RAD)

                    IF (PRESENT(FOURIER_MASK)) THEN
                        INCLUDE_THIS_VOXEL = FOURIER_MASK%real_values(I,J,K) .NE. 0.0
                    ELSE
                        INCLUDE_THIS_VOXEL = RAD .GE. RRAD_MIN .AND. RAD .LT. RRAD_MAX
                    ENDIF

                    IF (INCLUDE_THIS_VOXEL) THEN
                        CCLIN_NDR(1,IRAD)    =   CCLIN_NDR(1,IRAD) + &
                                                 DBLE(IMG1%complex_values(I,J,K)*CONJG(IMG2%complex_values(I,J,K))*WGT(IRAD))
                        CCLIN_NDR(2,IRAD)    =   CCLIN_NDR(2,IRAD) + DBLE(ABS(IMG1%complex_values(I,J,K))**2)
                        CCLIN_NDR(3,IRAD)    =   CCLIN_NDR(3,IRAD) + DBLE((ABS(IMG2%complex_values(I,J,K))*WGT(IRAD))**2)
                    ENDIF
                ENDDO
            ENDDO
        ENDDO

    END FUNCTION CCLIN_NUM_DENOMS_RINGS

!   !>  \brief  Correlation coefficient calculation based on Stewart & Grigorieff 2004
!   REAL FUNCTION CC3_OLD(IMG1,IMG2,RAD_MIN,RAD_MAX,RBFACT,ABSOLUTE)
!       !USE IMAGES
!       IMPLICIT NONE
!       ! Arguments
!       TYPE(IMAGE),            INTENT(IN)  ::  IMG1        !<  Noisy image
!       TYPE(IMAGE),            INTENT(IN)  ::  IMG2        !<  Reference image
!       INTEGER,    OPTIONAL,   INTENT(IN)  ::  RAD_MIN     !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
!       INTEGER,    OPTIONAL,   INTENT(IN)  ::  RAD_MAX     !<  Maximum radius
!       REAL,       OPTIONAL,   INTENT(IN)  ::  RBFACT      !<  B-factor (if not given, assumed to be 0.0)
!       LOGICAL,    OPTIONAL,   INTENT(IN)  ::  ABSOLUTE    !<  Whether to add the absolute value of each shell (turns CC3 into CC3M). Default to FALSE.
!       ! Private variables
!       INTEGER ::  RAD, I,J,K, RRAD_MIN, RRAD_MAX
!       LOGICAL ::  AABSOLUTE
!       REAL(KIND=8)    ::  COUNT(IMG1%logical_dimensions(1)/2+1) ! Count of voxels in shell
!       REAL(KIND=8)    ::  CP(IMG1%logical_dimensions(1)/2+1) ! Cross-product
!       REAL(KIND=8)    ::  SQ1(IMG1%logical_dimensions(1)/2+1),SQ2(IMG1%logical_dimensions(1)/2+1)   ! Sum of squares
!       REAL(KIND=8)    ::  W   ! Inverse of box dimension
!       REAL(KIND=8)    ::  R
!       REAL(KIND=8)    ::  SDPHA, SNPHA
!       REAL(KIND=8)    ::  WGT ! weight
!       REAL(KIND=8)    ::  RRBFACT ! B-factor (should be an argument)
!       LOGICAL ::  DNF(3)
!       ! Start work
!       ! Check both images are complex
!       IF (.NOT. (SAME_TYPE(IMG1,IMG2) .AND. IMG1%TYPE == 'COMP')) THEN
!           WRITE(*,'(3(A,X))') '**ERROR(CC3): images need to be complex.', IMG1%TYPE, IMG2%TYPE
!           STOP 'Fatal error in CC3'
!       ENDIF
!       ! Check images have same size
!       IF (.NOT. SAME_DIM(IMG1,IMG2)) THEN
!           WRITE(*,'(A,2(3(I3,X),X))') '**ERROR(CC3): images not same dimensions. ', IMG1%logical_dimensions, IMG2%logical_dimensions
!           STOP 'Fatal error in CC3'
!       ENDIF
!       ! B-factor
!       IF (PRESENT(RBFACT)) THEN
!           RRBFACT = DBLE(RBFACT)
!       ELSE
!           RRBFACT = 0.0D0
!       ENDIF
!       ! Low & high res. limits
!       IF (PRESENT(RAD_MIN)) THEN
!           RRAD_MIN = RAD_MIN
!       ELSE
!           RRAD_MIN = 1
!       ENDIF
!       IF (PRESENT(RAD_MAX)) THEN
!           RRAD_MAX = RAD_MAX
!       ELSE
!           RRAD_MAX = IMG1%logical_dimensions(1)/2
!       ENDIF
!
!       IF(PRESENT(ABSOLUTE)) THEN
!           AABSOLUTE = ABSOLUTE
!       ELSE
!           AABSOLUTE = .FALSE.
!       ENDIF
!
!       ! Loop to collect cross-correlation info
!       !PRINT *, '**debug(CC3): RRAD_MIN, RRAD_MAX = ', RRAD_MIN, RRAD_MAX
!       CP = 0.0D0
!       SQ1 = 0.0D0
!       SQ2 = 0.0D0
!       COUNT = 0.0D0
!       DNF = [.TRUE.,.FALSE.,.FALSE.]
!       DO K=1,IMG1%logical_dimensions(3)
!           DO J=1,IMG1%logical_dimensions(2)
!               DO I=1,IMG1%logical_dimensions(1)/2+1
!                   RAD = NINT(RADIUS(I,J,K,IMG1,DNF))
!                   IF (RAD .GE. RRAD_MIN .AND. RAD .LT. RRAD_MAX) THEN ! Maybe the check for the radius should be done in REALs, not INTGs?
!                       CP(RAD)  =  CP(RAD) + DBLE(IMG1%complex_values(I,J,K)*CONJG(IMG2%complex_values(I,J,K)))
!                       SQ1(RAD) = SQ1(RAD) + DBLE(ABS(IMG1%complex_values(I,J,K))**2)
!                       SQ2(RAD) = SQ2(RAD) + DBLE(ABS(IMG2%complex_values(I,J,K))**2)
!                       COUNT(RAD) = COUNT(RAD) + 1.0
!                   ENDIF
!               ENDDO
!           ENDDO
!       ENDDO
!
!       ! Debug
!       !PRINT *, '**debug(CC3_OLD): numerator, denom1, denom2 per shell before cubing:'
!       !DO RAD=1,IMG1%logical_dimensions(1)/2
!       !   PRINT *, RAD, CP(RAD), SQ1(RAD), SQ2(RAD)
!       !ENDDO
!
!       ! Loop over the shells
!       W = 1.0D0/DBLE(IMG1%logical_dimensions(1))     ! Asumption images are square/cube
!       SDPHA = 0.0D0
!       SNPHA = 0.0D0
!       DO RAD=2,IMG1%logical_dimensions(1)/2
!           IF (COUNT(RAD) .NE. 0.0) THEN
!               !PRINT '(A,4F12.3)', '**debug(CC3): CP, SQ1, SQ2 = ', CP(RAD), SQ1(RAD), SQ2(RAD), COUNT(RAD)
!               CP(RAD) = CP(RAD)/SQRT(SQ1(RAD)*SQ2(RAD))
!               !PRINT *, '**debug(CC3): CP for shell ', RAD, ' = ', CP(RAD)
!               IF (AABSOLUTE) THEN
!                   R = CP(RAD)**2/(W+ABS(CP(RAD))) ! Eq. 16 in Stewart & Grigorieff
!               ELSE
!                   R = CP(RAD)*ABS(CP(RAD))/(W+ABS(CP(RAD))) ! Eq. 16 in Stewart & Grigorieff
!               ENDIF
!               WGT = EXP(-RRBFACT*(RAD-1)**2)  ! B-factor weighting
!               SDPHA = SDPHA + R**3 * WGT  ! signal
!               SNPHA = SNPHA + WGT         ! noise
!           ENDIF
!       ENDDO
!       !PRINT *, '**debug(CC3): count of voxels = ', SUM(COUNT), (IMG1%logical_dimensions(1)/2+1)*IMG1%logical_dimensions(2)*IMG1%logical_dimensions(3)
!       !
!       IF (SDPHA .NE. 0.0D0) THEN
!           SDPHA = EXP(LOG(ABS(SDPHA)/SNPHA)/3.0D0)*SDPHA/ABS(SDPHA) ! ?
!       ENDIF
!       IF (SDPHA .GT.  1.0D0) SDPHA =  1.0D0
!       IF (SDPHA .LT. -1.0D0) SDPHA = -1.0D0
!       CC3_OLD = REAL(SDPHA)
!   END FUNCTION CC3_OLD

    !>  \brief  Correlation coefficient calculation based on Stewart & Grigorieff 2004
    REAL FUNCTION CC3(IMG1,IMG2,RAD_MIN,RAD_MAX,RBFACT,ABSOLUTE,FOURIER_MASK)
        !USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(IN)  ::  IMG1        !<  Noisy image
        TYPE(IMAGE),                INTENT(IN)  ::  IMG2        !<  Reference image
        REAL,           OPTIONAL,   INTENT(IN)  ::  RAD_MIN     !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
        REAL,           OPTIONAL,   INTENT(IN)  ::  RAD_MAX     !<  Maximum radius
        REAL,           OPTIONAL,   INTENT(IN)  ::  RBFACT      !<  B-factor (if not given, assumed to be 0.0)
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  ABSOLUTE    !<  Whether to add the absolute value of each shell (turns CC3 into CC3M). Default to FALSE.
        TYPE(IMAGE),    OPTIONAL,   INTENT(IN)  ::  FOURIER_MASK    !<  A real image which defines a Fourier-space mask.
        ! Private variables
        REAL(KIND=8),       ALLOCATABLE     ::  TMP(:,:)
        !LOGICAL,   PARAMETER               ::  SANITY_CHECK    =   .FALSE.
        !REAL                               ::  CC3_CHECK
        ! Start work

        TMP = CCLIN_NUM_DENOMS_RINGS(IMG1,IMG2,RAD_MIN,RAD_MAX,RBFACT=0.0,FOURIER_MASK=FOURIER_MASK)

        CC3 = CC3_FINALISE(TMP,IMG1,IMG2,RAD_MIN,RAD_MAX,RBFACT,ABSOLUTE)

        !IF (SANITY_CHECK) THEN
        !   CC3_CHECK = CC3_OLD(IMG1,IMG2,RAD_MIN,RAD_MAX,RBFACT,ABSOLUTE)
        !
        !   IF (CC3 .NE. CC3_CHECK) THEN
        !       WRITE(*,'(A,2(F12.6,X))') '**ERROR(CC3): sanity check failed: CC3, CC3_OLD = ', CC3, CC3_CHECK
        !       CALL TERMINATE('Fatal error in CC3')
        !   ENDIF
        !ENDIF

    END FUNCTION CC3

    !>  \brief  Given accumulated sums of crossproducts (numerator) and squares (denominators),
    !!          return the CC3 score
    FUNCTION CC3_FINALISE(NUM_DENOMS_RINGS,IMG1,IMG2,RAD_MIN,RAD_MAX,RBFACT,ABSOLUTE) RESULT(CC3_FINAL)
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(IN)  ::  IMG1                    !<  Noisy image
        TYPE(IMAGE),                INTENT(IN)  ::  IMG2                    !<  Reference image
        REAL(KIND=8),ALLOCATABLE,   INTENT(IN)  ::  NUM_DENOMS_RINGS(:,:)
        REAL,       OPTIONAL,       INTENT(IN)  ::  RAD_MIN                 !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
        REAL,       OPTIONAL,       INTENT(IN)  ::  RAD_MAX                 !<  Maximum radius
        REAL,       OPTIONAL,       INTENT(IN)  ::  RBFACT                  !<  B-factor (if not given, assumed to be 0.0)
        LOGICAL,    OPTIONAL,       INTENT(IN)  ::  ABSOLUTE                !<  Whether to add the absolute value of each shell (turns CC3 into CC3M). Default to FALSE.
        ! Result
        REAL(KIND=8)                            ::  CC3_FINAL
        ! Private variables
        REAL(KIND=8)            ::  TMP, W, R
        REAL                    ::  WGT, RRBFACT
        INTEGER                 ::  RAD
        REAL                    ::  RRAD_MIN, RRAD_MAX
        REAL(KIND=8)            ::  NUM, DENOM
        LOGICAL                 ::  AABSOLUTE
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.

        ! Start work

        !
        ! Deal with optional arguments
        !

        ! B-factor
        IF (PRESENT(RBFACT)) THEN
            RRBFACT = DBLE(RBFACT)
        ELSE
            RRBFACT = 0.0D0
        ENDIF

        ! Low & high res. limits
        RRAD_MIN = 2.0
        IF (PRESENT(RAD_MIN)) RRAD_MIN = RAD_MIN
        RRAD_MAX = IMG1%logical_dimensions(1)/2
        IF (PRESENT(RAD_MAX)) RRAD_MAX = RAD_MAX

        AABSOLUTE = .FALSE.
        IF(PRESENT(ABSOLUTE)) AABSOLUTE = ABSOLUTE


        ! This factor is 1/N in the 2D case, where NxN are the dimensions of the image. See Eq. 14 in Stewart & Grigorieff (2004).
        ! It is actually the standard deviation of the correlation of a pure noise NxN image with a reference image.
        W = 1.0D0 / DBLE(SQRT(REAL(IMG1%logical_dimensions(1)*IMG1%logical_dimensions(2)*IMG1%logical_dimensions(3))))


        ! Initialise
        NUM = 0.0D0
        DENOM = 0.0D0

        ! Debug
        !PRINT *, '**debug(CC3_FINALISE): numerator per shell before cubing:'
        !DO RAD=1,IMG1%logical_dimensions(1)/2
        !   PRINT *, RAD, NUM_DENOMS_RINGS(1,RAD), NUM_DENOMS_RINGS(2,RAD), NUM_DENOMS_RINGS(3,RAD)
        !ENDDO

        ! Loop over rings/shells
        DO RAD=NINT(RRAD_MIN),NINT(RRAD_MAX)-1
            WGT = EXP(-RRBFACT*(RAD-1)**2)
            TMP = NUM_DENOMS_RINGS(1,RAD)/SQRT(NUM_DENOMS_RINGS(2,RAD)*NUM_DENOMS_RINGS(3,RAD))
            IF (AABSOLUTE) THEN
                R = TMP**2/(W+ABS(TMP)) ! Eq. 16 in Stewart & Grigorieff
            ELSE
                R = TMP*ABS(TMP)/(W+ABS(TMP)) ! Eq. 16 in Stewart & Grigorieff
            ENDIF
            IF (DEBUG) THEN
                WRITE(*,'(A,I0,1X,3(G15.3,1X))') '**debug(CC3_FINALISE): rad, tmp, r, num_denoms_rings(3) = ', &
                                                RAD, TMP, R, NUM_DENOMS_RINGS(3,RAD)
            ENDIF
            ! Accumulate sums for the averaging over rings/shells
            WGT = EXP(-RRBFACT*(RAD-1)**2)  ! B-factor weighting
            NUM = NUM + R**3 * WGT
            DENOM = DENOM + WGT
        ENDDO

        ! Average
        CC3_FINAL = NUM / DENOM

        ! Cubic root
        !CC3_FINAL = CC3_FINAL**(1./3.) ! This is not defined for negative numbers!
        CC3_FINAL = EXP(LOG(ABS(NUM)/DENOM)/3.0D0)*NUM/ABS(NUM)

        ! In case we went beyond [-1.0,1.0]
        IF (CC3_FINAL .LT. -1.0)  CC3_FINAL = -1.0
        IF (CC3_FINAL .GT. 1.0)   CC3_FINAL = 1.0
    END FUNCTION CC3_FINALISE

    !>  \brief  Compute the real-space correlation between two images over a mask
    REAL FUNCTION CORRELATION(IMG1,IMG2,MASK,NORMALISE)
        use images
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(INOUT)   ::  IMG1, IMG2, MASK
        LOGICAL ,       INTENT(IN)  ::  NORMALISE
        ! Private variables
        INTEGER             ::  I,J,K,NUM_OF_VOXELS
        DOUBLE  PRECISION   ::  CROSS_PRODUCT, SUM_OF_SQUARES1, SUM_OF_SQUARES2, SUM1, SUM2
        LOGICAL             ::  DO_FTS
        ! Start work

        !
        ! Work out whether we're already in real space
        !
        DO_FTS = .FALSE.
        IF (.not. img1%is_in_real_space) THEN
            CALL BWD_FT(IMG1)
            CALL BWD_FT(IMG2)
            DO_FTS = .TRUE.
        ENDIF


        NUM_OF_VOXELS = 0
        CROSS_PRODUCT = 0.0
        SUM_OF_SQUARES1 = 0.0
        SUM_OF_SQUARES2 = 0.0
        SUM1 = 0.0
        SUM2 = 0.0
        DO K=1,MIN(IMG1%logical_dimensions(3),IMG2%logical_dimensions(3))
            DO J=1,MIN(IMG1%logical_dimensions(2),IMG2%logical_dimensions(2))
                DO I=1,MIN(IMG1%logical_dimensions(1),IMG2%logical_dimensions(1))
                    IF (MASK%real_values(I,J,K) .NE. 0.0) THEN
                        NUM_OF_VOXELS   = NUM_OF_VOXELS     +   1
                        CROSS_PRODUCT   = CROSS_PRODUCT     +   IMG1%real_values(I,J,K) * IMG2%real_values(I,J,K)
                        SUM_OF_SQUARES1 = SUM_OF_SQUARES1   +   IMG1%real_values(I,J,K) **2
                        SUM_OF_SQUARES2 = SUM_OF_SQUARES2   +   IMG2%real_values(I,J,K) **2
                        SUM1            = SUM1              +   IMG1%real_values(I,J,K)
                        SUM2            = SUM2              +   IMG2%real_values(I,J,K)
                    ENDIF
                ENDDO
            ENDDO
        ENDDO
        CROSS_PRODUCT   = CROSS_PRODUCT     /   NUM_OF_VOXELS
        SUM_OF_SQUARES1 = SUM_OF_SQUARES1   /   NUM_OF_VOXELS
        SUM_OF_SQUARES2 = SUM_OF_SQUARES2   /   NUM_OF_VOXELS
        SUM1            = SUM1              /   NUM_OF_VOXELS
        SUM2            = SUM2              /   NUM_OF_VOXELS

        ! The result
        IF (NORMALISE) THEN
            CORRELATION = (CROSS_PRODUCT - SUM1*SUM2) / (SQRT(SUM_OF_SQUARES1-SUM1**2)*SQRT(SUM_OF_SQUARES2-SUM2**2))
        ELSE
            CORRELATION = (CROSS_PRODUCT - SUM1*SUM2)
        ENDIF

        ! Images back to reciprocal space if that's how they were supplied
        IF (DO_FTS) THEN
            CALL FWD_FT(IMG1)
            CALL FWD_FT(IMG2)
        ENDIF

    END FUNCTION CORRELATION

END MODULE IMAGES_COMPARISON
