!>  \brief  Methods to perform operations on images which require helical thinking
MODULE IMAGES_HELICES

    USE IMAGES
    USE HELICES

    IMPLICIT NONE

    CONTAINS

    SUBROUTINE IMAGES_HELICES_UNIT_TESTS()
        IMPLICIT NONE
        CALL IMAGES_HELIX_UNIT_TEST_1()
        CALL IMAGES_HELIX_UNIT_TEST_2()
    END SUBROUTINE IMAGES_HELICES_UNIT_TESTS

    SUBROUTINE IMAGES_HELIX_UNIT_TEST_1()
        IMPLICIT NONE
        ! Private variables
        TYPE(HELICAL_SYMMETRY)  ::  HELIX
        TYPE(IMAGE)             ::  SELF
        REAL, ALLOCATABLE       ::  PHI(:),Z(:)
        INTEGER                 ::  I
        ! Start work
        SELF%logical_dimensions = [64,64,64]
        CALL HELICAL_SYMMETRY_INIT(HELIX,1.403,-22.03,PIXEL_SIZE=1.163)
        CALL VOLUME_LIST_H_SUBUNITS_POLAR(SELF%logical_dimensions(3),HELIX,PHI,Z)
        DO I=1,SIZE(PHI)
            WRITE(*,'(I3,1X,2(F7.2,1X))') I, PHI(I), Z(I)
        ENDDO
    END SUBROUTINE IMAGES_HELIX_UNIT_TEST_1

    SUBROUTINE IMAGES_HELIX_UNIT_TEST_2()
        use images
        IMPLICIT NONE
        ! Private variables
        TYPE(IMAGE)             ::  SELF, SELF_AMPL, SELF_PHASES
        TYPE(HELICAL_SYMMETRY)  ::  HELIX
        REAL,   PARAMETER       ::  PSIZE = 2.326
        ! Start work
        WRITE(*,'(A)') '**info(IMAGES_HELIX_UNIT_TEST_2): entering'
        CALL HELICAL_SYMMETRY_INIT(HELIX,1.403,-22.03,PIXEL_SIZE=PSIZE)
        SELF%logical_dimensions = [200,200,1]
        WRITE(*,'(A)') '**info(IMAGES_HELIX_UNIT_TEST_2): simulating layer lines'
        CALL IMAGE_SIMULATE_LAYER_LINES(SELF,HELIX,R_IN=10.0/PSIZE,R_OUT=105.0/PSIZE)
        !CALL IMAGE_SIMULATE_LAYER_LINES(SELF,HELIX,R_IN=49.0,R_OUT=51.0)

        ! Dump the amplitudes of the layer lines
        CALL AMPL(SELF,SELF_AMPL)
        CALL WRITE_TO_DISK(SELF_AMPL,'dbg_ampl.mrc')

        ! Dump the phases
        CALL SELF%ComputeAmplitudes(SELF_PHASES,PHASES=.TRUE.)
        CALL WRITE_TO_DISK(SELF_PHASES,'dbg_phases.mrc')

        ! Dump the back FT, which should look like a projection of a helix
        CALL SHIFT_REAL_SPACE_ORIGIN(SELF)
        CALL BWD_FT(SELF)
        CALL WRITE_TO_DISK(SELF,'dbg_real_space.mrc')
        WRITE(*,'(A)') '**info(IMAGES_HELIX_UNIT_TEST_2): done'
    END SUBROUTINE IMAGES_HELIX_UNIT_TEST_2

    !>  \brief  Simulate diffraction pattern of a helical lattice
    !!  \todo   Add linear interpolation to get better representation of layer lines that fall between pixels
    SUBROUTINE IMAGE_SIMULATE_LAYER_LINES(SELF,HELIX,R_IN,R_OUT)
#ifdef __INTEL_COMPILER
        USE IFPORT
#endif
        USE RUNTIME_PARAMETERS, ONLY : PI
        USE HELICES, ONLY : HELIX_FIND_LAYER_LINES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                    INTENT(INOUT)   ::  SELF        !<  Two-dimensional image
        TYPE(HELICAL_SYMMETRY),         INTENT(IN)      ::  HELIX       !<  Helical symmetry parameters
        REAL,                           INTENT(IN)      ::  R_IN        !<  Inner radius of filament (pixels)
        REAL,                           INTENT(IN)      ::  R_OUT       !<  Outer radius of filament (pixels)
        ! Private variables
        INTEGER,    ALLOCATABLE ::  NS(:)
        REAL,       ALLOCATABLE ::  BIGZS(:), SMALLZS(:), PHIS(:)
        REAL,       PARAMETER   ::  AXIAL_ROT   =   0.0 !<  Extra rotation of the helix around its axis
        REAL                    ::  R_MID
        INTEGER                 ::  I,II,LOGI(3), PHYS(3)
        COMPLEX                 ::  CURRENT_VALUE, CURRENT_VALUE_R
        REAL                    ::  BIGPSI, BIGZ, BIGR, SMALLR
        INTEGER                 ::  CURRENT_LAYER_LINE, N, CURRENT_H_SUB
        INTEGER                 ::  Z_SIGN
        LOGICAL,    PARAMETER   ::  DEBUG   =   .TRUE.
        REAL,       PARAMETER   ::  DEBUG_PSIZE = 2.326
        ! Start work

#ifdef NAGFOR_COMPILER
        WRITE(*,'(A)') '**WARNING(IMAGE_SIMULATE_LAYER_LINES): Bessel functions are not supported by nagfor. Will return blank.'
#endif
#ifdef PG_COMPILER
        WRITE(*,'(A)') '**WARNING(IMAGE_SIMULATE_LAYER_LINES): Bessel functions are not supported by pgfortran. Will return blank.'
#endif

        ! Make sure the image is set up OK
        IF (self%IsAVolume()) THEN
            CALL IMAGE_DEALLOCATE(SELF)
            SELF%logical_dimensions = [SELF%logical_dimensions(1),SELF%logical_dimensions(2),1]
        ENDIF
        IF (.NOT. SELF%IsAllocated()) CALL SELF%Allocate()
        SELF%real_values = 0.0
        SELF%is_in_real_space   = .false.

        ! Check the helical parameters are in pixel units
        IF (.NOT. HELIX%PIXEL_UNITS) THEN
            WRITE(*,'(A)') '**ERROR(IMAGE_SIMULATE_LAYER_LINES): helical parameters not given in pixel units'
            CALL TERMINATE('Fatal error in IMAGE_SIMULATE_LAYER_LINES')
        ENDIF

        ! Get the positions and Bessel order of the layer lines
        CALL HELIX_FIND_LAYER_LINES(HELIX,BIGZS,NS)

        IF (DEBUG) THEN
            DO I=1,SIZE(NS)
                IF (ABS(BIGZS(I)) .LE. 0.5) THEN
                IF (BIGZS(I) .NE. 0.) THEN
                    WRITE(*,'(I6,1X,2(F7.2,1X),I9)') I, BIGZS(I)/DEBUG_PSIZE, 1./BIGZS(I)*DEBUG_PSIZE, NS(I)
                ELSE
                    WRITE(*,'(I6,1X,2(F7.2,1X),I9)') I, BIGZS(I)/DEBUG_PSIZE,                    999., NS(I)
                ENDIF
                ENDIF
            ENDDO
        ENDIF

        ! Get the positions of subunits that fit into the volume
        CALL VOLUME_LIST_H_SUBUNITS_POLAR(SELF%logical_dimensions(1),HELIX,PHIS,SMALLZS)
        ! Convert phi coordinates to radians
        DO I=1,SIZE(PHIS)
            PHIS(I) = PHIS(I)/180.0 * PI
        ENDDO

        ! Mean radius
        R_MID = (R_OUT-R_IN)*0.5

        !
        BIGPSI = AXIAL_ROT/180.*PI

        ! Loop over the layer lines
        DO CURRENT_LAYER_LINE=1,SIZE(BIGZS)
            BIGZ=BIGZS(CURRENT_LAYER_LINE)
            IF (ABS(BIGZ) .LT. 0.5) THEN
                ! Loop over the sign of bigZ
                DO Z_SIGN=1,1
                    IF (Z_SIGN .EQ. 2) BIGZ=BIGZ*(-1.0)
                    N=NS(CURRENT_LAYER_LINE)
                    ! Loop over the reciprocal pixels along the current layer line
                    DO I=1,MAXIMUM_COMPLEX_ADDRESS_DIM1(SELF)
                        CURRENT_VALUE=(0.0,0.0)
                        II=I-1
                        BIGR=REAL(II)/REAL(SELF%logical_dimensions(1))
                        ! Loop over real-space radius, small r
                        DO SMALLR=CEILING(R_IN),FLOOR(R_OUT)
                            CURRENT_VALUE_R = (0.0,0.0)
                            ! Loop over helical subunits
                            DO CURRENT_H_SUB=1,SIZE(PHIS)
                                CURRENT_VALUE_R = CURRENT_VALUE_R &
                                                + EXP(CMPLX(0.0,2.0*PI*BIGZ*SMALLZS(CURRENT_H_SUB)-REAL(N)*PHIS(CURRENT_H_SUB)))
                            ENDDO
                            !print *, N, 2.0*PI*REAL(SMALLR)*BIGR, SMALLR, CURRENT_VALUE_R
                            !call flush(6)
! NAG doesn't implement Bessel functions yet, even though it's F2008 standard
#ifdef NAGFOR_COMPILER
                            CURRENT_VALUE_R = 0.0
#else
#ifdef PG_COMPILER
                            CURRENT_VALUE_R = 0.0
#else
                            CURRENT_VALUE_R = CURRENT_VALUE_R * REAL(SMALLR) * &
#ifdef __INTEL_COMPILER
                                                BesJN(  &
#else
                                                ! The Fortran 2008 intrinsic
                                                BESSEL_JN( &
#endif
                                                N,2.0*PI*REAL(SMALLR)*BIGR)
#endif
#endif
                            CURRENT_VALUE = CURRENT_VALUE + CURRENT_VALUE_R
                        ENDDO ! End of loop over small r
                        CURRENT_VALUE = CURRENT_VALUE * EXP(CMPLX(0.0,REAL(N)*(BIGPSI+PI*0.5)))
                        ! Now that we have the value for the layer line at this point, we need to add
                        ! it to the image, but since the layer line probably falls between pixels, we need to interpolate
                        ! As an initial test, we will cheat and do nearest-neighbour
                        LOGI = [II,NINT(BIGZ*SELF%logical_dimensions(2)),1]
                        CALL COMP_ADDR_PHYS_FROM_LOGI(SELF,LOGI,PHYS)
                        SELF%complex_values(PHYS(1),PHYS(2),PHYS(3)) = CURRENT_VALUE
                    ENDDO ! End of loop over pixels along the layer line
                ENDDO ! End of loop over bigZ sign
            ENDIF ! End of check for Z < Nyquist
        ENDDO ! End of loop over layer lines

    END SUBROUTINE IMAGE_SIMULATE_LAYER_LINES

    !>  \brief  List the polar coordinates of the helical subunits within a volume
    SUBROUTINE VOLUME_LIST_H_SUBUNITS_POLAR(Z_DIM,HELIX,PHI,Z)
        USE EULERANGLES, ONLY : DEGREES_M180_TO_180
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        INTEGER,                        INTENT(IN)      ::  Z_DIM       !<  Number of pixels
        TYPE(HELICAL_SYMMETRY),         INTENT(IN)      ::  HELIX       !<  Helical symmetry operator
        REAL,           ALLOCATABLE,    INTENT(INOUT)   ::  PHI(:)      !<  List of phi values (degrees)
        REAL,           ALLOCATABLE,    INTENT(INOUT)   ::  Z(:)        !<  List of Z values (pixels)
        ! Private variables
        INTEGER ::  ORI(3), NUM_SUBUNITS, NUM_SUBUNITS_ABOVE, NUM_SUBUNITS_BELOW,I,CURRENT_SUBUNIT
        ! Start work

        IF (.NOT. HELIX%PIXEL_UNITS) CALL TERMINATE('Need helix in pixel units')

        ! Work out the number of helical subunits that can fit in the volume
        ORI = INDEX_OF_CENTRAL_PIXEL_FROM_DIM(Z_DIM)
        NUM_SUBUNITS_ABOVE = FLOOR((Z_DIM-ORI(3))/HELIX%RISE)
        NUM_SUBUNITS_BELOW = FLOOR((ORI(3)-1)/HELIX%RISE)
        NUM_SUBUNITS = 1 + NUM_SUBUNITS_ABOVE + NUM_SUBUNITS_BELOW
        WRITE(*,'(A,I0)') '**debug(VOLUME_LIST_H_SUBUNITS_POLAR): expected num subunits = ', NUM_SUBUNITS

        ! Allocate memory for output arrays
        IF (ALLOCATED(PHI)) DEALLOCATE(PHI)
        IF (ALLOCATED(Z)) DEALLOCATE(Z)
        ALLOCATE(PHI(NUM_SUBUNITS),Z(NUM_SUBUNITS))

        ! Loop over the subunits
        DO I=-NUM_SUBUNITS_BELOW,NUM_SUBUNITS_ABOVE
            CURRENT_SUBUNIT=I+NUM_SUBUNITS_BELOW+1
            PHI(CURRENT_SUBUNIT) = DEGREES_M180_TO_180(REAL(I)*HELIX%TWIST)
            Z(CURRENT_SUBUNIT)   =                     REAL(I)*HELIX%RISE
        ENDDO

    END SUBROUTINE VOLUME_LIST_H_SUBUNITS_POLAR

    !>  \brief  Given a small cubic helically-symmetric volume, obtain an elongated projection without extending the helix in 3D
    !!
    !!  \todo   Make this work with proper subunit phasing in the projections (see comments)
    SUBROUTINE IMAGE_PROJECT_LONG_HELIX(SELF,PROJ,HELIX,PHI_AT_ORIGIN,INTERPOL_SCHEME,HELIX_MAX_RAD)
        USE INTERPOLATION
        USE IMAGES_THREED
        use images
        use UsefulFunctions, only : IsEven
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(INOUT)   ::  SELF            !<  Volume with helically-symmetric data (no Z-masking!). Expected to be cubic, with helical axis = 3rd axis.
        TYPE(IMAGE),                INTENT(INOUT)   ::  PROJ            !<  Projection image. Overwritten. Must have dimensions already set.
        TYPE(HELICAL_SYMMETRY),     INTENT(IN)      ::  HELIX           !<  Helical symmetry operator
        REAL,                       INTENT(IN)      ::  PHI_AT_ORIGIN   !<  Phi (twist) angle of projection at origin
        TYPE(INTERPOLATION_SCHEME), INTENT(IN)      ::  INTERPOL_SCHEME !<  Interpolation scheme to be used
        REAL,                       INTENT(IN)      ::  HELIX_MAX_RAD   !<  Max radius of helical object (pixels)
        ! Private variables
        INTEGER     ::  CURRENT_SHIFT
        INTEGER     ::  PROJ_MIN_LOGI(3), PROJ_TMP_MIN_LOGI(3)
        INTEGER     ::  PROJ_MIN_PHYS(3), PROJ_TMP_MIN_PHYS(3)
        INTEGER     ::  PROJ_MAX_LOGI(3), PROJ_TMP_MAX_LOGI(3)
        INTEGER     ::  PROJ_MAX_PHYS(3), PROJ_TMP_MAX_PHYS(3)
        INTEGER     ::  PROJ_ORIGIN(3), PROJ_TMP_ORIGIN(3)
        REAL        ::  CURRENT_PHI
        TYPE(IMAGE) ::  PROJ_TMP
        INTEGER     ::  NUM_TMP_IMAGES, CURRENT_TMP_IMAGE
        INTEGER     ::  SEGMENT_LENGTH
        TYPE(WINDOW)    ::  WIN
        ! Start work

        ! Check input 3D is cubic
        IF (.NOT. self%IsSquare()) THEN
            CALL TERMINATE('Fatal error in IMAGE_PROJECT_LONG_HELIX')
        ENDIF

        ! Check projection image has assigned dimensions and is allocated
        IF (.NOT. PROJ%IsAllocated() .OR. ANY(PROJ%logical_dimensions(1:2) .EQ. 1)) THEN
            WRITE(*,'(A,L1,1X,3(I0,1X))') '**ERROR(IMAGE_PROJECT_LONG_HELIX): projection image allocated? dim? ', &
                                        PROJ%IsAllocated(), PROJ%logical_dimensions
            CALL TERMINATE('Fatal error in IMAGE_PROJECT_LONG_HELIX')
        ENDIF

        ! Check 3D input dimension matches dim(1) of projection
        IF (SELF%logical_dimensions(1) .NE. PROJ%logical_dimensions(1)) THEN
            WRITE(*,'(A,2(3(I0,1X)))') '**ERROR(IMAGE_PROJECT_LONG_HELIX): incompatible dimensions: ', &
                                        SELF%logical_dimensions, PROJ%logical_dimensions
			CALL TERMINATE('Fatal error in IMAGE_PROJECT_LONG_HELIX')
        ENDIF

        ! Setup interpol window
        CALL ALLOCATE_WINDOW(WIN,INTERPOL_SCHEME%WINDOW_WIDTH,.FALSE.,.TRUE.)

        ! Setup temporary projection image
        PROJ_TMP%logical_dimensions = SELF%logical_dimensions
        PROJ_TMP%logical_dimensions(3) = 1
        CALL IMAGE_ALLOCATE(PROJ_TMP)

        ! Because the projection is only computed within the circle internal to the 3D cube, we will trim the ends
        ! of the projection so that only the central part of the projection (not affected by the circle) will
        ! be used
        SEGMENT_LENGTH = 2 * FLOOR(SQRT(PROJ_TMP%logical_dimensions(2)**2/4.0 - HELIX_MAX_RAD**2))
        WRITE(*,'(A,F0.4)') '**debug(IMAGE_PROJECT_LONG_HELIX): helix max rad = ', HELIX_MAX_RAD
        WRITE(*,'(A,I0)') '**debug(IMAGE_PROJECT_LONG_HELIX): segment length = ', SEGMENT_LENGTH
		WRITE(*,'(A,3(3(I0,1X)))') '**debug(IMAGE_PROJECT_LONG_HELIX): dims = ', SELF%logical_dimensions, &
		                            PROJ_TMP%logical_dimensions, PROJ%logical_dimensions


        ! Physical address of the origins
        if (proj%is_in_real_space .and. proj%origin_at_center) then
            proj_origin = proj%physical_address_of_box_center
        else
            proj_origin = [1,1,1]
        endif
        if (proj_tmp%is_in_real_space .and. proj_tmp%origin_at_center) then
            proj_tmp_origin = proj_tmp%physical_address_of_box_center
        else
            proj_tmp_origin = [1,1,1]
        endif

        ! How many images fit into the output?
        NUM_TMP_IMAGES = CEILING(REAL(PROJ%logical_dimensions(2)) / REAL(SEGMENT_LENGTH))
        IF (IsEven(NUM_TMP_IMAGES)) NUM_TMP_IMAGES = NUM_TMP_IMAGES + 1
        WRITE(*,'(A,I0)') '**debug(IMAGE_PROJECT_LONG_HELIX): number of projections = ', NUM_TMP_IMAGES

        DO CURRENT_TMP_IMAGE=1,NUM_TMP_IMAGES
            CURRENT_SHIFT = SEGMENT_LENGTH * (CURRENT_TMP_IMAGE - (NUM_TMP_IMAGES+1)/2)
            !
            WRITE(*,'(A,I0,A,I0,A,I0)') '**info(IMAGE_PROJECT_LONG_HELIX): on projection ', CURRENT_TMP_IMAGE, &
                                        ' of ', NUM_TMP_IMAGES, ', with shift = ', CURRENT_SHIFT
            ! Work out phi angle - THIS NEEDS TO BE IMPROVED
            CURRENT_PHI = PHI_AT_ORIGIN + REAL(CURRENT_SHIFT) * HELIX%TWIST / HELIX%RISE ! Does not phase subunits correctly!
            ! Project
            CALL VOLUME_PROJECT_LEAN(SELF,PROJ_TMP,DBLE(CURRENT_PHI),90.0D0,90.0D0,INTERPOL_SCHEME,WIN,MAX_CYL_RAD=HELIX_MAX_RAD)
            ! The bounds of the projection images
            CALL MIN_MAX_LOGICAL_ADDRESS(PROJ,      PROJ_MIN_LOGI,      PROJ_MAX_LOGI)
            CALL MIN_MAX_LOGICAL_ADDRESS(PROJ_TMP,  PROJ_TMP_MIN_LOGI,  PROJ_TMP_MAX_LOGI)
            ! We will only use the parts of PROJ_TMP not affected by circle mask
            ! (note: we know segment_length is even, since it's the result of 2*something)
            PROJ_TMP_MIN_LOGI(2) = -SEGMENT_LENGTH/2
            PROJ_TMP_MAX_LOGI(2) = SEGMENT_LENGTH/2-1
            ! Work out the addresses proj_tmp hovers over
            PROJ_MIN_PHYS = PROJ_ORIGIN + PROJ_TMP_MIN_LOGI
            PROJ_MAX_PHYS = PROJ_ORIGIN + PROJ_TMP_MAX_LOGI
            PROJ_TMP_MIN_PHYS = PROJ_TMP_ORIGIN + PROJ_TMP_MIN_LOGI
            PROJ_TMP_MAX_PHYS = PROJ_TMP_ORIGIN + PROJ_TMP_MAX_LOGI
            PROJ_MIN_PHYS(2) = PROJ_MIN_PHYS(2) + CURRENT_SHIFT
            PROJ_MAX_PHYS(2) = PROJ_MAX_PHYS(2) + CURRENT_SHIFT
            ! Maybe we're over one of the edges of the output image?
            IF (PROJ_MIN_PHYS(2) .LT. 1) THEN
                PROJ_TMP_MIN_PHYS(2) = PROJ_TMP_MIN_PHYS(2) + (1 - PROJ_MIN_PHYS(2))
                PROJ_MIN_PHYS(2) = 1
            ELSE IF (PROJ_MAX_PHYS(2) .GT. PROJ%logical_dimensions(2)) THEN
          		PROJ_TMP_MAX_PHYS(2) = PROJ_TMP_MAX_PHYS(2) - (PROJ_MAX_PHYS(2) - PROJ%logical_dimensions(2))
                PROJ_MAX_PHYS(2) = PROJ%logical_dimensions(2)
            ENDIF
            PROJ%real_values(PROJ_MIN_PHYS(1):PROJ_MAX_PHYS(1),PROJ_MIN_PHYS(2):PROJ_MAX_PHYS(2),1) = &
                        PROJ_TMP%real_values(PROJ_TMP_MIN_PHYS(1):PROJ_TMP_MAX_PHYS(1),PROJ_TMP_MIN_PHYS(2):PROJ_TMP_MAX_PHYS(2),1)
        ENDDO
        !CALL WRITE_TO_DISK(PROJ_TMP,'dbg_proj_tmp.mrc')
    END SUBROUTINE IMAGE_PROJECT_LONG_HELIX

    !>  \brief  Mask a helical density, for example in preparation for use as a reference or in an FSC experiment.
    !!
    !!
    !! Masking: 3 masks are calculated & combined:
    !! 1. A cylindrical mask, safely around the helix
    !! 2. A "molecular envelope mask"
    !! 3. A Z-edge mask, to ensure no discontinuity at the edges of the box
    !!
    !! The final step is to soften this combined mask.
    !!
    SUBROUTINE IMAGE_HELIX_MASK(SELF,HELIX,ASYM_UNIT_MW,PIXEL_SIZE,HELIX_MAX_RAD,F_MASK,F_SMOOTH,HARD_MASK,MASK_RESOLUTION)
        USE MASKS
        USE IMAGES_ARITHM
        use RUNTIME_PARAMETERS, only : DEFAULT_FALLOFF_PIXELS
        use Globals, only : protein_density
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                        INTENT(INOUT)   ::  SELF            !<  Volume to be masked. Overwritten with masked version.
        TYPE(HELICAL_SYMMETRY),             INTENT(IN)      ::  HELIX           !<  Helical symmetry of the reconstruction
        REAL,                               INTENT(IN)      ::  ASYM_UNIT_MW    !<  Expected molecular weight (Da) of the asymmetric unit
        REAL,                               INTENT(IN)      ::  PIXEL_SIZE      !<  Voxel size (Angstroms)
        REAL,                               INTENT(IN)      ::  HELIX_MAX_RAD   !<  Maximum radius of the helical assembly (pixels)
        REAL,                   OPTIONAL,   INTENT(OUT)     ::  F_MASK          !<  The fraction of voxels within the binary mask
        REAL,                   OPTIONAL,   INTENT(OUT)     ::  F_SMOOTH        !<  The fraction of voxels within the smooth mask (computed a la Sindelar, 2011)
        TYPE(IMAGE),            OPTIONAL,   INTENT(INOUT)   ::  HARD_MASK       !<  If present, the binary mask will be returned
        REAL,                   OPTIONAL,   INTENT(IN)      ::  MASK_RESOLUTION !<  In 1/pixel. If present, the mask will be limited to this resolution. If absent, half the Nyquist will be used as default.
        ! Private variables
        REAL                                        ::  TOTAL_MW, PCT_IN_MASK
        TYPE(IMAGE)                                 ::  MASK3D
        LOGICAL,        PARAMETER                   ::  DEBUG   =   .FALSE.
        REAL                                        ::  MMASK_RESOLUTION
        REAL                                        ::  Z_ENDS_MASK
        REAL,           PARAMETER                   ::  MAX_MASK_RESOLUTION = 5.0  !   In Angstroems, the highest resolution for the mask
        REAL                                        ::  COSINE_FALLOFF_WIDTH
        ! Start work

        ! If the image is blank, no need to do anything!
        IF (SELF%IsBlank()) THEN
            WRITE(*,'(A)') '**info(IMAGE_HELIX_MASK): input is blank, not doing anything'
        ELSE
            ! Check the helical parameters are in pixel units
            IF (.NOT. HELIX%PIXEL_UNITS) THEN
                WRITE(*,'(A)') '**ERROR(IMAGE_HELIX_MASK): was expecting helical parameters in pixel units'
                CALL TERMINATE('Fatal error in IMAGE_HELIX_MASK')
            ENDIF

            ! Check the image has origin at center
            IF (.NOT. SELF%ORIGIN_AT_CENTER) THEN
                CALL SHIFT_REAL_SPACE_ORIGIN(SELF)
            ENDIF

            ! Check the image is in real space
            IF (.not. self%is_in_real_space) THEN
                CALL BWD_FT(SELF)
            ENDIF

            ! What resolution will the mask be at? (in reciprocal pixels; 0.5 is Nyquist)
            IF (PRESENT(MASK_RESOLUTION)) THEN
                MMASK_RESOLUTION = MIN(MASK_RESOLUTION,0.5)
            ELSE
                MMASK_RESOLUTION = 0.25
            ENDIF
            IF (PIXEL_SIZE/MMASK_RESOLUTION .LT. MAX_MASK_RESOLUTION) MMASK_RESOLUTION = PIXEL_SIZE / MAX_MASK_RESOLUTION
            WRITE(*,'(A,F0.2,A)') '**info(IMAGE_HELIX_MASK): will use envelope mask with resolution of ', &
                                    PIXEL_SIZE/MMASK_RESOLUTION, ' Angstroems'

            ! The width of the cosine falloff
            COSINE_FALLOFF_WIDTH = DEFAULT_FALLOFF_PIXELS
            COSINE_FALLOFF_WIDTH = 1/MMASK_RESOLUTION


            ! Diagnostic dump
            IF (DEBUG) CALL WRITE_TO_DISK(SELF,'dbg_before_masking.mrc')

            ! First, impose the cylindrical mask
            IF (HELIX_MAX_RAD .EQ. 0.0) THEN
                CALL CYL_MASK(SELF,.9,NINT(COSINE_FALLOFF_WIDTH),ZERO_FLOAT_INSIDE=.TRUE.)
            ELSE
                CALL CYL_MASK(SELF,HELIX_MAX_RAD,NINT(COSINE_FALLOFF_WIDTH),ZERO_FLOAT_INSIDE=.TRUE.)
            ENDIF

            ! Debug
            IF (DEBUG) THEN
                PRINT *, '**debug(IMAGE_HELIX_MASK): HELIX_MAX_RAD =', HELIX_MAX_RAD
                CALL WRITE_TO_DISK(SELF,'dbg_cyl_mask.mrc')
            ENDIF

            ! Mask the Z edges
            !Z_ENDS_MASK = MIN(COSINE_FALLOFF_WIDTH,REAL(SELF%logical_dimensions(3)/2.*0.1))
            Z_ENDS_MASK = COSINE_FALLOFF_WIDTH
            Z_ENDS_MASK = MAX(Z_ENDS_MASK,5.0)
            IF (DEBUG) WRITE(*,'(A,F0.3)') '**debug(IMAGE_HELIX_MASK): z-ends mask width = ', Z_ENDS_MASK
            CALL IMAGE_MASK_Z_ENDS_INPLACE(SELF,NINT(Z_ENDS_MASK),SOFT=.TRUE.)

            ! Debug
            IF (DEBUG) THEN
                PRINT *, '**debug(IMAGE_HELIX_MASK): HELIX_MAX_RAD =', HELIX_MAX_RAD
                CALL WRITE_TO_DISK(SELF,'dbg_cylz_mask.mrc')
            ENDIF

            ! Masking, with an envelope this time
            IF (ASYM_UNIT_MW .EQ. 0.0) THEN
                PRINT '(A)', '**WARNING(IMAGE_HELIX_MASK): asymetric unit MW not given, not masking molecular envelope. ',  &
                                        'If you can estimate the MW of the asymetric unit, it is highly recommended you do so.'
                CALL IMAGE_ALLOCATE(MASK3D,SELF)
                MASK3D%real_values = 1.0
            ELSE
                ! First, work out the total MW of the helical assembly present in the box, taking into account
                ! the Z masking
                ! If the cosine-edge falloff width of the Z mask is N pixels, then this would be equivalent to
                ! zeroing N/2 slices of the volume completely, at each Z boundary; i.e. N in total.
                ! The formula below assumes that this is precisely the case (double-check this is true
                ! by looking carefully at the Z masking routine above)
                TOTAL_MW = ASYM_UNIT_MW * HELIX%AXIAL_SYMMETRY * HELIX%PERPENDICULAR_SYMMETRY * HELIX%NUM_STARTS * &
                                                                    (SELF%logical_dimensions(3)-Z_ENDS_MASK) / HELIX%RISE
                ! Work out the number of voxels in the mask corresponding to the MW
                PCT_IN_MASK = TOTAL_MW / (PROTEIN_DENSITY * PIXEL_SIZE**3)
                IF (DEBUG) THEN
                    !PRINT *, '**debug(): number of voxels = ', PCT_IN_MASK
                    !PRINT *, '**debug(): box volume = ', PRODUCT(SELF%logical_dimensions)
                ENDIF
                ! Convert to percentage of total number of voxels
                PCT_IN_MASK = PCT_IN_MASK * 100. / PRODUCT(SELF%logical_dimensions)
                IF (DEBUG) THEN
                    !PRINT *, '**debug(): total_mw = ', TOTAL_MW
                    !PRINT *, '**debug(): protein_density = ', PROTEIN_DENSITY
                    !PRINT *, '**debug(): pixel_size = ', PIXEL_SIZE
                    !PRINT *, '**debug(): protein_density * pixel_size = ', PROTEIN_DENSITY * PIXEL_SIZE
                ENDIF

                ! Diagnostics
                PRINT '(A,F0.2,A,F0.2,A)', '**info(IMAGE_HELIX_MASK): total MW (', TOTAL_MW, ') corresponds to ', &
                                            PCT_IN_MASK, '% of voxels'
                ! Calculate the molecular envelope
                WRITE(*,'(A,F0.3,A)') '**info(IMAGE_HELIX_MASK): computing low-pass envelope with resolution of ', &
                                        MMASK_RESOLUTION, '/pixel'
                CALL IMAGE_LOW_PASS_MASK(SELF,MASK3D,PCT_IN_MASK,SOFTEN=.FALSE.,LOW_PASS_FREQUENCY=MMASK_RESOLUTION)

                ! Debug
                IF (DEBUG) CALL WRITE_TO_DISK(MASK3D,'dbg_lp_mask.mrc')
            ENDIF


            ! Apply Z-ends mask
            CALL IMAGE_MASK_Z_ENDS_INPLACE(MASK3D,NINT(MAX(Z_ENDS_MASK,MASK3D%logical_dimensions(3)*0.2)),SOFT=.FALSE.)

            ! If requested, compute the fraction of voxels within the binary mask
            IF (PRESENT(F_MASK)) THEN
                F_MASK = MASK3D%SumOfVoxelValuesDouble() / PRODUCT(SELF%logical_dimensions)
            ENDIF

            ! If requested, return the combined mask before softening
            IF (PRESENT(HARD_MASK)) THEN
                !CALL IMAGE_ASSIGNMENT(HARD_MASK,MASK3D)
                HARD_MASK = MASK3D
            ENDIF

            ! Soften the combined mask
            WRITE(*,'(A,F0.2)') '**info(IMAGE_HELIX_MASK): adding cosine falloff to the mask, width = ', COSINE_FALLOFF_WIDTH
            IF (DEBUG) CALL WRITE_TO_DISK(MASK3D,'dbg_mask_before_cos.mrc')
            CALL MASK_ADD_COSINE_FALLOFF(MASK3D,COSINE_FALLOFF_WIDTH)
            IF (DEBUG) CALL WRITE_TO_DISK(MASK3D,'dbg_mask_after_cos.mrc')

            ! If requested, compute the fraction of voxels within softened mask
            IF (PRESENT(F_SMOOTH)) F_SMOOTH = FRACTION_OF_VOXELS_WITHIN_MASK(MASK3D)

            ! Applied the softened combined mask
            CALL MULTIPLY_IMAGES(SELF,MASK3D,SELF)
            WRITE(*,'(A)') '**info(IMAGE_HELIX_MASK): all done'

            ! Diagnostic dump
            IF (DEBUG) CALL WRITE_TO_DISK(SELF,'dbg_masked.mrc')
        ENDIF   ! End of test for blank image

    END SUBROUTINE IMAGE_HELIX_MASK

    !>  \brief  Mask (zeroe) voxels that are not in the helical asymetric unit
    SUBROUTINE IMAGE_MASK_HELICAL_SUBUNIT(SELF,HELIX,MAX_RADIUS)
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                        INTENT(INOUT)   ::  SELF
        TYPE(HELICAL_SYMMETRY),             INTENT(IN)      ::  HELIX
        REAL,                 OPTIONAL,     INTENT(IN)      ::  MAX_RADIUS      !<  In pixels, the maximum radius
        ! Private variables
!        REAL,   PARAMETER       ::  ASYM_UNIT_CENTER(3) = (/ 1.0, 0.0, 0.0 /)   ! The unit vector that points from the center of the
!                                                                                ! volume to the origin of the helical unit cell
        INTEGER                 ::  I,J,K, ORI(3)
        REAL(KIND=8)            ::  X,Y,Z,YSQ,R, PHI, PITCH
        REAL                    ::  MMAX_RADIUS
        TYPE(HELICAL_SYMMETRY)  ::  HELIX_ONESTART
        LOGICAL                 ::  IN_ASYM_UNIT
        ! Start work

        ! The address of the origin
        if (self%is_in_real_space .and. self%origin_at_center) then
            ori = self%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif

        ! A maximal radius
        MMAX_RADIUS = REAL(SELF%logical_dimensions(1))/2.0
        IF (PRESENT(MAX_RADIUS)) THEN
            MMAX_RADIUS=MIN(MMAX_RADIUS,MAX_RADIUS)
        ENDIF

        ! Ensure the helical operator is 1-start
        HELIX_ONESTART = HELIX
        CALL HELIX_TO_ONE_START(HELIX_ONESTART)

        ! Compute the pitch of the helix
        PITCH = HELIX_PITCH(HELIX_ONESTART)

        ! Loop over the image voxels
        DO K=1,SELF%logical_dimensions(3)
            Z = REAL(K-ORI(3),KIND=8)
            !ZZ = Z
            DO J=1,SELF%logical_dimensions(2)
                Y = REAL(J-ORI(2),KIND=8)
                YSQ = Y**2
                DO I=1,SELF%logical_dimensions(1)
                    X = REAL(I-ORI(1),KIND=8)
                    ! Radius
                    R = SQRT(X**2+YSQ)
                    IF (R .GT. MMAX_RADIUS) THEN
                        SELF%real_values(I,J,K) = 0.0
                        CYCLE
                    ELSE
                        ! Work out the longitude (phi)
                        PHI = HELIX_PHI_FROM_XY(X,Y)
                        ! Are we in the asymetric unit?
                        IN_ASYM_UNIT = HELIX_POINT_IS_IN_ASYM(HELIX_ONESTART,Z,PHI)
                        IF (.NOT. IN_ASYM_UNIT) SELF%real_values(I,J,K) = 0.0
                    ENDIF
                ENDDO
            ENDDO
        ENDDO

    END SUBROUTINE IMAGE_MASK_HELICAL_SUBUNIT

    !>  \brief  Impose helical symmetry on a volume
    !!
    !!  See derivation in notes 120328 for more details
    !!  \todo   Add pre-emptive filter to correct for interpolation when imposing odd axial symmetry
    SUBROUTINE IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW(SELF,HELIX,INTERPOL_SCHEME,MAX_RADIUS,PROGRESS_REPORT,QUIET)
        USE INTERPOLATION
        USE ROTATION_MATRICES
        use UsefulFunctions, only : IsOdd, IsEven
        use images, only : eval_window_bounds
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                        INTENT(INOUT)   ::  SELF                !<  Image (volume) to be symmetrised
        TYPE(HELICAL_SYMMETRY),             INTENT(IN)      ::  HELIX               !<  Helical symmetry operator
        TYPE(INTERPOLATION_SCHEME),         INTENT(IN)      ::  INTERPOL_SCHEME     !<  Interpolation scheme
        REAL,       OPTIONAL,               INTENT(IN)      ::  MAX_RADIUS          !<  Maximum outer radius of helix (pixels)
        LOGICAL,    OPTIONAL,               INTENT(IN)      ::  PROGRESS_REPORT     !<  Give some output regularly to let the user know we're working
        LOGICAL,    OPTIONAL,               INTENT(IN)      ::  QUIET               !<  Do not print anything
        ! Private variables
        TYPE(HELICAL_SYMMETRY)  ::  HELIX_ONESTART
        TYPE(WINDOW)            ::  WIN
        !REAL(KIND=8)            ::  PITCH
        INTEGER                 ::  I,J,K
        INTEGER                 ::  ORI(3)
        REAL(KIND=8)            ::  PHI,PPHI,X,Y,Z,YSQ,R
        LOGICAL                 ::  IN_ASYM_UNIT
        REAL(KIND=8)            ::  XX,YY,ZZ

        LOGICAL,    PARAMETER   ::  PREEMPTIVE_FILTER   =   .TRUE.  ! Whether to apply a pre-emptive high-pass filter to compensate for interpolation
        TYPE(IMAGE)             ::  TMPIMG
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        REAL                    ::  MMAX_RADIUS
        INTEGER                 ::  A_SYM
        REAL                    ::  RVALUE
        REAL                    ::  TMPVAL
        !INTEGER                 ::  COUNT_SUCCESSFUL_VOXELS, COUNT_TOTAL_VOXELS
        !LOGICAL                 ::  SUCCESS
        TYPE(IMAGE)             ::  SUBUNIT_VOXELS   ! DEBUG
        !INTEGER                 ::  DIFFICULTY_LEVEL, TOTAL_DIFFICULTY_LEVEL
        INTEGER,    PARAMETER   ::  FOURIER_PAD_FACTOR = 1
        INTEGER                 ::  NEW_DIM(3)
        LOGICAL                 ::  QQUIET
        ! Start work

        QQUIET = .FALSE.
        IF (PRESENT(QUIET)) QQUIET = QUIET

        ! Side dyad not implemented
        IF (HELIX%PERPENDICULAR_SYMMETRY .GT. 1) THEN
            WRITE(*,'(A)') '**ERROR(IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW): perpendicular symmetry not implemented yet'
            CALL TERMINATE('Feature not implemented')
        ENDIF

        ! Assume helix is in pixel units
        IF (.NOT. HELIX%PIXEL_UNITS) THEN
            WRITE(*,'(A)') '**ERROR(IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW): helical symmetry operator must be in pixel units'
            CALL TERMINATE('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW')
        ENDIF

        ! Padding in Fourier space
        IF (FOURIER_PAD_FACTOR .GT. 1) THEN
            WRITE(*,'(A,I0)') '**info(IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW): padding volume in Fourier space by factor of ', &
                                FOURIER_PAD_FACTOR
            IF (SELF%IsReal()) CALL SELF%ForwardFT()
            NEW_DIM = SELF%logical_dimensions * FOURIER_PAD_FACTOR
            CALL PAD(SELF,NEW_DIM)
            CALL SELF%BackwardFT()
        ENDIF


        ! Maximum radius
        MMAX_RADIUS = REAL(SELF%logical_dimensions(1))/2.0
        IF (PRESENT(MAX_RADIUS)) THEN
            IF (MAX_RADIUS .GT. 0.0) MMAX_RADIUS = MAX_RADIUS
        ENDIF
        MMAX_RADIUS = MMAX_RADIUS * REAL(FOURIER_PAD_FACTOR)


        ! With odd axial symmetries, there will be damping due to the interpolation, which can probably
        ! be corrected for, but I don't have the time to implement this now
        !IF (PREEMPTIVE_FILTER .AND. HELIX%AXIAL_SYMMETRY .GT. 1 .AND. IS_ODD(HELIX%AXIAL_SYMMETRY)) THEN
        !    WRITE(*,'(2A)') '**ERROR(IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW): pre-emptive filtering for odd', &
        !                    ' axial symmetry not yet implemented'
        !    CALL TERMINATE('Feature not implemented')
        !ENDIF

        IF (DEBUG) THEN
        !    FAILED_VOXELS = SELF
        !    FAILED_VOXELS%real_values = 0.0
            SUBUNIT_VOXELS = SELF
			SUBUNIT_VOXELS%real_values = 0.0
        ENDIF

        ! Ensure the helical operator is 1-start
        HELIX_ONESTART = HELIX
        CALL HELIX_TO_ONE_START(HELIX_ONESTART)
        IF (FOURIER_PAD_FACTOR .GT. 1) HELIX_ONESTART%RISE = HELIX_ONESTART%RISE * FOURIER_PAD_FACTOR
        IF (.NOT. QQUIET) THEN
            CALL HELICAL_SYMMETRY_PRINT_INFO(HELIX_ONESTART)
            IF (DEBUG) CALL HELICAL_SYMMETRY_PRINT_EXTRA_INFO(HELIX_ONESTART)
        ENDIF

        ! Compute the pitch of the helix
        !PITCH = HELIX_PITCH(HELIX_ONESTART)
        !IF (DEBUG) WRITE(*,'(A,F0.3)') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW): pitch of one-start helix = ', PITCH

        ! The boundaries of the helical subunit
        !IF (DEBUG) THEN
        !    WRITE(*,'(A,2(F0.3,1X))') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW): phi, z boundaries of h subunit = ', &
        !                        HELIX_ASYM_UNIT_PHI_LIMIT(HELIX_ONESTART,PITCH), HELIX_ASYM_UNIT_Z_LIMIT(HELIX_ONESTART,PITCH)
        !    CALL HELICAL_SYMMETRY_PRINT_EXTRA_INFO(HELIX_ONESTART)
        !ENDIF

        ! The address of the origin
        if (self%is_in_real_space .and. self%origin_at_center) then
            ori = self%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif

        ! Prepare the interpolation window
        CALL WIN%AllocateWindow(INTERPOL_SCHEME%WINDOW_WIDTH,self%IsComplex(),self%IsAVolume())

        ! Handle the axial symmetry first - need it applied to the asymetric unit
        IF (HELIX%AXIAL_SYMMETRY .GT. 1) THEN

            TMPIMG = SELF

            ! High-pass filter to compensate for low-pass-filtering effect of interpolation
            IF (PREEMPTIVE_FILTER .AND. IsOdd(HELIX%AXIAL_SYMMETRY)) THEN
                CALL TMPIMG%ForwardFT()
                CALL CORRECT_3D(TMPIMG,INTERPOL_SCHEME%INTERPOLATION_INSTR_FUNCTION,INTERPOL_SCHEME%WINDOWING_INSTR_FUNCTION, &
                                INTERPOL_SCHEME%WINDOW_WIDTH)
                CALL TMPIMG%BackwardFT()
            ELSE
                WRITE(*,'(A)') '**info(IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW): preemptive filter turned off because even sym'
            ENDIF

            ! Loop over voxels
            DO K=1,SELF%logical_dimensions(3)
                Z = REAL(K-ORI(3),KIND=8)
                ZZ = Z
                DO J=1,SELF%logical_dimensions(2)
                    Y = REAL(J-ORI(2),KIND=8)
                    YSQ = Y**2
                    DO I=1,SELF%logical_dimensions(1)
                        X = REAL(I-ORI(1),KIND=8)
                        ! Radius
                        R = SQRT(X**2+YSQ)
                        IF (R .GT. MMAX_RADIUS) CYCLE
                        ! Work out the longitude (phi)
                        PHI = HELIX_PHI_FROM_XY(X,Y)
                        ! Are we in the asymetric unit?
                        IN_ASYM_UNIT = HELIX_POINT_IS_IN_ASYM(HELIX_ONESTART,Z,PHI)
                        IF (IN_ASYM_UNIT) THEN
                            TMPVAL = SELF%real_values(I,J,K) / REAL(HELIX%AXIAL_SYMMETRY)
                            ! If we're in the asymetric unit, apply the axial symmetry
                            DO A_SYM = 2, HELIX%AXIAL_SYMMETRY
                                PPHI = PHI + (A_SYM-1)*(360./REAL(HELIX%AXIAL_SYMMETRY))
                                ! Convert back to Cartesian coordinates (see EULER_TO_ROTMAT)
                                CALL HELIX_XY_FROM_PHI_R(PHI,R,XX,YY)
                                ! Interpolate a value from those coordinates
                                WIN%CENTERED_ON = (/ XX,YY,ZZ /)
                                CALL EVAL_WINDOW_BOUNDS(WIN)
                                CALL WINDOW_FROM_IMAGE_REAL(TMPIMG,WIN)
                                CALL GET_REAL_FROM_WINDOW(  RVALUE,WIN,   &
                                                            INTERPOL_SCHEME%INTERPOLATION_FUNCTION,&
                                                            INTERPOL_SCHEME%WINDOWING_FUNCTION)
                                TMPVAL = TMPVAL + RVALUE / REAL(HELIX%AXIAL_SYMMETRY)
                            ENDDO
                            SELF%real_values(I,J,K) = TMPVAL
                        ENDIF
                    ENDDO
                ENDDO
            ENDDO
        ENDIF

        ! High-pass filter to compensate for low-pass-filtering effect of interpolation
        IF (PREEMPTIVE_FILTER) THEN
            TMPIMG = SELF
            CALL SELF%ForwardFT()
            CALL CORRECT_3D(SELF,INTERPOL_SCHEME%INTERPOLATION_INSTR_FUNCTION,INTERPOL_SCHEME%WINDOWING_INSTR_FUNCTION, &
                            INTERPOL_SCHEME%WINDOW_WIDTH)
            CALL SELF%BackwardFT()
        ELSE
            WRITE(*,'(A)') '**warning(IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW): preemptive filter turned off'
        ENDIF

        ! Loop over the output voxels
        !COUNT_TOTAL_VOXELS = 0
        !COUNT_SUCCESSFUL_VOXELS = 0
        !TOTAL_DIFFICULTY_LEVEL = 0
        DO K=1,SELF%logical_dimensions(3)
            Z = REAL(K-ORI(3),KIND=8)
            DO J=1,SELF%logical_dimensions(2)
                Y = REAL(J-ORI(2),KIND=8)
                YSQ = Y**2
                DO I=1,SELF%logical_dimensions(1)
                    X = REAL(I-ORI(1),KIND=8)
                    ! Radius
                    R = SQRT(X**2+YSQ)
                    IF (R .GT. MMAX_RADIUS) CYCLE
                    !COUNT_TOTAL_VOXELS = COUNT_TOTAL_VOXELS + 1
                    ! Work out the longitude (phi)
                    PHI = HELIX_PHI_FROM_XY(X,Y)
                    ! Are we in the asymetric unit?
                    IN_ASYM_UNIT = HELIX_POINT_IS_IN_ASYM(HELIX_ONESTART,Z,PHI)
                    ! If we were in the asymetric unit, we're done
                    IF (IN_ASYM_UNIT) THEN
                        ! Nothing to do
                        !COUNT_SUCCESSFUL_VOXELS = COUNT_SUCCESSFUL_VOXELS + 1
                        IF (DEBUG) SUBUNIT_VOXELS%real_values(I,J,K) = 1.0
                    ELSE
                        ! If we aren't in the asymetric unit, we need to find the corresponding point within the asymetric unit
                        ! ZZ and PHI are modified
                        CALL HELIX_POINT_TO_ASYM(HELIX_ONESTART,Z,PHI,ZZ,PPHI)
                        ! Convert back to Cartesian coordinates (see EULER_TO_ROTMAT)
                        CALL HELIX_XY_FROM_PHI_R(PPHI,R,XX,YY)
                        ! Interpolate a value from those coordinates
                        WIN%CENTERED_ON = (/ XX,YY,ZZ /)
                        CALL EVAL_WINDOW_BOUNDS(WIN)
                        CALL WINDOW_FROM_IMAGE_REAL(SELF,WIN)
                        CALL GET_REAL_FROM_WINDOW(  RVALUE,WIN,   &
                                                    INTERPOL_SCHEME%INTERPOLATION_FUNCTION,INTERPOL_SCHEME%WINDOWING_FUNCTION)
                        IF (DEBUG) THEN
                            IF (RVALUE .EQ. 0.0) THEN
                                WRITE(*,'(A,4(F0.2,1X))') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW): ZERO AT XX, YY, ZZ, R = ', &
                                                                XX,YY,ZZ, R
                            ENDIF
                        ENDIF
                        SELF%real_values(I,J,K) = RVALUE
                    ENDIF
                ENDDO
            ENDDO
        ENDDO

        IF (PREEMPTIVE_FILTER) THEN
            ! We need to restore the original asymetric unit, because currently, there is a high-passed version in its place
            DO K=1,SELF%logical_dimensions(3)
                Z = REAL(K-ORI(3),KIND=8)
                DO J=1,SELF%logical_dimensions(2)
                    Y = REAL(J-ORI(2),KIND=8)
                    YSQ = Y**2
                    DO I=1,SELF%logical_dimensions(1)
                        X = REAL(I-ORI(1),KIND=8)
                        ! Radius
                        R = SQRT(X**2+YSQ)
                        IF (R .GT. MMAX_RADIUS) CYCLE
                        ! Work out the longitude (phi)
                        PHI = HELIX_PHI_FROM_XY(X,Y)
                        ! Are we in the asymetric unit?
                        IN_ASYM_UNIT = HELIX_POINT_IS_IN_ASYM(HELIX_ONESTART,Z,PHI)
                        ! If we aren't in the asymetric unit, we need to find the corresponding point within the asymetric unit
                        IF (IN_ASYM_UNIT) THEN
                            ! Restore the data
                            SELF%real_values(I,J,K) = TMPIMG%real_values(I,J,K)
                        ENDIF
                    ENDDO
                ENDDO
            ENDDO
        ENDIF

        ! Undo the Fourier-space padding
        IF (FOURIER_PAD_FACTOR .GT. 1) THEN
            WRITE(*,'(A,I0)') '**info(IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW): cropping volume in Fourier space by factor of ', &
                                FOURIER_PAD_FACTOR
            IF (SELF%IsReal()) CALL SELF%ForwardFT()
            NEW_DIM = SELF%logical_dimensions / FOURIER_PAD_FACTOR
            CALL CROP(SELF,NEW_DIM)
            CALL SELF%BackwardFT()
            IF (DEBUG) THEN
                CALL SUBUNIT_VOXELS%ForwardFT()
                CALL CROP(SUBUNIT_VOXELS,NEW_DIM)
                CALL SUBUNIT_VOXELS%ForwardFT()
            ENDIF
        ENDIF

        IF (DEBUG) THEN
            CALL WRITE_TO_DISK(SUBUNIT_VOXELS, 'dbg_subunit.mrc')
        ENDIF
    END SUBROUTINE IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW

    !>  \brief  Equivalent to HEXTEND in frealix1/frealign, this subroutine symmetrises the volume using a helical symmetry operator
    !!
    !!          Warning: this subroutine has not been fully tested with helices involving axial or perpendicular symmetries
    !!
    !!  \todo   Test & debug symmetries (axial & perpendicular)
    SUBROUTINE IMAGE_IMPOSE_HELICAL_SYMMETRY(SELF,HELIX,INTERPOL_SCHEME,NUM_CENTRAL_SLICES,MAX_RADIUS,PROGRESS_REPORT)
        USE ROTATION_MATRICES
        USE INTERPOLATION
        USE OMP_LIB
        USE IMAGES_ARITHM
        USE IMAGES_THREED
        USE IMAGEFILES  ! debug
        USE UI, ONLY : INTERVAL_PRINT
        use UsefulFunctions, only : IsEven
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                        INTENT(INOUT)   ::  SELF                !<  Image (volume) to be symmetrised
        TYPE(HELICAL_SYMMETRY),             INTENT(IN)      ::  HELIX               !<  Helical symmetry operator
        TYPE(INTERPOLATION_SCHEME),         INTENT(IN)      ::  INTERPOL_SCHEME     !<  Interpolation scheme
        INTEGER,    OPTIONAL,               INTENT(IN)      ::  NUM_CENTRAL_SLICES  !<  The number of slices (starting from the central slice), which will be symmetrised into the volume. If not given, the full volume is symmetrised.
        REAL,       OPTIONAL,               INTENT(IN)      ::  MAX_RADIUS          !<  Maximum outer radius of helix (pixels)
        LOGICAL,    OPTIONAL,               INTENT(IN)      ::  PROGRESS_REPORT     !<  Give some output regularly to let the user know we're working
        ! Private variables
        INTEGER                 ::  Z_START, Z_END, CURRENT_Z   ! the first and last slices to include in the symmetrisation
        TYPE(IMAGE)             ::  OUTPUT_VOL, SLICE
        INTEGER                 ::  ORIGIN(3)
        INTEGER                 ::  A_SYM, P_SYM, Z_SHIFT
        REAL(KIND=8)            ::  Z_SHIFT_R1, Z_SHIFT_R2
        INTEGER                 ::  I,J,K
        REAL(KIND=8)            ::  ROT1,ROT2,ROT3, MATRIX3(4,4)
        LOGICAL                 ::  TTHREADING  ! Whether we're threading
        REAL(KIND=8)            ::  MMAX_RADIUS
        INTEGER                 ::  CURRENT_ITER,TOTAL_ITER
        INTEGER                 ::  NNUM_CENTRAL_SLICES
        INTEGER                 ::  FIRST_CENTRAL_SLICE,LAST_CENTRAL_SLICE
        REAL(KIND=8)            ::  RAD(2), MMAX_RADIUS2, AVERAGE
        INTEGER                 ::  VOXCOUNT
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        LOGICAL                 ::  PPROGRESS_REPORT
        LOGICAL,    PARAMETER   ::  DO_NOTHING = .FALSE.
        LOGICAL                 ::  GUESS_NUM_CENTRAL_SLICES
        TYPE(WINDOW),   TARGET  ::  WIN             !<  Interpolating window
        ! Start work
        ! Check the image is ready
        IF (.NOT. SELF%IsAllocated()) THEN
            WRITE(*,'(A)') '**ERROR(IMAGE_IMPOSE_HELICAL_SYMMETRY): image is not allocated'
            CALL TERMINATE('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
        ENDIF
        IF (ANY(SELF%logical_dimensions .EQ. 1)) THEN
            WRITE(*,'(A,3I4)') '**ERROR(IMAGE_IMPOSE_HELICAL_SYMMETRY): image is not 3D. Dim = ', SELF%logical_dimensions
            CALL TERMINATE('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
        ENDIF
        ! We need the helical rise parameter in pixels, not Angstroms
        IF (.NOT. HELIX%PIXEL_UNITS) THEN
            WRITE(*,'(A)') '**ERROR(IMAGE_IMPOSE_HELICAL_SYMMETRY): please supply helix rise parameter in pixels, not Angstroms.'
            CALL TERMINATE('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
        ENDIF
        ! Number of central slices
        GUESS_NUM_CENTRAL_SLICES = .FALSE.
        IF (.NOT. PRESENT(NUM_CENTRAL_SLICES)) THEN
            GUESS_NUM_CENTRAL_SLICES = .TRUE.
        ELSE
            GUESS_NUM_CENTRAL_SLICES = NUM_CENTRAL_SLICES .LE. 1
        ENDIF
        IF (PRESENT(NUM_CENTRAL_SLICES) .AND. .NOT. GUESS_NUM_CENTRAL_SLICES) THEN
            ! We will assume the user asks for an odd number of central slices
            IF (IsEven(NUM_CENTRAL_SLICES)) THEN
                WRITE(*,'(A,I0)') '**ERROR(IMAGE_IMPOSE_HELICAL_SYMMETRY): need an odd number of central slices. You asked for: ', &
                                    NUM_CENTRAL_SLICES
                CALL TERMINATE('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
            ENDIF
            ! The number of central slices should be .GE. the rise, otherwise, we don't even have a full description of the central helical repeat
            IF (NUM_CENTRAL_SLICES .LT. HELIX%RISE) THEN
                WRITE(*,'(A,I0,A,F0.3,A)') '**ERROR(IMAGE_IMPOSE_HELICAL_SYMMETRY): the number of central slices (',    &
                                             NUM_CENTRAL_SLICES, ') is less than the helical rise (', HELIX%RISE, ')'
                CALL TERMINATE('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
            ENDIF
            NNUM_CENTRAL_SLICES = NUM_CENTRAL_SLICES
        ELSE
            !NNUM_CENTRAL_SLICES = MAX(CEILING(1.1*HELIX%RISE),5)
            NNUM_CENTRAL_SLICES = CEILING(HELIX_GUESS_SUBUNIT_HEIGHT(HELIX))
            ! We cap at 1/2 of Z height of volume
            NNUM_CENTRAL_SLICES = MIN(NNUM_CENTRAL_SLICES,SELF%logical_dimensions(3)/2)
            IF (IsEven(NNUM_CENTRAL_SLICES)) NNUM_CENTRAL_SLICES = NNUM_CENTRAL_SLICES + 1 ! We make sure that NNUM_CENTRAL_SLICES is odd
        ENDIF
        WRITE(*,'(A,I0,A)') '**info(IMAGE_IMPOSE_HELICAL_SYMMETRY): imposing helical symmetry with ', &
                                                                NNUM_CENTRAL_SLICES, ' central slices'

        I = CEILING(HELIX_GUESS_SUBUNIT_HEIGHT(HELIX))
        IF (IsEven(I)) I = I+1
        IF (DEBUG) WRITE(*,'(A,I0)') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): subunit height = ', I
        IF (NNUM_CENTRAL_SLICES .LT. I) THEN
            WRITE(*,'(A,I0,A,I0)') '**info(IMAGE_IMPOSE_HELICAL_SYMMETRY): number of central slices (',NNUM_CENTRAL_SLICES   &
                                    ,') may be too small. Perhaps try ', I
        ENDIF

        ! Did the user ask for progress reports?
        IF (PRESENT(PROGRESS_REPORT)) THEN
            PPROGRESS_REPORT = PROGRESS_REPORT
        ELSE
            PPROGRESS_REPORT = .FALSE.
        ENDIF

        ! Assumption that the volume is even-dimensioned
        !IF (ANY(IS_ODD(SELF%logical_dimensions))) THEN
        !   WRITE(*,'(A,3I5)') '**ERROR(IMAGE_IMPOSE_HELICAL_SYMMETRY): volume dimensions need to be even: ', SELF%logical_dimensions
        !   CALL TERMINATE('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
        !ENDIF

        !
        ! Will we thread?
        !
        IF (PRODUCT(SELF%logical_dimensions) .GE. 64**3 .AND. SELF%logical_dimensions(3) .GT. 1) THEN
            TTHREADING = .TRUE.
        ELSE
            TTHREADING = .FALSE.
        ENDIF
        ! If we're already in a threaded region, override, since ifort does not support nested parallelism
        IF (OMP_GET_ACTIVE_LEVEL() .GT. 0) THEN
            !WRITE(*,'(A)') '**info(IMAGE_IMPOSE_HELICAL_SYMMETRY): already in a parallel scheme, so will disable local threading, since nesting not supported'
            TTHREADING = .FALSE.
        ENDIF

        !
        ! Maximum radius
        !
        IF (PRESENT(MAX_RADIUS)) THEN
            IF (DEBUG) PRINT '(A,F0.3)', '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): optional max_radius argument was present: ', &
                                        MAX_RADIUS
            IF (MAX_RADIUS .EQ. 0.0) THEN
                MMAX_RADIUS = MINVAL(SELF%logical_dimensions)*0.5 - 6.
            ELSE
                MMAX_RADIUS = MAX_RADIUS
            ENDIF
        ELSE
            MMAX_RADIUS = MINVAL(SELF%logical_dimensions)*0.5 - 6.
        ENDIF

        !
        ! Work out the index of the first and last central slices
        !
        if (self%is_in_real_space .and. self%origin_at_center) then
            origin = self%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif
        FIRST_CENTRAL_SLICE = ORIGIN(3)-(NNUM_CENTRAL_SLICES-1)/2
        LAST_CENTRAL_SLICE  = ORIGIN(3)+(NNUM_CENTRAL_SLICES-1)/2

        !
        ! Can't go beyond edges
        !
        IF (FIRST_CENTRAL_SLICE .LT. 1) THEN
            WRITE(*,'(A)') '**warning(IMAGE_IMPOSE_HELICAL_SYMMETRY): number of central slices to great. '//&
                                                                            'Probably programming error.'
            FIRST_CENTRAL_SLICE = 1
        ENDIF
        IF (LAST_CENTRAL_SLICE .GT. SELF%logical_dimensions(3)) THEN
            WRITE(*,'(A)') '**warning(IMAGE_IMPOSE_HELICAL_SYMMETRY): number of central slices to great. '//&
                                                                            'Probably programming error.'
            LAST_CENTRAL_SLICE = SELF%logical_dimensions(3)
        ENDIF

        !
        ! Work out the average value outside the cylinder in the central sections which will be used for symmetrisation.
        !
        MMAX_RADIUS2 = (MMAX_RADIUS)**2
        AVERAGE = 0.0
        VOXCOUNT = 0
        DO K=FIRST_CENTRAL_SLICE,LAST_CENTRAL_SLICE
            DO J=1,SELF%logical_dimensions(2)
                RAD(2) = REAL(J-ORIGIN(2))**2
                DO I=1,SELF%logical_dimensions(1)
                    RAD(1) = REAL(I-ORIGIN(1))**2 + RAD(2)
                    IF (RAD(1) .GT. MMAX_RADIUS2) THEN
                        AVERAGE = AVERAGE + SELF%real_values(I,J,K)
                        VOXCOUNT = VOXCOUNT + 1
                    ENDIF
                ENDDO
            ENDDO
        ENDDO
        AVERAGE = AVERAGE / REAL(VOXCOUNT)

        ! Subtract this value from the volume, so that the background goes to 0.0
        IF (DEBUG) WRITE(*,'(A,F0.6)') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): average value outside cylinder = ', AVERAGE
        SELF%real_values = SELF%real_values - AVERAGE

        !
        ! Zeroe voxels beyond the maximum radius
        !
        CALL CYL_MASK(SELF,REAL(MMAX_RADIUS)+3.,6)


        IF (DEBUG) THEN
            CALL WRITE_TO_DISK(SELF,'dbg_cyl_pre_h.mrc')
        ENDIF

        ! Allocate output image
        CALL IMAGE_ALLOCATE(OUTPUT_VOL,SELF)

        IF (DO_NOTHING) THEN
            WRITE(*,'(A)') '**warning(IMAGE_IMPOSE_HELICAL_SYMMETRY): not doing anything'
            OUTPUT_VOL%real_values = SELF%real_values
        ELSE

        ! We will have to scan the central region over the volume from bottom to top, such that every slice of the output will have
        ! "seen" NNUM_CENTRAL_SLICES additions to it. For this, the central region's center will start at Z_START and will finish at
        ! Z_END
        Z_START = 1 - (NNUM_CENTRAL_SLICES-1)/2
        Z_END = SELF%logical_dimensions(3) + (NNUM_CENTRAL_SLICES-1)/2
        TOTAL_ITER = Z_END-Z_START+1


        ! Deal with symmetry perpendicular to the helical/Z axis
        DO P_SYM=1,HELIX%PERPENDICULAR_SYMMETRY
            ! Rotation around an axis perpendicular to the H axis by (360/HELIX%PERPENDICULAR_SYMMETRY)*(P_SYM-1)
            ROT1 = (360/HELIX%PERPENDICULAR_SYMMETRY)*(P_SYM-1)

            ! Deal with symmetry around the helical/Z axis
            DO A_SYM = 1, HELIX%AXIAL_SYMMETRY
                ! Rotation around the H axis by (360/HELIX%AXIAL_SYMMETRY)*(A_SYM-1)
                ROT2 = (360/HELIX%AXIAL_SYMMETRY)*(A_SYM-1)

                ! Setup OMP threading

                !$OMP   PARALLEL    IF(TTHREADING)  &
                !$OMP               DEFAULT(NONE)   &
                !$OMP               PRIVATE(ROT3,MATRIX3,Z_SHIFT_R1,Z_SHIFT_R2,CURRENT_Z,CURRENT_ITER,WIN)  &
                !$OMP               PRIVATE(SLICE) &
                !$OMP               SHARED(TTHREADING,P_SYM,HELIX,Z_START,Z_END,SELF,INTERPOL_SCHEME,NNUM_CENTRAL_SLICES) &
                !$OMP               SHARED(OUTPUT_VOL,MMAX_RADIUS,PPROGRESS_REPORT,ROT1,ROT2,TOTAL_ITER,A_SYM,ORIGIN)

                !$OMP   SINGLE
                IF (TTHREADING .AND. OMP_GET_NUM_THREADS() .GT. 1 .AND. DEBUG) THEN
                    WRITE(*,'(A,I1,A)') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): OpenMP with ', OMP_GET_NUM_THREADS(), ' threads'
                ENDIF
                !$OMP   END SINGLE

                CALL WIN%AllocateWindow(INTERPOL_SCHEME%WINDOW_WIDTH,SELF%IsComplex(),SELF%IsAVolume())
                SLICE%logical_dimensions = SELF%logical_dimensions
                SLICE%logical_dimensions(3) = 1
                CALL IMAGE_ALLOCATE(SLICE)

                ! Move the central region from bottom to top of the volume (Z_SHIFT is the Z position within the output volume of its central slice, which is also the central slice of the original volume)
                !$OMP DO
                DO Z_SHIFT = Z_START, Z_END
                    CURRENT_ITER = Z_SHIFT - Z_START + 1
                    IF ((DEBUG .OR. PPROGRESS_REPORT) .AND. INTERVAL_PRINT(CURRENT_ITER,TOTAL_ITER)) THEN
                        !$OMP   CRITICAL        (PRINTOUT_PROGRESS_REPORT)
                        WRITE(*,'(4(A,I0))',ADVANCE='NO') '**info(IMAGE_IMPOSE_HELICAL_SYMMETRY): on ', CURRENT_ITER, &
                                                            ' of ', TOTAL_ITER, ' for axial symmetry ', A_SYM, &
                                                            ' and perpendicular symmetry ', P_SYM
                        IF (TTHREADING) THEN
                            WRITE(*,'(A,I0,A,I0)') ' on thread ', OMP_GET_THREAD_NUM()+1, ' of ', OMP_GET_NUM_THREADS()
                        ELSE
                            WRITE(*,'(A)') ' '
                        ENDIF
                        !$OMP   END CRITICAL    (PRINTOUT_PROGRESS_REPORT)
                    ENDIF
                    !PRINT '(A,I,A)', '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): thread ', OMP_GET_THREAD_NUM(), ' working...'

                    ! The distance from the center of the image to the center of the region of interest
                    Z_SHIFT_R1 = REAL(Z_SHIFT-ORIGIN(3))
                    ! The additional shift we'd need to move to the nearest repeat position
                    Z_SHIFT_R2 = (NINT(Z_SHIFT_R1 / HELIX%RISE) * HELIX%RISE) - Z_SHIFT_R1

                    ! Work out how much the image should be rotated by.
                    ! Note that the helical twist convention in frealix is such that a clockwise rotation
                    ! about the Z axis (when looking from +Z towards origin) is a positive angle, giving rise
                    ! to a left-handed helix.
                    ROT3 = (Z_SHIFT_R1 + Z_SHIFT_R2) * HELIX%TWIST/HELIX%RISE

                    ! Diagnostic
                    !PRINT *, '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): ', Z_SHIFT, ROT

                    ! Rotation around the helical axis (ROTATE_IMAGE expects the matrix that transforms from output to input, this is why REVERSE=.TRUE.)
                    MATRIX3 = EULER_TO_ROTMAT(ROT3+ROT2,ROT1,0.0D0,REVERSE=.TRUE.)

                    ! Add the rotated central region to the volume
                    DO I=1,NNUM_CENTRAL_SLICES
                        CURRENT_Z = Z_SHIFT + (I-((NNUM_CENTRAL_SLICES-1)/2+1))
                        IF (CURRENT_Z .LT. 1 .OR. CURRENT_Z .GT. SELF%logical_dimensions(3)) THEN
                            CYCLE
                        ELSE
                            MATRIX3(3,4) = -Z_SHIFT_R2 + REAL(I-((NNUM_CENTRAL_SLICES-1)/2+1))
                            ! Extract 2D slice from near the center of the input volume
                            CALL EXTRACT_SLICE(SELF,SLICE,0.0D0,0.0D0,0.0D0,INTERPOL_SCHEME,WIN,MATRIX3,RAD_MAX=MMAX_RADIUS)
                            !$OMP CRITICAL (IMG_IMP_H_SYM)
                            OUTPUT_VOL%real_values(:,:,CURRENT_Z) = OUTPUT_VOL%real_values(:,:,CURRENT_Z) + SLICE%real_values(:,:,1)
                            !$OMP END CRITICAL (IMG_IMP_H_SYM)
                        ENDIF
                    ENDDO
                ENDDO
                !$OMP   END DO
                CALL IMAGE_DEALLOCATE(SLICE)
                CALL DEALLOCATE_WINDOW(WIN)
                !$OMP   END PARALLEL
            ENDDO
        ENDDO

        ENDIF   ! End of check for DO_NOTHING


        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): about to normalise densities'
            WRITE(*,'(A,2L1)') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): SELF & OUTPUT_VOL same dim?, SELF allocated?: ', &
                                SELF%HasSameDimsAs(OUTPUT_VOL), SELF%IsAllocated()
            WRITE(*,'(A,2(I0,2X), 2L1)') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): SELF & OUTPUT_VOL size, allocated: ', &
                                    SIZE(SELF%real_values), SIZE(OUTPUT_VOL%real_values), &
                                    SELF%IsAllocated(), OUTPUT_VOL%IsAllocated()
            WRITE(*,'(2A,3(I0,2X))') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): NNUM_CENTRAL_SLICES, ',    &
                                    'HELIX%AXIAL_SYMMETRY, HELIX%PERPENDICULAR_SYMMETRY =', NNUM_CENTRAL_SLICES, &
                                    HELIX%AXIAL_SYMMETRY, HELIX%PERPENDICULAR_SYMMETRY

            CALL IMAGE_PRINT_INFO(SELF)
            CALL IMAGE_PRINT_INFO(OUTPUT_VOL)
        ENDIF
        ! Normalise the densities
        SELF%real_values = OUTPUT_VOL%real_values / REAL(NNUM_CENTRAL_SLICES* HELIX%AXIAL_SYMMETRY * HELIX%PERPENDICULAR_SYMMETRY)

        CALL IMAGE_DEALLOCATE(OUTPUT_VOL)

        ! Diagnostic
        IF (DEBUG) WRITE(*,'(A,3I4)') '**debug(IMAGE_IMPOSE_HELICAL_SYMMETRY): symmetrised image dimensions = ', &
                                        SELF%logical_dimensions
        WRITE(*,'(A)',ADVANCE='NO') '**info(IMAGE_IMPOSE_HELICAL_SYMMETRY): used following helical symmetry: '
        CALL HELICAL_SYMMETRY_PRINT_INFO(HELIX)
    END SUBROUTINE IMAGE_IMPOSE_HELICAL_SYMMETRY

    !>  \brief  Given an amplitude spectrum, look for meridional layer lines
    SUBROUTINE SCAN_FOR_MERIDONIAL_LAYER_LINE(  IMG_AMPL,IMG_PHAS,PLOT_ID,EXPECTED_FREQUENCY,EXPECTED_FREQUENCY_HALFRANGE, &
                                                TOP_SCORES,TOP_FREQUENCIES,TOP_PHASE_DIFFS)
        USE IMAGES
        USE PLOTS
        USE EULERANGLES, ONLY : ANGLE_DIFFERENCE
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(IN)      ::  IMG_AMPL                        !<  Amplitudes of image
        TYPE(IMAGE),                INTENT(IN)      ::  IMG_PHAS                        !<  Phases of image (must be in degrees)
        INTEGER,                    INTENT(IN)      ::  PLOT_ID                         !<  Index for filename of output png plot
        REAL,                       INTENT(IN)      ::  EXPECTED_FREQUENCY              !<  Frequency (1/pixel; 0.5 is Nyquist) at which layer line is expected
        REAL,                       INTENT(IN)      ::  EXPECTED_FREQUENCY_HALFRANGE    !<  Halfrange for frequency range in which layer line is expected
        REAL,   ALLOCATABLE,        INTENT(INOUT)   ::  TOP_SCORES(:)                   !<  Will return the scores (number of std dev above local background) of the layer line candidates
        REAL,   ALLOCATABLE,        INTENT(INOUT)   ::  TOP_FREQUENCIES(:)              !<  Will return the frequencies of the layer line candidates
        REAL,   ALLOCATABLE,        INTENT(INOUT)   ::  TOP_PHASE_DIFFS(:)              !<  The average phase difference between pixels on the lhs and rhs of the meridian
        ! Private variables
        INTEGER                 ::  I_LOGI,J_LOGI
        INTEGER                 ::  I_LOGI_MIN, I_LOGI_MAX, I_PHYS_MIN, I_PHYS_MAX
        INTEGER                 ::  J_LOGI_MIN, J_LOGI_MAX, J_PHYS_MIN, J_PHYS_MAX
        INTEGER                 ::  LBOUNDS(3), UBOUNDS(3)
        INTEGER                 ::  WINDOW_DIMENSIONS(2)
        INTEGER                 ::  IMG_AMPL_ORIGIN(3)
        REAL, ALLOCATABLE       ::  AVERAGE_DENSITY(:), AVERAGE_DENSITY_STD_SERIES(:)
        REAL, ALLOCATABLE       ::  AVERAGE_PHASE_DIFF(:)
        REAL(KIND=8)            ::  BACKG_DENSITY_AVE, BACKG_DENSITY_STD
        REAL, ALLOCATABLE       ::  SERIES_TO_PLOT(:,:)
        CHARACTER(LEN=64)       ::  PLOT_FILENAME
        INTEGER                 ::  NUM_IN_STATS
        REAL                    ::  J_FREQ
        INTEGER                 ::  J_PHYS
        INTEGER,    PARAMETER   ::  NUMBER_OF_LINES =   2   ! The top NUMBER_OF_LINES layer lines will be returned
        INTEGER,    PARAMETER   ::  LAYER_LINE_WIDTH  = 21   ! In pixels
        INTEGER,    PARAMETER   ::  LAYER_LINE_HEIGHT = 3
        REAL                    ::  CURRENT_SCORE
        INTEGER                 ::  NUM_TOP_SCORES_BEAT_BY_CURRENT, CURRENT_SCORE_POSITION
        INTEGER                 ::  POS
        CHARACTER(LEN=3)        ::  SIZE_TOP_SCORES
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        CALL MIN_MAX_LOGICAL_ADDRESS(IMG_AMPL,LBOUNDS,UBOUNDS)
        if (img_ampl%is_in_real_space .and. img_ampl%origin_at_center) then
            img_ampl_origin = img_ampl%physical_address_of_box_center
        else
            img_ampl_origin = [1,1,1]
        endif
        WINDOW_DIMENSIONS = [LAYER_LINE_WIDTH,LAYER_LINE_HEIGHT]
        !
        ALLOCATE(   AVERAGE_DENSITY(UBOUNDS(2)),    &
                    AVERAGE_DENSITY_STD_SERIES(UBOUNDS(2)), &
                    SERIES_TO_PLOT(UBOUNDS(2),1),   &
                    AVERAGE_PHASE_DIFF(UBOUNDS(2)))
        !
        I_LOGI_MIN = -(WINDOW_DIMENSIONS(1)-1)/2
        I_LOGI_MAX = -I_LOGI_MIN
        I_PHYS_MIN = I_LOGI_MIN + IMG_AMPL_ORIGIN(1)
        I_PHYS_MAX = I_LOGI_MAX + IMG_AMPL_ORIGIN(1)

        ! Loop along the second axis, from origin to edge
        BACKG_DENSITY_AVE = 0.0D0
        BACKG_DENSITY_STD = 0.0D0
        NUM_IN_STATS = 0
        DO J_LOGI = 0, UBOUNDS(2)
            ! The addresses over which we'll place our window
            J_LOGI_MIN = MAX(J_LOGI - (WINDOW_DIMENSIONS(2)-1)/2,0)
            J_LOGI_MAX = MIN(J_LOGI + (WINDOW_DIMENSIONS(2)-1)/2,UBOUNDS(2))
            J_PHYS_MIN = J_LOGI_MIN + IMG_AMPL_ORIGIN(2)
            J_PHYS_MAX = J_LOGI_MAX + IMG_AMPL_ORIGIN(2)
            ! Where we are in frequency
            J_FREQ = REAL(J_LOGI) / REAL(IMG_AMPL%logical_dimensions(2))
            !
            IF (J_LOGI .GE. 1) THEN
                ! Average amplitude at the current frequency
                AVERAGE_DENSITY(J_LOGI) = SUM(IMG_AMPL%real_values(I_PHYS_MIN:I_PHYS_MAX,J_PHYS_MIN:J_PHYS_MAX,1))   &
                                            / PRODUCT(WINDOW_DIMENSIONS)
                ! Average phase difference between pixels on lhs and pixels on rhs of meridian
                AVERAGE_PHASE_DIFF(J_LOGI) = 0.0
                DO I_LOGI=I_LOGI_MIN,-1 ! We loop over the lhs, and address the rhs in the 2nd arg to angle_difference
                    AVERAGE_PHASE_DIFF(J_LOGI) = AVERAGE_PHASE_DIFF(J_LOGI) + &
                        SUM(ANGLE_DIFFERENCE(DBLE(IMG_PHAS%real_values(IMG_AMPL_ORIGIN(1)+I_LOGI,J_PHYS_MIN:J_PHYS_MAX,1)),   &
                                             DBLE(IMG_PHAS%real_values(IMG_AMPL_ORIGIN(1)-I_LOGI,J_PHYS_MIN:J_PHYS_MAX,1)) )) &
                       / PRODUCT(WINDOW_DIMENSIONS)
                ENDDO
                ! If we're near where we expect the layer line
                IF (J_FREQ .GT. EXPECTED_FREQUENCY - 0.15 .AND. J_FREQ .LT. EXPECTED_FREQUENCY + 0.15) THEN
                    IF (     J_FREQ .LT. EXPECTED_FREQUENCY - EXPECTED_FREQUENCY_HALFRANGE   &
                        .OR. J_LOGI .GT. EXPECTED_FREQUENCY + EXPECTED_FREQUENCY_HALFRANGE) THEN
                        NUM_IN_STATS = NUM_IN_STATS + 1
                        BACKG_DENSITY_AVE = BACKG_DENSITY_AVE + AVERAGE_DENSITY(J_LOGI)
                        BACKG_DENSITY_STD = BACKG_DENSITY_STD + AVERAGE_DENSITY(J_LOGI)**2
                    ENDIF
                ENDIF
            ENDIF
        ENDDO

        ! Stats for the background near the layer line
        BACKG_DENSITY_AVE = BACKG_DENSITY_AVE / REAL(NUM_IN_STATS)
        BACKG_DENSITY_STD = BACKG_DENSITY_STD / REAL(NUM_IN_STATS)
        BACKG_DENSITY_STD = SQRT(BACKG_DENSITY_STD - BACKG_DENSITY_AVE**2)
        IF (DEBUG) WRITE(*,'(A,G15.5,1X,G15.5)') '**debug(): background average, std = ', BACKG_DENSITY_AVE, BACKG_DENSITY_STD

        AVERAGE_DENSITY_STD_SERIES = BACKG_DENSITY_AVE + 2.5*BACKG_DENSITY_STD

        ! Plot result
        WRITE(PLOT_FILENAME,'(A,I5.5,A)') 'ave_meri_density_',PLOT_ID,'.png'
        SERIES_TO_PLOT(:,1) = AVERAGE_DENSITY(:)
        CALL SERIES_TO_PNG(PLOT_FILENAME,SERIES_TO_PLOT,OVERLAY_YS=AVERAGE_DENSITY_STD_SERIES,Y_MIN=0.0,Y_MAX=3.5*.01)
        !write(*,'(I5,X,F10.3)') J_LOGI,AVERAGE_DENSITY

        ! Look for layer line candidates
        IF (ALLOCATED(TOP_SCORES)) DEALLOCATE(TOP_SCORES)
        IF (ALLOCATED(TOP_FREQUENCIES)) DEALLOCATE(TOP_FREQUENCIES)
        IF (ALLOCATED(TOP_PHASE_DIFFS)) DEALLOCATE(TOP_PHASE_DIFFS)
        ALLOCATE(TOP_SCORES(NUMBER_OF_LINES),TOP_FREQUENCIES(NUMBER_OF_LINES),TOP_PHASE_DIFFS(NUMBER_OF_LINES))
        TOP_SCORES = 0.0
        DO J_LOGI=1,UBOUNDS(2)
            J_PHYS = J_LOGI + IMG_AMPL_ORIGIN(2)
            ! Where we are in frequency
            J_FREQ = REAL(J_LOGI) / REAL(IMG_AMPL%logical_dimensions(2))
            ! If we are within frequency ranges where we've been told to look for layer lines
            IF (      J_FREQ .GE. EXPECTED_FREQUENCY - EXPECTED_FREQUENCY_HALFRANGE   &
                .AND. J_FREQ .LE. EXPECTED_FREQUENCY + EXPECTED_FREQUENCY_HALFRANGE) THEN
                ! The score is the number of standard deviations above the average for the local average amplitude
                CURRENT_SCORE = (AVERAGE_DENSITY(J_LOGI) - BACKG_DENSITY_AVE) / BACKG_DENSITY_STD
                NUM_TOP_SCORES_BEAT_BY_CURRENT = COUNT(CURRENT_SCORE .GT. TOP_SCORES)
                IF (NUM_TOP_SCORES_BEAT_BY_CURRENT .GE. 1) THEN
                    ! Current score makes it onto the list
                    CURRENT_SCORE_POSITION = 1 + (SIZE(TOP_SCORES) - NUM_TOP_SCORES_BEAT_BY_CURRENT)
                    ! Move all top scores and frequencies that are displaced
                    IF (CURRENT_SCORE_POSITION .LT. SIZE(TOP_SCORES)) THEN
                        DO POS=SIZE(TOP_SCORES),CURRENT_SCORE_POSITION+1,-1
                            TOP_SCORES(POS) = TOP_SCORES(POS-1)
                            TOP_FREQUENCIES(POS) = TOP_FREQUENCIES(POS-1)
                            TOP_PHASE_DIFFS(POS) = TOP_PHASE_DIFFS(POS-1)
                        ENDDO
                    ENDIF
                    ! Stick the current score and frequency in the list of hits
                    TOP_SCORES(CURRENT_SCORE_POSITION) = CURRENT_SCORE
                    TOP_FREQUENCIES(CURRENT_SCORE_POSITION) = J_FREQ
                    TOP_PHASE_DIFFS(CURRENT_SCORE_POSITION) = AVERAGE_PHASE_DIFF(J_LOGI)
                ENDIF
            ENDIF
        ENDDO

        ! Print out results
        IF (DEBUG) THEN
        WRITE(SIZE_TOP_SCORES,'(I0)') SIZE(TOP_SCORES)
        WRITE(*,'(A,'//SIZE_TOP_SCORES//'(F0.3,1X))')       '**debug(SCAN_FOR_MERIDIONIAL_LAYER_LINE): top scores        = ', &
                                                        TOP_SCORES
        WRITE(*,'(A,'//SIZE_TOP_SCORES//'(F0.3,1X))')  '**debug(SCAN_FOR_MERIDIONIAL_LAYER_LINE): top frequencies   = ', &
                                                        TOP_FREQUENCIES
        WRITE(*,'(A,'//SIZE_TOP_SCORES//'(F0.3,1X))') '**debug(SCAN_FOR_MERIDIONAL_LAYER_LINE): top average phase differences = ', &
                                                        TOP_PHASE_DIFFS
        ENDIF

    END SUBROUTINE SCAN_FOR_MERIDONIAL_LAYER_LINE

END MODULE IMAGES_HELICES
