!>  \brief  Module grouping function to go to/from 3D reconstructions and 2D images
MODULE IMAGES_THREED

    CONTAINS

    !>  \brief  Impose a missing wedge onto a 3D volume or 2D image. The wedge is
    SUBROUTINE IMPOSE_MISSING_WEDGE(IMG,PHI,THETA,PSI,WIDTH)
        USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(INOUT)   ::  IMG             !<  Volume
        REAL,           INTENT(IN)      ::  PHI, THETA, PSI !<  The wedge will be imposed after rotation by these Euler angles
        REAL,           INTENT(IN)      ::  WIDTH           !<  The "width" of the wedge, in degrees
        ! Private variables

        !

    END SUBROUTINE IMPOSE_MISSING_WEDGE

    !>  \brief  Project a 3D volume onto a 2D slice (real-space projection) (leaner version, which does not call ROTATE_IMAGE
    !!  \todo   Use euler angle object
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE VOLUME_PROJECT_LEAN(VOLUME,PROJECTION,PHI,THETA,PSI,INTERPOL_SCHEME,INTERPOL_WINDOW, &
                                    SUPER_LEAN,MAX_CYL_RAD,SHIFT_X,SHIFT_Y,Z_SHIFT,&
                                    HELICAL_SUBUNIT_ONLY,HELIX,PITCH)
        use interpolation, only  : INTERPOLATION_SCHEME
        USE IMAGES
        USE ROTATION_MATRICES
        USE IMAGES
        USE RUNTIME_PARAMETERS, ONLY : H_ASYM_UNIT_CENTER
        USE HELICES, ONLY : HELIX_POINT_IS_IN_ASYM, HELICAL_SYMMETRY, HELIX_PHI_FROM_XY
        USE OMP_LIB
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                        INTENT(IN)      ::  VOLUME          !<  Volume to be projected
        TYPE(IMAGE),                        INTENT(INOUT)   ::  PROJECTION      !<  Projection image
        REAL(KIND=8),                       INTENT(IN)      ::  PHI             !<  Phi Euler angle (degrees)
        REAL(KIND=8),                       INTENT(IN)      ::  THETA           !<  Theta Euler angle (degrees)
        REAL(KIND=8),                       INTENT(IN)      ::  PSI             !<  Psi Euler angle (degrees)
        TYPE(INTERPOLATION_SCHEME),         INTENT(IN)      ::  INTERPOL_SCHEME !<  Interpolation scheme
        TYPE(WINDOW),                       INTENT(INOUT)   ::  INTERPOL_WINDOW !<  Window for interpolation
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  SUPER_LEAN      !<  Whether we go super-lean, which basically means skipping all safety checks etc.
        REAL,                   OPTIONAL,   INTENT(IN)      ::  MAX_CYL_RAD     !<  If specified, any voxels beyond this radius will be ignored (pixels). The radius of a cylinder aligned with Z axis.
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  SHIFT_X         !<  In pixels, shift to be applied to the slice after projection
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  SHIFT_Y         !<  In pixels, shift to be applied to the slice after projection
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  Z_SHIFT         !<  In pixels, the volume will be shifted by this amount on along the Z axis before slice extraction (or this is what it'll look like)
        LOGICAL,                OPTIONAL,   INTENT(IN)      ::  HELICAL_SUBUNIT_ONLY    !<  Whether only the helical asymetric unit should be projected
        TYPE(HELICAL_SYMMETRY), OPTIONAL,   INTENT(IN)      ::  HELIX
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  PITCH
        ! Private variables
        REAL            ::  ROTATION(4,4)
        INTEGER         ::  I,J,K
        INTEGER         ::  ORIGIN(3)
        REAL(KIND=8)            ::  INPUT_COOS(4)
        REAL                    ::OUTPUT_COOS(4), RVALUE
        REAL    ::  MAX_RAD, RAD(3)
        LOGICAL ::  THREADING
        LOGICAL ::  SSUPER_LEAN
        REAL    ::  MMAX_CYL_RAD
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        REAL(KIND=8)    ::  SSHIFT_X, SSHIFT_Y
        LOGICAL ::  PROJECT_THIS_POINT
        ! Variables related to the projection of helical subunit only
        LOGICAL ::  HHELICAL_SUBUNIT_ONLY
        REAL(KIND=8)    ::  Z,H_PHI, ZZ_SHIFT
        ! Start work

        ! Helical subunit only?
        HHELICAL_SUBUNIT_ONLY = .FALSE.
        IF (PRESENT(HELICAL_SUBUNIT_ONLY)) HHELICAL_SUBUNIT_ONLY = HELICAL_SUBUNIT_ONLY
#ifndef SKIP_RUNTIME_CHECKS
        IF (HHELICAL_SUBUNIT_ONLY) THEN
            IF (.NOT. PRESENT(HELIX) .OR. .NOT. PRESENT(PITCH)) THEN
                WRITE(*,'(2A)') '**ERROR(VOLUME_PROJECT_LEAN): helical operator or pitch are missing. ', &
                                'Cannot implement helical_subunit_only option.'
                CALL TERMINATE('Fatal error in VOLUME_PROJECT_LEAN. Programming error')
            ENDIF
        ENDIF
#endif

        ! Shifs in XY?
        SSHIFT_X = 0.0D0
        SSHIFT_Y = 0.0D0
        ZZ_SHIFT = 0.0D0
        IF (PRESENT(SHIFT_X)) SSHIFT_X = SHIFT_X
        IF (PRESENT(SHIFT_Y)) SSHIFT_Y = SHIFT_Y
        IF (PRESENT(Z_SHIFT)) ZZ_SHIFT = Z_SHIFT

        !
        ! Are we going super lean?
        !
        SSUPER_LEAN = .FALSE.
        IF (PRESENT(SUPER_LEAN)) SSUPER_LEAN = SUPER_LEAN


        !
        ! Are we given a maximum radius?
        !
        MMAX_CYL_RAD = 99999999.99
        IF (PRESENT(MAX_CYL_RAD)) THEN
            MMAX_CYL_RAD = MAX_CYL_RAD**2
        ENDIF

#ifndef SKIP_RUNTIME_CHECKS
        !
        ! We only do the nice safety check if the user didn't ask for super lean
        !
        IF (.NOT. SSUPER_LEAN .OR. DEBUG) THEN

            ! Check input volume is in memory
            IF (.NOT. VOLUME%IN_MEMORY) THEN
                WRITE(*,'(A)') '**ERROR(VOLUME_PROJECT_LEAN): input volume is not in memory'
                CALL TERMINATE('Fatal error in VOLUME_PROJECT_LEAN')
            ENDIF

            ! Check input volume really is a volume
            IF ( ANY(VOLUME%logical_dimensions .EQ. 1) ) THEN
                WRITE(*,'(A,3(I4,1X))') '**ERROR(VOLUME_PROJECT_LEAN): input volume is not a volume, it has dimensions ', &
                                        VOLUME%logical_dimensions
                CALL TERMINATE('Fatal error in VOLUME_PROJECT_LEAN')
            ENDIF

            ! If slice not in memory, allocate to correct dimensions.
            IF ( .NOT. PROJECTION%IN_MEMORY) THEN
                IF ( ALL(PROJECTION%logical_dimensions .EQ. 1) ) THEN
                    PROJECTION%logical_dimensions = (/VOLUME%logical_dimensions(1),VOLUME%logical_dimensions(2),1/)
                ENDIF
                CALL IMAGE_ALLOCATE(PROJECTION)
                PROJECTION%is_in_real_space = VOLUME%is_in_real_space
                PROJECTION%ORIGIN_AT_CENTER = VOLUME%ORIGIN_AT_CENTER
            ENDIF

            ! Check dimensions match
            IF (ANY(VOLUME%logical_dimensions(1:2) .NE. PROJECTION%logical_dimensions(1:2))) THEN
                WRITE(*,'(A,3(I4,1X),A,3(I4,1X),A)') '**ERROR(VOLUME_PROJECT_LEAN): dimensions of volume (', &
                                                    VOLUME%logical_dimensions, &
                                                    ') do not match that of projection (', PROJECTION%logical_dimensions, ')'
                CALL TERMINATE('Fatal error in VOLUME_PROJECT_LEAN')
            ENDIF

            ! We assume that the origin is at the center of the volume
            IF (.NOT. VOLUME%ORIGIN_AT_CENTER) THEN
                WRITE(*,'(A)') '**ERROR(VOLUME_PROJECT_LEAN): input volume does not have origin at its center'
                CALL TERMINATE('Fatal error in VOLUME_PROJECT_LEAN')
                !IF (VOLUME%TYPE .EQ. 'REAL') CALL FWD_FT(VOLUME)
                !CALL SHIFT_REAL_SPACE_ORIGIN(VOLUME)
                !IF (VOLUME%TYPE .EQ. 'COMP') CALL BWD_FT(VOLUME)
            ENDIF


            ! We may need to go back to real space
            IF (.not. VOLUME%is_in_real_space) THEN
                WRITE(*,'(A)') '**ERROR(VOLUME_PROJECT_LEAN): input volume must be in real space'
                CALL TERMINATE('Fatal error in VOLUME_PROJECT_LEAN')
                !CALL BWD_FT(VOLUME)
            ENDIF


            ! We assume even dimensions
            IF (ANY(MOD(REAL(VOLUME%logical_dimensions),2.0) .NE. 0.0)) THEN
                WRITE(*,'(A,3(I4,1X))') '**ERROR(VOLUME_PROJECT_LEAN): input volume does not have even dimensions (',   &
                                        VOLUME%logical_dimensions, ')'
                CALL TERMINATE('Fatal error in VOLUME_PROJECT_LEAN')
            ENDIF

        ENDIF   !   End of check for .NOT. SSUPER_LEAN
#endif

        ! Check type
        IF (.NOT. projection%IsSameTypeAs(volume)) THEN
            PROJECTION%is_in_real_space = VOLUME%is_in_real_space
        ENDIF

        ! The projection will have origin at center, since we assume the input volume had origin at center
        PROJECTION%ORIGIN_AT_CENTER = .TRUE.

        !
        ! End of safety checks, start real work
        !

        ! Get the appropriate rotation matrix (ROTATE_IMAGE needs to matrix to go from output back to input, which is why we
        ! use REVERSE=.TRUE.)
        ROTATION = EULER_TO_ROTMAT(PHI,THETA,PSI,REVERSE=.TRUE.)

        ! Initialise the projection
        PROJECTION%real_values = 0.0

        ! Get the origin address for the volume
        if (volume%is_in_real_space .and. volume%origin_at_center) then
            ORIGIN = volume%physical_address_of_box_center
        else
            ORIGIN = [1,1,1]
        endif


        ! Coordinates initialised
        INPUT_COOS = 0.0
        OUTPUT_COOS = 0.0
        INPUT_COOS(4) = 1.0
        OUTPUT_COOS(4) = 1.0

        MAX_RAD = (REAL(PROJECTION%logical_dimensions(1))/2.-1.0 - REAL(INTERPOL_SCHEME%WINDOW_WIDTH)/2. - ABS(ZZ_SHIFT))**2
        RAD = 0.0

#ifdef LOW_LEVEL_THREADING
        THREADING = .TRUE.
#else
        THREADING = .FALSE.
#endif


#ifdef LOW_LEVEL_THREADING
        !$OMP   PARALLEL    IF(THREADING)   &
        !$OMP               DEFAULT(NONE)   &
        !$OMP               PRIVATE(I,J,K,OUTPUT_COOS,INPUT_COOS,RAD,RVALUE,H_PHI,Z,PROJECT_THIS_POINT)    &
        !$OMP               SHARED(PROJECTION,VOLUME,ROTATION,ORIGIN,INTERPOL_SCHEME,THREADING,MAX_RAD) &
        !$OMP               SHARED(MMAX_CYL_RAD,SSHIFT_X,SSHIFT_Y,ZZ_SHIFT,HHELICAl_SUBUNIT_ONLY,HELIX,PITCH)

        !$OMP   SINGLE
        IF (THREADING .AND. OMP_GET_NUM_THREADS() .GT. 1) THEN
            ! WRITE(*,'(A,I1,A)') '**debug(VOLUME_PROJECT_LEAN): OpenMP threading with ', OMP_GET_NUM_THREADS(), ' threads'
        ENDIF
        !$OMP   END SINGLE NOWAIT
#endif
        ! Prepare the window
        !CALL ALLOCATE_WINDOW(INTERPOL_WINDOW,INTERPOL_SCHEME%WINDOW_WIDTH,IS_COMP(VOLUME),IS_THREED(VOLUME))

        ! Loop over output pixels
#ifdef LOW_LEVEL_THREADING
        !$OMP   DO  !!SCHEDULE(STATIC,1)
#endif
        DO J=1,PROJECTION%logical_dimensions(2)
            OUTPUT_COOS(2) = REAL(J-ORIGIN(2)) - SSHIFT_Y
            RAD(2) = OUTPUT_COOS(2)**2
            DO I=1,PROJECTION%logical_dimensions(1)
                OUTPUT_COOS(1) = REAL(I-ORIGIN(1)) - SSHIFT_X
                RAD(1) = RAD(2) + OUTPUT_COOS(1)**2
                ! Check whether we are within the radius
                IF (RAD(1) .LE. MAX_RAD) THEN
                    ! Iterate over all the voxels in the ray which lands on the current pixel
                    DO K=1,VOLUME%logical_dimensions(3)
                        OUTPUT_COOS(3) = REAL(K-ORIGIN(3))
                        RAD(3) = RAD(1) + OUTPUT_COOS(3)**2
                        ! Check we are within the radius
                        IF(RAD(3) .LE. MAX_RAD) THEN
                            ! INPUT_COOS are the coordinates before Euler angle rotation is applied
                            INPUT_COOS = MATMUL(ROTATION,OUTPUT_COOS)
                            INPUT_COOS(3) = INPUT_COOS(3) - ZZ_SHIFT
                            ! Check the input voxel is within the cylinder we're interested in
                            IF ((INPUT_COOS(1))**2+(INPUT_COOS(2))**2 .LE. MMAX_CYL_RAD) THEN
                                ! Check whether the we are at a 0.0 voxel (note, this introduces slight artefacts near boundary of mask)
                                PROJECT_THIS_POINT = VOLUME%real_values( NINT(INPUT_COOS(1)+ORIGIN(1)),   &
                                                                    NINT(INPUT_COOS(2)+ORIGIN(2)),  &
                                                                    NINT(INPUT_COOS(3)+ORIGIN(3))) .NE. 0.0
                                PROJECT_THIS_POINT = .TRUE.
                                IF (PROJECT_THIS_POINT .AND. HHELICAL_SUBUNIT_ONLY) THEN
                                    H_PHI = HELIX_PHI_FROM_XY(INPUT_COOS(1),INPUT_COOS(2))
                                    Z     = REAL(INPUT_COOS(3),KIND=8)
                                    PROJECT_THIS_POINT =    PROJECT_THIS_POINT .AND. &
                                                            HELIX_POINT_IS_IN_ASYM(HELIX,Z,H_PHI)
                                ENDIF
                                IF (PROJECT_THIS_POINT) THEN
                                    ! Center the window on the input voxel
                                    INTERPOL_WINDOW%CENTERED_ON = INPUT_COOS(1:3)
                                    ! Define the bounds of the window over the output image
                                    ! These bounds can take any value, including 0, which means the address of the origin,
                                    ! as well as negative values.
                                    CALL EVAL_WINDOW_BOUNDS(INTERPOL_WINDOW)
                                    CALL WINDOW_FROM_IMAGE_REAL(VOLUME,INTERPOL_WINDOW)
                                    CALL GET_REAL_FROM_WINDOW(RVALUE,INTERPOL_WINDOW,   INTERPOL_SCHEME%INTERPOLATION_FUNCTION,&
                                                                                        INTERPOL_SCHEME%WINDOWING_FUNCTION)
                                    PROJECTION%real_values(I,J,1) = PROJECTION%real_values(I,J,1) + RVALUE
                                ENDIF
                            ENDIF
                        ENDIF
                    ENDDO
                ENDIF
            ENDDO
        ENDDO
#ifdef LOW_LEVEL_THREADING
        !$OMP   END DO

        !$OMP   END PARALLEL
#endif
    END SUBROUTINE VOLUME_PROJECT_LEAN


    !>  \brief  Extract a 2D slice from a 3D volume (equivalent to frealign's PEXTRACT, but more abstracted/generalised)
    !!  \todo   There is assumption that dimensions are even - add a check for this &/or use ORIGIN_ADDR calls
    !!  \todo   Use euler angles object
#ifdef SKIP_RUNTIME_CHECKS
    PURE &
#endif
    SUBROUTINE EXTRACT_SLICE(VOLUME,SLICE,PHI,THETA,PSI,INTERPOL_SCHEME,WIN,ROTATION,RAD_MAX,Z_SHIFT,X_SHIFT,Y_SHIFT)
        USE IMAGES, ONLY : IMAGE, WINDOW, highest_index_dim1, nearest_origin, min_max_logical_address
        use images, only : eval_window_bounds, window_from_image_real, get_real_from_window, get_comp_from_image
        USE INTERPOLATION, ONLY : INTERPOLATION_SCHEME
        USE ROTATION_MATRICES, ONLY : EULER_TO_ROTMAT
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE, DPI
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    TARGET,                 INTENT(IN)      ::  VOLUME          !<  Volume to extract slice from
        TYPE(IMAGE),                            INTENT(INOUT)   ::  SLICE           !<  Extracted slice
        REAL(KIND=8),                           INTENT(IN)      ::  PHI             !<  Phi Euler angle (degrees)
        REAL(KIND=8),                           INTENT(IN)      ::  THETA           !<  Theta Euler angle (degrees)
        REAL(KIND=8),                           INTENT(IN)      ::  PSI             !<  Psi Euler angle (degrees)
        TYPE(INTERPOLATION_SCHEME),             INTENT(IN)      ::  INTERPOL_SCHEME !<  Interpolation scheme
        TYPE(WINDOW),   TARGET,                 INTENT(INOUT)   ::  WIN             !<  Interpolating window
        REAL(KIND=8),   OPTIONAL,               INTENT(IN)      ::  ROTATION(4,4)   !<  Rotation matrix to position slice pixel into volume. If present, overrides Euler angles supplied in other arguments
        REAL(KIND=8),   OPTIONAL,               INTENT(IN)      ::  RAD_MAX         !<  In pixels. Pixels beyond this radius in the output will not be extracted.
        REAL(KIND=8),   OPTIONAL,               INTENT(IN)      ::  Z_SHIFT         !<  In pixels, the volume will be shifted by this amount on along the Z axis before slice extraction (or this is what it'll look like)
        REAL(KIND=8),   OPTIONAL,               INTENT(IN)      ::  X_SHIFT         !<  In pixels, the slice will be shifted by this amount along the first dimension (after slice extraction)
        REAL(KIND=8),   OPTIONAL,               INTENT(IN)      ::  Y_SHIFT         !<  In pixels, the slice will be shifted by this amount along the second dimension (after slice extraction)
        ! Private variables
        INTEGER ::  I,J
        INTEGER ::  ADDR_SLICE_MAX(3)       !<  Address of the the last element in the slice
        INTEGER ::  ADDR_SLICE_ORIGIN(3)    !<  Address of the origin of the slice
        !INTEGER    ::  ADDR_SLICE_MIN(2)       !<  Address of the first element in the slice
        INTEGER ::  LOGI_ADDR_IN_SLICE(2)       !<  Logical address of current element within the slice
        REAL    ::  LOGI_ADDR_IN_SLICE_R(4)
        REAL    ::  COOS_IN_VOLUME(4)       !<  Coordinates of current element in volume
        REAL    ::  RROTATION(4,4)          !<  Rotation matrix used to position pixel into volume
        INTEGER ::  VOLUME_LBOUNDS(3)        !<  Address of lower bound of volume relative to origin 0,0,0 at center
        INTEGER ::  VOLUME_UBOUNDS(3)
        INTEGER ::  VOLUME_MID_COOS(3)      !<  Address of origin
        REAL    ::  RVALUE
        COMPLEX ::  CVALUE
        REAL    ::  RRAD_MAX
        REAL(KIND=8)    ::  ZZ_SHIFT, PHASE_Z
        LOGICAL ::  ZZZ_SHIFT
        COMPLEX(KIND=8) ::  PHASE_SHIFT
        INTEGER ::  XX,YY
        LOGICAL ::  XY_SHIFT
        REAL(KIND=8)    ::  XX_SHIFT, YY_SHIFT, PHASE_X, PHASE_Y, PHASE
        REAL(KIND=8)    ::  SLICE_PHASE_FACTOR_X, SLICE_PHASE_FACTOR_Y
        REAL(KIND=8)    ::  VOLUME_PHASE_FACTOR_Z


        ! Start work
        LOGI_ADDR_IN_SLICE_R = 0.0

#ifndef SKIP_RUNTIME_CHECKS
        ! Check input volume is in memory
        IF (.NOT. VOLUME%IN_MEMORY) THEN
            WRITE(*,'(A)') '**ERROR(EXTRACT_SLICE): input volume is not in memory'
            CALL TERMINATE('Fatal error in EXTRACT_SLICE')
        ENDIF

        ! Check input volume really is a volume
        IF ( ANY(VOLUME%logical_dimensions .EQ. 1) ) THEN
            WRITE(*,'(A,3(I4,1X))') '**ERROR(EXTRACT_SLICE): input volume is not a volume, it has dimensions ', &
                                    VOLUME%logical_dimensions
            CALL TERMINATE('Fatal error in EXTRACT_SLICE')
        ENDIF

        ! Volume is assumed to be cubic
        IF (.NOT. volume%IsSquare()) THEN
            WRITE(*,'(A)') '**ERROR(EXTRACT_SLICE): non-cubic volumes are not supported'
            CALL TERMINATE('Fatal error in EXTRACT_SLICE')
        ENDIF
#endif

        ! If slice not in memory, allocate to correct dimensions.
        IF ( .NOT. SLICE%IN_MEMORY .OR. SLICE%logical_dimensions(1) .NE. VOLUME%logical_dimensions(1) .OR. &
                                        SLICE%logical_dimensions(2) .NE. VOLUME%logical_dimensions(2)) THEN
            SLICE%logical_dimensions = (/VOLUME%logical_dimensions(1),VOLUME%logical_dimensions(2),1/)
#ifndef SKIP_RUNTIME_CHECKS
            CALL SLICE%Allocate()
#endif
            SLICE%is_in_real_space = VOLUME%is_in_real_space
            SLICE%ORIGIN_AT_CENTER = VOLUME%ORIGIN_AT_CENTER
        ENDIF

        ! Check type
        IF (.NOT. slice%IsSameTypeAs(volume)) THEN
            !WRITE(*,'(5A)') '**WARNING(EXTRACT_SLICE): volume (', VOLUME%TYPE, ') and slice (', SLICE%TYPE, ') have different types'
            !WRITE(*,'(A)')  '**WARNING(EXTRACT_SLICE): changing type of slice to match that of volume'
            SLICE%is_in_real_space = VOLUME%is_in_real_space
        ENDIF
        SLICE%ORIGIN_AT_CENTER = VOLUME%ORIGIN_AT_CENTER

#ifndef SKIP_RUNTIME_CHECKS
        ! If we are in Fourier space, we probably want the origin at 1,1,1 - not center
        IF (.not. VOLUME%is_in_real_space .AND. VOLUME%ORIGIN_AT_CENTER) THEN
            WRITE(*,'(2A)') '**WARNING(EXTRACT_SLICE): working on a complex volume, ',   &
                                'yet real-space origin is in center. Are you sure?'
        ENDIF
        IF (.not. SLICE%is_in_real_space .AND. SLICE%ORIGIN_AT_CENTER) THEN
            WRITE(*,'(A)') '**WARNING(EXTRACT_SLICE): working on a complex slice, yet real-space origin is in center. Are you sure?'
        ENDIF
#endif

        ! Evaluate the address of the slice's origin and its extent
        IF (SLICE%is_in_real_space) THEN
            IF (SLICE%origin_at_center) THEN
                ADDR_SLICE_ORIGIN = SLICE%physical_address_of_box_center
            ELSE
                ADDR_SLICE_ORIGIN = [1,1,1]
            ENDIF
            ADDR_SLICE_MAX = SLICE%logical_dimensions
        ELSE
            ADDR_SLICE_ORIGIN = 1
            ! ADDR_SLICE_ORIGIN(2) is not actually determined until we know where we are, at which point we can use the NEAREST_ORIGIN routine
            ADDR_SLICE_MAX(1) = HIGHEST_INDEX_DIM1(SLICE%logical_dimensions(1))
            ADDR_SLICE_MAX(2) = SLICE%logical_dimensions(2)
        ENDIF

        ! Evaluate the bounds of the volume (with 0,0,0 origin at center)
        CALL MIN_MAX_LOGICAL_ADDRESS(VOLUME,VOLUME_LBOUNDS,VOLUME_UBOUNDS)

        ! Evaluate address of origin of volume
        if (VOLUME%is_in_real_space .and. VOLUME%origin_at_center) then
            VOLUME_MID_COOS = VOLUME%physical_address_of_box_center
        else
            VOLUME_MID_COOS = [1,1,1]
        endif

        ! Get the appropriate rotation matrix to go from slice coordinate to 3D coordinate
        IF (PRESENT(ROTATION)) THEN
            RROTATION = ROTATION
        ELSE
            RROTATION = EULER_TO_ROTMAT(PHI,THETA,PSI,REVERSE=.TRUE.)
        ENDIF

        ! This is used for phase shifts
        SLICE_PHASE_FACTOR_X = 2.0D0*DPI/REAL(SLICE%logical_dimensions(1),KIND=8)
        SLICE_PHASE_FACTOR_Y = 2.0D0*DPI/REAL(SLICE%logical_dimensions(2),KIND=8)
        VOLUME_PHASE_FACTOR_Z = 2.0D0*DPI/REAL(VOLUME%logical_dimensions(3),KIND=8)

        ! Additional Z shift?
        IF (PRESENT(Z_SHIFT)) THEN
#ifndef SKIP_RUNTIME_CHECKS
            IF (PRESENT(ROTATION)) THEN
                WRITE(*,'(A)') '**ERROR(EXTRACT_SLICE): Z_SHIFT and ROTATION options cannot be used together'
                CALL TERMINATE('Feature not implemented')
            ENDIF
            IF (SLICE%is_in_real_space) THEN
                WRITE(*,'(A)') '**ERROR(EXTRACT_SLICE): Z_SHIFT not implemented when extracting from a real volume'
                CALL TERMINATE('Feature not implemented')
            ENDIF
#endif
            ZZ_SHIFT = Z_SHIFT
            ZZ_SHIFT = ZZ_SHIFT*VOLUME_PHASE_FACTOR_Z
        ELSE
            ZZ_SHIFT = 0.0D0
        ENDIF

        ZZZ_SHIFT = ZZ_SHIFT .NE. 0.0D0

        ! X, Y shifts?
        XX_SHIFT = 0.0D0
        YY_SHIFT = 0.0D0
        IF (PRESENT(X_SHIFT)) XX_SHIFT = X_SHIFT*SLICE_PHASE_FACTOR_X
        IF (PRESENT(Y_SHIFT)) YY_SHIFT = Y_SHIFT*SLICE_PHASE_FACTOR_Y
        XY_SHIFT = XX_SHIFT .NE. 0.0D0 .OR. YY_SHIFT .NE. 0.0D0
#ifndef SKIP_RUNTIME_CHECKS
        IF (XY_SHIFT) THEN
            IF (SLICE%is_in_real_space) THEN
                WRITE(*,'(A)') '**ERROR(EXTRACT_SLICE): sorry X,Y shifts not implemented for real slice extraction'
                CALL TERMINATE('Fatal error in EXTRACT_SLICE')
            ENDIF
        ENDIF
#endif

        ! Maximum radius
        IF (PRESENT(RAD_MAX)) THEN
            RRAD_MAX = RAD_MAX**2
        ELSE
            RRAD_MAX = (REAL(SLICE%logical_dimensions(1))/2.)**2
        ENDIF



        ! Loop over pixels in the slice. We use separate loop structures for real and complex cases.
        IF (SLICE%is_in_real_space) THEN
            DO J=1,ADDR_SLICE_MAX(2)
                LOGI_ADDR_IN_SLICE(2) = J - ADDR_SLICE_ORIGIN(2)
                DO I=1,ADDR_SLICE_MAX(1)
                    LOGI_ADDR_IN_SLICE(1) = I - ADDR_SLICE_ORIGIN(1)
                    ! Are we within the max radius?
                    IF (REAL(LOGI_ADDR_IN_SLICE(1)**2 + LOGI_ADDR_IN_SLICE(2)**2) .GT. RRAD_MAX) THEN
                        SLICE%real_values(I,J,1) = 0.0
                        CYCLE
                    ENDIF
                    ! Work out where in the volume this pixel comes from
                    COOS_IN_VOLUME = MATMUL(RROTATION,(/REAL(LOGI_ADDR_IN_SLICE(1)),REAL(LOGI_ADDR_IN_SLICE(2)),0.0,1.0/))
                    ! Center the window around that point
                    WIN%CENTERED_ON = COOS_IN_VOLUME(1:3)
                    ! Work out the bounds of the window
                    CALL EVAL_WINDOW_BOUNDS(WIN)
                    ! If no part of the window falls within the volume (this shouldn't happen), the pixel is filled with 0 and we go to next
                    IF (ANY(WIN%LBOUND .GT. VOLUME_UBOUNDS) .OR. ANY(WIN%UBOUND .LT. VOLUME_LBOUNDS)) THEN
                        SLICE%real_values(I,J,1) = 0.0
                        CYCLE
                    ENDIF
                    ! Read relevant part of volume into window and interpolate the value we're after
                    CALL WINDOW_FROM_IMAGE_REAL(VOLUME,WIN)
                    CALL GET_REAL_FROM_WINDOW(RVALUE,WIN,INTERPOL_SCHEME%INTERPOLATION_FUNCTION,INTERPOL_SCHEME%WINDOWING_FUNCTION)
                    SLICE%real_values(I,J,1) = RVALUE
                ENDDO
            ENDDO
        ELSE    ! The COMP case
            DO J=1,ADDR_SLICE_MAX(2)
                LOGI_ADDR_IN_SLICE(2) = J - NEAREST_ORIGIN(J,SLICE%logical_dimensions(2))
                YY = LOGI_ADDR_IN_SLICE(2)**2
                DO I=1,ADDR_SLICE_MAX(1)
                    LOGI_ADDR_IN_SLICE(1) = I - ADDR_SLICE_ORIGIN(1)
                    XX = LOGI_ADDR_IN_SLICE(1)**2
                    !PRINT '(A,2I6)', '**debug(EXTRACT_SLICE): I,J = ', I,J
                    ! Are we within the maximum radius?
                    IF (REAL(XX+YY) .GT. RRAD_MAX) THEN
                        SLICE%complex_values(I,J,1) = (0.0,0.0)
                        CYCLE
                    ENDIF
                    ! Work out where in the volume this pixel comes from
                    LOGI_ADDR_IN_SLICE_R(1:2) = REAL(LOGI_ADDR_IN_SLICE(1:2))
                    COOS_IN_VOLUME = MATMUL(RROTATION,LOGI_ADDR_IN_SLICE_R)
                    ! Center the window around that point
                    WIN%CENTERED_ON = COOS_IN_VOLUME(1:3)
                    ! Work out the bounds of the window
                    call win%UpdateWindowBounds()
                    ! If no part of the window falls within the volume (this shouldn't happen), the pixel is filled with 0 and we go to next
                    IF (ANY(WIN%LBOUND .GT. VOLUME_UBOUNDS) .OR. ANY(WIN%UBOUND .LT. VOLUME_LBOUNDS)) THEN
                        SLICE%complex_values(I,J,1) = (0.0,0.0)
                        CYCLE
                    ENDIF
                    ! Read relevant part of volume into window and interpolate the value we're after
                    CALL GET_COMP_FROM_IMAGE(CVALUE,VOLUME,INTERPOL_SCHEME,WIN,VOLUME_LBOUNDS,VOLUME_UBOUNDS)
                    SLICE%complex_values(I,J,1) = CVALUE
                    !PRINT *, '**debug(EXTRACT_SLICE): CVALUE = ', CVALUE, SUM(WIN%CPOINTER)
                    ! If user asked for Z shift of volume, apply it to the slice (I think this is equivalent)
                    IF (ZZZ_SHIFT) THEN
                        PHASE_Z = ZZ_SHIFT*COOS_IN_VOLUME(3)
                        PHASE_SHIFT = CMPLX(COS(-PHASE_Z),SIN(-PHASE_Z),KIND=8)
                        SLICE%complex_values(I,J,1) = SLICE%complex_values(I,J,1) * PHASE_SHIFT
                    ENDIF
                    ! Apply XY shift to the slice
                    IF (XY_SHIFT) THEN
                        PHASE_X = XX_SHIFT*REAL(LOGI_ADDR_IN_SLICE(1),KIND=8)
                        PHASE_Y = YY_SHIFT*REAL(LOGI_ADDR_IN_SLICE(2),KIND=8)
                        PHASE = - PHASE_X - PHASE_Y
                        PHASE_SHIFT = CMPLX(COS(PHASE),SIN(PHASE),KIND=8)
                        SLICE%complex_values(I,J,1) = SLICE%complex_values(I,J,1) * PHASE_SHIFT
                    ENDIF
                ENDDO
            ENDDO
        ENDIF

    END SUBROUTINE EXTRACT_SLICE

    !>  \brief  Insert a 2D slice into a 3D volume.
    !!
    !!  Conceptually, the slice is first converted to an infinitely thin continuous function (a 2D plane), placed in a continuous
    !!  3D volume, which is then sampled. The sampling step is analoguous to a scanner, where each voxel integrates over its volume
    SUBROUTINE INSERT_SLICE_GET2D_PUT3D(VOLUME,SLICE,ROTATION,INTERPOL_SCHEME,OVERSAMPLING_FACTOR)
        USE IMAGES
        USE ROTATION_MATRICES
        use INTERPOLATION, only : INTERPOLATION_SCHEME
        USE EULERANGLES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                    TARGET,     INTENT(INOUT)           ::  VOLUME              !<  Volume to insert slice into
        TYPE(IMAGE),                                INTENT(IN)              ::  SLICE               !<  Input slice
        TYPE(ROTATION_MATRIX),                      INTENT(IN)              ::  ROTATION            !<  Rotation matrix to go from slice coordinates to 3D coordinates
        TYPE(INTERPOLATION_SCHEME),                 INTENT(IN)              ::  INTERPOL_SCHEME     !<  Interpolation scheme to use
        INTEGER,                        OPTIONAL,   INTENT(IN)              ::  OVERSAMPLING_FACTOR !<  Oversampling factor
        ! Private variables
        INTEGER                 ::  I,J
        INTEGER                 ::  ADDR_SLICE_MAX(3)       !<  Address of the the last element in the slice
        INTEGER                 ::  ADDR_SLICE_ORIGIN(3)    !<  Address of the origin of the slice
        !INTEGER                    ::  ADDR_SLICE_MIN(2)       !<  Address of the first element in the slice
        INTEGER                 ::  VOLUME_LBOUND(3)        !<  Address of lower bound of window relative to origin 0,0,0 at center
        INTEGER                 ::  VOLUME_UBOUND(3)
        INTEGER                 ::  SLICE_LBOUND(3), SLICE_UBOUND(3)
        INTEGER                 ::  VOLUME_MID_COOS(3)      !<  Address of origin
        !REAL                   ::  ROTATION(4,4)           !<  Rotation matrix used to position pixel into volume
        REAL                    ::  COOS_IN_SLICE(2)        !<  Coordinates in slice
        REAL                    ::  COOS_IN_VOLUME(4)
        TYPE(WINDOW), TARGET    ::  SLICE_WIN,VOLUME_WIN
        REAL                    ::  RVALUE
        COMPLEX                 ::  CVALUE
        INTEGER                 ::  OOVERSAMPLING_FACTOR

        ! Start work

        CALL TERMINATE('ERROR: not updated to take into account explicit Friedel mates - do not use until updated')

        ! Will we be oversampling?
        IF (PRESENT(OVERSAMPLING_FACTOR)) THEN
            OOVERSAMPLING_FACTOR = OVERSAMPLING_FACTOR
        ELSE
            OOVERSAMPLING_FACTOR = 1
        ENDIF

        ! Check input volume is in memory
        IF (.NOT. VOLUME%IN_MEMORY) THEN
            WRITE(*,'(A)') '**ERROR(INSERT_SLICE_GET2D_PUT3D): input volume is not in memory'
            CALL TERMINATE('Fatal error in INSERT_SLICE_GET2D_PUT3D')
        ENDIF

        ! Check input volume really is a volume
        IF ( ANY(VOLUME%logical_dimensions .EQ. 1) ) THEN
            WRITE(*,'(A,3(I4,1X))') '**ERROR(INSERT_SLICE_GET2D_PUT3D): input volume is not a volume, it has dimensions ', &
                                    VOLUME%logical_dimensions
            CALL TERMINATE('Fatal error in INSERT_SLICE_GET2D_PUT3D')
        ENDIF

        ! Check input slice is in memory
        IF (.NOT. SLICE%IN_MEMORY) THEN
            WRITE(*,'(A)') '**ERROR(INSERT_SLICE_GET2D_PUT3D): input slice is not in memory'
            CALL TERMINATE('Fatal error in INSERT_SLICE_GET2D_PUT3D')
        ENDIF


        ! Check type
        IF (.NOT. slice%IsSameTypeAs(volume)) THEN
            !WRITE(*,'(5A)') '**WARNING(INSERT_SLICE): volume (', VOLUME%TYPE, ') and slice (', SLICE%TYPE, ') have different types'
            !WRITE(*,'(A)')  '**WARNING(INSERT_SLICE): changing type of slice to match that of volume'
            !SLICE%TYPE = VOLUME%TYPE
            WRITE(*,'(A)') '**ERROR(INSERT_SLICE_GET2D_PUT3D): input slice and volume have different types'
            CALL TERMINATE('Fatal error in INSERT_SLICE_GET2D_PUT3D')
        ENDIF


        ! Evaluate the address of the slice's origin and its extent
        IF (SLICE%is_in_real_space) THEN
            IF (SLICE%origin_at_center) then
                ADDR_SLICE_ORIGIN = SLICE%physical_address_of_box_center
            else
                ADDR_SLICE_ORIGIN = [1,1,1]
            ENDIF
            ADDR_SLICE_MAX = SLICE%logical_dimensions
        ELSE
            ADDR_SLICE_ORIGIN(1) = 1
            ! ADDR_SLICE_ORIGIN(2) is not determined until we know where we are, at which point we can use the NEAREST_ORIGIN routine
            ADDR_SLICE_MAX(1) = HIGHEST_INDEX_DIM1(SLICE%logical_dimensions(1))
            ADDR_SLICE_MAX(2) = SLICE%logical_dimensions(2)
        ENDIF

        ! Evaluate the bounds of the slice and the volume (with 0,0,0 origin at center)
        CALL MIN_MAX_LOGICAL_ADDRESS(SLICE, SLICE_LBOUND, SLICE_UBOUND )
        CALL MIN_MAX_LOGICAL_ADDRESS(VOLUME,VOLUME_LBOUND,VOLUME_UBOUND)

        ! Evaluate address of origin of volume
        IF (VOLUME%is_in_real_space .and. VOLUME%origin_at_center) then
            VOLUME_MID_COOS = VOLUME%physical_address_of_box_center
        else
            VOLUME_MID_COOS = [1,1,1]
        ENDIF

        ! Prepare the windows
        CALL SLICE_WIN%AllocateWindow(INTERPOL_SCHEME%WINDOW_WIDTH,slice%IsComplex(),slice%IsAVolume(),OWN_MEMORY=.TRUE.)
        CALL VOLUME_WIN%AllocateWindow(2,VOLUME%IsComplex(),VOLUME%IsAVolume(),OWN_MEMORY=.TRUE.)

        ! Get the appropriate rotation matrix to go from slice coordinate to 3D coordinate
        !ROTATION = EULER_TO_ROTMAT(ANGLES%PHI,ANGLES%THETA,ANGLES%PSI,REVERSE=.TRUE.)

        ! Loop over the slice in 1/steps_per_pxl. We use separate loop structures for real and complex cases.
        IF (SLICE%is_in_real_space) THEN
            DO J=1,(ADDR_SLICE_MAX(2)-1)*OOVERSAMPLING_FACTOR+1
                COOS_IN_SLICE(2) = 1+REAL(J-1)/REAL(OOVERSAMPLING_FACTOR) - REAL(ADDR_SLICE_ORIGIN(2))
                DO I=1,(ADDR_SLICE_MAX(1)-1)*OOVERSAMPLING_FACTOR+1
                    COOS_IN_SLICE(1) = 1+REAL(I-1)/REAL(OOVERSAMPLING_FACTOR) - REAL(ADDR_SLICE_ORIGIN(1))
                    ! Work out where in the volume this pixel goes to
                    COOS_IN_VOLUME = MATMUL(ROTATION%RMATRIX,[COOS_IN_SLICE(1),COOS_IN_SLICE(2),0.0,1.0])
                    ! Center the window around the point in the slice
                    SLICE_WIN%CENTERED_ON(1:2) = COOS_IN_SLICE(1:2)
                    SLICE_WIN%CENTERED_ON(3) = 0.0
                    VOLUME_WIN%CENTERED_ON = COOS_IN_VOLUME(1:3)
                    ! Work out the bounds of the windows
                    CALL EVAL_WINDOW_BOUNDS(SLICE_WIN)
                    CALL EVAL_WINDOW_BOUNDS(VOLUME_WIN)
                    ! If no part of the window falls within the volume, nothing happens to that pixel
                    IF (ANY(VOLUME_WIN%LBOUND .GT. VOLUME_UBOUND) .OR. ANY(VOLUME_WIN%UBOUND .LT. VOLUME_LBOUND)) THEN
                        CYCLE
                    ENDIF
                    ! If no part of the window falls within the slice, nothing happens to that pixel    !! CHECK THIS
                    IF (ANY(SLICE_WIN%LBOUND .GT. SLICE_UBOUND) .OR. ANY(SLICE_WIN%UBOUND .LT. SLICE_LBOUND)) THEN
                        CYCLE
                    ENDIF
                    ! Get values from image into window
                    CALL WINDOW_FROM_IMAGE_REAL(SLICE,SLICE_WIN)
                    ! Interpolate value
                    CALL GET_REAL_FROM_WINDOW(RVALUE,SLICE_WIN,INTERPOL_SCHEME%INTERPOLATION_FUNCTION,  &
                                                INTERPOL_SCHEME%WINDOWING_FUNCTION)
                    ! Place it in window over volume
                    CALL PUT_REAL_INTO_WINDOW(RVALUE,VOLUME_WIN,INTERPOL_SCHEME%INTERPOLATION_FUNCTION, &
                                                INTERPOL_SCHEME%WINDOWING_FUNCTION)
                    ! Add value into the relevant voxel of the volume
                    CALL WINDOW_TO_IMAGE_REAL(VOLUME,VOLUME_WIN)
                ENDDO
            ENDDO
        ELSE    ! The COMP case
            DO J=1,(ADDR_SLICE_MAX(2)-1)*OOVERSAMPLING_FACTOR+1
                COOS_IN_SLICE(2) = 1+REAL(J-1)/REAL(OOVERSAMPLING_FACTOR) - REAL(NEAREST_ORIGIN(J,SLICE%logical_dimensions(2)))
                DO I=1,(ADDR_SLICE_MAX(1)-1)*OOVERSAMPLING_FACTOR+1
                    COOS_IN_SLICE(1) = 1+REAL(I-1)/REAL(OOVERSAMPLING_FACTOR) - REAL(ADDR_SLICE_ORIGIN(1))
                    ! Work out where in the volume this pixel goes to
                    COOS_IN_VOLUME = MATMUL(ROTATION%RMATRIX,[COOS_IN_SLICE(1),COOS_IN_SLICE(2),0.0,1.0])
                    ! Center the windows around that point
                    SLICE_WIN%CENTERED_ON(1:2) = COOS_IN_SLICE(1:2)
                    SLICE_WIN%CENTERED_ON(3) = 0.0
                    VOLUME_WIN%CENTERED_ON = COOS_IN_VOLUME(1:3)
                    ! Work out the bounds of the window
                    CALL EVAL_WINDOW_BOUNDS(SLICE_WIN)
                    CALL EVAL_WINDOW_BOUNDS(VOLUME_WIN)
                    ! If no part of the window falls within the volume, nothing happens to that pixel
                    IF (ANY(VOLUME_WIN%LBOUND .GT. VOLUME_UBOUND) .OR. ANY(VOLUME_WIN%UBOUND .LT. VOLUME_LBOUND)) THEN
                        CYCLE
                    ENDIF
                    ! If no part of the window falls within the slice, nothing happens to that pixel    !! CHECK THIS
                    IF (ANY(SLICE_WIN%LBOUND .GT. SLICE_UBOUND) .OR. ANY(SLICE_WIN%UBOUND .LT. SLICE_LBOUND)) THEN
                        CYCLE
                    ENDIF
                    ! Get values from image into window
                    CALL WINDOW_FROM_IMAGE_COMP(SLICE,SLICE_WIN)
                    ! Interpolate value
                    CALL GET_COMP_FROM_WINDOW(CVALUE,SLICE_WIN,INTERPOL_SCHEME%INTERPOLATION_FUNCTION,  &
                                                INTERPOL_SCHEME%WINDOWING_FUNCTION)
                    ! Place it in window over volume
                    CALL PUT_COMP_INTO_WINDOW(CVALUE,VOLUME_WIN,INTERPOL_SCHEME%INTERPOLATION_FUNCTION, &
                                                INTERPOL_SCHEME%WINDOWING_FUNCTION)
                    ! Add value into the relevant voxel of the volume
                    CALL WINDOW_TO_IMAGE_COMP(VOLUME,VOLUME_WIN)
                ENDDO
            ENDDO
        ENDIF

    END SUBROUTINE INSERT_SLICE_GET2D_PUT3D


    !>  \brief  Insert a 2D slice into a 3D volume (equivalent to frealign's PINSERT, but more abstracted/generalised)
    SUBROUTINE INSERT_SLICE_PUT(VOLUME,SLICE,ROTATION,INTERPOL_SCHEME,SUBTRACT,NUM_CONTRIB_ELEMS,AMPLITUDES_ONLY)
        USE IMAGES
        USE ROTATION_MATRICES
        use interpolation, only : INTERPOLATION_SCHEME
        USE EULERANGLES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                TARGET,     INTENT(INOUT)   ::  VOLUME              !<  Volume to insert slice into
        TYPE(IMAGE),                            INTENT(IN)      ::  SLICE               !<  Input slice
        TYPE(ROTATION_MATRIX),                  INTENT(IN)      ::  ROTATION            !<  Rotation matrix to go from slice coordinate to 3D coordinate
        TYPE(INTERPOLATION_SCHEME),             INTENT(IN)      ::  INTERPOL_SCHEME     !<  Interpolation scheme to use
        LOGICAL,                    OPTIONAL,   INTENT(IN)      ::  SUBTRACT            !<  If .TRUE., the slice will be subtracted from, rather than added to, the volume. .FALSE. by default.
        INTEGER,                    OPTIONAL,   INTENT(OUT)     ::  NUM_CONTRIB_ELEMS   !<  Return number of contributing elements
        LOGICAL,                    OPTIONAL,   INTENT(IN)      ::  AMPLITUDES_ONLY     !<  The phases are thrown away - only the amplitudes are added to the volume (applies only if input image is complex)
        ! Private variables
        INTEGER ::  I,J
        INTEGER ::  ADDR_SLICE_MAX(3)       !<  Address of the the last element in the slice
        INTEGER ::  ADDR_SLICE_ORIGIN(3)    !<  Address of the origin of the slice
        INTEGER ::  LOGI_ADDR_IN_SLICE(2)       !<  Address of current element within the slice
        REAL    ::  COOS_IN_VOLUME(4)       !<  Coordinates of current element in volume
        INTEGER ::  VOLUME_LBOUND(3)        !<  Address of lower bound of window relative to origin 0,0,0 at center
        INTEGER ::  VOLUME_UBOUND(3)
        INTEGER ::  VOLUME_MID_COOS(3)      !<  Address of origin
        TYPE(WINDOW), TARGET    ::  WIN
        INTEGER ::  NNUM_CONTRIB_ELEMS
        LOGICAL ::  AAMPLITUDES_ONLY
        COMPLEX ::  CVAL


        ! Start work

        AAMPLITUDES_ONLY = .FALSE.
        IF (PRESENT(AMPLITUDES_ONLY)) AAMPLITUDES_ONLY = .TRUE.

        ! Check input volume is in memory
        IF (.NOT. VOLUME%IN_MEMORY) THEN
            WRITE(*,'(A)') '**ERROR(INSERT_SLICE): input volume is not in memory'
            CALL TERMINATE('Fatal error in INSERT_SLICE')
        ENDIF

        ! Check input volume really is a volume
        IF ( ANY(VOLUME%logical_dimensions .EQ. 1) ) THEN
            WRITE(*,'(A,3(I4,1X))') '**ERROR(INSERT_SLICE): input volume is not a volume, it has dimensions ', &
                                    VOLUME%logical_dimensions
            CALL TERMINATE('Fatal error in INSERT_SLICE')
        ENDIF

        ! Check input slice is in memory
        IF (.NOT. SLICE%IN_MEMORY) THEN
            WRITE(*,'(A)') '**ERROR(INSERT_SLICE): input slice is not in memory'
            CALL TERMINATE('Fatal error in INSERT_SLICE')
        ENDIF


        ! Check type
        IF (.NOT. SLICE%IsSameTypeAs(VOLUME)) THEN
            !WRITE(*,'(5A)') '**WARNING(INSERT_SLICE): volume (', VOLUME%TYPE, ') and slice (', SLICE%TYPE, ') have different types'
            !WRITE(*,'(A)')  '**WARNING(INSERT_SLICE): changing type of slice to match that of volume'
            !SLICE%TYPE = VOLUME%TYPE
            WRITE(*,'(A)') '**ERROR(INSERT_SLICE): input slice and volume have different types'
            CALL TERMINATE('Fatal error in INSERT_SLICE')
        ENDIF

        ! Evaluate the address of the slice's origin and its extent
        IF (SLICE%is_in_real_space) THEN
            IF (SLICE%origin_at_center) then
                ADDR_SLICE_ORIGIN = SLICE%physical_address_of_box_center
            else
                ADDR_SLICE_ORIGIN = [1,1,1]
            ENDIF
            ADDR_SLICE_MAX = SLICE%logical_dimensions
        ELSE
            ADDR_SLICE_ORIGIN(1) = 1
            ! ADDR_SLICE_ORIGIN(2) is not determined until we know where we are, at which point we can use the NEAREST_ORIGIN routine
            ADDR_SLICE_MAX(1) = HIGHEST_INDEX_DIM1(SLICE%logical_dimensions(1))
            ADDR_SLICE_MAX(2) = SLICE%logical_dimensions(2)
        ENDIF

        ! Evaluate the bounds of the volume (with 0,0,0 origin at center)
        CALL MIN_MAX_LOGICAL_ADDRESS(VOLUME,VOLUME_LBOUND,VOLUME_UBOUND)

        ! Evaluate address of origin of volume
        IF (VOLUME%is_in_real_space .and. volume%origin_at_center) then
            volume_mid_coos = volume%physical_address_of_box_center
        else
            VOLUME_MID_COOS = [1,1,1]
        ENDIF

        ! Prepare the window
        CALL WIN%AllocateWindow(INTERPOL_SCHEME%WINDOW_WIDTH,volume%IsComplex(),volume%IsAVolume(),OWN_MEMORY=.TRUE.)

        ! Start counting
        NNUM_CONTRIB_ELEMS = 0

        ! Loop over pixels in the slice. We use separate loop structures for real and complex cases.
        IF (SLICE%is_in_real_space) THEN
            IF (AAMPLITUDES_ONLY) THEN
                WRITE(*,'(A)') '**ERROR(INSERT_SLICE_PUT): amplitudes-only option not implemented if the input slice is real'
                CALL TERMINATE('Fatal error in INSERT_SLICE_PUT')
            ENDIF
            DO J=1,ADDR_SLICE_MAX(2)
                LOGI_ADDR_IN_SLICE(2) = J - ADDR_SLICE_ORIGIN(2)
                DO I=1,ADDR_SLICE_MAX(1)
                    LOGI_ADDR_IN_SLICE(1) = I - ADDR_SLICE_ORIGIN(1)
                    ! Count the number of elements we're "putting"
                    NNUM_CONTRIB_ELEMS = NNUM_CONTRIB_ELEMS + 1
                    ! Work out where in the volume this pixel goes to
                    COOS_IN_VOLUME = MATMUL(ROTATION%RMATRIX,(/REAL(LOGI_ADDR_IN_SLICE(1)),REAL(LOGI_ADDR_IN_SLICE(2)),0.0,1.0/))
                    ! Center the window around that point
                    WIN%CENTERED_ON = COOS_IN_VOLUME(1:3)
                    ! Work out the bounds of the window
                    CALL EVAL_WINDOW_BOUNDS(WIN)
                    ! If no part of the window falls within the volume (this shouldn't happen), nothing happens to that pixel
                    IF (ANY(WIN%LBOUND .GT. VOLUME_UBOUND) .OR. ANY(WIN%UBOUND .LT. VOLUME_LBOUND)) THEN
                        CYCLE
                    ENDIF
                    ! Spread value into window
                    CALL PUT_REAL_INTO_WINDOW(SLICE%real_values(I,J,1),WIN,  &
                                                INTERPOL_SCHEME%INTERPOLATION_FUNCTION,INTERPOL_SCHEME%WINDOWING_FUNCTION)
                    ! Add values from window into volume
                    CALL WINDOW_TO_IMAGE_REAL(VOLUME,WIN,SUBTRACT=SUBTRACT)
                ENDDO
            ENDDO
        ELSE    ! The COMP case
            DO J=1,ADDR_SLICE_MAX(2)
                LOGI_ADDR_IN_SLICE(2) = J - NEAREST_ORIGIN(J,SLICE%logical_dimensions(2))
                DO I=1,ADDR_SLICE_MAX(1)
                    LOGI_ADDR_IN_SLICE(1) = I - ADDR_SLICE_ORIGIN(1)

                    ! At the origin of the first dimension, we have the unusual situation that Friedel mates are described
                    ! explicitely, except for the Nyquist components, which are only described in the negative frequencies.
                    ! To avoid duplication, we therefore cycle if we are on Friedel mates.
                    IF (LOGI_ADDR_IN_SLICE(1) .EQ. 0) THEN
                        IF (LOGI_ADDR_IN_SLICE(2) .GT. 0) THEN
                            CYCLE
                        ENDIF
                    ENDIF

                    NNUM_CONTRIB_ELEMS = NNUM_CONTRIB_ELEMS + 1

                    IF (SLICE%complex_values(I,J,1) .EQ. (0.0,0.0)) CYCLE

                    ! Work out where in the volume this pixel goes to
                    COOS_IN_VOLUME = MATMUL(ROTATION%RMATRIX,(/REAL(LOGI_ADDR_IN_SLICE(1)),REAL(LOGI_ADDR_IN_SLICE(2)),0.0,1.0/))
                    ! Center the window around that point
                    WIN%CENTERED_ON = COOS_IN_VOLUME(1:3)
                    ! Work out the bounds of the window
                    CALL EVAL_WINDOW_BOUNDS(WIN)
                    ! If no part of the window falls within the volume (this shouldn't happen), nothing happens to that pixel
                    IF (ANY(WIN%LBOUND .GT. VOLUME_UBOUND) .OR. ANY(WIN%UBOUND .LT. VOLUME_LBOUND)) THEN
                        CYCLE
                    ENDIF
                    ! Spread value into window
                    CVAL = SLICE%complex_values(I,J,1)
                    IF (AAMPLITUDES_ONLY) CVAL = CABS(CVAL)
                    CALL PUT_COMP_INTO_WINDOW(CVAL,WIN,INTERPOL_SCHEME%INTERPOLATION_FUNCTION,INTERPOL_SCHEME%WINDOWING_FUNCTION)
                    ! Add values from window into volume
                    CALL WINDOW_TO_IMAGE_COMP(VOLUME,WIN,SUBTRACT=SUBTRACT)
                ENDDO
            ENDDO
        ENDIF

        IF (PRESENT(NUM_CONTRIB_ELEMS)) NUM_CONTRIB_ELEMS = NNUM_CONTRIB_ELEMS

    END SUBROUTINE INSERT_SLICE_PUT

END MODULE IMAGES_THREED
