!>  \brief  A library of kernel functions used in interpolation
MODULE INTERPOLATION_KERNELS

CONTAINS

    !>  \brief  Returns the normalised sinc of ARG
    !!          If ARG = 0, SINC = 1. If ARG is integer-valued, SINC=0.
    PURE REAL FUNCTION SINC(ARG)
        USE RUNTIME_PARAMETERS, ONLY : PI
        IMPLICIT NONE
        REAL, INTENT(IN)    ::  ARG
        ! Private variables
        REAL    ::  ARG2
        REAL,   PARAMETER   ::  SMALL   =   1.0E-6
        ! Start work
        ARG2 = ABS(PI * ARG)
        IF (ARG2 .LE. SMALL) THEN
            SINC = 1.0
        ELSE IF (ABS(ARG2-AINT(ARG2)) .LE. SMALL) THEN
            SINC = 0.0
        ELSE
            SINC = SIN(ARG2)/ARG2
        ENDIF
    END FUNCTION SINC

    !>  \brief  Returns the square of the sinc of ARG
    PURE REAL FUNCTION SINC2(ARG)
        IMPLICIT NONE
        REAL,   INTENT(IN)  ::  ARG
        SINC2 = SINC(ARG)**2
    END FUNCTION SINC2

    !>  \brief  Returns the instrument function to SINC, which is a rectangle
    PURE REAL FUNCTION SINC_INSTR(ARG)
        IMPLICIT NONE
        REAL,   INTENT(IN)  ::  ARG
        SINC_INSTR = RECT(ARG,1)
    END FUNCTION SINC_INSTR

    !>  \brief  Returns the normalised linear function of ARG
    !!          If ARG = 0., LINEAR = 1. If ARG .GE. 1.0, LINEAR = 0.
    PURE REAL FUNCTION LINEAR(ARG)
        IMPLICIT NONE
        REAL, INTENT(IN)    ::  ARG
        LINEAR = TRIANGLE(ARG)
    END FUNCTION LINEAR

    !>  \brief  The instrument function to linear interpolation kernel
    PURE REAL FUNCTION LINEAR_INSTR(ARG)
        IMPLICIT NONE
        REAL, INTENT(IN)    ::  ARG
        LINEAR_INSTR = TRIANGLE_INSTR(ARG)
    END FUNCTION LINEAR_INSTR

    !>  \brief The triangle function, used in linear interpolation.
    PURE REAL FUNCTION TRIANGLE(ARG)
        IMPLICIT NONE
        REAL,   INTENT(IN)  ::  ARG
        ! Private variable
        REAL    ::  AARG
        ! Start work
        AARG = ABS(ARG)
        IF (AARG .GT. 1.0) THEN
            TRIANGLE = 0.0
        ELSE
            TRIANGLE = 1 - AARG
        ENDIF
    END FUNCTION TRIANGLE

    !>  \brief  Instrument function corresponding to the triangle function
    PURE REAL FUNCTION TRIANGLE_INSTR(ARG)
        IMPLICIT NONE
        REAL,   INTENT(IN)  ::  ARG
        !TRIANGLE_INSTR = SINC2_WITH_ALIASES(ARG)
        TRIANGLE_INSTR = SINC2(ARG)
    END FUNCTION TRIANGLE_INSTR

    !>  \brief  Returns 0.0 if ABS(ARG) .GT. 0.5, else 1.0
    !!          Warning: make sure the argument is positive!
    PURE REAL FUNCTION NEAREST(ARG)
        IMPLICIT NONE
        REAL, INTENT(IN)    ::  ARG
        NEAREST = RECT(ARG,1)
    END FUNCTION NEAREST

    PURE REAL FUNCTION NEAREST_INSTR(ARG)
        IMPLICIT NONE
        REAL,   INTENT(IN)  ::  ARG
        !NEAREST_INSTR = SINC_WITH_ALIASES(ARG) ! NEEDS TO BE DOUBLE-CHECKED
        NEAREST_INSTR = RECT_INSTR(ARG,1)
    END FUNCTION NEAREST_INSTR

    !>  \brief  Returns the Hann function of ARG (radians), given WIDTH. WIDTH is the full width
    !!          When ARG = WIDTH, HANN = 1.0 (this should never happen, since this function should not be called with ARG .GT. (WIDTH-1)/2)
    !!          When ARG = 0, HANN = 1.0
    !!          When ARG = WIDTH/2, HANN = 0
    PURE REAL FUNCTION HANN(ARG,WIDTH)
        USE RUNTIME_PARAMETERS, ONLY: PI
        IMPLICIT NONE
        REAL, INTENT(IN) :: ARG
        INTEGER, INTENT(IN) :: WIDTH !< Full width of the window (centered at ARG=0)
        ! Private variables
        REAL    ::  RWIDTH, HWIDTH
        ! Start work
        RWIDTH = REAL(WIDTH)
        HWIDTH = RWIDTH/2.0
        IF (ARG .LT. HWIDTH) THEN
            HANN = 0.5 * ( 1 + COS(PI*ARG/HWIDTH))
        ELSE
            HANN = 0.0
        ENDIF
    END FUNCTION

    !>  \brief  The instrument function corresponding to the Hann function
    !!  See http://mathworld.wolfram.com/ApodizationFunction.html and
    !!  http://mathworld.wolfram.com/HanningFunction.html
    !!  for more details
    PURE REAL FUNCTION HANN_INSTR(ARG,WIDTH)
        USE RUNTIME_PARAMETERS, ONLY: PI
        IMPLICIT NONE
        REAL,       INTENT(IN)  ::  ARG
        INTEGER,    INTENT(IN)  ::  WIDTH
        ! Private variable
        REAL :: HWIDTH, RWIDTH, TMP
        ! Start work
        RWIDTH = REAL(WIDTH)
        HWIDTH = RWIDTH / 2.0
        !TMP = 2*PI*ARG*HWIDTH
        !TMP = 2*ARG*RWIDTH
        TMP = ARG*RWIDTH
        !HANN_INSTR = HWIDTH*SINC(PI*RWIDTH*ARG)/(1-RWIDTH**2*ARG**2)   ! This is the quicker expression to evaluate, but it can lead to div/0.0 with ARG = 0.5
        !HANN_INSTR = HWIDTH*(SINC(TMP)+.5*SINC(TMP-PI)+.5*SINC(TMP+PI))
        !HANN_INSTR = SINC(TMP)+.5*SINC(TMP-PI)+.5*SINC(TMP+PI)
        HANN_INSTR = SINC(TMP)+.5*SINC(TMP-1.)+.5*SINC(TMP+1.)
        !PRINT '(A,F8.3,X,I2,X,F10.3)', '**debug(HANN_INSTR): ARG, WIDTH, VALUE = ', ARG, WIDTH, HANN_INSTR
    END FUNCTION HANN_INSTR

    !>  \brief  Top-hat / rectangle function is 1.0 if arg .le. width/2.0
    PURE REAL FUNCTION RECT(ARG,WIDTH)
        IMPLICIT NONE
        REAL,       INTENT(IN) :: ARG
        INTEGER,    INTENT(IN) :: WIDTH
        IF (ABS(ARG) .GT. REAL(WIDTH)/2.) THEN
            RECT = 0.0
        ELSE
            RECT = 1.0
        ENDIF
    END FUNCTION RECT

    !>  \brief  The instrument function corresponding to the rectangle window function
    PURE REAL FUNCTION RECT_INSTR(ARG,WIDTH)
        IMPLICIT NONE
        REAL,       INTENT(IN)  ::  ARG
        INTEGER,    INTENT(IN)  ::  WIDTH
        !RECT_INSTR = SINC_WITH_ALIASES(ARG,REAL(WIDTH))
        RECT_INSTR = SINC(ARG*REAL(WIDTH))
    END FUNCTION RECT_INSTR

    PURE REAL FUNCTION UNITY_WINDOW(ARG,WIDTH)
        IMPLICIT NONE
        REAL,       INTENT(IN)  ::  ARG
        INTEGER,    INTENT(IN)  ::  WIDTH
        UNITY_WINDOW = 1.0
    END FUNCTION UNITY_WINDOW

    PURE REAL FUNCTION UNITY_WINDOW_INSTR(ARG,WIDTH)
        IMPLICIT NONE
        REAL,       INTENT(IN)  ::  ARG
        INTEGER,    INTENT(IN)  ::  WIDTH
        UNITY_WINDOW_INSTR = DIRAC(ARG)
    END FUNCTION UNITY_WINDOW_INSTR

    !>  \brief  Always returns 1.0
    PURE REAL FUNCTION UNITY(ARG)
        IMPLICIT NONE
        REAL, INTENT(IN) :: ARG
        UNITY = 1.0
    END FUNCTION UNITY

    !>  \brief  The instrument function corresponding to the unity function
    PURE REAL FUNCTION UNITY_INSTR(ARG)
        IMPLICIT NONE
        REAL,   INTENT(IN)  ::  ARG
        UNITY_INSTR = DIRAC(ARG)
    END FUNCTION UNITY_INSTR

    !>  \brief  The Dirac delta / impulse 'improper function'
    PURE REAL FUNCTION DIRAC(ARG)
        IMPLICIT NONE
        REAL,   INTENT(IN)  ::  ARG
        IF (ARG .EQ. 0.0) THEN
            DIRAC = 1.0
        ELSE
            DIRAC = 0.0
        ENDIF
    END FUNCTION DIRAC


END MODULE INTERPOLATION_KERNELS
