MODULE LINEAR_ALGEBRA
    IMPLICIT NONE

    CONTAINS

    !>  \brief  Solves a set of linear equations by calling LU decomposition routine from GSL
    !!
    !!  (A)*(X) = (B)
    !!
    !!  (A) is a matrix of doubles (known, input), (X) is the vector of unknowns, (B) is right-hand-side vector (known)
    !!
    !!  Internally, these are "converted" to fgsl data types
    !!
    SUBROUTINE LINALG_SOLVE_LU_FGSL(AF,XF,BF,NF)
        USE FGSL
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   ALLOCATABLE,    INTENT(IN)      ::  AF(:,:) !<  Matrix (N*N)
        REAL(KIND=8),   ALLOCATABLE,    INTENT(INOUT)   ::  XF(:)   !<  Unknown vector (output) (N)
        REAL(KIND=8),   ALLOCATABLE,    INTENT(IN)      ::  BF(:)   !<  Right-hand-side of the equation (N)
        INTEGER,                        INTENT(IN)      ::  NF      !<  Dimension of matrix & vectors
        ! Private variables
        TYPE(FGSL_MATRIX)       ::  A
        TYPE(FGSL_VECTOR)       ::  B,X
        TYPE(FGSL_PERMUTATION)  ::  P
        INTEGER(FGSL_INT)       ::  STATUS, SIGNUM
        INTEGER(FGSL_SIZE_T)    ::  N
        ! Start work
        N = NF
        ! Initialise arrays and matrix
        A = FGSL_MATRIX_INIT(TYPE=1.0_fgsl_double)
        B = FGSL_VECTOR_INIT(TYPE=1.0_fgsl_double)
        X = FGSL_VECTOR_INIT(TYPE=1.0_fgsl_double)
        P = FGSL_PERMUTATION_ALLOC(N)
        ! Convert fortran vectors / matrices to FGSL types
        STATUS = FGSL_MATRIX_ALIGN(AF,N,N,N,A)
        STATUS = FGSL_VECTOR_ALIGN(BF,N,B,N,0_fgsl_size_t,1_fgsl_size_t)
        STATUS = FGSL_VECTOR_ALIGN(XF,N,X,N,0_fgsl_size_t,1_fgsl_size_t)
        ! Decompose & solve
        STATUS = FGSL_LINALG_LU_DECOMP(A,P,SIGNUM)
        STATUS = FGSL_LINALG_LU_SOLVE(A,P,B,X)
        ! Copy results over to Fortran double !! NOT SURE IF THIS CORRECT WAY TO DO IT
        XF = X
        ! Free up FGSL objects
        CALL FGSL_MATRIX_FREE(A)
        CALL FGSL_VECTOR_FREE(B)
        CALL FGSL_VECTOR_FREE(X)
        CALL FGSL_PERMUTATION_FREE(P)
    END SUBROUTINE

    !>  \brief  Solves a set of linear equations by calling LU decomposition routine from LAPACK
    SUBROUTINE LINALG_SOLVE_LAPACK(A,X,B)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   ALLOCATABLE,    INTENT(INOUT)   ::  A(:,:) !<  Matrix (N*N)
        REAL(KIND=8),   ALLOCATABLE,    INTENT(INOUT)   ::  X(:)   !<  Unknown vector (output) (N)
        REAL(KIND=8),   ALLOCATABLE,    INTENT(IN)      ::  B(:)   !<  Right-hand-side of the equation (N)
        ! Private variables
        INTEGER ::  N, LDA, LDB,NRHS
        INTEGER, ALLOCATABLE :: IPIV(:)
        INTEGER ::  INFO
        ! Start work

        ! The lapack routine will overwrite the 'B' argument with the solution X
        X = B
        ! The other arguments to the LAPACK routine
        N = SIZE(A,2)
        NRHS=1
        LDA = SIZE(A,1)
        LDB = SIZE(B,1)
        ALLOCATE(IPIV(N))
        !
        CALL DGESV(N,NRHS,A,LDA,IPIV,X,LDB,INFO)
        ! Check return status
        IF (INFO .GT. 0) THEN
            WRITE(*,'(2A,I0)') '**ERROR(LINALG_SOLVE_LAPACK): factorisation was completed', &
                            ' but the factor U is exactly singular so the solution could not be computed, I = ', INFO
            CALL TERMINATE('Fatal error in LINALG_SOLVE_LAPACK')
        ELSE IF (INFO .LT. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(LINALG_SOLVE_LAPACK): this argument had an illegal value: ', INFO
            CALL TERMINATE('Fatal error in LINALG_SOLVE_LAPACK')
        ENDIF
    END SUBROUTINE LINALG_SOLVE_LAPACK

    REAL FUNCTION MATRIX_DET(M)
        USE FGSL
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),   ALLOCATABLE,    INTENT(IN)      ::  M(:,:)
        ! Private variables
        TYPE(FGSL_MATRIX)   ::  MM
        INTEGER(FGSL_INT)       ::  STATUS, SIGNUM
        INTEGER(FGSL_SIZE_T)    ::  N
        ! Start work
        MM = FGSL_MATRIX_INIT(TYPE=1.0_FGSL_DOUBLE)
        N = SIZE(M,1)
        STATUS = FGSL_MATRIX_ALIGN(M,N,N,N,MM)
        MATRIX_DET = FGSL_LINALG_LU_DET(MM,SIGNUM)
    END FUNCTION MATRIX_DET
END MODULE LINEAR_ALGEBRA

