!>  \brief  Module with masking methods
!!  \todo   Improve naming convention for subroutines, and also consistency of arguments.
MODULE MASKS

    CONTAINS


    !>  \brief  Convenience function to generate a Fourier mask suited to diffraction from projected amyloid fibrils
    SUBROUTINE FOURIER_MASK_INIT_AMYLOID(SELF,MOULD,PIXEL_SIZE,FIL_MAX_RAD,MAX_SPACING_ALIGN,ANGLES)
        USE IMAGES
        USE EULERANGLES, ONLY : EULER
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(INOUT)   ::  SELF                !<  The Fourier mask
        TYPE(IMAGE),                INTENT(IN)      ::  MOULD               !<  The mask will be compatible with images of this dimension
        REAL,                       INTENT(IN)      ::  PIXEL_SIZE          !<  Pixel size in Angstroms
        REAL,                       INTENT(IN)      ::  FIL_MAX_RAD         !<  Maximum radius of the filament (Angstroms)
        REAL,                       INTENT(IN)      ::  MAX_SPACING_ALIGN   !<  The maximum spacing (Angstroms) used during alignments
        TYPE(EULER),    OPTIONAL,   INTENT(IN)      ::  ANGLES              !<  Euler angles describing the orientation of the filament
        ! Private variables
        REAL,       PARAMETER   ::  LAYER_LINE_SPACING              =   4.8             ! Angstroms
        REAL,       PARAMETER   ::  LAYER_LINE_SPACING_HALF_RANGE   =   0.2             ! Angstroms
        !REAL,       PARAMETER   ::  HP_SPACING                      =   200.0           ! Angstroms - this is for Ab(40), which have diameter of ~180A
        !REAL,       PARAMETER   ::  LP_SPACING                      =   8.0             ! Angstroms
        LOGICAL,    PARAMETER   ::  DEBUG                           =   .FALSE.
        TYPE(IMAGE)             ::  SELF_COPY   ! used when debugging
        ! Start work

        ! Initialise a Fourier mask
        CALL FOURIER_MASK_INIT(SELF,MOULD)

        ! Band-pass the low frequencies
        !CALL FOURIER_MASK_ADD_HOLLOW_SPHERE(    SELF,FREQ_INNER=PIXEL_SIZE/HP_SPACING,FREQ_OUTER=PIXEL_SIZE/LP_SPACING)
        CALL FOURIER_MASK_ADD_HOLLOW_SPHEROID(  SELF,   &
                                                FREQ_OUTER_MAIN     = PIXEL_SIZE / 15.0,  &
                                                FREQ_INNER_PERP     = PIXEL_SIZE / (FIL_MAX_RAD*1.2),    &
                                                FREQ_OUTER_PERP     = PIXEL_SIZE / MAX_SPACING_ALIGN,   &
                                                MAIN_AXIS_ANGLES    = ANGLES )

        ! Band-pass the layer line
        CALL FOURIER_MASK_ADD_HOLLOW_CYLINDER(  SELF,   &
                                                FREQ_INNER = 0.0,   &
                                                FREQ_OUTER = PIXEL_SIZE/25.0, &
                                                FREQ_BASE  = PIXEL_SIZE/(LAYER_LINE_SPACING+LAYER_LINE_SPACING_HALF_RANGE), &
                                                FREQ_TOP   = PIXEL_SIZE/(LAYER_LINE_SPACING-LAYER_LINE_SPACING_HALF_RANGE), &
                                                ANGLES     = ANGLES )

        ! Dump to disk a version of the Fourier mask easy to look at
        IF (DEBUG) THEN
            CALL WRITE_TO_DISK(SELF,'dbg_fourier_mask_amyloid.mrc')
            SELF_COPY = SELF
            CALL AMPL_FROM_HALF_TO_FULL(SELF_COPY)
            CALL WRITE_TO_DISK(SELF_COPY,'dbg_fourier_mask_amyloid_full.mrc')
        ENDIF
    END SUBROUTINE FOURIER_MASK_INIT_AMYLOID

    !>  \brief  Intialise a real-space image with dimensions suitable for multiplication with Fourier-transformed images so that
    !!          it can act as a mask in Fourier space
    SUBROUTINE FOURIER_MASK_INIT(SELF,MOULD)
        USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(INOUT)   ::  SELF            !<  The Fourier mask to be initialised
        TYPE(IMAGE),                INTENT(IN)      ::  MOULD           !<  The mask will be suitable for images with the mould's dimensions
        ! Private variables

        ! Start work

        ! Deallocate if necessary
        IF (self%IsAllocated()) CALL IMAGE_DEALLOCATE(SELF)

        ! Work out logical dimensions
        SELF%logical_dimensions(1)   = HIGHEST_INDEX_DIM1(MOULD%logical_dimensions(1))
        SELF%logical_dimensions(2:3) = MOULD%logical_dimensions(2:3)

        ! Mask will be real
        SELF%is_in_real_space = .true.

        ! Allocate memory (this also sets data arrays to 0)
        CALL IMAGE_ALLOCATE(SELF)

    END SUBROUTINE FOURIER_MASK_INIT

    !>  \brief  Add an ellipsoid
    !!
    SUBROUTINE FOURIER_MASK_ADD_HOLLOW_SPHEROID(SELF,FREQ_OUTER_MAIN,FREQ_INNER_PERP,FREQ_OUTER_PERP,   &
                                                MAIN_AXIS_ANGLES)
        USE IMAGES
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE EULERANGLES, ONLY : EULER
        USE ROTATION_MATRICES, ONLY : EULER_TO_ROTMAT, PRINT_ROTMAT
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(INOUT)   ::  SELF                !<  Fourier mask to be modified
        REAL,                       INTENT(IN)      ::  FREQ_OUTER_MAIN     !<  Spatial frequency (1/pixel; Nyquist is 0.5) of outer radius along the main axis
        REAL,                       INTENT(IN)      ::  FREQ_INNER_PERP     !<  Spatial frequency of inner radius along the two normal axes
        REAL,                       INTENT(IN)      ::  FREQ_OUTER_PERP     !<  Spatial frequency of outer radius along the two normal axes
        TYPE(EULER),    OPTIONAL,   INTENT(IN)      ::  MAIN_AXIS_ANGLES    !<  Euler angles defining the main spheroid axis
        ! Private variables
        INTEGER ::  I,J,K
        REAL    ::  X,Y,Z
        REAL    ::  SPH_SF
        LOGICAL ::  IN_SPHEROID
        TYPE(EULER) :: AANGLES
        REAL    ::  COOS(4)
        REAL    ::  ROT(4,4)
        ! Start work

        ! The Fourier mask should be a real image
        IF (.NOT. self%IsReal()) THEN
            WRITE(*,'(A)') '**ERROR(FOURIER_MASK_ADD_HOLLOW_SPHEROID): Fourier mask images should be real'
            CALL TERMINATE('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHEROID')
        ENDIF

        ! Check frequencies make sense
        IF (FREQ_OUTER_PERP .LT. FREQ_INNER_PERP) THEN
            WRITE(*,'(A)') '**ERROR(FOURIER_MASK_ADD_HOLLOW_SPHEROID): top frequency must be greater than base frequency'
            CALL TERMINATE('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHEROID')
        ENDIF

        ! The axis
        AANGLES%PHI   = 0.0
        AANGLES%THETA = 90.0
        AANGLES%PSI   = 90.0
        IF (PRESENT(MAIN_AXIS_ANGLES)) AANGLES = MAIN_AXIS_ANGLES
        ROT = EULER_TO_ROTMAT(AANGLES%PHI,AANGLES%THETA,AANGLES%PSI,REVERSE=.TRUE.)

        ! Loop over voxels
        DO K=1,SELF%logical_dimensions(3)
            Z = REAL(K-NEAREST_ORIGIN(K,SELF%logical_dimensions(3)))/REAL(SELF%logical_dimensions(3))
            DO J=1,SELF%logical_dimensions(2)
                Y = REAL(J-NEAREST_ORIGIN(J,SELF%logical_dimensions(2)))/REAL(SELF%logical_dimensions(2))
                DO I=1,SELF%logical_dimensions(1)
                    X = REAL(I-1)/REAL(SELF%logical_dimensions(1)-1)*0.5
                    !
                    COOS = MATMUL(ROT,[X,Y,0.0,1.0])
                    !
                    SPH_SF = SQRT(   COOS(1)**2  &
                                   + COOS(2)**2  &
                                   + COOS(3)**2 * (FREQ_OUTER_PERP/FREQ_OUTER_MAIN)**2)
                    !

                    IN_SPHEROID = SPH_SF .GE. FREQ_INNER_PERP .AND. SPH_SF .LE. FREQ_OUTER_PERP
                    !
                    IF (IN_SPHEROID) SELF%real_values(I,J,K) = 1.0
                ENDDO
            ENDDO
        ENDDO
    END SUBROUTINE FOURIER_MASK_ADD_HOLLOW_SPHEROID

    !>  \brief  Add a hollow sphere (ring in the 2D case) centered on the origin to a Fourier filter
    !!          (equivalent to a hard-edge band-pass filter)
    SUBROUTINE FOURIER_MASK_ADD_HOLLOW_SPHERE(SELF,FREQ_INNER,FREQ_OUTER)
        USE IMAGES
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(INOUT)   ::  SELF        !<  Fourier mask to be modified
        REAL,                       INTENT(IN)      ::  FREQ_INNER  !<  Spatial frequency (1/pixel; Nyquist is 0.5) of inner radius
        REAL,                       INTENT(IN)      ::  FREQ_OUTER  !<  Spatial frequency (1/pixel; Nyquist is 0.5) of outer radius
        ! Private variables
        INTEGER ::  I,J,K
        REAL    ::  X,Y,Z
        REAL    ::  CURRENT_SF
        ! Start work

        ! The Fourier mask should be a real image
        IF (.NOT. self%IsReal()) THEN
            WRITE(*,'(A)') '**ERROR(FOURIER_MASK_ADD_HOLLOW_SPHERE): Fourier mask images should be real'
            CALL TERMINATE('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHERE')
        ENDIF

        ! Check frequencies make sense
        IF (FREQ_OUTER .LT. FREQ_INNER) THEN
            WRITE(*,'(A)') '**ERROR(FOURIER_MASK_ADD_HOLLOW_SPHERE): outer frequency must be greater than inner frequency'
            CALL TERMINATE('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHERE')
        ENDIF

        ! Loop over the image voxels
        DO K=1,SELF%logical_dimensions(3)
            Z = (REAL(K-NEAREST_ORIGIN(K,SELF%logical_dimensions(3)))/REAL(SELF%logical_dimensions(3)))**2
            DO J=1,SELF%logical_dimensions(2)
                Y = (REAL(J-NEAREST_ORIGIN(J,SELF%logical_dimensions(2)))/REAL(SELF%logical_dimensions(2)))**2
                DO I=1,SELF%logical_dimensions(1)
                    X = (REAL(I-1)/REAL(SELF%logical_dimensions(1)-1)*0.5)**2
                    ! The current spatial frequency
                    CURRENT_SF = SQRT(X+Y+Z)
                    !
                    IF (CURRENT_SF .GE. FREQ_INNER .AND. CURRENT_SF .LE. FREQ_OUTER) SELF%real_values(I,J,K) = 1.0

                ENDDO
            ENDDO
        ENDDO
    END SUBROUTINE FOURIER_MASK_ADD_HOLLOW_SPHERE

    !>  \brief  Add a hollow cylinder to a Fourier mask. The cylinder axis is defined by the Euler angles.
    SUBROUTINE FOURIER_MASK_ADD_HOLLOW_CYLINDER(SELF,FREQ_INNER,FREQ_OUTER,FREQ_BASE,FREQ_TOP,ANGLES)
        use Images
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE EULERANGLES, ONLY : EULER
        USE ROTATION_MATRICES, ONLY : EULER_TO_ROTMAT, PRINT_ROTMAT
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(INOUT)   ::  SELF        !<  Fourier mask to be modified
        REAL,                       INTENT(IN)      ::  FREQ_INNER  !<  Spatial frequency (1/pixel; Nyquist is 0.5) of inner radius
        REAL,                       INTENT(IN)      ::  FREQ_OUTER  !<  Spatial frequency (1/pixel; Nyquist is 0.5) of outer radius
        REAL,                       INTENT(IN)      ::  FREQ_BASE   !<  Spatial frequency (along the Y axis) for the base of the cylinder
        REAL,                       INTENT(IN)      ::  FREQ_TOP    !<  Spatial frequency (along the Y axis) for the top of the cylinder
        TYPE(EULER),    OPTIONAL,   INTENT(IN)      ::  ANGLES      !<  Euler angles defining the cylinder axis
        ! Private variables
        INTEGER ::  I,J,K
        REAL    ::  X,Y,Z
        REAL    ::  CYL_SF
        LOGICAL ::  IN_CYLINDER
        TYPE(EULER) :: AANGLES
        REAL    ::  COOS(4)
        REAL    ::  ROT(4,4)
        ! Start work

        ! The Fourier mask should be a real image
        IF (.NOT. self%IsReal()) THEN
            WRITE(*,'(A)') '**ERROR(FOURIER_MASK_ADD_HOLLOW_SPHERE): Fourier mask images should be real'
            CALL TERMINATE('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHERE')
        ENDIF

        ! Check frequencies make sense
        IF (FREQ_OUTER .LT. FREQ_INNER) THEN
            WRITE(*,'(A)') '**ERROR(FOURIER_MASK_ADD_HOLLOW_CYLINDER): outer frequency must be greater than inner frequency'
            CALL TERMINATE('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHERE')
        ENDIF
        IF (FREQ_TOP .LT. FREQ_BASE) THEN
            WRITE(*,'(A)') '**ERROR(FOURIER_MASK_ADD_HOLLOW_CYLINDER): top frequency must be greater than base frequency'
            CALL TERMINATE('Fatal error in FOURIER_MASK_ADD_HOLLOW_CYLINDER')
        ENDIF

        ! The axis
        AANGLES%PHI   = 0.0
        AANGLES%THETA = 90.0
        AANGLES%PSI   = 90.0
        IF (PRESENT(ANGLES)) AANGLES = ANGLES
        ROT = EULER_TO_ROTMAT(AANGLES%PHI,AANGLES%THETA,AANGLES%PSI,REVERSE=.TRUE.)

        ! Loop over voxels
        DO K=1,SELF%logical_dimensions(3)
            Z = REAL(K-NEAREST_ORIGIN(K,SELF%logical_dimensions(3)))/REAL(SELF%logical_dimensions(3))
            DO J=1,SELF%logical_dimensions(2)
                Y = REAL(J-NEAREST_ORIGIN(J,SELF%logical_dimensions(2)))/REAL(SELF%logical_dimensions(2))
                DO I=1,SELF%logical_dimensions(1)
                    X = REAL(I-1)/REAL(SELF%logical_dimensions(1)-1)*0.5
                    !
                    COOS = MATMUL(ROT,[X,Y,0.0,1.0])
                    !
                    CYL_SF = SQRT(COOS(1)**2+COOS(2)**2)
                    !
                    IN_CYLINDER =           ABS(COOS(3)) .GE. FREQ_BASE  &
                                    .AND.   ABS(COOS(3)) .LE. FREQ_TOP   &
                                    .AND.   CYL_SF       .GE. FREQ_INNER &
                                    .AND.   CYL_SF       .LE. FREQ_OUTER
                    !
                    IF (IN_CYLINDER) SELF%real_values(I,J,K) = 1.0
                ENDDO
            ENDDO
        ENDDO
    END SUBROUTINE FOURIER_MASK_ADD_HOLLOW_CYLINDER

    !>  \brief  Mask the first and last few slices in a volume. This is useful for example to remove the discontinuity in a helical reconstruction
    !!          where the Z dimension isn't an integer number of twists.
    !!  \todo   Check conversion of radius as fraction of maximum inner radius to number of pixels
    !!  \todo   Why isn't this done in place???
    SUBROUTINE  IMAGE_MASK_Z_ENDS(IMAGE_IN,MASK,RAD,FALLOFF,SOFTEN)
        USE IMAGES
        use UsefulFunctions, only : IsEven
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(IN)      ::  IMAGE_IN        !<  Input image
        TYPE(IMAGE),                INTENT(OUT)     ::  MASK            !<  Output mask
        REAL,                       INTENT(IN)      ::  RAD             !<  Radius (from center of volume) for the masking. Pixels, or fraction of inner radius.
        REAL,                       INTENT(IN)      ::  FALLOFF         !<  Falloff width for soft edge of mask.
        LOGICAL,    OPTIONAL,       INTENT(IN)      ::  SOFTEN          !<  Whether to "soften" (~low-pass filter) the mask. True by default.
        ! Private variables
        REAL    ::  RRAD, FFALLOFF
        INTEGER ::  I, DIST, ORIGIN
        LOGICAL ::  SSOFTEN
        ! Start work

        ! This is really only meant for 3D images
        IF (IMAGE_IN%logical_dimensions(3) .LE. 1) THEN
            WRITE(*,'(A,3(I4,1X))') '**ERROR(IMAGE_MASK_Z_ENDS): this is only meant for 3D images. Input dimensions = ', &
                                     IMAGE_IN%logical_dimensions
            CALL TERMINATE('Fatal error in IMAGE_MASK_Z_ENDS')
        ENDIF

        ! Will we be softening the mask?
        IF (PRESENT(SOFTEN)) THEN
            SSOFTEN = SOFTEN
        ELSE
            SSOFTEN = .TRUE.
        ENDIF

        ! Allocate the mask
        CALL IMAGE_ALLOCATE(MASK,IMAGE_IN)

        ! Initialise the mask
        MASK%real_values = 1.0

        ! Convert the radius from fraction to pixels if necessary
        IF (RAD .LE. 1.0) THEN
            RRAD = RAD * (MASK%logical_dimensions(3)/2-1)  ! Is this exactly the right conversion from radius as a fraction of internal radius?
        ELSE
            RRAD = RAD
        ENDIF

        ! Convert the falloff from fraction to pixels if necessary
        IF (FALLOFF .LE. 1.0) THEN
            FFALLOFF = FALLOFF * (MASK%logical_dimensions(3)/2-1)
        ELSE
            FFALLOFF = FALLOFF
        ENDIF

        ! Work out origin
        IF (IsEven(MASK%logical_dimensions(3))) THEN
            ORIGIN = MASK%logical_dimensions(3)/2 +1
        ELSE
            ORIGIN = (MASK%logical_dimensions(3)-1)/2
        ENDIF

        ! Loop along the Z dimension, blank appropriate slices
        DO I=1,MASK%logical_dimensions(3)
            DIST = ABS(ORIGIN-I)
            IF (DIST .GE. RRAD) MASK%real_values(:,:,I) = 0.0
        ENDDO

        ! Now soften the mask
        CALL MASK_ADD_COSINE_FALLOFF(MASK,FFALLOFF)
    END SUBROUTINE IMAGE_MASK_Z_ENDS

    !>  \brief  Mask the first and last few slices in a volume. This is useful for example to remove the discontinuity
    !!          in a helical reconstruction where the Z dimension isn't an integer number of twists.
    SUBROUTINE IMAGE_MASK_Z_ENDS_INPLACE(SELF,WIDTH,SOFT)
        USE IMAGES
        USE INTERPOLATION, ONLY : HANN
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(INOUT)   ::  SELF           !<  Input/output image
        INTEGER,                    INTENT(IN)      ::  WIDTH          !<  Number of Z slices to be masked at either end of the volume
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  SOFT           !<  Either WIDTH slices are blanked, or a WIDTH-wide
        ! Private variables
        LOGICAL ::  SSOFT
        INTEGER ::  I
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        ! This is really only meant for 3D images
        IF (SELF%logical_dimensions(3) .LE. 1) THEN
            WRITE(*,'(A,3(I4,1X))') '**ERROR(IMAGE_MASK_Z_ENDS_INPLACE): this is only meant for 3D images. Input dimensions = ',&
                                     SELF%logical_dimensions
            CALL TERMINATE('Fatal error in IMAGE_MASK_Z_ENDS_INPLACE')
        ENDIF

        ! Will we be softening the mask?
        IF (PRESENT(SOFT)) THEN
            SSOFT = SOFT
        ELSE
            SSOFT = .TRUE.
        ENDIF

        IF (DEBUG) THEN
            WRITE(*,'(A,I0,1X,L1)') '**debug(IMAGE_MASK_Z_ENDS_INPLACE): width, soft? ', WIDTH, SSOFT
        ENDIF


        ! Loop along the Z dimension, blank appropriate slices
        IF (SSOFT) THEN
            ! At i=1, we expect hann = 0.0
            ! At i=width+1, we expect hann = 1.0
            DO I=1,WIDTH
                SELF%real_values(:,:,I) = SELF%real_values(:,:,I) * HANN(REAL(WIDTH-I+1),WIDTH=WIDTH*2)
            ENDDO
            ! At i=self%logical_dimensions(3), we expect hann = 0.0
            ! At i=self%logical_dimensions(3)-width, we expect hann = 1.0
            DO I=SELF%logical_dimensions(3)-WIDTH,SELF%logical_dimensions(3)
                SELF%real_values(:,:,I) = SELF%real_values(:,:,I) * HANN(REAL(I-SELF%logical_dimensions(3)+WIDTH),WIDTH=WIDTH*2)
            ENDDO
        ELSE
            DO I=1,WIDTH
                SELF%real_values(:,:,I) = 0.0
            ENDDO
            DO I=SELF%logical_dimensions(3)-WIDTH+1,SELF%logical_dimensions(3)
                SELF%real_values(:,:,I) = 0.0
            ENDDO
        ENDIF

    END SUBROUTINE IMAGE_MASK_Z_ENDS_INPLACE



    !>  \brief  Calculate a mask for the input image, based on a low-pass filtered version of the image
    !!  Note: in principle this is similar to the D3MASK subroutine from frealign/frealix1, so the results should be similar.
    !!  There are several major differences though: 1. the filtering here is done with a cosine-edged filter, whereas frealix1/frealign uses a Gaussian;
    !!  2. Rather than ask the user for a number of standard deviations, we ask for a percentage of voxels expected to be in the mask
    SUBROUTINE IMAGE_LOW_PASS_MASK(IMAGE_IN,MASK, PCT_IN_MASK,FALLOFF_TYPE,LOW_PASS_FREQUENCY,FALLOFF_WIDTH, SOFTEN, INVERSION)
        USE IMAGES
        USE PARSERS
        USE RUNTIME_PARAMETERS

        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                            INTENT(IN)      ::  IMAGE_IN            !<  Input image
        TYPE(IMAGE),                            INTENT(OUT)     ::  MASK                !<  Output mask
        REAL,                                   INTENT(IN)      ::  PCT_IN_MASK         !<  Percentage of voxels which need to end up with 1.0 in mask
        REAL,                       OPTIONAL,   INTENT(IN)      ::  LOW_PASS_FREQUENCY  !<  Low-pass frequency. In reciprocal pixels (0.5 is Nyquist). Defaults to 0.25.
        REAL,                       OPTIONAL,   INTENT(IN)      ::  FALLOFF_WIDTH       !<  Mask soft falloff width
        CHARACTER(LEN=WORD_LEN),    OPTIONAL,   INTENT(IN)      ::  FALLOFF_TYPE        !<  How and if to have a falloff around the edge of the mask
        LOGICAL,                    OPTIONAL,   INTENT(IN)      ::  SOFTEN              !<  Whether to soften the mask. Defaults to TRUE.
        LOGICAL,                    OPTIONAL,   INTENT(IN)      ::  INVERSION           !<  Whether to invert the mask, letting through background.

        ! Private variables
        REAL                    ::  LLOW_PASS_FREQUENCY
        REAL,       ALLOCATABLE ::  HISTOGRAM_BOUNDARIES(:)
        INTEGER,    ALLOCATABLE ::  HISTOGRAM(:)
        REAL                    ::  THRESHOLD
        INTEGER                 ::  I, J, K, NUM_IN_MASK_TARGET,NUM_IN_MASK_ACTUAL
        LOGICAL                 ::  SSOFTEN
        LOGICAL                 ::  EEXTEND
        LOGICAL                 ::  IINVERSION
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        ! Assumption that incoming image is real
        IF (.not. IMAGE_IN%is_in_real_space) THEN
            WRITE(*,'(A)') '**ERROR(IMAGE_LOW_PASS_MASK): expected real image'
            CALL TERMINATE('Fatal error in IMAGE_LOW_PASS_MASK')
        ENDIF

        ! Low-pass filter radius
        IF (PRESENT(LOW_PASS_FREQUENCY)) THEN
            LLOW_PASS_FREQUENCY = LOW_PASS_FREQUENCY
        ELSE
            LLOW_PASS_FREQUENCY = 0.25
        ENDIF

        ! Soften the mask?
        IF (PRESENT(SOFTEN)) THEN
            SSOFTEN = SOFTEN
        ELSE
            SSOFTEN = .TRUE.
        ENDIF

        ! Extend / soften the mask?
        EEXTEND = .FALSE.
        IF (PRESENT(FALLOFF_TYPE)) THEN
            SELECT CASE (FALLOFF_TYPE)
                CASE('COS_EDGE')
                    SSOFTEN=.TRUE.
                    EEXTEND=.FALSE.
                CASE('HARD_EXTEND')
                    SSOFTEN=.FALSE.
                    EEXTEND=.TRUE.
                CASE('NONE')
                    SSOFTEN=.FALSE.
                    EEXTEND=.FALSE.
                CASE DEFAULT
                WRITE(*,'(2A)') '**ERROR(MODE_MASK): unrecognised falloff type: ', FALLOFF_TYPE
                CALL TERMINATE('Fatal error in MODE_MASK')
            END SELECT
        ENDIF

        !Invert mask?
        IF (PRESENT(INVERSION)) THEN
            IINVERSION = INVERSION
        ELSE
            IINVERSION = .FALSE.
        ENDIF

        ! Copy incoming image
        CALL IMAGE_ALLOCATE(MASK,IMAGE_IN)
        MASK%real_values = IMAGE_IN%real_values

        ! Low pass filter
        CALL MASK%ApplyBandPassFilter(0.0,LLOW_PASS_FREQUENCY,COSINE_FALLOFF=0.025)

        ! Debug dump
        IF (DEBUG) CALL WRITE_TO_DISK(MASK,'dbg_lp.mrc')

        ! Get histogram for the image
        CALL MASK%ComputeHistogram(HISTOGRAM,HISTOGRAM_BOUNDARIES,NUM_OF_BINS=100,NEGS_IN_FIRST_BIN=.TRUE.)

        ! Target number of voxels in mask
        NUM_IN_MASK_TARGET = NINT(REAL(PRODUCT(MASK%logical_dimensions))*PCT_IN_MASK/100.)
        !PRINT *, '**debug(IMAGE_LOW_PASS_MASK): NUM_IN_MASK_TARGET = ', NUM_IN_MASK_TARGET

        ! Work out a threshold at which PCT_IN_MASK % of voxels are included
        THRESHOLD = 0.0
        NUM_IN_MASK_ACTUAL = 0
        DO I=SIZE(HISTOGRAM),2,-1   ! we skip the first bin, cos it only has negative values in it.
            !PRINT '(3(A,I10))', '**debug(IMAGE_LOW_PASS_MASK): on bin ', I, ' of ', SIZE(HISTOGRAM), ': NUM_IN_MASK_ACTUAL = ', NUM_IN_MASK_ACTUAL

            IF (NUM_IN_MASK_TARGET .GE. NUM_IN_MASK_ACTUAL .AND. NUM_IN_MASK_TARGET .LT. NUM_IN_MASK_ACTUAL + HISTOGRAM(I)) THEN
                ! The appropriate threshold is within the current bin - use linear interpolation to get a value
                THRESHOLD = HISTOGRAM_BOUNDARIES(I) + (HISTOGRAM_BOUNDARIES(I+1)-HISTOGRAM_BOUNDARIES(I))   &
                                                        *(NUM_IN_MASK_TARGET-NUM_IN_MASK_ACTUAL)/HISTOGRAM(I)
                !PRINT '(A,ES10.3,A,ES10.3)', '**debug(IMAGE_LOW_PASS_MASK): threshold should be between ', HISTOGRAM_BOUNDARIES(I), ' and ', HISTOGRAM_BOUNDARIES(I+1)
            ELSE
                ! The appropriate threshold is not within the current bin, move on to the next bin

            ENDIF
            NUM_IN_MASK_ACTUAL = NUM_IN_MASK_ACTUAL + HISTOGRAM(I)
        ENDDO

        ! If the threshold hasn't be set, something went wrong
        IF (THRESHOLD .EQ. 0.0) THEN
            WRITE(*,'(A)') '**ERROR(IMAGE_LOW_PASS_MASK): threshold was not found'
            CALL TERMINATE('Fatal error in IMAGE_LOW_PASS_MASK')
        ELSE
            IF (DEBUG) WRITE(*,'(A,ES10.3)') '**debug(IMAGE_LOW_PASS_MASK): threshold set to ', THRESHOLD
        ENDIF

        ! Binarise the filtered map
        DO K=1,MASK%logical_dimensions(3)
            DO J=1,MASK%logical_dimensions(2)
                DO I=1,MASK%logical_dimensions(1)
                    IF (MASK%real_values(I,J,K) .LT. THRESHOLD) THEN
                        MASK%real_values(I,J,K) = 0.0
                    ELSE
                        MASK%real_values(I,J,K) = 1.0
                    ENDIF
                ENDDO
            ENDDO
        ENDDO

        ! Debug: dump the binary image
        IF (DEBUG) THEN
            PRINT *, '**debug(IMAGE_LOW_PASS_MASK): dumping dbg_mask_out.mrc'
            CALL WRITE_TO_DISK(MASK,'dbg_lp_mask_out.mrc')
        ENDIF

        ! Soften the mask
        IF (SSOFTEN) THEN
            CALL MASK_ADD_COSINE_FALLOFF(MASK,5.0)
            IF (DEBUG) CALL WRITE_TO_DISK(MASK,'dbg_mask_out_soft.mrc')
        ENDIF

        IF (EEXTEND) THEN
            WRITE (*,'(A,F0.3)') '**debug(IMAGE_LOW_PASS_MASK): extending mask with the following falloff width: ', FALLOFF_WIDTH
            CALL EXTEND_MASK(MASK, FALLOFF_WIDTH)
        ENDIF

        IF (IINVERSION) THEN
            CALL INVERT_MASK(MASK)
            WRITE (*,*) 'inverting mask'
        ENDIF

    END SUBROUTINE IMAGE_LOW_PASS_MASK

    !>  \brief  Compute the fraction of voxels within a mask image (binary or smooth)
    REAL FUNCTION FRACTION_OF_VOXELS_WITHIN_MASK(MASK) RESULT(FRAC)
        USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(IN)  ::  MASK
        ! Private variables
        REAL(KIND=8)    ::  F_MASK
        INTEGER         ::  I,J,K
        ! Start work

        ! Image is expected to be real
        IF (.NOT. MASK%is_in_real_space) THEN
            WRITE(*,'(A)') '**ERROR(FRACTION_OF_VOXELS_WITHIN_MASK): image is not real'
            CALL TERMINATE('Fatal error in FRACTION_OF_VOXELS_WITHIN_MASK')
        ENDIF

        F_MASK = 0.0D0
        ! Image should only have values between 0 and 1
        ! Loop over voxels
        DO K=1,MASK%logical_dimensions(3)
            DO J=1,MASK%logical_dimensions(2)
                DO I=1,MASK%logical_dimensions(1)
                    IF (MASK%real_values(I,J,K) .GT. 1.0 .OR. MASK%real_values(I,J,K) .LT. 0.0) THEN
                        WRITE(*,'(A,G15.8)') '**ERROR(FRACTION_OF_VOXELS_WITHIN_MASK): unexpected value in mask volume: ', &
                                            MASK%real_values(I,J,K)
                        CALL TERMINATE('Fatal error FRACTION_OF_VOXELS_WITHIN_MASK')
                    ENDIF
                    F_MASK = F_MASK + MASK%real_values(I,J,K)**2
                ENDDO
            ENDDO
        ENDDO
        FRAC = F_MASK / PRODUCT(MASK%logical_dimensions)
    END FUNCTION FRACTION_OF_VOXELS_WITHIN_MASK

    !>  \brief  Add a cosine-edged falloff to a mask. The mask should consists of 1's and 0's only. The 1's will be left untouched.
    !!  \todo   Use stricter definition of max inner radius etc.
    PURE SUBROUTINE MASK_ADD_COSINE_FALLOFF(MASK,FALLOFF_WIDTH)
        USE IMAGES
        USE INTERPOLATION, ONLY  :   HANN
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(INOUT)   ::  MASK            !<  Input mask
        REAL,           INTENT(IN)      ::  FALLOFF_WIDTH   !<  Width of falloff. Pixels.
        ! Private variables
        INTEGER ::  I,J,K
        INTEGER ::  II,JJ,KK
        INTEGER ::  IMIN,IMAX,JMIN,JMAX,KMIN,KMAX
        INTEGER ::  WINDOW_RADIUS
        INTEGER ::  IFALLOFF_WIDTH
        REAL    ::  VAL, RAD, RAD_SQ, RAD_K, RAD_J, RAD_I
        LOGICAL ::  NEAR_EDGE
        REAL    ::  FFALLOFF_WIDTH, FFALLOFF_WIDTH_SQ
        ! Start work


        ! If the falloff parameter is <1.0, assume it's a fraction of inner radius
        IF (FALLOFF_WIDTH .LT. 1.0) THEN
            FFALLOFF_WIDTH = FALLOFF_WIDTH * MASK%logical_dimensions(1) / 2.0  ! This is done the lazy way, assuming square etc. Need to call proper functions to find out max inner radius etc.
        ELSE
            FFALLOFF_WIDTH = FALLOFF_WIDTH
        ENDIF

        FFALLOFF_WIDTH_SQ = FFALLOFF_WIDTH**2

        ! Integer value used as parameter to the hann function
        IFALLOFF_WIDTH = NINT(2.*FFALLOFF_WIDTH)

        ! Work out the size of the "window" to check around each mask voxel
        WINDOW_RADIUS = CEILING(ABS(FFALLOFF_WIDTH))


        ! Loop over the volume
        DO K=1,MASK%logical_dimensions(3)
            DO J=1,MASK%logical_dimensions(2)
                DO I=1,MASK%logical_dimensions(1)
                    ! If the current voxel is "in the mask" (i.e. has value 1.0), proceed
                    IF (MASK%real_values(I,J,K) .EQ. 1.0) THEN
                        ! Find out whether the current voxel has an immediate neighbour which is 0.0
                        KMIN = K-1
                        IF (KMIN .LT. 1) KMIN = 1
                        JMIN = J-1
                        IF (JMIN .LT. 1) JMIN = 1
                        IMIN = I-1
                        IF (IMIN .LT. 1) IMIN = 1
                        KMAX = K+1
                        IF (KMAX .GT. MASK%logical_dimensions(3)) KMAX = MASK%logical_dimensions(3)
                        JMAX = J+1
                        IF (JMAX .GT. MASK%logical_dimensions(2)) JMAX = MASK%logical_dimensions(2)
                        IMAX = I+1
                        IF (IMAX .GT. MASK%logical_dimensions(1)) IMAX = MASK%logical_dimensions(1)
                        NEAR_EDGE = .FALSE.
                        EDGE_CHECK: DO KK=KMIN,KMAX
                            DO JJ=JMIN,JMAX
                                DO II=IMIN,IMAX
                                    IF (MASK%real_values(II,JJ,KK) .LT. 1.0) THEN
                                        NEAR_EDGE = .TRUE.
                                        EXIT EDGE_CHECK
                                    ENDIF
                                ENDDO
                            ENDDO
                        ENDDO EDGE_CHECK
                        IF (NEAR_EDGE) THEN
                            ! Loop over the neighbours
                            KMIN = K-WINDOW_RADIUS
                            IF (KMIN .LT. 1) KMIN = 1
                            JMIN = J-WINDOW_RADIUS
                            IF (JMIN .LT. 1) JMIN = 1
                            IMIN = I-WINDOW_RADIUS
                            IF (IMIN .LT. 1) IMIN = 1
                            KMAX = K+WINDOW_RADIUS
                            IF (KMAX .GT. MASK%logical_dimensions(3)) KMAX = MASK%logical_dimensions(3)
                            JMAX = J+WINDOW_RADIUS
                            IF (JMAX .GT. MASK%logical_dimensions(2)) JMAX = MASK%logical_dimensions(2)
                            IMAX = I+WINDOW_RADIUS
                            IF (IMAX .GT. MASK%logical_dimensions(1)) IMAX = MASK%logical_dimensions(1)
                            DO KK=KMIN,KMAX
                                RAD_K = REAL((KK-K)**2)
                                DO JJ=JMIN,JMAX
                                    RAD_J = REAL((JJ-J)**2)
                                    DO II=IMIN,IMAX
                                        ! If the current neighbour has a value less than 1.0, we'll investigate further
                                        IF (MASK%real_values(II,JJ,KK) .LT. 1.0) THEN
                                            RAD_I = REAL((II-I)**2)
                                            ! How far the current neighbour is from the mask voxel
                                            !RAD = SQRT(REAL((KK-K)**2+(JJ-J)**2+(II-I)**2))
                                            RAD_SQ = RAD_I + RAD_J + RAD_K
                                            IF (RAD_SQ .LE. FFALLOFF_WIDTH_SQ) THEN
                                                RAD = SQRT(RAD_SQ)
                                                VAL = HANN(RAD,IFALLOFF_WIDTH)
                                                IF (MASK%real_values(II,JJ,KK) .LT. VAL) THEN
                                                    MASK%real_values(II,JJ,KK) = VAL
                                                ENDIF
                                            ENDIF
                                        ENDIF
                                    ENDDO
                                ENDDO
                            ENDDO
                        ENDIF ! End of test of whether we are near the edge of the binary mask
                    ENDIF ! End of test of whether current voxel is in mask (values of 1.0)
                ENDDO
            ENDDO
        ENDDO


    END SUBROUTINE MASK_ADD_COSINE_FALLOFF

    !>  \brief  Extend binary mask
    SUBROUTINE EXTEND_MASK(MASK, FALLOFF_WIDTH)
        USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(INOUT)   ::  MASK            !<  Input mask
        REAL,           INTENT(IN)      ::  FALLOFF_WIDTH   !<  Width of extension. Pixels.
        ! Private variables
        INTEGER ::  I,J,K
        INTEGER ::  II,JJ,KK
        INTEGER ::  IMIN,IMAX,JMIN,JMAX,KMIN,KMAX
        INTEGER ::  WINDOW_RADIUS
        REAL    ::  VAL, RAD_SQ, RAD_K, RAD_J, RAD_I
        LOGICAL ::  NEAR_EDGE
        REAL    ::  FFALLOFF_WIDTH, FFALLOFF_WIDTH_SQ
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.



        IF (DEBUG) WRITE(*,'(A)') '**debug(EXTEND_MASK): entering'

        ! If the falloff parameter is <1.0, assume it's a fraction of inner radius
        IF (FALLOFF_WIDTH .LT. 1.0) THEN
            FFALLOFF_WIDTH = FALLOFF_WIDTH * MASK%logical_dimensions(1) / 2.0  ! This is done the lazy way, assuming square etc. Need to call proper functions to find out max inner radius etc.
        ELSE
            FFALLOFF_WIDTH = FALLOFF_WIDTH
        ENDIF

        FFALLOFF_WIDTH_SQ = FFALLOFF_WIDTH**2

        ! Work out the size of the "window" to check around each mask voxel
        WINDOW_RADIUS = CEILING(ABS(FFALLOFF_WIDTH))

        ! Loop over the volume
        DO K=1,MASK%logical_dimensions(3)
            DO J=1,MASK%logical_dimensions(2)
                DO I=1,MASK%logical_dimensions(1)
                    ! If the current voxel is "in the mask" (i.e. has value 1.0), proceed
                    IF (MASK%real_values(I,J,K) .EQ. 1.0) THEN
                        ! Find out whether the current voxel has an immediate neighbour which is 0.0
                        KMIN = K-1
                        IF (KMIN .LT. 1) KMIN = 1
                        JMIN = J-1
                        IF (JMIN .LT. 1) JMIN = 1
                        IMIN = I-1
                        IF (IMIN .LT. 1) IMIN = 1
                        KMAX = K+1
                        IF (KMAX .GT. MASK%logical_dimensions(3)) KMAX = MASK%logical_dimensions(3)
                        JMAX = J+1
                        IF (JMAX .GT. MASK%logical_dimensions(2)) JMAX = MASK%logical_dimensions(2)
                        IMAX = I+1
                        IF (IMAX .GT. MASK%logical_dimensions(1)) IMAX = MASK%logical_dimensions(1)
                        NEAR_EDGE = .FALSE.
                        EDGE_CHECK: DO KK=KMIN,KMAX
                            DO JJ=JMIN,JMAX
                                DO II=IMIN,IMAX
                                    IF (MASK%real_values(II,JJ,KK) .LT. 1.0) THEN
                                        NEAR_EDGE = .TRUE.
                                        EXIT EDGE_CHECK
                                    ENDIF
                                ENDDO
                            ENDDO
                        ENDDO EDGE_CHECK
                        IF (NEAR_EDGE) THEN
                            ! Loop over the neighbours
                            KMIN = K-WINDOW_RADIUS
                            IF (KMIN .LT. 1) KMIN = 1
                            JMIN = J-WINDOW_RADIUS
                            IF (JMIN .LT. 1) JMIN = 1
                            IMIN = I-WINDOW_RADIUS
                            IF (IMIN .LT. 1) IMIN = 1
                            KMAX = K+WINDOW_RADIUS
                            IF (KMAX .GT. MASK%logical_dimensions(3)) KMAX = MASK%logical_dimensions(3)
                            JMAX = J+WINDOW_RADIUS
                            IF (JMAX .GT. MASK%logical_dimensions(2)) JMAX = MASK%logical_dimensions(2)
                            IMAX = I+WINDOW_RADIUS
                            IF (IMAX .GT. MASK%logical_dimensions(1)) IMAX = MASK%logical_dimensions(1)
                            DO KK=KMIN,KMAX
                                RAD_K = REAL((KK-K)**2)
                                DO JJ=JMIN,JMAX
                                    RAD_J = REAL((JJ-J)**2)
                                    DO II=IMIN,IMAX
                                        ! If the current neighbour has a value less than 1.0, we'll investigate further
                                        IF (MASK%real_values(II,JJ,KK) .LT. 1.0) THEN
                                            RAD_I = REAL((II-I)**2)
                                            ! How far the current neighbour is from the mask voxel
                                            !RAD = SQRT(REAL((KK-K)**2+(JJ-J)**2+(II-I)**2))
                                            RAD_SQ = RAD_I + RAD_J + RAD_K
                                            IF (RAD_SQ .LE. FFALLOFF_WIDTH_SQ) THEN
                                                ! Distinguish new mask voxels from old mask voxels by giving them value of 2.0
                                                VAL = 2.0
                                                IF (MASK%real_values(II,JJ,KK) .LT. 1.0) THEN
                                                    MASK%real_values(II,JJ,KK) = VAL
                                                ENDIF
                                            ENDIF
                                        ENDIF
                                    ENDDO
                                ENDDO
                            ENDDO
                        ENDIF
                    ENDIF
                ENDDO
            ENDDO
        ENDDO

        ! New mask voxels are now 1.0
        DO K=1,MASK%logical_dimensions(3)
            DO J=1,MASK%logical_dimensions(2)
                DO I=1,MASK%logical_dimensions(1)
                    IF (MASK%real_values(I,J,K) .EQ. 2.0) THEN
                        MASK%real_values(I,J,K)=1.0
                    ENDIF
                ENDDO
            ENDDO
        ENDDO


        IF (DEBUG) WRITE(*,'(A)') '**debug(EXTEND_MASK): all done'
    END SUBROUTINE EXTEND_MASK


    !>  \brief Invert a mask
    SUBROUTINE INVERT_MASK(MASK)
        USE IMAGES
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),    INTENT(INOUT)   ::  MASK            !<  Input mask
        INTEGER ::  I,J,K

        ! Check that the mask is binary, if not, crash out
        IF (.NOT. ALL(  MASK%real_values(1:MASK%logical_dimensions(1),1:MASK%logical_dimensions(2),1:MASK%logical_dimensions(3)) &
                                                              .EQ. 1.0  .OR. &
                        MASK%real_values(1:MASK%logical_dimensions(1),1:MASK%logical_dimensions(2),1:MASK%logical_dimensions(3)) &
                                                              .EQ. 0.0 ) ) THEN
            WRITE(*,'(A)') '**ERROR(INVERT_MASK): this subroutine cannot deal with soft masks'
            CALL TERMINATE('Feature not implemented yet')
        ENDIF

        !
        DO K=1,MASK%logical_dimensions(3)
            DO J=1,MASK%logical_dimensions(2)
                DO I=1,MASK%logical_dimensions(1)
                    IF (MASK%real_values(I,J,K) .LT. 0.5) THEN
                        MASK%real_values(I,J,K)=1.0
                    ELSE
                        MASK%real_values(I,J,K)=0.0
                    ENDIF
                ENDDO
            ENDDO
        ENDDO
    END SUBROUTINE INVERT_MASK

END MODULE MASKS
