!>  \brief  Methods to minimise functions
MODULE MINIMISERS
    USE, INTRINSIC :: ISO_C_BINDING
    USE :: FGSL
    USE :: DIFFERENTIAL_EVOLUTION
    IMPLICIT NONE
    ! By default, stuff is hidden from the outside world
    PRIVATE
    ! These interfaces are public
    PUBLIC :: DS_RUN, DS_INIT, DS_FREE, DS_SET_START_STEP
    PUBLIC :: FGSL_MINIMISABLE_FUNCTION, DS_PRINT_PROGRESS_SUMMARY


    !
    ! Abstract interface for function used by the FGSL DS minimiser
    !

    !>  \brief  What a minimisable function should look like
    ABSTRACT INTERFACE
        FUNCTION FGSL_MINIMISABLE_FUNCTION(VALUES,PARAMS) BIND(C)
            IMPORT :: C_DOUBLE, C_PTR   ! Because we're in an abstract interface, we normally don't have access to all the stuff from the module, so we need to IMPORT
            TYPE(C_PTR), VALUE  ::  VALUES
            TYPE(C_PTR), VALUE  ::  PARAMS
            REAL(KIND=C_DOUBLE) ::  FGSL_MINIMISABLE_FUNCTION
        END FUNCTION FGSL_MINIMISABLE_FUNCTION
    END INTERFACE

    !
    ! Parameters for the downhill simplex
    !
    REAL,       PARAMETER,  PUBLIC  ::  SEARCH_ACCURACY_TO_DS_SIZE_RATIO    =   1.0
    INTEGER,    PARAMETER,  PUBLIC  ::  DS_MAX_ITERATIONS                   =   1000

    !
    ! Abstract interfaces for functions used by the FGSL SA minimiser
    !

    !>  \brief  What an energy function should look like (used by FGSL simulated annealing for scoring function, for example)
    ABSTRACT INTERFACE
        FUNCTION FGSL_ENERGY_FUNCTION(PARAMS) BIND(C)
            IMPORT              ::  C_DOUBLE, C_PTR
            TYPE(C_PTR), VALUE  ::  PARAMS
            REAL(KIND=C_DOUBLE) ::  FGSL_ENERGY_FUNCTION
        END FUNCTION FGSL_ENERGY_FUNCTION
    END INTERFACE

    !>  \brief  What a stepping function should look like (used by FGSL simulated annealing for modifying the current configuration)
    ABSTRACT INTERFACE
        SUBROUTINE FGSL_STEPPING_FUNCTION(RNG,CONF,MAX_STEP) BIND(C)
            IMPORT  ::  C_PTR, FGSL_DOUBLE
            TYPE(C_PTR),        VALUE   ::  RNG         !<  Pointer to random number generator
            TYPE(C_PTR),        VALUE   ::  CONF        !<  Pointer to configuration to alter
            REAL(FGSL_DOUBLE),  VALUE   ::  MAX_STEP    !<  Maximum step size
        END SUBROUTINE FGSL_STEPPING_FUNCTION
    END INTERFACE

    !>  \brief  What a metric function should look like (used by FGSL SA to measure the distance between two configurations)
    ABSTRACT INTERFACE
        FUNCTION FGSL_METRIC_FUNCTION(CONF1,CONF2) BIND(C)
            IMPORT              ::  FGSL_DOUBLE, C_PTR
            TYPE(C_PTR), VALUE  ::  CONF1
            TYPE(C_PTR), VALUE  ::  CONF2
            REAL(FGSL_DOUBLE)   ::  FGSL_METRIC_FUNCTION
        END FUNCTION FGSL_METRIC_FUNCTION
    END INTERFACE

    !>  \brief  What a printing function should look like (used by FGSL SA to printout information about the current configuration)
    ABSTRACT INTERFACE
        SUBROUTINE FGSL_PRINT_FUNCTION(CONF) BIND(C)
            IMPORT  ::  C_PTR
            TYPE(C_PTR),    VALUE   ::  CONF
        END SUBROUTINE FGSL_PRINT_FUNCTION
    END INTERFACE

    !>  \brief  What a configuration constructor should look like
    !ABSTRACT INTERFACE
    !   SUBROUTINE FGSL_CONF_CONSTRUCTOR(CONF) BIND(C)
    !       IMPORT  ::  C_PTR
    !       TYPE(C_PTR),    VALUE   ::  CONF
    !   END SUBROUTINE
    !END INTERFACE

    !>  \brief  What a configuration constructor should look like
    !ABSTRACT INTERFACE
    !   SUBROUTINE FGSL_CONF_DESTRUCTOR(CONF) BIND(C)
    !       IMPORT  ::  C_PTR
    !       TYPE(C_PTR),    VALUE   ::  CONF
    !   END SUBROUTINE
    !END INTERFACE

    !>  \brief  What a configuration copy function should look like
    !ABSTRACT INTERFACE
    !   SUBROUTINE  FGSL_CONF_COPIER(CONF_SOURCE,CONF_DEST) BIND(C)
    !       IMPORT  ::  C_PTR
    !       TYPE(C_PTR),    VALUE   ::  CONF_SOURCE
    !       TYPE(C_PTR),    VALUE   ::  CONF_DEST
    !   END SUBROUTINE
    !END INTERFACE

    !
    ! Minimiser types
    !

    !>  \brief  The basic minimiser type
    TYPE, ABSTRACT ::   MINIMISER
        INTEGER ::  N   !<  Number of dimensions
        LOGICAL ::  IN_MEMORY = .FALSE. !<  Whether allocated memory
    END TYPE

    !>  \brief  Brute-force (exhaustive search) minimiser
    TYPE, PUBLIC, EXTENDS(MINIMISER)    ::  MINIMISER_BF
        PROCEDURE(FGSL_MINIMISABLE_FUNCTION), POINTER, NOPASS   ::  FUNCTION => null()  !<  A pointer to the function the minimiser aims to minimise.
                                                                                        !!  NO PASS because there an assumption by default that type-bound procedures have an object of the type to which they are bound as first argument.
                                                                                        !!  This default assumption is overriden by NO PASS.
        REAL, ALLOCATABLE   ::  STARTING_VALUE(:)   !<  For each dimension, the starting value
        REAL, ALLOCATABLE   ::  HALF_RANGE(:)       !<  For each dimension, the half-range to search over
        REAL, ALLOCATABLE   ::  STEP_SIZE(:)        !<  For each dimension, the stepping size during the search
        INTEGER             ::  NUM_ITERATIONS      !<  The total number of iterations needed to complete the exhaustive search
        !CONTAINS
        !   PROCEDURE   ::  FREE => BF_FREE
    END TYPE

    !
    ! Downhill simplex types
    !

    !>  \brief  Downhill simplex multi-dimensional minimiser
    TYPE, PUBLIC, EXTENDS(MINIMISER) :: MINIMISER_DS
        PROCEDURE(FGSL_MINIMISABLE_FUNCTION), POINTER, NOPASS   ::  FUNCTION => null()  !<  A pointer to the function the minimiser aims to minimise.
                                                                                        !!  NO PASS because there an assumption by default that type-bound procedures have an object of the type to which they are bound as first argument.
                                                                                        !!  This default assumption is overriden by NO PASS.
        REAL,   ALLOCATABLE    ::   BEST_SCORE(:)                          !<  For each iteration, a record of the best score
        REAL,   ALLOCATABLE    ::   SIMPLEX_SIZE(:)                        !<  For each iteration, a record of the simplex size
        INTEGER                 ::  MAX_NUM_ITER  =     DS_MAX_ITERATIONS  !<  Maximum number of iteration
        LOGICAL                 ::  DOUBLE_RUNS   =     .TRUE.             !<  Start a second minimisation run from the end of the first
    END TYPE


    !>  \brief  DS minimiser which relies on the FGSL
    TYPE, PUBLIC, EXTENDS(MINIMISER_DS) :: MINIMISER_DS_FGSL
        INTEGER(FGSL_SIZE_T)                                ::  N_FGSL
        TYPE(FGSL_MULTIMIN_FMINIMIZER)                      ::  MINIMISER_FGSL
        CHARACTER(LEN=FGSL_STRMAX,KIND=FGSL_CHAR)           ::  NAME
        TYPE(FGSL_MULTIMIN_FUNCTION)                        ::  FUNCTION_FGSL
        REAL(FGSL_DOUBLE),                       POINTER    ::  CURRENT_BEST_FGSL(:)        !<  The search parameter values that gave the best score
        REAL(FGSL_DOUBLE)                                   ::  CURRENT_MINIMUM_FGSL        !<  The value of the best score
    END TYPE MINIMISER_DS_FGSL

    !
    ! Simulated annealing types
    !

    !>  \brief  Simulated annealing minimiser
    TYPE, PUBLIC, EXTENDS(MINIMISER)    ::  MINIMISER_SA
        TYPE(FGSL_RNG)      ::  RNG             !<  Random number generator (provided by the FGSL)
        INTEGER             ::  ITERS_FIXED_T   !<  How many iterations for each temperature?
        REAL                ::  KB              !<  Boltzmann constant
        REAL                ::  T_INITIAL       !<  Initial temperature
        REAL                ::  T_DAMPING       !<  Damping factor
        REAL                ::  T_MIN           !<  Minimum temperature
        REAL                ::  CURRENT_ENERGY  !<  Current energy
        REAL                ::  T_CURRENT       !<  Current temperature
    END TYPE MINIMISER_SA

    !>  \brief  SA (simulated annealing) minimiser which relies on the FGSL
    TYPE, PUBLIC, EXTENDS(MINIMISER_SA) ::  MINIMISER_SA_FGSL
        INTEGER(FGSL_SIZE_T)        ::  N_FGSL                  !<  Number of dimensions (as an FGSL integer)
        TYPE(FGSL_SIMAN_PARAMS_T)   ::  SA_PARAMS               !<  Data structure that holds all parameters that control the SA run
        INTEGER(FGSL_INT)           ::  NUM_TRIES_FGSL          !<  How many points do we try before stepping
        INTEGER(FGSL_INT)           ::  ITERS_FIXED_T_FGSL      !<  How many iterations for each temperature?
        REAL(FGSL_DOUBLE)           ::  MAX_STEP_SIZE_FGSL      !<  Maximum step size in random walk
        REAL(FGSL_DOUBLE)           ::  KB_FGSL                 !<  Boltzmann constant
        REAL(FGSL_DOUBLE)           ::  T_INITIAL_FGSL          !<  Initial temperature
        REAL(FGSL_DOUBLE)           ::  T_DAMPING_FGSL          !<  Damping factor
        REAL(FGSL_DOUBLE)           ::  T_MIN_FGSL              !<  Minimum temperature
        TYPE(C_PTR)                 ::  CURRENT_CONF            !<  Current configuration (pointer to object with all information needed to return an energy/score) (at initialisation, starting position; upon solvation, search result)
        PROCEDURE(FGSL_ENERGY_FUNCTION),    POINTER,    NOPASS  ::  ENERGY_FUNCTION     !<  Pointer to energy function
        PROCEDURE(FGSL_STEPPING_FUNCTION),  POINTER,    NOPASS  ::  STEPPING_FUNCTION   !<  Pointer to stepping function
        PROCEDURE(FGSL_METRIC_FUNCTION),    POINTER,    NOPASS  ::  METRIC_FUNCTION     !<  Pointer to metric function
        PROCEDURE(FGSL_PRINT_FUNCTION),     POINTER,    NOPASS  ::  PRINT_FUNCTION      !<  Pointer to a print function
        !PROCEDURE(FGSL_CONF_CONSTRUCTOR),  POINTER,    NOPASS  ::  CONF_CONSTRUCTOR    !<  Function to allocate/construct a configuration (state) object
        !PROCEDURE(FGSL_CONF_DESTRUCTOR),   POINTER,    NOPASS  ::  CONF_DESTRUCTOR     !<  Function to free/deallocate/destroy a configuration object
        !PROCEDURE(FGSL_CONF_COPIER),       POINTER,    NOPASS  ::  CONF_COPIER         !<  Function to copy one configuration (state) to another
        INTEGER                     ::  SIZE_OF_CONF            !<  Size (in bytes) of the configuration object CURRENT_CONF
    END TYPE MINIMISER_SA_FGSL

    !
    ! Interfaces
    !

    INTERFACE DS_RUN
        MODULE PROCEDURE FGSL_DS_RUN
    END INTERFACE

    INTERFACE  DS_INIT
        MODULE PROCEDURE FGSL_DS_INIT
    END INTERFACE

    INTERFACE   DS_SET_START_STEP
        MODULE PROCEDURE FGSL_DS_SET_START_STEP
    END INTERFACE

    INTERFACE DS_UPDATE_CURRENT_BEST
        MODULE PROCEDURE FGSL_DS_UPDATE_CURRENT_BEST
    END INTERFACE

    INTERFACE DS_FREE
        MODULE PROCEDURE FGSL_DS_FREE
    END INTERFACE

    INTERFACE SA_INIT
        MODULE PROCEDURE MINIMISER_SA_FGSL_INIT
    END INTERFACE

    INTERFACE SA_SOLVE
        MODULE PROCEDURE MINIMISER_SA_FGSL_SOLVE
    END INTERFACE

    CONTAINS

    !>  \brief  Solve the simulated annealing problem
    SUBROUTINE MINIMISER_SA_FGSL_SOLVE(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(MINIMISER_SA_FGSL),    INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER(FGSL_SIZE_T)    ::  SIZE_OF_CONF_FGSL
        ! Start work
        SIZE_OF_CONF_FGSL = SELF%SIZE_OF_CONF
        !
        CALL TERMINATE('Not implemented')
        ! The portland compiler (v 13.3) does not like this call below
        !CALL FGSL_SIMAN_SOLVE(  SELF%RNG,SELF%CURRENT_CONF, &
        !                        SELF%ENERGY_FUNCTION,SELF%STEPPING_FUNCTION,SELF%METRIC_FUNCTION,SELF%PRINT_FUNCTION,   &
        !                        ELEMENT_SIZE=SIZE_OF_CONF_FGSL,PARAMS=SELF%SA_PARAMS)
    END SUBROUTINE MINIMISER_SA_FGSL_SOLVE

    !>  \brief  Initialise the SA minimiser
    SUBROUTINE MINIMISER_SA_FGSL_INIT(  SELF,N,STARTING_CONF, SIZE_OF_CONF, &
                                        ENERGY_FUNCTION, STEPPING_FUNCTION, METRIC_FUNCTION, &
                                        PRINT_FUNCTION, &
                                        !CONF_CONSTRUCTOR, CONF_DESTRUCTOR, CONF_COPIER,    &
                                        NUM_TRIES,ITERS_FIXED_T,MAX_STEP_SIZE,  &
                                        KB,T_INITIAL,T_DAMPING,T_MIN)
        USE, INTRINSIC  ::  ISO_C_BINDING
        IMPLICIT NONE
        ! Arguments
        TYPE(MINIMISER_SA_FGSL),    INTENT(INOUT)   ::  SELF
        INTEGER,                    INTENT(IN)      ::  N                   !<  Number of dimensions
        TYPE(C_PTR),                INTENT(IN)      ::  STARTING_CONF       !<  Starting configuration (pointer to an object that contains all information needed to score)
        INTEGER,                    INTENT(IN)      ::  SIZE_OF_CONF        !<  Size (in bytes) of the configuration object (this can easily be worked out by using the sizeof() inquuiry intrinsic function
        PROCEDURE(FGSL_ENERGY_FUNCTION)             ::  ENERGY_FUNCTION     !<  Energy/scoring function
        PROCEDURE(FGSL_STEPPING_FUNCTION)           ::  STEPPING_FUNCTION   !<  Function to alter the configuration by performing a step
        PROCEDURE(FGSL_METRIC_FUNCTION)             ::  METRIC_FUNCTION     !<  Function to measure distance between two configurations / states
        PROCEDURE(FGSL_PRINT_FUNCTION)              ::  PRINT_FUNCTION      !<  Function to print out information about the current configuration / state
        !PROCEDURE(FGSL_CONF_CONSTRUCTOR)           ::  CONF_CONSTRUCTOR    !<  Function to allocate/construct a configuration (state) object
        !PROCEDURE(FGSL_CONF_DESTRUCTOR)                ::  CONF_DESTRUCTOR     !<  Function to free/deallocate/destroy a configuration object
        !PROCEDURE(FGSL_CONF_COPIER)                    ::  CONF_COPIER         !<  Function to copy one configuration (state) to another
        INTEGER,                    INTENT(IN)      ::  NUM_TRIES           !<  How many points do we try before stepping
        INTEGER,                    INTENT(IN)      ::  ITERS_FIXED_T       !<  How many iterations for each temperature?
        REAL,                       INTENT(IN)      ::  MAX_STEP_SIZE       !<  Maximum step size in random walk
        REAL,                       INTENT(IN)      ::  KB                  !<  Boltzmann constant
        REAL,                       INTENT(IN)      ::  T_INITIAL           !<  Initial temperature
        REAL,                       INTENT(IN)      ::  T_DAMPING           !<  Damping factor for temperature
        REAL,                       INTENT(IN)      ::  T_MIN               !<  Minimum temperature
        ! Private variables
        TYPE(FGSL_RNG_TYPE) ::  T

        ! Start work
        ! The dimensionality
        SELF%N      =   N
        SELF%N_FGSL =   SELF%N
        ! Setup the random number generator
        T           =   FGSL_RNG_ENV_SETUP()
        SELF%RNG    =   FGSL_RNG_ALLOC(T)
        ! Setup SA parameters
        SELF%NUM_TRIES_FGSL     =   NUM_TRIES
        SELF%ITERS_FIXED_T_FGSL =   ITERS_FIXED_T
        SELF%MAX_STEP_SIZE_FGSL =   MAX_STEP_SIZE
        SELF%KB_FGSL            =   KB
        SELF%T_INITIAL_FGSL     =   T_INITIAL
        SELF%T_DAMPING_FGSL     =   T_DAMPING
        SELF%T_MIN_FGSL         =   T_MIN
        SELF%CURRENT_CONF       =   STARTING_CONF
        SELF%SIZE_OF_CONF       =   SIZE_OF_CONF
        SELF%ENERGY_FUNCTION    =>  ENERGY_FUNCTION
        SELF%STEPPING_FUNCTION  =>  STEPPING_FUNCTION
        SELF%METRIC_FUNCTION    =>  METRIC_FUNCTION
        SELF%PRINT_FUNCTION     =>  PRINT_FUNCTION
        !SELF%CONF_CONSTRUCTOR  =>  CONF_CONSTRUCTOR
        !SELF%CONF_DESTRUCTOR   =>  CONF_DESTRUCTOR
        !SELF%CONF_COPIER       =>  CONF_COPIER

        ! Do the initialisation
        CALL FGSL_SIMAN_PARAMS_INIT(    SELF%SA_PARAMS, &
                                        SELF%NUM_TRIES_FGSL,SELF%ITERS_FIXED_T_FGSL,SELF%MAX_STEP_SIZE_FGSL,    &
                                        SELF%KB_FGSL, SELF%T_INITIAL_FGSL,SELF%T_DAMPING_FGSL,SELF%T_MIN_FGSL)

    END SUBROUTINE MINIMISER_SA_FGSL_INIT


    !>  \brief  Run the minimisation
    SUBROUTINE FGSL_DS_RUN(SELF,TARGET_SIZE)
        USE UI, ONLY: INTERVAL_PRINT
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        use DatesAndTimes
        USE OMP_LIB
        IMPLICIT NONE
        ! Arguments
        TYPE(MINIMISER_DS_FGSL),    INTENT(INOUT)   ::  SELF                !<  Minimiser object
        REAL,                       INTENT(IN)      ::  TARGET_SIZE         !<  Target size that the
        ! Private variables
        INTEGER(FGSL_INT)   ::  STATUS
        INTEGER             ::  I
        REAL(FGSL_DOUBLE)   ::  CURRENT_SIZE
        REAL(FGSL_DOUBLE)   ::  TARGET_SIZE_FGSL
        REAL                ::  CURRENT_BEST_SCORE = 0.0
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        type(DateAndTime)   ::  myTime
        ! Start work
        IF (DEBUG) WRITE(*,'(A)') '**debug(FGSL_DS_RUN): entered'
        ! Check the minimiser's been initialised
        IF (.NOT. SELF%IN_MEMORY) THEN
            WRITE(*,'(A)') '**ERROR(FGSL_DS_RUN): minimiser not initialised'
            CALL TERMINATE('Fatal error in FGSL_DS_RUN')
        ENDIF
        ! Target size of the Simplex. When the simplex has shrunk to this size, the iterations stop.
        TARGET_SIZE_FGSL = TARGET_SIZE
        ! Work out how often we'll print info
        !OUTPUT_INTERVAL = NINT(REAL(MAX_NUM_ITERATIONS)/100.)
        !IF (OUTPUT_INTERVAL .LT. 1) OUTPUT_INTERVAL = 1

        ! Start iterations
        I = 0
        DO
            I = I + 1

            ! Do an iteration
            STATUS = FGSL_MULTIMIN_FMINIMIZER_ITERATE(SELF%MINIMISER_FGSL)
            IF (STATUS .NE. FGSL_SUCCESS) THEN
                WRITE(*,'(A,I4)') '**ERROR(FGSL_DS_RUN): failure at iteration ', I
                CALL TERMINATE('Fatal error in FGSL_DS_RUN')
            ENDIF
            !PRINT '(A,I4,F12.6)', '**debug(FGSL_DS_RUN): iteration, best score: ', I, FGSL_MULTIMIN_FMINIMIZER_MINIMUM(SELF%MINIMISER_FGSL)

            ! Check the size of the simplex
            CURRENT_SIZE = FGSL_MULTIMIN_FMINIMIZER_SIZE(SELF%MINIMISER_FGSL)

            ! Check the current best score
            CURRENT_BEST_SCORE = FGSL_MULTIMIN_FMINIMIZER_MINIMUM(SELF%MINIMISER_FGSL)

            ! Diagnostics
            !IF (MOD(I-1,OUTPUT_INTERVAL) .EQ. 0) THEN
            !IF (INTERVAL_PRINT(I,MAX_NUM_ITERATIONS)) THEN
            IF (OMP_GET_ACTIVE_LEVEL() .EQ. 0) THEN
                !$OMP   CRITICAL (FGSL_DS_RUN_OMP_CRIT)
                PRINT '(2A,2(A,I5),A,E13.6,A,ES10.3)',  '**info(FGSL_DS_RUN): ', myTime%DateAndTimeAsString(), &
                                                        ': on iteration ', I , &
                                                        ' of a maximum of ', SELF%MAX_NUM_ITER, '. Current best score = ', &
                                                        CURRENT_BEST_SCORE, '. Current simplex size = ', CURRENT_SIZE
                !$OMP   END CRITICAL (FGSL_DS_RUN_OMP_CRIT)
            ENDIF

            ! Keep track of best score and simplex size by growing the relevant arrays
            CALL GROW_ALLOCATABLE_ARRAY_REAL(SELF%BEST_SCORE)
            CALL GROW_ALLOCATABLE_ARRAY_REAL(SELF%SIMPLEX_SIZE)
            SELF%BEST_SCORE(SIZE(SELF%BEST_SCORE))    =  -CURRENT_BEST_SCORE
            SELF%SIMPLEX_SIZE(SIZE(SELF%SIMPLEX_SIZE))  =   CURRENT_SIZE
            IF (I .EQ. 1) SELF%SIMPLEX_SIZE(1) = CURRENT_SIZE

            ! The best score should never decrease
            IF (SIZE(SELF%BEST_SCORE) .GT. 1) THEN
                IF (SELF%BEST_SCORE(SIZE(SELF%BEST_SCORE)) - SELF%BEST_SCORE(SIZE(SELF%BEST_SCORE)-1) .LT. -.01) THEN
                    WRITE(*,'(A,F0.5,A,F0.5)') '**ERROR(FGSL_DS_RUN): best score went from ', &
                                                SELF%BEST_SCORE(SIZE(SELF%BEST_SCORE)-1), ' to ', &
                                                SELF%BEST_SCORE(SIZE(SELF%BEST_SCORE))
                    IF (I .EQ. 1) THEN
                        ! This is an ugly work-around for a situation encountered 2012-aug-23 whereby
                        ! the filament_get_path routine returns 1 more waypoint when called by the minimiser
                        ! than when called by fmc_eval when getting the starting_score. It looks like this is due
                        ! to changes of < 0.001 pixels in the path length calculation, but I've not been able to find
                        ! a proper fix
                        WRITE(*,'(A)') '**info(FGSL_DS_RUN): ignoring decreasing-score error because we are in the first iteration'
                    ELSE
                        CALL TERMINATE('Fatal error in FGSL_DS_RUN')
                    ENDIF
                ENDIF
            ENDIF

            ! Check the status of the simplex
            STATUS = FGSL_MULTIMIN_TEST_SIZE(CURRENT_SIZE,TARGET_SIZE_FGSL)
            IF (STATUS .EQ. FGSL_SUCCESS .OR. I .EQ. SELF%MAX_NUM_ITER) THEN
                EXIT
            ENDIF
        ENDDO
        ! Diagnostics
        IF (I .GE. SELF%MAX_NUM_ITER) THEN
            PRINT '(2A,2(A,I0),A,F0.6)', '**info(FGSL_DS_RUN): ', myTime%DateAndTimeAsString(), ': reached iteration ', I , &
                                        ' of a maximum of ', SELF%MAX_NUM_ITER, '. Best score = ', CURRENT_BEST_SCORE
        ELSE
            PRINT '(2A,2(A,I0),A,F0.6)', '**info(FGSL_DS_RUN): ', myTime%DateAndTimeAsString(), ': &
                                         target size reached at iteration ', &
                                         I , ' of a maximum of ', SELF%MAX_NUM_ITER, '. Best score = ', CURRENT_BEST_SCORE
        ENDIF

        ! Update the best values
        CALL FGSL_DS_UPDATE_CURRENT_BEST(SELF)
        !PRINT *, '**debug(FGSL_DS_RUN): with current values for the variables = ', SELF%CURRENT_BEST_FGSL
    END SUBROUTINE FGSL_DS_RUN

    !>  \brief  Print out best score and size per iteration
    SUBROUTINE DS_PRINT_PROGRESS_SUMMARY(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(MINIMISER_DS), INTENT(IN)  ::  SELF
        ! Private variables
        INTEGER ::  I
        ! Start work
        WRITE(*,'(3(A12,1X))') 'Iteration #', 'Simplex size', 'Best score'
        DO I=1,SIZE(SELF%BEST_SCORE)
            WRITE(*,'(I12,1X,F12.5,1X,F12.5)') I-1, SELF%SIMPLEX_SIZE(I), SELF%BEST_SCORE(I)
        ENDDO
    END SUBROUTINE DS_PRINT_PROGRESS_SUMMARY

    SUBROUTINE GROW_ALLOCATABLE_ARRAY_REAL(SELF)
        IMPLICIT NONE
        ! Arguments
        REAL,   ALLOCATABLE,    INTENT(INOUT) ::  SELF(:)
        ! Private variables
        REAL,   ALLOCATABLE ::  TMP(:)
        INTEGER ::  NEW_SIZE
        ! Start work
        IF (ALLOCATED(SELF)) THEN
            NEW_SIZE = SIZE(SELF) + 1
        ELSE
            NEW_SIZE = 1
        ENDIF
        ALLOCATE(TMP(NEW_SIZE))
        IF (ALLOCATED(SELF)) THEN
            TMP(1:SIZE(SELF)) = SELF
        ELSE
            TMP = 0.0
        ENDIF
        CALL MOVE_ALLOC(FROM=TMP,TO=SELF)
    END SUBROUTINE GROW_ALLOCATABLE_ARRAY_REAL

    !>  \brief  Initialise a FGSL downhill simplex minimiser
    SUBROUTINE FGSL_DS_INIT(SELF,FUNCTION_TO_MINIMISE,N,PARAMETERS,INITIAL_SCORE,RANDOM,MAX_NUM_ITER)
        USE, INTRINSIC  ::  ISO_C_BINDING
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(MINIMISER_DS_FGSL),    INTENT(INOUT)   ::  SELF
        PROCEDURE(FGSL_MINIMISABLE_FUNCTION)        ::  FUNCTION_TO_MINIMISE
        INTEGER,                    INTENT(IN)      ::  N                           !<  Number of dimensions
        TYPE(C_PTR),                INTENT(IN)      ::  PARAMETERS                  !<  Parameters (a C pointer to an object, an array, whatever... ;
                                                                                    !!  probably makes sense to have it point to a derived-type object that
                                                                                    !!  contains useful stuff for the function to be minimised)
        REAL                                        ::  INITIAL_SCORE
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  RANDOM                      !<  Whether the starting setup of the DS should be randomised. False by default.
        INTEGER,        OPTIONAL,   INTENT(IN)      ::  MAX_NUM_ITER                !<  Maximum number of iterations
        ! Private variables
        LOGICAL                     ::  SIMPLEX_RANDOM_INIT
        LOGICAL,    PARAMETER       ::  DEBUG   =   .FALSE.
        ! Start work

        IF (DEBUG) WRITE(*,'(A,F0.6)') '**debug(FGSL_DS_INIT): entry, with initial score = ', INITIAL_SCORE

        ! Will we use the random version of the DS
        IF (PRESENT(RANDOM)) THEN
            SIMPLEX_RANDOM_INIT = RANDOM
        ELSE
            SIMPLEX_RANDOM_INIT = .FALSE.
        ENDIF

        ! Initialise the minimizer
        SELF%N = N
        SELF%N_FGSL = SELF%N
        IF (SIMPLEX_RANDOM_INIT) THEN
            WRITE(*,'(A)') '**info(FGSL_DS_INIT): initialising the DS using random basis vectors instead of fixed coordinate axes'
            SELF%MINIMISER_FGSL = FGSL_MULTIMIN_FMINIMIZER_ALLOC(FGSL_MULTIMIN_FMINIMIZER_NMSIMPLEX2RAND,SELF%N_FGSL)
        ELSE
            SELF%MINIMISER_FGSL = FGSL_MULTIMIN_FMINIMIZER_ALLOC(FGSL_MULTIMIN_FMINIMIZER_NMSIMPLEX2,SELF%N_FGSL)
        ENDIF
        SELF%NAME = FGSL_MULTIMIN_FMINIMIZER_NAME(SELF%MINIMISER_FGSL)
        !PRINT *, '**debug(FGSL_DS_INIT): initialising minimiser named ', TRIM(SELF%NAME)
        ! Initialise the function to be minimised
        SELF%FUNCTION_FGSL = FGSL_MULTIMIN_FUNCTION_INIT(FUNCTION_TO_MINIMISE,SELF%N_FGSL,PARAMETERS)
        SELF%FUNCTION => FUNCTION_TO_MINIMISE

        !CALL FGSL_DS_SET_START_STEP(SELF,STARTING_VALUES,FIRST_STEP)

        ! The GSL minimisers maintain a current best estimate of the minimum at all times, which can be accessed as follows:
        !STARTING_VALUES_VEC = FGSL_MULTIMIN_FMINIMIZER_X(SELF%MINIMISER_FGSL)
        !STATUS = FGSL_VECTOR_ALIGN(SELF%CURRENT_BEST_FGSL,STARTING_VALUES_VEC)
        !IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('Failed vector align for current_best in FGSL_DS_INIT')
        !PRINT *, '**debug(FGSL_DS_INIT): done. current best params = ', SELF%CURRENT_BEST_FGSL

        ! Remember the initial score
        IF (ALLOCATED(SELF%BEST_SCORE)) DEALLOCATE(SELF%BEST_SCORE)
        IF (ALLOCATED(SELF%SIMPLEX_SIZE)) DEALLOCATE(SELF%SIMPLEX_SIZE)
        IF (.NOT. ALLOCATED(SELF%BEST_SCORE)) THEN
            ALLOCATE(SELF%BEST_SCORE(1))
            ALLOCATE(SELF%SIMPLEX_SIZE(1))
            SELF%BEST_SCORE(1) = INITIAL_SCORE
            SELF%SIMPLEX_SIZE  = 0.0
        ENDIF

        !
        IF (PRESENT(MAX_NUM_ITER)) SELF%MAX_NUM_ITER = MAX_NUM_ITER

        ! Set the IN_MEMORY flag
        SELF%IN_MEMORY = .TRUE.
    END SUBROUTINE FGSL_DS_INIT

    !>  \brief  Set the starting point and first step for a minimiser object
    SUBROUTINE FGSL_DS_SET_START_STEP(SELF,STARTING_VALUES,FIRST_STEP)
        USE, INTRINSIC  ::  ISO_C_BINDING
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(MINIMISER_DS_FGSL),    INTENT(INOUT)   ::  SELF
        REAL(KIND=8),               INTENT(IN)      ::  STARTING_VALUES(:)
        REAL(KIND=8),               INTENT(IN)      ::  FIRST_STEP(:)
        ! Private variables
        INTEGER(FGSL_INT)           ::  STATUS
        TYPE(FGSL_VECTOR)           ::  STARTING_VALUES_VEC, FIRST_STEP_VEC
        REAL(KIND=FGSL_DOUBLE), TARGET      ::  STARTING_VALUES_FGSL(SELF%N)
        REAL(KIND=FGSL_DOUBLE), TARGET      ::  FIRST_STEP_FGSL(SELF%N)
        REAL    ::  STARTING_SCORE
        LOGICAL, PARAMETER :: DEBUG = .FALSE.
        ! Start work

        IF (DEBUG) THEN
            WRITE (*,*) '**debug(FGSL_DS_SET_START_STEP): starting values = ', STARTING_VALUES
            WRITE (*,*) '**debug(FGSL_DS_SET_START_STEP):      first step = ', FIRST_STEP
        ENDIF

        ! Deal with best score history array
        IF (ALLOCATED(SELF%BEST_SCORE)) THEN
            IF (DEBUG) WRITE(*,'(A)') '**debug(FGSL_DS_SET_START_STEP): best score array already allocated, will reset it'
            STARTING_SCORE = SELF%BEST_SCORE(1)
            IF (DEBUG) WRITE(*,'(A,F0.6)') '**debug(FGSL_DS_SET_START_STEP): starting score = ', STARTING_SCORE
            DEALLOCATE(SELF%BEST_SCORE)
            IF (ALLOCATED(SELF%SIMPLEX_SIZE)) DEALLOCATE(SELF%SIMPLEX_SIZE)
            ALLOCATE(SELF%BEST_SCORE(1))
            ALLOCATE(SELF%SIMPLEX_SIZE(1))
            SELF%BEST_SCORE(1) = STARTING_SCORE
            SELF%SIMPLEX_SIZE(1) = 0.0
        ENDIF

        ! Check sizes of start, step arrays
        IF ((SIZE(STARTING_VALUES) .NE. SELF%N) .OR. (SIZE(FIRST_STEP) .NE. SELF%N)) THEN
            WRITE(*,'(A,3(I0,A))') '**ERROR(FGSL_DS_SET_START_STEP): size of starting values (', SIZE(STARTING_VALUES), &
                                    ') and first step (', SIZE(FIRST_STEP), &
                                    ') arrays do not match the number of dimensions (', SELF%N, ')'
            CALL TERMINATE('Fatal error in FGSL_DS_SET_START_STEP')
        ENDIF

        ! Convert incoming starting values to FGSL types
        STARTING_VALUES_FGSL = STARTING_VALUES
        STARTING_VALUES_VEC = FGSL_VECTOR_INIT(1.0_FGSL_DOUBLE)
        STATUS = FGSL_VECTOR_ALIGN(STARTING_VALUES_FGSL,SELF%N_FGSL,STARTING_VALUES_VEC,SELF%N_FGSL,0_FGSL_SIZE_T,1_FGSL_SIZE_T)
        IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('Failed vector align in FGSL_DS_SET_START_STEP')
        ! Convert incoming first step to FGSL types
        FIRST_STEP_FGSL = FIRST_STEP
        FIRST_STEP_VEC = FGSL_VECTOR_INIT(1.0_FGSL_DOUBLE)
        STATUS = FGSL_VECTOR_ALIGN(FIRST_STEP_FGSL,SELF%N_FGSL,FIRST_STEP_VEC,SELF%N_FGSL,0_FGSL_SIZE_T,1_FGSL_SIZE_T)
        IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('Failed vector align in FGSL_DS_SET_START_STEP')
        ! Set the starting point
        !PRINT *, '**debug(FGSL_DS_SET_START_STEP): setting the starting point...'
        STATUS = FGSL_MULTIMIN_FMINIMIZER_SET(SELF%MINIMISER_FGSL,SELF%FUNCTION_FGSL,STARTING_VALUES_VEC,FIRST_STEP_VEC)    ! This is the line that needs to be called again on the same object to get a new, random first step direction.
        !PRINT *, '**debug(FGSL_DS_SET_START_STEP): setting the starting point... done'
        IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('Failed fminimizer_set in FGSL_DS_SET_START_STEP')
        ! Free up FGSL-versions of the starting value and first step
        CALL FGSL_VECTOR_FREE(STARTING_VALUES_VEC)
        CALL FGSL_VECTOR_FREE(FIRST_STEP_VEC)
    END SUBROUTINE FGSL_DS_SET_START_STEP

    !>  \brief  Update the current best parameters in the minimiser object
    SUBROUTINE FGSL_DS_UPDATE_CURRENT_BEST(SELF)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(MINIMISER_DS_FGSL),    INTENT(INOUT)   ::  SELF
        ! Private variables
        TYPE(FGSL_VECTOR) :: VALUES_VEC
        REAL(FGSL_DOUBLE)   ::  MIN
        INTEGER(FGSL_INT) :: STATUS
        LOGICAL,    PARAMETER   ::  DEBUG = .FALSE.
        CHARACTER(LEN=6)    ::  N_C
        ! Start work
        VALUES_VEC = FGSL_MULTIMIN_FMINIMIZER_X(SELF%MINIMISER_FGSL)
        STATUS = FGSL_VECTOR_ALIGN(SELF%CURRENT_BEST_FGSL,VALUES_VEC)
        IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('Failed vector align for current_best in FGSL_DS_UPDATE_CURRENT_BEST')

        MIN = FGSL_MULTIMIN_FMINIMIZER_MINIMUM(SELF%MINIMISER_FGSL)
        SELF%CURRENT_MINIMUM_FGSL = MIN

        IF (DEBUG) THEN
            WRITE(N_C,'(I0)') SELF%N
            PRINT '(A'//N_C//'(F0.3,1X))', '**debug(FGSL_DS_UPDATE_CURRENT_BEST): updating best values to ', SELF%CURRENT_BEST_FGSL
            PRINT '(A,F0.6)', '**debug(FGSL_DS_UPDATE_CURRENT_BEST): updating best score to ', SELF%CURRENT_MINIMUM_FGSL
        ENDIF
    END SUBROUTINE FGSL_DS_UPDATE_CURRENT_BEST

    !>  \brief Deallocate memory taken up by the minimiser
    SUBROUTINE FGSL_DS_FREE(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(MINIMISER_DS_FGSL),    INTENT(INOUT)   ::  SELF
        ! Private variables

        ! Start work
        NULLIFY(SELF%FUNCTION)
        NULLIFY(SELF%CURRENT_BEST_FGSL)
        CALL FGSL_MULTIMIN_FMINIMIZER_FREE(SELF%MINIMISER_FGSL)
        CALL FGSL_MULTIMIN_FUNCTION_FREE(SELF%FUNCTION_FGSL)
        !
        SELF%IN_MEMORY = .FALSE.
    END SUBROUTINE FGSL_DS_FREE



    !>  \brief  Run the brute force minimiser's exhaustive search (NOT CODED)
    SUBROUTINE BF_RUN(SELF)
        USE RUNTIME_PARAMETERS
        use DatesAndTimes
        USE, INTRINSIC :: ISO_C_BINDING
        IMPLICIT NONE
        ! Arguments
        TYPE(MINIMISER_BF),     TARGET, INTENT(INOUT)   ::  SELF    !<  Brute-force minimiser (must be initialised)
        ! Private variables
        INTEGER                     ::  IERR, I
        REAL,   TARGET, ALLOCATABLE ::  CURRENT_VALUES(:)
        REAL,           ALLOCATABLE ::  BEST_VALUES(:)
        LOGICAL                     ::  DIMENSION_AT_MAX(SELF%N)
        LOGICAL                     ::  SEARCH_COMPLETED
        INTEGER                     ::  CURRENT_ITERATION
        REAL                        ::  SECONDS_SINCE_START, SECONDS_PER_ITERATION, SECONDS_TO_FINISH
        INTEGER                     ::  CLOCK_COUNT_AT_START, CLOCK_COUNT_CURRENT,CLOCK_COUNTS_PER_SECOND
        LOGICAL,    PARAMETER       ::  OUTPUT_AT_EVERY_ITERATION   =   .TRUE.
        REAL                        ::  CURRENT_SCORE, BEST_SCORE
        TYPE(C_PTR)                 ::  CURRENT_VALUES_CPTR
        TYPE(C_PTR)                 ::  SELF_CPTR
        type(DateAndTime)           ::  myTime
        ! Start work

        ! Check the minimiser is ready
        IF (.NOT. SELF%IN_MEMORY) THEN
            WRITE(*,'(A)') '**ERROR(BF_RUN): minimiser was not initialised, call BF_INIT first!'
            CALL TERMINATE('Fatal error in BF_RUN')
        ENDIF

        ! Allocate memory for the current values
        ALLOCATE(CURRENT_VALUES(SELF%N),BEST_VALUES(SELF%N),STAT=IERR)
        IF (IERR .NE. 0) CALL TERMINATE('Fatal error in BF_RUN')

        ! Set up C pointer to the minimiser
        SELF_CPTR = C_LOC(SELF)

        ! Start work

        ! The time when we start
        CALL SYSTEM_CLOCK(CLOCK_COUNT_AT_START,CLOCK_COUNTS_PER_SECOND)

        ! The starting point for the brute-force search
        CURRENT_VALUES = SELF%STARTING_VALUE - SELF%HALF_RANGE

        ! First, score the current values
        CURRENT_VALUES_CPTR = C_LOC(CURRENT_VALUES)
        BEST_SCORE = SELF%FUNCTION(CURRENT_VALUES_CPTR,SELF_CPTR)
        WRITE(*,'(A,F6.3)') '**info(BF_RUN): starting score = ', CURRENT_SCORE

        ! Start the brute-force search iterations
        DIMENSION_AT_MAX = .FALSE.
        CURRENT_ITERATION = 1
        DO
            ! We're now starting a new iteration
            CURRENT_ITERATION = CURRENT_ITERATION + 1
            IF (OUTPUT_AT_EVERY_ITERATION) THEN
                ! Add a printout with a time estimate
                CALL SYSTEM_CLOCK(CLOCK_COUNT_CURRENT)
                SECONDS_SINCE_START = REAL(CLOCK_COUNT_CURRENT - CLOCK_COUNT_AT_START) / REAL(CLOCK_COUNTS_PER_SECOND)
                SECONDS_PER_ITERATION = SECONDS_SINCE_START / (CURRENT_ITERATION - 1)
                SECONDS_TO_FINISH = (SELF%NUM_ITERATIONS - CURRENT_ITERATION + 1) * SECONDS_PER_ITERATION
                WRITE(*,'(3A,I6,A,I6,A,F6.3,2A)') '**info(BF_RUN): ', myTime%DateAndTimeAsString(), ': starting iteration ', &
                                                    CURRENT_ITERATION, ' of ', SELF%NUM_ITERATIONS, '. Current best score : ',&
                                                     BEST_SCORE, '. Estimated time to finish: ', &
                                                     myTime%DurationAsString(SECONDS_TO_FINISH)
            ENDIF

            ! Try the current parameters by calling the scoring function
            CURRENT_VALUES_CPTR = C_LOC(CURRENT_VALUES)
            CURRENT_SCORE =  SELF%FUNCTION(CURRENT_VALUES_CPTR,SELF_CPTR)

            ! If the score is the best we've seen so far, remember the values and the score
            IF (CURRENT_SCORE .LT. BEST_SCORE) THEN
                BEST_SCORE = CURRENT_SCORE
                BEST_VALUES = CURRENT_VALUES
            ENDIF


            ! Do the increment
            ! First, we find the first dimension that hasn't reached the maximum
            SEARCH_COMPLETED = .TRUE.
            DO I=1,SELF%N
                ! If we haven't reached the max for this dimension, increment it, and reset all previous dimensions to their starting point
                IF (.NOT. DIMENSION_AT_MAX(I)) THEN
                    ! If we got here, it means our search is not over yet
                    SEARCH_COMPLETED = .FALSE.
                    ! Increment the ith dimension
                    CURRENT_VALUES(I) = CURRENT_VALUES(I) + SELF%STEP_SIZE(I)
                    ! Reset all previous dimensions to their starting values, and reset their DIMENSION_AT_MAX flags to .FALSE.
                    IF (I .GT. 1) THEN
                        CURRENT_VALUES(1:I-1) = SELF%STARTING_VALUE(1:I-1) - SELF%HALF_RANGE(1:I-1)
                        IF (ANY( .NOT. DIMENSION_AT_MAX(1:I-1))) CALL TERMINATE('Failed sanity check in BF_RUN')
                        DIMENSION_AT_MAX(1:I-1) = .FALSE.
                    ENDIF
                    ! If the ith dimension has reached or gone over its max, set it at the max, and set its logical flag DIMENSION_AT_MAX to .TRUE.
                    IF (CURRENT_VALUES(I) .GE. SELF%STARTING_VALUE(I) + SELF%HALF_RANGE(I)) THEN
                        CURRENT_VALUES(I) = SELF%STARTING_VALUE(I) + SELF%HALF_RANGE(I)
                        DIMENSION_AT_MAX(I) = .TRUE.
                    ENDIF
                ENDIF
            ENDDO
            IF (SEARCH_COMPLETED) EXIT
        ENDDO

        ! Search is done
        WRITE(*,'(2(A,I6))') '**info(BF_RUN): brute-force search completed in ', CURRENT_ITERATION, &
                            ' iterations (predicted number of iterations was ', SELF%NUM_ITERATIONS, ')'

    END SUBROUTINE BF_RUN


    !>  \brief  Initialise a brute-force minimiser
    SUBROUTINE BF_INIT(SELF,FUNCTION_TO_MINIMISE,NUM_DIM,STARTING_VALUE,HALF_RANGE,STEP_SIZE)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(MINIMISER_BF),     INTENT(INOUT)   ::  SELF                    !<  The brute-force minimiser to initialise / allocate
        PROCEDURE(FGSL_MINIMISABLE_FUNCTION)    ::  FUNCTION_TO_MINIMISE    !<  The function to minimise
        INTEGER,                INTENT(IN)      ::  NUM_DIM                 !<  Number of dimensions
        REAL, ALLOCATABLE,      INTENT(IN)      ::  STARTING_VALUE(:)       !<  Starting point for the minimiser
        REAL, ALLOCATABLE,      INTENT(IN)      ::  HALF_RANGE(:)           !<  Half-range to search over
        REAL, ALLOCATABLE,      INTENT(IN)      ::  STEP_SIZE(:)            !<  Step size of the exhaustive search
        ! Private variables
        INTEGER ::  IERR, I
        ! Start work

        ! If already allocated, deallocate
        IF (SELF%IN_MEMORY) THEN
            !CALL SELF%FREE()
            CALL BF_FREE(SELF)
        ENDIF

        ! Allocate memory
        SELF%N = NUM_DIM
        ALLOCATE(SELF%STARTING_VALUE(SELF%N),SELF%HALF_RANGE(SELF%N),SELF%STEP_SIZE(SELF%N),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I3)') '**ERROR(BF_INIT): memory allocation failed with error ', IERR
            CALL TERMINATE('Fatal error in BF_INIT')
        ENDIF

        ! Check the incoming starting values, half range and step size have expected dimensionality
        IF (SIZE(STARTING_VALUE) .NE. NUM_DIM .OR. SIZE(HALF_RANGE) .NE. NUM_DIM .OR. SIZE(STEP_SIZE) .NE. NUM_DIM) THEN
            WRITE(*,'(A,I5,A,3I5)') '**ERROR(BF_INIT): incoming arrays do not have expected size of ', NUM_DIM, ': ',       &
                                    SIZE(STARTING_VALUE), SIZE(HALF_RANGE), SIZE(STEP_SIZE)
            CALL TERMINATE('Fatal error in BF_INIT')
        ENDIF

        ! Copy starting values, half range and step size over
        SELF%STARTING_VALUE = STARTING_VALUE
        SELF%HALF_RANGE = HALF_RANGE
        SELF%STEP_SIZE = STEP_SIZE

        ! Work out how many iterations the exhaustive search will take
        ! We will start at lowest value for each dimension, then go in step sizes, until we get to upper limit.
        ! We will also do extra iterations for the starting value and the upper limit
        SELF%NUM_ITERATIONS = 2**SELF%N
        DO I=1,SELF%N
            SELF%NUM_ITERATIONS = SELF%NUM_ITERATIONS * FLOOR(SELF%HALF_RANGE(I)*2/SELF%STEP_SIZE(I))
        ENDDO
        WRITE(*,'(A,I8)') '**info(BF_INIT): predicted number of steps in the exhaustive search = ', SELF%NUM_ITERATIONS

        ! Set up pointer to function to minimise
        SELF%FUNCTION => FUNCTION_TO_MINIMISE

        ! Set the memory flag
        SELF%IN_MEMORY = .TRUE.
    END SUBROUTINE BF_INIT

    !>  \brief  Deallocate a brute-force minimiser
    SUBROUTINE BF_FREE(SELF)
        IMPLICIT NONE
        ! Argument
        CLASS(MINIMISER_BF),    INTENT(INOUT)   ::  SELF    !<  Brute-force minimiser to deallocate
        ! Private variables

        ! Start work
        IF (.NOT. SELF%IN_MEMORY) THEN
            WRITE(*,'(A)') '**ERROR(BF_FREE): cannot free a minimiser not in memory'
            STOP 'Fatal error in BF_FREE'
        ENDIF
        ! Deallocate component arrays
        DEALLOCATE(SELF%STARTING_VALUE,SELF%HALF_RANGE,SELF%STEP_SIZE)
        ! Reset pointer
        NULLIFY(SELF%FUNCTION)
        ! Reset values
        SELF%N = 0
        SELF%NUM_ITERATIONS = 0
        SELF%IN_MEMORY = .FALSE.
    END SUBROUTINE BF_FREE




END MODULE MINIMISERS
