MODULE MODES
    USE RUNTIME_PARAMETERS, ONLY : WORD_LEN, LINE_LEN
    USE RTPS
    IMPLICIT NONE
    PRIVATE

    !>  \brief  Definition of a subroutine which constitutes a frealix2 mode
    ABSTRACT INTERFACE
        SUBROUTINE MODE_SUBROUTINE()
        END SUBROUTINE MODE_SUBROUTINE
    END INTERFACE

    TYPE, PRIVATE :: RTP_REAL_P
        TYPE(RTP_REAL), POINTER ::  P
    END TYPE
    TYPE, PRIVATE :: RTP_INTG_P
        TYPE(RTP_INTG), POINTER ::  P
    END TYPE
    TYPE, PRIVATE :: RTP_LOGI_P
        TYPE(RTP_LOGI), POINTER ::  P
    END TYPE
    TYPE, PRIVATE :: RTP_CHAR_P
        TYPE(RTP_CHAR), POINTER ::  P
    END TYPE

    TYPE, PUBLIC :: MODE
        PRIVATE
        CHARACTER(LEN=WORD_LEN)                     ::  COMMAND
        PROCEDURE(MODE_SUBROUTINE), POINTER, NOPASS ::  SUBROUTINE                      !<  Subroutine to call when the mode is in effect
        CHARACTER(LEN=LINE_LEN)                     ::  SHORT_DESCRIPTION               !<  A short description of what the mode does
        ! Run-time parameters associated with this mode
        TYPE(RTP_REAL_P), ALLOCATABLE               ::  REAL_RTPS(:)                    !<  The runtime parameters needed to run the mode
        TYPE(RTP_INTG_P), ALLOCATABLE               ::  INTG_RTPS(:)
        TYPE(RTP_LOGI_P), ALLOCATABLE               ::  LOGI_RTPS(:)
        TYPE(RTP_CHAR_P), ALLOCATABLE               ::  CHAR_RTPS(:)
        INTEGER                                     ::  NUMBER_OF_RTPS      =   0       !<  The number of runtime parameters associated with this mode
        CONTAINS
            PROCEDURE   ::  INIT                        =>  MODE_INIT
            PROCEDURE   ::  ADD_RTP                     =>  MODE_ADD_RTP
            PROCEDURE   ::  PRINT_INFO                  =>  MODE_PRINT_INFO
            PROCEDURE   ::  PRINT_WELCOME               =>  MODE_PRINT_WELCOME
            PROCEDURE   ::  PRINT_COMMAND_FILE          =>  MODE_PRINT_COMMAND_FILE
            PROCEDURE   ::  GET_RTP_VALUES_FROM_USER    =>  MODE_GET_RTP_VALUES_FROM_USER
            PROCEDURE   ::  RUN                         =>  MODE_RUN
    END TYPE MODE


CONTAINS

    !>  \brief  Run a frealix mode
    SUBROUTINE MODE_RUN(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(MODE),        INTENT(INOUT)      ::  SELF
        ! Private variables
        ! Start work
        WRITE(*,'(A)') '<mode command='//TRIM(ADJUSTL(SELF%COMMAND))//'>'
        ! Get RTP values from user
        CALL SELF%GET_RTP_VALUES_FROM_USER()
        ! Print welcome message
        CALL SELF%PRINT_WELCOME()
        ! Run the actual routine
        CALL SELF%SUBROUTINE()
        WRITE(*,'(A)') '</mode>'
    END SUBROUTINE MODE_RUN


    !>  \brief  Initialise a frealix mode
    SUBROUTINE MODE_INIT(SELF,COMMAND,SUBROUTINE,SHORT_DESCRIPTION)
        IMPLICIT NONE
        ! Arguments
        CLASS(MODE),        INTENT(INOUT)   ::  SELF
        CHARACTER(LEN=*),   INTENT(IN)      ::  COMMAND
        PROCEDURE(MODE_SUBROUTINE)          ::  SUBROUTINE
        CHARACTER(LEN=*),   INTENT(IN)      ::  SHORT_DESCRIPTION
        ! Private variables

        ! Start work
        SELF%COMMAND            =   COMMAND
        SELF%SHORT_DESCRIPTION  =   SHORT_DESCRIPTION
        SELF%SUBROUTINE         =>  SUBROUTINE
        SELF%NUMBER_OF_RTPS     =   0
        IF (ALLOCATED(SELF%REAL_RTPS)) DEALLOCATE(SELF%REAL_RTPS)
    END SUBROUTINE MODE_INIT


    !>  \brief  Get all values for RTPs from user
    SUBROUTINE MODE_GET_RTP_VALUES_FROM_USER(SELF)
        IMPLICIT NONE
        ! Argument
        CLASS(MODE),        INTENT(INOUT)      ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_RTP, RTP_TYPE, RTP_INDEX
        ! Start work
        DO CURRENT_RTP=1,SELF%NUMBER_OF_RTPS
            ! We need to find the RTP with order value CURRENT_RTP
            CALL MODE_FIND_RTP_FROM_ORDER(SELF,CURRENT_RTP,RTP_TYPE,RTP_INDEX)
            !
            SELECT CASE (RTP_TYPE)
                CASE (1)
                    CALL SELF%REAL_RTPS(RTP_INDEX)%P%GET_VALUE_FROM_USER()
                CASE (2)
                    CALL SELF%INTG_RTPS(RTP_INDEX)%P%GET_VALUE_FROM_USER()
                CASE (3)
                    CALL SELF%LOGI_RTPS(RTP_INDEX)%P%GET_VALUE_FROM_USER()
                CASE (4)
                    CALL SELF%CHAR_RTPS(RTP_INDEX)%P%GET_VALUE_FROM_USER()
            END SELECT
        ENDDO
    END SUBROUTINE MODE_GET_RTP_VALUES_FROM_USER

    !>  \brief  Print welcome message for the mode
    SUBROUTINE MODE_PRINT_WELCOME(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(MODE),        INTENT(IN)      ::  SELF
        ! Private variables

        ! Start work
        WRITE(*,'(3A)') 'Frealix is running in its ', TRIM(ADJUSTL(SELF%COMMAND)), ' mode'
        CALL MODE_PRINT_INFO(SELF)
        ! Flush standard output
        FLUSH(6) ! This is a standard F2003 statement
    END SUBROUTINE MODE_PRINT_WELCOME

    !>  \brief  Print information about the mode
    SUBROUTINE MODE_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(MODE),        INTENT(IN)      ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_RTP, RTP_TYPE, RTP_INDEX
        ! Start work
        !WRITE(*,'(3A)') '** ', TRIM(ADJUSTL(SELF%COMMAND)) ,' mode information **'
        !WRITE(*,'(A)') ' '
        WRITE(*,'(2A)') 'Mode description: ', TRIM(ADJUSTL(SELF%SHORT_DESCRIPTION))
        WRITE(*,'(A)') 'Run-time parameters:'

        CALL RTP_PRINT_INFO_HEADER()

        ! Print information about the mode's RTPs (loop over them)
        DO CURRENT_RTP=1,SELF%NUMBER_OF_RTPS
            ! We need to find the RTP with order value CURRENT_RTP
            CALL MODE_FIND_RTP_FROM_ORDER(SELF,CURRENT_RTP,RTP_TYPE,RTP_INDEX)
            !
            SELECT CASE (RTP_TYPE)
                CASE (1)
                    CALL SELF%REAL_RTPS(RTP_INDEX)%P%PRINT_INFO()
                CASE (2)
                    CALL SELF%INTG_RTPS(RTP_INDEX)%P%PRINT_INFO()
                CASE (3)
                    CALL SELF%LOGI_RTPS(RTP_INDEX)%P%PRINT_INFO()
                CASE (4)
                    CALL SELF%CHAR_RTPS(RTP_INDEX)%P%PRINT_INFO()
            END SELECT
        ENDDO

        WRITE(*,'(A)') ' '
    END SUBROUTINE MODE_PRINT_INFO

    !>  \brief  Print out a command file for the mode
    SUBROUTINE MODE_PRINT_COMMAND_FILE(SELF,COMMAND_FILENAME)
        USE RUNTIME_PARAMETERS, ONLY : AVAILABLE_LUN, LUN_IS_NOW_AVAILABLE, FILENAME_LEN, TERMINATE
        use DatesAndTimes
        IMPLICIT NONE
        ! Arguments
        CLASS(MODE),                    INTENT(IN)  ::  SELF
        CHARACTER(LEN=*),               INTENT(IN)  ::  COMMAND_FILENAME
        ! Private variables
        INTEGER ::  CURRENT_RTP, RTP_TYPE, RTP_INDEX
        INTEGER ::  LUN
        CHARACTER(LEN=32)   ::  FREALIX_MODE
        INTEGER ::  IOS
        type(DateAndTime)   ::  myTime
        ! Start work
        ! Get logical unit number
        LUN = AVAILABLE_LUN()
        ! Open file
        OPEN(UNIT=LUN,FILE=TRIM(ADJUSTL(COMMAND_FILENAME)),STATUS='REPLACE',IOSTAT=IOS)
        IF (IOS .NE. 0) THEN
            WRITE(*,'(3A,I0)') '**ERROR(MODE_PRINT_COMMAND_FILE): file opening failed for file ', &
                                    TRIM(ADJUSTL(COMMAND_FILENAME)), ' with error ', IOS
            CALL TERMINATE('Fatal error in MODE_PRINT_COMMAND_FILE')
        ENDIF
        FREALIX_MODE = 'FREALIX_MODE'
        WRITE(LUN,'(A)') '!'
        WRITE(LUN,'(2A)') '! Example frealix command file, generated ', myTime%DateAndTimeAsString()
        WRITE(LUN,'(A)') '!'
        WRITE(LUN,'(A32,A)') ADJUSTL(FREALIX_MODE), SELF%COMMAND
        WRITE(LUN,'(A)') '!'
        ! Print information about the mode's RTPs (loop over them)
        DO CURRENT_RTP=1,SELF%NUMBER_OF_RTPS
            ! We need to find the RTP with order value CURRENT_RTP
            CALL MODE_FIND_RTP_FROM_ORDER(SELF,CURRENT_RTP,RTP_TYPE,RTP_INDEX)
            !
            SELECT CASE (RTP_TYPE)
                CASE (1)
                    CALL SELF%REAL_RTPS(RTP_INDEX)%P%PRINT_TO_COMMAND_FILE(LUN)
                CASE (2)
                    CALL SELF%INTG_RTPS(RTP_INDEX)%P%PRINT_TO_COMMAND_FILE(LUN)
                CASE (3)
                    CALL SELF%LOGI_RTPS(RTP_INDEX)%P%PRINT_TO_COMMAND_FILE(LUN)
                CASE (4)
                    CALL SELF%CHAR_RTPS(RTP_INDEX)%P%PRINT_TO_COMMAND_FILE(LUN)
            END SELECT
        ENDDO

        CALL LUN_IS_NOW_AVAILABLE(LUN)
    END SUBROUTINE MODE_PRINT_COMMAND_FILE

    !>  \brief  Given the order index for an RTP, find it
    SUBROUTINE MODE_FIND_RTP_FROM_ORDER(SELF,ORDER,RTP_TYPE,RTP_INDEX)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(MODE),     INTENT(IN)  ::  SELF
        INTEGER,        INTENT(IN)  ::  ORDER       !<  The order value for the RTP we're looking for
        INTEGER,        INTENT(OUT) ::  RTP_TYPE    !<  Return 1, 2, 3, or 4 to indicate that the RTP is real, integer, logical or character
        INTEGER,        INTENT(OUT) ::  RTP_INDEX   !<  Return the index of the RTP within the appropriate array
        ! Private variables
        LOGICAL ::  FOUND_RTP
        INTEGER ::  I
        INTEGER ::  SUM_OF_SIZES
        ! Start work
        FOUND_RTP = .FALSE.
        ! No point looking if the order value is larger than the total number of RTPS
        IF (ORDER .GT. SELF%NUMBER_OF_RTPS) THEN
            WRITE(*,'(A,I0,A,I0,A)')    '**ERROR(MODE_FIND_RTP_FROM_ORDER): you asked to lookup RTP #', ORDER, &
                                        ', but there are only ', SELF%NUMBER_OF_RTPS, ' RTPS in this mode'
            CALL TERMINATE('Fatal error in MODE_FIND_RTP_FROM_ORDER. Programming error.')
        ENDIF
        ! Check the declared number of RTPS matches the total of the RTP array sizes
        SUM_OF_SIZES = 0
        IF (ALLOCATED(SELF%REAL_RTPS)) SUM_OF_SIZES = SUM_OF_SIZES + SIZE(SELF%REAL_RTPS)
        IF (ALLOCATED(SELF%INTG_RTPS)) SUM_OF_SIZES = SUM_OF_SIZES + SIZE(SELF%INTG_RTPS)
        IF (ALLOCATED(SELF%LOGI_RTPS)) SUM_OF_SIZES = SUM_OF_SIZES + SIZE(SELF%LOGI_RTPS)
        IF (ALLOCATED(SELF%CHAR_RTPS)) SUM_OF_SIZES = SUM_OF_SIZES + SIZE(SELF%CHAR_RTPS)
        IF (SUM_OF_SIZES .NE. SELF%NUMBER_OF_RTPS) THEN
            WRITE(*,'(A,I0,A,I0)')  '**ERROR(MODE_FIND_RTP_FROM_ORDER): there are ', SELF%NUMBER_OF_RTPS,    &
                                    ' declared RTPS in the mode, but the RTP arrays have a total size of ', SUM_OF_SIZES
            CALL TERMINATE('Fatal error in MODE_FIND_RTP_FROM_ORDER. Programming error.')
        ENDIF

        IF (ALLOCATED(SELF%REAL_RTPS)) THEN
            ! First we try the real RTPS
            DO I=1,SIZE(SELF%REAL_RTPS)
                IF (SELF%REAL_RTPS(I)%P%ORDER .EQ. ORDER) THEN
                    FOUND_RTP = .TRUE.
                    RTP_TYPE = 1
                    RTP_INDEX = I
                    RETURN
                ENDIF
            ENDDO
        ENDIF
        IF (ALLOCATED(SELF%INTG_RTPS)) THEN
            ! Next, we try integers
            DO I=1,SIZE(SELF%INTG_RTPS)
                IF (SELF%INTG_RTPS(I)%P%ORDER .EQ. ORDER) THEN
                    FOUND_RTP = .TRUE.
                    RTP_TYPE = 2
                    RTP_INDEX = I
                    RETURN
                ENDIF
            ENDDO
        ENDIF
        IF (ALLOCATED(SELF%LOGI_RTPS)) THEN
            ! Next, we try logicals
            DO I=1,SIZE(SELF%LOGI_RTPS)
                IF (SELF%LOGI_RTPS(I)%P%ORDER .EQ. ORDER) THEN
                    FOUND_RTP = .TRUE.
                    RTP_TYPE = 3
                    RTP_INDEX = I
                    RETURN
                ENDIF
            ENDDO
        ENDIF
        IF (ALLOCATED(SELF%CHAR_RTPS)) THEN
            ! Next, we try characters
            DO I=1,SIZE(SELF%CHAR_RTPS)
                IF (SELF%CHAR_RTPS(I)%P%ORDER .EQ. ORDER) THEN
                    FOUND_RTP = .TRUE.
                    RTP_TYPE = 4
                    RTP_INDEX = I
                    RETURN
                ENDIF
            ENDDO
        ENDIF

        IF (.NOT. FOUND_RTP) THEN
            WRITE(*,'(A,I0,A,I0,A)') '**ERROR(MODE_FIND_RTP_FROM_ORDER): could not find RTP #', ORDER, &
                                    '. There are ', SELF%NUMBER_OF_RTPS, ' RTPS in this mode.'
            CALL TERMINATE('Fatal error in MODE_FIND_RTP_FROM_ORDER')
        ENDIF

    END SUBROUTINE MODE_FIND_RTP_FROM_ORDER

    !>  \brief  Add an RTP to the mode
    SUBROUTINE MODE_ADD_RTP(SELF,NEW_RTP)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        CLASS(MODE),                    INTENT(INOUT)   ::  SELF
        CLASS(RTP),     TARGET,         INTENT(IN)      ::  NEW_RTP
        ! Private variables
        LOGICAL,    PARAMETER    :: DEBUG   =   .FALSE.
        ! Start work
        SELECT TYPE(NEW_RTP)
            TYPE IS (RTP_REAL)
                ! Grow array
                IF (ALLOCATED(SELF%REAL_RTPS)) THEN
                    SELF%REAL_RTPS = RESHAPE(SELF%REAL_RTPS,SHAPE=[SIZE(SELF%REAL_RTPS)+1],PAD=SELF%REAL_RTPS)
                ELSE
                    ALLOCATE(SELF%REAL_RTPS(1))
                ENDIF
                ! Copy RTP over
                SELF%REAL_RTPS(SIZE(SELF%REAL_RTPS))%P => NEW_RTP
                ! Keep track of the order RTPs were added
                SELF%REAL_RTPS(SIZE(SELF%REAL_RTPS))%P%ORDER = SELF%NUMBER_OF_RTPS + 1
            TYPE IS (RTP_INTG)
                ! Grow array
                IF (ALLOCATED(SELF%INTG_RTPS)) THEN
                    SELF%INTG_RTPS = RESHAPE(SELF%INTG_RTPS,SHAPE=[SIZE(SELF%INTG_RTPS)+1],PAD=SELF%INTG_RTPS)
                ELSE
                    ALLOCATE(SELF%INTG_RTPS(1))
                ENDIF
                ! Copy RTP over
                SELF%INTG_RTPS(SIZE(SELF%INTG_RTPS))%P => NEW_RTP
                ! Keep track of the order RTPs were added
                SELF%INTG_RTPS(SIZE(SELF%INTG_RTPS))%P%ORDER = SELF%NUMBER_OF_RTPS + 1
                IF (DEBUG) THEN
                    WRITE(*,'(A,3I0)') '**debug(MODE_ADD_RTP): added intg rtp: ', SELF%INTG_RTPS(SIZE(SELF%INTG_RTPS))%P%ORDER, &
                                                                                  SELF%NUMBER_OF_RTPS, SIZE(SELF%INTG_RTPS)
                ENDIF
            TYPE IS (RTP_LOGI)
                ! Grow array
                IF (ALLOCATED(SELF%LOGI_RTPS)) THEN
                    SELF%LOGI_RTPS = RESHAPE(SELF%LOGI_RTPS,SHAPE=[SIZE(SELF%LOGI_RTPS)+1],PAD=SELF%LOGI_RTPS)
                ELSE
                    ALLOCATE(SELF%LOGI_RTPS(1))
                ENDIF
                ! Copy RTP over
                SELF%LOGI_RTPS(SIZE(SELF%LOGI_RTPS))%P => NEW_RTP
                ! Keep track of the order RTPs were added
                SELF%LOGI_RTPS(SIZE(SELF%LOGI_RTPS))%P%ORDER = SELF%NUMBER_OF_RTPS + 1
            TYPE IS (RTP_CHAR)
                ! Grow array
                IF (ALLOCATED(SELF%CHAR_RTPS)) THEN
                    SELF%CHAR_RTPS = RESHAPE(SELF%CHAR_RTPS,SHAPE=[SIZE(SELF%CHAR_RTPS)+1],PAD=SELF%CHAR_RTPS)
                ELSE
                    ALLOCATE(SELF%CHAR_RTPS(1))
                ENDIF
                ! Copy RTP over
                SELF%CHAR_RTPS(SIZE(SELF%CHAR_RTPS))%P => NEW_RTP
                ! Keep track of the order RTPs were added
                SELF%CHAR_RTPS(SIZE(SELF%CHAR_RTPS))%P%ORDER = SELF%NUMBER_OF_RTPS + 1
            CLASS DEFAULT
                CALL TERMINATE('RTP type not implemented')
        END SELECT
        SELF%NUMBER_OF_RTPS = SELF%NUMBER_OF_RTPS + 1
    END SUBROUTINE MODE_ADD_RTP


    !
    ! Routines below probably not useful
    !


    !>  \brief  Initialise an RTP and add it to the mode
    SUBROUTINE MODE_INIT_AND_ADD_REAL_RTP(SELF,DEFAULT_VALUE,KEYWORD,DESCRIPTION,HELP)
        IMPLICIT NONE
        ! Arguments
        CLASS(MODE),                    INTENT(INOUT)   ::  SELF
        REAL,                           INTENT(IN)      ::  DEFAULT_VALUE   !<  Default value for the parameter
        CHARACTER(LEN=*),               INTENT(IN)      ::  KEYWORD         !<  Keyword used in frealix command files to identify the parameter
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  DESCRIPTION     !<  Description used in the question and as comment in frealix command files
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  HELP            !<  Help to be given to user to describe what the parameter means / does
        ! Private variables
        TYPE(RTP_REAL)  ::  RTP
        ! Start work
        CALL RTP%INIT(DEFAULT_VALUE,KEYWORD,DESCRIPTION=DESCRIPTION,HELP=HELP)
        CALL MODE_ADD_RTP(SELF,RTP)
    END SUBROUTINE MODE_INIT_AND_ADD_REAL_RTP

    !>  \brief  Initialise an RTP and add it to the mode
    SUBROUTINE MODE_INIT_AND_ADD_INTG_RTP(SELF,DEFAULT_VALUE,KEYWORD,DESCRIPTION,HELP)
        IMPLICIT NONE
        ! Arguments
        CLASS(MODE),                    INTENT(INOUT)   ::  SELF
        INTEGER,                        INTENT(IN)      ::  DEFAULT_VALUE   !<  Default value for the parameter
        CHARACTER(LEN=*),               INTENT(IN)      ::  KEYWORD         !<  Keyword used in frealix command files to identify the parameter
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  DESCRIPTION     !<  Description used in the question and as comment in frealix command files
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  HELP            !<  Help to be given to user to describe what the parameter means / does
        ! Private variables
        TYPE(RTP_INTG)  ::  RTP
        ! Start work
        CALL RTP%INIT(DEFAULT_VALUE,KEYWORD,DESCRIPTION=DESCRIPTION,HELP=HELP)
        CALL MODE_ADD_RTP(SELF,RTP)
    END SUBROUTINE MODE_INIT_AND_ADD_INTG_RTP

    !>  \brief  Initialise an RTP and add it to the mode
    SUBROUTINE MODE_INIT_AND_ADD_CHAR_RTP(SELF,DEFAULT_VALUE,KEYWORD,DESCRIPTION,HELP)
        IMPLICIT NONE
        ! Arguments
        CLASS(MODE),                    INTENT(INOUT)   ::  SELF
        CHARACTER(LEN=*),               INTENT(IN)      ::  DEFAULT_VALUE   !<  Default value for the parameter
        CHARACTER(LEN=*),               INTENT(IN)      ::  KEYWORD         !<  Keyword used in frealix command files to identify the parameter
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  DESCRIPTION     !<  Description used in the question and as comment in frealix command files
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  HELP            !<  Help to be given to user to describe what the parameter means / does
        ! Private variables
        TYPE(RTP_CHAR)  ::  RTP
        ! Start work
        CALL RTP%INIT(DEFAULT_VALUE,KEYWORD,DESCRIPTION=DESCRIPTION,HELP=HELP)
        CALL MODE_ADD_RTP(SELF,RTP)
    END SUBROUTINE MODE_INIT_AND_ADD_CHAR_RTP

    !>  \brief  Initialise an RTP and add it to the mode
    SUBROUTINE MODE_INIT_AND_ADD_LOGI_RTP(SELF,DEFAULT_VALUE,KEYWORD,DESCRIPTION,HELP)
        IMPLICIT NONE
        ! Arguments
        CLASS(MODE),                    INTENT(INOUT)   ::  SELF
        LOGICAL,                        INTENT(IN)      ::  DEFAULT_VALUE   !<  Default value for the parameter
        CHARACTER(LEN=*),               INTENT(IN)      ::  KEYWORD         !<  Keyword used in frealix command files to identify the parameter
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  DESCRIPTION     !<  Description used in the question and as comment in frealix command files
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  HELP            !<  Help to be given to user to describe what the parameter means / does
        ! Private variables
        TYPE(RTP_LOGI)  ::  RTP
        ! Start work
        CALL RTP%INIT(DEFAULT_VALUE,KEYWORD,DESCRIPTION=DESCRIPTION,HELP=HELP)
        CALL MODE_ADD_RTP(SELF,RTP)
    END SUBROUTINE MODE_INIT_AND_ADD_LOGI_RTP



END MODULE MODES
