MODULE MODE_AMYLOID_LAYER_LINE
    ! Modules
    USE MODES
    USE RTPS

    IMPLICIT NONE
    TYPE(MODE) :: AMYLOID_LAYER_LINE

    TYPE(RTP_CHAR), TARGET  ::  INPUT_STACK_FILENAME  =               &
        RTP_CHAR(   VALUE='input_stack.mrc',                  &
                    KEYWORD='INPUT_STACK_FILENAME',           &
                    DESCRIPTION='Input stack of filament segments')


    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_STACK_FILENAME  =               &
        RTP_CHAR(   VALUE='output_stack.mrc',                  &
                    KEYWORD='OUTPUT_STACK_FILENAME',           &
                    DESCRIPTION='Stack of filtered amplitude spectra')


    TYPE(RTP_REAL), TARGET  ::  PIXEL_SIZE  =               &
        RTP_REAL(   VALUE=1.17,                  &
                    KEYWORD='PIXEL_SIZE',           &
                    DESCRIPTION='Pixel size in Angstroms')

CONTAINS

SUBROUTINE MODE_AMYLOID_LAYER_LINE_INIT_AND_RUN
    IMPLICIT NONE
    ! Intialise the mode
    CALL AMYLOID_LAYER_LINE%INIT(COMMAND='AMYLOID_LAYER_LINE',SUBROUTINE=MODE_AMYLOID_LAYER_LINE_SLAVE, &
                                SHORT_DESCRIPTION='Given a stack of filament segments, compute amplitude spectra')

    ! Add RTPs to the mode
    CALL AMYLOID_LAYER_LINE%ADD_RTP(INPUT_STACK_FILENAME)
    CALL AMYLOID_LAYER_LINE%ADD_RTP(OUTPUT_STACK_FILENAME)
    CALL AMYLOID_LAYER_LINE%ADD_RTP(PIXEL_SIZE)

    ! Run the mode
    CALL AMYLOID_LAYER_LINE%RUN()
END SUBROUTINE MODE_AMYLOID_LAYER_LINE_INIT_AND_RUN

SUBROUTINE MODE_AMYLOID_LAYER_LINE_SLAVE
    USE RUNTIME_PARAMETERS
    use Images
    USE IMAGES_HELICES, ONLY : SCAN_FOR_MERIDONIAL_LAYER_LINE
    USE PARSERS, ONLY : FILENAME_ADD_SUFFIX
    use ImageFiles
    IMPLICIT NONE
    ! Variables
    TYPE(IMAGE)     ::  IMG, IMG_AMPL, IMG_PHAS, AMPL_AVG
    TYPE(IMAGEFILE) ::  IMGFILE_IN,IMGFILE_OUT
    INTEGER ::  STACK_SIZE
    INTEGER ::  I, INPUT_DIM(3)
    INTEGER, ALLOCATABLE     :: GOOD_LOCS(:)
    INTEGER ::  CURRENT_OUTPUT_LOC
    REAL,   ALLOCATABLE     ::  SCORES_TMP(:), FREQS_TMP(:), PDIFFS_TMP(:)
    REAL,   ALLOCATABLE     ::  SCORES_ALL(:), FREQS_ALL(:), PDIFFS_ALL(:)
    CHARACTER(LEN=FILENAME_LEN) :: AMPL_AVE_FILENAME


    GOOD_LOCS = [473,474,475,483,486,490,491,494,504,506,516,517,529,582,678,745,1433,1434,1435,1438]

    CALL IMGFILE_IN%INIT(INPUT_STACK_FILENAME%VALUE)
    CALL IMGFILE_IN%PrintInfo()

    ! What is the stack size?
    STACK_SIZE = IMGFILE_IN%GetStackSize()

    ! Initialise an output imagefile with the same dimensions
    !CALL IMGFILE_OUT%INIT(OUTPUT_STACK_FILENAME%VALUE,'OUT',MOULD=IMGFILE_IN)
    INPUT_DIM = IMGFILE_IN%GetDimensions()
    CALL IMGFILE_OUT%INIT(OUTPUT_STACK_FILENAME%VALUE,DIM_X=INPUT_DIM(1),DIM_Y=INPUT_DIM(2),DIM_Z=1)
    CALL IMGFILE_OUT%PrintInfo()

	! Initialise an image to hold the average amplitude spectrum
	AMPL_AVG%logical_dimensions = IMGFILE_IN%GetDimensions()
	AMPL_AVG%logical_dimensions(3) = 1
	CALL IMAGE_ALLOCATE(AMPL_AVG)

	! Allocate memory for the results
	ALLOCATE(SCORES_ALL(STACK_SIZE),FREQS_ALL(STACK_SIZE),PDIFFS_ALL(STACK_SIZE))

    CURRENT_OUTPUT_LOC = 0
    DO I=1,STACK_SIZE
        IF (.TRUE.) THEN !ANY(GOOD_LOCS .EQ. I)) THEN
            WRITE(*,'(A,I0,A,I0)') '**info(MODE_AMYLOID_LAYER_LINE): working on location ', I, ' of ', STACK_SIZE

            CURRENT_OUTPUT_LOC = CURRENT_OUTPUT_LOC + 1

            ! Read the image data from disk to image object
            call img%ReadFromDisk(imgfile_in,wanted_location=i)

            ! Compute the amplitudes & phases
            CALL IMG%ComputeAmplitudes(IMG_AMPL)
            CALL IMG%ComputeAmplitudes(IMG_PHAS,PHASES=.TRUE.)

            ! High-pass filter the amplitude spectrum
            CALL IMG_AMPL%ApplyBandPassFilter(1./60.,0.0,COSINE_FALLOFF=.05)

            CALL SCAN_FOR_MERIDONIAL_LAYER_LINE(    IMG_AMPL,IMG_PHAS,I,PIXEL_SIZE%VALUE/4.8,    &
                                                    (PIXEL_SIZE%VALUE/4.0)-(PIXEL_SIZE%VALUE/4.8),  &
                                                    SCORES_TMP,FREQS_TMP,PDIFFS_TMP)

            WRITE(*,'(A,I0,A,F0.5,A,F0.5,A,A,F0.2)') '**info(MODE_AMYLOID_LAYER_LINE): for location ', I, &
                                                ' the prime candidate for a layer line (score = ', SCORES_TMP(1), &
                                                ') occured at frequency of ', PIXEL_SIZE%VALUE/FREQS_TMP(1), ' Angstroems', &
                                                ' with an average phase difference across the meridian of ', &
                                                PDIFFS_TMP(1)
            ! Remember the scores
            SCORES_ALL(I) = SCORES_TMP(1)
            FREQS_ALL(I)  = PIXEL_SIZE%VALUE/FREQS_TMP(1)
            PDIFFS_ALL(I) = PDIFFS_TMP(1)

            ! Beam stop
            CALL IMG_AMPL%ApplyFilter(DISC_INVERSE,10.0)

			! Accumulate average of amplitude spectra
			AMPL_AVG%real_values = AMPL_AVG%real_values + IMG_AMPL%real_values

			! Write out the filtered amplitude spectrum for the current image
            call img_ampl%WriteToDisk(imgfile_out,wanted_location=CURRENT_OUTPUT_LOC)
        ENDIF
    ENDDO


    CALL IMGFILE_OUT%PrintInfo()

    CALL IMGFILE_IN%CLOSE()
    CALL IMGFILE_OUT%CLOSE()

    ! Write out the average amplitude spectrum
    AMPL_AVG%real_values = AMPL_AVG%real_values / STACK_SIZE
    CALL SCAN_FOR_MERIDONIAL_LAYER_LINE(    IMG_AMPL,IMG_PHAS,I,PIXEL_SIZE%VALUE/4.59,    &
                                            (PIXEL_SIZE%VALUE/4.45)-(PIXEL_SIZE%VALUE/4.59),  &
                                            SCORES_TMP,FREQS_TMP,PDIFFS_TMP)

    WRITE(*,'(A,A,F0.5,A,F0.5,A)')  '**info(MODE_AMYLOID_LAYER_LINE): for the average spectrum, ', &
                                    ' the prime candidate for a layer line (score = ', SCORES_TMP(1), &
                                    ') occured at frequency of ', PIXEL_SIZE%VALUE/FREQS_TMP(1), ' Angstroems'
    CALL FILENAME_ADD_SUFFIX(OUTPUT_STACK_FILENAME%VALUE,'_ave',AMPL_AVE_FILENAME)
    CALL AMPL_AVG%WriteToDisk(AMPL_AVE_FILENAME,PIXEL_SIZE=PIXEL_SIZE%VALUE)

    ! Write out results
    WRITE(*,'(A)') '**info(MODE_AMYLOID_LAYER_LINE): summary of results'
    WRITE(*,'(A)') 'Location in stack,Score,Layer line spacing (A),Phase difference (degrees)'
    DO I=1,STACK_SIZE
        WRITE(*,'(I0,A,F0.4,A,F0.2,A,F0.2)') I, ',', SCORES_ALL(I), ',',FREQS_ALL(I),',',PDIFFS_ALL(I)
    ENDDO
    WRITE(*,'(/A)') '**info(MODE_AMYLOID_LAYER_LINE): all done'


END SUBROUTINE MODE_AMYLOID_LAYER_LINE_SLAVE

END MODULE MODE_AMYLOID_LAYER_LINE
