!>  \brief  Mode to bin or unbin images, volumes, frealix parameter files or boxer coordinates files.
MODULE MODE_BIN
    USE MODES
    USE RTPS
    IMPLICIT NONE
    ! Declare mode object
    TYPE(MODE)  ::  BIN
    ! Declare RTPs
    TYPE(RTP_CHAR), TARGET  ::  INPUT_FILENAME  =       &
        RTP_CHAR(   VALUE='input.mrc',                  &
                    KEYWORD='INPUT_FILENAME',           &
                    DESCRIPTION='Name of input file, to be binned. Can be image or frealix alignment file')

    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_FILENAME =       &
        RTP_CHAR(   VALUE='output.mrc',                 &
                    KEYWORD='OUTPUT_FILENAME',          &
                    DESCRIPTION='Name of output file')

    TYPE(RTP_INTG), TARGET  ::  BINNING_FACTOR  =       &
        RTP_INTG(   VALUE=4,                            &
                    KEYWORD='BINNING_FACTOR',           &
                    DESCRIPTION='Images or alignment parameters will be binned/decimated/coarsened '//  &
                    'by this factor in each dimension')

    TYPE(RTP_LOGI), TARGET  ::  UNBIN           =       &
        RTP_LOGI(   VALUE=.FALSE.,                      &
                    KEYWORD='UNBIN',                    &
                    DESCRIPTION='If true, the images or parameters will be unbinned, not binned')

    TYPE(RTP_LOGI), TARGET  ::  PRESERVE_CENTER  =      &
        RTP_LOGI(   VALUE=.TRUE.,                       &
                    KEYWORD='PRESERVE_CENTER',          &
                    DESCRIPTION='Ensure that the origin is still centered. Actual binning '//   &
                                '(real-space averaging of neighbouring pixels) would not do this.')
CONTAINS

SUBROUTINE MODE_BIN_INIT_AND_RUN
    IMPLICIT NONE
    ! Initialise the mode
    CALL BIN%INIT(  COMMAND='BIN',              &
                    SUBROUTINE=MODE_BIN_SLAVE,  &
                    SHORT_DESCRIPTION='Bin or unbin images, volumes, frealix alignment files or boxer coordinates files')
    ! Add RTPs to the mode
    CALL BIN%ADD_RTP(INPUT_FILENAME)
    CALL BIN%ADD_RTP(OUTPUT_FILENAME)
    CALL BIN%ADD_RTP(BINNING_FACTOR)
    CALL BIN%ADD_RTP(UNBIN)
    CALL BIN%ADD_RTP(PRESERVE_CENTER)
    ! Run the mode
    CALL BIN%RUN()
END SUBROUTINE MODE_BIN_INIT_AND_RUN

SUBROUTINE MODE_BIN_SLAVE
    USE IMAGES
    USE IMAGEFILES
    USE PARSERS, ONLY : IMAGE_FILENAME, EXTENSION_FROM_FILENAME
    USE FREALIX_PARAMETERS
    USE COOS
    USE UI, ONLY : INTERVAL_PRINT
    IMPLICIT NONE
    ! Variables
    TYPE(IMAGE)     ::  VOLIN,VOLOUT
    TYPE(BOXER_DB)  ::  DBIN,DBOUT
    TYPE(IMAGEFILE) ::  IMGFILE, IMGFILEOUT
    INTEGER         ::  CURRENT_LOC, STACK_SIZE
    REAL            ::  PIXEL_SIZE_OLD, PIXEL_SIZE_NEW

    ! Start work
    ! If the input filename has an image extension (mrc,spi,hed or img), we treat it as an image/volume
    IF (IMAGE_FILENAME(INPUT_FILENAME%VALUE)) THEN
        WRITE(*,'(A)') '**info(MODE_BIN): input file is image file'
        CALL IMGFILE%INIT(INPUT_FILENAME%VALUE)
        PIXEL_SIZE_OLD = IMGFILE%GetPixelSize()
        IF (UNBIN%VALUE) THEN
            PIXEL_SIZE_NEW = PIXEL_SIZE_OLD / BINNING_FACTOR%VALUE
        ELSE
            PIXEL_SIZE_NEW = PIXEL_SIZE_OLD * BINNING_FACTOR%VALUE
        ENDIF
        CALL IMGFILE%PrintInfo()
        CALL IMGFILEOUT%INIT(OUTPUT_FILENAME%VALUE,PIXEL_SIZE=PIXEL_SIZE_NEW)
        STACK_SIZE = IMGFILE%GetStackSize()
        IF (IMGFILE%GetDimension(2) .eq. IMGFILE%GetDimension(1) .and. IMGFILE%GetDimension(1) .eq. IMGFILE%GetDimension(3)) then
            STACK_SIZE = 1
        ENDIF
        DO CURRENT_LOC = 1,STACK_SIZE
            IF (INTERVAL_PRINT(CURRENT_LOC,STACK_SIZE)) THEN
                WRITE(*,'(2(A,I0))') '**info(MODE_BIN): working on location ', &
                                        CURRENT_LOC, ' of ', STACK_SIZE
            ENDIF
            CALL VOLIN%ReadFromDisk(IMGFILE,wanted_location=CURRENT_LOC,read_volume_not_slice=(STACK_SIZE.EQ.1))
            CALL BIN(VOLIN,VOLOUT,BINNING_FACTOR%VALUE,UNBIN=UNBIN%VALUE,PRESERVE_CENTER=PRESERVE_CENTER%VALUE)
            CALL VOLOUT%WriteToDisk(IMGFILEOUT,wanted_location=CURRENT_LOC)
        ENDDO
        CALL IMGFILE%CLOSE()
        CALL IMGFILEOUT%PrintInfo()
        CALL IMGFILEOUT%CLOSE()
    ! Maybe the input file was a boxer "database" (list of coordinates)
    ELSE IF (EXTENSION_FROM_FILENAME(INPUT_FILENAME%VALUE) .EQ. 'box') THEN
        WRITE(*,'(A)') '**info(MODE_BIN): input file is Boxer coordinates file'
        CALL BOXER_DB_READ_FROM_DISK(DBIN,INPUT_FILENAME%VALUE)
        CALL BOXER_DB_BIN(DBIN,DBOUT,BINNING_FACTOR%VALUE,UNBIN=UNBIN%VALUE)
        CALL BOXER_DB_WRITE_TO_DISK(DBOUT,OUTPUT_FILENAME%VALUE)
    ! Otherwise, we treat the file as a parameter file
    ELSE
        WRITE(*,'(A)') '**info(MODE_BIN): input file is frealix alignment file'
        CALL FREALIX2_PAR_BIN(INPUT_FILENAME%VALUE,OUTPUT_FILENAME%VALUE,BINNING_FACTOR%VALUE,UNBIN=UNBIN%VALUE)
    ENDIF

END SUBROUTINE MODE_BIN_SLAVE

END MODULE MODE_BIN
