!>  \brief  Combine 3D reconstructions obtained from subset of the input data into a single 3D reconstruction.
!!          A set of 3D reconstructions and a set of 3D weights are required as input.
!!
!!  Note: this mode uses the FLIP option of FWD_FT/BWD_FT, which is being depreciated. Come up with a workaround.
MODULE MODE_COMBINE_3DS
    USE MODES
    USE RTPS
    IMPLICIT NONE
    PRIVATE
    PUBLIC :: MODE_COMBINE_3DS_INIT_AND_RUN

    ! Declare run-time parameters
    TYPE(RTP_CHAR), TARGET  ::  REC_LIST_FILE  =               &
        RTP_CHAR(   VALUE='default_3Drec_list',                  &
                    KEYWORD='REC_LIST_FILE',           &
                    DESCRIPTION='Name of file containing list of 3D reconstructions to combine')


    TYPE(RTP_CHAR), TARGET  ::  WGT_LIST_FILE  =               &
        RTP_CHAR(   VALUE='default_3Dwgt_list',                  &
                    KEYWORD='WGT_LIST_FILE',           &
                    DESCRIPTION='List of 3D weights to use for combination')


    TYPE(RTP_CHAR), TARGET  ::  WGT_PARAM_FILE  =               &
        RTP_CHAR(   VALUE='default_3Dwgt_param',                  &
                    KEYWORD='WGT_PARAM_FILE',           &
                    DESCRIPTION='File containing weight parameters from frealix')


    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_REC_FILE  =               &
        RTP_CHAR(   VALUE='default_output_rec_file',                  &
                    KEYWORD='OUTPUT_REC_FILE',           &
                    DESCRIPTION='Output 3D reconstruction')


    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_WGT_FILE  =               &
        RTP_CHAR(   VALUE='default_output_wgt_file',                  &
                    KEYWORD='OUTPUT_WGT_FILE',           &
                    DESCRIPTION='Output weights file')


    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_CENTRAL_SLICE_FILE  =               &
        RTP_CHAR(   VALUE='default_output_central_slice_file',                  &
                    KEYWORD='OUTPUT_CENTRAL_SLICE_FILE',           &
                    DESCRIPTION='Output file containing a stack of 2D images, ' //&
                                'each being the central slice from one of the input 3Ds')


    TYPE(RTP_INTG), TARGET  ::  SYM_REDUN  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='SYM_REDUN',           &
                    DESCRIPTION='The symmetry redundancy, e.g. 1 for C1, 12 for D6')


    TYPE(RTP_LOGI), TARGET  ::  CALC_FSC  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='CALC_FSC',           &
                    DESCRIPTION='')


    TYPE(RTP_LOGI), TARGET  ::  MASK_WITH_CYLINDER  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='MASK_WITH_CYLINDER',           &
                    DESCRIPTION='Whether to mask output 3D with a cylinder')


    TYPE(RTP_REAL), TARGET  ::  CYLINDER_RADIUS  =               &
        RTP_REAL(   VALUE=1000.0,                  &
                    KEYWORD='CYLINDER_RADIUS',           &
                    DESCRIPTION='Radius of cylinder (pixels)')


    TYPE(RTP_INTG), TARGET  ::  CYLINDER_SOFT  =               &
        RTP_INTG(   VALUE=5,                  &
                    KEYWORD='CYLINDER_SOFT',           &
                    DESCRIPTION='Width of soft edge of cylinder')
CONTAINS

SUBROUTINE MODE_COMBINE_3DS_INIT_AND_RUN()
    IMPLICIT NONE
    TYPE(MODE) :: COMBINE_3DS

    ! Intialise the mode
    CALL COMBINE_3DS%INIT(COMMAND='COMBINE_3DS',SUBROUTINE=MODE_COMBINE_3DS_SLAVE,SHORT_DESCRIPTION='Add description here')

    ! Add RTPs to the mode
    CALL COMBINE_3DS%ADD_RTP(REC_LIST_FILE)
    CALL COMBINE_3DS%ADD_RTP(WGT_LIST_FILE)
    CALL COMBINE_3DS%ADD_RTP(WGT_PARAM_FILE)
    CALL COMBINE_3DS%ADD_RTP(OUTPUT_REC_FILE)
    CALL COMBINE_3DS%ADD_RTP(OUTPUT_WGT_FILE)
    CALL COMBINE_3DS%ADD_RTP(OUTPUT_CENTRAL_SLICE_FILE)
    CALL COMBINE_3DS%ADD_RTP(SYM_REDUN)
    CALL COMBINE_3DS%ADD_RTP(CALC_FSC)
    CALL COMBINE_3DS%ADD_RTP(MASK_WITH_CYLINDER)
    CALL COMBINE_3DS%ADD_RTP(CYLINDER_RADIUS)
    CALL COMBINE_3DS%ADD_RTP(CYLINDER_SOFT)

    ! Run the mode
    CALL COMBINE_3DS%RUN()
END SUBROUTINE MODE_COMBINE_3DS_INIT_AND_RUN

SUBROUTINE MODE_COMBINE_3DS_SLAVE
    USE RUNTIME_PARAMETERS      ! This gives us access to user-defined or default values for runtime parameters
    USE IMAGES               ! This defines the 'Image' derived type
    USE IMAGEFILES              ! To read/write images to/from disk
    USE IMAGES_ARITHM            ! This defines arithmetics on Image objects
    USE IMAGES_COMPARISON        ! This contains useful FSC-related functions
    USE PARSERS                 ! This contains the parser for lists of files
    USE FSCS
    IMPLICIT NONE
    ! Variables
    TYPE(IMAGE) ::  REC_IN, REC_OUT(2), WGT_IN, WGT_OUT(2)                  ! If the user asked for FSC, we'll need two 3Ds
    TYPE(IMAGE) ::  REC_OUT_CROP(2)
    INTEGER     ::  CROP_DIM(3)
    CHARACTER(LEN=FILENAME_LEN),    ALLOCATABLE ::  REC_FILES(:)            ! Array of filenames for the 3D reconstructions
    CHARACTER(LEN=FILENAME_LEN),    ALLOCATABLE ::  WGT_FILES(:)            ! Array of filenames for the 3D weights
    INTEGER,                        ALLOCATABLE ::  NUM_CONTRIB_ELEMS(:)    ! Number of elements contributing to each 3D rec.
    INTEGER :: I,J
    INTEGER ::  IFIRST,ILAST
    INTEGER :: NUMBER_OF_RECS                   ! The number of reconstructions to be combined
    INTEGER :: NUMBER_OF_WGTS                   ! The number of weights files
    REAL(KIND=8)    ::  CTF_VAR                 ! The CTF variance from all incoming slices (sum of CTF**2 / num. contrib. elems)
    REAL(KIND=8)    ::  WC2_FUDGE, SYM_REDUN_D  ! Real(kind=8) is the same as Double Precision
    INTEGER(KIND=8) :: TOTAL_NUM_CONTRIB_ELEMS(2)
    REAL    :: SINCLUT(2000)                    ! Sinc function look-up table
    CHARACTER(LEN=FILENAME_LEN) ::  FSC_VOL1, FSC_VOL2
    LOGICAL ::  MIRROR_ARRAY(3) ! To indicate which direction we'll be mirroring
    INTEGER ::  PAR_LUN !<  Logical unit number to write to

    INTEGER ::  NUM_COMBINED_3DS
    REAL, ALLOCATABLE   ::  SHELLCORR(:)
    INTEGER ::  ISPLIT
    TYPE(IMAGE) ::  CENTRAL_SLICE

    ! Set some parameter values
!   PARAMETER   (WC2_FUDGE  =   0.1D0)
    PARAMETER   (WC2_FUDGE  =   0.1000000015)

    ! Get runtime parameters from control file


    ! Get an available LUN
    PAR_LUN = AVAILABLE_LUN()

    ! This will be used in double arithmetics later
    SYM_REDUN_D = DBLE(SYM_REDUN%VALUE)

    ! Read & parse file list for 3D reconstructions
    CALL PARSE_FILELIST_FILE(REC_LIST_FILE%VALUE,REC_FILES)
    NUMBER_OF_RECS = SIZE(REC_FILES)
    WRITE(*,'(A,I3,A)') 'COMBINE_3DS: ',NUMBER_OF_RECS,' 3D reconstructions will be processed'

    ! Read & parse file list for 3D weights.
    CALL PARSE_FILELIST_FILE(WGT_LIST_FILE%VALUE,WGT_FILES)
    NUMBER_OF_WGTS = SIZE(WGT_FILES)


    ! Crash out if we don't have same number of 3Ds and weights.
    IF ( NUMBER_OF_RECS .NE. NUMBER_OF_WGTS ) THEN
        WRITE(*,'(5A,I0,A,I0)')     '**ERROR(COMBINE_3DS: The files ', TRIM(REC_LIST_FILE%VALUE), &
                                    ' and ', TRIM(WGT_LIST_FILE%VALUE),   &
                                    ' have different numbers of files listed: ', NUMBER_OF_RECS,' & ', NUMBER_OF_WGTS
        CALL TERMINATE('Fatal error in COMBINE_3DS')
    ENDIF


    ! Get the number of elements contributing to each reconstruction
    CALL GET_INTS_FROM_COLUMN(WGT_PARAM_FILE%VALUE,2,NUM_CONTRIB_ELEMS)

    ! Crash out if we don't have the same number of WGT params as reconstructions
    IF ( SIZE(NUM_CONTRIB_ELEMS) .NE. NUMBER_OF_RECS ) THEN
        WRITE(*,'(5A,I0,A,I0)') '**ERROR(COMBINE_3DS): The files ',                     &
                                        TRIM(REC_LIST_FILE%VALUE), ' and ', TRIM(WGT_PARAM_FILE%VALUE),     &
                                        ' have different numbers of files listed: ',            &
                                        NUMBER_OF_RECS,' & ', SIZE(NUM_CONTRIB_ELEMS)
        CALL TERMINATE('Fatal error in COMBINE_3DS')
    ENDIF


    ! If the user asked for an FSC, there will be 2 combined volumes
    IF (CALC_FSC%VALUE) THEN
        NUM_COMBINED_3DS = 2
    ELSE
        NUM_COMBINED_3DS = 1
    ENDIF

    ! Prepare the output volumes. They need to be same dim as input. To find out input dim, we read from disk.
    call rec_in%ReadFromDisk(REC_FILES(1))
    call wgt_in%ReadFromDisk(WGT_FILES(1))
    IF (WGT_IN%logical_dimensions(1) .NE. HIGHEST_INDEX_DIM1(REC_IN%logical_dimensions(1))) THEN
        WRITE(*,'(A,2(3(I0,1X),3X))')    '**ERROR(COMBINE_3DS): I expected the weights DIM1 to be half of rec. DIM1, but instead:', &
                                        'WGT_IN%logical_dimensions, REC_IN%logical_dimensions =', &
                                        WGT_IN%logical_dimensions, REC_IN%logical_dimensions
        CALL TERMINATE('Error in COMBINE_3DS')
    ENDIF
    REC_OUT%is_in_real_space = .false.

    ! Allocate memory for the output volumes and initialise them
    DO J = 1, NUM_COMBINED_3DS
        REC_OUT(J)%logical_dimensions = REC_IN%logical_dimensions
        WGT_OUT(J)%logical_dimensions = WGT_IN%logical_dimensions
        CALL IMAGE_ALLOCATE(REC_OUT(J))
        CALL IMAGE_ALLOCATE(WGT_OUT(J))
        REC_OUT(J)%complex_values = (0.0,0.0)
        WGT_OUT(J)%real_values = 0.0
    ENDDO

    ! Allocate memory for central slice image
    CENTRAL_SLICE%logical_dimensions(1:2) = REC_IN%logical_dimensions(1)
    CALL IMAGE_ALLOCATE(CENTRAL_SLICE)

    ! Prepare the Sinc look-up table
    CALL CALCSINC(SINCLUT,2000)

    ! We will need to mirror weights volumes in the X direction
    MIRROR_ARRAY = (/.TRUE.,.FALSE.,.FALSE./)

    ! Work out total number of contributing elements
    TOTAL_NUM_CONTRIB_ELEMS = 0
    DO I=1,NUMBER_OF_RECS
        TOTAL_NUM_CONTRIB_ELEMS(1) = TOTAL_NUM_CONTRIB_ELEMS(1) + NUM_CONTRIB_ELEMS(I)
    ENDDO
    ! Work out how to split the reconstructions in two so that each
    ! has a similar number of contributing elements
    IF (CALC_FSC%VALUE) THEN
        DO I=1,NUMBER_OF_RECS
            TOTAL_NUM_CONTRIB_ELEMS(2) = TOTAL_NUM_CONTRIB_ELEMS(2) + NUM_CONTRIB_ELEMS(I)
            IF ( TOTAL_NUM_CONTRIB_ELEMS(2) .GE. TOTAL_NUM_CONTRIB_ELEMS(1)/2) THEN
                ISPLIT = I
                EXIT
            ENDIF
        ENDDO
        TOTAL_NUM_CONTRIB_ELEMS(1) = TOTAL_NUM_CONTRIB_ELEMS(1) - TOTAL_NUM_CONTRIB_ELEMS(2)
        WRITE(*,'(A,I3,2(1X,I12))') '**debug(COMBINE_3DS): isplit, elems1, elems2: ', ISPLIT, &
                                    TOTAL_NUM_CONTRIB_ELEMS(1), TOTAL_NUM_CONTRIB_ELEMS(2)
    ENDIF


    ! Initialise the count of contrib elements
    TOTAL_NUM_CONTRIB_ELEMS(1:2) = 0

    ! Now start work
    DO J = 1, NUM_COMBINED_3DS
        PRINT '(A,I2,A,I2)', '**debug(COMBINE_3DS): processing out 3D ', J, ' of ', NUM_COMBINED_3DS
        IF (J == 1) THEN
            IF (CALC_FSC%VALUE) THEN
                IFIRST = 1
                ILAST  = ISPLIT
            ELSE
                IFIRST = 1
                ILAST  = NUMBER_OF_RECS
            ENDIF
        ELSE
            IFIRST = ILAST + 1
            ILAST  = NUMBER_OF_RECS
        ENDIF
        DO I = IFIRST, ILAST
            WRITE(*,'(A,I3,A,I10,A)') '**debug(COMBINE_3DS): processing 3D rec ', I, ' with ', &
                                        NUM_CONTRIB_ELEMS(I), ' contrib. elements'
            IF (NUM_CONTRIB_ELEMS(I) == 0) CYCLE        ! This can happen if a process only had images with bad scores which never went to 3D
            ! Read current 3D / WGT from disk
            call rec_in%ReadFromDisk(rec_files(i))
            call wgt_in%ReadFromDisk(wgt_files(i))
            ! Dump central slice of input 3D to a diagnostic output file
            CENTRAL_SLICE%real_values(1:REC_IN%logical_dimensions(1),1:REC_IN%logical_dimensions(2),1) = &
                                    REC_IN%real_values( 1:REC_IN%logical_dimensions(1),  &
                                                        1:REC_IN%logical_dimensions(2), &
                                                        REC_IN%logical_dimensions(3)/2+1)
            call CENTRAL_SLICE%WriteToDisk(OUTPUT_CENTRAL_SLICE_FILE%VALUE,wanted_location=I)
            ! The weights need to be mirrored about the X axis
            CALL MIRROR_IMAGE(WGT_IN,MIRROR_ARRAY)
            ! Cancel the effect of Frealix's CORRECT3D
            !CALL CORRECT3D(REC_IN,SINCLUT,REVERSE=.TRUE.)
            ! The 3D needs to be FT'ed
            CALL TERMINATE('Broken - need to add step to shift origin')
            !CALL FWD_FT(REC_IN,FLIP=.TRUE.)
            CALL FWD_FT(REC_IN)
            ! The output weights is just the sum of the inputs
            CALL ADD_IMAGES(WGT_OUT(J),WGT_IN,WGT_OUT(J))
            ! Keep track of the number of contributing elements
            TOTAL_NUM_CONTRIB_ELEMS(J) = TOTAL_NUM_CONTRIB_ELEMS(J) + NUM_CONTRIB_ELEMS(I)
            ! The CTF variance (sum of squares) for the current input 3D
            CTF_VAR = WGT_IN%SumOfVoxelValuesDouble() / DBLE(NUM_CONTRIB_ELEMS(I))
            ! The un-weighted reconstruction
            REC_OUT(J)%complex_values =     CMPLX(REC_OUT(J)%complex_values,KIND=8)   &
                     + ( CMPLX(REC_IN%complex_values,KIND=8) * ( DBLE(WGT_IN%real_values) + CTF_VAR * SYM_REDUN_D * WC2_FUDGE ) )
        ENDDO

        ! Debug dump
!       PRINT *, '**debug(COMBINE_3DS): dumping before weighting...'
!       WRITE(FSC_VOL1,'(A,I1,A)') 'junk_before_weight_',J,'.mrc'
!       CALL BWD_FT(REC_OUT(J),FLIP=.TRUE.)
!       CALL WRITE_TO_DISK(REC_OUT(J),FSC_VOL1)
!       CALL FWD_FT(REC_OUT(J),FLIP=.TRUE.)

        ! The CTF variance (sum of squares) for the full set
        CTF_VAR = WGT_OUT(J)%SumOfVoxelValuesDouble() / DBLE(TOTAL_NUM_CONTRIB_ELEMS(J))

        ! The final output is obtained by diving by the weights
        REC_OUT(J)%complex_values = CMPLX(REC_OUT(J)%complex_values,KIND=8) / &
                                    ( DBLE(WGT_OUT(J)%real_values) + CTF_VAR * SYM_REDUN_D * WC2_FUDGE )

        ! Reverse FT the ouput reconstruction
        CALL TERMINATE('Need to add shift origin')
        !CALL BWD_FT(REC_OUT(J),FLIP=.TRUE.)
        CALL BWD_FT(REC_OUT(J))

        ! Debug dump
!       PRINT *, '**debug(COMBINE_3DS): dumping...'
!       WRITE(FSC_VOL1,'(A,I1,A)') 'junk_before_correct3d_',J,'.mrc'
!       CALL WRITE_TO_DISK(REC_OUT(J),FSC_VOL1)

        ! Reapply the CORRECT3D correction for fall-off, unless we're goign to do an FSC
        IF (.NOT. CALC_FSC%VALUE) THEN
            CALL REC_OUT(J)%Correct3D(SINCLUT,REVERSE=.FALSE.)
        ENDIF

        ! Debug dump
!       PRINT *, '**debug(COMBINE_3DS): dumping...'
!       WRITE(FSC_VOL1,'(A,I1,A)') 'junk_',J,'.mrc'
!       CALL WRITE_TO_DISK(REC_OUT(J),FSC_VOL1)

        ! The output weights file needs to be mirrored back to original Frealix orientation
        CALL MIRROR_IMAGE(WGT_OUT(J),MIRROR_ARRAY)
    ENDDO


    ! If we wanted an FSC, let's do it. If possible/necessary we will first crop the reconstruction
    IF (CALC_FSC%VALUE) THEN
        ! We may need/want to crop.
        ! If so, we need to know the radius of the reconstruction

        !REC_OUT_CROP(1:2) = REC_OUT(1:2) !! THIS DOESN"T WORK PROPERLY (23-Oct-09) - can't find the bug
        DO I=1,2
            CALL IMAGE_ALLOCATE(REC_OUT_CROP(I),REC_OUT(I))
            REC_OUT_CROP(I)%real_values = REC_OUT(I)%real_values
        END DO
        IF (MASK_WITH_CYLINDER%VALUE) THEN
            IF (REC_OUT(1)%logical_dimensions(1) .GT. CYLINDER_RADIUS%VALUE * 3.0) THEN
                CROP_DIM = NINT(CYLINDER_RADIUS%VALUE * 3)
                IF (MOD(CROP_DIM(1),2) .NE. 0.0) CROP_DIM = CROP_DIM + 1
                WRITE(*,'(A,3(I4,1X))') '(COMBINE_3DS): cropping FSC volumes to new dimensions: ', CROP_DIM
                CALL CROP(REC_OUT_CROP(1),CROP_DIM)
                CALL CROP(REC_OUT_CROP(2),CROP_DIM)
            ENDIF
        ENDIF

        ! Save the volumes used for the FSC
        PRINT *, '**debug(COMBINE_3DS): dumped FSC volumes will not be CORRECT3Ded'
        FSC_VOL1 = TRIM(OUTPUT_REC_FILE%VALUE) // '_fsc1'
        FSC_VOL2 = TRIM(OUTPUT_REC_FILE%VALUE) // '_fsc2'
        call rec_out_crop(1)%WriteToDisk(FSC_VOL1)
        call rec_out_crop(2)%WriteToDisk(FSC_VOL2)
        CALL FWD_FT(REC_OUT_CROP(1))
        CALL FWD_FT(REC_OUT_CROP(2))
        CALL COMPUTE_FSC(REC_OUT_CROP(1),REC_OUT_CROP(2),SHELLCORR)
        WRITE(*,'(/A)') 'FSC between approx. half datasets'
        CALL PRINT_FSC(SHELLCORR)

        ! Now we are done with the FSC, we can sum the two volumes together to obtain the final output volume (see comments above for details of how we do this)
        DO I = 1, 2
            !CALL CORRECT3D(REC_OUT(I),SINCLUT,REVERSE=.TRUE.)
            CALL TERMINATE('Need to add shift origin')
            !CALL FWD_FT(REC_OUT(I),FLIP=.TRUE.)
            CALL FWD_FT(REC_OUT(I))
            ! unweigh the reconstructions
            CALL MIRROR_IMAGE(WGT_OUT(I),MIRROR_ARRAY)
            CTF_VAR = WGT_OUT(I)%SumOfVoxelValuesDouble() / DBLE(TOTAL_NUM_CONTRIB_ELEMS(I))
            REC_OUT(I)%complex_values = CMPLX(REC_OUT(I)%complex_values,KIND=8) *  &
                                ( DBLE(WGT_OUT(I)%real_values) + CTF_VAR * SYM_REDUN_D * WC2_FUDGE )
        ENDDO
        ! Final weight is sum of weights of both volumes
        CALL ADD_IMAGES(WGT_OUT(1),WGT_OUT(2),WGT_OUT(1))
        ! Final 3D is sum of unweighted 3Ds and will be re-weighted later
        CALL ADD_IMAGES(REC_OUT(1),REC_OUT(2),REC_OUT(1))
        ! Total number of elements
        TOTAL_NUM_CONTRIB_ELEMS(1) = TOTAL_NUM_CONTRIB_ELEMS(1) + TOTAL_NUM_CONTRIB_ELEMS(2)


        ! Now get ready to weigh the final reconstruction
        CTF_VAR     = WGT_OUT(1)%SumOfVoxelValuesDouble()     / DBLE(TOTAL_NUM_CONTRIB_ELEMS(1))
        REC_OUT(1)%complex_values      = CMPLX(REC_OUT(1)%complex_values,KIND=8) / &
                                ( DBLE(WGT_OUT(1)%real_values) + CTF_VAR * SYM_REDUN_D * WC2_FUDGE )
        CALL TERMINATE('Need to add shift origin')
        !CALL BWD_FT(REC_OUT(1),FLIP=.TRUE.)
        CALL BWD_FT(REC_OUT(1))
        CALL REC_OUT(1)%CORRECT3D(SINCLUT,REVERSE=.FALSE.)
        ! Mirror output weights back to original Frealix orientation
        CALL MIRROR_IMAGE(WGT_OUT(1),MIRROR_ARRAY)
    ENDIF

    IF ( MASK_WITH_CYLINDER%VALUE) THEN
        WRITE(*,'(A,F5.2,A,I3)') '**debug(COMBINE_3DS): masking with cylinder of radius ', &
                                    CYLINDER_RADIUS%VALUE, ' and soft edge of ', CYLINDER_SOFT%VALUE
        CALL CYL_MASK(REC_OUT(1),CYLINDER_RADIUS%VALUE,CYLINDER_SOFT%VALUE)
    ENDIF

    ! Write output volumes to disk
    call rec_out(1)%WriteToDisk(output_rec_file%value)
    call wgt_out(1)%WriteToDisk(output_wgt_file%value)


    WRITE(*,'(A,A)') 'COMBINE_3DS: All done. Output 3D reconstruction is ', TRIM(OUTPUT_REC_FILE%VALUE)

END SUBROUTINE MODE_COMBINE_3DS_SLAVE

END MODULE MODE_COMBINE_3DS
