!>  \brief  Estimate the persistence length by measuring mean tangent correlations
MODULE MODE_ESTIMATE_PERSISTENCE_LENGTH
    USE MODES
    USE RTPS
    IMPLICIT NONE

    TYPE(MODE)  ::  ESTIMATE_PL

    TYPE(RTP_CHAR), TARGET  ::  INPUT_FILENAME = &
                        RTP_CHAR(   VALUE = 'waypoints.ali',    &
                                    KEYWORD = 'INPUT_FILENAME', &
                                    DESCRIPTION = 'Filename for input waypoints')

    TYPE(RTP_REAL), TARGET  ::  PIXEL_SIZE = &
                        RTP_REAL(   VALUE = 0.0,    &
                                    KEYWORD = 'PIXEL_SIZE',  &
                                    DESCRIPTION = 'Pixel size (Angstroms)')

    TYPE(RTP_LOGI), TARGET  ::  IGNORE_FAL_WPS = &
                        RTP_LOGI(   VALUE = .FALSE.,    &
                                    KEYWORD = 'IGNORE_FAL_WPS',  &
                                    DESCRIPTION = 'Ignore the first and last waypoints of each filament')

    TYPE(RTP_REAL), TARGET  ::  MAXIMUM_FILAMENT_LENGTH = &
                        RTP_REAL(   VALUE = 0.0,    &
                                    KEYWORD = 'MAXIMUM_FILAMENT_LENGTH',  &
                                    DESCRIPTION = 'Ignore filamnents longer than this length (in Angstroms)')

CONTAINS

SUBROUTINE MODE_ESTIMATE_PL_INIT_AND_RUN
    IMPLICIT NONE
    ! Intialise the mode
    CALL ESTIMATE_PL%INIT(COMMAND='ESTIMATE_PERSISTENCE_LENGTH',SUBROUTINE=MODE_ESTIMATE_PL_SLAVE,  &
                                SHORT_DESCRIPTION='Estimate the persistence length by measuring tangent correlations')

    ! Add RTPs to the mode
    CALL ESTIMATE_PL%ADD_RTP(INPUT_FILENAME)
    CALL ESTIMATE_PL%ADD_RTP(PIXEL_SIZE)
    CALL ESTIMATE_PL%ADD_RTP(IGNORE_FAL_WPS)
    CALL ESTIMATE_PL%ADD_RTP(MAXIMUM_FILAMENT_LENGTH)

    ! Run the mode
    CALL ESTIMATE_PL%RUN()
END SUBROUTINE MODE_ESTIMATE_PL_INIT_AND_RUN

SUBROUTINE MODE_ESTIMATE_PL_SLAVE
    USE COOS, ONLY : WAYPOINTS_ADD_DISORDER
    USE FILAMENTS_BASE
    USE FILAMENTS_PATH, ONLY : FILAMENT_GET_PATH_PIXEL_STEPS
    USE FILAMENTS_GEOMETRY, ONLY : FILAMENT_ARC_LENGTH
    USE FILAMENTS_PARAMETRISATION, ONLY : FILAMENT_PARAMETRISATION_UPDATE
    USE FREALIX_PARAMETERS
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE
    ! Private variables
    TYPE(FILAMENT), ALLOCATABLE ::  FILS(:)
    TYPE(WAYPOINT), ALLOCATABLE ::  WP(:)
    INTEGER                     ::  CURRENT_FIL
    REAL                        ::  MAX_FIL_LENGTH
    REAL(KIND=8),   PARAMETER   ::  S_MIN = 200.0D0
    REAL                        ::  S_MAX
    INTEGER                     ::  NUM_INTERVALS
    REAL(KIND=8),   ALLOCATABLE ::  AVE(:),SQ_AVE(:)
    REAL,           ALLOCATABLE ::  S(:)
    INTEGER,        ALLOCATABLE ::  N(:), NF(:)
    INTEGER                     ::  I, CURRENT_WP,WP_FROM,WP_TO
    REAL(KIND=8)                ::  D
    ! Start work

    ! Parse the input waypoints file
    CALL FILAMENTS_FROM_FREALIX_PAR_NOINIT(FILS,INPUT_FILENAME%VALUE,PIXEL_SIZE%VALUE)

    ! Loop over filaments to initialise them and find the longest one
    MAX_FIL_LENGTH = 0.0
    DO CURRENT_FIL=1,SIZE(FILS)
        ! Initialise the filament
        WP = FILS(CURRENT_FIL)%WP
        CALL FILAMENT_ALLOCATE(FILS(CURRENT_FIL),SIZE(WP))
        FILS(CURRENT_FIL)%WP = WP
        CALL FILAMENT_PARAMETRISATION_UPDATE(FILS(CURRENT_FIL))
        ! Print out information for this filament
        CALL FILAMENT_PRINT_INFO(FILS(CURRENT_FIL))
        ! Is this the longest filament we've seen so far?
        IF (FILAMENT_ARC_LENGTH(FILS(CURRENT_FIL),IGNORE_FAL_WPS%VALUE) .GT. MAX_FIL_LENGTH) THEN
            MAX_FIL_LENGTH = FILAMENT_ARC_LENGTH(FILS(CURRENT_FIL),IGNORE_FAL_WPS%VALUE)
        ENDIF
    ENDDO
    WRITE(*,'(A,F0.1,A)') '**info(MODE_ESTIMATE_PERSISTENCE_LENGTH): longest filament is ', MAX_FIL_LENGTH , ' pixels'

    ! If we were given a maximum filament length by the user, heed it
    IF (MAXIMUM_FILAMENT_LENGTH%SET_BY_USER) THEN
        MAX_FIL_LENGTH = MIN(MAX_FIL_LENGTH,MAXIMUM_FILAMENT_LENGTH%VALUE/PIXEL_SIZE%VALUE)
        WRITE(*,'(A,F0.1,A)') '**info(MODE_ESTIMATE_PERSISTENCE_LENGTH): will only use filaments up to ', &
                                                                        MAX_FIL_LENGTH, ' pixels long'
    ENDIF

    ! Allocate arrays for our measurements
    S_MAX = AINT(MAX_FIL_LENGTH / S_MIN) * S_MIN
    NUM_INTERVALS = INT((S_MAX - S_MIN) / S_MIN) + 1
    ALLOCATE(S(NUM_INTERVALS),AVE(NUM_INTERVALS),SQ_AVE(NUM_INTERVALS),N(NUM_INTERVALS),NF(NUM_INTERVALS))
    WRITE(*,'(A,3(A,F0.1),A,I0,A)') '**info(MODE_ESTIMATE_PERSISTENCE_LENGTH): ', &
                                    'will measure tangent correlations with separations from ',&
                                    S_MIN,' to ', S_MAX , ' in ', S_MIN , ' increments (', &
                                    NUM_INTERVALS, ' interval lengths)'

    ! Initialise the arrays
    S      = 0.0 ! length of this interval (in pixels)
    AVE    = 0.0D0
    SQ_AVE = 0.0D0
    N      = 0  ! number of measurements that contributed to this interval's stats
    NF     = 0  ! number of filaments that contributed to this interval's stats

    ! Setup the S array - the length of each interval
    DO I=1,NUM_INTERVALS
        S(I) = S_MIN*I
    ENDDO

    ! Collect the data
    DO CURRENT_FIL=1,SIZE(FILS)
        IF (MAXIMUM_FILAMENT_LENGTH%SET_BY_USER) THEN
            IF (FILAMENT_ARC_LENGTH(FILS(CURRENT_FIL),IGNORE_FAL_WPS%VALUE) .GT. MAX_FIL_LENGTH) THEN
                WRITE(*,'(A,2(I0,A))') '**info(MODE_ESTIMATE_PERSISTENCE_LENGTH): skipping filament ', CURRENT_FIL, ' of ', &
                                                                                   SIZE(FILS), ' because it is too long'
                CYCLE
            ENDIF
        ENDIF
        WRITE(*,'(A,I0,A,I0)') '**info(MODE_ESTIMATE_PERSISTENCE_LENGTH): working on filament ', CURRENT_FIL, ' of ', SIZE(FILS)

        ! Get waypoints at regular intervals along the filament
        CALL FILAMENT_GET_PATH_PIXEL_STEPS( FILS(CURRENT_FIL),S_MIN,WP,  &
                                            IGNORE_FAL_WPS=IGNORE_FAL_WPS%VALUE,INCLUDE_FINISH_WP=.FALSE.)
        ! We should have at least 2 waypoints, if not something went wrong
        IF (SIZE(WP) .LT. 2) THEN
            WRITE(*,'(A)') '**ERROR(MODE_ESTIMATE_PERSISTENCE_LENGTH): less than 2 waypoints found'
            CALL TERMINATE('Fatal error in MODE_ESTIMATE_PERSISTENCE_LENGTH. Failed sanity check.')
        ENDIF

        ! This filament will contribute to all these interval's stats
        NF(1:SIZE(WP)-1) = NF(1:SIZE(WP)-1) + 1

        ! Loop over the intervals
        DO I=1,SIZE(WP)-1
            ! Loop over the waypoints
            DO CURRENT_WP=1,SIZE(WP)-I,I
                WP_FROM = CURRENT_WP
                WP_TO   = WP_FROM + I
                ! Compute the dot product
                D = DOT_PRODUCT(WP(WP_FROM)%DCOO,WP(WP_TO)%DCOO)
                ! Accumulate data
                N(I)      = N(I) + 1
                AVE(I)    = AVE(I) + D
                SQ_AVE(I) = SQ_AVE(I) + D**2
            ENDDO
        ENDDO

    ENDDO

    ! Print out results
    WRITE(*,'(A)') '**info(MODE_ESTIMATE_PERSISTENCE_LENGTH): tangent correlation statistics follow'
    WRITE(*,'(5(A17,1X))') 'Interval (pixels)', 'Num. measurements', 'Num. filaments', 'Sample mean', 'Sample variance'
    DO I=1,NUM_INTERVALS
        AVE(I)    = AVE(I) / N(I)
        SQ_AVE(I) = SQ_AVE(I) / N(I)
        WRITE(*,'(F17.1,1X,2(I17,1X),1X,2(G19.5,1X))') S(I), N(I), NF(I), AVE(I), SQ_AVE(I)-AVE(I)**2
    ENDDO

    ! All done
    WRITE(*,'(A)') '**info(MODE_ESTIMATE_PERSISTENCE_LENGTH): all done.'

END SUBROUTINE MODE_ESTIMATE_PL_SLAVE


END MODULE MODE_ESTIMATE_PERSISTENCE_LENGTH
