!>  \brief  Given a model 3D, refine way point parameters. The input and outputs are frealix2 parameter files.
!!
!!  Note: this mode can be used to just do a reconstruction without refinement, by setting the "refine?" flag in the input parameter file
!!  to F for each waypoint.
!!
MODULE MODE_FILAMENT_REFINE
!DEC$ OPTIMIZE:0
    USE MODES
    USE RTPS

    USE MINIMISERS, ONLY : DS_MAX_ITERATIONS


    IMPLICIT NONE
    TYPE(MODE) :: FILAMENT_REFINE

    TYPE(RTP_CHAR), TARGET  ::  INPUT_PAR_FILENAME  =               &
        RTP_CHAR(   VALUE='waypoints_old.ali',                  &
                    KEYWORD='INPUT_PAR_FILENAME',           &
                    DESCRIPTION='')


    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_PAR_FILENAME  =               &
        RTP_CHAR(   VALUE='waypoints.ali',                  &
                    KEYWORD='OUTPUT_PAR_FILENAME',           &
                    DESCRIPTION='')

    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_PAR_FILENAME_SUBUNITS =  &
        RTP_CHAR(   VALUE='none',                       &
                    KEYWORD='OUTPUT_PAR_FILENAME_SUBUNITS', &
                    DESCRIPTION='Unless ''none'', filename for output alignment parameter file, '//&
                                'with one line per helical asymetric unit')


    TYPE(RTP_CHAR), TARGET  ::  MODEL_3D_FILENAME  =               &
        RTP_CHAR(   VALUE='model3d.mrc',                  &
                    KEYWORD='MODEL_3D_FILENAME',           &
                    DESCRIPTION='3D reconstruction to be used as reference, or model, during refinement')


    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_3D_FILENAME  =               &
        RTP_CHAR(   VALUE='model3d_out.mrc',                  &
                    KEYWORD='OUTPUT_3D_FILENAME',           &
                    DESCRIPTION='Filename for new 3D reconstruction following refinement')


    TYPE(RTP_CHAR), TARGET  ::  CTFTILT_FILES  =               &
        RTP_CHAR(   VALUE='none',                  &
                    KEYWORD='CTFTILT_FILES',           &
                    DESCRIPTION='')


    TYPE(RTP_CHAR), TARGET  ::  CTFFIND_FILES  =               &
        RTP_CHAR(   VALUE='none',                  &
                    KEYWORD='CTFFIND_FILES',           &
                    DESCRIPTION='')


    TYPE(RTP_CHAR), TARGET  ::  FILM_FILES  =               &
        RTP_CHAR(   VALUE='film_????.mrc',                  &
                    KEYWORD='FILM_FILES',           &
                    DESCRIPTION='')


    TYPE(RTP_CHAR), TARGET  ::  MATCHING_PROJECTIONS_FILES  =               &
        RTP_CHAR(   VALUE='match_proj_??????.mrc',                  &
                    KEYWORD='MATCHING_PROJECTIONS_FILES',           &
                    DESCRIPTION='')


    TYPE(RTP_CHAR), TARGET  ::  FILAMENT_TRACE_FILES  =               &
        RTP_CHAR(   VALUE='trace_??????_??????.box',                  &
                    KEYWORD='FILAMENT_TRACE_FILES',           &
                    DESCRIPTION='')


    TYPE(RTP_REAL), TARGET  ::  PIXEL_SIZE  =               &
        RTP_REAL(   VALUE=9.99,                  &
                    KEYWORD='PIXEL_SIZE',           &
                    DESCRIPTION='')


    TYPE(RTP_REAL), TARGET  ::  TWIST_PER_SUBUNIT  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='TWIST_PER_SUBUNIT',           &
                    DESCRIPTION='In degrees')


    TYPE(RTP_REAL), TARGET  ::  RISE_PER_SUBUNIT  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='RISE_PER_SUBUNIT',           &
                    DESCRIPTION='In Angstroems')


    TYPE(RTP_INTG), TARGET  ::  SYMMETRY_AXIAL  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='SYMMETRY_AXIAL',           &
                    DESCRIPTION='Rotational symmetry redundancy about the helical axis')


    TYPE(RTP_INTG), TARGET  ::  SYMMETRY_PERP  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='SYMMETRY_PERP',           &
                    DESCRIPTION='Rotational symmetry redundancy perpendicular to the helical axis (e.g. 2 for DNA)')


    TYPE(RTP_LOGI), TARGET  ::  WPS_ARE_CROSSOVERS  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='WPS_ARE_CROSSOVERS',           &
                    DESCRIPTION='The waypoints are crossovers (typically of an amyloid beta fibril or actin filament)')


    TYPE(RTP_LOGI), TARGET  ::  FAL_WPS_ARE_CROSSOVERS  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='FAL_WPS_ARE_CROSSOVERS',           &
                    DESCRIPTION='The first and last waypoints are also crossovers. Typically, it''s recommended extra'//&
                                ' FAL waypoints are picked before and after the first and last waypoint. '//&
                                'Therefore, this would typically false.')


    TYPE(RTP_LOGI), TARGET  ::  AMYLOID_BETA  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='AMYLOID_BETA',           &
                    DESCRIPTION='The filaments are amyloid beta fibrils')


    TYPE(RTP_LOGI), TARGET  ::  MICROTUBULE  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='MICROTUBULE',           &
                    DESCRIPTION='The filaments are microtubules')


    TYPE(RTP_INTG), TARGET  ::  MT_PF_NUMBER  =               &
        RTP_INTG(   VALUE=0,                  &
                    KEYWORD='MT_PF_NUMBER',           &
                    DESCRIPTION='For microtubules only: number of protofilaments (e.g. 13)')


    TYPE(RTP_INTG), TARGET  ::  MT_RISE  =               &
        RTP_INTG(   VALUE=0,                  &
                    KEYWORD='MT_RISE',           &
                    DESCRIPTION='For microtubules only: how much further along the filament (distance measured '//&
                                'in tubulin subunits) the end of a MT_PF_NUMBER-element "washer" is compared to '//&
                                'its beginning. The 13-pf microtobules have a value of 3.')


    TYPE(RTP_REAL), TARGET  ::  HELIX_MAX_RADIUS  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='HELIX_MAX_RADIUS',           &
                    DESCRIPTION='Maximum outer radius of the helix, in Angstroms')


    TYPE(RTP_INTG), TARGET  ::  NUM_HELICAL_STARTS  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='NUM_HELICAL_STARTS',           &
                    DESCRIPTION='Number of helical starts. The helical parameters given above must describe one '//&
                                'of these starts. Not including symmetry (e.g. could be describing a 1-start helix '//&
                                'with an additional 2-fold axial symmetry).')

    TYPE(RTP_LOGI), TARGET  ::  STACK  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='STACK',           &
                    DESCRIPTION='The assembly is a stack rather than a true helix')


    TYPE(RTP_REAL), TARGET  ::  ASYM_UNIT_MW  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='ASYM_UNIT_MW',           &
                    DESCRIPTION='Expected molecular weight (Da) of the asymetric unit')

    TYPE(RTP_LOGI), TARGET  ::  CROSS_BETA_FILTER  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='CROSS_BETA_FILTER',           &
                    DESCRIPTION='Filter the data in Fourier space to use the cross-beta reflection at 4.8A')


    TYPE(RTP_LOGI), TARGET  ::  USE_CONJUGATE_GRADIENT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='USE_CONJUGATE_GRADIENT',           &
                    DESCRIPTION='Use conjugate gradient (rather than downhill simplex) for refinement')


    TYPE(RTP_LOGI), TARGET  ::  USE_DIFFERENTIAL_EVOLUTION  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='USE_DIFFERENTIAL_EVOLUTION',           &
                    DESCRIPTION='Use differential evolution (rather than downhill simplex) for refinement')


    TYPE(RTP_LOGI), TARGET  ::  ADD_FAL_WPS  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='ADD_FAL_WPS',           &
                    DESCRIPTION='Add one waypoint at each end of the filament, by spline extrapolation.')


    TYPE(RTP_LOGI), TARGET  ::  FAL_WPS_IN_SCORE  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='FAL_WPS_IN_SCORE',           &
                    DESCRIPTION='The parts of the filament between the first and second as well as between the'//&
                                ' penultimate and last waypoints should be included in the scoring function')


    TYPE(RTP_LOGI), TARGET  ::  FAL_WPS_IN_THREED  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='FAL_WPS_IN_THREED',           &
                    DESCRIPTION='The parts of the filament between the first and second as well as between the '//&
                                'penultimate and last waypoints should be included in the 3D reconstruction')


    TYPE(RTP_REAL), TARGET  ::  FORCE_WP_TO_WP_DISTANCE  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='FORCE_WP_TO_WP_DISTANCE',           &
                    DESCRIPTION='Each filament''s waypoints will be recalculated so that they are spaced by '//&
                                'this distance (in Angstroms). If set to 0.0, the waypoints will be left as in input.')

    TYPE(RTP_REAL), TARGET  ::  FORCE_WP_TO_WP_PHI  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='FORCE_WP_TO_WP_PHI',           &
                    DESCRIPTION='Each filament''s waypoints will be recalculated so that they are spaced by '//&
                                'this angle (in degrees). If set to 0.0, the waypoints will be left as in input.')


    TYPE(RTP_LOGI), TARGET  ::  REVERSE_FILM_CONTRAST  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='REVERSE_FILM_CONTRAST',           &
                    DESCRIPTION='Reverse micrograph contrast internally. It is assumed that protein density is '//&
                                'darker (lower values) than the background. If it isn''t, set this to T')


    TYPE(RTP_LOGI), TARGET  ::  CALC_3D_RECONSTRUCTION  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='CALC_3D_RECONSTRUCTION',           &
                    DESCRIPTION='Calculate a 3D reconstruction using the refined waypoint parameters.')


    TYPE(RTP_LOGI), TARGET  ::  CALC_FSC  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='CALC_FSC',           &
                    DESCRIPTION='Calculate an FSC between two half-dataset reconstructions.')


    TYPE(RTP_LOGI), TARGET  ::  ESTIMATE_SSNR  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='ESTIMATE_SSNR',           &
                    DESCRIPTION='Estimate the SSNR and use it during Wiener filtering of the reconstruction.')


!    TYPE(RTP_LOGI), TARGET  ::  ESTIMATE_PHI_ANGLES  =               &
!        RTP_LOGI(   VALUE=.FALSE.,                  &
!                    KEYWORD='ESTIMATE_PHI_ANGLES',           &
!                    DESCRIPTION='Set this to TRUE if the Phi (twist) angles are unknown. Except in the case of crossovers')


    TYPE(RTP_LOGI), TARGET  ::  EXTRA_CCF_PER_WP  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='EXTRA_CCF_PER_WP',           &
                    DESCRIPTION='')


    TYPE(RTP_LOGI), TARGET  ::  BFACTOR_DURING_3D_RECONSTRUCTION  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='BFACTOR_DURING_3D_RECONSTRUCTION',           &
                    DESCRIPTION='Apply a B-factor filter to images when they are added to 3D reconstructions, '//&
                                'to reflect their quality (score)')

   TYPE(RTP_LOGI), TARGET  ::  BFACTOR_PER_SUBUNIT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='BFACTOR_PER_SUBUNIT',           &
                    DESCRIPTION='During reconstruction, B factor is computer per subunit rather than per filament')


    TYPE(RTP_REAL), TARGET  ::  MODEL_3D_FILTER_BFACTOR  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='MODEL_3D_FILTER_BFACTOR',           &
                    DESCRIPTION='Apply this B factor (in A^2) to the reference before it is used by the scoring function')


    TYPE(RTP_REAL), TARGET  ::  AVERAGE_SCORE  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='AVERAGE_SCORE',           &
                    DESCRIPTION='Average score of filaments after refinement. Used to calculate a Bfactor to '//&
                                'apply to images during reconstruction.')


    TYPE(RTP_LOGI), TARGET  ::  FINALISE_3D_RECONSTRUCTION  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='FINALISE_3D_RECONSTRUCTION',           &
                    DESCRIPTION='Compute a finalised 3D volume. If .FALSE., the unweighted 3D and a CTF^2 volume '//&
                                'will be dumped to the disk and can be finalised later.')

    TYPE(RTP_LOGI), TARGET  ::  RECONSTRUCT_AMPLITUDES  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='RECONSTRUCT_AMPLITUDES',           &
                    DESCRIPTION='Compute a reconstruction using only amplitudes.')


    TYPE(RTP_LOGI), TARGET  ::  APPLY_FOM_FILTER  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='APPLY_FOM_FILTER',           &
                    DESCRIPTION='Apply a Figure-Of-Merit (FOM) filter to the final reconstruction (equivalent to '//&
                                'FCREF in Frealign; see Henderson & Rosenthal, 2003)')


    TYPE(RTP_REAL), TARGET  ::  MIN_SCORE_FIL_TO_3D  =               &
        RTP_REAL(   VALUE=-1.0,                  &
                    KEYWORD='MIN_SCORE_FIL_TO_3D',           &
                    DESCRIPTION='Only filaments with a score greather than this will be included in the 3D reconstruction')


!    TYPE(RTP_INTG), TARGET  ::  NUM_MINIM_ROUNDS  =               &
!        RTP_INTG(   VALUE=1,                  &
!                    KEYWORD='NUM_MINIM_ROUNDS',           &
!                    DESCRIPTION='')


    TYPE(RTP_LOGI), TARGET  ::  POLARITY_CHECK  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='POLARITY_CHECK',           &
                    DESCRIPTION='Test both filament polarities (the two polarities are related by a reversal of '//&
                                'the order of waypoints, equivalent to a change of 180 degrees in out-of-plane tilt)')


    TYPE(RTP_LOGI), TARGET  ::  PHI_POLARITY_CHECK  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='PHI_POLARITY_CHECK',           &
                    DESCRIPTION='Test both twist polarities of the filament (the two polarities are related by an '//&
                                'axial rotation of 180 degrees)')


    TYPE(RTP_LOGI), TARGET  ::  REAL_SPACE_PROJECTION  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='REAL_SPACE_PROJECTION',           &
                    DESCRIPTION='Use real space projection. If .FALSE., reciprocal space slice extraction is '//&
                                'used instead (which is faster).')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_X  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='REFINE_X',           &
                    DESCRIPTION='Refine X')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_Y  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='REFINE_Y',           &
                    DESCRIPTION='Refine Y')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_Z  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='REFINE_Z',           &
                    DESCRIPTION='Refine Z (defocus)')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_PHI  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='REFINE_PHI',           &
                    DESCRIPTION='Refine phi (twist; axial rotation)')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_THETA  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='REFINE_THETA',           &
                    DESCRIPTION='Refine theta (out-of-plane rotation)')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_PSI  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='REFINE_PSI',           &
                    DESCRIPTION='Refine psi (in-plane rotation)')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_SUBUNIT_NUM  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='REFINE_SUBUNIT_NUM',           &
                    DESCRIPTION='Refine subunit number (axial positioning)')

    TYPE(RTP_LOGI), TARGET  ::  USE_HERMITE_SPLINES =    &
       RTP_LOGI(    VALUE=.TRUE.,               &
                    KEYWORD='USE_HERMITE_SPLINES',  &
                    DESCRIPTION='Use Hermite splines, where the tangents (i.e. psi & theta) are used as contraints')


    TYPE(RTP_LOGI), TARGET  ::  RIGID_X  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='RIGID_X',           &
                    DESCRIPTION='X is rigid (per-filament rather than per-waypoint)')


    TYPE(RTP_LOGI), TARGET  ::  RIGID_Y  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='RIGID_Y',           &
                    DESCRIPTION='Y is rigid (per-filament rather than per-waypoint)')


    TYPE(RTP_LOGI), TARGET  ::  RIGID_Z  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='RIGID_Z',           &
                    DESCRIPTION='Z is rigid (per-filament rather than per-waypoint)')


    TYPE(RTP_LOGI), TARGET  ::  RIGID_PHI  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='RIGID_PHI',           &
                    DESCRIPTION='Phi is rigid (per-filament rather than per-waypoint)')


    TYPE(RTP_LOGI), TARGET  ::  RIGID_THETA  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='RIGID_THETA',           &
                    DESCRIPTION='Theta is rigid (per-filament rather than per-waypoint)')


    TYPE(RTP_LOGI), TARGET  ::  RIGID_PSI  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='RIGID_PSI',           &
                    DESCRIPTION='Psi is rigid (per-filament rather than per-waypoint)')


    TYPE(RTP_LOGI), TARGET  ::  RIGID_SUBUNIT_NUM  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='RIGID_SUBUNIT_NUM',           &
                    DESCRIPTION='Subunit number (~ axial phasing) is rigid (per-filament rather than per-waypoint).'//&
                                ' Rigid implies constant rise per subunit.')


    TYPE(RTP_LOGI), TARGET  ::  HOLD_Z_AVERAGE  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='HOLD_Z_AVERAGE',           &
                    DESCRIPTION='The average Z value of each filament will be held constant')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_SUBUNIT_NUM_INCREMENTS  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='REFINE_SUBUNIT_NUM_INCREMENTS',           &
                    DESCRIPTION='Refine wp-on-wp increment in subunit_num rather than actual subunit_num values')


    TYPE(RTP_LOGI), TARGET  ::  SUBUNIT_NUM_INTEGER_CHANGES_ONLY  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='SUBUNIT_NUM_INTEGER_CHANGES_ONLY',           &
                    DESCRIPTION='Only allow integer shifts in subunit_num values')

    TYPE(RTP_LOGI), TARGET  ::  FRAGMENTED_PHI  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='FRAGMENTED_PHI',           &
                    DESCRIPTION='Phi function is discontinuous, with breaks between waypoints')

    TYPE(RTP_LOGI), TARGET  ::  FRAGMENTED_SUBUNIT_NUM  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='FRAGMENTED_SUBUNIT_NUM',           &
                    DESCRIPTION='Subunit_num function is discontinuous, with breaks between waypoints')

    TYPE(RTP_LOGI), TARGET  ::  CONSTANT_TWIST  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='CONSTANT_TWIST',           &
                    DESCRIPTION='The twist per unit length is constant.')


    TYPE(RTP_LOGI), TARGET  ::  CONSTANT_RISE_PER_SUBUNIT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='CONSTANT_RISE_PER_SUBUNIT',           &
                    DESCRIPTION='The rise per subunit is constant. Note that non-cumulative axial disorder is not accounted for.')


    TYPE(RTP_REAL), TARGET  ::  SEARCH_HALF_RANGE_X  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='SEARCH_HALF_RANGE_X',           &
                    DESCRIPTION='Pixels. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    TYPE(RTP_REAL), TARGET  ::  SEARCH_HALF_RANGE_Y  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='SEARCH_HALF_RANGE_Y',           &
                    DESCRIPTION='Pixels. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    TYPE(RTP_REAL), TARGET  ::  SEARCH_HALF_RANGE_Z  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='SEARCH_HALF_RANGE_Z',           &
                    DESCRIPTION='Pixels. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    TYPE(RTP_REAL), TARGET  ::  SEARCH_HALF_RANGE_PHI  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='SEARCH_HALF_RANGE_PHI',           &
                    DESCRIPTION='Degrees. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    TYPE(RTP_REAL), TARGET  ::  SEARCH_HALF_RANGE_THETA  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='SEARCH_HALF_RANGE_THETA',           &
                    DESCRIPTION='Degrees. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    TYPE(RTP_REAL), TARGET  ::  SEARCH_HALF_RANGE_PSI  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='SEARCH_HALF_RANGE_PSI',           &
                    DESCRIPTION='Degrees. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    TYPE(RTP_REAL), TARGET  ::  SEARCH_HALF_RANGE_SUBUNIT_NUM  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='SEARCH_HALF_RANGE_SUBUNIT_NUM',           &
                    DESCRIPTION='Number of subunits. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    TYPE(RTP_LOGI), TARGET  ::  SEARCH_FILAMENT_TILT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='SEARCH_FILAMENT_TILT',           &
                    DESCRIPTION='Do a brute-force search for filament out-of-plane tilts '//&
                                '(whole filament is tilted as a rigid body)')


    TYPE(RTP_LOGI), TARGET  ::  SUBTRACT_FIL_FROM_REF  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='SUBTRACT_FIL_FROM_REF',           &
                    DESCRIPTION='Subtract the original state of each filament from the 3D reference model, '//&
                                'to reduce bias to original state, before the filament is processed.')

    TYPE(RTP_LOGI), TARGET  ::  SINGLE_FILAMENT_REFINEMENT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='SINGLE_FILAMENT_REFINEMENT',           &
                    DESCRIPTION='Refine each filament against itself.')


    TYPE(RTP_INTG), TARGET  ::  NUM_MINIM_RUNS  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='NUM_MINIM_RUNS',           &
                    DESCRIPTION='Number of minimisations to run')


    TYPE(RTP_LOGI), TARGET  ::  DS_RANDOM_START_CONF  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='DS_RANDOM_START_CONF',           &
                    DESCRIPTION='Randomise the starting state of the dowhill simplex minimiser')


    TYPE(RTP_LOGI), TARGET  ::  MINIM_RANDOM_START_POS  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='MINIM_RANDOM_START_POS',           &
                    DESCRIPTION='Randomise the starting position of the minimiser (with '//&
                                'Gaussian noise, sigma = search half-range)')


    TYPE(RTP_INTG), TARGET  ::  DS_MAX_NUMBER_OF_ITERATIONS  =               &
        RTP_INTG(   VALUE=DS_MAX_ITERATIONS,                  &
                    KEYWORD='DS_MAX_NUMBER_OF_ITERATIONS',           &
                    DESCRIPTION='Maximum number of iterations of the Downhill Simplex per run')


    TYPE(RTP_INTG), TARGET  ::  PATH_OVERSAMPLING  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='PATH_OVERSAMPLING',           &
                    DESCRIPTION='On average, each pixel will be used this number of times in the scoring function. 1 by default.')


    TYPE(RTP_LOGI), TARGET  ::  SKIP_REFINEMENT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='SKIP_REFINEMENT',           &
                    DESCRIPTION='Do not do any refinement.')


    TYPE(RTP_LOGI), TARGET  ::  MODEL_3D_AVAILABLE  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='MODEL_3D_AVAILABLE',           &
                    DESCRIPTION='A model 3D reconstruction is available as input.')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_WPS_SIMULTANEOUSLY  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='REFINE_WPS_SIMULTANEOUSLY',           &
                    DESCRIPTION='Refine all waypoints together, rather than sequentially. Default is simultaneous refinement.')


    TYPE(RTP_LOGI), TARGET  ::  SEARCH_SUBUNIT_NUM  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='SEARCH_SUBUNIT_NUM',           &
                    DESCRIPTION='Do a brute-force search in integer steps around the current subunit number values.')


    TYPE(RTP_LOGI), TARGET  ::  SEARCH_MT_SEAM  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='SEARCH_MT_SEAM',           &
                    DESCRIPTION='For microtubules only: search for the seam position')


    TYPE(RTP_LOGI), TARGET  ::  AVERAGE_MODEL_CYLINDRICALLY  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='AVERAGE_MODEL_CYLINDRICALLY',           &
                    DESCRIPTION='Average cylindrically the input 3D model before refinement starts')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_IN_FOURIER_SPACE  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='REFINE_IN_FOURIER_SPACE',           &
                    DESCRIPTION='Refine the filament model''s Fourier transform rather than the model itself')


    TYPE(RTP_LOGI), TARGET  ::  REFINE_NORMAL_MODES  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='REFINE_NORMAL_MODES',           &
                    DESCRIPTION='Refine the filament model''s normal mode amplitudes rather than the model itself')


    TYPE(RTP_LOGI), TARGET  ::  STRAIGHTEN_FILAMENT_BEFORE_REFINEMENT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='STRAIGHTEN_FIL_BEFORE_REFINEMENT',           &
                    DESCRIPTION='Straighten the filament (and impose constant helical parameters) before refinement')


    TYPE(RTP_LOGI), TARGET  ::  MOVE_WAYPOINTS_TO_NEAREST_REPEAT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='MOVE_WAYPOINTS_TO_NEAREST_REPEAT',           &
                    DESCRIPTION='On output, move (refined) waypoints to the nearest helical repeat')

    TYPE(RTP_LOGI), TARGET  ::  AVERAGE_Z_TO_CTFTILT_PLANE  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='AVERAGE_Z_TO_CTFTILT_PLANE',           &
                    DESCRIPTION='Move each filament in the Z direction so that its average defocus matches '//&
                                'the expectation from CTFTILT information')


    TYPE(RTP_LOGI), TARGET  ::  USE_WEIGHTED_CORRELATION  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='USE_WEIGHTED_CORRELATION',           &
                    DESCRIPTION='Use a weighted (rather than linear) correlation function. See Stewart & Grigorieff (2004).')


    TYPE(RTP_LOGI), TARGET  ::  USE_AVERAGE_OF_CUBED_SEGMENT_SCORES  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='USE_AVE_OF_CUBED_SEGMENT_SCORES',           &
                    DESCRIPTION='When in full-filament mode: if false, full-filament normalisation is used')


    TYPE(RTP_LOGI), TARGET  ::  USE_CURVATURE_RESTRAINT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='USE_CURVATURE_RESTRAINT',           &
                    DESCRIPTION='Use the curvature of each filament as a restraint on its refinement')

    TYPE(RTP_LOGI), TARGET  ::  USE_TORSION_RESTRAINT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='USE_TORSION_RESTRAINT',           &
                    DESCRIPTION='Use the torsion of each filament as a restraint on its refinement')

    TYPE(RTP_LOGI), TARGET  ::  USE_STRETCH_RESTRAINT  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='USE_STRETCH_RESTRAINT',           &
                    DESCRIPTION='Use the stretch of each filament as a restraint on its refinement')


    TYPE(RTP_REAL), TARGET  ::  PERSISTENCE_LENGTH  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='PERSISTENCE_LENGTH',           &
                    DESCRIPTION='Persistence length of the filament (Angstroms)')


    TYPE(RTP_REAL), TARGET  ::  TORSION_PERSISTENCE_LENGTH  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='TORSION_PERSISTENCE_LENGTH',           &
                    DESCRIPTION='Torsional persistence length (Angstroms)')


    TYPE(RTP_REAL), TARGET  ::  STRETCH_PERSISTENCE_LENGTH  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='STRETCH_PERSISTENCE_LENGTH',           &
                    DESCRIPTION='Stretching persistence length (Angstroms)')


    TYPE(RTP_REAL), TARGET  ::  TARGET_RESOLUTION  =               &
        RTP_REAL(   VALUE=25.0,                  &
                    KEYWORD='TARGET_RESOLUTION',           &
                    DESCRIPTION='Target resolution for the refinements (Angstroms)')


    TYPE(RTP_REAL), TARGET  ::  LOW_RESOLUTION_CUTON  =               &
        RTP_REAL(   VALUE=1000.0,                  &
                    KEYWORD='LOW_RESOLUTION_CUTON',           &
                    DESCRIPTION='Lowest resolution to be used during refinements (Angstroms)')


    TYPE(RTP_REAL), TARGET  ::  HIGH_RESOLUTION_CUTOFF  =               &
        RTP_REAL(   VALUE=2.0,                  &
                    KEYWORD='HIGH_RESOLUTION_CUTOFF',           &
                    DESCRIPTION='Highest resolution to be used during refinements (Angstroms)')

    TYPE(RTP_LOGI), TARGET  ::  WPS_AS_SPS  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='WPS_AS_SPS',           &
                    DESCRIPTION='Treat filament waypoints (WPs) as independent single particles (SPs)')


    TYPE(RTP_REAL), TARGET  ::  SP_SEGMENT_LENGTH  =               &
        RTP_REAL(   VALUE=700.0,                  &
                    KEYWORD='SP_SEGMENT_LENGTH',           &
                    DESCRIPTION='In Angstroms. Single particles are defined as segments of this length along '//&
                                'the filament. Only used if WPS_AS_SPS is .TRUE..')


    TYPE(RTP_LOGI), TARGET  ::  CACHE_PROJECTIONS  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='CACHE_PROJECTIONS',           &
                    DESCRIPTION='Cache 2D projections of 3D models. Could lead to significant time savings, '//&
                                'especially when using expensive interpolation schemes. However, could also '//&
                                'lead to loss in quality of alignments.')


    TYPE(RTP_LOGI), TARGET  ::  SINGLE_FILAMENT_RECONSTRUCTIONS  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='SINGLE_FILAMENT_RECONSTRUCTIONS',           &
                    DESCRIPTION='For each individual filament, write to disk the central slice of its 3D reconstruction')


    TYPE(RTP_LOGI), TARGET  ::  WRITE_MATCHING_PROJECTIONS  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='WRITE_MATCHING_PROJECTIONS',           &
                    DESCRIPTION='Write stacks of images & matching projections to disk after refinement')


    TYPE(RTP_LOGI), TARGET  ::  WRITE_FILAMENT_TRACES  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='WRITE_FILAMENT_TRACES',           &
                    DESCRIPTION='Write coordinates (boxer) files showing waypoint positions and filament traces')


    TYPE(RTP_CHAR), TARGET  ::  INTERPOL_FUNCTION  =               &
        RTP_CHAR(   VALUE='LINEAR',                  &
                    KEYWORD='INTERPOL_FUNCTION',           &
                    DESCRIPTION='The interpolation function to use')


    TYPE(RTP_CHAR), TARGET  ::  WINDOW_FUNCTION  =               &
        RTP_CHAR(   VALUE='RECT',                  &
                    KEYWORD='WINDOW_FUNCTION',           &
                    DESCRIPTION='The windowing function to use')


    TYPE(RTP_INTG), TARGET  ::  WINDOW_WIDTH  =               &
        RTP_INTG(   VALUE=2,                  &
                    KEYWORD='WINDOW_WIDTH',           &
                    DESCRIPTION='Window width for interpolation (kernel support)')

	TYPE(RTP_LOGI), TARGET ::   SP_TO_FF = 	&
		RTP_LOGI(	VALUE=.FALSE.,			&
					KEYWORD='SP_TO_FF',		&
					DESCRIPTION='Input parameters are from SP analysis, but FF analsysis should be done')

CONTAINS

SUBROUTINE MODE_FILAMENT_REFINE_INIT_AND_RUN
    IMPLICIT NONE
    ! Intialise the mode
    CALL FILAMENT_REFINE%INIT(COMMAND='FILAMENT_REFINE',SUBROUTINE=MODE_FILAMENT_REFINE_SLAVE, &
                                SHORT_DESCRIPTION='Refine filament waypoint parameters')

    ! Add RTPs to the mode
    CALL FILAMENT_REFINE%ADD_RTP(INPUT_PAR_FILENAME)
    CALL FILAMENT_REFINE%ADD_RTP(OUTPUT_PAR_FILENAME)
    CALL FILAMENT_REFINE%ADD_RTP(OUTPUT_PAR_FILENAME_SUBUNITS)
    CALL FILAMENT_REFINE%ADD_RTP(MODEL_3D_FILENAME)
    CALL FILAMENT_REFINE%ADD_RTP(OUTPUT_3D_FILENAME)
    CALL FILAMENT_REFINE%ADD_RTP(CTFTILT_FILES)
    CALL FILAMENT_REFINE%ADD_RTP(CTFFIND_FILES)
    CALL FILAMENT_REFINE%ADD_RTP(FILM_FILES)
    CALL FILAMENT_REFINE%ADD_RTP(MATCHING_PROJECTIONS_FILES)
    CALL FILAMENT_REFINE%ADD_RTP(FILAMENT_TRACE_FILES)
    CALL FILAMENT_REFINE%ADD_RTP(PIXEL_SIZE)
    CALL FILAMENT_REFINE%ADD_RTP(TWIST_PER_SUBUNIT)
    CALL FILAMENT_REFINE%ADD_RTP(RISE_PER_SUBUNIT)
    CALL FILAMENT_REFINE%ADD_RTP(SYMMETRY_AXIAL)
    CALL FILAMENT_REFINE%ADD_RTP(SYMMETRY_PERP)
    CALL FILAMENT_REFINE%ADD_RTP(WPS_ARE_CROSSOVERS)
    CALL FILAMENT_REFINE%ADD_RTP(FAL_WPS_ARE_CROSSOVERS)
    CALL FILAMENT_REFINE%ADD_RTP(AMYLOID_BETA)
    CALL FILAMENT_REFINE%ADD_RTP(MICROTUBULE)
    CALL FILAMENT_REFINE%ADD_RTP(MT_PF_NUMBER)
    CALL FILAMENT_REFINE%ADD_RTP(MT_RISE)
    CALL FILAMENT_REFINE%ADD_RTP(HELIX_MAX_RADIUS)
    CALL FILAMENT_REFINE%ADD_RTP(NUM_HELICAL_STARTS)
    CALL FILAMENT_REFINE%ADD_RTP(STACK)
    CALL FILAMENT_REFINE%ADD_RTP(ASYM_UNIT_MW)
    CALL FILAMENT_REFINE%ADD_RTP(CROSS_BETA_FILTER)
    CALL FILAMENT_REFINE%ADD_RTP(USE_CONJUGATE_GRADIENT)
    CALL FILAMENT_REFINE%ADD_RTP(USE_DIFFERENTIAL_EVOLUTION)
    CALL FILAMENT_REFINE%ADD_RTP(ADD_FAL_WPS)
    CALL FILAMENT_REFINE%ADD_RTP(FAL_WPS_IN_SCORE)
    CALL FILAMENT_REFINE%ADD_RTP(FAL_WPS_IN_THREED)
    CALL FILAMENT_REFINE%ADD_RTP(FORCE_WP_TO_WP_DISTANCE)
    CALL FILAMENT_REFINE%ADD_RTP(FORCE_WP_TO_WP_PHI)
    CALL FILAMENT_REFINE%ADD_RTP(REVERSE_FILM_CONTRAST)
    CALL FILAMENT_REFINE%ADD_RTP(CALC_3D_RECONSTRUCTION)
    CALL FILAMENT_REFINE%ADD_RTP(CALC_FSC)
    CALL FILAMENT_REFINE%ADD_RTP(ESTIMATE_SSNR)
    CALL FILAMENT_REFINE%ADD_RTP(BFACTOR_DURING_3D_RECONSTRUCTION)
    CALL FILAMENT_REFINE%ADD_RTP(BFACTOR_PER_SUBUNIT)
    CALL FILAMENT_REFINE%ADD_RTP(MODEL_3D_FILTER_BFACTOR)
    CALL FILAMENT_REFINE%ADD_RTP(AVERAGE_SCORE)
    CALL FILAMENT_REFINE%ADD_RTP(FINALISE_3D_RECONSTRUCTION)
    CALL FILAMENT_REFINE%ADD_RTP(RECONSTRUCT_AMPLITUDES)
    CALL FILAMENT_REFINE%ADD_RTP(APPLY_FOM_FILTER)
    CALL FILAMENT_REFINE%ADD_RTP(MIN_SCORE_FIL_TO_3D)
    CALL FILAMENT_REFINE%ADD_RTP(POLARITY_CHECK)
    CALL FILAMENT_REFINE%ADD_RTP(PHI_POLARITY_CHECK)
    !CALL FILAMENT_REFINE%ADD_RTP(REAL_SPACE_PROJECTION)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_X)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_Y)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_Z)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_PHI)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_THETA)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_PSI)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_SUBUNIT_NUM)
    CALL FILAMENT_REFINE%ADD_RTP(USE_HERMITE_SPLINES)
    CALL FILAMENT_REFINE%ADD_RTP(RIGID_X)
    CALL FILAMENT_REFINE%ADD_RTP(RIGID_Y)
    CALL FILAMENT_REFINE%ADD_RTP(RIGID_Z)
    CALL FILAMENT_REFINE%ADD_RTP(RIGID_PHI)
    CALL FILAMENT_REFINE%ADD_RTP(RIGID_THETA)
    CALL FILAMENT_REFINE%ADD_RTP(RIGID_PSI)
    CALL FILAMENT_REFINE%ADD_RTP(RIGID_SUBUNIT_NUM)
    CALL FILAMENT_REFINE%ADD_RTP(HOLD_Z_AVERAGE)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_SUBUNIT_NUM_INCREMENTS)
    CALL FILAMENT_REFINE%ADD_RTP(SUBUNIT_NUM_INTEGER_CHANGES_ONLY)
    CALL FILAMENT_REFINE%ADD_RTP(FRAGMENTED_PHI)
    CALL FILAMENT_REFINE%ADD_RTP(FRAGMENTED_SUBUNIT_NUM)
    CALL FILAMENT_REFINE%ADD_RTP(CONSTANT_TWIST)
    CALL FILAMENT_REFINE%ADD_RTP(CONSTANT_RISE_PER_SUBUNIT)
    CALL FILAMENT_REFINE%ADD_RTP(SEARCH_HALF_RANGE_X)
    CALL FILAMENT_REFINE%ADD_RTP(SEARCH_HALF_RANGE_Y)
    CALL FILAMENT_REFINE%ADD_RTP(SEARCH_HALF_RANGE_Z)
    CALL FILAMENT_REFINE%ADD_RTP(SEARCH_HALF_RANGE_PHI)
    CALL FILAMENT_REFINE%ADD_RTP(SEARCH_HALF_RANGE_THETA)
    CALL FILAMENT_REFINE%ADD_RTP(SEARCH_HALF_RANGE_PSI)
    CALL FILAMENT_REFINE%ADD_RTP(SEARCH_HALF_RANGE_SUBUNIT_NUM)
    CALL FILAMENT_REFINE%ADD_RTP(SEARCH_FILAMENT_TILT)
    CALL FILAMENT_REFINE%ADD_RTP(SUBTRACT_FIL_FROM_REF)
    CALL FILAMENT_REFINE%ADD_RTP(SINGLE_FILAMENT_REFINEMENT)
    CALL FILAMENT_REFINE%ADD_RTP(NUM_MINIM_RUNS)
    CALL FILAMENT_REFINE%ADD_RTP(DS_RANDOM_START_CONF)
    CALL FILAMENT_REFINE%ADD_RTP(MINIM_RANDOM_START_POS)
    CALL FILAMENT_REFINE%ADD_RTP(DS_MAX_NUMBER_OF_ITERATIONS)
    CALL FILAMENT_REFINE%ADD_RTP(PATH_OVERSAMPLING)
    CALL FILAMENT_REFINE%ADD_RTP(SKIP_REFINEMENT)
    CALL FILAMENT_REFINE%ADD_RTP(MODEL_3D_AVAILABLE)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_WPS_SIMULTANEOUSLY)
    CALL FILAMENT_REFINE%ADD_RTP(SEARCH_SUBUNIT_NUM)
    CALL FILAMENT_REFINE%ADD_RTP(SEARCH_MT_SEAM)
    CALL FILAMENT_REFINE%ADD_RTP(AVERAGE_MODEL_CYLINDRICALLY)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_IN_FOURIER_SPACE)
    CALL FILAMENT_REFINE%ADD_RTP(REFINE_NORMAL_MODES)
    CALL FILAMENT_REFINE%ADD_RTP(STRAIGHTEN_FILAMENT_BEFORE_REFINEMENT)
    CALL FILAMENT_REFINE%ADD_RTP(MOVE_WAYPOINTS_TO_NEAREST_REPEAT)
    CALL FILAMENT_REFINE%ADD_RTP(AVERAGE_Z_TO_CTFTILT_PLANE)
    CALL FILAMENT_REFINE%ADD_RTP(USE_WEIGHTED_CORRELATION)
    CALL FILAMENT_REFINE%ADD_RTP(USE_AVERAGE_OF_CUBED_SEGMENT_SCORES)
    CALL FILAMENT_REFINE%ADD_RTP(USE_CURVATURE_RESTRAINT)
    CALL FILAMENT_REFINE%ADD_RTP(USE_TORSION_RESTRAINT)
    CALL FILAMENT_REFINE%ADD_RTP(USE_STRETCH_RESTRAINT)
    CALL FILAMENT_REFINE%ADD_RTP(PERSISTENCE_LENGTH)
    CALL FILAMENT_REFINE%ADD_RTP(TORSION_PERSISTENCE_LENGTH)
    CALL FILAMENT_REFINE%ADD_RTP(STRETCH_PERSISTENCE_LENGTH)
    CALL FILAMENT_REFINE%ADD_RTP(TARGET_RESOLUTION)
    CALL FILAMENT_REFINE%ADD_RTP(LOW_RESOLUTION_CUTON)
    CALL FILAMENT_REFINE%ADD_RTP(HIGH_RESOLUTION_CUTOFF)
    CALL FILAMENT_REFINE%ADD_RTP(WPS_AS_SPS)
    CALL FILAMENT_REFINE%ADD_RTP(SP_SEGMENT_LENGTH)
    !CALL FILAMENT_REFINE%ADD_RTP(CACHE_PROJECTIONS)
    CALL FILAMENT_REFINE%ADD_RTP(SINGLE_FILAMENT_RECONSTRUCTIONS)
    CALL FILAMENT_REFINE%ADD_RTP(WRITE_MATCHING_PROJECTIONS)
    CALL FILAMENT_REFINE%ADD_RTP(WRITE_FILAMENT_TRACES)
    CALL FILAMENT_REFINE%ADD_RTP(INTERPOL_FUNCTION)
    CALL FILAMENT_REFINE%ADD_RTP(WINDOW_FUNCTION)
    CALL FILAMENT_REFINE%ADD_RTP(WINDOW_WIDTH)
	CALL FILAMENT_REFINE%ADD_RTP(SP_TO_FF)

    ! Run the mode
    CALL FILAMENT_REFINE%RUN()
END SUBROUTINE MODE_FILAMENT_REFINE_INIT_AND_RUN


SUBROUTINE MODE_FILAMENT_REFINE_SLAVE
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE, PSF_RADIUS, PSF_RADIUS_PIXELS, PBC, FILENAME_LEN, DEFAULT_FALLOFF_PIXELS
    use UsefulFunctions, only : IsOdd
    USE PARSERS, ONLY : STRING_EQUAL, FILE_NUMBER_FROM_FILENAME, FILENAME_FROM_WILD, FILENAME_ADD_SUFFIX
    USE PARSERS, ONLY : FILENAME_REMOVE_EXTENSION, FILENAME_CONTAINS_QMARKS
    USE FREALIX_PARAMETERS
    USE PROCFS, ONLY : PRINT_MEMORY_USAGE_STATS
    ! "Science" modules
    USE HELICES, ONLY : HELICAL_SYMMETRY, HELICAL_SYMMETRY_INIT, HELICAL_SYMMETRY_PRINT_INFO
    USE FILAMENTS_BASE, ONLY : FILAMENT, FILAMENT_DEALLOCATE, FILAMENT_ASSIGN
    USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_EXTRA_INFO
    USE FILAMENTS_MODEL_MODIFY, ONLY : FILAMENT_SHIFT_ALL_WPS, FILAMENT_FLIP_POLARITY, FILAMENT_ADD_FAL_WPS, FILAMENT_STRAIGHTEN
    USE FILAMENTS_MODEL_MODIFY, ONLY : FILAMENT_MOVE_WPS_TO_NEAREST_SUBUNIT
    USE FILAMENTS, ONLY : FILAMENT_RESAMPLE_WAYPOINTS, FILAMENT_PRINT_DIFF
    USE FMCS_EVALUATION, ONLY : FMC_EVAL, FMC_MODEL3D_PREP
    USE FILAMENTS_MICROTUBULES, ONLY : IMAGE_IMPOSE_MICROTUBULE_SYMMETRY
    USE FILAMENTS_MINIMISATION, ONLY : FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX
    USE IMAGES
    USE FMCS, ONLY : FMC_BF_TILT, FMC_MAXIMISE, FMC_BF_FIND_MT_SEAM, FMC_BF_SUBUNIT_NUM_INTG_REFINE
    USE FMCS_BASE, ONLY : FILAMENT_MODEL_COMPARISON, FMC_BOX_SIZE, FMC_THREED_SIZE
    USE FMCS_BASE, ONLY : FMC_INIT, FMC_NEW_FILAMENT, FMC_UPDATE_FIL_THREED, FMC_PRINT_INFO, FMC_DEALLOCATE
    USE FILAMENTS, ONLY : FILAMENT_BOXER_DB_TO_DISK, FILAMENT_CALC_THREED, FILAMENT_PRINT_INFO, FILAMENT_TRACE_TO_DISK_CSV
    USE INTERPOLATION, ONLY : INTERPOLATION_SCHEME, INTERPOLATION_SCHEME_INIT
    USE THREEDS_BASE, ONLY : THREED, THREED_INIT, THREED_DEALLOCATE
    USE IMAGES_HELICES, ONLY : IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW, IMAGE_HELIX_MASK
    USE POINT_SYMMETRIES, ONLY : POINT_SYMMETRY_SYMBOL
    USE THREEDS, ONLY : THREED_EVAL_RECONSTRUCTION, THREED_ADD, THREED_WRITE_TO_DISK
    USE FSCS, ONLY : COMPUTE_FSC
    USE IMAGES_COMPARISON, ONLY : PRINT_FSC
    use ImageFiles

    IMPLICIT NONE
    !
    ! Private variables
    !
    TYPE(HELICAL_SYMMETRY)                  ::  HELIX_AVERAGE                   !   Average helical parameters
    LOGICAL                                 ::  USE_CTFTILT                     !   Whether we'll be using ctftilt or ctffind information
    TYPE(FILAMENT), ALLOCATABLE, TARGET     ::  FILS(:)                         !   Array of filament objects
    TYPE(FILAMENT), ALLOCATABLE             ::  FIL_STATES(:,:,:)               !   Temporary array of filament states (1st dim: polarity, 2dim: phi polarity, 3dim: before/after refinement)
    TYPE(FILAMENT), ALLOCATABLE             ::  FILS_FINAL(:)                   !   Array of final states of filament objects
    TYPE(IMAGE), TARGET                     ::  MODEL_3DVOL                     !   Input 3D model
    TYPE(IMAGE)                             ::  NEW_3DVOL(2)                    !   Output 3D reconstruction volume(s) (one for each half-reconstruction, if we're doing FSC)
    TYPE(IMAGE)                             ::  NEW_3DVOL_AMPL(2)               !   Output 3D reconstruction volume(s) containing only amplitudes
    TYPE(THREED)                            ::  NEW_3DREC                       !   3D reconstruction object
    TYPE(THREED)                            ::  CURRENT_3DREC                   !   Temporary 3D reconstruction object for individual filaments
    TYPE(THREED), TARGET                    ::  REF_3DREC                       !   Input reference 3D reconstruction
    TYPE(IMAGE)                             ::  CURRENT_3DVOL                   !   Temporary finalised 3D reconstruction
    TYPE(IMAGE)                             ::  CURRENT_3DVOL_AMPL
    TYPE(IMAGE)                             ::  CURRENT_3DVOL_SLICE             !   Central XY slice from temporary finalised 3D reconstruction
    INTEGER                                 ::  IERR                            !   Status
    TYPE(FILAMENT_MODEL_COMPARISON)         ::  FMC                             !   An object containing relevant information re. the comparison
    TYPE(IMAGE), TARGET                     ::  CURRENT_FILM_IMAGE              !   The image of the film being currently analysed
    LOGICAL                                 ::  NEW_FILM                        !   Whether we need to read a new film image from disk
    INTEGER                                 ::  CURRENT_FILAMENT                !   Filament we're currently looping over
    LOGICAL                                 ::  REFINE_PARAMS(7)                !   Which parameters (X,Y,Z,Phi,Theta,Psi) will we refine?
    TYPE(INTERPOLATION_SCHEME)              ::  INTERPOL_SCHEME                 !   Interpolation scheme to be used
    INTEGER                                 ::  I                               !   Loop counter
    REAL                                    ::  HELIX_MAX_RADIUS_PIX            !   Maximum outer radius of helix (pixels)
    CHARACTER(LEN=FILENAME_LEN)             ::  MATCH_PROJ_FILENAME             !   Filename for matching projection stacks
    INTEGER                                 ::  CURRENT_FILM_NUMBER             !   Number of film to which current filament belongs
    CHARACTER(LEN=FILENAME_LEN)             ::  FILAMENT_TRACE_FILENAME         !   Filename for coordinates of waypoints and trace
    CHARACTER(LEN=FILENAME_LEN)             ::  TEMP_FILENAME                   !   Temporary file name
    CHARACTER(LEN=FILENAME_LEN)             ::  OUTPUT_3D_AMPL_FILENAME         !   Filename for output amplitude reconstruction
    REAL                                    ::  FILM_HP_PARAM                   !   High-pass filter parameter for film images
    REAL,           ALLOCATABLE             ::  FOUSHELL(:)                     !   Fourier shell correlation
    REAL                                    ::  SEARCH_HALF_RANGE_OVERRIDE(7)   !   Override of search half-range by the user
    LOGICAL                                 ::  RREFINE_NORMAL_MODES(7)         !   Some parameters may be refined via their normal modes
    INTEGER                                 ::  MIN_SAFE_BOX_SIZE               !   Safe box size
    REAL                                    ::  DUMMY                           !   For debugging
    LOGICAL,        PARAMETER               ::  DEBUG   =   .FALSE.             !   Debugging flag
    INTEGER                                 ::  CROP_DIM(3)
    ! Scores
    REAL,   ALLOCATABLE                     ::  SCORE(:,:,:)                    !   For each polarity (1st dim), for each phi polarity (2nd dim) and before/after minimisation (values 1 or 2 3rd dim), the score
    REAL                                    ::  SCORE_TMP
    ! Polarity checks
    INTEGER                                 ::  NUM_POLARITIES_TO_TEST          !   2 if we check both possible polarities of each filament
    INTEGER                                 ::  NUM_PHI_POLARITIES_TO_TEST      !   2 if we check both phi polarities, 1 otherwise
    INTEGER                                 ::  CURRENT_POLARITY                !   Polarity of filament we're currently checking
    INTEGER                                 ::  CURRENT_PHI_POLARITY            !
    INTEGER                                 ::  BEST_POLARITY   =   1           !   Keep track of best polarity for current filament
    INTEGER                                 ::  BEST_PHI_POLARITY(2)    =   1   !   For each filament polarity, the best phi polarity
    ! Imagefile objects
    TYPE(IMAGEFILE)                         ::  FILM_FILE
    TYPE(IMAGEFILE)                         ::  THREED_FILE
    !TYPE(IMAGEFILE)                         ::  SLICES_FILE
    !
    LOGICAL                                 ::  RIGID_PARAMS(7)
    LOGICAL                                 ::  HOLD_PARAM_AVERAGE(7)
    LOGICAL                                 ::  REFINE_PARAM_INCREMENTS(7)
    LOGICAL                                 ::  INTEGER_CHANGES_ONLY(7)
    TYPE(POINT_SYMMETRY_SYMBOL)             ::  SYM_SYMBOL


    !
    ! Get variable values from control file
    !


    !
    ! Start work
    !

    ! Memory footprint diag
    IF (DEBUG) THEN
        CALL PRINT_MEMORY_USAGE_STATS()
    ENDIF

    ! This is a helical mode
    SYM_SYMBOL%SYMBOL = 'H'

    !
    ! Sanity check: target resolution cannot be lower than the Nyquist!
    !
    IF (TARGET_RESOLUTION%VALUE .LT. PIXEL_SIZE%VALUE * 2.0) THEN
        WRITE(*,'(A,F0.2,A,F0.3,A)')    '**ERROR(MODE_FILAMENT_REFINE): target resolution (', TARGET_RESOLUTION%VALUE, &
                                        ') is incompatible with pixel size (', PIXEL_SIZE%VALUE, ')'
        CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE')
    ENDIF

    !
    ! Now we know the pixel size, convert the PSF_RADIUS to PSF_RADIUS_PIXELS
    !
    WRITE(*,'(A,F0.1,A)') '**info(MODE_FILAMENT_REFINE): will take PSF radius to be ', PSF_RADIUS, ' A'
    PSF_RADIUS_PIXELS = PSF_RADIUS / PIXEL_SIZE%VALUE

    !
    ! Initialise the interpolation scheme
    !
    CALL INTERPOLATION_SCHEME_INIT(INTERPOL_SCHEME,INTERPOL_FUNCTION%VALUE,WINDOW_FUNCTION%VALUE,WINDOW_WIDTH%VALUE)


    !
    ! Work out a safe box size, bearing in mind the maximum object radius
    !
    MIN_SAFE_BOX_SIZE = FMC_BOX_SIZE(INTERPOL_SCHEME,HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE)


    !
    ! Are we given CTFFIND files or CTFTILT files?
    !
    USE_CTFTILT = .FALSE.
    IF (STRING_EQUAL(CTFTILT_FILES%VALUE,'none')) THEN
        IF (STRING_EQUAL(CTFFIND_FILES%VALUE,'none')) THEN
            WRITE(*,'(A)') '**ERROR(MODE_FILAMENT_REFINE): neither CTFFIND nor CTFTILT files were supplied'
            CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE')
        ELSE
            USE_CTFTILT = .FALSE.
        ENDIF
    ELSE
        USE_CTFTILT = .TRUE.
    ENDIF
    ! Diagnostic
    IF (.NOT. USE_CTFTILT) THEN
        PRINT '(A)', '**info(MODE_FILAMENT_REFINE): no CTFTILT files given, will use CTFFIND info'
    ENDIF

    !
    ! Will we be checking the polarity of the filament?
    !
    IF (POLARITY_CHECK%VALUE) THEN
        NUM_POLARITIES_TO_TEST = 2
    ELSE
        NUM_POLARITIES_TO_TEST = 1
    ENDIF

    !
    ! Will we be checking the phi polarity of the filament?
    !
    IF (PHI_POLARITY_CHECK%VALUE) THEN
        NUM_PHI_POLARITIES_TO_TEST = 2
    ELSE
        NUM_PHI_POLARITIES_TO_TEST = 1
    ENDIF

    !
    ! Which parameters will we refine
    !
    REFINE_PARAMS = [REFINE_X%VALUE,REFINE_Y%VALUE,REFINE_Z%VALUE,  &
                    REFINE_PHI%VALUE,REFINE_THETA%VALUE,REFINE_PSI%VALUE, &
                    REFINE_SUBUNIT_NUM%VALUE]

    ! Can only refine psi & theta when using Hermite splines
    IF ((REFINE_PSI%VALUE .OR. REFINE_THETA%VALUE) .AND. .NOT. USE_HERMITE_SPLINES%VALUE) THEN
        WRITE(*,'(A)') '**ERROR(MODE_FILAMENT_REFINE): psi and theta can only be refined when Hermite splines are used'
        CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE. Bad choice of parameters.')
    ENDIF

    !
    ! Which parameters will be held rigid (per-filament rather than per-waypoint)
    !
    RIGID_PARAMS = [RIGID_X%VALUE,RIGID_Y%VALUE,RIGID_Z%VALUE,  &
                    RIGID_PHI%VALUE,RIGID_THETA%VALUE,RIGID_PSI%VALUE,  &
                    RIGID_SUBUNIT_NUM%VALUE]

    !
    ! Will we be holding the average value for any parameter constant?
    !
    HOLD_PARAM_AVERAGE = [.FALSE.,.FALSE.,HOLD_Z_AVERAGE%VALUE,.FALSE.,.FALSE.,.FALSE.,.FALSE.]

    !
    ! Will we refine subunit_num wp-on-wp increments rather than absolute values?
    !
    REFINE_PARAM_INCREMENTS = (/ .FALSE., .FALSE., .FALSE., .FALSE., .FALSE., .FALSE., REFINE_SUBUNIT_NUM_INCREMENTS%VALUE /)

    !
    ! Will we allow only integer-valued changes during refinement?
    !
    INTEGER_CHANGES_ONLY = (/ .FALSE., .FALSE., .FALSE., .FALSE., .FALSE., .FALSE., SUBUNIT_NUM_INTEGER_CHANGES_ONLY%VALUE /)

    IF (ANY(INTEGER_CHANGES_ONLY)) THEN
        IF (.NOT. SKIP_REFINEMENT%VALUE) THEN
            IF (.NOT. USE_DIFFERENTIAL_EVOLUTION%VALUE) THEN
                WRITE(*,'(A)') '**ERROR(MODE_FILAMENT_REFINE): integer-only changes in subunit_num are only implemented '// &
                                'in conjunction with the differential evolution minimiser'
                CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE. Bad choice of parameters.')
            ENDIF
            IF (REFINE_NORMAL_MODES%VALUE) THEN
                WRITE(*,'(A)') '**ERROR(MODE_FILAMENT_REFINE): integer-only changes in subunit_num are not '// &
                                'compatible with normal mode refinement'
                CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE. Bad choice of parameters.')
            ENDIF
        ENDIF
    ENDIF

    ! It doesn't make sense to refine the defocus unless we are aiming for high resolution
    IF (REFINE_Z%VALUE .AND. .NOT. HOLD_Z_AVERAGE%VALUE) THEN
        IF (TARGET_RESOLUTION%VALUE .GT. 5.0) THEN
            WRITE(*,'(A,I0,A)') '**warning(MODE_FILAMENT_REFINE): refining Z without fixing the average Z per filament amounts '// &
                            'to defocus refinement, but your target resolution is ', TARGET_RESOLUTION%VALUE, ', which may be '// &
                            'too low. You may want to use option HOLD_Z_AVERAGE'
        ENDIF
    ENDIF

    ! I don't think it is possible to do sequential refinement of waypoints while refining a rigid parameter (I'm not sure)
    IF (ANY(REFINE_PARAMS .AND. RIGID_PARAMS) .AND. .NOT. REFINE_WPS_SIMULTANEOUSLY%VALUE) THEN
        WRITE(*,'(A)') '**ERROR(MODE_FILAMENT_REFINE): I do not know how to refine waypoints sequentially '//&
                                                     'while also refining a rigid parameters. Not implemented.'
        CALL TERMINATE('Fatal error in in MODE_FILAMENT_REFINE. Feature not implemented.')
    ENDIF

    ! Normal modes refinement implies all waypoints are refined simultaneously
    IF (.NOT. REFINE_WPS_SIMULTANEOUSLY%VALUE .AND. REFINE_NORMAL_MODES%VALUE) THEN
        WRITE(*,'(A)') '**ERROR(MODE_FILAMENT_REFINE): normal mode refinement requires that waypoints be refined simultaneously'
        CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE. Bad choice of run-time parameter values.')
    ENDIF

    !
    ! Any search half-range override from the user
    !
    SEARCH_HALF_RANGE_OVERRIDE = [  SEARCH_HALF_RANGE_X%VALUE,SEARCH_HALF_RANGE_Y%VALUE,SEARCH_HALF_RANGE_Z%VALUE,  &
                                    SEARCH_HALF_RANGE_PHI%VALUE,SEARCH_HALF_RANGE_THETA%VALUE,SEARCH_HALF_RANGE_PSI%VALUE,&
                                    SEARCH_HALF_RANGE_SUBUNIT_NUM%VALUE   ]

    !
    ! Work out the outer radius of helix (pixels)
    !
    HELIX_MAX_RADIUS_PIX = HELIX_MAX_RADIUS%VALUE / PIXEL_SIZE%VALUE

    !
    ! Check for incompatible settings of CONSTANT_RISE_PER_SUBUNIT%VALUE and RIGID_SUBUNIT_NUM%VALUE
    !
    IF (CONSTANT_RISE_PER_SUBUNIT%VALUE .AND. .NOT. (RIGID_SUBUNIT_NUM%VALUE .OR. FRAGMENTED_SUBUNIT_NUM%VALUE)) THEN
        WRITE(*,'(2A)') '**ERROR(MODE_FILAMENT_REFINE): you have CONSTANT_RISE_PER_SUBUNIT ',&
                        'set to true and RIGID_SUBUNIT_NUM set to false. These settings are incompatible.'
        WRITE(*,'(2A)') '**ERROR(MODE_FILAMENT_REFINE): if you want to keep a constant rise per subunit,',   &
                        ' you should set RIGID_SUBUNIT_NUM to true.'
        CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE')
    ENDIF

    !
    ! Which parameters will we refine via their normal modes?
    !
    RREFINE_NORMAL_MODES = .FALSE.
    IF (REFINE_NORMAL_MODES%VALUE) THEN
        RREFINE_NORMAL_MODES = .TRUE.
        IF (FRAGMENTED_SUBUNIT_NUM%VALUE) THEN
            RREFINE_NORMAL_MODES(7) = .FALSE.
        ENDIF
        IF (FRAGMENTED_PHI%VALUE) THEN
            RREFINE_NORMAL_MODES(4) = .FALSE.
        ENDIF
    ENDIF

    !
    ! We need to know the average score if we're to do B-factor filtering during
    ! 3D reconstruction
    !
    IF (BFACTOR_DURING_3D_RECONSTRUCTION%VALUE) THEN
        IF (.NOT. AVERAGE_SCORE%SET_BY_USER) THEN
            WRITE(*,'(2A)') '**ERROR(MODE_FILAMENT_REFINE): average score not specified on input. ', &
                            'Needed for B factor calculation during 3D reconstruction.'
            CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE')
        ENDIF
        IF (PBC .EQ. 0.0) THEN
            WRITE(*,'(A)') '**ERROR(MODE_FILAMENT_REFINE): PBC value cannot be 0.0. Please set it in the input parameter file.'
            WRITE(*,'(A)') 'Typical PBC values may range from 5.0 (strong weighting) to 100.0 (very weak weighting).'
            CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE')
        ENDIF
    ENDIF

    !
    ! Initialise the average helical parameters
    !
    CALL HELICAL_SYMMETRY_INIT(HELIX_AVERAGE,RISE_PER_SUBUNIT%VALUE,TWIST_PER_SUBUNIT%VALUE,    &
                                SYMMETRY_AXIAL%VALUE,SYMMETRY_PERP%VALUE,PIXEL_SIZE%VALUE,NUM_STARTS=NUM_HELICAL_STARTS%VALUE, &
                                STACK=STACK%VALUE)
    WRITE(*,'(A)',ADVANCE='NO') '**info(MODE_FILAMENT_REFINE): initialised helical symmetry operator: '
    CALL HELICAL_SYMMETRY_PRINT_INFO(HELIX_AVERAGE)

    !
    ! Read in the frealix parameter file and intialise an array of filament objects
    !
    WRITE(*,'(A)') '<parse_ali_file>'
    WRITE(*,'(/3A)') '**info(MODE_FILAMENT_REFINE): reading input parameter file (', &
                                            ADJUSTL(TRIM(INPUT_PAR_FILENAME%VALUE)),') and initialising filaments'
    IF (USE_CTFTILT) THEN
        CALL FILAMENTS_FROM_FREALIX_PAR(    FILS,INPUT_PAR_FILENAME%VALUE,FILM_FILES%VALUE,PIXEL_SIZE%VALUE,HELIX_AVERAGE,  &
                                            CTFTILT_FILENAME_WILD = CTFTILT_FILES%VALUE,                                    &
                                            WPS_ARE_CROSSOVERS = WPS_ARE_CROSSOVERS%VALUE,                                  &
                                            FAL_WPS_ARE_CROSSOVERS = FAL_WPS_ARE_CROSSOVERS%VALUE,                          &
                                            MAX_RADIUS = HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE,                           &
                                            MICROTUBULE=MICROTUBULE%VALUE,MT_PF_NUMBER=MT_PF_NUMBER%VALUE,MT_RISE = MT_RISE%VALUE,&
                                            RIGID_PARAMS = RIGID_PARAMS,                                                    &
                                            CONSTANT_RISE_PER_SUBUNIT = CONSTANT_RISE_PER_SUBUNIT%VALUE,                    &
                                            CONSTANT_TWIST_PER_LENGTH = CONSTANT_TWIST%VALUE,                               &
                                            WPS_AS_SPS = WPS_AS_SPS%VALUE,REFINE_PARAMS = REFINE_PARAMS,                    &
                                            PERSISTENCE_LENGTH = PERSISTENCE_LENGTH%VALUE/PIXEL_SIZE%VALUE,                       &
                                            TORSION_PERSISTENCE_LENGTH = TORSION_PERSISTENCE_LENGTH%VALUE/PIXEL_SIZE%VALUE,       &
                                            STRETCH_PERSISTENCE_LENGTH = STRETCH_PERSISTENCE_LENGTH%VALUE/PIXEL_SIZE%VALUE,       &
                                            FAL_WPS_IN_SCORE = FAL_WPS_IN_SCORE%VALUE,FAL_WPS_IN_THREED = FAL_WPS_IN_THREED%VALUE,&
                                            REFINE_NORMAL_MODES = RREFINE_NORMAL_MODES,                                     &
                                            AVERAGE_Z_TO_CTFTILT_PLANE = AVERAGE_Z_TO_CTFTILT_PLANE%VALUE,                  &
                                            USE_HERMITE_SPLINES = USE_HERMITE_SPLINES%VALUE,                                &
                                            FRAGMENTED_PHI = FRAGMENTED_PHI%VALUE,                                          &
                                            FRAGMENTED_SUBUNIT_NUM = FRAGMENTED_SUBUNIT_NUM%VALUE,                          &
                                            AMYLOID_BETA=AMYLOID_BETA%VALUE)
    ELSE
        CALL FILAMENTS_FROM_FREALIX_PAR(    FILS,INPUT_PAR_FILENAME%VALUE,FILM_FILES%VALUE,PIXEL_SIZE%VALUE,HELIX_AVERAGE,  &
                                            CTFFIND_FILENAME_WILD = CTFFIND_FILES%VALUE,                                    &
                                            WPS_ARE_CROSSOVERS = WPS_ARE_CROSSOVERS%VALUE,                                  &
                                            FAL_WPS_ARE_CROSSOVERS = FAL_WPS_ARE_CROSSOVERS%VALUE,                          &
                                            MAX_RADIUS = HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE,                           &
                                            MICROTUBULE =MICROTUBULE%VALUE,MT_PF_NUMBER=MT_PF_NUMBER%VALUE,MT_RISE=MT_RISE%VALUE,  &
                                            RIGID_PARAMS = RIGID_PARAMS,                                                    &
                                            CONSTANT_RISE_PER_SUBUNIT = CONSTANT_RISE_PER_SUBUNIT%VALUE,                    &
                                            CONSTANT_TWIST_PER_LENGTH = CONSTANT_TWIST%VALUE,                               &
                                            WPS_AS_SPS = WPS_AS_SPS%VALUE,REFINE_PARAMS =REFINE_PARAMS,                     &
                                            PERSISTENCE_LENGTH = PERSISTENCE_LENGTH%VALUE/PIXEL_SIZE%VALUE,                       &
                                            TORSION_PERSISTENCE_LENGTH = TORSION_PERSISTENCE_LENGTH%VALUE/PIXEL_SIZE%VALUE,       &
                                            STRETCH_PERSISTENCE_LENGTH = STRETCH_PERSISTENCE_LENGTH%VALUE/PIXEL_SIZE%VALUE,       &
                                            FAL_WPS_IN_SCORE = FAL_WPS_IN_SCORE%VALUE,FAL_WPS_IN_THREED = FAL_WPS_IN_THREED%VALUE,&
                                            REFINE_NORMAL_MODES = RREFINE_NORMAL_MODES,                                     &
                                            AVERAGE_Z_TO_CTFTILT_PLANE = AVERAGE_Z_TO_CTFTILT_PLANE%VALUE,                  &
                                            USE_HERMITE_SPLINES = USE_HERMITE_SPLINES%VALUE,                                &
                                            FRAGMENTED_PHI = FRAGMENTED_PHI%VALUE,                                          &
                                            FRAGMENTED_SUBUNIT_NUM = FRAGMENTED_SUBUNIT_NUM%VALUE,                          &
                                            AMYLOID_BETA=AMYLOID_BETA%VALUE)
    ENDIF
    WRITE(*,'(A,I0,A/)') '**info(MODE_FILAMENT_REFINE): ', SIZE(FILS), ' filaments were initialised'
    WRITE(*,'(A)') '</parse_ali_file>'



    !
    ! Allocate memory to array of final filament states
    !
    ALLOCATE(FILS_FINAL(SIZE(FILS)),STAT=IERR)
    IF (IERR .NE. 0) THEN
        WRITE(*,'(A,I0)') '**ERROR(MODE_FILAMENT_REFINE): memory allocation for fils_final array failed with error ', IERR
        CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE')
    ENDIF

    !
    ! Allocate memory to array which will keep track of scores & positions as we minimise
    !
    ! The score array will keep track of the score for for each oritentation (flip) and for each iteration of the refinement
    !
    ALLOCATE(       SCORE(NUM_POLARITIES_TO_TEST,NUM_PHI_POLARITIES_TO_TEST,2),        &
                    FIL_STATES(NUM_POLARITIES_TO_TEST,NUM_PHI_POLARITIES_TO_TEST,2),   &
                    STAT=IERR   )
    IF (IERR .NE. 0) THEN
        WRITE(*,'(A,I0)') '**ERROR(MODE_FILAMENT_REFINE): memory allocation for score array failed with error ', IERR
        CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE')
    ENDIF

    ! Memory footprint diag
    IF (DEBUG) THEN
        WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage after filaments, scores etc, but before 3Ds'
        CALL PRINT_MEMORY_USAGE_STATS()
    ENDIF


    !
    ! Read the 3D model from disk
    !
    WRITE(*,'(A)') '<read_model_3d_from_disk>'
    IF (MODEL_3D_AVAILABLE%VALUE) THEN
        CALL THREED_FILE%INIT(MODEL_3D_FILENAME%VALUE,PIXEL_SIZE=PIXEL_SIZE%VALUE)
        call MODEL_3DVOL%ReadFromDisk(THREED_FILE,read_volume_not_slice=.true.)
        CALL THREED_FILE%CLOSE()
        ! Check the 3D model is indeed 3D
        IF (ANY(MODEL_3DVOL%logical_dimensions .LE. 1)) THEN
            WRITE(*,'(A,3(I0,1X))') '**ERROR(MODE_FILAMENT_REFINE): input 3D model is not 3D! Dimensions = ', &
                                    MODEL_3DVOL%logical_dimensions
            CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE')
        ENDIF
        ! For now, we assume the 3D model will be cube
        IF (.NOT. IS_SQUARE(MODEL_3DVOL)) THEN
            WRITE(*,'(A,3I5)')  '**ERROR(MODE_FILAMENT_REFINE): sorry, 3D model must be in a cubic box. '//&
                                'Image supplied is ', MODEL_3DVOL%logical_dimensions
            CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE')
        ENDIF
        ! Also read in the full 3D reconstruction information (required when SUBTRACT_FIL_FROM_REF%VALUE)
        IF (SUBTRACT_FIL_FROM_REF%VALUE .OR. SINGLE_FILAMENT_REFINEMENT%VALUE) THEN
            CALL THREED_INIT(   REF_3DREC,                                              &
                                FILENAME_ROOT       =   MODEL_3D_FILENAME%VALUE,        &
                                INTERPOL_SCHEME     =   INTERPOL_SCHEME,                &
                                OVERSAMPLING_FACTOR =   1,                              &
                                SYM_SYMBOL          =   SYM_SYMBOL,                     &
                                HELIX               =   HELIX_AVERAGE,                  &
                                ASYM_UNIT_MW        =   ASYM_UNIT_MW%VALUE,             &
                                PIXEL_SIZE          =   PIXEL_SIZE%VALUE,               &
                                HELIX_MAX_RAD       =   HELIX_MAX_RADIUS_PIX,           &
                                COMPUTE_FSC         =   CALC_FSC%VALUE,                 &
                                USE_SSNR_ESTIMATE   =   ESTIMATE_SSNR%VALUE,            &
                                FOM_FILTER          =   APPLY_FOM_FILTER%VALUE,         &
                                RECONSTRUCT_AMPLITUDES = RECONSTRUCT_AMPLITUDES%VALUE )
        ENDIF
    ELSE
        IF (DEBUG) WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): model 3D not available, therefore not read in!'
        IF (.NOT. SKIP_REFINEMENT%VALUE) THEN
            WRITE(*,'(A)') '**info(MODE_FILAMENT_REFINE): no model 3D is available, so no refinement will be done'
            SKIP_REFINEMENT%VALUE = .TRUE.
        ENDIF
        IF (WRITE_MATCHING_PROJECTIONS%VALUE) THEN
            WRITE(*,'(A)') '**warning(MODE_FILAMENT_REFINE): since no model 3D volume is available, '// &
                            'no matching projections will be generated'
            WRITE_MATCHING_PROJECTIONS%VALUE = .FALSE.
        ENDIF
    ENDIF
    WRITE(*,'(A)') '</read_model_3d_from_disk>'

    ! Memory footprint diagnostics
    IF (DEBUG) THEN
        WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage after 3D read from file'
        CALL PRINT_MEMORY_USAGE_STATS()
    ENDIF

    !
    ! If necessary, initialise a 3D reconstruction object
    !
    WRITE(*,'(A)') '<initalise_3d_reconstruction>'
    IF (CALC_3D_RECONSTRUCTION%VALUE) THEN
        WRITE(*,'(A)') '**info(MODE_FILAMENT_REFINE): initialising output 3D reconstruction'
        NEW_3DREC%logical_dimensions(1:3) = FMC_THREED_SIZE(INTERPOL_SCHEME,HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE)
        CALL THREED_INIT(   NEW_3DREC,                                              &
                            INTERPOL_SCHEME     =   INTERPOL_SCHEME,                &
                            SYM_SYMBOL          =   SYM_SYMBOL,                     &
                            HELIX               =   HELIX_AVERAGE,                  &
                            ASYM_UNIT_MW        =   ASYM_UNIT_MW%VALUE,             &
                            PIXEL_SIZE          =   PIXEL_SIZE%VALUE,               &
                            HELIX_MAX_RAD       =   HELIX_MAX_RADIUS_PIX,           &
                            COMPUTE_FSC         =   CALC_FSC%VALUE,                 &
                            USE_SSNR_ESTIMATE   =   ESTIMATE_SSNR%VALUE,            &
                            FOM_FILTER          =   APPLY_FOM_FILTER%VALUE,         &
                            RECONSTRUCT_AMPLITUDES = RECONSTRUCT_AMPLITUDES%VALUE )
        ! Memory footprint diag
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage after  new 3drec'
            CALL PRINT_MEMORY_USAGE_STATS()
        ENDIF
        ! Initialise another 3D rec object, to be used temporarily
        WRITE(*,'(A)') '**info(MODE_FILAMENT_REFINE): initialising temporary 3D reconstruction'
        CURRENT_3DREC%logical_dimensions(1:3) = FMC_THREED_SIZE(INTERPOL_SCHEME,HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE)
        CALL THREED_INIT(   CURRENT_3DREC,                                          &
                            INTERPOL_SCHEME     =   INTERPOL_SCHEME,                &
                            SYM_SYMBOL          =   SYM_SYMBOL,                     &
                            HELIX               =   HELIX_AVERAGE,                  &
                            ASYM_UNIT_MW        =   ASYM_UNIT_MW%VALUE,             &
                            PIXEL_SIZE          =   PIXEL_SIZE%VALUE,               &
                            HELIX_MAX_RAD       =   HELIX_MAX_RADIUS_PIX,           &
                            COMPUTE_FSC         =   CALC_FSC%VALUE,                 &
                            USE_SSNR_ESTIMATE   =   ESTIMATE_SSNR%VALUE,            &
                            FOM_FILTER          =   APPLY_FOM_FILTER%VALUE,         &
                            RECONSTRUCT_AMPLITUDES = RECONSTRUCT_AMPLITUDES%VALUE      )
        ! Memory footprint diag
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage after  current 3drec'
            CALL PRINT_MEMORY_USAGE_STATS()
        ENDIF
        IF (SINGLE_FILAMENT_RECONSTRUCTIONS%VALUE) THEN
            CURRENT_3DVOL_SLICE%logical_dimensions(1:2) = FMC_THREED_SIZE(INTERPOL_SCHEME,HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE)
            CURRENT_3DVOL_SLICE%logical_dimensions(3) = 1
            CALL IMAGE_ALLOCATE(CURRENT_3DVOL_SLICE)
        ENDIF
    ENDIF
    WRITE(*,'(A)') '</initalise_3d_reconstruction>'

    ! Memory footprint diag
    IF (DEBUG) THEN
        WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage after 3Ds loaded'
        CALL PRINT_MEMORY_USAGE_STATS()
    ENDIF


    !
    ! If we're in "single-particle mode" (i.e. where waypoints are treated as independent single particles), let the user know
    !
    IF (WPS_AS_SPS%VALUE) THEN
        WRITE(*,'(2A)') '**info(MODE_FILAMENT_REFINE): will be treating waypoints as independent single particles, ',    &
                        'so full-filament scoring & refinement will not take place'
    ENDIF

    !
    ! Loop over filaments
    !
    WRITE(*,'(A,I0,A)') '<loop_over_filaments num_of_fils=',SIZE(FILS),'>'
    DO CURRENT_FILAMENT=1,SIZE(FILS)

        WRITE(*,'(A,I0,A)') '<filament num=',CURRENT_FILAMENT,'>'

        ! Did the user ask us to straighten the filament on entry?
        IF (STRAIGHTEN_FILAMENT_BEFORE_REFINEMENT%VALUE) THEN
            WRITE(*,'(A)') '**info(MODE_FILAMENT_REFINE): straightening filament before refinement begins'
            CALL FILAMENT_STRAIGHTEN(FILS(CURRENT_FILAMENT))
        ENDIF

        ! Experimental: tweak Z coordinates to minimise discrepency from canonical helical operator
        IF (SP_TO_FF%VALUE) THEN
			IF (.NOT. WPS_AS_SPS%VALUE) THEN
				WRITE(*,'(A)') '**ERROR(MODE_FILAMENT_REFINE): SP_TO_FF flag can only be used when WPS_AS_SPS is true'
				WRITE(*,'(A)') '**ERROR(MODE_FILAMENT_REFINE): it is broken otherwise'
				CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE')
			ENDIF
			CALL FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX(FILS(CURRENT_FILAMENT))
        ENDIF

        ! If requested, add new waypoints at either end of the filament
        IF (ADD_FAL_WPS%VALUE) THEN
            WRITE(*,'(A,I0)') '**info(MODE_FILAMENT_REFINE): adding waypoints to the ends of filament ', FILS(CURRENT_FILAMENT)%ID
            CALL FILAMENT_ADD_FAL_WPS(FILS(CURRENT_FILAMENT))
        ENDIF

        ! If requested, resample waypoints along the path of the filament
        IF (FORCE_WP_TO_WP_DISTANCE%VALUE .GT. 0.0) THEN
            CALL FILAMENT_RESAMPLE_WAYPOINTS(FILS(CURRENT_FILAMENT),REAL(FORCE_WP_TO_WP_DISTANCE%VALUE/PIXEL_SIZE%VALUE,KIND=8))
        ENDIF
        IF (FORCE_WP_TO_WP_PHI%VALUE .GT. 0.0) THEN
            CALL FILAMENT_RESAMPLE_WAYPOINTS(FILS(CURRENT_FILAMENT),REAL(FORCE_WP_TO_WP_PHI%VALUE,KIND=8),STEP_IN_PHI=.TRUE.)
        ENDIF
        IF (FORCE_WP_TO_WP_DISTANCE%VALUE .GT. 0.0 .AND. FORCE_WP_TO_WP_PHI%VALUE .GT. 0.0) THEN
            WRITE(*,'(A)') '**ERROR(MODE_FILAMENT_REFINE): force_wp_to_wp_distance and force_wp_to_wp_phi can''t be used together'
            CALL TERMINATE('Fatal error in MODE_FILAMENT_REFINE. Bad choice of parameters.')
        ENDIF


        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): filament info with statistics follows'
            CALL FILAMENT_PRINT_INFO(FILS(CURRENT_FILAMENT))
            CALL FILAMENT_PRINT_EXTRA_INFO(FILS(CURRENT_FILAMENT))
        ENDIF


        ! Will we need to read a new film image from disk?
        IF (CURRENT_FILAMENT .EQ. 1) THEN
            NEW_FILM    =   .TRUE.
        ELSE IF (FILS(CURRENT_FILAMENT)%FILM%FILENAME .NE. FILS(CURRENT_FILAMENT-1)%FILM%FILENAME) THEN
            NEW_FILM    =   .TRUE.
        ELSE IF (FILS(CURRENT_FILAMENT)%FILM%LOCATION_IN_FILE .NE. FILS(CURRENT_FILAMENT-1)%FILM%LOCATION_IN_FILE) THEN
            NEW_FILM    =   .TRUE.
        ELSE
            NEW_FILM    =   .FALSE.
        ENDIF
        IF (CALC_3D_RECONSTRUCTION%VALUE .OR. SUBTRACT_FIL_FROM_REF%VALUE .OR. SINGLE_FILAMENT_REFINEMENT%VALUE) THEN
            NEW_FILM = .TRUE. ! This is because when 3D reconstruction is computed, the film is read in again but not filtered
        ENDIF
        ! Note that some time could be saved by working out whether the correct film image is alread loaded in memory and
        ! therefore only the filtering remains to be done. However, to keep the code simpler for now, I sacrifice some run time to IO.
        ! In the case where we subtract fil 3D from reference, this will be done by FMC_INIT or FMC_NEW_FILAMENT, so the film
        ! must be bandpassed after that.

        ! If necessary, read a new film image from disk & high-pass filter it
        FILM_HP_PARAM =   PIXEL_SIZE%VALUE / (3.0 * HELIX_MAX_RADIUS%VALUE)
        IF (NEW_FILM .AND. MODEL_3D_AVAILABLE%VALUE) THEN
            IF (DEBUG) PRINT '(2A)', '**debug(MODE_FILAMENT_REFINE): reading micrograph ', &
                                    TRIM(ADJUSTL(FILS(CURRENT_FILAMENT)%FILM%FILENAME))
            CALL FILM_FILE%INIT(FILS(CURRENT_FILAMENT)%FILM%FILENAME)
            CALL FILM_FILE%PrintInfo()
            CALL CURRENT_FILM_IMAGE%ReadFromDisk(FILM_FILE,FILS(CURRENT_FILAMENT)%FILM%LOCATION_IN_FILE)
            CALL FILM_FILE%CLOSE()
            IF (REVERSE_FILM_CONTRAST%VALUE) THEN
                CURRENT_FILM_IMAGE%real_values = CURRENT_FILM_IMAGE%real_values * (-1.0)
            ENDIF
        ENDIF

        ! Memory footprint diag
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage after film'
            CALL PRINT_MEMORY_USAGE_STATS()
        ENDIF


        ! Initialise a filament-model comparison (FMC) object if necessary
        IF (MODEL_3D_AVAILABLE%VALUE) THEN
            IF (.NOT. FMC%INITIALISED) THEN
                CALL FMC_INIT(                                                                                  &
                                FMC,FILS(CURRENT_FILAMENT),CURRENT_FILM_IMAGE,                                  &
                                MODEL_3DVOL,                                                                    &
                                HELIX_AVERAGE,                                                                  &
                                TARGET_RESOLUTION           =   TARGET_RESOLUTION%VALUE,                        &
                                INTERPOL                    =   INTERPOL_SCHEME,                                &
                                ASYM_UNIT_MW                =   ASYM_UNIT_MW%VALUE,                             &
                                REAL_SPACE_PROJ             =   REAL_SPACE_PROJECTION%VALUE,                    &
                                SUBTRACT_FIL_FROM_REF       =   SUBTRACT_FIL_FROM_REF%VALUE,                    &
                                MODEL3D                     =   REF_3DREC,                                      &
                                WPS_AS_SPS                  =   WPS_AS_SPS%VALUE,                               &
                                SP_SEGMENT_LENGTH           =   SP_SEGMENT_LENGTH%VALUE,                        &
                                SEARCH_HALF_RANGE_OVERRIDE  =   SEARCH_HALF_RANGE_OVERRIDE,                     &
                                NUM_MINIM_RUNS              =   NUM_MINIM_RUNS%VALUE,                           &
                                DS_RANDOM_START_CONF        =   DS_RANDOM_START_CONF%VALUE,                     &
                                MINIM_RANDOM_START_POS      =   MINIM_RANDOM_START_POS%VALUE,                   &
                                DS_MAX_NUM_ITER             =   DS_MAX_NUMBER_OF_ITERATIONS%VALUE,              &
                                PATH_OVERSAMPLING           =   PATH_OVERSAMPLING%VALUE,                        &
                                LOW_RESOLUTION_CUTON        =   MIN(LOW_RESOLUTION_CUTON%VALUE,PIXEL_SIZE%VALUE/FILM_HP_PARAM), &
                                HIGH_RESOLUTION_CUTOFF      =   HIGH_RESOLUTION_CUTOFF%VALUE,                   &
                                USE_CC3                     =   USE_WEIGHTED_CORRELATION%VALUE,                 &
                                CUBED_AVERAGE               =   USE_AVERAGE_OF_CUBED_SEGMENT_SCORES%VALUE,      &
                                CURVATURE_RESTRAINT         =   USE_CURVATURE_RESTRAINT%VALUE,                  &
                                TORSION_RESTRAINT           =   USE_TORSION_RESTRAINT%VALUE,                    &
                                STRETCH_RESTRAINT           =   USE_STRETCH_RESTRAINT%VALUE,                    &
                                MODEL_AVE_CYL               =   AVERAGE_MODEL_CYLINDRICALLY%VALUE,              &
                                CACHE_PROJECTIONS           =   CACHE_PROJECTIONS%VALUE,                        &
                                MODEL3DVOL_BFACTOR          =   MODEL_3D_FILTER_BFACTOR%VALUE,                  &
                                HOLD_PARAM_AVERAGE          =   HOLD_PARAM_AVERAGE,                             &
                                USE_CONJUGATE_GRADIENT      =   USE_CONJUGATE_GRADIENT%VALUE,                   &
                                USE_DIFFERENTIAL_EVOLUTION  =   USE_DIFFERENTIAL_EVOLUTION%VALUE,               &
                                REFINE_PARAM_INCREMENTS     =   REFINE_PARAM_INCREMENTS,                        &
                                INTEGER_CHANGES_ONLY        =   INTEGER_CHANGES_ONLY,                           &
                                CROSS_BETA_FILTER           =   CROSS_BETA_FILTER%VALUE,                        &
                                REFINE_WPS_SIMULTANEOUSLY   =   REFINE_WPS_SIMULTANEOUSLY%VALUE,                &
                                SINGLE_FILAMENT_REFINEMENT  =   SINGLE_FILAMENT_REFINEMENT%VALUE                &
                            )
            ELSE
                ! The FMC object is already initialised
                FMC%INTERPOL = INTERPOL_SCHEME
                CALL FMC_NEW_FILAMENT(FMC,FILS(CURRENT_FILAMENT))
                IF (FMC%SUBTRACT_FIL_FROM_REF .OR. FMC%SINGLE_FILAMENT_REFINEMENT) THEN
                    ! Compute the reconstruction of the filament from the original state, so that it can later
                    ! be subtracted from the reference 3D
                    CALL FMC_UPDATE_FIL_THREED(FMC)
                ENDIF
            ENDIF
            CALL FMC_PRINT_INFO(FMC)
        ENDIF

        IF (NEW_FILM  .AND. MODEL_3D_AVAILABLE%VALUE) THEN
            ! High-pass filter the film image to bring average density to 0.0 and remove density ramps
            WRITE(*,'(A,F0.2,A)') '**info(MODE_FILAMENT_REFINE): high-pass filtering the micrograph. Resolution = ', &
                                    PIXEL_SIZE%VALUE/FILM_HP_PARAM, ' A'
            CALL BAND_PASS(CURRENT_FILM_IMAGE,FILM_HP_PARAM,0.0,FILM_HP_PARAM/2.0)
        ENDIF

        IF (DEBUG .AND. MODEL_3D_AVAILABLE%VALUE) THEN
            WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): filament info with statistics follows (2)'
            CALL FILAMENT_PRINT_INFO(FMC%FIL)
            CALL FILAMENT_PRINT_EXTRA_INFO(FMC%FIL)
        ENDIF

        ! Memory footprint diag
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage after FMC'
            CALL PRINT_MEMORY_USAGE_STATS()
        ENDIF


        !
        ! If none of the waypoints of the current filament are marked as to-be-refined, we can skip to the next filament
        !
        IF (.NOT. ANY(FILS(CURRENT_FILAMENT)%WP%REFINE) .OR. SKIP_REFINEMENT%VALUE) THEN

            WRITE(*,'(A,I0,A)') '**info(MODE_FILAMENT_REFINE): none of the waypoints of filament ', FILS(CURRENT_FILAMENT)%ID, &
                           ' need to be refined, or you indicated that refinement should be skipped. Skipping filament refinement.'
            FMC%INTERPOL = INTERPOL_SCHEME
            CALL FMC_NEW_FILAMENT(FMC,FILS(CURRENT_FILAMENT))
            CALL FILAMENT_ASSIGN(FILS_FINAL(CURRENT_FILAMENT),FILS(CURRENT_FILAMENT))

        !
        ! If at least one of the waypoints of the current filament are marked as to-be-refined
        !
        ELSE
            !
            ! Loop over filament polarities
            !
            !
            ! TODO: introduce OpenMP parallelisation around polarity loops?
            !
            WRITE(*,'(A)') '<polarity_loop>'
            POLARITY_LOOP : DO CURRENT_POLARITY=1,NUM_POLARITIES_TO_TEST

                WRITE(*,'(A,I0,A)') '<polarity num=', CURRENT_POLARITY, '>'
                ! Let the user know if we're checking polarities
                IF (NUM_POLARITIES_TO_TEST .GT. 1) THEN
                    WRITE(*,'(2(A,I1),A,I0)') '**info(MODE_FILAMENT_REFINE): checking polarity ', CURRENT_POLARITY, ' of ', &
                                                NUM_POLARITIES_TO_TEST, ' for filament ', FMC%FIL%ID
                ENDIF


                ! Loop over phi polarities
                WRITE(*,'(A)') '<phi_polarity_loop>'
                DO CURRENT_PHI_POLARITY=1,NUM_PHI_POLARITIES_TO_TEST

                    WRITE(*,'(A,I0,A)') '<phi_polarity num=', CURRENT_PHI_POLARITY, '>'
                    IF (CURRENT_PHI_POLARITY .GT. 1) THEN
                        WRITE(*,'(A,I0,A,I0)') '**info(MODE_FILAMENT_REFINE): on phi polarity ', CURRENT_PHI_POLARITY, &
                                                ' of ', NUM_PHI_POLARITIES_TO_TEST
                    ENDIF

                    ! If necessary, reset the filament to the starting point and flip its polarity
                    CALL FILAMENT_ASSIGN(FMC%FIL,FILS(CURRENT_FILAMENT))
                    IF (DEBUG) THEN
                        WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINEMENT): fil info after resetting to fils(current_filament):'
                        CALL FILAMENT_PRINT_INFO(FMC%FIL)
                    ENDIF
                    IF (CURRENT_POLARITY .EQ. 2) THEN
                        !PRINT '(A)', '**debug(MODE_FILAMENT_STARTUP): flipping polarity of filament'
                        CALL FILAMENT_FLIP_POLARITY(FMC%FIL)
                        FMC%FIL_ORIGINAL = FMC%FIL
                    ENDIF
                    IF (CURRENT_PHI_POLARITY .EQ. 2) THEN
                        CALL FILAMENT_SHIFT_ALL_WPS(FMC%FIL,[0.,0.,0.,180.,0.,0.],UPDATE_MODEL=.TRUE.)
                        FMC%FIL_ORIGINAL = FMC%FIL
                    ENDIF

                    ! Print out filament info for current polarity
                    WRITE(*,'(3(A,I0),A)') '**info(MODE_FILAMENT_REFINE): info for filament ', FMC%FIL%ID, &
                                            ' (polarity ', CURRENT_POLARITY, '; phi polarity ', CURRENT_PHI_POLARITY, &
                                            '), before refinement: '
                    CALL FILAMENT_PRINT_INFO(FMC%FIL)

                    ! Remember the starting state of the filament
                    FIL_STATES(CURRENT_POLARITY,CURRENT_PHI_POLARITY,1) = FMC%FIL


                    ! Prepare the model 3D (pad/crop, impose helical symmetry, mask, subtract current fil's 3D...)
                    ! TODO: Could this be done outside the polarity loops? (the model should be constant)
                    WRITE(*,'(A)') '<fmc_model3d_prep>'
                    CALL FMC_MODEL3D_PREP(FMC)
                    WRITE(*,'(A)') '</fmc_model3d_prep>'


                    ! Memory footprint diag
                    IF (DEBUG) THEN
                        WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage after fmc_model3d_prep'
                        CALL PRINT_MEMORY_USAGE_STATS()
                    ENDIF

                    !
                    ! Score the starting state
                    !
                    PRINT '(/A,I0,A,I0,A,I0,A)',    '**info(MODE_FILAMENT_REFINE): scoring starting state for filament ', &
                                                    FILS(CURRENT_FILAMENT)%ID, ' (polarity ', CURRENT_POLARITY, &
                                                    '; phi polarity ', CURRENT_PHI_POLARITY, ')'
                    WRITE(TEMP_FILENAME,'(A,I1,A)') 'start_pol_', CURRENT_POLARITY, '.mrc'
                    SCORE(CURRENT_POLARITY,CURRENT_PHI_POLARITY,1) = FMC_EVAL(FMC,FMC%FIL,WRITE_STACKS=.FALSE.,&
                                                                                STACKS_ROOTNAME=TEMP_FILENAME,&
                                                                                DETAILLED_SCORE_INFO=.TRUE.)
                    PRINT '(A,I0,A,I1,A,F0.3/)',    '**info(MODE_FILAMENT_REFINE): starting state score for filament ', &
                                                    FILS(CURRENT_FILAMENT)%ID, ' (polarity ', CURRENT_POLARITY, &
                                                    ') is ', SCORE(CURRENT_POLARITY,CURRENT_PHI_POLARITY,1)


                    ! Memory footprint diag
                    IF (DEBUG) THEN
                        WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage after first scoring function'
                        CALL PRINT_MEMORY_USAGE_STATS()
                    ENDIF




                    IF (FMC%WPS_AS_SPS) THEN
                        ! We've been told to treat WPs as single particles...
                        CALL FMC_MAXIMISE(FMC)
                        IF (DEBUG .OR. .TRUE.) THEN
                            WRITE(*,'(2A,I0,A)') '**debug(MODE_FILAMENT_REFINE): detailled score after single ', &
                                                'particle refinement, filament #', FMC%FIL%ID, ':'
                            DUMMY = FMC_EVAL(FMC,FMC%FIL,DETAILLED_SCORE_INFO=.TRUE.)
                        ENDIF
                    ELSE
                        ! We are not treating waypoints as single particles
                        IF (SEARCH_SUBUNIT_NUM%VALUE) THEN
                            PRINT '(A)', '**info(MODE_FILAMENT_REFINE): trying BF subunit_num search before DS. Experimental.'
                            CALL FMC_BF_SUBUNIT_NUM_INTG_REFINE(FMC)
                        ENDIF


                        ! If requested, do a filament-tilt brute-force search
                        ! TODO: it should be possible to incorporate this search logic in the framework of the RIGID_PARAMS
                        IF (SEARCH_FILAMENT_TILT%VALUE) THEN
                            WRITE(*,'(2A,F0.2,A,F0.2,A)') '**info(MODE_FILAMENT_REFINE): doing rigid-body filament ',&
                                                            'tilt brute force search over +/- ', FMC%SEARCH_HALF_RANGE(5), &
                                                            ' degrees in ',FMC%SEARCH_SAMPLING(5),' degree steps'
                            CALL FMC_BF_TILT(FMC,TILT_HALF_RANGE=FMC%SEARCH_HALF_RANGE(5),TILT_SAMPLING=FMC%SEARCH_SAMPLING(5))
						ELSE
							! Now find the nearby maximum in the scoring function
	                        ! (this is the workhorse of refinements)
	                        CALL FMC_MAXIMISE(FMC)
						ENDIF

                        ! If requested, do seam search
                        IF (SEARCH_MT_SEAM%VALUE) THEN
                            IF (FMC%FIL%MICROTUBULE) THEN
                                WRITE(*,'(A)') '**info(MODE_FILAMENT_REFINE): will search for microtubule seam'
                                CALL FMC_BF_FIND_MT_SEAM(FMC)
                            ELSE
                                IF (DEBUG) THEN
                                    WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): filament is not microtubule, '// &
                                                    'will not search for seam.'
                                ENDIF
                            ENDIF
                        ENDIF


                        ! Test: to give the flipped polarity a better chance, we do another maximisation
                        IF (CURRENT_POLARITY .EQ. 2) THEN
                            !WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): giving the flipped polarity a second and third run at the maximisation'
                            !CALL FMC_MAXIMISE(FMC)
                        ENDIF
                    ENDIF   ! End of test for WPS_AS_SPS%VALUE

                    ! Evaluate the score at the end of the iteration
                    WRITE(*,'(A,I0,A)') '**info(MODE_FILAMENT_REFINE): scoring information for filament ', &
                                            FILS(CURRENT_FILAMENT)%ID,  ' following minimisation follows'
                    WRITE(TEMP_FILENAME,'(A,I1)') 'after_ds_polarity', CURRENT_POLARITY
                    SCORE(CURRENT_POLARITY,CURRENT_PHI_POLARITY,2) = FMC_EVAL(FMC,FMC%FIL, &
                                                                        WRITE_STACKS=.FALSE.,STACKS_ROOTNAME=TEMP_FILENAME, &
                                                                        DETAILLED_SCORE_INFO=.TRUE.)

                    ! Store the state of the filament at the end of the refinement
                    FIL_STATES(CURRENT_POLARITY,CURRENT_PHI_POLARITY,2) = FMC%FIL

                    ! It may be interesting to get a FF score when we're doing SP refinement, to see how "compatible"
                    ! the two modes are
                    IF (FMC%WPS_AS_SPS) THEN
                        WRITE(*,'(2A)') '**info(MODE_FILAMENT_REFINE): below is score with ',   &
                                        'full-filament (rather than WPS as SPS) function: '
                        FMC%WPS_AS_SPS = .FALSE.
                        DUMMY = FMC_EVAL(FMC,FMC%FIL,WRITE_STACKS=.FALSE.,  &
                                        STACKS_ROOTNAME=TEMP_FILENAME,DETAILLED_SCORE_INFO=.TRUE.)
                        FMC%WPS_AS_SPS = .TRUE.
                    ENDIF

                    PRINT '(3(A,I0),A,F0.4)',   '**info(MODE_FILAMENT_REFINE): score for filament ', FILS(CURRENT_FILAMENT)%ID , &
                                                ' (polarity ', CURRENT_POLARITY, '; phi polarity ', CURRENT_PHI_POLARITY, '): ', &
                                                SCORE(CURRENT_POLARITY,CURRENT_PHI_POLARITY,2)
                    CALL FILAMENT_PRINT_INFO(FMC%FIL)
                    IF (.NOT. WPS_AS_SPS%VALUE) CALL FILAMENT_PRINT_EXTRA_INFO(FMC%FIL)


                    WRITE(*,'(A)') '</phi_polarity>'
                ENDDO   ! End of loop over filament phi polarities
                WRITE(*,'(A)') '</phi_polarity_loop>'

                ! Decide which phi polarity was better for the current filament in the current polarity
                IF (PHI_POLARITY_CHECK%VALUE) THEN
                    IF (SCORE(CURRENT_POLARITY,1,2) .GT. SCORE(CURRENT_POLARITY,2,2)) THEN
                        BEST_PHI_POLARITY(CURRENT_POLARITY) = 1
                    ELSE
                        BEST_PHI_POLARITY(CURRENT_POLARITY) = 2
                    ENDIF

                    WRITE(*,'(3(A,I0),A,F0.4)')     '**info(MODE_FILAMENT_REFINE): for filament ', FILS(CURRENT_FILAMENT)%ID,   &
                                                    ' in polarity ', CURRENT_POLARITY,                                          &
                                                    ', phi polarity ', BEST_PHI_POLARITY(CURRENT_POLARITY),                     &
                                                    ' was best, with a score of ', &
                                                    SCORE(CURRENT_POLARITY,BEST_PHI_POLARITY(CURRENT_POLARITY),2)
                ELSE
                    BEST_PHI_POLARITY(CURRENT_POLARITY) = 1
                ENDIF

                WRITE(*,'(A)') '</polarity>'
            ENDDO   POLARITY_LOOP ! End of loop over filament polarities
            WRITE(*,'(A)') '</polarity_loop>'

            !
            ! Decide which polarity was best for the current filament
            !
            IF (POLARITY_CHECK%VALUE) THEN
                IF (SCORE(1,BEST_PHI_POLARITY(1),2) .GT. SCORE(2,BEST_PHI_POLARITY(2),2)) THEN
                    BEST_POLARITY = 1
                ELSE
                    BEST_POLARITY = 2
                ENDIF

                WRITE(*,'(A,I0,A,I0,A,F0.4)')   '**info(MODE_FILAMENT_REFINE): for filament ', FILS(CURRENT_FILAMENT)%ID,   &
                                                ', polarity ', BEST_POLARITY,                                       &
                                                ' was best, with a score of ', &
                                                SCORE(BEST_POLARITY,BEST_PHI_POLARITY(BEST_POLARITY),2)
                ! Reset the filament object in the FMC to the starting point of refinement for the winning polarity
                CALL FMC_NEW_FILAMENT(FMC,FIL_STATES(BEST_POLARITY,BEST_PHI_POLARITY(BEST_POLARITY),1))
            ELSE
                BEST_POLARITY = 1
            ENDIF

            !
            ! Remember the results of the search for the current filament
            !
            !FILS_FINAL(CURRENT_FILAMENT)        =   FIL_STATES(BEST_POLARITY,BEST_PHI_POLARITY(BEST_POLARITY),2)
            CALL FILAMENT_ASSIGN(FILS_FINAL(CURRENT_FILAMENT),FIL_STATES(BEST_POLARITY,BEST_PHI_POLARITY(BEST_POLARITY),2))
            FILS_FINAL(CURRENT_FILAMENT)%SCORE  =   SCORE(BEST_POLARITY,BEST_PHI_POLARITY(BEST_POLARITY),2)
        !
        ! End of test of whether any wps needed to be refined
        !
        ENDIF


        !
        ! Now that we're all done with refinement, compute the final score and output any files the user may
        ! have asked for (matching projections etc.)
        !
        IF (MODEL_3D_AVAILABLE%VALUE) THEN
            IF (FMC%INITIALISED) THEN
                ! Prepare the model 3D (pad/crop, impose helical symmetry, mask..). In case
                ! we didn't do any refinement, this won't have been done yet.
                CALL FMC_MODEL3D_PREP(FMC)
                CALL FILENAME_FROM_WILD(MATCHING_PROJECTIONS_FILES%VALUE,FILS_FINAL(CURRENT_FILAMENT)%ID,MATCH_PROJ_FILENAME)
                IF (WRITE_MATCHING_PROJECTIONS%VALUE) THEN
                    WRITE(*,'(A,I0,2A)') '**info(MODE_FILAMENT_REFINE): writing stacks of images and matching projections '//&
                                        'for filament ', FILS_FINAL(CURRENT_FILAMENT)%ID, &
                                        '. Filename root: ', TRIM(ADJUSTL(MATCH_PROJ_FILENAME))
                ENDIF
                ! Print out final state of filament
                WRITE(*,'(A,I0,A)') '**info(MODE_FILAMENT_REFINE): final state of filament ', FILS(CURRENT_FILAMENT)%ID, ':'
                CALL FILAMENT_PRINT_INFO(FILS_FINAL(CURRENT_FILAMENT))
                WRITE(*,'(A,I0,A)') '**info(MODE_FILAMENT_REFINE): final statistics for filament ', FILS(CURRENT_FILAMENT)%ID, ':'
                CALL FILAMENT_PRINT_EXTRA_INFO(FILS_FINAL(CURRENT_FILAMENT))
                ! The easiest way to get matching projection stacks is by calling
                ! the scoring function (threading off so that matching projections are in correct order)
                WRITE(*,'(A,I0,A)') '**info(MODE_FILAMENT_REFINE): detailled scoring information for final state of filament ', &
                                    FILS_FINAL(CURRENT_FILAMENT)%ID, ' follows:'
                SCORE_TMP = FMC_EVAL(   FMC,FILS_FINAL(CURRENT_FILAMENT),WRITE_STACKS=WRITE_MATCHING_PROJECTIONS%VALUE,  &
                                        STACKS_ROOTNAME=MATCH_PROJ_FILENAME,   &
                                        DETAILLED_SCORE_INFO=.TRUE.,THREADING=.FALSE.,WRITE_FRC_PLOTS=.TRUE.)
                IF (FMC%STEP .LT. DEFAULT_FALLOFF_PIXELS + 1.) then
                    WRITE(*,'(3A)') '**warning(MODE_FILAMENT_REFINE): the step size for the scoring function was small, which ', &
                                   ' suggests filament deformations are significant relative to the pixel size. Check results ', &
                                   ' carefully and if applicable, move to a smaller pixel size'
                ENDIF

                ! Sanity check
                IF (ABS(SCORE_TMP-FILS_FINAL(CURRENT_FILAMENT)%SCORE) .GT. 0.02) THEN
                    WRITE(*,'(A,2(F0.4))') '**warning(MODE_FILAMENT_REFINE): final score is inconsistent: ', &
                                            SCORE_TMP, FILS_FINAL(CURRENT_FILAMENT)%SCORE
                    CALL TERMINATE('Failed sanity check in MODE_FILAMENT_REFINE')
                ENDIF
                ! If we used projection cache, also give detailled scoring info with the cache de-activated
                IF (FMC%CACHE_PROJECTIONS) THEN
                    WRITE(*,'(A)') '**info(MODE_FILAMENT_REFINE): parameters were refined using cached projections. '// &
                                    'Below is the detailled scoring information of the filament again, '//&
                                    'this time without using cached projections.'
                    DUMMY = FMC_EVAL(   FMC,FILS_FINAL(CURRENT_FILAMENT),WRITE_STACKS=WRITE_MATCHING_PROJECTIONS%VALUE,  &
                                        STACKS_ROOTNAME=MATCH_PROJ_FILENAME,   &
                                        DETAILLED_SCORE_INFO=.TRUE.,THREADING=.FALSE.,WRITE_FRC_PLOTS=.FALSE.,DONT_USE_CACHE=.TRUE.)
                ENDIF
                ! We also do a sanity check
                IF (ANY(FILS(CURRENT_FILAMENT)%WP%REFINE) .AND. .NOT. SKIP_REFINEMENT%VALUE) THEN
                    IF ( ABS( SCORE(BEST_POLARITY,BEST_PHI_POLARITY(BEST_POLARITY),2)    &
                             -FILS_FINAL(CURRENT_FILAMENT)%SCORE) .GT. 0.02 ) THEN
                        WRITE(*,'(2(A,F8.5))')  '**WARNING(MODE_FILAMENT_REFINE): mismatch between scores - '//&
                                                'programming error? Scores =', &
                                                SCORE(BEST_POLARITY,BEST_PHI_POLARITY(BEST_POLARITY),2), ';', &
                                                FILS_FINAL(CURRENT_FILAMENT)%SCORE
                        CALL TERMINATE('Failed sanity check in MODE_FILAMENT_REFINE')
                    ENDIF
                ENDIF
            ELSE
                WRITE(*,'(A)')  '**warning(MODE_FILAMENT_REFINE): unexpectedly, FMC was never initialised, '//&
                                'not doing the final scoring etc.'
            ENDIF
        ENDIF ! End of test for model 3d availability

        !
        ! If the user asked, write boxer files with new waypoint positions, as well as traces of the filaments.
        !
        IF (WRITE_FILAMENT_TRACES%VALUE) THEN
            IF (FILENAME_CONTAINS_QMARKS(FILM_FILES%VALUE)) THEN
                ! Find out the number of the film to which the current filament belongs
                CURRENT_FILM_NUMBER =  FILE_NUMBER_FROM_FILENAME(FILS_FINAL(CURRENT_FILAMENT)%FILM%FILENAME,FILM_FILES%VALUE)
                ! Work out the filename we'll write to
                CALL FILENAME_FROM_WILD(FILAMENT_TRACE_FILES%VALUE,CURRENT_FILM_NUMBER,FILAMENT_TRACE_FILENAME)
                CALL FILENAME_FROM_WILD(FILAMENT_TRACE_FILENAME,FILS(CURRENT_FILAMENT)%ID,FILAMENT_TRACE_FILENAME)
                ! Write the coordinates to disk
                PRINT '(2A)', '**info(MODE_FILAMENT_REFINE): dumping trace to ', TRIM(FILAMENT_TRACE_FILENAME)
                CALL FILAMENT_BOXER_DB_TO_DISK(FILS_FINAL(CURRENT_FILAMENT),FILAMENT_TRACE_FILENAME)

                ! Write out trace to CSV file
                CALL FILENAME_REMOVE_EXTENSION(FILAMENT_TRACE_FILENAME,FILAMENT_TRACE_FILENAME)
                FILAMENT_TRACE_FILENAME = TRIM(FILAMENT_TRACE_FILENAME)//'.csv'
                PRINT '(2A)', '**info(MODE_FILAMENT_REFINE): dumping per-subunit trace to CSV file ', FILAMENT_TRACE_FILENAME
                CALL FILAMENT_TRACE_TO_DISK_CSV(FILS_FINAL(CURRENT_FILAMENT),FILAMENT_TRACE_FILENAME)

                ! Write out just the way points without trace
                CALL FILENAME_REMOVE_EXTENSION(FILAMENT_TRACE_FILENAME,TEMP_FILENAME)
                FILAMENT_TRACE_FILENAME = TRIM(ADJUSTL(TEMP_FILENAME))//'_wp.box'
                CALL FILAMENT_BOXER_DB_TO_DISK(FILS_FINAL(CURRENT_FILAMENT),FILAMENT_TRACE_FILENAME,WITH_TRACE=.FALSE.)
            ELSE
                WRITE(*,'(A)') '**warning(mode_filament_refine): writing of filament traces is disabled because input '//&
                                'micrograph stack was used, rather than set of separate files'
            ENDIF
        ENDIF

        ! Print out changes in filament alignment
        WRITE(*,'(A)') '**info(MODE_FILAMENT_REFINE): summary of changes in filament during refinement: '
        CALL FILAMENT_PRINT_DIFF(FILS(CURRENT_FILAMENT),FILS_FINAL(CURRENT_FILAMENT))



        !
        ! Calculate a new 3D reconstruction
        !
        ! Note (1-Nov-2010):
        ! Watch out: if all the odd-numbered filaments have low scores, one of the FSC volumes could end up empty!
        ! The logic below could be re-arranged to avoid this.
        ! The score would be checked against the threshold first, and then if the score is good enough,
        ! the filament would be added to the next group.
        IF (CALC_3D_RECONSTRUCTION%VALUE) THEN

            ! Some of the output 3Ds will be croped to save diskspace
            CROP_DIM = NINT(2.*1.1*HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE)
            IF (IsOdd(CROP_DIM(1))) CROP_DIM = CROP_DIM + 1 ! Not necessary for Frealix, but may be so for other software

            IF (     (FILS_FINAL(CURRENT_FILAMENT)%SCORE .GT. MIN_SCORE_FIL_TO_3D%VALUE)  &
                .OR. (SINGLE_FILAMENT_RECONSTRUCTIONS%VALUE)    ) THEN
                WRITE(*,'(A,I0,A)') '**info(MODE_FILAMENT_REFINE): about to compute 3D reconstruction for filament ', &
                                    FILS_FINAL(CURRENT_FILAMENT)%ID, '.'
                WRITE(*,'(2A)') '**info(MODE_FILAMENT_REFINE): reading in the micrograph image. It will be used without', &
                                ' any filtering for the filament reconstruction'
                CALL FILM_FILE%INIT(FILS(CURRENT_FILAMENT)%FILM%FILENAME)
                CALL FILM_FILE%PrintInfo()
                CALL CURRENT_FILM_IMAGE%ReadFromDisk(FILM_FILE,FILS(CURRENT_FILAMENT)%FILM%LOCATION_IN_FILE)
                CALL FILM_FILE%CLOSE()
                IF (REVERSE_FILM_CONTRAST%VALUE) THEN
                    CURRENT_FILM_IMAGE%real_values = CURRENT_FILM_IMAGE%real_values * (-1.0)
                ENDIF
                IF (WPS_AS_SPS%VALUE) THEN
                    CALL FILAMENT_CALC_THREED(  FILS_FINAL(CURRENT_FILAMENT),CURRENT_FILM_IMAGE,CURRENT_3DREC,      &
                                                SAFE_BOX_SIZE=MIN_SAFE_BOX_SIZE,FINALISE=.FALSE.,WIPE=.TRUE.,       &
                                                SP_MODE=WPS_AS_SPS%VALUE,SP_SEGMENT_LENGTH=FMC%SP_SEGMENT_LENGTH,         &
                                                SP_HELIX=HELIX_AVERAGE,                                             &
                                                APPLY_A_BFACTOR=BFACTOR_DURING_3D_RECONSTRUCTION%VALUE,   &
                                                AVERAGE_SCORE=AVERAGE_SCORE%VALUE, &
                                                BFACTOR_PER_SUBUNIT=BFACTOR_PER_SUBUNIT%VALUE)
                ELSE
                    CALL FILAMENT_CALC_THREED(  FILS_FINAL(CURRENT_FILAMENT),CURRENT_FILM_IMAGE,CURRENT_3DREC,      &
                                                SAFE_BOX_SIZE=MIN_SAFE_BOX_SIZE,FINALISE=.FALSE.,WIPE=.TRUE.,       &
                                                SP_MODE=WPS_AS_SPS%VALUE,SP_SEGMENT_LENGTH=FMC%SP_SEGMENT_LENGTH,         &
                                                APPLY_A_BFACTOR=BFACTOR_DURING_3D_RECONSTRUCTION%VALUE,   &
                                                AVERAGE_SCORE=AVERAGE_SCORE%VALUE, &
                                                BFACTOR_PER_SUBUNIT=BFACTOR_PER_SUBUNIT%VALUE)
                ENDIF
                ! If the user asked for it, finalise this recontruction from the current filament
                ! and write the central slice to disk
                IF (SINGLE_FILAMENT_RECONSTRUCTIONS%VALUE) THEN
                    WRITE(*,'(A,I0,A)') '**info(MODE_FILAMENT_REFINE): finalising single-filament 3D reconstruction for filament ',&
                                        FILS_FINAL(CURRENT_FILAMENT)%ID, '.'
                    CALL THREED_EVAL_RECONSTRUCTION(CURRENT_3DREC,CURRENT_3DVOL,IMG_AMPL=CURRENT_3DVOL_AMPL)
                    CALL CROP(CURRENT_3DVOL,CROP_DIM)
                    ! Extract YZ slice
                    CALL IMAGE_GET_SLICE(CURRENT_3DVOL,CURRENT_3DVOL_SLICE,AXIS=1,ADDRESS=CURRENT_3DVOL%logical_dimensions(1)/2+1)
                    CALL FILENAME_ADD_SUFFIX(OUTPUT_3D_FILENAME%VALUE,'_slice_yz_??????',TEMP_FILENAME)
                    CALL FILENAME_FROM_WILD(TEMP_FILENAME,FILS_FINAL(CURRENT_FILAMENT)%ID,TEMP_FILENAME)
                    PRINT '(2A)', '**info(MODE_FILAMENT_REFINE): writing slice to file ', TRIM(ADJUSTL(TEMP_FILENAME))
                    CALL CURRENT_3DVOL_SLICE%WriteToDisk(TEMP_FILENAME,PIXEL_SIZE=PIXEL_SIZE%VALUE)
                    ! Extract XY slice
                    CALL IMAGE_GET_SLICE(CURRENT_3DVOL,CURRENT_3DVOL_SLICE,AXIS=3,ADDRESS=CURRENT_3DVOL%logical_dimensions(3)/2+1)
                    CALL FILENAME_ADD_SUFFIX(OUTPUT_3D_FILENAME%VALUE,'_slice_xy_??????',TEMP_FILENAME)
                    CALL FILENAME_FROM_WILD(TEMP_FILENAME,FILS_FINAL(CURRENT_FILAMENT)%ID,TEMP_FILENAME)
                    PRINT '(2A)', '**info(MODE_FILAMENT_REFINE): writing slice to file ', TRIM(ADJUSTL(TEMP_FILENAME))
                    CALL CURRENT_3DVOL_SLICE%WriteToDisk(TEMP_FILENAME,PIXEL_SIZE=PIXEL_SIZE%VALUE)
                ENDIF

                ! Add the current filament's reconstruction to the total reconstruction
                ! only if the filament's score is over the user-supplied threshold
                WRITE(*,'(A,I0,A,F0.3,A)',ADVANCE='NO') '**info(MODE_FILAMENT_REFINE): score for filament ',    &
                                                        FILS_FINAL(CURRENT_FILAMENT)%ID, ' (',                  &
                                                        FILS_FINAL(CURRENT_FILAMENT)%SCORE,') is '
                IF (FILS_FINAL(CURRENT_FILAMENT)%SCORE .GT. MIN_SCORE_FIL_TO_3D%VALUE) THEN
                    WRITE(*,'(A)',ADVANCE='NO') 'above '
                ELSE
                    WRITE(*,'(A)',ADVANCE='NO') 'below '
                ENDIF
                WRITE(*,'(A,F0.3,A)') 'threshold for inclusion into final 3D (',MIN_SCORE_FIL_TO_3D%VALUE, ').'

                IF (FILS_FINAL(CURRENT_FILAMENT)%SCORE .GT. MIN_SCORE_FIL_TO_3D%VALUE) THEN
                    CALL THREED_ADD(NEW_3DREC,CURRENT_3DREC,NEW_3DREC)
                ENDIF
            ENDIF ! End of test of whether filament should be reconstructed
        ENDIF ! End of test of whether user asked for a 3D reconstruction

        !
        ! Move waypoints to the nearest helical repeats
        !
        IF (MOVE_WAYPOINTS_TO_NEAREST_REPEAT%VALUE) THEN
            WRITE(*,'(A)') '**info(MODE_FILAMENT_REFINE): moving output waypoints to the nearest helical repeats'
            CALL FILAMENT_MOVE_WPS_TO_NEAREST_SUBUNIT(FILS_FINAL(CURRENT_FILAMENT))
        ENDIF

        WRITE(*,'(A)') '</filament>'

    ENDDO   ! End of loop over filaments
    WRITE(*,'(A)') '</loop_over_filaments>'

    ! Memory footprint diag
    IF (DEBUG) THEN
        WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage after loop over filaments'
        CALL PRINT_MEMORY_USAGE_STATS()
    ENDIF

    !
    ! Write out alignment parameter file
    !
    CALL FILAMENTS_TO_FREALIX_PAR(FILS_FINAL,OUTPUT_PAR_FILENAME%VALUE)

    !
    ! Write out alignment parameter file with each asymetric unit
    !
    IF (.NOT. STRING_EQUAL(OUTPUT_PAR_FILENAME_SUBUNITS%VALUE,'none')) THEN
        CALL FILAMENTS_TO_FREALIX_PAR(FILS_FINAL,OUTPUT_PAR_FILENAME_SUBUNITS%VALUE,SUBUNITS=.TRUE.)
    ENDIF

    !
    ! Finalise the 3D reconstruction & write it to disk
    !
    IF (CALC_3D_RECONSTRUCTION%VALUE) THEN
        WRITE(*,'(A)') '<calc_3d_rec>'
        IF (CALC_FSC%VALUE) THEN
            WRITE(*,'(A)') '<calc_fsc>'
            ! Evaluate / finalise the two half reconstructions
            CALL THREED_EVAL_RECONSTRUCTION(NEW_3DREC,NEW_3DVOL(1),REC_NUM=1,PRESERVE_HIGH_FREQUENCIES=.TRUE.,SKIP_AMPL=.TRUE.)
            CALL THREED_EVAL_RECONSTRUCTION(NEW_3DREC,NEW_3DVOL(2),REC_NUM=2,PRESERVE_HIGH_FREQUENCIES=.TRUE.,SKIP_AMPL=.TRUE.)
            ! Impose symmetry
            CALL IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW( NEW_3DVOL(1),HELIX_AVERAGE,  &
                                                    INTERPOL_SCHEME,MAX_RADIUS=HELIX_MAX_RADIUS_PIX,PROGRESS_REPORT=.TRUE.)
            CALL IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW( NEW_3DVOL(2),HELIX_AVERAGE, &
                                                    INTERPOL_SCHEME,MAX_RADIUS=HELIX_MAX_RADIUS_PIX,PROGRESS_REPORT=.TRUE.)
            ! Mask the reconstructions
            CALL IMAGE_HELIX_MASK(NEW_3DVOL(1),HELIX_AVERAGE,ASYM_UNIT_MW%VALUE,PIXEL_SIZE%VALUE,HELIX_MAX_RADIUS_PIX, &
                                    MASK_RESOLUTION = PIXEL_SIZE%VALUE / 20.0 )
            CALL IMAGE_HELIX_MASK(NEW_3DVOL(2),HELIX_AVERAGE,ASYM_UNIT_MW%VALUE,PIXEL_SIZE%VALUE,HELIX_MAX_RADIUS_PIX, &
                                    MASK_RESOLUTION = PIXEL_SIZE%VALUE / 20.0 )
            ! Dump reconstruction 1 to disk
            CALL FILENAME_ADD_SUFFIX(OUTPUT_3D_FILENAME%VALUE,'_fsc1',TEMP_FILENAME)
            CALL NEW_3DVOL(1)%WriteToDisk(TEMP_FILENAME,PIXEL_SIZE=PIXEL_SIZE%VALUE)
            ! Dump reconstruction 2 to disk
            CALL FILENAME_ADD_SUFFIX(OUTPUT_3D_FILENAME%VALUE,'_fsc2',TEMP_FILENAME)
            CALL NEW_3DVOL(2)%WriteToDisk(TEMP_FILENAME,PIXEL_SIZE=PIXEL_SIZE%VALUE)
            ! Calculate & printout the FSC
            CALL FWD_FT(NEW_3DVOL(1))
            CALL FWD_FT(NEW_3DVOL(2))
            CALL COMPUTE_FSC(NEW_3DVOL(1),NEW_3DVOL(2),FOUSHELL)
            CALL PRINT_FSC(FOUSHELL)
            WRITE(*,'(A)') '</calc_fsc>'
        ENDIF

        ! Dump the 3D reconstruction to disk, without finalising. The unweighted volume, the CTF^2 volumes are written separately.
        ! (We always write the unfinished 3D to disk, so that we have flexibility later on to subtract from it, for example)
        CALL THREED_WRITE_TO_DISK(NEW_3DREC,OUTPUT_3D_FILENAME%VALUE)

        ! If we're finalising, also dump to disk the final reconstruction
        IF (FINALISE_3D_RECONSTRUCTION%VALUE) THEN
            WRITE(*,'(A)') '<finalise_3d>'
            ! Finalise the total recontruction
            IF (SINGLE_FILAMENT_RECONSTRUCTIONS%VALUE .AND. SIZE(FILS) .EQ. 1) THEN
                WRITE(*,'(2A)') '**info(MODE_FILAMENT_REFINE): since we were only refining 1 filament, ',    &
                                'for which we also already evaluated a 3D reconstruction, we will use this ',   &
                                'as the output 3D reconstruction, without recalculating one'
                NEW_3DVOL(1) = CURRENT_3DVOL
				IF (RECONSTRUCT_AMPLITUDES%VALUE) NEW_3DVOL_AMPL(1) = CURRENT_3DVOL_AMPL
            ELSE
                WRITE(*,'(A,I0,A)') '**info(MODE_FILAMENT_REFINE): will finalise the combined 3D reconstruction from the ', &
                                        SIZE(FILS), ' filaments'
                CALL THREED_EVAL_RECONSTRUCTION(NEW_3DREC,NEW_3DVOL(1),PRESERVE_HIGH_FREQUENCIES=.TRUE.,IMG_AMPL=NEW_3DVOL_AMPL(1))
            ENDIF
            ! If microtubule, rebuild the seam
            IF (MICROTUBULE%VALUE) THEN
                WRITE(*,'(A)') '**info(MODE_FILAMENT_REFINE): imposing microtubule symmetry in real space'
                CALL IMAGE_IMPOSE_MICROTUBULE_SYMMETRY(FILS_FINAL(1),NEW_3DVOL(1))
            ENDIF
            ! Write the final reconstruction to disk
            IF (RECONSTRUCT_AMPLITUDES%VALUE) THEN
                CALL FILENAME_ADD_SUFFIX(OUTPUT_3D_FILENAME%VALUE,'_ampl',OUTPUT_3D_AMPL_FILENAME)
                CALL NEW_3DVOL_AMPL(1)%WriteToDisk(OUTPUT_3D_AMPL_FILENAME,PRINT_FILE_INFO=.TRUE.,PIXEL_SIZE=PIXEL_SIZE%VALUE)
            ENDIF
!
            !
            WRITE(*,'(2A)') '**info(MODE_FILAMENT_REFINE): preparing a cropped, symmetrised and masked ',    &
                            'version of the final 3D reconstruction'
            CALL CROP(NEW_3DVOL(1),CROP_DIM)
            CALL IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW(NEW_3DVOL(1),HELIX_AVERAGE,INTERPOL_SCHEME,  &
                                                    MAX_RADIUS=HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE)
            CALL IMAGE_HELIX_MASK(NEW_3DVOL(1),HELIX_AVERAGE,ASYM_UNIT_MW%VALUE,PIXEL_SIZE%VALUE,HELIX_MAX_RADIUS%VALUE)
            CALL FILENAME_ADD_SUFFIX(OUTPUT_3D_FILENAME%VALUE,'_hm',OUTPUT_3D_FILENAME%VALUE)
            CALL NEW_3DVOL(1)%WriteToDisk(OUTPUT_3D_FILENAME%VALUE,PIXEL_SIZE=PIXEL_SIZE%VALUE,PRINT_FILE_INFO=.TRUE.)

            WRITE(*,'(A)') '</finalise_3d>'
        ENDIF
        WRITE(*,'(A)') '</calc_3d_rec>'
    ENDIF


    ! Cleanup
    IF (ALLOCATED(FILS)) THEN
        DO I=1,SIZE(FILS)
            CALL FILAMENT_DEALLOCATE(FILS(I))
        ENDDO
        DEALLOCATE(FILS)
    ENDIF

    ! Deallocate image objects (I don't think this should be necessary,
    ! but it seems ifort is not runnin the finalisation subroutines as I
    ! expected)
    CALL IMAGE_DEALLOCATE(NEW_3DVOL(1))
    CALL IMAGE_DEALLOCATE(NEW_3DVOL(2))
    CALL IMAGE_DEALLOCATE(MODEL_3DVOL)
    CALL IMAGE_DEALLOCATE(CURRENT_3DVOL)
    CALL IMAGE_DEALLOCATE(CURRENT_3DVOL_SLICE)
    CALL IMAGE_DEALLOCATE(CURRENT_FILM_IMAGE)
    ! Deallocate threed objects
    CALL THREED_DEALLOCATE(CURRENT_3DREC)
    CALL THREED_DEALLOCATE(NEW_3DREC)
    !
    CALL FMC_DEALLOCATE(FMC)

    ! Memory footprint diag
    IF (DEBUG) THEN
        WRITE(*,'(A)') '**debug(MODE_FILAMENT_REFINE): memory usage on exit'
        CALL PRINT_MEMORY_USAGE_STATS()
    ENDIF

END SUBROUTINE MODE_FILAMENT_REFINE_SLAVE

END MODULE MODE_FILAMENT_REFINE
