!>  \brief  Combine 3D reconstructions obtained from subsets of the input data into a single 3D reconstruction.
!!          This mode is meant for use with frealix2, whereas mode COMBINE_3Ds is meant for use with frealix1 & chuff.
MODULE MODE_FINALISE_3DS
    USE MODES
    USE RTPS
    IMPLICIT NONE


    TYPE(RTP_CHAR), TARGET  ::  INPUT_THREED_FILES  =               &
        RTP_CHAR(   VALUE='threed_???.mrc',                  &
                    KEYWORD='INPUT_THREED_FILES',           &
                    DESCRIPTION='Root filename (with wildcard) for input 3D reconstruction files')


    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_THREED_FILENAME  =               &
        RTP_CHAR(   VALUE='threed.mrc',                  &
                    KEYWORD='OUTPUT_THREED_FILENAME',           &
                    DESCRIPTION='Output 3D reconstruction filename')


    TYPE(RTP_CHAR), TARGET  ::  INPUT_THREED_SELECTION  =               &
        RTP_CHAR(   VALUE='1-100',                  &
                    KEYWORD='INPUT_THREED_SELECTION',           &
                    DESCRIPTION='Selection of 3D reconstructions to be summed & finalised')


    TYPE(RTP_LOGI), TARGET  ::  CALC_FSC  =               &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='CALC_FSC',           &
                    DESCRIPTION='Whether to calculate an FSC between halves of the dataset')


    TYPE(RTP_LOGI), TARGET  ::  APPLY_FOM_FILTER  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='APPLY_FOM_FILTER',           &
                    DESCRIPTION='Whether a Figure-Of-Merit filter should be applied to the final reocnstruction')


    TYPE(RTP_CHAR), TARGET  ::  INTERPOL_FUNCTION  =               &
        RTP_CHAR(   VALUE='LINEAR',                  &
                    KEYWORD='INTERPOL_FUNCTION',           &
                    DESCRIPTION='Which interpolation function to use')


    TYPE(RTP_CHAR), TARGET  ::  WINDOW_FUNCTION  =               &
        RTP_CHAR(   VALUE='RECT',                  &
                    KEYWORD='WINDOW_FUNCTION',           &
                    DESCRIPTION='Which windowing function to use')


    TYPE(RTP_INTG), TARGET  ::  WINDOW_WIDTH  =               &
        RTP_INTG(   VALUE=2,                  &
                    KEYWORD='WINDOW_WIDTH',           &
                    DESCRIPTION='Window width for interpolation (kernel support)')


    TYPE(RTP_LOGI), TARGET  ::  HELICAL_MASKING  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='HELICAL_MASKING',           &
                    DESCRIPTION='')


    TYPE(RTP_REAL), TARGET  ::  TWIST_PER_SUBUNIT  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='TWIST_PER_SUBUNIT',           &
                    DESCRIPTION='')


    TYPE(RTP_REAL), TARGET  ::  RISE_PER_SUBUNIT  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='RISE_PER_SUBUNIT',           &
                    DESCRIPTION='')


    TYPE(RTP_INTG), TARGET  ::  SYMMETRY_AXIAL  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='SYMMETRY_AXIAL',           &
                    DESCRIPTION='')


    TYPE(RTP_INTG), TARGET  ::  SYMMETRY_PERP  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='SYMMETRY_PERP',           &
                    DESCRIPTION='')


    TYPE(RTP_REAL), TARGET  ::  HELIX_MAX_RADIUS  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='HELIX_MAX_RADIUS',           &
                    DESCRIPTION='Maximum outer radius of the helix, in Angstroms')


    TYPE(RTP_REAL), TARGET  ::  ASYM_UNIT_MW  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='ASYM_UNIT_MW',           &
                    DESCRIPTION='Expected molecular weight (Da) of the asymetric unit')


    TYPE(RTP_REAL), TARGET  ::  PIXEL_SIZE  =               &
        RTP_REAL(   VALUE=1.0,                  &
                    KEYWORD='PIXEL_SIZE',           &
                    DESCRIPTION='Pixel size (Angstrom)')


CONTAINS
SUBROUTINE MODE_FINALISE_3DS_INIT_AND_RUN
    IMPLICIT NONE
    TYPE(MODE) :: FINALISE_3DS

    ! Intialise the mode
    CALL FINALISE_3DS%INIT(COMMAND='FINALISE_3DS',SUBROUTINE=MODE_FINALISE_3DS_SLAVE,SHORT_DESCRIPTION='Add description here')

    ! Add RTPs to the mode
    CALL FINALISE_3DS%ADD_RTP(INPUT_THREED_FILES)
    CALL FINALISE_3DS%ADD_RTP(OUTPUT_THREED_FILENAME)
    CALL FINALISE_3DS%ADD_RTP(INPUT_THREED_SELECTION)
    CALL FINALISE_3DS%ADD_RTP(CALC_FSC)
    CALL FINALISE_3DS%ADD_RTP(APPLY_FOM_FILTER)
    CALL FINALISE_3DS%ADD_RTP(INTERPOL_FUNCTION)
    CALL FINALISE_3DS%ADD_RTP(WINDOW_FUNCTION)
    CALL FINALISE_3DS%ADD_RTP(WINDOW_WIDTH)
    CALL FINALISE_3DS%ADD_RTP(HELICAL_MASKING)
    CALL FINALISE_3DS%ADD_RTP(TWIST_PER_SUBUNIT)
    CALL FINALISE_3DS%ADD_RTP(RISE_PER_SUBUNIT)
    CALL FINALISE_3DS%ADD_RTP(SYMMETRY_AXIAL)
    CALL FINALISE_3DS%ADD_RTP(SYMMETRY_PERP)
    CALL FINALISE_3DS%ADD_RTP(HELIX_MAX_RADIUS)
    CALL FINALISE_3DS%ADD_RTP(ASYM_UNIT_MW)
    CALL FINALISE_3DS%ADD_RTP(PIXEL_SIZE)

    ! Run the mode
    CALL FINALISE_3DS%RUN()
END SUBROUTINE MODE_FINALISE_3DS_INIT_AND_RUN

SUBROUTINE MODE_FINALISE_3DS_SLAVE
    USE RUNTIME_PARAMETERS      ! This gives us access to user-defined or default values for runtime parameters
    USE PARSERS                 ! This contains the parser for lists of files
    USE THREEDS
    USE FSCS
    USE IMAGES_COMPARISON
    USE HELICES
    USE IMAGES_HELICES
    USE INTERPOLATION
    USE IMAGEFILES
    use DatesAndTimes
    IMPLICIT NONE
    ! Private variables
    INTEGER,            ALLOCATABLE ::  LIST_OF_THREEDS(:)
    INTEGER                         ::  I, J, THREED_NUM, CENTRAL_SLICE
    CHARACTER(LEN=FILENAME_LEN)     ::  INPUT_THREED_BASE_FILENAME, TMP_FILENAME, OUTPUT_AMPL_FILENAME
    TYPE(THREED)                    ::  INPUT_THREED, THREED_SUM(2)
    TYPE(IMAGE)                     ::  THREED_FINAL(2), CENTRAL_SLICES(2), AMPL_FINAL(2)
    REAL,   ALLOCATABLE             ::  FOUSHELL(:)
    TYPE(HELICAL_SYMMETRY)          ::  HELIX
    LOGICAL,    PARAMETER           ::  HELICAL_SYMMETRISE  =   .TRUE.
    TYPE(INTERPOLATION_SCHEME)      ::  INTERPOL_SCHEME
    LOGICAL,            PARAMETER   ::  DEBUG   =   .FALSE.
    INTEGER                         ::  CROP_DIM(3)
    INTEGER                         ::  INPUT_THREED_NUM_THREEDS, NUM_SUM_THREEDS
    TYPE(IMAGE)                     ::  FSC_VOLUMES(2)
    TYPE(POINT_SYMMETRY_SYMBOL)     ::  SYM_SYMBOL
    CHARACTER(LEN=6)                ::  SIZE_LIST_C
    type(DateAndTime)               ::  myTime
    ! Get runtime parameters from control file


    ! Work out the list of 3D reconstructions
    LIST_OF_THREEDS = SET_TO_LIST(INPUT_THREED_SELECTION%VALUE)
    WRITE(SIZE_LIST_C,'(I6)') SIZE(LIST_OF_THREEDS)
#ifdef NAGFOR_COMPILER
    IF (SIZE(LIST_OF_THREEDS) .LT. 100) THEN
#endif
    PRINT '(A,I0,A,'//SIZE_LIST_C//'(I0,1X))', '**info(MODE_FINALISE_3DS): The following ', SIZE(LIST_OF_THREEDS), &
                                                    ' 3D reconstructions are selected: ', LIST_OF_THREEDS
#ifdef NAGFOR_COMPILER
    ENDIF
#endif

    ! Initialise point symmetry
    SYM_SYMBOL%SYMBOL = 'H'

    ! Initialise helical parameters
    CALL HELICAL_SYMMETRY_INIT( HELIX,RISE_PER_SUBUNIT%VALUE,TWIST_PER_SUBUNIT%VALUE,    &
                                SYMMETRY_AXIAL%VALUE,SYMMETRY_PERP%VALUE,PIXEL_SIZE%VALUE)

    ! Intialise interpolation scheme in case it is needed
    CALL INTERPOLATION_SCHEME_INIT(INTERPOL_SCHEME,'LINEAR','RECT',2)

    ! Initialise a 3D reconstruction object for the input 3Ds
    THREED_NUM = LIST_OF_THREEDS(1)
    CALL FILENAME_FROM_WILD(INPUT_THREED_FILES%VALUE,THREED_NUM,INPUT_THREED_BASE_FILENAME)
    CALL THREED_INIT(       INPUT_THREED,                                           &
                            FILENAME_ROOT       =   INPUT_THREED_BASE_FILENAME,     &
                            INTERPOL_DESCR      =   INTERPOL_FUNCTION%VALUE,              &
                            WINDOW_DESCR        =   WINDOW_FUNCTION%VALUE,                &
                            WINDOW_WIDTH        =   WINDOW_WIDTH%VALUE,                   &
                            SYM_SYMBOL          =   SYM_SYMBOL,                     &
                            HELIX               =   HELIX,                          &
                            ASYM_UNIT_MW        =   ASYM_UNIT_MW%VALUE,                   &
                            PIXEL_SIZE          =   PIXEL_SIZE%VALUE,                     &
                            HELIX_MAX_RAD       =   HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE,    &
                            RECONSTRUCT_AMPLITUDES =  THREED_ON_DISK_AMPL_VOL(INPUT_THREED_BASE_FILENAME)  )
    INPUT_THREED_NUM_THREEDS = THREED_NUM_OF_RECS(INPUT_THREED)


    ! How many sums do we need?
    IF (SIZE(LIST_OF_THREEDS) .GT. 1) THEN
        IF (INPUT_THREED_NUM_THREEDS .EQ. 1) THEN
            ! There are multiple input 3Ds, each with 1 internal 3D
            IF (CALC_FSC%VALUE .OR. APPLY_FOM_FILTER%VALUE) THEN
                ! We need 2 running sums
                NUM_SUM_THREEDS = 2
            ELSE
                NUM_SUM_THREEDS = 1
            ENDIF
        ELSE
            ! There are multiple input 3Ds, each with 2 internal 3Ds
            NUM_SUM_THREEDS = 1
        ENDIF
    ELSE
        IF (INPUT_THREED_NUM_THREEDS .EQ. 1) THEN
            ! There is only one input 3D, and it only has 1 internal 3D
            NUM_SUM_THREEDS = 1
            IF (CALC_FSC%VALUE .OR. APPLY_FOM_FILTER%VALUE) THEN
                ! FSC/FOM not possible if only one 3D
                WRITE(*,'(A)') '**info(MODE_FINALISE_3DS): only one 3D reconstruction, therefore no FSC or FOM filter'
                CALC_FSC%VALUE = .FALSE.
                APPLY_FOM_FILTER%VALUE = .FALSE.
            ENDIF
        ELSE
            ! There is only one input 3D, with 2 internal 3Ds
            NUM_SUM_THREEDS = 1
        ENDIF
    ENDIF

    ! Initialise the "sum" 3Ds
    DO J=1,NUM_SUM_THREEDS
        THREED_SUM(J)%logical_dimensions = INPUT_THREED%logical_dimensions
        CALL THREED_INIT(   THREED_SUM(J),                                          &
                            INTERPOL_DESCR      =   INTERPOL_FUNCTION%VALUE,              &
                            WINDOW_DESCR        =   WINDOW_FUNCTION%VALUE,                &
                            WINDOW_WIDTH        =   WINDOW_WIDTH%VALUE,                   &
                            SYM_SYMBOL          =   SYM_SYMBOL,                     &
                            HELIX               =   HELIX,                          &
                            ASYM_UNIT_MW        =   ASYM_UNIT_MW%VALUE,                   &
                            PIXEL_SIZE          =   PIXEL_SIZE%VALUE,                     &
                            HELIX_MAX_RAD       =   HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE,    &
                            NUM_OF_RECS         =   INPUT_THREED_NUM_THREEDS,       &
                            FOM_FILTER          =   APPLY_FOM_FILTER%VALUE,&! .AND. (NUM_SUM_THREEDS .EQ. 1),    &
                            COMPUTE_FSC         =   CALC_FSC%VALUE,    &! .AND. (NUM_SUM_THREEDS .EQ. 1)             &
                            RECONSTRUCT_AMPLITUDES = INPUT_THREED%RECONSTRUCT_AMPLITUDES &
                                                                              )
        CALL THREED_WIPE(THREED_SUM(J))
    ENDDO


    ! Loop over the 3Ds and sum them
    DO I=1,SIZE(LIST_OF_THREEDS)
        THREED_NUM = LIST_OF_THREEDS(I)

        ! Read in the 3D reconstruction
        CALL FILENAME_FROM_WILD(INPUT_THREED_FILES%VALUE,THREED_NUM,INPUT_THREED_BASE_FILENAME)
        WRITE(*,'(2A,3(A,I0),2A)') '**info(MODE_FINALISE_3DS): ', myTime%DateAndTimeAsString(), ': adding 3D ', THREED_NUM, &
                                    ' (', I, ' of ', SIZE(LIST_OF_THREEDS), '): ', TRIM(INPUT_THREED_BASE_FILENAME)

        CALL THREED_READ_FROM_DISK(INPUT_THREED,INPUT_THREED_BASE_FILENAME)
        IF (DEBUG) WRITE(*,'(3A)') '**debug(MODE_FINALISE_3DS): ', myTime%DateAndTimeAsString(), &
                                    ': reading from disk completed. About to add.'

        ! If we're accumulating 2 sum reconstructions, the even numbered 3Ds go into a second volume
        J = 1
        IF (NUM_SUM_THREEDS .EQ. 2) THEN
            IF (MOD(I,2) .EQ. 0) THEN
                J = 2
            ENDIF
        ENDIF
        ! Accumulate sums of the volumes
        CALL THREED_ADD(THREED_SUM(J),INPUT_THREED,THREED_SUM(J))
    ENDDO

    ! We can now finalise the 3D
    IF (NUM_SUM_THREEDS .EQ. 2) THEN
        ! There were 2 sums collected. They should each contain only one reconstruction
        IF (INPUT_THREED_NUM_THREEDS .EQ. 1) THEN
            CALL THREED_MERGE(THREED_SUM(1),THREED_SUM(2))
        ELSE
            WRITE(*,'(A)') '**ERROR(MODE_FINALISE_3DS): programming error'
            CALL TERMINATE('Fatal error in MODE_FINALISE_3DS')
        ENDIF
    ENDIF
    IF (CALC_FSC%VALUE) THEN
        CALL THREED_EVAL_RECONSTRUCTION(THREED_SUM(1),THREED_FINAL(1),FSC_VOLUMES=FSC_VOLUMES,IMG_AMPL=AMPL_FINAL(1))
    ELSE
        CALL THREED_EVAL_RECONSTRUCTION(THREED_SUM(1),THREED_FINAL(1),IMG_AMPL=AMPL_FINAL(1))
    ENDIF

    ! Deal with FSC
    IF (CALC_FSC%VALUE) THEN
        ! Write the 2 FSC volumes to disk
        CALL FILENAME_ADD_SUFFIX(OUTPUT_THREED_FILENAME%VALUE,'_fsc1',TMP_FILENAME)
        PRINT '(2A)', '**info(MODE_FINALISE_3DS): writing FSC volume to ', TRIM(TMP_FILENAME)
        CALL WRITE_TO_DISK(FSC_VOLUMES(1),TMP_FILENAME)

        CALL FILENAME_ADD_SUFFIX(OUTPUT_THREED_FILENAME%VALUE,'_fsc2',TMP_FILENAME)
        PRINT '(2A)', '**info(MODE_FINALISE_3DS): writing FSC volume to ', TRIM(TMP_FILENAME)
        CALL WRITE_TO_DISK(FSC_VOLUMES(2),TMP_FILENAME)

        ! Print out the FSC
        CALL THREED_PRINT_FSC(THREED_SUM(1))

        ! Do an FRC between the central slices
        CENTRAL_SLICE = INDEX_OF_CENTRAL_PIXEL_FROM_DIM(THREED_FINAL(1)%logical_dimensions(3))
        CALL IMAGE_GET_SLICE(FSC_VOLUMES(1),CENTRAL_SLICES(1),3,CENTRAL_SLICE)
        CALL IMAGE_GET_SLICE(FSC_VOLUMES(2),CENTRAL_SLICES(2),3,CENTRAL_SLICE)
        CALL COMPUTE_FSC(CENTRAL_SLICES(1),CENTRAL_SLICES(2),FOUSHELL)
        WRITE(*,'(A)') '**info(MODE_FINALISE_3DS): FRC between central slices of two half-dataset reconstructions follows'
        CALL PRINT_FSC(FOUSHELL)
    ENDIF


    ! We can write it to disk
    CALL THREED_WRITE_TO_DISK(THREED_SUM(1),OUTPUT_THREED_FILENAME%VALUE)
    call threed_final(1)%WriteToDisk(OUTPUT_THREED_FILENAME%VALUE,PIXEL_SIZE=PIXEL_SIZE%VALUE,PRINT_FILE_INFO=.TRUE.)
    IF (INPUT_THREED%RECONSTRUCT_AMPLITUDES) THEN
        CALL FILENAME_ADD_SUFFIX(OUTPUT_THREED_FILENAME%VALUE,'_ampl',OUTPUT_AMPL_FILENAME)
        CALL AMPL_FINAL(1)%WriteToDisk(OUTPUT_AMPL_FILENAME,PIXEL_SIZE=PIXEL_SIZE%VALUE,PRINT_FILE_INFO=.TRUE.)
    ENDIF

    ! Let's also prepare a version which is cropped, symmetrised & masked
    WRITE(*,'(A)') '**info(MODE_FINALISE_3DS): preparing a cropped, symmetrised and masked version of the final 3D reconstruction'
    CROP_DIM = NINT(2.*1.1*HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE)
    IF (IS_ODD(CROP_DIM(1))) CROP_DIM = CROP_DIM + 1 ! Not necessary for Frealix, but may be so for other software
    CALL CROP(THREED_FINAL(1),CROP_DIM)
    CALL IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW(THREED_FINAL(1),HELIX,INTERPOL_SCHEME,MAX_RADIUS=HELIX_MAX_RADIUS%VALUE/PIXEL_SIZE%VALUE)
    IF (HELICAL_MASKING%VALUE) THEN
        CALL IMAGE_HELIX_MASK(THREED_FINAL(1),HELIX,ASYM_UNIT_MW%VALUE,PIXEL_SIZE%VALUE,HELIX_MAX_RADIUS%VALUE)
        CALL FILENAME_ADD_SUFFIX(OUTPUT_THREED_FILENAME%VALUE,'_hm',OUTPUT_THREED_FILENAME%VALUE)
    ELSE
        CALL FILENAME_ADD_SUFFIX(OUTPUT_THREED_FILENAME%VALUE,'_h',OUTPUT_THREED_FILENAME%VALUE)
    ENDIF
    CALL THREED_FINAL(1)%WriteToDisk(OUTPUT_THREED_FILENAME%VALUE,PIXEL_SIZE=PIXEL_SIZE%VALUE)

END SUBROUTINE MODE_FINALISE_3DS_SLAVE

END MODULE MODE_FINALISE_3DS
