!>  \brief  Given a model 3D, find occurences of the filament in a micrograph
!!
MODULE MODE_FIND_FILAMENTS
    USE RUNTIME_PARAMETERS
    USE MODES
    USE RTPS


    IMPLICIT NONE
    TYPE(MODE) :: FIND_FILAMENTS

    TYPE(RTP_CHAR), TARGET  ::  INPUT_3D_MODEL  =               &
        RTP_CHAR(   VALUE='model3d.mrc',                  &
                    KEYWORD='INPUT_3D_MODEL',           &
                    DESCRIPTION='Model 3D reconstruction')


    TYPE(RTP_CHAR), TARGET  ::  INPUT_FILM_FILE  =               &
        RTP_CHAR(   VALUE='film.mrc',                  &
                    KEYWORD='INPUT_FILM_FILE',           &
                    DESCRIPTION='Micrograph to pick from')

CONTAINS

SUBROUTINE MODE_FIND_FILAMENTS_INIT_AND_RUN
    IMPLICIT NONE
    ! Intialise the mode
    CALL FIND_FILAMENTS%INIT(COMMAND='FIND_FILAMENTS',SUBROUTINE=MODE_FIND_FILAMENTS_SLAVE,SHORT_DESCRIPTION='Add description here')

    ! Add RTPs to the mode
    CALL FIND_FILAMENTS%ADD_RTP(INPUT_3D_MODEL)
    CALL FIND_FILAMENTS%ADD_RTP(INPUT_FILM_FILE)

    ! Run the mode
    CALL FIND_FILAMENTS%RUN()
END SUBROUTINE MODE_FIND_FILAMENTS_INIT_AND_RUN

SUBROUTINE MODE_FIND_FILAMENTS_SLAVE
    USE IMAGES
    USE IMAGES_COMPARISON
    USE IMAGES_ARITHM
    USE IMAGES_THREED
    USE ROTATION_MATRICES
    use INTERPOLATION, only : INTERPOLATION_SCHEME, INTERPOLATION_SCHEME_INIT
    use ImageFiles
    IMPLICIT NONE
    !
    ! Private variables
    !

    ! Imagefile objects
    TYPE(IMAGEFILE) ::  MODEL_FILE
    TYPE(IMAGEFILE) ::  FILM_FILE
    TYPE(IMAGEFILE) ::  CCFS_FILE
    TYPE(IMAGEFILE) ::  REFS_FILE
    TYPE(IMAGEFILE) ::  CCF_SUM_FILE
    TYPE(IMAGEFILE) ::  CCF_MAX_FILE
    TYPE(IMAGEFILE) ::  CCF_VAR_FILE

    ! Image objects
    TYPE(IMAGE)     ::  MODEL_3D
    TYPE(IMAGE)     ::  MODEL_PROJECTION, MODEL_PROJECTION_ROT
    TYPE(IMAGE)     ::  MICROGRAPH
    TYPE(IMAGE)     ::  CCF, CCF_SUM, CCF_MAX, CCF2_SUM, CCF_VAR

    ! Variables for the histogram
    INTEGER,    ALLOCATABLE ::  HISTOGRAM(:)
    REAL,       ALLOCATABLE ::  BIN_BOUNDARIES(:)

    ! Search parameters
    REAL,       PARAMETER   ::  IN_PLANE_ANGLE_STEP_SIZE = 10.0
    REAL                    ::  CURRENT_IN_PLANE_ROTATION

    ! Other variables
    REAL                        ::  MATRIX(4,4)
    TYPE(INTERPOLATION_SCHEME)  ::  INTERPOL_SCHEME
    TYPE(WINDOW)                ::  INTERPOL_WINDOW

    !
    ! Get variable values from control file
    !


    !
    ! Start work
    !

    ! Set up interpolation
    CALL INTERPOLATION_SCHEME_INIT(INTERPOL_SCHEME,'LINEAR','RECT',2)
    CALL INTERPOL_WINDOW%AllocateWindow(INTERPOL_SCHEME%WINDOW_WIDTH,COMP=.FALSE.,THREED=.FALSE.)

    ! Read in the 3D model
    CALL MODEL_FILE%INIT(INPUT_3D_MODEL%VALUE)
    call model_3d%ReadFromDisk(model_file,read_volume_not_slice=.true.)
    CALL MODEL_FILE%CLOSE()
    ! Read in the film
    CALL FILM_FILE%INIT(INPUT_FILM_FILE%VALUE)
    call micrograph%ReadFromDisk(film_file)
    CALL FILM_FILE%CLOSE()
    ! Band pass the film
    CALL BAND_PASS(MICROGRAPH,.05,.3,.025)
    CALL WRITE_TO_DISK(MICROGRAPH,'dbg_film_bp.mrc')
    ! Project the 3D model in a single direction
    CALL VOLUME_PROJECT_LEAN(MODEL_3D,MODEL_PROJECTION,0.0D0,90.0D0,0.0D0,INTERPOL_SCHEME,INTERPOL_WINDOW)
    ! Here, multiply the projection by CTF^2 (in the meantime, just multiply by -1)
    CALL IMAGE_MULTIPLY_BY_CONSTANT(MODEL_PROJECTION,-1.0)
    ! Dump the projection
    CALL WRITE_TO_DISK(MODEL_PROJECTION,'dbg_proj.mrc')

    ! Initialise output files
    CALL CCFS_FILE%INIT('dbg_ccfs.mrc')
    CALL CCF_SUM_FILE%INIT('dbg_ccf_sum.mrc')
    CALL CCF_MAX_FILE%INIT('dbg_ccf_max.mrc')
    CALL CCF_VAR_FILE%INIT('dbg_ccf_var.mrc')
    CALL REFS_FILE%INIT('dbg_refs.mrc')

    ! Loop over in-plane rotations of the reference
    CALL IMAGE_ALLOCATE(CCF_MAX,MICROGRAPH)
    CCF_MAX%real_values = -99999.99999
    CURRENT_IN_PLANE_ROTATION = 0.0
    DO
        PRINT *, '**debug(MODE_FIND_FILAMENT): in-plane rotation = ', CURRENT_IN_PLANE_ROTATION
        ! Get rotation matrix for current rotation
        MATRIX = EULER_TO_ROTMAT(0.0D0,0.0D0,DBLE(CURRENT_IN_PLANE_ROTATION),REVERSE=.TRUE.)
        CALL PRINT_ROTMAT(MATRIX)
        ! Apply current rotation
        CALL ROTATE_IMAGE(MODEL_PROJECTION,MODEL_PROJECTION_ROT,MATRIX,INTERPOL_SCHEME,INTERPOL_WINDOW)
        call MODEL_PROJECTION_ROT%WriteToDisk(REFS_FILE)
        ! Pad the projection to match the dimensions of the micrograph
        CALL PAD(MODEL_PROJECTION_ROT,MICROGRAPH%logical_dimensions)
        ! Compute a CCF
        CALL FAST_CORRELATION(MODEL_PROJECTION_ROT,MICROGRAPH,CCF)
        call ccf%WriteToDisk(CCFS_FILE)
        ! Prepare sum of CCFs
        IF (.NOT. CCF_SUM%IsAllocated()) CALL CCF_SUM%Allocate(CCF)
        IF (.NOT. CCF2_SUM%IsAllocated()) CALL CCF2_SUM%Allocate(CCF)
        ! Accumulate sum of CCFs
        CALL ADD_IMAGES(CCF,CCF_SUM,CCF_SUM)
        ! Accumulate sum of CCFs^2
        CCF2_SUM%real_values = CCF2_SUM%real_values + CCF%real_values**2
        ! Keep track of max CCF for every pixel
        CALL GT_IMAGE(CCF,CCF_MAX,CCF_MAX)
        ! Increment the in-plane rotation
        CURRENT_IN_PLANE_ROTATION = CURRENT_IN_PLANE_ROTATION + IN_PLANE_ANGLE_STEP_SIZE
        IF(CURRENT_IN_PLANE_ROTATION .GE. 360.0) EXIT
    ENDDO
    ! Do the variance calculation
    CALL IMAGE_ALLOCATE(CCF_VAR,CCF)
    CCF_VAR%real_values = CCF2_SUM%real_values - CCF_SUM%real_values**2
    ! Dump the sum & max of CCFs
    call ccf_sum%WriteToDisk(ccf_sum_file)
    call ccf_max%WriteToDisk(ccf_max_file)
    call ccf2_sum%WriteToDisk(ccf_var_file)
    ! Compute histogram of CCF & print it out
    PRINT *, '**histogram of ccf_max'
    CALL CCF_MAX%ComputeHistogram(HISTOGRAM,BIN_BOUNDARIES,NUM_OF_BINS=300,PRINT=.TRUE.)
    PRINT *, '**histogram of ccf_var'
    CALL CCF_VAR%ComputeHistogram(HISTOGRAM,BIN_BOUNDARIES,NUM_OF_BINS=300,PRINT=.TRUE.)

    ! Close output files
    CALL CCFS_FILE%CLOSE()
    CALL CCF_SUM_FILE%CLOSE()
    CALL CCF_MAX_FILE%CLOSE()
    CALL CCF_VAR_FILE%CLOSE()
    CALL REFS_FILE%CLOSE()

END SUBROUTINE MODE_FIND_FILAMENTS_SLAVE

END MODULE MODE_FIND_FILAMENTS
