MODULE MODE_FREALIGN_STACK_TO_FREALIX_ALI
    USE MODES
    USE RTPS
    IMPLICIT NONE
    ! Declare the mode
    TYPE(MODE)  ::  M

    ! Declare run-time parameters
    TYPE(RTP_CHAR), TARGET      ::  FREALIGN_PAR_FILENAME   =   RTP_CHAR(   &
                                    VALUE='frealign.par',                   &
                                    KEYWORD='FREALIGN_PAR_FILENAME',        &
                                    DESCRIPTION='Name of input frealign parameter file')

    TYPE(RTP_CHAR), TARGET      ::  STACK_FILENAME          =   RTP_CHAR(   &
                                    VALUE='stack.mrc',                      &
                                    KEYWORD='STACK_FILENAME',               &
                                    DESCRIPTION='The stack of 2D images processed by Frealign')

    TYPE(RTP_CHAR), TARGET      ::  FILM_FILENAMES          =   RTP_CHAR(   &
                                    VALUE='film_????.mrc',                  &
                                    KEYWORD='FILM_FILENAMES',               &
                                    DESCRIPTION='Pattern for film filenames. Question marks replace single characters.')

    TYPE(RTP_CHAR), TARGET      ::  FILM_SELECTION          =   RTP_CHAR(   &
                                    VALUE='1,6,7-9',                        &
                                    KEYWORD='FILM_SELECTION',               &
                                    DESCRIPTION='Which films to analyse. Comma-separated list where - indicates range.')

    TYPE(RTP_CHAR), TARGET      ::  ALI_FILENAME            =   RTP_CHAR(   &
                                    VALUE='waypoints.ali',                  &
                                    KEYWORD='ALI_FILENAME',                 &
                                    DESCRIPTION='Output alignment file name')

    TYPE(RTP_REAL), TARGET      ::  PIXEL_SIZE              =   RTP_REAL(   &
                                    VALUE=1.0,                              &
                                    KEYWORD='PIXEL_SIZE',                   &
                                    DESCRIPTION='Pixel size of the Frealign parameter file, in Angstroms')

    TYPE(RTP_INTG), TARGET      ::  Y_MAX                   =   RTP_INTG(   &
                                    VALUE=0,                                &
                                    KEYWORD='Y_MAX',                        &
                                    DESCRIPTION='Maximum Y address at which to look for single particle. '//&
                                    'Useful to avoid film label.')

    TYPE(RTP_INTG), TARGET      ::  BINNING_FACTOR          =   RTP_INTG(   &
                                    VALUE=1,                                &
                                    KEYWORD='BINNING_FACTOR',               &
                                    DESCRIPTION='Images (stack & films) are binned by this factor relative to the '//&
                                    'Frealign parameter file')

    TYPE(RTP_LOGI), TARGET      ::  CHECK_180_ROTATION      =   RTP_LOGI(   &
                                    VALUE=.FALSE.,                          &
                                    KEYWORD='CHECK_180_ROTATION',           &
                                    DESCRIPTION='For each image, look for its 180-degree-rotated version in the film images also')

CONTAINS

SUBROUTINE MODE_FREALIGN_STACK_TO_FREALIX_ALI_INIT_AND_RUN
    IMPLICIT NONE
    ! Initialise the mode
    CALL M%INIT(    COMMAND='FREALIGN_STACK_TO_FREALIX_ALI',                &
                    SUBROUTINE=MODE_FREALIGN_STACK_TO_FREALIX_ALI_SLAVE,    &
                    SHORT_DESCRIPTION='Convert alignment parameters from Frealign to Frealix')

    ! Add RTPs to the mode
    CALL M%ADD_RTP(FREALIGN_PAR_FILENAME)
    CALL M%ADD_RTP(STACK_FILENAME)
    CALL M%ADD_RTP(FILM_FILENAMES)
    CALL M%ADD_RTP(FILM_SELECTION)
    CALL M%ADD_RTP(ALI_FILENAME)
    CALL M%ADD_RTP(PIXEL_SIZE)
    CALL M%ADD_RTP(Y_MAX)
    CALL M%ADD_RTP(BINNING_FACTOR)
    CALL M%ADD_RTP(CHECK_180_ROTATION)
    ! Run the mode
    CALL M%RUN()
END SUBROUTINE MODE_FREALIGN_STACK_TO_FREALIX_ALI_INIT_AND_RUN


SUBROUTINE MODE_FREALIGN_STACK_TO_FREALIX_ALI_SLAVE
    USE SINGLE_PARTICLES
    USE FREALIGN_PARAMETERS
    USE FREALIX_PARAMETERS
    USE PARSERS
    USE RUNTIME_PARAMETERS
    USE IMAGES
    USE IMAGES_ARITHM
    USE PEAKS
    USE FILAMENTS_MODEL_BASE, ONLY : FILAMENT_PRINT_INFO
    USE UI, ONLY : INTERVAL_PRINT
    USE FILAMENTS
    USE ROTATION_MATRICES, ONLY : EULER_TO_ROTMAT
    USE INTERPOLATION
    use ImageFiles
	IMPLICIT NONE
    ! Private variables
    TYPE(SINGLE_PARTICLE),  ALLOCATABLE ::  SP(:)
    INTEGER,                ALLOCATABLE ::  LIST_OF_FILMS(:)
    INTEGER                             ::  I
    INTEGER                             ::  FILM_NUM
    CHARACTER(LEN=FILENAME_LEN)         ::  FILM_FILENAME
    TYPE(IMAGE)                         ::  FILM_IMG
    INTEGER                             ::  CURRENT_SP
    TYPE(IMAGEFILE)                     ::  SP_IMAGEFILE, FILM_IMAGEFILE
    TYPE(IMAGE)                         ::  SP_IMG
    TYPE(IMAGE)                         ::  TMP_IMG
    TYPE(IMAGE)                         ::  CORR
    TYPE(PEAK)                          ::  BEST_PEAK
    TYPE(PEAK),             ALLOCATABLE ::  PEAKS_FOUND(:,:,:)  ! First dimension: location in stack, second: film index, third: 180-deg rotation (1 if not, 2 if so)
    LOGICAL,                PARAMETER   ::  DEBUG   =   .FALSE.
    TYPE(IMAGE)                         ::  FILM_LOCAL_AVERAGE
    TYPE(IMAGE)                         ::  FILM_LOCAL_STD_DEV
    TYPE(IMAGE)                         ::  FILM_SQUARED
    INTEGER                             ::  SP_DIM(3), FILM_ORI(3)
    INTEGER                             ::  FOUND_FILM_INDEX
    INTEGER                             ::  NUMBER_OF_FILAMENTS, CURRENT_FILAMENT
    TYPE(FILAMENT),         ALLOCATABLE ::  FILS(:)
    TYPE(WAYPOINT)                      ::  CURRENT_WP
    TYPE(WAYPOINT),         ALLOCATABLE ::  WPS(:)
    LOGICAL                             ::  NEW_FILAMENT
    LOGICAL                             ::  LAST_WP_IN_FILAMENT
    INTEGER                             ::  ROTATIONS_TO_CHECK, CURRENT_ROTATION, BEST_ROTATION
    REAL                                ::  ROTMAT(4,4)
    TYPE(IMAGE)                         ::  TMPIMG
    TYPE(INTERPOLATION_SCHEME)          ::  INTERPOL
    TYPE(WINDOW)                        ::  INTERPOL_WINDOW
    ! Start work

    ! Grab values from the input Frealign parameter file
    SP = FREALIGN_PAR_TO_SP(FREALIGN_PAR_FILENAME%VALUE)

    ! Work out how many filaments are described (Frealign 8.09 stores
    ! filament number in the FILM column)
    NUMBER_OF_FILAMENTS = 1
    DO CURRENT_SP=2,SIZE(SP)
        IF (SP(CURRENT_SP)%FILM_NUM .NE. SP(CURRENT_SP-1)%FILM_NUM) NUMBER_OF_FILAMENTS = NUMBER_OF_FILAMENTS + 1
    ENDDO
    WRITE(*,'(A,I0)') '**info(MODE_FREALIGN_STACK_TO_FREALIX_ALI): number of filaments = ', NUMBER_OF_FILAMENTS
    ALLOCATE(FILS(NUMBER_OF_FILAMENTS))

    ! Work out how many rotations we'll check
    ROTATIONS_TO_CHECK = 1
    IF (CHECK_180_ROTATION%VALUE) ROTATIONS_TO_CHECK=2
    ROTMAT = EULER_TO_ROTMAT(0.0D0,0.0D0,180.0D0,REVERSE=.TRUE.)

    ! Initialise interpolation
    CALL INTERPOLATION_SCHEME_INIT(INTERPOL,'LINEAR','RECT',2)


    ! Open the stack file
    CALL SP_IMAGEFILE%INIT(FILENAME=STACK_FILENAME%VALUE)

    !
    LIST_OF_FILMS = SET_TO_LIST(FILM_SELECTION%VALUE)

    ALLOCATE(PEAKS_FOUND(SIZE(SP),SIZE(LIST_OF_FILMS),ROTATIONS_TO_CHECK))

    !
    ! Loop over films / micrographs, and over single particles
    ! and find the best correlation peak between all combinations
    !
    DO I=1,SIZE(LIST_OF_FILMS)

        WRITE(*,'(2(A,I0))') '**info(MODE_FREALIGN_STACK_TO_FREALIX_ALI): on film ', I, ' of ', SIZE(LIST_OF_FILMS)

        FILM_NUM = LIST_OF_FILMS(I)

        ! Check the micrograph file exists
        CALL FILENAME_FROM_WILD(FILM_FILENAMES%VALUE,FILM_NUM,FILM_FILENAME)
        IF (.NOT. FILE_EXISTS(FILM_FILENAME)) THEN
            WRITE(*,'(2A)') '**ERROR(MODE_FREALIGN_STACK_TO_FREALIX_ALI): file does not exist: ', FILM_FILENAME
            CALL TERMINATE('Fatal error in MODE_FREALIGN_STACK_TO_FREALIX_ALI')
        ENDIF

        ! Read the micrograph into memory
        WRITE(*,'(A)') '**info(MODE_FREALIGN_STACK_TO_FREALIX_ALI): reading film into memory'
        CALL FILM_IMAGEFILE%INIT(FILM_FILENAME)
        CALL FILM_IMAGEFILE%PrintInfo()
        call FILM_IMG%ReadFromDisk(FILM_IMAGEFILE)
        CALL FILM_IMAGEFILE%CLOSE()
        if (film_img%is_in_real_space .and. film_img%origin_at_center) then
            film_ori = film_img%physical_address_of_box_center
        else
            film_ori = [1,1,1]
        endif

		! Pre-process the micrograph
		CALL COSINE_RECT(FILM_IMG,REAL(FILM_IMG%logical_dimensions(1:2)-26),    &
		                 AZIMUTH=0.0,FALLOFF=10.0,WIPE_INPUT=.FALSE.,MASK_WITH_AVERAGE=.TRUE.)
		IF (Y_MAX%SET_BY_USER .AND. Y_MAX%VALUE .NE. 0 .AND. Y_MAX%VALUE .LT. FILM_IMG%logical_dimensions(2)) THEN
			FILM_IMG%real_values(:,Y_MAX%VALUE:FILM_IMG%logical_dimensions(2),:) = AVERAGE_REAL(FILM_IMG)
		ENDIF
		!CALL WRITE_TO_DISK(FILM_IMG,'dbg_film_rect.mrc')

		! Compute the square of the film image
		CALL IMAGE_ALLOCATE(FILM_SQUARED,FILM_IMG) ! We will need this later
        FILM_SQUARED%real_values = FILM_IMG%real_values ** 2


        WRITE(*,'(2A)') '**info(MODE_FREALIGN_STACK_TO_FREALIX_ALI): computing the micrograph''s local', &
                                                ' average and standard deviation'

        ! Prepare the film & its squared version (FT, shift origin)
        CALL FWD_FT(FILM_IMG)
        CALL FWD_FT(FILM_SQUARED)
        CALL SHIFT_REAL_SPACE_ORIGIN(FILM_IMG)
        CALL SHIFT_REAL_SPACE_ORIGIN(FILM_SQUARED)

        ! Compute the local average at each point on the film
        ! (where "local" is defined by the mask over the single particles)
        ! (NOTE: we are going to compute the local normalised correlation coefficient, as
        ! described in Roseman, 2003)

        ! First, we create an image of the "mask", in our case the full box
        call sp_img%ReadFromDisk(SP_IMAGEFILE,wanted_location=1)
        SP_IMG%real_values = 1.0
        SP_IMG%ORIGIN_AT_CENTER = .TRUE.
        SP_DIM = SP_IMG%logical_dimensions

        ! Prepare the mask for convolution with the film
        CALL PAD(SP_IMG,FILM_IMG%logical_dimensions)
        CALL FWD_FT(SP_IMG)
        CALL SHIFT_REAL_SPACE_ORIGIN(SP_IMG)

        ! Compute the convolution
        CALL IMAGE_ALLOCATE(FILM_LOCAL_AVERAGE,FILM_IMG)
        FILM_LOCAL_AVERAGE%complex_values = FILM_IMG%complex_values * SP_IMG%complex_values

        ! Scale to give local average
        CALL SHIFT_REAL_SPACE_ORIGIN(FILM_LOCAL_AVERAGE)
        CALL BWD_FT(FILM_LOCAL_AVERAGE)
        FILM_LOCAL_AVERAGE%real_values = FILM_LOCAL_AVERAGE%real_values * &
                                        REAL(PRODUCT(FILM_IMG%logical_dimensions)) / REAL(PRODUCT(SP_DIM)) ! multiplication by film%logical_dimensions is to get correct scaling of the convolution
        !IF (DEBUG) THEN
        !    CALL WRITE_TO_DISK(FILM_LOCAL_AVERAGE,'dbg_film_local_average.mrc')
        !ENDIF

        ! Compute the local standard deviation at each point on the film
        CALL IMAGE_ALLOCATE(FILM_LOCAL_STD_DEV,FILM_IMG)
        FILM_LOCAL_STD_DEV%complex_values = FILM_SQUARED%complex_values * SP_IMG%complex_values
        CALL SHIFT_REAL_SPACE_ORIGIN(FILM_LOCAL_STD_DEV)
        CALL BWD_FT(FILM_LOCAL_STD_DEV)
        FILM_LOCAL_STD_DEV%real_values = FILM_LOCAL_STD_DEV%real_values * &
                                        REAL(PRODUCT(FILM_IMG%logical_dimensions)) / REAL(PRODUCT(SP_DIM)) &
                                        - FILM_LOCAL_AVERAGE%real_values
        WHERE (FILM_LOCAL_STD_DEV%real_values .GT. 0.0)
            FILM_LOCAL_STD_DEV%real_values = SQRT(FILM_LOCAL_STD_DEV%real_values)
        ELSEWHERE
            FILM_LOCAL_STD_DEV%real_values = 0.0
        END WHERE
        IF (DEBUG) THEN
            CALL WRITE_TO_DISK(FILM_LOCAL_STD_DEV,'dbg_film_local_std_dev.mrc')
        ENDIF
        CALL IMAGE_DEALLOCATE(FILM_SQUARED)  ! Won't need this anymore

        CALL IMAGE_DEALLOCATE(SP_IMG)   ! This will be re-allocated later, in OpenMP parallel region

        !$OMP   PARALLEL            &
        !$OMP   DEFAULT(NONE)           &
        !$OMP   PRIVATE(CURRENT_SP,SP_IMG,CORR,CURRENT_ROTATION,TMP_IMG,TMPIMG,INTERPOL_WINDOW)   &
        !$OMP   SHARED(SP,SP_IMAGEFILE,FILM_IMG,FILM_LOCAL_STD_DEV,SP_DIM,PEAKS_FOUND,Y_MAX,I,ROTATIONS_TO_CHECK,FILM_NUM,INTERPOL)

        CALL IMAGE_INIT(CORR)
        CALL IMAGE_INIT(SP_IMG)
        CALL IMAGE_INIT(TMP_IMG)
        CALL IMAGE_INIT(TMPIMG)
        CALL ALLOCATE_WINDOW(INTERPOL_WINDOW,INTERPOL%WINDOW_WIDTH,COMP=.FALSE.,THREED=.FALSE.)

        ! Loop over the stack of single-particle images
        !$OMP   DO
        DO CURRENT_SP=1,SIZE(SP)
		!DO CURRENT_SP=1,1
            IF (INTERVAL_PRINT(CURRENT_SP,SIZE(SP))) THEN
                !$OMP   CRITICAL    (WRITE_STATEMENT_OMP_CRIT)
                WRITE(*,'(3(A,I0),A)') '**info(MODE_FREALIGN_STACK_TO_FREALIX_ALI): on particle ', CURRENT_SP, ' of ', SIZE(SP), &
                                     ' (loc in stack: ', SP(CURRENT_SP)%LOC_IN_STACK, ')'
                !$OMP   END CRITICAL    (WRITE_STATEMENT_OMP_CRIT)
            ENDIF

            DO CURRENT_ROTATION=1,ROTATIONS_TO_CHECK

                ! Load the single particle image into memory, and give it 0.0 mean and std dev of 1.0 (see Roseman 2003)
                !$OMP   CRITICAL    (READ_IMAGE_OMP_CRIT)
                call sp_img%ReadFromDisk(sp_imagefile,wanted_location=sp(current_sp)%LOC_IN_STACK)
                !$OMP   END CRITICAL    (READ_IMAGE_OMP_CRIT)
				!CALL COSINE_DISC(SP_IMG,RAD=SP_IMG%logical_dimensions(1)*0.2,FALLOFF=5.0,WIPE_INPUT=.FALSE.,MASK_WITH_AVERAGE=.TRUE.)
				SP_IMG%ORIGIN_AT_CENTER = .TRUE.
                SP_IMG%real_values = SP_IMG%real_values - AVERAGE_REAL(SP_IMG)
                SP_IMG%real_values = SP_IMG%real_values / STD_DEV_REAL(SP_IMG)
                IF (DEBUG) THEN
                    CALL WRITE_TO_DISK(SP_IMG,'dbg_sp_image_normalised.mrc')
                ENDIF
                IF (CURRENT_ROTATION .EQ. 2) THEN
                    ! Rotate the single particle image by 180 degrees
                    !CALL IMAGE_ROTATE_180(SP_IMG) ! This doesn't preserve the centering of the image
                    TMP_IMG = SP_IMG
                    CALL ROTATE_IMAGE(TMP_IMG,SP_IMG,REAL(EULER_TO_ROTMAT(0.0D0,0.0D0,180.0D0)),INTERPOL,INTERPOL_WINDOW)
                ENDIF
                CALL PAD(SP_IMG,FILM_IMG%logical_dimensions)
                CALL FWD_FT(SP_IMG)
                CALL SHIFT_REAL_SPACE_ORIGIN(SP_IMG)

                ! Compute the correlation
                CALL IMAGE_ALLOCATE(CORR,FILM_IMG)
                CORR%is_in_real_space = .false.
                CORR%ORIGIN_AT_CENTER = .FALSE.
                CORR%complex_values = FILM_IMG%complex_values * CONJG(SP_IMG%complex_values)
                CALL SHIFT_REAL_SPACE_ORIGIN(CORR)
                CALL BWD_FT(CORR)
                CORR%real_values = CORR%real_values * REAL(PRODUCT(CORR%logical_dimensions))

                ! Normalise the correlation
                IF (.not. CORR%is_in_real_space) CALL TERMINATE('OOPS')
                WHERE (FILM_LOCAL_STD_DEV%real_values .lt. 0.001)
                    CORR%real_values = 0.0
                ELSEWHERE
                    CORR%real_values =   CORR%real_values / REAL(PRODUCT(SP_DIM)) / FILM_LOCAL_STD_DEV%real_values
                END WHERE


                ! Find the peak
                !$OMP   CRITICAL    (RECORD_PEAK_OMP_CRIT)
                !PEAKS_FOUND(CURRENT_SP,I,CURRENT_ROTATION) = PEAK_FROM_IMAGE_INTG(CORR,Y_MAX=Y_MAX%VALUE,Y_MIN=30)
                PEAKS_FOUND(CURRENT_SP,I,CURRENT_ROTATION) = PEAK_FROM_IMAGE(CORR,Y_MAX=Y_MAX%VALUE,Y_MIN=10)
                !$OMP   END CRITICAL    (RECORD_PEAK_OMP_CRIT)

                !
                !IF (DEBUG) THEN
                    !$OMP   CRITICAL    (WRITE_STATEMENT_OMP_CRIT)
                    WRITE(*,'(4(A,I0),A)',ADVANCE='NO') '**debug(MODE_FREALIGN_STACK_TO_FREALIX_ALI): single particle ', &
                                                        CURRENT_SP, ' of ', SIZE(SP), ' vs film # ', FILM_NUM, ' (', &
													    I , '); peak info: '
                    CALL PEAK_PRINT_INFO(PEAKS_FOUND(CURRENT_SP,I,CURRENT_ROTATION))
                    !$OMP   END CRITICAL    (WRITE_STATEMENT_OMP_CRIT)
                !ENDIF

                ! Debug: window out image
                IF (DEBUG) THEN
                    TMPIMG%logical_dimensions = SP_DIM
                    CALL SHIFT_REAL_SPACE_ORIGIN(FILM_IMG)
                    CALL BWD_FT(FILM_IMG)
                    print *, '**debug(): boxing out at coos', NINT(PEAKS_FOUND(CURRENT_SP,I,CURRENT_ROTATION)%COO)
                    CALL WINDOW_IMG(FILM_IMG,TMPIMG,NINT(PEAKS_FOUND(CURRENT_SP,I,CURRENT_ROTATION)%COO))
                    CALL WRITE_TO_DISK(TMPIMG,'dbg_out.mrc')
                    CALL FWD_FT(FILM_IMG)
                    CALL SHIFT_REAL_SPACE_ORIGIN(FILM_IMG)
                    CALL WRITE_TO_DISK(CORR,'dbg_corr.mrc')
					CALL SHIFT_REAL_SPACE_ORIGIN(SP_IMG)
					CALL BWD_FT(SP_IMG)
					CALL WRITE_TO_DISK(SP_IMG,'dbg_sp.mrc')
					print *, '**debug(): dumped correlation function and boxed image'
					call terminate('debug')
                ENDIF

            ENDDO   ! End of loop over rotations

        ENDDO   ! End of loop over single particles
        !$OMP   ENDDO

        ! Deallocate any objects that were allocated within parallel region
        CALL IMAGE_DEALLOCATE(SP_IMG)
        CALL IMAGE_DEALLOCATE(CORR)
        CALL IMAGE_DEALLOCATE(TMP_IMG)
        CALL IMAGE_DEALLOCATE(TMPIMG)
        CALL DEALLOCATE_WINDOW(INTERPOL_WINDOW)

        !$OMP   END PARALLEL

    ENDDO   ! End of loop over films

    ! Close image files
    CALL SP_IMAGEFILE%CLOSE()

    !
    ! Now look for the best peak for each location in the stack of single particles,
    ! and turn single particles into filament waypoints
    !
    CURRENT_FILAMENT = 0
    DO CURRENT_SP=1,SIZE(SP)

        ! Work out whether we're starting a new filament
        NEW_FILAMENT = .FALSE.
        IF (CURRENT_SP .EQ. 1) THEN
            NEW_FILAMENT = .TRUE.
        ELSE
            IF (SP(CURRENT_SP)%FILM_NUM .NE. SP(CURRENT_SP-1)%FILM_NUM) THEN
                ! We've just moved on to a new filament
                NEW_FILAMENT = .TRUE.
            ENDIF
        ENDIF

        ! If we're starting a new filament, reset/prepare array of waypoints
        IF (NEW_FILAMENT) THEN
            ! Increment
            IF (DEBUG) WRITE(*,'(A,I0)') '**debug(MODE_FREALIGN_STACK_TO_FREALIX_ALI): working on new filament, ID ', &
                                            SP(CURRENT_SP)%FILM_NUM
            CURRENT_FILAMENT = CURRENT_FILAMENT + 1
            ! Reset the array of waypoints
            IF (ALLOCATED(WPS)) DEALLOCATE(WPS)
        ENDIF

        ! Look for the film to which this single particle belongs
        BEST_PEAK%rvalue = -10.0
        BEST_ROTATION = 1
        DO I=1,SIZE(LIST_OF_FILMS)
            DO CURRENT_ROTATION=1,ROTATIONS_TO_CHECK
                IF (PEAKS_FOUND(CURRENT_SP,I,CURRENT_ROTATION)%rvalue .GT. BEST_PEAK%rvalue) THEN
                    BEST_PEAK = PEAKS_FOUND(CURRENT_SP,I,CURRENT_ROTATION)
                    FOUND_FILM_INDEX = I
                    BEST_ROTATION =  CURRENT_ROTATION
                ENDIF
            ENDDO
        ENDDO ! End of loop over films

        ! Print out result
        WRITE(*,'(A,I0,A,I0)',ADVANCE='NO') '**info(MODE_FREALIGN_STACK_TO_FREALIX_ALI): particle ', &
                                                CURRENT_SP, ' belongs to film ', FOUND_FILM_INDEX
        IF (BEST_ROTATION .EQ. 2) WRITE(*,'(A)',ADVANCE='NO') ' (with 180-deg rotation) '
        WRITE(*,'(A)',ADVANCE='NO')  ' peak info: '
        CALL PEAK_PRINT_INFO(BEST_PEAK)

        ! Convert the peak to a waypoint
        CURRENT_WP%COO = BEST_PEAK%COO
        CALL CTF_ANG_TO_PIX(SP(CURRENT_SP)%CTF,PIXEL_SIZE=PIXEL_SIZE%VALUE*REAL(BINNING_FACTOR%VALUE))
        CURRENT_WP%COO(3) = Z_FROM_CTF(SP(CURRENT_SP)%CTF)
        ! Apply the shifts from the single particle
        CURRENT_WP%COO(1) = CURRENT_WP%COO(1) - SP(CURRENT_SP)%SHIFT_X/REAL(BINNING_FACTOR%VALUE)
        CURRENT_WP%COO(2) = CURRENT_WP%COO(2) - SP(CURRENT_SP)%SHIFT_Y/REAL(BINNING_FACTOR%VALUE)
        CURRENT_WP%EULER  = SP(CURRENT_SP)%ANGLES
        IF (BEST_ROTATION .EQ. 2) THEN
            CURRENT_WP%COO(1) = - CURRENT_WP%COO(1)
            CURRENT_WP%COO(2) = - CURRENT_WP%COO(2)
            CURRENT_WP%EULER%PSI = CURRENT_WP%EULER%PSI + 180.
        ENDIF

        ! Add the waypoint to the array
        CALL WAYPOINTS_GROW_ARRAY(WPS)
        WPS(SIZE(WPS)) = CURRENT_WP

        ! Work out whether we've just processed the last waypoint in a filament
        LAST_WP_IN_FILAMENT = .FALSE.
        IF (CURRENT_SP .EQ. SIZE(SP)) THEN
            LAST_WP_IN_FILAMENT = .TRUE.
        ELSE
            IF (SP(CURRENT_SP)%FILM_NUM .NE. SP(CURRENT_SP+1)%FILM_NUM) THEN
                LAST_WP_IN_FILAMENT = .TRUE.
            ENDIF
        ENDIF

        ! If that was the last waypoint in the filament, we can populate the filament object
        IF (LAST_WP_IN_FILAMENT) THEN
            CALL FILAMENT_FROM_WPS( FILS(CURRENT_FILAMENT),WPS, &
                                    ID=SP(CURRENT_SP)%FILM_NUM, &
                                    FILM_ID=LIST_OF_FILMS(FOUND_FILM_INDEX  ),   &
                                    PIXEL_SIZE=PIXEL_SIZE%VALUE*REAL(BINNING_FACTOR%VALUE))
            IF (DEBUG) THEN
                WRITE(*,'(A)') '**debug(MODE_FREALIGN_STACK_TO_FREALIX_ALI): just populated the following filament:'
                CALL FILAMENT_PRINT_INFO(FILS(CURRENT_FILAMENT))
            ENDIF
        ENDIF

    ENDDO ! End of loop over particles

    ! Now that we have prepared the array of filaments, let's write a .ali file to disk
    CALL FILAMENTS_TO_FREALIX_PAR(FILS,ALI_FILENAME%VALUE)

END SUBROUTINE MODE_FREALIGN_STACK_TO_FREALIX_ALI_SLAVE

END MODULE MODE_FREALIGN_STACK_TO_FREALIX_ALI
