!>  \brief  Mode to measure the helical symmetry of an input image
MODULE MODE_HMEASURE
    ! Modules
    USE MODES
    USE RTPS
    IMPLICIT NONE
    TYPE(MODE) :: HMEASURE

    TYPE(RTP_CHAR), TARGET  ::  INPUT_FILENAME  =               &
        RTP_CHAR(   VALUE='test_in',                  &
                    KEYWORD='INPUT_FILENAME',           &
                    DESCRIPTION='')


    TYPE(RTP_REAL), TARGET  ::  TWIST  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='TWIST_PER_SUBUNIT',           &
                    DESCRIPTION='Degrees')


    TYPE(RTP_REAL), TARGET  ::  RISE  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='RISE_PER_SUBUNIT',           &
                    DESCRIPTION='Angstroms')


    TYPE(RTP_REAL), TARGET  ::  PIXEL_SIZE  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='PIXEL_SIZE',           &
                    DESCRIPTION='')


    TYPE(RTP_REAL), TARGET  ::  MAX_RAD  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='HELIX_MAX_RADIUS',           &
                    DESCRIPTION='Maximum radius (Angstroms)')


    TYPE(RTP_INTG), TARGET  ::  SYMMETRY_AXIAL  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='SYMMETRY_AXIAL',           &
                    DESCRIPTION='')


    TYPE(RTP_INTG), TARGET  ::  SYMMETRY_PERP  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='SYMMETRY_PERP',           &
                    DESCRIPTION='')

    TYPE(RTP_INTG), TARGET  ::  HELICAL_STARTS  =               &
        RTP_INTG(   VALUE=1,                  &
                    KEYWORD='NUM_HELICAL_STARTS',           &
                    DESCRIPTION='Number of helical starts')

  CONTAINS

SUBROUTINE MODE_HMEASURE_INIT_AND_RUN
    ! Intialise the mode
    CALL HMEASURE%INIT(  COMMAND='HMEASURE',SUBROUTINE=MODE_HMEASURE_SLAVE,    &
                        SHORT_DESCRIPTION='Estimate the helical symmetry parameters')

    ! Add RTPs to the mode
    CALL HMEASURE%ADD_RTP(INPUT_FILENAME)
    CALL HMEASURE%ADD_RTP(TWIST)
    CALL HMEASURE%ADD_RTP(RISE)
    CALL HMEASURE%ADD_RTP(PIXEL_SIZE)
    CALL HMEASURE%ADD_RTP(MAX_RAD)
    CALL HMEASURE%ADD_RTP(SYMMETRY_AXIAL)
    CALL HMEASURE%ADD_RTP(SYMMETRY_PERP)
    CALL HMEASURE%ADD_RTP(HELICAL_STARTS)

    ! Run the mode
    CALL HMEASURE%RUN()
END SUBROUTINE MODE_HMEASURE_INIT_AND_RUN

SUBROUTINE MODE_HMEASURE_SLAVE
    USE RUNTIME_PARAMETERS
    use Images
    USE HELICES
    USE IMAGES_HELICES, ONLY : IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW
    USE IMAGES_ARITHM, ONLY : CYL_MASK
    USE IMAGEFILES
    USE INTERPOLATION
    USE IMAGES_COMPARISON, ONLY : CORRELATION
    IMPLICIT NONE
    ! Variables
    TYPE(IMAGE)                 ::  IMGIN, IMGOUT, IMGMASK
    TYPE(HELICAL_SYMMETRY)      ::  HELIX
    TYPE(INTERPOLATION_SCHEME)  ::  INTERPOL_SCHEME
    !TYPE(IMAGEFILE)             ::  INPUT_FILE, OUTPUT_FILE, ASYM_UNIT_FILE
    !TYPE(IMAGE)                 ::  ASYM_UNIT_VOL
    REAL                        ::  SCORE
    REAL,   PARAMETER           ::  RISE_HALF_RANGE = 0.05
    REAL,   PARAMETER           ::  TWIST_HALF_RANGE = 0.05
    REAL,   PARAMETER           ::  RISE_STEP_SIZE = 0.005
    REAL,   PARAMETER           ::  TWIST_STEP_SIZE = 0.01
    INTEGER                     ::  RISE_STEP, TWIST_STEP
    REAL                        ::  CURRENT_RISE, CURRENT_TWIST
    REAL                        ::  BEST_SCORE, BEST_RISE, BEST_TWIST


    ! Start work

    ! Convert the maximum radius to pixels
    IF (MAX_RAD%VALUE .NE. 0.0) THEN
        MAX_RAD%VALUE = MAX_RAD%VALUE / PIXEL_SIZE%VALUE
    ENDIF



    ! Setup interpolation scheme
    CALL INTERPOLATION_SCHEME_INIT(INTERPOL_SCHEME,'LINEAR','RECT',2)

    ! Read the image from disk
    call imgin%ReadFromDisk(input_filename%value)

    ! Create a mask for the correlation measurement
    CALL IMGMASK%Allocate(MOULD=IMGIN)
    IMGMASK%real_values = 1.0
    CALL CYL_MASK(IMGMASK,MAX_RAD%VALUE,3)

    !
    BEST_SCORE = -HUGE(1.0)
    DO TWIST_STEP=-2,2! Loop over twists
        CURRENT_TWIST = TWIST%VALUE + TWIST_STEP * TWIST_STEP_SIZE
        DO RISE_STEP=-2,2! Loop over rises
            CURRENT_RISE = RISE%VALUE + RISE_STEP * RISE_STEP_SIZE

            ! Setup the helical symmetry operator
            CALL HELICAL_SYMMETRY_INIT(HELIX,CURRENT_RISE,CURRENT_TWIST,SYMMETRY_AXIAL%VALUE,SYMMETRY_PERP%VALUE,&
                                        PIXEL_SIZE=PIXEL_SIZE%VALUE,&
                                        NUM_STARTS=HELICAL_STARTS%VALUE)

            ! Symmetrise
            IMGOUT = IMGIN
            CALL IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW(IMGOUT,HELIX,INTERPOL_SCHEME,MAX_RADIUS=MAX_RAD%VALUE,QUIET=.TRUE.)

            ! Measure the correlation between the volumes before and after symmetrisation
            SCORE =  CORRELATION(IMGIN,IMGOUT,IMGMASK,NORMALISE=.TRUE.)

            WRITE(*,'(A,3(F0.5,1X))') '**info(MODE_HMEASURE_SLAVE): twist, rise, score = ', CURRENT_TWIST, &
                                                                                            CURRENT_RISE , SCORE
            IF (SCORE .GT. BEST_SCORE) THEN
                BEST_RISE = CURRENT_RISE
                BEST_TWIST = CURRENT_TWIST
                BEST_SCORE = SCORE
            ENDIF
        ENDDO
    ENDDO

    WRITE(*,'(A,3(F0.5,1X))') '**info(MODE_HMEASURE_SLAVE): best twist, rise, score = ', BEST_TWIST, BEST_RISE , BEST_SCORE

END SUBROUTINE MODE_HMEASURE_SLAVE

END MODULE MODE_HMEASURE
