MODULE MODE_MASK

    !
    ! Modules we'll need
    !
    USE RUNTIME_PARAMETERS
    USE MODES
    USE RTPS

    IMPLICIT NONE
    TYPE(MODE) :: MASK

    TYPE(RTP_CHAR), TARGET  ::  INPUT_FILENAME  =               &
        RTP_CHAR(   VALUE='input.mrc',                  &
                    KEYWORD='INPUT_FILENAME',           &
                    DESCRIPTION='Input image / volume')


    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_FILENAME  =               &
        RTP_CHAR(   VALUE='threed.mrc',                  &
                    KEYWORD='OUTPUT_FILENAME',           &
                    DESCRIPTION='Output image / volume')


    TYPE(RTP_CHAR), TARGET  ::  MASK_TYPE  =               &
        RTP_CHAR(   VALUE='CIRCLE',                  &
                    KEYWORD='MASK_TYPE',           &
                    DESCRIPTION='One of CIRCLE, CYLINDER, Z_EDGES, LOW_PASS, COSINE_DISC or AUTO')


    TYPE(RTP_CHAR), TARGET  ::  FALLOFF_TYPE  =               &
        RTP_CHAR(   VALUE='NONE',                  &
                    KEYWORD='FALLOFF_TYPE',           &
                    DESCRIPTION='One of COS_EDGE, HARD_EXTEND, or NONE. Used by LOW_PASS mask only.')


    TYPE(RTP_REAL), TARGET  ::  MASK_RADIUS  =               &
        RTP_REAL(   VALUE=0.5,                  &
                    KEYWORD='MASK_RADIUS',           &
                    DESCRIPTION='Mask radius (pixels or reciprocal pixels)')

    TYPE(RTP_REAL), TARGET  ::  MASK_INNER_RADIUS  =               &
        RTP_REAL(   VALUE=0.0,                  &
                    KEYWORD='MASK_INNER_RADIUS',           &
                    DESCRIPTION='Mask inner radius (pixels or reciprocal pixels) (used to give hollow cylinder)')


    TYPE(RTP_REAL), TARGET  ::  MASK_FALLOFF  =               &
        RTP_REAL(   VALUE=0.05,                  &
                    KEYWORD='MASK_FALLOFF',           &
                    DESCRIPTION='Mask soft falloff width (pixels or reciprocal pixels)')


    TYPE(RTP_REAL), TARGET  ::  PCT_IN_MASK  =               &
        RTP_REAL(   VALUE=2.0,                  &
                    KEYWORD='PCT_IN_MASK',           &
                    DESCRIPTION='Percentage of voxels in the mask (not including the voxels in soft falloff around the mask)')


    TYPE(RTP_LOGI), TARGET  ::  INVERTED  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='INVERTED',           &
                    DESCRIPTION='If true, flips 1 to 0 and vice versa at the end of masking routine to let through only background')

    TYPE(RTP_LOGI), TARGET  ::  MASK_WITH_AVERAGE =     &
        RTP_LOGI(   VALUE=.TRUE.,                  &
                    KEYWORD='MASK_WITH_AVERAGE',           &
                    DESCRIPTION='Use average value within the mask to fill the areas outside of the mask aperture. '//&
                                'Not implemented for all types of masks')
CONTAINS

SUBROUTINE MODE_MASK_INIT_AND_RUN
    IMPLICIT NONE
    ! Intialise the mode
    CALL MASK%INIT(COMMAND='MASK',SUBROUTINE=MODE_MASK_SLAVE,SHORT_DESCRIPTION='Add description here')

    ! Add RTPs to the mode
    CALL MASK%ADD_RTP(INPUT_FILENAME)
    CALL MASK%ADD_RTP(OUTPUT_FILENAME)
    CALL MASK%ADD_RTP(MASK_TYPE)
    CALL MASK%ADD_RTP(FALLOFF_TYPE)
    CALL MASK%ADD_RTP(MASK_RADIUS)
    CALL MASK%ADD_RTP(MASK_INNER_RADIUS)
    CALL MASK%ADD_RTP(MASK_FALLOFF)
    CALL MASK%ADD_RTP(PCT_IN_MASK)
    CALL MASK%ADD_RTP(INVERTED)
    CALL MASK%ADD_RTP(MASK_WITH_AVERAGE)

    ! Run the mode
    CALL MASK%RUN()
END SUBROUTINE MODE_MASK_INIT_AND_RUN

SUBROUTINE MODE_MASK_SLAVE
    USE IMAGES
    USE IMAGES_ARITHM
    USE MASKS
    IMPLICIT NONE
    !
    ! Private variables
    !
    TYPE(IMAGE) ::  INPUT_IMAGE, MASK_IMAGE
    REAL    ::  RAD,FALLOFF, RAD_IN

    !
    ! Get variable values from control file
    !


    !
    ! Start work
    !

    ! Read input image
    call input_image%ReadFromDisk(input_filename%value,read_volume_not_slice=.true.)

    ! Depending on the MASK_TYPE%VALUE chosen by the user, apply the appropriate mask
    SELECT CASE (MASK_TYPE%VALUE)
        CASE ('CIRCLE')
            WRITE(*,'(2A)') '**info(MODE_MASK): circular mask'
            IF (MASK_RADIUS%VALUE .LT. 1.0) THEN
                RAD = MASK_RADIUS%VALUE*(INPUT_IMAGE%logical_dimensions(1)/2.)
                FALLOFF = MASK_FALLOFF%VALUE*(INPUT_IMAGE%logical_dimensions(1)/2.)
            ELSE
                RAD = MASK_RADIUS%VALUE
                FALLOFF = MASK_FALLOFF%VALUE
            ENDIF
            CALL CIRC_MASK(INPUT_IMAGE,RAD)
            CALL WRITE_TO_DISK(INPUT_IMAGE,OUTPUT_FILENAME%VALUE)
        CASE ('CYLINDER')
            WRITE(*,'(2A)') '**info(MODE_MASK): cylindrical mask'
            IF (MASK_RADIUS%VALUE .LT. 1.0) THEN
                RAD = MASK_RADIUS%VALUE*(INPUT_IMAGE%logical_dimensions(1)/2.)
                RAD_IN = MASK_INNER_RADIUS%VALUE*(INPUT_IMAGE%logical_dimensions(1)/2.)
                FALLOFF = MASK_FALLOFF%VALUE*(INPUT_IMAGE%logical_dimensions(1)/2.)
            ELSE
                RAD = MASK_RADIUS%VALUE
                RAD_IN = MASK_INNER_RADIUS%VALUE
                FALLOFF = MASK_FALLOFF%VALUE
            ENDIF
            CALL CYL_MASK(INPUT_IMAGE,RAD,NINT(FALLOFF),RAD_IN)
            CALL WRITE_TO_DISK(INPUT_IMAGE,OUTPUT_FILENAME%VALUE)
        CASE ('Z_EDGES')
            ! Apply the mask
            CALL IMAGE_MASK_Z_ENDS_INPLACE(INPUT_IMAGE,NINT(MASK_RADIUS%VALUE),SOFT=.TRUE.)
            ! Write output
            CALL WRITE_TO_DISK(INPUT_IMAGE,OUTPUT_FILENAME%VALUE)
        CASE ('LOW_PASS')
            WRITE(*,'(A)') '**info(MODE_MASK): low pass mask'
            CALL IMAGE_LOW_PASS_MASK(INPUT_IMAGE,MASK_IMAGE, PCT_IN_MASK%VALUE,FALLOFF_TYPE%VALUE,  &
                                    LOW_PASS_FREQUENCY=MASK_RADIUS%VALUE, FALLOFF_WIDTH=MASK_FALLOFF%VALUE, &
                                    INVERSION=INVERTED%VALUE)
            ! Apply the mask
            INPUT_IMAGE%real_values = INPUT_IMAGE%real_values * MASK_IMAGE%real_values
            ! Write the masked image to disk
            CALL WRITE_TO_DISK(INPUT_IMAGE,OUTPUT_FILENAME%VALUE)
        CASE ('COSINE_DISC')
            WRITE(*,'(A)') '**info(MODE_MASK): cosine-edged disc mask'
            CALL INPUT_IMAGE%ApplyCosineDisc( RAD=MASK_RADIUS%VALUE,FALLOFF=MASK_FALLOFF%VALUE,WIPE_INPUT=.FALSE.,   &
                                MASK_WITH_AVERAGE=MASK_WITH_AVERAGE%VALUE)
            CALL WRITE_TO_DISK(INPUT_IMAGE,OUTPUT_FILENAME%VALUE)
        CASE DEFAULT
            WRITE(*,'(2A)') '**ERROR(MODE_MASK): unrecognised mask type: ', TRIM(MASK_TYPE%VALUE)
            CALL TERMINATE('Fatal error in MODE_MASK')
    END SELECT

END SUBROUTINE MODE_MASK_SLAVE

END MODULE MODE_MASK
