!>  \brief  Mode to align frames from a movie
MODULE MODE_MOVIE_ALIGN
    USE MODES
    USE RTPS
    IMPLICIT NONE
    ! Declare mode object
    TYPE(MODE)  ::  MOVIE_ALIGN
    ! Declare RTPs
    TYPE(RTP_CHAR), TARGET  ::  INPUT_FILENAME  =       &
        RTP_CHAR(   VALUE='input.mrc',                  &
                    KEYWORD='INPUT_FILENAME',           &
                    DESCRIPTION='Name of input file, to be binned. Can be image or frealix alignment file')

    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_FILENAME =       &
        RTP_CHAR(   VALUE='output.mrc',                 &
                    KEYWORD='OUTPUT_FILENAME',          &
                    DESCRIPTION='Name of output file')

    TYPE(RTP_INTG), TARGET  ::  FIRST_FRAME  =       &
        RTP_INTG(   VALUE=1,                            &
                    KEYWORD='FIRST_FRAME',           &
                    DESCRIPTION='')

    TYPE(RTP_INTG), TARGET  ::  LAST_FRAME  =       &
        RTP_INTG(   VALUE=10000,                            &
                    KEYWORD='LAST_FRAME',           &
                    DESCRIPTION='')

    TYPE(RTP_REAL), TARGET  ::  DOSE_RATE  =       &
        RTP_REAL(   VALUE=5.0,                            &
                    KEYWORD='DOSE_RATE',           &
                    DESCRIPTION='Dose rate in e/A^2')

    TYPE(RTP_REAL), TARGET  ::  PIXEL_SIZE  =       &
        RTP_REAL(   VALUE=5.0,                            &
                    KEYWORD='PIXEL_SIZE',           &
                    DESCRIPTION='Pixel size in Angstroms')

    TYPE(RTP_REAL), TARGET  ::  ACCELERATION_VOLTAGE  =       &
        RTP_REAL(   VALUE=300.0,                            &
                    KEYWORD='ACCELERATION_VOLTAGE',           &
                    DESCRIPTION='Acceleration voltage of the electrons (in kV)')

    TYPE(RTP_REAL), TARGET  ::  SEARCH_RADIUS_MIN  =       &
        RTP_REAL(   VALUE=1.0,                            &
                    KEYWORD='SEARCH_RADIUS_MIN',           &
                    DESCRIPTION='Minimum shift during initial correlation alignment')

    TYPE(RTP_REAL), TARGET  ::  BFACTOR  =       &
        RTP_REAL(   VALUE=0.0,                            &
                    KEYWORD='BFACTOR',           &
                    DESCRIPTION='B factor applied to frames for alignment (in A^2)')

CONTAINS

SUBROUTINE MODE_MOVIE_ALIGN_INIT_AND_RUN
    IMPLICIT NONE
    ! Initialise the mode
    CALL MOVIE_ALIGN%INIT(  COMMAND='BIN',              &
                    SUBROUTINE=MODE_MOVIE_ALIGN_SLAVE,  &
                    SHORT_DESCRIPTION='Bin or unbin images, volumes, frealix alignment files or boxer coordinates files')
    ! Add RTPs to the mode
    CALL MOVIE_ALIGN%ADD_RTP(INPUT_FILENAME)
    CALL MOVIE_ALIGN%ADD_RTP(OUTPUT_FILENAME)
    CALL MOVIE_ALIGN%ADD_RTP(FIRST_FRAME)
    CALL MOVIE_ALIGN%ADD_RTP(LAST_FRAME)
    CALL MOVIE_ALIGN%ADD_RTP(DOSE_RATE)
    CALL MOVIE_ALIGN%ADD_RTP(PIXEL_SIZE)
    CALL MOVIE_ALIGN%ADD_RTP(SEARCH_RADIUS_MIN)
    CALL MOVIE_ALIGN%ADD_RTP(BFACTOR)
    ! Run the mode
    CALL MOVIE_ALIGN%RUN()
END SUBROUTINE MODE_MOVIE_ALIGN_INIT_AND_RUN

SUBROUTINE MODE_MOVIE_ALIGN_SLAVE
    USE IMAGES
    USE PARSERS, ONLY : IMAGE_FILENAME, EXTENSION_FROM_FILENAME
    USE IMAGES_COMPARISON, ONLY : FIND_SHIFT
    use ImageFiles
    IMPLICIT NONE
    ! Variables
    TYPE(IMAGEFILE)             ::  FRAMESFILE
    TYPE(IMAGE), ALLOCATABLE    ::  FRAMES(:)
    TYPE(IMAGE)                 ::  REFERENCE
    INTEGER                     ::  NUM_FRAMES_TOTAL, NUM_FRAMES
    INTEGER                     ::  CURRENT_FRAME
    INTEGER                     ::  COUNTER
    REAL,       ALLOCATABLE     ::  LAST_SHIFT_X(:), LAST_SHIFT_Y(:)
    REAL,       ALLOCATABLE     ::  OVERALL_SHIFT_X(:), OVERALL_SHIFT_Y(:)
    REAL                        ::  CURRENT_SHIFT(3)
    INTEGER,    PARAMETER       ::  NUMBER_OF_ALIGNMENT_ITERATIONS = 2
    INTEGER                     ::  CURRENT_ALIGNMENT_ITERATION

    ! Start work
    CALL FRAMESFILE%INIT(INPUT_FILENAME%VALUE)
    CALL FRAMESFILE%PrintInfo()
    NUM_FRAMES_TOTAL = FRAMESFILE%GetStackSize()

    ! Check first and last frame are within range
    IF ((FIRST_FRAME%VALUE .GT. LAST_FRAME%VALUE) .OR. (LAST_FRAME%VALUE .GT. NUM_FRAMES_TOTAL)) THEN
        WRITE(*,'(A,2(I0,1X))') '**ERROR(MODE_MOVIE_ALIGN_SLAVE): bad first, last frames: ', FIRST_FRAME%VALUE,LAST_FRAME%VALUE
        CALL TERMINATE('Fatal error in MODE_MOVIE_ALIGN_SLAVE')
    ENDIF

    ! How many frames will we be processing?
    NUM_FRAMES = LAST_FRAME%VALUE-FIRST_FRAME%VALUE+1


    ! Allocate memory for frames
    ALLOCATE(FRAMES(NUM_FRAMES))

    ! Allocate memory for shifts
    ALLOCATE(LAST_SHIFT_X(NUM_FRAMES),LAST_SHIFT_Y(NUM_FRAMES))
    ALLOCATE(OVERALL_SHIFT_X(NUM_FRAMES),OVERALL_SHIFT_Y(NUM_FRAMES))
    OVERALL_SHIFT_X = 0.0
    OVERALL_SHIFT_Y = 0.0

    ! Loop over frames, load them in memory
    WRITE(*,'(A)') '**info(MODE_MOVIE_ALIGN): loading frames into memory'
    DO CURRENT_FRAME=FIRST_FRAME%VALUE,LAST_FRAME%VALUE
        COUNTER = CURRENT_FRAME-FIRST_FRAME%VALUE+1
        call frames(counter)%ReadFromDisk(framesfile)
    ENDDO


    WRITE(*,'(A)') '**info(MODE_MOVIE_ALIGN): computing Fourier transforms'
    !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(CURRENT_FRAME,COUNTER)
    !$OMP DO
    DO CURRENT_FRAME=1,NUM_FRAMES
        CALL FWD_FT(FRAMES(CURRENT_FRAME))
    ENDDO
    !$OMP ENDDO
    !$OMP END PARALLEL

    ! Close file
    CALL FRAMESFILE%CLOSE()
    WRITE(*,'(A)') '**info(MODE_MOVIE_ALIGN): done loading frames into memory'

    ! Initially, middle frame is reference
    REFERENCE = FRAMES(NUM_FRAMES/2)

    DO CURRENT_ALIGNMENT_ITERATION=1,NUMBER_OF_ALIGNMENT_ITERATIONS

        WRITE(*,'(2(A,I0))') '**info(MODE_MOVIE_ALIGN): alignment iteration ', CURRENT_ALIGNMENT_ITERATION, &
                                                                       ' of ', NUMBER_OF_ALIGNMENT_ITERATIONS

        ! Apply B factor to the reference
        CALL APPLY_BFACTOR(REFERENCE,BFACTOR%VALUE)

        !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(CURRENT_FRAME,CURRENT_SHIFT)
        ! Loop over frames
        !$OMP DO
        DO CURRENT_FRAME=1,NUM_FRAMES
            ! Compute
            IF (CURRENT_ALIGNMENT_ITERATION .EQ. 1) THEN
                CURRENT_SHIFT = FIND_SHIFT(FRAMES(CURRENT_FRAME),REFERENCE,MIN_RADIUS=SEARCH_RADIUS_MIN%VALUE)
            ELSE
                CURRENT_SHIFT = FIND_SHIFT(FRAMES(CURRENT_FRAME),REFERENCE)
            ENDIF
            ! Remember shift
            LAST_SHIFT_X(CURRENT_FRAME) = CURRENT_SHIFT(1)
            LAST_SHIFT_Y(CURRENT_FRAME) = CURRENT_SHIFT(2)
            OVERALL_SHIFT_X(CURRENT_FRAME) = OVERALL_SHIFT_X(CURRENT_FRAME) + LAST_SHIFT_X(CURRENT_FRAME)
            OVERALL_SHIFT_Y(CURRENT_FRAME) = OVERALL_SHIFT_Y(CURRENT_FRAME) + LAST_SHIFT_Y(CURRENT_FRAME)
            ! Apply shift
            CALL PSHIFT_IMAGE(FRAMES(CURRENT_FRAME),-OVERALL_SHIFT_X(CURRENT_FRAME),-OVERALL_SHIFT_Y(CURRENT_FRAME),0.0)
        ENDDO
        !$OMP END DO
        !$OMP END PARALLEL

        ! Update the reference
        REFERENCE%real_values = 0.0
        DO CURRENT_FRAME=1,NUM_FRAMES
            WRITE(*,'(2(A,I0),2(A,F0.3))')  '**info(MODE_MOVIE_ALIGN): frame ', CURRENT_FRAME, &
                                            ' of ', NUM_FRAMES, '. last shifts: ',&
                                            LAST_SHIFT_X(CURRENT_FRAME), ',', LAST_SHIFT_Y(CURRENT_FRAME)
            REFERENCE%complex_values = REFERENCE%complex_values + FRAMES(CURRENT_FRAME)%complex_values
        ENDDO

    ENDDO ! End of loop over alignment iterations

    CALL BWD_FT(REFERENCE)
    CALL WRITE_TO_DISK(REFERENCE,OUTPUT_FILENAME%VALUE)



END SUBROUTINE MODE_MOVIE_ALIGN_SLAVE

END MODULE MODE_MOVIE_ALIGN
