MODULE MODE_STACK_ALIGN_TO_SUM

    ! Modules
    USE RUNTIME_PARAMETERS
    USE MODES
    USE RTPS
    IMPLICIT NONE
    TYPE(MODE) :: STACK_ALIGN_TO_SUM

    TYPE(RTP_CHAR), TARGET  ::  INPUT_FILENAME  =       &
        RTP_CHAR(   VALUE='input.mrc',                  &
                    KEYWORD='INPUT_STACK',              &
                    DESCRIPTION='')


    TYPE(RTP_CHAR), TARGET  ::  OUTPUT_FILENAME  =      &
        RTP_CHAR(   VALUE='output.mrc',                 &
                    KEYWORD='OUTPUT_STACK',             &
                    DESCRIPTION='')


    TYPE(RTP_INTG), TARGET  ::  NUM_ITERATIONS  =       &
        RTP_INTG(   VALUE=3,                            &
                    KEYWORD='NUM_ITERATIONS',           &
                    DESCRIPTION='Number of alignment & averaging iterations')


    TYPE(RTP_LOGI), TARGET  ::  SUBTRACT_FROM_SUM  =    &
        RTP_LOGI(   VALUE=.TRUE.,                       &
                    KEYWORD='SUBTRACT_FROM_SUM',        &
                    DESCRIPTION='Whether to substract the input image from the average image before doing the alignment')

    TYPE(RTP_LOGI), TARGET  ::  AMYLOID_BETA  =               &
        RTP_LOGI(   VALUE=.FALSE.,                  &
                    KEYWORD='AMYLOID_BETA',           &
                    DESCRIPTION='Input stack consists of projections of amyloid-beta fibrils')

    TYPE(RTP_REAL), TARGET  ::  PIXEL_SIZE  =               &
        RTP_REAL(   VALUE=1.17,                  &
                    KEYWORD='PIXEL_SIZE',           &
                    DESCRIPTION='Pixel size (Angstroms)')

CONTAINS

SUBROUTINE MODE_STACK_ALIGN_TO_SUM_INIT_AND_RUN
    IMPLICIT NONE
    ! Intialise the mode
    CALL STACK_ALIGN_TO_SUM%INIT(COMMAND='STACK_ALIGN_TO_SUM',SUBROUTINE=MODE_STACK_ALIGN_TO_SUM_SLAVE, &
                             SHORT_DESCRIPTION='Iteratively compute the average of a stack and align each image of the stack to it')

    ! Add RTPs to the mode
    CALL STACK_ALIGN_TO_SUM%ADD_RTP(INPUT_FILENAME)
    CALL STACK_ALIGN_TO_SUM%ADD_RTP(OUTPUT_FILENAME)
    CALL STACK_ALIGN_TO_SUM%ADD_RTP(NUM_ITERATIONS)
    CALL STACK_ALIGN_TO_SUM%ADD_RTP(SUBTRACT_FROM_SUM)
    CALL STACK_ALIGN_TO_SUM%ADD_RTP(AMYLOID_BETA)
    CALL STACK_ALIGN_TO_SUM%ADD_RTP(PIXEL_SIZE)

    ! Run the mode
    CALL STACK_ALIGN_TO_SUM%RUN()


END SUBROUTINE MODE_STACK_ALIGN_TO_SUM_INIT_AND_RUN

SUBROUTINE MODE_STACK_ALIGN_TO_SUM_SLAVE
    use Images
    USE IMAGEFILES
    USE EULERANGLES, ONLY : EULER
    USE ROTATION_MATRICES, ONLY : EULER_TO_ROTMAT, IDENTITY_ROTMAT
    USE IMAGES_COMPARISON, ONLY : CORRELATION, FIND_SHIFT_EULER
    USE INTERPOLATION, ONLY : INTERPOLATION_SCHEME, INTERPOLATION_SCHEME_INIT
    USE IMAGES_ARITHM, ONLY : IMAGES_AVERAGE
    USE MASKS, ONLY : FOURIER_MASK_INIT_AMYLOID
    IMPLICIT NONE
    ! Variables
    TYPE(IMAGE)                     ::  STACK_AVE, MASK, REF
    TYPE(IMAGE),    ALLOCATABLE     ::  STACK(:), STACK_ORIGINAL(:)
    TYPE(IMAGE)                     ::  FMASK, FMASK_COPY
    INTEGER                         ::  I, IERR, ITERATION
    REAL                            ::  SHIFT(3), MATRIX(4,4)
    REAL,           ALLOCATABLE     ::  STACK_MATRICES(:,:,:)
    TYPE(EULER)                     ::  ROT
    INTEGER                         ::  CURRENT_SHIFT, BEST_SHIFT
    REAL                            ::  CC, BEST_CC
    INTEGER                         ::  PAR_LUN
    TYPE(INTERPOLATION_SCHEME)      ::  INTERPOL_SCHEME
    TYPE(WINDOW)                    ::  INTERPOL_WINDOW
    TYPE(IMAGEFILE)                 ::  INPUT_STACK_FILE, OUTPUT_STACK_FILE
    INTEGER                         ::  LOCS_IN_STACK
    INTEGER                         ::  CURRENT_LOC
    LOGICAL,    PARAMETER           ::  DEBUG                   = .TRUE.
    LOGICAL,    PARAMETER           ::  REAL_SPACE_ALIGNMENT    = .FALSE.


    ! Start work

    PAR_LUN =   AVAILABLE_LUN()

    ! Setup interpolation scheme
    CALL INTERPOLATION_SCHEME_INIT(INTERPOL_SCHEME,'LINEAR','RECT',2)
    CALL INTERPOL_WINDOW%AllocateWindow(INTERPOL_SCHEME%WINDOW_WIDTH,COMP=.FALSE.,THREED=.FALSE.)


    ! Open the input file & find out the number of locations in the stack
    CALL INPUT_STACK_FILE%INIT(INPUT_FILENAME%VALUE)
    CALL INPUT_STACK_FILE%PrintInfo()
    LOCS_IN_STACK = INPUT_STACK_FILE%GetStackSize()

    ! Allocate memory so that we can hold the stack in core twice
    ALLOCATE(STACK(LOCS_IN_STACK),STACK_ORIGINAL(LOCS_IN_STACK),STAT=IERR)
    IF (IERR .NE. 0) THEN
        WRITE(*,'(A,I0)') '**ERROR(MODE_STACK_ALIGN_TO_SUM): memory allocation failed for stack, with error ', IERR
        CALL TERMINATE('Fatal error in MODE_STACK_ALIGN_TO_SUM')
    ENDIF

    ! Read the stack image data from disk
    DO CURRENT_LOC=1,LOCS_IN_STACK
        call stack(i)%ReadFromDisk(INPUT_STACK_FILE,wanted_location=current_loc)
        STACK_ORIGINAL(I) = STACK(I)
    ENDDO

    ! Open the output file
    CALL OUTPUT_STACK_FILE%INIT(OUTPUT_FILENAME%VALUE,MOULD=INPUT_STACK_FILE)

    ! Close the input file
    CALL INPUT_STACK_FILE%CLOSE()


    ! Allocate memory for & initialise the rotation matrices. We will store the transformation matrix
    ! for each image from the original stack to the latest alignment
    ALLOCATE(STACK_MATRICES(4,4,SIZE(STACK)),STAT=IERR)
    IF (IERR .NE. 0) THEN
        WRITE(*,'(A,I0)') '**ERROR(MODE_STACK_ALIGN_TO_SUM): memory allocation failed for rotation matrices with error ', IERR
        CALL TERMINATE('Fatal error in MODE_STACK_ALIGN_TO_SUM')
    ENDIF
    FORALL(I=1:SIZE(STACK))
        STACK_MATRICES(:,:,I) = IDENTITY_ROTMAT()
    END FORALL


    ! Prepare a Fourier mask
    IF (AMYLOID_BETA%VALUE) THEN
        CALL FOURIER_MASK_INIT_AMYLOID(FMASK,STACK(1),PIXEL_SIZE%VALUE,FIL_MAX_RAD=100.0,MAX_SPACING_ALIGN=8.0)
        FMASK_COPY = FMASK
        CALL AMPL_FROM_HALF_TO_FULL(FMASK_COPY)
        CALL WRITE_TO_DISK(FMASK_COPY,'dbg_fourier_mask.mrc')
        CALL IMAGE_DEALLOCATE(FMASK_COPY)
        IF (REAL_SPACE_ALIGNMENT) THEN
            WRITE(*,'(A)') '**ERROR(MODE_STACK_ALIGN_TO_SUM): Fourier masks not implemented yet in real-space CCC calculation'
            CALL TERMINATE('Fatal error in MODE_STACK_ALIGN_TO_SUM. Feature not implemented yet')
        ENDIF
    ENDIF


    !
    ! Start iterative alignment. STACK keeps latest aligned state of the stack, while STACK_ORIGINAL keeps original input stack
    !
    DO ITERATION=1,NUM_ITERATIONS%VALUE

        WRITE(*,'(/A,I0,A,I0)') '**info(MODE_STACK_ALIGN_TO_SUM): on iteration ', ITERATION, ' of ', NUM_ITERATIONS%VALUE

        ! Calculate new average, from current alignment
        CALL IMAGES_AVERAGE(STACK,STACK_AVE)
        STACK_AVE%real_values = 0.0

        ! Align each image
        DO I=1,SIZE(STACK)
            ! Calculate the reference for this image
            REF = STACK_AVE
            IF (SUBTRACT_FROM_SUM%VALUE) THEN
                REF%real_values = STACK_AVE%real_values - STACK(I)%real_values
            ENDIF

            ! Do the alignment
            IF (REAL_SPACE_ALIGNMENT) THEN
                ! Find alignment the Chuck way, by doing shifts and calculating real-space normalised ccc
                BEST_CC = -1.0
                DO CURRENT_SHIFT=0,3
                    STACK(I)%real_values = STACK_ORIGINAL(I)%real_values
                    CALL FWD_FT(STACK(I))
                    CALL PSHIFT_IMAGE(STACK(I),0.0,REAL(CURRENT_SHIFT),0.0)
                    CALL BWD_FT(STACK(I))
                    CC = CORRELATION(STACK(I),REF,MASK,NORMALISE=.TRUE.)
                    IF (CC .GT. BEST_CC) THEN
                        BEST_CC = CC
                        BEST_SHIFT = CURRENT_SHIFT
                    ENDIF
                    !PRINT '(A,I2,A,I2,X,F12.8)', '**debug(): loc ', I, '; current shift, cc = ', CURRENT_SHIFT, CC
                ENDDO
                SHIFT = [0.0,REAL(BEST_SHIFT),0.0]
                !PRINT '(A,I2,A,I2,X,F12.8)', '**debug(): loc ', I, '; best shift by real space, cc = ', BEST_SHIFT, BEST_CC
            ELSE
                CALL FIND_SHIFT_EULER(STACK(I),REF,SHIFT,ROT,SHIFT_ONLY=.TRUE.,MAX_RAD_SHIFT=4.0,VERTICAL_ONLY=.TRUE.)
            ENDIF

            ! Print out result of the alignemnt
            WRITE(*,'(2(A,I0),A,3(F0.2,1X))') '**info(MODE_STACK_ALIGN_TO_SUM): image ', I, ' of ', SIZE(STACK), &
                                                '; shift & rot = ', SHIFT(1:2), ROT%PSI

            ! Shift & rotate the original image
            MATRIX = EULER_TO_ROTMAT(ROT%PHI,ROT%THETA,ROT%PSI,SHIFT=SHIFT,REVERSE=.TRUE.)
            STACK_MATRICES(:,:,I) = MATMUL(MATRIX,STACK_MATRICES(:,:,I))    ! Not 100% whether this is the correct order for the matrix multiplication
            CALL ROTATE_IMAGE(STACK_ORIGINAL(I),STACK(I),STACK_MATRICES(:,:,I),INTERPOL_SCHEME,INTERPOL_WINDOW)

        ENDDO

    ENDDO

    !
    ! Calculate new sum, from final alignment
    !
    CALL IMAGES_AVERAGE(STACK,STACK_AVE)

    ! Write out the final stack
    DO I=1,SIZE(STACK)
        call stack(i)%WriteToDisk(output_stack_file,wanted_location=i)
    ENDDO
    CALL OUTPUT_STACK_FILE%PrintInfo()
    CALL OUTPUT_STACK_FILE%CLOSE()

END SUBROUTINE MODE_STACK_ALIGN_TO_SUM_SLAVE


END MODULE MODE_STACK_ALIGN_TO_SUM
