!>  \brief  Classes and methods to handle peaks of functions
MODULE PEAKS

    USE COOS    !   The coos (coordinates) module defines the POINT derived type
    IMPLICIT NONE

    PRIVATE

    PUBLIC  ::  PEAK, PEAK_FROM_IMAGE_INTG, PEAK_FROM_IMAGE


    !>  \brief  Defines a peak list
    TYPE    ::  PEAK_LIST
        TYPE(PEAK), ALLOCATABLE ::  PEAK(:)
        INTEGER                 ::  N               =   1       !<  Number of peaks in peak list
        REAL                    ::  MIN_PEAK_DIST   =   0.0     !<  Minimum distance (pixels) between peaks
    END TYPE

    CONTAINS

    !>  \brief  Return the coordinates of the highest peak within the image/volume.
    !!
    !!  First, the voxel with the highest value is found. Then the center of mass of the
    !!  neighbourhood is computed. This is where the output peak is.
    !!  I lifted this center-of-mass idea from EMAN2.06's calc_max_location_wrap_intp routine.
    TYPE(PEAK) FUNCTION PEAK_FROM_IMAGE(IMG,MAX_RAD,X_ONLY,Y_ONLY,Z_ONLY,Y_MAX,Y_MIN) RESULT(HIGHEST_PEAK)
        USE IMAGES
        USE IMAGES_ARITHM, ONLY : IMAGE_CENTER_OF_MASS
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(IN)  ::  IMG     !<  Image in which to look for the peak
        REAL,           OPTIONAL,   INTENT(IN)  ::  MAX_RAD !<  Maximum radius (in pixels) within which to look for the peak
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  X_ONLY  !<  Search for peak only on equator  (2nd dimension = origin)
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  Y_ONLY  !<  Search for peak only on meridian (1st dimension = origin)
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  Z_ONLY  !<  Search for peak only along the Z axis
        INTEGER,        OPTIONAL,   INTENT(IN)  ::  Y_MAX   !<  Maximum address in the second dimension to look for a peak
		INTEGER,		OPTIONAL,	INTENT(IN)	::	Y_MIN	!<	Minimum address in the second dimension to look for a peak
        ! Private variables
        TYPE(PEAK)  ::  INTG_PEAK
        INTEGER,    PARAMETER     ::  COM_RAD = 2 !<    Radius for the center-of-mass computation
        INTEGER ::  LBOUNDS(3),UBOUNDS(3)
        REAL    ::  COM_COO(3)
        ! Start work

        ! Get integer peak
        INTG_PEAK = PEAK_FROM_IMAGE_INTG(IMG,MAX_RAD,X_ONLY,Y_ONLY,Z_ONLY,Y_MAX,Y_MIN)
        HIGHEST_PEAK = INTG_PEAK

        ! Let's define the boundaries for our center-of-mass computation
        LBOUNDS = NINT(INTG_PEAK%COO) - 2
        UBOUNDS = NINT(INTG_PEAK%COO) + 2
        IF (IMG%logical_dimensions(3) .EQ. 1) THEN
            LBOUNDS(3) = 1
            UBOUNDS(3) = 1
        ENDIF

        ! Compute the center of mass
        CALL IMAGE_CENTER_OF_MASS(IMG,COM_COO,LBOUNDS=LBOUNDS,UBOUNDS=UBOUNDS)

        ! Modify the output peak's coordinates
        HIGHEST_PEAK%COO = COM_COO

    END FUNCTION PEAK_FROM_IMAGE





    !>  \brief  Return the integer coordinates of the voxel with the heighest value
    TYPE(PEAK) FUNCTION PEAK_FROM_IMAGE_INTG(IMG,MAX_RAD,X_ONLY,Y_ONLY,Z_ONLY,Y_MAX,Y_MIN,MIN_RAD) RESULT(HIGHEST_PEAK)
        USE IMAGES
        USE IMAGEFILES
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(IMAGE),                INTENT(IN)  ::  IMG     !<  Image in which to look for the peak
        REAL,           OPTIONAL,   INTENT(IN)  ::  MIN_RAD !<  Minimum radius (in pixels) within which to look for the peak
        REAL,           OPTIONAL,   INTENT(IN)  ::  MAX_RAD !<  Maximum radius (in pixels) within which to look for the peak
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  X_ONLY  !<  Search for peak only on equator  (2nd dimension = origin)
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  Y_ONLY  !<  Search for peak only on meridian (1st dimension = origin)
        LOGICAL,        OPTIONAL,   INTENT(IN)  ::  Z_ONLY  !<  Search for peak only along the Z axis
        INTEGER,        OPTIONAL,   INTENT(IN)  ::  Y_MAX   !<  Maximum address in the second dimension to look for a peak
		INTEGER,		OPTIONAL,	INTENT(IN)	::	Y_MIN	!<	Minimum address in the second dimension to look for a peak
        ! Private variables
        INTEGER         ::  I, J, K, RAD1, RAD2, RAD3
        REAL            ::  MMIN_RAD,MMAX_RAD
        INTEGER         ::  ORIGIN(3), SEARCH_LBOUND(3), SEARCH_UBOUND(3)
        LOGICAL         ::  XX_ONLY, YY_ONLY, ZZ_ONLY
        REAL(KIND=8)    ::  S, SS
        REAL(KIND=8)    ::  STD
        INTEGER         ::  YY_MIN, YY_MAX
        LOGICAL, PARAMETER  ::  DEBUG   =   .FALSE.
        ! Start work

        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(PEAK_FROM_IMAGE_INTG): entering. Image info:'
            CALL IMAGE_PRINT_INFO(IMG)
        ENDIF

        !
		YY_MIN = 1
        YY_MAX = IMG%logical_dimensions(2)
        IF (PRESENT(Y_MAX)) THEN
            IF (Y_MAX .GT. 0) YY_MAX = Y_MAX
        ENDIF
		IF (PRESENT(Y_MIN)) THEN
			IF (Y_MIN .GT. 1) YY_MIN = Y_MIN
		ENDIF

        ! Restricted to one of the axes?
        XX_ONLY = .FALSE.
        YY_ONLY = .FALSE.
        ZZ_ONLY = .FALSE.
        IF (PRESENT(X_ONLY)) XX_ONLY = X_ONLY
        IF (PRESENT(Y_ONLY)) YY_ONLY = Y_ONLY
        IF (PRESENT(Z_ONLY)) ZZ_ONLY = Z_ONLY

        ! X, Y and Z_ONLY options mutually exclusive
        IF ((XX_ONLY .AND. YY_ONLY) .OR. (XX_ONLY .AND. ZZ_ONLY) .OR. (YY_ONLY .AND. ZZ_ONLY)) THEN
            WRITE(*,'(A,3L1)') '**ERROR(PEAK_FROM_IMAGE_INTG): x, y, z only flags are mutually exclusive: ',    &
                                    XX_ONLY, YY_ONLY, ZZ_ONLY
            CALL TERMINATE('Fatal error in PEAK_FROM_IMAGE_INTG')
        ENDIF

        ! Maximum radius?
        IF (PRESENT(MAX_RAD)) THEN
            MMAX_RAD = MAX_RAD**2
        ELSE
            ! If horiz or vert is active, we set the maximum radius if it wasn't set already
            IF (XX_ONLY .OR. YY_ONLY .OR. ZZ_ONLY) THEN
                MMAX_RAD = (MAXVAL(IMG%logical_dimensions))**2
            ELSE
                MMAX_RAD = 0.0
            ENDIF
        ENDIF

        ! Minimum radius?
        IF (PRESENT(MIN_RAD)) THEN
            MMIN_RAD = MIN_RAD**2
        ELSE
            MMIN_RAD = 0.0
        ENDIF

        IF (DEBUG) WRITE(*,'(A,F0.4,1X,L1)') '**debug(PEAK_FROM_IMAGE_INTG): max rad, zero? = ', MMAX_RAD, MMAX_RAD .EQ. 0.0

        ! Check image is real
        IF (.NOT. IMG%IsReal()) THEN
            WRITE(*,'(2A)') '**ERROR(PEAK_FROM_IMAGE_INTG): input image is not real: ', IMG%is_in_real_space
            CALL TERMINATE('Fatal error in PEAK_FROM_IMAGE_INTG')
        ENDIF

        ! Find the peak
        HIGHEST_PEAK%rvalue = -HUGE(1.0)
        IF (DEBUG) WRITE(*,'(A,G20.10)') '**debug(PEAK_FROM_IMAGE_INTG): peak init with value: ', HIGHEST_PEAK%rvalue
        S = 0.0D0
        SS = 0.0D0

        !PRINT '(A,F6.3)', '**debug(PEAK_FROM_IMAGE_INTG): looking for peak with maximum radius restriction of ', MAX_RAD
        if (img%is_in_real_space .and. img%origin_at_center) then
            origin = img%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif
        !PRINT '(A,3I4)', '**debug(PEAK_FROM_IMAGE_INTG): origin = ', ORIGIN
        IF (YY_ONLY) THEN
            SEARCH_LBOUND = (/ORIGIN(1),1,         ORIGIN(3) /)
            SEARCH_UBOUND = (/ORIGIN(1),IMG%logical_dimensions(2),ORIGIN(3) /)
        ELSE IF (XX_ONLY) THEN
            SEARCH_LBOUND = (/1,         ORIGIN(2),ORIGIN(3) /)
            SEARCH_UBOUND = (/IMG%logical_dimensions(1),ORIGIN(2),ORIGIN(3) /)
        ELSE IF (ZZ_ONLY) THEN
            SEARCH_LBOUND = (/ORIGIN(1), ORIGIN(2),1         /)
            SEARCH_UBOUND = (/ORIGIN(1), ORIGIN(2),IMG%logical_dimensions(3)/)
        ELSE
            SEARCH_LBOUND = [1,1,1]
            SEARCH_UBOUND = IMG%logical_dimensions
        ENDIF

        ! User-specified bounds
        SEARCH_UBOUND(2) = MIN(SEARCH_UBOUND(2),YY_MAX)
		SEARCH_LBOUND(2) = MAX(SEARCH_LBOUND(2),YY_MIN)
        !print *, '**debug(PEAK_FROM_IMAGE_INTG): search bounds 2nd dim ', SEARCH_LBOUND(2), SEARCH_UBOUND(2)


        ! Loop
        DO K=SEARCH_LBOUND(3),SEARCH_UBOUND(3)
            RAD3 = (K-ORIGIN(3))**2
            DO J=SEARCH_LBOUND(2),SEARCH_UBOUND(2)
                RAD2 = RAD3 + (J-ORIGIN(2))**2
                DO I=SEARCH_LBOUND(1),SEARCH_UBOUND(1)
                    RAD1 = RAD2 + (I-ORIGIN(1))**2
                    IF (        (RAD1 .LE. MMAX_RAD .OR. MMAX_RAD .EQ. 0.0) &
                        .AND.   (RAD1 .GE. MMIN_RAD .OR. MMIN_RAD .EQ. 0.0)) THEN
                        S = S + IMG%real_values(I,J,K)
                        SS = SS + IMG%real_values(I,J,K)**2
                        IF (IMG%real_values(I,J,K) .GT. HIGHEST_PEAK%rvalue) THEN
                            HIGHEST_PEAK%rvalue = IMG%real_values(I,J,K)
                            HIGHEST_PEAK%COO = REAL([I,J,K])
                        ENDIF
                    ENDIF
                ENDDO
            ENDDO
        ENDDO
        IF (DEBUG) PRINT '(A,3(i0,1X))', '**debug(PEAK_FROM_IMAGE_INTG): peak was found with i,j,k = ', INT(HIGHEST_PEAK%COO)


        ! The standard deviation of the image
        STD = S**2/REAL(PRODUCT(IMG%logical_dimensions))
        STD = (SS - STD)/REAL(PRODUCT(IMG%logical_dimensions))
        STD = SQRT(STD)
        IF (STD .LT. TINY(1.0)) THEN
            WRITE(*,'(A,G20.10)') '**ERROR(PEAK_FROM_IMAGE_INTG): image has strange std dev = ', STD
            IF (DEBUG) CALL WRITE_TO_DISK(IMG,'dbg_peak_from_image_intg.mrc')
            CALL TERMINATE('Fatal error in PEAK_FROM_IMAGE_INTG')
        ELSE
            HIGHEST_PEAK%NSIGMA = (HIGHEST_PEAK%rvalue - S/PRODUCT(IMG%logical_dimensions)) / STD
        ENDIF
    END FUNCTION PEAK_FROM_IMAGE_INTG




END MODULE PEAKS
