!>  \brief  Classes and methods to deal with single particle point group symmetries
!!
MODULE POINT_SYMMETRIES
    USE ROTATION_MATRICES

    IMPLICIT NONE
    PRIVATE
    PUBLIC  ::  POINT_SYMMETRY, POINT_SYMMETRY_SYMBOL, POINT_SYMMETRY_INIT, POINT_SYMMETRY_PRINT

    !>  \brief  Describes point group symmetry using Sch�nflies notation
    !!
    TYPE    ::  POINT_SYMMETRY_SYMBOL
        CHARACTER(LEN=3)    ::  SYMBOL = 'C1 '      !<  Point group symmetry using Sch�nflies notation
    END TYPE

    !>  \brief  Describes point group symmetry as a set of rotation matrices
    !!
    TYPE    ::  POINT_SYMMETRY
        INTEGER                     ::  IREDUN = 0                  !<  Total symmetry redundancy
        TYPE(ROTATION_MATRIX),      ALLOCATABLE ::  SYMOP(:)                    !<  Symmetry operators (rotation matrices)
        TYPE(POINT_SYMMETRY_SYMBOL) ::  SYMBOL                      !<  Sch�nflies notation
        LOGICAL     ::  HELICAL = .FALSE.
    END TYPE

    CONTAINS


        !>  \brief  Print out point group symmetry matrices
        SUBROUTINE POINT_SYMMETRY_PRINT(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(POINT_SYMMETRY),   INTENT(IN)  ::  SELF
            ! Private variables
            INTEGER ::  I,NSYM
            ! Start work
            NSYM = SIZE(SELF%SYMOP)
            WRITE(*,'(/3A)') '**info(POINT_SYMMETRY_PRINT): symmetry operator matrices for ',   &
                                SELF%SYMBOL%SYMBOL, ' point group follow'
            DO I=1,NSYM
                WRITE(*,'(3(A,I2),A)') 'Symmetry operator', I, ' of ', NSYM, ' with ', SELF%SYMOP(I)%JMAX, '-fold redundancy:'
                CALL ROTATION_MATRIX_PRINT(SELF%SYMOP(I))
            ENDDO
            WRITE(*,'(3A,I3/)') '**info(POINT_SYMMETRY_PRINT): Total symmetry redundancy for point group ', &
                                    SELF%SYMBOL%SYMBOL, ' is ', SELF%IREDUN
        END SUBROUTINE POINT_SYMMETRY_PRINT

        !>  \brief  Initialise a point_symmetry object, given Sch�nflies symmetry symbol
        !!
        SUBROUTINE POINT_SYMMETRY_INIT(SELF,SYM_SYMBOL)
            USE RUNTIME_PARAMETERS, ONLY : PI,TERMINATE
            USE ROTATION_MATRICES, ONLY : IDENTITY_ROTMAT
            IMPLICIT NONE
            ! Arguments
            TYPE(POINT_SYMMETRY),       INTENT(OUT) ::  SELF                !<  Point group symmetry matrices to initialise
            TYPE(POINT_SYMMETRY_SYMBOL),    INTENT(IN)  ::  SYM_SYMBOL          !<  Point group symmetry using Sch�nflies notation
            ! Private variables
            INTEGER                 ::  J,K,L
            INTEGER                 ::  NGET,NSYM,NASYM
            INTEGER                 ::  N_CDTOI
            INTEGER                 ::  IREDUN
            INTEGER,PARAMETER,DIMENSION(4,6)    ::  NSYMTYPE =  RESHAPE(    SOURCE=    &
                                                                                            (/  8,0,0,0,  &
                                                                                                8,5,0,0,  &
                                                                                                4,5,6,0,  &
                                                                                                3,7,4,0,  &
                                                                                                1,4,5,6,  &
                                                                                                2,4,5,6   /), &
                                                                            SHAPE =         (/  4,6/) &
                                                                        )
            INTEGER,PARAMETER,DIMENSION(6)      ::  NSYM_CDTOI = [1,2,3,3,4,4]
            REAL                    ::  ANGLE
            REAL,DIMENSION(3,3,8)           ::  SYMSTORE =      RESHAPE(    SOURCE= &
                                        (/  0.309017,0.809017,-0.5,     -0.809017,0.5,0.309017,         0.5,0.309017,0.809017,  &
                                            0.809017,0.309017,0.5,      0.309017,0.5,-0.809017,         -0.5,0.809017,0.309017, &
                                            0.,  1.,  0.,               -1.,  0.,  0.,                  0.,  0.,  1.,           &
                                            0.,  0.,  1.,               1.,  0.,  0.,                   0.,  1.,  0.,           &
                                            1.,  0.,  0.,               0., -1.,  0.,                   0.,  0., -1.,           &
                                            -1.,  0.,  0.,              0.,  1.,  0.,                   0.,  0., -1.,           &
                                            0.,  1.,  0.,               1.,  0.,  0.,                   0.,  0., -1.,           &
                                            0.,  0.,  0.,               0.,  0.,  0.,                   0.,  0.,  1.    /),     &
                                        SHAPE= (/3,3,8/) &
                                                                        )
            CHARACTER(LEN=1)            ::  SYMBOL1
            CHARACTER(LEN=1),PARAMETER,DIMENSION(16)::  ASYMTEST = [' ','C','D','T','O','I','0','1','2','3','4','5','6','7','8','9']
            ! Start work
            SELF%SYMBOL = SYM_SYMBOL
            CALL PARSE_SYM_SYMBOL(SYM_SYMBOL,SYMBOL1,NASYM)
            IF (NASYM.GE.0 .AND. SYMBOL1 .NE. 'H') THEN
                WRITE(*,'(2A)') '**info(POINT_SYMMETRY_INIT): initializing symmetry matrices for point group ', SYM_SYMBOL%SYMBOL
                N_CDTOI = 0
                IREDUN = 1
                DO J = 2,5
                    IF(SYMBOL1.EQ.ASYMTEST(J)) N_CDTOI = J-1
                ENDDO
                IF (SYMBOL1.EQ.ASYMTEST(6)) THEN
                    ! We have icosahedral symmetry defined as in the Int Tables...
                    N_CDTOI = 5
                    ! .. or as defined by Tony Crowther
                    IF (NASYM.EQ.2) N_CDTOI = 6
                ENDIF

                ! Calculate N-fold rotation matrix (symbols C and D)
                IF ((N_CDTOI.EQ.1).OR.(N_CDTOI.EQ.2)) THEN
                    ANGLE = 2.0*PI/NASYM
                    SYMSTORE(1,1,8) = COS(ANGLE)
                    SYMSTORE(1,2,8) = SIN(ANGLE)
                    SYMSTORE(2,1,8) =-SIN(ANGLE)
                    SYMSTORE(2,2,8) = COS(ANGLE)
                ENDIF

                NSYM = NSYM_CDTOI(N_CDTOI)
                ALLOCATE(SELF%SYMOP(NSYM))
                DO J = 1,NSYM
                    NGET = NSYMTYPE(J,N_CDTOI)
                    DO K = 1,4
                        DO L = 1,4
                            IF ((K.LE.3).AND.(L.LE.3)) THEN
                                SELF%SYMOP(J)%RMATRIX(K,L) = SYMSTORE(K,L,NGET)
                            ELSE
                                SELF%SYMOP(J)%RMATRIX(K,L) = 0.0
                                IF (K.EQ.L) SELF%SYMOP(J)%RMATRIX(K,L) = 1.0
                            ENDIF
                        ENDDO
                    ENDDO
                ENDDO
                CALL CHECK_POINT_SYMMETRY(SELF)
                IF (SELF%IREDUN.LT.0) THEN
                    DO J = 1,NSYM
                        IF (SELF%SYMOP(J)%JMAX.LT.0) THEN
                            WRITE(*,'(A,I0)') '**ERROR(POINT_SYMMETRY_PRINT): Repeated application of symmetry ',    &
                                            'matrix does not produce identity, ISYM = ', J
                            CALL TERMINATE('Fatal error in POINT_SYMMETRY_PRINT')
                        ENDIF
                    ENDDO
                ENDIF
            ELSE IF (SYMBOL1 .EQ. 'H') THEN
                IREDUN = 1
                ALLOCATE(SELF%SYMOP(1))
                SELF%SYMOP(1)%RMATRIX = IDENTITY_ROTMAT()
                CALL CHECK_POINT_SYMMETRY(SELF)
            ELSE IF (SYMBOL1 .NE. 'H') THEN
                WRITE(*,'(A,A)') '**ERROR(POINT_SYMMETRY_INIT): Invalid symmetry, SYM_SYMBOL = ', SYM_SYMBOL%SYMBOL
                CALL TERMINATE('Fatal error in POINT_SYMMETRY_INIT')
            ENDIF

            ! Is this a helical symmetry operator?
            SELF%HELICAL = (SYM_SYMBOL%SYMBOL .EQ. 'H')
        END SUBROUTINE POINT_SYMMETRY_INIT


        SUBROUTINE PARSE_SYM_SYMBOL(SELF,SYMBOL1,NASYM)
            USE RUNTIME_PARAMETERS, ONLY : TERMINATE
            IMPLICIT NONE
            ! Arguments
            TYPE(POINT_SYMMETRY_SYMBOL),    INTENT(IN)  ::  SELF                !<  Point group symmetry using Sch�nflies notation
            CHARACTER(LEN=1),       INTENT(OUT) ::  SYMBOL1             !<  First character of the symmetry symbol
            INTEGER,            INTENT(OUT) ::  NASYM               !<  Integer number contained in the symmetry symbol
            ! Private variables
            INTEGER                 ::  J
            INTEGER                 ::  KSYMN1,KSYMN2,KSYMN3
            CHARACTER(LEN=1)            ::  SYMBOL2
            CHARACTER(LEN=1)            ::  SYMBOL3
            CHARACTER(LEN=1),PARAMETER,DIMENSION(17)::  ASYMTEST =  &
                                                        [' ','C','D','T','O','I','0','1','2','3','4','5','6','7','8','9','H']
            LOGICAL                 ::  ASYMSTORE
            ! Start work
            KSYMN1 = -1
            KSYMN2 = -1
            KSYMN3 = -1
            SYMBOL1 = SELF%SYMBOL(1:1)
            SYMBOL2 = SELF%SYMBOL(2:2)
            SYMBOL3 = SELF%SYMBOL(3:3)

            ASYMSTORE = .FALSE.

            DO J = 2,17
                IF ((SYMBOL3.EQ.ASYMTEST(J)).AND.(J.LE.6)) THEN
                    ! No letters allowed
                    WRITE(*,'(A)') '**ERROR(PARSE_SYM_SYMBOL): 3rd character in symmetry symbol invalid'
                    CALL TERMINATE('Fatal error in PARSE_SYM_SYMBOL')
                ENDIF
                IF ((SYMBOL3.EQ.ASYMTEST(J)).AND.(J.GT.6)) KSYMN3 = J-7
                IF ((SYMBOL2.EQ.ASYMTEST(J)).AND.(J.LE.6)) THEN
                    ! No letters allowed
                    WRITE(*,'(A)') '**ERROR(PARSE_SYM_SYMBOL): 2nd character in symmetry symbol invalid'
                    CALL TERMINATE('Fatal error in PARSE_SYM_SYMBOL')
                            ENDIF
                IF ((SYMBOL2.EQ.ASYMTEST(J)).AND.(J.GT.6)) KSYMN2 = J-7
                IF (SYMBOL1.EQ.ASYMTEST(J)) THEN
                    KSYMN1 = J-1
                    IF(J.LE.6) ASYMSTORE = .TRUE.
                    IF(J.GT.6) KSYMN1 = J-7
                ENDIF
            ENDDO

            IF (KSYMN1.EQ.-1) THEN
                WRITE(*,'(A)') '**ERROR(PARSE_SYM_SYMBOL): Symmetry symbol invalid'
                CALL TERMINATE('Fatal error in PARSE_SYM_SYMBOL')
            ENDIF

            IF (ASYMSTORE ) THEN
                IF(KSYMN3.NE.-1) THEN
                    NASYM=KSYMN3+10*KSYMN2
                ELSE
                    IF(KSYMN2.NE.-1) THEN
                        NASYM=KSYMN2
                    ELSE
                        NASYM=0
                    ENDIF
                ENDIF
                IF ((KSYMN1.EQ.1).OR.(KSYMN1.EQ.2)) THEN
                    IF (NASYM.EQ.0) THEN
                        WRITE(*,'(A)') '**ERROR(PARSE_SYM_SYMBOL): Point groups C and D must be followed by an integer'
                        CALL TERMINATE('Fatal error in PARSE_SYM_SYMBOL')
                    ENDIF
                ENDIF
            ELSE IF (SYMBOL1 .NE. 'H') THEN
                WRITE(*,'(A)') '**ERROR(PARSE_SYM_SYMBOL): Input of symmetry matrices not yet implemented'
                CALL TERMINATE('Fatal error in PARSE_SYM_SYMBOL')
            ENDIF
        END SUBROUTINE PARSE_SYM_SYMBOL


        SUBROUTINE CHECK_POINT_SYMMETRY(SELF)
            IMPLICIT NONE
            ! Arguments
            TYPE(POINT_SYMMETRY),       INTENT(INOUT)   ::  SELF        !<  Point group symmetry matrices
            ! Private variables
            INTEGER                     ::  I,J,K,N,NSYM
            INTEGER                     ::  SYMCOUNT,IR
            REAL                        ::  DIFF
            REAL,PARAMETER                  ::  DELTA = 0.01
            REAL                        ::  TM(4,4)
            ! Start work
            NSYM = SIZE(SELF%SYMOP)
            IR = 1
            DO N = 1,NSYM
                SELF%SYMOP(N)%JMAX = 0
            ENDDO
            DO N = 1,NSYM
                DO I = 1,4
                    DO J = 1,4
                        TM(I,J) = 0.0
                    ENDDO
                ENDDO
                TM(1,1) = 1.0
                TM(2,2) = 1.0
                TM(3,3) = 1.0
                TM(4,4) = 1.0

                SYMCOUNT = 0
                DO K = 1,100
                    SYMCOUNT = SYMCOUNT+1
                    TM = MATMUL(TM,SELF%SYMOP(N)%RMATRIX)
                    DIFF = 0.0
                    DO I = 1,3
                        DO J = 1,3
                            IF (I.EQ.J) THEN
                                DIFF = DIFF+ABS(TM(I,J)-1.0)
                            ELSE
                                DIFF = DIFF+ABS(TM(I,J))
                            ENDIF
                        ENDDO
                    ENDDO
                    IF (DIFF.LT.DELTA) EXIT
                ENDDO
                IF (SYMCOUNT.EQ.100) THEN
                    IR = -1
                    SELF%SYMOP(N)%JMAX = -1
                    EXIT
                ELSE
                    IR = IR*SYMCOUNT
                    SELF%SYMOP(N)%JMAX = SYMCOUNT
                ENDIF
            ENDDO
            SELF%IREDUN = IR
        END SUBROUTINE CHECK_POINT_SYMMETRY


END MODULE POINT_SYMMETRIES
