!>  \brief  Module to interface with the proc filesystem, if available
MODULE PROCFS
    CHARACTER(LEN=*),   PARAMETER   ::  PROCFS_PATH = '/proc'

CONTAINS

    !>  \brief  Printout memory usage statistics
    SUBROUTINE PRINT_MEMORY_USAGE_STATS()
        use DatesAndTimes
        IMPLICIT NONE
        ! Arguments

        ! Private variables
        INTEGER ::  PEAK_VM
        type(DateAndTime) :: myTime
        ! Start work
        WRITE(*,'(2A)')         '**info(PRINT_MEMORY_USAGE_STATS): date & time = ', myTime%DateAndTimeAsString()
        WRITE(*,'(A,I12,A)')    '**info(PRINT_MEMORY_USAGE_STATS): current virtual memory size = ', GET_VM_SIZE(),         ' kB'
        PEAK_VM = GET_PEAK_VM_SIZE()
        IF (PEAK_VM .NE. 0) THEN
        WRITE(*,'(A,I12,A)')    '**info(PRINT_MEMORY_USAGE_STATS):    peak virtual memory size = ', GET_PEAK_VM_SIZE(),    ' kB'
        ENDIF
    END SUBROUTINE PRINT_MEMORY_USAGE_STATS

    !>  \brief  Get current virtual memory size, in kB
    INTEGER FUNCTION GET_VM_SIZE(PID )
        IMPLICIT NONE
        ! Arguments
        INTEGER,    OPTIONAL,   INTENT(IN)  ::  PID     !<  Process ID
        ! Private variables

        ! Start work
        GET_VM_SIZE = PROCFS_STATUS_GET_VALUE('VmSize',PID)
    END FUNCTION GET_VM_SIZE

    !>  \brief  Get current virtual memory size, in kB
    INTEGER FUNCTION GET_PEAK_VM_SIZE(PID)
        IMPLICIT NONE
        ! Arguments
        INTEGER,    OPTIONAL,   INTENT(IN)  ::  PID     !<  Process ID
        ! Private variables

        ! Start work
        GET_PEAK_VM_SIZE = PROCFS_STATUS_GET_VALUE('VmPeak',PID)
    END FUNCTION GET_PEAK_VM_SIZE

    !>  \brief  Get current virtual memory size, in kB
    INTEGER FUNCTION GET_SWAP_SIZE(PID)
        IMPLICIT NONE
        ! Arguments
        INTEGER,    OPTIONAL,   INTENT(IN)  ::  PID     !<  Process ID
        ! Private variables

        ! Start work
        GET_SWAP_SIZE = PROCFS_STATUS_GET_VALUE('VmSwap',PID)
    END FUNCTION GET_SWAP_SIZE

    !>  \brief  Guess system swap usage by parsing procfs's meminfo file
    INTEGER FUNCTION PROCFS_MEMINFO_SYSTEM_SWAP()
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        ! Private variables
        ! Start work
        PROCFS_MEMINFO_SYSTEM_SWAP = 0
        CALL TERMINATE('Not implemented yet')
        ! Relevant keywords in /proc/meminfo
        ! are SwapTotal and SwapFree

    END FUNCTION PROCFS_MEMINFO_SYSTEM_SWAP

    !>  \brief  Get value for given parameter in procfs status file
    INTEGER FUNCTION PROCFS_STATUS_GET_VALUE(KEY,PID) RESULT(VALUE)
        USE PARSERS, ONLY : GREP, SPLIT, IS_BLANK
        USE RUNTIME_PARAMETERS, ONLY : GET_PROCESS_ID, LINE_LEN, WORD_LEN
        IMPLICIT NONE
        ! Arguments
        CHARACTER(LEN=*),               INTENT(IN)  ::  KEY     !<  Keyword we're looking for in the status file
        INTEGER,            OPTIONAL,   INTENT(IN)  ::  PID     !<  Process ID
        ! Private variables
        CHARACTER(LEN=LINE_LEN)                 ::  LINE
        CHARACTER(LEN=WORD_LEN)                 ::  VALUE_C
        CHARACTER(LEN=LINE_LEN),   ALLOCATABLE  ::  WORDS(:)
        INTEGER ::  PPID
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work
        IF (PRESENT(PID)) THEN
            PPID = PID
        ELSE
            PPID = GET_PROCESS_ID()
        ENDIF
        VALUE = 0
        IF (PROCFS_AVAILABLE()) THEN
            LINE = GREP(PROCFS_STATUS_FILE_PATH(PPID),KEY)
            IF (IS_BLANK(LINE)) THEN
                VALUE = 0
            ELSE
                WORDS = SPLIT(LINE)
                VALUE_C = TRIM(ADJUSTL(WORDS(2)))
                IF (DEBUG) WRITE(*,'(2A)') '**debug(PROCFS_STATUS_GET_VALUE): found value of ', TRIM(ADJUSTL(VALUE_C))
                READ(VALUE_C,'(I12)') VALUE
            ENDIF
        ELSE
            VALUE = 0
        ENDIF
    END FUNCTION PROCFS_STATUS_GET_VALUE

    !>  \brief  Get the path of the status file for a process
    FUNCTION PROCFS_STATUS_FILE_PATH(PID) RESULT(PATH)
        USE RUNTIME_PARAMETERS, ONLY : GET_PROCESS_ID, LINE_LEN
        IMPLICIT NONE
        ! Arguments
        INTEGER,    OPTIONAL,   INTENT(IN)  ::  PID     !<  Process ID
        ! Result
        CHARACTER(LEN=LINE_LEN)             ::  PATH
        ! Private variables
        CHARACTER(LEN=6)    ::  PIDC
        INTEGER             ::  PPID
        ! Start work
        IF (PRESENT(PID)) THEN
            PPID = PID
        ELSE
            PPID = GET_PROCESS_ID()
        ENDIF
        ! Convert PID to character string
        WRITE(PIDC,'(I0)') PPID

        PATH = PROCFS_PATH // '/' // TRIM(ADJUSTL(PIDC)) // '/status'
    END FUNCTION PROCFS_STATUS_FILE_PATH

    !>  \brief  Check whether the proc filesystem is available
    LOGICAL FUNCTION PROCFS_AVAILABLE()
        USE PARSERS, ONLY : FILE_EXISTS
        USE RUNTIME_PARAMETERS, ONLY : LINE_LEN
        IMPLICIT NONE
        ! Arguments
        ! Private variables
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        CHARACTER(LEN=LINE_LEN) ::  FILE_PATH
        ! Start work
        FILE_PATH = PROCFS_PATH // '/1/status'

        PROCFS_AVAILABLE = FILE_EXISTS(FILE_PATH)
        IF (DEBUG) WRITE(*,'(3A,L1)') '**debug(PROCFS_AVAILABLE): file, exists?: ', FILE_PATH, ', ', PROCFS_AVAILABLE
    END FUNCTION PROCFS_AVAILABLE
END MODULE PROCFS
