!>  \brief  Classes and methods for caching 2D projections of 3D images
MODULE PROJECTION_CACHES

    USE IMAGES
    USE EULERANGLES
    USE CTFS

    !>  \brief  A cache of projection images
    TYPE PROJECTION_CACHE
        TYPE(EULER)                             ::  ANGLES_TOL          !<  Euler angles will be considered identical when
                                                                        !!  within this tolerance
        TYPE(CTF_PARAMETERS)                    ::  CTF_TOL             !<  CTF parameters will be considered indentical
                                                                        !!  within this tolerance
        REAL                                    ::  VOL_Z_SHIFT_TOL     !<  Shifts will be considered identical when
                                                                        !!  within this tolerance
        TYPE(EULER),            ALLOCATABLE     ::  ANGLES(:)           !<  For each projection in cache,
                                                                        !!  the Euler angles at which it was calculated
        TYPE(CTF_PARAMETERS),   ALLOCATABLE     ::  CTF(:)              !<  For each projection in cache, the CTF parameters
        REAL,                   ALLOCATABLE     ::  VOL_Z_SHIFT(:)      !<  For each projection in cache, the shifts applied to
                                                                        !!  the volume before projection
        INTEGER,                ALLOCATABLE     ::  TIMES_USED(:)       !<  The number of times a given projection was requested
        TYPE(IMAGE),            ALLOCATABLE     ::  PROJ(:)             !<  The image data for each projection
        TYPE(IMAGE),            ALLOCATABLE     ::  ENV_MASK(:)         !<  Smooth envelope mask
        CONTAINS
            FINAL   ::  PROJECTION_CACHE_CLEAN
    END TYPE PROJECTION_CACHE

    CONTAINS

    !>  \brief  Add projection to the cache
    SUBROUTINE PROJECTION_CACHE_ADD_PROJ(SELF,ANGLES,VOL_Z_SHIFT,CTF,PROJ,ENV_MASK)
        IMPLICIT NONE
        ! Arguments
        TYPE(PROJECTION_CACHE),     INTENT(INOUT)   ::  SELF        !<  Cache object
        TYPE(EULER),                INTENT(IN)      ::  ANGLES      !<  Euler angles at which the projection was computed
        REAL,                       INTENT(IN)      ::  VOL_Z_SHIFT !<  Z shift
        TYPE(CTF_PARAMETERS),       INTENT(IN)      ::  CTF         !<  CTF parameters for the projection
        TYPE(IMAGE),                INTENT(IN)      ::  PROJ        !<  Projection image
        TYPE(IMAGE),                INTENT(IN)      ::  ENV_MASK    !<  Envelope mask
        ! Private variables

        ! Start work

        IF (PROJECTION_CACHE_FIND_PROJ(SELF,ANGLES,VOL_Z_SHIFT,CTF) .EQ. 0) THEN
            !$OMP   CRITICAL (PROJ_CACHE_OMP_CRIT)
            IF (ALLOCATED(SELF%ANGLES)) THEN
                ! Append to the list. The arrays are dyanmically reallocated (assuming compiler option
                ! -assume realloc_lhs)
                SELF%TIMES_USED     = (/ SELF%TIMES_USED, 1 /)
                SELF%ANGLES         = (/ SELF%ANGLES, EULER_BTWN_0_AND_360(ANGLES)  /)
                SELF%VOL_Z_SHIFT    = (/ SELF%VOL_Z_SHIFT, VOL_Z_SHIFT /)
                SELF%CTF            = (/ SELF%CTF,  CTF /)
                CALL IMAGE_APPEND_TO_ARRAY(SELF%PROJ,PROJ)
                CALL IMAGE_APPEND_TO_ARRAY(SELF%ENV_MASK,ENV_MASK)
            ELSE
                ! It looks like this is the first time we're adding this set of angles
                ALLOCATE(SELF%ANGLES(1),SELF%VOL_Z_SHIFT(1),SELF%CTF(1),SELF%TIMES_USED(1),SELF%PROJ(1),SELF%ENV_MASK(1))
                SELF%ANGLES(1)      = EULER_BTWN_0_AND_360(ANGLES)
                SELF%VOL_Z_SHIFT(1) = VOL_Z_SHIFT
                SELF%CTF            = CTF
                SELF%TIMES_USED(1)  = 1
                SELF%PROJ(1)        = PROJ
                SELF%ENV_MASK(1)    = ENV_MASK
            ENDIF
            !$OMP   END CRITICAL (PROJ_CACHE_OMP_CRIT)
        ELSE
            WRITE(*,'(A)') '**ERROR(PROJECTION_CACHE_ADD_PROJ): a projection is already cached with similar angles'
            CALL TERMINATE('Fatal error in PROJECTION_CACHE_ADD_PROJ')
        ENDIF
    END SUBROUTINE PROJECTION_CACHE_ADD_PROJ

    !>  \brief  Retrieve projection image
    SUBROUTINE PROJECTION_CACHE_RETRIEVE_PROJ(SELF,PROJ_POSITION,PROJ,ENV_MASK)
        IMPLICIT NONE
        ! Arguments
        TYPE(PROJECTION_CACHE), INTENT(INOUT)   ::  SELF                !<  Projection cache object
        INTEGER,                INTENT(IN)      ::  PROJ_POSITION       !<  Position of the projection image to be retrieved
        TYPE(IMAGE),            INTENT(OUT)     ::  PROJ                !<  Retrieved projection image
        TYPE(IMAGE),            INTENT(OUT)     ::  ENV_MASK            !<  Retrieved envelope mask
        ! Private variables

        ! Start work
        IF (PROJ_POSITION .GT. SIZE(SELF%PROJ)) THEN
            WRITE(*,'(A,I0,1X,I0)') '**ERROR(PROJECTION_CACHE_RETRIEVE_PROJ): position greater than size of cache: ', &
                                    PROJ_POSITION, SIZE(SELF%PROJ)
            CALL TERMINATE('Fatal error in PROJECTION_CACHE_RETRIEVE_PROJ')
        ELSE
            PROJ = SELF%PROJ(PROJ_POSITION)
            ENV_MASK = SELF%ENV_MASK(PROJ_POSITION)
            SELF%TIMES_USED(PROJ_POSITION) = SELF%TIMES_USED(PROJ_POSITION) + 1
        ENDIF

    END SUBROUTINE PROJECTION_CACHE_RETRIEVE_PROJ

    !>  \brief  Return the position within the cache of the existing projection at the given Euler angles. Return 0
    !!          if no existing projection matches
    INTEGER FUNCTION PROJECTION_CACHE_FIND_PROJ(SELF,ANGLES,VOL_Z_SHIFT,CTF) RESULT (PROJ_POSITION)
        IMPLICIT NONE
        ! Arguments
        TYPE(PROJECTION_CACHE), INTENT(IN)  ::  SELF                !<  Projection cache object
        TYPE(EULER),            INTENT(IN)  ::  ANGLES
        REAL,                   INTENT(IN)  ::  VOL_Z_SHIFT
        TYPE(CTF_PARAMETERS),   INTENT(IN)  ::  CTF
        ! Private variables
        INTEGER ::  I
        ! Start work
        PROJ_POSITION = 0
        IF (ALLOCATED(SELF%ANGLES)) THEN
            ! Let's loop through existing angles to see if any of them are within tolerance of the "new" angles
            DO I=1,SIZE(SELF%ANGLES)
                IF (ABS(EULER_DIFFERENCE(ANGLES,SELF%ANGLES(I))) .LT. SELF%ANGLES_TOL .AND. &
                    ABS(VOL_Z_SHIFT-SELF%VOL_Z_SHIFT(I)) .LT. SELF%VOL_Z_SHIFT_TOL  .AND. &
                    ABS(CTF-SELF%CTF(I)) .LT. SELF%CTF_TOL) THEN
                    PROJ_POSITION = I
                    EXIT
                ENDIF
            ENDDO
        ENDIF
    END FUNCTION PROJECTION_CACHE_FIND_PROJ

    !>  \brief  Print information about the cache of projections
    SUBROUTINE PROJECTION_CACHE_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(PROJECTION_CACHE), INTENT(INOUT)   ::  SELF                !<  Projection cache object
        ! Private variables

        ! Start work
        ! Private variables
        LOGICAL,    PARAMETER   ::  PRINT_FULL_LIST = .TRUE.
        INTEGER ::  I
        ! Start work

        IF (ALLOCATED(SELF%ANGLES)) THEN
            WRITE(*,'(A,I0)') '**info(PROJECTION_CACHE_PRINT_INFO): number of distinct projections = ', SIZE(SELF%ANGLES)
            WRITE(*,'(A,I0)') '**info(PROJECTION_CACHE_PRINT_INFO): total number of projections = ', SUM(SELF%TIMES_USED)
        ENDIF

        WRITE(*,'(A,5(F0.3,1X))') '**info(PROJECTION_CACHE_PRINT_INFO): tolerance (phi,theta,psi,volZ,defocus) = ',   &
                                                                            SELF%ANGLES_TOL%PHI, &
                                                                            SELF%ANGLES_TOL%THETA, &
                                                                            SELF%ANGLES_TOL%PSI, &
                                                                            SELF%VOL_Z_SHIFT_TOL, &
                                                                            SELF%CTF_TOL%DFMID1

        IF (ALLOCATED(SELF%ANGLES)) THEN
            WRITE(*,'(A,F0.3)') '**info(PROJECTION_CACHE_PRINT_INFO): number of times each projection was used (average) = ', &
                                                            REAL(SUM(SELF%TIMES_USED)) / REAL(SIZE(SELF%TIMES_USED))
            WRITE(*,'(A,I0)') '**info(PROJECTION_CACHE_PRINT_INFO): number of times each projection was used (maximum) = ', &
                                                            MAXVAL(SELF%TIMES_USED)

            ! Now print out the whole list
            IF (PRINT_FULL_LIST) THEN
                WRITE(*,'(A)') '**info(PROJECTION_CACHE_PRINT_INFO): full list of Euler angles follows'
                WRITE(*,'(6(A8,1X))') '#', 'Phi', 'Theta', 'Psi', 'Z shift','Defocus','Count'
                DO I=1,SIZE(SELF%TIMES_USED)
                    WRITE(*,'(I8,1X,4(F8.3,1X),1(F8.1,1X),I8,1X)')  I, SELF%ANGLES(I)%PHI, SELF%ANGLES(I)%THETA, &
                                                                SELF%ANGLES(I)%PSI,&
                                                                SELF%VOL_Z_SHIFT(I), SELF%CTF(I)%DFMID1, &
                                                                SELF%TIMES_USED(I)
                ENDDO
            ENDIF
        ENDIF
    END SUBROUTINE PROJECTION_CACHE_PRINT_INFO

    !>  \brief  Initialise a cache of projections
    SUBROUTINE PROJECTION_CACHE_INIT(SELF,EULER_TOLERANCE,VOL_Z_SHIFT_TOLERANCE,CTF_TOLERANCE)
        IMPLICIT NONE
        ! Arguments
        TYPE(PROJECTION_CACHE), INTENT(INOUT)   ::  SELF                    !<  Projection cache object
        TYPE(EULER),            INTENT(IN)      ::  EULER_TOLERANCE         !<  Euler angles will be considered identical
                                                                            !!  within this tolerance
        REAL,                   INTENT(IN)      ::  VOL_Z_SHIFT_TOLERANCE   !<  Z shifts will be considered identical
                                                                            !!  within this tolerance
        TYPE(CTF_PARAMETERS),   INTENT(IN)      ::  CTF_TOLERANCE           !<  CTF parameters will be considered identical
                                                                            !!  within this tolerance
        ! Private variables

        ! Start work
        SELF%ANGLES_TOL         = EULER_TOLERANCE
        SELF%VOL_Z_SHIFT_TOL    = VOL_Z_SHIFT_TOLERANCE
        SELF%CTF_TOL            = CTF_TOLERANCE
    END SUBROUTINE PROJECTION_CACHE_INIT

    !>  \brief  Clean-up after a cache of projections is no longer needed
    SUBROUTINE PROJECTION_CACHE_CLEAN(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(PROJECTION_CACHE), INTENT(INOUT)   ::  SELF    !<  Projection cache object
        ! Private variables
        INTEGER ::  I
        ! Start work

        ! Deallocate
        !$OMP   CRITICAL (PROJ_CACHE_OMP_CRIT)
        IF (ALLOCATED(SELF%ANGLES)) DEALLOCATE(SELF%ANGLES)
        IF (ALLOCATED(SELF%VOL_Z_SHIFT)) DEALLOCATE(SELF%VOL_Z_SHIFT)
        IF (ALLOCATED(SELF%CTF)) DEALLOCATE(SELF%CTF)
        IF (ALLOCATED(SELF%TIMES_USED)) DEALLOCATE(SELF%TIMES_USED)
        IF (ALLOCATED(SELF%PROJ)) THEN
            DO I=1,SIZE(SELF%PROJ)
                CALL IMAGE_DEALLOCATE(SELF%PROJ(I))
            ENDDO
            DEALLOCATE(SELF%PROJ)
        ENDIF
        IF (ALLOCATED(SELF%ENV_MASK)) THEN
            DO I=1,SIZE(SELF%ENV_MASK)
                CALL IMAGE_DEALLOCATE(SELF%ENV_MASK(I))
            ENDDO
            DEALLOCATE(SELF%ENV_MASK)
        ENDIF
        !$OMP   END CRITICAL (PROJ_CACHE_OMP_CRIT)
    END SUBROUTINE PROJECTION_CACHE_CLEAN
END MODULE PROJECTION_CACHES
