!>  \brief  Class & methods defining run-time parameters and their behaviour
MODULE RTPS


    USE RUNTIME_PARAMETERS
    IMPLICIT NONE
    PRIVATE
    PUBLIC :: RTP_INIT, RTP_GET_VALUE, RTP_PRINT_INFO_HEADER

    INTEGER,            PARAMETER   ::  COLWIDTH(5)         =   [36,45,12,64,15]
    CHARACTER(LEN=*),   PARAMETER   ::  COLWIDTH_C(5)       =   ['36','45','12','64','15']
    INTEGER,            PARAMETER   ::  COM_COLWIDTH(3)     =   [32,FILENAME_LEN,LINE_LEN]
    CHARACTER(LEN=4),   PARAMETER   ::  COM_COLWIDTH_C(3)   =   ['32  ',FILENAME_LEN_C,LINE_LEN_C]

    TYPE,   ABSTRACT,   PUBLIC  ::  RTP
        CHARACTER(LEN=32)       ::  KEYWORD                 !<  Keyword used in frealix command files to identify the parameter
        CHARACTER(LEN=LINE_LEN) ::  DESCRIPTION             !<  Description used in the question and as comment in frealix command files
        LOGICAL                 ::  SET_BY_USER =   .FALSE. !<  Flag to indicate whether the RTP has been set by the user
        INTEGER                 ::  ORDER       =   0       !<  Runtime parameters used within modes are indexed
        CONTAINS
            PROCEDURE   ::  GET_VALUE_FROM_USER =>  RTP_GET_VALUE
    END TYPE

    TYPE, EXTENDS(RTP), PUBLIC  ::  RTP_REAL
        !REAL    ::  DEFAULT_VALUE   !<  Default value for the parameter
        REAL            ::  VALUE           !<  Current value for the parameter
        CONTAINS
            PROCEDURE   ::  INIT                =>  RTP_INIT_REAL
            PROCEDURE   ::  PRINT_INFO          =>  RTP_REAL_PRINT_INFO
            PROCEDURE   ::  PRINT_TO_COMMAND_FILE   =>  RTP_REAL_PRINT_TO_COMMAND_FILE
    END TYPE

    TYPE, EXTENDS(RTP), PUBLIC  ::  RTP_INTG
        !INTEGER ::  DEFAULT_VALUE   !<  Default value for the parameter
        INTEGER ::  VALUE           !<  Current value for the parameter
        CONTAINS
            PROCEDURE   ::  INIT                =>  RTP_INIT_INTG
            PROCEDURE   ::  PRINT_INFO          =>  RTP_INTG_PRINT_INFO
            PROCEDURE   ::  PRINT_TO_COMMAND_FILE   =>  RTP_INTG_PRINT_TO_COMMAND_FILE
    END TYPE

    TYPE, EXTENDS(RTP), PUBLIC  ::  RTP_LOGI
        !LOGICAL ::  DEFAULT_VALUE   !<  Default value for the parameter
        LOGICAL ::  VALUE           !<  Current value for the parameter
        CONTAINS
            PROCEDURE   ::  INIT                =>  RTP_INIT_LOGI
            PROCEDURE   ::  PRINT_INFO          =>  RTP_LOGI_PRINT_INFO
            PROCEDURE   ::  PRINT_TO_COMMAND_FILE   =>  RTP_LOGI_PRINT_TO_COMMAND_FILE
    END TYPE

    TYPE, EXTENDS(RTP), PUBLIC  ::  RTP_CHAR
        !CHARACTER(LEN=LINE_LEN)    ::  DEFAULT_VALUE   !<  Default value for the parameter
        CHARACTER(LEN=LINE_LEN)    ::  VALUE           !<  Current value for the parameter
        CONTAINS
            PROCEDURE   ::  INIT                =>  RTP_INIT_CHAR
            PROCEDURE   ::  PRINT_INFO          =>  RTP_CHAR_PRINT_INFO
            PROCEDURE   ::  PRINT_TO_COMMAND_FILE   =>  RTP_CHAR_PRINT_TO_COMMAND_FILE
    END TYPE

    INTERFACE   RTP_ASSIGN
        MODULE PROCEDURE    RTP_ASSIGN_REAL
        MODULE PROCEDURE    RTP_ASSIGN_CHAR
        MODULE PROCEDURE    RTP_ASSIGN_INTG
        MODULE PROCEDURE    RTP_ASSIGN_LOGI
    END INTERFACE

    INTERFACE RTP_INIT
        MODULE PROCEDURE    RTP_INIT_REAL
        MODULE PROCEDURE    RTP_INIT_INTG
        MODULE PROCEDURE    RTP_INIT_CHAR
        MODULE PROCEDURE    RTP_INIT_LOGI
    END INTERFACE RTP_INIT

    CONTAINS

    !>  \brief  Print RTP information in format suitable for command file
    SUBROUTINE RTP_REAL_PRINT_TO_COMMAND_FILE(SELF,LUN)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_REAL),    INTENT(IN)  ::  SELF
        INTEGER,            INTENT(IN)  ::  LUN
        ! Private variables

        ! Start work
        WRITE(LUN,'(A'//COM_COLWIDTH_C(1)//',F0.6,10X,A,1X,A)')   ADJUSTL(SELF%KEYWORD), SELF%VALUE,    &
                                                                '! ', TRIM(ADJUSTL(SELF%DESCRIPTION))
    END SUBROUTINE RTP_REAL_PRINT_TO_COMMAND_FILE

    !>  \brief  Print RTP information in format suitable for command file
    SUBROUTINE RTP_INTG_PRINT_TO_COMMAND_FILE(SELF,LUN)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_INTG),    INTENT(IN)  ::  SELF
        INTEGER,            INTENT(IN)  ::  LUN
        ! Private variables

        ! Start work
        WRITE(LUN,'(A'//COM_COLWIDTH_C(1)//',I0,10X,A,1X,A)') ADJUSTL(SELF%KEYWORD), SELF%VALUE, '! ', &
                                                            TRIM(ADJUSTL(SELF%DESCRIPTION))
    END SUBROUTINE RTP_INTG_PRINT_TO_COMMAND_FILE

    !>  \brief  Print RTP information in format suitable for command file
    SUBROUTINE RTP_CHAR_PRINT_TO_COMMAND_FILE(SELF,LUN)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_CHAR),    INTENT(IN)  ::  SELF
        INTEGER,            INTENT(IN)  ::  LUN
        ! Private variables

        ! Start work
        WRITE(LUN,'(A'//COM_COLWIDTH_C(1)//',A,10X,A,1X,A)') ADJUSTL(SELF%KEYWORD), TRIM(ADJUSTL(SELF%VALUE)), '! ', &
                                                            TRIM(ADJUSTL(SELF%DESCRIPTION))
    END SUBROUTINE RTP_CHAR_PRINT_TO_COMMAND_FILE

    !>  \brief  Print RTP information in format suitable for command file
    SUBROUTINE RTP_LOGI_PRINT_TO_COMMAND_FILE(SELF,LUN)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_LOGI),    INTENT(IN)  ::  SELF
        INTEGER,            INTENT(IN)  ::  LUN
        ! Private variables

        ! Start work
        WRITE(LUN,'(A'//COM_COLWIDTH_C(1)//',L1,10X,A,1X,A)') ADJUSTL(SELF%KEYWORD), SELF%VALUE, '! ',   &
                                                            TRIM(ADJUSTL(SELF%DESCRIPTION))
    END SUBROUTINE RTP_LOGI_PRINT_TO_COMMAND_FILE

    !>  \brief  Print information about a real RTP
    SUBROUTINE RTP_REAL_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_REAL), INTENT(IN)  ::  SELF
        ! Private variables

        ! Start work
        CALL RTP_PRINT_INFO_SLAVE_1(SELF)
        WRITE(*,'(F'//COLWIDTH_C(2)//'.4,1X)',ADVANCE='NO') SELF%VALUE!, SELF%DEFAULT_VALUE
        CALL RTP_PRINT_INFO_SLAVE_2(SELF)
        CALL RTP_PRINT_INFO_SLAVE_3(SELF)
        WRITE(*,'(A)') ' '
    END SUBROUTINE RTP_REAL_PRINT_INFO

    !>  \brief  Print information about a real RTP
    SUBROUTINE RTP_INTG_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_INTG), INTENT(IN)  ::  SELF
        ! Private variables

        ! Start work
        CALL RTP_PRINT_INFO_SLAVE_1(SELF)
        WRITE(*,'(I'//COLWIDTH_C(2)//',1X)',ADVANCE='NO') SELF%VALUE!, SELF%DEFAULT_VALUE
        CALL RTP_PRINT_INFO_SLAVE_2(SELF)
        CALL RTP_PRINT_INFO_SLAVE_3(SELF)
        WRITE(*,'(A)') ' '
    END SUBROUTINE RTP_INTG_PRINT_INFO

    !>  \brief  Print information about a real RTP
    SUBROUTINE RTP_LOGI_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_LOGI), INTENT(IN)  ::  SELF
        ! Private variables

        ! Start work
        CALL RTP_PRINT_INFO_SLAVE_1(SELF)
        WRITE(*,'(L'//COLWIDTH_C(2)//',1X)',ADVANCE='NO') SELF%VALUE!, SELF%DEFAULT_VALUE
        CALL RTP_PRINT_INFO_SLAVE_2(SELF)
        CALL RTP_PRINT_INFO_SLAVE_3(SELF)
        WRITE(*,'(A)') ' '
    END SUBROUTINE RTP_LOGI_PRINT_INFO

    !>  \brief  Print information about a real RTP
    SUBROUTINE RTP_CHAR_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_CHAR), INTENT(IN)  ::  SELF
        ! Private variables

        ! Start work
        CALL RTP_PRINT_INFO_SLAVE_1(SELF)
        WRITE(*,'(A'//COLWIDTH_C(2)//',1X)',ADVANCE='NO') TRIM(SELF%VALUE)
        CALL RTP_PRINT_INFO_SLAVE_2(SELF)
        CALL RTP_PRINT_INFO_SLAVE_3(SELF)
        WRITE(*,'(A)') ' '
    END SUBROUTINE RTP_CHAR_PRINT_INFO

    !>  \brief  Print basic information (not the value) of an RTP
    SUBROUTINE RTP_PRINT_INFO_SLAVE_1(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),     INTENT(IN)  ::  SELF
        ! Private variables

        ! Start work

        WRITE(*,'(A'//COLWIDTH_C(1)//',1X)',ADVANCE='NO') TRIM(ADJUSTL(SELF%KEYWORD))
    END SUBROUTINE RTP_PRINT_INFO_SLAVE_1

    !>  \brief  Print basic information (not the value) of an RTP
    SUBROUTINE RTP_PRINT_INFO_SLAVE_2(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),     INTENT(IN)  ::  SELF
        ! Private variables

        ! Start work

        WRITE(*,'(L'//COLWIDTH_C(3)//',1X)',ADVANCE='NO') SELF%SET_BY_USER
    END SUBROUTINE RTP_PRINT_INFO_SLAVE_2

    !>  \brief  Print the description of an RTP, over multiple lines if necessary
    SUBROUTINE RTP_PRINT_INFO_SLAVE_3(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),     INTENT(IN)  ::  SELF
        ! Private variables
        INTEGER ::  DESCR_LENGTH, DESCR_LEAD_BLANKS
        INTEGER :: NUM_LINES, CURRENT_LINE, NUM_SPACES
        CHARACTER(LEN=5)    ::  NUM_SPACES_HERE
        INTEGER ::  BEGIN,END
        CHARACTER(LEN=:), ALLOCATABLE :: TMP_STR
        LOGICAL ::  NO_SPACE
        ! Start work

!        WRITE(*,'(A<COLWIDTH(4)>,X)',ADVANCE='NO') adjustl(SELF%DESCRIPTION)

        DESCR_LENGTH = LEN_TRIM(ADJUSTL(SELF%DESCRIPTION))
        DESCR_LEAD_BLANKS = LEN(SELF%DESCRIPTION) - LEN_TRIM(SELF%DESCRIPTION)
        NUM_LINES = CEILING(REAL(DESCR_LENGTH)/REAL(COLWIDTH(4)))
        NUM_SPACES = COLWIDTH(1) + 1 + COLWIDTH(2) + 1 + COLWIDTH(3) + 1

        ALLOCATE(CHARACTER(LEN=COLWIDTH(4)) :: TMP_STR)


        DO CURRENT_LINE=1,NUM_LINES
            IF (CURRENT_LINE .GT. 1) THEN
                ! We need a new line
                WRITE(*,'(A)') ' '
            ENDIF
            !
            BEGIN = (CURRENT_LINE-1)* COLWIDTH(4) + 1
            END   =  CURRENT_LINE   * COLWIDTH(4)
            !
            !BEGIN = BEGIN + DESCR_LEAD_BLANKS
            !END   = END   + DESCR_LEAD_BLANKS
            !
            END   = MIN(END,DESCR_LEAD_BLANKS+DESCR_LENGTH)
            TMP_STR = SELF%DESCRIPTION(BEGIN:END)
            NO_SPACE = .FALSE.
            IF (CURRENT_LINE .GT. 1) THEN
                WRITE(NUM_SPACES_HERE,'(I5)') NUM_SPACES
            ELSE
                NUM_SPACES_HERE = '0'
                NO_SPACE = .TRUE.
            ENDIF
            IF (NO_SPACE) THEN
                WRITE(*,                       '(A'//COLWIDTH_C(4)//',1X)',ADVANCE='NO') ADJUSTL(TMP_STR)
            ELSE
                WRITE(*,'('//NUM_SPACES_HERE//'X,A'//COLWIDTH_C(4)//',1X)',ADVANCE='NO') ADJUSTL(TMP_STR)
            ENDIF
        ENDDO
    END SUBROUTINE RTP_PRINT_INFO_SLAVE_3

    !>  \brief  Print out header for RTP descriptions
    SUBROUTINE RTP_PRINT_INFO_HEADER()
        IMPLICIT NONE
        ! Arguments

        ! Private variables
        CHARACTER(LEN=COLWIDTH(4))  ::  DESCRIPTION
        CHARACTER(LEN=COLWIDTH(5))  ::  HELP
        ! Start work
        DESCRIPTION = 'Description'
        HELP = 'Help'
        WRITE(*,'(A'//COLWIDTH_C(1)//',1X,A'//COLWIDTH_C(2)//',1X,A'//COLWIDTH_C(3)//',1X,A'//COLWIDTH_C(4)//')') &
                                                                                'Key', 'Value', 'Set by user?', ADJUSTL(DESCRIPTION)
    END SUBROUTINE RTP_PRINT_INFO_HEADER

    !>  \brief  Initialise a runtime parameter
    SUBROUTINE RTP_INIT_REAL(SELF,DEFAULT_VALUE,KEYWORD,DESCRIPTION,HELP)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_REAL),                INTENT(INOUT)   ::  SELF
        REAL,                           INTENT(IN)      ::  DEFAULT_VALUE   !<  Default value for the parameter
        CHARACTER(LEN=*),               INTENT(IN)      ::  KEYWORD         !<  Keyword used in frealix command files to identify the parameter
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  DESCRIPTION     !<  Description used in the question and as comment in frealix command files
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  HELP            !<  Help to be given to user to describe what the parameter means / does
        ! Start work
        CALL RTP_INIT_SLAVE(SELF,KEYWORD,DESCRIPTION=DESCRIPTION,HELP=HELP)
        !SELF%DEFAULT_VALUE  =   DEFAULT_VALUE
        SELF%VALUE          =   DEFAULT_VALUE
        SELF%SET_BY_USER    =   .FALSE.
    END SUBROUTINE RTP_INIT_REAL

    !>  \brief  Initialise a runtime parameter
    SUBROUTINE RTP_INIT_INTG(SELF,DEFAULT_VALUE,KEYWORD,DESCRIPTION,HELP)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_INTG),                INTENT(INOUT)   ::  SELF
        INTEGER,                        INTENT(IN)      ::  DEFAULT_VALUE   !<  Default value for the parameter
        CHARACTER(LEN=*),               INTENT(IN)      ::  KEYWORD         !<  Keyword used in frealix command files to identify the parameter
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  DESCRIPTION     !<  Description used in the question and as comment in frealix command files
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  HELP            !<  Help to be given to user to describe what the parameter means / does
        ! Start work
        CALL RTP_INIT_SLAVE(SELF,KEYWORD,DESCRIPTION=DESCRIPTION,HELP=HELP)
        !SELF%DEFAULT_VALUE  =   DEFAULT_VALUE
        SELF%VALUE          =   DEFAULT_VALUE
        SELF%SET_BY_USER    =   .FALSE.
    END SUBROUTINE RTP_INIT_INTG

    !>  \brief  Initialise a runtime parameter
    SUBROUTINE RTP_INIT_LOGI(SELF,DEFAULT_VALUE,KEYWORD,DESCRIPTION,HELP)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_LOGI),                INTENT(INOUT)   ::  SELF
        LOGICAL,                        INTENT(IN)      ::  DEFAULT_VALUE   !<  Default value for the parameter
        CHARACTER(LEN=*),               INTENT(IN)      ::  KEYWORD         !<  Keyword used in frealix command files to identify the parameter
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  DESCRIPTION     !<  Description used in the question and as comment in frealix command files
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  HELP            !<  Help to be given to user to describe what the parameter means / does
        ! Start work
        CALL RTP_INIT_SLAVE(SELF,KEYWORD,DESCRIPTION=DESCRIPTION,HELP=HELP)
        !SELF%DEFAULT_VALUE  =   DEFAULT_VALUE
        SELF%VALUE          =   DEFAULT_VALUE
        SELF%SET_BY_USER    =   .FALSE.
    END SUBROUTINE RTP_INIT_LOGI

    !>  \brief  Initialise a runtime parameter
    SUBROUTINE RTP_INIT_CHAR(SELF,DEFAULT_VALUE,KEYWORD,DESCRIPTION,HELP)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP_CHAR),                INTENT(INOUT)   ::  SELF
        CHARACTER(LEN=*),               INTENT(IN)      ::  DEFAULT_VALUE   !<  Default value for the parameter
        CHARACTER(LEN=*),               INTENT(IN)      ::  KEYWORD         !<  Keyword used in frealix command files to identify the parameter
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  DESCRIPTION     !<  Description used in the question and as comment in frealix command files
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  HELP            !<  Help to be given to user to describe what the parameter means / does
        ! Start work
        CALL RTP_INIT_SLAVE(SELF,KEYWORD,DESCRIPTION=DESCRIPTION,HELP=HELP)
        !SELF%DEFAULT_VALUE  =   DEFAULT_VALUE
        SELF%VALUE          =   DEFAULT_VALUE
        SELF%SET_BY_USER    =   .FALSE.
    END SUBROUTINE RTP_INIT_CHAR



    !>  \brief  Initialise a runtime parameter
    SUBROUTINE RTP_INIT_SLAVE(SELF,KEYWORD,DESCRIPTION,HELP)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),                     INTENT(INOUT)   ::  SELF
        CHARACTER(LEN=*),               INTENT(IN)      ::  KEYWORD     !<  Keyword used in frealix command files to identify the parameter
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  DESCRIPTION !<  Description used in the question and as comment in frealix command files
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)      ::  HELP        !<  Help to be given to user to describe what the parameter means / does
        ! Private variables

        ! Start work
        SELF%KEYWORD    =   KEYWORD
        IF (PRESENT(DESCRIPTION)) THEN
            SELF%DESCRIPTION = DESCRIPTION
        ENDIF

    END SUBROUTINE RTP_INIT_SLAVE

    !>  \brief  The RTP's value pointer is pointed at the relevant variable
    SUBROUTINE RTP_ASSIGN_REAL(SELF,VAR)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),         INTENT(INOUT)   ::  SELF
        REAL,               INTENT(IN)      ::  VAR
        ! Private variables
        ! Start work
        SELECT TYPE(SELF)
            TYPE IS (RTP_REAL)
                SELF%VALUE = VAR
            CLASS DEFAULT
                WRITE(*,'(A)') '**ERROR(RTP_ASSIGN_REAL): RTP / variable type mismatch'
                CALL TERMINATE('Fatal error in RTP_ASSIGN_REAL')
        END SELECT
    END SUBROUTINE RTP_ASSIGN_REAL

    !>  \brief  The RTP's value pointer is pointed at the relevant variable
    SUBROUTINE RTP_ASSIGN_INTG(SELF,VAR)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),             INTENT(INOUT)   ::  SELF
        INTEGER,        TARGET, INTENT(IN)      ::  VAR
        ! Private variables
        ! Start work
        SELECT TYPE(SELF)
            TYPE IS (RTP_INTG)
                SELF%VALUE = VAR
            CLASS DEFAULT
                WRITE(*,'(A)') '**ERROR(RTP_ASSIGN_INTG): RTP / variable type mismatch'
                CALL TERMINATE('Fatal error in RTP_ASSIGN_INTG')
        END SELECT
    END SUBROUTINE RTP_ASSIGN_INTG

    !>  \brief  The RTP's value pointer is pointed at the relevant variable
    SUBROUTINE RTP_ASSIGN_LOGI(SELF,VAR)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),             INTENT(INOUT)   ::  SELF
        LOGICAL,        TARGET, INTENT(IN)      ::  VAR
        ! Private variables
        ! Start work
        SELECT TYPE(SELF)
            TYPE IS (RTP_LOGI)
                SELF%VALUE = VAR
            CLASS DEFAULT
                WRITE(*,'(A)') '**ERROR(RTP_ASSIGN_LOGI): RTP / variable type mismatch'
                CALL TERMINATE('Fatal error in RTP_ASSIGN_LOGI')
        END SELECT
    END SUBROUTINE RTP_ASSIGN_LOGI

    !>  \brief  The RTP's value pointer is pointed at the relevant variable
    SUBROUTINE RTP_ASSIGN_CHAR(SELF,VAR)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),                     INTENT(INOUT)   ::  SELF
        CHARACTER(LEN=*),       TARGET, INTENT(IN)      ::  VAR
        ! Private variables
        ! Start work
        SELECT TYPE(SELF)
            TYPE IS (RTP_CHAR)
                SELF%VALUE = VAR
            CLASS DEFAULT
                WRITE(*,'(A)') '**ERROR(RTP_ASSIGN_CHAR): RTP / variable type mismatch'
                CALL TERMINATE('Fatal error in RTP_ASSIGN_CHAR')
        END SELECT
    END SUBROUTINE RTP_ASSIGN_CHAR



    !>  \brief  Fetch the value for a runtime parameter, either via user interface or by parsing command file
    SUBROUTINE RTP_GET_VALUE(SELF)
        USE UI
        USE PARSERS
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),                 INTENT(INOUT)   ::  SELF
        ! Private variables
        CHARACTER(LEN=LINE_LEN) ::  QUESTION
        ! Start work
        QUESTION = 'Please give ' // TRIM(SELF%KEYWORD)
        IF (INTERACTIVE) THEN
            ! Eventually, the UI routines should be rewritten to take RTP objects as arguments, so that the select type will be done at a lower level
            SELECT TYPE(SELF)
                TYPE IS (RTP_CHAR)
                    CALL UI_GET_CHAR(QUESTION,SELF%VALUE)
                TYPE IS (RTP_LOGI)
                    CALL UI_GET_LOGI(QUESTION,SELF%VALUE)
                TYPE IS (RTP_REAL)
                    CALL UI_GET_REAL(QUESTION,SELF%VALUE)
                TYPE IS (RTP_INTG)
                    CALL UI_GET_INTG(QUESTION,SELF%VALUE)
            END SELECT
            SELF%SET_BY_USER = .TRUE.
        ELSE
            SELECT TYPE(SELF)
                TYPE IS (RTP_CHAR)
                    CALL FILE_GET_CHAR(CONTROL_FILENAME,SELF%KEYWORD,SELF%VALUE,FOUND=SELF%SET_BY_USER)
                TYPE IS (RTP_LOGI)
                    CALL FILE_GET_LOGI(CONTROL_FILENAME,SELF%KEYWORD,SELF%VALUE,FOUND=SELF%SET_BY_USER)
                TYPE IS (RTP_REAL)
                    CALL FILE_GET_REAL(CONTROL_FILENAME,SELF%KEYWORD,SELF%VALUE,FOUND=SELF%SET_BY_USER)
                TYPE IS (RTP_INTG)
                    CALL FILE_GET_INTG(CONTROL_FILENAME,SELF%KEYWORD,SELF%VALUE,FOUND=SELF%SET_BY_USER)
            END SELECT
        ENDIF
    END SUBROUTINE RTP_GET_VALUE

    !>  \brief  Given an array of RTPs, and a string of characters, find the RTP value in the string if possible
    SUBROUTINE  RTP_FROM_STRING(RTPS,STRING)
        USE PARSERS
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),         ALLOCATABLE,    INTENT(INOUT)   ::  RTPS(:) !<  Array of runtime parameters to check the line against
        CHARACTER(LEN=WORD_LEN),            INTENT(IN)      ::  STRING  !<  A line (string) of characters
        ! Private variables
        CHARACTER(LEN=WORD_LEN),    ALLOCATABLE ::  WORDS(:)
        CHARACTER(LEN=WORD_LEN),    ALLOCATABLE ::  LABEL, VALUE
        INTEGER ::  I
        ! Start work

        !
        ! Analyse the line. The first word is the label, the second is the value
        !
        WORDS = SPLIT(STRING)
        LABEL = WORDS(1)
        VALUE = WORDS(2)

        !
        ! Loop over the array of RTPS, to find the one, if any, which is designated by the string
        !
        DO I=1,SIZE(RTPS)
            IF (STRING_EQUAL(LABEL,RTPS(I)%KEYWORD)) THEN
                CALL RTP_VALUE_FROM_STRING(RTPS(I),VALUE)
                RTPS(I)%SET_BY_USER = .TRUE.
            ENDIF
        ENDDO
    END SUBROUTINE RTP_FROM_STRING

    !>  \brief  Set RTP's value to that described by the supplied string
    SUBROUTINE RTP_VALUE_FROM_STRING(SELF,STRING)
        IMPLICIT NONE
        ! Arguments
        CLASS(RTP),                         INTENT(INOUT)   ::  SELF    !<  Runtime parameter for which the value should be set
        CHARACTER(LEN=WORD_LEN),            INTENT(IN)      ::  STRING  !<  Character string with value for the RTP
        ! Private variables

        ! Start work
        SELECT TYPE(SELF)
            TYPE IS (RTP_CHAR)
                READ(STRING,*) SELF%VALUE
            TYPE IS (RTP_LOGI)
                READ(STRING,*) SELF%VALUE
            TYPE IS (RTP_REAL)
                READ(STRING,*) SELF%VALUE
            TYPE IS (RTP_INTG)
                READ(STRING,*) SELF%VALUE
        END SELECT
    END SUBROUTINE RTP_VALUE_FROM_STRING
END MODULE RTPS
