!>    \brief Collection of all parameters that affect runtime behaviour
!!
!!    This module contains declarations for all variables which control run-time behaviour,
!!    most of which can be amended at run time by the user by using a control file.
!!    Default values for all variables are specified such that the user does not need
!!    to supply their own values.
!!
MODULE RUNTIME_PARAMETERS
    use DatesAndTimes
    IMPLICIT NONE

    ! The user cannot change the value of these variables
    INTEGER,            PARAMETER               ::  FILENAME_LEN            =   200             !< Maximum number of characters in a file name
    CHARACTER(LEN=4),   PARAMETER               ::  FILENAME_LEN_C          =   '200 '
    INTEGER,            PARAMETER               ::  RTP_LEN                 =   64              !< Maximum number of characters in a run-time parameter (RTP) name
    INTEGER,            PARAMETER               ::  LINE_LEN                =   1024            !< Maximum number of characters in a line of a file being read/written
    CHARACTER(LEN=4),   PARAMETER               ::  LINE_LEN_C              =   '1024'
    INTEGER,            PARAMETER               ::  WORD_LEN                =   64              !< Maximum number of characters in a word (from a flat-text file)
    CHARACTER(LEN=4),   PARAMETER               ::  WORD_LEN_C              =   '64  '

    ! The control file name will be set by the first command-line argument
    CHARACTER(LEN=FILENAME_LEN)                 ::  CONTROL_FILENAME

    ! Frealix mode. See documentation for allowed values.
    CHARACTER(LEN=FILENAME_LEN)                 ::  FREALIX_MODE

    ! Default file format
    CHARACTER(LEN=1)                            ::  DEFAULT_FILE_FORMAT     =   'I'             !< The default file format:
                                                                                                !! 'I', 'M' or 'S' for imagic, mrc, spider
    ! Version string
    CHARACTER(LEN=15),  PARAMETER               ::  VERSION                 =   '1.0.0'         !< Version number   (0.0.0 indicates development version)

    ! SVN revision info
#if defined(SVN_REV)
    CHARACTER(LEN=*),   PARAMETER               ::  SVN_REVISION            =   SVN_REV
#else
    CHARACTER(LEN=*),   PARAMETER               ::  SVN_REVISION            =   'NoSVNinfo'
#endif

    ! Compile date
#ifdef NAGFOR_COMPILER
    CHARACTER(LEN=*),   PARAMETER               ::  COMPILE_DATE            =    'Unknown'
#else
    CHARACTER(LEN=*),   PARAMETER               ::  COMPILE_DATE            =    __DATE__       !<    The preprocessor subsitutes the date here at compile time
#endif
    ! Whether this is an interactive session
    LOGICAL                                     ::  INTERACTIVE             =   .FALSE.         !< Whether we are running an interactive FREALIX session

    ! FFTW parameters
    LOGICAL                                     ::  USE_FFTW_WISDOM         =   .FALSE.         !< Whether FFTW wisdom should be used
    LOGICAL                                     ::  USE_FFTW_OPENMP         =   .FALSE.         !< Whether FFTW OpenMP threading should be used

    ! Constants
    REAL,               PARAMETER               ::  PI                      =   ACOS(-1.0)
    REAL(KIND=8),       PARAMETER               ::  DPI                     =   ACOS(-1.0D0)
    !REAL,               PARAMETER               ::  PROTEIN_DENSITY         =   0.86            !<    Protein density in Dalton / cubic Angstrom. Converted from 1.43 g/cm^3, see
    !                                                                                            !!  Quillin, M. L. and Matthews, B. W., "Accurate calculation of the density of proteins.", Acta Crystallogr D Biol Crystallogr, vol. 56, pp. 791--794, 2000
    REAL                                        ::  PSF_RADIUS              =   400             !<    In Angstroms. The approximate radius of point spread function.
    REAL                                        ::  PSF_RADIUS_PIXELS       =   0.0             !<    To be set at runtime, when the pixel size is known
    REAL,               PARAMETER               ::  DEFAULT_FALLOFF_PIXELS  =   10.0            !< Default falloff, in pixels

    REAL                                        ::  PBC                     =   0.0             !<  "Fudge" parameter which relates an alignment score to a B factor to be applied during 3D reconstruction


    REAL,               PARAMETER               ::  H_ASYM_UNIT_CENTER(3)   = (/ 1.0, 0.0, 0.0 /)   !< The unit vector that points from the center of the
                                                                                                    !! volume to the origin of the helical unit cell


    ! Start and finish time
    INTEGER                                     ::  DATE_START(8)
    INTEGER                                     ::  DATE_FINISH(8)

    ! File handles, a.k.a. LUNs (logical unit number)
    INTEGER,            PARAMETER               ::  TMP_FILE_HANDLE         =   99
    INTEGER,            PARAMETER               ::  MAX_LUN                 =   200             !<    This is imposed by the frealign/frealix IO routines (which could do with a rewrite!)
    INTEGER,            PARAMETER               ::  MIN_LUN                 =   10              !<    This is to stay out of the way of default LUNs, which for ifort are 0,5,6. I'm not sure what the correct behaviour is, but I'm pretty sure 10 is a safe number.
    LOGICAL                                     ::  LUNS(MIN_LUN:MAX_LUN)   =   .TRUE.          !<    For each LUN at our disposal, .TRUE. if it is available, .FALSE. if it is already used.

    !
    ! The direction of the Z axis.
    !
    ! There are 2 possibilities:
    !
    ! A. Z_AXIS_DIRECTION is -1.0. Increasing Z reduces underfocus. A cryo-EM micrograph collected at 2.3 um underfocus has a Z height of -23000pxl if the pixel size is 1 A/pxl. As Z increases, features become closer to focus.
    ! or
    ! B. Z_AXIS_DIRECTION is 1.0. Increasing Z increases underfocus.
    !
    ! Since underfocus corresponds to the case where the objective focal plain is further away from the electron source than the specimen is, A and B can be described as:
    !
    ! A. The Z axis goes from the electron source to the detector. Z=0 is the focal plain.
    ! or
    ! B. The Z axis goes from the detector to the electron source. Z=0 is the focal plain.
    !
    !
    REAL,               PARAMETER            ::    Z_AXIS_DIRECTION         =   -1.0

    ! Microtubule-specific parameters
    REAL                                    ::    MT_PHI_FIRST_PF           =   0.0             !<    For microtubules only: the latitude (phi angle; degrees) at which protofilament 1 is to be found in the geometry of the 3D model/reconstruction.
                                                                                                !!Protofilaments are labelled 1 to MT_PF_NUMBER. For a given "washer", the tubulin of pf 1 are nearer the - end of the filament.
    ! Endenianness of the system (not checked)
    LOGICAL,            PARAMETER            ::     BIG_ENDIAN              =   ICHAR(TRANSFER(1,'a')) == 0 !<  Test whether the system is big-endian

    INTERFACE SAFE_FOR_SQRT
        MODULE PROCEDURE SAFE_FOR_SQRT_REAL
        MODULE PROCEDURE SAFE_FOR_SQRT_DBLE
    END INTERFACE

    CONTAINS

!        !>    Convert degrees to radians
!        ELEMENTAL PURE REAL FUNCTION RADIANS(DEGREES)
!            IMPLICIT NONE
!            REAL, INTENT(IN) :: DEGREES
!            RADIANS = DEGREES * PI / 180.
!        END FUNCTION RADIANS
!
!        !>    Convert radians to degrees
!        ELEMENTAL PURE REAL FUNCTION DEGREES(RADIANS)
!            IMPLICIT NONE
!            REAL, INTENT(IN) :: RADIANS
!            DEGREES = RADIANS / PI * 180
!        END FUNCTION DEGREES


    !>    \brief    Return the next available LUN
    INTEGER FUNCTION AVAILABLE_LUN()
        IMPLICIT NONE
        ! Private variables
        INTEGER    ::    I
        LOGICAL    ::    SUCCESS
        !$OMP    CRITICAL        (LUN_OMP_CRIT)
        ! Start work
        SUCCESS = .FALSE.
        ! Loop through the LUNS array until an available unit is found
        DO I=MIN_LUN,MAX_LUN
            IF (LUNS(I)) THEN
                ! LUN I is available
                SUCCESS = .TRUE.
                AVAILABLE_LUN = I
                EXIT
            ENDIF
        ENDDO
        IF (SUCCESS) THEN
            ! Now we have found an available LUN, make it unavailable for the next call before we return
            LUNS(AVAILABLE_LUN) = .FALSE.
        ELSE
            WRITE(*,'(A)') '**ERROR(AVAILABLE_LUN): was unable to find an available LUN'
            CALL TERMINATE('Fatal error in AVAILABLE_LUN')
        ENDIF
        !$OMP    END CRITICAL    (LUN_OMP_CRIT)
    END FUNCTION AVAILABLE_LUN

    !>    \brief    Mark a given LUN as available again
    SUBROUTINE LUN_IS_NOW_AVAILABLE(LUN)
        IMPLICIT NONE
        ! Arguments
        INTEGER,    INTENT(IN)    ::    LUN    !<    Number of the LUN to "free up"
        ! Start work
        IF (LUN .LT. MIN_LUN .OR. LUN .GT. MAX_LUN) THEN
            WRITE(*,'(A,I4)') '**WARNING: attempt to free up a LUN which is not in range: ', LUN
        ELSE
            !$OMP    CRITICAL        (LUN_OMP_CRIT)
            LUNS(LUN) = .TRUE.
            !$OMP    END CRITICAL    (LUN_OMP_CRIT)
        ENDIF
    END SUBROUTINE LUN_IS_NOW_AVAILABLE

    !>    \brief    Crash out with error message and useful information, like time/date and traceback
    SUBROUTINE TERMINATE(STRING,DONT_STOP)
        use DatesAndTimes
#if defined(__INTEL_COMPILER)
        USE IFCORE
        USE IFPORT
#elif defined(NAGFOR_COMPILER)
        USE F90_UNIX, ONLY : FLUSH
#endif
        IMPLICIT NONE
        ! Arguments
        CHARACTER(LEN=*),   OPTIONAL,   INTENT(IN)    ::    STRING      !<  Message to be printed out
        LOGICAL,            OPTIONAL,   INTENT(IN)    ::    DONT_STOP   !<  Whether an abort signal should be raised so that a core dump occurs
        ! Private variables
        LOGICAL    ::    DDONT_STOP
        type(DateAndTime)   ::  myTime
#if defined(NAGFOR_COMPILER)
        real                ::  dummy
#endif
        ! Start work
        DDONT_STOP = .FALSE.
        IF (PRESENT(DONT_STOP)) DDONT_STOP = DONT_STOP

        DATE_FINISH = myTime%DateAndTimeAsIntegers()

#if defined(__GFORTRAN__)
        IF (.NOT. DDONT_STOP) THEN
            WRITE(*,'(A)') '</frealix>'
            STOP
        ENDIF
#elif defined(__INTEL_COMPILER)
        IF (PRESENT(STRING)) THEN
            WRITE(*,'(/A/)') STRING
        ELSE
            WRITE(*,'(/2A/)')   'Error termination of frealix, traceback below ', &
                                '(if you do not see a stack trace below, check the environment variable FOR_DISABLE_STACK_TRACE).'
        ENDIF
        CALL TRACEBACKQQ(USER_EXIT_CODE=-1)
        WRITE(*,'(/2A/)') 'Termination date & time : ', myTime%DateAndTimeAsString()

        WRITE(*,'(/4A/)') myTime%DateAndTimeAsString(), ': FREALIX says sorry (total execution time: ', &
                                                myTime%DurationAsString(REAL(myTime%SecondsBetweenDates(DATE_START,DATE_FINISH))), ')'
        IF (DDONT_STOP) THEN
            WRITE(*,'(A)') '**debug(TERMINATE): abnormal termination, see error messages above'
        ELSE
            WRITE(*,'(A)') '</frealix>'
            STOP
        ENDIF
#elif defined(NAGFOR_COMPILER)
        CALL FLUSH(6)
        IF (PRESENT(STRING)) THEN
            WRITE(*,'(/A/)') STRING
        ELSE
            WRITE(*,'(/2A/)')   'Error termination of frealix, traceback below ', &
                                '(if you do not see a stack trace below, check the environment variable FOR_DISABLE_STACK_TRACE).'
        ENDIF
        ! Now we generate a trace back the ugly way
        WRITE(*,'(A)') '**info(TERMINATE): Frealix is terminating in error state, see error message(s) above for details.'
        WRITE(*,'(A)') '**info(TERMINATE): will now attempt to generate a back trace by trigerring a floating point exception'
        WRITE(*,'(/4A/)') myTime%DateAndTimeAsString(), ': FREALIX says sorry (total execution time: ', &
                                        myTime%DurationAsString(REAL(myTime%SecondsBetweenDates(DATE_START,DATE_FINISH))), ')'
        CALL FLUSH(6)
        DUMMY = 0.0
        !DUMMY = 100.0/DUMMY
		DUMMY = ATAN2(DUMMY,DUMMY)
		IF (DDONT_STOP) THEN
            WRITE(*,'(A)') '**debug(TERMINATE): abnormal termination, see error messages above'
        ELSE
            WRITE(*,'(A)') '</frealix>'
            STOP
        ENDIF
#else
        IF (.NOT. DDONT_STOP) THEN
            WRITE(*,'(A)') '</frealix>'
            STOP
        ENDIF
#endif
    END SUBROUTINE TERMINATE

    !>    \brief    Register signal handlers
    SUBROUTINE REGISTER_SIGNAL_HANDLERS()
#if defined(__INTEL_COMPILER)
        USE IFPORT
#endif
        IMPLICIT NONE
        ! Arguments

        ! Private variables
        INTEGER    ::    H_RET
        ! Start work
#if defined(__INTEL_COMPILER)
        H_RET = SIGNALQQ(SIG$ABORT,SIGNAL_HANDLER)
        H_RET = SIGNALQQ(SIG$FPE,FPE_HANDLER) + H_RET
        H_RET = SIGNALQQ(SIG$ILL,SIGNAL_HANDLER) + H_RET
        H_RET = SIGNALQQ(SIG$INT,SIGNAL_HANDLER) + H_RET
        H_RET = SIGNALQQ(SIG$SEGV,SIGNAL_HANDLER) + H_RET
        H_RET = SIGNALQQ(SIG$TERM,SIGNAL_HANDLER) + H_RET
        IF (H_RET .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(REGISTER_SIGNAL_HANDLERS): errors reported by signalqq. h_ret = ', H_RET
            CALL TERMINATE('Failed when initialising abort signal handler')
        ENDIF
#else
        ! Signal handlers not implemented (yet)
#endif
    END SUBROUTINE REGISTER_SIGNAL_HANDLERS


    !>    \brief    Signal handler for ABORT signal (only available when built with Intel compiler)
#if defined(__INTEL_COMPILER)
    FUNCTION SIGNAL_HANDLER(SIGNUM)
        USE IFPORT
        IMPLICIT NONE
        ! Argument
        INTEGER(KIND=2),    INTENT(IN)    ::    SIGNUM    !< Identifies the signal
        ! Result
        !DEC$ ATTRIBUTES C :: SIGNAL_HANDLER
        INTEGER(KIND=4)    ::    SIGNAL_HANDLER
        ! Private variables
        LOGICAL,    PARAMETER    ::    DEBUG    =    .FALSE.
        ! Start work
        WRITE(*,'(A)') ' '
        IF (DEBUG) THEN
            WRITE(*,'(A,I0)') '**debug(SIGNAL_HANDLER): caught signal # ', SIGNUM
        ENDIF
        ! Work out which signal it is
        ! Here is the full list, from ifport.f90:
        !INTEGER(4), PARAMETER :: SIG$ERR   = -1
        !INTEGER(4), PARAMETER :: SIG$INT   =  2
        !INTEGER(4), PARAMETER :: SIG$ILL   =  4
        !INTEGER(4), PARAMETER :: SIG$FPE   =  8
        !INTEGER(4), PARAMETER :: SIG$SEGV  = 11
        !INTEGER(4), PARAMETER :: SIG$TERM  = 15
        !INTEGER(4), PARAMETER :: SIG$USR1  = 16
        !INTEGER(4), PARAMETER :: SIG$USR2  = 17
        !INTEGER(4), PARAMETER :: SIG$USR3  = 20
        !INTEGER(4), PARAMETER :: SIG$BREAK = 21
        !INTEGER(4), PARAMETER :: SIG$ABORT = 22
        !INTEGER(4), PARAMETER :: SIG$NSIG  = 23
        SELECT CASE (SIGNUM)
            CASE (SIG$ABORT)
                WRITE(*,'(A)') '**info(SIGNAL_HANDLER): caught abnormal termination signal (SIGABORT). Will terminate.'
                CALL TERMINATE('Abnormal termination signal received',DONT_STOP=.TRUE.)
            CASE (SIG$FPE)
                WRITE(*,'(2A)') '**ERROR(SIGNAL_HANDLER): received floating-point error signal (SIGFPE), ', &
                                'but this should have been handled by another routine!'
                WRITE(*,'(A)') '**info(SIGNAL_HANDLER): caught floating-point error signal (SIGFPE). Will terminate.'
                CALL TERMINATE('Floating-point error signal received by wrong signal handler',DONT_STOP=.TRUE.)
            CASE (SIG$ILL)
                WRITE(*,'(A)')  '**info(SIGNAL_HANDLER): caught illegal instruction error signal (SIGILL). Will terminate.'
                CALL TERMINATE('Illegal instruction signal received',DONT_STOP=.TRUE.)
            CASE (SIG$INT)
                WRITE(*,'(2A)') '**info(SIGNAL_HANDLER): caught interrupt signal (SIGINT). ',    &
                                'This may be because you used Ctrl-C. Will terminate.'
                CALL TERMINATE('Interrupt signal received')
            CASE (SIG$SEGV)
                WRITE(*,'(2A)') '**info(SIGNAL_HANDLER): caught illegal storage access signal (SIGSEGV), ', &
                                'aka segmentation fault. Will terminate.'
                CALL TERMINATE('Segmentation fault signal received',DONT_STOP=.TRUE.)
            CASE (SIG$TERM)
                WRITE(*,'(A)') '**info(SIGNAL_HANDLER): caught termination request signal (SIGTERM). Will terminate.'
                CALL TERMINATE('Termination request received')
            CASE DEFAULT
                WRITE(*,'(A,I0,A)') '**ERROR(SIGNAL_HANDLER): caught unknown signal number ', SIGNUM, '. Will terminate.'
                CALL TERMINATE('Unknown signal received')
        END SELECT
        !
        SIGNAL_HANDLER = 1
    END FUNCTION SIGNAL_HANDLER
#endif

    !>    \brief    Signal handler for floating point error signals (only available when built with Intel compiler)
#if defined(__INTEL_COMPILER)
    FUNCTION FPE_HANDLER(SIGNUM,SIGINFO)
        USE IFCORE
        USE IFPORT
        IMPLICIT NONE
        ! Argument
        INTEGER(KIND=2),    INTENT(IN)    ::    SIGNUM    !< Identifies the signal
        INTEGER(KIND=2),    INTENT(IN)    ::    SIGINFO
        ! Result
        !DEC$ ATTRIBUTES C :: FPE_HANDLER
        INTEGER(KIND=4)    ::    FPE_HANDLER
        ! Private variables
        LOGICAL,    PARAMETER    ::    DEBUG    =    .FALSE.
        ! Start work
        WRITE(*,'(A)') ' '
        IF (DEBUG) THEN
            WRITE(*,'(A,2(I0,1X))') '**debug(FPE_HANDLER): caught signal #, info ', SIGNUM, SIGINFO
        ENDIF
        ! Work out which signal it is
        SELECT CASE (SIGNUM)
            CASE (SIG$FPE)
                WRITE(*,'(A)',ADVANCE='NO') '**info(FPE_HANDLER): caught the following floating-point error signal (SIGFPE): '
                SELECT CASE (SIGINFO)
                    CASE (FPE$INVALID)
                        WRITE(*,'(A)',ADVANCE='NO') 'invalid operation'
                    CASE (FPE$ZERODIVIDE)
                        WRITE(*,'(A)',ADVANCE='NO') 'division by zero'
                    CASE (FPE$OVERFLOW)
                        WRITE(*,'(A)',ADVANCE='NO') 'numeric overflow'
                    CASE (FPE$UNDERFLOW)
                        WRITE(*,'(A)',ADVANCE='NO') 'numeric underflow'
                    CASE (FPE$INEXACT)
                        WRITE(*,'(A)',ADVANCE='NO') 'inexact precision'
                    CASE DEFAULT
                        WRITE(*,'(A)',ADVANCE='NO') 'unknown'
                END SELECT
                WRITE(*,'(A)') '. Will terminate.'
                CALL CORE_DUMP_INSTRUCTIONS()
                CALL TERMINATE('Floating-point error signal received',DONT_STOP=.TRUE.)
            CASE DEFAULT
                WRITE(*,'(A,I0,A)') '**ERROR(FPE_HANDLER): caught unknown signal number ', SIGNUM, '. Will terminate.'
                CALL TERMINATE('Unknown signal received')
        END SELECT
        !
        FPE_HANDLER = 1
    END FUNCTION FPE_HANDLER
#endif

    !>    \brief    Print out instructions on how to get a core dump
    SUBROUTINE CORE_DUMP_INSTRUCTIONS()
        IMPLICIT NONE
        ! Argumnets

        ! Private variables

        ! Start work
        WRITE(*,'(/2A)')'**info(CORE_DUMP_INSTRUCTIONS): if you would like to get a core dump for debugging purposes, ',   &
                        'issue the following commands and re-run the program:'
        WRITE(*,'(A)')  '**info(CORE_DUMP_INSTRUCTIONS): C-shell: limit coredumpsize unlimited; setenv decfort_dump_flag y'
        WRITE(*,'(A/)') '**info(CORE_DUMP_INSTRUCTIONS): Bash: ulimit -c unlimited; export decfort_dump_flag=y'
    END SUBROUTINE CORE_DUMP_INSTRUCTIONS

    !>    \brief    Check whether the real supplied as argument is safe for square-rooting. If not, crash out.
    !!    \todo    Come up with a proper tolerance
    SUBROUTINE SAFE_FOR_SQRT_REAL(SELF,NEGATIVE_TO_ZERO)
!#if defined(__INTEL_COMPILER)
!        USE IFPORT
!#endif
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=4),                INTENT(INOUT)  ::    SELF
        LOGICAL,        OPTIONAL,    INTENT(IN)     ::    NEGATIVE_TO_ZERO    !<    If true and if the input value is negative, it'll be zeroed
        ! Private variables
        REAL(KIND=4),   PARAMETER        ::    TOLERANCE    =    -.0000001
        LOGICAL    ::    NNEGATIVE_TO_ZERO
        ! Start work
        IF (PRESENT(NEGATIVE_TO_ZERO)) THEN
            NNEGATIVE_TO_ZERO = NEGATIVE_TO_ZERO
        ELSE
            NNEGATIVE_TO_ZERO = .FALSE.
        ENDIF

        ! We only need to worry about the variable being negative
        IF (SELF .LT. 0.) THEN
            IF (SELF .GT. TOLERANCE .OR. NNEGATIVE_TO_ZERO) THEN
                SELF = 0.
            ELSE
                WRITE(*,'(A,G15.7)') '**ERROR(SAFE_FOR_SQRT_REAL): cannot compute square root of ', SELF
                CALL TERMINATE('Fatal error in SAFE_FOR_SQRT_REAL'    &
!#if defined(__INTEL_COMPILER)
!                , DONT_STOP=.TRUE.    &
!#endif
                )
!#if defined(__INTEL_COMPILER)
!                IDUMMY = RAISEQQ(SIG$FPE)
!#endif
            ENDIF
        ENDIF
    END SUBROUTINE SAFE_FOR_SQRT_REAL

    !>    \brief    Check whether the double supplied as argument is safe for square-rooting. If not, crash out.
    !!    \todo    Come up with a proper tolerance
    SUBROUTINE SAFE_FOR_SQRT_DBLE(SELF,NEGATIVE_TO_ZERO)
!#if defined(__INTEL_COMPILER)
!        USE IFPORT
!#endif
        IMPLICIT NONE
        ! Arguments
        REAL(KIND=8),                INTENT(INOUT)    ::    SELF
        LOGICAL,        OPTIONAL,    INTENT(IN)        ::    NEGATIVE_TO_ZERO    !<    If true and if the input value is negative, it'll be zeroed
        ! Private variables
        REAL(KIND=8),    PARAMETER        ::    TOLERANCE    =    -.000000001D0
        LOGICAL    ::    NNEGATIVE_TO_ZERO
        ! Start work
        IF (PRESENT(NEGATIVE_TO_ZERO)) THEN
            NNEGATIVE_TO_ZERO = NEGATIVE_TO_ZERO
        ELSE
            NNEGATIVE_TO_ZERO = .FALSE.
        ENDIF

        ! We only need to worry about the variable being negative
        IF (SELF .LT. 0.D0) THEN
            IF (SELF .GT. TOLERANCE .OR. NNEGATIVE_TO_ZERO) THEN
                SELF = 0.D0
            ELSE
                WRITE(*,'(A,G15.7)') '**ERROR(SAFE_FOR_SQRT_DBLE): cannot compute square root of ', SELF
                CALL TERMINATE('Fatal error in SAFE_FOR_SQRT_DBLE'    &
!#if defined(__INTEL_COMPILER)
!                , DONT_STOP=.TRUE.    &
!#endif
                )
!#if defined(__INTEL_COMPILER)
!                IDUMMY = RAISEQQ(SIG$FPE)
!#endif
            ENDIF
        ENDIF
    END SUBROUTINE SAFE_FOR_SQRT_DBLE


    !>    \brief    Print a welcome message to the terminal
    SUBROUTINE PRINT_WELCOME()
        IMPLICIT NONE
        ! Arguments

        ! Private variables
        CHARACTER(LEN=2),    PARAMETER    ::    TEXT_WIDTH    =    '30'
        type(DateAndTime)   ::  myTime
        ! Start work
        WRITE(*,'(/18X,A/)')                '** Welcome to Frealix **'
        WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')        'Version:',                    VERSION
        WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')        'SVN revision:',            SVN_REVISION
        WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')        'Build date:',                COMPILE_DATE
        WRITE(*,'(A'//TEXT_WIDTH//',1X)',ADVANCE='NO')        'Debug symbols?:'
#if defined (WITH_DEBUG_SYMBOLS)
        WRITE(*,'(A)')                        'Yes'
#else
        WRITE(*,'(A)')                        'No'
#endif
        WRITE(*,'(A)')                         ' '
        IF (INTERACTIVE) THEN
        WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')        'Interactive mode?:', 'Yes'
        ELSE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')        'Control filename:',    TRIM(ADJUSTL(CONTROL_FILENAME))
        ENDIF
        WRITE(*,'(A)')                         ' '
        CALL PRINT_ENVIRONMENT_INFO()
        WRITE(*,'(A)')                         ' '
        WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')        'Date & time:', myTime%DateAndTimeAsString()
        WRITE(*,'(A)')                         ' '
        ! Flush standard output
        FLUSH(6) ! This is a standard F2003 statement
    END SUBROUTINE PRINT_WELCOME

    !>    \brief    Find values for enviroment variables which may affect frealix at runtime and print them out
    SUBROUTINE PRINT_ENVIRONMENT_INFO()
#ifdef __INTEL_COMPILER
        USE IFPORT
#endif
        IMPLICIT NONE
        ! Arguments

        ! Private variables
        CHARACTER(LEN=2),    PARAMETER          ::    TEXT_WIDTH    =    '30'
        CHARACTER(LEN=WORD_LEN)                 ::    HOME, HOSTNAME, PWD, USERNAME, HOSTTYPE, MACHTYPE
        CHARACTER(LEN=LINE_LEN)                 ::    COMMAND_LINE
        INTEGER    ::    IERR,CMD_LEN,RES
        ! Start work
        CALL GET_ENV_VAR('HOME',HOME)
#ifdef __INTEL_COMPILER
        RES = HOSTNAM(HOSTNAME)
        IF (RES .NE. 0) CALL GET_ENV_VAR('HOST',HOSTNAME)
#else
        CALL GET_ENV_VAR('HOST',HOSTNAME)
#endif
        CALL GET_ENV_VAR('PWD',PWD)
        CALL GET_ENV_VAR('USER',USERNAME)
        CALL GET_ENV_VAR('HOSTTYPE',HOSTTYPE)
        CALL GET_ENV_VAR('MACHTYPE',MACHTYPE)
        CALL GET_COMMAND(COMMAND_LINE,LENGTH=CMD_LEN,STATUS=IERR)
        IF (IERR .NE. 0) CALL TERMINATE('Could not get command line')

        WRITE(*,'(A'//TEXT_WIDTH//',1X,A,1X,4A)')   'User name (home directory):',      TRIM(ADJUSTL(USERNAME)), '(', &
                                                                                        TRIM(ADJUSTL(HOME)), ')'
        WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')         'Working directory:',               TRIM(ADJUSTL(PWD))
        WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')         'Command line:',                    TRIM(ADJUSTL(COMMAND_LINE))
        WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')         'Host name:',                       TRIM(ADJUSTL(HOSTNAME))
        WRITE(*,'(A'//TEXT_WIDTH//',1X,2A)')        'Host, machine type:',              TRIM(ADJUSTL(HOSTTYPE))//', '//&
                                                                                        TRIM(ADJUSTL(MACHTYPE))

        CALL PRINT_OPENMP_INFO()

    END SUBROUTINE PRINT_ENVIRONMENT_INFO

    !>    \brief    Print information on OpenMP which may affect runtime behaviour
    SUBROUTINE PRINT_OPENMP_INFO()
        USE OMP_LIB
        IMPLICIT NONE
        ! Arguments

        ! Private variables
        CHARACTER(LEN=2),  PARAMETER     ::     TEXT_WIDTH      =   '30'
        CHARACTER(LEN=WORD_LEN)          ::     ENV_OMP_NUM_THREADS
        LOGICAL    ::    ENV_OMP_NUM_THREADS_IS_SET
        LOGICAL    ::    OMP_IS_ON
        ! Start work

        OMP_IS_ON = OPENMP_IS_AVAILABLE()

        IF (OMP_IS_ON) THEN
            ! We manually turn thread nesting off, to keep things simple
            CALL OMP_SET_NESTED(.FALSE.)
            WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')    'OpenMP threading available?:', 'Yes'
            ! First printout relevant environment variables the user may have set
            CALL GET_ENV_VAR('OMP_NUM_THREADS',ENV_OMP_NUM_THREADS,IS_SET=ENV_OMP_NUM_THREADS_IS_SET,QUIET=.TRUE.)
            IF (ENV_OMP_NUM_THREADS_IS_SET) THEN
                WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')        'Omp_num_threads env. var.:',     TRIM(ADJUSTL(ENV_OMP_NUM_THREADS))
            ENDIF
            WRITE (*,'(A'//TEXT_WIDTH//',1X,I0)') 'Max number of OpenMP threads:', OMP_GET_MAX_THREADS()
            WRITE (*,'(A'//TEXT_WIDTH//',1X,I0)') 'Number of processors:', OMP_GET_NUM_PROCS()
            IF (OMP_GET_NESTED()) THEN
                WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')    'Nested threading available?:', 'Yes'
            ELSE
                WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')    'Nested threading available?:', 'No'
            ENDIF
        ELSE
            WRITE(*,'(A'//TEXT_WIDTH//',1X,A)')    'OpenMP threading available?:', 'No'
        ENDIF


    END SUBROUTINE PRINT_OPENMP_INFO

    !>    \brief    Test whether OpenMP is available
    LOGICAL FUNCTION OPENMP_IS_AVAILABLE()
        USE OMP_LIB
        IMPLICIT NONE
        ! Arguments

        ! Private variables

        ! Start work

        !$OMP    PARALLEL    &
        !$OMP          SHARED(OPENMP_IS_AVAILABLE)
        !$OMP    SINGLE
        OPENMP_IS_AVAILABLE = OMP_GET_MAX_THREADS() .GT. 1
        !$OMP    END SINGLE
        !$OMP    END PARALLEL
    END FUNCTION OPENMP_IS_AVAILABLE

    !>  \brief  Return the ID of the current process
    INTEGER FUNCTION GET_PROCESS_ID()
#if defined(__INTEL_COMPILER)
        USE IFPORT
#endif
        IMPLICIT NONE
        ! Arguments
        ! Private variables
        ! Start work
#if defined(__INTEL_COMPILER)
        GET_PROCESS_ID = GETPID()
#else
        GET_PROCESS_ID = 0
#endif
    END FUNCTION GET_PROCESS_ID

    !>    \brief    Find out the value of an environment variable at runtime
    SUBROUTINE GET_ENV_VAR(NAME,VALUE,IS_SET,QUIET)
        IMPLICIT NONE
        ! Arguments
        CHARACTER(LEN=*),                   INTENT(IN)      ::    NAME      !<    Status code returned by GET_ENVIRONMENT_VARIABLE
        CHARACTER(LEN=*),                   INTENT(OUT)     ::    VALUE
        LOGICAL,            OPTIONAL,       INTENT(OUT)     ::    IS_SET    !<    Whether the environment variable exists / is set
        LOGICAL,            OPTIONAL,       INTENT(IN)      ::    QUIET     !<    Do not print warnings. True by default.
        ! Private variables
        INTEGER                            ::    LENGTH
        INTEGER                            ::    STATUS
        LOGICAL                            ::    IIS_SET, QQUIET
        CHARACTER(LEN=WORD_LEN)    ::    VAL
        ! Start work

        ! Will we be quiet?
        IF (PRESENT(QUIET)) THEN
            QQUIET = QUIET
        ELSE
            QQUIET = .TRUE.
        ENDIF

        !
        CALL GET_ENVIRONMENT_VARIABLE(NAME,VAL,LENGTH,STATUS,.TRUE.)
        IIS_SET = .FALSE.
        IF (STATUS .GE. 2) THEN
            WRITE(*,'(A,I0)') '**ERROR(GET_ENV_VAR): get_environment_variable failed with error ', STATUS
            CALL TERMINATE('Fatal error')
        ELSE IF (STATUS .EQ. 1) THEN
            IF (.NOT. QQUIET) WRITE(*,'(3A)') '**warning(GET_ENV_VAR): environment variable "', &
                                                TRIM(ADJUSTL(NAME)), '" does not exist'
        ELSE IF (STATUS .EQ. -1) THEN
            IF (.NOT. QQUIET) WRITE(*,'(2A,2(A,I0))') '**warning(GET_ENV_VAR): environment variable "', &
                                                TRIM(ADJUSTL(NAME)), '" length (', LENGTH, ') was trimmed to ', LEN(VAL)
        ELSE IF (LENGTH .EQ. 0) THEN
            IF (.NOT. QQUIET) WRITE(*,'(2A,2(A,I0))') '**warning(GET_ENV_VAR): environment variable "', &
                                                TRIM(ADJUSTL(NAME)), '" exists but has no value'
        ELSE
            IIS_SET = .TRUE.
        ENDIF

        !
        VAL = ADJUSTL(VAL)
        WRITE(VALUE,'(A)') TRIM(ADJUSTL(VAL))

        ! The user asked to know whether that variable was set, so we let them know
        IF (PRESENT(IS_SET)) THEN
            IS_SET = IIS_SET
        ENDIF
    END SUBROUTINE GET_ENV_VAR


    !>    \brief    Returns true if the argument is even
    PURE LOGICAL FUNCTION IS_EVEN_SLOW(INT)
        IMPLICIT NONE
        ! Argument
        INTEGER,    INTENT(IN)    ::    INT
        ! Start work
        IF (MOD(INT,2) .EQ. 0.) THEN
            IS_EVEN_SLOW = .TRUE.
        ELSE
            IS_EVEN_SLOW = .FALSE.
        ENDIF
    END FUNCTION IS_EVEN_SLOW

    !>    \brief    Returns true if the argument is even
    PURE ELEMENTAL LOGICAL FUNCTION IS_EVEN(INT)
        IMPLICIT NONE
        ! Argument
        INTEGER,    INTENT(IN)    ::    INT
        ! Start work
        ! Test bit 0 of number. If it is 0, the number is even
        IS_EVEN = .NOT. BTEST(INT,0)
    END FUNCTION IS_EVEN

    !>    \brief    Returns true if the argument is odd
    PURE ELEMENTAL LOGICAL FUNCTION IS_ODD(INT)
        IMPLICIT NONE
        ! Argument
        INTEGER,    INTENT(IN)    ::    INT
        ! Start work
        IS_ODD = BTEST(INT,0)
    END FUNCTION IS_ODD

    !>    \brief    Returns true if the argument is odd
    PURE LOGICAL FUNCTION IS_ODD_SLOW(INT)
        IMPLICIT NONE
        ! Argument
        INTEGER,    INTENT(IN)    ::    INT
        ! Start work
        IF (IS_EVEN_SLOW(INT)) THEN
            IS_ODD_SLOW = .FALSE.
        ELSE
            IS_ODD_SLOW = .TRUE.
        ENDIF
    END FUNCTION IS_ODD_SLOW

    !>    \brief    Return the array of unique elements present in the input array
    FUNCTION INT_UNIQUE(ARRAY)
        IMPLICIT NONE
        ! Argument
        INTEGER,    INTENT(IN)    ::    ARRAY(:)
        ! Result
        INTEGER,    ALLOCATABLE    ::    INT_UNIQUE(:)
        ! Private variables
        INTEGER    ::    NUMBER_OF_UNIQUE_ELEMENTS
        INTEGER    ::    TMP_RESULT(SIZE(ARRAY))
        INTEGER    ::    I, J, IERR
        LOGICAL    ::    UNIQUE
        ! Start work

        ! Initialise the output
        NUMBER_OF_UNIQUE_ELEMENTS = 1
        TMP_RESULT(1) = ARRAY(1)

        ! Loop over the input array
        LOOP_OVER_INPUT: DO I=2,SIZE(ARRAY)
            UNIQUE = .TRUE.
            ! Loop over the input array
            DO J=1,NUMBER_OF_UNIQUE_ELEMENTS
                IF (ARRAY(I) .EQ. TMP_RESULT(J)) THEN
                    UNIQUE = .FALSE.
                    CYCLE LOOP_OVER_INPUT
                ENDIF
            ENDDO
            ! If we got here, it means the element is unique - add it to the output
            NUMBER_OF_UNIQUE_ELEMENTS = NUMBER_OF_UNIQUE_ELEMENTS + 1
            TMP_RESULT(NUMBER_OF_UNIQUE_ELEMENTS) = ARRAY(I)
        ENDDO LOOP_OVER_INPUT

        ! Allocate memory for the output
        ALLOCATE(INT_UNIQUE(NUMBER_OF_UNIQUE_ELEMENTS),STAT=IERR)
        IF (IERR .NE. 0) THEN
            CALL TERMINATE('Memory allocation failed in INT_UNIQUE')
        ENDIF

        ! Copy values over to the output array
        INT_UNIQUE(1:NUMBER_OF_UNIQUE_ELEMENTS) = TMP_RESULT(1:NUMBER_OF_UNIQUE_ELEMENTS)

    END FUNCTION INT_UNIQUE




END MODULE RUNTIME_PARAMETERS
