!>  \brief  Classes & methods for dealing with single particles
MODULE SINGLE_PARTICLES

    USE COOS
    USE EULERANGLES
    USE CTFS

    ! Portland compiler (v 13.3) does not like extending a type across modules, it seems
    !TYPE, EXTENDS(POINT) :: SINGLE_PARTICLE
    TYPE SINGLE_PARTICLE
        REAL(KIND=8)            ::  COO(3)  =   [0.0D0,0.0D0,0.0D0] !<  X, Y, Z coordinates
        INTEGER                 ::  LOC_IN_STACK            !<  Location number in stack; first image in stack has number 1
        TYPE(EULER)             ::  ANGLES
        TYPE(CTF_PARAMETERS)    ::  CTF
        REAL(KIND=8)            ::  SHIFT_X
        REAL(KIND=8)            ::  SHIFT_Y
        INTEGER                 ::  FILM_NUM
        REAL(KIND=8)            ::  SUBUNIT_NUM =   0.0D0   !<  When dealing with helical assemblies, the position of subunit nearest to the origin
        TYPE(HELICAL_SYMMETRY)  ::  HELIX                   !<  Helical symmetry operator when dealing with segments of helical assembles
        REAL                    ::  SCORE
    END TYPE SINGLE_PARTICLE

    INTERFACE OPERATOR (+)
        MODULE PROCEDURE SP_PLUS_SP
    END INTERFACE

    INTERFACE OPERATOR (-)
        MODULE PROCEDURE SP_MINUS_SP
    END INTERFACE

    CONTAINS

    PURE ELEMENTAL FUNCTION SP_PLUS_SP(SP1,SP2)
        IMPLICIT NONE
        !
        TYPE(SINGLE_PARTICLE), INTENT(IN)  ::  SP1, SP2
        !
        TYPE(SINGLE_PARTICLE)  ::  SP_PLUS_SP
        !
        SP_PLUS_SP%ANGLES      = SP1%ANGLES + SP2%ANGLES
        SP_PLUS_SP%CTF         = SP1%CTF + SP2%CTF
        SP_PLUS_SP%SHIFT_X     = SP1%SHIFT_X + SP2%SHIFT_X
        SP_PLUS_SP%SHIFT_Y     = SP1%SHIFT_Y + SP2%SHIFT_Y
        SP_PLUS_SP%FILM_NUM    = SP1%FILM_NUM + SP2%FILM_NUM
        SP_PLUS_SP%SUBUNIT_NUM = SP1%SUBUNIT_NUM + SP2%SUBUNIT_NUM
        SP_PLUS_SP%HELIX       = SP1%HELIX + SP2%HELIX
        SP_PLUS_SP%SCORE       = SP1%SCORE + SP2%SCORE
    END FUNCTION SP_PLUS_SP

    PURE ELEMENTAL FUNCTION SP_MINUS_SP(SP1,SP2)
        IMPLICIT NONE
        !
        TYPE(SINGLE_PARTICLE), INTENT(IN)  ::  SP1, SP2
        !
        TYPE(SINGLE_PARTICLE)  ::  SP_MINUS_SP
        !
        SP_MINUS_SP%ANGLES      = SP1%ANGLES - SP2%ANGLES
        SP_MINUS_SP%CTF         = SP1%CTF - SP2%CTF
        SP_MINUS_SP%SHIFT_X     = SP1%SHIFT_X - SP2%SHIFT_X
        SP_MINUS_SP%SHIFT_Y     = SP1%SHIFT_Y - SP2%SHIFT_Y
        SP_MINUS_SP%FILM_NUM    = SP1%FILM_NUM - SP2%FILM_NUM
        SP_MINUS_SP%SUBUNIT_NUM = SP1%SUBUNIT_NUM - SP2%SUBUNIT_NUM
        SP_MINUS_SP%HELIX       = SP1%HELIX - SP2%HELIX
        SP_MINUS_SP%SCORE       = SP1%SCORE - SP2%SCORE
    END FUNCTION SP_MINUS_SP



    !>  \brief  Initialise a single particle object
    SUBROUTINE SP_INIT(SELF,CTF,ANGLES,SHIFT_X,SHIFT_Y,PHI,THETA,PSI,SUBUNIT_NUM,HELIX)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        CLASS(SINGLE_PARTICLE),             INTENT(INOUT)   ::  SELF        !<  Single particle object
        TYPE(CTF_PARAMETERS),               INTENT(IN)      ::  CTF         !<  CTF parameters object
        TYPE(EULER),            OPTIONAL,   INTENT(IN)      ::  ANGLES      !<  Euler angles. If not given, and separate angles are not given either, angles will be set to 0.0.
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  SHIFT_X     !<  X shift (pixels)
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  SHIFT_Y     !<  Y shift (pixels)
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  PHI         !<  Phi Euler angles (degrees)
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  THETA       !<  Theta Euler angles (degree)
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  PSI         !<  Psi Euler angles (degrees)
        REAL(KIND=8),           OPTIONAL,   INTENT(IN)      ::  SUBUNIT_NUM !<  Subunit number at the origin (always 0.0 for non-helical objects)
        TYPE(HELICAL_SYMMETRY), OPTIONAL,   INTENT(IN)      ::  HELIX       !<  Helical operator. Required when doing SP processing of filament segments.
        ! Private variables

        ! Start work

        ! Deal with angles
        IF (PRESENT(ANGLES)) THEN
            SELF%ANGLES = ANGLES
            IF (PRESENT(PHI) .OR. PRESENT(THETA) .OR. PRESENT(PSI)) THEN
                WRITE(*,'(A)') '**ERROR(SP_INIT): either give an Euler object, or separate angles, not both'
                CALL TERMINATE('Fatal error in SP_INIT - programming error')
            ENDIF
        ELSE
            IF (PRESENT(PHI)) THEN
                SELF%ANGLES%PHI = PHI
            ELSE
                SELF%ANGLES%PHI = 0.0D0
            ENDIF
            IF (PRESENT(THETA)) THEN
                SELF%ANGLES%THETA = THETA
            ELSE
                SELF%ANGLES%THETA = 0.0D0
            ENDIF
            IF (PRESENT(PSI)) THEN
                SELF%ANGLES%PSI = PSI
            ELSE
                SELF%ANGLES%PSI = 0.0D0
            ENDIF
        ENDIF

        ! Deal with shifts
        IF (PRESENT(SHIFT_X)) THEN
            SELF%SHIFT_X = SHIFT_X
        ELSE
            SELF%SHIFT_X = 0.0D0
        ENDIF
        IF (PRESENT(SHIFT_Y)) THEN
            SELF%SHIFT_Y = SHIFT_Y
        ELSE
            SELF%SHIFT_Y = 0.0D0
        ENDIF

        ! Deal with subunit_num
        IF (PRESENT(SUBUNIT_NUM)) THEN
            SELF%SUBUNIT_NUM = SUBUNIT_NUM
        ELSE
            SELF%SUBUNIT_NUM = 0.0D0
        ENDIF

        ! Deal with helical operator
        IF (PRESENT(HELIX)) THEN
            SELF%HELIX = HELIX
        ELSE
            CALL HELICAL_SYMMETRY_INIT(SELF%HELIX,RISE=0.0,TWIST=0.0)
        ENDIF

        ! Copy the CTF parameters over
        SELF%CTF = CTF
    END SUBROUTINE SP_INIT

    SUBROUTINE SPS_PRINT_INFO(SELF)
        IMPLICIT NONE
        TYPE(SINGLE_PARTICLE),  INTENT(IN)  ::  SELF(:)
        INTEGER ::  I
        DO I=1,SIZE(SELF)
            CALL SP_PRINT_INFO(SELF(I))
        ENDDO
    END SUBROUTINE SPS_PRINT_INFO

    !>  \brief  Print useful info to the terminal which describes the single-particle parameters
    SUBROUTINE SP_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(SINGLE_PARTICLE), INTENT(IN)  ::  SELF    !<  Single particle object
        ! Private variables
        LOGICAL,    PARAMETER   ::  ANNOTATE = .TRUE.
        ! Start work
        IF (ANNOTATE) THEN
            WRITE(*,'(3(F0.3,1X,A,1X),3(F0.3,1X,A,1X),2(F0.3,1X,A,1X),(F0.3,1X,A,1X))')                 &
                                                        SELF%ANGLES%PHI,    '(phi)',            &
                                                        SELF%ANGLES%THETA,  '(theta)',          &
                                                        SELF%ANGLES%PSI,    '(psi)',            &
                                                        SELF%SHIFT_X,       '(x)',              &
                                                        SELF%SHIFT_Y,       '(y)',              &
                                                        SELF%SUBUNIT_NUM,   '(subunit_num)',    &
                                                        SELF%CTF%DFMID1,    '(df1)',            &
                                                        SELF%CTF%DFMID2,    '(df2)',            &
                                                        SELF%SCORE,         '(score)'
        ELSE
            WRITE(*,'(3(F8.3,1X),3(F8.3,1X),2(F8.3,1X))') SELF%ANGLES, SELF%SHIFT_X, SELF%SHIFT_Y, SELF%SUBUNIT_NUM,   &
                                                        SELF%CTF%DFMID1, SELF%CTF%DFMID2
        ENDIF
    END SUBROUTINE SP_PRINT_INFO

END MODULE SINGLE_PARTICLES

!>  \brief  Classes & methods for dealing with comparisons between singles particles and 3D models
MODULE SP_MODEL_COMPARISONS

    USE SINGLE_PARTICLES, ONLY : SINGLE_PARTICLE
    use Images
    USE INTERPOLATION, ONLY : INTERPOLATION_SCHEME
    USE, INTRINSIC :: ISO_C_BINDING

    IMPLICIT NONE

    PRIVATE

    PUBLIC  ::  SP_MODEL_COMPARISON, SPMC_INIT, SPMC_PRINT_INFO, SPMC_EVAL, SPMC_GET_SP, SPMC_DS_MAXIMISE


    TYPE SP_MODEL_COMPARISON
        PRIVATE
        TYPE(SINGLE_PARTICLE)       ::  SP                                      !<  Single-particle object, describes the single particle parameters (angles, shifts, CTFs)
        TYPE(IMAGE),        POINTER ::  MODEL3D                                 !<  3D image of the model
        TYPE(IMAGE)                 ::  SP_IMAGE                                !<  Image of the single particle (before shifts are applied)
        REAL                        ::  PIXEL_SIZE                              !<  Pixel size (in Angstroms)
        ! We hold in memory a few intermediate steps to avoid recalculating them everytime we want to score a comparison
        LOGICAL                     ::  FIRST_USE               =   .TRUE.      !<  Whether intermediate images have already been computed. They still
                                                                                !!  may need to be recomputed even if this is set to .FALSE., for example
                                                                                !!  if we now need to score parameters which have changed.
        TYPE(SINGLE_PARTICLE)       ::  SP_READY                                !<  The single particle parameters (CTF, shifts, Euler angles) for which the intermediate images were calculated
        TYPE(IMAGE)                 ::  CTF_IMAGE_READY                         !<  Image of the CTF function, which will be applied to the image and the projection
        TYPE(IMAGE), PUBLIC         ::  PROJ_READY                              !<  Projection image ready for similarity measure
        TYPE(IMAGE), PUBLIC         ::  SP_IMAGE_READY                          !<  Single particle image ready for similarity measure
        TYPE(IMAGE)                 ::  ENV_MASK_READY                          !<  Envelope mask prepared from the projection
        ! The components below affect behaviour when optimising the SP/model comparison
        REAL                        ::  TARGET_RESOLUTION       =   0.0         !<  Target resolution (A) for the refinement
        REAL                        ::  FREQ_MAX                =   0.5         !<  Highest frequency (1/pixel units, 0.5 is Nyquist) to use when scoring using correlation functions.
        REAL                        ::  FREQ_MIN                =   0.0         !<  Lowest frequency (1/pixel units, 0.5 is Nyquist) to use when scoring using correlation functions.
        REAL                        ::  OBJ_MAX_RAD                             !<  Maximum (spherical) radius of the 3D object (pixels)
        REAL                        ::  OBJ_MAX_RAD_CYL         =   0.0         !<  Maximum (cylindrical) radius of the 3D object (pixels). Will be ignored if set to 0.0.
        LOGICAL                     ::  REFINE_PARAMS(7)        =   .FALSE.     !<  Which of the 6 degrees of freedom to refine (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi)
        REAL                        ::  PARAM_SCALE(7)          =   0.0         !<  For each parameter, a scaling factor to ensure that all parameters have comparable values
        REAL                        ::  SEARCH_ACCURACY(7)      =   0.0         !<  The desired accuracy of the search result (pixels, degrees).
        REAL                        ::  SEARCH_HALF_RANGE(7)    =   0.0         !<  The half range (degrees, pixels) of exhaustive search (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi)
        REAL                        ::  SEARCH_SAMPLING(7)      =   0.0         !<  The sampling step size (pixels, degrees) during exhaustive search (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi)
        LOGICAL                     ::  REAL_SPACE_PROJ         =   .FALSE.     !<  Whether to use real space projection. If .FALSE., reciprocal space slice extraction is used instead (which is faster).
        INTEGER                     ::  NUM_MINIM_RUNS          =   1           !<  The number of times the minimiser should be run. Only meant for dowhill simplex (DS), which can be started several times from the same position but with a random first step.
        LOGICAL                     ::  DS_RANDOM_START         =   .TRUE.      !<  Whether the starting setup of the DS should be randomised (see GSL documentation for gsl_multimin_fminimizer_nmsimplex2rand
                                                                                !!  at http://www.gnu.org/software/gsl/manual/html_node/Multimin-Algorithms-without-Derivatives.html)
        !
        TYPE(INTERPOLATION_SCHEME)  ::  INTERPOL                                !<  Interpolation scheme
        ! Work area memory
        TYPE(WINDOW)                ::  INTERPOL_WINDOW                         !<  Window image for interpolation
        TYPE(IMAGE)                 ::  SP_IMAGE_TMP
        !
        TYPE(C_PTR)                 ::  CPTR                                    !<  A C-pointer pointing to itself. This is needed when passing the object through the FGSL.
        !CONTAINS
        !    PROCEDURE,  PUBLIC      ::  INIT        =>  SPMC_INIT
        !    PROCEDURE,  PUBLIC      ::  PRINT_INFO  =>  SPMC_PRINT_INFO
        !    PROCEDURE,  PUBLIC      ::  EVAL        =>  SPMC_EVAL
        !    PROCEDURE,  PUBLIC      ::  GET_SP      =>  SPMC_GET_SP
        !    PROCEDURE,  PUBLIC      ::  DS_MAXIMISE =>  SPMC_DS_MAXIMISE
    END TYPE SP_MODEL_COMPARISON

    CONTAINS

    !>  \brief  Brute-force maximisation of the similarity measure between the single particle
    !!  image and the 3D mdel by varying the refinement paramenters
    SUBROUTINE SPMC_BF_MAXIMISE(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(SP_MODEL_COMPARISON),      INTENT(INOUT)   ::  SELF
        ! Private variables

        ! Start work
    END SUBROUTINE SPMC_BF_MAXIMISE

    !>  \brief  Return a single-particle object with the current state of the SP-model comparison
    FUNCTION SPMC_GET_SP(SELF) RESULT(SP)
        IMPLICIT NONE
        ! Arguments
        CLASS(SP_MODEL_COMPARISON),     INTENT(IN)  ::  SELF    !<  SP-model comparison object
        ! Result
        TYPE(SINGLE_PARTICLE)   ::  SP
        ! Private variables

        ! Start work
        SP = SELF%SP

    END FUNCTION SPMC_GET_SP

    !>  \brief  Dowhill simplex maximisation of the similarity measure between the single particle image and the 3D model by varying the refinement parameters
    SUBROUTINE SPMC_DS_MAXIMISE(SELF)
        USE MINIMISERS, ONLY : MINIMISER_DS_FGSL, SEARCH_ACCURACY_TO_DS_SIZE_RATIO, DS_INIT, DS_RUN, DS_SET_START_STEP
        USE SINGLE_PARTICLES
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        use images
        IMPLICIT NONE
        ! Arguments
        CLASS(SP_MODEL_COMPARISON),     INTENT(INOUT)   ::  SELF    !<  Single particle (SP) model comparison object
        ! Private variables
        REAL                        ::  DS_TARGET_SIZE
        REAL(KIND=8),   ALLOCATABLE ::  START_VALUES(:), FINAL_VALUES(:,:), BEST_FINAL_VALUES(:), FINAL_SCORE(:)
        REAL(KIND=8),   ALLOCATABLE ::  FIRST_STEP(:)
        INTEGER                     ::  I, IERR
        TYPE(MINIMISER_DS_FGSL)     ::  DS_MINIMISER    !<  A minimiser object (we use a FGSL minimiser - this is all we have as of Jan2010)
        INTEGER                     ::  NUM_DIM_TOTAL
        INTEGER                     ::  CURRENT_MINIM_RUN
        LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
        INTEGER                     ::  BEST_MINIM_RUN(1)
        REAL                        ::  INITIAL_SCORE
        TYPE(SINGLE_PARTICLE)       ::  SP_START
        CHARACTER(LEN=6)            ::  NUM_DIM_C
        ! Start work

        ! Do we have any work to do?
        IF (.NOT. ANY(SELF%REFINE_PARAMS)) THEN
            WRITE(*,'(A)') '**WARNING(SPMC_DS_MAXIMISE): nothing to do, since no parameters are to be refined'
            RETURN
        ENDIF

        ! If only the in-plane shifts are to be optimised, then the user should call a CCF-based routine
        IF (ALL(SELF%REFINE_PARAMS .EQV. [.TRUE.,.TRUE.,.FALSE.,.FALSE.,.FALSE.,.FALSE.,.FALSE.])) THEN
            WRITE(*,'(A)') '**ERROR(SPMC_DS_MAXIMISE): refinement of X and Y shifts only is done more ',    &
                            'efficiently using cross-correlation functions'
            CALL TERMINATE('Fatal error in SPMC_DS_MAXIMISE')
        ENDIF

        ! We may need to prepare the reference volume
        !$OMP   CRITICAL
        IF (.NOT. SELF%REAL_SPACE_PROJ) THEN
            IF (SELF%MODEL3D%is_in_real_space) THEN
                WRITE(*,'(A)') '**warning(SPMC_DS_MAXIMISE): FWD_FTing the volume'
                CALL FWD_FT(SELF%MODEL3D)
            ENDIF
            IF (SELF%MODEL3D%ORIGIN_AT_CENTER) THEN
                WRITE(*,'(A)') '**warning(SPMC_DS_MAXIMISE): shifting the real-space origin of the volume'
                CALL SHIFT_REAL_SPACE_ORIGIN(SELF%MODEL3D)
            ENDIF
        ENDIF
        !$OMP   END CRITICAL

        ! Work out a target size for the simplex
        DS_TARGET_SIZE = 0.0
        DO I=1,7
            IF (SELF%REFINE_PARAMS(I)) THEN
                DS_TARGET_SIZE = DS_TARGET_SIZE + (SELF%SEARCH_ACCURACY(I)/SELF%PARAM_SCALE(I))/SEARCH_ACCURACY_TO_DS_SIZE_RATIO
            ENDIF
        ENDDO
        DS_TARGET_SIZE = DS_TARGET_SIZE / COUNT(SELF%REFINE_PARAMS)
        !$OMP   SINGLE
        WRITE(*,'(A,F0.4)') '**info(SPMC_DS_MAXIMISE): target size for simplex = ', DS_TARGET_SIZE
        !$OMP   END SINGLE  NOWAIT

        ! Get the start values for each parameter
        CALL SPMC_GET_VALUES(SELF,START_VALUES)

        ! The first step is essentially the half range, scaled appropriately
        FIRST_STEP = SPMC_DS_GET_FIRST_STEP(SELF)

        ! Starting score?
        INITIAL_SCORE = SPMC_EVAL(SELF)

        ! Print out initial single particle state
        !$OMP   CRITICAL        (PRINT_OUT_STARTING_SP_DS_OMP_CRIT)
        SP_START = SELF%SP
        WRITE(*,'(A)',ADVANCE='NO') '**info(SPMC_DS_MAXIMISE): prior to downhill simplex minimisation, ',   &
                                    'the single particle parameters are: '
        CALL SP_PRINT_INFO(SELF%SP)
        WRITE(*,'(A,F0.5)') '**info(SPMC_DS_MAXIMISE): and the starting score is ', INITIAL_SCORE
        !$OMP   END CRITICAL    (PRINT_OUT_STARTING_SP_DS_OMP_CRIT)

        ! Initialise the DS minimiser
        CALL DS_INIT(DS_MINIMISER,SPMC_EVAL_WRAP,COUNT(SELF%REFINE_PARAMS),SELF%CPTR,   &
                    INITIAL_SCORE=INITIAL_SCORE, &
                    RANDOM=(SELF%DS_RANDOM_START .OR. (SELF%NUM_MINIM_RUNS .GT. 1)))

        ! Allocate memory for the results of the minimiser runs
        NUM_DIM_TOTAL = COUNT(SELF%REFINE_PARAMS)
        WRITE(NUM_DIM_C,'(I6)') NUM_DIM_TOTAL
        ALLOCATE(   FINAL_VALUES(NUM_DIM_TOTAL,SELF%NUM_MINIM_RUNS),    &
                    FINAL_SCORE(SELF%NUM_MINIM_RUNS),   &
                    BEST_FINAL_VALUES(NUM_DIM_TOTAL),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(SPMC_DS_MAXIMISE): memory allocation failed with error ', IERR
            CALL TERMINATE('Fatal error in SPMC_DS_MAXIMISE')
        ENDIF


        DO CURRENT_MINIM_RUN = 1, SELF%NUM_MINIM_RUNS

            ! Set the starting position and the first step for the DS
            CALL DS_SET_START_STEP(DS_MINIMISER,START_VALUES,FIRST_STEP)

            ! Run the DS minimiser
            CALL DS_RUN(DS_MINIMISER,DS_TARGET_SIZE)

            ! Get the last best values from the minimiser
            FINAL_VALUES(:,CURRENT_MINIM_RUN) = DS_MINIMISER%CURRENT_BEST_FGSL(:)
            FINAL_SCORE(CURRENT_MINIM_RUN) = DS_MINIMISER%CURRENT_MINIMUM_FGSL
            IF (DEBUG) THEN
                PRINT '(A,I0,A,I0,A,'//NUM_DIM_C//'(F0.3,1X))',  &
                        '**debug(SPMC_DS_MAXIMISE): final score for run ', CURRENT_MINIM_RUN, ' of ' , SELF%NUM_MINIM_RUNS, &
                        ' = ', FINAL_SCORE(CURRENT_MINIM_RUN)
                PRINT '(A,I0,A,I0,A,'//NUM_DIM_C//'(F0.3,1X))',  &
                        '**debug(SPMC_DS_MAXIMISE): final values for run ', CURRENT_MINIM_RUN, ' of ' , SELF%NUM_MINIM_RUNS, &
                        ' = ', FINAL_VALUES(:,CURRENT_MINIM_RUN)
            ENDIF
        ENDDO

        ! Decide which run of the minimiser gave the best results
        BEST_MINIM_RUN = MINLOC(FINAL_SCORE)
        WRITE(*,'(A,I0,A)') '**info(FMC_DS_MAXIMISE): minimisation run ', BEST_MINIM_RUN, ' gave the best results'
        BEST_FINAL_VALUES = FINAL_VALUES(:,BEST_MINIM_RUN(1))

        ! Plug those back into the the single particle
        CALL SPMC_SET_VALUES(SELF,BEST_FINAL_VALUES)
        SELF%SP%SCORE = -FINAL_SCORE(BEST_MINIM_RUN(1))

        ! Return the reference volume to its original state
        !!$OMP  CRITICAL
        !IF (.NOT. SELF%REAL_SPACE_PROJ) THEN
        !   IF (.NOT. SELF%MODEL3D%ORIGIN_AT_CENTER) CALL SHIFT_REAL_SPACE_ORIGIN(SELF%MODEL3D)
        !   IF (SELF%MODEL3D%TYPE .NE. 'REAL') CALL BWD_FT(SELF%MODEL3D)
        !ENDIF
        !!$OMP  END CRITICAL

        ! Print out final single particle step
        !$OMP   CRITICAL        (PRINT_OUT_FINAL_SP_DS)
        WRITE(*,'(A)',ADVANCE='NO') '**info(SPMC_DS_MAXIMISE): all done. Following downhill simplex minimisation, ',    &
                                    'the single particle parameters are: '
        CALL SP_PRINT_INFO(SELF%SP)
        SP_START = SELF%SP - SP_START
        WRITE(*,'(A)',ADVANCE='NO') '**info(SPMC_DS_MAXIMISE): The single-particle parameters changed by this much: '
        CALL SP_PRINT_INFO(SP_START)
        !$OMP   END CRITICAL    (PRINT_OUT_FINAL_SP_DS)
    END SUBROUTINE SPMC_DS_MAXIMISE

    !>  \brief  Return the first step for the DS minimiser
    FUNCTION SPMC_DS_GET_FIRST_STEP(SELF) RESULT(FIRST_STEP)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(SP_MODEL_COMPARISON),  INTENT(IN)  ::  SELF
        ! Result
        REAL(KIND=8),   ALLOCATABLE ::  FIRST_STEP(:)
        ! Private variables
        INTEGER ::  CURRENT_PARAM
        INTEGER ::  CURRENT_VALUE
        INTEGER ::  IERR
        ! Start work

        ! Allocate memory
        ALLOCATE(FIRST_STEP(COUNT(SELF%REFINE_PARAMS)),STAT=IERR)
        IF (IERR .NE. 0) CALL TERMINATE('Memory allocation failed in SPMC_DS_GET_FIRST_STEP')

        !
        CURRENT_VALUE = 0
        DO CURRENT_PARAM = 1,7
            IF (SELF%REFINE_PARAMS(CURRENT_PARAM)) THEN
                CURRENT_VALUE = CURRENT_VALUE + 1
                FIRST_STEP(CURRENT_VALUE) = SELF%SEARCH_HALF_RANGE(CURRENT_PARAM) / SELF%PARAM_SCALE(CURRENT_PARAM)
            ENDIF
        ENDDO

        ! Sanity
        IF (CURRENT_VALUE .NE. COUNT(SELF%REFINE_PARAMS)) THEN
            CALL TERMINATE('Failed sanity check in SPMC_DS_GET_FIRST_STEP')
        ENDIF
    END FUNCTION SPMC_DS_GET_FIRST_STEP

    !>  \brief  Wrapper to be called by the FGSL DS minimiser
    !!  The values and parameters are passed as C pointers so that this subroutine can be called by the (F)GSL.
    FUNCTION SPMC_EVAL_WRAP(VALUES,PARAMS) BIND(C)
        USE, INTRINSIC :: ISO_C_BINDING
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE FGSL, ONLY : FGSL_DOUBLE, FGSL_VECTOR, FGSL_OBJ_C_PTR, FGSL_VECTOR_ALIGN, FGSL_INT, FGSL_SUCCESS
        IMPLICIT NONE
        ! Arguments
        TYPE(C_PTR),    VALUE   ::  VALUES              !<  C pointer to current values to try out
        TYPE(C_PTR),    VALUE   ::  PARAMS              !<  C pointer to a SPMC object
        ! Result
        REAL(KIND=C_DOUBLE)     ::  SPMC_EVAL_WRAP
        ! Private variables
        TYPE(SP_MODEL_COMPARISON),  POINTER     ::  SPMC
        REAL(FGSL_DOUBLE),          POINTER     ::  VAL_FGSL(:)
        REAL(KIND=8),               ALLOCATABLE ::  VAL(:)
        TYPE(FGSL_VECTOR)                       ::  VAL_VEC
        INTEGER                                 ::  EXPECTED_NUM_VALUES
        INTEGER(FGSL_INT)                       ::  STATUS
        INTEGER                                 ::  IERR
        ! Start work

        ! Unpack the SPMC object
        CALL C_F_POINTER(PARAMS,SPMC)

        ! Work out how many dimensions we're minimising
        EXPECTED_NUM_VALUES = COUNT(SPMC%REFINE_PARAMS)

        ! Unpack the values we are going to try
        CALL FGSL_OBJ_C_PTR(VAL_VEC,VALUES)
        STATUS = FGSL_VECTOR_ALIGN(VAL_FGSL,VAL_VEC)
        IF (STATUS .NE. FGSL_SUCCESS) CALL TERMINATE('FGSL_VECTOR_ALIGN failed in SPMC_EVAL_WRAP')

        ! Check we got the expected number of variables
        IF (SIZE(VAL_FGSL) .NE. EXPECTED_NUM_VALUES) THEN
            WRITE(*,'(A,2I5)') '**ERROR(SPMC_EVAL_WRAP): did not get expected number of variables ', &
                                SIZE(VAL_FGSL), EXPECTED_NUM_VALUES
            CALL TERMINATE('Fatal error in SPMC_EVAL_WRAP')
        ENDIF

        ! Convert from FGSL_DOUBLE to REAL(KIND=8)
        ALLOCATE(VAL(EXPECTED_NUM_VALUES),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(SPMC_EVAL_WRAP): allocation of real values array failed with error', IERR
            CALL TERMINATE('Fatal error in SPMC_EVAL_WRAP')
        ENDIF
        VAL = VAL_FGSL

        ! Modify the single particle object to have the desired values
        CALL SPMC_SET_VALUES(SPMC,VAL)

        ! Now call the actual scoring function
        SPMC_EVAL_WRAP = - SPMC_EVAL(SPMC)

    END FUNCTION SPMC_EVAL_WRAP

    !>  \brief  Change the values of SP parameters in a SPMC object
    SUBROUTINE SPMC_SET_VALUES(SELF,VALUES)
        USE CTFS
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        CLASS(SP_MODEL_COMPARISON),                 INTENT(INOUT)   ::  SELF
        REAL(KIND=8),               ALLOCATABLE,    INTENT(IN)      ::  VALUES(:)
        ! Private variables
        INTEGER ::  CURRENT_PARAM, CURRENT_VALUE
        ! Start work
        IF (SIZE(VALUES) .NE. COUNT(SELF%REFINE_PARAMS)) THEN
            WRITE(*,'(A)') '**ERROR(SPMC_SET_VALUES): number of values and number of refinement parameters ', &
                            SIZE(VALUES), COUNT(SELF%REFINE_PARAMS)
            CALL TERMINATE('Fatal error in SPMC_SET_VALUES')
        ENDIF

        ! Loop over parameters
        CURRENT_VALUE = 0
        DO CURRENT_PARAM = 1,7
            IF (SELF%REFINE_PARAMS(CURRENT_PARAM)) THEN
                CURRENT_VALUE = CURRENT_VALUE + 1
                SELECT CASE (CURRENT_PARAM)
                    CASE(1)
                        SELF%SP%SHIFT_X         =   VALUES(CURRENT_VALUE) * SELF%PARAM_SCALE(CURRENT_PARAM)
                    CASE(2)
                        SELF%SP%SHIFT_Y         =   VALUES(CURRENT_VALUE) * SELF%PARAM_SCALE(CURRENT_PARAM)
                    CASE(3)
                        CALL CTF_AT_Z(SELF%SP%CTF,VALUES(CURRENT_VALUE) * SELF%PARAM_SCALE(CURRENT_PARAM))
                    CASE(4)
                        SELF%SP%ANGLES%PHI      =   VALUES(CURRENT_VALUE) * SELF%PARAM_SCALE(CURRENT_PARAM)
                    CASE(5)
                        SELF%SP%ANGLES%THETA    =   VALUES(CURRENT_VALUE) * SELF%PARAM_SCALE(CURRENT_PARAM)
                    CASE(6)
                        SELF%SP%ANGLES%PSI      =   VALUES(CURRENT_VALUE) * SELF%PARAM_SCALE(CURRENT_PARAM)
                    CASE(7)
                        SELF%SP%SUBUNIT_NUM     =   VALUES(CURRENT_VALUE) * SELF%PARAM_SCALE(CURRENT_PARAM)
                END SELECT
            ENDIF
        END DO

        ! Sanity check
        IF (CURRENT_VALUE .NE. SIZE(VALUES)) THEN
            WRITE(*,'(A)') '**ERROR(SPMC_SET_VALUES): likely programming error'
            CALL TERMINATE('Fatal error in SPMC_SET_VALUES')
        ENDIF
    END SUBROUTINE SPMC_SET_VALUES

    !>  \brief  Read out the values of PS parameters in a SPMC object
    SUBROUTINE SPMC_GET_VALUES(SELF,VALUES)
        USE CTFS
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        CLASS(SP_MODEL_COMPARISON),                     INTENT(INOUT)   ::  SELF
        REAL(KIND=8),                   ALLOCATABLE,    INTENT(OUT)     ::  VALUES(:)
        ! Private variables
        INTEGER ::  CURRENT_PARAM, CURRENT_VALUE
        INTEGER ::  IERR
        ! Start work

        ! Allocate memory to the values array
        ALLOCATE(VALUES(COUNT(SELF%REFINE_PARAMS)),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(SPMC_GET_VALUES): memory allocation failed with IERR = ', IERR
            CALL TERMINATE('Fatal error in SPMC_GET_VALUES')
        ENDIF

        ! Loop over parameters
        CURRENT_VALUE = 0
        DO CURRENT_PARAM = 1,7
            IF (SELF%REFINE_PARAMS(CURRENT_PARAM)) THEN
                CURRENT_VALUE = CURRENT_VALUE + 1
                SELECT CASE(CURRENT_PARAM)
                    CASE(1)
                        VALUES(CURRENT_VALUE)   =   SELF%SP%SHIFT_X
                    CASE(2)
                        VALUES(CURRENT_VALUE)   =   SELF%SP%SHIFT_Y
                    CASE(3)
                        CALL CTF_ANG_TO_PIX(SELF%SP%CTF,PIXEL_SIZE=SELF%PIXEL_SIZE)
                        VALUES(CURRENT_VALUE)   =   Z_FROM_CTF(SELF%SP%CTF)
                    CASE(4)
                        VALUES(CURRENT_VALUE)   =   SELF%SP%ANGLES%PHI
                    CASE(5)
                        VALUES(CURRENT_VALUE)   =   SELF%SP%ANGLES%THETA
                    CASE(6)
                        VALUES(CURRENT_VALUE)   =   SELF%SP%ANGLES%PSI
                    CASE(7)
                        VALUES(CURRENT_VALUE)   =   SELF%SP%SUBUNIT_NUM
                END SELECT
                VALUES(CURRENT_VALUE) = VALUES(CURRENT_VALUE) / SELF%PARAM_SCALE(CURRENT_PARAM)
            ENDIF
        ENDDO

        ! Sanity check
        IF (CURRENT_VALUE .NE. SIZE(VALUES)) THEN
            WRITE(*,'(A)') '**ERROR(SPMC_GET_VALUES): likely programming error'
            CALL TERMINATE('Fatal error in SPMC_GET_VALUES')
        ENDIF

    END SUBROUTINE SPMC_GET_VALUES


    !>  \brief  Score the current state of the single particle model comparison object. Essentially,
    !!          this involves comparing the FT of the single particle image with the corresponding
    !!          central section of the model's FT
    REAL FUNCTION SPMC_EVAL(SELF)
        USE IMAGES_COMPARISON
        USE IMAGES_CTF
        USE EULERANGLES
        USE SINGLE_PARTICLES, ONLY : SINGLE_PARTICLE, SP_PRINT_INFO
        USE, INTRINSIC :: IEEE_ARITHMETIC, ONLY : IEEE_IS_NAN
        IMPLICIT NONE
        ! Arguments
        CLASS(SP_MODEL_COMPARISON),     INTENT(INOUT)   ::  SELF
        ! Private variables
        LOGICAL                 ::  NEW_PARAMS(7)       !<  For each parameter (1-X, 2-Y, 3-Z/defocus/CTF, 4-Phi,
                                                        !!  5-Theta,6.Psi,7.SubunitNum), whether we are dealing
                                                        !!  with new values compared to the prepared intermediate
                                                        !!  steps already in memory
        LOGICAL,    PARAMETER   ::  DEBUG           =   .FALSE.
        REAL(KIND=8)            ::  SUBUNIT_NUM_SHIFT, PHI_SHIFT, Z_SHIFT
        TYPE(EULER)             ::  ANGLES
        ! Start work


        ! If we're doing slice extraction (rather than real space projection), check the 3D model is in Fourier space, with real-space origin at the corner
        !$OMP   CRITICAL
        IF (SELF%MODEL3D%is_in_real_space) THEN
            WRITE(*,'(2A)') '**warning(SPMC_EVAL): model 3D was not in Fourier space - apply FT. ',  &
                            'Do this externally to save time if possible.'
            CALL FWD_FT(SELF%MODEL3D)
        ENDIF
        IF (SELF%MODEL3D%ORIGIN_AT_CENTER) THEN
            WRITE(*,'(2A)') '**warning(SPMC_EVAL): model 3D had real-space origin at center - apply phase shift. ',  &
                            'Do this externally to save time if possible.'
            CALL SHIFT_REAL_SPACE_ORIGIN(SELF%MODEL3D)
        ENDIF
        !$OMP   END CRITICAL


        !
        ! Check what parameters have changed
        !
        IF (SELF%FIRST_USE) THEN
            ! All parameters are new
            NEW_PARAMS = .TRUE.
            SELF%FIRST_USE = .FALSE.
        ELSE
            NEW_PARAMS = .FALSE.
            IF (SELF%SP%ANGLES .NE. SELF%SP_READY%ANGLES) THEN
                NEW_PARAMS(3:6) = .TRUE.
            ENDIF
            IF (SELF%SP%CTF .NE. SELF%SP_READY%CTF) THEN
                NEW_PARAMS(3) = .TRUE.
            ENDIF
            IF (SELF%SP%SHIFT_X .NE. SELF%SP_READY%SHIFT_X .OR. SELF%SP%SHIFT_Y .NE. SELF%SP_READY%SHIFT_Y) THEN
                NEW_PARAMS(1:2) = .TRUE.
            ENDIF
            IF (SELF%SP%SUBUNIT_NUM .NE. SELF%SP_READY%SUBUNIT_NUM) THEN
                NEW_PARAMS(7) = .TRUE.
            ENDIF
        ENDIF

        !
        ! Depending on which parameters have changed, we may need to recalculate intermediate images
        !
        IF (ANY(NEW_PARAMS)) THEN
            IF (NEW_PARAMS(3)) THEN
                ! We need to (re)calcuate the CTF IMAGE
                IF (DEBUG) WRITE(*,'(A)') '**debug(SPMC_EVAL): recomputing CTF image'
                SELF%CTF_IMAGE_READY%logical_dimensions = SELF%SP_IMAGE%logical_dimensions
                CALL CTF_TO_IMG(SELF%SP%CTF,SELF%CTF_IMAGE_READY,PIXEL_SIZE=SELF%PIXEL_SIZE)
            ENDIF
            IF (ANY(NEW_PARAMS(1:2)) .OR. ANY(NEW_PARAMS(3:6)) .OR. NEW_PARAMS(7)) THEN
                ! We need to compute a new projection and a new envelope mask
                IF (DEBUG) WRITE(*,'(A)') '**debug(SPMC_EVAL): recomputing projection and envelope mask'
                ! Work out any changes due to the non-integer value of subunit_num
                ! (we may have to change phi and z)
                SUBUNIT_NUM_SHIFT = SELF%SP%SUBUNIT_NUM - NINT(SELF%SP%SUBUNIT_NUM)
                PHI_SHIFT = - SUBUNIT_NUM_SHIFT * SELF%SP%HELIX%TWIST
                Z_SHIFT = - SUBUNIT_NUM_SHIFT * SELF%SP%HELIX%RISE
                ANGLES = SELF%SP%ANGLES
                ANGLES%PHI = ANGLES%PHI + PHI_SHIFT
                ! Compute the projection
                CALL PROJECTION_PREPARE(    SELF%MODEL3D,SELF%PROJ_READY,SELF%SP_IMAGE%logical_dimensions,&
                                            ANGLES,SELF%CTF_IMAGE_READY,  &
                                            PROJ_MAX_RAD=SELF%OBJ_MAX_RAD,  &
                                            SHIFT_X=SELF%SP%SHIFT_X,SHIFT_Y=SELF%SP%SHIFT_Y,  &
                                            INTERPOL=SELF%INTERPOL,WIN=SELF%INTERPOL_WINDOW,  &
                                            ENV_MASK=SELF%ENV_MASK_READY,REAL_SPACE_PROJ=SELF%REAL_SPACE_PROJ,&
                                            VOL_SHIFT_Z = Z_SHIFT,RECT_MASK_WIDTH=2.*SELF%OBJ_MAX_RAD_CYL)
            ENDIF
            IF (NEW_PARAMS(3)) THEN
                ! We need to (re) apply the CTF to the image, and remask it with the envelope mask
                IF (DEBUG) WRITE(*,'(A)') '**debug(SPMC_EVAL): apply CTF & mask to the sp image'
                SELF%SP_IMAGE_READY = SELF%SP_IMAGE
                CALL IMAGE_PREPARE( SELF%SP_IMAGE_READY,SELF%SP_IMAGE_TMP,SELF%CTF_IMAGE_READY,SELF%OBJ_MAX_RAD,    &
                                    ENV_MASK=SELF%ENV_MASK_READY)
            ENDIF
        ELSE
            WRITE(*,'(A)') '**warning(SPMC_EVAL): evaluating something that was already evaluated - no parameters changed!'
        ENDIF

        ! Now that we've recalculated intermediate steps, remember we've done so
        SELF%SP_READY = SELF%SP

        ! Debug dump
        !IF (DEBUG) THEN
        !   IF (SELF%SP_IMAGE_READY%TYPE .EQ. 'COMP') CALL BWD_FT(SELF%SP_IMAGE_READY)
        !   IF (SELF%PROJ_READY%TYPE .EQ. 'COMP') CALL BWD_FT(SELF%PROJ_READY)
        !   CALL WRITE_TO_DISK(SELF%SP_IMAGE_READY,'dbg_sp.mrc')
        !   CALL WRITE_TO_DISK(SELF%PROJ_READY,'dbg_proj.mrc')
        !   CALL FWD_FT(SELF%SP_IMAGE_READY)
        !   CALL FWD_FT(SELF%PROJ_READY)
        !   !CALL TERMINATE('Testing SPMC_EVAL')
        !ENDIF

        !
        ! Now all we have to do is measure the similarity between PROJ_READY and SP_IMAGE_READY
        !
        !IF (DEBUG) WRITE(*,'(A,I0)') '**debug(SPMC_EVAL): about to call CC3 with RAD_MAX = ', NINT(SELF%SP_IMAGE_READY%logical_dimensions(1)*SELF%PIXEL_SIZE/SELF%TARGET_RESOLUTION)
        IF (IS_REAL(SELF%SP_IMAGE_READY)) CALL FWD_FT(SELF%SP_IMAGE_READY)
        IF (IS_REAL(SELF%PROJ_READY)) CALL FWD_FT(SELF%PROJ_READY)
        SPMC_EVAL = CCLIN(  SELF%SP_IMAGE_READY,SELF%PROJ_READY,  &
                            RAD_MAX=REAL(SELF%SP_IMAGE_READY%logical_dimensions(1))*SELF%FREQ_MAX,   &
                            RAD_MIN=REAL(SELF%SP_IMAGE_READY%logical_dimensions(1))*SELF%FREQ_MIN)

        SELF%SP%SCORE = SPMC_EVAL
        SELF%SP_READY%SCORE = SPMC_EVAL
        ! Debug
        IF (DEBUG) THEN
            WRITE(*,'(A,F0.5,A)',ADVANCE='NO') '**debug(SPMC_EVAL): score of ', SPMC_EVAL, &
                                                ' for the following single-particle parameters: '
            CALL SP_PRINT_INFO(SELF%SP)
            IF (IEEE_IS_NAN(SPMC_EVAL)) THEN
                CALL TERMINATE('Fatal error in SPMC_EVAL: returning NaN')
            ENDIF
            !CALL TERMINATE('Testing SPMC_EVAL')
        ENDIF
    END FUNCTION SPMC_EVAL

    !>  \brief  Deallocate work memory for the FMC object
    SUBROUTINE SPMC_DEALLOCATE_WORK_MEMORY(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(SP_MODEL_COMPARISON),    INTENT(INOUT)   ::  SELF
        ! Private variables

        ! Start work
        CALL DEALLOCATE_WINDOW(SELF%INTERPOL_WINDOW)
        CALL IMAGE_DEALLOCATE(SELF%SP_IMAGE_TMP)
        CALL IMAGE_DEALLOCATE(SELF%PROJ_READY)
        CALL IMAGE_DEALLOCATE(SELF%CTF_IMAGE_READY)
        CALL IMAGE_DEALLOCATE(SELF%SP_IMAGE_READY)
        CALL IMAGE_DEALLOCATE(SELF%ENV_MASK_READY)
    END SUBROUTINE SPMC_DEALLOCATE_WORK_MEMORY

    !>  \brief  Allocate work memory for the FMC object
    SUBROUTINE SPMC_ALLOCATE_WORK_MEMORY(SELF)
        use images
        IMPLICIT NONE
        ! Arguments
        TYPE(SP_MODEL_COMPARISON),    INTENT(INOUT)   ::  SELF
        ! Private variables

        ! Start work
        CALL SPMC_DEALLOCATE_WORK_MEMORY(SELF)
        CALL SELF%INTERPOL_WINDOW%AllocateWindow(SELF%INTERPOL%WINDOW_WIDTH,COMP=.NOT. SELF%REAL_SPACE_PROJ,THREED=.TRUE.)
        CALL IMAGE_ALLOCATE(SELF%SP_IMAGE_TMP,SELF%SP_IMAGE)
        CALL IMAGE_ALLOCATE(SELF%PROJ_READY,SELF%SP_IMAGE)
        CALL IMAGE_ALLOCATE(SELF%CTF_IMAGE_READY,SELF%SP_IMAGE)
        CALL IMAGE_ALLOCATE(SELF%SP_IMAGE_READY,SELF%SP_IMAGE)
        CALL IMAGE_ALLOCATE(SELF%ENV_MASK_READY,SELF%SP_IMAGE)
    END SUBROUTINE SPMC_ALLOCATE_WORK_MEMORY

    !>  \brief  Initialise a single particle model comparison object
    SUBROUTINE SPMC_INIT(   SELF,SP,MODEL3D,SP_IMAGE,REFINE_PARAMS,PIXEL_SIZE, &
                            TARGET_RESOLUTION,OBJ_MAX_RAD,INTERPOL,REAL_SPACE_PROJ, &
                            SEARCH_HALF_RANGE_OVERRIDE,NUM_MINIM_RUNS,DS_RANDOM_START, &
                            LOW_RESOLUTION_CUTON,HIGH_RESOLUTION_CUTOFF,OBJ_MAX_RAD_CYL)
        USE IMAGES_COMPARISON
        USE CTFS, ONLY : CTF_PIX_TO_ANG
        USE,    INTRINSIC   :: ISO_C_BINDING
        IMPLICIT NONE
        ! Arguments
        TYPE(SP_MODEL_COMPARISON),  TARGET,     INTENT(INOUT)   ::  SELF                            !<  Single particle - model comparison object to be initialised
        TYPE(SINGLE_PARTICLE),                  INTENT(IN)      ::  SP                              !<  Single particle object
        TYPE(IMAGE),                TARGET,     INTENT(IN)      ::  MODEL3D                         !<  3D model to which the 2D image will be compared
        TYPE(IMAGE),                            INTENT(IN)      ::  SP_IMAGE                        !<  2D projection image of the single particle (presumably boxed out from a noisy micrograph)
        LOGICAL,                                INTENT(IN)      ::  REFINE_PARAMS(7)                !<  Refine parameters (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi)
        REAL,                                   INTENT(IN)      ::  PIXEL_SIZE                      !<  Pixel size (Angstroms)
        REAL,                                   INTENT(IN)      ::  TARGET_RESOLUTION               !<  Target resolution (Angstroms)
        REAL,                                   INTENT(IN)      ::  OBJ_MAX_RAD                     !<  Maximum radius of the object (pixels)
        TYPE(INTERPOLATION_SCHEME),             INTENT(IN)      ::  INTERPOL                        !<  Interpolation scheme
        LOGICAL,                    OPTIONAL,   INTENT(IN)      ::  REAL_SPACE_PROJ                 !<  Whether to use real space projection. If .FALSE., reciprocal space slice extraction is used instead (which is faster). Defaults to .FALSE.
        REAL,                       OPTIONAL,   INTENT(IN)      ::  SEARCH_HALF_RANGE_OVERRIDE(7)   !<  For each refinement dimension (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi), 0.0 if default search half-range
                                                                                                    !!  should be used, or other value to override the default.
        INTEGER,                    OPTIONAL,   INTENT(IN)      ::  NUM_MINIM_RUNS                  !<  The number of times the minimiser should be run. Only meant for dowhill simplex (DS), which can be started several times
                                                                                                    !!  from the same position but with a random first step.
        LOGICAL,                    OPTIONAL,   INTENT(IN)      ::  DS_RANDOM_START                 !<  Whether the starting setup of the DS should be randomised (see GSL documentation for gsl_multimin_fminimizer_nmsimplex2rand
                                                                                                    !!  at http://www.gnu.org/software/gsl/manual/html_node/Multimin-Algorithms-without-Derivatives.html)
        REAL,                       OPTIONAL,   INTENT(IN)      ::  LOW_RESOLUTION_CUTON            !<  Lowest resolution (in Angstroms) to be used by scoring functions
        REAL,                       OPTIONAL,   INTENT(IN)      ::  HIGH_RESOLUTION_CUTOFF          !<  Highest resolution (in Angstroms) to be used by scoring functions
        REAL,                       OPTIONAL,   INTENT(IN)      ::  OBJ_MAX_RAD_CYL                 !<  Maximum (cylindrical) radius of the 3D object (pixels). Useful refining images of filaments.
        ! Private variables
        LOGICAL ::  BAD_DIMENSIONS
        ! Start work

        ! The 3D model and the single particle image are expected to have the same dimensions
        BAD_DIMENSIONS = .FALSE.
        IF (.NOT. IS_SQUARE(MODEL3D) .OR. .NOT. IS_SQUARE(SP_IMAGE)) THEN
            BAD_DIMENSIONS = .TRUE.
        ENDIF
        IF (MODEL3D%logical_dimensions(1) .NE. SP_IMAGE%logical_dimensions(1)) THEN
            BAD_DIMENSIONS = .TRUE.
        ENDIF
        IF (BAD_DIMENSIONS) THEN
            WRITE(*,'(A)') '**ERROR(SPMC_INIT): image and 3D model need to be square and of same dimensions, but they are not.'
            WRITE(*,'(A,3(I5,1X))') '**ERROR(SPMC_INIT): image dimensions = ', SP_IMAGE%logical_dimensions
            WRITE(*,'(A,3(I5,1X))') '**ERROR(SPMC_INIT): model dimensions = ', MODEL3D%logical_dimensions
            CALL TERMINATE('Fatal error in SPMC_INIT')
        ENDIF

        SELF%SP                     =   SP
        SELF%MODEL3D                =>  MODEL3D
        SELF%SP_IMAGE               =   SP_IMAGE
        SELF%REFINE_PARAMS          =   REFINE_PARAMS
        SELF%PIXEL_SIZE             =   PIXEL_SIZE
        SELF%OBJ_MAX_RAD            =   OBJ_MAX_RAD
        SELF%FIRST_USE              =   .TRUE.
        ! Min and max resolutions
        SELF%FREQ_MAX               =   0.5
        SELF%FREQ_MIN               =   0.0
        IF (PRESENT(LOW_RESOLUTION_CUTON))      SELF%FREQ_MIN = PIXEL_SIZE / LOW_RESOLUTION_CUTON
        IF (PRESENT(HIGH_RESOLUTION_CUTOFF))    SELF%FREQ_MAX = PIXEL_SIZE / HIGH_RESOLUTION_CUTOFF
        ! Target resolution
        SELF%TARGET_RESOLUTION      =   TARGET_RESOLUTION

        ! Work out the desired accuracy of the search, as well as the search half-range and sampling (we need to the wavelength in angstroms)
        IF (SELF%SP%CTF%PIX_UNITS) CALL CTF_PIX_TO_ANG(SELF%SP%CTF,SELF%PIXEL_SIZE)
        SELF%SEARCH_ACCURACY(1:6)   =   ACCURACY_FROM_RESOLUTION(   SELF%TARGET_RESOLUTION,SELF%OBJ_MAX_RAD*SELF%PIXEL_SIZE,   &
                                                                    SELF%MODEL3D%logical_dimensions(1),    &
                                                                    SELF%SP%CTF%WL,SELF%PIXEL_SIZE,AKV_IS_WL=.TRUE.)
        PRINT '(A,F0.5,A)', '**warning(SPMC_INIT): hard setting of subunit_num search accuracy at 0.1 (was ', &
                            SELF%SEARCH_ACCURACY(7), ')'
        SELF%SEARCH_ACCURACY(7)     =   0.1
        SELF%SEARCH_HALF_RANGE      =   SELF%SEARCH_ACCURACY * 2.
        SELF%SEARCH_SAMPLING        =   SELF%SEARCH_ACCURACY / 4.

        ! If the user supplied overrides, apply them
        IF (PRESENT(SEARCH_HALF_RANGE_OVERRIDE)) THEN
            WHERE (SEARCH_HALF_RANGE_OVERRIDE .NE. 0.0)
                SELF%SEARCH_HALF_RANGE = SEARCH_HALF_RANGE_OVERRIDE
            END WHERE
        ENDIF

        ! For subunit_num, it doesn't make sense to look beyond +- 0.5 subunits
        !SELF%SEARCH_HALF_RANGE(7)   =   MIN(SELF%SEARCH_HALF_RANGE(7),0.5)
        WRITE(*,'(A)') '**warning(SPMC_INIT): setting half-range for subunit_num to 0.5'
        SELF%SEARCH_HALF_RANGE(7)   =   0.5

        ! Work out the scaling of the various parameters
        SELF%PARAM_SCALE(1:6)       =   PARAM_SCALING_FROM_ACCURACY(SELF%SEARCH_ACCURACY(1:6),SELF%OBJ_MAX_RAD)
        WRITE(*,'(A)') '**warning(SPMC_INIT): setting scale for subunit_num to 1.0'
        SELF%PARAM_SCALE(7)         =   1.0

        ! Interpolation scheme
        SELF%INTERPOL               =   INTERPOL

        ! Real space projection?
        IF (PRESENT(REAL_SPACE_PROJ)) THEN
            SELF%REAL_SPACE_PROJ    =   REAL_SPACE_PROJ
        ELSE
            SELF%REAL_SPACE_PROJ    =   .FALSE.
        ENDIF

        ! Deal with DS options for random & multiple starting conditions
        IF (PRESENT(NUM_MINIM_RUNS)) SELF%NUM_MINIM_RUNS = NUM_MINIM_RUNS
        IF (PRESENT(DS_RANDOM_START)) SELF%DS_RANDOM_START = DS_RANDOM_START

        !
        IF (PRESENT(OBJ_MAX_RAD_CYL)) SELF%OBJ_MAX_RAD_CYL = OBJ_MAX_RAD_CYL

        ! Allocate work area memory
        CALL SPMC_ALLOCATE_WORK_MEMORY(SELF)

        ! Set up the C pointer to itself
        SELF%CPTR                   =   C_LOC(SELF)
    END SUBROUTINE SPMC_INIT

    !>  \brief  Print out information describing the single particle model comparison object
    SUBROUTINE SPMC_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        CLASS(SP_MODEL_COMPARISON), INTENT(IN)  ::  SELF        !<  Single particle - model comparison object
        ! Private variables
        CHARACTER(LEN=2),    PARAMETER   ::  TEXT_WIDTH  =   '35'
        ! Start work

        !$OMP   CRITICAL        (PRINT_OUT_SPMC_INFO)

        ! Print header line
        WRITE(*,'(/A)') 'Single-particle-model comparison parameters'
        ! Print information
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F7.3)')       'Pixel size (Angstroems)',          SELF%PIXEL_SIZE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F7.3)')       'Target resolution (A)',            SELF%TARGET_RESOLUTION
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F7.3)')       'Minimum frequency (1/pixel)',      SELF%FREQ_MIN
        WRITE(*,'(A'//TEXT_WIDTH//',1X,F7.3)')       'Maximum frequency (1/pixel)',      SELF%FREQ_MAX
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')         'Real-space projection?',           SELF%REAL_SPACE_PROJ
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(A8,1X))')   'Search paramameters',              'X','Y','Z/def','Phi','Theta','Psi',&
                                                                                       'Subunit#'
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(L8,1X))')   'Parameter to be refined?',         SELF%REFINE_PARAMS
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(F8.3,1X))') 'Search accuracy',                  SELF%SEARCH_ACCURACY
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(F8.3,1X))') 'Search half-range',                SELF%SEARCH_HALF_RANGE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(F8.3,1X))') 'Search sampling',                  SELF%SEARCH_SAMPLING
        WRITE(*,'(A'//TEXT_WIDTH//',1X,7(F8.3,1X))') 'Parameter scales',                 SELF%PARAM_SCALE
        WRITE(*,'(A'//TEXT_WIDTH//',1X,L1)')         'Randomised DS start?',             SELF%DS_RANDOM_START
        WRITE(*,'(A'//TEXT_WIDTH//',1X,I0)')         'Number of (DS) minimisation runs', SELF%NUM_MINIM_RUNS
        IF (SELF%OBJ_MAX_RAD_CYL .NE. 0.0) THEN
            WRITE(*,'(A'//TEXT_WIDTH//',1X,F7.3)')       'Maximum (cylindrical) radius (pixel)',      SELF%OBJ_MAX_RAD_CYL
        ENDIF

        ! Add blank line
        WRITE(*,'(A)') ' '

        !$OMP   END CRITICAL    (PRINT_OUT_SPMC_INFO)
    END SUBROUTINE SPMC_PRINT_INFO


END MODULE SP_MODEL_COMPARISONS


