!>  \brief  Module gathering spline-fitting / interpolating routines
!!
!! For details, see file "Splines.pdf" in the documents folder.
!!  \todo   Allocate fgsl objects necessary for solving splines directly within the spline object so that we don't have allocation and deallocation everytime we solve splines (also, spline solving can then become pure subroutine)
MODULE SPLINES

    USE FGSL

    ! Spline object based on the GNU scientific library (GSL) and its Fortran interface FGSL
    TYPE GSL_SPLINE
        PRIVATE
        TYPE(FGSL_INTERP_ACCEL)                 ::  ACC
        TYPE(FGSL_SPLINE)                       ::  SPLINE
        INTEGER                                 ::  STATUS
        INTEGER                                 ::  NUM_POINTS
        REAL(FGSL_DOUBLE), ALLOCATABLE, PUBLIC  ::  X(:),Y(:)
    END TYPE GSL_SPLINE

    ! Homegrown spline object for interpolation, built following
    ! http://online.redwoods.cc.ca.us/instruct/darnold/laproj/Fall98/SkyMeg/Proj.PDF
    TYPE INTERPOL_SPLINE
        INTEGER ::  N       ! number of points
        INTEGER ::  K   = 3 ! degree of spline (assumed to be 3, for cubic)
        REAL(KIND=8), ALLOCATABLE, PUBLIC   ::  X(:), Y(:), DY(:), DDY(:) ! ordinates, values and first and seconds derivatives
        REAL(KIND=8), ALLOCATABLE           ::  W(:,:)  ! weights - first dim is point number, second is degree of the weight. So a cubic polynomial is Y(:) = W(:,1)*X(:)**3 + W(:,2)*X(:)**2 + W(:,3)*X(:) + W(:,4)
        LOGICAL ::  NATURAL ! Spline is natural, that is DDY(1) = DDY(N) = 0
        LOGICAL ::  PARABOLIC ! Spline has parabolic runout, as in DDY(1) = DDY(2) and DDY(N) = DDY(N-1)
        LOGICAL ::  CUBIC ! Spline has cubic runout, as in DDY(1) = 2*DDY(2)-DDY(3)
        LOGICAL ::  KNOWN_D ! First derivatives are known at all knots
        LOGICAL ::  IN_MEMORY = .FALSE.! Whether spline is allocated
        INTEGER ::  LAST_SEGMENT_USED   =   1   ! Remember in which segment the spline was last evaluated
        REAL(KIND=8), ALLOCATABLE   ::  M(:,:)  !< Matrix for linear equations (SPLINE%N-2,SPLINE%N-2)
        REAL(KIND=8), ALLOCATABLE   ::  DDV(:)  !< second derivative vector (unknown)
        REAL(KIND=8), ALLOCATABLE   ::  RHS(:)  !< right-hand-side of equation when solving the spline
        ! Commenting all TBPs for now
        CONTAINS
#ifndef PG_COMPILER
            ! The pgfortran compiler (13.3) crashes with ICE when the finalisation routine is there
            FINAL       ::  DEALLOCATE_INTERPOL_SPLINE
#endif
        !   PROCEDURE   ::  ALLOCATE    =>  ALLOCATE_INTERPOL_SPLINE
        !   PROCEDURE   ::  DEALLOCATE  =>  DEALLOCATE_INTERPOL_SPLINE
        !   PROCEDURE   ::  EVAL        =>  INTERPOL_SPLINE_EVAL
        !   PROCEDURE   ::  SET_TYPE    =>  SPLINE_SET_TYPE
        !   PROCEDURE   ::  SOLVE       =>  SOLVE_INTERPOL_SPLINE
        !   PROCEDURE   ::  EVAL_D      =>  INTERPOL_SPLINE_EVAL_D
    END TYPE INTERPOL_SPLINE

    INTERFACE INTERPOL_SPLINE_ALLOCATE
        MODULE PROCEDURE ALLOCATE_INTERPOL_SPLINE
    END INTERFACE

    INTERFACE INTERPOL_SPLINE_DEALLOCATE
        MODULE PROCEDURE DEALLOCATE_INTERPOL_SPLINE
    END INTERFACE

    INTERFACE INTERPOL_SPLINE_SOLVE
        MODULE PROCEDURE SOLVE_INTERPOL_SPLINE
    END INTERFACE

    INTERFACE IS_ALLOCATED
        MODULE PROCEDURE INTERPOL_SPLINE_ALLOCATED
    END INTERFACE

    INTERFACE ASSIGNMENT(=)
        MODULE PROCEDURE INTERPOL_SPLINE_ASSIGN
    END INTERFACE

CONTAINS

!>	\brief	Return the value of the spline's third derivative at X
PURE FUNCTION INTERPOL_SPLINE_EVAL_DDD(SELF,X) RESULT(DDDY)
    USE RUNTIME_PARAMETERS, ONLY: TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(IN)  ::  SELF
    REAL(KIND=8),           INTENT(IN)  ::  X
    ! Result
    REAL(KIND=8)    ::  DDDY
    ! Private variables
    INTEGER                 ::  SEG
    ! Start work


    ! Find which function we need to use
    IF (X .LT. SELF%X(1)) THEN
		SEG = 1
	ELSE IF (X .GT. SELF%X(SELF%N)) THEN
        SEG = SELF%N-1
    ELSE
        SEG = INTERPOL_SPLINE_FIND_SEG(SELF,X)
    ENDIF

    ! Now calculate the value
    DDDY = 6.*SELF%W(SEG,1)
END FUNCTION INTERPOL_SPLINE_EVAL_DDD

!>  \brief  Return the value of the spline's second derivative at X.
PURE FUNCTION INTERPOL_SPLINE_EVAL_DD(SELF,X) RESULT(DDY)
    USE RUNTIME_PARAMETERS, ONLY: TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(IN)  ::  SELF
    REAL(KIND=8),           INTENT(IN)  ::  X
    ! Result
    REAL(KIND=8)    ::  DDY
    ! Private variables
    INTEGER                 ::  SEG
    ! Start work


    ! Find which function we need to use
    IF (X .LT. SELF%X(1)) THEN
        SEG = 1
    ELSE IF (X .GT. SELF%X(SELF%N)) THEN
        SEG = SELF%N-1
    ELSE
        SEG = INTERPOL_SPLINE_FIND_SEG(SELF,X)
    ENDIF

    ! Now calculate the value
    DDY = 0.0D0
!    PRINT *, '**DEBUG(INTERPOL_SPLINE_EVAL_DD): SELF%W(SEG,1) = ', SELF%W(SEG,1)
!    PRINT *, '**DEBUG(INTERPOL_SPLINE_EVAL_DD): (X-SELF%X(SEG)) = ', (X-SELF%X(SEG))
!    PRINT *, '**DEBUG(INTERPOL_SPLINE_EVAL_DD): SELF%W(SEG,2) = ', SELF%W(SEG,2)
    DDY = 6.*SELF%W(SEG,1)*(X-SELF%X(SEG)) + 2.*SELF%W(SEG,2)
    ! This should be rewritten to work with any degree of spline (see the other EVAL function)

    ! Debug
    !IF (DEBUG) WRITE(*,'(A,I0,X,F0.4,X,F0.4)') '**debug(INTERPOL_SPLINE_EVAL_D): SEG, X, DY: ', SEG, X, DY
END FUNCTION INTERPOL_SPLINE_EVAL_DD

!>  \brief  Return the value of the spline's first derivative at X. Can do extrapolation too.
PURE FUNCTION INTERPOL_SPLINE_EVAL_D(SELF,X) RESULT(DY)
    USE RUNTIME_PARAMETERS, ONLY: TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(IN)  ::  SELF
    REAL(KIND=8),           INTENT(IN)  ::  X
    ! Result
    REAL(KIND=8)    ::  DY
    ! Private variables
    INTEGER                 ::  SEG
    !LOGICAL,   PARAMETER   ::  DEBUG   =   .FALSE.
    ! Start work
    !IF (SELF%K .NE. 3) THEN
    !   WRITE(*,'(A,I)') '**ERROR(INTERPOL_SPLINE_EVAL_D): only written for cubic splines. K = ', SELF%K
    !   CALL TERMINATE('Fatal error in INTERPOL_SPLINE_EVAL_D')
    !ENDIF

    ! Old error from when we didn't do extrapolation
    !IF (X .LT. SELF%X(1) .OR. X .GT. SELF%X(SELF%N)) THEN
    !   WRITE(*,'(3(A,F0.4),A)') '**WARNING(INTERPOL_SPLINE_EVAL_D): input value (', X, ') is outside range of the spline (', SELF%X(1), ' to ', SELF%X(SELF%N), '). Will be extrapolating.'
    !   !STOP 'Fatal error in INTERPOL_SPLINE_EVAL_D'
    !ENDIF

    ! Find which function we need to use
    IF (X .LT. SELF%X(1)) THEN
        SEG = 1
    ELSE IF (X .GT. SELF%X(SELF%N)) THEN
        SEG = SELF%N-1
    ELSE
        SEG = INTERPOL_SPLINE_FIND_SEG(SELF,X)
    ENDIF

    ! Now calculate the value
    DY = 0.0D0
    DY = 3.*SELF%W(SEG,1)*(X-SELF%X(SEG))**2 + 2.*SELF%W(SEG,2)*(X-SELF%X(SEG)) + SELF%W(SEG,3)
    ! This should be rewritten to work with any degree of spline (see the other EVAL function)

    ! Debug
    !IF (DEBUG) WRITE(*,'(A,I0,X,F0.4,X,F0.4)') '**debug(INTERPOL_SPLINE_EVAL_D): SEG, X, DY: ', SEG, X, DY
END FUNCTION INTERPOL_SPLINE_EVAL_D

!>  \brief  Return the value of the spline at X. Can do extrapolation too.
PURE SUBROUTINE INTERPOL_SPLINE_EVAL(SPLINE,X,Y,FIRST_AND_LAST_UNKNOWN)
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),      INTENT(INOUT)   ::  SPLINE
    REAL(KIND=8),               INTENT(IN)      ::  X
    REAL(KIND=8),               INTENT(OUT)     ::  Y                       !<  Result of the evaluation
    LOGICAL,        OPTIONAL,   INTENT(IN)      ::  FIRST_AND_LAST_UNKNOWN  !<  If .TRUE., the first and last segments' polynomial are unknown and we reuse the second and penultimate's segments polynomials instead.
    ! Private variables
    INTEGER                 ::  SEG                             ! the relevant segment of the spline
    INTEGER                 ::  DEG                             ! degree
    LOGICAL, PARAMETER      ::  HORNER          =   .TRUE.      ! Whether to use Horner's method for evaluating the polynomial (default) or the naive method
    LOGICAL                 ::  FFIRST_AND_LAST_UNKNOWN
    LOGICAL,    PARAMETER   ::  DEBUG           =   .FALSE.
    ! Start work

    ! Do we have well defined first/last segments?
    IF (PRESENT(FIRST_AND_LAST_UNKNOWN)) THEN
        FFIRST_AND_LAST_UNKNOWN = FIRST_AND_LAST_UNKNOWN
    ELSE
        FFIRST_AND_LAST_UNKNOWN = .FALSE.
    ENDIF


    ! If X is before the first point (or after the last point) of the interpolation, we will extrapolate using the first polynomial
    IF (X .LT. SPLINE%X(1)) THEN
        !IF (DEBUG) THEN
        !    PRINT '(A,F0.5)', '**debug(INTERPOL_SPLINE_EVAL): extrapolation (before 1st point). X = ', X
        !    CALL INTERPOL_SPLINE_PRINT_INFO(SPLINE)
        !    CALL TERMINATE('Debug INTERPOL_SPLINE_EVAL')
        !ENDIF
        SEG = 1
    ELSE IF (X .GT. SPLINE%X(SPLINE%N)) THEN
        SEG = SPLINE%N-1
    ELSE
        ! Find which function we need to use
        SEG = INTERPOL_SPLINE_FIND_SEG(SPLINE,X)
        SPLINE%LAST_SEGMENT_USED = SEG
    ENDIF

    ! If we don't know the first and last segment properly
    IF (FFIRST_AND_LAST_UNKNOWN) THEN
        !PRINT *, '**debug(INTERPOL_SPLINE_EVAL): SPLINE%N =', SPLINE%N
        IF (SEG .EQ. 1) SEG = 2
        IF (SEG .EQ. SPLINE%N-1) SEG = SPLINE%N-2
        !IF (DEBUG) THEN
        !    PRINT *, '**debug(INTERPOL_SPLINE_EVAL): first and last unknown. SEG =', SEG
        !    PRINT *, '**debug(INTERPOL_SPLINE_EVAL): SPLINE%W(SEG,1:4) =', SPLINE%W(SEG,1:4)
        !ENDIF
    ENDIF

    ! Now calculate the value
    Y = 0.0D0
    IF (HORNER) THEN
        Y = SPLINE%W(SEG,1)
        DO DEG = 1, SPLINE%K
            Y = Y*(X-SPLINE%X(SEG)) + SPLINE%W(SEG,DEG+1)
        ENDDO
    ELSE
        DO DEG = 1, SPLINE%K+1
            Y = Y + SPLINE%W(SEG,DEG) * (X-SPLINE%X(SEG))**DBLE(SPLINE%K-DEG+1)
        ENDDO
    ENDIF
END SUBROUTINE INTERPOL_SPLINE_EVAL

!>  \brief  Return the value of the spline at X, as well as its first and second derivatives. Uses Horner's method.
!!
!!  Algorithm from
!!  Horner's Method for Evaluating and Deflating Polynomials
!!  C. Sidney Burrus, James W. Fox, Gary A. Sitton, and Sven Treitel
!!  Rice University
!!  November 26, 2003
PURE SUBROUTINE INTERPOL_SPLINE_HORNER_EVAL_ALL(SPLINE,X,Y,D,DD,FIRST_AND_LAST_UNKNOWN)
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),      INTENT(INOUT)   ::  SPLINE
    REAL(KIND=8),               INTENT(IN)      ::  X
    REAL(KIND=8),               INTENT(OUT)     ::  Y                       !<  Result of the evaluation
    REAL(KIND=8),               INTENT(OUT)     ::  D                       !<  First derivative
    REAL(KIND=8),               INTENT(OUT)     ::  DD                      !<  Second derivative
    LOGICAL,        OPTIONAL,   INTENT(IN)      ::  FIRST_AND_LAST_UNKNOWN  !<  If .TRUE., the first and last segments' polynomial are unknown and we reuse the second and penultimate's segments polynomials instead.
    ! Private variables
    INTEGER                 ::  SEG                             ! the relevant segment of the spline
    INTEGER                 ::  DEG                             ! degree
    LOGICAL                 ::  FFIRST_AND_LAST_UNKNOWN
    LOGICAL,    PARAMETER   ::  DEBUG           =   .FALSE.
    REAL(KIND=8)            ::  XX
    ! Start work

    ! Do we have well defined first/last segments?
    IF (PRESENT(FIRST_AND_LAST_UNKNOWN)) THEN
        FFIRST_AND_LAST_UNKNOWN = FIRST_AND_LAST_UNKNOWN
    ELSE
        FFIRST_AND_LAST_UNKNOWN = .FALSE.
    ENDIF


    ! If X is before the first point (or after the last point) of the interpolation, we will extrapolate using the first polynomial
    IF (X .LT. SPLINE%X(1)) THEN
        SEG = 1
    ELSE IF (X .GT. SPLINE%X(SPLINE%N)) THEN
        SEG = SPLINE%N-1
    ELSE
        ! Find which function we need to use
        SEG = INTERPOL_SPLINE_FIND_SEG(SPLINE,X)
        SPLINE%LAST_SEGMENT_USED = SEG
    ENDIF

    ! If we don't know the first and last segment properly
    IF (FFIRST_AND_LAST_UNKNOWN) THEN
        IF (SEG .EQ. 1) SEG = 2
        IF (SEG .EQ. SPLINE%N-1) SEG = SPLINE%N-2
    ENDIF

    ! Now calculate the value, the derivative and the second derivative
    Y  = 0.0D0
    D  = 0.0D0
    DD = 0.0D0
    XX = X-SPLINE%X(SEG)
    DO DEG = 1, SPLINE%K-1
        Y  = XX*Y  + SPLINE%W(SEG,DEG)
        D  = XX*D  + Y
        DD = XX*DD + D
    ENDDO
    Y  = XX*Y + SPLINE%W(SEG,SPLINE%K)
    D  = XX*D + Y
    Y  = XX*Y + SPLINE%W(SEG,SPLINE%K+1)
    DD = 2*DD

END SUBROUTINE INTERPOL_SPLINE_HORNER_EVAL_ALL


!>  \brief  Find the roots X of the spline when its value is an integer multiple of Y
FUNCTION INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE(SELF,Y,Y_OFFSET) RESULT(X)
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),              INTENT(IN)  ::  SELF        !<  Interpolating spline
    REAL(KIND=8),                       INTENT(IN)  ::  Y           !<  Value of the spline. The roots of the spline when its value is a multiple of Y will be returned
    REAL(KIND=8),           OPTIONAL,   INTENT(IN)  ::  Y_OFFSET    !<  If present, the roots of the spline when its value is Y_OFFSET + n*Y will be returned.
    ! Result
    REAL(KIND=8),           ALLOCATABLE ::  X(:)
    ! Private variables
    !REAL(KIND=8)   ::  MIN_X,MAX_X
    REAL(KIND=8)    ::  MIN_Y,MAX_Y
    REAL(KIND=8)    ::  CURRENT_Y
    INTEGER         ::  NUMBER_OF_ROOTS, CURRENT_ROOT, CURRENT_Y_MUL
    INTEGER         ::  MIN_Y_MUL, MAX_Y_MUL
    INTEGER         ::  IERR
    REAL(KIND=8)    ::  YY_OFFSET
    ! Start work

    ! Y offset
    IF (PRESENT(Y_OFFSET)) THEN
        YY_OFFSET = Y_OFFSET
    ELSE
        YY_OFFSET = 0.0D0
    ENDIF

    ! For simplification, we only treat the case where the spline is monotonic
    IF (.NOT. (INTERPOL_SPLINE_IS_MONOTONIC_INCREASING(SELF) .OR. INTERPOL_SPLINE_IS_MONOTONIC_DECREASING(SELF))) THEN
        WRITE(*,'(A)') '**ERROR(INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE): spline is not monotonic increasing or decreasing'
        WRITE(*,'(A,2L1)') '**ERROR(INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE):  ALL(SELF%DY .GE. 0.0D0), ALL(SELF%DDY .GE. 0.0D0): ', &
                            ALL(SELF%DY .GE. 0.0D0), ALL(SELF%DDY .GE. 0.0D0)
        CALL INTERPOL_SPLINE_PRINT_INFO(SELF)
        CALL TERMINATE('Fatal error in INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE')
    ENDIF

    ! Find the minimum and maximum values of the spline
    MIN_Y = MIN(SELF%Y(1),SELF%Y(SELF%N))
    MAX_Y = MAX(SELF%Y(1),SELF%Y(SELF%N))

    ! Find the multiple of Y nearest to, but greater than, MIN_Y
    MIN_Y_MUL = CEILING((MIN_Y-YY_OFFSET)/Y)
    MAX_Y_MUL = FLOOR((MAX_Y-YY_OFFSET)/Y)

    ! How many roots?
    NUMBER_OF_ROOTS = MAX_Y_MUL - MIN_Y_MUL + 1

    ! Allocate memory for the results
    IF (ALLOCATED(X)) DEALLOCATE(X)
    ALLOCATE(X(NUMBER_OF_ROOTS),STAT=IERR)
    IF (IERR .NE. 0) THEN
        CALL TERMINATE('Memory allocation failed in INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE')
    ENDIF

    ! Find the solutions
    DO CURRENT_ROOT = 1, NUMBER_OF_ROOTS
        CURRENT_Y_MUL = MIN_Y_MUL + CURRENT_ROOT - 1
        CURRENT_Y = DBLE(CURRENT_Y_MUL) * Y + YY_OFFSET
        X(CURRENT_ROOT) = INTERPOL_SPLINE_FIND_ROOT(SELF,CURRENT_Y)
    ENDDO

END FUNCTION INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE

!>  \brief  Find the root X of the spline when its value is Y
FUNCTION INTERPOL_SPLINE_FIND_ROOT(SELF,Y) RESULT(X)
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    USE, INTRINSIC :: IEEE_ARITHMETIC, ONLY : IEEE_IS_NAN
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(IN)  ::  SELF
    REAL(KIND=8),           INTENT(IN)  ::  Y
    ! Result
    REAL(KIND=8)                        ::  X
    ! Private variables
    INTEGER ::  I, SEG
    LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
    ! Start work

    IF (DEBUG) THEN
        WRITE(*,'(A)') '**debug(INTERPOL_SPLINE_FIND_ROOT): spline info follows'
        CALL INTERPOL_SPLINE_PRINT_INFO(SELF)
    ENDIF

    ! Find which segment we're interested in
    ! While we're at it, check for special case where the solution is one of the waypoints
    SEG = 0
    IF (INTERPOL_SPLINE_IS_MONOTONIC_INCREASING(SELF)) THEN
        DO I=1,SELF%N-1
            IF (Y .EQ. SELF%Y(I)) THEN
                X = SELF%X(I)
                RETURN
            ELSE IF (Y .EQ. SELF%Y(I+1)) THEN
                X = SELF%X(I+1)
                RETURN
            ELSE IF (Y .GE. SELF%Y(I) .AND. Y .LE. SELF%Y(I+1)) THEN
                SEG = I
                EXIT
            ENDIF
        ENDDO
    ELSE IF (INTERPOL_SPLINE_IS_MONOTONIC_DECREASING(SELF)) THEN
        DO I=1,SELF%N-1
            IF (Y .EQ. SELF%Y(I)) THEN
                X = SELF%X(I)
                RETURN
            ELSE IF (Y .EQ. SELF%Y(I+1)) THEN
                X = SELF%X(I+1)
                RETURN
            ELSE IF (Y .LE. SELF%Y(I) .AND. Y .GE. SELF%Y(I+1)) THEN
                SEG = I
                EXIT
            ENDIF
        ENDDO
    ELSE
        WRITE(*,'(A)') '**ERROR(INTERPOL_SPLINE_FIND_ROOT): spline is not monotonic increasing or decreasing'
        WRITE(*,'(A,2L1)') '**ERROR(INTERPOL_SPLINE_FIND_ROOT):  ALL(SELF%DY .GE. 0.0D0), ALL(SELF%DDY .GE. 0.0D0): ', &
                            ALL(SELF%DY .GE. 0.0D0), ALL(SELF%DDY .GE. 0.0D0)
        CALL INTERPOL_SPLINE_PRINT_INFO(SELF)
        CALL TERMINATE('Fatal error in INTERPOL_SPLINE_FIND_ROOT')
    ENDIF

    IF (SEG .EQ. 0) THEN
        WRITE(*,'(3(A,F0.3),A)')    '**ERROR(INTERPOL_SPLINE_FIND_ROOT): value Y (', Y, &
                                    ') does not appear to lie within the values of the spline (from ', &
                                    SELF%Y(1), ' to ', SELF%Y(SELF%N), ')'
        CALL INTERPOL_SPLINE_PRINT_INFO(SELF)
        CALL TERMINATE('Fatal error in INTERPOL_SPLINE_FIND_ROOT')
    ENDIF


    !
    X = SELF%X(SEG) + GSL_SOLVE_CUBIC_POLYNOMIAL(SELF%W(SEG,1),SELF%W(SEG,2),SELF%W(SEG,3),SELF%W(SEG,4),Y, &
                                                XMIN=SELF%X(SEG)-SELF%X(SEG),XMAX=SELF%X(SEG+1)-SELF%X(SEG))

    IF (IEEE_IS_NAN(X)) THEN
        WRITE(*,'(A)') '**ERROR(INTERPOL_SPLINE_FIND_ROOT): found NaN value'
        CALL TERMINATE('Fatal error in INTERPOL_SPLINE_FIND_ROOT')
    ENDIF
END FUNCTION INTERPOL_SPLINE_FIND_ROOT

!>  \brief  Return whether the spline is strictly monotonic
LOGICAL FUNCTION INTERPOL_SPLINE_IS_MONOTONIC(SELF)
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(IN)  ::  SELF
    ! Private variables
    REAL(KIND=8)    ::  D_ROOTS(2)
    INTEGER         ::  NUM_D_ROOTS
    INTEGER         ::  SEG
    LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
    ! Start work

    IF (DEBUG) THEN
        WRITE(*,'(A)') '**debug(INTERPOL_SPLINE_IS_MONOTONIC): spline info on entry'
        CALL INTERPOL_SPLINE_PRINT_INFO(SELF,CHECK_MONO=.FALSE.)
    ENDIF

    INTERPOL_SPLINE_IS_MONOTONIC = .TRUE.
    ! A function is definitely monotonic if its derivative is never 0.0.
    ! So for each segment, let's check whether we can find roots for the derivative.
    DO SEG=1,SELF%N-1
        NUM_D_ROOTS = GSL_SOLVE_QUADRATIC(3*SELF%W(SEG,1),2*SELF%W(SEG,2),SELF%W(SEG,3),D_ROOTS,  &
                                            XMIN=0.0D0,XMAX=SELF%X(SEG+1)-SELF%X(SEG))
        IF (DEBUG) THEN
			WRITE(*,'(A,I0)') '**DEBUG(INTERPOL_SPLINE_IS_MONOTONIC): num_d_roots = ', NUM_D_ROOTS
			WRITE(*,*) 'SEG, D_ROOTS = ', SEG, D_ROOTS
		ENDIF
        IF (NUM_D_ROOTS .GT. 0) THEN
			! The function may still be monotonic, but not strictly so
			INTERPOL_SPLINE_IS_MONOTONIC = .FALSE.
		ENDIF
    ENDDO

END FUNCTION INTERPOL_SPLINE_IS_MONOTONIC

!>  \brief  Return whether the spline is strictly monotonic increasing
LOGICAL FUNCTION INTERPOL_SPLINE_IS_MONOTONIC_INCREASING(SELF)
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(IN)  ::  SELF
    ! Private variables
    ! Start work
    INTERPOL_SPLINE_IS_MONOTONIC_INCREASING = INTERPOL_SPLINE_IS_MONOTONIC(SELF) .AND. SELF%Y(SELF%N) .GT. SELF%Y(1)
END FUNCTION INTERPOL_SPLINE_IS_MONOTONIC_INCREASING

LOGICAL FUNCTION INTERPOL_SPLINE_IS_MONOTONIC_DECREASING(SELF)
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(IN)  ::  SELF
    ! Private variables
    ! Start work
    INTERPOL_SPLINE_IS_MONOTONIC_DECREASING =   INTERPOL_SPLINE_IS_MONOTONIC(SELF) .AND. SELF%Y(SELF%N) .LT. SELF%Y(1)
END FUNCTION INTERPOL_SPLINE_IS_MONOTONIC_DECREASING


!>  \brief  Given an ordinate X, find out which segment of the spline it "belongs" to
PURE FUNCTION INTERPOL_SPLINE_FIND_SEG(SELF,X) RESULT(SEG)
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),              INTENT(IN)  ::  SELF
    REAL(KIND=8),                       INTENT(IN)  ::  X
    ! Result
    INTEGER ::  SEG
    ! Private variables
    INTEGER ::  I
    ! Start work

    SEG = 0
    DO I=SELF%LAST_SEGMENT_USED,SELF%N-1
        IF (X .GT. SELF%X(I) .AND. X .LE. SELF%X(I+1)) THEN
            SEG = I
            EXIT
        ENDIF
    ENDDO
    ! If we still haven't found the segment, start again from the beginning
    IF (SEG .EQ. 0) THEN
        DO I=1,SELF%LAST_SEGMENT_USED-1
            IF (X .GT. SELF%X(I) .AND. X .LE. SELF%X(I+1)) THEN
                SEG = I
                EXIT
            ENDIF
        ENDDO
    ENDIF
	IF (SEG .EQ. 0) THEN
		IF (X .EQ. SELF%X(1)) SEG = 1
	ENDIF
END FUNCTION INTERPOL_SPLINE_FIND_SEG


!>  \brief  Return the value of the spline at X, but don't use weights - instead use second derivatives
FUNCTION INTERPOL_SPLINE_EVAL2(SPLINE,X) RESULT(Y)
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(IN)  ::  SPLINE
    REAL(KIND=8),           INTENT(IN)  ::  X
    REAL(KIND=8)    ::  Y
    ! Private variables
    INTEGER ::  I
    INTEGER ::  SEG ! the relevant segment of the spline
    REAL(KIND=8)    ::  A,B,C,D
    ! Start work
    !IF (X .LT. SPLINE%X(1) .OR. X .GT. SPLINE%X(SPLINE%N)) THEN
    !   WRITE(*,'(3(A,F6.2),A)') '**ERROR(INTERPOL_SPLINE_EVAL2): input value (', X, ') is outside range of the spline (',  &
    !                                                               SPLINE%X(1), ' to ', SPLINE%X(SPLINE%N), ')'
    !   CALL TERMINATE('Fatal error in INTERPOL_SPLINE_EVAL2')
    !ENDIF
    IF (SPLINE%K .NE. 3) THEN
        WRITE(*, '(A)') '**ERROR(INTERPOL_SPLINE_EVAL2): routine only works on cubic splines'
        CALL TERMINATE('Fatal error in INTERPOL_SPLINE_EVAL2')
    ENDIF
    ! Find which function we need to use
    DO I=1,SPLINE%N-1
        IF (X .GE. SPLINE%X(I) .AND. X .LE. SPLINE%X(I+1)) THEN
            SEG = I
            EXIT
        ENDIF
    ENDDO
    ! Now calculate the value (see numerical recipes chapter on cublic spline interpolation)
    A = (SPLINE%X(SEG+1) - X) / (SPLINE%X(SEG+1)-SPLINE%X(SEG))
    B = 1.0D0 - A
    C = (A**3-A)*(SPLINE%X(SEG+1)-SPLINE%X(SEG))**2/6.0D0
    D = (B**3-B)*(SPLINE%X(SEG+1)-SPLINE%X(SEG))**2/6.0D0
    Y = A*SPLINE%Y(SEG) + B*SPLINE%Y(SEG+1) + C*SPLINE%DDY(SEG) + D*SPLINE%DDY(SEG+1)
END FUNCTION INTERPOL_SPLINE_EVAL2

!>  \brief  Set type of spline
SUBROUTINE SPLINE_SET_TYPE(SPLINE,TYPE)
    IMPLICIT NONE
    TYPE(INTERPOL_SPLINE)   ::  SPLINE
    CHARACTER(LEN=*)        ::  TYPE    !<  Character string - we'll only need to check first letter
    ! Private variable

    ! Start work
    SPLINE%NATURAL      =   .FALSE.
    SPLINE%PARABOLIC    =   .FALSE.
    SPLINE%CUBIC        =   .FALSE.
    SPLINE%KNOWN_D      =   .FALSE.
    SELECT CASE(TYPE(1:1))
        CASE('N')
            SPLINE%NATURAL      =   .TRUE.
        CASE ('P')
            SPLINE%PARABOLIC    =   .TRUE.
        CASE ('C')
            SPLINE%CUBIC        =   .TRUE.
        CASE ('K')
            SPLINE%KNOWN_D      =   .TRUE.
        CASE DEFAULT
            WRITE(*,'(2A)') '**WARNING(SPLINE_SET_TYPE): did not understand what type you meant: ', TYPE
            WRITE(*,'(A)')  '**WARNING(SPLINE_SET_TYPE): setting type to natural'
            SPLINE%NATURAL      =   .TRUE.
    END SELECT
END SUBROUTINE SPLINE_SET_TYPE

!>  \brief  Solve the interpolation spline, using the FGSL for the linear algebra and following the method described in
!!      PDF in the doc folder
SUBROUTINE SOLVE_INTERPOL_SPLINE(SELF)
    USE LINEAR_ALGEBRA
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(INOUT)   ::  SELF  !<  Spline object
    ! Private variables

    REAL(KIND=8)    ::  H, HM
    INTEGER ::  I,J
    CHARACTER(LEN=6) :: N_C
    LOGICAL, PARAMETER :: DEBUG = .FALSE.
    LOGICAL, PARAMETER :: USE_LAPACK = .TRUE.
    ! Start work
#ifndef SKIP_RUNTIME_CHECKS
    WRITE(N_C,'(I6)') SELF%N
    ! Check spline is allocated
    IF (.NOT. SELF%IN_MEMORY) THEN
        WRITE(*,'(A)') '**ERROR(SOLVE_INTERPOL_SPLINE): spline not allocated on entry', IS_ALLOCATED(SELF)
        CALL TERMINATE('Fatal error in SOLVE_INTERPOL_SPLINE')
    ENDIF
    ! Check values of X are increasing
    DO I=2,SELF%N
        IF (SELF%X(I) .LE. SELF%X(I-1)) THEN
            WRITE(*,'(A)') '**ERROR(SOLVE_INTERPOL_SPLINE): values of X must increase'
            WRITE(*,'('//N_C//'(F0.2,1X))') SELF%X
            CALL TERMINATE('Fatal error in SOLVE_INTERPOL_SPLINE')
        ENDIF
    ENDDO
#endif


    ! Setup the matrix
    ! This is done according to the document included in the documentation.
    !
    ! The spline is solved by solving this set of linear equations:
    !   (M)*(DDV) = (RHS)
    !
    IF (SELF%NATURAL .OR. SELF%PARABOLIC .OR. SELF%CUBIC) THEN
        DO I=2,SELF%N-1
            H  = SELF%X(I+1)-SELF%X(I)
            HM = SELF%X(I)-SELF%X(I-1)
            DO J=2,SELF%N-1
                IF (J .EQ. I) THEN
                    SELF%M(I-1,J-1) = 2.0D0 * (HM + H)
                ELSE IF (J .EQ. I+1) THEN
                    SELF%M(I-1,J-1) = H
                ELSE IF (J .EQ. I-1) THEN
                    SELF%M(I-1,J-1) = HM
                ELSE
                    SELF%M(I-1,J-1) = 0.0D0
                ENDIF
            ENDDO
            SELF%RHS(I-1) = 6.0D0*((SELF%Y(I+1)-SELF%Y(I))/H-(SELF%Y(I)-SELF%Y(I-1))/HM)
        ENDDO
        IF (SELF%PARABOLIC) THEN
            ! 1st modification
            I  = 2
            HM = SELF%X(I)-SELF%X(I-1)
            J  = 2
            SELF%M(I-1,J-1) = SELF%M(I-1,J-1) + HM
            ! 2nd modification
            I = SELF%N-1
            HM = SELF%X(I)-SELF%X(I-1)
            H  = SELF%X(I+1)-SELF%X(I)
            J  = I
            SELF%M(I-1,J-1) = SELF%M(I-1,J-1) + H
        ELSE IF (SELF%CUBIC) THEN
            ! 1st modification
            I  = 2
            HM = SELF%X(I)-SELF%X(I-1)
            H  = SELF%X(I+1)-SELF%X(I)
            J  = 2
            SELF%M(I-1,J-1) = SELF%M(I-1,J-1) + 2.0D0*HM
            J  = 3
            SELF%M(I-1,J-1) = SELF%M(I-1,J-1) - HM
            ! 2nd modification
            I = SELF%N-1
            HM = SELF%X(I)-SELF%X(I-1)
            H  = SELF%X(I+1)-SELF%X(I)
            J  = I
            SELF%M(I-1,J-1) = SELF%M(I-1,J-1) + 2.0D0*H
            J  = I-1
            SELF%M(I-1,J-1) = SELF%M(I-1,J-1) - H
        ENDIF ! End of modifications for non-natural
        ! Initialise the unknown vector
        SELF%DDV(1:SELF%N-2) = 0.0D0

#ifndef SKIP_RUNTIME_CHECKS
        ! Diagnostics
        IF (DEBUG) THEN
            PRINT *, '**debug(SOLVE_INTERPOL_SPLINE): SELF%M   = ', SELF%M
            PRINT *, '**debug(SOLVE_INTERPOL_SPLINE): det = ', MATRIX_DET(SELF%M)
            PRINT *, '**debug(SOLVE_INTERPOL_SPLINE): SELF%N-2 = ', SELF%N-2
            PRINT *, '**debug(SOLVE_INTERPOL_SPLINE): SELF%RHS = ', SELF%RHS
        ENDIF
#endif

        ! To solve the equations, we will use LU decomposition, as provided by GSL
        ! (Gaussian decomposition may be more efficient?)
        IF (SELF%N .EQ. 3) THEN
            SELF%DDV(1) = SELF%RHS(1)/SELF%M(1,1)
        ELSE
            IF (USE_LAPACK) THEN
                CALL LINALG_SOLVE_LAPACK(SELF%M,SELF%DDV,SELF%RHS)
            ELSE
                CALL LINALG_SOLVE_LU_FGSL(SELF%M,SELF%DDV,SELF%RHS,SELF%N-2)
            ENDIF
        ENDIF
        !PRINT *, '**debug(SOLVE_INTERPOL_SPLINE): DDV = ', DDV

        ! Copy the results (second derivatives) over into the spline object; the boundaries are worked out separately.
        DO I=2,SELF%N-1
            SELF%DDY(I) = SELF%DDV(I-1)
        ENDDO
        ! The second derivative at the boundaries
        IF (SELF%NATURAL) THEN
            SELF%DDY(1) = 0.0D0
            SELF%DDY(SELF%N) = 0.0D0
        ELSE IF (SELF%PARABOLIC) THEN
            SELF%DDY(1) = SELF%DDY(2)
            SELF%DDY(SELF%N) = SELF%DDY(SELF%N-1)
        ELSE IF (SELF%CUBIC) THEN
            SELF%DDY(1) = 2*SELF%DDY(2) - SELF%DDY(3)
            SELF%DDY(SELF%N) = 2*SELF%DDY(SELF%N-1) - SELF%DDY(SELF%N-2)
        ENDIF


        ! Now we know the second derivatives at each knot (DDY), we can work out the weights
        DO I=1,SELF%N-1
            H = SELF%X(I+1)-SELF%X(I)

            SELF%W(I,4) = SELF%Y(I)
            IF (SELF%KNOWN_D) THEN
                SELF%W(I,3) = SELF%DY(I)
            ELSE
                SELF%W(I,3) = (SELF%Y(I+1)-SELF%Y(I))/H       &
                                - (SELF%DDY(I+1) + 2*SELF%DDY(I))*H/6.0D0
            ENDIF
            SELF%W(I,2) = SELF%DDY(I) / 2.0D0
            SELF%W(I,1) = (SELF%DDY(I+1)-SELF%DDY(I))/(6.0D0*H)
        ENDDO

        ! Now we can work out the first derivative
        DO I=1,SELF%N-1
            SELF%DY(I) = SELF%W(I,3)
        ENDDO

        ! Work out the first derivative on the last point
        SELF%DY(SELF%N) = INTERPOL_SPLINE_EVAL_D(SELF,SELF%X(SELF%N))

    ELSE IF (SELF%KNOWN_D) THEN
        IF (SELF%K .EQ. 3) THEN
            DO I=1,SELF%N-1
                H = SELF%X(I+1)-SELF%X(I)
                SELF%W(I,1) = -2.0D0*(SELF%Y(I+1)-SELF%Y(I)-(SELF%DY(I+1)-SELF%DY(I))*H/2.0D0 - SELF%DY(I)*H) /H**3
                SELF%W(I,2) = ( SELF%DY(I+1) - SELF%DY(I) - 3.0D0*H**2*SELF%W(I,1) ) / (2.0D0*H)
                SELF%W(I,3) = SELF%DY(I)
                SELF%W(I,4) = SELF%Y(I)
                SELF%DDY(I) = 2.0D0 * SELF%W(I,2)
            ENDDO
        ELSE
#ifndef SKIP_RUNTIME_CHECKS
            WRITE(*,'(A)') '**ERROR(SOLVE_INTERPOL_SPLINE): KNOWN_D only available for cubic spline'
            CALL TERMINATE('Fatal error in SOLVE_INTERPOL_SPLINE')
#endif
        ENDIF
    ELSE
#ifndef SKIP_RUNTIME_CHECKS
        WRITE(*,'(A)') '**ERROR(SOLVE_INTERPOL_SPLINE): unknown spline type!'
        CALL TERMINATE('Fatal error in SOLVE_INTERPOL_SPLINE')
#endif
    ENDIF !

    ! Now that we have solved the spline, we know its first derivative (added 18-Aug-2011)
    SELF%KNOWN_D      =   .TRUE.
    SELF%NATURAL      =   .FALSE.
    SELF%PARABOLIC    =   .FALSE.
    SELF%CUBIC        =   .FALSE.

    ! Diagnostics
    !CALL INTERPOL_SPLINE_PRINT_INFO(SPLINE)
END SUBROUTINE SOLVE_INTERPOL_SPLINE

!>  \brief  Print summary info to terminal
SUBROUTINE INTERPOL_SPLINE_PRINT_INFO(SELF,CHECK_MONO)
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(IN)  ::  SELF
	LOGICAL, OPTIONAL,      INTENT(IN)  ::  CHECK_MONO
    ! Private variables
    REAL    ::  EVAL2
	LOGICAL ::  CCHECK_MONO
	INTEGER ::  I
    ! Start work
	CCHECK_MONO = .TRUE.
	IF (PRESENT(CHECK_MONO)) CCHECK_MONO =  CHECK_MONO
    PRINT *
    PRINT '(A)', '**Spline information**'
    WRITE(*,'(A,I0)') 'Number of points: ', SELF%N
    WRITE(*,'(A,5(L1,1X))') 'Natural? Parabolic? Cubic? Known D? In memory?', SELF%NATURAL, SELF%PARABOLIC, &
                                                                            SELF%CUBIC, SELF%KNOWN_D, SELF%IN_MEMORY
    WRITE(*,'(A7,1X,9(A12,1X))') 'Point #','W1', 'W2', 'W3', 'W4', 'X', 'Y', 'S(X)(2)', 'DY', 'DDY'
    DO I=1,SELF%N
        !CALL INTERPOL_SPLINE_EVAL(SELF,SELF%X(I),EVAL)
        EVAL2 = INTERPOL_SPLINE_EVAL2(SELF,SELF%X(I))
        WRITE(*,'(I7,1X,4(G12.5,1X) ,2(F12.2,1X),3(G12.5,1X))') I, SELF%W(I,1:4), SELF%X(I), SELF%Y(I),     &
                                                EVAL2,    &
                                                SELF%DY(I), SELF%DDY(I)
    ENDDO
	IF (CCHECK_MONO) THEN
	    WRITE(*,'(A,2(L1,1X))') 'Monotonic increasing? decreasing? ', INTERPOL_SPLINE_IS_MONOTONIC_INCREASING(SELF), &
                            INTERPOL_SPLINE_IS_MONOTONIC_DECREASING(SELF)
	ENDIF
	PRINT *
END SUBROUTINE INTERPOL_SPLINE_PRINT_INFO

!>  \brief  Subroutine used to overload the assignment operator
SUBROUTINE INTERPOL_SPLINE_ASSIGN(LHS,RHS)
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(INOUT)   ::  LHS
    TYPE(INTERPOL_SPLINE),  INTENT(IN)      ::  RHS
    ! Private variables

    ! Start work
    IF (IS_ALLOCATED(LHS) .AND. ((LHS%N .NE. RHS%N) .OR. (LHS%K .NE. RHS%K))) CALL DEALLOCATE_INTERPOL_SPLINE(LHS)
    IF (.NOT. IS_ALLOCATED(LHS)) CALL ALLOCATE_INTERPOL_SPLINE(LHS,RHS%N,RHS%K)
    LHS%K           =   RHS%K
    IF (ALLOCATED(RHS%X)) THEN
        LHS%X(:)    =   RHS%X  ! We use (:) on the lhs to avoid reallocation by the compiler
    ENDIF
    IF (ALLOCATED(RHS%Y)) THEN
        LHS%Y(:)    =   RHS%Y
    ENDIF
    IF (ALLOCATED(RHS%DY)) THEN
        LHS%DY(:)   =   RHS%DY
    ENDIF
    IF (ALLOCATED(RHS%DDY)) THEN
    LHS%DDY(:)      =   RHS%DDY
    ENDIF
    IF (ALLOCATED(RHS%W)) THEN
    LHS%W(:,:)      =   RHS%W
    ENDIF
    LHS%NATURAL     =   RHS%NATURAL
    LHS%PARABOLIC   =   RHS%PARABOLIC
    LHS%CUBIC       =   RHS%CUBIC
    LHS%KNOWN_D     =   RHS%KNOWN_D
    LHS%IN_MEMORY   =   RHS%IN_MEMORY
    LHS%LAST_SEGMENT_USED   =   RHS%LAST_SEGMENT_USED
    IF (ALLOCATED(RHS%M)) then
        LHS%M       =   RHS%M
    ENDIF
    IF (ALLOCATED(RHS%DDV)) then
        LHS%DDV     =   RHS%DDV
    endif
    IF (allocated(RHS%RHS)) then
        LHS%RHS     =   RHS%RHS
    endif
END SUBROUTINE INTERPOL_SPLINE_ASSIGN

!>  \brief  Deallocate memory from interpol_spline object
SUBROUTINE DEALLOCATE_INTERPOL_SPLINE(SELF)
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(INOUT)   ::  SELF
    ! Private variables

    ! Start work
    IF (IS_ALLOCATED(SELF)) THEN
        DEALLOCATE(SELF%X,SELF%Y,SELF%DY,SELF%DDY,SELF%W)
        DEALLOCATE(SELF%M,SELF%DDV,SELF%RHS)
		SELF%IN_MEMORY = .FALSE.
    ELSE
        !WRITE(*,'(A)') '**ERROR(DEALLOCATE_INTERPOL_SPLINE): cannot deallocate an object which is not allocated'
        !CALL TERMINATE('Fatal error in DEALLOCATE_INTERPOL_SPLINE')
    ENDIF

END SUBROUTINE DEALLOCATE_INTERPOL_SPLINE

!>  \brief  Check wither an interpol_spline object is currently allocated
PURE LOGICAL FUNCTION INTERPOL_SPLINE_ALLOCATED(SELF)
    !USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(IN)  ::  SELF
    ! Start work
    IF (SELF%IN_MEMORY) THEN
        IF (ALLOCATED(SELF%X) .AND. ALLOCATED(SELF%Y) .AND. ALLOCATED(SELF%DY) .AND. ALLOCATED(SELF%DDY)) THEN
            INTERPOL_SPLINE_ALLOCATED = .TRUE.
        ELSE
            !CALL TERMINATE('Incoherent allocation status')
        ENDIF
    ELSE
        INTERPOL_SPLINE_ALLOCATED = .FALSE.
    ENDIF
END FUNCTION INTERPOL_SPLINE_ALLOCATED

!>  \brief  Allocate memory to interpol_spline object
SUBROUTINE ALLOCATE_INTERPOL_SPLINE(SELF,NUM_POINTS,DEGREE)
    USE RUNTIME_PARAMETERS, ONLY: TERMINATE
    IMPLICIT NONE
    ! Arguments
    TYPE(INTERPOL_SPLINE),  INTENT(INOUT)   ::  SELF        !< Spline object
    INTEGER,                INTENT(IN)      ::  NUM_POINTS  !< Number of points to interpolate through
    INTEGER,                INTENT(IN)      ::  DEGREE      !< Degree of the splines
    ! Private variables
    INTEGER ::  IERR
    ! Start work
    !PRINT *, '**debug(ALLOCATE_INTERPOL_SPLINE): on entry, ', ALLOCATED(SELF), ALLOCATED(SELF%X), SIZE(SELF%X), ALLOCATED(SELF%Y)
    IF (DEGREE .NE. 3) THEN
        WRITE(*, '(A)') '**ERROR(ALLOCATE_INTERPOL_SPLINE): only cubic splines supported'
        CALL TERMINATE('Fatal error in ALLOCATE_INTERPOL_SPLINE')
    ENDIF
    ! Check whether already allocated
    IF (IS_ALLOCATED(SELF)) THEN
        PRINT '(A)', '**WARNING(ALLOCATE_INTERPOL_SPLINE): already allocated, will deallocate first'
        !CALL SELF%DEALLOCATE()
        CALL INTERPOL_SPLINE_DEALLOCATE(SELF)
    ENDIF
    !PRINT *, '**DEBUG(ALLOCATE_INTERPOL_SPLINE): ', ALLOCATED(SELF%X), ALLOCATED(SELF%Y), ALLOCATED(SELF%DY), ALLOCATED(SELF%DDY),ALLOCATED(SELF%W)
    !PRINT *, '**DEBUG(ALLOCATE_INTERPOL_SPLINE): ', SIZE(SELF%X)
    ! Check a sensible number of points was requested
    IF (NUM_POINTS .LT. 1) THEN
        WRITE(*,'(A,I0,A)') '**WARNING(ALLOCATE_INTERPOL_SPLINE): allocating spline with ', NUM_POINTS, ' knots'
    ENDIF
    ! Allocate
    ALLOCATE(   SELF%X(NUM_POINTS),SELF%Y(NUM_POINTS),SELF%DY(NUM_POINTS),SELF%DDY(NUM_POINTS),         &
                SELF%W(NUM_POINTS,DEGREE+1),    &
                STAT=IERR)
    ! Set N and K
    SELF%N = NUM_POINTS
    SELF%K = DEGREE
    !
    SELF%LAST_SEGMENT_USED = 1
    ! Error case
    IF (IERR .NE. 0) THEN
        WRITE(*,'(A,I3,1X,I3)')  '**ERROR(ALLOCATE_INTERPOL_SPLINE): error in memory allocation; num_points, degree = ',     &
                                NUM_POINTS, DEGREE
        CALL TERMINATE('Fatal error in ALLOCATE_INTERPOL_SPLINE')
    ENDIF
    ! Allocate memory for local variables used when solving the spline
    ALLOCATE(SELF%M(SELF%N-2,SELF%N-2),SELF%DDV(SELF%N-2),SELF%RHS(SELF%N-2),STAT=IERR)
    IF (IERR .NE. 0) THEN
        WRITE(*,'(A)') '**ERROR(ALLOCATE_INTERPOL_SPLINE): memory allocation failed for local variables'
        CALL TERMINATE('Fatal error in ALLOCATE_INTERPOL_SPLINE')
    ENDIF
    ! Initialise all the arrays
    SELF%X = 0.0D0
    SELF%Y = 0.0D0
    SELF%DY = 0.0D0
    SELF%DDY = 0.0D0
    SELF%W = 0.0D0
    SELF%M = 0.0D0
    SELF%DDV = 0.0D0
    SELF%RHS = 0.0D0
    ! Set allocated flag
    SELF%IN_MEMORY = .TRUE.

END SUBROUTINE ALLOCATE_INTERPOL_SPLINE

!>  \brief  Use the GSL to solve (find the roots of) a quadratic equation
!!
!!  The equation is: a x^2 + b x + c = 0. The number of roots is returned. The roots themselves are returned in the ROOTS array, if there's only one root, it'll be the first element of the array.
INTEGER FUNCTION GSL_SOLVE_QUADRATIC(A,B,C,ROOTS,XMIN,XMAX) RESULT(NUM_ROOTS)
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE
    ! Arguments
    REAL(KIND=8),               INTENT(IN)  ::  A,B,C
    REAL(KIND=8),               INTENT(OUT) ::  ROOTS(2)
    REAL(KIND=8),   OPTIONAL,   INTENT(IN)  ::  XMIN,XMAX       !<  The minimum and maximum values allowed for the roots to be returned
    ! Private variables
    REAL(KIND=FGSL_DOUBLE)      ::  AA,BB,CC
    LOGICAL                     ::  ROOTS_IN_RANGE(2)
    LOGICAL,    PARAMETER       ::  DEBUG   =   .FALSE.
    ! Start work
    ROOTS_IN_RANGE = .TRUE.
    AA = A
    BB = B
    CC = C
    NUM_ROOTS = FGSL_POLY_SOLVE_QUADRATIC(AA,BB,CC,ROOTS(1),ROOTS(2))

    IF (DEBUG) THEN
        WRITE(*,'(A,I0)') '**debug(GSL_SOLVE_QUADRATIC): num_roots = ', NUM_ROOTS
        WRITE(*,'(A,2(F8.2,1X))') '**debug(GSL_SOLVE_QUADRATIC): roots = ', ROOTS
        IF (PRESENT(XMIN)) WRITE(*,'(A,F8.2)') '**debug(GSL_SOLVE_QUADRATIC): xmin = ', XMIN
        IF (PRESENT(XMAX)) WRITE(*,'(A,F8.2)') '**debug(GSL_SOLVE_QUADRATIC): xmax = ', XMAX
    ENDIF


    ! Now we check whether the roots are within range
    IF (NUM_ROOTS .EQ. 0) THEN
        ROOTS_IN_RANGE = .FALSE.
    ELSEIF (NUM_ROOTS .EQ. 1) THEN
        ROOTS_IN_RANGE = [.TRUE.,.FALSE.]
    ELSEIF (NUM_ROOTS .EQ. 2) THEN
        ROOTS_IN_RANGE = .TRUE.
    ENDIF

    ! Check the second root
    IF (NUM_ROOTS .EQ. 2) THEN
        IF (PRESENT(XMIN)) THEN
            IF (ROOTS(2) .LT. XMIN) ROOTS_IN_RANGE(2) = .FALSE.
        ENDIF
        IF (PRESENT(XMAX)) THEN
            IF (ROOTS(2) .GT. XMAX) ROOTS_IN_RANGE(2) = .FALSE.
        ENDIF
    ENDIF

    ! Check the first root
    IF (NUM_ROOTS .GE. 1) THEN
        IF (PRESENT(XMIN)) THEN
            IF (ROOTS(1) .LT. XMIN) ROOTS_IN_RANGE(1) = .FALSE.
        ENDIF
        IF (PRESENT(XMAX)) THEN
            IF (ROOTS(1) .GT. XMAX) ROOTS_IN_RANGE(1) = .FALSE.
        ENDIF
    ENDIF

    ! The final number of roots
    NUM_ROOTS = COUNT(ROOTS_IN_RANGE)

    ! If the only root in range is second one, move it to the first position
    IF (NUM_ROOTS .EQ. 1 .AND. ROOTS_IN_RANGE(2)) THEN
        ROOTS(1) = ROOTS(2)
        ROOTS_IN_RANGE = [.TRUE.,.FALSE.]
    ENDIF

END FUNCTION GSL_SOLVE_QUADRATIC

!>  \brief  Use the GSL to solve (find a root of) a cubic polynomial
!!
!!  The  equation is: a x^3 + b  x^2 + c  x + d  = y
!!  the GSL wants   :   x^3 + aa x^2 + bb x + cc = 0
FUNCTION GSL_SOLVE_CUBIC_POLYNOMIAL(A,B,C,D,Y,XMIN,XMAX) RESULT(X)
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    USE, INTRINSIC :: IEEE_ARITHMETIC, ONLY : IEEE_IS_NAN
    IMPLICIT NONE
    ! Arguments
    REAL(KIND=8),   INTENT(IN)  ::  A,B,C,D,Y
    REAL(KIND=8),   INTENT(IN)  ::  XMIN,XMAX   !<  Minimum and maximum values allowed/expected for the solution
    ! Result
    REAL(KIND=8)    ::  X
    ! Private variables
    INTEGER ::  NUM_ROOTS
    REAL(KIND=FGSL_DOUBLE)      ::  TMPX(3)
    !COMPLEX(KIND=FGSL_DOUBLE)  ::  TMPZ(3)
    REAL(KIND=FGSL_DOUBLE)      ::  AA,BB,CC
    LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
    LOGICAL                     ::  ROOTS_IN_RANGE(3), ROOTS_IN_RANGE_WITH_TOLERANCE(3)
    REAL,           PARAMETER   ::  RANGE_TOLERANCE = .05
    INTEGER                     ::  I
    REAL(KIND=8),   PARAMETER   ::  SMALL_NUMBER = 0.000000000001
    LOGICAL                     ::  CUBIC
    LOGICAL                     ::  PRINT_ERROR, PRINT_ERROR_TOL
    ! Start work

    TMPX = -999999.9999
    PRINT_ERROR = .FALSE.
    PRINT_ERROR_TOL = .FALSE.
    ROOTS_IN_RANGE = .FALSE.
    ROOTS_IN_RANGE_WITH_TOLERANCE = .FALSE.

    IF (A .NE. 0.0D0) THEN
        CUBIC = .TRUE.
        AA = B/A
        BB = C/A
        CC = (D-Y)/A
        ! Solve the equation
        NUM_ROOTS =  FGSL_POLY_SOLVE_CUBIC(AA,BB,CC,TMPX(1),TMPX(2),TMPX(3))
        ! Diagnostics
        IF (DEBUG) THEN
            WRITE(*,'(A,5(F0.8,1X))') '**debug(GSL_SOLVE_CUBIC_POLYNOMIAL): A,B,C,D,Y = ', A,B,C,D,Y
            WRITE(*,'(A,3(F0.8,1X))') '**debug(GSL_SOLVE_CUBIC_POLYNOMIAL): AA, BB, CC = ', AA, BB, CC
            WRITE(*,'(A,3(F0.8,1X))') '**debug(GSL_SOLVE_CUBIC_POLYNOMIAL): TMPX = ', TMPX
            !WRITE(*,'(A,3(2(F0.3,X),X))') '**debug(GSL_SOLVE_CUBIC_POLYNOMIAL): TMPZ = ', TMPZ
        ENDIF
    ELSE
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(GSL_SOLVE_CUBIC_POLYNOMIAL): actually dealing with a quadratic'
        ENDIF
        CUBIC = .FALSE.
        ! If A == 0.0, we are dealing with a quadratic equation
        AA = B
        BB = C
        CC = (D-Y)
        ! Solve the equation
        NUM_ROOTS =  FGSL_POLY_SOLVE_QUADRATIC(AA,BB,CC,TMPX(1),TMPX(2))
    ENDIF



    !
    IF (NUM_ROOTS .EQ. 1) THEN
        X = TMPX(1)
    ELSE IF (.NOT. CUBIC .AND. NUM_ROOTS .EQ. 2) THEN
        ! We found two solutions to the quadratic equation
        ! Are they within range?
        ROOTS_IN_RANGE(1:2) = TMPX(1:2) .GE. XMIN .AND. TMPX(1:2) .LE. XMAX
        IF (COUNT(ROOTS_IN_RANGE(1:2)) .EQ. 1) THEN
            ! Only one of the two solutions were within range
            DO I=1,2
                IF (ROOTS_IN_RANGE(I)) THEN
                    X = TMPX(I)
                ENDIF
            ENDDO
        ELSE IF (COUNT(ROOTS_IN_RANGE(1:2)) .EQ. 0) THEN
            ! None of the solutions were within the range, so we try to see if they're within the range with some added tolerance
            ROOTS_IN_RANGE_WITH_TOLERANCE(1:2) =        TMPX(1:2) .GE. XMIN-(XMAX-XMIN)*RANGE_TOLERANCE &
                                                .AND.   TMPX(1:2) .LE. XMAX+(XMAX-XMIN)*RANGE_TOLERANCE
            IF(COUNT(ROOTS_IN_RANGE_WITH_TOLERANCE(1:2)) .EQ. 1) THEN
                ! Great, only one of them was within range with tolerance
                DO I=1,2
                    IF (ROOTS_IN_RANGE_WITH_TOLERANCE(I)) THEN
                        X = TMPX(I)
                    ENDIF
                ENDDO
            ELSE
                ! OK, we give up and printout information
                PRINT_ERROR_TOL = .TRUE.
            ENDIF
        ELSE IF (COUNT(ROOTS_IN_RANGE(1:2)) .EQ. 2) THEN
            PRINT_ERROR = .TRUE.
        ELSE
            WRITE(*,'(A,I0)') '**ERROR(GSL_SOLVE_CUBIC_POLYNOMIAL): unexpected number of roots found: ', COUNT(ROOTS_IN_RANGE(1:2))
            CALL TERMINATE('Fatal error in GSL_SOLVE_CUBIC_POLYNOMIAL')
        ENDIF
    ELSE IF (CUBIC .AND. NUM_ROOTS .EQ. 3) THEN
        ! We found three solutions to the cubic
        ROOTS_IN_RANGE = TMPX .GE. XMIN .AND. TMPX .LE. XMAX
        ! We need to look for the correct root. Chances are only one of them will be within tolerable values of X
        IF (COUNT(ROOTS_IN_RANGE) .EQ. 1) THEN
            DO I=1,3
                IF (ROOTS_IN_RANGE(I)) THEN
                    X = TMPX(I)
                ENDIF
            ENDDO
        ELSE IF (COUNT(ROOTS_IN_RANGE) .EQ. 0) THEN
            ROOTS_IN_RANGE_WITH_TOLERANCE =       TMPX .GE. XMIN-(XMAX-XMIN)*RANGE_TOLERANCE  &
                                            .AND. TMPX .LE. XMAX+(XMAX-XMIN)*RANGE_TOLERANCE
            IF (COUNT(ROOTS_IN_RANGE_WITH_TOLERANCE) .EQ. 1) THEN
                DO I=1,3
                    IF (ROOTS_IN_RANGE_WITH_TOLERANCE(I)) THEN
                        X = TMPX(I)
                    ENDIF
                ENDDO
            ELSE
                PRINT_ERROR_TOL = .TRUE.
            ENDIF
        ELSE
            PRINT_ERROR = .TRUE.
        ENDIF
    ELSE
        WRITE(*,'(A,I0)') '**ERROR(GSL_SOLVE_CUBIC_POLYNOMIAL): unexpected number of roots: ', NUM_ROOTS
        WRITE(*,'(A,L1)') '**ERROR(GSL_SOLVE_CUBIC_POLYNOMIAL): cubic? ', CUBIC
        CALL TERMINATE('Fatal error in GSL_SOLVE_CUBIC_POLYNOMIAL')
    ENDIF

    ! Printout errors
    IF (PRINT_ERROR) THEN
        WRITE(*,'(A,I0,A,3(F0.8,1X))')  '**ERROR(GSL_SOLVE_CUBIC_POLYNOMIAL): ', NUM_ROOTS, ' roots were returned: ', TMPX
        WRITE(*,'(A,I0,A,2(F0.8,1X))')  '**ERROR(GSL_SOLVE_CUBIC_POLYNOMIAL): and ', COUNT(ROOTS_IN_RANGE), &
                                        ' of them were found to be within range: ', XMIN, XMAX
        WRITE(*,'(A,5(F0.8,1X))') '**debug(GSL_SOLVE_CUBIC_POLYNOMIAL): A,B,C,D,Y = ', A,B,C,D,Y
        WRITE(*,'(A,3(F0.8,1X))') '**debug(GSL_SOLVE_CUBIC_POLYNOMIAL): AA, BB, CC = ', AA, BB, CC
        WRITE(*,'(A,L1)') '**ERROR(GSL_SOLVE_CUBIC_POLYNOMIAL): cubic? ', CUBIC
        CALL TERMINATE('Fatal error in GSL_SOLVE_CUBIC_POLYNOMIAL')
    ENDIF
    IF (PRINT_ERROR_TOL) THEN
        WRITE(*,'(A,I0,A,3(F0.8,1X))')  '**ERROR(GSL_SOLVE_CUBIC_POLYNOMIAL): ', NUM_ROOTS, ' roots were returned: ', TMPX
        WRITE(*,'(A,I0,A,2(F0.8,1X))')  '**ERROR(GSL_SOLVE_CUBIC_POLYNOMIAL): and ', COUNT(ROOTS_IN_RANGE_WITH_TOLERANCE), &
                                        ' of them were found to be within range with tolerance: ', &
                                        XMIN-(XMAX-XMIN)*RANGE_TOLERANCE, XMAX+(XMAX-XMIN)*RANGE_TOLERANCE
        WRITE(*,'(A,5(F0.8,1X))') '**debug(GSL_SOLVE_CUBIC_POLYNOMIAL): A,B,C,D,Y = ', A,B,C,D,Y
        WRITE(*,'(A,3(F0.8,1X))') '**debug(GSL_SOLVE_CUBIC_POLYNOMIAL): AA, BB, CC = ', AA, BB, CC
        WRITE(*,'(A,L1)') '**ERROR(GSL_SOLVE_CUBIC_POLYNOMIAL): cubic? ', CUBIC
        CALL TERMINATE('Fatal error in GSL_SOLVE_CUBIC_POLYNOMIAL')
    ENDIF

    ! Check for NaN
    IF (IEEE_IS_NAN(X)) THEN
        WRITE(*,'(A)') '**ERROR(GSL_SOLVE_CUBIC_POLYNOMIAL): found NaN'
        CALL TERMINATE('Fatal error in GSL_SOLVE_CUBIC_POLYNOMIAL')
    ENDIF
END FUNCTION GSL_SOLVE_CUBIC_POLYNOMIAL

FUNCTION GSL_INTERPOL_SPLINE_EVAL(SPLINE,X) RESULT(Y)
    IMPLICIT NONE
    ! Arguments
    TYPE(GSL_SPLINE),   INTENT(INOUT)   ::  SPLINE
    REAL,               INTENT(IN)      ::  X
    REAL                                ::  Y
    ! Private variables
    REAL(FGSL_DOUBLE)   ::  XX
    ! Start work
    XX = X
    Y = FGSL_SPLINE_EVAL(SPLINE%SPLINE,XX,SPLINE%ACC)
END FUNCTION GSL_INTERPOL_SPLINE_EVAL

SUBROUTINE GSL_INTERPOL_SPLINE_INIT(SPLINE,NUM_POINTS)
    IMPLICIT NONE
    ! Arguments
    TYPE(GSL_SPLINE),   INTENT(INOUT)   ::  SPLINE
    INTEGER,            INTENT(IN)      ::  NUM_POINTS
    ! Private variables
    INTEGER(FGSL_SIZE_T)    ::  NNUM_POINTS
    ! Start work
    IF (.NOT. (ALLOCATED(SPLINE%X) .AND. ALLOCATED(SPLINE%Y))) THEN
        PRINT *, '**ERROR(GSL_INTERPOL_SPLINE_INIT): X and Y arrays are not allocated'
        STOP 'Fatal error in GSL_INTERPOL_SPLINE_INIT'
    ENDIF
    IF (SIZE(SPLINE%X) .NE. SIZE(SPLINE%Y)) THEN
        PRINT *, '**ERROR(GSL_INTERPOL_SPLINE_INIT): X and Y arrays are not allocated with same sizes'
        STOP 'Fatal error in GSL_INTERPOL_SPLINE_INIT'
    ENDIF
    IF (SIZE(SPLINE%X) .NE. NUM_POINTS) THEN
        PRINT *, '**WARNING(GSL_INTERPOL_SPLINE_INIT): NUM_POINTS (', NUM_POINTS, ') .NE. SIZE(X) (', SIZE(SPLINE%X), ')'
        STOP 'Fatal error in GSL_INTERPOL_SPLINE_INIT'
    ENDIF
    ! Convert integer to FGSL type
    NNUM_POINTS = NUM_POINTS
    ! Initialise the spline
    SPLINE%ACC = FGSL_INTERP_ACCEL_ALLOC()
    SPLINE%SPLINE = FGSL_SPLINE_ALLOC(FGSL_INTERP_CSPLINE,NNUM_POINTS)
    SPLINE%STATUS = FGSL_SPLINE_INIT(SPLINE%SPLINE,SPLINE%X,SPLINE%Y,NNUM_POINTS)
END SUBROUTINE GSL_INTERPOL_SPLINE_INIT

SUBROUTINE GSL_INTERPOL_SPLINE_FREE(SPLINE)
    ! Arguments
    TYPE(GSL_SPLINE),   INTENT(INOUT)   ::  SPLINE
    ! Start work
    CALL FGSL_SPLINE_FREE(SPLINE%SPLINE)
    CALL FGSL_INTERP_ACCEL_FREE(SPLINE%ACC)
END SUBROUTINE GSL_INTERPOL_SPLINE_FREE

SUBROUTINE GSL_CSPLINE_TEST
    use fgsl
    implicit none
    integer :: i,index
    real(fgsl_double) :: x(7),y(7),dy(7),xa
    type(fgsl_interp_accel) :: acc
    type(fgsl_spline) :: spline
    character(len=fgsl_strmax) :: name
    integer(fgsl_int) :: status
    real(fgsl_double) :: ra, rda, rd2a, ri
    integer, parameter :: nstep = 100
    type(interpol_spline) :: home_spline
    !integer(fgsl_int)  ::  npoints
    integer(fgsl_size_t)    ::  npoints

    ! build the signal
    print *, '#m=0,S=2'
    !npoints = 7
    !do i=1,npoints
    !   x(i) = i + 0.5 * sin(real(i))
    !   y(i) = i + cos (real(i)**2)
    !end do
    npoints = 7
    x  = [1.0,2.0,3.0,4.6,5.5,6.8,7.4]
    y  = [1.0,3.7,-4.0,3.0,14.1,11.0,15.60]
    dy = [0.0,0.0,10.0,-5.0,0.0 ,0.0,0.0]


    ! print out the signal
    do i=1,npoints
        print '(2g15.5)', x(i), y(i)
    end do

    !
    acc = fgsl_interp_accel_alloc()
    spline = fgsl_spline_alloc(fgsl_interp_cspline,npoints)
    name = fgsl_spline_name(spline)
    status = fgsl_spline_init(spline,x,y,npoints)
    !print *, 'FGSL spline init status = ', status

    print *, '#m=2,S=0'
    do i=1,nstep
        xa = x(1) + (x(npoints)-x(1))*i/nstep
        ra = fgsl_spline_eval(spline, xa, acc)
        rda = fgsl_spline_eval_deriv(spline, xa, acc)
        rd2a = fgsl_spline_eval_deriv2(spline, xa, acc)
        ri = fgsl_spline_eval_integ(spline, 0.0_fgsl_double, xa, acc)
        !print *, i, ra, rda, rd2a, ri
        print '(2g15.5)', xa, ra
    enddo
    index = fgsl_spline_min_size(spline)
    !print *, 'index = ', index
    call fgsl_spline_free(spline)

    ! Now do the same with the homegrown spline object to see whether we get the same
    call interpol_spline_allocate(home_spline,int(npoints),3)
    call spline_set_type(home_spline,'NATURAL')
    home_spline%x = x
    home_spline%y = y
    home_spline%dy = dy
    call solve_interpol_spline(home_spline)
    print *, '#m=1,S=0'
    do i=1,nstep
        xa = x(1) + (x(npoints)-x(1))*i/nstep
        ra = interpol_spline_eval2(home_spline,xa)
        print '(2g15.5)', xa, ra
    enddo
END SUBROUTINE GSL_CSPLINE_TEST


END MODULE SPLINES
