!>  \brief  Defines the THREED class, which handles 3D reconstruction from a set of images,
!!          including appropriate filters etc.
!!
MODULE THREEDS
    USE THREEDS_BASE
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE

    CONTAINS

    !>  \brief  Test functionality of the 3D reconstruction module
    SUBROUTINE THREED_UNIT_TESTS()
        IMPLICIT NONE
        CALL THREED_UNIT_TEST_1()
        CALL THREED_UNIT_TEST_2()
    END SUBROUTINE THREED_UNIT_TESTS

    !>  \brief  Single slice insertion with no interpolation
    SUBROUTINE THREED_UNIT_TEST_1()
        USE MASKS, ONLY : MASK_ADD_COSINE_FALLOFF
        USE CTFS, ONLY : CTF_PARAMETERS_INIT, CTF_PARAMETERS, CTF_ANG_TO_PIX
        USE IMAGES_CTF, ONLY : IMAGE_APPLY_CTF
        USE IMAGES_ARITHM, ONLY : CIRC_MASK
        USE IMAGES_COMPARISON, ONLY : CORRELATION
        USE EULERANGLES
        use images
        IMPLICIT NONE
        ! Private variables
        TYPE(IMAGE) ::  TEST_2D, TEST_2D_ORIGINAL, TEST_3D, CHECK_2D, MASK
        INTEGER     ::  ORI(3), TEST_DIM
        TYPE(CTF_PARAMETERS) :: CTF
        TYPE(EULER) ::  ANGLES
        TYPE(THREED)    ::  SELF
        REAL,   PARAMETER   ::  PIXEL_SIZE = 1.0
        REAL    ::  CC
        ! Start work

        WRITE(*,'(A)') '**info(THREED_UNIT_TEST_1): entering'

        TEST_DIM = 128

        ! Prepare a CTF object
        CALL CTF_PARAMETERS_INIT(CTF,AKV=200.0,CS=2.0,AMPL_CONT=0.07,DFMID1=1.3)
        CALL CTF_ANG_TO_PIX(CTF,PIXEL_SIZE=PIXEL_SIZE)

        ! Create a test pattern (a cosine bell) and apply the microscope CTF
        TEST_2D%logical_dimensions = [TEST_DIM,TEST_DIM,1]
        CALL IMAGE_ALLOCATE(TEST_2D)
        TEST_2D%real_values = 0.0
        if (test_2d%is_in_real_space .and. test_2d%origin_at_center) then
            ori = test_2d%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif
        TEST_2D%real_values(ORI(1),ORI(2),ORI(3)) = 1.0
        CALL MASK_ADD_COSINE_FALLOFF(TEST_2D,5.0)
        CALL WRITE_TO_DISK(TEST_2D,'threed_unit_test_1_test_2d.mrc')
        TEST_2D_ORIGINAL = TEST_2D
        CALL IMAGE_APPLY_CTF(TEST_2D,CTF)
        IF (test_2d%IsComplex()) CALL BWD_FT(TEST_2D)
        CALL WRITE_TO_DISK(TEST_2D,'threed_unit_test_1_test_2d_ctf.mrc')



        ! Prepare Euler angles
        ANGLES%PHI = 0.0
        ANGLES%THETA = 0.0
        ANGLES%PSI = 0.0

        ! Prepare a 3D reconstruction
        SELF%logical_dimensions = [TEST_DIM,TEST_DIM,TEST_DIM]
        CALL THREED_INIT(SELF)

        ! Add slice(s) to 3D reconstruction
        IF (test_2d%IsReal()) CALL FWD_FT(TEST_2D)
        IF (TEST_2D%ORIGIN_AT_CENTER) CALL SHIFT_REAL_SPACE_ORIGIN(TEST_2D)
        CALL THREED_ADD_SLICE(SELF,TEST_2D,CTF,ANGLES)



        ! Write 3D object to disk
        CALL THREED_WRITE_TO_DISK(SELF,'threed_unit_test_1_threed.mrc')

        ! Finalise the 3D reconstruction
        CALL THREED_EVAL_RECONSTRUCTION(SELF,TEST_3D)
        CALL WRITE_TO_DISK(TEST_3D,'threed_unit_test_1_test_3d.mrc')


        ! Check we got the expected result: extract the central slice in real space and compare it to our test pattern
        CALL IMAGE_GET_SLICE(TEST_3D,CHECK_2D,3,ORI(1))
        CALL WRITE_TO_DISK(CHECK_2D,'threed_unit_test_1_check_2d.mrc')
        CALL IMAGE_ALLOCATE(MASK,TEST_2D_ORIGINAL)
        MASK%real_values = 1.0
        CALL CIRC_MASK(MASK,RAD=TEST_DIM/3.0)
        CC = CORRELATION(TEST_2D_ORIGINAL,CHECK_2D,MASK,NORMALISE=.TRUE.)
        WRITE(*,'(A,F0.4)') '**info(THREED_UNIT_TEST_1): this normalised cross-correlation score should be very high: ', CC
        IF (CC .LT. 0.95) THEN
            WRITE(*,'(A)') '**ERROR(THREED_UNIT_TEST_1): failed unit test'
            CALL TERMINATE('Fatal error in THREED_UNIT_TESTS')
        ENDIF

        WRITE(*,'(A)') '**info(THREED_UNIT_TEST_1): test passed'

    END SUBROUTINE THREED_UNIT_TEST_1

    !>  \brief  Many slices inserted, requiring interpolation. Comparison of interpolation-free direction
    SUBROUTINE THREED_UNIT_TEST_2()
        USE MASKS, ONLY : MASK_ADD_COSINE_FALLOFF
        USE CTFS, ONLY : CTF_PARAMETERS_INIT, CTF_PARAMETERS, CTF_ANG_TO_PIX
        USE IMAGES_CTF, ONLY : IMAGE_APPLY_CTF
        USE IMAGES_ARITHM, ONLY : CIRC_MASK
        USE IMAGES_COMPARISON, ONLY : CORRELATION
        USE EULERANGLES
        use images
        IMPLICIT NONE
        ! Private variables
        TYPE(IMAGE) ::  TEST_2D, TEST_2D_ORIGINAL, TEST_3D, CHECK_2D, MASK
        INTEGER     ::  ORI(3), TEST_DIM
        TYPE(CTF_PARAMETERS) :: CTF
        TYPE(EULER) ::  ANGLES
        TYPE(THREED)    ::  SELF
        REAL,   PARAMETER   ::  PIXEL_SIZE = 1.0
        REAL    ::  CC
        INTEGER, PARAMETER ::  NUMBER_OF_INSERTIONS = 100
        INTEGER ::  CURRENT_INSERTION
        ! Start work

        WRITE(*,'(A)') '**info(THREED_UNIT_TEST_2): entering'

        TEST_DIM = 128

        ! Prepare a CTF object
        CALL CTF_PARAMETERS_INIT(CTF,AKV=200.0,CS=2.0,AMPL_CONT=0.07,DFMID1=1.3)
        CALL CTF_ANG_TO_PIX(CTF,PIXEL_SIZE=PIXEL_SIZE)

        ! Create a test pattern (a cosine bell) and apply the microscope CTF
        TEST_2D%logical_dimensions = [TEST_DIM,TEST_DIM,1]
        CALL IMAGE_ALLOCATE(TEST_2D)
        TEST_2D%real_values = 0.0
        if (test_2d%is_in_real_space .and. test_2d%origin_at_center) then
            ori = test_2d%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif
        TEST_2D%real_values(ORI(1),ORI(2),ORI(3)) = 1.0
        CALL MASK_ADD_COSINE_FALLOFF(TEST_2D,5.0)
        CALL WRITE_TO_DISK(TEST_2D,'threed_unit_test_2_test_2d.mrc')
        TEST_2D_ORIGINAL = TEST_2D
        CALL IMAGE_APPLY_CTF(TEST_2D,CTF)
        IF (test_2d%IsComplex()) CALL BWD_FT(TEST_2D)
        CALL WRITE_TO_DISK(TEST_2D,'threed_unit_test_2_test_2d_ctf.mrc')



        ! Prepare Euler angles
        ANGLES%PHI = 0.0
        ANGLES%THETA = 0.0
        ANGLES%PSI = 0.0

        ! Prepare a 3D reconstruction
        SELF%logical_dimensions = [TEST_DIM,TEST_DIM,TEST_DIM]
        CALL THREED_INIT(SELF)

        ! Add slice(s) to 3D reconstruction
        IF (test_3d%IsReal()) CALL FWD_FT(TEST_2D)
        IF (TEST_2D%ORIGIN_AT_CENTER) CALL SHIFT_REAL_SPACE_ORIGIN(TEST_2D)
        DO CURRENT_INSERTION=1,NUMBER_OF_INSERTIONS
            ANGLES%THETA = REAL(CURRENT_INSERTION) / REAL(NUMBER_OF_INSERTIONS) * 180. + 1.23
            WRITE(*,'(A,I0,A,I0,A,F0.2)') '**info(THREED_UNIT_TEST_2): inserting slice ', CURRENT_INSERTION, ' of ', &
                                            NUMBER_OF_INSERTIONS, ' with theta = ', ANGLES%THETA
            CALL THREED_ADD_SLICE(SELF,TEST_2D,CTF,ANGLES)
        ENDDO



        ! Write 3D object to disk
        CALL THREED_WRITE_TO_DISK(SELF,'threed_unit_test_2_threed.mrc')

        ! Finalise the 3D reconstruction
        CALL THREED_EVAL_RECONSTRUCTION(SELF,TEST_3D)
        CALL WRITE_TO_DISK(TEST_3D,'threed_unit_test_2_test_3d.mrc')


        ! Check we got the expected result: extract the central slice in real space and compare it to our test pattern
        CALL IMAGE_GET_SLICE(TEST_3D,CHECK_2D,3,ORI(1))
        CALL WRITE_TO_DISK(CHECK_2D,'threed_unit_test_2_check_2d.mrc')
        CALL IMAGE_ALLOCATE(MASK,TEST_2D_ORIGINAL)
        MASK%real_values = 1.0
        CALL CIRC_MASK(MASK,RAD=TEST_DIM/3.0)
        CC = CORRELATION(TEST_2D_ORIGINAL,CHECK_2D,MASK,NORMALISE=.TRUE.)
        WRITE(*,'(A,F0.4)') '**info(THREED_UNIT_TEST_2): this normalised cross-correlation score should be very high: ', CC
        IF (CC .LT. 0.99) THEN
            WRITE(*,'(A)') '**ERROR(THREED_UNIT_TEST_2): failed unit test'
            CALL TERMINATE('Fatal error in THREED_UNIT_TESTS')
        ENDIF

        WRITE(*,'(A)') '**info(THREED_UNIT_TEST_2): test passed'

    END SUBROUTINE THREED_UNIT_TEST_2

    !>  \brief  Return a real 3D volume reconstruction from the THREED object - i.e. apply weighting & correct for the interpolation
    !!          artefacts (a la CORRECT3D / gridding)
    SUBROUTINE THREED_EVAL_RECONSTRUCTION(  SELF,IMG, &
                                            WIENER_FILTER,REC_NUM,PSSNR_FUDGE,PRESERVE_HIGH_FREQUENCIES,    &
                                            CYLINDER_MASK,SKIP_FOM_FILTER,FSC_VOLUMES,SKIP_FSC,IMG_AMPL,SKIP_AMPL)
        USE IMAGES_ARITHM, ONLY : CIRC_MASK
        USE IMAGEFILES ! for debug
        use images
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),               INTENT(INOUT)   ::  SELF
        TYPE(IMAGE),                INTENT(INOUT)   ::  IMG                         !<  Output reconstruction image
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  WIENER_FILTER               !<  Whether a Wiener filter should be used. Defaults to true. If not, a simple division by CTF^2 is performed.
        INTEGER,        OPTIONAL,   INTENT(IN)      ::  REC_NUM                     !<  Which internal reconstruction should be used. If not specified and there are two internal reconstructions, the sum will be returned.
        REAL,           OPTIONAL,   INTENT(IN)      ::  PSSNR_FUDGE                 !<  Scaling fudge factor for the PSSNR term
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  PRESERVE_HIGH_FREQUENCIES   !<  Even if they are dominated by noise (i.e. the SSNR is low), do not zero high frequencies.
                                                                                    !!  This will probably lead to more noisy reconstructions, but may be used if the Wiener filter is thought to be too stringent. Defaults to false.
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  CYLINDER_MASK               !<  Cylindrical, rather than spherical, mask
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  SKIP_FOM_FILTER             !<  Do not apply FOM filter
        TYPE(IMAGE),    OPTIONAL,   INTENT(INOUT)   ::  FSC_VOLUMES(2)              !<  If present, the volumes between which the FSC is computed will be returned
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  SKIP_FSC                    !<  Do not compute FSC
        TYPE(IMAGE),    OPTIONAL,   INTENT(INOUT)   ::  IMG_AMPL                    !<  Output amplitudes reconstruction image
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  SKIP_AMPL                   !<  Do not evaluate the amplitude reconstruction
        ! Private variables
        LOGICAL,        PARAMETER       ::  DEBUG   =   .FALSE.
        LOGICAL     ::  WWIENER_FILTER
        INTEGER     ::  RREC_NUM    ! 0 indicates sum of both
        LOGICAL     ::  PPRESERVE_HIGH_FREQUENCIES
        LOGICAL     ::  SSKIP_FOM_FILTER
        LOGICAL     ::  SSKIP_FSC
        LOGICAL     ::  CCYLINDER_MASK
        LOGICAL     ::  SSKIP_AMPL
        ! Start work

        ! Will we be using a Wiener filter?
        WWIENER_FILTER = .TRUE.
        IF (PRESENT(WIENER_FILTER)) WWIENER_FILTER = WIENER_FILTER

        SSKIP_FOM_FILTER = .FALSE.
        IF (PRESENT(SKIP_FOM_FILTER)) SSKIP_FOM_FILTER = SKIP_FOM_FILTER

        SSKIP_FSC = .FALSE.
        IF (PRESENT(SKIP_FSC)) SSKIP_FSC = SKIP_FSC

        CCYLINDER_MASK = .FALSE.
        IF (PRESENT(CYLINDER_MASK)) CCYLINDER_MASK = CYLINDER_MASK

        SSKIP_AMPL = .FALSE.
        IF (PRESENT(SKIP_AMPL)) SSKIP_AMPL = SKIP_AMPL

        ! Which reconstruction are we asked to evaluate?
        IF (PRESENT(REC_NUM)) THEN
            RREC_NUM = REC_NUM
        ELSE
            IF (THREED_NUM_OF_RECS(SELF) .EQ. 2) THEN
                RREC_NUM = 0    ! we'll evaluate the sum of both
            ELSE
                RREC_NUM = 1    ! there is only one reconstruction available, that's the one we'll eval
            ENDIF
        ENDIF

        ! By default, we will not preserve high-frequencies if their SSNR is too low
        PPRESERVE_HIGH_FREQUENCIES = .FALSE.
        IF (PRESENT(PRESERVE_HIGH_FREQUENCIES)) PPRESERVE_HIGH_FREQUENCIES = PRESERVE_HIGH_FREQUENCIES

        ! Check that the reconstruction we are asked to evaluate exists
        IF (RREC_NUM .GT. THREED_NUM_OF_RECS(SELF)) THEN
            WRITE(*,'(A,I0,A,I0,A)') '**ERROR(THREED_EVAL_RECONSTRUCTION): was asked to evaluate reconstruction ', RREC_NUM, &
                                        ' yet there are only ', THREED_NUM_OF_RECS(SELF), ' reconstructions available'
            CALL TERMINATE('Fatal error in THREED_EVAL_RECONSTRUCTION. Programming error?')
        ELSE IF (RREC_NUM .EQ. 0 .AND. THREED_NUM_OF_RECS(SELF) .EQ. 1) THEN
            WRITE(*,'(2A,I0,A)') '**ERROR(THREED_EVAL_RECONSTRUCTION): was asked to evaluate the sum of',    &
                                ' internal reconstructions yet there is only ', &
                                    THREED_NUM_OF_RECS(SELF), ' reconstruction available'
            CALL TERMINATE('Fatal error in THREED_EVAL_RECONSTRUCTION. Programming error?')
        ENDIF

        !
        IF (THREED_NUM_OF_RECS(SELF) .EQ. 2) THEN
            IF (.NOT. THREED_TWO_RECS(SELF)) THEN
                WRITE(*,'(A)') '**ERROR(THREED_EVAL_RECONSTRUCTION): expecting, but did not find, two internal reconstructions'
                CALL TERMINATE('Fatal error in THREED_EVAL_RECONSTRUCTION')
            ENDIF
        ENDIF

        !
        IF (SELF%RECONSTRUCT_AMPLITUDES .AND. .NOT. PRESENT(IMG_AMPL) .AND. .NOT. SSKIP_AMPL) THEN
            WRITE(*,'(2A)') '**ERROR(THREED_EVAL_RECONSTRUCTION): was asked to reconstruct amplitudes, ',    &
                                'but not given image volume to write to'
            CALL TERMINATE('Fatal error in THREED_EVAL_RECONSTRUCTION')
        ENDIF

        ! Allocate memory for output
        CALL IMAGE_ALLOCATE(IMG,SELF%REC(1)%VOL_IMAGE)
        IF (SELF%RECONSTRUCT_AMPLITUDES .AND. PRESENT(IMG_AMPL)) THEN
            CALL IMAGE_ALLOCATE(IMG_AMPL,SELF%REC(1)%VOL_AMPL)
        ENDIF

        ! If the incoming threed object had no data, we return a blank volume
        IF (SUM(SELF%REC%NUM_CONTRIB_ELEMS) .EQ. 0) THEN
            PRINT '(A)', '**info(THREED_EVAL_RECONSTRUCTION): empty 3D reconstruction!'
            IMG%real_values = 0.0
            img%is_in_real_space = .true.
            IF (SELF%RECONSTRUCT_AMPLITUDES .AND. PRESENT(IMG_AMPL)) THEN
                IMG_AMPL%real_values = 0.0
                IMG_AMPL%is_in_real_space = .true.
            ENDIF
        ! The normal case, where the 3D reconstruction isn't blank
        ELSE

            ! Diag
            IF (DEBUG) THEN
                IF (SELF%GET2D_PUT3D) THEN
                    PRINT '(A)', '**debug(THREED_EVAL_RECONSTRUCTION): using GET2D_PUT3D logic'
                ELSE
                    PRINT '(A)', '**debug(THREED_EVAL_RECONSTRUCTION): using PUT3D logic'
                ENDIF
            ENDIF

            ! Debug dumps
            IF (DEBUG) THEN
                IMG = SELF%REC(1)%VOL_IMAGE
                CALL SHIFT_REAL_SPACE_ORIGIN(IMG)
                !CALL AMPL(IMG,TMPVOL)
                !CALL WRITE_TO_DISK(TMPVOL,'debug_before_3dctf_ampl.mrc')
                !TMPVOL%real_values = TMPVOL%real_values**2
                !CALL WRITE_TO_DISK(TMPVOL,'debug_before_3dctf_pow.mrc')
                CALL BWD_FT(IMG)
                CALL WRITE_TO_DISK(IMG,'debug_before_3dctf.mrc')
                IF (DEBUG) WRITE(*,'(A,L1)') '**debug(THREED_EVAL_RECONSTRUCTION): before 3dctf, volume origin at center? ', &
                                                IMG%ORIGIN_AT_CENTER
                CALL FWD_FT(IMG)
                !CALL SHIFT_REAL_SPACE_ORIGIN(IMG)
            ENDIF

            ! Prepare the CTF^2 volume
            IF (DEBUG) CALL WRITE_TO_DISK(SELF%REC(1)%VOL_CTF2,'debug_3dctf.mrc')

            !CALL CTFVOL_SHIFT(SELF%REC(1)%VOL_CTF2,CTF2_HALFVOL)
            !IF (DEBUG) THEN
            !    CALL WRITE_TO_DISK(CTF2_HALFVOL,'debug_3dctf2.mrc')
            !ENDIF

            ! Compute FSC
            IF (SELF%COMPUTE_FSC .AND. .NOT. SSKIP_FSC) THEN
                WRITE(*,'(A)') '**info(THREED_EVAL_RECONSTRUCTION): computing FSC'
                CALL THREED_UPDATE_FSC(SELF,FSC_VOLUMES=FSC_VOLUMES)
            ENDIF

            ! Apply Wiener filter
            IF (WWIENER_FILTER) THEN
                WRITE(*,'(A)') '**info(THREED_EVAL_RECONSTRUCTION): will apply a Wiener-like filter'
                CALL THREED_APPLY_WIENER_FILTER(SELF,IMG,REC_NUM=RREC_NUM,PSSNR_FUDGE=PSSNR_FUDGE,  &
                                                PRESERVE_HIGH_FREQUENCIES=PPRESERVE_HIGH_FREQUENCIES)
                IF (SELF%RECONSTRUCT_AMPLITUDES .AND. PRESENT(IMG_AMPL)) THEN
                    CALL THREED_APPLY_WIENER_FILTER(SELF,IMG_AMPL,REC_NUM=RREC_NUM,PSSNR_FUDGE=PSSNR_FUDGE, &
                                                    PRESERVE_HIGH_FREQUENCIES=PPRESERVE_HIGH_FREQUENCIES)
                ENDIF
            ELSE
                WRITE(*,'(A)') '**info(THREED_EVAL_RECONSTRUCTION): no Wiener filter - will divide by CTF^2 instead'
                CALL THREED_CTF2_DIVISION(SELF,IMG,REC_NUM=RREC_NUM)
                IF (SELF%RECONSTRUCT_AMPLITUDES .AND. PRESENT(IMG_AMPL)) THEN
                     CALL THREED_CTF2_DIVISION(SELF,IMG,REC_NUM=RREC_NUM)
                ENDIF
            ENDIF

            ! Apply FOM filter
            IF (SELF%FOM_FILTER .AND. .NOT. SSKIP_FOM_FILTER) THEN
                WRITE(*,'(A)') '**info(THREED_EVAL_RECONSTRUCTION): applying FOM filter'
                CALL THREED_APPLY_FOM_FILTER(SELF,IMG,PRINT_FILTER=.TRUE.)
                IF (SELF%RECONSTRUCT_AMPLITUDES) THEN
                    WRITE(*,'(A)') '**info(THREED_EVAL_RECONSTRUCTION): not applying FOM filter to amplitudes reconstruction'
                ENDIF
            ENDIF

            ! The volume should be in Fourier space
            IF (img%is_in_real_space) CALL FWD_FT(IMG)

            ! The Nyquist elements are not reliable, so we remove them
            CALL CIRC_MASK(IMG,0.5)

            ! Also, zero the average
            IMG%complex_values(1,1,1) = (0.,0.)

            ! The image is ready to be back-FT'ed, but first may need to be shifted
            IF (.NOT. IMG%ORIGIN_AT_CENTER) CALL SHIFT_REAL_SPACE_ORIGIN(IMG)
            CALL BWD_FT(IMG)

            ! Correct for falloff due to interpolation in reciprocal space
            WRITE(*,'(A)') '**info(THREED_EVAL_RECONSTRUCTION): will correct damping artefact due to interpolation kernel'
            IF (DEBUG) CALL WRITE_TO_DISK(IMG,'debug_before_correct_3d.mrc')
            IF (SELF%GET2D_PUT3D) THEN
                IF (DEBUG) PRINT '(A)', '**debug(THREED_EVAL_RECONSTRUCTION): correcting for interpolation kernel LINEAR RECT 2'
                CALL CORRECT_3D(IMG,SELF%INTERPOL_SCHEME%INTERPOLATION_INSTR_FUNCTION,  &
                                    SELF%INTERPOL_SCHEME%WINDOWING_INSTR_FUNCTION,    &
                                    SELF%INTERPOL_SCHEME%WINDOW_WIDTH,REVERSE=.FALSE.)
                IF (SELF%RECONSTRUCT_AMPLITUDES .AND. PRESENT(IMG_AMPL) .AND. .NOT. SSKIP_AMPL) THEN
                    CALL CORRECT_3D(IMG_AMPL,SELF%INTERPOL_SCHEME%INTERPOLATION_INSTR_FUNCTION, &
                                    SELF%INTERPOL_SCHEME%WINDOWING_INSTR_FUNCTION,    &
                                    SELF%INTERPOL_SCHEME%WINDOW_WIDTH,REVERSE=.FALSE.)
                ENDIF
            ELSE
                IF (DEBUG) PRINT '(A)', '**debug(THREED_EVAL_RECONSTRUCTION): correcting for appropriate interpolation kernel'
                ! Generalised CORRECT_3D
                CALL CORRECT_3D(IMG,SELF%INTERPOL_SCHEME%INTERPOLATION_INSTR_FUNCTION,  &
                                    SELF%INTERPOL_SCHEME%WINDOWING_INSTR_FUNCTION,    &
                                    SELF%INTERPOL_SCHEME%WINDOW_WIDTH,REVERSE=.FALSE.)
                IF (SELF%RECONSTRUCT_AMPLITUDES .AND. PRESENT(IMG_AMPL) .AND. .NOT. SSKIP_AMPL) THEN
                    CALL CORRECT_3D(IMG_AMPL,SELF%INTERPOL_SCHEME%INTERPOLATION_INSTR_FUNCTION, &
                                    SELF%INTERPOL_SCHEME%WINDOWING_INSTR_FUNCTION,    &
                                    SELF%INTERPOL_SCHEME%WINDOW_WIDTH,REVERSE=.FALSE.)
                ENDIF
            ENDIF


            ! Add a mask - correct_3d can leave funny artefacts, and we can flatten the solvent anyway. A cosine-edged sphere should do the trick.
            CALL COSINE_DISC(IMG,SELF%MASK_RAD,SELF%MASK_FALLOFF,WIPE_INPUT=.FALSE.,MASK_WITH_AVERAGE=.FALSE.,  &
                                CYLINDER=CCYLINDER_MASK .OR. SELF%POINT_SYM%HELICAL)
            IF (SELF%RECONSTRUCT_AMPLITUDES .AND. .NOT. SSKIP_AMPL) THEN
                WRITE(*,'(A)') '**info(THREED_EVAL_RECONSTRUCTION): not applying cosine-edge disc mask to amplitudes reconstruction'
            ENDIF
        ENDIF   ! End of test for blank reconstruction
    END SUBROUTINE THREED_EVAL_RECONSTRUCTION

    !>  \brief  Apply a Figure-Of-Merit (FOM) filter to the reconstruction, a la Frealign (flag FCREF) and Henderson & Rosenthal (2003)
    !!
    !!  We actually use Eq 11 of Sindelar & Grigorieff (2012), which is the square of Henderson & Rosenthal (2003). Apparently, this is
    !!  better suited to our use, where the FSC is calcuated between masked volumes
    subroutine threed_apply_fom_filter(self,img,print_filter)
        use images
        implicit none
        ! arguments
        type(threed),               intent(inout)   ::  self            !<  3D reconstruction object
        type(image),                intent(inout)   ::  img             !<  Evaluated 3d reconstruction volume (should already have been wiener filtered)
        logical,        optional,   intent(in)      ::  print_filter    !<  Print out filter values
        ! private variables
        logical,    parameter   ::  dnf(3) = (/.true.,.false.,.false./)
        integer                 ::  i,j,k, current_shell
        integer                 ::  last_printed_shell
        logical                 ::  pprint_filter, print_filter_value
        real                    ::  filter_value
        logical,    parameter   ::  use_sindelar_2012   =   .true.
        logical,    parameter   ::  oonce_at_zero_stay_there = .true.
        integer                 ::  first_shell_where_filter_is_zero
        integer                 ::  dont_zero_before_this_shell
        ! start work

        pprint_filter = .false.
        if (present(print_filter)) pprint_filter = print_filter
        last_printed_shell = -1

        ! update the fsc if necessary
        if (.not. self%fsc_is_up_to_date) then
            call threed_update_fsc(self)
        endif

        ! the image needs to be in fourier space
        if (img%is_in_real_space) call fwd_ft(img)

        if (pprint_filter) then
            write(*,'(a)') '**info(threed_apply_fom_filter): fsc, fom, filter values for reconstruction:'
            write(*,'(3(a6,1x))') 'Shell', 'FSC', 'FOM'
        endif

        first_shell_where_filter_is_zero = 0
        dont_zero_before_this_shell = size(self%fsc) / 4 ! We're allowed to zero before then, but not to be forced to stay at 0

        ! loop over the volume and apply the filter
        do k=1,img%logical_dimensions(3)
            do j=1,img%logical_dimensions(2)
                do i=1,lowest_index_for_negative_frequencies(img%logical_dimensions(1))-1
                    current_shell = img%ComputeShellNumberFromPhysicalAddress(i,j,k,do_not_fold=dnf)+1
                    print_filter_value = (pprint_filter .and. current_shell-1 .eq. last_printed_shell+1)
                    if (print_filter_value) last_printed_shell = current_shell-1
                    if (current_shell .le. size(self%fsc)) then
                        if (self%fsc(current_shell) .le. 0.0e0) then
                            filter_value = 0.0e0
                            if (first_shell_where_filter_is_zero .eq. 0 .and. current_shell .ge. dont_zero_before_this_shell) then
                                first_shell_where_filter_is_zero = current_shell
                            endif
                        else
                            filter_value =  2.*self%fsc(current_shell)/(1.+self%fsc(current_shell))
                            if (.not. use_sindelar_2012) filter_value = sqrt(filter_value)
                        endif
                        if (oonce_at_zero_stay_there .and. current_shell .ge. first_shell_where_filter_is_zero &
                                                     .and. first_shell_where_filter_is_zero .ne. 0) then
                            filter_value = 0.0e0
                        endif
                        img%complex_values(i,j,k) = img%complex_values(i,j,k) * filter_value
                        if (print_filter_value) then
                            write(*,'(i6,1x,2(f6.3,1x))') current_shell, self%fsc(current_shell), filter_value
                        endif
                    else
                        img%complex_values(i,j,k) = (0.0,0.0)
                    endif
                enddo
            enddo
        enddo
        if (pprint_filter) write(*,'(a)') ''
    end subroutine threed_apply_fom_filter

    !>  \brief  Divide reconstruction by CTF^2 volume to obtain least-squares estimate of the reconstruction in the absence of SNR estimates
    SUBROUTINE THREED_CTF2_DIVISION(SELF,IMG,REC_NUM)
        USE IMAGES_CTF
        use images
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),               INTENT(INOUT)   ::  SELF            !<  3D reconstruction object (input)
        TYPE(IMAGE),                INTENT(INOUT)   ::  IMG             !<  Resulting volume
        INTEGER,                    INTENT(IN)      ::  REC_NUM         !<  Which internal reconstruction should be used. If 0, and there are two internal reconstructions, the sum will be returned.
        ! Private variables
        INTEGER                         ::  I,J,K,I_REAL,J_REAL,K_REAL
        LOGICAL,    PARAMETER           ::  DIVIDE_THE_SACHSE_WAY = .FALSE.  !  Use maxval(ctfsqavg)*0.01 as an added constant in the denominator to avoid division by zero
                                                                             !  Tested on 20-March-2013 and it makes reconstructions worse, so leaving this off.
        REAL                            ::  MAX_VAL
        ! Start work

        ! Make sure the image is already allocated
        IF (.NOT. IS_ALLOCATED(IMG)) THEN
            IMG%logical_dimensions = SELF%logical_dimensions
            CALL IMAGE_ALLOCATE(IMG)
        ENDIF

        ! Ensure the reconstruction volume is in Fourier space
        IF (SELF%REC(1)%VOL_IMAGE%is_in_real_space) THEN
            CALL FWD_FT(SELF%REC(1)%VOL_IMAGE)
            CALL FWD_FT(SELF%REC(2)%VOL_IMAGE)
        ENDIF

        ! Ensure the origin is not in center (is this really necessary?)
        IF (SELF%REC(1)%VOL_IMAGE%ORIGIN_AT_CENTER) THEN
            CALL SHIFT_REAL_SPACE_ORIGIN(SELF%REC(1)%VOL_IMAGE)
            CALL SHIFT_REAL_SPACE_ORIGIN(SELF%REC(2)%VOL_IMAGE)
        ENDIF

        ! Work out the constant value to avoid divide by zero
        IF (DIVIDE_THE_SACHSE_WAY) THEN
            ! Work out the maximum value
            MAX_VAL = -HUGE(1.0)
            DO K=1,IMG%logical_dimensions(3)
                K_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3(IMG,K)
                IF (K_REAL .EQ. IMG%logical_dimensions(3)+1) CYCLE
                DO J=1,IMG%logical_dimensions(2)
                    J_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3(IMG,J)
                    IF (J_REAL .EQ. IMG%logical_dimensions(2)+1) CYCLE
                    DO I=1,IMG%logical_dimensions(1)/2+1
                        I_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM1(IMG,I)
                        IF (REC_NUM .EQ. 0) THEN
                            IF (  SELF%REC(1)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL)+   &
                                  SELF%REC(2)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL)    &
                                  .GT. MAX_VAL) THEN
                                MAX_VAL =   SELF%REC(1)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL)+   &
                                            SELF%REC(2)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL)
                            ENDIF
                        ELSE
                            IF (SELF%REC(REC_NUM)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL) .GT. MAX_VAL) THEN
                                MAX_VAL =   SELF%REC(REC_NUM)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL)
                            ENDIF
                        ENDIF
                    ENDDO
                ENDDO
            ENDDO
        ENDIF

        ! Do the division
        DO K=1,IMG%logical_dimensions(3)
            K_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3(IMG,K)
            IF (K_REAL .EQ. IMG%logical_dimensions(3)+1) CYCLE
            DO J=1,IMG%logical_dimensions(2)
                J_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3(IMG,J)
                IF (J_REAL .EQ. IMG%logical_dimensions(2)+1) CYCLE
                DO I=1,IMG%logical_dimensions(1)/2+1
                    I_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM1(IMG,I)
                    IF (REC_NUM .EQ. 0) THEN
                        IMG%complex_values(I,J,K) =    SELF%REC(1)%VOL_IMAGE%complex_values(I,J,K) + &
                                                       SELF%REC(2)%VOL_IMAGE%complex_values(I,J,K)
                        IF (DIVIDE_THE_SACHSE_WAY) THEN
                            IMG%complex_values(I,J,K) =    IMG%complex_values(I,J,K) / &
                                                            (SELF%REC(1)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL) +   &
                                                             SELF%REC(2)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL) +   &
                                                             MAX_VAL*0.01)
                        ELSE
                            IMG%complex_values(I,J,K) =    IMG%complex_values(I,J,K) / &
                                                            MAX(SELF%REC(1)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL)+   &
                                                                SELF%REC(2)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL),  &
                                                                0.0001)    ! To avoid division by zero, we use the max function
                        ENDIF
                    ELSE
                        IMG%complex_values(I,J,K) =    SELF%REC(REC_NUM)%VOL_IMAGE%complex_values(I,J,K)
                        IF (DIVIDE_THE_SACHSE_WAY) THEN
                            IMG%complex_values(I,J,K) =    IMG%complex_values(I,J,K)    &
                                                 / (SELF%REC(REC_NUM)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL)+MAX_VAL*0.01)
                        ELSE
                            IMG%complex_values(I,J,K) =    IMG%complex_values(I,J,K)    &
                                                 / MAX(SELF%REC(REC_NUM)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL),0.0001)    ! To avoid division by zero, we use the max function
                        ENDIF
                    ENDIF
                ENDDO
            ENDDO
        ENDDO
    END SUBROUTINE THREED_CTF2_DIVISION


    !>  \brief  Apply Wiener filter to 3D reconstruction
    SUBROUTINE THREED_APPLY_WIENER_FILTER(SELF,IMG,REC_NUM,PSSNR_FUDGE,PRESERVE_HIGH_FREQUENCIES)
        USE IMAGEFILES ! for debug
        USE IMAGES_CTF
        USE IMAGES_ARITHM, ONLY : CIRC_MASK
        use images
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),               INTENT(INOUT)   ::  SELF                        !<  3D reconstruction object
        TYPE(IMAGE),                INTENT(INOUT)   ::  IMG                         !<  Resulting volume
        INTEGER,                    INTENT(IN)      ::  REC_NUM                     !<  Which internal reconstruction should be used. If 0, and there are two internal reconstructions, the sum will be returned.
        REAL,           OPTIONAL,   INTENT(IN)      ::  PSSNR_FUDGE                 !<  A scaling fudge factor applied to the Wiener term
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  PRESERVE_HIGH_FREQUENCIES   !<  Even if they are dominated by noise (i.e. the SSNR is low), do not zero high frequencies.
                                                                                    !!  This will probably lead to more noisy reconstructions, but may be used if the Wiener filter is thought to be too stringent. Defaults to false.
        ! Private variables
        LOGICAL,        PARAMETER       ::  DEBUG   =   .FALSE.
        !TYPE(IMAGE)                    ::  TMPVOL
        INTEGER                         ::  I,J,K
        INTEGER                         ::  I_REAL,J_REAL,K_REAL
        REAL,           ALLOCATABLE     ::  WIENER_FILTER_TERM(:)
        LOGICAL,        PARAMETER       ::  DNF(3) = [.TRUE.,.FALSE.,.FALSE.]
        INTEGER                         ::  CURRENT_SHELL, NUM_OF_SHELLS
        REAL                            ::  PPSSNR_FUDGE
        LOGICAL                         ::  PPRESERVE_HIGH_FREQUENCIES
        ! Start work

        ! Fudge factor
        IF (PRESENT(PSSNR_FUDGE)) THEN
            WRITE(*,'(A,G15.3)') '**debug(THREED_APPLY_WIENER_FILTER): applying fudge factor of ', PSSNR_FUDGE
            PPSSNR_FUDGE = PSSNR_FUDGE
        ELSE
            PPSSNR_FUDGE = 1.0
            WRITE(*,'(A,G15.3)') '**debug(THREED_APPLY_WIENER_FILTER): no fudge factor specified, defaulting to ', PPSSNR_FUDGE
        ENDIF

        IF (.NOT. HAS_EVEN_DIMS(IMG)) THEN
            WRITE(*,'(A,3(I0,1X))') '**ERROR(THREED_APPLY_WIENER_FILTER): odd dimensions not supported: ', IMG%logical_dimensions
            CALL TERMINATE('Feature not yet implemented')
        ENDIF

        ! Debug dumps
        !IF (DEBUG) THEN
        !    IMG = SELF%VOL_IMAGE
        !    CALL SHIFT_REAL_SPACE_ORIGIN(IMG)
        !    CALL AMPL(IMG,TMPVOL)
        !    CALL WRITE_TO_DISK(TMPVOL,'debug_before_3dctf_ampl.mrc')
        !    TMPVOL%real_values = TMPVOL%real_values**2
        !    CALL WRITE_TO_DISK(TMPVOL,'debug_before_3dctf_pow.mrc')
        !   CALL BWD_FT(IMG)
        !    CALL WRITE_TO_DISK(IMG,'debug_before_3dctf.mrc')
        !    CALL FWD_FT(IMG)
        !    CALL SHIFT_REAL_SPACE_ORIGIN(IMG)
        !ENDIF

        ! Make sure the image is already allocated
        IF (.NOT. IS_ALLOCATED(IMG)) THEN
            IMG%logical_dimensions = SELF%logical_dimensions
            CALL IMAGE_ALLOCATE(IMG)
        ENDIF

        ! We assume the image is square/cube
        IF (.NOT. img%IsSquare()) THEN
            WRITE(*,'(A,3(I0,1X))') '**ERROR(THREED_APPLY_WIENER_FILTER): output image is not cubic: ', IMG%logical_dimensions
            CALL TERMINATE('Fatal error in THREED_APPLY_WIENER_FILTER')
        ENDIF

        ! By default, we will not preserve high-frequencies if their SSNR is too low
        IF (PRESENT(PRESERVE_HIGH_FREQUENCIES)) THEN
            PPRESERVE_HIGH_FREQUENCIES = PRESERVE_HIGH_FREQUENCIES
        ELSE
            PPRESERVE_HIGH_FREQUENCIES = .FALSE.
        ENDIF

        ! Ensure the reconstruction volume is in Fourier space
        IF (SELF%REC(1)%VOL_IMAGE%is_in_real_space) THEN
            CALL FWD_FT(SELF%REC(1)%VOL_IMAGE)
            IF (SIZE(SELF%REC) .EQ. 2) CALL FWD_FT(SELF%REC(2)%VOL_IMAGE)
        ENDIF

        ! Ensure the origin is not in center (is this really necessary?)
        IF (SELF%REC(1)%VOL_IMAGE%ORIGIN_AT_CENTER) THEN
            CALL SHIFT_REAL_SPACE_ORIGIN(SELF%REC(1)%VOL_IMAGE)
            IF (SIZE(SELF%REC) .EQ. 2) CALL SHIFT_REAL_SPACE_ORIGIN(SELF%REC(2)%VOL_IMAGE)
        ENDIF

        ! Work out the number of shells to the Nyquist
        ! (maybe should instead work out the number of shells to the corners of the cube?)
        NUM_OF_SHELLS = NUMBER_OF_SHELLS(IMG)

        ! Calculate the Wiener filter term (constant or per-resolution-shell)
        IF (SELF%USE_SSNR_ESTIMATE) THEN
            IF (REC_NUM .NE. 0 .AND. SIZE(SELF%REC) .GT. 1) THEN
                WRITE(*,'(2A)') '**warning(THREED_APPLY_WIENER_FILTER): will apply a Wiener filter based on ',   &
                                'FSC-estimated SSNR, even though it will only be applied to one of the reconstructions'
            ENDIF
            IF (SELF%SSNR_IS_UP_TO_DATE) THEN
                WRITE(*,'(2A)') '**info(THREED_APPLY_WIENER_FILTER): we already have an up-to-date estimate of ',   &
                                'the SSNR, so we will not re-estimate it'
            ELSE
                WRITE(*,'(2A)') '**info(THREED_APPLY_WIENER_FILTER): we do not have an up-to-date estimate of ',    &
                                'the SSNR, so we will estimate it now'
                CALL THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS(SELF,FSC_ESTIMATE_ABOVE_NOISE=PPRESERVE_HIGH_FREQUENCIES)
            ENDIF
            WIENER_FILTER_TERM = SELF%SSNR_ESTIMATE
            ! The Wiener filter term is actually 1/PSSNR
            DO I=1,SIZE(WIENER_FILTER_TERM)
                WIENER_FILTER_TERM(I) = 1.0 / MAX(WIENER_FILTER_TERM(I),.001)
            ENDDO
        ELSE
            WIENER_FILTER_TERM = THREED_GUESS_WIENER_FILTER_TERM(SELF,REC_NUM=REC_NUM)
        ENDIF

        ! Apply fudge
        WIENER_FILTER_TERM = WIENER_FILTER_TERM * PPSSNR_FUDGE

        IF (DEBUG) THEN
            !WRITE(*,'(A,<SIZE(WIENER_FILTER_TERM)>(F0.3,X))') '**debug(THREED_APPLY_WIENER_FILTER): wiener filter term = ', &
            !                                                            WIENER_FILTER_TERM
            !CALL IMAGE_PRINT_INFO(SELF%REC(1)%VOL_IMAGE)
            !CALL IMAGE_PRINT_INFO(SELF%REC(2)%VOL_IMAGE)
            !CALL IMAGE_PRINT_INFO(SELF%REC(1)%VOL_CTF2)
            !CALL IMAGE_PRINT_INFO(SELF%REC(2)%VOL_CTF2)
        ENDIF

        ! Loop over the complex elements of the reconstruction volume and apply the filter
        DO K=1,IMG%logical_dimensions(3)
            K_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3(IMG,K)
            IF (K_REAL .EQ. IMG%logical_dimensions(3)+1) CYCLE
            DO J=1,IMG%logical_dimensions(2)
                J_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3(IMG,J)
                IF (J_REAL .EQ. IMG%logical_dimensions(2)+1) CYCLE
                DO I=1,IMG%logical_dimensions(1)/2+1
                    I_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM1(IMG,I)
                    CURRENT_SHELL = img%ComputeShellNumberFromPhysicalAddress(I,J,K,DO_NOT_FOLD=DNF)
                    IF (CURRENT_SHELL+1 .GT. NUM_OF_SHELLS) CYCLE
                    IF (REC_NUM .EQ. 0) THEN
                        IMG%complex_values(I,J,K) = (SELF%REC(1)%VOL_IMAGE%complex_values(I,J,K) + &
                                                     SELF%REC(2)%VOL_IMAGE%complex_values(I,J,K)) / &
                                                    (SELF%REC(1)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL) + &
                                                     SELF%REC(2)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL) &
                                                    + WIENER_FILTER_TERM(CURRENT_SHELL+1))
                    ELSE
                        IMG%complex_values(I,J,K) =  SELF%REC(REC_NUM)%VOL_IMAGE%complex_values(I,J,K) / &
                                                    (SELF%REC(REC_NUM)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL) &
                                                    + WIENER_FILTER_TERM(CURRENT_SHELL+1))
                    ENDIF
                ENDDO
            ENDDO
        ENDDO

        ! This leaves the 3D reconstruction with origin not at center
        IMG%ORIGIN_AT_CENTER = .FALSE.

        ! The Nyquist elements are not reliable, so we remove them
        CALL CIRC_MASK(IMG,0.5)

        ! The image is ready to be back-FT'ed, but first needs to be shifted
        CALL SHIFT_REAL_SPACE_ORIGIN(IMG)
        CALL BWD_FT(IMG)

    END SUBROUTINE THREED_APPLY_WIENER_FILTER

    !>  \brief  Estimate the SSNR of a reconstruction based on the comparison of two reconstructions from half datasets.
    SUBROUTINE THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS(SELF,FSC_ESTIMATE_ABOVE_NOISE)
        USE IMAGES_HELICES
        USE IMAGES_CTF
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),                   INTENT(INOUT)       ::  SELF
        LOGICAL,        OPTIONAL,       INTENT(IN)          ::  FSC_ESTIMATE_ABOVE_NOISE        !<  If true, the FSC estimate is not allowed below a sigma threshold, which will prevent subsequent Wiener filters from being too harsh. Defaults to false.
        ! Private variables
        INTEGER                     ::  CURRENT_REC
        TYPE(IMAGE)                 ::  HALF_DATASET_REC(2)
        REAL(KIND=8),   ALLOCATABLE ::  HALF_DATASET_FSC(:), SUM_CTF2(:), FSC_NUM(:),FSC_DENOM1(:),FSC_DENOM2(:)
        INTEGER,        ALLOCATABLE ::  VOXELS_IN_SHELL(:)
        INTEGER                     ::  I,J,K, I_REAL,J_REAL,K_REAL
        INTEGER                     ::  NUM_OF_SHELLS, CURRENT_SHELL
        LOGICAL,        PARAMETER   ::  IIGNORE_FRIEDEL_MATES   =   .TRUE.  ! Some Friedel mates are described explicelty by the FFTW array, we should not count them twice!
        INTEGER                     ::  IERR
        LOGICAL,        PARAMETER   ::  DNF(3) = [.TRUE.,.FALSE.,.FALSE.]
        REAL                        ::  CURRENT_FSC
        REAL,           PARAMETER   ::  THREED_PSSNR_FUDGE_FACTOR    =  1.0    !1.8    ! To be replaced by proper equation later (when Chuck works it all out)
        INTEGER                     ::  FSC_BELOW_THRESH_SHELL  ! Shell number at which the FSC falls below 0.143
        REAL,           PARAMETER   ::  SIGMA_FACTOR    =   2.
        REAL                        ::  NOISE_FSC_SIGMA
        LOGICAL                     ::  FFSC_ESTIMATE_ABOVE_NOISE    ! If true, the FSC estimate is not allowed below a sigma threshold, which will prevent subsequent Wiener filters from being too harsh
        LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work

        ! Check that two reconstructions have been prepared
        IF (THREED_NUM_OF_RECS(SELF) .NE. 2) THEN
            WRITE(*,'(A)') '**ERROR(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): number of internal reconstructions = ',  &
                                THREED_NUM_OF_RECS(SELF)
            CALL TERMINATE('Fatal error in THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS. Programming error.')
        ENDIF

        ! Check neither reconstruction is blank
        IF (ANY(SELF%REC%NUM_CONTRIB_ELEMS .EQ. 0)) THEN
            WRITE(*,'(2A,2(I0,1X))') '**ERROR(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): at least one of the internal ',    &
                                    'reconstructions is blank: ', SELF%REC%NUM_CONTRIB_ELEMS
            CALL TERMINATE('Fatal error in THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS')
        ENDIF

        ! Will we allow the FSC to go below n sigma?
        IF (PRESENT(FSC_ESTIMATE_ABOVE_NOISE)) THEN
            FFSC_ESTIMATE_ABOVE_NOISE = FSC_ESTIMATE_ABOVE_NOISE
        ELSE
            FFSC_ESTIMATE_ABOVE_NOISE = .FALSE.
        ENDIF



        ! Check whether we are re-calculating SSNR unnecessarily
        IF (SELF%SSNR_IS_UP_TO_DATE) THEN
            WRITE(*,'(2A)') '**warning(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): SSNR estimate appears to ',  &
                            'already be up-to-date. You may be wasting time by calling this again.'
        ENDIF

        WRITE(*,'(2A)') '**info(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): will estimate the SSNR by ',    &
                        'computing the FSC between the two internal reconstructions'

        IF (FFSC_ESTIMATE_ABOVE_NOISE) THEN
            WRITE(*,'(A,F0.1,A)') '**info(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): the FSC will not be allowed below ', &
                                        SIGMA_FACTOR, ' standard derivations of noise FSCs'
        ENDIF

        ! Debug
        IF (DEBUG) THEN
            CALL THREED_WRITE_TO_DISK(SELF,'dbg_3d_raw.mrc')
        ENDIF


        ! Loop over each reconstruction
        DO CURRENT_REC=1,THREED_NUM_OF_RECS(SELF)
            WRITE(*,'(2(A,I0))') '**info(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): evaluating internal reconstruction ', &
                                    CURRENT_REC, ' of ', THREED_NUM_OF_RECS(SELF)
            ! Compute two reconstructions from the half datasets, by straightforward division by CTF^2
            CALL THREED_EVAL_RECONSTRUCTION(SELF,HALF_DATASET_REC(CURRENT_REC),WIENER_FILTER=.FALSE.,REC_NUM=CURRENT_REC)
            IF (.NOT. HALF_DATASET_REC(CURRENT_REC)%ORIGIN_AT_CENTER) CALL SHIFT_REAL_SPACE_ORIGIN(HALF_DATASET_REC(CURRENT_REC))
            IF (DEBUG) THEN
                IF (CURRENT_REC==1) CALL WRITE_TO_DISK(HALF_DATASET_REC(CURRENT_REC),'dbg_3d_ssnrest_1_nomask.mrc')
                IF (CURRENT_REC==2) CALL WRITE_TO_DISK(HALF_DATASET_REC(CURRENT_REC),'dbg_3d_ssnrest_2_nomask.mrc')
            ENDIF
            ! Compute  & apply a soft mask for each reconstruction
            WRITE(*,'(2(A,I0))') '**info(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): masking internal reconstruction ', &
                                CURRENT_REC, ' of ', THREED_NUM_OF_RECS(SELF)
            IF (SELF%POINT_SYM%HELICAL) THEN
                ! The mask corresponds to molecular envelope at resolution of ~20A
                CALL IMAGE_HELIX_MASK(  HALF_DATASET_REC(CURRENT_REC),SELF%HELIX,SELF%ASYM_UNIT_MW,SELF%PIXEL_SIZE,   &
                                        SELF%HELIX_MAX_RAD,F_MASK=SELF%F_MASK,F_SMOOTH=SELF%F_SMOOTH,   &
                                        MASK_RESOLUTION=SELF%PIXEL_SIZE/20.)
            ELSE
                WRITE(*,'(A)') '**ERROR(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): masking not yet implemented',  &
                                ' for single-particle reconstructions'
                CALL TERMINATE('Feature not implemented yet')
            ENDIF
            IF (DEBUG) THEN
                IF (CURRENT_REC==1) CALL WRITE_TO_DISK(HALF_DATASET_REC(CURRENT_REC),'dbg_3d_ssnrest_1_mask.mrc')
                IF (CURRENT_REC==2) CALL WRITE_TO_DISK(HALF_DATASET_REC(CURRENT_REC),'dbg_3d_ssnrest_2_mask.mrc')
            ENDIF
            ! Go to Fourier space
            CALL FWD_FT(HALF_DATASET_REC(CURRENT_REC))
        ENDDO

        ! Check that F_MASK and F_SMOOTH have reasonable values
        IF (SELF%F_MASK .EQ. 0.0 .OR. SELF%F_SMOOTH .EQ. 0.0) THEN
            WRITE(*,'(A,2(F0.3,1X))') '**ERROR(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): F_SMOOTH, F_MASK = ', &
                                    SELF%F_SMOOTH, SELF%F_MASK
            CALL TERMINATE('Fatal error in THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS')
        ENDIF
        IF (DEBUG) THEN
            WRITE(*,'(A,2(F0.3,1X))') '**debug(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): F_SMOOTH, F_MASK = ', &
                                        SELF%F_SMOOTH, SELF%F_MASK
        ENDIF


        ! For the following steps, the reconstructions need to be in Fourier space
        IF (HALF_DATASET_REC(1)%is_in_real_space) CALL FWD_FT(HALF_DATASET_REC(1))
        IF (HALF_DATASET_REC(2)%is_in_real_space) CALL FWD_FT(HALF_DATASET_REC(2))

        ! Also, I think we need them to be with real space origin not at center
        IF (HALF_DATASET_REC(1)%ORIGIN_AT_CENTER) CALL SHIFT_REAL_SPACE_ORIGIN(HALF_DATASET_REC(1))
        IF (HALF_DATASET_REC(2)%ORIGIN_AT_CENTER) CALL SHIFT_REAL_SPACE_ORIGIN(HALF_DATASET_REC(2))

        ! Allocate memory for output arrays
        NUM_OF_SHELLS = NUMBER_OF_SHELLS(HALF_DATASET_REC(1))
        IF (ALLOCATED(SELF%SSNR_ESTIMATE)) DEALLOCATE(SELF%SSNR_ESTIMATE)
        ALLOCATE(HALF_DATASET_FSC(NUM_OF_SHELLS), SUM_CTF2(NUM_OF_SHELLS), SELF%SSNR_ESTIMATE(NUM_OF_SHELLS),&
                FSC_NUM(NUM_OF_SHELLS),FSC_DENOM1(NUM_OF_SHELLS),FSC_DENOM2(NUM_OF_SHELLS),VOXELS_IN_SHELL(NUM_OF_SHELLS),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(2A,I0)') '**ERROR(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): memory allocation for', &
                                ' intermediate and result arrays failed with error ', IERR
            CALL TERMINATE('Fatal error in THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS')
        ENDIF
        HALF_DATASET_FSC    =   0.0
        SUM_CTF2            =   0.0
        SELF%SSNR_ESTIMATE  =   0.0
        FSC_NUM             =   0.0
        FSC_DENOM1          =   0.0
        FSC_DENOM2          =   0.0
        VOXELS_IN_SHELL     =   0

        ! Loop over the volume and accumulate terms for the estimation of the SSNR as in Equation 19 of Sindelar (2011):
        ! SSNR(r) = (1/F_MASK) * N(r) / SUM_CTF2(r) * 2. * F_SMOOTH * FSC(r) / (1.-FSC(r))
        WRITE(*,'(2A)') '**info(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): accumulating ',   &
                                'statistics for FSC and SSNR calculations'
        DO I=1,SELF%logical_dimensions(1)/2+1
            I_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM1(SELF%REC(1)%VOL_IMAGE,I)
            DO J=1,SELF%logical_dimensions(2)
                IF (IIGNORE_FRIEDEL_MATES .AND. I == 1 .AND. J .GT. 1 .AND. J .LT. SELF%logical_dimensions(2)/2+1) CYCLE   ! Skip voxels whose Friedel mates are explicit in the FFTW description
                J_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3(SELF%REC(1)%VOL_IMAGE,J)
                IF (J_REAL .EQ. SELF%REC(1)%VOL_IMAGE%logical_dimensions(2)+1) CYCLE ! Positive Nyquist component is not described in the VOL_CTF2 array
                DO K=1,SELF%logical_dimensions(3)
                    IF (IIGNORE_FRIEDEL_MATES .AND. I == 1 .AND. J == 1 .AND. K .GT. 1 .AND. K .LT. SELF%logical_dimensions(3)/2+1) CYCLE
                    K_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3(SELF%REC(1)%VOL_IMAGE,K)
                    IF (K_REAL .EQ. SELF%REC(1)%VOL_IMAGE%logical_dimensions(3)+1) CYCLE

                    CURRENT_SHELL = HALF_DATASET_REC(1)%ComputeShellNumberFromPhysicalAddress(I,J,K,DO_NOT_FOLD=DNF) + 1 ! Shell 0 (the origin voxel) needs index 1 (Fortran arrays are addressed from 1)
                    IF (CURRENT_SHELL .LE. NUM_OF_SHELLS) THEN
                        FSC_NUM(CURRENT_SHELL)          =   FSC_NUM(CURRENT_SHELL)    &
                                                            + DBLE(HALF_DATASET_REC(1)%complex_values(I,J,K) &
                                                            * CONJG(HALF_DATASET_REC(2)%complex_values(I,J,K)))
                        FSC_DENOM1(CURRENT_SHELL)       =   FSC_DENOM1(CURRENT_SHELL) &
                                                            + DBLE(ABS(HALF_DATASET_REC(1)%complex_values(I,J,K)))**2
                        FSC_DENOM2(CURRENT_SHELL)       =   FSC_DENOM2(CURRENT_SHELL) &
                                                            + DBLE(ABS(HALF_DATASET_REC(2)%complex_values(I,J,K)))**2
                        SUM_CTF2(CURRENT_SHELL)         =   SUM_CTF2(CURRENT_SHELL)     &
                                                            + DBLE(SELF%REC(1)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL)) &
                                                            + DBLE(SELF%REC(2)%VOL_CTF2%real_values(I_REAL,J_REAL,K_REAL))
                        VOXELS_IN_SHELL(CURRENT_SHELL)  =   VOXELS_IN_SHELL(CURRENT_SHELL) + 1
                    ENDIF
                ENDDO
            ENDDO
        ENDDO

        ! Loop over shells
        FSC_BELOW_THRESH_SHELL = -1
        WRITE(*,'(2(A,I0))') '**info(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): computing SSNR'
        WRITE(*,'(A5,1X,A6,1X,A6,1X,A6)') 'Shell', 'FSC', 'Sigma', 'PSSNR'
        DO CURRENT_SHELL=1,SIZE(SELF%SSNR_ESTIMATE)
            CURRENT_FSC = FSC_NUM(CURRENT_SHELL) / SQRT(FSC_DENOM1(CURRENT_SHELL)*FSC_DENOM2(CURRENT_SHELL))
            ! To avoid division by 0 later on, we cap the FSC at 0.999
            CURRENT_FSC = MIN(CURRENT_FSC,0.999)
            ! Work out the expected variance of FSC between noise volumes at this shell (see for example Eq19 in Van Heel 2005)
            NOISE_FSC_SIGMA = SIGMA_FACTOR / SQRT(REAL(VOXELS_IN_SHELL(CURRENT_SHELL))) / SQRT(2.) / SELF%F_MASK
            ! If we've just fallen below 0.143, remember what shell we're on
            IF (CURRENT_FSC .LT. 0.143 .AND. FSC_BELOW_THRESH_SHELL .EQ. -1) FSC_BELOW_THRESH_SHELL = CURRENT_SHELL
            ! If requested, do not allow FSC to fall below n standard deviations of the expected FSC between noise volumes
            IF (FFSC_ESTIMATE_ABOVE_NOISE) CURRENT_FSC = MAX(CURRENT_FSC,MIN(NOISE_FSC_SIGMA,0.999))
            IF (SUM_CTF2(CURRENT_SHELL) .GT. TINY(1.0)) THEN
                ! Equation 19 of Sindelar (2011)
                SELF%SSNR_ESTIMATE(CURRENT_SHELL) = (1/SELF%F_MASK) * (VOXELS_IN_SHELL(CURRENT_SHELL)/SUM_CTF2(CURRENT_SHELL)) &
                                                                * (2.*SELF%F_SMOOTH*CURRENT_FSC/(1-CURRENT_FSC))
            ELSE
                SELF%SSNR_ESTIMATE(CURRENT_SHELL) = 0.0
            ENDIF
            ! Apply fudge factor
            SELF%SSNR_ESTIMATE(CURRENT_SHELL) = SELF%SSNR_ESTIMATE(CURRENT_SHELL) * THREED_PSSNR_FUDGE_FACTOR
            ! SSNR cannot be negative
            SELF%SSNR_ESTIMATE(CURRENT_SHELL) = MAX(0.0,SELF%SSNR_ESTIMATE(CURRENT_SHELL))
            ! Output
            WRITE(*,'(I5,1X,F6.3,1X,F6.3,1X,F0.3)') CURRENT_SHELL, CURRENT_FSC, NOISE_FSC_SIGMA, SELF%SSNR_ESTIMATE(CURRENT_SHELL)
        ENDDO

        ! Let the user know the estimated internal resolution
        WRITE(*,'(A,I0,A,I0)') '**info(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): the FSC fell below 0.143 in shell ', &
                                FSC_BELOW_THRESH_SHELL-1, ' of ', SIZE(SELF%SSNR_ESTIMATE)-1

        ! Deallocate arrays holding intermediate results
        DEALLOCATE(HALF_DATASET_FSC, SUM_CTF2,FSC_NUM,FSC_DENOM1,FSC_DENOM2,VOXELS_IN_SHELL)

        ! Deallocate
        CALL IMAGE_DEALLOCATE(HALF_DATASET_REC(1))
        CALL IMAGE_DEALLOCATE(HALF_DATASET_REC(2))

        ! Now we have an updated PSSNR estimate
        SELF%SSNR_IS_UP_TO_DATE = .TRUE.

        WRITE(*,'(2(A,I0))') '**info(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): all done'

    END SUBROUTINE THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS

    !>  \brief  Compute the FSC between half-dataset reconstructions
    SUBROUTINE THREED_UPDATE_FSC(SELF,FSC_VOLUMES)
        USE IMAGES_HELICES
        USE IMAGES_CTF, ONLY : CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3, CTFVOL_ADDR_REAL_FROM_COMP_DIM1
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),               INTENT(INOUT)   ::  SELF
        TYPE(IMAGE),    OPTIONAL,   INTENT(INOUT)   ::  FSC_VOLUMES(2)  !<  If given, the final FSC volumes will be returned
        ! Private variables
        INTEGER                     ::  CURRENT_REC
        TYPE(IMAGE)                 ::  HALF_DATASET_REC(2)
        REAL(KIND=8),   ALLOCATABLE ::  FSC_NUM(:),FSC_DENOM1(:),FSC_DENOM2(:)
        INTEGER                     ::  NUM_OF_SHELLS, CURRENT_SHELL, I,J,K, I_REAL,J_REAL,K_REAL
        LOGICAL,        PARAMETER   ::  IIGNORE_FRIEDEL_MATES   =   .TRUE.  ! Some Friedel mates are described explicelty by the FFTW array, we should not count them twice!
        LOGICAL,        PARAMETER   ::  DNF(3) = (/.TRUE.,.FALSE.,.FALSE./)
        LOGICAL,        PARAMETER   ::  DEBUG   =   .FALSE.
        LOGICAL,        PARAMETER   ::  HELICAL_SYMMETRISE = .TRUE.
        REAL                        ::  DENOM
        ! Start work

        ! Check that two reconstructions have been prepared
        IF (THREED_NUM_OF_RECS(SELF) .NE. 2) THEN
            WRITE(*,'(A)') '**ERROR(THREED_UPDATE_FSC): number of internal reconstructions = ',  &
                                THREED_NUM_OF_RECS(SELF)
            CALL TERMINATE('Fatal error in THREED_UPDATE_FSC. Programming error.')
        ENDIF

        ! Check neither reconstruction is blank
        IF (ANY(SELF%REC%NUM_CONTRIB_ELEMS .EQ. 0)) THEN
            WRITE(*,'(A,2(I0,1X))') '**ERROR(THREED_UPDATE_FSC): at least one of the internal reconstructions is blank: ', &
                                    SELF%REC%NUM_CONTRIB_ELEMS
            CALL TERMINATE('Fatal error in THREED_UPDATE_FSC')
        ENDIF

        WRITE(*,'(A)') '**info(THREED_UPDATE_FSC): will compute FSC between half-dataset reconstructions'

        ! Check whether we need to be here in the first place
        IF (SELF%FSC_IS_UP_TO_DATE) THEN
            WRITE(*,'(A)') '**warning(THREED_UPDATE_FSC): FSC appears to be up-to-date already'
        ENDIF

        ! Loop over each reconstruction
        DO CURRENT_REC=1,THREED_NUM_OF_RECS(SELF)
            WRITE(*,'(2(A,I0))') '**info(THREED_UPDATE_FSC): evaluating internal reconstruction ', &
                                    CURRENT_REC, ' of ', THREED_NUM_OF_RECS(SELF)
            ! Compute two reconstructions from the half datasets
            CALL THREED_EVAL_RECONSTRUCTION(SELF,HALF_DATASET_REC(CURRENT_REC),REC_NUM=CURRENT_REC, &
                                            SKIP_FOM_FILTER=.TRUE.,SKIP_FSC=.TRUE.,SKIP_AMPL=.TRUE.)
            IF (.NOT. HALF_DATASET_REC(CURRENT_REC)%ORIGIN_AT_CENTER) THEN
                CALL SHIFT_REAL_SPACE_ORIGIN(HALF_DATASET_REC(CURRENT_REC))
            ENDIF
            IF (DEBUG) THEN
                IF (CURRENT_REC==1) CALL WRITE_TO_DISK(HALF_DATASET_REC(CURRENT_REC),'dbg_3d_ssnrest_1_nomask.mrc')
                IF (CURRENT_REC==2) CALL WRITE_TO_DISK(HALF_DATASET_REC(CURRENT_REC),'dbg_3d_ssnrest_2_nomask.mrc')
            ENDIF
            ! Compute  & apply a soft mask for each reconstruction
            WRITE(*,'(2(A,I0))') '**info(THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS): masking internal reconstruction ', &
                                    CURRENT_REC, ' of ', THREED_NUM_OF_RECS(SELF)
            IF (SELF%POINT_SYM%HELICAL) THEN
                IF (HELICAL_SYMMETRISE) THEN
                    IF (DEBUG) WRITE(*,'(A,F0.3)') '**debug(THREED_UPDATE_FSC): about to hextend with max rad = ', &
                                                    SELF%HELIX_MAX_RAD
                    CALL IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW(HALF_DATASET_REC(CURRENT_REC),SELF%HELIX,SELF%INTERPOL_SCHEME,&
                                                        MAX_RADIUS=SELF%HELIX_MAX_RAD)
                ENDIF
                ! The mask corresponds to molecular envelope at resolution of ~20A
                IF (DEBUG) WRITE(*,'(A,F0.3)') '**debug(THREED_UPDATE_FSC): about to hmask with max rad = ', SELF%HELIX_MAX_RAD
                CALL IMAGE_HELIX_MASK(  HALF_DATASET_REC(CURRENT_REC),SELF%HELIX,SELF%ASYM_UNIT_MW,SELF%PIXEL_SIZE,   &
                                        SELF%HELIX_MAX_RAD,F_MASK=SELF%F_MASK,F_SMOOTH=SELF%F_SMOOTH,   &
                                        MASK_RESOLUTION=SELF%PIXEL_SIZE/20.)
            ELSE
                WRITE(*,'(A)') '**ERROR(THREED_UPDATE_FSC): masking not yet implemented for single-particle reconstructions'
                CALL TERMINATE('Feature not implemented yet')
            ENDIF
            IF (DEBUG) THEN
                IF (CURRENT_REC==1) CALL WRITE_TO_DISK(HALF_DATASET_REC(CURRENT_REC),'dbg_3d_ssnrest_1_mask.mrc')
                IF (CURRENT_REC==2) CALL WRITE_TO_DISK(HALF_DATASET_REC(CURRENT_REC),'dbg_3d_ssnrest_2_mask.mrc')
            ENDIF
            IF (PRESENT(FSC_VOLUMES)) THEN
                FSC_VOLUMES(CURRENT_REC) = HALF_DATASET_REC(CURRENT_REC)
            ENDIF
            ! Go to Fourier space
            CALL FWD_FT(HALF_DATASET_REC(CURRENT_REC))
        ENDDO

        ! Allocate memory for FSC array
        NUM_OF_SHELLS = NUMBER_OF_SHELLS(HALF_DATASET_REC(1))
        IF (ALLOCATED(SELF%FSC)) DEALLOCATE(SELF%FSC)
        ALLOCATE(SELF%FSC(NUM_OF_SHELLS),FSC_NUM(NUM_OF_SHELLS),FSC_DENOM1(NUM_OF_SHELLS),FSC_DENOM2(NUM_OF_SHELLS))
        SELF%FSC = 0.0
        FSC_NUM = 0.0D0
        FSC_DENOM1 = 0.0D0
        FSC_DENOM2 = 0.0D0
        WRITE(*,'(2(A,I0))') '**info(THREED_UPDATE_FSC): accumulating statistics for FSC computation'
        DO I=1,LOWEST_INDEX_FOR_NEGATIVE_FREQUENCIES(SELF%logical_dimensions(1))
            I_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM1(SELF%REC(1)%VOL_IMAGE,I)
            DO J=1,SELF%logical_dimensions(2)
                IF (IIGNORE_FRIEDEL_MATES .AND. I == 1 .AND. J .GT. 1 &
                   .AND. J .LT. LOWEST_INDEX_FOR_NEGATIVE_FREQUENCIES(SELF%logical_dimensions(2))) CYCLE   ! Skip voxels whose Friedel mates are explicit in the FFTW description
                J_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3(SELF%REC(1)%VOL_IMAGE,J)
                IF (J_REAL .EQ. SELF%REC(1)%VOL_IMAGE%logical_dimensions(2)+1) CYCLE ! Positive Nyquist component is not described in the VOL_CTF2 array
                DO K=1,SELF%logical_dimensions(3)
                    IF (IIGNORE_FRIEDEL_MATES .AND. I == 1 .AND. J == 1 .AND. K .GT. 1 &
                        .AND. K .LT. LOWEST_INDEX_FOR_NEGATIVE_FREQUENCIES(SELF%logical_dimensions(3))) CYCLE
                    K_REAL = CTFVOL_ADDR_REAL_FROM_COMP_DIM2OR3(SELF%REC(1)%VOL_IMAGE,K)
                    IF (K_REAL .EQ. SELF%REC(1)%VOL_IMAGE%logical_dimensions(3)+1) CYCLE

                    CURRENT_SHELL = HALF_DATASET_REC(1)%ComputeShellNumberFromPhysicalAddress(I,J,K,DO_NOT_FOLD=DNF) + 1 ! Shell 0 (the origin voxel) needs index 1 (Fortran arrays are addressed from 1)
                    IF (CURRENT_SHELL .LE. NUM_OF_SHELLS) THEN
                        FSC_NUM(CURRENT_SHELL)          =   FSC_NUM(CURRENT_SHELL)    &
                                                            + DBLE(HALF_DATASET_REC(1)%complex_values(I,J,K) &
                                                            * CONJG(HALF_DATASET_REC(2)%complex_values(I,J,K)))
                        FSC_DENOM1(CURRENT_SHELL)       =   FSC_DENOM1(CURRENT_SHELL) &
                                                            + DBLE(ABS(HALF_DATASET_REC(1)%complex_values(I,J,K)))**2
                        FSC_DENOM2(CURRENT_SHELL)       =   FSC_DENOM2(CURRENT_SHELL) &
                                                            + DBLE(ABS(HALF_DATASET_REC(2)%complex_values(I,J,K)))**2
                    ENDIF
                ENDDO
            ENDDO
        ENDDO

!        CALL BWD_FT(HALF_DATASET_REC(1))
!        CALL BWD_FT(HALF_DATASET_REC(2))
!        CALL WRITE_TO_DISK(HALF_DATASET_REC(1),'dbg_half1.mrc')
!        CALL WRITE_TO_DISK(HALF_DATASET_REC(2),'dbg_half2.mrc')

        ! Loop over shells
        WRITE(*,'(2(A,I0))') '**info(THREED_UPDATE_FSC): computing FSC'
        DO CURRENT_SHELL=1,SIZE(SELF%FSC)
            DENOM = SQRT(FSC_DENOM1(CURRENT_SHELL)*FSC_DENOM2(CURRENT_SHELL))
            IF (DENOM .EQ. 0.0) THEN
                WRITE(*,'(A,I0,A)') '**WARNING(THREED_UPDATE_FSC): FSC denominator for shell ', CURRENT_SHELL, ' is 0.0'
                SELF%FSC(CURRENT_SHELL) = 0.0
            ELSE
                SELF%FSC(CURRENT_SHELL) = FSC_NUM(CURRENT_SHELL) / DENOM
            ENDIF
        ENDDO

        ! Remember we did this
        SELF%FSC_IS_UP_TO_DATE = .TRUE.
    END SUBROUTINE THREED_UPDATE_FSC

    !>  \brief  Guess a constant (i.e. not frequency-dependent) Wiener filter term, as in Grigorieff (1998) and Frealign or as in Sachse (2007)
    FUNCTION THREED_GUESS_WIENER_FILTER_TERM(SELF,REC_NUM) RESULT(WIENER_FILTER_TERM)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),               INTENT(IN)  ::  SELF
        INTEGER,                    INTENT(IN)  ::  REC_NUM
        ! Results
        REAL,               ALLOCATABLE         ::  WIENER_FILTER_TERM(:)
        ! Private variables
        LOGICAL,            PARAMETER   ::  DEBUG = .FALSE.
        LOGICAL,            PARAMETER   ::  INCLUDE_NSR_TERM_IN_WIENER  =   .FALSE.
        LOGICAL,            PARAMETER   ::  A_LA_SACHSE_2007 = .FALSE. !  Tested on 20-March-2013 and it makes reconstructions worse, so leaving this off.
        REAL(KIND=8)                    ::  SIG_CTF
        REAL                            ::  MAX_VAL
        REAL                            ::  SYM_REDUN
        INTEGER                         ::  NUM_OF_SHELLS
        INTEGER ::  I,J,K, IERR
        INTEGER ::  ORI(3)
        ! Start work

        ! Allocate memory for output arrays
        NUM_OF_SHELLS = NUMBER_OF_SHELLS(SELF%REC(1)%VOL_IMAGE)
        IF (ALLOCATED(WIENER_FILTER_TERM)) DEALLOCATE(WIENER_FILTER_TERM)
        ALLOCATE(WIENER_FILTER_TERM(NUM_OF_SHELLS),STAT=IERR)
        IF (IERR .NE. 0) THEN
            WRITE(*,'(A,I0)') '**ERROR(THREED_GUESS_WIENER_FILTER_TERM): memory allocation for result arrays failed with error ', &
                                IERR
            CALL TERMINATE('Fatal error in THREED_GUESS_WIENER_FILTER_TERM')
        ENDIF


        IF (A_LA_SACHSE_2007) THEN
            ! The constant will be 0.01* the maximum value in the CTF2 vol
            MAX_VAL = -HUGE(1.0)
            if (self%rec(1)%vol_ctf2%is_in_real_space .and. self%rec(1)%vol_ctf2%origin_at_center) then
                ori = self%rec(1)%vol_ctf2%physical_address_of_box_center
            else
                ori = [1,1,1]
            endif
            DO K=1,SELF%REC(1)%VOL_CTF2%logical_dimensions(3)
                DO J=1,SELF%REC(1)%VOL_CTF2%logical_dimensions(2)
                    DO I=1,ORI(1)
                        IF (I .EQ. ORI(1) .AND. (J .GT. ORI(2) .OR. K .GT. ORI(3))) CYCLE
                        IF (REC_NUM .EQ. 0) THEN
                            IF (  (SELF%REC(1)%VOL_CTF2%real_values(I,J,K) + &
                                   SELF%REC(2)%VOL_CTF2%real_values(I,J,K)) .GT. MAX_VAL ) THEN
                                MAX_VAL = SELF%REC(1)%VOL_CTF2%real_values(I,J,K) + SELF%REC(2)%VOL_CTF2%real_values(I,J,K)
                            ENDIF
                        ELSE
                            IF (  SELF%REC(REC_NUM)%VOL_CTF2%real_values(I,J,K) .GT. MAX_VAL ) THEN
                                MAX_VAL = SELF%REC(REC_NUM)%VOL_CTF2%real_values(I,J,K)
                            ENDIF
                        ENDIF
                    ENDDO
                ENDDO
            ENDDO
            WIENER_FILTER_TERM = 0.01*MAX_VAL
        ELSE
            ! Accumulate some statistics which we will use for our fake 1/ssnr term
            SIG_CTF = 0.0D0
            if (self%rec(1)%vol_ctf2%is_in_real_space .and. self%rec(1)%vol_ctf2%origin_at_center) then
                ori = self%rec(1)%vol_ctf2%physical_address_of_box_center
            else
                ori = [1,1,1]
            endif
            DO K=1,SELF%REC(1)%VOL_CTF2%logical_dimensions(3)
                DO J=1,SELF%REC(1)%VOL_CTF2%logical_dimensions(2)
                    DO I=1,ORI(1)
                        IF (I .EQ. ORI(1) .AND. (J .GT. ORI(2) .OR. K .GT. ORI(3))) CYCLE
                        IF (REC_NUM .EQ. 0) THEN
                            SIG_CTF = SIG_CTF + SELF%REC(1)%VOL_CTF2%real_values(I,J,K) + SELF%REC(2)%VOL_CTF2%real_values(I,J,K)
                        ELSE
                            SIG_CTF = SIG_CTF + SELF%REC(REC_NUM)%VOL_CTF2%real_values(I,J,K)
                        ENDIF
                    ENDDO
                ENDDO
            ENDDO
            IF (DEBUG) THEN
                PRINT '(A,G15.6)', '**debug(THREED_GUESS_WIENER_FILTER_TERM): REAL(SIG_CTF) before div by nce = ', REAL(SIG_CTF)
            ENDIF
            IF (REC_NUM .EQ. 0) THEN
                SIG_CTF = SIG_CTF / DBLE(SELF%REC(1)%NUM_CONTRIB_ELEMS + SELF%REC(2)%NUM_CONTRIB_ELEMS)
            ELSE
                SIG_CTF = SIG_CTF / DBLE(SELF%REC(REC_NUM)%NUM_CONTRIB_ELEMS)
            ENDIF


            SYM_REDUN = SELF%POINT_SYM%IREDUN
            IF (DEBUG) THEN
                PRINT '(A,I0)',  '**debug(THREED_GUESS_WIENER_FILTER_TERM): SELF%NUM_CONTRIB_ELEMS = ',SELF%REC(1)%NUM_CONTRIB_ELEMS
                !PRINT '(A,G0.6)', '**debug(THREED_GUESS_WIENER_FILTER_TERM): SUM_DOUBLE(CTF2_VOL) = ', SUM_DOUBLE(CTF2_VOL)
                PRINT '(A,G15.6)', '**debug(THREED_GUESS_WIENER_FILTER_TERM): REAL(SIG_CTF) = ', REAL(SIG_CTF)
                PRINT '(A,G15.3)', '**debug(THREED_GUESS_WIENER_FILTER_TERM): REAL(SYM_REDUM) = ', REAL(SYM_REDUN)
                PRINT '(A,G15.3)', '**debug(THREED_GUESS_WIENER_FILTER_TERM): REAL(SELF%NON_SYMMETRY_REDUNDANCY) = ', &
                                                                            REAL(SELF%NON_SYMMETRY_REDUNDANCY)
            ENDIF

            ! The actual calculation
            WIENER_FILTER_TERM = REAL(SIG_CTF)*REAL(SYM_REDUN)*0.1

        ENDIF   ! End of test for "a la Sachse 2007"

        ! We may or may not want to include an extra term for non-symmetry redundancy
        IF (INCLUDE_NSR_TERM_IN_WIENER) THEN
            WRITE(*,'(A)') '**debug(THREED_GUESS_WIENER_FILTER_TERM): including non-symmetry redundancy',   &
                            ' in calculation of Wiener-filter term'
            WIENER_FILTER_TERM = WIENER_FILTER_TERM*REAL(SELF%NON_SYMMETRY_REDUNDANCY)
        ENDIF

        IF (DEBUG) THEN
            !WRITE(*,'(A,<SIZE(WIENER_FILTER_TERM)>(G0.3,X))') '**debug(THREED_GUESS_WIENER_FILTER_TERM): wiener_filter_term = ', &
            !                                                    WIENER_FILTER_TERM
        ENDIF
    END FUNCTION THREED_GUESS_WIENER_FILTER_TERM



    !>  \brief  Add a slice (image) to a 3D reconstruction
    SUBROUTINE THREED_ADD_SLICE(SELF,SLICE_IN,CTF,ANGLES,SHIFT_X,SHIFT_Y,SUBTRACT,BFACTOR,APPLY_CTF_TO_SLICE,SLICE_PSI_ZERO,    &
                                ADD_TO_THREED_NUM)
        USE CTFS
        USE IMAGES_CTF
        USE IMAGES_THREED
        USE IMAGES
        USE EULERANGLES
        USE IMAGES_ARITHM
        USE ROTATION_MATRICES
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        USE OMP_LIB
        USE IMAGEFILES  ! debug only
        use UsefulFunctions, only : IsOdd, IsEven
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),               INTENT(INOUT)   ::  SELF                !<  3D reconstruction to update
        TYPE(IMAGE),                INTENT(IN)      ::  SLICE_IN            !<  Image data to insert (will be FT'ed internally if necessary)
        TYPE(CTF_PARAMETERS),       INTENT(IN)      ::  CTF                 !<  CTF parameters for that slice
        TYPE(EULER),                INTENT(IN)      ::  ANGLES              !<  Euler angles to insert slice
        REAL(KIND=8),   OPTIONAL,   INTENT(IN)      ::  SHIFT_X             !<  Shift to apply in X to slice before insertion
        REAL(KIND=8),   OPTIONAL,   INTENT(IN)      ::  SHIFT_Y             !<  Shift to apply in Y to slice before insertion
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  SUBTRACT            !<  If .TRUE., the slice (image) is subtracted from the reconstruction,
                                                                            !!  rather than added to it. .FALSE. by default.
        REAL,           OPTIONAL,   INTENT(IN)      ::  BFACTOR             !<  Optionally, apply this Bfactor (unitless) to the image when
                                                                            !!  inserting it into the reconstruction.
                                                                            !!  See Grigorieff (1998), Eq 3, and Grigorieff (2007), Eq 7.
                                                                            !!  B factor is expected to be unitless (to get this, take the B factor
                                                                            !!  in A^2 and divide by the square of the pixel size)
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  APPLY_CTF_TO_SLICE  !<  Set this to false if you've already applied the CTF
        LOGICAL,        OPTIONAL,   INTENT(IN)      ::  SLICE_PSI_ZERO      !<  Do not rotate the slice in plane (the CTF will still be rotated)
        INTEGER,        OPTIONAL,   INTENT(IN)      ::  ADD_TO_THREED_NUM   !<  Can be 1 or 2 (ignored unless an FSC is being computed)
        ! Private variables
        TYPE(IMAGE)             ::  SLICE_CTF
        TYPE(IMAGE)             ::  PROJ_CTF
        TYPE(ROTATION_MATRIX)   ::  MATRIX                      !   Rotation matrix to go from slice coordinates to 3D coordinates
        TYPE(ROTATION_MATRIX)   ::  MMATRIX                     !   Rotation matrix modified with point group symmetry for multiple slice insertion
        TYPE(ROTATION_MATRIX)   ::  MATRIX_PSI_ZERO, MMATRIX_PSI_ZERO
        LOGICAL                 ::  SSUBTRACT
        INTEGER                 ::  SUBTRACT_FACTOR
        INTEGER                 ::  I,J,K,NSYM,N
        INTEGER,ALLOCATABLE     ::  ISYM(:)
        INTEGER                 ::  SLICE_DIM(3)
        LOGICAL,    PARAMETER   ::  WIPE_THREED_ON_INPUT    =   .FALSE.
        REAL                    ::  SSHIFT_X,SSHIFT_Y
        LOGICAL,    PARAMETER   ::  DEBUG                   =   .FALSE.
        LOGICAL,    PARAMETER   ::  IGNORE_BFACTOR          =   .FALSE.
        LOGICAL,    PARAMETER   ::  FSC_EVERY_OTHER_SLICE   =   .FALSE.
        INTEGER,    PARAMETER   ::  PAD_FACTOR_FOR_CTF_MULT =   1           ! If >1, the slice will be padded by that factor before CTF is applied to it
        INTEGER                 ::  ADD_SLICE_TO_REC_NUM
        TYPE(IMAGE)             ::  SLICE
        INTEGER                 ::  NCE
        LOGICAL                 ::  AAPPLY_CTF_TO_SLICE, SSLICE_PSI_ZERO
        ! Start work

        SLICE = SLICE_IN

        ! Will we apply CTF to the slice internally?
        AAPPLY_CTF_TO_SLICE = .TRUE.
        IF (PRESENT(APPLY_CTF_TO_SLICE)) AAPPLY_CTF_TO_SLICE = APPLY_CTF_TO_SLICE

        ! Will we zero the slice's psi angle
        SSLICE_PSI_ZERO = .FALSE.
        IF (PRESENT(SLICE_PSI_ZERO)) SSLICE_PSI_ZERO = SLICE_PSI_ZERO

        ! Work out which reconstruction this slice will be added to
        IF (SELF%COMPUTE_FSC) THEN
            IF (THREED_NUM_OF_RECS(SELF) .NE. 2) THEN
                WRITE(*,'(2A,I0)') '**ERROR(THREED_ADD_SLICE): compute_fsc flag is T, but number of internal ', &
                                                            'reconstructions is not 2: ', &
                                                            THREED_NUM_OF_RECS(SELF)
                CALL TERMINATE('Fatal error in THREED_ADD_SLICE')
            ENDIF
            !$OMP   CRITICAL (THREED_ADD_SLICE_DECIDE_WHICH_THREED)
            IF (PRESENT(ADD_TO_THREED_NUM)) THEN
                ADD_SLICE_TO_REC_NUM = ADD_TO_THREED_NUM
            ELSE
                ADD_SLICE_TO_REC_NUM = SELF%ADD_NEXT_SLICE_TO_REC_NUM
                ! Update object so that next slice is added to other reconstruction
                IF (SELF%ADD_NEXT_SLICE_TO_REC_NUM .EQ. 1) THEN
                    SELF%ADD_NEXT_SLICE_TO_REC_NUM = 2
                ELSE IF (SELF%ADD_NEXT_SLICE_TO_REC_NUM .EQ. 2) THEN
                    SELF%ADD_NEXT_SLICE_TO_REC_NUM = 1
                ELSE
                    WRITE(*,'(A)') '**ERROR(THREED_ADD_SLICE): unexpected value of add_next_slice_to_rec_num = ', &
                                    SELF%ADD_NEXT_SLICE_TO_REC_NUM
                    CALL TERMINATE('Fatal error in THREED_ADD_SLICE. Programming error.')
                ENDIF
            ENDIF
            !$OMP   END CRITICAL (THREED_ADD_SLICE_DECIDE_WHICH_THREED)
        ELSE
            ADD_SLICE_TO_REC_NUM = 1
        ENDIF

        IF (DEBUG) WRITE(*,'(A,3(F0.2,1X),A,I0)') '**debug(THREED_ADD_SLICE): adding slice with euler angles ', &
                                                ANGLES%PHI, ANGLES%THETA, ANGLES%PSI, ' to reconstruction ', ADD_SLICE_TO_REC_NUM

        ! Will we be ignoring the B factor? (debugging)
        IF (IGNORE_BFACTOR .AND. PRESENT(BFACTOR)) THEN
            WRITE(*,'(A)') '**debug(THREED_ADD_SLICE): will ignore B factor'
        ENDIF

        ! Check CTF is in pixels
        IF (.NOT. CTF%PIX_UNITS) THEN
            WRITE(*,'(A)') '**ERROR(THREED_ADD_SLICE): need CTF in pixel units.'
            CALL TERMINATE('Fatal error in THREED_ADD_SLICE')
        ENDIF
        ! Check the THREED object is allocated
        IF (.NOT. THREED_ALLOCATED(SELF)) THEN
            WRITE(*,'(A)') '**ERROR(THREED_ADD_SLICE): 3D object not allocated'
            CALL TERMINATE('Fatal error in THREED_ADD_SLICE')
        ENDIF
        ! Check the IMAGE object is allocated
        IF (.NOT. IS_ALLOCATED(SLICE)) THEN
            WRITE(*,'(A)') '**ERROR(THREED_ADD_SLICE): slice object not allocated'
            CALL TERMINATE('Fatal error in THREED_ADD_SLICE')
        ENDIF
        ! Check the first dimension is even
        IF (IsOdd(SLICE%logical_dimensions(1))) THEN
            WRITE(*,'(A,I0)') '**ERROR(THREED_ADD_SLICE): odd first dimension for slice is not supported: ', &
                                SLICE%logical_dimensions(1)
            CALL TERMINATE('Fatal error in THREED_ADD_SLICE')
        ENDIF
        ! Are we subtracting?
        SSUBTRACT = .FALSE.
        IF (PRESENT(SUBTRACT))  SSUBTRACT = SUBTRACT

        ! If we're subtracting, we need to have a -1 later
        SUBTRACT_FACTOR = 1
        IF (SSUBTRACT) SUBTRACT_FACTOR = -1

        ! Shifts in X and Y default to 0.0 if not given
        SSHIFT_X = 0.0
        SSHIFT_Y = 0.0
        IF (PRESENT(SHIFT_X)) SSHIFT_X = SHIFT_X
        IF (PRESENT(SHIFT_Y)) SSHIFT_Y = SHIFT_Y


        !
        ! Wipe 3D on input if required
        !
        IF (WIPE_THREED_ON_INPUT) THEN
            ! Note: I am not sure that the SINGLE OpenMP directive garantees that the wipe will be done only by the first thread to get here.
            ! Therefore, I wouldn't necessarily trust results from this routine when run with threading
            IF (OMP_GET_ACTIVE_LEVEL() .GT. 0) THEN
                WRITE(*,'(3A)') '**WARNING(THREED_ADD_SLICE): this portion of the code has not been tested in parallel. ',   &
                                    'Results may be wrong. ', &
                                    'See in-source comment for more details.'
            ENDIF
            !$OMP   SINGLE
            CALL THREED_WIPE(SELF)
            !$OMP   END SINGLE
        ENDIF

        ! Pad if necessary
        IF (PAD_FACTOR_FOR_CTF_MULT .GT. 1 .AND. AAPPLY_CTF_TO_SLICE) THEN
            IF (.not. SLICE%is_in_real_space) CALL BWD_FT(SLICE)
            CALL PAD_BY_FACTOR(SLICE,PAD_FACTOR_FOR_CTF_MULT,PAD_WITH_VALUE=EDGE_AVERAGE(SLICE))
        ENDIF

        ! If the input image needs to be FT'ed, do it
        IF (SLICE%is_in_real_space) THEN
            CALL FWD_FT(SLICE)
        ENDIF

        !
        ! Apply the CTF to the slice
        !
        IF (AAPPLY_CTF_TO_SLICE) THEN
            PROJ_CTF%logical_dimensions(1) = MAXIMUM_COMPLEX_ADDRESS_DIM1(SLICE)
            PROJ_CTF%logical_dimensions(2) = SLICE%logical_dimensions(2)
            PROJ_CTF%logical_dimensions(3) = 1
            PROJ_CTF%is_in_real_space = .true.
            PROJ_CTF%ORIGIN_AT_CENTER = .FALSE.
            CALL IMAGE_ALLOCATE(PROJ_CTF)
            CALL CTF_TO_IMG(CTF,PROJ_CTF,ORIGIN_AT_CENTER = .FALSE.)
            FORALL(I=1:PROJ_CTF%logical_dimensions(1),J=1:PROJ_CTF%logical_dimensions(2),K=1:PROJ_CTF%logical_dimensions(3))
                SLICE%complex_values(I,J,K) = SLICE%complex_values(I,J,K) * PROJ_CTF%real_values(I,J,K)
            END FORALL
            CALL IMAGE_DEALLOCATE(PROJ_CTF)
        ENDIF

        ! Crop back to original size if necessary
        IF (PAD_FACTOR_FOR_CTF_MULT .GT. 1 .AND. AAPPLY_CTF_TO_SLICE) THEN
            IF (.not. SLICE%is_in_real_space) CALL BWD_FT(SLICE)
            CALL CROP_BY_FACTOR(SLICE,PAD_FACTOR_FOR_CTF_MULT)
            CALL FWD_FT(SLICE)
        ENDIF


        !
        ! Apply B-factor to the slice
        !
        IF (PRESENT(BFACTOR) .AND. .NOT. IGNORE_BFACTOR) THEN
            IF (BFACTOR .NE. 0.) THEN
                IF (SLICE%is_in_real_space) CALL FWD_FT(SLICE)
                IF (DEBUG) WRITE(*,'(A,F0.3,A)') '**debug(THREED_ADD_SLICE): applying B factor of ', BFACTOR, &
                                                                ' to slice before adding it to the reconstruction'
                CALL APPLY_BFACTOR(SLICE,BFACTOR,LP_COSINE=.FALSE.,PRINT_FILTER_SHAPE=.FALSE.,SQUARED=.TRUE.)
            ENDIF
        ENDIF

        !
        ! Remove anything beyond the Nyquist frequency in the slice
        !
        IF (.NOT. IsEven(SLICE%logical_dimensions(1))) CALL TERMINATE('Code only written for even dimensions')
        IF (SLICE%is_in_real_space) CALL FWD_FT(SLICE)
        CALL CIRC_MASK(SLICE,0.5)

        !
        ! Apply shift to the slice
        !
        IF ((ABS(SSHIFT_X) + ABS(SSHIFT_Y)) .GT. 0.0) THEN
            CALL PSHIFT_IMAGE(SLICE,SSHIFT_X,SSHIFT_Y,0.0)
        ENDIF


        ! For the next few operations, we need the slice to have its real-space origin in the center
        IF (.NOT. SLICE%ORIGIN_AT_CENTER) CALL SHIFT_REAL_SPACE_ORIGIN(SLICE)

        !
        ! Reverse FT
        !
        CALL BWD_FT(SLICE)

        !
        ! Crop or pad the box if necessary to match the dimension of the 3D reconstruction
        !
        SLICE_DIM = SELF%logical_dimensions
        SLICE_DIM(3) = 1
        CALL RESIZE(SLICE,SLICE_DIM)

        ! Check the slice has even dimensions
        IF (.NOT. HAS_EVEN_DIMS(SLICE)) THEN
            WRITE(*,'(A,2(I0,1X))') '**ERROR(THREED_ADD_SLICE): slice does not have even dimensions: ', &
                                    SLICE%logical_dimensions(1:2)
            CALL TERMINATE('Fatal error in THREED_ADD_SLICE')
        ENDIF

        !
        ! Apply "degridding" if necessary
        !
        IF (SELF%GET2D_PUT3D) THEN
            CALL CORRECT_3D(SLICE,SELF%INTERPOL_SCHEME%INTERPOLATION_INSTR_FUNCTION,    &
                            SELF%INTERPOL_SCHEME%WINDOWING_INSTR_FUNCTION,  &
                            SELF%INTERPOL_SCHEME%WINDOW_WIDTH)
        ENDIF


        !
        ! Forward FT back to reciprocal space
        !
        CALL FWD_FT(SLICE)

        !
        ! Phase shift the slice so its real-space origin is not in the center (needed for the insertion into the 3D vol)
        !
        IF (SLICE%ORIGIN_AT_CENTER) CALL SHIFT_REAL_SPACE_ORIGIN(SLICE)

        !
        ! Generate the CTF image corresponding to that slice
        !
        SLICE_CTF%logical_dimensions = SLICE%logical_dimensions
        SLICE_CTF%ORIGIN_AT_CENTER = .TRUE.
        SLICE_CTF%is_in_real_space = .true.
        CALL IMAGE_ALLOCATE(SLICE_CTF)
        CALL CTF_TO_IMG(CTF,SLICE_CTF,ORIGIN_AT_CENTER=.TRUE.)
        IF (DEBUG) THEN
            CALL CTF_PRINT_INFO(CTF)
            CALL WRITE_TO_DISK(SLICE_CTF,'debug_slice_ctf1.mrc')
        ENDIF
        ! Square the CTF
        !! \todo add an optional argument to CTF_TO_IMG so that it returns squared CTF already
        SLICE_CTF%real_values = SLICE_CTF%real_values**2
        ! Apply Bfactor to the CTF^2 slice
        IF (PRESENT(BFACTOR) .AND. .NOT. IGNORE_BFACTOR) THEN
            IF (BFACTOR .NE. 0.) THEN
                IF (DEBUG) WRITE(*,'(A,F0.3,A)') '**debug(THREED_ADD_SLICE): applying B factor of ', BFACTOR, &
                                                ' to slice CTF before adding it to the reconstruction'
                CALL APPLY_BFACTOR(SLICE_CTF,BFACTOR,LP_COSINE=.FALSE.,PRINT_FILTER_SHAPE=.FALSE.,SQUARED=.TRUE.)
            ENDIF
        ENDIF
        ! Mask the CTF to remove stuff beyond Nyq
        CALL CIRC_MASK(SLICE_CTF,REAL(SLICE_CTF%logical_dimensions(1))/2.)

        IF (DEBUG) THEN
            CALL WRITE_TO_DISK(SLICE_CTF,'debug_slice_ctf.mrc')
        ENDIF

        !
        ! Work out the rotation matrix for the insertion
        !
        MATRIX%RMATRIX          = EULER_TO_ROTMAT(ANGLES%PHI,ANGLES%THETA,ANGLES%PSI,REVERSE=.TRUE.)
        MATRIX_PSI_ZERO%RMATRIX = EULER_TO_ROTMAT(ANGLES%PHI,ANGLES%THETA,0.0D0     ,REVERSE=.TRUE.)

        !
        ! Do the insertions; insert IREDUN times for symmetries not equal to C1
        !
        NSYM = SIZE(SELF%POINT_SYM%SYMOP)
        ALLOCATE(ISYM(NSYM))
        DO I = 1,NSYM
            ISYM(I) = 0
        ENDDO

        N = 1
        DO WHILE (N.LE.NSYM)
            MMATRIX = MATRIX
            MMATRIX_PSI_ZERO = MATRIX_PSI_ZERO
            DO I = 1,NSYM
                IF (ISYM(I).NE.SELF%POINT_SYM%SYMOP(I)%JMAX) THEN
                    DO J = 1,ISYM(I)
                        MMATRIX%RMATRIX          =  MATMUL(SELF%POINT_SYM%SYMOP(I)%RMATRIX,MMATRIX%RMATRIX)
                        MMATRIX_PSI_ZERO%RMATRIX =  MATMUL(SELF%POINT_SYM%SYMOP(I)%RMATRIX,MMATRIX_PSI_ZERO%RMATRIX)
                    ENDDO
                ENDIF
            ENDDO

            ! Debug: we expect the volume to have its real-space origin in the corner
            IF (SELF%REC(ADD_SLICE_TO_REC_NUM)%VOL_IMAGE%ORIGIN_AT_CENTER) THEN
                WRITE(*,'(A)') '**warning(THREED_ADD_SLICE): unexpectedly, volume has real-space origin in center'
            ENDIF


            ! We are putting/spreading into 3D
            IF (SSLICE_PSI_ZERO) THEN
                CALL INSERT_SLICE_PUT(SELF%REC(ADD_SLICE_TO_REC_NUM)%VOL_IMAGE, SLICE,     MMATRIX_PSI_ZERO, &
                                                SELF%INTERPOL_SCHEME,SUBTRACT=SSUBTRACT,NUM_CONTRIB_ELEMS=NCE)
            ELSE
                CALL INSERT_SLICE_PUT(SELF%REC(ADD_SLICE_TO_REC_NUM)%VOL_IMAGE, SLICE,     MMATRIX, &
                                                SELF%INTERPOL_SCHEME,SUBTRACT=SSUBTRACT,NUM_CONTRIB_ELEMS=NCE)
            ENDIF
            IF (SELF%RECONSTRUCT_AMPLITUDES) THEN
            IF (SSLICE_PSI_ZERO) THEN
                CALL INSERT_SLICE_PUT(SELF%REC(ADD_SLICE_TO_REC_NUM)%VOL_AMPL, SLICE,     MMATRIX_PSI_ZERO, &
                                           SELF%INTERPOL_SCHEME,SUBTRACT=SSUBTRACT,NUM_CONTRIB_ELEMS=NCE,AMPLITUDES_ONLY=.TRUE.)
            ELSE
                CALL INSERT_SLICE_PUT(SELF%REC(ADD_SLICE_TO_REC_NUM)%VOL_AMPL, SLICE,     MMATRIX, &
                                           SELF%INTERPOL_SCHEME,SUBTRACT=SSUBTRACT,NUM_CONTRIB_ELEMS=NCE,AMPLITUDES_ONLY=.TRUE.)
            ENDIF
            ENDIF
            CALL INSERT_SLICE_PUT(SELF%REC(ADD_SLICE_TO_REC_NUM)%VOL_CTF2,  SLICE_CTF, MMATRIX, &
                                            SELF%INTERPOL_SCHEME,SUBTRACT=SSUBTRACT)


            ! Update the number of contributing elements
            !$OMP   CRITICAL        (THREED_SUM_NUM_CONTRIB_ELEMS)
            SELF%REC(ADD_SLICE_TO_REC_NUM)%NUM_CONTRIB_ELEMS = SELF%REC(ADD_SLICE_TO_REC_NUM)%NUM_CONTRIB_ELEMS + NCE
            !$OMP   END CRITICAL    (THREED_SUM_NUM_CONTRIB_ELEMS)

            N = 1
            DO WHILE (N.LE.NSYM)
                ISYM(N) = ISYM(N)+1
                IF (ISYM(N).GE.SELF%POINT_SYM%SYMOP(N)%JMAX) THEN
                    ISYM(N) = 0
                    N = N+1
                ELSE
                    EXIT
                ENDIF
            ENDDO
        ENDDO

        ! Now that we've added data, our PSSNR estimate is probably thrown off and so is our FSC
        SELF%SSNR_IS_UP_TO_DATE = .FALSE.
        SELF%FSC_IS_UP_TO_DATE = .FALSE.
    END SUBROUTINE THREED_ADD_SLICE


END MODULE THREEDS
