!>  \brief  Defines the THREED class, which handles 3D reconstruction from a set of images,
!!          including appropriate filters etc.
!!
MODULE THREEDS_BASE
    use Images
    USE INTERPOLATION
    USE POINT_SYMMETRIES
    USE HELICES, ONLY : HELICAL_SYMMETRY
    USE RUNTIME_PARAMETERS, ONLY : TERMINATE
    IMPLICIT NONE

    !PRIVATE :: THREED_RECONSTRUCTION

    TYPE THREED_RECONSTRUCTION
        !PRIVATE
        TYPE(IMAGE)                 ::  VOL_IMAGE
        TYPE(IMAGE)                 ::  VOL_CTF2
        TYPE(IMAGE)                 ::  VOL_AMPL                                !<  Volume for the reconstruction of the amplitudes (optional)
        INTEGER(KIND=8)             ::  NUM_CONTRIB_ELEMS       =   0           !<  Number of contributing elements
    END TYPE THREED_RECONSTRUCTION

    !>  \brief  A threed object includes all necessary data to generate a final 3D reconstruction, but keep track of intermediate
    !!          stages, for example, CTF**2 volume, half-dataset reconstructions etc.
    !!
    TYPE THREED
        !PRIVATE

        TYPE(THREED_RECONSTRUCTION),    ALLOCATABLE ::  REC(:)                  !<  Component reconstructions. Usually 2 (one for each half-dataset reconstructions) or just 1.

        INTEGER, PUBLIC             ::  logical_dimensions(3)   =   [1,1,1]     !<  Dimensions of the 3D reconstrction
        TYPE(POINT_SYMMETRY)        ::  POINT_SYM                               !<  Point group symmetry matrices
        INTEGER                     ::  NON_SYMMETRY_REDUNDANCY =   0           !<  Non-symmetry redundancy. For example, if the same slice is inserted several times for pseudo-symmetry.
        LOGICAL                     ::  IN_MEMORY               =   .FALSE.     !<  Whether memory has been allocated to the 3D object
        TYPE(INTERPOLATION_SCHEME)  ::  INTERPOL_SCHEME                         !<  The interpolation functions & parameters used to calculate the 3D reconstruction
        REAL                        ::  MASK_RAD                =   .99         !<  The radius of the mask to impose on the 3D reconstruction. In pixels.
        REAL                        ::  MASK_FALLOFF            =   .1          !<  The width (in pixels) of the dropoff in the mask from 1.0 to 0.0. The mask reaches 0.0 and MASK_RAD.
        ! SSNR
        LOGICAL                     ::  COMPUTE_FSC             =   .FALSE.     !<  Whether an FSC should be calculated. If so, two reconstructions will be accumulated.
        LOGICAL                     ::  FOM_FILTER              =   .FALSE.     !<  Whether a Figure of Merit filter should be applied to the reconstruction (see Henderson & Rosenthal 2003; equivalent flag in Frealign = FCREF)
        LOGICAL                     ::  USE_SSNR_ESTIMATE       =   .FALSE.     !<  Whether an estimate of the SSNR should be used as part of the Wiener filter
        INTEGER                     ::  ADD_NEXT_SLICE_TO_REC_NUM   =   1       !<  To which internal reconstruction should the next slice be added? Only useful in the context of half-dataset reconstructions, FSCs, SSNR estimationg etc.
        REAL,       ALLOCATABLE     ::  FSC(:)                                  !<  The FSC computed between half-dataset reconstructions
        LOGICAL                     ::  FSC_IS_UP_TO_DATE       =   .FALSE.     !<  Whether the FSC array is up to date
        REAL,       ALLOCATABLE     ::  SSNR_ESTIMATE(:)                        !<  An estimate of the SSNR of particle images that went into the 3D reconstruction.
        LOGICAL                     ::  SSNR_IS_UP_TO_DATE      =   .FALSE.     !<  Whether the PSSNR estimate is up-to-date
        REAL                        ::  F_MASK                  =   0.0         !<  The fraction of voxels within the binary mask
        REAL                        ::  F_SMOOTH                =   0.0         !<  The fraction of voxels within the smooth mask (computed a la Sindelar, 2011)
        ! Masking
        TYPE(HELICAL_SYMMETRY)      ::  HELIX                                   !<  Helical symmetry of the reconstruction
        REAL                        ::  ASYM_UNIT_MW            =   0.0         !<  Expected molecular weight (Da) of the asymmetric unit
        REAL                        ::  PIXEL_SIZE              =   0.0         !<  Voxel size (Angstroms)
        REAL                        ::  HELIX_MAX_RAD           =   0.0         !<  Maximum radius of the helical assembly (pixels)
        ! Amplitudes reconstruction
        LOGICAL                     ::  RECONSTRUCT_AMPLITUDES  =   .FALSE.     !<  Compute a reconstruction using only the input's amplitudes
        ! Experimental properties
        INTEGER                     ::  OVERSAMPLING_FACTOR     =   1           !<  Oversampling factor during 3D reconstruction.
                                                                                !!  This can be achieved, for example, by padding in real space (if reconstruction done in reciprocal space). Value of 1 means no oversampling.
        LOGICAL                     ::  GET2D_PUT3D             =   .FALSE.     !<  Whether to use the experimental way, which interpolates in the slice's grid, and then samples in 3D using nearest-neighbour. If .FALSE., the regular way is used.
        !CONTAINS
        !   PROCEDURE   ::  ADD_SLICE   =>  THREED_ADD_SLICE
        !   PROCEDURE   ::  EVAL_RECONSTRUCTION =>  THREED_EVAL_RECONSTRUCTION
        !    PROCEDURE   ::  GET_INTERPOL_SCHEME =>  THREED_GET_INTERPOL_SCHEME
        !    PROCEDURE   ::  GET_WINDOW_WIDTH    =>  THREED_GET_WINDOW_WIDTH
        !    FINAL       ::  THREED_DEALLOCATE
    END TYPE THREED

    INTERFACE IS_ALLOCATED
        MODULE PROCEDURE THREED_ALLOCATED
    END INTERFACE

    !INTERFACE DEALLOCATE
    !   MODULE PROCEDURE THREED_DEALLOCATE
    !END INTERFACE

    !INTERFACE ASSIGNMENT(=)
    !    MODULE PROCEDURE THREED_ASSIGNMENT
    !    MODULE PROCEDURE THREED_RECONSTRUCTION_ASSIGNMENT
    !END INTERFACE

    INTERFACE INIT
        MODULE PROCEDURE THREED_INIT
    END INTERFACE

    CONTAINS

    !>  \brief  Print out basic information about the THREED object
    SUBROUTINE THREED_PRINT_INFO(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),                   INTENT(IN)      ::  SELF
        ! Private variables

        ! Start work
        WRITE(*,'(A,3(I0,1X))') '**info(THREED_PRINT_INFO): dim = ', SELF%logical_dimensions
    END SUBROUTINE THREED_PRINT_INFO

    SUBROUTINE THREED_ASSIGNMENT(LHS,RHS)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),                   INTENT(INOUT)   ::  LHS
        TYPE(THREED),                   INTENT(IN)      ::  RHS
        ! Private variables
        INTEGER ::  CURRENT_REC
        ! Start work

        !
        LHS%logical_dimensions                         =   RHS%logical_dimensions
        CALL THREED_ALLOCATE(LHS,THREED_NUM_OF_RECS(RHS))

        DO CURRENT_REC=1,SIZE(LHS%REC)
            CALL THREED_RECONSTRUCTION_ASSIGNMENT(LHS%REC(CURRENT_REC),RHS%REC(CURRENT_REC))
        ENDDO

        LHS%POINT_SYM                   =   RHS%POINT_SYM
        LHS%NON_SYMMETRY_REDUNDANCY     =   RHS%NON_SYMMETRY_REDUNDANCY
        LHS%IN_MEMORY                   =   RHS%IN_MEMORY
        LHS%INTERPOL_SCHEME             =   RHS%INTERPOL_SCHEME
        LHS%MASK_RAD                    =   RHS%MASK_RAD
        LHS%MASK_FALLOFF                =   RHS%MASK_FALLOFF


        LHS%COMPUTE_FSC                 =   RHS%COMPUTE_FSC
        LHS%FOM_FILTER                  =   RHS%FOM_FILTER
        LHS%USE_SSNR_ESTIMATE           =   RHS%USE_SSNR_ESTIMATE
        LHS%ADD_NEXT_SLICE_TO_REC_NUM   =   RHS%ADD_NEXT_SLICE_TO_REC_NUM
        IF (ALLOCATED(RHS%FSC)) THEN
            IF (ALLOCATED(LHS%FSC)) DEALLOCATE(LHS%FSC)
            ALLOCATE(LHS%FSC(SIZE(RHS%FSC)))
            LHS%FSC                         =   RHS%FSC
        ENDIF
        LHS%FSC_IS_UP_TO_DATE           =   RHS%FSC_IS_UP_TO_DATE
        IF (ALLOCATED(RHS%SSNR_ESTIMATE)) THEN
            IF (ALLOCATED(LHS%SSNR_ESTIMATE)) DEALLOCATE(LHS%SSNR_ESTIMATE)
            ALLOCATE(LHS%SSNR_ESTIMATE(SIZE(RHS%SSNR_ESTIMATE)))
            LHS%SSNR_ESTIMATE               =   RHS%SSNR_ESTIMATE
        ENDIF
        LHS%SSNR_IS_UP_TO_DATE          =   RHS%SSNR_IS_UP_TO_DATE
        LHS%F_MASK                      =   RHS%F_MASK
        LHS%F_SMOOTH                    =   RHS%F_SMOOTH


        LHS%HELIX                       =   RHS%HELIX
        LHS%ASYM_UNIT_MW                =   RHS%ASYM_UNIT_MW
        LHS%PIXEL_SIZE                  =   RHS%PIXEL_SIZE
        LHS%HELIX_MAX_RAD               =   RHS%HELIX_MAX_RAD

        LHS%RECONSTRUCT_AMPLITUDES      =   RHS%RECONSTRUCT_AMPLITUDES

        LHS%OVERSAMPLING_FACTOR         =   RHS%OVERSAMPLING_FACTOR

        LHS%GET2D_PUT3D                 =   RHS%GET2D_PUT3D

    END SUBROUTINE THREED_ASSIGNMENT

    SUBROUTINE THREED_RECONSTRUCTION_ASSIGNMENT(LHS,RHS)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED_RECONSTRUCTION),    INTENT(INOUT)   ::  LHS
        TYPE(THREED_RECONSTRUCTION),    INTENT(IN)      ::  RHS
        ! Private variables
        LOGICAL     ::  RECONSTRUCT_AMPS
        ! Start work
        RECONSTRUCT_AMPS = PRODUCT(RHS%VOL_AMPL%logical_dimensions) .GT. 1
        CALL THREED_RECONSTRUCTION_DEALLOCATE(LHS)
        CALL THREED_RECONSTRUCTION_ALLOCATE(LHS,RHS%VOL_IMAGE%logical_dimensions,RECONSTRUCT_AMPS)
        LHS%VOL_IMAGE         = RHS%VOL_IMAGE
        LHS%VOL_CTF2          = RHS%VOL_CTF2
        IF (RECONSTRUCT_AMPS) LHS%VOL_AMPL          = RHS%VOL_AMPL
        LHS%NUM_CONTRIB_ELEMS = RHS%NUM_CONTRIB_ELEMS
    END SUBROUTINE THREED_RECONSTRUCTION_ASSIGNMENT

    SUBROUTINE THREED_RECONSTRUCTION_DEALLOCATE(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED_RECONSTRUCTION),    INTENT(INOUT)   ::  SELF
        ! Private variables

        ! Start work
        CALL IMAGE_DEALLOCATE(SELF%VOL_IMAGE)
        CALL IMAGE_DEALLOCATE(SELF%VOL_CTF2)
        CALL IMAGE_DEALLOCATE(SELF%VOL_AMPL)
        SELF%NUM_CONTRIB_ELEMS = 0
    END SUBROUTINE THREED_RECONSTRUCTION_DEALLOCATE

    SUBROUTINE THREED_RECONSTRUCTION_ALLOCATE(SELF,DIMS,RECONSTRUCT_AMPLITUDES)
        use images
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED_RECONSTRUCTION),    INTENT(INOUT)   ::  SELF
        INTEGER,                        INTENT(IN)      ::  DIMS(3)
        LOGICAL,                        INTENT(IN)      ::  RECONSTRUCT_AMPLITUDES
        ! Private variables
        LOGICAL,                        PARAMETER       ::  DEBUG   =   .FALSE.
        ! Start work
        ! Allocate memory for the image volume, which is complex
        IF (DEBUG) WRITE(*,'(A,3(I0,1X))') '**debug(THREED_RECONSTRUCTION_ALLOCATE): on entry, dims= ', DIMS
        SELF%VOL_IMAGE%logical_dimensions = DIMS
        SELF%VOL_IMAGE%is_in_real_space = .false.
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(THREED_RECONSTRUCTION_ALLOCATE): about to allocate image vol'
            CALL IMAGE_PRINT_INFO(SELF%VOL_IMAGE)
        ENDIF
        CALL IMAGE_ALLOCATE(SELF%VOL_IMAGE)
        ! Allocate memory to the CTF**2 volume, which is real
        SELF%VOL_CTF2%logical_dimensions = DIMS
        SELF%VOL_CTF2%is_in_real_space = .true.
        IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_RECONSTRUCTION_ALLOCATE): about to allocate ctf2 vol'
        CALL IMAGE_ALLOCATE(SELF%VOL_CTF2)
        ! Allocate memory to the amplitudes volume, which is real
        IF (RECONSTRUCT_AMPLITUDES) THEN
            SELF%VOL_AMPL%logical_dimensions = DIMS
            SELF%VOL_AMPL%is_in_real_space = .false.
            IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_RECONSTRUCTION_ALLOCATE): about to allocate ampl vol'
            CALL IMAGE_ALLOCATE(SELF%VOL_AMPL)
        ENDIF

    END SUBROUTINE THREED_RECONSTRUCTION_ALLOCATE

    PURE LOGICAL FUNCTION THREED_RECONSTRUCTION_ALLOCATED(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED_RECONSTRUCTION),    INTENT(IN)  ::  SELF
        ! Private variables

        ! Start work
        THREED_RECONSTRUCTION_ALLOCATED = SELF%VOL_IMAGE%IsAllocated() .AND. &
                                          SELF%VOL_CTF2%IsAllocated()
    END FUNCTION THREED_RECONSTRUCTION_ALLOCATED


    PURE LOGICAL FUNCTION THREED_ALLOCATED(SELF)
        IMPLICIT NONE
        ! Argument
        TYPE(THREED),   INTENT(IN)  ::  SELF
        ! Start work
        IF (ALLOCATED(SELF%REC)) THEN
            THREED_ALLOCATED = THREED_RECONSTRUCTION_ALLOCATED(SELF%REC(1))
            IF (SIZE(SELF%REC) .EQ. 2) THEN
                THREED_ALLOCATED = THREED_ALLOCATED .AND. THREED_RECONSTRUCTION_ALLOCATED(SELF%REC(2))
            ENDIF
            THREED_ALLOCATED = THREED_ALLOCATED .AND. SELF%IN_MEMORY
        ELSE
            THREED_ALLOCATED = .FALSE.
        ENDIF
    END FUNCTION THREED_ALLOCATED

        !>  \brief  Allocate memory and initialise a 3D reconstruction object. The interpolation can be passed as an object, or descriptors can be supplied.
    SUBROUTINE THREED_INIT( SELF,MASK_RAD,MASK_FALLOFF,INTERPOL_DESCR,WINDOW_DESCR,WINDOW_WIDTH,INTERPOL_SCHEME,    &
                            OVERSAMPLING_FACTOR, &
                            SYM_SYMBOL,NON_SYMMETRY_REDUNDANCY,FILENAME_ROOT,USE_SSNR_ESTIMATE,COMPUTE_FSC, &
                            HELIX,ASYM_UNIT_MW,PIXEL_SIZE,HELIX_MAX_RAD,NUM_OF_RECS,FOM_FILTER,RECONSTRUCT_AMPLITUDES)
        USE RUNTIME_PARAMETERS, ONLY : IS_ODD, TERMINATE, DEFAULT_FALLOFF_PIXELS
        IMPLICIT NONE
        ! Argument
        TYPE(THREED),                               INTENT(INOUT)   ::  SELF                    !<  3D reconstruction object to allocate & initialise
        REAL,                           OPTIONAL,   INTENT(IN)      ::  MASK_RAD                !<  The radius of the mask to impose on the slices and the 3D reconstruction. In pixels.
        REAL,                           OPTIONAL,   INTENT(IN)      ::  MASK_FALLOFF            !<  The width (in pixels) of the dropoff in the mask from 1.0 to 0.0. The mask reaches 0.0 and MASK_RAD.
        CHARACTER(LEN=*),               OPTIONAL,   INTENT(IN)      ::  INTERPOL_DESCR          !<  A string which describes the interpolation function to use for the 3D
        CHARACTER(LEN=*),               OPTIONAL,   INTENT(IN)      ::  WINDOW_DESCR            !<  A string which describes the windowing function to use in the 3D reconstruction
        INTEGER,                        OPTIONAL,   INTENT(IN)      ::  WINDOW_WIDTH            !<  The window width to use during 3D reconstruction
        TYPE(INTERPOLATION_SCHEME),     OPTIONAL,   INTENT(IN)      ::  INTERPOL_SCHEME         !<  Interpolation scheme (contains information about interpolation
        INTEGER,                        OPTIONAL,   INTENT(IN)      ::  OVERSAMPLING_FACTOR     !<  Oversampling factor (only used when following GET2D_PUT3D logic)
        TYPE(POINT_SYMMETRY_SYMBOL),    OPTIONAL,   INTENT(IN)      ::  SYM_SYMBOL              !<  Point group symmetry using Schonflies notation
        INTEGER,                        OPTIONAL,   INTENT(IN)      ::  NON_SYMMETRY_REDUNDANCY !<  Non-symmetry redundancy. For example, if the same slice is inserted several times for pseudo-symmetry.
        CHARACTER(LEN=*),               OPTIONAL,   INTENT(IN)      ::  FILENAME_ROOT           !<  Root filename for the 3D reconstruction. If given, will be read from disk.
        LOGICAL,                        OPTIONAL,   INTENT(IN)      ::  USE_SSNR_ESTIMATE       !<  Whether an estimate of the SSNR should be used in the Wiener filter
        LOGICAL,                        OPTIONAL,   INTENT(IN)      ::  COMPUTE_FSC             !<  Whether an FSC should be computed internally.
        TYPE(HELICAL_SYMMETRY),         OPTIONAL,   INTENT(IN)      ::  HELIX                   !<  Helical symmetry of the reconstruction
        REAL,                           OPTIONAL,   INTENT(IN)      ::  ASYM_UNIT_MW            !<  Expected molecular weight (Da) of the asymmetric unit
        REAL,                           OPTIONAL,   INTENT(IN)      ::  PIXEL_SIZE              !<  Voxel size (Angstroms)
        REAL,                           OPTIONAL,   INTENT(IN)      ::  HELIX_MAX_RAD           !<  Maximum radius of the helical assembly (pixels)
        INTEGER,                        OPTIONAL,   INTENT(IN)      ::  NUM_OF_RECS             !<  Use this to override the number of internal reconstructions
        LOGICAL,                        OPTIONAL,   INTENT(IN)      ::  FOM_FILTER              !<  Whether a FOM filter should be applied to the 3D reconstruction (a la Frelaign FCREF; Henderson & Rosenthal 2003)
        LOGICAL,                        OPTIONAL,   INTENT(IN)      ::  RECONSTRUCT_AMPLITUDES  !<  Compute a 3D reconstruction using only the 2D amplitudes
        ! Private variables
        TYPE(POINT_SYMMETRY_SYMBOL)                                 ::  SSYM_SYMBOL             !<  Point group symmetry using Schonflies notation, local
        INTEGER                                                     ::  NUM_OF_RECS_NEEDED, CURRENT_REC
        INTEGER                                                     ::  NUM_OF_RECS_ON_DISK, NNUM_OF_RECS
        INTEGER                                                     ::  IERR
        LOGICAL,                        PARAMETER                   ::  DEBUG   =   .FALSE.
        ! Start work

        ! Check how the user defined the interpolation scheme, and if necessary initialise the interpolation_scheme object
        IF (PRESENT(INTERPOL_SCHEME)) THEN
            SELF%INTERPOL_SCHEME = INTERPOL_SCHEME
        ELSE
            IF (PRESENT(WINDOW_DESCR) .AND. PRESENT(WINDOW_WIDTH) .AND. PRESENT(INTERPOL_DESCR)) THEN
                CALL INTERPOLATION_SCHEME_INIT(SELF%INTERPOL_SCHEME,INTERPOL_DESCR,WINDOW_DESCR,WINDOW_WIDTH)
            ELSE
                WRITE(*,'(2A)') '**info(THREED_INIT): no interpolation scheme information supplied.',    &
                                ' Make sure all required arguments are given. Defaulting to linear, rect, 2'
                CALL INTERPOLATION_SCHEME_INIT(SELF%INTERPOL_SCHEME,'LINEAR','RECT',2)
            ENDIF
        ENDIF

        ! Use SSNR estimate?
        IF (PRESENT(USE_SSNR_ESTIMATE))  SELF%USE_SSNR_ESTIMATE = USE_SSNR_ESTIMATE

        ! Compute FSC?
        IF (PRESENT(COMPUTE_FSC))  SELF%COMPUTE_FSC = COMPUTE_FSC

        ! Apply FOM filter?
        IF (PRESENT(FOM_FILTER)) SELF%FOM_FILTER = FOM_FILTER
        IF (SELF%FOM_FILTER) SELF%COMPUTE_FSC = .TRUE.

        ! Reconstruct amplitudes?
        IF (PRESENT(RECONSTRUCT_AMPLITUDES)) SELF%RECONSTRUCT_AMPLITUDES = RECONSTRUCT_AMPLITUDES

        ! We will need to compute an FSC to estimate the SSNR
        IF (SELF%USE_SSNR_ESTIMATE .AND. .NOT. SELF%COMPUTE_FSC) THEN
            WRITE(*,'(A)') '**warning(THREED_INIT): will need to compute an FSC to estimate the SSNR.', &
                            ' Overriding and setting COMPUTE_FSC to true'
            SELF%COMPUTE_FSC = .TRUE.
        ENDIF

        ! Will we need two reconstructions (e.g. to compute FSC)?
        IF (SELF%COMPUTE_FSC) THEN
            NUM_OF_RECS_NEEDED = 2
        ELSE
            NUM_OF_RECS_NEEDED = 1
        ENDIF

        ! If we'll be reading from disk, how many internal reconstructions are stored there?
        IF (PRESENT(FILENAME_ROOT)) THEN
            NUM_OF_RECS_ON_DISK = THREED_ON_DISK_NUM_OF_RECS(FILENAME_ROOT)
        ELSE
            NUM_OF_RECS_ON_DISK = 0
        ENDIF

        ! How many internal reconstructions will the 3D have?
        IF (PRESENT(NUM_OF_RECS)) THEN
            NNUM_OF_RECS = NUM_OF_RECS
        ELSE
            NNUM_OF_RECS = MAX(NUM_OF_RECS_NEEDED,NUM_OF_RECS_ON_DISK)
        ENDIF

        ! Now that we know how many internal reconstructions we have, we may overwrite the default value for USE_SSNR_ESTIMATE
!        IF (.NOT. PRESENT(USE_SSNR_ESTIMATE)) THEN
!            IF (NNUM_OF_RECS .EQ. 2) THEN
!                WRITE(*,'(A)') '**warning(THREED_INIT): 2 internal reconstructions therefore will use SSNR estimate via FSC. Overriding COMPUTE_FSC and USE_SSNR_ESTIMATE'
!                SELF%USE_SSNR_ESTIMATE = .TRUE.
!                SELF%COMPUTE_FSC = .TRUE.
!            ELSE
!                SELF%USE_SSNR_ESTIMATE = .FALSE.
!            ENDIF
!        ENDIF

        ! Deallocate memory
        IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_INIT): about to call threed_deallocate'
        CALL THREED_DEALLOCATE(SELF)

        ! Read from disk if a filename was given
        IF (PRESENT(FILENAME_ROOT)) THEN
            IF (NUM_OF_RECS_ON_DISK .LT. NUM_OF_RECS_NEEDED) THEN
                WRITE(*,'(A,I0,A,A,A,I0,A)') '**ERROR(THREED_INIT): ', NUM_OF_RECS_NEEDED , &
                                            ' 3D reconstructions are needed, but the reconstruction on disk (root filename ', &
                                            TRIM(ADJUSTL(FILENAME_ROOT)), ') only has ', NUM_OF_RECS_ON_DISK , ' reconstruction'
                CALL TERMINATE('Fatal error in THREED_INIT')
            ENDIF
            IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_INIT): about to allocate rec array'
            ALLOCATE(SELF%REC(NNUM_OF_RECS),STAT=IERR)
            IF (IERR .NE. 0) THEN
                WRITE(*,'(A,I0)') '**ERROR(THREED_INIT): rec array allocation failed with error ', IERR
                CALL TERMINATE('Failed memory allocation in THREED_INIT')
            ENDIF
            ! Read from disk
            CALL THREED_READ_FROM_DISK(SELF,FILENAME_ROOT)
        ELSE
            ! If the 3D reconstruction is not being read from disk, we need to set it up manually
            ! Check the dimensions have been set
            IF (ALL(SELF%logical_dimensions .EQ. [1,1,1])) THEN
                WRITE(*,'(A)') '**ERROR(THREED_INIT): dimensions for THREED object not set'
                CALL TERMINATE('Fatal error in THREED_INIT')
            ELSE IF ((ANY(SELF%logical_dimensions .GT. 1000000)) .OR. ANY(SELF%logical_dimensions .LE. 0)) THEN
                WRITE(*,'(A,3(I0,1X))') '**ERROR(THREED_INIT): strange dimensions for THREED object: ', &
                                        SELF%logical_dimensions
                CALL TERMINATE('Fatal error in THREED_INIT')
            ENDIF
            ! Allocate memory
            IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_INIT): about to call threed_allocate'
            CALL THREED_ALLOCATE(SELF,NNUM_OF_RECS)
            ! Zero the volumes
            CALL THREED_WIPE(SELF)
        ENDIF


        ! We'll need to keep track of the origin of the reconstruction
        DO CURRENT_REC=1,THREED_NUM_OF_RECS(SELF)
            SELF%REC(CURRENT_REC)%VOL_IMAGE%ORIGIN_AT_CENTER = .FALSE.
            SELF%REC(CURRENT_REC)%VOL_CTF2%ORIGIN_AT_CENTER = .TRUE.
            SELF%REC(CURRENT_REC)%VOL_AMPL%ORIGIN_AT_CENTER = .FALSE.
        ENDDO

        ! Only cubic volumes with even dimensions are supported
        IF (SELF%logical_dimensions(1) .NE. SELF%logical_dimensions(2) .OR. &
            SELF%logical_dimensions(1) .NE. SELF%logical_dimensions(3)) THEN
            WRITE(*,'(A,3(I0,1X))') '**ERROR(THREED_INIT): non-cubic reconstruction volumes are not yet supported: ', &
                                    SELF%logical_dimensions
            CALL TERMINATE('Fatal error in THREED_INIT')
        ENDIF
        IF (ANY(IS_ODD(SELF%logical_dimensions))) THEN
            WRITE(*,'(A,3(I0,1X))') '**ERROR(THREED_INIT): reconstruction volumes with odd dimensions are not yet supported: ', &
                                    SELF%logical_dimensions
            CALL TERMINATE('Fatal error in THREED_INIT')
        ENDIF

        ! Set the masking parameters
        IF (PRESENT(MASK_RAD)) THEN
            SELF%MASK_RAD       =   MASK_RAD
        ELSE
            SELF%MASK_RAD       =   REAL(MINVAL(SELF%logical_dimensions))/2. - 1.0
        ENDIF
        IF (DEBUG) PRINT *, '**debug(THREED_INIT): MASK_RAD = ', SELF%MASK_RAD
        IF (PRESENT(MASK_FALLOFF)) THEN
            SELF%MASK_FALLOFF   =   MASK_FALLOFF
        ELSE
            SELF%MASK_FALLOFF   =   DEFAULT_FALLOFF_PIXELS
        ENDIF

        ! Set oversampling factor
        IF (PRESENT(OVERSAMPLING_FACTOR)) THEN
            SELF%OVERSAMPLING_FACTOR    =   OVERSAMPLING_FACTOR
        ELSE
            SELF%OVERSAMPLING_FACTOR    =   1
        ENDIF
        ! Oversampling is currently only used in the case of GET2D_PUT3D
        IF (SELF%OVERSAMPLING_FACTOR .GT. 1 .AND. .NOT. SELF%GET2D_PUT3D) THEN
            PRINT *, '**warning(THREED_INIT): oversampling will be ignored, since not using GET2D_PUT3D logic'
        ENDIF
        ! Set symmetry matrices
        IF (PRESENT(SYM_SYMBOL)) THEN
            SSYM_SYMBOL = SYM_SYMBOL
        ELSE
            SSYM_SYMBOL%SYMBOL = 'C1 '
        ENDIF
        CALL POINT_SYMMETRY_INIT(SELF%POINT_SYM,SSYM_SYMBOL)
        IF (.NOT. SELF%POINT_SYM%HELICAL) CALL POINT_SYMMETRY_PRINT(SELF%POINT_SYM)
        ! Non-symmetry redundancy
        IF (PRESENT(NON_SYMMETRY_REDUNDANCY)) THEN
            SELF%NON_SYMMETRY_REDUNDANCY = NON_SYMMETRY_REDUNDANCY
        ELSE
            SELF%NON_SYMMETRY_REDUNDANCY = 1
        ENDIF
        PRINT '(A,I0)', '**info(THREED_INIT): non-symmetry redundancy = ', SELF%NON_SYMMETRY_REDUNDANCY

        ! Parameters needed for helical masking
        IF (PRESENT(HELIX)) SELF%HELIX = HELIX
        IF (PRESENT(ASYM_UNIT_MW)) SELF%ASYM_UNIT_MW = ASYM_UNIT_MW
        IF (PRESENT(PIXEL_SIZE)) SELF%PIXEL_SIZE = PIXEL_SIZE
        IF (PRESENT(HELIX_MAX_RAD)) SELF%HELIX_MAX_RAD = HELIX_MAX_RAD

        ! If the user asked to estimate SSNR, and the reconstruction is helical, we should check all relevant parameters have been set
        IF (SELF%USE_SSNR_ESTIMATE .AND. SELF%COMPUTE_FSC) THEN
            WRITE(*,'(2A)') '**info(THREED_INIT): two 3D reconstruction will be accumulated internally, ',   &
                            'and their masked FSC will be used to evaluate the SSNR'
            IF (SELF%POINT_SYM%HELICAL) THEN
                IF (.NOT. (PRESENT(HELIX) .AND. PRESENT(ASYM_UNIT_MW) .AND. PRESENT(PIXEL_SIZE) .AND. PRESENT(HELIX_MAX_RAD))    &
                    .OR. SELF%ASYM_UNIT_MW .EQ. 0. .OR. SELF%PIXEL_SIZE .EQ. 0. .OR. SELF%HELIX_MAX_RAD .EQ. 0.) THEN
                    WRITE(*,'(2A)') '**ERROR(THREED_INIT): SSNR estimation via FSC for helical reconstructions ',    &
                                    'requires that helical parameters be given to the threed object.'
                    CALL TERMINATE('Fatal error in THREED_INIT. Programming error')
                ENDIF
            ELSE
                WRITE(*,'(A)') '**ERROR(THREED_INIT): non-helical SSNR estimation via FSC not implemented yet'
                CALL TERMINATE('Feature not implemented yet')
            ENDIF
        ENDIF

        ! We don't have an SSNR estimate or FSC  yet
        SELF%SSNR_IS_UP_TO_DATE = .FALSE.
        SELF%FSC_IS_UP_TO_DATE = .FALSE.

        IF (DEBUG) WRITE(*,'(A,I0)') '**debug(THREED_INIT): all done. Number of internal 3D reconstructions = ', &
                                    THREED_NUM_OF_RECS(SELF)

    END SUBROUTINE THREED_INIT


    SUBROUTINE THREED_DEALLOCATE(SELF)
        IMPLICIT NONE
        ! Argument
        TYPE(THREED),   INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_REC
        LOGICAL, PARAMETER :: DEBUG = .FALSE.
        ! Start work
        IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_DEALLOCATE): entering'
        IF (THREED_ALLOCATED(SELF)) THEN
            DO CURRENT_REC=1,THREED_NUM_OF_RECS(SELF)
                IF (DEBUG) WRITE(*,'(2(A,I0))') '**debug(THREED_DEALLOCATE): about to deallocate rec ', &
                                            CURRENT_REC, ' of ', THREED_NUM_OF_RECS(SELF)
                CALL THREED_RECONSTRUCTION_DEALLOCATE(SELF%REC(CURRENT_REC))
            ENDDO
            IF (DEBUG) WRITE(*,'(A,I0)') '**debug(THREED_DEALLOCATE): deallocating rec array, size ', SIZE(SELF%REC)
            DEALLOCATE(SELF%REC)
            IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_DEALLOCATE): deallocated rec array'
        ELSE
            IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_DEALLOCATE): attempt to deallocate THREED which was not allocated'
        ENDIF
        IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_DEALLOCATE): about to deallocate rec array'
        IF (ALLOCATED(SELF%REC)) DEALLOCATE(SELF%REC)
        ! Set the allocated flag
        SELF%IN_MEMORY = .FALSE.
        !
        SELF%SSNR_IS_UP_TO_DATE = .FALSE.
        SELF%FSC_IS_UP_TO_DATE = .FALSE.
        IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_DEALLOCATE): deallocating real arrays'
        IF (ALLOCATED(SELF%SSNR_ESTIMATE)) DEALLOCATE(SELF%SSNR_ESTIMATE)
        IF (ALLOCATED(SELF%FSC)) DEALLOCATE(SELF%FSC)
        IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_DEALLOCATE): exiting'
    END SUBROUTINE THREED_DEALLOCATE

    SUBROUTINE THREED_ALLOCATE(SELF,NUM_RECS)
        USE RUNTIME_PARAMETERS, ONLY : TERMINATE
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),   INTENT(INOUT)   ::  SELF
        INTEGER,        INTENT(IN)      ::  NUM_RECS
        ! Private variables
        INTEGER                         ::  CURRENT_REC
        LOGICAL,        PARAMETER       ::  DEBUG   =   .FALSE.
        ! Start work
        IF (NUM_RECS .LT. 1 .OR. NUM_RECS .GT. 2) THEN
            WRITE(*,'(A,I0)') '**ERROR(THREED_ALLOCATE): number of reconstructions should be 1 or 2, not ', NUM_RECS
            CALL TERMINATE('Fatal error in THREED_ALLOCATE')
        ENDIF
        IF (DEBUG) THEN
            WRITE(*,'(A)') '**debug(THREED_ALLOCATE): threed info on entry'
            CALL THREED_PRINT_INFO(SELF)
        ENDIF
        IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_ALLOCATE): about to call threed_deallocate'
        CALL THREED_DEALLOCATE(SELF)
        IF (DEBUG) WRITE(*,'(A)') '**debug(THREED_ALLOCATE): about to allocate rec array'
        ALLOCATE(SELF%REC(NUM_RECS))
        DO CURRENT_REC=1,NUM_RECS
            IF (DEBUG) WRITE(*,'(A,I0)') '**debug(THREED_ALLOCATE): about to allocate rec # ', CURRENT_REC
            CALL THREED_RECONSTRUCTION_ALLOCATE(SELF%REC(CURRENT_REC),SELF%logical_dimensions,SELF%RECONSTRUCT_AMPLITUDES)
        ENDDO
        SELF%IN_MEMORY = .TRUE.
    END SUBROUTINE THREED_ALLOCATE

        !>  \brief  Return the window width of the interpolation scheme used by the THREED object
    PURE INTEGER FUNCTION THREED_GET_WINDOW_WIDTH(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),  INTENT(IN)      ::  SELF
        ! Start work
        THREED_GET_WINDOW_WIDTH = SELF%INTERPOL_SCHEME%WINDOW_WIDTH
    END FUNCTION THREED_GET_WINDOW_WIDTH

    !>  \brief  Return the interpolation scheme used by the THREED object
    !!  This function segfaults with ifort 11.1
    FUNCTION THREED_GET_INTERPOL_SCHEME(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),  INTENT(IN)      ::  SELF
        ! Result
        TYPE(INTERPOLATION_SCHEME)      ::  THREED_GET_INTERPOL_SCHEME
        ! Private variables

        ! Start work
        THREED_GET_INTERPOL_SCHEME = SELF%INTERPOL_SCHEME
    END FUNCTION THREED_GET_INTERPOL_SCHEME


    !>  \brief  Return the number of reconstructions accumulated (1 or 2)
    PURE INTEGER FUNCTION THREED_NUM_OF_RECS(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),       INTENT(IN)  ::  SELF
        ! Start work
        THREED_NUM_OF_RECS = SIZE(SELF%REC)
    END FUNCTION THREED_NUM_OF_RECS

    !>  \brief  Check the number of internal 3D reconstructions in a threed object stored on disk
    INTEGER FUNCTION THREED_ON_DISK_NUM_OF_RECS(FILENAME_ROOT)
        USE PARSERS
        USE RUNTIME_PARAMETERS, ONLY : FILENAME_LEN
        IMPLICIT NONE
        ! Argument
        CHARACTER(LEN=*),   INTENT(IN)      ::  FILENAME_ROOT   !<  Root filename for the files
        ! Private variables
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_IMAGE, FILENAMES_IMAGE(2)
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_CTF2!, FILENAMES_CTF2(2)
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_AMPL
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_STATS!, FILENAMES_STATS(2)  ! this file will contain two lines, 1. the number of contributing elements, 2. the non-symmetry redundancy

        ! Start work
        ! Get filenames for component files
        CALL THREED_FILENAMES_FROM_ROOT(FILENAME_ROOT,FILENAME_IMAGE,FILENAME_CTF2,FILENAME_STATS,FILENAME_AMPL)
        ! Check whether the input files are for one or two reconstructions
        CALL THREED_FILENAME_ADD_SUFFIXES(FILENAME_IMAGE,FILENAMES_IMAGE)
        IF (FILE_EXISTS(FILENAMES_IMAGE(1)) .AND. FILE_EXISTS(FILENAMES_IMAGE(2))) THEN
            THREED_ON_DISK_NUM_OF_RECS = 2
        ELSE
            THREED_ON_DISK_NUM_OF_RECS = 1
        ENDIF
    END FUNCTION THREED_ON_DISK_NUM_OF_RECS

    !>  \brief  Check whether a 3D object stored on disk includes the amplitude reconstruction
    LOGICAL FUNCTION THREED_ON_DISK_AMPL_VOL(FILENAME_ROOT)
        USE PARSERS
        USE RUNTIME_PARAMETERS, ONLY : FILENAME_LEN
        IMPLICIT NONE
        ! Argument
        CHARACTER(LEN=*),   INTENT(IN)  ::  FILENAME_ROOT
        ! Private variables
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_IMAGE
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_CTF2
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_AMPL, FILENAMES_AMPL(2)
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_STATS
        ! Start work
        ! Get filenames for component files
        CALL THREED_FILENAMES_FROM_ROOT(FILENAME_ROOT,FILENAME_IMAGE,FILENAME_CTF2,FILENAME_STATS,FILENAME_AMPL)
        IF (THREED_ON_DISK_NUM_OF_RECS(FILENAME_ROOT) .EQ. 2) THEN
            CALL THREED_FILENAME_ADD_SUFFIXES(FILENAME_AMPL,FILENAMES_AMPL)
            THREED_ON_DISK_AMPL_VOL = FILE_EXISTS(FILENAMES_AMPL(1)) .AND. FILE_EXISTS(FILENAMES_AMPL(2))
        ELSE
            THREED_ON_DISK_AMPL_VOL = FILE_EXISTS(FILENAME_AMPL)
        ENDIF
    END FUNCTION THREED_ON_DISK_AMPL_VOL


    !>  \brief  Set the non-symmetry redundancy (NSR) of the 3D reconstruction
    SUBROUTINE THREED_SET_NSR(SELF,NSR)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),  INTENT(INOUT)   ::  SELF
        INTEGER,        INTENT(IN)      ::  NSR
        ! Private variables

        ! Start work
        ! The 3D reconstruction object keeps track of the NSR
        IF (SELF%NON_SYMMETRY_REDUNDANCY .NE.1 .AND. SELF%NON_SYMMETRY_REDUNDANCY .NE. NSR) THEN
            WRITE(*,'(A,2(I5,1X))') '**ERROR(THREED_SET_NSR): mismatch between argument non-symmetry redundancy ',   &
                                    '(NSR) and NSR of incoming 3D reconstruction: ', &
                                    NSR, SELF%NON_SYMMETRY_REDUNDANCY
            CALL TERMINATE('Fatal error in THREED_SET_NSR')
        ENDIF
        SELF%NON_SYMMETRY_REDUNDANCY = NSR
    END SUBROUTINE THREED_SET_NSR

    !>  \brief  Check whether two threed reconstructions have been accumulated
    PURE LOGICAL FUNCTION THREED_TWO_RECS(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),       INTENT(IN)  ::  SELF    !<  3D reconstruction object
        ! Private variables

        ! Start work
        THREED_TWO_RECS  =                      ALL(IS_ALLOCATED(SELF%REC%VOL_IMAGE))   &
                                        .AND.   ALL(IS_ALLOCATED(SELF%REC%VOL_CTF2))    &
                                        .AND.   ALL(SELF%REC%NUM_CONTRIB_ELEMS .NE. 0)
    END FUNCTION THREED_TWO_RECS

    !>  \brief  Check the dimensions of the components of the threed object are consistent
    SUBROUTINE THREED_CHECK_DIMENSIONS(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),       INTENT(IN)  ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_REC
        ! Start work
        DO CURRENT_REC=1,THREED_NUM_OF_RECS(SELF)
            IF (ANY(SELF%REC(CURRENT_REC)%VOL_IMAGE%logical_dimensions .NE. SELF%logical_dimensions)) THEN
                WRITE(*,'(A,I0,A,3(I0,1X),A,3(I0,1X),A)') '**ERROR(THREED_CHECK_DIMENSIONS): dimensions for image volume ', &
                                                        CURRENT_REC, ' (',SELF%REC(CURRENT_REC)%VOL_IMAGE%logical_dimensions, &
                                                        ') do not match dimensions of threed object (', SELF%logical_dimensions, ')'
                CALL TERMINATE('Fatal error in THREED_CHECK_DIMENSIONS')
            ENDIF
            IF (ANY(SELF%REC(CURRENT_REC)%VOL_CTF2%logical_dimensions .NE. SELF%logical_dimensions)) THEN
                WRITE(*,'(A,I0,A,3(I0,1X),A,3(I0,1X),A)') '**ERROR(THREED_CHECK_DIMENSIONS): dimensions for CTF^2 volume ', &
                                                        CURRENT_REC, ' (',SELF%REC(CURRENT_REC)%VOL_IMAGE%logical_dimensions, &
                                                        ') do not match dimensions of threed object (', SELF%logical_dimensions, ')'
                CALL TERMINATE('Fatal error in THREED_CHECK_DIMENSIONS')
            ENDIF
        ENDDO
    END SUBROUTINE THREED_CHECK_DIMENSIONS

    !>  \brief  Merge the two internal reconstructions into a single one
    SUBROUTINE THREED_TO_SINGLE(SELF)
        use images
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),                   INTENT(INOUT)   ::  SELF
        ! Private variables
        TYPE(THREED_RECONSTRUCTION) ::  BAK(2)
        ! Start work

        ! Backup data
        CALL THREED_RECONSTRUCTION_ASSIGNMENT(BAK(1),SELF%REC(1))
        IF (SIZE(SELF%REC) .EQ. 2) CALL THREED_RECONSTRUCTION_ASSIGNMENT(BAK(2),SELF%REC(2))

        ! Reallocate
        CALL THREED_RECONSTRUCTION_DEALLOCATE(SELF%REC(1))
        CALL THREED_RECONSTRUCTION_DEALLOCATE(SELF%REC(2))
        DEALLOCATE(SELF%REC)
        ALLOCATE(SELF%REC(1))
        CALL IMAGE_ALLOCATE(SELF%REC(1)%VOL_IMAGE,BAK(1)%VOL_IMAGE)
        CALL IMAGE_ALLOCATE(SELF%REC(1)%VOL_CTF2, BAK(1)%VOL_CTF2)


        SELF%REC(1)%VOL_IMAGE%real_values  = BAK(1)%VOL_IMAGE%real_values  + BAK(2)%VOL_IMAGE%real_values
        SELF%REC(1)%VOL_CTF2%real_values   = BAK(1)%VOL_CTF2%real_values   + BAK(2)%VOL_CTF2%real_values
        SELF%REC(1)%NUM_CONTRIB_ELEMS = BAK(1)%NUM_CONTRIB_ELEMS + BAK(2)%NUM_CONTRIB_ELEMS
        IF (SELF%RECONSTRUCT_AMPLITUDES) THEN
        CALL IMAGE_ALLOCATE(SELF%REC(1)%VOL_AMPL,BAK(1)%VOL_AMPL)
        SELF%REC(1)%VOL_AMPL%real_values   = BAK(1)%VOL_AMPL%real_values   + BAK(2)%VOL_AMPL%real_values
        ENDIF
        SELF%IN_MEMORY = .TRUE.
    END SUBROUTINE THREED_TO_SINGLE

    !>  \brief  Merge two threeds each with one internal reconstruction
    !!          into a single threed object with two internal
    SUBROUTINE THREED_MERGE(INPUT1,INPUT2)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),                   INTENT(INOUT)   ::  INPUT1
        TYPE(THREED),                   INTENT(IN)      ::  INPUT2
        ! Private variables
        TYPE(THREED_RECONSTRUCTION) ::  TMP_REC(2)
        ! Start work

        ! Check the input threeds only have one internal reconstruction
        IF (THREED_NUM_OF_RECS(INPUT1) .NE. 1 .OR. &
            THREED_NUM_OF_RECS(INPUT2) .NE. 1 ) THEN
            WRITE(*,'(2A,2(I0,1X))')  '**ERROR(THREED_MERGE): input 3D reconstruction objects should have one ',  &
                                    'and only one internal reconstruction, but they have: ', &
                                    THREED_NUM_OF_RECS(INPUT1), &
                                    THREED_NUM_OF_RECS(INPUT2)
            CALL TERMINATE('Fatal error in THREED_MERGE. Programming error.')
        ENDIF

!        ! Initialise the output threed
!        CALL THREED_INIT(   OUTPUT,                                 &
!                            MASK_RAD=INPUT1%MASK_RAD,               &
!                            MASK_FALLOFF=INPUT1%MASK_FALLOFF,       &
!                            INTERPOL_SCHEME=INPUT1%INTERPOL_SCHEME, &
!                            OVERSAMPLING_FACTOR=INPUT1%OVERSAMPLING_FACTOR,&
!                            SYM_SYMBOL=INPUT1%POINT_SYM%SYMBOL,     &
!                            NON_SYMMETRY_REDUNDANCY=INPUT1%NON_SYMMETRY_REDUNDANCY, &
!                            USE_SSNR_ESTIMATE=INPUT1%USE_SSNR_ESTIMATE, &
!                            COMPUTE_FSC=INPUT1%COMPUTE_FSC,         &
!                            HELIX=INPUT1%HELIX,                     &
!                            ASYM_UNIT_MW=INPUT1%ASYM_UNIT_MW,       &
!                            PIXEL_SIZE=INPUT1%PIXEL_SIZE,           &
!                            HELIX_MAX_RAD=INPUT1%HELIX_MAX_RAD,     &
!                            NUM_OF_RECS=2,                          &
!                            FOM_FILTER=INPUT1%FOM_FILTER  )

        ! Copy the data over
        TMP_REC(1)%VOL_IMAGE = INPUT1%REC(1)%VOL_IMAGE
        TMP_REC(1)%VOL_CTF2 = INPUT1%REC(1)%VOL_CTF2
        TMP_REC(1)%NUM_CONTRIB_ELEMS = INPUT1%REC(1)%NUM_CONTRIB_ELEMS
        IF (INPUT1%RECONSTRUCT_AMPLITUDES) TMP_REC(1)%VOL_AMPL = INPUT1%REC(1)%VOL_AMPL
        TMP_REC(2)%VOL_IMAGE = INPUT2%REC(1)%VOL_IMAGE
        TMP_REC(2)%VOL_CTF2 = INPUT2%REC(1)%VOL_CTF2
        TMP_REC(2)%NUM_CONTRIB_ELEMS = INPUT2%REC(1)%NUM_CONTRIB_ELEMS
        IF (INPUT2%RECONSTRUCT_AMPLITUDES) TMP_REC(2)%VOL_AMPL = INPUT2%REC(1)%VOL_AMPL

        ! Reallocate
        CALL IMAGE_DEALLOCATE(INPUT1%REC(1)%VOL_IMAGE)
        CALL IMAGE_DEALLOCATE(INPUT1%REC(1)%VOL_CTF2)
        IF (INPUT1%RECONSTRUCT_AMPLITUDES) CALL IMAGE_DEALLOCATE(INPUT1%REC(1)%VOL_AMPL)
        DEALLOCATE(INPUT1%REC)
        ALLOCATE(INPUT1%REC(2))
        INPUT1%REC(1)%VOL_IMAGE = TMP_REC(1)%VOL_IMAGE
        INPUT1%REC(1)%VOL_CTF2 = TMP_REC(1)%VOL_CTF2
        INPUT1%REC(1)%NUM_CONTRIB_ELEMS = TMP_REC(1)%NUM_CONTRIB_ELEMS
        INPUT1%REC(2)%VOL_IMAGE = TMP_REC(2)%VOL_IMAGE
        INPUT1%REC(2)%VOL_CTF2 = TMP_REC(2)%VOL_CTF2
        INPUT1%REC(2)%NUM_CONTRIB_ELEMS = TMP_REC(2)%NUM_CONTRIB_ELEMS
        IF (INPUT1%RECONSTRUCT_AMPLITUDES) INPUT1%REC(1)%VOL_AMPL = TMP_REC(1)%VOL_AMPL
        IF (INPUT2%RECONSTRUCT_AMPLITUDES) INPUT1%REC(2)%VOL_AMPL = TMP_REC(2)%VOL_AMPL


    END SUBROUTINE THREED_MERGE

    !>  \brief  Sum two threeds
    !!
    !!  Input 2 is added to (or subtracted from) INPUT1, to give OUTPUT
    SUBROUTINE THREED_ADD(INPUT1,INPUT2,OUTPUT,SUBTRACT)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),                   INTENT(IN)      ::  INPUT1
        TYPE(THREED),                   INTENT(IN)      ::  INPUT2
        TYPE(THREED),                   INTENT(INOUT)   ::  OUTPUT
        LOGICAL,        OPTIONAL,       INTENT(IN)      ::  SUBTRACT    !<  Whether to subtract INPUT2 from INPUT1, rather than add. .FALSE. by default.
        ! Private variables
        REAL            ::  SSIGN
        INTEGER         ::  NUM_OF_RECS
        INTEGER         ::  I,J,K, CURRENT_REC
        TYPE(THREED)    ::  INPUT1_TMP
        ! Start work

        ! Are the two 3Ds of same dimensions
        IF (ANY(INPUT1%logical_dimensions .NE. INPUT2%logical_dimensions)) THEN
            WRITE(*,'(2(A,3(I0,1X)))') '**ERROR(THREED_ADD): the two 3D reconstructions do not have the same dimensions: ', &
                                        INPUT1%logical_dimensions, ' and ', INPUT2%logical_dimensions
        ENDIF

        ! Either both or neither reconstructions has an amplitude volume
        IF ((INPUT1%RECONSTRUCT_AMPLITUDES .AND. .NOT. INPUT2%RECONSTRUCT_AMPLITUDES) .OR. &
            (INPUT2%RECONSTRUCT_AMPLITUDES .AND. .NOT. INPUT1%RECONSTRUCT_AMPLITUDES)) THEN
            WRITE(*,'(A)') '**ERROR(THREED_ADD): only one of the 3D reconstructions includes an amplitude-only volume'
            CALL TERMINATE('Fatal error in THREED_ADD')
        ENDIF

        ! Are we subtracting or adding? (adding by default)
        SSIGN = 1.0
        IF (PRESENT(SUBTRACT)) THEN
            IF (SUBTRACT) THEN
                SSIGN = -1.0
            ENDIF
        ENDIF

        ! How many reconstructions?
        NUM_OF_RECS = THREED_NUM_OF_RECS(INPUT1)
        IF (NUM_OF_RECS .EQ. 2 .AND. (THREED_NUM_OF_RECS(INPUT2) .EQ. 1) .AND. (THREED_NUM_OF_RECS(OUTPUT) .EQ. 1)) THEN
            !WRITE(*,'(A)') '**debug(THREED_ADD): first but not second 3D has two reconstructions. Merging to single rec'
            CALL THREED_ASSIGNMENT(INPUT1_TMP,INPUT1)
            CALL THREED_TO_SINGLE(INPUT1_TMP)
        ELSE
            CALL THREED_ASSIGNMENT(INPUT1_TMP,INPUT1)
        ENDIF
        NUM_OF_RECS = THREED_NUM_OF_RECS(INPUT1_TMP)
        IF (THREED_NUM_OF_RECS(INPUT1_TMP) .NE. THREED_NUM_OF_RECS(INPUT2) .OR. NUM_OF_RECS .NE. THREED_NUM_OF_RECS(OUTPUT)) THEN
            WRITE(*,'(2A,3(I0,1X))') '**ERROR(THREED_ADD): input and output 3D reconstruction objects do not contain ',   &
                                    'the same number of 3D reconstructions: ', &
                                    THREED_NUM_OF_RECS(INPUT1_TMP), THREED_NUM_OF_RECS(INPUT2), THREED_NUM_OF_RECS(OUTPUT)
            CALL TERMINATE('Fatal error in THREED_ADD')
        ENDIF

        ! Check allocation
        IF (.NOT. THREED_ALLOCATED(INPUT1_TMP) .OR. .NOT. THREED_ALLOCATED(INPUT2) .OR. .NOT. THREED_ALLOCATED(OUTPUT)) THEN
            WRITE(*,'(A,3L1)') '**ERROR(THREED_ADD): not all threed objects allocated: ',   THREED_ALLOCATED(INPUT1_TMP),   &
                                                                                            THREED_ALLOCATED(INPUT2),       &
                                                                                            THREED_ALLOCATED(OUTPUT)
            CALL TERMINATE('Fatal error in THREED_ADD')
        ENDIF

        ! Do the actual addition
        !WRITE(*,'(A,I0)') '**debug(THREED_ADD): about to do actual addition, NUM_OF_RECS = ', NUM_OF_RECS
        !WRITE(*,'(A,3(I0,1X))') '**debug(THREED_ADD): dims = ', INPUT1_TMP%logical_dimensions
        DO CURRENT_REC=1,NUM_OF_RECS
            DO K=1,INPUT1_TMP%logical_dimensions(3)
                DO J=1,INPUT1_TMP%logical_dimensions(2)
                    DO I=1,INPUT1_TMP%logical_dimensions(1)
                        IF (I .LE. MAXIMUM_COMPLEX_ADDRESS_DIM1(INPUT1_TMP%REC(CURRENT_REC)%VOL_IMAGE)) THEN
                        OUTPUT%REC(CURRENT_REC)%VOL_IMAGE%complex_values(I,J,K) = &
                                INPUT1_TMP%REC(CURRENT_REC)%VOL_IMAGE%complex_values(I,J,K) &
                              + INPUT2%REC(CURRENT_REC)%VOL_IMAGE%complex_values(I,J,K) * SSIGN
                        IF (INPUT1_TMP%RECONSTRUCT_AMPLITUDES) THEN
                        OUTPUT%REC(CURRENT_REC)%VOL_AMPL%complex_values(I,J,K) =  &
                                INPUT1_TMP%REC(CURRENT_REC)%VOL_AMPL%complex_values(I,J,K)     &
                              + INPUT2%REC(CURRENT_REC)%VOL_AMPL%complex_values(I,J,K) * SSIGN
                        ENDIF
                        ENDIF
                        OUTPUT%REC(CURRENT_REC)%VOL_CTF2%real_values(I,J,K) = &
                                INPUT1_TMP%REC(CURRENT_REC)%VOL_CTF2%real_values(I,J,K) + &
                                INPUT2%REC(CURRENT_REC)%VOL_CTF2%real_values(I,J,K) * SSIGN

                    ENDDO
                ENDDO
            ENDDO
            OUTPUT%REC(CURRENT_REC)%NUM_CONTRIB_ELEMS = INPUT1_TMP%REC(CURRENT_REC)%NUM_CONTRIB_ELEMS + &
                                                        INPUT2%REC(CURRENT_REC)%NUM_CONTRIB_ELEMS * INT(SSIGN)
        ENDDO

        ! Deal with the non-symmetry redundancy
        IF (INPUT1_TMP%NON_SYMMETRY_REDUNDANCY .EQ. 1 .OR. INPUT2%NON_SYMMETRY_REDUNDANCY .EQ. 1) THEN
            IF (INPUT1_TMP%NON_SYMMETRY_REDUNDANCY .EQ. 1) THEN
                OUTPUT%NON_SYMMETRY_REDUNDANCY = INPUT2%NON_SYMMETRY_REDUNDANCY
            ELSE IF (INPUT2%NON_SYMMETRY_REDUNDANCY .EQ. 1) THEN
                OUTPUT%NON_SYMMETRY_REDUNDANCY = INPUT1_TMP%NON_SYMMETRY_REDUNDANCY
            ENDIF
        ELSE
            IF (INPUT1_TMP%NON_SYMMETRY_REDUNDANCY .NE. INPUT2%NON_SYMMETRY_REDUNDANCY) THEN
                WRITE(*,'(A,2(I5,1X))') '**ERROR(THREED_ADD): the two input 3Ds do not have the same non-symmetry redundancy: ', &
                                        INPUT1_TMP%NON_SYMMETRY_REDUNDANCY, INPUT2%NON_SYMMETRY_REDUNDANCY
                CALL TERMINATE('Fatal error in THREED_ADD')
            ENDIF
        ENDIF
        IF (    INPUT1_TMP%NON_SYMMETRY_REDUNDANCY .NE. OUTPUT%NON_SYMMETRY_REDUNDANCY .AND. &
                INPUT1_TMP%NON_SYMMETRY_REDUNDANCY .EQ. INPUT2%NON_SYMMETRY_REDUNDANCY) THEN
            WRITE(*,'(2A,2(I5,1X))') '**ERROR(THREED_ADD): the input 3Ds do not have the same ',  &
                                    'non-symmetry redundancy as the output 3D: ',    &
                                    INPUT1_TMP%NON_SYMMETRY_REDUNDANCY, OUTPUT%NON_SYMMETRY_REDUNDANCY
            CALL TERMINATE('Fatal error in THREED_ADD')
        ENDIF

        CALL THREED_DEALLOCATE(INPUT1_TMP)

        ! The output's PSSNR estimate & FSC are not up to date anymore
        OUTPUT%SSNR_IS_UP_TO_DATE = .FALSE.
        OUTPUT%FSC_IS_UP_TO_DATE = .FALSE.
    END SUBROUTINE THREED_ADD

    !>  \brief  Fills the 3D's volumes and data to 0.
    SUBROUTINE THREED_WIPE(SELF)
        IMPLICIT NONE
        ! Argument
        TYPE(THREED),   INTENT(INOUT)   ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_REC
        ! Start work
        DO CURRENT_REC=1,THREED_NUM_OF_RECS(SELF)
            SELF%REC(CURRENT_REC)%VOL_CTF2%real_values = 0.0
            SELF%REC(CURRENT_REC)%VOL_IMAGE%complex_values = (0.0,0.0)
            SELF%REC(CURRENT_REC)%NUM_CONTRIB_ELEMS = 0
            IF (SELF%RECONSTRUCT_AMPLITUDES) SELF%REC(CURRENT_REC)%VOL_AMPL%complex_values = (0.0,0.0)
        ENDDO
        SELF%SSNR_IS_UP_TO_DATE = .FALSE.
    END SUBROUTINE THREED_WIPE

    SUBROUTINE THREED_FILENAME_ADD_SUFFIXES(FILENAME,FILENAMES)
        USE PARSERS
        IMPLICIT NONE
        ! Arguments
        CHARACTER(LEN=*),   INTENT(IN)  ::  FILENAME
        CHARACTER(LEN=*),   INTENT(OUT) ::  FILENAMES(2)
        ! Private variables

        ! Start work
        CALL FILENAME_ADD_SUFFIX(FILENAME,'_1of2',FILENAMES(1))
        CALL FILENAME_ADD_SUFFIX(FILENAME,'_2of2',FILENAMES(2))
    END SUBROUTINE THREED_FILENAME_ADD_SUFFIXES

    !>  \brief  Given a root filename for a 3D reconstruction object, return the actual filenames for the component files
    SUBROUTINE THREED_FILENAMES_FROM_ROOT(FILENAME_ROOT,FILENAME_IMAGE,FILENAME_CTF2,FILENAME_STATS,FILENAME_AMPL)
        USE PARSERS
        USE RUNTIME_PARAMETERS, ONLY : FILENAME_LEN
        IMPLICIT NONE
        ! Arguments
        CHARACTER(LEN=*),   INTENT(IN)  ::  FILENAME_ROOT   !<  Root filename
        CHARACTER(LEN=*),   INTENT(OUT) ::  FILENAME_IMAGE  !<  Filename for volume containing image data
        CHARACTER(LEN=*),   INTENT(OUT) ::  FILENAME_CTF2   !<  Filename for volume containing squared CTF data
        CHARACTER(LEN=*),   INTENT(OUT) ::  FILENAME_STATS  !<  Filename for text file containing 3D reconstruction statistics
        CHARACTER(LEN=*),   INTENT(OUT) ::  FILENAME_AMPL   !<  Filename for the amplitudes volume
        ! Private variables
        CHARACTER(LEN=FILENAME_LEN)     ::  TMP_STRING, TMP_STRING_NOEXT
        CHARACTER(LEN=3)                ::  TMP_STRING_EXT
        ! Start work

        TMP_STRING = TRIM(ADJUSTL(FILENAME_ROOT))
        TMP_STRING_EXT = EXTENSION_FROM_FILENAME(TMP_STRING)
        IF (STRING_EQUAL(TMP_STRING_EXT,'   ')) THEN
            ! If the filename_root has no extension
            TMP_STRING_NOEXT = TMP_STRING
            TMP_STRING_EXT = 'mrc'
        ELSE
            TMP_STRING_NOEXT = TMP_STRING(1:LEN_TRIM(TMP_STRING)-4)
        ENDIF
        WRITE(FILENAME_IMAGE,'(A,A,A1,A3)') TRIM(ADJUSTL(TMP_STRING_NOEXT)), '_unweighted', '.', TMP_STRING_EXT
        WRITE(FILENAME_CTF2 ,'(A,A,A1,A3)') TRIM(ADJUSTL(TMP_STRING_NOEXT)), '_ctfweights', '.', TMP_STRING_EXT
        WRITE(FILENAME_AMPL,'(A,A,A1,A3)') TRIM(ADJUSTL(TMP_STRING_NOEXT)), '_ampl','.', TMP_STRING_EXT
        WRITE(FILENAME_STATS  ,'(A,A,A1,A3)') TRIM(ADJUSTL(TMP_STRING_NOEXT)), '_stats', '.', 'dat'
    END SUBROUTINE THREED_FILENAMES_FROM_ROOT


    !>  \brief  Read a (non-finalised) 3D reconstruction from disk (i.e. both parts, the unweighted image data, and the CTF^2 weights).
    !!          You need to supply a filename with extension.
    !!  Warning: the resulting 3D reconstruction object is not fully initialised (it is missing the interpolation scheme, as well as masking parameters etc.)
    SUBROUTINE THREED_READ_FROM_DISK(SELF,FILENAME_ROOT)
        USE PARSERS
        USE RUNTIME_PARAMETERS, ONLY : FILENAME_LEN, AVAILABLE_LUN, LUN_IS_NOW_AVAILABLE
        use Images
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),      INTENT(INOUT)   ::  SELF            !<  3D reconstruction object to save to disk
        CHARACTER(LEN=*),   INTENT(IN)      ::  FILENAME_ROOT   !<  Root filename for the files
        ! Private variables
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_IMAGE, FILENAMES_IMAGE(2)
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_CTF2, FILENAMES_CTF2(2)
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_AMPL, FILENAMES_AMPL(2)
        CHARACTER(LEN=FILENAME_LEN) ::  FILENAME_STATS, FILENAMES_STATS(2)  ! this file will contain two lines, 1. the number of contributing elements, 2. the non-symmetry redundancy

        INTEGER ::  IOS ! I/O status
        INTEGER ::  CURRENT_REC
        INTEGER ::  FILE_HANDLE
        LOGICAL,    PARAMETER   ::  DEBUG   =   .FALSE.
        ! Start work
        IF (IS_BLANK(EXTENSION_FROM_FILENAME(FILENAME_ROOT))) THEN
            WRITE(*,'(A)') '**ERROR(THREED_WRITE_TO_DISK): code not written for case where FILENAME_ROOT does not have an extension'
            CALL TERMINATE('Fatal error in THREED_WRITE_TO_DISK')
        ELSE
            ! Get an available file handle
            FILE_HANDLE = AVAILABLE_LUN()
            ! Get filenames for component files
            CALL THREED_FILENAMES_FROM_ROOT(FILENAME_ROOT,FILENAME_IMAGE,FILENAME_CTF2,FILENAME_STATS,FILENAME_AMPL)
            ! Check whether the input files are for one or two reconstructions
            IF (THREED_ON_DISK_NUM_OF_RECS(FILENAME_ROOT) .EQ. 2) THEN
                IF (THREED_NUM_OF_RECS(SELF) .NE. 2) THEN
                    WRITE(*,'(2A,I0)') '**ERROR(THREED_READ_FROM_DISK): reading from files describing two internal ',    &
                                        'reconstructions into a threed object with ', &
                                        THREED_NUM_OF_RECS(SELF)
                    CALL TERMINATE('Fatal error in THREED_READ_FROM_DISK')
                ENDIF
            ENDIF
            !
            IF ((THREED_ON_DISK_AMPL_VOL(FILENAME_ROOT) .AND. .NOT. SELF%RECONSTRUCT_AMPLITUDES) .OR. &
                (.NOT. THREED_ON_DISK_AMPL_VOL(FILENAME_ROOT) .AND. SELF%RECONSTRUCT_AMPLITUDES)) THEN
                IF (SELF%RECONSTRUCT_AMPLITUDES) THEN
                    WRITE(*,'(A)') '**ERROR(THREED_READ_FROM_DISK): amplitude reconstruction(s) on disk expected but not found'
                ELSE
                    WRITE(*,'(A)') '**ERROR(THREED_READ_FROM_DISK): amplitude reconstruction(s) on disk found but not expected'
                ENDIF
                CALL TERMINATE('Fatal error in THREED_READ_FROM_DISK')
            ENDIF

            ! If needed, get filenames for both reconstructions
            IF (THREED_NUM_OF_RECS(SELF) .EQ. 2) THEN
                CALL THREED_FILENAME_ADD_SUFFIXES(FILENAME_IMAGE,FILENAMES_IMAGE)
                CALL THREED_FILENAME_ADD_SUFFIXES(FILENAME_CTF2,FILENAMES_CTF2)
                CALL THREED_FILENAME_ADD_SUFFIXES(FILENAME_STATS,FILENAMES_STATS)
                CALL THREED_FILENAME_ADD_SUFFIXES(FILENAME_AMPL,FILENAMES_AMPL)
            ELSE IF (THREED_NUM_OF_RECS(SELF) .EQ. 1) THEN
                FILENAMES_IMAGE = FILENAME_IMAGE
                FILENAMES_CTF2 = FILENAME_CTF2
                FILENAMES_STATS = FILENAME_STATS
                FILENAMES_AMPL = FILENAME_AMPL
            ELSE
                WRITE(*,'(A)') '**ERROR(THREED_READ_FROM_DISK): unexpected THREED_NUM_OF_RECS(SELF) = ', THREED_NUM_OF_RECS(SELF)
                CALL TERMINATE('Programming error')
            ENDIF
            DO CURRENT_REC=1,THREED_NUM_OF_RECS(SELF)
                IF (DEBUG) WRITE(*,'(A,I0,A,I0)') '**debug(THREED_READ_FROM_DISK): on internal reconstruction ', CURRENT_REC, &
                                                                                ' of ', THREED_NUM_OF_RECS(SELF)
                IF (DEBUG) WRITE(*,'(5(A,1X))') '**debug(THREED_READ_FROM_DISK): filenames:',    &
                                                        TRIM(ADJUSTL(FILENAMES_IMAGE(CURRENT_REC))), &
                                                        TRIM(ADJUSTL(FILENAMES_CTF2(CURRENT_REC))), &
                                                        TRIM(ADJUSTL(FILENAMES_STATS(CURRENT_REC))), &
                                                        TRIM(ADJUSTL(FILENAMES_AMPL(CURRENT_REC)))
                ! Read image data
                !CALL READ_FROM_DISK(SELF%REC(CURRENT_REC)%VOL_IMAGE,FILENAMES_IMAGE(CURRENT_REC))
                CALL SELF%REC(CURRENT_REC)%VOL_IMAGE%ReadFromDisk(FILENAMES_IMAGE(CURRENT_REC),read_volume_not_slice=.true.)
                IF (DEBUG) WRITE(*,'(A,3(I0,1X))') '**debug(THREED_READ_FROM_DISK): image dimensions: ', &
                                                        SELF%REC(CURRENT_REC)%VOL_IMAGE%logical_dimensions
                ! Check whether it is blank
                IF (IMAGE_IS_BLANK(SELF%REC(CURRENT_REC)%VOL_IMAGE)) THEN
                    WRITE(*,'(3A)') '**warning(THREED_READ_FROM_DISK): image volume (', &
                                    TRIM(ADJUSTL(FILENAMES_IMAGE(CURRENT_REC))),&
                                    ') is blank'
                ENDIF
                ! Check whether it has NaN value(s)
                IF (HAS_NAN(SELF%REC(CURRENT_REC)%VOL_IMAGE)) THEN
                    WRITE(*,'(3A)') '**warning(THREED_READ_FROM_DISK): image volume (', &
                                    TRIM(ADJUSTL(FILENAMES_IMAGE(CURRENT_REC))),&
                                    ') has at least one NaN voxel value'
                ENDIF
                CALL FWD_FT(SELF%REC(CURRENT_REC)%VOL_IMAGE)
                CALL SHIFT_REAL_SPACE_ORIGIN(SELF%REC(CURRENT_REC)%VOL_IMAGE)

                ! Read in amplitudes
                IF (SELF%RECONSTRUCT_AMPLITUDES) THEN
                    !CALL READ_FROM_DISK(SELF%REC(CURRENT_REC)%VOL_AMPL, FILENAMES_AMPL(CURRENT_REC))
                    CALL SELF%REC(CURRENT_REC)%VOL_AMPL%ReadFromDisk(FILENAMES_AMPL(CURRENT_REC),read_volume_not_slice=.true.)
                    ! Check whether it is blank
                    IF (IMAGE_IS_BLANK(SELF%REC(CURRENT_REC)%VOL_AMPL)) THEN
                        WRITE(*,'(3A)') '**warning(THREED_READ_FROM_DISK): AMPL volume (', &
                                        TRIM(ADJUSTL(FILENAMES_AMPL(CURRENT_REC))),&
                                        ') is blank'
                    ENDIF
                    ! Check whether it has NaN value(s)
                    IF (HAS_NAN(SELF%REC(CURRENT_REC)%VOL_AMPL)) THEN
                        WRITE(*,'(3A)') '**warning(THREED_READ_FROM_DISK): AMPL volume (', &
                                        TRIM(ADJUSTL(FILENAMES_AMPL(CURRENT_REC))),&
                                        ') has at least one NaN voxel value'
                    ENDIF
                    CALL FWD_FT(SELF%REC(CURRENT_REC)%VOL_AMPL)
                    CALL SHIFT_REAL_SPACE_ORIGIN(SELF%REC(CURRENT_REC)%VOL_AMPL)
                ENDIF

                ! Read CTF2 data
                !CALL READ_FROM_DISK(SELF%REC(CURRENT_REC)%VOL_CTF2, FILENAMES_CTF2(CURRENT_REC))
                CALL SELF%REC(CURRENT_REC)%VOL_CTF2%ReadFromDisk(FILENAMES_CTF2(CURRENT_REC),read_volume_not_slice=.true.)
                ! Check whether it is blank
                IF (IMAGE_IS_BLANK(SELF%REC(CURRENT_REC)%VOL_CTF2)) THEN
                    WRITE(*,'(3A)') '**warning(THREED_READ_FROM_DISK): CTF2 volume (', TRIM(ADJUSTL(FILENAMES_CTF2(CURRENT_REC))),&
                                    ') is blank'
                ENDIF
                ! Check whether it has NaN value(s)
                IF (HAS_NAN(SELF%REC(CURRENT_REC)%VOL_CTF2)) THEN
                    WRITE(*,'(3A)') '**warning(THREED_READ_FROM_DISK): CTF2 volume (', TRIM(ADJUSTL(FILENAMES_CTF2(CURRENT_REC))),&
                                    ') has at least one NaN voxel value'
                ENDIF

                ! Read in stats
                OPEN(UNIT=FILE_HANDLE,FILE=TRIM(ADJUSTL(FILENAMES_STATS(CURRENT_REC))),IOSTAT=IOS)
                IF (IOS .NE. 0) THEN
                    WRITE(*,'(A)') '**ERROR(THREED_READ_FROM_DISK): error opening stats file ', IOS
                    CALL TERMINATE('Fatal error in THREED_READ_FROM_DISK')
                ELSE
                    READ(FILE_HANDLE,'(I12)') SELF%REC(CURRENT_REC)%NUM_CONTRIB_ELEMS
                    READ(FILE_HANDLE,'(I12)') SELF%NON_SYMMETRY_REDUNDANCY
                ENDIF
                CLOSE(UNIT=FILE_HANDLE)
            ENDDO
            IF (DEBUG) WRITE(*,'(A,3(i0,1x))') '**debug(THREED_READ_FROM_DISK): setting threed dimensions to ', &
                                                SELF%REC(1)%VOL_IMAGE%logical_dimensions
            SELF%logical_dimensions = SELF%REC(1)%VOL_IMAGE%logical_dimensions
            IF (DEBUG) WRITE(*,'(A,3(i0,1x))') '**debug(THREED_READ_FROM_DISK): set threed dimensions to ', &
                                                SELF%REC(1)%VOL_IMAGE%logical_dimensions
            CALL THREED_CHECK_DIMENSIONS(SELF)
            SELF%IN_MEMORY = .TRUE.
            ! Release file handle
            CALL LUN_IS_NOW_AVAILABLE(FILE_HANDLE)
        ENDIF
    END SUBROUTINE THREED_READ_FROM_DISK

    !>  \brief  Check whether the 3D reconstruction has any NaN values in there
    LOGICAL FUNCTION THREED_HAS_NAN(SELF)
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),   INTENT(IN)  ::  SELF
        ! Private variables
        INTEGER ::  CURRENT_REC
        ! Start work
        THREED_HAS_NAN = .FALSE.
        DO CURRENT_REC=1,THREED_NUM_OF_RECS(SELF)
            ! Check whether it has NaN value(s)
            IF (HAS_NAN(SELF%REC(CURRENT_REC)%VOL_IMAGE)) THEN
                THREED_HAS_NAN = .TRUE.
                RETURN
            ENDIF
            IF (HAS_NAN(SELF%REC(CURRENT_REC)%VOL_CTF2)) THEN
                THREED_HAS_NAN = .TRUE.
                RETURN
            ENDIF
        ENDDO
    END FUNCTION THREED_HAS_NAN


    !>  \brief  Dump a (non-finalised) 3D reconstruction to disk (i.e. both parts, the unweighted image data, and the CTF^2 weights).
    !!          You need to supply a filename with extension. The number of contributing elements is written into a separate text file.
    SUBROUTINE THREED_WRITE_TO_DISK(SELF,FILENAME_ROOT)
        USE PARSERS, ONLY : IS_BLANK, EXTENSION_FROM_FILENAME
        USE IMAGEFILES, ONLY : IMAGEFILE
        USE RUNTIME_PARAMETERS, ONLY : FILENAME_LEN, AVAILABLE_LUN, LUN_IS_NOW_AVAILABLE
        !USE IMAGES_CTF, ONLY : CTFVOL_SHIFT ! DEBUG
        use images
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),      INTENT(INOUT)   ::  SELF            !<  3D reconstruction object to save to disk
        CHARACTER(LEN=*),   INTENT(IN)      ::  FILENAME_ROOT   !<  Root filename for the files
        ! Private variables
        CHARACTER(LEN=FILENAME_LEN)     ::  FILENAME_IMAGE, FILENAMES_IMAGE(2)
        CHARACTER(LEN=FILENAME_LEN)     ::  FILENAME_CTF2, FILENAMES_CTF2(2)
        CHARACTER(LEN=FILENAME_LEN)     ::  FILENAME_AMPL, FILENAMES_AMPL(2)
        CHARACTER(LEN=FILENAME_LEN)     ::  FILENAME_STATS, FILENAMES_STATS(2)  ! this file will contain two lines,
                                                                                ! 1. the number of contributing elements,
                                                                                ! 2. the non-symmetry redundancy
        INTEGER                         ::  IOS ! I/O status

        INTEGER                         ::  CURRENT_REC
        INTEGER                         ::  FILE_HANDLE
        !TYPE(IMAGE)                     ::  TMPCTF
        ! Start work
        IF (IS_BLANK(EXTENSION_FROM_FILENAME(FILENAME_ROOT))) THEN
            WRITE(*,'(A)') '**ERROR(THREED_WRITE_TO_DISK): code not written for case where FILENAME_ROOT does not have an extension'
            CALL TERMINATE('Fatal error in THREED_WRITE_TO_DISK')
        ELSE
            ! Get an available file handle
            FILE_HANDLE = AVAILABLE_LUN()
            ! Work out filenames to write to
            CALL THREED_FILENAMES_FROM_ROOT(FILENAME_ROOT,FILENAME_IMAGE,FILENAME_CTF2,FILENAME_STATS,FILENAME_AMPL)
            ! If needed, get filenames for both reconstructions
            IF (THREED_NUM_OF_RECS(SELF) .EQ. 2) THEN
                CALL THREED_FILENAME_ADD_SUFFIXES(FILENAME_IMAGE,FILENAMES_IMAGE)
                CALL THREED_FILENAME_ADD_SUFFIXES(FILENAME_CTF2,FILENAMES_CTF2)
                CALL THREED_FILENAME_ADD_SUFFIXES(FILENAME_AMPL,FILENAMES_AMPL)
                CALL THREED_FILENAME_ADD_SUFFIXES(FILENAME_STATS,FILENAMES_STATS)
            ELSE IF (THREED_NUM_OF_RECS(SELF) .EQ. 1) THEN
                FILENAMES_IMAGE = FILENAME_IMAGE
                FILENAMES_CTF2 = FILENAME_CTF2
                FILENAMES_STATS = FILENAME_STATS
                FILENAMES_AMPL = FILENAME_AMPL
            ELSE
                WRITE(*,'(A)') '**ERROR(THREED_WRITE_TO_DISK): unexpected THREED_NUM_OF_RECS(SELF) = ', THREED_NUM_OF_RECS(SELF)
                CALL TERMINATE('Programming error')
            ENDIF

            DO CURRENT_REC=1,THREED_NUM_OF_RECS(SELF)
                ! Write the unweighted 3D reconstruction to disk
                !WRITE(*,'(A,L1)') '**debug(THREED_WRITE_TO_DISK): about to shift origin. Currently at center? ', SELF%REC(CURRENT_REC)%VOL_IMAGE%ORIGIN_AT_CENTER
                CALL SHIFT_REAL_SPACE_ORIGIN(SELF%REC(CURRENT_REC)%VOL_IMAGE)
                CALL BWD_FT(SELF%REC(CURRENT_REC)%VOL_IMAGE)
                ! Check whether it is blank
                IF (IMAGE_IS_BLANK(SELF%REC(CURRENT_REC)%VOL_IMAGE)) THEN
                    WRITE(*,'(3A)') '**warning(THREED_WRITE_TO_DISK): image volume (', TRIM(ADJUSTL(FILENAMES_IMAGE(CURRENT_REC))),&
                                    ') is blank'
                ENDIF
                ! Check whether it has NaN value(s)
                IF (HAS_NAN(SELF%REC(CURRENT_REC)%VOL_IMAGE)) THEN
                    WRITE(*,'(3A)') '**warning(THREED_WRITE_TO_DISK): image volume (', TRIM(ADJUSTL(FILENAMES_IMAGE(CURRENT_REC))),&
                                    ') has at least one NaN voxel value'
                ENDIF
                CALL SELF%REC(CURRENT_REC)%VOL_IMAGE%WriteToDisk(FILENAMES_IMAGE(CURRENT_REC),    &
                                            PIXEL_SIZE=SELF%PIXEL_SIZE)
                CALL FWD_FT(SELF%REC(CURRENT_REC)%VOL_IMAGE)
                CALL SHIFT_REAL_SPACE_ORIGIN(SELF%REC(CURRENT_REC)%VOL_IMAGE)

                ! Write the amplitudes reconstruction to disk
                IF (SELF%RECONSTRUCT_AMPLITUDES) THEN
                    CALL SHIFT_REAL_SPACE_ORIGIN(SELF%REC(CURRENT_REC)%VOL_AMPL)
                    CALL BWD_FT(SELF%REC(CURRENT_REC)%VOL_AMPL)
                    ! Check whether it is blank
                    IF (IMAGE_IS_BLANK(SELF%REC(CURRENT_REC)%VOL_AMPL)) THEN
                        WRITE(*,'(3A)') '**warning(THREED_WRITE_TO_DISK): AMPL volume (', &
                                        TRIM(ADJUSTL(FILENAMES_AMPL(CURRENT_REC))),&
                                        ') is blank'
                    ENDIF
                    ! Check whether it has NaN value(s)
                    IF (HAS_NAN(SELF%REC(CURRENT_REC)%VOL_AMPL)) THEN
                        WRITE(*,'(3A)') '**warning(THREED_WRITE_TO_DISK): AMPL volume (', &
                                        TRIM(ADJUSTL(FILENAMES_AMPL(CURRENT_REC))),&
                                        ') has at least one NaN voxel value'
                    ENDIF

                    CALL SELF%REC(CURRENT_REC)%VOL_AMPL%WriteToDisk(FILENAMES_AMPL(CURRENT_REC),&
                                                PIXEL_SIZE=SELF%PIXEL_SIZE)
                    CALL FWD_FT(SELF%REC(CURRENT_REC)%VOL_AMPL)
                    CALL SHIFT_REAL_SPACE_ORIGIN(SELF%REC(CURRENT_REC)%VOL_AMPL)
                ENDIF

                ! Write the CTF^2 volume to disk
                ! Check whether it is blank
                IF (IMAGE_IS_BLANK(SELF%REC(CURRENT_REC)%VOL_CTF2)) THEN
                    WRITE(*,'(3A)') '**warning(THREED_WRITE_TO_DISK): CTF2 volume (', TRIM(ADJUSTL(FILENAMES_CTF2(CURRENT_REC))),&
                                    ') is blank'
                ENDIF
                ! Check whether it has NaN value(s)
                IF (HAS_NAN(SELF%REC(CURRENT_REC)%VOL_CTF2)) THEN
                    WRITE(*,'(3A)') '**warning(THREED_WRITE_TO_DISK): CTF2 volume (', TRIM(ADJUSTL(FILENAMES_CTF2(CURRENT_REC))),&
                                    ') has at least one NaN voxel value'
                ENDIF
                !CALL CTFVOL_SHIFT(SELF%REC(CURRENT_REC)%VOL_CTF2,TMPCTF)
                !CALL CTF2_FILE%WRITE_IMAGE(TMPCTF,FILENAME=FILENAMES_CTF2(CURRENT_REC),&
                !                            PIXEL_SIZE=SELF%PIXEL_SIZE)
                CALL SELF%REC(CURRENT_REC)%VOL_CTF2%WriteToDisk(FILENAMES_CTF2(CURRENT_REC),&
                                            PIXEL_SIZE=SELF%PIXEL_SIZE)


                ! Write the number of contributing elements to a flat text file
                OPEN(UNIT=FILE_HANDLE,FILE=TRIM(ADJUSTL(FILENAMES_STATS(CURRENT_REC))),STATUS='REPLACE',IOSTAT=IOS)
                IF (IOS .NE. 0) THEN
                    WRITE(*,'(A)') '**ERROR(THREED_WRITE_TO_DISK): error opening STATS file ', IOS
                    CALL TERMINATE('Fatal error in THREED_WRITE_TO_DISK')
                ELSE
                    WRITE(FILE_HANDLE,'(I0)') SELF%REC(CURRENT_REC)%NUM_CONTRIB_ELEMS
                    WRITE(FILE_HANDLE,'(I0)') SELF%NON_SYMMETRY_REDUNDANCY
                ENDIF
                CLOSE(UNIT=FILE_HANDLE)
            ENDDO
            ! Release file handle
            CALL LUN_IS_NOW_AVAILABLE(FILE_HANDLE)
        ENDIF
    END SUBROUTINE THREED_WRITE_TO_DISK

    !>  \brief  Print the 3D's FSC to terminal
    SUBROUTINE THREED_PRINT_FSC(SELF)
        USE IMAGES_COMPARISON, ONLY : PRINT_FSC
        IMPLICIT NONE
        ! Arguments
        TYPE(THREED),   INTENT(IN)  ::  SELF
        ! Private variables

        ! Start work

        IF (.NOT. SELF%FSC_IS_UP_TO_DATE) THEN
            WRITE(*,'(A)') '**warning(THREED_PRINT_FSC): FSC is not up-to-date'
        ENDIF

        WRITE(*,'(A)') '**info(THREED_PRINT_FSC): FSC between half-dataset reconstructions follows'
        CALL PRINT_FSC(SELF%FSC)

    END SUBROUTINE THREED_PRINT_FSC





END MODULE THREEDS_BASE
