!> \brief Module describing the image derived type
!!
!!  Warning: when USEing this module, and if using an ONLY clause to the USE statement, make sure you include the IMAGE_ASSIGNMENT
!!  routine, otherwise the = operator may not do what you expect. It may be simpler to just "use Images" without an ONLY
!!  clause
!!
!!
!!  Some definitions:
!!  - logical address: set of 3 integer index values which define a voxel, such that the origin has logical adress (0,0,0). Positive and negative.
!!  - physical address: set of 3 integer index values which define the position within the Fortran 3D array of a voxel (i,j,k). Allowed values for the nth dimension are from 1 to self%logical_dimensions(n).
!!  - coordinates (shortened to coos): set of 3 real values which define the position of a point within the volume, relative to the origin. In units of pixels or 1/pixels
!!
!!  Note that at the time of writing (August 2011), most routines in Frealix
!!  do not follow this nomenclature. In particular, COOS is often used to mean
!!  logical address. This (mis-)use of the term coordinates/COOS is deprecated.
!!
!!  !!  Here is an example of how to generate a hard sphere and use a gaussian filter to soften it:
!!  ! First, we have to generate the mask, which is a hard sphere
!!  CALL ALLOCATE(MASK3D,MODEL_3D_READY)
!!  FILTER3D%ORIGIN_AT_CENTER = .TRUE.
!!  CALL GENERATE_FILTER_IMAGE(MASK3D,FILTER=DISC,FILTER_PARAM=MASK3D%logical_dimensions(1)/2.0*0.8)
!!  ! We now prepare a gaussian filter to soften the sphere's edge
!!  CALL ALLOCATE(FILTER3D,MASK3D)
!!  FILTER3D%ORIGIN_AT_CENTER = .FALSE.
!!  CALL GENERATE_FILTER_IMAGE(FILTER3D,FILTER=GAUSS_FWHM,FILTER_PARAM=FILTER3D%logical_dimensions(1)/2.0*0.8)
!!  ! We now apply the filter to generate the soft mask
!!  CALL FWD_FT(MASK3D)
!!  CALL APPLY_FILTER(MASK3D,FILTER3D)
!!  CALL BWD_FT(MASK3D)
!!  ! Debug dump
!!  !CALL WRITE_TO_DISK(MASK3D,'debug_soft_sphere.mrc')
!!  ! Free up some space
!!  CALL DEALLOCATE(FILTER3D)
!!
!!
!!  \todo   Switch all functions over to subroutines to avoid sub-optimal performance when OpenMP threading
!!  \todo   make as much of this as possible private, to move towards full object-oriented programming
module Images

    use Globals
    use, intrinsic :: iso_c_binding
    use fftw33
    use runtime_parameters, only : terminate

    ! Eventually, this will need to be private
    !private

    public  ::  highest_index_dim1, lowest_index_for_negative_frequencies, comp_coo_from_phys_index_dim2or3, comp_coo_from_phys_index_dim1
    public  ::  nearest_origin, maximum_complex_address_dim1, index_of_central_pixel_from_dim
    public  ::  disc_inverse

    ! These public statements will need to be removed eventually
    public  ::  get_comp_from_image, min_max_logical_address, eval_window_bounds, window_from_image_real, get_real_from_window
    public  ::  window_to_image_real, window_to_image_comp
    public  ::  thon_rings
    !public  ::  bwd_ft, fwd_ft

    !

    type, public :: image
        integer                                 ::  logical_dimensions(3)               =   (/1,1,1/)   !< Logical (x,y,z) dimensions of the image. Note that this does not necessarily correspond to memory allocation (i.e. physical) dimensions.
        logical                                 ::  is_in_real_space                    =   .true.      !< The image is in real space, not Fourier space.
        logical                                 ::  origin_at_center                    =   .true.      !< Whether the real-space origin of the image is at the center of the image. note that in fourier space,
                                                                                                        !! the origin is always at the first corner pixel (address 1,1,1). when an image is complex,
                                                                                                        !! the origin_at_center flag still refers to the real-space origin of the image (if it were back-ft'ed).
        integer                                 ::  physical_address_of_box_center(3)   =   (/0,0,0/)   !<  In each dimension, the address of the pixel at the origin
        type(c_ptr)                             ::  p                                   =   c_null_ptr  !< C pointer to the image data array
        real(kind=c_float), pointer             ::  real_values(:,:,:)                  =>  null()      !< Real array to hold values for real images.
        complex(kind=c_float_complex), pointer  ::  complex_values(:,:,:)               =>  null()      !< Complex array to hold values for comp images.
        logical                                 ::  in_memory                           =   .false.     !< Whether image values are in-memory, in other words whether the image has memory space allocated to its data array. default = .false.

        ! fftw-specific flags
        type(c_ptr)                             ::  plan_fwd                            =   c_null_ptr  !<  FFTW plan for the image (fwd)
        type(c_ptr)                             ::  plan_bwd                            =   c_null_ptr  !<  FFTW plan for the image (bwd)
        logical                                 ::  planned                             =   .false.     !<  Whether the plan has been setup by/for fftw
        contains
            !
            procedure           ::  IsAVolume => is_threed
            ! Fast Fourier transforms
            procedure,  public  ::  ForwardFT => fwd_ft
            procedure,  public  ::  BackwardFT => bwd_ft
            ! Amplitudes & phases
            procedure           ::  ampl
            procedure,  public  ::  ComputeAmplitudes => ampl
            ! Rotational averages
            procedure           ::  rot_average
            procedure, public   ::  Retun1DRotationalAverage => rot_average
            ! Disk I/O
            procedure           ::  ReadFromImageFile
            procedure           ::  ReadFromDiskGivenFilename
            procedure           ::  WriteToImageFile
            procedure           ::  WriteToDiskGivenFilename
            generic, public     ::  ReadFromDisk => ReadFromImageFile, ReadFromDiskGivenFilename
            generic, public     ::  WriteToDisk  => WriteToImageFile,  WriteToDiskGivenFilename
            ! Query functions
            procedure           ::  is_square
            procedure, public   ::  IsSquare => is_square
            procedure           ::  is_real
            procedure, public   ::  IsReal => is_real
            procedure           ::  is_comp
            procedure, public   ::  IsComplex => is_comp
            procedure           ::  is_binarised
            procedure, public   ::  IsBinarised => is_binarised
            procedure           ::  image_is_blank
            procedure, public   ::  IsBlank => image_is_blank
            procedure           ::  has_even_dims
            procedure, public   ::  HasEvenDims => has_even_dims
            ! Stats
            procedure           ::  histogram_real
            procedure, public   ::  ComputeHistogram => histogram_real
            procedure           ::  sum_double
            procedure, public   ::  SumOfVoxelValuesDouble => sum_double
            procedure           ::  variance_real
            procedure, public   ::  ReturnVariance => variance_real
            ! Image comparisons
            procedure           ::  same_dim
            procedure, public   ::  HasSameDimsAs => same_dim
            procedure           ::  same_type
            procedure, public   ::  IsSameTypeAs => same_type
            ! Looping
            procedure           ::  shell
            procedure, public   ::  ComputeShellNumberFromPhysicalAddress => shell
            procedure           ::  radius
            procedure, public   ::  ComputeRadiusFromPhysicalAddress => radius
            ! Filters & masks
            procedure           ::  band_pass
            procedure, public   ::  ApplyBandPassFilter => band_pass
            procedure           ::  apply_bfactor
            procedure, public   ::  ApplyBFactor => apply_bfactor
            procedure           ::  cosine_disc
            procedure, public   ::  ApplyCosineDisc => cosine_disc
            procedure           ::  rect_mask
            procedure, public   ::  ApplyRectMask => rect_mask
            procedure           ::  apply_filter
            procedure, public   ::  ApplyFilter => apply_filter
            procedure, public   ::  correct3d
            ! Resampling, resizing
            procedure           ::  bin
            procedure, public   ::  BinTo => bin
            ! Density value manipulations
            procedure, public   ::  NormaliseVariance
            ! Noise
            procedure           ::  image_add_gauss_noise
            procedure, public   ::  AddGaussianNoise => image_add_gauss_noise
            ! Memory management
            procedure           ::  image_allocated
            procedure, public   ::  IsAllocated => image_allocated
            procedure           ::  Allocate => image_allocate
            ! Destruction
            final   ::  image_deallocate
    end type image

    !>  \brief  a window is a special kind of image object which windows a part of larger image.
    !!          it is not suitable for operations such as fwd_ft, because it doesn't have the dim/2+1 trick built in.
    type, public, extends(image)    ::  window
        integer                 ::  width       =   2   !<  width of the window (pixels) - window is always square/cube. for linear, width is 2.
        real                    ::  centered_on(3)      !<  coordinates within the image around which the window is centered
        real                    ::  center(3)           !<  coordinate of the center relative to the window (with origin 1,1,1 at top left corner of window)
        integer                 ::  lbound(3)           !<  address of top-left voxel in image
        integer                 ::  ubound(3)           !<  address of bottom-right voxel in image
        logical                 ::  own_memory          !<  whether the window should have its own area in memory. if not, it will just point to another area of memory.
        real,       pointer     ::  rpointer(:,:,:)     !<  real pointer used when the window does not have its own area in memory.
        complex,    pointer     ::  cpointer(:,:,:)     !<  complex pointer used when the window does not have its own area in memory.
        contains
            !
            procedure           ::  eval_window_bounds
            procedure, public   ::  UpdateWindowBounds => eval_window_bounds
            ! Memory management
            procedure           ::  allocate_window
            procedure, public   ::  AllocateWindow => allocate_window
            ! Destruction
            final               ::  deallocate_window
    end type window

    !>  \brief  this defines what a filter function should look like.
    abstract interface
        pure real function filter_function(arg,parameter)
            real,   intent(in)  ::  arg
            real,   intent(in)  ::  parameter
        end function filter_function
    end interface

    interface is_allocated
        module procedure image_allocated
    end interface

    interface assignment(=)
        module procedure image_assignment
        module procedure real_to_image_assignment
        module procedure complex_to_image_assignment
    end interface

    ! eventually, i intend to phase out the origin_addr alias
    !interface origin_addr
    !    module procedure address_of_origin
    !end interface

    interface nearest_orig
        module procedure index_nearest_origin
    end interface

    interface min_max_logical_address
        module procedure bounds_from_center
    end interface

    interface nearest_origin
        module procedure index_nearest_origin
    end interface nearest_origin

    interface put_dens_into_window
        module procedure put_real_into_window
        module procedure put_comp_into_window
    end interface

    interface get_dens_from_window
        module procedure get_real_from_window
        module procedure get_comp_from_window
    end interface

    contains

    subroutine image_base_unit_tests()
        implicit none
        !
        call image_base_unit_test_1()
        call image_base_unit_test_2()
        call image_base_unit_test_3()
    end subroutine image_base_unit_tests

    !>  \brief  check image allocation and finalisation
    subroutine image_base_unit_test_1()
        implicit none
        ! private variables
        type(image) ::  self,self_copy
        ! start work
        write(*,'(a)') '**info(image_base_unit_test_1): entering'
        self%logical_dimensions = [32,32,32]
        call image_allocate(self)
        self_copy = self
        write(*,'(a)') '**info(image_base_unit_test_1): exiting'
    end subroutine image_base_unit_test_1

    !>  \brief  check image allocate, deallocation and assignements
    subroutine image_base_unit_test_2()
        implicit none
        ! private variables
        type(image) ::  self, self_copy
        !
        write(*,'(a)') '**info(image_base_unit_test_2): checking allocation of even-dimension 3d image'
        self%logical_dimensions = [32,32,32]
        call image_allocate(self)
        call image_deallocate(self)
        !
        write(*,'(a)') '**info(image_base_unit_test_2): checking allocation of odd-dimension 3d image'
        self%logical_dimensions = [55,55,55]
        call image_allocate(self)
        call image_deallocate(self)
        !
        write(*,'(a)') '**info(image_base_unit_test_2): checking allocation of arbirtrary dimension 3d image'
        self%logical_dimensions = [32,55,120]
        call image_allocate(self)
        call image_deallocate(self)
        !
        write(*,'(a)') '**info(image_base_unit_test_2): checking image assignments (1)'
        call image_allocate(self)
        self_copy = self
        call image_deallocate(self)
        call random_number(self_copy%real_values) ! should still be able to use self_copy
        call image_deallocate(self_copy)
        !
        write(*,'(a)') '**info(image_base_unit_test_2): checking image assignments (2)'
        call image_allocate(self)
        self_copy = self
        call image_deallocate(self_copy)
        call random_number(self%real_values) ! should still be able to use self
        call image_deallocate(self)
        !
        write(*,'(a)') '**info(image_base_unit_test_2): checking image assignments (3)'
        self%logical_dimensions = [32,32,1]
        call image_allocate(self)
        self_copy = self
        call image_deallocate(self_copy)
        call random_number(self%real_values) ! should still be able to use self
        call image_deallocate(self)
    end subroutine image_base_unit_test_2

    !>  \brief  check that an image has fsc of 1.0 with itself
    subroutine image_base_unit_test_3()
        use runtime_parameters, only : terminate
        implicit none
        ! private variables
        type(image) ::  self, self_copy
        !real, allocatable :: fsc_values(:)
        ! start work
        write(*,'(a)') '**info(image_base_unit_test_3): entering'
        self%logical_dimensions = [13,13,13]
        call image_allocate(self)
        call random_number(self%real_values)
        self_copy = self
        !
        !call compute_fsc(self,self_copy,fsc_values)
        !allocate(fsc_values(7))
        !fsc_values = 1.0
        !
        !if (any(fsc_values .lt. 0.999)) then
        !    write(*,'(a)') '**error(fsc_unit_test_1): not all fsc values are 1.0'
        !    write(*,*) fsc_values
        !    call terminate('failed unit test')
        !else
        !    write(*,'(a)') '**info(fsc_unit_test_1): passed unit test'
        !endif
        write(*,'(a)') '**info(image_base_unit_test_3): exiting'
    end subroutine image_base_unit_test_3

    !>  \brief  test functionalistyt of the image_fft module
    subroutine image_fft_unit_tests()
        implicit none
        call image_fft_unit_test_1()
        call image_fft_unit_test_2()
    end subroutine image_fft_unit_tests

    !>  \brief  check that we can go back and forth between real and complex image and that values are kept
    subroutine image_fft_unit_test_1()
        use runtime_parameters, only : terminate
        implicit none
        ! private variables
        type(image) ::  self, self_bak
        real, allocatable :: diff(:,:,:)
        ! start work
        write(*,'(a)') '**info(image_fft_unit_test_1): entering'
        self%logical_dimensions = [ 32,32,1 ]
        call image_allocate(self)
        call image_allocate(self_bak,mould=self)
        ! use fortran's intrinsic random number generator
        call random_number(self%real_values)
        ! back up initial state
        self_bak = self
        call self%WriteToDisk('image_fft_unit_1_in.mrc')
        ! go to fourier space and back
        call fwd_ft(self)
        call bwd_ft(self)
        !
        call self%WriteToDisk('image_fft_unit_1_out.mrc')
        ! check we still have the same thing on return
        allocate(diff(self%logical_dimensions(1),self%logical_dimensions(2),self%logical_dimensions(3)))
        diff =  abs(self    %real_values(1:self%logical_dimensions(1),1:self%logical_dimensions(2),1:self%logical_dimensions(3)) &
                -   self_bak%real_values(1:self%logical_dimensions(1),1:self%logical_dimensions(2),1:self%logical_dimensions(3)) )
        if (any(diff .gt. 0.001)) then
            write(*,'(a)') '**error(image_fft_unit_test_1): test failed'
            call terminate('fatal error in image_fft_unit_test_1')
        else
            write(*,'(a)') '**info(image_fft_unit_test_1): test passed'
        endif
        !
        write(*,'(a)') '**info(image_fft_unit_test_1): exiting'
    end subroutine image_fft_unit_test_1

    !> \brief   check we can do an ft on an image that was allocated by assignment
    subroutine image_fft_unit_test_2()
        implicit none
        ! private variables
        type(image) ::  self,self_copy
        ! start work
        write(*,'(a)') '**info(image_fft_unit_test_2): entering'
        self%logical_dimensions = [32,32,1]
        call image_allocate(self)
        self_copy = self
        call image_deallocate(self)
        !
        write(*,'(a)') '**info(image_fft_unit_test_2): checking ft works on image allocated via assignment after source deallocated'
        call fwd_ft(self_copy)
        call bwd_ft(self_copy)
        !
        write(*,'(a)') '**info(image_fft_unit_test_2): exiting'
    end subroutine image_fft_unit_test_2


    !
    ! --- CONVENIENCE FUNCTIONS, TO BE DEPRECATED ---
    !

    subroutine write_to_disk(self,filename)
        type(Image),                    intent(in)      ::  self
        character(len=*),               intent(in)      ::  filename
        call self%WriteToDisk(filename)
    end subroutine write_to_disk

    subroutine read_from_disk(self,filename)
        type(Image),                    intent(inout)   ::  self
        character(len=*),               intent(in)      ::  filename
        call self%ReadFromDisk(filename)
    end subroutine read_from_disk

    subroutine fwd_ft_tbp(self)
        class(Image),   intent(inout)   ::  self
        call fwd_ft(self)
    end subroutine
    subroutine bwd_ft_tbp(self)
        class(Image),   intent(inout)   ::  self
        call bwd_ft(self)
    end subroutine

    subroutine image_allocate_tbp(self,mould,dims)
        class(Image),               intent(inout)   ::  self
        type(image),    optional,   intent(in)      ::  mould
        integer,        optional,   intent(in)      ::  dims(3)
        call image_allocate(self,mould,dims)
    end subroutine



    !
    ! --- END OF CONVENIENCE FUNCTIONS ---
    !

    !>  \brief  Write an image to disk, given a filename
    subroutine WriteToDiskGivenFilename(self,filename,wanted_location,pixel_size,print_file_info,delete_if_already_exists)
        use ImageFiles
        ! Arguments
        class(Image),               intent(in)      ::  self
        character(len=*),           intent(in)      ::  filename                        !<  Name of file to write image to
        integer,        optional,   intent(in)      ::  wanted_location                 !<  Which location (section) in the file should we write to?
        real,           optional,   intent(in)      ::  pixel_size                      !<  Pixel size (in Angstroms) to be written to the file header
        logical,        optional,   intent(in)      ::  print_file_info                 !<  Print out information about the file
        logical,        optional,   intent(in)      ::  delete_if_already_exists    !<  If the file already exists on disk, replace it
        ! Private variables
        type(ImageFile)                             ::  my_imagefile
        ! Start work

        ! Initialise an ImageFile object
        call my_imagefile%Init(filename,pixel_size=pixel_size,delete_if_already_exists=delete_if_already_exists)

        ! Write the data
        call self%WriteToImageFile(my_imagefile,wanted_location)

        ! Print out file information
        if (present(print_file_info)) then
            if (print_file_info) then
                call my_imagefile%PrintInfo()
            endif
        endif

        ! The destructor should take care of closing the file etc.

    end subroutine WriteToDiskGivenFilename


    !>  \brief  Write an image to disk, given an initialised ImageFile object
    subroutine WriteToImageFile(self,my_imagefile,wanted_location)
        use ImageFiles
        ! Arguments
        class(Image),               intent(in)      ::  self
        class(ImageFile),           intent(inout)   ::  my_imagefile                    !<  An initialised ImageFile object
        integer,        optional,   intent(in)      ::  wanted_location                 !<  Which location (section) in the file should we write to?
        ! Private variables
        integer                                     ::  first_slice
        integer                                     ::  last_slice
        integer                                     ::  wwanted_location
        ! Start work

        ! Which location
        wwanted_location = 1
        if (present(wanted_location)) wwanted_location = wanted_location

        ! What slices/locations will we write to
        if (self%IsAVolume()) then
            if (wwanted_location .gt. 1) then
                call this_program%TerminateWithFatalError('Image::WriteToImageFile','Stacks of volumes not supported yet')
            endif
            first_slice = 1
            last_slice = self%logical_dimensions(3)
        else
            first_slice = wwanted_location
            last_slice = wwanted_location
        endif

        ! We always write real-space images to disk
        if (.not. self%is_in_real_space) call terminate('Need image in real space to write it to disk')

        ! Write to disk
        call my_imagefile%WriteSlicesToDisk(first_slice,last_slice,self%real_values,self%logical_dimensions(1))

    end subroutine WriteToImageFile

    !>  \brief  Read an image from disk, given a filename
    subroutine ReadFromDiskGivenFilename(self,filename,wanted_location,read_volume_not_slice)
        use ImageFiles
        ! Arguments
        class(Image),               intent(inout)   ::  self
        character(len=*),           intent(in)      ::  filename                        !<  Name of the file to read image from
        integer,        optional,   intent(in)      ::  wanted_location                 !<  Which location (section) in the file should we read into memory?
        logical,        optional,   intent(in)      ::  read_volume_not_slice           !<  Read a volume, not a slice
        ! Private variables
        type(ImageFile)                             ::  my_imagefile
        ! Start work

        ! Initialise an ImageFile object
        call my_imagefile%Init(filename)

        ! Read the data from disk
        call self%ReadFromImageFile(my_imagefile,wanted_location,read_volume_not_slice)
    end subroutine ReadFromDiskGivenFilename

    !>  \brief  Read an image from disk, given an initialised ImageFile object
    subroutine ReadFromImageFile(self,my_imagefile,wanted_location,read_volume_not_slice)
        use ImageFiles
        ! Arguments
        class(Image),               intent(inout)   ::  self
        class(ImageFile),           intent(inout)   ::  my_imagefile                    !<  An initialised ImageFile object
        integer,        optional,   intent(in)      ::  wanted_location                 !<  Which location (section) in the file should we read into memory?
        logical,        optional,   intent(in)      ::  read_volume_not_slice           !<  Read a volume, not a slice
        ! Private variables
        logical                                     ::  rread_volume_not_slice
        integer                                     ::  new_image_dimensions(3)
        integer                                     ::  first_slice
        integer                                     ::  last_slice
        integer                                     ::  wwanted_location
        ! Start work

        ! Reading 3D or 2D?
        rread_volume_not_slice = .false.
        if (present(read_volume_not_slice)) rread_volume_not_slice = read_volume_not_slice

        ! Which location?
        wwanted_location = 1
        if (present(wanted_location)) wwanted_location = wanted_location

        ! The image will have these dimensions
        new_image_dimensions = my_imagefile%GetDimensions()
        if (.not. rread_volume_not_slice) new_image_dimensions(3) = 1

        ! Allocate image
        call self%Allocate(dims=new_image_dimensions)

        ! The slices we will read
        if (rread_volume_not_slice) then
            if (wwanted_location .gt. 1) then
                call this_program%TerminateWithFatalError('Image::ReadFromImageFile','Stacks of volumes not supported yet')
            endif
            first_slice = 1
            last_slice = my_imagefile%GetStackSize()
        else
            first_slice = wwanted_location
            last_slice = wwanted_location
        endif

        ! Do the reading from disk
        call my_imagefile%ReadSlicesFromDisk(first_slice,last_slice,self%real_values)

        ! We assume that we read in real data
        self%is_in_real_space = .true.
    end subroutine ReadFromImageFile


    !>  \brief  apply a cosine edged band-pass filter to the image, and return it in the space it was given.
    !!          note: set parameters to 0.0 to get a low-pass- or high-pass-only filter.
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine band_pass(img,hp_freq,lp_freq,cosine_falloff)
        use runtime_parameters, only : pi, terminate
        implicit none
        ! arguments
        class(image),       intent(inout)   ::  img             !<  input image. can be real or complex, either way filter is applied in reciprocal space and image is returned in same space as given.
        real,               intent(in)      ::  hp_freq         !<  frequencies at this radius and below will be 0.0 in output. in reciprocal pixel units (from nyquist is 0.5).
        real,               intent(in)      ::  lp_freq         !<  frequencies at this radius and above will be 0.0 in output. in reciprocal pixel units (from nyquist is 0.5).
        real,   optional,   intent(in)      ::  cosine_falloff  !<  width of cosine falloff. in reciprocal pixel units (frequencies at hp_freq + cosine_falloff will be left intact)
        ! private variables
        logical                 ::  do_ft
        real                    ::  ccosine_falloff
        logical,    parameter   ::  debug   =   .false.
        integer                 ::  i,j,k
        real                    ::  freq, coos(3), falloff_factor
        ! start work

#ifndef SKIP_RUNTIME_CHECKS
        ! check filter parameters
        if (hp_freq .ge. 0.5) then
            write(*,'(a,f5.1,a)') '**error(band_pass): high-pass frequency is higher than nyquist (', &
                                    hp_freq, '). if you meant to do a low-pass-only filter, set high-pass parameter to 0.0'
            call terminate('fatal error in band_pass')
        endif
        if (lp_freq .lt. 0.0 .or. hp_freq .lt. 0.0) then
            write(*,'(a,2(f0.4,1x))') '**error(band_pass): negative filter parameters do not make sense: ', lp_freq, hp_freq
            call terminate('fatal error in band_pass')
        endif
        if (hp_freq .gt. lp_freq .and. hp_freq .ne. 0.0 .and. lp_freq .ne. 0.0) then
            write(*,'(a,2(f0.4,1x))') '**error(band_pass): high-pass frequency cannot be lower than low-pass frequency: ', &
                                        lp_freq, hp_freq
            call terminate('fatal error in band_pass')
        endif
#endif

        ! cosine falloff is 0.05 unless otherwise specified
        if (present(cosine_falloff)) then
            ccosine_falloff = cosine_falloff
        else
            ccosine_falloff = 0.05
        endif

        ! if necessary, ft the input image
        if (img%is_in_real_space) then
            do_ft = .true.
            call fwd_ft(img)
        else
            do_ft = .false.
        endif

        ! apply the filter, loop over image
        do k=1,img%logical_dimensions(3)
            ! the distance from the origin to the current pixel, in reciprocal pixel units
            coos(3) = comp_coo_from_phys_index_dim2or3(img%logical_dimensions(3),k)**2
            do j=1,img%logical_dimensions(2)
                coos(2) = comp_coo_from_phys_index_dim2or3(img%logical_dimensions(2),j)**2
                do i=1,highest_index_dim1(img%logical_dimensions(1))
                    coos(1) = comp_coo_from_phys_index_dim1(img%logical_dimensions(1),i)**2
                    ! compute radius, in units of reciprocal pixels
                    freq = sqrt(sum(coos))
                    ! apply the high-pass
                    if (hp_freq .ne. 0.0) then
                        if (freq .le. hp_freq) then
                            img%complex_values(i,j,k) = (0.0,0.0)
                        else if (freq .le. hp_freq + ccosine_falloff) then
                            falloff_factor = ((freq - hp_freq)/ccosine_falloff)*pi  ! from 0 to pi
                            falloff_factor = cos(falloff_factor) ! from 1 to -1
                            falloff_factor = -falloff_factor ! from -1 to 1
                            falloff_factor = (falloff_factor + 1.0)/2.  ! from 0 to 1
                            img%complex_values(i,j,k) = img%complex_values(i,j,k) * falloff_factor
                        endif
                    endif
                    ! apply the low-pass
                    if (lp_freq .ne. 0.0) then
                        if (freq .ge. lp_freq) then
                            img%complex_values(i,j,k) = 0.0
                        else if (freq .ge. lp_freq - ccosine_falloff) then
                            falloff_factor = ((freq - (lp_freq - ccosine_falloff))/ccosine_falloff)*pi ! from 0 to pi
                            falloff_factor = cos(falloff_factor) ! from 1 to -1
                            falloff_factor = (falloff_factor + 1.0)/2. ! from 1 to 0
                            img%complex_values(i,j,k) = img%complex_values(i,j,k) * falloff_factor
                        endif
                    endif
                enddo
            enddo
        enddo

        ! if the image was given as a real image, return it as a real image
        if (do_ft) then
            call bwd_ft(img)
        endif
    end subroutine band_pass

    !>  \brief  apply a filter (ie. any real image with origin at corner (1,1,1) and values from 0.0 to 1.0) to an image
    subroutine apply_filter_img(img,filter_img)
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  img         !<  image to be filtered (in-place)
        type(image),    intent(inout)   ::  filter_img  !<  filter to be applied
        ! private variables
        integer ::  i,j,k
        ! check dimensions
        if (.not. same_dim(img,filter_img)) then
            write(*,'(a,3i4,a,3i4)') '**error(apply_filter): image and filter do not have same dimension: ', &
                                     img%logical_dimensions, '&', filter_img%logical_dimensions
            call terminate('fatal error in apply_filter')
        endif
        ! check types
        if (img%is_in_real_space) then
            write(*,'(2a)') '**warning(apply_filter): attempting to apply filter to an image of type', 'real'
        endif
        if (.not. filter_img%is_in_real_space) then
            write(*,'(a)') '**error(apply_filter): filter must be of type real, it cannot be of type ', 'comp'
            call terminate('fatal error in apply_filter')
        endif
        ! check origins are in the expected place
        if (.not. img%origin_at_center) then
            write(*,'(3a)') '**error(apply_filter): image should have real-space origin at center pixel'
            call terminate('fatal error in apply_filter')
        endif
        if (filter_img%origin_at_center) then
            write(*,'(2a)') '**debug(apply_filter): filter image does not have origin at (1,1,1) corner - ', &
                            'will apply appropriate shift'
            call shift_real_space_origin(filter_img)
        endif


        ! now do the actual multiplication
        if (.not. img%is_in_real_space) then
            do k=1,img%logical_dimensions(3)
                do j=1,img%logical_dimensions(2)
                    do i=1,highest_index_dim1(img%logical_dimensions(1))
                        img%complex_values(i,j,k) = img%complex_values(i,j,k) * filter_img%real_values(i,j,k)
                    enddo
                enddo
            enddo
        else
            do k=1,img%logical_dimensions(3)
                do j=1,img%logical_dimensions(2)
                    do i=1,img%logical_dimensions(1)
                        img%real_values(i,j,k) = img%real_values(i,j,k) * filter_img%real_values(i,j,k)
                    enddo
                enddo
            enddo
        endif

    end subroutine apply_filter_img

    !>  \brief  mask real image with a cosine-falloff-edged rectangle. rectangle is filled with 1s near the center, and 0s near the edge.
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine cosine_rect(img,rect_dim,azimuth,falloff,wipe_input,mask_with_average,shift_x,shift_y, &
                            mask_with_average_edge_value)
        use runtime_parameters, only : pi
#ifndef SKIP_RUNTIME_CHECKS
        use runtime_parameters, only : terminate
#endif
        use rotation_matrices, only : radians
        implicit none
        ! arguments
        type(image),            intent(inout)   ::  img                 !<  image to mask
        real,                   intent(in)      ::  rect_dim(2)         !<  dimensions of the rectangle, in pixels
        real,                   intent(in)      ::  azimuth             !<  in degrees, the clockwise angle that specifies the first axis of the rectangle. 0.0 indicates the x axis.
        real,                   intent(in)      ::  falloff             !<  distance from the start of the falloff (1.0) to the end of the falloff(0.0). the mask reaches 0.0 at rect_dim/2. in pixels.
        logical,    optional,   intent(in)      ::  wipe_input          !<  wipe / initialise the input to 1.0. .true. by default
        logical,    optional,   intent(in)      ::  mask_with_average   !<  if .true., the mask will go to the average value within the mask, rather than 0.0.
        real,       optional,   intent(in)      ::  shift_x,shift_y     !<  the center of the rectangle should be shifted by this much in x and y
        logical,    optional,   intent(in)      ::  mask_with_average_edge_value    !<  Mask with average grey value of edge pixels
        ! private variables
        integer         ::  smallest_dim
        logical         ::  wwipe_input, mmask_with_average, mmask_with_average_edge_value
        real            ::  mask_value
        logical,    parameter   ::  debug   =   .false.
        ! start work

        ! only coded for 2d images
#ifndef SKIP_RUNTIME_CHECKS
        if (img%logical_dimensions(3) .ne. 1) then
            write(*,'(a,i0)') '**error(cosine_disc): input image is not 2d, it has 3rd dimension of ', img%logical_dimensions(3)
            call terminate('fatal error in cosine_rect')
        endif
#endif

        ! wipe input?
        if (present(wipe_input)) then
            wwipe_input = wipe_input
        else
            wwipe_input = .true.
        endif

        ! mask with average?
        mmask_with_average = .false.
        if (present(mask_with_average)) mmask_with_average = mask_with_average
        mmask_with_average_edge_value = .false.
        if (present(mask_with_average_edge_value)) mmask_with_average_edge_value = mask_with_average_edge_value

#ifndef SKIP_RUNTIME_CHECKS
        if (mmask_with_average .and. mmask_with_average_edge_value) then
            write(*,'(a)') '**ERROR(cosine_rect): programming error. Cannot mask with average within mask AND with edge average'
            call terminate('Fatal error in cosine_rect')
        endif

        ! mask with average is not compatible with wipe input
        if (mmask_with_average .and. wwipe_input) then
            write(*,'(a)') '**error(cosine_rect): options mask_with_average and wipe_input are not compatible'
            call terminate('fatal error in cosine_rect - programming error')
        endif

        ! check type
        if (.not. img%is_in_real_space) then
            write(*,'(2a)') '**error(cosine_rect): image type not supported ', 'real'
            call terminate('fatal error in cosine_rect')
        endif

        ! check the rectangle dimensions make sense
        if (img%logical_dimensions(3) .eq. 1) then
            smallest_dim = minval(img%logical_dimensions(1:2))
        else
            smallest_dim = minval(img%logical_dimensions(1:3))
        endif
        if (any(rect_dim .le. 0.0)) then
            write(*,'(a,2(f0.3))') '**error(cosine_rect): negative dimension(s): ', rect_dim
            call terminate('fatal error in cosine_rect')
        else if (maxval(rect_dim) .gt. maxval(img%logical_dimensions(1:2))) then
            write(*,'(a,2(f0.3,1x),3(i0,1x))') '**error(cosine_rect): dimensions too large: ', rect_dim, img%logical_dimensions
            call terminate('fatal error in cosine_rect')
        else if (maxval(rect_dim) .gt. smallest_dim) then
            write(*,'(2a,2(f0.3,1x),3(i0,1x))') '**warning(cosine_rect): dimensions may be too large, ',   &
                                                'depending on the rectangle azimuth: ', rect_dim, img%logical_dimensions
        endif

        ! debug
        if (debug) then
            write(*,'(a,3(f0.3,1x))') '**debug(cosine_rect): rectangle dimensions, falloff: ', rect_dim, falloff
        endif
#endif

        ! if required, wipe the input
        if (wwipe_input) then
            img%real_values = 1.0
        endif

        ! Work out the mask value
        if (mmask_with_average) then
            ! loop over image to compute the average value within the mask
            call cosine_rect_slave( img,rect_dim,azimuth,falloff,get_stats=.true.,mask_value=mask_value, &
                                    shift_x=shift_x,shift_y=shift_y)
        else if (mmask_with_average_edge_value) then
            mask_value = edge_average(img)
        else
            mask_value = 0.0
        endif

        ! loop over pixels to apply the mask
        call cosine_rect_slave(img,rect_dim,azimuth,falloff,get_stats=.false.,mask_value=mask_value,shift_x=shift_x,shift_y=shift_y)

    end subroutine cosine_rect

    !>  \brief  convenience routine containing the looping structure to define a smooth-edged rectangle. can be called
    !!          either to compile statistics to determine average value within mask, or to apply mask
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine cosine_rect_slave(img,rect_dim,azimuth,falloff,get_stats,mask_value,shift_x,shift_y)
        use rotation_matrices, only : radians
        use runtime_parameters, only : pi
#ifndef SKIP_RUNTIME_CHECKS
        use runtime_parameters, only : terminate
#endif
        implicit none
        ! arguments
        type(image),            intent(inout)   ::  img                 !<  image to mask
        real,                   intent(in)      ::  rect_dim(2)         !<  dimensions of the rectangle, in pixels
        real,                   intent(in)      ::  azimuth             !<  in degrees, the clockwise angle that specifies the first axis of the rectangle. 0.0 indicates the x axis.
        real,                   intent(in)      ::  falloff             !<  distance from the start of the falloff (1.0) to the end of the falloff(0.0). the mask reaches 0.0 at rect_dim/2. in pixels.
        logical,                intent(in)      ::  get_stats           !<  if true, we'll compute average value within mask rather than apply a mask
        real,                   intent(inout)   ::  mask_value          !<  greyvalue for the mask. if get_stats, this will be an output.
        real,       optional,   intent(in)      ::  shift_x             !<  center of the rectangle should be shifted by this much in x
        real,       optional,   intent(in)      ::  shift_y             !<  center of the rectangle should be shifted by this much in y
        ! private variables
        real    ::  origin(3)
        real    ::  cos_az, sin_az
        real    ::  rect_min(2), rect_max(2), edge_min(2), edge_max(2)
        real(kind=8)    ::  sum_in_mask
        real(kind=8)    ::  count_in_mask
        integer         ::  i,j
        real            ::  x,y,xx,yy,xx_tmp,yy_tmp
        real            ::  rad
        real            ::  falloff_factor
        ! start work

        ! values we'll need when looping
        origin      = [1.0,1.0,1.0]
        if (img%origin_at_center .and. img%is_in_real_space) origin = real(img%physical_address_of_box_center)
        if (present(shift_x)) origin(1) = origin(1) + shift_x
        if (present(shift_y)) origin(2) = origin(2) + shift_y
        cos_az      =  cos(radians(azimuth))
        sin_az      =  sin(radians(azimuth))
        rect_min    = -rect_dim/2.
        rect_max    =  rect_dim/2.
        edge_min    =  rect_min + falloff
        edge_max    =  rect_max - falloff

        ! loop over pixels and work out whether we are within the rectangle
        sum_in_mask = 0.0d0
        count_in_mask = 0
        do j=1,img%logical_dimensions(2)
            y = real(j - origin(2))
            xx_tmp = - sin_az * y
            yy_tmp = + cos_az * y
            do i=1,img%logical_dimensions(1)
                x = real(i - origin(1))
                ! work out coordinates before rotation
                xx = cos_az * x + xx_tmp
                yy = sin_az * x + yy_tmp
                ! is this point within the rectangle?
                if (        xx .ge. rect_min(1) .and. xx .le. rect_max(1)   &
                    .and.   yy .ge. rect_min(2) .and. yy .le. rect_max(2))  then
                    ! we are within the rectangle
                    ! now, check whether we are within the cosine-edge region of the rectangle
                    if (        xx .le. edge_min(1) .or. xx .ge. edge_max(1)    &
                        .or.    yy .le. edge_min(2) .or. yy .ge. edge_max(2)) then
                        ! we are within the falloff edge of the rectangle
                        ! we need to work out the falloff factor
                        ! we follow different logic depending on whether we are near the corners of the rectangle
                        if (xx .le. edge_min(1)) then
                            if (yy .le. edge_min(2)) then
                                !  we are in a corner. work out radius to corner
                                rad = sqrt((xx-edge_min(1))**2+(yy-edge_min(2))**2)
                            else if (yy .ge. edge_max(2)) then
                                ! we are in a corner
                                rad = sqrt((xx-edge_min(1))**2+(yy-edge_max(2))**2)
                            else
                                ! we are along an edge
                                rad = edge_min(1) - xx
                            endif
                        else if (xx .ge. edge_max(1)) then
                            if (yy .le. edge_min(2)) then
                                !  we are in a corner. work out radius to corner
                                rad = sqrt((xx-edge_max(1))**2+(yy-edge_min(2))**2)
                            else if (yy .ge. edge_max(2)) then
                                !  we are in a corner. work out radius to corner
                                rad = sqrt((xx-edge_max(1))**2+(yy-edge_max(2))**2)
                            else
                                ! we are along an edge
                                rad = xx-edge_max(1)
                            endif
                        else if (yy .le. edge_min(2)) then
                            ! we are along an edge
                            rad = edge_min(2)-yy
                        else if (yy .ge. edge_max(2)) then
                            ! we are along an edge
                            rad = yy-edge_max(2)
                        else
#ifndef SKIP_RUNTIME_CHECKS
                            ! we must have got lost!
                            write(*,'(a)') '**error(cosine_rect_slave): programming error.'
                            call terminate('fatal error in cosine_rect_slave')
#endif
                        endif
                        ! now we've wored out how far we are from the start of the falloff, we can work out the falloff factor
                        if (rad .le. falloff) then
                            falloff_factor = (cos((rad/falloff)*pi)+1.)/2.0
                        else
                            falloff_factor = 0.0d0
                        endif
                    else
                        ! we are within the rectangle, not on its edge
                        falloff_factor = 1.0d0
                    endif ! end of test for whether we're in the rectangle and within its border region

                    ! now we've worked out the falloff factor for where we are, let's act on it
                    if (get_stats) then
                        ! update stats
                        sum_in_mask = sum_in_mask + dble(img%real_values(i,j,1) * falloff_factor)
                        count_in_mask = count_in_mask + dble(falloff_factor)
                    else
                        ! apply mask
                        img%real_values(i,j,1) = img%real_values(i,j,1) * falloff_factor + (1.-falloff_factor)*mask_value
                    endif
                else
                    ! the point is not within the rectangle at all
                    if (get_stats) then
                        ! nothing to do
                    else
                        img%real_values(i,j,1) = mask_value
                    endif
                endif
            enddo   ! end of loop over first dimension
        enddo ! end of loop over second dimension

        if (get_stats) then
            mask_value = real(sum_in_mask / count_in_mask)
        endif

    end subroutine cosine_rect_slave


    !>  \brief  mask real image with a rectangle.
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine rect_mask(img,rect_dim,azimuth,wipe_input,mask_with_average,shift_x,shift_y)
        use runtime_parameters, only : pi, terminate
        use rotation_matrices, only : radians
        implicit none
        ! arguments
        class(image),           intent(inout)   ::  img                 !<  image to mask
        real,                   intent(in)      ::  rect_dim(2)         !<  dimensions of the rectangle, in pixels. length (along azimuth) first.
        real,                   intent(in)      ::  azimuth             !<  in degrees, the clockwise angle that specifies the first axis of the rectangle. 0.0 indicates the x axis.
        logical,    optional,   intent(in)      ::  wipe_input          !<  wipe / initialise the input to 1.0. .true. by default
        logical,    optional,   intent(in)      ::  mask_with_average   !<  if .true., the mask will go to the average value within the mask, rather than 0.0.
        real,       optional,   intent(in)      ::  shift_x,shift_y     !<  the center of the rectangle should be shifted by this much in x and y
        ! private variables
        integer         ::  smallest_dim
        logical         ::  wwipe_input, mmask_with_average
        real            ::  mask_value
        logical,    parameter   ::  debug   =   .false.
        ! start work

#ifndef SKIP_RUNTIME_CHECKS
        ! only coded for 2d images
        if (img%logical_dimensions(3) .ne. 1) then
            write(*,'(a,i0)') '**error(rect_mask): input image is not 2d, it has 3rd dimension of ', img%logical_dimensions(3)
            call terminate('fatal error in rect_mask')
        endif
#endif

        ! wipe input?
        if (present(wipe_input)) then
            wwipe_input = wipe_input
        else
            wwipe_input = .true.
        endif

        ! mask with average?
        if (present(mask_with_average)) then
            mmask_with_average = mask_with_average
        else
            mmask_with_average = .false.
        endif

#ifndef SKIP_RUNTIME_CHECKS
        ! mask with average is not compatible with wipe input
        if (mmask_with_average .and. wwipe_input) then
            write(*,'(a)') '**error(rect_mask): options mask_with_average and wipe_input are not compatible'
            call terminate('fatal error in rect_mask - programming error')
        endif

        ! check type
        if (.not. img%is_in_real_space) then
            write(*,'(2a)') '**error(rect_mask): image type not supported ', 'comp'
            call terminate('fatal error in rect_mask')
        endif

        ! check the rectangle dimensions make sense
        if (img%logical_dimensions(3) .eq. 1) then
            smallest_dim = minval(img%logical_dimensions(1:2))
        else
            smallest_dim = minval(img%logical_dimensions(1:3))
        endif
        if (any(rect_dim .le. 0.0)) then
            write(*,'(a,2(f0.3,1x))') '**error(rect_mask): negative dimension(s): ', rect_dim
            call terminate('fatal error in rect_mask')
        else if (maxval(rect_dim) .gt. maxval(img%logical_dimensions(1:2))) then
            write(*,'(a,2(f0.3,1x),3(i0,1x))') '**error(rect_mask): dimensions too large: ', rect_dim, img%logical_dimensions
            call terminate('fatal error in rect_mask')
        else if (maxval(rect_dim) .gt. smallest_dim) then
            write(*,'(2a,2(f0.3,1x),3(i0,1x))') '**warning(rect_mask): dimensions may be too large, ', &
                                                'depending on the rectangle azimuth: ', rect_dim, img%logical_dimensions
        endif

        ! debug
        if (debug) then
            write(*,'(a,2(f0.3,1x))') '**debug(rect_mask): rectangle dimensions: ', rect_dim
        endif
#endif

        ! if required, wipe the input
        if (wwipe_input) then
            img%real_values = 1.0
        endif

        if (mmask_with_average) then
            ! loop over image to compute the average value within the mask
            call rect_mask_slave(img,rect_dim,azimuth,get_stats=.true.,mask_value=mask_value,shift_x=shift_x,shift_y=shift_y)
        else
            mask_value = 0.0
        endif

        ! loop over pixels to apply the mask
        call rect_mask_slave(img,rect_dim,azimuth,get_stats=.false.,mask_value=mask_value,shift_x=shift_x,shift_y=shift_y)

    end subroutine rect_mask

    !>  \brief  zero pixels outside specified rectangle
    pure subroutine rect_mask_slave(img,rect_dim,azimuth,get_stats,mask_value,shift_x,shift_y)
        use rotation_matrices, only : radians
        use runtime_parameters, only : pi, terminate
        implicit none
        ! arguments
        type(image),            intent(inout)   ::  img                 !<  image to mask
        real,                   intent(in)      ::  rect_dim(2)         !<  dimensions of the rectangle, in pixels. first is length (along azimuth), second is width.
        real,                   intent(in)      ::  azimuth             !<  in degrees, the clockwise angle that specifies the first axis of the rectangle. 0.0 indicates the x axis.
        logical,                intent(in)      ::  get_stats           !<  if true, we'll compute average value within mask rather than apply a mask
        real,                   intent(inout)   ::  mask_value          !<  greyvalue for the mask. if get_stats, this will be an output.
        real,       optional,   intent(in)      ::  shift_x             !<  center of the rectangle should be shifted by this much in x
        real,       optional,   intent(in)      ::  shift_y             !<  center of the rectangle should be shifted by this much in y
        ! private variables
        real    ::  origin(3)
        real    ::  cos_az, sin_az
        real    ::  rect_min(2), rect_max(2)
        real(kind=8)    ::  sum_in_mask
        real(kind=8)    ::  count_in_mask
        integer         ::  i,j
        real            ::  x,y,xx,yy,xx_tmp,yy_tmp
        real            ::  falloff_factor
        ! start work

        ! values we'll need when looping
        origin      = [1.0,1.0,1.0]
        if (img%origin_at_center .and. img%is_in_real_space) origin = real(img%physical_address_of_box_center)
        if (present(shift_x)) origin(1) = origin(1) + shift_x
        if (present(shift_y)) origin(2) = origin(2) + shift_y
        cos_az      =  cos(radians(azimuth))
        sin_az      =  sin(radians(azimuth))
        rect_min    = -rect_dim/2.
        rect_max    =  rect_dim/2.

        ! loop over pixels and work out whether we are within the rectangle
        sum_in_mask = 0.0d0
        count_in_mask = 0
        do j=1,img%logical_dimensions(2)
            y = real(j - origin(2))
            xx_tmp = - sin_az * y
            yy_tmp = + cos_az * y
            do i=1,img%logical_dimensions(1)
                x = real(i - origin(1))
                ! work out coordinates before rotation
                xx = cos_az * x + xx_tmp
                yy = sin_az * x + yy_tmp
                ! is this point within the rectangle?
                if (        xx .ge. rect_min(1) .and. xx .le. rect_max(1)   &
                    .and.   yy .ge. rect_min(2) .and. yy .le. rect_max(2))  then
                    ! we are within the rectangle
                    falloff_factor = 1.0
                    if (get_stats) then
                        ! update stats
                        sum_in_mask = sum_in_mask + dble(img%real_values(i,j,1) * falloff_factor)
                        count_in_mask = count_in_mask + dble(falloff_factor)
                    endif
                else
                    falloff_factor = 0.0
                    if (get_stats) then
                        ! nothing to do
                    else
                        img%real_values(i,j,1) = mask_value
                    endif
                endif
            enddo   ! end of loop over first dimension
        enddo ! end of loop over second dimension

        if (get_stats) then
            mask_value = real(sum_in_mask / count_in_mask)
        endif

    end subroutine rect_mask_slave

    !>  \brief  fill real image with a cosine-falloff-edged disc. disc is filled with 1s near the center, and 0s near the edge
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine cosine_disc(img,rad,falloff,wipe_input,mask_with_average,cylinder,inverse)
        use runtime_parameters, only : pi, terminate
        implicit none
        ! arguments
        class(image),           intent(inout)   ::  img                 !<  image to fill
        real,                   intent(in)      ::  rad                 !<  radius at which the falloff reaches 0.0 (in pixels, or fraction of max radius)
        real,                   intent(in)      ::  falloff             !<  distance from the start of the falloff (1.0) to the end of the falloff(0.0). the mask reaches 0.0 at rad. in pixels or fraction.
        logical,    optional,   intent(in)      ::  wipe_input          !<  whether to wipe / initialise the input to 1.0. .true. by default
        logical,    optional,   intent(in)      ::  mask_with_average   !<  if .true., the mask will go to the average value within the mask, rather than 0.0. if the mask is inverted, this refers to the average value in the area that will be masked out.
        logical,    optional,   intent(in)      ::  cylinder            !<  cosine-edge cylinder rather than sphere
        logical,    optional,   intent(in)      ::  inverse             !<  invert the filter.
        ! private variables
        integer         ::  i,j,k
        integer         ::  smallest_dim
        real            ::  dist, radm, rrad
        logical         ::  wwipe_input, mmask_with_average
        logical         ::  fraction
        real(kind=8)    ::  sum_in_mask
        real(kind=8)    ::  count_in_mask
        real            ::  mask_value
        real            ::  falloff_factor
        integer         ::  origin(3)
        real            ::  x,y,z
        integer         ::  ix,iy,iz
        integer         ::  max_rad(3)
        logical         ::  ccylinder
        logical         ::  iinverse
        ! start work

#ifndef SKIP_RUNTIME_CHECKS
        ! built-in assumption the image is square/cubic
        if (.not. is_square(img)) then
            write(*,'(a,3(i0,1x))') '**error(cosine_disc): input image is not square / cubic: ', img%logical_dimensions
            call terminate('fatal error in cosine_disc')
        endif
#endif

        ! wipe input?
        if (present(wipe_input)) then
            wwipe_input = wipe_input
        else
            wwipe_input = .true.
        endif

        ! mask with average?
        if (present(mask_with_average)) then
            mmask_with_average = mask_with_average
        else
            mmask_with_average = .false.
        endif

        ! cylinder?
        ccylinder = .false.
        if (present(cylinder)) ccylinder = cylinder

        ! invert?
        iinverse = .false.
        if (present(inverse)) iinverse = inverse

#ifndef SKIP_RUNTIME_CHECKS
        ! mask with average is not compatible with wipe input
        if (mmask_with_average .and. wwipe_input) then
            write(*,'(a)') '**error(cosine_disc): options mask_with_average and wipe_input are not compatible'
            call terminate('fatal error in cosine_disc - programming error')
        endif

        ! check type
        if (.not. img%is_in_real_space) then
            write(*,'(2a)') '**error(cosine_disc): image type not supported ', 'comp'
            call terminate('fatal error in cosine_disc')
        endif

        ! check the radius makes sense
        if (img%logical_dimensions(3) .eq. 1) then
            smallest_dim = minval(img%logical_dimensions(1:2))
        else
            smallest_dim = minval(img%logical_dimensions(1:3))
        endif

        if (rad .le. 0.0) then
            write(*,'(a,f12.3)') '**error(cosine_disc): radius is too small: ', rad
            call terminate('fatal error in cosine_disc')
        else if (rad .gt. smallest_dim) then
            write(*,'(a,f12.3,3i4)') '**error(cosine_disc): radius is too large: ', rad, img%logical_dimensions
            call terminate('fatal error in cosine_disc')
        endif
#endif

        ! are we dealing with a fraction? if so and the image isn't square/cubic, we'll get an elipse rather than disc
        if (rad .gt. 0.0 .and. rad .le. 1.0) then
#ifndef SKIP_RUNTIME_CHECKS
            if (falloff .gt. 1.0) then
                write(*,'(a,f0.1,a,f8.2,a)') '**error(cosine_disc): radius was specified as a fraction (', rad,     &
                                            '), but the falloff was not (', falloff, ')'
                call terminate('fatal error in cosine_disc')
            endif
#endif
            fraction = .true.
        else
            fraction = .false.
        endif

        rrad = rad
        radm = rrad-falloff

        ! if required, wipe the input
        if (wwipe_input) then
            img%real_values = 1.0
        endif

        ! origin
        origin      = [1.0,1.0,1.0]
        if (img%origin_at_center .and. img%is_in_real_space) origin = real(img%physical_address_of_box_center)

        ! maximum radius (squared)
        if (fraction) then
            if (img%origin_at_center .and. img%is_in_real_space) then
                max_rad = (origin-1)**2
            else
                max_rad = real(int(img%logical_dimensions/2))**2
            endif
        endif

        if (mmask_with_average) then
            ! loop over pixels to determine average value within mask
            sum_in_mask = 0.0d0
            count_in_mask = 0
            if (fraction) then
                do k=1,img%logical_dimensions(3)
                    z = real(k - origin(3))**2 / max_rad(3)
                    do j=1,img%logical_dimensions(2)
                        y = real(j - origin(2))**2 / max_rad(2)
                        do i=1,img%logical_dimensions(1)
                            x = real(i - origin(1))**2 / max_rad(1)
                            ! work out distance from origin (fraction of max radius)
                            if (ccylinder) then
                                dist = sqrt(x+y)
                            else
                                dist = sqrt(x+y+z)
                            endif
                            if (dist .ge. radm .and. dist .le. rrad) then
                                falloff_factor = (cos(((dist-radm)/falloff)*pi)+1.)/2.0
                                sum_in_mask = sum_in_mask + dble(img%real_values(i,j,k) * falloff_factor)
                                count_in_mask = count_in_mask + dble(falloff_factor)
                            else if (dist .lt. radm) then
                                sum_in_mask = sum_in_mask + dble(img%real_values(i,j,k))
                                count_in_mask = count_in_mask + 1.0d0
                            endif
                        enddo
                    enddo
                enddo
            else
                do k=1,img%logical_dimensions(3)
                    iz = (k - origin(3))**2
                    do j=1,img%logical_dimensions(2)
                        iy = (j - origin(2))**2
                        do i=1,img%logical_dimensions(1)
                            ix = (i - origin(1))**2
                            ! work out distance from origin
                            if (ccylinder) then
                                dist = sqrt(real(ix+iy))
                            else
                                dist = sqrt(real(ix+iy+iz))
                            endif
                            ! accumulate statistics
                            if (dist .ge. radm .and. dist .le. rrad) then
                                falloff_factor = (cos(((dist-radm)/falloff)*pi)+1.)/2.0
                                sum_in_mask = sum_in_mask + dble(img%real_values(i,j,k) * falloff_factor)
                                count_in_mask = count_in_mask + dble(falloff_factor)
                            else if (dist .lt. radm) then
                                sum_in_mask = sum_in_mask + dble(img%real_values(i,j,k))
                                count_in_mask = count_in_mask + 1.0d0
                            endif
                        enddo
                    enddo
                enddo
            endif   ! end of test for fraction
            mask_value = real(sum_in_mask / count_in_mask)
        else
            mask_value = 0.0
        endif

        ! loop over pixels to apply the mask
        if (fraction) then
            do k=1,img%logical_dimensions(3)
                z = real(k - origin(3))**2 / max_rad(3)
                do j=1,img%logical_dimensions(2)
                    y = real(j - origin(2))**2 / max_rad(2)
                    do i=1,img%logical_dimensions(1)
                        x = real(i - origin(1))**2 / max_rad(1)
                        ! work out distance from origin (fraction of max radius)
                        if (ccylinder) then
                            dist = sqrt(x+y)
                        else
                            dist = sqrt(x+y+z)
                        endif
                        if (dist .ge. radm .and. dist .le. rrad) then
                            falloff_factor = (cos(((dist-radm)/falloff)*pi)+1.)/2.0
                            if (iinverse) falloff_factor = (falloff_factor-1.0)*(-1.0)
                            img%real_values(i,j,k) = img%real_values(i,j,k) * falloff_factor + (1.-falloff_factor)*mask_value
                        else if (dist .gt. rrad .and. .not. iinverse) then
                            img%real_values(i,j,k) = mask_value
                        else if (dist .le. rrad .and. iinverse) then
                            img%real_values(i,j,k) = mask_value
                        endif
                    enddo
                enddo
            enddo
        else
            do k=1,img%logical_dimensions(3)
                z = real(k - origin(3))**2
                do j=1,img%logical_dimensions(2)
                    y = real(j - origin(2))**2
                    do i=1,img%logical_dimensions(1)
                        x = real(i - origin(1))**2
                        ! work out distance from origin
                        if (ccylinder) then
                            dist = sqrt(x+y)
                        else
                            dist = sqrt(x+y+z)
                        endif
                        if (dist .ge. radm .and. dist .le. rrad) then
                            falloff_factor = (cos(((dist-radm)/falloff)*pi)+1.)/2.0
                            if (iinverse) falloff_factor = (falloff_factor-1.0)*(-1.0)
                            img%real_values(i,j,k) = img%real_values(i,j,k) * falloff_factor + (1.-falloff_factor)*mask_value
                        else if (dist .gt. rrad .and. .not. iinverse) then
                            img%real_values(i,j,k) = mask_value
                        else if (dist .le. rrad .and. iinverse) then
                            img%real_values(i,j,k) = mask_value
                        endif
                    enddo
                enddo
            enddo
        endif

    end subroutine cosine_disc

    !>  \brief  fill real image with a disc.
    subroutine hard_disc(img,rad)
        implicit none
        ! arguments
        type(image),    intent(inout)           ::  img
        real,           intent(in)              ::  rad
        ! private variables

        ! start work
        call generate_filter_image(img,disc,rad)

    end subroutine hard_disc

    !>  \brief  return the b factor that should be applied to an image before it is inserted into a 3d reconstruction, given its score after refinement, and the average score of the dataset
    real function bfactor_from_score(score,average_score) result(bfactor)
        use runtime_parameters, only : pbc
        use rotation_matrices, only : degrees
        implicit none
        ! arguments
        real,       intent(in)      ::  score           !<  score of the image after refinement
        real,       intent(in)      ::  average_score   !<  average score of all images after refinement
        ! private variables
        logical,    parameter       ::  debug   =   .false.

        ! start work
        if (pbc .eq. 0.0) then
            bfactor = 0.0
        else
            if (debug) write(*,'(a,3(f0.4,1x))') '**debug(bfactor_from_score): average score, score, pbc: ', &
                                                degrees(acos(average_score)), degrees(acos(score)), pbc
            ! we want to return a negative b factor if the image has a better-than-average score
            bfactor = (degrees(acos(score)) - degrees(acos(average_score))) / pbc
            bfactor = max(bfactor,-2.0) ! a la frealign, we do not allow very negative b factors
            bfactor = bfactor * 16. ! this factor is there to match frealign. see notes of 29-apr-2012.
        endif

        if (debug) then
            write(*,'(a,3(f0.4,1x))') '**debug(bfactor_from_score): average score, score, bfactor: ', average_score, score, bfactor
        endif

    end function bfactor_from_score

    !>  \brief  apply a bfactor to the image / volume, in place
    subroutine apply_bfactor(   img,bfactor,   &
                                lp_cosine,lp_cosine_freq,lp_cosine_falloff, &
                                hp_cosine,hp_cosine_freq,hp_cosine_falloff, &
                                print_filter_shape,squared,filter_shape)
        use runtime_parameters, only : pi, terminate
        implicit none
        ! arguments
        class(image),                       intent(inout)   ::  img
        real,                               intent(in)      ::  bfactor             !<  b factor, unitless (to do this, take the b factor in a^2 and divide by the square of the pixel size). using the crystallographic convention, a la henderson & grigorieff
        logical,                optional,   intent(in)      ::  lp_cosine           !<  whether also to apply a cosine-edge low-pass filter. false by default.
        real,                   optional,   intent(in)      ::  lp_cosine_freq      !<  spatial frequency (in 1/pixel; 0.5 is nyquist) at which the cosine falloff ends. in other words, there is a guarantee that the falloff is complete by that point.
        real,                   optional,   intent(in)      ::  lp_cosine_falloff   !<  width (in 1/pixels) of the cosine falloff
        logical,                optional,   intent(in)      ::  hp_cosine           !<  whether also to apply a cosine-edge high-pass filter. false by default.
        real,                   optional,   intent(in)      ::  hp_cosine_freq      !<  spatial frequency (in 1/pixel; 0.5 is nyquist) at which the cosine falloff ends. in other words, the filter reaches 1.0 by that point.
        real,                   optional,   intent(in)      ::  hp_cosine_falloff   !<  width (in 1/pixels) of the cosine falloff
        logical,                optional,   intent(in)      ::  print_filter_shape  !<  print the values of the filter to output. defaults to false.
        logical,                optional,   intent(in)      ::  squared             !<  whether the weighting should be squared (useful for e.g. equation 1 of grigorieff 1998)
        real,   allocatable,    optional,   intent(inout)   ::  filter_shape(:)     !<  if present, the filter values will be returned in this array
        ! private variables
        integer ::  i,j,k
        logical ::  llp_cosine
        real    ::  llp_cosine_freq, llp_cosine_freq_min, llp_cosine_falloff, lp_cosine_attenuation
        logical ::  hhp_cosine
        real    ::  hhp_cosine_freq, hhp_cosine_freq_min, hhp_cosine_falloff, hp_cosine_attenuation
        logical ::  pprint_filter_shape
        real    ::  squaring_factor
        logical,    parameter   ::  debug   =   .false.
        real    ::  filter_value
        integer ::  orig(3)
        real    ::  coos(3), freq_sq
        real    ::  freq
        real,   allocatable ::  ffilter_shape(:)
        ! start work

        ! will we be squaring?
        squaring_factor = 1.0
        if (present(squared)) then
            if (squared) then
                squaring_factor = 2.0
            endif
        endif
        if (debug) write(*,'(a,f0.1)') '**debug(apply_bfactor): squaring factor = ', squaring_factor

        ! will we be doing a cosine-edge low-pass filter too?
        llp_cosine = .false.
        if (present(lp_cosine)) llp_cosine = lp_cosine
        hhp_cosine = .false.
        if (present(hp_cosine)) hhp_cosine = hp_cosine


        ! if we are doing a cosine-edge low-pass filter, check we have all the information we need
        if (llp_cosine) then
            if (present(lp_cosine_freq) .and. present(lp_cosine_falloff)) then
                llp_cosine_freq = lp_cosine_freq
                llp_cosine_falloff = lp_cosine_falloff
            else
                write(*,'(a)') '**error(apply_bfactor): you asked for a cosine-edged low-pass filter',  &
                                ' but did not supply needed arguments'
                call terminate('fatal error in apply_bfactor')
            endif
        else
            llp_cosine_freq = maxval(img%logical_dimensions(1:3)) + 2.
            llp_cosine_falloff = 1.
        endif

        ! if we are doing a cosine-edge low-pass filter, check we have all the information we need
        if (hhp_cosine) then
            if (present(hp_cosine_freq) .and. present(hp_cosine_falloff)) then
                hhp_cosine_freq = hp_cosine_freq
                hhp_cosine_falloff = hp_cosine_falloff
            else
                write(*,'(a)') '**error(apply_bfactor): you asked for a cosine-edged high-pass filter but ',    &
                                'did not supply needed arguments'
                call terminate('fatal error in apply_bfactor')
            endif
        else
            hhp_cosine_freq = 0.0
            hhp_cosine_falloff = 1.
        endif

        ! will we be printing out the shape of the filter?
        if (present(print_filter_shape)) then
            pprint_filter_shape = print_filter_shape
        else
            pprint_filter_shape = .false.
        endif

        ! check the image is in memory
        if (.not. is_allocated(img)) then
            write(*,'(a)') '**error(apply_factor): image not allocated'
            call terminate('fatal error in apply_bfactor')
        endif

        ! the square of the dimension of the image, times 4 because we follow the crystallography / henderson / grigorieff convention
!       nsam_sq = (img%logical_dimensions(1)*img%logical_dimensions(2)) * 4

        ! work out parameters for the cosine filters
        llp_cosine_freq_min =   max(llp_cosine_freq - llp_cosine_falloff,0.0)
        hhp_cosine_freq_min =   max(hhp_cosine_freq - hhp_cosine_falloff,0.0)

        if (debug) then
            write(*,'(a,3(f0.3,1x))') '**debug(apply_bfactor): lp: ', llp_cosine_freq_min, llp_cosine_freq, llp_cosine_falloff
            write(*,'(a,3(f0.3,1x))') '**debug(apply_bfactor): hp: ', hhp_cosine_freq_min, hhp_cosine_freq, hhp_cosine_falloff
        endif

        ! origin
        orig = [1,1,1]
        if (img%origin_at_center .and. img%is_in_real_space) orig = img%physical_address_of_box_center
        if (debug) write(*,'(a,3(i0,1x))') '**debug(apply_bfactor): origin = ', orig

        ! allocate filter_shape
        allocate(ffilter_shape(highest_index_dim1(img%logical_dimensions(1))))


        if (debug) write(*,'(a)',advance='no') '**debug(apply_bfactor): bfactor (not squared) = '
        if (img%is_in_real_space) then
            if (img%origin_at_center) then
                do k=1,img%logical_dimensions(3)
                    coos(3) = (real(k - orig(3))/real(img%logical_dimensions(3)))**2
                    do j=1,img%logical_dimensions(2)
                        coos(2) = (real(j - orig(2))/real(img%logical_dimensions(2)))**2
                        do i=1,img%logical_dimensions(1)
                            coos(1) = (real(i - orig(1))/real(img%logical_dimensions(1)))**2
                            ! compute squared radius, in units of reciprocal pixels
                            freq_sq = sum(coos)
                            freq = sqrt(freq_sq)
                            ! apply the low-pass filter
                            if (freq .lt. llp_cosine_freq_min) then
                                lp_cosine_attenuation = 1.0
                            else if (freq .lt. llp_cosine_freq) then
                                lp_cosine_attenuation = (cos(((freq-llp_cosine_freq_min)/llp_cosine_falloff)*pi)+1.)/2.0
                            else
                                lp_cosine_attenuation = 0.0
                            endif
                            ! apply the high-pass filter
                            if (freq .lt. hhp_cosine_freq_min) then
                                hp_cosine_attenuation = 0.0
                            else if (freq .lt. hhp_cosine_freq) then
                                hp_cosine_attenuation = (sin(((freq-hhp_cosine_freq_min)/hhp_cosine_falloff)*pi)+1.)/2.0
                            else
                                hp_cosine_attenuation = 1.0
                            endif
                            filter_value =      exp(-bfactor * squaring_factor * freq_sq / 4.0)  &
                                            *   lp_cosine_attenuation * hp_cosine_attenuation! 4.0 because we follow the crystallography / henderson / grigorieff convention
                            img%real_values(i,j,k) = img%real_values(i,j,k) * filter_value
                            if (j .eq. orig(2) .and. k .eq. orig(3) .and. i .ge. orig(1)) then
                                if (debug) write(*,'(g15.3,1x)',advance='no') filter_value
                                ffilter_shape(i-orig(1)+1) = filter_value
                            endif
                        enddo
                    enddo
                enddo
            else
                ! we assume that the image with the origin at pixel (1,1,1) has smaller extent along dimension 1
                if (img%logical_dimensions(1) .eq. img%logical_dimensions(2) .or. &
                    img%logical_dimensions(1) .eq. img%logical_dimensions(3)) then
                    call terminate('unexpected real image dimensions')
                endif
                do k=1,img%logical_dimensions(3)
                    ! the distance from the origin to the current pixel, in reciprocal pixel units
                    coos(3) = (comp_coo_from_phys_index_dim2or3(img%logical_dimensions(3),k)    &
                                                                        /real(img%logical_dimensions(3)))**2
                    do j=1,img%logical_dimensions(2)
                        coos(2) = (comp_coo_from_phys_index_dim2or3(img%logical_dimensions(2),j)    &
                                                                        /real(img%logical_dimensions(2)))**2
                        do i=1,img%logical_dimensions(1)
                            coos(1) = (comp_coo_from_phys_index_dim1(img%logical_dimensions(1),i)   &
                                                                        /real(img%logical_dimensions(1)))**2
                            ! compute squared radius, in units of reciprocal pixels
                            freq_sq = sum(coos)
                            freq = sqrt(freq_sq)
                            ! apply the low-pass filter
                            if (freq .lt. llp_cosine_freq_min) then
                                lp_cosine_attenuation = 1.0
                            else if (freq .lt. llp_cosine_freq) then
                                lp_cosine_attenuation = (cos(((freq-llp_cosine_freq_min)/llp_cosine_falloff)*pi)+1.)/2.0
                            else
                                lp_cosine_attenuation = 0.0
                            endif
                            ! apply the high-pass filter
                            if (freq .lt. hhp_cosine_freq_min) then
                                hp_cosine_attenuation = 0.0
                            else if (freq .lt. hhp_cosine_freq) then
                                hp_cosine_attenuation = (sin(((freq-hhp_cosine_freq_min)/hhp_cosine_falloff)*pi)+1.)/2.0
                            else
                                hp_cosine_attenuation = 1.0
                            endif
                            filter_value =      exp(-bfactor * squaring_factor * freq_sq / 4.0)  &
                                            *   lp_cosine_attenuation * hp_cosine_attenuation! 4.0 because we follow the crystallography / henderson / grigorieff convention
                            img%real_values(i,j,k) = img%real_values(i,j,k) * filter_value
                            if (j .eq. orig(2) .and. k .eq. orig(3)) then
                                if (debug) write(*,'(g15.3,1x)',advance='no') filter_value
                                ffilter_shape(i-orig(1)+1) = filter_value
                            endif
                        enddo
                    enddo
                enddo
            endif
        else
            ! image is complex
            do k=1,img%logical_dimensions(3)
                ! the distance from the origin to the current pixel, in reciprocal pixel units
                coos(3) = comp_coo_from_phys_index_dim2or3(img%logical_dimensions(3),k)**2
                do j=1,img%logical_dimensions(2)
                    coos(2) = comp_coo_from_phys_index_dim2or3(img%logical_dimensions(2),j)**2
                    do i=1,highest_index_dim1(img%logical_dimensions(1))
                        coos(1) = comp_coo_from_phys_index_dim1(img%logical_dimensions(1),i)**2
                        ! compute squared radius, in units of reciprocal pixels
                        freq_sq = sum(coos)
                        freq = sqrt(freq_sq)
                        ! apply the low-pass filter
                        if (freq .lt. llp_cosine_freq_min) then
                            lp_cosine_attenuation = 1.0
                        else if (freq .lt. llp_cosine_freq) then
                            lp_cosine_attenuation = (cos(((freq-llp_cosine_freq_min)/llp_cosine_falloff)*pi)+1.)/2.0
                        else
                            lp_cosine_attenuation = 0.0
                        endif
                        ! apply the high-pass filter
                        if (freq .lt. hhp_cosine_freq_min) then
                            hp_cosine_attenuation = 0.0
                        else if (freq .lt. hhp_cosine_freq) then
                            hp_cosine_attenuation = (sin(((freq-hhp_cosine_freq_min)/hhp_cosine_falloff)*pi)+1.)/2.0
                        else
                            hp_cosine_attenuation = 1.0
                        endif
                        filter_value =      exp(-bfactor * squaring_factor * freq_sq / 4.0)  &
                                        *   lp_cosine_attenuation * hp_cosine_attenuation! 4.0 because we follow the crystallography / henderson / grigorieff convention
                        img%complex_values(i,j,k) = img%complex_values(i,j,k) * filter_value
                        if (j .eq. orig(2) .and. k .eq. orig(3)) then
                            if (debug) write(*,'(g15.3,1x)',advance='no') filter_value
                            ffilter_shape(i-orig(1)+1) = filter_value
                        endif
                    enddo
                enddo
            enddo
        endif
        if (debug) write(*,'(a)') '.'

        ! print out shape of the filter if user asked for it
        if (pprint_filter_shape) then
            write(*,'(/a)') 'weight of combined b-factor & low-pass filters'
            write(*,'(a16,5x,a10)') 'freq (1/pixel)', 'weight'
            do i=1,size(ffilter_shape)
                freq = real(i-1) / real(img%logical_dimensions(2))
                write(*,'(f16.3,5x,f10.4)') freq, ffilter_shape(i)
            enddo
        endif

        !
        if (present(filter_shape)) filter_shape = ffilter_shape

    end subroutine apply_bfactor

    !>  \brief  apply a filter function to an image
    subroutine apply_filter(img,filter,filter_param)
        use runtime_parameters, only : terminate
        ! arguments
        class(image),       intent(inout)   ::  img
        procedure(filter_function)          ::  filter
        real,               intent(in)      ::  filter_param
        ! private variables
        integer ::  i,j,k
        ! start work

        ! check image is ready
        if (.not. img%in_memory) then
            write(*,'(a)') '**error(apply_filter): image not initialised / allocated'
            call terminate('fatal error in apply_filter')
        endif

        if (img%is_in_real_space) then
            do k=1,img%logical_dimensions(3)
                do j=1,img%logical_dimensions(2)
                    do i=1,img%logical_dimensions(1)
                        img%real_values(i,j,k) = img%real_values(i,j,k) * filter(radius(img,i,j,k),filter_param)
                    enddo
                enddo
            enddo
        else
            do k=1,img%logical_dimensions(3)
                do j=1,img%logical_dimensions(2)
                    do i=1,img%logical_dimensions(1)
                        img%complex_values(i,j,k) = img%complex_values(i,j,k) * filter(radius(img,i,j,k),filter_param)
                    enddo
                enddo
            enddo
        endif

    end subroutine apply_filter

    !>  \brief  generate a filter
    subroutine generate_filter_image(img,filter,filter_param)
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        type(image),        intent(inout)   ::  img
        procedure(filter_function)          ::  filter
        real,               intent(in)      ::  filter_param
        ! private variables
        integer ::  i,j,k
        ! start work

        ! check image is ready
        if (.not. img%in_memory) then
            write(*,'(a)') '**error(generate_filter_image): image not initialised / allocated'
            call terminate('fatal error in generate_filter_image')
        endif

        do k=1,img%logical_dimensions(3)
            do j=1,img%logical_dimensions(2)
                do i=1,img%logical_dimensions(1)
                    img%real_values(i,j,k) = filter(radius(img,i,j,k),filter_param)
                enddo
            enddo
        enddo
    end subroutine generate_filter_image


          ! ! ! ! ! ! ! ! ! ! ! ! ! !
    ! ! ! ! filter functions below  ! ! ! !
          ! ! ! ! ! ! ! ! ! ! ! ! ! !

    !>  \brief  gaussian function, but the width is defined by the full width at half maximum
    pure real function gauss_fwhm(arg,fwhm)
        implicit none
        real,   intent(in)  ::  arg     !<  argument
        real,   intent(in)  ::  fwhm    !<  full width at half maximum

        gauss_fwhm = gauss(arg,fwhm/(2.*sqrt(2.*log(2.))))
    end function gauss_fwhm

    !>  \brief  gaussian function centered at arg=0.0 with height 1.0
    pure real function gauss(arg,stddev)
        implicit none
        real,   intent(in)  ::  arg !<  argument
        real,   intent(in)  ::  stddev  !<  standard deviation of the gaussian bell

        gauss = exp(-arg**2/(2*stddev**2))
    end function gauss

    !>  \brief  a rectangle (in 1d), a disk (in 2d) or a sphere (in 3d) function. the radius is given by rad.
    pure real function disc(arg,rad)
        implicit none
        real,   intent(in)  ::  arg !<  argument
        real,   intent(in)  ::  rad !<  disc/sphere radius

        if (arg .le. rad) then
            disc = 1.0
        else
            disc = 0.0
        endif
    end function disc

    !>  \brief  a rectangle (in 1d), a disk (in 2d) or a sphere (in 3d) function. the radius is given by rad.
    pure real function disc_inverse(arg,rad)
        implicit none
        real,   intent(in)  ::  arg !<  argument
        real,   intent(in)  ::  rad !<  disc/sphere radius

        if (arg .gt. rad) then
            disc_inverse = 1.0
        else
            disc_inverse = 0.0
        endif
    end function disc_inverse



    !>  \brief  rotate/shift image. out of place. the rotation matrix defines the transformation from output to input.
    !!
    !!
    !!  \todo   change the convention, so that the matrix defines the transformation from input to output! (requires that every single call to rotate_image is checked carefully)
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine rotate_image(imgin,imgout,matrix,interpol_scheme,interpol_window,threading,rad_max_cyl,skip_checks)
        use interpolation
#ifndef SKIP_RUNTIME_CHECKS
        use rotation_matrices, only : print_rotmat
#endif
        implicit none
        ! arguments
        type(image),                            intent(in)      ::  imgin           !<  input image (target because window pointing at it)
        type(image),                            intent(inout)   ::  imgout          !<  output image (status is inout so that no reallocation happens)
        real,                                   intent(in)      ::  matrix(4,4)     !<  3d rotation & shift matrix to go from output to input.
        type(interpolation_scheme),             intent(in)      ::  interpol_scheme !<  interpolation scheme
        type(window),                           intent(inout)   ::  interpol_window !<  interpolation window
        logical,                    optional,   intent(in)      ::  threading       !<  whether to force threading
        real,                       optional,   intent(in)      ::  rad_max_cyl     !<  only output voxels within this radius (pixels) from the z axis will be computed
        logical,                    optional,   intent(in)      ::  skip_checks     !<  skip all safety checks on entry. true by default. only set to false once you're sure you're calling this correctly.
        ! private variables
        integer                 ::  i,j,k
        integer                 ::  addr_min(3), addr_max(3), imgout_ubound(3)
        integer                 ::  mid_coos(3)
        real                    ::  input_coos(4), output_coos(4), rvalue
        logical                 ::  two_d   ! whether we are dealing with 2d images
        complex                 ::  cvalue
        logical                 ::  tthreading
        real                    ::  rrad_max_cyl
        logical                 ::  sskip_checks
        real                    ::  background_value
        logical,    parameter   ::  debug = .false.

#ifndef SKIP_RUNTIME_CHECKS
        ! are we doing safety checks?
        if (present(skip_checks)) then
            sskip_checks = skip_checks
        else
            sskip_checks = .false.
        endif

        if (.not. sskip_checks) then
            ! check image type
            ! if we are dealing with a complex image, this subroutine only does rotation
            if (.not. imgin%is_in_real_space .and. any(matrix(1:3,4) .ne. 0.0) ) then
                write(*,'(a)') '**error(rotate_image): cannot shift complex image. use pshift to do shift.'
                call terminate('fatal error in rotate_image')
            endif

            ! we will only deal with real where the origin is at the center
            if (imgin%is_in_real_space .and. .not. imgin%origin_at_center) then
                write(*,'(a)') '**error(rotate_image): cannot operate on real images with origin not at center'
                call terminate('fatal error in rotate_image')
            endif

            ! debug
            if (debug) then
                write(*,'(a,l1,1x,l1)') '**debug(rotate_image): image in real space, origin_at_center = ', imgin%is_in_real_space, imgin%origin_at_center
            endif

            ! check allocation status of input
            if (.not. imgin%in_memory) then
                write(*,'(a)') '**error(rotate_image): input image is not allocated'
                call terminate('fatal error in rotate_image')
            endif

            ! check input and output images have the same type
            if (.not. same_type(imgin,imgout)) then
                imgout%is_in_real_space = imgin%is_in_real_space
            endif

            ! output must be same dimensions as input
            if (.not. same_dim(imgin,imgout)) then
                if (imgout%in_memory) call image_deallocate(imgout)
            endif

            ! if output image is not allocated, allocate it
            if (.not. imgout%in_memory) then
                imgout%is_in_real_space = imgin%is_in_real_space
                imgout%logical_dimensions = imgin%logical_dimensions
                imgout%origin_at_center = imgin%origin_at_center
                call image_allocate(imgout,imgin)
            endif


            ! check whether we are dealing with 2d images
            if (imgin%logical_dimensions(3) .eq. 1) then
                two_d = .true.
            else
                two_d = .false.
            endif
        endif   ! end of safety checks
#endif

        ! do we have a maximum cylinder radius?
        if (present(rad_max_cyl)) then
            rrad_max_cyl = rad_max_cyl**2
        else
            rrad_max_cyl = 0.0
        endif
#ifndef SKIP_RUNTIME_CHECKS
        if (rrad_max_cyl .ne. 0.0) then
            if (.not. imgin%is_in_real_space) then
                call terminate('cylinder radius option not implemented for complex images')
            endif
        endif

        ! debug
        if (debug) then
            !write(*,*) '**debug(rotate_image): input first line'
            !call print_line(imgin,1,1)
            print '(a)', '**debug(rotate_image): rotating by matrix:'
            call print_rotmat(matrix)
        endif
#endif

        ! mid_coos has the address of the origin
        mid_coos      = [1,1,1]
        if (imgin%origin_at_center .and. imgin%is_in_real_space) mid_coos = imgin%physical_address_of_box_center

        ! imgout_ubound(1:3) define the upper bound of the input image
        if (.not. imgout%is_in_real_space) then
            imgout_ubound(1) = highest_index_dim1(imgout%logical_dimensions(1))
        else
            imgout_ubound(1) = imgout%logical_dimensions(1)
        endif
        imgout_ubound(2:3) = imgout%logical_dimensions(2:3)

        ! addr_min(1:3) and addr_max(1:3) define the minimum and maximum addresses in each direction
        ! when the image is centered at address (/0,0,0/)
        call bounds_from_center(imgout,addr_min,addr_max)


        ! work out whether we'll be threading
        if (present(threading)) then
            tthreading = threading
        else
            if (imgin%logical_dimensions(1)*imgin%logical_dimensions(2)*imgin%logical_dimensions(3) .ge. 128**3 &
                    .and. imgin%logical_dimensions(3) .gt. 1) then
                tthreading = .true.
            else
                tthreading = .false.
            endif
        endif


        ! loop over output pixels, (i,j,k) is the output address, output_coos are the input coordinates (as though origin 0,0,0 at center),
        ! and input_coos are the input coordinates
        output_coos(4) = 1.0
        if (imgout%is_in_real_space) then
            ! choose a background value
            background_value = imgin%real_values(1,1,1)
            do k = 1, imgout_ubound(3)
                output_coos(3) = real(k-mid_coos(3))
                do j = 1, imgout_ubound(2) ! this loop (over lines) could be threaded
                    output_coos(2) = real(j-mid_coos(2))
                    do i = 1, imgout_ubound(1)
                        output_coos(1) = real(i-mid_coos(1))
                        ! input_coos(1:3) is (/0.,0.,0./) at origin of image

                        ! wipe the output
                        imgout%real_values(i,j,k) = background_value

                        ! check we're within the bounds of the cylinder
                        if (rrad_max_cyl .ne. 0.0) then
                            if (output_coos(1)**2+output_coos(2)**2 .gt. rrad_max_cyl) cycle
                        endif

                        ! calculate input coordinates
                        input_coos = matmul(matrix,output_coos)
                        !print *, '**debug(rotate_image): output_coos, input_coos = ', output_coos, input_coos
                        ! output_coos(1:3) is (/0.,0.,0./) at origin of image

                        ! skip this voxel if it is outside input image, even though part of the window may still fall within input image
                        if (any(input_coos(1:3) .gt. addr_max(1:3)) .or. any(input_coos(1:3) .lt. addr_min(1:3))) then
                            cycle
                        endif

                        ! if the voxel is 0.0-valued, skip it
                        if (imgin%real_values(nint(input_coos(1)+mid_coos(1)),nint(input_coos(2)+mid_coos(2)),   &
                                                         nint(input_coos(3)+mid_coos(3))) .eq. 0.0) then
                            cycle
                        endif

                        ! center the window around the output coos
                        interpol_window%centered_on = input_coos(1:3)

                        ! define the bounds of the window over the output image
                        ! these bounds can take any value, including 0, which means the address of the origin,
                        ! as well as negative values.
                        call eval_window_bounds(interpol_window)

                        ! if no part of the window falls onto the output image, skip to the next input voxel
                        if (     any(interpol_window%lbound(1:3) .gt. addr_max(1:3)) &
                            .or. any(interpol_window%ubound(1:3) .lt. addr_min(1:3))) then
                            !write(*,'(a,3(i3,x))') '**debug(move_image): window outside image (i,j,k): ', i,j,k
                            cycle
                        endif


                        ! copy image data to the window, get interpolated value from window
                        if (interpol_window%in_memory .and. interpol_window%own_memory) interpol_window%real_values = 0.0
                        rvalue = 0.0
                        call window_from_image_real(imgin,interpol_window)
                        !print *, '**debug(rotate_image): win%rpointer =', win%rpointer
                        call get_real_from_window(rvalue,interpol_window,interpol_scheme%interpolation_function,   &
                                                                interpol_scheme%windowing_function)
                        ! does the following line need to be critical? it shouldn't since there is only one write to each position in imgout%real_values.
                        imgout%real_values(i,j,k) = rvalue
                    enddo
                enddo
            enddo
        else
            do k = 1, imgout_ubound(3)
                output_coos(3) = real(k-nearest_orig(k,imgout%logical_dimensions(3)))
                ! input_coos(1:3) is (/0.,0.,0./) at origin of image
                do j = 1, imgout_ubound(2)
                    output_coos(2) = real(j-nearest_orig(j,imgout%logical_dimensions(2)))
                    do i = 1, imgout_ubound(1)
                        !if (debug) write(*,'(a,i3,a,i3)') '**debug(rotate_image): i = ', i, ' of ', imgin_ubound(1)
                        output_coos(1) = real(i-mid_coos(1))

                        ! calculate output coordinates
                        input_coos = matmul(matrix,output_coos)
                        ! output_coos(1:3) is (/0.,0.,0./) at origin of image

                        ! center the window around the output coos
                        interpol_window%centered_on = input_coos(1:3)

                        ! define the bounds of the window over the output image
                        ! these bounds can take any value, including 0, which means the address of the origin,
                        ! as well as negative values.
                        call eval_window_bounds(interpol_window)

                        ! if no part of the window falls onto the output image, skip to the next input voxel
                        if (     any(interpol_window%lbound(1:3) .gt. addr_max(1:3))     &
                            .or. any(interpol_window%ubound(1:3) .lt. addr_min(1:3))) then
                            imgout%complex_values(i,j,k) = (0.0,0.0)
                            cycle
                        endif

                        ! copy image data to the window, get interpolated value from window
                        call window_from_image_comp(imgin,interpol_window)
                        call get_comp_from_window(cvalue,interpol_window,interpol_scheme%interpolation_function,   &
                                                            interpol_scheme%windowing_function)
                        imgout%complex_values(i,j,k) = cvalue

                    enddo
                enddo
            enddo
        endif

    end subroutine rotate_image

    !>  \brief  re-arrange the data in the real array so that the image is effectively rotated by 180 degrees,
    !!          though not around the usual rotation center
    subroutine image_rotate_180(self)
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  self
        ! private variables

        ! start work
        self%real_values(1:self%logical_dimensions(1),1:self%logical_dimensions(2),1:self%logical_dimensions(3)) = &
                self%real_values(self%logical_dimensions(1):1:-1,self%logical_dimensions(2):1:-1,self%logical_dimensions(3):1:-1)
    end subroutine image_rotate_180


    !>  \brief  shift the real-space origin of the image. if it is currently at the center pixel, move it to the first pixel (corner with address 1,1,1), and vice versa.
    !!  the image is always returned in the space it was given
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine shift_real_space_origin(self)
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  self    !<  image object
        ! private variables
        logical ::  real_on_entry
        real    ::  shift(3)
        ! start work

        ! check image type on entry
        real_on_entry = self%is_in_real_space

        ! work out shift to apply
        shift(1:3) = real(index_of_central_pixel_from_dim(self%logical_dimensions(1:3))) - 1.0
        if (self%logical_dimensions(3) .eq. 1) shift(3) = 0.0

        ! apply appropriate phase shift
        if (real_on_entry) call fwd_ft(self)
        call pshift_image(self,-shift(1),-shift(2),-shift(3))
        if (real_on_entry) call bwd_ft(self)

        ! keep track of where the real-space origin is now
        self%origin_at_center = .not. self%origin_at_center
    end subroutine shift_real_space_origin

    !>  \brief  complex image translation using phase shift. shifts are given in pixels.
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine pshift_image(img,shift_x,shift_y,shift_z)
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  img     !< image to be shifted
        real,           intent(in   )   ::  shift_x !< shift in x direction (pixels)
        real,           intent(in   )   ::  shift_y !< shift in y direction (pixels)
        real,           intent(in   )   ::  shift_z !< shift in z direction (pixels)
        ! private variables
        integer         ::  i,j,k
        integer         ::  iori,jori,kori  ! origin coordinates
        real(kind=8)    ::  phase_x,phase_y,phase_z,phase
        complex(kind=8) ::  phaseshift
        logical,    parameter   ::  debug   =   .false.

        ! start work
#ifndef SKIP_RUNTIME_CHECKS
        if (debug) write(*,'(a,3(f0.2,1x))') '**debug(pshift_image): shifting image by ', shift_x, shift_y, shift_z

        ! check image type
        if (img%is_in_real_space) then
            write(*,'(3a)') '**error(pshift_image): can only operate on images of type comp. type ', 'real', ' not supported'
            call terminate('fatal error in pshift_image')
        endif
#endif

        ! loop over voxels
        phase = 0.0d0
        do k=1,img%logical_dimensions(3)
            kori = nearest_orig(k,img%logical_dimensions(3))
            phase_z = phase_from_shift(dble(shift_z),k-kori,img%logical_dimensions(3))
            do j=1,img%logical_dimensions(2)
                jori = nearest_orig(j,img%logical_dimensions(2))
                phase_y = phase_from_shift(dble(shift_y),j-jori,img%logical_dimensions(2))
                do i=1,highest_index_dim1(img%logical_dimensions(1))
                    iori = 1
                    phase_x = phase_from_shift(dble(shift_x),i-iori,img%logical_dimensions(1))
                    ! calculate the phase shift
                    phaseshift = phase_shift_3d(phase_x,phase_y,phase_z)
                    ! apply the phase shift
                    img%complex_values(i,j,k)= img%complex_values(i,j,k) * phaseshift
                enddo
            enddo
        enddo
    end subroutine pshift_image

    !>  \brief  compute the phase shift, given the phase in each dimension
    pure function phase_shift_3d(phase_x,phase_y,phase_z)
        implicit none
        ! arguemnts
        real(kind=8),   intent(in)  ::  phase_x,phase_y,phase_z
        ! result
        complex(kind=8) ::  phase_shift_3d
        ! private variables
        real(kind=8)    ::  phase_tmp
        ! start work
        phase_tmp = - phase_x - phase_y - phase_z
        phase_shift_3d = cmplx(cos(phase_tmp),sin(phase_tmp),kind=8)
    end function phase_shift_3d

    !>  \brief  compute the phase to be applied to an element dist_from_origin away from the origin so that the real-space shift
    !!          shift is applied
    pure function phase_from_shift(shift,dist_from_origin,ldim)
        use runtime_parameters, only : dpi
        implicit none
        ! arguments
        real(kind=8),   intent(in)      ::  shift               !<  real space shift (in pixels)
        integer,        intent(in)      ::  dist_from_origin    !<  distance from origin in fourier transform (in pixels) along the axis of the shift
        integer,        intent(in)      ::  ldim                !<  the total extent of the image in the dimension relevant to the shift
        ! result
        real(kind=8)    ::  phase_from_shift
        ! private variables

        ! start work
        phase_from_shift = shift*real(dist_from_origin,kind=8)*2.0d0*dpi/real(ldim,kind=8)
    end function phase_from_shift



    !>  \brief  add noise to an image, modelling micrograph noise
    subroutine image_add_detection_noise(self,dose,counting)
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  self        !<  image to which noise will be added
        real,           intent(in)      ::  dose        !<  number of electrons incident on specimen per pixel (assume this is a constant)
        logical,        intent(in)      ::  counting    !<  whether we are simulating a counting image
        ! private variables
        logical,        parameter       ::  debug   =   .true.
        ! start work

        if (debug) call write_to_disk(self,'dbg_micr_input.mrc')

        ! scale the density values to number of electrons
        self%real_values = self%real_values *1000.0

        ! electron statistics
        call image_poissonify(self)

        if (debug) call write_to_disk(self,'dbg_micr_poisson.mrc')

        ! linear mode
        if (.not. counting) then
            call image_landauify(self)
            if (debug) call write_to_disk(self,'dbg_micr_landau.mrc')
        endif

        if (debug) call write_to_disk(self,'dbg_micr_output.mrc')
    end subroutine image_add_detection_noise

    !>  \brief  replace each value in an image by a value randomly selected from a poisson distribution
    !!          with parameter equal to the input value. output therefore has integer values.
    subroutine image_poissonify(self)
        use fgsl
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  self
        ! private variables
        type(fgsl_rng)          ::  rng
        type(fgsl_rng_type)     ::  default_rng
        integer                 ::  i,j,k
        ! start work

        ! create a generator (the algorithm is chosen by the environment variable gsl_rng_type
        default_rng = fgsl_rng_env_setup()
        rng         = fgsl_rng_alloc(default_rng)

        ! loop over the voxels
        do k=1,self%logical_dimensions(3)
            do j=1,self%logical_dimensions(2)
                do i=1,self%logical_dimensions(1)
                    self%real_values(i,j,k) = real(fgsl_ran_poisson(rng,real(self%real_values(i,j,k),fgsl_double)))
                enddo
            enddo
        enddo
    end subroutine image_poissonify

    !>  \brief  for each count in input image, add a value selected from a landau distribution
    subroutine image_landauify(self)
        use fgsl
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  self
        ! private variables
        type(fgsl_rng)          ::  rng
        type(fgsl_rng_type)     ::  default_rng
        integer                 ::  i,j,k,e
        integer                 ::  n_electrons
        ! start work

        ! create a generator (the algorithm is chosen by the environment variable gsl_rng_type
        default_rng = fgsl_rng_env_setup()
        rng         = fgsl_rng_alloc(default_rng)

        ! loop over the voxels
        do k=1,self%logical_dimensions(3)
            do j=1,self%logical_dimensions(2)
                do i=1,self%logical_dimensions(1)
                    n_electrons = int(self%real_values(i,j,k))
                    self%real_values(i,j,k) = 0.0
                    do e=1,n_electrons
                        self%real_values(i,j,k) = self%real_values(i,j,k) + fgsl_ran_landau(rng)
                    enddo
                enddo
            enddo
        enddo
    end subroutine image_landauify

    !>  \brief adds gaussian noise to an image. stdev_noise is a real.
    subroutine image_add_gauss_noise(img, stdev_noise,seed)
        use fgsl
        implicit none
        ! arguments
        class(image) ,          intent(inout)   ::  img
        real        ,           intent(in)      ::  stdev_noise
        integer,    optional,   intent(in)      ::  seed        !<  if none supplied, default to 0
        ! private variables
        integer                         :: i, j, k
        type(fgsl_rng)          ::  rng
        type(fgsl_rng_type)     ::  default_rng
        real(fgsl_double)       ::  stdev_noise_fgsl
        integer(kind=8)         ::  sseed

        sseed = 0
        if (present(seed)) sseed = seed

        !if you inputted snr, this is how noise variance would be calculated.
        !grey_var = variance_real(imgin)
        !noise_var=grey_var/snr

        stdev_noise_fgsl=stdev_noise
        default_rng= fgsl_rng_env_setup()
        rng=fgsl_rng_alloc(default_rng)
        call fgsl_rng_set(rng,sseed)

        do i=1,img%logical_dimensions(1)
            do j=1,img%logical_dimensions(2)
                do k=1,img%logical_dimensions(3)
                    img%real_values(i,j,k)=img%real_values(i,j,k)+fgsl_ran_gaussian(rng, stdev_noise_fgsl)
                enddo
            enddo
        enddo

        call fgsl_rng_free(rng)
    end subroutine image_add_gauss_noise




    !>  \brief  a generalised routine for the compensating for the effects of the interpolation kernel used during  slice insertion
    !!
    !!  you need to supply the appropriate instrument function corresponding to the interpolation kernel used during
    !!  slice insertion. this should be the instrument function corresponding to the apodization function:
    !!  http://mathworld.wolfram.com/apodizationfunction.html
    !!  for example, if a rectangle function was used, you need to correct with a sinc
    !!  the image_interpol module should have an _instr function for each interpol/window function to make this easier for you!
    !!
#if defined(SKIP_RUNTIME_CHECKS)
    pure &
#endif
    subroutine correct_3d(img,interpol_function_instr,window_function_instr,window_width,reverse)
        use interpolation, only : interpolation_function, WINDOWING_FUNCTION, INTERPOL_WINDOW_CONVOLUTION
#if defined (SKIP_RUNTIME_CHECKS)
#else
		use runtime_parameters, only : terminate
#endif
        implicit none
        ! arguments
        type(image),    intent(inout)               ::  img                     !<  volume to be corrected
        procedure(interpolation_function),  pointer ::  interpol_function_instr !<  the instrument function corresponding to the interpolation function used to calculate the 3d reconstruction
        procedure(windowing_function),      pointer ::  window_function_instr   !<  the instrument function corresponding to the windowing function used in calculating the 3d
        integer,    intent(in)                      ::  window_width            !<  window width used
        logical,    optional,   intent(in)          ::  reverse                 !<  unless reverse=.true., the volume will be divided by the intrument function
        ! private variables
        logical ::  rreverse
        integer ::  i,j,k
        real    ::  rad
        real    ::  boxsize(3)
        real    ::  w(3)
        integer ::  origin(3)
        integer ::  largest_dim
        integer ::  max_irad
        real,       allocatable     ::  convoluted_function(:,:)
        logical,    parameter       ::  debug   =   .false.
        ! start work
#if defined(SKIP_RUNTIME_CHECKS)
#else
        if (debug) print '(a,3(i3,1x))', '**debug(correct_3d): img%logical_dimensions = ', img%logical_dimensions
        ! this is only meant with images with origin at center
        if (img%is_in_real_space .and. .not. img%origin_at_center) then
            write(*,'(a)') '**error(correct_3d): cannot deal with real images whose origin is not at the center'
            call terminate('fatal error in correct_3d')
        endif
#endif

        ! check whether we need to correct for, or apply the fall-off
        if (present(reverse)) then
            rreverse = reverse
        else
            rreverse = .false.
        endif

        ! which is the largest dimension?
        largest_dim = maxloc(img%logical_dimensions,dim=1)

        ! the box size
        boxsize = real(img%logical_dimensions)

        ! the origin
        if (img%is_in_real_space .and. img%origin_at_center) then
            origin = img%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif

        ! the maximum radius (including the origin)
        max_irad = max( img%logical_dimensions(largest_dim) - origin(largest_dim) + 1, &
                        origin(largest_dim) )

#if defined(SKIP_RUNTIME_CHECKS)
#else
        if (debug) write(*,'(a,f0.2,1x,2(i0,1x))') '**debug(correct_3d): boxsize, origin, max_irad = ', boxsize, origin(1), max_irad
#endif

        ! work out values for the convolution of window_instr with interpol_instr
        allocate(convoluted_function(max_irad,3))
        do j=1,3
        do k=1,max_irad
            rad  = real(k)
            convoluted_function(k,j) = interpol_window_convolution( rad/boxsize(j),interpol_function_instr,window_function_instr, &
                                                                    window_width,img%logical_dimensions(j))

#if defined(SKIP_RUNTIME_CHECKS)
#else
            if (debug) write(*,'(a,f0.6,1x,f0.6)') '**debug(correct_3d): rad, convoluted_function(rad): ',   &
                                                    rad, convoluted_function(k,1)
#endif
        enddo
        enddo

        !
        if (img%is_in_real_space) then
            do k=1,img%logical_dimensions(3)
                w(3) = convoluted_function(abs(k-origin(3))+1,3)
                do j=1,img%logical_dimensions(2)
                    w(2) = convoluted_function(abs(j-origin(2))+1,2)
                    do i=1,img%logical_dimensions(1)
                        w(1) = convoluted_function(abs(i-origin(1))+1,1)
                        if (rreverse) then
                            img%real_values(i,j,k) = img%real_values(i,j,k) * w(1)*w(2)*w(3)
                        else
                            img%real_values(i,j,k) = img%real_values(i,j,k) / (w(1)*w(2)*w(3))
                        endif
                    enddo
                enddo
            enddo
        else
            ! in fourier space
            do k=1,img%logical_dimensions(3)
                w(3) = convoluted_function(abs(k-index_nearest_origin(k,img%logical_dimensions(3)))+1,3)
                do j=1,img%logical_dimensions(2)
                    w(2) = convoluted_function(abs(j-index_nearest_origin(j,img%logical_dimensions(2)))+1,2)
                    do i=1,highest_index_dim1(img%logical_dimensions(1))
                        w(1) = convoluted_function(i,1)
                        if (rreverse) then
                            img%complex_values(i,j,k) = img%complex_values(i,j,k) * w(1)*w(2)*w(3)
                        else
                            img%complex_values(i,j,k) = img%complex_values(i,j,k) / (w(1)*w(2)*w(3))
                        endif
                    enddo
                enddo
            enddo
        endif
    end subroutine correct_3d

    !>  \brief  get a complex value from an image by interpolation
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine get_comp_from_image(cvalue,img,interpol,win,coos_min,coos_max)
        use INTERPOLATION, only : interpolation_scheme
        implicit none
        ! arguments
        complex,                            intent(out)         ::  cvalue      !<  value extracted from the image
        type(image),                        intent(in)          ::  img
        type(interpolation_scheme),         intent(in)          ::  interpol
        type(window),                       intent(inout)       ::  win
        integer,                            intent(in)          ::  coos_min(3) !<  the minimum logical address in the image
        integer,                            intent(in)          ::  coos_max(3) !<  the maximum logical address in the image
        ! private variables
        logical,    parameter   ::  use_hardcoded_interpol_if_available = .true.
        integer ::  i,j,k,ii,jj,kk
        real    ::  dist(3), weight(3)
        integer ::  tmp_coos(3), tmp_coos_in(3)
        ! start work
        if (use_hardcoded_interpol_if_available .and. interpol%id .eq. 'line') then
            cvalue = (0.0,0.0)

            ! we want to interpolate to the coordinates win%center(1:3) using linear interpolation
            !
            ! in one dimension, suppose we are 0.3 away from point 0 (and therefore 0.7 away from point 1)
            ! the result should be value0*(1-0.3) + value1*0.3
            ! = value0 - 0.3 value0 + 0.3 value1
            ! = value0 + 0.3 (value1-value0)
            !
            ! so we need the distance from the real coordinates to the previous pixel
            !dist = win%center - floor(win%center)

            ! we also need to the values of the nearest neighbours
            !logi_nn_000 = floor(win%center)
            !phys_nn_000 = comp_addr_phys_from_logi(img,logi_nn_000)



            ! iterate over window voxels. i,j,k is address within window, ii,jj,kk is logical address within image (with origin 0,0,0 at center),
            ! and iii,jjj,kkk is physical address within image (with origin 1,1,1) at top-left
            do k=1,2
                kk  = k + win%lbound(3) - 1
                if (kk .lt. coos_min(3) .or. kk .gt. coos_max(3)) cycle
                dist(3) = abs(k - win%center(3))
                weight(3) = interpol%interpolation_function(dist(3))
                do j=1,2
                    jj = j + win%lbound(2) - 1
                    if (jj .lt. coos_min(2) .or. jj .gt. coos_max(2)) cycle
                    dist(2) = abs(j - win%center(2))
                    weight(2) = interpol%interpolation_function(dist(2))
                    do i=1,2
                        ii = i + win%lbound(1) - 1
                        if (ii .lt. coos_min(1) .or. ii .gt. coos_max(1)) cycle
                        dist(1) = abs(i - win%center(1))
                        weight(1) = interpol%interpolation_function(dist(1))
                        ! get address within complex image which corresponds to voxel coordinate
                        tmp_coos_in = (/ii,jj,kk/)
                        call comp_addr_phys_from_logi(img,tmp_coos_in,tmp_coos)
                        ! now copy the voxel value over
                        if ( ii .ge. 0) then
                            cvalue = cvalue + img%complex_values(tmp_coos(1),tmp_coos(2),tmp_coos(3)) &
                                                                            * weight(1) * weight(2) * weight(3)
                        else
                            cvalue = cvalue + conjg(img%complex_values(tmp_coos(1),tmp_coos(2),tmp_coos(3)))    &
                                                                            * weight(1) * weight(2) * weight(3)
                        endif
                    enddo
                enddo
            enddo

        else
            call window_from_image_comp(img,win)
            call get_comp_from_window(cvalue,win,interpol%interpolation_function,interpol%windowing_function)
        endif
    end subroutine get_comp_from_image


    !>  \brief  get a complex value from the window by interpolation
#if defined(SKIP_RUNTIME_CHECKS)
    pure &
#endif
    subroutine get_comp_from_window(cvalue,win,interpol_function,window_function,square)
        use interpolation, only : interpolation_function, windowing_function
        implicit none
        ! arguments
        complex,                intent(out)             ::  cvalue              !< the value to extract from the window
        type(window),           intent(in)              ::  win                 !< the window to extract the value from
        procedure(interpolation_function),  pointer     ::  interpol_function
        procedure(windowing_function),      pointer     ::  window_function
        logical,    optional,   intent(in)              ::  square              !<  whether the interpolated value should be squared
        ! private variables
        integer ::  i,j,k
        real    ::  dist(3), weight(3)
        logical ::  ssquare

        ! if all the values in the window are zero, don't need to do anything!
        if (sum(win%cpointer) .eq. (0.0,0.0)) then
            cvalue = (0.0,0.0)
            return
        endif

        ! will we be squaring?
        if (present(square)) then
            ssquare = square
        else
            ssquare = .false.
        endif

#if defined(SKIP_RUNTIME_CHECKS)
#else
        if (ssquare) then
            print *, '**warning(get_comp_from_window): squaring! are you sure you want to do that?'
            call terminate('get_comp_from_window')
        endif
#endif

        ! start work
        cvalue = (0.0,0.0)
        do k=1,win%logical_dimensions(3)
            dist(3) = abs(k - win%center(3))
            weight(3) = interpol_function(dist(3)) * window_function(dist(3),win%width)
            do j=1,win%logical_dimensions(2)
                dist(2) = abs(j - win%center(2))
                weight(2) = interpol_function(dist(2)) * window_function(dist(2),win%width)
                do i=1,win%logical_dimensions(1)
                    if (win%cpointer(i,j,k) .eq. (0.0,0.0)) cycle
                    dist(1) = abs(i - win%center(1))
                    weight(1) = interpol_function(dist(1)) * window_function(dist(1),win%width)
                    cvalue = cvalue + win%cpointer(i,j,k) * weight(1) * weight(2) * weight(3)
                enddo
            enddo
        enddo

        if (ssquare) then
            cvalue = cvalue**2
        endif
    end subroutine get_comp_from_window


    !>  \brief  get a real value from the window by interpolation
    pure subroutine get_real_from_window(rvalue,win,interpol_function,window_function,square)
        use interpolation, only : interpolation_function, windowing_function
        implicit none
        ! arguments
        real,                   intent(out)             ::  rvalue              !< the value to extract from the window
        type(window),           intent(in)              ::  win                 !< the window to extract the value from
        procedure(interpolation_function),  pointer     ::  interpol_function
        procedure(windowing_function),      pointer     ::  window_function
        logical,    optional,   intent(in)              ::  square              !<  whether the interpolated value should be squared
        ! private variables
        integer ::  i,j,k
        real    ::  dist(3), weight(3)

        ! start work
        rvalue = 0.0
        do k=1,win%logical_dimensions(3)
            dist(3) = abs(k - win%center(3))
            weight(3) = interpol_function(dist(3)) * window_function(dist(3),win%width)
            do j=1,win%logical_dimensions(2)
                dist(2) = abs(j - win%center(2))
                weight(2) = interpol_function(dist(2)) * window_function(dist(2),win%width)
                do i=1,win%logical_dimensions(1)
                    if (win%rpointer(i,j,k) .eq. 0.0) cycle
                    dist(1) = abs(i - win%center(1))
                    weight(1) = interpol_function(dist(1)) * window_function(dist(1),win%width)
                    rvalue = rvalue + win%rpointer(i,j,k) * weight(1) * weight(2) * weight(3)
                enddo
            enddo
        enddo
        if (present(square)) then
            if (square) then
                rvalue = rvalue**2
            endif
        endif
    end subroutine get_real_from_window


    !>  \brief  spread a value into a window
#if defined(SKIP_RUNTIME_CHECKS)
    pure &
#endif
    subroutine put_real_into_window(rvalue,win,interpol_function,window_function)
        use interpolation, only : interpolation_function, windowing_function
        implicit none
        ! arguments
        real,   intent(in)                              ::  rvalue                  !<  the value to spread into the window
        type(window),   intent(inout)                   ::  win                     !<  the window to spread the value into
        procedure(interpolation_function),  pointer     ::  interpol_function       !<  the interpolation function to use
        procedure(windowing_function),      pointer     ::  window_function         !<  the windowing function to use
        ! private variables
        integer ::  i,j,k
        real    ::  dist(3), weight(3)

        if (win%own_memory) then
            win%real_values = 0.0
        else
#if defined(SKIP_RUNTIME_CHECKS)
#else
            write(*,'(a)') '**error(put_real_into_window): can only work on windows with own memory'
            call terminate('fatal error in put_real_into_window')
#endif
        endif

        ! if input value is 0.0, no work to do
        if (rvalue .eq. 0.0) then
            return
        endif

        ! start work
        do k=1,win%logical_dimensions(3)
            dist(3) = abs(k - win%center(3))
            weight(3) = interpol_function(dist(3)) * window_function(dist(3),win%width)
            do j=1,win%logical_dimensions(2)
                dist(2) = abs(j - win%center(2))
                weight(2) = interpol_function(dist(2)) * window_function(dist(2),win%width)
                do i=1,win%logical_dimensions(1)
                    dist(1) = abs(i - win%center(1))
                    weight(1) = interpol_function(dist(1)) * window_function(dist(1),win%width)
                    !win%real_values(i,j,k) = rvalue * weight(1) * weight(2) * weight(3)
                    win%rpointer(i,j,k) = win%rpointer(i,j,k) + rvalue * weight(1) * weight(2) * weight(3)
                enddo
            enddo
        enddo
    end subroutine put_real_into_window

    !>  \brief spread a complex value into a window
#if defined(SKIP_RUNTIME_CHECKS)
    pure &
#endif
    subroutine put_comp_into_window(cvalue,win,interpol_function,window_function)
        use interpolation, only : interpolation_function, windowing_function
        implicit none
        ! arguments
        complex,                    intent(in)          ::  cvalue              !< the value to spread into the window
        type(window),               intent(inout)       ::  win                 !< the window to spread the value into
        procedure(interpolation_function),  pointer     ::  interpol_function
        procedure(windowing_function),      pointer     ::  window_function
        ! private variables
        integer ::  i,j,k
        real    ::  dist(3), weight(3)

        ! diagnostics
        !write(*,'(a,2es12.3,a,3f6.2)') '**debug(put_comp_into_window): putting ', cvalue, ' into window center at ', win%center
        !write(*,'(a,3(i3,x))') '**debug(put_comp_into_window): window dimensions =', win%logical_dimensions
        !write(*,'(a,i3)') '**debug(put_comp_into_window): window width =', win%width

#if defined(SKIP_RUNTIME_CHECKS)
#else
        if (.not. win%own_memory) then
            write(*,'(a)') '**error(put_comp_into_window): can only work on windows with own memory'
            call terminate('fatal error in put_comp_into_window')
        endif
#endif

        ! start work
        do k=1,win%logical_dimensions(3)
            dist(3) = abs(real(k) - win%center(3))
            weight(3) = interpol_function(dist(3)) * window_function(dist(3),win%width)
            do j=1,win%logical_dimensions(2)
                dist(2) = abs(real(j) - win%center(2))
                weight(2) = interpol_function(dist(2)) * window_function(dist(2),win%width)
                do i=1,win%logical_dimensions(1)
                    dist(1) = abs(real(i) - win%center(1))
                    weight(1) = interpol_function(dist(1)) * window_function(dist(1),win%width)
                    win%cpointer(i,j,k) = cvalue * weight(1) * weight(2) * weight(3)
                enddo
            enddo
        enddo
    end subroutine put_comp_into_window

    !>  \brief  fill the output image with the value of the interpolation function with argument = radius of image pixel
    subroutine image_from_interpol_function(img,interpol_function)
        use interpolation, only : interpolation_function
        implicit none
        type(image),    intent(inout)   ::  img !<  image to be convoluted with the kernel - overwritten
        procedure(interpolation_function)       ::  interpol_function
        ! private variables
        integer         ::  i,j,k
        real            ::  rad
        logical         ::  dnf(3)

        ! start work
        if (img%is_in_real_space) then
            write(*,*) 'here'
            do k=1,img%logical_dimensions(3)
                do j=1,img%logical_dimensions(2)
                    write(*,*) '**debug: j=', j, ' of ', img%logical_dimensions(2)
                    do i=1,img%logical_dimensions(1)
                        rad = radius(img,i,j,k)
                        img%real_values(i,j,k) = interpol_function(rad)
                    enddo
                enddo
            enddo
        else
            if (.not. is_square(img)) then
                call terminate('not implemented for non-cube/square images')
            endif
            dnf = (/.true.,.false.,.false./)
            do k=1,img%logical_dimensions(3)
                do j=1,img%logical_dimensions(2)
                    do i=1,highest_index_dim1(img%logical_dimensions(1))
                        rad = radius(img,i,j,k,do_not_fold=dnf)
                        img%complex_values(i,j,k) = interpol_function(rad)
                    enddo
                enddo
            enddo
        endif
    end subroutine image_from_interpol_function

    !>  \brief  generate an image of a windowing function
    subroutine image_from_window_function(img,window_function,window_width)
        use interpolation, only : windowing_function
        implicit none
        type(image),    intent(inout)   ::  img !<  image to be convoluted with the kernel - overwritten
        procedure(windowing_function)           ::  window_function
        integer,    intent(in)  ::  window_width
        ! private variables
        integer         ::  i,j,k
        logical         ::  dnf(3)
        real            ::  rad

        ! start work
        if (img%is_in_real_space) then
            do k=1,img%logical_dimensions(3)
                do j=1,img%logical_dimensions(2)
                    do i=1,img%logical_dimensions(1)
                        rad = radius(img,i,j,k)
                        img%real_values(i,j,k) = window_function(rad,window_width)
                    enddo
                enddo
            enddo
        else
            if (.not. is_square(img)) then
                call terminate('not implemented for non-cube/square images')
            endif
            dnf = (/.true.,.false.,.false./)
            do k=1,img%logical_dimensions(3)
                do j=1,img%logical_dimensions(2)
                    do i=1,highest_index_dim1(img%logical_dimensions(1))
                        rad = radius(img,i,j,k,do_not_fold=dnf)
                        img%complex_values(i,j,k) = window_function(rad,window_width)
                    enddo
                enddo
            enddo
        endif
    end subroutine image_from_window_function



    !>  \brief  window part of an image and copy the contents into a window object
    subroutine window_from_image(img,win,coos)
        implicit none
        ! arguments
        type(image),    intent(in)      ::  img     !<  image to extract data from
        type(window),   intent(inout)   ::  win     !<  window to store the data
        real,           intent(in)      ::  coos(3) !<  coordinates around which to extract the window
        ! private variables

        ! start work

        ! find the bounds of the window
        win%centered_on = coos
        call eval_window_bounds(win)

        if (.not. img%is_in_real_space) then
            call window_from_image_comp(img,win)
        else
            call window_from_image_real(img,win)
        endif
    end subroutine window_from_image

    !>  \brief  window object is populated with data from image object over which it is hovering
    !!  \todo   remove built-in assumption that image has even dimensions by calling origin_addr
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine window_from_image_real(img,win)
        implicit none
        ! arguments
        type(image),    target, intent(in)      ::  img !<  image to extract data from
        type(window),   target, intent(inout)   ::  win !<  window to store the data
        ! private variables
        integer ::  i,j,k
        integer ::  ii,jj,kk
        integer ::  iii,jjj,kkk
        integer ::  coos_min(3),coos_max(3)
        integer ::  lbnd(3), ubnd(3), img_origin(3)
        ! start work
        call bounds_from_center(img,coos_min,coos_max)
#ifndef SKIP_RUNTIME_CHECKS
        if (win%own_memory) then
#endif
            win%real_values = 0.0
            ! iterate over window voxels. i,j,k is physical address within window,
            ! ii,jj,kk is logical address within image (with origin 0,0,0 at center),
            ! and iii,jjj,kkk is physical address within image (with origin 1,1,1) at top-left
            do k=1,win%logical_dimensions(3)
                kk  = k + win%lbound(3) - 1
                if (kk .lt. coos_min(3) .or. kk .gt. coos_max(3)) cycle
                kkk = kk + img%logical_dimensions(3)/2 + 1
                if (img%logical_dimensions(3) .eq. 1) kkk = 1
                do j=1,win%logical_dimensions(2)
                    jj = j + win%lbound(2) - 1
                    if (jj .lt. coos_min(2) .or. jj .gt. coos_max(2)) cycle
                    jjj = jj + img%logical_dimensions(2)/2 + 1
                    do i=1,win%logical_dimensions(1)
                        ii = i + win%lbound(1) - 1
                        if (ii .lt. coos_min(1) .or. ii .gt. coos_max(1)) cycle
                        iii = ii + img%logical_dimensions(1)/2 + 1
                        ! copy voxel value over
                        win%real_values(i,j,k) = img%real_values(iii,jjj,kkk)
                    enddo
                enddo
            enddo
            win%rpointer => win%real_values
#ifndef SKIP_RUNTIME_CHECKS
        else
            if (img%is_in_real_space .and. img%origin_at_center) then
                img_origin = img%physical_address_of_box_center
            else
                img_origin = [1,1,1]
            endif
            !print *, '**debug(window_from_image_real): image dim, origin = ', img%logical_dimensions, img_origin
            !print *, '**debug(window_from_image_real): coos_min, coos_max = ', coos_min, coos_max
            !print *, '**debug(window_from_image_real): win%logical_dimensions = ', win%logical_dimensions
            !print *, '**debug(window_from_image_real): win%lbound = ', win%lbound
            do i=1,3
                lbnd(i) = max(coos_min(i),1+win%lbound(i)-1)
                lbnd(i) = lbnd(i) + img_origin(i)
                ubnd(i) = min(coos_max(i),win%logical_dimensions(i) + win%lbound(i) - 1)
                ubnd(i) = ubnd(i) + img_origin(i)
                if (lbnd(i) .eq. 1) then
                    ubnd(i) = lbnd(i) + win%logical_dimensions(i) - 1
                else if (ubnd(i) .eq. img%logical_dimensions(i)) then
                    lbnd(i) = ubnd(i) - win%logical_dimensions(i) + 1
                endif
            enddo
            if (img%logical_dimensions(3) .eq. 1) then
                lbnd(3) = 1
                ubnd(3) = 1
            endif
            !print *, '**debug(window_from_image_real): rpointer points to ', lbnd,ubnd
            win%rpointer => img%real_values(lbnd(1):ubnd(1),lbnd(2):ubnd(2),lbnd(3):ubnd(3))
        endif
#endif
    end subroutine window_from_image_real

    !>  \brief  window object is populated with data from image object over which it is hovering
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine window_from_image_comp(img,win)
        implicit none
        ! arguments
        type(image),    target, intent(in)      ::  img
        type(window),   intent(inout)   ::  win
        ! private variables
        integer ::  i,j,k
        integer ::  ii,jj,kk
        integer ::  iii,jjj,kkk
        integer ::  tmp_coos(3), tmp_coos_in(3)
        integer ::  coos_min(3),coos_max(3) !< the bounds of the image
        integer ::  lbnd(3), ubnd(3), img_origin(3)

        ! start work
        call bounds_from_center(img,coos_min,coos_max)      ! this should be passed to the routine, not recalculated here.


        ! in the complex case, we always have to use own_memory for a window, because it is not possible to use a pointer when we
        ! are near the frequency origin in the first dimension (which is "folded"). this is why there is 1==1 here.
#ifndef SKIP_RUNTIME_CHECKS
        if (win%own_memory .or. 1==1) then
#endif
            win%complex_values = (0.0,0.0)
            ! iterate over window voxels. i,j,k is address within window, ii,jj,kk is address within image (with origin 0,0,0 at center),
            ! and iii,jjj,kkk is address within image (with origin 1,1,1) at top-left
            do k=1,win%logical_dimensions(3)
                kk  = k + win%lbound(3) - 1
                if (kk .lt. coos_min(3) .or. kk .gt. coos_max(3)) cycle
                do j=1,win%logical_dimensions(2)
                    jj = j + win%lbound(2) - 1
                    if (jj .lt. coos_min(2) .or. jj .gt. coos_max(2)) cycle
                    do i=1,win%logical_dimensions(1)
                        ii = i + win%lbound(1) - 1
                        if (ii .lt. coos_min(1) .or. ii .gt. coos_max(1)) cycle
                        ! get address within complex image which corresponds to voxel coordinate
                        tmp_coos_in = (/ii,jj,kk/)
                        call comp_addr_phys_from_logi(img,tmp_coos_in,tmp_coos)
                        iii = tmp_coos(1)
                        jjj = tmp_coos(2)
                        kkk = tmp_coos(3)
                        ! now copy the voxel value over
                        if ( ii .ge. 0) then
                            win%complex_values(i,j,k) = img%complex_values(iii,jjj,kkk)
                        else
                            win%complex_values(i,j,k) = conjg(img%complex_values(iii,jjj,kkk))  ! this is the line that cannot be done with pointers
                        endif
                    enddo
                enddo
            enddo
            !win%cpointer => win%complex_values ! 25-june-2012 - this pointer should already be associated (by window_allocate)
#ifndef SKIP_RUNTIME_CHECKS
        !
        ! the complex case without own_memory does not work, because of the case where only part of the window is in positive frequencies in the first dimension
        !
        else
            if (img%is_in_real_space .and. img%origin_at_center) then
                img_origin = img%physical_address_of_box_center
            else
                img_origin = [1,1,1]
            endif
            !print *, '**debug(window_from_image_comp): image dim, origin = ', img%logical_dimensions, img_origin
            !print *, '**debug(window_from_image_comp): coos_min, coos_max = ', coos_min, coos_max
            !print *, '**debug(window_from_image_comp): win%logical_dimensions = ', win%logical_dimensions
            !print *, '**debug(window_from_image_comp): win%lbound = ', win%lbound
            lbnd = max(coos_min,1+win%lbound-1)
            lbnd = lbnd + img_origin
            ubnd = min(coos_max,win%logical_dimensions + win%lbound - 1)
            ubnd = ubnd + img_origin
            !print *, '**debug(window_from_image_comp): lbnd, ubnd = ', lbnd, ubnd
            do i=1,3
                if (lbnd(i) .eq. 1) then
                    ubnd(i) = lbnd(i) + win%logical_dimensions(i) - 1
                else if (ubnd(i) .eq. img%logical_dimensions(i)) then
                    lbnd(i) = ubnd(i) - win%logical_dimensions(i) + 1
                endif
            enddo
            if (img%logical_dimensions(3) .eq. 1) then
                lbnd(3) = 1
                ubnd(3) = 1
            endif
            !print *, '**debug(window_from_image_comp): rpointer points to ', lbnd,ubnd
            win%cpointer => img%complex_values(lbnd(1):ubnd(1),lbnd(2):ubnd(2),lbnd(3):ubnd(3)) ! this does not work when the window hovers over the origin in the first dimension
        endif
#endif
    end subroutine window_from_image_comp

    !>  \brief  window object data is copied into image object
    pure subroutine window_to_image_real(img,win,subtract)
        implicit none
        ! arguments
        type(image),                intent(inout)   ::  img         !<  image to extract data from
        type(window),               intent(in)      ::  win         !<  window to store the data
        logical,        optional,   intent(in)      ::  subtract    !<  if .true., the values in the window are subtracted from the image. .false. by default.
        ! private variables
        integer ::  i,j,k
        integer ::  ii,jj,kk
        integer ::  iii,jjj,kkk
        real    ::  sign
        ! start work
        if (present(subtract)) then
            if (subtract) then
                sign = -1.0
            else
                sign = 1.0
            endif
        else
            sign = 1.0
        endif

        !$omp   critical        (copy_window_to_volume_real_outer)
        if (win%own_memory) then
            ! iterate over window voxels. i,j,k is address within window, ii,jj,kk is address within image (with origin 0,0,0 at center),
            ! and iii,jjj,kkk is address within image (with origin 1,1,1) at top-left
            do k=1,win%logical_dimensions(3)
                kk  = k + win%lbound(3) - 1
                kkk = kk + img%logical_dimensions(3)/2 + 1
                if (img%logical_dimensions(3) .eq. 1) kkk = 1
                if (kkk .lt. 1 .or. kkk .gt. img%logical_dimensions(3)) cycle
                do j=1,win%logical_dimensions(2)
                    jj = j + win%lbound(2) - 1
                    jjj = jj + img%logical_dimensions(2)/2 + 1
                    if (jjj .lt. 1 .or. jjj .gt. img%logical_dimensions(2)) cycle
                    do i=1,win%logical_dimensions(1)
                        ii = i + win%lbound(1) - 1
                        iii = ii + img%logical_dimensions(1)/2 + 1
                        if (iii .lt. 1 .or. iii .gt. img%logical_dimensions(1)) cycle
                        ! copy voxel value over
                        img%real_values(iii,jjj,kkk) = img%real_values(iii,jjj,kkk) + win%rpointer(i,j,k) * sign
                    enddo
                enddo
            enddo
        else
            ! nothing to do
        endif
        !$omp   end critical    (copy_window_to_volume_real_outer)
    end subroutine window_to_image_real

    !>  \brief  window object data is copied into image object
    subroutine window_to_image_comp(img,win,subtract)
        implicit none
        ! arguments
        type(image),                intent(inout)       ::  img
        type(window),               intent(in)          ::  win
        logical,        optional,   intent(in)          ::  subtract    !<  if .true., the values in the window are subtracted from the image. .false. by default.
        ! private variables
        integer ::  i,j,k
        integer ::  ii,jj,kk
        integer ::  iii,jjj,kkk
        integer ::  tmp_coos(3), tmp_coos_in(3)
        integer ::  coos_min(3),coos_max(3) !< the bounds of the iamge with origin 0,0,0 at center
        real    ::  sign

        ! start work
        call bounds_from_center(img,coos_min,coos_max)      ! this should be passed to the routine, not recalculated here.

        ! are we subtracting or adding?
        if (present(subtract)) then
            if (subtract) then
                sign = -1.0
            else
                sign = 1.0
            endif
        else
            sign = 1.0
        endif

        !$omp   critical    (copy_window_to_volume_comp_outer)

        ! iterate over window voxels. i,j,k is address within window, ii,jj,kk is address within image (with origin 0,0,0 at center),
        ! and iii,jjj,kkk is address within image (with origin 1,1,1 at top-left)
        !write(*,*) '**debug: win%lbound =', win%lbound
        !write(*,*) '**debug: window%centered_on =', win%centered_on
        if (win%own_memory) then
            do k=1,win%logical_dimensions(3)
                kk  = k + win%lbound(3) - 1
                if (kk .lt. coos_min(3) .or. kk .gt. coos_max(3)) cycle
                do j=1,win%logical_dimensions(2)
                    jj = j + win%lbound(2) - 1
                    if (jj .lt. coos_min(2) .or. jj .gt. coos_max(2)) cycle
                    do i=1,win%logical_dimensions(1)
                        ii = i + win%lbound(1) - 1
                        if (ii .lt. coos_min(1) .or. ii .gt. coos_max(1)) cycle

                        ! get address within complex image which corresponds to voxel coordinate
                        tmp_coos_in = (/ii,jj,kk/)
                        !write(*,*) '**debug: coos =', tmp_coos
                        call comp_addr_phys_from_logi(img,tmp_coos_in,tmp_coos)
                        !write(*,*) '**debug: addr =', tmp_coos
                        iii = tmp_coos(1)
                        jjj = tmp_coos(2)
                        kkk = tmp_coos(3)

                        ! now copy the voxel value over
                        !write(*,'(a,2(f7.3,x))') '**debug(window_to_image_comp): cvalue =', win%complex_values(i,j,k)
                        !write(*,'(a,3(i3,x))') '**debug(window_to_image_comp): copying value over: iii,jjj,kkk =', iii,jjj,kkk
                        if (ii .ge. 0) then
                            img%complex_values(iii,jjj,kkk) = img%complex_values(iii,jjj,kkk) + win%cpointer(i,j,k) * sign
                        else
                            img%complex_values(iii,jjj,kkk) = img%complex_values(iii,jjj,kkk) + conjg(win%cpointer(i,j,k)) * sign
                        endif

                        ! at the origin of the first dimension, we have the unusual situation that friedel mates
                        ! are described explicitely, except for the nyquist components, which are only
                        ! described in the negative frequencies. therefore if a value lands anywhere on
                        ! the origin of the first dimension, we must make sure it is inserted a second time.
                        !if (ii .eq. 0 .and. jj .ne. 0 .and. (win%logical_dimensions(3) .eq. 1 .or. kk .ne. 0)) then
                        if (ii .eq. 0 .and. jj .ne. coos_min(2) .and. &
                                (win%logical_dimensions(3) .eq. 1 .or. kk .ne. coos_min(3))) then
                            ! need to also add a friedel/hermitian mate if we are on the origin in 1st dimension
                            tmp_coos_in = (/ii,-jj,-kk/)
                            call comp_addr_phys_from_logi(img,tmp_coos_in,tmp_coos)
                            iii = tmp_coos(1)
                            jjj = tmp_coos(2)
                            kkk = tmp_coos(3)
                            img%complex_values(iii,jjj,kkk) = img%complex_values(iii,jjj,kkk) + conjg(win%cpointer(i,j,k)) * sign
                        endif
                    enddo
                enddo
            enddo
        else
            ! nothing to do
        endif
        !$omp   end critical    (copy_window_to_volume_comp_outer)
    end subroutine window_to_image_comp

    !>  \brief  compute the lower and upper bound addresses of a window object. not the assumption that
    !!          central coordinates of window are given relative to origin (0,0,0) at center of image. also,
    !!          compute the coordinates of the window's center (i.e. where we're interpolating to / convoluting from),
    !!          with origin (1,1,1) at the top-left corner voxel.
    pure subroutine eval_window_bounds(win)
        implicit none
        ! arguments
        class(window),   intent(inout)   ::  win !<  window object
        ! private variables
        integer ::  l
        ! start work

        ! define the bounds of the window
        ! (note the window width is 2 for linear, .ge. 2 for sinc)
        ! these bounds can take any value, including 0, which means the address of the origin,
        ! and negative values
        do l=1,3
            win%lbound(l) = floor  (win%centered_on(l)) - (win%logical_dimensions(l)/2 - 1)
            win%ubound(l) = ceiling(win%centered_on(l)) + (win%logical_dimensions(l)/2 - 1)
            ! there is a special case in which the output coordinates are exactly on an output voxel. in this case, the center is
            ! at n/2+1
            if ( win%centered_on(l) .eq. anint(win%centered_on(l))) then ! we've fallen bang on a voxel coordinate
                win%lbound(l) = int(win%centered_on(l)) - (win%logical_dimensions(l)/2)
                win%ubound(l) = int(win%centered_on(l)) + (win%logical_dimensions(l)/2 - 1)
            endif
            ! now work out the coordinates of the "center"
            win%center(l) = win%centered_on(l) - real(win%lbound(l)) + 1.0
        enddo
        if (win%logical_dimensions(3) .eq. 1) then
            win%ubound(3) = 0
            win%lbound(3) = 0
            win%center(3) = 1.0
        endif

    end subroutine eval_window_bounds

    !>  \brief  deallocate memory assigned to a window object
    subroutine deallocate_window(self)
        use fftw33
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        type(window),   intent(inout)   ::  self
        ! private variables
        logical,        parameter       ::  debug   =   .false.
        ! start work
        if (debug)  write(*,'(a)') '**debug(deallocate_window): entering'
        if (self%in_memory) then
            if (self%own_memory) then
                if (associated(self%rpointer)) nullify(self%rpointer)
                if (associated(self%cpointer)) nullify(self%cpointer)
                if (associated(self%real_values)) nullify(self%real_values)
                if (associated(self%complex_values)) nullify(self%complex_values)
                !$omp   critical        (fftw_omp_crit)
                call fftwf_free(self%p)
                self%in_memory=.false.
                !$omp   end critical    (fftw_omp_crit)
            endif
            self%in_memory=.false.
        else
            !write(*,'(a)') '**info(deallocate_window): attempt at deallocating window which was not allocated'
            !call terminate('test deallocate_window')
        endif
        if (debug)  write(*,'(a)') '**debug(deallocate_window): exiting'
    end subroutine deallocate_window

    !>  \brief  allocate memory to the window object
    subroutine allocate_window(win,width,comp,threed,own_memory)
        use iso_c_binding
        use fftw33
        use runtime_parameters, only : is_odd, terminate
        implicit none
        ! arguments
        class(window),  target, intent(inout)   ::  win         !<  the window to allocate memory to
        integer,                intent(in)      ::  width       !<  window width in pixels
        logical,                intent(in)      ::  comp        !<  is it complex or real?
        logical,                intent(in)      ::  threed      !<  3-dimensional?
        logical,    optional,   intent(in)      ::  own_memory  !<  whether the window should have its own area in memory.
        ! private variables
        integer ::  alloc_dim(3)
        ! start work

        ! copy type over from image
        if (comp) then
            win%is_in_real_space = .false.
        else
            win%is_in_real_space = .true.
        endif

        ! will the window have its own area in memory?
        if (present(own_memory)) then
            win%own_memory = own_memory
        else
            if (win%is_in_real_space) then
                win%own_memory = .false.
            else
                win%own_memory = .true.
            endif
        endif


        ! own memory is compulsory for comp windows
        if (.not. win%is_in_real_space) then
            win%own_memory = .true.
        endif

        ! own memory is compulsory if we are skipping runtime checks, to make things faster
#ifdef SKIP_RUNTIME_CHECKS
        win%own_memory = .true.
#endif

        ! own memory is compulsory for comp windows
        ! if the window is complex, we cannot do the same tricks with pointers that we can with real, because of the "folding" of the
        ! first dimension, which cannot be accomodated by window_from_image_comp
        if (.not. win%own_memory .and. .not. win%is_in_real_space) then
            write(*,'(a)') '**error(allocate_window): complex windows must have their own memory'
            call terminate('fatal error in allocate_window')
        endif

        ! if window already allocated
        if (win%own_memory .and. win%in_memory) then
            write(*,'(a)') '**error(allocate_window): window already in memory'
            call terminate('fatal error in allocate_window')
        endif

        ! check width is even and positive
        if (is_odd(width) .or. width .le. 0) then
            write(*,'(a,i0)') '**error(allocate_window): window width must be even positive: ', width
            call terminate('fatal error in allocate_window')
        endif

        ! set dimensions of window
        win%width       =   width
        win%logical_dimensions(1:3)    =   width
        if (.not. threed) win%logical_dimensions(3) = 1

        ! work out memory allocation size (we will allocate complex array)
        alloc_dim = win%logical_dimensions
        if (win%is_in_real_space) then
            alloc_dim(1) = alloc_dim(1) / 2
        endif
        if (win%logical_dimensions(3) .eq. 1) alloc_dim(3) = 1

        ! allocate memory if necessary
        if (win%own_memory) then
            ! do the allocation
            !$omp   critical        (fftw_omp_crit)
            !win%p = fftwf_alloc_complex(int(product(alloc_dim),c_size_t))
            win%p = fftwf_malloc(int(product(alloc_dim)*8,c_size_t))
            win%in_memory = .true.
            !$omp   end critical    (fftw_omp_crit)
            call c_f_pointer(win%p,win%complex_values,alloc_dim)
            alloc_dim(1) = alloc_dim(1) * 2
            call c_f_pointer(win%p,win%real_values,alloc_dim)

            ! finally, initialise the contents
            win%real_values = 0.0

            ! and setup the pointers
            win%rpointer => win%real_values
            win%cpointer => win%complex_values
        else
            ! the window does not have its own memory. rather,
            ! when it comes to hovering over an image/volume,
            ! the r and c pointers will be associated with sections
            ! of the image/volume's data arrays, thus avoiding
            ! copying of data from image to window
        endif
        win%in_memory = .true.
    end subroutine allocate_window


    !>  \brief  compute the coordinate along the second or third axis of an fted image, given a physical index along that axis. result is in 1/pixels, from -0.5 to 0.5
    pure function comp_coo_from_phys_index_dim1(dim,index) result(coo)
        implicit none
        ! arguments
        integer,    intent(in)  ::  dim     !<  length of image along the approriate axis
        integer,    intent(in)  ::  index   !<  logical index of voxel along the axis
        ! result
        real                    ::  coo     !<  coordinate (1/pixels)
        !

        ! start work
        coo  = comp_coo_from_logi_index(dim,index-1)
    end function comp_coo_from_phys_index_dim1

    !>  \brief  compute the coordinate along the second or third axis of an fted image, given a physical index along that axis. result is in 1/pixels, from -0.5 to 0.5
    pure function comp_coo_from_phys_index_dim2or3(dim,index) result(coo)
        implicit none
        ! arguments
        integer,    intent(in)  ::  dim     !<  length of image along the approriate axis
        integer,    intent(in)  ::  index   !<  logical index of voxel along the axis
        ! result
        real                    ::  coo     !<  coordinate (1/pixels)
        !

        ! start work
        coo  = comp_coo_from_logi_index(dim,comp_index_logi_from_phys(dim,index))
    end function comp_coo_from_phys_index_dim2or3

    !>  \brief  compute the coordinate along an axis, given a logical index along that axis. result is in 1/pixels, from -0.5 to 0.5.
    pure function comp_coo_from_logi_index(dim,index) result(coo)
        implicit none
        ! arguments
        integer,    intent(in)  ::  dim     !<  length of image along the approriate axis
        integer,    intent(in)  ::  index   !<  logical index of voxel along the axis
        ! result
        real                    ::  coo     !<  coordinate (1/pixels)
        ! private variables

        ! start work
        coo = real(index) / real(dim)
    end function comp_coo_from_logi_index

    pure subroutine update_physical_address_of_box_center(self)
        ! arguments
        type(image),    intent(inout)   ::  self
        ! private variables
        integer ::  i
        ! start work
        do i=1,3
            self%physical_address_of_box_center(i) = index_of_central_pixel_from_dim(self%logical_dimensions(i))
        enddo
    end subroutine update_physical_address_of_box_center

!    !>  \brief  returns the physical address of the image origin
!    !!  \todo   ensure this is called in all relevant subroutines, which should help smooth behaviour with even/odd dimensions
!    pure function address_of_origin_deprecate(self) result (address_of_origin)
!        use runtime_parameters, only : is_even
!        implicit none
!        ! arguments
!        type(image),   intent(in)  ::  self
!        ! result
!        integer ::  address_of_origin(3)
!        ! private variables
!        integer ::  i
!        ! start work
!        if (self%origin_at_center .and. self%is_in_real_space) then
!            do i=1,3
!                address_of_origin(i) = index_of_central_pixel_from_dim(self%logical_dimensions(i))
!            enddo
!        else
!            address_of_origin = [1,1,1]
!        endif
!    end function address_of_origin_deprecate

    !>  \brief given a logical dimension of an image, return the index for the central pixel
    pure elemental function index_of_central_pixel_from_dim(dim) result(i)
        use runtime_parameters, only : is_even
        implicit none
        ! arguments
        integer,    intent(in)  ::  dim     !<  logical dimension
        ! results
        integer                 ::  i       !<  index of central voxel
        ! start work
        if (is_even(dim)) then
            ! dimension is even
            i = dim/2+1
        else
            ! dimension is odd
            i = (dim-1)/2+1
        endif
    end function index_of_central_pixel_from_dim

!    !>  \brief  returns the physical address of a voxel given its logical address.
!    pure function image_address_phys_from_logi(self,logi) result (phys)
!        implicit none
!        ! arguments
!        type(image),    intent(in)  ::  self    !<  image object
!        integer,        intent(in)  ::  logi(3) !<  logical address of voxel
!        ! result
!        integer                     ::  phys(3) !<  physical address of voxel
!        ! start work
!        if (self%is_in_real_space) then
!            ! image is real
!            phys = real_addr_phys_from_logi(self,logi)
!        else
!            ! image is complex
!            call comp_addr_phys_from_logi(self,logi,phys)
!        endif
!    end function image_address_phys_from_logi
!
!    pure function real_addr_phys_from_logi(img,logi) result(phys)
!        implicit none
!        ! arguments
!        type(image),    intent(in)  ::  img
!        integer,        intent(in)  ::  logi(3)
!        ! result
!        integer                     ::  phys(3)
!        ! start work
!        phys = address_of_origin(img) + logi
!    end function real_addr_phys_from_logi

!    !>  \brief  return the coordinates of the voxel with given address
!    pure function real_addr_logi_from_phys(img,phys) result(logi)
!        implicit none
!        ! arguments
!        type(image),    intent(in)  ::  img
!        integer,        intent(in)  ::  phys(3)
!        ! result
!        integer                     ::  logi(3)
!        ! start work
!        logi = phys - address_of_origin(img)
!    end function real_addr_logi_from_phys



!    !>  \brief  return the physical addresses of the voxel with given logical address
!    !!          as well as its 7 neighbours
!    !!
!    !!  this is useful when we have the logical address of the bottom left corner of the cube of
!    pure function comp_addr_phys_cube_from_logi(img,logi) result(phys)
!        implicit none
!        type(image),    intent(in)  ::  img
!        integer,        intent(in)  ::  logi(3)
!        integer                     ::  phys(3,8)
!!        !
!        phys = 0
!!        if (logi(1) .ge. 0) then
!!            ! first dimension
!!            phys(1) = logi(1) + 1
!!            ! second & third dimensions
!!            do i=2,3
!!                if (logi(i) .ge. 0) then
!!                    phys(i) = logi(i) + 1
!!                else
!!                    phys(i) = img%logical_dimensions(i) + 1 + logi(i)
!!                endif
!!            enddo
!!        else
!!            ! first dimension
!!            phys(1) = -logi(1) + 1
!!            ! second & third dimensions
!!            do i=2,3
!!                if (logi(i) .gt. 0) then
!!                    phys(i) = img%logical_dimensions(i) - logi(i) + 1
!!                else
!!                    phys(i) = - logi(i) + 1
!!                endif
!!            enddo
!!        endif
!!        if (img%logical_dimensions(3) .eq. 1) phys(3) = 1
!    end function

    pure integer function phys_from_logi_dim1(logi)
        implicit none
        ! arguments
        integer,    intent(in)  ::  logi
        ! start work
        if (logi .ge. 0) then
            phys_from_logi_dim1 = logi+1
        else
            phys_from_logi_dim1 =-logi+1
        endif
    end function phys_from_logi_dim1



    !>  \brief  returns the physical address of voxel with given logical address.
    !!
    !!          the address returned is for a complex image with origin at address (1,1,1),
    !!          dimension 1 from 1 to dim(1)/2+1 (division by 2 rounded down), and other
    !!          2 dimensions from 1 to dim, folded halfway, the fftw way.
    pure subroutine comp_addr_phys_from_logi(img,logi,phys)
        implicit none
        ! arguments
        type(image),    intent(in)  ::  img
        integer,        intent(in)  ::  logi(3)
        ! result
        integer,        intent(out) ::  phys(3)
        ! private variables
        integer ::  i
        ! start work
        if (logi(1) .ge. 0) then
            ! first dimension
            phys(1) = logi(1) + 1
            ! second & third dimensions
            do i=2,3
                if (logi(i) .ge. 0) then
                    phys(i) = logi(i) + 1
                else
                    phys(i) = img%logical_dimensions(i) + 1 + logi(i)
                endif
            enddo
        else
            ! first dimension
            phys(1) = -logi(1) + 1
            ! second & third dimensions
            do i=2,3
                if (logi(i) .gt. 0) then
                    phys(i) = img%logical_dimensions(i) - logi(i) + 1
                else
                    phys(i) = - logi(i) + 1
                endif
            enddo
        endif
        if (img%logical_dimensions(3) .eq. 1) phys(3) = 1
    end subroutine comp_addr_phys_from_logi

    !>  \brief  get the logical address within a complex image (result of ft of real image), given a physical address
    pure function comp_addr_logi_from_phys(img,phys) result(logi)
        implicit none
        ! arguments
        type(image),    intent(in)  ::  img         !<  image object
        integer,        intent(in)  ::  phys(3)     !<  physical address of voxel
        ! result
        integer                     ::  logi(3)     !<  logical address of voxel
        ! private variables
        integer     ::  i
        ! start work
        !
        logi(1) = phys(1) - 1
        do i=2,3
            logi(i) = comp_index_logi_from_phys(img%logical_dimensions(i),phys(i))
        enddo
    end function comp_addr_logi_from_phys

    !>  \brief  compute the logical index of a voxel given its physical index and the logical size of the image. only useful for 2nd and 3rd dimension.
    pure function comp_index_logi_from_phys(size,phys) result(logi)
        implicit none
        ! arguments
        integer,        intent(in)  ::  size    !<  logical dimension of image/array
        integer,        intent(in)  ::  phys    !<  physical index of voxel
        ! result
        integer                     ::  logi    !<  logical index of voxel
        ! start work
        if (phys .ge. lowest_index_for_negative_frequencies(size)) then
            ! negative frequency
            logi = phys - (size + 1)
        else
            ! positive frequency
            logi = phys - 1
        endif
    end function comp_index_logi_from_phys

    !>  \brief  the maximum address along the first dimension of a complex image
    pure function maximum_complex_address_dim1(self) result (i)
        use runtime_parameters, only : is_even
        implicit none
        ! arguments
        type(image),    intent(in)  ::  self
        ! result
        integer                     ::  i
        ! start work
        if (is_even(self%logical_dimensions(1))) then
            i = self%logical_dimensions(1)/2 + 1
        else
            i = (self%logical_dimensions(1)-1)/2 + 1
        endif
    end function maximum_complex_address_dim1

    !>  \brief  given the logical dimension of a complex image, find the index on that dimension of the first voxel that stores negative frequencies
    pure function lowest_index_for_negative_frequencies(dim) result(i)
        use runtime_parameters, only : is_even
        implicit none
        ! arguments
        integer,    intent(in)  ::  dim !<  logical dimension of image
        ! result
        integer                 ::  i
        ! private variables

        ! start work
        if (is_even(dim)) then
            i = dim/2 + 1
        else
            i = (dim+3)/2
        endif
    end function lowest_index_for_negative_frequencies

    !>  \brief  return the index of the nearest origin given an input index and the length along the appropriate dimension.
    !!          if the index corresponds to a positive frequency, the nearest origin is 1, otherwise it is dim + 1.
    !!          note that this is only useful for cases where origin is in the top left corner (i.e. not origin_at_center).
    !!          this function assumes that we are in a complex ft image, where the first dimension is half of the others. do not use this on real images with origin at (1,1,1).
    pure integer function index_nearest_origin(i,dim)
        use runtime_parameters, only : is_even
        implicit none
        ! arguments
        integer, intent(in) :: i    !< input index
        integer, intent(in) :: dim  !< image length along appropriate dimension
        ! start work
        if (i .ge. lowest_index_for_negative_frequencies(dim)) then
            index_nearest_origin = dim + 1 ! of course that point wraps around to the actual origin, which is at 1.
        else
            index_nearest_origin = 1
        endif
    end function index_nearest_origin


    !>  \brief  define the minimum and maximum logical address in each direction
    !!          (by definition, origin has logical address (0,0,0))
    pure subroutine bounds_from_center(img,lbounds,ubounds)
        use runtime_parameters, only : is_even
        implicit none
        ! arguments
        type(image),    intent(in)  ::  img         !<  input image
        integer,        intent(out) ::  lbounds(3)  !<  minimum logical address along each direction
        integer,        intent(out) ::  ubounds(3)  !<  maximum logical address along each direction
        ! private variables
        integer ::  i
        ! start work
        if (img%is_in_real_space) then
            do i=1,3
                if (is_even(img%logical_dimensions(i))) then
                    ! even dimension
                    lbounds(i) = -img%logical_dimensions(i)/2
                    ubounds(i) = img%logical_dimensions(i)/2-1
                else
                    ! odd dimension
                    lbounds(i) = -(img%logical_dimensions(i)-1)/2
                    ubounds(i) = (img%logical_dimensions(i)-1)/2
                endif
            enddo
        else
            ! complex image
            if (is_even(img%logical_dimensions(1))) then
                ! even first dimension
                lbounds(1) = -(img%logical_dimensions(1)/2)
                ubounds(1) = img%logical_dimensions(1)/2
            else
                ! odd first dimension
                lbounds(1) = -(img%logical_dimensions(1)-1)/2
                ubounds(1) = (img%logical_dimensions(1)-1)/2
            endif
            do i=2,3
                if (is_even(img%logical_dimensions(i))) then
                    ! even dimension
                    lbounds(i) = -img%logical_dimensions(i)/2
                    ubounds(i) = img%logical_dimensions(i)/2-1
                else
                    ! odd dimension
                    lbounds(i) = -(img%logical_dimensions(i)-1)/2
                    ubounds(i) = (img%logical_dimensions(i)-1)/2
                endif
            enddo
        endif
        if (img%logical_dimensions(3) .eq. 1) then
            lbounds(3) = 0
            ubounds(3) = 0
        endif
    end subroutine bounds_from_center

    !>  \brief  given an address, return the address of the voxel diametrically opposed
    subroutine addr_friedel_mate(self,i,j,k)
        implicit none
        ! arguments
        type(image),    intent(in)      ::  self
        integer,        intent(inout)   ::  i,j,k
        ! private variables
        integer ::  icoo,jcoo,kcoo
        integer ::  ori(3)
        ! start work
        if (self%is_in_real_space .and. self%origin_at_center) then
            ori = self%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif
        ! coordinates (relative to origin)
        icoo = i - ori(1)
        jcoo = j - ori(2)
        kcoo = k - ori(3)
        ! the friedel mate is diametrically opposed
        icoo = -icoo
        jcoo = -jcoo
        kcoo = -kcoo
        ! address
        i = icoo + ori(1)
        j = jcoo + ori(2)
        k = kcoo + ori(3)
    end subroutine addr_friedel_mate



    !
    !==== the routines below have not been written to take odd dimensions and non-cubic volumes ====!
    !       and/or they do not adhere to proper nomenclature
    !

    !>  \brief  returns the shell to which the voxel with address i,j,k belongs
    pure integer function shell(img,i,j,k,do_not_fold)
        implicit none
        ! arguments
        class(image),               intent(in)  ::  img
        integer,                    intent(in)  ::  i,j,k
        logical,        optional,   intent(in)  ::  do_not_fold(3)  !<  when dealing with corner origins, specifies if a dimension should behave as though it only had one origin
        ! start work
        if (present(do_not_fold)) then
            shell = nint(radius(img,i,j,k,do_not_fold))
        else
            shell = nint(radius(img,i,j,k))
        endif
    end function shell

    !>  \brief return the radius from the origin to the voxel with address i,j,k
    pure real function radius(img,i,j,k,do_not_fold)
        implicit none
        ! arguments
        class(image),               intent(in)  ::  img
        integer,                    intent(in)  ::  i,j,k
        logical,        optional,   intent(in)  ::  do_not_fold(3)  !<  when dealing with corner origins, specifies if a dimension should behave as though it only had one origin
        ! start work
        radius = sqrt( radius_sq(i,j,k,img,do_not_fold) )
    end function radius

    !>  \brief  returns the square of the radius from origin to the voxel with address i,j,k
    pure real function radius_sq(i,j,k,img,do_not_fold)
        implicit none
        ! arguments
        integer,                    intent(in)  ::  i,j,k
        type(image),                intent(in)  ::  img
        logical,        optional,   intent(in)  ::  do_not_fold(3)  !<  when dealing with corner origins, specifies if a dimension should behave as though it only had one origin
        ! private variables
        integer ::  mid(3)
        logical ::  ddo_not_fold(3)
        ! start work

        ! check do_not_fold
        if (present(do_not_fold)) then
            ddo_not_fold = do_not_fold
        else
            ddo_not_fold = (/.false.,.false.,.false./)
        endif

        if (img%origin_at_center .and. img%is_in_real_space) then
            mid = img%physical_address_of_box_center
        else
            ! dimension 1
            if (ddo_not_fold(1)) then
                mid(1) = 1
            else
                mid(1) = nearest_orig(i,img%logical_dimensions(1))
            endif
            ! dimension 2
            if (ddo_not_fold(2)) then
                mid(2) = 1
            else
                mid(2) = nearest_orig(j,img%logical_dimensions(2))
            endif
            ! dimension 3
            if (ddo_not_fold(3)) then
                mid(3) = 1
            else
                mid(3) = nearest_orig(k,img%logical_dimensions(3))
            endif
        endif
        radius_sq = real((i - mid(1))**2 + (j - mid(2))**2 + (k - mid(3))**2)
    end function radius_sq



    !>  \brief  return the radius as a fraction of maximum inner radius
    !!  \todo   it looks like this could be optimised - perhaps call radius function?
    real function radius_fraction(i,j,k,img,do_not_fold)
        implicit none
        ! arguments
        integer,                intent(in)  ::  i,j,k           !<  address to calculate the radius for
        type(image),            intent(in)  ::  img             !<  image object, provides dimensions & origin
        logical,    optional,   intent(in)  ::  do_not_fold(3)  !<  when dealing with corner origins, specifies if a dimension should behave as though it only had one origin
        ! private variable
        integer ::  addr_origin(3)  !<  address of the origin
        integer ::  max_rad(3)      !<  maximum inner radius in each dimension
        integer ::  current_dim
        logical ::  ddo_not_fold(3)
        ! start work

        ! check do_not_fold
        if (present(do_not_fold)) then
            ddo_not_fold = do_not_fold
        else
            ddo_not_fold = (/.false.,.false.,.false./)
        endif

        ! address of origin
        if (img%origin_at_center .and. img%is_in_real_space) then
            addr_origin = img%physical_address_of_box_center
        else
            ! dimension 1
            if (ddo_not_fold(1)) then
                addr_origin(1) = 1
            else
                addr_origin(1) = nearest_orig(i,img%logical_dimensions(1))
            endif
            ! dimension 2
            if (ddo_not_fold(2)) then
                addr_origin(2) = 1
            else
                addr_origin(2) = nearest_orig(j,img%logical_dimensions(2))
            endif
            ! dimension 3
            if (ddo_not_fold(3)) then
                addr_origin(3) = 1
            else
                addr_origin(3) = nearest_orig(k,img%logical_dimensions(3))
            endif
        endif
        if (img%logical_dimensions(3) .eq. 1) addr_origin(3) = 1 ! if we are dealing with a plane, not volume

        ! maximum radius calculation
        if (img%origin_at_center .and. img%is_in_real_space) then
            max_rad(1:3) = (addr_origin(1:3) - 1)**2
        else
            do current_dim=1,3
                max_rad(current_dim) = real(int(img%logical_dimensions(current_dim)/2))**2
            enddo
        endif
        if (img%logical_dimensions(3) .eq. 1) max_rad(3) = 1

        radius_fraction = sqrt(     &
                                real((i-addr_origin(1))**2)/max_rad(1) +    &
                                real((j-addr_origin(2))**2)/max_rad(2) +    &
                                real((k-addr_origin(3))**2)/max_rad(3)      &
                                )
        ! check for bad values in the 2d case
        !if (radius_fraction .lt. 0.0 .or. radius_fraction .gt. 1.42) then
        !   print *, '**error(radius_fraction): oops, result is wrong: ', radius_fraction
        !   call terminate('fatal error in radius_fraction')
        !endif
    end function radius_fraction

    !>  \brief returns the angle of coordinate vector. 0.0 means 3 o'clock. positive is ccw.
    !!
    !!  warning: this function has only been tested on 2d images - please remove this when tested in 3d
    pure function angle(i,j,k,img)
        implicit none
        ! arguments
        integer,        intent(in)  ::  i,j,k       !<  input address
        type(image),    intent(in)  ::  img
        ! result
        real                        ::  angle(2)    !<  angle (radians). first angle is from x axis to y coordinate, second is from x axis to z coordinate. ccw positive.
        ! private variables
        integer                     ::  mid(3)
        integer                     ::  x, y, z
        ! start work
        if (img%origin_at_center .and. img%is_in_real_space) then
            mid = img%physical_address_of_box_center
        else
            mid(1) = 1
            mid(2) = nearest_orig(j,img%logical_dimensions(2))
            mid(3) = nearest_orig(k,img%logical_dimensions(3))
        endif
        if (img%logical_dimensions(3) .eq. 1) mid(3) = 1
        x = i - mid(1)
        y = j - mid(2)
        z = k - mid(3)
        angle(1) = atan2(real(y),real(x))
        angle(2) = atan2(real(z),real(x))
    end function angle

    !>  \brief  work out the number of shells to the nyquist
    !!  \todo   add option to work out the number of shells to the corners of the cube?
    pure integer function number_of_shells(self)
        implicit none
        ! arguments
        type(image),    intent(in)  ::  self
        ! private varibles

        ! start work
        if (has_even_dims(self)) then
            number_of_shells = self%logical_dimensions(1)/2+1   ! including the origin voxel, which is like shell #0
        else
            number_of_shells = (self%logical_dimensions(1)-1)/2 + 1
        endif

    end function number_of_shells



    !> \brief   apply a forward ft to the image object. the ft is scaled.
    !!          the dc component is at (img%logical_dimensions(1)/2+1,img%logical_dimensions(2)/2+1,img%logical_dimensions(3)/2+1) (assuming even dimensions) or at (1,1,1) by default.
    !!
    !!
    !! for details on fftw, see http://www.fftw.org/
    !! a helpful page for understanding the output format: http://www.dsprelated.com/showmessage/102675/1.php
    !! \todo use wisdom: http://www.fftw.org/fftw2_doc/fftw_2.html#sec13
    !! \todo use openmp: http://www.fftw.org/fftw2_doc/fftw_4.html#sec48
    !!  \todo   check this: http://objectmix.com/fortran/371439-ifort-openmp-fftw-problem.html
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine fwd_ft(img,scale)
#ifndef SKIP_RUNTIME_CHECKS
        use runtime_parameters, only : terminate
#endif
        implicit none
        ! arguments
        class(image), intent(inout) :: img       !<  image object to ft
        logical, intent(in), optional :: scale  !<  whether to scale the values so that a bwd_ft of the output would yield the original image back, rather than a scaled version.
        ! private variables
        logical                 ::  sscale
        ! start work

#ifndef SKIP_RUNTIME_CHECKS
        ! check the image is in memory
        if (.not. img%in_memory) then
            write(*,'(a)') '**error(fwd_ft): image not in memory'
            call terminate('fatal error in fwd_ft')
        endif

        ! check that the image is currently real
        if (.not. img%is_in_real_space) then
            write(*,'(a)') ' **error(fwd_ft): fwd_ft called on a non-real image'
            call terminate('fatal error in fwd_ft')
        endif
#endif

        ! scale?
        sscale = .true.
        if (present(scale)) sscale = scale

        ! do the actual ft
        call fftwf_execute_dft_r2c(img%plan_fwd,img%real_values,img%complex_values)

        ! scale the fourier transform
        if (sscale) then
            !img%complex_values = img%complex_values / sqrt(real(img%logical_dimensions(1) * img%logical_dimensions(2)* img%logical_dimensions(3)))
            img%complex_values = img%complex_values / &
                                 real(img%logical_dimensions(1) * img%logical_dimensions(2)* img%logical_dimensions(3))
        endif

        ! set the image type
        img%is_in_real_space = .false.

    end subroutine fwd_ft

    !>  \brief apply a backward ft to the image object.
    !!
#ifdef SKIP_RUNTIME_CHECKS
    pure &
#endif
    subroutine bwd_ft(img,scale)
#ifndef SKIP_RUNTIME_CHECKS
        use runtime_parameters, only : terminate
#endif
        use omp_lib

        implicit none

        class(image), intent(inout) :: img
        logical, intent(in), optional :: scale  !<  whether to scale the values so that a bwd_ft of the output would yield the original image back, rather than a scaled version.

        logical :: sscale

#ifndef SKIP_RUNTIME_CHECKS
        ! check the image is in memory
        if (.not. img%in_memory) then
            write(*,'(a)') '**error(bwd_ft): image not in memory'
            call terminate('fatal error in bwd_ft')
        endif

        ! check that the image is currently complex
        if (img%is_in_real_space) then
            write(*,'(a)') ' **error(bwd_ft): bwd_ft called on a non-complex image'
            call terminate('fatal error in bwd_ft')
        endif
#endif

        ! scale?
        sscale = .true.
        if (present(scale)) sscale = scale


        ! do the actual ft
        call fftwf_execute_dft_c2r(img%plan_bwd,img%complex_values,img%real_values)


        ! scale the output
        if (sscale) then
            !img%complex_values = img%complex_values / sqrt(real(img%logical_dimensions(1) * img%logical_dimensions(2)* img%logical_dimensions(3)))
        endif

        ! set the image type
        img%is_in_real_space = .true.

    end subroutine bwd_ft



    !>  \brief  initialise the values of an image object. useful when image objects are private in openmp parallel
    !!          regions, for example.
    pure subroutine image_init(self)
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  self        !<  image object
        ! private variables

        ! start work
        self%logical_dimensions                = (/1,1,1/)
        self%is_in_real_space   = .true.
        self%origin_at_center   = .true.
        !
        self%p                  = c_null_ptr
        self%real_values             => null()
        self%complex_values             => null()
        self%in_memory          = .false.
        !
        self%plan_fwd           = c_null_ptr
        self%plan_bwd           = c_null_ptr
        self%planned            = .false.
    end subroutine image_init

    !>  \brief  get a 2d slice from a 3d volume along one of the axes
    subroutine image_get_slice(self,slice,axis,address)
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        type(image),    intent(in)      ::  self        !<  3d image
        type(image),    intent(inout)   ::  slice       !<  2d image
        integer,        intent(in)      ::  axis        !<  a slice perpendicular to this axis will be extracted
        integer,        intent(in)      ::  address     !<  the index of the slice along the chosen axis
        ! private variables

        ! start work

        ! check the input image is 3d
        if (.not. is_threed(self)) then
            write(*,'(a,3(i0,1x))') '**error(image_get_slice): input image is not 3d: ', self%logical_dimensions
            call terminate('fatal error in image_get_slice')
        endif

        ! check axis number is 1, 2 or 3
        if (axis .lt. 1 .or. axis .gt. 3) then
            write(*,'(a,i0)') '**error(image_get_slice): axis number out of range (1-3): ', axis
            call terminate('fatal error in image_get_slice')
        endif

        ! deallocate output image if necessary
        call image_deallocate(slice)

        ! set output image dimensions
        select case (axis)
            case (1)
                slice%logical_dimensions(1) = self%logical_dimensions(2)
                slice%logical_dimensions(2) = self%logical_dimensions(3)
            case (2)
                slice%logical_dimensions(1) = self%logical_dimensions(1)
                slice%logical_dimensions(2) = self%logical_dimensions(3)
            case (3)
                slice%logical_dimensions(1) = self%logical_dimensions(1)
                slice%logical_dimensions(2) = self%logical_dimensions(2)
        end select

        ! allocate output image
        call image_allocate(slice)

        ! copy pixel data
        select case(axis)
            case (1)
                slice%real_values(1:slice%logical_dimensions(1),1:slice%logical_dimensions(2),1) = &
                    self%real_values(address,1:self%logical_dimensions(2),1:self%logical_dimensions(3))
            case (2)
                slice%real_values(1:slice%logical_dimensions(1),1:slice%logical_dimensions(2),1) = &
                    self%real_values(1:self%logical_dimensions(1),address,1:self%logical_dimensions(3))
            case (3)
                slice%real_values(1:slice%logical_dimensions(1),1:slice%logical_dimensions(2),1) = &
                    self%real_values(1:self%logical_dimensions(1),1:self%logical_dimensions(2),address)
        end select
    end subroutine image_get_slice

    !>  \brief  print info about the image object. used mostly for debugging.
    subroutine image_print_info(self)
        use fftw33, only : fftwf_print_plan
        implicit none
        ! arguments
        type(image),    intent(in)      ::  self
        ! private variables
        character(len=1),   parameter   ::  tab = '2'
        character(len=2),   parameter   ::  text_width = '50'
        logical,            parameter   ::  print_fftw_plans = .false.
        logical,            parameter   ::  print_rvalue_stats = .true.
        ! start work
        write(*,'(a)')                                      ' '
        write(*,'(a)')                                      'image info:'
        write(*,'('//tab//'x,a'//text_width//',3(i0,1x))')  'logical dimensions: ', self%logical_dimensions
        write(*,'('//tab//'x,a'//text_width//',l1)')        'in real space?: ', self%is_in_real_space
        write(*,'('//tab//'x,a'//text_width//',l1)')        'origin at center? ', self%origin_at_center
        write(*,'('//tab//'x,a'//text_width//',2l1)')       'real, complex arrays allocated, associated? ', &
                                                                                        associated(self%real_values), &
                                                                                        associated(self%complex_values)
        if (associated(self%real_values)) then
        write(*,'('//tab//'x,a'//text_width//',3(i0,1x))')  'real array dimensions: ',  size(self%real_values,dim=1), &
                                                                                        size(self%real_values,dim=2), &
                                                                                        size(self%real_values,dim=3)
        endif
        if (associated(self%complex_values)) then
        write(*,'('//tab//'x,a'//text_width//',3(i0,1x))')  'comp array dimensions: ',  size(self%complex_values,dim=1), &
                                                                                        size(self%complex_values,dim=2), &
                                                                                        size(self%complex_values,dim=3)
        endif
        write(*,'('//tab//'x,a'//text_width//',l1)')        'image in memory? ',        self%in_memory
        write(*,'('//tab//'x,a'//text_width//',l1)')        'fftw plans ready? ',       self%planned
        if (print_fftw_plans) then
            write(*,'(a)') 'forward plan:'
            call fftwf_print_plan(self%plan_fwd)
            call flush(6)
            write(*,'(a)') 'backward plan:'
            call fftwf_print_plan(self%plan_bwd)
            call flush(6)
        endif
        if (print_rvalue_stats .and. is_real(self)) then
            write(*,'('//tab//'x,a'//text_width//',2(f0.3,1x))')  'real array min, max values: ', image_min_val(self), &
                                                                                                            image_max_val(self)
            write(*,'('//tab//'x,a'//text_width//',2(f0.3,1x))')  'real array mean, std values: ', average_real(self), &
                                                                                                               std_dev_real(self)
            write(*,'('//tab//'x,a'//text_width//',l1)') 'any nan values? ', has_nan(self)
        endif
        write(*,'(a)')                                  ' '
    end subroutine image_print_info

    !>  \brief  returns .true. if image is allocated/in memory
    elemental pure logical function image_allocated(self)
        implicit none
        ! argument
        class(image),    intent(in)  ::  self
        ! start work
        image_allocated = self%in_memory
    end function image_allocated

    !>  \brief check if both images have same dimensions, return true if they do.
    pure logical function same_dim(img1,img2)
        implicit none
        class(image),   intent(in)  ::  img1
        type(image),    intent(in)  ::  img2
        if (        (img1%logical_dimensions(1) .eq. img2%logical_dimensions(1)) &
            .and.   (img1%logical_dimensions(2) .eq. img2%logical_dimensions(2)) &
            .and.   (img1%logical_dimensions(3) .eq. img2%logical_dimensions(3)) ) then
            same_dim = .true.
        else
            same_dim = .false.
        endif
    end function same_dim

    !>  \brief check if both image have same type, return true if they do.
    pure logical function same_type(img1,img2)
        implicit none
        class(image), intent(in) :: img1,img2
        same_type = img1%is_in_real_space .and. img2%is_in_real_space .or. &
                    (.not. img1%is_in_real_space) .and. (.not. img2%is_in_real_space)
    end function same_type


    !>  \brief  determine whether an image is double or single precision
!#if defined(SKIP_RUNTIME_CHECKS)
!    pure &
!#endif
!    logical function is_double(img)
!        use runtime_parameters, only : terminate
!        implicit none
!        type(image), intent(in) :: img
!        is_double = .false.
!    end function is_double

    pure logical function is_real(self)
        class(image),    intent(in)  ::self
        is_real = self%is_in_real_space
    end function
    pure logical function is_comp(self)
        class(image),    intent(in)  ::  self
        is_comp = .not. self%is_in_real_space
    end function

    !>  \brief  determine whether image is square / cubic
    pure logical function is_square(img)
        implicit none
        class(image), intent(in) :: img
        if ( img%logical_dimensions(3) .eq. 1) then
            if ( img%logical_dimensions(1) .ne. img%logical_dimensions(2) ) then
                is_square = .false.
            else
                is_square = .true.
            endif
        else
            if ( img%logical_dimensions(1) .ne. img%logical_dimensions(2) .or. &
                 img%logical_dimensions(1) .ne. img%logical_dimensions(3) ) then
                is_square = .false.
            else
                is_square = .true.
            endif
        endif
    end function is_square




    !>  \brief  check whether the image is 3 dimensional
    pure logical function is_threed(self)
        implicit none
        ! arguments
        class(image),    intent(in)  ::  self
        ! private variables

        ! start work
        is_threed = .not. any(self%logical_dimensions .eq. 1)
    end function is_threed

    !>  \brief  determine whether image has even dimensions
    pure logical function has_even_dims(img)
        use runtime_parameters, only : is_even
        implicit none
        class(image), intent(in) :: img
        has_even_dims = dims_are_even(img%logical_dimensions)
    end function has_even_dims

    !>  \brief  determine whether dimensions are even
    pure logical function dims_are_even(dimensions)
        use runtime_parameters, only : is_even
        implicit none
        integer,    intent(in)  ::  dimensions(3)
        dims_are_even = .false.
        if (dimensions(3) .eq. 1) then
            if (is_even(dimensions(1)) .and. is_even(dimensions(2))) dims_are_even = .true.
        else
            if (is_even(dimensions(1)) .and. is_even(dimensions(2)) .and. is_even(dimensions(3))) dims_are_even = .true.
        endif
    end function dims_are_even

    !>  \brief  determine whether an image has been binarised (meaning it only contains 1's and 0's)
    logical function is_binarised(img)
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        class(image),    intent(in)  ::  img
        ! private variables
        integer ::  i,j

        ! start work

        ! only makes sense for real images
        if (.not. img%is_in_real_space) then
            write(*,'(a)') '**error(is_binarised): only real images are supported'
            call terminate('fatal error in is_binarised')
        endif

        !
        is_binarised = .true.

        ! if any value is other than 0.0 or 1.0, then not binarised
        do j=1,img%logical_dimensions(2)
            do i=1,img%logical_dimensions(1)
                if (img%real_values(i,j,1) .ne. 0.0 .and. (img%real_values(i,j,1)-1.0) .gt. tiny(img%real_values(i,j,1))) then
                    is_binarised = .false.
                    return
                endif
            enddo
        enddo
    end function is_binarised

    !>  \brief  determine whether an image is blank/zero
    pure logical function image_is_blank(img) result(is_blank)
        implicit none
        ! arguments
        class(image),    intent(in)  ::  img
        ! start work
        if (img%is_in_real_space) then
            if (any( img%real_values(1:img%logical_dimensions(1),1:img%logical_dimensions(2),1:img%logical_dimensions(3)) .ne. &
                                     img%real_values(1,1,1) )   ) then
                is_blank = .false.
            else
                is_blank = .true.
            endif
        else
            if (any( img%complex_values .ne. (0.0,0.0) )   ) then
                is_blank = .false.
            else
                is_blank = .true.
            endif
        endif
    end function image_is_blank

        !>  \brief  return the minimum value in an image
    real function image_min_val(self)
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        type(image),    intent(in)  ::  self
        ! private variables
        ! start work
        if (self%is_in_real_space) then
            image_min_val = minval(self%real_values(1:self%logical_dimensions(1),   &
                                                    1:self%logical_dimensions(2),   &
                                                    1:self%logical_dimensions(3)))
        else
            write(*,'(a)') '**error(image_min_val): not implemented for complex images'
            call terminate('fatal error in image_min_val')
        endif
    end function image_min_val

    !>  \brief  return the minimum value in an image
    real function image_max_val(self)
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        type(image),    intent(in)  ::  self
        ! private variables
        ! start work
        if (self%is_in_real_space) then
            image_max_val = maxval(self%real_values(1:self%logical_dimensions(1),   &
                                                    1:self%logical_dimensions(2),   &
                                                    1:self%logical_dimensions(3)))
        else
            write(*,'(a)') '**error(image_max_val): not implemented for complex images'
            call terminate('fatal error in image_max_val')
        endif
    end function image_max_val

    !>  \brief  return the sum of real values. for large image, use sum_double instead!
    real function sum_real(img)
        implicit none
        type(image),    intent(in)  ::  img
        sum_real = sum_double(img)
    end function sum_real

    !>  \brief  return the average of real values of an image
    real function average_real(img)
        implicit none
        ! argument
        type(image),    intent(in)  ::  img
        ! start work
        average_real = sum_real(img) / real(product(img%logical_dimensions))
    end function average_real

    !>  \brief return the average of real values at the edges of the image
    pure real function edge_average(self)
        implicit none
        ! argument
        type(image),    intent(in)  ::  self
        ! private variables
        integer         ::  i,j,k
        real(kind=8)    ::  s
        integer(kind=8) ::  c
        ! start work
        s = 0.0d0
        c = 0
        do k=1,self%logical_dimensions(3),max(self%logical_dimensions(3)-1,1)
            do j=1,self%logical_dimensions(2),max(self%logical_dimensions(2)-1,1)
                do i=1,self%logical_dimensions(1),max(self%logical_dimensions(1)-1,1)
                    s = s + self%real_values(i,j,k)
                    c = c + 1
                enddo
            enddo
        enddo
        edge_average = s/c
    end function edge_average

    !>  \brief  return the variance of the real values of an image
    !!  note: this algorithm is not precise and should not be used. see
    !!  http://en.wikipedia.org/wiki/algorithms_for_calculating_variance
    !!  \todo replace with recommended robust algorithm
    pure real function variance_real(img)
        implicit none
        ! argument
        class(image),   intent(in)  ::  img
        ! private variables
        integer                     ::  i,j,k
        real(kind=8)                ::  sum, sum2
        ! start work
        !variance_real = sum(img%real_values(1:img%logical_dimensions(1),1:img%logical_dimensions(2),1:img%logical_dimensions(3))**2) - sum_real(img)**2
        sum = 0.0d0
        sum2 = 0.0d0
        do k=1,img%logical_dimensions(3)
            do j=1,img%logical_dimensions(2)
                do i=1,img%logical_dimensions(1)
                    sum = sum + img%real_values(i,j,k)
                    sum2 = sum2 + img%real_values(i,j,k)**2
                enddo
            enddo
        enddo
        variance_real = (sum2 - (sum**2/product(img%logical_dimensions)))/product(img%logical_dimensions)
    end function variance_real

    !>  \brief  return the standard deviation of the real values of an image
    pure real function std_dev_real(img)
        implicit none
        ! argument
        type(image),    intent(in)  ::  img
        ! start work
        std_dev_real = sqrt(variance_real(img))
    end function std_dev_real

    !>  \brief  Normalize the variance of an image to 1.0 (leave its average unchanged)
    subroutine NormaliseVariance(self)
        ! argument
        class(image),   intent(inout)   ::  self
        ! private variable
        real    ::  variance
        ! start work
        variance = self%ReturnVariance()
        self%real_values  = self%real_values / variance
    end subroutine NormaliseVariance




    !>  \brief  return the sum of real values as a double precision value.
    !!
    !!  the option ignore_friedel_mates is meant for use when summing ctf^2 volumes, which are real but represent reciprocal-space information arranged in the fftw manner,
    !!  whereby in the first dimension, friedel mates are described explicitely except for the nyquist frequency element.
    double precision function sum_double(img,ignore_friedel_mates)
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        class(image),           intent(in)  ::  img
        logical,    optional,   intent(in)  ::  ignore_friedel_mates    !<  if .true., friedel mates (which are descibed explicitely in the first dimension of a ctf^2 volume) will be ignored when computing the sum
        ! private variables
        integer ::  i,j,k
        logical ::  iignore_friedel_mates
        ! start work

        ! are we ignoring friedel mates?
        if (present(ignore_friedel_mates)) then
            iignore_friedel_mates = .true.
        else
            iignore_friedel_mates = .false.
        endif

        if (iignore_friedel_mates) then
            ! the logic for this is only written for even dimension images, because i'm too lazy to think about odd dimensions right now.
            if (mod(img%logical_dimensions(2),2) .ne. 0. .or. mod(img%logical_dimensions(2),2) .ne. 0.) then
                write(*,'(2a,3(i5,1x))')    '**error(sum_double): images with odd 2nd or 3rd dimensions ', &
                                            'are not compatible. dimensions = ', img%logical_dimensions
                call terminate('fatal error in sum_double')
            endif
            ! this ignore_friedel_mates options is only really meant for ctf^2 volumes with specific dimensions. perhaps this restriction can be lifted, if needed.
            if (img%logical_dimensions(2) .ne. img%logical_dimensions(3) .or. &
                img%logical_dimensions(1) .ne. img%logical_dimensions(2)/2+1) then
                write(*,'(a,3(i5,1x))') '**error(sum_double): unexpected input image dimensions: ', img%logical_dimensions
                call terminate('fatal error in sum_double')
            endif
        endif

        sum_double = 0.0d0
        if (img%is_in_real_space) then
            do i=1,img%logical_dimensions(1)
                do j=1, img%logical_dimensions(2)
                    if (iignore_friedel_mates .and. i == 1 .and. j .gt. 1 .and. j .lt. img%logical_dimensions(2)/2+1) cycle
                    do k=1, img%logical_dimensions(3)
                        if (iignore_friedel_mates .and. i == 1 .and. j == 1 .and. k .gt. 1 .and. &
                            k .lt. img%logical_dimensions(3)/2+1) cycle
                        sum_double = sum_double + dble(img%real_values(i,j,k))
                    enddo
                enddo
            enddo
        else
            write(*,'(a)') '**error(sum_double): image must be in real space'
            call terminate('fatal error in sum_double')
        endif
    end function sum_double


    !>  \brief  check whether any of the volume voxels are nan
    pure logical function has_nan(self)
		use, intrinsic :: ieee_arithmetic
        implicit none
        ! arguments
        type(image),    intent(in)  ::  self
        ! private variables
        integer ::  i,j,k
        ! start work
        has_nan = .false.
        if (self%is_in_real_space) then
            do k=1,self%logical_dimensions(3)
                do j=1,self%logical_dimensions(2)
                    do i=1,self%logical_dimensions(1)
                        if (ieee_is_nan(self%real_values(i,j,k))) then
                            has_nan = .true.
                            return
                        endif
                    enddo
                enddo
            enddo
        else
            do k=1,self%logical_dimensions(3)
                do j=1,self%logical_dimensions(2)
                    do i=1,highest_index_dim1(self%logical_dimensions(1))
                        if (ieee_is_nan(aimag(self%complex_values(i,j,k))) .or. ieee_is_nan(real(self%complex_values(i,j,k)))) then
                            has_nan = .true.
                            return
                        endif
                    enddo
                enddo
            enddo
        endif
    end function has_nan

    !>  \brief  return the physical memory index of the last pixel in the first dimension of a ft'ed image
    pure function highest_index_dim1(dim) result(i)
        use runtime_parameters, only : is_even
        implicit none
        ! arguments
        integer,    intent(in)  ::  dim !<  logical extent of the image in dimension 1
        ! result
        integer                 ::  i
        ! start work
        if (is_even(dim)) then
            i = dim/2 + 1
        else
            i = (dim-1)/2 + 1
        endif
    end function highest_index_dim1

    !>  \brief  add an image to the end of an array of images
    subroutine image_append_to_array(self,new_img)
        implicit none
        ! arguments
        type(image),    allocatable,    intent(inout)   ::  self(:)     !<  array of images
        type(image),                    intent(in)      ::  new_img     !<  this image will be appended to the array of images
        ! private variables
        type(image),    allocatable ::  tmp(:)
        ! start work
        if (allocated(tmp)) deallocate(tmp)
        allocate(tmp(size(self)+1))
        tmp(1:size(self)) = self(1:size(self))
        tmp(size(tmp)) = new_img
        call move_alloc(from=tmp,to=self) ! f2003 intrinsic
    end subroutine image_append_to_array

    !>  \brief  assignment (=) operation
    !!  imgin is effectively copied to imgout
    subroutine image_assignment(imgout,imgin)
        implicit none
        ! arguments
        type(image), intent(inout) :: imgout
        type(image), intent(in   ) :: imgin
        ! private variables

        ! start work

        ! allocate memory for the output image
        call image_allocate(imgout,mould=imgin)

        ! copy greyvalues over
        imgout%real_values = imgin%real_values

    end subroutine image_assignment

    subroutine real_to_image_assignment(imgout,realin)
        implicit none
        type(image), intent(inout) :: imgout
        real, intent(in) :: realin
        imgout%real_values = realin
    end subroutine real_to_image_assignment

    subroutine complex_to_image_assignment(imgout,compin)
        implicit none
        type(image), intent(inout) :: imgout
        complex, intent(in) :: compin
        imgout%complex_values = compin
    end subroutine complex_to_image_assignment

    !> \brief allocate memory for the image object. must know dimensions!
    !!
    !!  \todo   should all image properties really be carried over? isn't that what the assignment operation is for?
    subroutine image_allocate(img,mould,dims)
        use iso_c_binding
        use omp_lib
        !use fftw_include   ! legacy fftw-fortran interface
        use fftw33          ! modern fftw-fortran interface
        use runtime_parameters, only : terminate, is_even
        implicit none
        ! arguments
        class(image),               intent(inout)   ::  img
        type(image),    optional,   intent(in)      ::  mould       !<  the image will be allocated with this image's dimensions
        integer,        optional,   intent(in)      ::  dims(3)     !<  dimensions of the image
        ! private variables
        integer                 ::  array_shape(3)
        logical,    parameter   ::  debug   =   .false.

        ! start work

        ! check whether memory is already allocated
        if (img%in_memory) then
            if (debug) write(*,'(2a)') '**debug(image_allocate): tyring to allocate memory to image already in memory, ',    &
                                        'will deallocate first'
            call image_deallocate(img)
        endif

        if (present(mould)) then
            if (debug) write(*,'(a)') '**debug(image_allocate): using dimensions, type and origin_at_center from mould'
            img%logical_dimensions = mould%logical_dimensions
            img%is_in_real_space = mould%is_in_real_space
            img%origin_at_center = mould%origin_at_center
        endif

        if (present(dims)) img%logical_dimensions = dims


        ! check the dimensions of the image have been set
        if (all (img%logical_dimensions .eq. 1)) then
            write(*,'(a)') '**error(image_allocate): allocating memory to an image with dimensions (1,1,1)'
            call terminate('fatal error in image_allocate')
        endif

        ! workout the desired shape of the complex array
        if (is_even(img%logical_dimensions(1))) then
            array_shape(1) = img%logical_dimensions(1)/2 + 1
        else
            array_shape(1) = (img%logical_dimensions(1)-1)/2 + 1
        endif
        array_shape(2) = img%logical_dimensions(2)
        array_shape(3) = img%logical_dimensions(3)

        ! do the allocation, the first dimension gets a +2 because of the fftw
        ! letting fftw do the allocation in c ensures that we will be using aligned memory
        !$omp   critical    (fftw_omp_crit)
        !img%p = fftwf_alloc_complex(int(product(array_shape), c_size_t))
        img%p = fftwf_malloc(int(product(array_shape)*8, c_size_t))
        ! remember that we did the allocation
        img%in_memory = .true.
        !$omp   end critical    (fftw_omp_crit)

        ! check whether the allocation went smoothly
        ! don't know how to do this when using fftw_malloc

        ! set up the cvalue pointer
        call c_f_pointer(img%p,img%complex_values,array_shape)

        ! workout the shape of the real array
        array_shape(1) = 2 * array_shape(1)

        ! set up the rvalue pointer
        call c_f_pointer(img%p,img%real_values,array_shape)

        ! initialise the contents
        img%complex_values = (0.0,0.0)

        ! prepare the plans for fftw
        if (.not. img%planned) then
            !$omp   critical    (fftw_omp_crit)
            if ( img%logical_dimensions(3) .gt. 1) then
                img%plan_fwd = fftwf_plan_dft_r2c_3d(img%logical_dimensions(3), img%logical_dimensions(2), &
                                                    img%logical_dimensions(1), &
                                                    img%real_values, img%complex_values, fftw_estimate )
                img%plan_bwd = fftwf_plan_dft_c2r_3d(img%logical_dimensions(3), img%logical_dimensions(2), &
                                                    img%logical_dimensions(1), &
                                                    img%complex_values, img%real_values, fftw_estimate )
            else
                img%plan_fwd = fftwf_plan_dft_r2c_2d(img%logical_dimensions(2), img%logical_dimensions(1),             &
                                                    img%real_values, img%complex_values, fftw_estimate )
                img%plan_bwd = fftwf_plan_dft_c2r_2d(img%logical_dimensions(2), img%logical_dimensions(1),             &
                                                    img%complex_values, img%real_values, fftw_estimate )
            endif
            img%planned = .true.
            !$omp   end critical    (fftw_omp_crit)
        else
            if (debug) write(*,'(a)') '**debug(image_allocate): strange, the fftw plans were already there...'
            call terminate('fatal error in image_allocate: plans were already setup')
        endif

        ! Update object properties
        call update_physical_address_of_box_center(img)

    end subroutine image_allocate

    !>  \brief  deallocate memory allocated to an image
    subroutine image_deallocate(img)
        use omp_lib
        use fftw33
        implicit none
        ! arguments
        type(image),    intent(inout) ::  img
        ! private variables
        logical,    parameter   ::  debug   =   .false.
        ! start work
        if (debug) write(*,'(a)') '**debug(image_deallocate): entering'

        ! deallocate image data
        if (img%in_memory) then
            if (debug) write(*,'(a)') '**debug(image_deallocate): about to disassociate complex & real pointer arrays'
            if (associated(img%complex_values)) nullify(img%complex_values)     ! dissociate the cvalue pointer from the memory
            if (associated(img%real_values)) nullify(img%real_values)   ! dissociate the actual memory
            ! free the memory
            if (debug) write(*,'(a)') '**debug(image_deallocate): about to deallocate memory'
            !$omp   critical    (fftw_omp_crit)
            call fftwf_free(img%p)
            img%p = c_null_ptr
            img%in_memory   =   .false.
            !$omp   end critical    (fftw_omp_crit)
        else
            if (debug) write(*,'(2a,3(i0,1x))')  '**debug(image_deallocate): tyring to deallocate an image ', &
                                                'which was not in memory anyway, dim = ', img%logical_dimensions
        endif

        ! deal with fftw plans
        if (img%planned) then
            if (debug) write(*,'(a)') '**debug(image_deallocate): about to destroy fftw plans'
            ! destroy the plan
            !if (debug) write(*,*) ' **debug(image_deallocate): destroying in thread', omp_get_thread_num()+1
            !$omp   critical    (fftw_omp_crit)
            call fftwf_destroy_plan(img%plan_fwd)
            img%plan_fwd = c_null_ptr
            call fftwf_destroy_plan(img%plan_bwd)
            img%plan_bwd = c_null_ptr
            img%planned     =   .false.
            !$omp   end critical    (fftw_omp_crit)
            !if (debug) write(*,*) ' **debug(image_deallocate): destroying done in thread', omp_get_thread_num()+1
        endif
        if (debug) write(*,'(a)') '**debug(image_deallocate): exiting'
    end subroutine image_deallocate



    !>  \brief  corrects 3d volume for fall-off at edges due to interpolation kernel in pextract. adapted from frealign.
    !!
    !!  specific for nearest-neighbour correction. superseded by correct_3d in the mod_threed module, which works with arbitrary interpolation schemes.
    subroutine correct3d(img,sinclut,reverse)
        implicit none
        ! arguments
        class(image),   intent(inout)           ::  img
        real,           intent(in)              ::  sinclut(2000)
        logical,        intent(in), optional    ::  reverse
        ! private variables
        logical ::  rreverse
        real    ::  rad(3)  ! distance from center along each dimension
        integer ::  mid(3)  ! mid point along each direction
        integer ::  i,j,k
        real    ::  boxft_lut

        ! this is only meant for real images
        if (.not. img%is_in_real_space) then
            write(*,'(a,a)') '**error(correct3d): cannot work on images of type ', 'comp'
            !stop 'fatal error in correct3d'
            call terminate()
        endif

        ! check whether we need to correct for, or apply the fall-off
        if (present(reverse)) then
            rreverse = reverse
        else
            rreverse = .false.
        endif

        ! calculate midpoints along each dimension
        do i=1,3
            mid(i) = img%logical_dimensions(i)/2+1
        enddo

        ! loop over voxels, apply / correct for the falloff
        do i=1,img%logical_dimensions(1)
            rad(1) = real(i-mid(1))/img%logical_dimensions(1)
            do j=1,img%logical_dimensions(2)
                rad(2) = real(i-mid(2))/img%logical_dimensions(2)
                do k=1,img%logical_dimensions(3)
                    rad(3) = real(i-mid(3))/img%logical_dimensions(3)
                    if (rreverse) then
                        img%real_values(i,j,k) = img%real_values(i,j,k) * boxft_lut(rad,sinclut)
                    else
                        img%real_values(i,j,k) = img%real_values(i,j,k) / boxft_lut(rad,sinclut)
                    endif
                enddo
            enddo
        enddo
    end subroutine correct3d


    !>  \brief  create a small image by windowing a large image
    subroutine window_img(imgin,imgout,coo,window_to_img)
        implicit none
        ! arguments
        type(image),                intent(inout)   ::  imgin               !<  input (large) image
        type(image),                intent(inout)   ::  imgout              !<  output (small) image
        integer,                    intent(in)      ::  coo(3)              !<  center coordinates
        logical,        optional,   intent(in)      ::  window_to_img       !<  defaults to false. if true, image data are copied from the small image into the large one
        ! private variables
        integer                 ::  i,imin,jmin,kmin,imax,jmax,kmax
        integer                 ::  imin_out,imax_out,jmin_out,jmax_out,kmin_out,kmax_out
        logical                 ::  wwindow_to_img
        logical,    parameter   ::  debug   =   .false.

        ! check types
        if (.not. same_type(imgin,imgout)) then
            write(*,'(a)') '**error(window_img): images of different types (real/Fourier)'
            call terminate('fatal error in window_img')
        endif

        ! for now, we only deal with real images
        if (.not. imgin%is_in_real_space) then
            write(*,'(a)') '**error(window_img): cannot operate on images in Fourier space'
            call terminate('fatal error in window_img')
        endif

        ! which way are we going?
        if (present(window_to_img)) then
            wwindow_to_img = window_to_img
        else
            wwindow_to_img = .false.
        endif

        ! check input is allocated
        if (.not. is_allocated(imgin)) then
            write(*,'(a)') '**error(window_img): input image is not in memory'
            call terminate('fatal error in window_img')
        endif

        ! we may also need the small window image to be allocated
        if (wwindow_to_img .and. .not. is_allocated(imgout)) then
            write(*,'(a)') '**error(window_img): small window image is not in memory'
            call terminate('fatal error in window_img')
        endif

        ! check dimensions
        do i=1,3
            if (imgout%logical_dimensions(i) .gt. imgin%logical_dimensions(i)) then
                write(*,'(a,i0,a,i0,a,i0)')     '**error(window_img): dimension ', i,   &
                                                ' of output image is greater than of input image: ',&
                                                imgout%logical_dimensions(i), ' > ', imgin%logical_dimensions(i)
                call terminate('fatal error in window_img')
            endif
        enddo

        ! check dimensions of output image have been set
        if (all(imgout%logical_dimensions(1:2) .eq. 1)) then
            write(*,'(a,3(i3,1x))') '**error(window_img): dimensions of output image have not been set: ', imgout%logical_dimensions
            call terminate('fatal error in window_img')
        endif

        ! check dimensions of input image have been set
        if (all(imgin%logical_dimensions(1:2) .eq. 1)) then
            write(*,'(a,3(i3,1x))') '**error(window_img): dimensions of input image have not been set: ', imgin%logical_dimensions
            call terminate('fatal error in window_img')
        endif

        ! check the output dimensions are even
        if (.not. has_even_dims(imgout)) then
            write(*,'(a,3(i3,1x))') '**error(window_img): dimensions of output image are not even: ', imgout%logical_dimensions
            call terminate('fatal error in window_img')
        endif

        ! check allocation status of output
        if ( .not. is_allocated(imgout)) then
            call image_allocate(imgout)
        endif

        ! work out min & max coordinates in each direction
        imin = coo(1) - imgout%logical_dimensions(1)/2
        jmin = coo(2) - imgout%logical_dimensions(2)/2
        kmin = coo(3) - imgout%logical_dimensions(3)/2
        imax = imin + imgout%logical_dimensions(1) - 1
        jmax = jmin + imgout%logical_dimensions(2) - 1
        kmax = kmin + imgout%logical_dimensions(3) - 1
        if (imgout%logical_dimensions(3) .eq. 1) then
            kmin = 1
            kmax = 1
        endif

        ! if we're over the edge of the input image, we leave those rows/columns alone
        if (imin .lt. 1) then
            imin_out =  1 + (1-imin)
            imin = 1
        else
            imin_out = 1
        endif
        if (jmin .lt. 1) then
            jmin_out = 1 + (1-jmin)
            jmin = 1
        else
            jmin_out = 1
        endif
        if (kmin .lt. 1) then
            kmin_out = 1 + (1-kmin)
            kmin = 1
        else
            kmin_out = 1
        endif
        if (imax .gt. imgin%logical_dimensions(1)) then
            imax_out = imgout%logical_dimensions(1) - (imax-imgin%logical_dimensions(1))
            imax = imgin%logical_dimensions(1)
        else
            imax_out = imgout%logical_dimensions(1)
        endif
        if (jmax .gt. imgin%logical_dimensions(2)) then
            jmax_out = imgout%logical_dimensions(2) - (jmax-imgin%logical_dimensions(2))
            jmax = imgin%logical_dimensions(2)
        else
            jmax_out = imgout%logical_dimensions(2)
        endif
        if (kmax .gt. imgin%logical_dimensions(3)) then
            kmax_out = imgout%logical_dimensions(3) - (kmax-imgin%logical_dimensions(3))
            kmax = imgin%logical_dimensions(3)
        else
            kmax_out = imgout%logical_dimensions(3)
        endif
        if (imgout%logical_dimensions(3) .eq. 1) then
            kmin_out = 1
            kmax_out = 1
        endif

        ! window
        if (debug) then
            !print '(a,2i5)', '**debug(window_img): imin,     imax     :', imin,imax
            !print '(a,2i5)', '**debug(window_img): jmin,     jmax     :', jmin,jmax
            !print '(a,2i5)', '**debug(window_img): imin_out, imax_out :', imin_out,imax_out
            !print '(a,2i5)', '**debug(window_img): jmin_out, jmax_out :', jmin_out,jmax_out
        endif
        if (wwindow_to_img) then
            ! the unusual case, where we are copying from the window to the large image
            imgin%real_values(imin:imax,jmin:jmax,kmin:kmax) =   imgin%real_values(imin:imax,jmin:jmax,kmin:kmax) + &
                                                    imgout%real_values(imin_out:imax_out,jmin_out:jmax_out,kmin_out:kmax_out)
        else
            ! the default mode, where we actual window a small part of the input image
            imgout%real_values = 0.0
            imgout%real_values(imin_out:imax_out,jmin_out:jmax_out,kmin_out:kmax_out) = &
                                                                            imgin%real_values(imin:imax,jmin:jmax,kmin:kmax)
            ! check for nan values
            if (has_nan(imgout)) then
                write(*,'(a,2(f0.3,1x))') '**warning(window_img): windowed image has at least one nan value. coordinates = ', coo
            endif
        endif
    end subroutine window_img



    !>  \brief  print a line of density values
    subroutine print_line(img,lineno,sliceno)
        implicit none
        type(image),    intent(in)  ::  img
        integer,        intent(in)  ::  lineno
        integer,        intent(in)  ::  sliceno
        if (img%is_in_real_space) then
                write(*,*) img%real_values(1:img%logical_dimensions(1),lineno,sliceno)
        else
                write(*,'(a,a)') '**error (print_line): unsupported image type ','comp'
        endif
    end subroutine print_line

    !>  \brief  return a histogram (an integer array) of values in the image.
    !!  voxels contribute to a bin if their value is .ge. the lower limit .and. .lt. the upper limit. the exception is the last bin, which goes all the way up to the upper limit
    subroutine histogram_real(img,histogram,bin_boundaries,num_of_bins, bin_condense_outliers, negs_in_first_bin,print)
        implicit none
        ! arguments
        class(image),               intent(in)  ::  img                     !<  input image
        integer,    allocatable,    intent(out) ::  histogram(:)            !<  for each bin, the number of voxels whose values are within the bin boundaries
        real,       allocatable,    intent(out) ::  bin_boundaries(:)       !<  the lower boundary for each bin, as well as the upper limit for the last bin.
        integer,        optional,   intent(in)  ::  num_of_bins             !<  number of bins in histogram. defaults to 100.
        logical,        optional,   intent(in)  ::  bin_condense_outliers   !<  whether to condense all of the outliers into one final bin at the end. defaults to true.
        logical,        optional,   intent(in)  ::  negs_in_first_bin       !<  all negative values in the first bin. useful if you're interested in positive values only. false by default.
        logical,        optional,   intent(in)  ::  print                   !<  whether the print the histogram
        ! private variables
        integer                 ::  nnum_of_bins
        integer                 ::  ierr, i, j, k, b
        integer                 ::  found_bin
        real                    ::  bin_width
        integer                 ::  tmp
        logical                 ::  nnegs_in_first_bin, pprint, bbin_condense_outliers
        real                    ::  grey_min, grey_max, grey_ave, grey_var
        logical,    parameter   ::  debug   =   .false.
        ! start work

        ! image should be real
        if (.not. img%is_in_real_space) then
            write(*,'(2a)') '**error(histogram_real): image of unsupported type ', 'comp'
            call terminate('fatal error in histogram_real')
        endif

        ! number of bins
        if (present(num_of_bins)) then
            nnum_of_bins = num_of_bins
        else
            nnum_of_bins = 100
        endif

        ! put all the outliers in final bin?
        if (present(bin_condense_outliers)) then
            bbin_condense_outliers= bin_condense_outliers
        else
            bbin_condense_outliers= .true.
        endif

        ! concentrate on positive values?
        if (present(negs_in_first_bin)) then
            nnegs_in_first_bin = negs_in_first_bin
        else
            nnegs_in_first_bin = .false.
        endif

        ! print to console?
        if (present(print)) then
            pprint = print
        else
            pprint = .false.
        endif

        ! allocate the output arrays
        allocate(histogram(nnum_of_bins),bin_boundaries(nnum_of_bins+1),stat=ierr)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**error(histogram_real): memory allocation failed with error ', ierr
            call terminate('fatal error in histogram_real')
        endif

        ! get basic statistics
        grey_min = minval(img%real_values(1:img%logical_dimensions(1),1:img%logical_dimensions(2),1:img%logical_dimensions(3)))
        grey_max = maxval(img%real_values(1:img%logical_dimensions(1),1:img%logical_dimensions(2),1:img%logical_dimensions(3)))
        grey_ave = average_real(img)
        grey_var = variance_real(img)
        if (debug) then
            print '(a,2(es13.6,1x))', '**debug(histogram_real): min, max = ', grey_min, grey_max
            print '(a,3(es13.6,1x))', '**debug(histogram_real): ave, var, std = ', grey_ave, grey_var, sqrt(grey_var)
        endif

        ! initialise the output arrays
        histogram(:) = 0
        bin_boundaries(:) = 0.0

        ! work out the bin boundaries
        bin_boundaries(1) = grey_min
        bin_boundaries(nnum_of_bins+1) = grey_max


        !bin_boundaries(2)              =   max(grey_min,(grey_ave-10.*sqrt(grey_var)))
        !bin_boundaries(nnum_of_bins)   =   min(grey_max,(grey_ave+10.*sqrt(grey_var)))
        !! the lower boundary of the first bin and the higher boundary of the last bin are always set to the minimum and maximum values
        !bin_boundaries(1) = grey_min
        !bin_boundaries(nnum_of_bins+1) = grey_max

        ! if we're placing all negative values in the first bin, then the second bin's lower bound should be 0.0
        if (nnegs_in_first_bin) then
            bin_boundaries(2) = max(0.0,bin_boundaries(2))
            bin_width = (bin_boundaries(nnum_of_bins+1) - bin_boundaries(2)) / (nnum_of_bins-1)
            do i=3,nnum_of_bins
                bin_boundaries(i) = bin_boundaries(2) + bin_width*(i-1)
            enddo
        else
            bin_width = (bin_boundaries(nnum_of_bins+1) - bin_boundaries(1)) / nnum_of_bins
            if (debug) print *, '**debug(histogram_real): initial bin_width = ', bin_width

            if (bbin_condense_outliers) then
                bin_boundaries(2)               =   max(bin_boundaries(1)+bin_width,            (grey_ave-5.*sqrt(grey_var)))
                bin_boundaries(nnum_of_bins)    =   min(bin_boundaries(nnum_of_bins+1)-bin_width,   (grey_ave+5.*sqrt(grey_var)))
            else
                bin_boundaries(2)               =   bin_boundaries(1)+bin_width
                bin_boundaries(nnum_of_bins)    =   bin_boundaries(nnum_of_bins+1)-bin_width
            endif

            bin_width = (bin_boundaries(nnum_of_bins) - bin_boundaries(2)) / (nnum_of_bins-2)
            if (debug) print *, '**debug(histogram_real): new bin_width = ', bin_width
            do i=3,nnum_of_bins-1
                bin_boundaries(i) = bin_boundaries(i-1) + bin_width
            enddo
        endif

        if (debug) print *, '**debug(histogram_real): bin_width = ', bin_width

        ! loop over the image to accumulate the histogram statistics
        ! (note that in the case where there are more bins than pixels, it would makes sense to loop over the bins. however, i expect this to be a rare occurence.
        do k=1,img%logical_dimensions(3)
            do j=1,img%logical_dimensions(2)
                do i=1,img%logical_dimensions(1)
                    found_bin = 0
                    ! check if the current voxel belongs to a bin other than the last one
                    do b=1,nnum_of_bins-1
                        if (img%real_values(i,j,k) .ge. bin_boundaries(b) .and. &
                            img%real_values(i,j,k) .lt.  bin_boundaries(b+1)) then
                            found_bin = b
                            exit
                        endif
                    enddo
                    ! if we still haven't found the bin the voxel belongs to, check the last bin
                    if (found_bin .eq. 0) then
                        b = nnum_of_bins
                        if (img%real_values(i,j,k) .ge. bin_boundaries(b) .and. &
                            img%real_values(i,j,k) .le.  bin_boundaries(b+1)) then
                                found_bin = b
                        endif
                    endif
                    ! if we still haven't found the bin the voxel belongs to, something went wrong
                    if (found_bin .eq. 0) then
                        write(*,'(a,3i4,3(a,f10.3),a)') '**error(histogram_real): could not find bin for voxel at coordinates (', &
                                                        i, j, k, ') with value ', img%real_values(i,j,k), &
                                                        '. the min and max of the histogram are ',bin_boundaries(1), &
                                                        ' and ', bin_boundaries(nnum_of_bins+1), '.'
                        call terminate('fatal error in histogram_real')
                    else
                        histogram(found_bin) = histogram(found_bin) + 1
                    endif
                enddo
            enddo
        enddo

        ! optional: print out the histogram
        if (pprint) then
            tmp = 0
            write(*,'(/4(a10,1x))') 'from', 'to', 'count','total'
            do b=1,nnum_of_bins
                tmp = tmp + histogram(b)
                write(*,'(2(es10.3,1x),i10,1x,i10)') bin_boundaries(b), bin_boundaries(b+1), histogram(b),tmp
            enddo
            write(*,'(a)') ' '
        endif
    end subroutine histogram_real





    !>  \brief  mirror image in specified dimensions
    subroutine mirror_image(img,flipmask)
        implicit none
        type(image),    intent(inout)               ::  img !<  image to be mirrored
        logical,        intent(in),     optional    ::  flipmask(3) !<  for each dimension, a logical flag to indicate whether to flip
        ! private variables
        logical         ::  fflipmask(3)
        integer         ::  i,j,k!,ii,jj,kk
        !integer            ::  cnt_max(3)
        real,allocatable::  rtemp(:,:,:)

        ! check the type of the input image
        if (.not. img%is_in_real_space) then
            write(*,*) '**error(mirror_image): sorry, complex images not handled yet'
            call terminate()
        endif

        ! check whether flipmask was given as argument
        if (present(flipmask)) then
            fflipmask = flipmask
        else
            fflipmask = (/.true.,.true.,.true./)
        endif

        !write(*,*) '**debug: mirror along dimension 1'
        ! mirror along dimension 1
        if (fflipmask(1)) then
            if (allocated(rtemp)) deallocate(rtemp)
            allocate(rtemp(1,img%logical_dimensions(2),img%logical_dimensions(3)))
            do i=1,img%logical_dimensions(1)/2
                rtemp(1,:,:) = img%real_values(i,:,:)
                img%real_values(i,:,:) = img%real_values(img%logical_dimensions(1)-i+1,:,:)
                img%real_values(img%logical_dimensions(1)-i+1,:,:) = rtemp(1,:,:)
            enddo
        endif

        !write(*,*) '**debug: mirror along dimension 2', img%logical_dimensions
        ! mirror along dimension 2
        if (fflipmask(2)) then
            if (allocated(rtemp)) deallocate(rtemp)
            allocate(rtemp(img%logical_dimensions(1),1,img%logical_dimensions(3)))
            do j=1,img%logical_dimensions(2)/2
                rtemp(:,1,:) = img%real_values(:,j,:)
                img%real_values(:,j,:) = img%real_values(:,img%logical_dimensions(2)-j+1,:)
                img%real_values(1:img%logical_dimensions(1),img%logical_dimensions(2)-j+1,:) = rtemp(:,1,:)
            enddo
        endif

        !write(*,*) '**debug: mirror along dimension 3', img%logical_dimensions
        ! mirror along dimension 3
        if (fflipmask(3)) then
            if (allocated(rtemp)) deallocate(rtemp)
            allocate(rtemp(img%logical_dimensions(1),img%logical_dimensions(2),1))
            do k=1,img%logical_dimensions(3)/2
                rtemp(:,:,1) = img%real_values(:,:,k)
                img%real_values(:,:,k) = img%real_values(:,:,img%logical_dimensions(3)-k+1)
                img%real_values(1:img%logical_dimensions(1),:,img%logical_dimensions(3)-k+1) = rtemp(:,:,1)
            enddo
        endif

        if (allocated(rtemp)) deallocate(rtemp)
    end subroutine mirror_image

    !>  \brief  calculate the amplitude (or phase, or power, or log power...) spectrum of an image. out-of-place.
    subroutine ampl(imgin,imgout,square,log,phases)
        use rotation_matrices, only : degrees
        implicit none
        class(image),           intent(in)      ::  imgin       !<  input image
        type(image),            intent(out)     ::  imgout      !<  output image
        logical,    optional,   intent(in)      ::  square      !<  return the squared amplitudes, i.e. the power
        logical,    optional,   intent(in)      ::  log         !<  return log (base 10) of the amplitudes
        logical,    optional,   intent(in)      ::  phases      !<  return the phases (rather than amplitudes; in degrees) of the image
        ! private variables
        type(image) ::  imgtmp
        logical     ::  ssquare, llog, pphases
        ! start work

        ! optional arguments
        ssquare = .false.
        if (present(square)) ssquare = square

        llog = .false.
        if (present(log)) llog = log

        pphases = .false.
        if (present(phases)) pphases = phases

        ! check allocation status
        if (.not. imgin%in_memory) then
            write(*,'(a)') '**error(ampl): input image not allocated'
            call terminate('fatal error in ampl')
        endif

        ! allocate output
        imgout%logical_dimensions(1) = highest_index_dim1(imgin%logical_dimensions(1))
        imgout%logical_dimensions(2:3) = imgin%logical_dimensions(2:3)
        call image_allocate(imgout)

        ! allocate temp, and copy image over
        call image_allocate(imgtmp,imgin)


        ! check type. if needed, ft the input
        if (imgin%is_in_real_space) then
            imgtmp%real_values = imgin%real_values
            call fwd_ft(imgtmp)
        else
            imgtmp%complex_values = imgin%complex_values
        endif

        ! take amplitudes
        if(ssquare) then
            imgout%real_values(1:imgout%logical_dimensions(1),:,:) = cabs(imgtmp%complex_values)**2
        else if (llog) then
            imgout%real_values(1:imgout%logical_dimensions(1),:,:) = log10(cabs(imgtmp%complex_values))
        else if (pphases) then
            where(imgtmp%complex_values .ne. (0.0,0.0))
                imgout%real_values(1:imgout%logical_dimensions(1),:,:) = &
                        degrees(atan2(aimag(imgtmp%complex_values),real(imgtmp%complex_values)))
            end where
        else
            imgout%real_values(1:imgout%logical_dimensions(1),:,:) = cabs(imgtmp%complex_values)
        endif
        imgout%origin_at_center = .false.

        ! convert to full image
        call ampl_from_half_to_full(imgout)

        imgout%is_in_real_space = .true.
    end subroutine ampl

    !> \brief   move the origin from (1,1,1) to the center of the image. useful when the phase or amplitude
    !!          components of an image ft were used to create a real image, and one wants the full representation
    !!          of the ft, including negative frequencies
    subroutine  ampl_from_half_to_full(img)
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  img
        ! private variables
        type(image) ::  imgout
        integer     ::  old_dim(3), new_dim(3)
        integer     ::  imgout_origin(3)
        integer     ::  i,j,k
        integer     ::  img_phys_addr(3)
        integer     ::  imgout_logi_addr(3)
        logical     ::  symmetry_mate

        ! start work

        ! work out old and new dimensions
        old_dim = img%logical_dimensions
        new_dim = [(old_dim(1)-1)*2,old_dim(2),old_dim(3)]
        if (new_dim(1) .eq. old_dim(1)-1) new_dim(1) = old_dim(1)

        ! prepare output image
        imgout%logical_dimensions = new_dim
        call image_allocate(imgout)

        ! origin of output image
        if (imgout%is_in_real_space .and. imgout%origin_at_center) then
            imgout_origin = imgout%physical_address_of_box_center
        else
            imgout_origin = [1,1,1]
        endif

        ! loop over output image
        do i=1,imgout%logical_dimensions(1)
            symmetry_mate = i .lt. imgout_origin(1) ! we are in the negative frequencies, so we'll have to go for the symmetry mate
            imgout_logi_addr(1) = i - imgout_origin(1) ! logical address
            if (symmetry_mate) imgout_logi_addr(1) = - imgout_logi_addr(1)
            do j=1,imgout%logical_dimensions(2)
                imgout_logi_addr(2) = j - imgout_origin(2) ! logical address
                if (symmetry_mate) imgout_logi_addr(2) = - imgout_logi_addr(2)
                do k=1,imgout%logical_dimensions(3)
                    imgout_logi_addr(3) = k - imgout_origin(3) ! logical address
                    if (symmetry_mate) imgout_logi_addr(3) = - imgout_logi_addr(3)
                    ! get the physical address in the input image corresponding to the output voxel
                    call comp_addr_phys_from_logi(imgout,imgout_logi_addr,img_phys_addr)
                    ! copy value over
                    imgout%real_values(i,j,k) = img%real_values(img_phys_addr(1),img_phys_addr(2),img_phys_addr(3))
                enddo
            enddo
        enddo

        ! replace input with output
        call image_allocate(img,imgout)
        img = imgout

        img%origin_at_center = .true.

    end subroutine ampl_from_half_to_full



    !>  \brief  returns the power in each resolution shell
    function rot_power_spectrum_old(self,log_scale) result(rot_power_spectrum)
        implicit none
        ! arguments
        type(image),    intent(in)  ::  self        !<  input image
        logical,        optional    ::  log_scale   !<  if true, log of power spectrum will be returned
        ! result
        real,   allocatable ::  rot_power_spectrum(:)
        ! private variables
        type(image) ::  imgtmp
        logical     ::  llog_scale
        ! start work
        if (present(log_scale)) then
            llog_scale = log_scale
        else
            llog_scale = .false.
        endif
        call ampl(self,imgtmp)
        imgtmp%real_values = imgtmp%real_values**2
        rot_power_spectrum = rot_average(imgtmp)
        if (llog_scale) then
            where (rot_power_spectrum == 0.) rot_power_spectrum = 0.0000001
            rot_power_spectrum = log10(rot_power_spectrum)
        endif
    end function rot_power_spectrum_old

    !>  \brief  returns the power in each resolution shell
    function rot_power_spectrum(self,log_scale)
        use UsefulFunctions, only : IsOdd
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  self        !<  input image
        logical,        optional        ::  log_scale   !<  if true, log of power spectrum will be returned
        ! result
        real(kind=8),   allocatable ::  rot_power_spectrum(:)
        ! private variables
        logical             ::  llog_scale
        integer             ::  i,j,k
        integer             ::  num_output_elements
        real,   allocatable ::  count_per_shell(:)
        real                ::  coos(3)
        real                ::  rad
        integer             ::  current_shell
        ! start work
        if (present(log_scale)) then
            llog_scale = log_scale
        else
            llog_scale = .false.
        endif
        ! make sure the image is in fourier space
        if (self%is_in_real_space) then
            call fwd_ft(self)
        endif


        ! allocate memory for output arrays
        if (is_threed(self)) then
            num_output_elements = minval(self%logical_dimensions)
        else
            num_output_elements = minval(self%logical_dimensions(1:2))
        endif
        if (IsOdd(num_output_elements)) num_output_elements = num_output_elements + 1
        num_output_elements = num_output_elements / 2
        num_output_elements = num_output_elements + 1   ! because we need the first element to have the origin
        allocate(rot_power_spectrum(num_output_elements))
        allocate(count_per_shell(num_output_elements))

        ! initialise output arrays
        rot_power_spectrum = 0.
        count_per_shell = 0

        ! loop over image
        do k=1,self%logical_dimensions(3)
            ! the distance from the origin to the current pixel, in reciprocal pixel units
            coos(3) = comp_coo_from_phys_index_dim2or3(self%logical_dimensions(3),k)**2
            do j=1,self%logical_dimensions(2)
                coos(2) = comp_coo_from_phys_index_dim2or3(self%logical_dimensions(2),j)**2
                do i=1,highest_index_dim1(self%logical_dimensions(1))
                    coos(1) = comp_coo_from_phys_index_dim1(self%logical_dimensions(1),i)**2
                    ! compute radius, in units of reciprocal pixels
                    rad = sqrt(sum(coos))
                    ! if we are beyond nyquist, cycle
                    if (rad .gt. .5) cycle
                    ! what shell are we in
                    current_shell = nint(2.*rad*real(num_output_elements-1))+1 ! +1 is because the origin has rad = 0.0
                    ! debug
                    !write(*,'(a,3(f0.3,x),x,f0.3,x,i0)') '**debug(rot_power_spectrum): coos, rad, shell: ', coos, rad, current_shell
                    ! accumulate statistics
                    count_per_shell(current_shell) = count_per_shell(current_shell) + 1
                    rot_power_spectrum(current_shell) = rot_power_spectrum(current_shell) + cabs(self%complex_values(i,j,k))**2
                enddo
            enddo
        enddo

        ! average
        rot_power_spectrum = rot_power_spectrum / dble(count_per_shell)

        ! apply log scale
        if (llog_scale) then
            where (rot_power_spectrum == 0.) rot_power_spectrum = 0.0000001
            rot_power_spectrum = log10(rot_power_spectrum)
        endif
    end function rot_power_spectrum


    !>  \brief  returns the power in each resolution shell
    function thon_rings(self,log_scale)
        use UsefulFunctions, only : IsOdd
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  self        !<  input image
        logical,        optional        ::  log_scale   !<  if true, log of power spectrum will be returned
        ! result
        real(kind=8),   allocatable ::  thon_rings(:)
        ! private variables
        type(image)         ::  imgtmp
        logical             ::  llog_scale
        integer             ::  i,j,k
        integer             ::  num_output_elements
        real,   allocatable ::  count_per_shell(:)
        real                ::  rad
        integer             ::  current_shell
        integer             ::  ori(3), ii,jj,kk
        ! start work
        if (present(log_scale)) then
            llog_scale = log_scale
        else
            llog_scale = .false.
        endif
        ! make sure the image is in fourier space
        if (self%is_in_real_space) then
            call fwd_ft(self)
        endif

        ! compute amplitude
        call ampl(self,imgtmp,square=.false.,log=.true.)
        self = imgtmp

        ! mask the center
        !call circ_mask(self,rad=10.0,inverse=.true.)
        call cosine_disc(   self,rad=self%logical_dimensions(1)/12.,falloff=self%logical_dimensions(1)/32.,  &
                            wipe_input=.false.,inverse=.true.,mask_with_average=.true.)
        !call cosine_disc(self,rad=5.,falloff=2.,wipe_input=.false.,inverse=.true.,mask_with_average=.true.)

        !call write_to_disk(self,'dbg_ampl.mrc')

        ! filter
        call band_pass(self,hp_freq=0.005,lp_freq=.15,cosine_falloff=.001)

        call cosine_disc(   self,rad=self%logical_dimensions(1)/10.,falloff=self%logical_dimensions(1)/32.,   &
                            wipe_input=.false.,inverse=.true.,mask_with_average=.true.)

        !call write_to_disk(self,'dbg_filtered_ampl.mrc')

        ! address of origin
        if (self%is_in_real_space .and. self%origin_at_center) then
            ori = self%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif

        ! allocate memory for output arrays
        if (is_threed(self)) then
            num_output_elements = minval(self%logical_dimensions)
        else
            num_output_elements = minval(self%logical_dimensions(1:2))
        endif
        num_output_elements = min(num_output_elements,1024)
        if (IsOdd(num_output_elements)) num_output_elements = num_output_elements + 1
        num_output_elements = num_output_elements / 2
        num_output_elements = num_output_elements + 1   ! because we need the first element to have the origin
        allocate(thon_rings(num_output_elements))
        allocate(count_per_shell(num_output_elements))

        ! initialise output arrays
        thon_rings = 0.
        count_per_shell = 0

        ! loop over image
        do k=1,self%logical_dimensions(3)
            kk=(k-ori(3))**2
            do j=1,self%logical_dimensions(2)
                jj=(j-ori(2))**2
                do i=1,highest_index_dim1(self%logical_dimensions(1))
                    ii=(i-ori(1))**2
                    ! compute radius, as a fraction of image's linear dimension
                    rad = sqrt(real(ii+jj+kk))/real(self%logical_dimensions(1))
                    ! if we are beyond edge, cycle
                    if (rad .gt. .5) cycle
                    ! what shell are we in
                    current_shell = nint(2.*rad*real(num_output_elements-1))+1 ! +1 is because the origin has rad = 0.0
                    ! debug
                    !write(*,'(a,3(f0.3,x),x,f0.3,x,i0)') '**debug(rot_power_spectrum): coos, rad, shell: ', coos, rad, current_shell
                    ! accumulate statistics
                    count_per_shell(current_shell) = count_per_shell(current_shell) + 1
                    thon_rings(current_shell) = thon_rings(current_shell) + self%real_values(i,j,k)**2
                enddo
            enddo
        enddo

        ! average
        thon_rings = thon_rings / dble(count_per_shell)

        ! apply log scale
        if (llog_scale) then
            where (thon_rings == 0.) thon_rings = 0.0000001
            thon_rings = log10(thon_rings)
        endif
    end function thon_rings

    !>  \brief  returns the rotational average of an image.
    !!
    !!  \todo   implement for comp images, where a rotational amplitude spectrum should be returned.
    function rot_average(img) result(rot_ave)
        implicit none
        class(image),   intent(in)  ::  img
        real,           allocatable ::  rot_ave(:)
        ! private variables
        integer ::  i,j,k
        integer ::  sh
        integer,    allocatable ::  count_per_shell(:)
        ! start work
        if (.not. img%is_in_real_space) then
            write(*,'(a)') '**error(rot_average): can only work on real images'
            call terminate('fatal error in rot_average')
        endif
        if (.not. is_square(img)) then
            write(*,'(a)') '**error(rot_average): have not checked this works properly for non-square cube images'
            call terminate('fatal error in rot_average')
        endif
        if (.not. has_even_dims(img)) then
            write(*,'(a,3(i0,1x))') '**error(rot_average): odd dimensions not supported: ', img%logical_dimensions
            call terminate('feature not implemented')
        endif

        ! allocate
        allocate(rot_ave(img%logical_dimensions(1)/2+1))
        allocate(count_per_shell(size(rot_ave)))

        ! initialise
        rot_ave = 0.0
        count_per_shell = 0

        do k=1,img%logical_dimensions(3)
            do j=1,img%logical_dimensions(2)
                do i=1,img%logical_dimensions(1)
                    sh = shell(img,i,j,k)
                    if (sh .lt. img%logical_dimensions(1)/2+1) then
                        rot_ave(sh+1) = rot_ave(sh+1) + img%real_values(i,j,k)
                        count_per_shell(sh+1) = count_per_shell(sh+1) + 1
                    endif
                enddo
            enddo
        enddo

        rot_ave = rot_ave / count_per_shell

    end function rot_average

    !>  \brief  average 3d image cylindrically
    subroutine image_cyl_average(self)
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  self

        ! private variables
        real,       allocatable ::  average_value(:)            !  average value as a function of radius
        integer,    allocatable ::  num_of_values(:)
        integer,    parameter   ::  num_central_slices  =   23  !   number of central slices over which the average should be measured
        integer,    parameter   ::  radial_oversampling =   4
        integer ::  start_slice, finish_slice
        integer ::  i,j,k,ori(3), addr_logi(3)
        integer ::  num_of_shells, current_shell
        integer ::  ierr
        logical,    parameter   ::  debug   =   .false.

        ! start work

        ! only meant for real images
        if (.not. self%is_in_real_space) then
            write(*,'(a)') '**error(image_cyl_average): image is not in real space'
            call terminate('fatal error in image_cyl_average')
        endif

        ! get origin address
        if (self%is_in_real_space .and. self%origin_at_center) then
            ori = self%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif

        ! work out the start and finish slices
        if (self%logical_dimensions(3) .le. num_central_slices) then
            start_slice = 1
            finish_slice = self%logical_dimensions(3)
        else
            start_slice  = ori(3) - (num_central_slices-1)/2
            finish_slice = ori(3) + (num_central_slices-1)/2
        endif

        ! work out dimensions for allocatable array
        num_of_shells = maxval(self%logical_dimensions(1:2)-ori(1:2)+1)*radial_oversampling

        ! allocate arrays
        allocate(average_value(num_of_shells),num_of_values(num_of_shells),stat=ierr)
        if (ierr .ne. 0) call terminate('failed memory allocation')

        ! initialise arrays
        average_value = 0.0
        num_of_values = 0

        if (debug) then
            write(*,'(a)') '**debug(image_cyl_average): first loop over image...'
        endif

        ! loop over image to compute average value as a function of radius
        do k=start_slice,finish_slice
            do j=1,self%logical_dimensions(2)
                addr_logi(2) = (j - ori(2))**2
                do i=1,self%logical_dimensions(1)
                    addr_logi(1) = (i - ori(1))**2
                    ! what shell are in?
                    current_shell = nint(sqrt(real(sum(addr_logi(1:2)))+1.0)*real(radial_oversampling))
                    ! if we are within our limits, accumulate statistics
                    if (current_shell .le. size(num_of_values)) then
                        num_of_values(current_shell) = num_of_values(current_shell) + 1
                        average_value(current_shell) = average_value(current_shell) + self%real_values(i,j,k)
                    endif
                enddo
            enddo
        enddo

        if (debug) then
            write(*,'(a)') '**debug(image_cyl_average): computing average as a funciton of radius...'
        endif

        ! do the actual averaging
        where (num_of_values == 0)
            average_value = 0.0
        else where
            average_value = average_value / real(num_of_values)
        end where

        if (debug) then
            write(*,'(a)') '**debug(image_cyl_average): replacing values in image...'
        endif

        ! loop over the image to replace values with cylindrical average
        do k=1,self%logical_dimensions(3)
            do j=1,self%logical_dimensions(2)
                addr_logi(2) = (j - ori(2))**2
                do i=1,self%logical_dimensions(1)
                    addr_logi(1) = (i - ori(1))**2
                    ! what shell are in?
                    current_shell = nint(sqrt(real(sum(addr_logi(1:2)))+1.0)*real(radial_oversampling))
                    ! replace value
                    if (current_shell .le. size(num_of_values)) then
                        self%real_values(i,j,k) = average_value(current_shell)
                    else
                        ! self%real_values   =   0.0
                    endif
                enddo
            enddo
        enddo

        if (debug) then
            write(*,'(a)') '**debug(image_cyl_average): all done'
        endif

    end subroutine image_cyl_average

    !>  \brief  print voxel values of an image to monitor. only useful for very small images, for debugging purposes.
    subroutine print_image(img)
        implicit none
        ! arguments
        type(image),    intent(in)  ::  img
        ! private variables
        integer ::  j,k,imax
        character(len=50)   ::  fmt
        ! start work

        ! check the image is small enough
        if (any (img%logical_dimensions .gt. 9)) then
            write(*,'(a)') '**error(print_image): sorry, can only handle images of dimensions .le. 9'
            call terminate('fatal error in print_image')
        endif

        ! how many values per line?
        if (.not. img%is_in_real_space) then
            imax = img%logical_dimensions(1)*2
        else
            imax = img%logical_dimensions(1)
        endif

        ! format
        write(fmt,'(a,i2,a)') '(a,i2,a,', imax, '(se12.3,x))'

        ! loop & print
        do k=1,img%logical_dimensions(3)
            write(*,'(/a,i1)') 'slice #', k
            do j=1,img%logical_dimensions(2)
                if (img%is_in_real_space) then
                    write(*,fmt) 'line #', j, ' | ', img%real_values(1:imax,j,k)
                else
                    write(*,fmt) 'line #', j, ' | ', img%complex_values(1:imax/2,j,k)
                endif
            enddo
        enddo
    end subroutine print_image

    !>  \brief  print the profile (1d projection) of the image along dimension n
    subroutine print_profile(img,n)
        implicit none
        type(image),    intent(in)  ::  img
        integer,        intent(in)  ::  n
        ! private variables
        integer ::  i
        ! start work
        if (.not. img%is_in_real_space) then
            write(*,'(a)') '**error(print_profile): can only work on real images for now'
            !stop 'fatal error in print_profile'
            call terminate()
        endif
        select case (n)
            case(1)
                do i=1,img%logical_dimensions(1)
                    write(*,'(es12.5)') sum(img%real_values(i,:,:))
                enddo
            case (2)
                do i=1,img%logical_dimensions(2)
                    write(*,'(es12.5)') sum(img%real_values(1:img%logical_dimensions(1),i,:))
                enddo
            case(3)
                do i=1,img%logical_dimensions(2)
                    write(*,'(es12.5)') sum(img%real_values(1:img%logical_dimensions(1),:,i))
                enddo
        end select
    end subroutine print_profile


    !>  \brief  pad the input image with 0, to make it factor times bigger
    subroutine pad_by_factor(img,factor,pad_with_value)
        implicit none
        ! arguments
        type(image),            intent(inout)   ::  img
        integer,                intent(in)      ::  factor
        real,       optional,   intent(in)      ::  pad_with_value
        ! private variables
        integer ::  newdim(3)

        ! start work
        newdim = img%logical_dimensions * factor
        if (img%logical_dimensions(3) .eq. 1) then
            newdim(3) = 1
        endif

        call pad(img,newdim,pad_with_value=pad_with_value)
    end subroutine pad_by_factor

    !>  \brief  crop the input image, to make it factor times smaller
    subroutine crop_by_factor(img,factor)
        implicit none
        ! arguments
        type(image),            intent(inout)   ::  img
        integer,                intent(in)      ::  factor
        ! private variables
        integer ::  newdim(3)

        ! start work
        newdim = img%logical_dimensions / factor
        if (img%logical_dimensions(3) .eq. 1) then
            newdim(3) = 1
        endif

        call crop(img,newdim)
    end subroutine crop_by_factor

    !>  \brief  resize the input image to new dimensions. in place.
    !!  \todo   add capability to resize even if some of the dimensions are larger and some are smaller
    subroutine resize(img,new_dim)
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  img !<  image to be resized
        integer,        intent(in)      ::  new_dim(3)
        ! private variables
        integer ::  old_dim(3)
        integer ::  last_dim
        ! start work
        old_dim = img%logical_dimensions

        ! are we working in 2d or 3d?
        if (img%logical_dimensions(3) .eq. 1) then
            last_dim = 2
        else
            last_dim = 3
        endif

        ! work out whether we are padding, cropping, a bit of both or nothing
        if ( all(old_dim(1:last_dim) .eq. new_dim(1:last_dim)) ) then
            ! do nothing
        else if (all(old_dim(1:last_dim) .le. new_dim(1:last_dim))) then
            ! pad
            call pad(img,new_dim)
        else if (all(old_dim(1:last_dim) .ge. new_dim(1:last_dim))) then
            ! crop
            call crop(img,new_dim)
        else
            ! some dimensions are larger, some dimensions are smaller
            ! we don't know how to handle this yet
            write(*,'(1a,2(3(i0,1x),3x))')  '**error(resize): resizing from old to new dimensions is',    &
                                            ' not a simple pad or crop operation. not implemented yet. old, new dim = ', &
                                            old_dim, new_dim
            call terminate('fatal error in resize. feature not implemented yet.')
        endif
    end subroutine resize

    !>  \brief  bin/coarsen an image, i.e. a number of input voxel are averaged into each output voxel
    !!
    !!  the following text from chuck helps to get the centering of the output correctly when doing real-space binning...
    !!
    !!  below i convert from an unbinned (x,y) to a binned (x',y') coordinate system.
    !!  strategy:
    !!   (1) figure out how far x' is from binned origin, in unbinned coordinates; let this be dx
    !!   (2) convert dx to binned coordinates (dx');
    !!   (3) then compensate for (1,1) rather than (0,0) origin
    !!
    !!  (1) in unbinned coordinates: dx = (x - (1 + bin/2 - 0.5))
    !!    nb: this is from a prior analysis, which i am not bothering to repeat here
    !!  (2) in binned coordinates:   dx' = (x - (1 + bin/2 - 0.5))/bin
    !!  (3) in binned coordinate system: x' = 1 + dx' = 1 + (x - (1 + bin/2 - 0.5))/bin
    !!   = 1 + x/bin - (1 + bin/2 - 0.5)/bin
    !!   = 1 + x/bin - (bin + 1)/(2*bin)
    !!   = 1 + x/bin - 1/2 - 1/*(2*bin)
    !!   = 1 + (x - 1/2)/bin - 1/2
    !!   = 1/2 + (x - 1/2)/bin
    !!
    !!  unbinned -> binned case:
    !!     x = 0.5 + (x - 0.5)/bin;
    !!     y = 0.5 + (y - 0.5)/bin;
    !!
    !!  binned -> unbinned case:
    !!    x = 0.5 + (x - 0.5)*bin
    !!    y = 0.5 + (y - 0.5)*bin
    !!
    !!
    subroutine bin(imgin,imgout,binning_factor,unbin,preserve_center)
        use INTERPOLATION, only : interpolation_scheme, interpolation_scheme_init
        use UsefulFunctions, only : IsEven
        implicit none
        ! arguments
        class(image),           intent(in)      ::  imgin           !<  input image to be binned
        type(image),            intent(inout)   ::  imgout          !<  output, binned, image
        integer,                intent(in)      ::  binning_factor  !<  binning factor
        logical,    optional,   intent(in)      ::  unbin           !<  if .true., unbinning will occur. .false. by default
        logical,    optional,   intent(in)      ::  preserve_center !<  whether to shift the output image so that the origin (at center) coincides with that of the input image
        ! private variable
        integer ::  i,j,k, ii,jj,kk
        real    ::  shift(3)
        logical ::  ppreserve_center
        logical,    parameter   ::  bin_in_fourier_space = .true.
        logical ::  uunbin
        integer ::  newdim(3)
        type(interpolation_scheme)  ::  interpol_scheme
        ! start work

        ! check allocation status
        if (.not. imgin%in_memory) then
            write(*,'(a)') '**error(bin): input image not allocated'
            call terminate('fatal error in bin')
        endif

        ! are we unbinning?
        uunbin = .false.
        if (present(unbin)) uunbin = unbin

        ! will we conserve the centering of the image?
        ppreserve_center = .true.
        if (present(preserve_center)) ppreserve_center = preserve_center


        ! prepare an interpolation scheme in case we need it
        if (ppreserve_center .and. .not. bin_in_fourier_space) then
            call interpolation_scheme_init(interpol_scheme,'linear','rect',2)
        endif


        ! this is only meant for real images
        if (is_real(imgin)) then
            if (uunbin) then
                !
                ! unbinning of real image
                !

                ! set output image dimensions
                newdim = imgin%logical_dimensions * binning_factor
                if (imgin%logical_dimensions(3) .eq. 1) then
                    ! the 2d case
                    newdim(3) = 1
                endif
                !
                !   unbinning of a real image can be done by padding its fourier transform with 0.0 (shift correction
                !   is not needed when using that method)
                !
                call image_allocate(imgout,imgin)
                imgout%real_values = imgin%real_values
                call fwd_ft(imgout)
                call pad(imgout,newdim)
                !
                write(*,'(a)') '**info(bin): image unbinning was done by padding the fourier transform'
                if (ppreserve_center) then
                    write(*,'(a)') '**info(bin): image centering was preserved'
                else
                    !!  binned -> unbinned case:
                    !!    x = 0.5 + (x - 0.5)*bin
                    !!    y = 0.5 + (y - 0.5)*bin
                    shift(1:3) = +0.5
                    if (imgout%logical_dimensions(3) .eq. 1) shift(3) = 0.0
                    do i=1,3
                        if (IsEven(imgout%logical_dimensions(i))) then
                            shift(i) = shift(i) - (0.5 * real(binning_factor))
                        endif
                    enddo
                    call pshift_image(imgout,shift_x=-shift(1),shift_y=-shift(2),shift_z=-shift(3))
                    write(*,'(2a,3(f0.2,1x))') '**info(bin): image was phase shifted to simulate real-space binning algorithm. ',  &
                                                'centering not preserved. applied shift: ', shift
                endif
                call bwd_ft(imgout)
            else
                !
                ! binning of real image
                !
                if (bin_in_fourier_space) then
                    !
                    ! real image binning in fourier space
                    !

                    ! set output image dimensions
                    newdim = ceiling(imgin%logical_dimensions / real(binning_factor))
                    if (imgin%logical_dimensions(3) .eq. 1) then
                        ! the 2d case
                        newdim(3) = 1
                    endif
                    call image_allocate(imgout,imgin)
                    imgout%real_values = imgin%real_values
                    call fwd_ft(imgout)
                    call crop(imgout,newdim)
                    write(*,'(a)') '**info(bin): image binning was done by cropping the fourier transform'
                    if (ppreserve_center) then
                        write(*,'(a)') '**info(bin): image centering was preserved'
                    else
                        ! work out the shift to apply
                        !!  unbinned -> binned case:
                        !!     x = 0.5 + (x - 0.5)/bin;
                        !!     y = 0.5 + (y - 0.5)/bin;
                        shift(1:3) = +0.5
                        do i=1,3
                            if (IsEven(imgout%logical_dimensions(i))) then
                                shift(i) = shift(i) - (0.5/ real(binning_factor))
                            endif
                        enddo
                        if (imgout%logical_dimensions(3) .eq. 1) shift(3) = 0.0
                        shift = -shift
                        call pshift_image(imgout,shift_x=shift(1),shift_y=shift(2),shift_z=shift(3))
                   write(*,'(2a,3(f0.2,1x))') '**info(bin): image was phase-shifted to simulate real-space binning algorithm. ',  &
                                                'centering not preserved. applied shift: ', shift

                    endif
                    call bwd_ft(imgout)
                else
                    !
                    ! real image binning in real space
                    !

                    ! set output image dimensions
                    imgout%logical_dimensions = ceiling(imgin%logical_dimensions / real(binning_factor))
                    if (imgin%logical_dimensions(3) .eq. 1) then
                        ! the 2d case
                        imgout%logical_dimensions(3) = 1
                    endif
                    ! allocate & blank
                    call image_allocate(imgout)
                    imgout%real_values = 0.0
                    !
                    ! do the binning
                    !
                    do k=1,imgin%logical_dimensions(3)
                        kk = ceiling(real(k)/real(binning_factor))
                        if (imgin%logical_dimensions(3) .eq. 1) kk = 1
                        do j=1,imgin%logical_dimensions(2)
                            jj = ceiling(real(j)/real(binning_factor))
                            do i=1,imgin%logical_dimensions(1)
                                ii = ceiling(real(i)/real(binning_factor))
                                ! copy value over
                                imgout%real_values(ii,jj,kk) = imgout%real_values(ii,jj,kk) + imgin%real_values(i,j,k)
                            enddo
                        enddo
                    enddo
                    !print *, '**debug(bin): dumping'
                    !call write_to_disk(imgout,'dbg_bin_prenorm.mrc')

                    ! normalise
                    imgout%real_values = imgout%real_values / binning_factor**2
                    if (imgin%logical_dimensions(3) .gt. 1) then
                        imgout%real_values = imgout%real_values / binning_factor
                    endif

                    ! shift to get correct origin (maybe this should be done on the input, larger image to get more accuracy?)
                    if (ppreserve_center) then
                        ! work out the shift to apply
                        !!  unbinned -> binned case:
                        !!     x = 0.5 + (x - 0.5)/bin;
                        !!     y = 0.5 + (y - 0.5)/bin;
                        shift(1:3) = +0.5
                        if (imgout%logical_dimensions(3) .eq. 1) shift(3) = 0.0
                        do i=1,3
                            if (IsEven(imgout%logical_dimensions(i))) then
                                shift(i) = shift(i) - (0.5/ real(binning_factor))
                            endif
                        enddo
                        write(*,'(a,3f10.4)') '**info(bin): shifting by ', -shift

                        ! apply the shift
                        call fwd_ft(imgout)

                        if (imgout%logical_dimensions(3) .eq. 1) then
                            call pshift_image(imgout,shift(1),shift(2),0.0)
                        else
                            call pshift_image(imgout,shift(1),shift(2),shift(3))
                        endif
                        call bwd_ft(imgout)
                    endif ! end of test for correct shift
                endif   ! end of test for bin in fourier space
            endif   ! end of test for unbinning
        else
            !
            ! the complex case
            !
            write(*,'(a)') '**error(bin): Fourier-space images not supported'
            call terminate('fatal error in bin')
        endif
    end subroutine bin


    !>  \brief  pad the input image with 0, and give it new dimensions newdim
    subroutine pad(img,newdim,shift,pad_with_value)
        implicit none
        ! arguments
        type(image),                intent(inout)   ::  img
        integer,                    intent(in)      ::  newdim(3)
        real(kind=8),   optional,   intent(in)      ::  shift(3)    !<  when padding in fourier space, a real-space shift can be done
        real,           optional,   intent(in)      ::  pad_with_value  !<  value to pad with
        ! private variables
        type(image) ::  imgtmp
        integer ::  old_dim(3)
        integer ::  addr_start(3)
        integer ::  addr_end(3)
        integer ::  i,j,k,ii,jj,kk,coos(3),new_addr(3)
        integer ::  origin_old(3), origin_new(3)
        real(kind=8)    ::  phase_x,phase_y,phase_z, sshift(3)
        complex(kind=8) ::  pshift
        real            ::  ppad_with_value
        logical,    parameter   ::  debug   =   .false.
        ! start work

        if (.not. img%origin_at_center) then
            write(*,'(a)') '**warning(pad): behaviour not well defined when the image does not have its origin at its center'
        endif

        sshift = (/0.0d0,0.0d0,0.0d0/)
        if (present(shift)) then
            if (is_real(img)) then
                write(*,'(a)') '**error(pad): shift not implemented for real images'
                call terminate('fatal error in pad')
            endif
            sshift = shift
        endif

        ! value to pad with
        ppad_with_value = 0.0
        if (present(pad_with_value)) then
            if (.not. img%is_in_real_space) then
                write(*,'(a)') '**error(pad): cannot specify value to pad with for complex image'
                call terminate('fatal error in pad')
            endif
            ppad_with_value = pad_with_value
        endif

        ! old origin
        if (img%is_in_real_space .and. img%origin_at_center) then
            origin_old = img%physical_address_of_box_center
        else
            origin_old = [1,1,1]
        endif
        if (debug) print *, '**debug(pad): old origin = ', origin_old

        old_dim = img%logical_dimensions


        ! copy contents of image to temporary image
        call image_allocate(imgtmp,img)
        imgtmp%real_values = img%real_values


        ! reallocate image to correct dimensions
        if (debug) print '(a,3i5)', '**debug(pad): before reallocation: ', img%logical_dimensions
        call image_deallocate(img)
        if (debug) write(*,*) '**debug: output image should have dim', newdim
        img%logical_dimensions = newdim
        call image_allocate(img)
        if (debug) print '(a,3i5)', '**debug(pad): after  reallocation: ', img%logical_dimensions

        ! new origin
        if (img%is_in_real_space .and. img%origin_at_center) then
            origin_new = img%physical_address_of_box_center
        else
            origin_new = [1,1,1]
        endif
        if (debug) print *, '**debug(pad): new origin = ', origin_new

        ! initialise image
        img%real_values = ppad_with_value

        !
        ! place the original voxel values into the image at the correct place
        ! we use different logic in the real and comp cases
        !
        if (img%is_in_real_space) then
            addr_start = origin_new - origin_old + 1
            !addr_start = (newdim-old_dim)/2+1
            addr_end   = addr_start + old_dim - 1
            if (img%logical_dimensions(3) .eq. 1) then
                addr_start(3) = 1
                addr_end(3) = 1
            endif
            if (debug) then
                write(*,*) '**debug(pad): will copy from, to: ', '1,', old_dim(1), ' to ', addr_start(1), ',', addr_end(1)
                write(*,*) '**debug(pad): will copy from, to: ', '1,', old_dim(2), ' to ', addr_start(2), ',', addr_end(2)
                write(*,*) '**debug(pad): will copy from, to: ', '1,', old_dim(3), ' to ', addr_start(3), ',', addr_end(3)
            endif
            img%real_values(addr_start(1):addr_end(1),addr_start(2):addr_end(2),addr_start(3):addr_end(3)) = &
            imgtmp%real_values(1:old_dim(1),1:old_dim(2),1:old_dim(3))
        else
            ! loop over all voxels of the input, and for each one, work out its coordinates relative
            ! to the 1,1,1 origin. the coordinates are the same in the output image, so we then just
            ! have to convert these coordinates to a new address, bearing in mind the new dimensions.
            do k=1,imgtmp%logical_dimensions(3)
                kk = k - nearest_orig(k,imgtmp%logical_dimensions(3))
                phase_z = phase_from_shift(sshift(3),kk,imgtmp%logical_dimensions(3))
                do j=1,imgtmp%logical_dimensions(2)
                    jj = j - nearest_orig(j,imgtmp%logical_dimensions(2))
                    phase_y = phase_from_shift(sshift(2),jj,imgtmp%logical_dimensions(2))
                    do i=1,highest_index_dim1(imgtmp%logical_dimensions(1))
                        ii = i - 1
                        phase_x = phase_from_shift(sshift(1),ii,imgtmp%logical_dimensions(1))
                        pshift = phase_shift_3d(phase_x,phase_y,phase_z)
                        coos = [ii,jj,kk]
                        call comp_addr_phys_from_logi(img,coos,new_addr)
                        img%complex_values(new_addr(1),new_addr(2),new_addr(3)) = imgtmp%complex_values(i,j,k) * pshift
                    enddo
                enddo
            enddo
        endif
        call image_deallocate(imgtmp)
    end subroutine pad

    !>  \brief  crop the image to the dimensions specified.
    subroutine crop(img,newdim)
        implicit none
        ! arguments
        type(image),    intent(inout)   ::  img         !<  image to be cropped
        integer,        intent(in)      ::  newdim(3)   !<  dimensions to crop the image to
        ! private variables
        type(image) ::  imgtmp
        integer     ::  addr_start(3), addr_end(3), origin_old(3), origin_new(3)
        integer     ::  i,j,k,ii,jj,kk,new_coos(3),old_addr(3)
        ! start work
        if (any(img%logical_dimensions .lt. newdim)) then
            write(*,'(a,2(3(i0,1x),3x))') '**error(crop): output dimension(s) greater than input dimensions:', &
                                            img%logical_dimensions, newdim
            call terminate('fatal error in crop')
        endif

        ! allocate temporary image to hold the output densities
        imgtmp%logical_dimensions = newdim
        imgtmp%is_in_real_space = img%is_in_real_space
        call image_allocate(imgtmp)

        ! old and new origins
        if (img%is_in_real_space .and. img%origin_at_center) then
            origin_old = img%physical_address_of_box_center
        else
            origin_old = [1,1,1]
        endif
        if (imgtmp%is_in_real_space .and. imgtmp%origin_at_center) then
            origin_new = imgtmp%physical_address_of_box_center
        else
            origin_new = [1,1,1]
        endif


        if (img%is_in_real_space) then
            ! work out the pixels that need to be copied over
            addr_start = origin_old - origin_new + 1
            addr_end   = addr_start + newdim - 1
            if (img%logical_dimensions(3) .eq. 1) then
                addr_start(3) = 1
                addr_end(3) = 1
            endif

            ! copy density values over
            imgtmp%real_values(1:imgtmp%logical_dimensions(1),1:imgtmp%logical_dimensions(2),1:imgtmp%logical_dimensions(3)) =    &
                            img%real_values(addr_start(1):addr_end(1),addr_start(2):addr_end(2),addr_start(3):addr_end(3))
        else
            ! we're dealing with a complex image
            ! loop over the new image and copy corresponding pixels from the old image
            do k=1,imgtmp%logical_dimensions(3)
                kk = k - nearest_orig(k,imgtmp%logical_dimensions(3))
                do j=1,imgtmp%logical_dimensions(2)
                    jj = j - nearest_orig(j,imgtmp%logical_dimensions(2))
                    do i=1,highest_index_dim1(imgtmp%logical_dimensions(1))
                        ii = i-1
                        new_coos = (/ii,jj,kk/)
                        call comp_addr_phys_from_logi(img,new_coos,old_addr)
                        imgtmp%complex_values(i,j,k) = img%complex_values(old_addr(1),old_addr(2),old_addr(3))
						if (k .eq. lowest_index_for_negative_frequencies(imgtmp%logical_dimensions(3)) .or. &
					        j .eq. lowest_index_for_negative_frequencies(imgtmp%logical_dimensions(2)) ) then
							!imgtmp%complex_values(i,j,k) = real(imgtmp%complex_values(i,j,k))
						endif
                    enddo
                enddo
            enddo
        endif


        ! the input image is overwritten with the output
        call image_deallocate(img)
        call image_allocate(img,imgtmp)
        img%real_values = imgtmp%real_values
        call image_deallocate(imgtmp)

    end subroutine crop

end module images
